/**
 * @file count_time_server.c
 * @brief 
 * @author Airead Fan <fgh1987168@gmail.com>
 * @date 2012/03/21 11:58:14
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

struct time_wait {
    int wait;
    int count_time;
    long int dead_time;
    int last_count_time;
    int flag;
};

int time_wait_init(struct time_wait *tw, int wait_time) 
{
    tw->wait = wait_time;
    tw->dead_time = time(NULL) + wait_time;
    tw->count_time = 0;
    tw->last_count_time = 0;

    return 0;
}

/**
 * @brief time check
 * @param tw
 * @return 0 for nothing, 1 for count change, 2 for timeout
 */
int time_wait_check(struct time_wait *tw)
{
    long int t;
    
    tw->flag = 0;

    t = time(NULL);
    tw->count_time = tw->dead_time - t;
    if (tw->count_time < 0) {
        tw->dead_time = t + tw->wait;
        tw->flag = 2;               /* timeout */
    }
    
    if (tw->count_time != tw->last_count_time && tw->count_time >= 0) {
        tw->last_count_time = tw->count_time;
        tw->flag = 1;                /* count changed */
    }

    return 0;
}

static int client_num;
static int clients[10];

int get_listen_socket(char *ip, unsigned short port, int listen_backlog)
{
    int listening_socket;
    int optval;
    struct sockaddr_in local_ipv4_address;

    /* Create a socket */
    //int socket(int domain, int type, int protocol);
    if ((listening_socket = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
        fprintf(stderr, "Failed to create new TCP socket: %s\n", strerror(errno));
        exit(1);
    }

    /* set socket option */
    optval = 1;
    //int setsockopt(int sockfd, int level, int optname, const void *optval, socklen_t optlen);
    setsockopt(listening_socket, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));

#ifdef SO_REUSEPORT
    optval = 1;
    setsockopt(listening_socket, SOL_SOCKET, SO_REUSEPORT, &optval, sizeof(optval));
#endif

    /* bind */
    memset(&local_ipv4_address, 0, sizeof(local_ipv4_address));

    local_ipv4_address.sin_family = AF_INET;
    local_ipv4_address.sin_port = htons(port);
    inet_pton(AF_INET, ip, &local_ipv4_address.sin_addr);

    if (bind(listening_socket, (struct sockaddr *)&local_ipv4_address, sizeof(local_ipv4_address)) < 0) {
        fprintf(stderr, "Bind to %s:%d failed: %s\n", ip, port, strerror(errno));
        close(listening_socket);
        exit(1);
    }

    fprintf(stdout, "Bound to %s:%d successfully.\n", ip, port);

    /* Listen */
    if (listen(listening_socket, listen_backlog) < 0) {
        fprintf(stderr, "Listen on %s:%d failed: %s\n", ip, port, strerror(errno));
        close(listening_socket);
        exit(1);
    }

    fprintf(stdout, "Listen on %s:%d successfully, waiting for incoming connections ...\n", ip, port);

    return listening_socket;
}

int add_client(int listen_socket)
{
    int connfd;
    char peer_ip[INET_ADDRSTRLEN];
    unsigned short port;
    struct sockaddr_in peer_addr;
    socklen_t len;

    len = sizeof(peer_addr);
    connfd = accept(listen_socket, (struct sockaddr *)&peer_addr, &len);
    if (connfd < 0) {
        fprintf(stderr, "accept failed: %s\n", strerror(errno));
        return -1;
    }
    
    len = sizeof(peer_ip);
    //const char *inet_ntop(int af, const void *src, char *dst, socklen_t size);
    inet_ntop(AF_INET, &peer_addr.sin_addr, peer_ip, len);
    if (peer_ip == NULL) {
        fprintf(stderr, "inet_ntop failed\n");
        return -1;
    }
    port = ntohs(peer_addr.sin_port);

    printf("receive client form %s:%d\n", peer_ip, port);

    clients[client_num] = connfd;
    client_num++;

    return connfd;
}

int remove_client(int fd)
{
    int i;                    /* location */

    for (i = 0; i < client_num; i++) {
        if (clients[i] == fd) {
            break;
        }
    }

    if (i < client_num) {
        for (; i < client_num - 1; i++) {
            clients[i] = clients[i + 1];
        }
        client_num--;
    }
    
    return 0;
}

int deal_client_fd(fd_set *readfds)
{
    int i;

    for (i = 0; i < client_num; i++) {
        if (FD_ISSET(clients[i], readfds)) {
            close(clients[i]);
            remove_client(clients[i]);
        }
    }

    return 0;
}

int register_select(int fd, fd_set *readfds, int *nfds)
{
    FD_SET(fd, readfds);
    if (*nfds <= fd) {
        *nfds = fd + 1;
    }

    return 0;
}

int main(int argc, char *argv[])
{
    fd_set readfds;
    int nfds;
    int i, ret;
    int listening_socket;
    char *ip;
    struct time_wait tw;
    unsigned short port;
    int listen_backlog;
    struct timeval tv;

    if (argc < 4) {
        fprintf(stderr, "usage: %s <ip_addr> <port> <backlog>\n", argv[0]);
        exit(1);
    }

    ip = argv[1];
    port = (unsigned short)strtol(argv[2], NULL, 10);
    listen_backlog = (int) strtol(argv[3], NULL, 10);

    if ((listening_socket = get_listen_socket(ip, port, listen_backlog)) < 0) {
        fprintf(stderr, "get listen socket failed\n");
        exit(1);
    }

    nfds = 0;
    time_wait_init(&tw, 5);
    /* accept */
    for(;;) {
        FD_ZERO(&readfds);
        register_select(listening_socket, &readfds, &nfds);
        for (i = 0; i < client_num; i++) {
            register_select(clients[i], &readfds, &nfds);
        }

        tv.tv_sec = 0;
        tv.tv_usec = 100000;
        ret = select(nfds, &readfds, NULL, NULL, &tv);
        if (ret < 0) {
            fprintf(stderr, "select failed: %s\n", strerror(errno));
        } else if (ret == 0) {
            time_wait_check(&tw);

            if (tw.flag == 1) {
                printf("time: %d\n", tw.count_time);
                for (i = 0; i < client_num; i++) {
                    write(clients[i], &tw.count_time, sizeof(tw.count_time));
                }
            }
            
            if (tw.flag == 2) {
                printf("=================================\n");
                printf("collect data from client\n");
                printf("=================================\n");
            }
        } else {
            if (FD_ISSET(listening_socket, &readfds)) {
                add_client(listening_socket);
            }
            deal_client_fd(&readfds);
        }
    }//for end

    close(listening_socket);

    return 0;
}

