/**
 * @file server.c
 * @brief server functions
 * @author Airead Fan <fgh1987168@gmail.com>
 * @date 2012  2月 22 12:00:18 CST
 */

#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <assert.h>

#include "server.h"
#include "select.h"
#include "client.h"
#include "count_time.h"
#include "common_debug.h"

/**
 * @brief get server listening socket file describer
 * @param self
 * @return on error returns -1, otherwise returns sockect file describer
 */
int server_get_listen_socket(struct server *self)
{
    int listening_socket;
    int optval;
    int listen_backlog;
    struct sockaddr_in local_ipv4_address;
    char *ip_addr;
    unsigned short port;

    STEP_CHECK();

    assert(self);

    if (self->ip_addr == NULL) {
        PRINTERR("ip_addr is NULL!\n");
        return -1;
    }

    ip_addr = self->ip_addr;
    port = self->port;

    /* Create a socket */
    //int socket(int domain, int type, int protocol);
    if ((listening_socket = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
        PRINTERR("Failed to create new TCP socket: %s\n", strerror(errno));
        return -1;
    }

    /* set socket option for reuse */
    optval = 1;
    //int setsockopt(int sockfd, int level, int optname, const void *optval, socklen_t optlen);
    setsockopt(listening_socket, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));

#ifdef SO_REUSEPORT
    optval = 1;
    setsockopt(listening_socket, SOL_SOCKET, SO_REUSEPORT, &optval, sizeof(optval));
#endif

    /* bind */
    memset(&local_ipv4_address, 0, sizeof(local_ipv4_address));

    local_ipv4_address.sin_family = AF_INET;
    local_ipv4_address.sin_port = htons(port);
    inet_pton(AF_INET, ip_addr, &local_ipv4_address.sin_addr);

    if (bind(listening_socket, (struct sockaddr *)&local_ipv4_address, sizeof(local_ipv4_address)) < 0) {
        PRINTERR("Bind to %s:%d failed: %s\n", ip_addr, port, strerror(errno));
        close(listening_socket);
        return -1;
    }

    DBG("Bound to %s:%d successfully.\n", ip_addr, port);

    /* Listen */
    listen_backlog = 10;
    if (listen(listening_socket, listen_backlog) < 0) {
        PRINTERR("Listen on %s:%d failed: %s\n", ip_addr, port, strerror(errno));
        close(listening_socket);
        return -1;
    }

    DBG("Listen on %s:%d successfully, waiting for incoming connections ...\n", ip_addr, port);

    return listening_socket;
}

/**
 * @brief collect data from client
 * @param self
 * @return on success returns 0, otherwise returns -1
 */
int server_collect_data(struct server *self)
{
    int i;
    struct command cmd;
    struct client *client;

    client = self->c_array;

    for (i = 0; i < self->client_num; i++) {
        cmd.type = CMD_GET_CLIENT_DATA;
        /* FIXME: error handle */
        write(client[i].id, &cmd, sizeof(cmd));
        read(client[i].id, &cmd, sizeof(cmd));
        if (cmd.type != CMD_SEND_CLIENT_DATA) {
            PRINTERR("get client error\n");
            return -1;
        }
        client[i].data = cmd.client_data;
        DBG("receive [%d] data.three: %d\n", i, client[i].data.three);
    }

    return 0;
}

/**
 * @brief send server data to all clients
 * @param self
 * @return on success returns 0, otherwise returns -1
 */
int server_send_result(struct server *self)
{
    int i;
    struct command cmd;
    struct client *client;

    client = self->c_array;

    for (i = 0; i < self->client_num; i++) {
        cmd.type = CMD_SEND_SERVER_DATA;
        cmd.server_data = self->data;
        write(client[i].id, &cmd, sizeof(cmd));
    }

    return 0;
}

/**
 * @brief control server with command user interface
 * @param self
 * @param sdata
 * @return on success returns 0, otherwise returns -1
 */
int server_cui_control(struct server *self, struct select_data *sdata)
{
    char buf[BUFF_SIZE];
    ssize_t read_n;

    if (FD_ISSET(self->input_fd, &sdata->readfds)) {
        memset(buf, 0, BUFF_SIZE);
        //ssize_t read(int fd, void *buf, size_t count);
        read_n = read(self->input_fd, buf, BUFF_SIZE - 1);
        if (read_n < 0) {
            PRINTERR("read data error: %s\n", strerror(errno));
            return -1;
        } else if (read_n == 0) { /* remote close */

        } else {
            printf("send %s \n", buf);
            if (!strcmp(buf, "get\n")) {
                if (server_collect_data(self) < 0) {
                    PRINTERR("collect data failed\n");
                    return -1;
                }
            } else if (!strcmp(buf, "data\n")) {
                server_get_result(self);
                server_send_result(self);
            }
        }
    }    

    return 0;
}

/**
 * @brief collect client data
 * @param self
 * @param sdata
 * @return on success returns 0, otherwise returns -1
 */
int server_deal_client_data(struct server *self, struct select_data *sdata)
{
    struct client *client;
    int i;
    ssize_t read_n;
    struct command cmd;

    client = self->c_array;

    for (i = 0; i < self->client_num; i++) {
        if (FD_ISSET(client[i].id, &sdata->readfds)) {
            //ssize_t read(int fd, void *buf, size_t count);
            read_n = read(client[i].id, &cmd, sizeof(cmd));
            if (read_n < 0) {
                PRINTERR("read data error: %s\n", strerror(errno));
                return -1;
            } else if (read_n == 0) { /* remote close */
                close(client[i].id);
                server_remove_client(self, client[i].id);
            } else {
                switch (cmd.type) {
                case CMD_SEND_CLIENT_DATA:
                    client[i].data = cmd.client_data;
                    DBG("recevice client_data.three: %d\n", client[i].data.three);
                    break;
                default:
                    PRINTERR("error, can't reach here\n");
                }
            }
        }
    }

    return 0;
}
/**
 * @brief the first swicth function
 * @param self
 * @param sdata
 * @return on success returns 0, otherwise returns -1
 */
static int server_main_switch(struct server *self, struct select_data *sdata)
{

    server_cui_control(self, sdata);

    server_deal_client_data(self, sdata);

    
    return 0;
}

/**
 * @brief initialize server
 * @param self :
 * @param ip_addr : binded ip
 * @param prot : binded port
 * @return on success returns 0, otherwise returns -1
 */
int server_init(struct server *self, char *ip_addr, unsigned short port)
{
    if (ip_addr == NULL) {
        PRINTERR("ip_addr is NULL!\n");
        return -1;
    }

    self->ip_addr = ip_addr;
    self->port = port;
    self->input_fd = 0;
    self->client_num = 0;

    if ((self->listen_socket = server_get_listen_socket(self)) < 0) {
        PRINTERR("server get listen socket failed\n");
        return -1;
    }
    
    return 0;
}

/**
 * @brief 
 * @param 
 * @return 
 */
int server_time_control(struct server *self, struct count_time *ct)
{
    int i;
    struct command cmd;

    if (count_time_check(ct) == COUNT_TIME_DISABLE) {
        return 0;
    }

    switch (ct->flag) {
    case COUNT_TIME_NOTHING:
        break;
    case COUNT_TIME_CHANGED:
        printf("time: %d\n", ct->count_time);
        for (i = 0; i < self->client_num; i++) {
            cmd.type = CMD_COUNT_TIME;
            self->data.count_time = ct->count_time;
            cmd.server_data = self->data;
            write(self->c_array[i].id, &cmd, sizeof(cmd));
        }
        break;
    case COUNT_TIME_TIMEOUT:
        printf("=================================\n");
        printf("collect data from client\n");
        printf("=================================\n");
        if (server_collect_data(self) < 0) {
            PRINTERR("collect data failed\n");
            return -1;
        }
        server_get_result(self);
        server_send_result(self);
        break;
    default:
        PRINTERR("error: can't reach here\n");
    }
            
    return 0;
}

/**
 * @brief server main function
 * @param ip_addr : bind ip address.
 * @param port : bind port
 * @return on error returns -1
 *
 * receive all client information, and return the result.
 */
int server_main(char *ip_addr, unsigned short port)
{
    struct server server;
    struct select_data *sdata;
    struct count_time ct;
    int ret;

    STEP_CHECK();

    if ((sdata = select_new()) == NULL) {
        PRINTERR("get select data failed\n");
        return -1;
    }    

    if (server_init(&server, ip_addr, port) < 0) {
        PRINTERR("server init failed\n");
        return -1;
    }

    count_time_init(&ct, COUNT_TIME_VALUE);
    while (1) {
        select_init(sdata);
        sdata->timeout.tv_sec = 0;
        sdata->timeout.tv_usec = 100000;
        server_register_select(&server, sdata);
        /* int select(int nfds, fd_set *readfds, fd_set *writefds,
         *           fd_set *exceptfds, struct timeval *timeout); */
        ret = select(sdata->nfds, &sdata->readfds, &sdata->writefds, 
                     &sdata->exceptfds, &sdata->timeout);
        if (ret < 0) {
            PRINTERR("select failed: %s\n", strerror(errno));
        } else if (ret == 0) {  /* select time out */
            server_time_control(&server, &ct);
        } else {
            if (FD_ISSET(server.listen_socket, &sdata->readfds)) {
                DBG("server_add_client\n");
                server_add_client(&server);
            }
            server_main_switch(&server, sdata);
        }
//        server_print(&server);
    }

    return 0;
}

/**
 * @brief add a client to client_list
 * @param self :
 * @return on error returns -1, otherwise returns connected file describer
 *
 * This not only accept the connection but add the connected file describer to
 * the fd_set (readfds) for select. if any fail occurs, It should close the
 * connection which have connected.
 */
int server_add_client(struct server *self)
{
    int connfd;
    char peer_ip[INET_ADDRSTRLEN];
    unsigned short port;
    struct sockaddr_in peer_addr;
    socklen_t len;

    STEP_CHECK();

    assert(self);

    len = sizeof(peer_addr);
    connfd = accept(self->listen_socket, (struct sockaddr *)&peer_addr, &len);
    if (connfd < 0) {
        PRINTERR("accept failed: %s\n", strerror(errno));
        return -1;
    }
    
    len = sizeof(peer_ip);
    //const char *inet_ntop(int af, const void *src, char *dst, socklen_t size);
    inet_ntop(AF_INET, &peer_addr.sin_addr, peer_ip, len);
    if (peer_ip == NULL) {
        PRINTERR("inet_ntop failed\n");
        return -1;
    }
    port = ntohs(peer_addr.sin_port);

    DBG("receive client form %s:%d\n", peer_ip, port);

    memset(&self->c_array[self->client_num], 0, sizeof(struct client));
    self->c_array[self->client_num].id = connfd;
    self->client_num++;

    return connfd;
}

/**
 * @brief remove a client from client_list
 * @param self :
 * @param id : client id
 * @return on success returns 0, otherwise returns -1
 */
int server_remove_client(struct server *self, int id)
{
    int i;                    /* location */
    struct client *pc;          /* point client */

    pc = self->c_array;
    for (i = 0; i < self->client_num; i++) {
        if (pc[i].id == id) {
            break;
        }
    }

    if (i < self->client_num) {
        for (; i < self->client_num; i++) {
            pc[i] = pc[i + 1];
        }
        self->client_num--;
    }
    
    return 0;
}

/**
 * @brief print server all information for debug
 * @param self
 * @return void
 */
void server_print(struct server *self)
{
    int i;

    DBG("=================================\n");
    DBG("listen socket: %d, ip_addr: %s, port: %d， client_num: %d\n", 
        self->listen_socket, self->ip_addr, self->port, self->client_num);
    
    for (i = 0; i < self->client_num; i++) {
        client_print(&self->c_array[i]);
    }

}

/**
 * @brief register listen socket and connected socket of clients to select
 * @param self
 * @return on success returns 0, otherwise returns -1
 */
int server_register_select(struct server *self, struct select_data *sdata)
{
    int i;

//    STEP_CHECK();

    select_register_readfds(sdata, self->input_fd);

    select_register_readfds(sdata, self->listen_socket);
    
    for (i = 0; i < self->client_num ; i++) {
        select_register_readfds(sdata, self->c_array[i].id);
    }

    return 0;
}

/**
 * @brief get result from clients
 * @param self
 * @return on success returns 0, otherwise returns -1
 */
int server_get_result(struct server *self)
{
    int i;
    int sum;

    sum = 0;
    for (i = 0; i < self->client_num; i++) {
        sum += self->c_array[i].data.three;
    }

    self->data.th = sum;

    return 0;
}
