//
//  runtime_opengl.c
//  FrameworkPrototype
//
//  Created by Wayne Hoferkamp on 1/5/12.
//  Copyright 2012 NA. All rights reserved.
//

#include <stdbool.h>

#include "SDL.h"
#include "SDL_opengl.h"
#include "SDL_image.h"
#include "SDL_ttf.h"

#include "draw.h"
#include "framework.h"
#include "runtime.h"

// public

runtime_t runtime;

// private

void runtime_graphics_init(void);
void runtime_fonts_init(void);
void runtime_graphics_open(void);
void runtime_event_loop(void);
void runtime_screen_render(void);
void runtime_event_clear(input_state_ptr_t input_state_ptr);
void runtime_event(SDL_Event* event, input_state_ptr_t input_state_ptr);
mod_key_t runtime_event_mod_key(SDLMod sdl_mod_key);
void runtime_fonts_exit(void);
void runtime_graphics_exit(void);

// private utility

// ...

// implementation

void runtime_init()
{
    /*** debug ***/ fprintf(stdout, "[Trace] runtime_init()\n");
    
    runtime.screen_width = 0;
    runtime.screen_height = 0;
    runtime.fps = 0;
    runtime.resource_dir_ptr = RESOURCE_DIR;
    runtime.surface_ptr = NULL;
    runtime.frames = 0;
    
    srand(0);
    
    runtime_graphics_init();
    runtime_fonts_init();

    // framework must set screen size and fps on pre-init
    framework_alloc();
    framework_pre_init();
    runtime_graphics_open();
    framework_init();
}

void runtime_run()
{
    /*** debug ***/ fprintf(stdout, "[Trace] runtime_run()\n");
    
    runtime_event_loop();
}

void runtime_exit()
{
    /*** debug ***/ fprintf(stdout, "[Trace] runtime_exit()\n");
    
    framework_exit();
    framework_free();

    runtime_fonts_exit();
    runtime_graphics_exit();
}

void runtime_graphics_init()
{
    /*** debug ***/ fprintf(stdout, "[Trace] runtime_graphics_init()\n");
    
    Uint32 init_flags = SDL_INIT_VIDEO;
    
    // initialize the SDL library
    if (SDL_Init(init_flags) < 0)
    {
        fprintf(stderr, "[Error] SDL_Init(): %s\n", SDL_GetError());
        exit(-1);
    }

    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
}

void runtime_fonts_init()
{
    /*** debug ***/ fprintf(stdout, "[Trace] runtime_fonts_init()\n");
    
    if (TTF_Init() < 0)
    {
        fprintf(stderr, "[Error] TTF_init(): %s\n", TTF_GetError());
        exit(-1);
    }
}

void runtime_graphics_open()
{
    /*** debug ***/ fprintf(stdout, "[Trace] runtime_graphics_open()\n");
    
    Uint8 video_bpp = 32;
    // Uint32 video_flags = SDL_HWSURFACE | SDL_DOUBLEBUF;
    Uint32 video_flags = SDL_OPENGL | SDL_GL_DOUBLEBUFFER;
	
    // set video mode
    // framework must set screen size and fps on init
    /*** debug ***/ fprintf(stdout, "[Debug] runtime.screen_width = %d\n", runtime.screen_width);
    /*** debug ***/ fprintf(stdout, "[Debug] runtime.screen_height = %d\n", runtime.screen_height);
    /*** debug ***/ fprintf(stdout, "[Debug] runtime.fps = %d\n", runtime.fps);
    runtime.surface_ptr = SDL_SetVideoMode(runtime.screen_width, runtime.screen_height, video_bpp, video_flags);
    if (runtime.surface_ptr == NULL)
    {
        fprintf(stderr, "[Error] SDL_SetVideoMode: %s\n", SDL_GetError());
        SDL_Quit();
        exit(-1);
    }

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glShadeModel(GL_SMOOTH);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    glViewport(0, 0, runtime.screen_width, runtime.screen_height);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0, (float) runtime.screen_width / (float) runtime.screen_height, 1.0, 100.0);    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void runtime_event_loop()
{
    /*** debug ***/ fprintf(stdout, "[Trace] runtime_event_loop()\n");
    
	SDL_Event event;
    input_state_ptr_t input_state_ptr = input_state_alloc();
	
	/*** debug ***/ int start = SDL_GetTicks();
	while (framework_get_screen() != NULL)
	{
		// *** TODO: should render be at top of loop?
		// 1. render frame
		// 2. vsync
		// 3. user sees screen and reacts
		// 4. poll input
		// 5. update model
		runtime_screen_render();
		// handle the next pending event
        runtime_event_clear(input_state_ptr);
		if (SDL_PollEvent(&event))
		{
			switch (event.type)
			{
				case SDL_MOUSEMOTION:
				case SDL_MOUSEBUTTONDOWN:
                case SDL_MOUSEBUTTONUP:
				case SDL_KEYDOWN:
                case SDL_KEYUP:
                    runtime_event(&event, input_state_ptr);
                    break;
				case SDL_QUIT:
                    // force the game to quit
                    game_set_screen(NULL);
					break;
				default:
					break;
			}
		}
		// respond to user action and update model (as long as there was no quit event)
        if (framework_get_screen() != NULL)
        {
            framework_input_task(input_state_ptr);
        }
	}
	/*** debug ***/ int end = SDL_GetTicks();
	/*** debug ***/ fprintf(stdout, "[Debug] frames = %d\n", runtime.frames);
	/*** debug ***/ fprintf(stdout, "[Debug] seconds = %f\n", (end - start) / 1000.0);
	/*** debug ***/ fprintf(stdout, "[Debug] frames per second = %f\n", runtime.frames / ((end - start) / 1000.0));
    
    input_state_free(&input_state_ptr);
}

void runtime_screen_render()
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_screen_render()\n");
    
	// render current state of model
    // call framework active screen render (isolated from SDL, calls runtime)
    framework_render_frame();
	// *** TODO: does SDL_GL_SwapBuffers wait for vsync?
	SDL_GL_SwapBuffers();
	runtime.frames++;
	// wait until next frame
    /**/
	// *** TODO: is this delay necessary if SDL_Flip waits for vsync?
	double ticks_per_frame = 1000.0 / runtime.fps;
	Uint32 ticks = SDL_GetTicks();
	Uint32 multiple = (ticks + ticks_per_frame + 0.5) / ticks_per_frame;
	Uint32 next_frame = multiple * ticks_per_frame + 0.5;
	Uint32 delay = next_frame - ticks;
    /**/
	/*** debug ***/ // fprintf(stderr, "next_frame = %d\n", next_frame);
    /**/
	SDL_Delay(delay);
    /**/
}

void runtime_event_clear(input_state_ptr_t input_state_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_event_clear()\n");
    
    if (input_state_ptr->event == event_button_up)
    {
        input_state_ptr->button = 0;
    }
    else if (input_state_ptr->event == event_key_up)
    {
        input_state_ptr->key = 0;
        input_state_ptr->mod_key = mod_key_none;
    }
    input_state_ptr->event = event_none;
}

void runtime_event(SDL_Event* event, input_state_ptr_t input_state_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_event()\n");
    
	// convert SDL_Event to input_state_t
    switch (event->type)
    {
        case SDL_MOUSEMOTION:
            input_state_ptr->event = event_move;
            input_state_ptr->x = event->motion.x;
            input_state_ptr->y = event->motion.y;
            break;
        case SDL_MOUSEBUTTONDOWN:
            /*** debug ***/ fprintf(stdout, "[Debug] (down) button  = %d\n", event->button.button);
            input_state_ptr->event = event_button_down;
            input_state_ptr->button = event->button.button;
            break;
        case SDL_MOUSEBUTTONUP:
            /*** debug ***/ fprintf(stdout, "[Debug] (up)   button  = %d\n", event->button.button);
            input_state_ptr->event = event_button_up;
            input_state_ptr->button = event->button.button;
            break;
        case SDL_KEYDOWN:
            /*** debug ***/ fprintf(stdout, "[Debug] (down) key sym = %d\n", event->key.keysym.sym);
            /*** debug ***/ fprintf(stdout, "[Debug] (down) key mod = %d\n", event->key.keysym.mod);
            input_state_ptr->event = event_key_down;
            input_state_ptr->key = event->key.keysym.sym;
            input_state_ptr->mod_key = runtime_event_mod_key(event->key.keysym.mod);
            break;
        case SDL_KEYUP:
            /*** debug ***/ fprintf(stdout, "[Debug] (up)   key sym = %d\n", event->key.keysym.sym);
            /*** debug ***/ fprintf(stdout, "[Debug] (down) key mod = %d\n", event->key.keysym.mod);
            input_state_ptr->event = event_key_up;
            input_state_ptr->key = event->key.keysym.sym;
            input_state_ptr->mod_key = runtime_event_mod_key(event->key.keysym.mod);
            break;
    }
}

mod_key_t runtime_event_mod_key(SDLMod sdl_mod_key)
{
    /*** debug ***/ fprintf(stdout, "[Trace] runtime_event_mod_key()\n");
    
    mod_key_t mod_key = mod_key_none;
    switch (sdl_mod_key)
    {
        case KMOD_LSHIFT:
        case KMOD_RSHIFT:
            /*** debug ***/ fprintf(stdout, "[Debug] mod key = shift\n");
            mod_key = mod_key_shift;
            break;
        case KMOD_LCTRL:
        case KMOD_RCTRL:
            /*** debug ***/ fprintf(stdout, "[Debug] mod key = control\n");
            mod_key = mod_key_control;
            break;
        case KMOD_LALT:
        case KMOD_RALT:
            /*** debug ***/ fprintf(stdout, "[Debug] mod key = alt\n");
            mod_key = mod_key_alt;
            break;
        case KMOD_LMETA:
        case KMOD_RMETA:
            /*** debug ***/ fprintf(stdout, "[Debug] mod key = command\n");
            mod_key = mod_key_command;
            break;
        default:
            break;
    }
    return mod_key;
}

void runtime_fonts_exit()
{
    /*** debug ***/ fprintf(stdout, "[Trace] runtime_fonts_exit()\n");
    
	TTF_Quit();
}

void runtime_graphics_exit()
{
    /*** debug ***/ fprintf(stdout, "[Trace] runtime_graphics_exit()\n");
    
	SDL_Quit();
}

// utility

// runtime functions necessary for sprites, text, menus (to be called by framework)

int runtime_alloc_file_surface(surface_ptr_t* surface_ptr_ptr, char* file_name_ptr, color_ptr_t color_key_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_alloc_file_surface()\n");

	SDL_Surface* file_surface_ptr;
    char dir_file_name_array[strlen(runtime.resource_dir_ptr) + 1 + strlen(file_name_ptr) + 1];
    sprintf(dir_file_name_array, "%s/%s", runtime.resource_dir_ptr, file_name_ptr);
	file_surface_ptr = IMG_Load(dir_file_name_array);
	if (file_surface_ptr == NULL)
	{
		fprintf(stderr, "[Error] IMG_Load(): %s\n", SDL_GetError());
        *(surface_ptr_ptr) = NULL;
		return -1;
	}
    *(surface_ptr_ptr) = malloc(sizeof(GLuint));
    glGenTextures(1, *(surface_ptr_ptr));
    glBindTexture(GL_TEXTURE_2D, *(*((GLuint**) surface_ptr_ptr)));
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    GLenum texture_format;
    if (file_surface_ptr->format->BytesPerPixel == 4)
    {
        if (file_surface_ptr->format->Rmask == 0x000000ff)
        {
            texture_format = GL_RGBA;
        }
        else
        {
            texture_format = GL_BGRA;
        }
    }
    else if (file_surface_ptr->format->BytesPerPixel == 3)
    {
        if (file_surface_ptr->format->Rmask == 0x000000ff)
        {
            texture_format = GL_RGB;
        }
        else
        {
            texture_format = GL_BGR;
        }
    }
    else
    {
        fprintf(stderr, "[Error] Unsupported image format: %s\n", dir_file_name_array);
        SDL_Quit();
        exit(-1);
    }
    glTexImage2D(GL_TEXTURE_2D, 0, file_surface_ptr->format->BytesPerPixel, file_surface_ptr->w, file_surface_ptr->h, 0, texture_format, GL_UNSIGNED_BYTE, file_surface_ptr->pixels);
	SDL_FreeSurface(file_surface_ptr);
    return 0;
}

void runtime_alloc_color_surface(surface_ptr_t* surface_ptr_ptr, color_ptr_t color_ptr, int width, int height, color_ptr_t color_key_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_alloc_color_surface()\n");

    *(surface_ptr_ptr) = malloc(sizeof(GLuint));
    glGenTextures(1, *(surface_ptr_ptr));
    /*** debug ***/ fprintf(stdout, "[Debug] blank texture id = %u\n", *(*((GLuint**) surface_ptr_ptr)));
    glBindTexture(GL_TEXTURE_2D, *(*((GLuint**) surface_ptr_ptr)));
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, 4, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
}

int runtime_alloc_font_surface(surface_ptr_t* surface_ptr_ptr, char* font_name_ptr, int font_size, color_ptr_t color_ptr, char* string_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_alloc_font_surface()\n");

    // *** TODO: optimize font loading by caching and resusing fonts (and freeing at exit)
	TTF_Font* font_ptr;
    char dir_file_name_array[strlen(runtime.resource_dir_ptr) + 1 + strlen(font_name_ptr) + 4 + 1];
    sprintf(dir_file_name_array, "%s/%s.ttf", runtime.resource_dir_ptr, font_name_ptr);
	font_ptr = TTF_OpenFont(dir_file_name_array, font_size);
	if (font_ptr == NULL)
	{
		fprintf(stderr, "[Error] TTF_OpenFont(): %s\n", TTF_GetError());
        *(surface_ptr_ptr) = NULL;
		return -1;
	}
    SDL_Color sdl_color = { color_ptr->r, color_ptr->g, color_ptr->b, 0 };
    SDL_Surface* font_surface_ptr = TTF_RenderText_Blended(font_ptr, string_ptr, sdl_color);
    TTF_CloseFont(font_ptr);
    *(surface_ptr_ptr) = malloc(sizeof(GLuint));
    glGenTextures(1, *(surface_ptr_ptr));
    glBindTexture(GL_TEXTURE_2D, *(*((GLuint**) surface_ptr_ptr)));
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    SDL_Surface* temp_surface_ptr = SDL_CreateRGBSurface(0, font_surface_ptr->w, font_surface_ptr->h, 32, 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000);
    SDL_BlitSurface(font_surface_ptr, 0, temp_surface_ptr, 0);
    glTexImage2D(GL_TEXTURE_2D, 0, 4, temp_surface_ptr->w, temp_surface_ptr->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, temp_surface_ptr->pixels);
    SDL_FreeSurface(temp_surface_ptr);
    SDL_FreeSurface(font_surface_ptr);
    return 0;
}

void runtime_free_surface(surface_ptr_t* surface_ptr_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_free_surface()\n");
    
    /*** debug ***/ fprintf(stdout, "[Debug] free texture id = %u\n", *(*((GLuint**) surface_ptr_ptr)));
    glDeleteTextures(1, *(surface_ptr_ptr));
    free(*(surface_ptr_ptr));
    *(surface_ptr_ptr) = NULL;
}

void runtime_blit_surface(surface_ptr_t surface_ptr, int x, int y)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_blit_surface()\n");

    int w = runtime_surface_width(surface_ptr);
    int h = runtime_surface_height(surface_ptr);

    // switch to 2D
    glDisable(GL_DEPTH_TEST);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, runtime.screen_width, runtime.screen_height, 0, -1, 1);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_COLOR, GL_ONE_MINUS_SRC_ALPHA);

    // render texture
    glBindTexture(GL_TEXTURE_2D, *((GLuint*) surface_ptr));
    glBegin(GL_QUADS);
    glTexCoord2i(0, 0);
    glVertex3i(x, y, 0);
    glTexCoord2i(1, 0);
    glVertex3i(x + w, y, 0);
    glTexCoord2i(1, 1);
    glVertex3i(x + w, y + h, 0);
    glTexCoord2i(0, 1);
    glVertex3i(x, y + h, 0);
    glEnd();

    // switch to 3D
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0, (float) runtime.screen_width / (float) runtime.screen_height, 0.1, 50.0);    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
}

void runtime_blit_surface_to_surface(surface_ptr_t src_surface_ptr, surface_ptr_t dest_surface_ptr, int x, int y)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_blit_surface_to_surface()\n");

    /*
    SDL_Surface* sdl_src_surface_ptr = src_surface_ptr;
    SDL_Surface* sdl_dest_surface_ptr = dest_surface_ptr;
    SDL_Rect dest_rec;
    dest_rec.x = x;
    dest_rec.y = y;
    dest_rec.w = 0;
    dest_rec.h = 0;
    SDL_BlitSurface(sdl_src_surface_ptr, NULL, sdl_dest_surface_ptr, &dest_rec);
    */
    fprintf(stderr, "[Error] Unsupported function: runtime_blit_surface_to_surface\n");
    SDL_Quit();
    exit(-1);
}

void runtime_clear_background()
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_clear_background()\n");

    glClearColor(0.5, 0.5, 0.5, 0.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

int runtime_surface_width(surface_ptr_t surface_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_surface_width()\n");

    glBindTexture(GL_TEXTURE_2D, *((GLuint*) surface_ptr));
    int width;
    glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &width);
    return width;
}

int runtime_surface_height(surface_ptr_t surface_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_surface_height()\n");

    glBindTexture(GL_TEXTURE_2D, *((GLuint*) surface_ptr));
    int height;
    glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &height);
    return height;
}

int runtime_rand(int range)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_rand()\n");

    return rand() / (RAND_MAX + 1.0) * range;
}

void runtime_draw_line(surface_ptr_t* surface_ptr, int x1, int y1, int x2, int y2, color_ptr_t color_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_draw_line()\n");

    /*
    SDL_Surface* sdl_surface_ptr = (SDL_Surface*) surface_ptr;
    Uint32 sdl_color = SDL_MapRGB(sdl_surface_ptr->format, color_ptr->r, color_ptr->g, color_ptr->b);
    SDL_LockSurface(sdl_surface_ptr);
    draw_line(sdl_surface_ptr, x1, y1, x2, y2, sdl_color);
    SDL_UnlockSurface(sdl_surface_ptr);
    */
    
    fprintf(stderr, "[Error] Unsupported function: runtime_draw_line\n");
    SDL_Quit();
    exit(-1);
}

void runtime_draw_circle(surface_ptr_t* surface_ptr, int x, int y, int r, color_ptr_t color_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_draw_circle()\n");

    /*
    SDL_Surface* sdl_surface_ptr = (SDL_Surface*) surface_ptr;
    Uint32 sdl_color = SDL_MapRGB(sdl_surface_ptr->format, color_ptr->r, color_ptr->g, color_ptr->b);
    SDL_LockSurface(sdl_surface_ptr);
    draw_circle(sdl_surface_ptr, x, y, r, sdl_color);
    SDL_UnlockSurface(sdl_surface_ptr);
    */
    
    fprintf(stderr, "[Error] Unsupported function: runtime_draw_circle\n");
    SDL_Quit();
    exit(-1);
}

surface_ptr_t runtime_scale_surface(surface_ptr_t surface_ptr, int width, int height)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_scale_surface()\n");
    
    /*
    SDL_Surface* unscaled_surface_ptr = (SDL_Surface*) surface_ptr;
    float x_scale = (float) width / (float) unscaled_surface_ptr->w;
    float y_scale = (float) height / (float) unscaled_surface_ptr->h;
    SDL_Surface* scaled_surface_ptr = SDL_CreateRGBSurface(unscaled_surface_ptr->flags, width, height, unscaled_surface_ptr->format->BitsPerPixel, unscaled_surface_ptr->format->Rmask, unscaled_surface_ptr->format->Gmask, unscaled_surface_ptr->format->Bmask, unscaled_surface_ptr->format->Amask);
    SDL_LockSurface(scaled_surface_ptr);
    draw_rotate_scale_surface(scaled_surface_ptr, unscaled_surface_ptr, 0.0, x_scale, y_scale);
    SDL_UnlockSurface(scaled_surface_ptr);
    return scaled_surface_ptr;
    */

    fprintf(stderr, "[Error] Unsupported function: runtime_scale_surface\n");
    SDL_Quit();
    exit(-1);
}

void runtime_fill_surface(surface_ptr_t surface_ptr, color_ptr_t color_ptr)
{
    /*** debug ***/ // fprintf(stdout, "[Trace] runtime_fill_surface()\n");

    /*
    SDL_Surface* sdl_surface_ptr = (SDL_Surface*) surface_ptr;
    Uint32 sdl_color = SDL_MapRGB(sdl_surface_ptr->format, color_ptr->r, color_ptr->g, color_ptr->b);
    SDL_LockSurface(sdl_surface_ptr);
    SDL_FillRect(sdl_surface_ptr, 0, sdl_color);
    SDL_UnlockSurface(sdl_surface_ptr);
    */
    
    fprintf(stderr, "[Error] Unsupported function: runtime_fill_surface\n");
    SDL_Quit();
    exit(-1);
}
