#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <glib.h>

#include "ports.h"
#include "repository.h"
#include "collection.h"
#include "driver_base.h"
#include "debug.h"
#include "common.h"

// cvsup
static const char* DEFAULT_BASE = "*default base=\0";
// httpup
static const char* ROOT_DIR = "ROOT_DIR=\0";
// rsync
static const char* DESTINATION_DIR = "destination=\0";

static const char* SHELL = "/bin/bash\0";


static int driver_exec(Ports* ports, const char* cmd, char** stdresult)
{
	pid_t pid;
	int fdpipe[2];
	int status, res, flags;
	int ret = 0;

	pipe (fdpipe);
	pid = fork ();
	
	if (pid < 0)
		return (-1);
        else if (pid == 0) {
		/* Child process */
		close(fdpipe[0]);
		dup2(fdpipe[1], 1);
		dup2(fdpipe[1], 2);
		execv (cmd, NULL);
		_exit(EXIT_FAILURE);
	} else if (pid > 0) {
		/* Parent process */
		int lastbuflen, buflen, wpval, bytes;
		char buf[BUFSIZ];
		FILE* fpipe0;
		
		close(fdpipe[1]);
		
		*stdresult = NULL;
		while ((wpval = waitpid (pid, &status, WNOHANG)) == 0) {
			usleep(1);
		}
		buflen = 0;
		while ((bytes = read (fdpipe[0], buf, BUFSIZ-1)) != 0) {
			buf[bytes] = 0;
			lastbuflen = buflen;
			buflen += bytes;
			*stdresult = (char*)realloc(*stdresult, buflen+1);
			memset ((*stdresult)+lastbuflen, 0, bytes+1);
			strcat(*stdresult, buf);
		}
		
		ret = buflen;
		fflush (stdout);
		fflush (stderr);
	}
	return ret;
}

static int driver_exec_shell(Ports* ports, const char* cmd, char** stdresult)
{
	pid_t pid;
	int fdpipe[2];
	int status, res, flags;
	int ret = 0;

	pipe (fdpipe);
	pid = fork ();
	
	if (pid < 0) 
		return (-1);
        else if (pid == 0) {
		/* Child process */
		close(fdpipe[0]);
		dup2(fdpipe[1], 1);
		dup2(fdpipe[1], 2);
		execl( SHELL, SHELL, "-c", cmd, NULL);
		_exit(EXIT_FAILURE);
	} else if (pid > 0) {
		/* Parent process */
		int lastbuflen, buflen, wpval, bytes;
		char buf[BUFSIZ];
		FILE* fpipe0;
		
		close(fdpipe[1]);
		ports_on_refresh(ports);
		
		buflen = 0;
		*stdresult = NULL;
		do {
			bytes = read (fdpipe[0], buf, BUFSIZ-1);
			buf[bytes] = 0;
			lastbuflen = buflen;
			buflen += bytes;
			*stdresult = (char*)realloc(*stdresult, buflen+1);
			memset ((*stdresult)+lastbuflen, 0, bytes+1);
			strcat(*stdresult, buf);

			ports_on_refresh(ports);
			wpval = waitpid (pid, &status, WNOHANG);
		} while (bytes != 0 && !WIFEXITED(status));
		
		ret = buflen;
		fflush (stdout);
		fflush (stderr);
	}
	return ret;
}

DriverBase* driver_base_new(DriverType driver_type)
{
	DriverBase* driver = NULL;
	
	driver = (DriverBase*)calloc(1, sizeof(DriverBase));
	if(!driver) {
		debug_log(ERROR, "driver_base_new(): Insuffisant Memory");
		return NULL;
	}
	
	driver->driver_type = driver_type;
	driver->exec = (exec_func)driver_exec;
	driver->exec_shell = (exec_func)driver_exec_shell;
	switch(driver_type) {
		case CVSUP:
			strcpy(driver->command, drivers_dir);
			strcat(driver->command, "/cvsup");
			driver->load_config_file = cvsup_driver_load_config_file;
			break;
		case HTTPUP:
			strcpy(driver->command, drivers_dir);
			strcat(driver->command, "/httpup");
			driver->load_config_file = httpup_driver_load_config_file;
			break;
		case RSYNC:
			strcpy(driver->command, drivers_dir);
			strcat(driver->command, "/rsync");
			driver->load_config_file = rsync_driver_load_config_file;
			break;
	}
	
	return driver;
}

int cvsup_driver_load_config_file(Repository* repository)
{
	FILE* fh;
	char* buffer;
	int buflen = 0;
	char* newline;
	char base_directory[BUFSIZ];
	char collection[256];
	
	repository->collections = NULL;
	
	fh = fopen(repository->config_file, "r");
	if (!fh) {
		debug_log(ERROR, "cvsup_driver_load_config_file(): could not open %s file", repository->config_file);
		return (-1);
	}
	
	memset(base_directory, 0, sizeof(base_directory));
	while (getline(&buffer, &buflen, fh) != -1) {
		if (*buffer == '*' && strncmp(buffer, DEFAULT_BASE, strlen(DEFAULT_BASE)) == 0) {
			strcpy(base_directory, buffer+strlen(DEFAULT_BASE));
		}
		if (*buffer != '#' && *buffer != '*' && *buffer != 0) {
			char *wspos = strchr(buffer, ' ');
			if (wspos)
				*wspos = 0;
			repository->collections = g_list_append(repository->collections, collection_new(repository, buffer));
		}
	}
	fclose (fh);
	if (buffer)
		free(buffer);
	
	if (*base_directory == 0) {
		debug_log(ERROR, "cvsup_driver_load_config_file(): could not load base directory from %s file", repository->config_file);
		return (-1);
	}
	if (!repository->collections) {
		debug_log(ERROR, "cvsup_driver_load_config_file(): could not collection(s) from %s file", repository->config_file);
		return (-1);
	}
	strcpy(repository->base_directory, base_directory);
	repository->collections = g_list_sort (repository->collections, compare_collection_list_data);
	
	return 0;
}

int httpup_driver_load_config_file(Repository* repository)
{
	FILE* fh;
	char* buffer;
	int buflen = 0;
	char* newline;
	char base_directory[BUFSIZ];
	char collection[256];
	char* lastslashpos;
	
	repository->collections = NULL;
	
	fh = fopen(repository->config_file, "r");
	if (!fh) {
		debug_log(ERROR, "httpup_driver_load_config_file(): could not open %s file", repository->config_file);
		return (-1);
	}
	
	memset(base_directory, 0, sizeof(base_directory));
	while (getline(&buffer, &buflen, fh) != -1) {
		newline = strchr(buffer, '\n');
		if (newline)
			*newline = 0;
		if (*buffer != '#' && strncmp(buffer, ROOT_DIR, strlen(ROOT_DIR)) == 0) {
			strcpy(base_directory, buffer+strlen(ROOT_DIR));
			lastslashpos = strrchr(base_directory, '/');
			*lastslashpos = 0;
			repository->collections = g_list_append(repository->collections, collection_new(repository, lastslashpos+1));
		}
	}
	fclose (fh);
	if (buffer)
		free(buffer);
	
	if (*base_directory == 0) {
		debug_log(ERROR, "httpup_driver_load_config_file(): could not load base directory from %s file", repository->config_file);
		return (-1);
	}
	if (!repository->collections) {
		debug_log(ERROR, "httpup_driver_load_config_file(): could not collection(s) from %s file", repository->config_file);
		return (-1);
	}
	strcpy(repository->base_directory, base_directory);
	repository->collections = g_list_sort (repository->collections, compare_collection_list_data);

	return 0;
}


int rsync_driver_load_config_file(Repository* repository)
{
	FILE* fh;
	char* buffer;
	int buflen = 0;
	char* newline;
	char base_directory[BUFSIZ];
	char collection[256];
	char* lastslashpos;
	
	repository->collections = NULL;
	
	fh = fopen(repository->config_file, "r");
	if (!fh) {
		debug_log(ERROR, "rsync_driver_load_config_file(): could not open %s file", repository->config_file);
		return (-1);
	}
	
	memset(base_directory, 0, sizeof(base_directory));
	while (getline(&buffer, &buflen, fh) != -1) {
		newline = strchr(buffer, '\n');
		if (newline)
			*newline = 0;
		if (*buffer != '#' && strncmp(buffer, DESTINATION_DIR, strlen(DESTINATION_DIR)) == 0) {
			strcpy(base_directory, buffer+strlen(DESTINATION_DIR));
			lastslashpos = strrchr(base_directory, '/');
			*lastslashpos = 0;
			repository->collections = g_list_append(repository->collections, collection_new(repository, lastslashpos+1));
		}
	}
	fclose (fh);
	if (buffer)
		free(buffer);
	
	if (*base_directory == 0) {
		debug_log(ERROR, "rsync_driver_load_config_file(): could not load base directory from %s file", repository->config_file);
		return (-1);
	}
	if (!repository->collections) {
		debug_log(ERROR, "rsync_driver_load_config_file(): could not collection(s) from %s file", repository->config_file);
		return (-1);
	}
	strcpy(repository->base_directory, base_directory);
	repository->collections = g_list_sort (repository->collections, compare_collection_list_data);

	return 0;
}

gint compare_collection_list_data(gconstpointer object1, gconstpointer object2)
{
	return strcmp(((Collection*)object1)->name, ((Collection*)object2)->name);
}
