/*--------------------------------------------------------------------------------*
 *               Document Generator Project
 *			        by Peter Antoine
 *
 *    file: utf8
 *    desc: This file holds the UTF-8 handling functions.
 *
 *  author: 
 *    date: 25/09/2012 17:50:23
 *--------------------------------------------------------------------------------*
 *                     Copyright (c) 2012 Peter Antoine
 *                            All rights Reserved.
 *                    Released Under the Artistic Licence
 *--------------------------------------------------------------------------------*/

#include "utf8.h"

static unsigned char	utf8_valid_start_byte[] =
{
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
	0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 
	0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 
	0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x05, 0x05, 0x05, 0x05, 0x06, 0x06, 0x00, 0x00
};

/*--------------------------------------------------------------------------------*
 * We don't really care what is in the UTF-8 char not actually doing any rendering
 * or decoding, just need to be able to read the file so it can be parsed. 
 *
 * Valid UTF-8 character patterns:
 *
 *  7	U+007F		0xxxxxxx
 *  11 U+07FF		110xxxxx 10xxxxxx
 *  16 U+FFFF		1110xxxx 10xxxxxx 10xxxxxx
 *  21 U+1FFFFF		11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
 *  26 U+3FFFFFF	111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
 *  31 U+7FFFFFFF	1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 
 *
 *  overlong byte patterns: (these must not be accepted)
 *
 *  1100000x 		  (10xxxxxx) 
 *  11100000 100xxxxx (10xxxxxx) 
 *  11110000 1000xxxx (10xxxxxx 10xxxxxx) 
 *  11111000 10000xxx (10xxxxxx 10xxxxxx 10xxxxxx) 
 *  11111100 100000xx (10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx)
 *--------------------------------------------------------------------------------*/
#define UTF8_1_BYTE_MASK	((unsigned char) 0x7F)
#define UTF8_1_BYTE_VALUE	((unsigned char) 0x00)

#define UTF8_2_BYTE_MASK	((unsigned char) 0xE0)
#define UTF8_2_BYTE_VALUE	((unsigned char) 0xC0)

#define UTF8_3_BYTE_MASK	((unsigned char) 0xF0)
#define UTF8_3_BYTE_VALUE	((unsigned char) 0xE0)

#define UTF8_4_BYTE_MASK	((unsigned char) 0xF8)
#define UTF8_4_BYTE_VALUE	((unsigned char) 0xF0)

#define UTF8_5_BYTE_MASK	((unsigned char) 0xFC)
#define UTF8_5_BYTE_VALUE	((unsigned char) 0xF8)

#define UTF8_6_BYTE_MASK	((unsigned char) 0xFE)
#define UTF8_6_BYTE_VALUE	((unsigned char) 0xFC)

#define UTF8_CONTIN_MASK	((unsigned char) 0xC0)
#define UTF8_CONTIN_VALUE	((unsigned char) 0x80)

/*----- FUNCTION -----------------------------------------------------------------*
 *  name: utf8_char_size
 *  desc: This function will get the chartacter size of the utf8 in the buffer.
 *        It returns zero if the character is not a valid utf start char. It does
 *        not validate the utf8.
 *--------------------------------------------------------------------------------*/
unsigned char	utf8_char_size(unsigned char* buffer)
{
	unsigned char	result = 0;

	result = utf8_valid_start_byte[buffer[0]];

	return result;
}



