/*
Copyright (c) 2011, Christopher Utz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of SimpleHttp nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTOPHER UTZ BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Author: Christopher Utz <utz.m.christopher@gmail.com>
*/

#ifndef SHTTP_HASH_H_
#define SHTTP_HASH_H_

#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>

#include <shttp_error.h>

typedef unsigned (*shttp_hash_func_t)(void *key, size_t key_size);
typedef bool (*shttp_hash_key_comparator_t)(void *key1, size_t key1_size, void *key2, size_t key2_size);
typedef void (*shttp_hash_key_destructor_t)(void *key, size_t key_size);
typedef void (*shttp_hash_value_destructor_t)(void *value);

typedef struct shttp_hash_stats_t {
    double avg_dist;
    int max_dist;
    int num_allocated;
} shttp_hash_stats_t;

typedef struct shttp_hash_ops_t {
    shttp_hash_func_t hash_func;
    shttp_hash_key_comparator_t key_comparator;
    shttp_hash_key_destructor_t key_destructor;
    shttp_hash_value_destructor_t value_destructor;
} shttp_hash_ops_t;

typedef struct shttp_hash_elem_t {
    void *key;
    size_t key_size;
    unsigned key_hash;
    void *value;
} shttp_hash_elem_t;

typedef struct shttp_hash_t {
    shttp_hash_elem_t *elems;
    int length;
    int num_allocated;
    shttp_hash_ops_t ops;
} shttp_hash_t;

typedef struct shttp_hash_iter_t {
    int index;
    shttp_hash_t *hash;
} shttp_hash_iter_t;

shttp_error_t       shttp_hash_init                    (shttp_hash_t *hash,
                                                        int length_hint, 
                                                        shttp_hash_ops_t ops);

void                shttp_hash_cleanup                 (shttp_hash_t *hash);

bool                shttp_hash_set                     (shttp_hash_t *hash, 
                                                        void *key, 
                                                        size_t key_size, 
                                                        void *value);

void *              shttp_hash_get                     (shttp_hash_t *hash, 
                                                        void *key, 
                                                        size_t key_size);

void *              shttp_hash_remove                  (shttp_hash_t *hash, 
                                                        void *key, 
                                                        size_t key_size);

int                 shttp_hash_length                  (shttp_hash_t *hash);

void                shttp_hash_clear                   (shttp_hash_t *hash);

void                shttp_hash_stats                   (shttp_hash_t *hash, 
                                                        shttp_hash_stats_t *stats);

shttp_error_t       shttp_hash_iter_init               (shttp_hash_iter_t *iter,
                                                        shttp_hash_t *hash);

bool                shttp_hash_iter_valid              (shttp_hash_iter_t *iter);

bool                shttp_hash_iter_next               (shttp_hash_iter_t *iter);

bool                shttp_hash_iter_this               (shttp_hash_iter_t *iter, 
                                                        void **key, 
                                                        size_t *key_size, 
                                                        void **value);

unsigned            shttp_oat_hash                     (void *key, 
                                                        size_t key_size);

bool                shttp_compare_memory               (void *key1, 
                                                        size_t key1_size, 
                                                        void *key2, 
                                                        size_t key2_size);

#endif // SHTTP_HASH_H_
