/*
Copyright (c) 2011, Christopher Utz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of SimpleHttp nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTOPHER UTZ BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Author: Christopher Utz <utz.m.christopher@gmail.com>
*/

#include <shttp_buffer.h>

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <shttp_mem.h>

shttp_error_t
shttp_buffer_init(shttp_buffer_t *buffer, size_t allocate)
{
    assert(buffer != NULL);

    shttp_error_t ret;

    buffer->data = shttp_malloc(allocate);
    
    if (!buffer->data) {
        ret = SHTTP_MEM_ERROR;
        goto malloc_failed;
    }
    
    buffer->length = 0;
    buffer->allocated = allocate;
    
    return SHTTP_OK;

malloc_failed:
    return ret;
}

void
shttp_buffer_cleanup(shttp_buffer_t *buffer)
{
    if (buffer) {
        shttp_free(buffer->data);
        buffer->data = NULL;
    }
}

void
shttp_buffer_clear(shttp_buffer_t *buffer)
{
    assert(buffer != NULL);
    
    buffer->length = 0;
}

void
shttp_buffer_rtrim(shttp_buffer_t *buffer, const char *trim_chars)
{
    assert(buffer != NULL);
    assert(trim_chars != NULL);
    
    while (buffer->length > 0) {
        const char *trim_char;
        for (trim_char=trim_chars; *trim_char; trim_char++)
            if (*trim_char == buffer->data[buffer->length-1])
                break;
        if (!*trim_char)
            break;
        buffer->length--;
    }
}

bool
shttp_buffer_append(shttp_buffer_t *buffer, const char *data, size_t length)
{
    assert(buffer != NULL);
    assert(data != NULL);
    
    size_t new_length = buffer->length + length;
    
    if (new_length > buffer->allocated) {
        char *new_data = shttp_realloc(buffer->data, new_length);
        
        if (!new_data)
            return false;
        
        buffer->data = new_data;
        buffer->allocated = new_length;
    }
    
    memcpy(buffer->data+buffer->length, data, length);
    buffer->length += length;
    
    return true;
}

size_t
shttp_buffer_find_char(const shttp_buffer_t *buffer, size_t start_pos, char ch)
{
    assert(buffer != NULL);
    assert(start_pos < buffer->length);
    
    for (size_t i=start_pos; i < buffer->length; i++)
        if (buffer->data[i] == ch)
            return i;
    
    return buffer->length;
}

