/*
Copyright (c) 2011, Christopher Utz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of SimpleHttp nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTOPHER UTZ BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Author: Christopher Utz <utz.m.christopher@gmail.com>
*/

#include <shttp_error.h>

#include <stdio.h>
#include <assert.h>
#include <errno.h>
#include <string.h>

#include <shttp_mem.h>
#include <shttp_string.h>

#define SHTTP_LOG_MSG_SIZE 4096

static void
shttp_default_log_callback(shttp_log_level_t level, int line, char *file, 
    const char *msg)
{
    fprintf(stderr, 
            "[%s][%s:%d] - %s\n", 
            shttp_log_level_str(level), 
            file,
            line,
            msg);
}

static shttp_log_callback_t shttp_log_callback = shttp_default_log_callback;

void
shttp_log_(shttp_log_level_t level, int line, char *file, const char *fmt, ...)
{
    assert(fmt != NULL);

    va_list ap;
    char msg[SHTTP_LOG_MSG_SIZE];

    va_start(ap, fmt);
    vsnprintf(msg, sizeof(msg), fmt, ap);
    va_end(ap);
    
    shttp_log_callback(level, line, file, msg);
}

void
shttp_log_errno_(shttp_log_level_t level, int code, int line, char *file, 
    const char *fmt, ...)
{
    assert(fmt != NULL);
    
    va_list ap;
    char msg1[SHTTP_LOG_MSG_SIZE];
    char msg2[SHTTP_LOG_MSG_SIZE];
    char msg3[SHTTP_LOG_MSG_SIZE];
    
    va_start(ap, fmt);
    vsnprintf(msg1, sizeof(msg1), fmt, ap);
    va_end(ap);
    
    strerror_r(code, msg2, sizeof(msg2));

    snprintf(msg3, sizeof(msg3), "%s: %s", msg1, msg2);
    
    shttp_log_callback(level, line, file, msg3);
}

void
shttp_set_log_callback(shttp_log_callback_t callback)
{
    if (callback)
        shttp_log_callback = callback;
    else
        shttp_log_callback = shttp_default_log_callback;
}

const char *
shttp_log_level_str(shttp_log_level_t level)
{
    switch (level) {
        case SHTTP_DEBUG:
            return "DEBUG";
        case SHTTP_INFO:
            return "INFO";
        case SHTTP_WARN:
            return "WARN";
        case SHTTP_ERR:
            return "ERR";
        default:
            return "?";
    }
}
