/*
Copyright (c) 2011, Christopher Utz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of SimpleHttp nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTOPHER UTZ BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Author: Christopher Utz <utz.m.christopher@gmail.com>
*/

#include <shttp_hash.h>

#include <string.h>
#include <assert.h>

#include <shttp_mem.h>

static void *shttp_hash_elem_deleted = "";

#define XX_HASH_MAX_LOAD_FACTOR 0.7

#define shttp_hash_load_factor(_hash) \
    ((_hash)->length / ((float)(_hash)->num_allocated))

#define shttp_hash_min_elems(_length_hint) \
    shttp_max(4, 1 + ((int)((_length_hint) * (1.0 / XX_HASH_MAX_LOAD_FACTOR))));

static void
shttp_hash_free_elems(shttp_hash_t *hash)
{
    shttp_hash_elem_t *curr;

    // Don't waste time iterating if we aren't gonna do anything.
    if (!hash->ops.key_destructor && !hash->ops.value_destructor) {
        return;
    }

    for (int i=0; i < hash->num_allocated; i++) {
        curr = &hash->elems[i];

        if (curr->key && curr->key != shttp_hash_elem_deleted) {
            if (hash->ops.key_destructor) {
                hash->ops.key_destructor(curr->key, curr->key_size);
            }
            if (hash->ops.value_destructor) {
                hash->ops.value_destructor(curr->value);
            }
        }
    }
}

static bool
shttp_hash_rehash(shttp_hash_t *hash)
{
    int new_num_allocated = hash->num_allocated * 2;

    shttp_hash_elem_t *new_elems = shttp_malloc(
        sizeof(shttp_hash_elem_t) * new_num_allocated);

    if (!new_elems) {
        goto shttp_malloc_failed;
    }

    memset(new_elems, 0, sizeof(shttp_hash_elem_t) * new_num_allocated);

    shttp_hash_elem_t *curr;

    for (int i=0; i < hash->num_allocated; i++) {
        curr = &hash->elems[i];

        if (curr->key == shttp_hash_elem_deleted) {
            continue;
        } else if (curr->key) {
            unsigned i = curr->key_hash % new_num_allocated; 

            while (new_elems[i].key) {
                i = ((i == new_num_allocated-1) ? 0 : i+1);
            }
            
            new_elems[i].key = curr->key;
            new_elems[i].key_size = curr->key_size;
            new_elems[i].value = curr->value;
            new_elems[i].key_hash = curr->key_hash;
        }
    }

    shttp_free(hash->elems);

    hash->elems = new_elems;
    hash->num_allocated = new_num_allocated;

    return true;

shttp_malloc_failed:
    return false;
}

shttp_error_t
shttp_hash_init(shttp_hash_t *hash, int length_hint, shttp_hash_ops_t ops)
{
    assert(length_hint > 0);

    hash->length = 0;
    hash->num_allocated = 0;

    hash->elems = shttp_malloc(sizeof(shttp_hash_elem_t) * hash->num_allocated);
    
    if (!hash->elems) {
        goto malloc_failed;
    }

    hash->num_allocated = shttp_hash_min_elems(length_hint);

    memset(hash->elems, 0, sizeof(shttp_hash_elem_t) * hash->num_allocated);

    hash->ops = ops;

    if (!hash->ops.hash_func) {
        hash->ops.hash_func = shttp_hash_default_hash_func;
    }

    if (!hash->ops.key_comparator) {
        hash->ops.key_comparator = shttp_hash_default_key_comparator;
    }

    return SHTTP_OK;

shttp_malloc_failed:
    return SHTTP_MEM_ERROR;
}

void
shttp_hash_cleanup(shttp_hash_t *hash)
{
    if (hash) {
        shttp_hash_free_elems(hash);
        shttp_free(hash->elems);
    }
}


bool
shttp_hash_set(shttp_hash_t *hash, void *key, size_t key_size, void *value)
{
    assert(hash != NULL);
    assert(key != NULL);
    assert(key_size >= 0);

    if (shttp_hash_load_factor(hash) > XX_HASH_MAX_LOAD_FACTOR) {
        if (!shttp_hash_rehash(hash)) {
            return false;
        }
    }

    unsigned key_hash = hash->ops.hash_func(key, key_size);
    unsigned i = key_hash % hash->num_allocated; 

    shttp_hash_elem_t *curr;

    while (hash->elems[i].key) {
        curr = &hash->elems[i];

        if (curr->key == shttp_hash_elem_deleted) {
            break;
        } else if (hash->ops.key_comparator(curr->key, curr->key_size, key, 
            key_size)) {

            if (hash->ops.key_destructor) {
                hash->ops.key_destructor(curr->key, curr->key_size);
            }

            if (hash->ops.value_destructor) {
                hash->ops.value_destructor(curr->value);
            }

            curr->key = key;
            curr->key_size = key_size;
            curr->value = value;
            curr->key_hash = key_hash;

            return true;
        }

        i = ((i == hash->num_allocated-1) ? 0 : i+1);
    }

    hash->elems[i].key = key;
    hash->elems[i].key_size = key_size;
    hash->elems[i].value = value;
    hash->elems[i].key_hash = key_hash;

    hash->length++;

    return true;
}

void *
shttp_hash_get(shttp_hash_t *hash, void *key, size_t key_size)
{
    assert(hash != NULL);
    assert(key != NULL);
    assert(key_size >= 0);

    assert(hash->length < hash->num_allocated);

    unsigned key_hash = hash->ops.hash_func(key, key_size);
    unsigned i = key_hash % hash->num_allocated; 

    shttp_hash_elem_t *curr;

    while (hash->elems[i].key) {
        curr = &hash->elems[i];
        
        if (curr->key != shttp_hash_elem_deleted &&
            hash->ops.key_comparator(curr->key, curr->key_size, key, 
            key_size)) {
            return curr->value;
        }

        i = ((i == hash->num_allocated-1) ? 0 : i+1);
    }

    return NULL;
}

void *
shttp_hash_remove(shttp_hash_t *hash, void *key, size_t key_size)
{
    assert(hash != NULL);
    assert(key != NULL);
    assert(key_size >= 0);

    assert(hash->length < hash->num_allocated);

    unsigned key_hash = hash->ops.hash_func(key, key_size);
    unsigned i = key_hash % hash->num_allocated; 

    shttp_hash_elem_t *curr;

    while (hash->elems[i].key) {
        curr = &hash->elems[i];
        
        if (curr->key != shttp_hash_elem_deleted &&
            hash->ops.key_comparator(curr->key, curr->key_size, key, 
            key_size)) {

            if (hash->ops.key_destructor) {
                hash->ops.key_destructor(curr->key, curr->key_size);
            }

            curr->key = shttp_hash_elem_deleted;
            hash->length--;

            return curr->value;
        }

        i = ((i == hash->num_allocated-1) ? 0 : i+1);
    }

    return NULL;
}

int
shttp_hash_length(shttp_hash_t *hash)
{
    assert(hash != NULL);

    return hash->length;
}

void
shttp_hash_clear(shttp_hash_t *hash)
{
    assert(hash != NULL);

    hash->length = 0;

    shttp_hash_free_elems(hash);
    memset(hash->elems, 0, sizeof(shttp_hash_elem_t) * hash->num_allocated);
}

void
shttp_hash_stats(shttp_hash_t *hash, shttp_hash_stats_t *stats)
{
    assert(hash != NULL);

    shttp_hash_elem_t *curr;
    int ideal, dist;

    stats->avg_dist = 0;
    stats->max_dist = 0;
    stats->num_allocated = hash->num_allocated;

    for (int i=0; i < hash->num_allocated; i++) {
        curr = &hash->elems[i];

        if (curr->key && curr->key != shttp_hash_elem_deleted) {
            ideal = curr->key_hash % hash->num_allocated;

            if (i >= ideal) {
                dist = i - ideal;
            } else {
                dist = i - ideal + hash->num_allocated;
            }

            stats->max_dist = shttp_max(dist, stats->max_dist);
            stats->avg_dist += dist;
        }
    }

    if (hash->length > 0) {
        stats->avg_dist /= hash->length;
    }
}

shttp_error_t
shttp_hash_iter_init(shttp_hash_iter_t *iter, shttp_hash_t *hash)
{
    assert(iter != NULL);
    assert(hash != NULL);

    for(iter->index=0; iter->index < hash->num_allocated; iter->index++) {
        if (hash->elems[iter->index].key &&
            hash->elems[iter->index].key != shttp_hash_elem_deleted) {
            break;
        }
    }

    iter->hash = hash;

    return SHTTP_OK;
}

bool
shttp_hash_iter_valid(shttp_hash_iter_t *iter)
{
    assert(iter != NULL);

    return iter->index < iter->hash->num_allocated;
}

bool
shttp_hash_iter_next(shttp_hash_iter_t *iter)
{
    assert(iter != NULL);

    iter->index++;

    for(; iter->index < iter->hash->num_allocated; iter->index++) {
        if (iter->hash->elems[iter->index].key &&
            iter->hash->elems[iter->index].key != shttp_hash_elem_deleted) {
            break;
        }
    }

    return shttp_hash_iter_valid(iter);
}

bool
shttp_hash_iter_this(shttp_hash_iter_t *iter, void **key, size_t *key_size, void **value)
{
    assert(iter != NULL);

    if (shttp_hash_iter_valid(iter)) {
        if (key) {
            *key = iter->hash->elems[iter->index].key;
        }
        if (key_size) {
            *key_size = iter->hash->elems[iter->index].key_size;
        }
        if (value) {
            *value = iter->hash->elems[iter->index].value;
        }
    } else {
        return false;
    }
}



unsigned
shttp_hash_oat(void *key, size_t key_size)
{
    uint32_t hash = 0;

    for(size_t i=0; i < key_size; i++) {
        hash += key[i];
        hash += (hash << 10);
        hash ^= (hash >> 6);
    }

    hash += (hash << 3);
    hash ^= (hash >> 11);
    hash += (hash << 15);

    return (unsigned) hash;
}

bool
shttp_hash_default_key_comparator(void *key1, size_t key1_size, void *key2, 
    size_t key2_size)
{
    return key1_size == key2_size && memcmp(key1, key2, key1_size) == 0;
}
