/*
Copyright (c) 2011, Christopher Utz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of SimpleHttp nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTOPHER UTZ BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Author: Christopher Utz <utz.m.christopher@gmail.com>
*/

#include <shttp_http_server.h>

#include <sys/socket.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <shttp_mem.h>
#include <shttp_http_server_connection.h>

#define SHTTP_DEFAULT_HTTP_SERVER_IP_ADDR "127.0.0.1"
#define SHTTP_DEFAULT_HTTP_SERVER_PORT 8080

static void
shttp_server_socket_readable_callback(struct ev_loop *loop, ev_io *w, 
    int revents)
{
    shttp_http_server_t *server = (shttp_http_server_t *) w->data;

    shttp_client_socket_t client_socket;
    shttp_error_t ret;

    ret = shttp_server_socket_accept(&server->server_socket, &client_socket);

    if (ret == SHTTP_NOTHING)
        return;
    else if (ret != SHTTP_OK) {
        shttp_log(SHTTP_ERR, "failed accepting http request");
        return;
    }

    shttp_log(SHTTP_DEBUG,
              "accepted connection");

    shttp_http_server_connection_t *conn = shttp_new(
        shttp_http_server_connection_t);

    if (!conn) {
        shttp_log(SHTTP_ERR,
            "failed allocating memory for http server connection");
        return;
    }

    shttp_http_server_connection_init(conn, &client_socket);
}

shttp_error_t
shttp_http_server_init(shttp_http_server_t *server, 
    shttp_http_server_config_t *config)
{
    assert(server != NULL);
    assert(config != NULL);

    shttp_error_t ret;

    const char *ip_addr = config->ip_addr ? config->ip_addr :
        SHTTP_DEFAULT_HTTP_SERVER_IP_ADDR;

    unsigned short port = config->port > 0 ? config->port :
        SHTTP_DEFAULT_HTTP_SERVER_PORT;

    int backlog = config->backlog > 0 ? config->backlog : SOMAXCONN;

    ret = shttp_server_socket_init(&server->server_socket, 
                                   ip_addr, 
                                   port, 
                                   backlog);

    if (ret != SHTTP_OK)
        goto server_socket_init_failed;

    return SHTTP_OK;

server_socket_init_failed:
    return ret;
}

void
shttp_http_server_cleanup(shttp_http_server_t *server)
{
    if (server) {
        shttp_server_socket_cleanup(&server->server_socket);
    }
}

shttp_error_t
shttp_http_server_start(shttp_http_server_t *server, struct ev_loop *loop)
{
    assert(server != NULL);
    assert(loop != NULL); 

    shttp_error_t ret;

    ret = shttp_server_socket_listen(&server->server_socket);

    if (ret != SHTTP_OK)
        goto server_socket_listen_failed;

    ev_io_init(&server->server_socket_readable,
               shttp_server_socket_readable_callback,
               server->server_socket.fd,
               EV_READ);

    server->server_socket_readable.data = server;

    ev_io_start(loop, &server->server_socket_readable);

    return SHTTP_OK;

server_socket_listen_failed:
    return ret;
}

shttp_error_t
shttp_http_server_run(shttp_http_server_config_t *config)
{
    assert(config != NULL);

    shttp_error_t ret;
    shttp_http_server_t server;

    ret = shttp_http_server_init(&server, config);

    if (ret != SHTTP_OK)
        goto http_server_init_failed;

    struct ev_loop *loop = ev_default_loop(0);

    if (!loop) {
        shttp_log(SHTTP_ERR, "ev_default_loop() failed (failed"
            " initializing libev)");
        ret = SHTTP_RUNTIME_ERROR;
        goto ev_default_loop_failed;
    }

    ret = shttp_http_server_start(&server, loop);

    if (ret != SHTTP_OK)
        goto http_server_start_failed;

    ev_run(loop, 0);

    shttp_http_server_cleanup(&server);

    return SHTTP_OK;

http_server_start_failed:
ev_default_loop_failed:
    shttp_http_server_cleanup(&server);
http_server_init_failed:
    return ret;
}
