/*
Copyright (c) 2011, Christopher Utz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of SimpleHttp nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTOPHER UTZ BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Author: Christopher Utz <utz.m.christopher@gmail.com>
*/

#include <shttp_request_parser.h>

#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>

#include <shttp_mem.h>
#include <shttp_util.h>
#include <shttp_tokenize.h>

static inline size_t
shttp_find_crlf(const char *str, size_t length)
{
    for (size_t i=0; i < length-1; i++) {
        if (str[i] == '\r' && str[i+1] == '\n') {
            return i;
        }
    }

    return length;
}

static inline size_t
shttp_find_colon(const char *str, size_t length)
{
    for (size_t i=0; i < length-1; i++) {
        if (str[i] == ':') {
            return i;
        }
    }

    return length;
}

static inline void
shttp_skip_leading_cr_and_lf(char **out_str, size_t *out_length, 
    char *str, size_t length)
{
    while (length > 0) {
        if (shttp_is_cr_or_lf(*str)) {
            str++;
            length--;
        } else {
            break;
        }
    }

    *out_str = str;
    *out_length = length;
}

static size_t
shttp_request_parser_parse_request_line(shttp_request_parser_t *parser, 
    const char *data, size_t length)
{    
    // TODO: enforce a maximum request line length.

    shttp_error_t res;

    size_t crlf_idx = shttp_find_crlf(data, length);
    
    if (crlf_idx < length) {
        shttp_buffer_append(&parser->staging_buffer, data, crlf_idx);

        char *request_line;
        size_t request_line_length;

        // Skip leading CR/LF characters, as per 
        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4
        shttp_skip_leading_cr_and_lf(&request_line, &request_line_length, 
            parser->staging_buffer.data, parser->staging_buffer.length);

        shttp_token_t tokens[4];

        int num_tokens = shttp_tokenize(request_line,
            request_line_length, " ", tokens, 4);

        if (num_tokens != 3) {
            parser->error = SHTTP_MALFORMED_REQUEST;
            goto parse_error;
        }

        // NULL terminate token 0 and 1
        request_line[(tokens[0].chars-request_line)+tokens[0].length] = '\0';
        request_line[(tokens[1].chars-request_line)+tokens[1].length] = '\0';

        parser->method = shttp_method_from_str(tokens[0].chars);

        if (parser->method == SHTTP_UNKNOWN_METHOD)
            parser->error = SHTTP_INVALID_REQUEST;

        res = shttp_uri_parse(&parser->uri, tokens[1].chars);

        if (res == SHTTP_MEM_ERROR) {
            parser->error = SHTTP_MEM_ERROR;
            goto parse_error;
        }
        else if (res == SHTTP_MALFORMED_URI)
            parser->error = SHTTP_INVALID_REQUEST;

        if (tokens[2].length != SHTTP_STR_LEN("HTTP/1.1"))
            parser->error = SHTTP_INVALID_REQUEST;
        else if (strncmp(tokens[2].chars, SHTTP_STR_AND_LEN("HTTP/1.1")) == 0)
            parser->version = SHTTP_1_1;
        else if (strncmp(tokens[2].chars, SHTTP_STR_AND_LEN("HTTP/1.0")) == 0)
            parser->version = SHTTP_1_0;
        else {
            parser->error = SHTTP_INVALID_REQUEST;
            parser->version = SHTTP_1_1;
        }

        shttp_buffer_clear(&parser->staging_buffer); 

        parser->state = SHTTP_PARSE_HEADERS;
        return crlf_idx + 2;

parse_error:
        shttp_buffer_clear(&parser->staging_buffer); 
        parser->state = SHTTP_PARSE_DONE;
        return crlf_idx + 2;
    } else {
        shttp_buffer_append(&parser->staging_buffer, data, length);
        return length;
    }
}

static bool
shttp_request_parser_interpret_headers(shttp_request_parser_t *parser)
{

    if (shttp_has_body(parser->method)) {
        // 1) If transfer-encoding != identity, chunked is last 
        //    transfer-encoding. Proceed with decoding chunked transfer 
        //    encoding.
        // 2) Use content-length.
        // 3) error = SHTTP_MALFORMED_REQUEST

        // TODO: add support for chunked transfer encoding? Nginx doesn't
        // support chunked requests out of the box.

        shttp_table_entry_t *transfer_encoding_hdr = shttp_table_find(
            &parser->headers, "transfer-encoding");

        if (transfer_encoding_hdr &&
            strcmp(transfer_encoding_hdr->value, "identity") != 0) {
            parser->state = SHTTP_PARSE_DONE;
            parser->error = SHTTP_UNKNOWN_TRANSFER_ENCODING;
            return false;
        }

        shttp_table_entry_t *content_length_hdr = shttp_table_find(
            &parser->headers, "content-length");

        if (!content_length_hdr) {
            parser->state = SHTTP_PARSE_DONE;
            parser->error = SHTTP_INVALID_REQUEST;
            return false;
        }

        const char *curr = content_length_hdr->value;
        size_t content_length = 0;

        // TODO: what if it overflows?
        while (shttp_is_digit(*curr)) {
            content_length = content_length * 10 + shttp_digit(*curr);
            curr++;
        }

        if (*curr != '\0') {
            parser->state = SHTTP_PARSE_DONE;
            parser->error = SHTTP_INVALID_REQUEST;
            return false;
        }

        parser->content_length = content_length;
        parser->content_remaining = content_length;
    }

    return true;
}

static void
shttp_request_parser_parse_header(shttp_request_parser_t *parser)
{
    assert(parser->header_buffer.length > 0);

    char *header = parser->header_buffer.data;
    size_t header_length = parser->header_buffer.length;

    size_t colon_idx = shttp_find_colon(header, header_length);

    if (colon_idx == header_length) {
        parser->error = SHTTP_INVALID_REQUEST;
        return;
    }

    size_t name_length = colon_idx;

    if (name_length == 0) {
        parser->error = SHTTP_INVALID_REQUEST;
        return;
    }

    size_t value_length = header_length - colon_idx - 1;
    char *name = shttp_malloc(name_length + 1);
    char *final_value = shttp_malloc(value_length + 1);

    if (!name || !final_value) {
        parser->error = SHTTP_MEM_ERROR;
        goto error;
    }

    size_t j=0;

    for (size_t i=colon_idx + 1; i < header_length; i++) {
        if (shttp_is_lws(header[i])) {
            if (j > 0 && final_value[j-1] != ' ') {
                final_value[j++] = ' ';
            }
        } else if (shttp_is_control(header[i])) {
            parser->error = SHTTP_INVALID_REQUEST;
            goto error;
        } else {
            final_value[j++] = header[i];
        }
    }

    final_value[j] = '\0';

    while (j > 0 && shttp_is_lws(final_value[j-1])) {
        final_value[--j] = '\0';
    }

    for (size_t i=0; i < name_length; i++) {
        if (!shttp_is_token(header[i])) {
            parser->error = SHTTP_INVALID_REQUEST;
            goto error;
        } else {
            if (shttp_is_upper(header[i])) {
                name[i] = shttp_lower(header[i]);
            } else {
                name[i] = header[i];
            }
        }
    }

    name[name_length] = '\0';

    if (shttp_table_append(&parser->headers, name, 
        final_value, ',') != SHTTP_OK) {
        parser->error = SHTTP_MEM_ERROR;
        goto error;
    }

    return;

error:
    free(name);
    free(final_value);
}

static size_t
shttp_request_parser_parse_headers(shttp_request_parser_t *parser,
    const char *data, size_t length)
{
    // TODO: enforce a maximum header length.

    size_t crlf_idx = shttp_find_crlf(data, length);
    
    if (crlf_idx < length) {
        shttp_buffer_append(&parser->staging_buffer, data, crlf_idx);

        if (parser->staging_buffer.length > 0) {
            if (shttp_is_continuation(parser->staging_buffer.data[0])) {
                shttp_buffer_append(&parser->header_buffer, 
                    parser->staging_buffer.data, parser->staging_buffer.length);
            } else {
                if (parser->header_buffer.length > 0)
                    shttp_request_parser_parse_header(parser);

                shttp_buffer_clear(&parser->header_buffer);

                shttp_buffer_append(&parser->header_buffer, 
                    parser->staging_buffer.data, parser->staging_buffer.length);
            }
        } else {
            // All headers fed in. Move on to parsing the body if the method
            // has one.

            if (parser->header_buffer.length > 0)
                shttp_request_parser_parse_header(parser);

            if (!shttp_request_parser_interpret_headers(parser)) {
                return crlf_idx + 2;
            }

            if (shttp_has_body(parser->method))
                parser->state = SHTTP_PARSE_BODY;
            else
                parser->state = SHTTP_PARSE_DONE;
        }

        shttp_buffer_clear(&parser->staging_buffer);

        return crlf_idx + 2;
    } else {
        shttp_buffer_append(&parser->staging_buffer, data, length);
        return length;
    }
}

static size_t
shttp_request_parser_parse_body(shttp_request_parser_t *parser,
    const char *data, size_t length)
{
    size_t amt_to_consume = SHTTP_MIN(length, parser->content_remaining);

    for (int i=0; i < parser->body_callbacks.length; i++) {
        shttp_request_parser_body_callback_t callback = 
            *shttp_array_at(&parser->body_callbacks, 
                            i, 
                            shttp_request_parser_body_callback_t);

        callback(data, amt_to_consume);
    }

    parser->content_remaining -= amt_to_consume;

    if (parser->content_remaining == 0) {
        parser->state = SHTTP_PARSE_DONE;
    }

    return amt_to_consume;
}

shttp_error_t
shttp_request_parser_init(shttp_request_parser_t *parser)
{
    parser->state = SHTTP_PARSE_REQUEST_LINE;
    parser->error = SHTTP_OK;

    shttp_error_t res = shttp_buffer_init(&parser->staging_buffer, 
        SHTTP_BUFFER_RECOMMENDED_ALLOCATE);

    if (res != SHTTP_OK)
        goto buffer_init_failed;

    res = shttp_uri_init(&parser->uri);

    if (res != SHTTP_OK)
        goto uri_init_failed;

    res = shttp_table_init(&parser->headers, SHTTP_TABLE_RECOMMENDED_ALLOCATE);

    if (res != SHTTP_OK)
        goto table_init_failed;

    res = shttp_buffer_init(&parser->header_buffer, 
        SHTTP_BUFFER_RECOMMENDED_ALLOCATE);

    if (res != SHTTP_OK)
        goto buffer_init_2_failed;

    res = shttp_array_init(&parser->body_callbacks,
                           sizeof(shttp_request_parser_body_callback_t),
                           SHTTP_ARRAY_RECOMMENDED_ALLOCATE);

    if (res != SHTTP_OK)
       goto array_init_failed; 

    return SHTTP_OK;

array_init_failed:
    shttp_buffer_cleanup(&parser->header_buffer);
buffer_init_2_failed:
    shttp_table_cleanup(&parser->headers);
table_init_failed:
    shttp_uri_cleanup(&parser->uri);
uri_init_failed:
    shttp_buffer_cleanup(&parser->staging_buffer);
buffer_init_failed:
    return SHTTP_MEM_ERROR;
}

void
shttp_request_parser_cleanup(shttp_request_parser_t *parser)
{
    if (parser) {
        shttp_buffer_cleanup(&parser->staging_buffer);
        shttp_uri_cleanup(&parser->uri);
        shttp_buffer_cleanup(&parser->header_buffer);
        shttp_table_cleanup(&parser->headers);
    }
}

shttp_error_t
shttp_request_parser_add_body_callback(shttp_request_parser_t *parser,
    shttp_request_parser_body_callback_t callback)
{
    assert(parser != NULL);
    assert(callback != NULL);

    return shttp_array_append(&parser->body_callbacks, &callback);
}

size_t
shttp_request_parser_feed(shttp_request_parser_t *parser, 
    const char *data, size_t length)
{
    assert(parser != NULL);
    assert(data != NULL);

    size_t pos = 0;
    
    while (pos < length) {
        const char *remaining_data = data + pos;
        size_t remaining_length = length - pos;

        switch(parser->state) {
            case SHTTP_PARSE_REQUEST_LINE:
                pos += shttp_request_parser_parse_request_line(parser, 
                    remaining_data, remaining_length);
                break;
            case SHTTP_PARSE_HEADERS:
                pos += shttp_request_parser_parse_headers(parser,
                    remaining_data, remaining_length);
                break;
            case SHTTP_PARSE_BODY:
                pos += shttp_request_parser_parse_body(parser,
                    remaining_data, remaining_length);
            default:
                goto done;
        }

        assert(pos <= length);
    }

done:
    return pos;
}
