/*
Copyright (c) 2011, Christopher Utz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of SimpleHttp nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTOPHER UTZ BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Author: Christopher Utz <utz.m.christopher@gmail.com>
*/

#include <shttp_server_socket.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <shttp_mem.h>

static shttp_error_t
shttp_set_nonblocking(int fd)
{
    int flags = fcntl(fd, F_GETFL);

    if (flags < 0) {
        shttp_log_errno(SHTTP_ERR, errno, "fcntl() failed");
        return SHTTP_RUNTIME_ERROR;
    }

    flags |= O_NONBLOCK;

    if (fcntl(fd, F_SETFL, flags) == -1) {
        shttp_log_errno(SHTTP_ERR, errno, "fcntl() failed");
        return SHTTP_RUNTIME_ERROR;
    }

    return SHTTP_OK;
}

shttp_error_t
shttp_server_socket_init(shttp_server_socket_t *socket, const char *ip_addr, 
    unsigned short port, int backlog)
{
    assert(socket != NULL);
    assert(ip_addr != NULL);
    assert(port >= 1);

    socket->port = port;
    socket->fd = -1;
    socket->backlog = backlog;

    socket->ip_addr = shttp_strdup(ip_addr);

    if (!socket->ip_addr)
        goto strdup_failed;

    return SHTTP_OK;

strdup_failed:
    return SHTTP_MEM_ERROR;
}

void
shttp_server_socket_cleanup(shttp_server_socket_t *socket)
{
    if (socket) {
        if (socket->fd >= 0)
            close(socket->fd);

        shttp_free(socket->ip_addr);

        socket->ip_addr = NULL;
        socket->fd = -1;
    }
}

shttp_error_t
shttp_server_socket_listen(shttp_server_socket_t *sock)
{
    assert(sock != NULL);
    assert(sock->fd < 0);    
    
    struct addrinfo *addr_info_res = NULL, hints;
    shttp_error_t ret;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE;

    char port_str[32];
    snprintf(port_str, sizeof(port_str), "%d", sock->port);

    int ecode = getaddrinfo(sock->ip_addr,
        port_str,
        &hints,
        &addr_info_res);

    if (ecode != 0) {
        const char *gai_str = gai_strerror(ecode);
        shttp_log(SHTTP_ERR, "getaddrinfo() failed: %s", gai_str);
        ret = SHTTP_RUNTIME_ERROR;
        goto getaddrinfo_failed;
    }

    sock->fd = socket(addr_info_res->ai_family, 
        addr_info_res->ai_socktype, 
        addr_info_res->ai_protocol);

    if (sock->fd < 0) {
        shttp_log_errno(SHTTP_ERR, errno, "listen() failed");
        ret = SHTTP_RUNTIME_ERROR;
        goto socket_failed;
    }   

    int yes = 1;

    if (setsockopt(sock->fd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) < 0) {
        shttp_log_errno(SHTTP_ERR, errno, "setsockopt() failed");
        ret = SHTTP_RUNTIME_ERROR;
        goto setsockopt_failed;
    }   

    if (bind(sock->fd, addr_info_res->ai_addr, addr_info_res->ai_addrlen) < 0) {
        shttp_log_errno(SHTTP_ERR, errno, "bind() failed");
        ret = SHTTP_RUNTIME_ERROR;
        goto bind_failed;
    }

    if (listen(sock->fd, sock->backlog) < 0) {
        shttp_log_errno(SHTTP_ERR, errno, "listen() failed");
        ret = SHTTP_RUNTIME_ERROR;
        goto listen_failed;
    }   

    ret = shttp_set_nonblocking(sock->fd);

    if (ret != SHTTP_OK)
        goto set_nonblocking_failed;

    return SHTTP_OK;

set_nonblocking_failed:
listen_failed:
bind_failed:
setsockopt_failed:
socket_failed:
    freeaddrinfo(addr_info_res);
getaddrinfo_failed:
    return ret;
}

shttp_error_t
shttp_server_socket_accept(shttp_server_socket_t *socket, 
    shttp_client_socket_t *client_socket)
{
    assert(socket != NULL);
    assert(client_socket != NULL);

    int client_fd;
    shttp_error_t ret;
    
    do {
        client_fd = accept(socket->fd, NULL, NULL);
    } while(client_fd < 0 && (errno == EINTR || errno == ECONNABORTED));

    if (client_fd < 0 && (errno == EAGAIN || errno == EWOULDBLOCK)) {
        return SHTTP_NOTHING;
    }
    
    if (client_fd < 0) {
        ret = SHTTP_NETWORK_ERROR;
        shttp_log_errno(SHTTP_ERR, errno, "accept() failed");
        goto accept_failed;
    }

    ret = shttp_client_socket_init(client_socket, client_fd);

    if (ret != SHTTP_OK) {
        shttp_log(SHTTP_ERR, "failed initializing client socket");
        goto client_socket_init_failed;
    }
    
    return SHTTP_OK;
    
client_socket_init_failed:
    close(client_fd);
accept_failed:
    return ret;
}
