/*
Copyright (c) 2011, Christopher Utz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of SimpleHttp nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTOPHER UTZ BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Author: Christopher Utz <utz.m.christopher@gmail.com>
*/

#include <shttp_table.h>

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <shttp_mem.h>

static void
shttp_table_free_entries(shttp_table_t *table)
{
    for (int i=0; i < table->entries.length; i++) {
        shttp_table_entry_t *entry = 
            shttp_array_at(&table->entries, i, shttp_table_entry_t); 

        shttp_free(entry->name);
        shttp_free(entry->value);
    }
}

shttp_error_t
shttp_table_init(shttp_table_t *table, int allocate)
{
    assert(table != NULL);
    assert(allocate >= 0);

    shttp_error_t res = shttp_array_init(&table->entries, 
        sizeof(shttp_table_entry_t), allocate);

    if (res != SHTTP_OK)
        return SHTTP_MEM_ERROR;

    return SHTTP_OK;
}

void
shttp_table_cleanup(shttp_table_t *table)
{
    if (table) {
        shttp_table_free_entries(table);
        shttp_array_cleanup(&table->entries);
    }
}

shttp_table_entry_t *
shttp_table_find(shttp_table_t *table, char *name)
{
    assert(table != NULL);
    assert(name != NULL);

    for (int i=0; i < table->entries.length; i++) {
        shttp_table_entry_t *entry = 
            shttp_array_at(&table->entries, i, shttp_table_entry_t); 

        if (strcmp(entry->name, name) == 0) {
            return entry;
        }
    }

    return NULL;
}

shttp_error_t
shttp_table_add(shttp_table_t *table, char *name, char *value)
{
    assert(table != NULL);
    assert(name != NULL);
    assert(value != NULL);

    shttp_table_entry_t entry = {
        .name = name,
        .value = value
    };

    return shttp_array_append(&table->entries, &entry);
}

shttp_error_t
shttp_table_append(shttp_table_t *table, char *name, char *value, char sep)
{
    assert(table != NULL);
    assert(name != NULL);
    assert(value != NULL);

    shttp_table_entry_t *existing = shttp_table_find(table, name);

    if (existing) {
        int existing_len = strlen(existing->value);
        int new_len = existing_len + strlen(value) + 2;

        char *new_value = shttp_realloc(existing->value, new_len);

        if (!new_value)
            return SHTTP_MEM_ERROR;

        new_value[existing_len] = sep;
        strcpy(&new_value[existing_len+1], value);
        new_value[new_len] = '\0';

        existing->value = new_value;

        return SHTTP_OK;
    } else {
        return shttp_table_add(table, name, value);
    }
}
