/*
Copyright (c) 2011, Christopher Utz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of SimpleHttp nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTOPHER UTZ BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Author: Christopher Utz <utz.m.christopher@gmail.com>
*/

#include <shttp_uri.h>

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <shttp_util.h>
#include <shttp_mem.h>

static shttp_error_t
shttp_uri_parse_query(shttp_uri_t *uri, const char *head)
{
    uri->query = shttp_strdup(head+1);

    return SHTTP_OK;
}

static shttp_error_t
shttp_uri_parse_path(shttp_uri_t *uri, const char *head)
{
    const char *sep = strpbrk(head, "?");

    if (!sep) {
        uri->path = shttp_strdup(head);

        if (!uri->path)
            return SHTTP_MEM_ERROR;

        return SHTTP_OK;
    } else {
        uri->path = shttp_strndup(head, sep - head);

        if (!uri->path)
            return SHTTP_MEM_ERROR;

        return shttp_uri_parse_query(uri, sep);
    }
}

static shttp_error_t
shttp_uri_parse_port(shttp_uri_t *uri, const char *head)
{
    const char *curr = head+1;

    while (shttp_is_digit(*curr)) {
        uri->port = uri->port * 10 + shttp_digit(*curr);
        curr++;
    }

    switch (*curr) {
        case '/':
            return shttp_uri_parse_path(uri, curr);
        case '?':
            return shttp_uri_parse_query(uri, curr);
        case '\0':
            return SHTTP_OK;
        default:
            return SHTTP_MALFORMED_URI;
    }
}

static shttp_error_t
shttp_uri_parse_host(shttp_uri_t *uri, const char *head)
{
    const char *sep = strpbrk(head, ":/?");

    if (!sep) {
        uri->host = shttp_strdup(head);

        if (!uri->host)
            return SHTTP_MEM_ERROR;
    } else {
        uri->host = shttp_strndup(head, sep - head);

        if (!uri->host)
            return SHTTP_MEM_ERROR;

        switch (*sep) {
            case ':':
                return shttp_uri_parse_port(uri, sep);
            case '/':
                return shttp_uri_parse_path(uri, sep);
            case '?':
                return shttp_uri_parse_query(uri, sep);
            default:
                assert(0);
                return SHTTP_RUNTIME_ERROR;
        }
    }

    return SHTTP_OK;
}

static shttp_error_t
shttp_uri_parse_abs_uri(shttp_uri_t *uri, const char *uri_str)
{
    uri->type = SHTTP_ABS_URI;
    uri->scheme = SHTTP_HTTP;

    shttp_error_t ret = shttp_uri_parse_host(uri, 
        uri_str + SHTTP_STR_LEN("http://"));

    if (ret != SHTTP_OK)
        return ret;

    if (*uri->host == '\0')
        return SHTTP_MALFORMED_URI;

    if (uri->port == 0)
        uri->port = 80;
    else if (uri->port < 0 || uri->port >= (1<<15))
        return SHTTP_MALFORMED_URI;

    if (!uri->path) {
        uri->path = shttp_strdup("/");

        if (!uri->path)
            return SHTTP_MEM_ERROR;
    }

    return SHTTP_OK;
}

static shttp_error_t
shttp_uri_parse_abs_path(shttp_uri_t *uri, const char *uri_str)
{
    uri->type = SHTTP_ABS_PATH;

    return shttp_uri_parse_path(uri, uri_str);
}

shttp_error_t
shttp_uri_init(shttp_uri_t *uri)
{
    assert(uri != NULL);

    memset(uri, 0, sizeof(shttp_uri_t));

    return SHTTP_OK;
}

shttp_error_t
shttp_uri_parse(shttp_uri_t *uri, const char *uri_str)
{
    assert(uri != NULL);
    assert(uri_str != NULL);

    if (strncmp(uri_str, SHTTP_STR_AND_LEN("http://")) == 0)
        return shttp_uri_parse_abs_uri(uri, uri_str);
    else if (strncmp(uri_str, SHTTP_STR_AND_LEN("/")) == 0)
        return shttp_uri_parse_abs_path(uri, uri_str);
    else
        return SHTTP_MALFORMED_URI;
}

void
shttp_uri_cleanup(shttp_uri_t *uri)
{
    if (uri) {
        free(uri->host);
        free(uri->path);
        free(uri->query);

        memset(uri, 0, sizeof(shttp_uri_t));
    }
}
