#include <strings.h>

#include "std/debug.h"
#include "std/memory.h"
#include "std/ptrarray.h"

static void DeletePtrArray(PtrArrayT *self) {
  if (self->managed)
    PtrArrayForEach(self, (IterFuncT)MemUnref, NULL);
  MemUnref(self->data);
}

TYPEDECL(PtrArrayT, (FreeFuncT)DeletePtrArray);

PtrArrayT *NewPtrArray(size_t reserved, bool managed) {
  PtrArrayT *array = NewInstance(PtrArrayT);

  array->managed = managed;
  array->reserved = 0;

  PtrArrayResize(array, reserved);

  return array;
}

/*
 * Helper functions.
 */

static void MaybeGrow(PtrArrayT *self, size_t count);
static void MaybeShrink(PtrArrayT *self, size_t count);

static inline void ClearRange(PtrArrayT *self, size_t begin, size_t end) {
  bzero(&self->data[begin], (end - begin + 1) * sizeof(PtrT));
}

static inline void FreeRange(PtrArrayT *self, size_t begin, size_t end) {
  if (self->managed)
    PtrArrayForEachInRange(self, begin, end, (IterFuncT)MemUnref, NULL);
}

static inline void MoveRange(PtrArrayT *self,
                             size_t to, size_t first, size_t last)
{
  size_t remainding = last - first + 1;

  memmove(&self->data[to],
          &self->data[first],
          remainding * sizeof(PtrT));
}

static void RemoveRange(PtrArrayT *self asm("a0"),
                        size_t first asm("d0"), size_t last asm("d1"))
{
  size_t count = last - first + 1;

  FreeRange(self, first, last);
  MoveRange(self, first, last + 1, self->size - 1);
  ClearRange(self, self->size - count, self->size - 1);
  MaybeShrink(self, count);
}

static inline void RemoveFast(PtrArrayT *self, PtrT *item) {
  PtrT *last = &self->data[self->size - 1];

  if (self->managed)
      MemUnref(*item);

  if (item != last)
    *item = *last;

  *last = NULL;

  self->size--;
}

/*
 * @brief Check if index is valid and normalise it.
 *
 * Accepts negative indices which point to elements starting from the end of
 * array (eg. -1 is the last element).  Terminates program if index is invalid.
 */

static size_t CheckIndex(PtrArrayT *self asm("a0"), ssize_t index asm("d0")) {
  ASSERT((index < self->size) && (index <= -self->size),
         "Index %d out of bound.", index);

  return (index < 0) ? (self->size - index) : index;
}

/*
 * Iteration functions.
 */

void PtrArrayForEach(PtrArrayT *self, IterFuncT func, PtrT data) {
  if (self->size)
    PtrArrayForEachInRange(self, 0, self->size - 1, func, data);
}

void PtrArrayForEachInRange(PtrArrayT *self, ssize_t begin, ssize_t end,
                            IterFuncT func, PtrT data)
{
  begin = CheckIndex(self, begin);
  end = CheckIndex(self, end);

  ASSERT(begin <= end, "Invalid range of elements specified [%d..%d]!",
         begin, end);

  {
    PtrT *item = &self->data[begin];
    PtrT *last = &self->data[end];

    do { func(*item++, data); } while (item <= last);
  }
}

/*
 * Element adding functions.
 */

PtrT *PtrArrayInsertFast(PtrArrayT *self, ssize_t index, PtrT data) {
  MaybeGrow(self, 1);

  {
    PtrT *item = &self->data[0];
    PtrT *last = &self->data[self->size - 1];

    *last = *item;
    *item = data;

    return item;
  }
}

PtrT *PtrArrayInsert(PtrArrayT *self, ssize_t index, PtrT data) {
  return PtrArrayInsertElements(self, index, data ? &data : NULL, 1);
}

PtrT *PtrArrayInsertElements(PtrArrayT *self, ssize_t index,
                             PtrT *data, size_t count)
{
  PtrT *item;

  index = CheckIndex(self, index);
  MaybeGrow(self, count);
  MoveRange(self, index + count, index, self->size - count - 1);

  item = &self->data[index];

  if (data)
    memcpy(item, data, count * sizeof(PtrT));

  return item;
}

PtrT *PtrArrayAppend(PtrArrayT *self, PtrT data) {
  return PtrArrayAppendElements(self, data ? &data : NULL, 1);
}

PtrT *PtrArrayAppendElements(PtrArrayT *self, PtrT *data, size_t count) {
  PtrT *item;

  MaybeGrow(self, count);

  item = &self->data[self->size - count];

  if (data)
    memcpy(item, data, count * sizeof(PtrT));

  return item;
}

/*
 * Element removal functions.
 */

void PtrArrayRemoveFast(PtrArrayT *self, size_t index) {
  PtrT *item = &self->data[index];

  RemoveFast(self, item);
}

void PtrArrayRemove(PtrArrayT *self, ssize_t index) {
  PtrArrayRemoveRange(self, index, index);
}

void PtrArrayRemoveRange(PtrArrayT *self, ssize_t first, ssize_t last) {
  first = CheckIndex(self, first);
  last = CheckIndex(self, last);

  ASSERT(first <= last, "Invalid range of elements specified [%d..%d]!",
         first, last);

  RemoveRange(self, first, last);
}

void PtrArrayFilterFast(PtrArrayT *self, PredicateT func) {
  PtrT *item = self->data;
  size_t index = 0;

  while (index < self->size) {
    if (func(*item)) {
      item++;
      index++;
    } else {
      RemoveFast(self, item);
    }
  }
}

/*
 * @brief Change maximum number of element the array can hold.
 *
 * 1) newSize > current size => extra space will be allocated.
 * 2) newSize < current size => some elements will be removed and
 *    reserved space will be shrinked.
 *
 * Note: Allocates one more element that will serve the purpose of temporary
 * element for some operations (i.e. sort, swap).
 */
void PtrArrayResize(PtrArrayT *self, size_t newSize) {
  if (self->data) {
    if (newSize == self->size)
      return;

    if (newSize < self->size) {
      FreeRange(self, newSize, self->size - 1);
      self->size = newSize;
    }
  }

  if (!self->data) {
    self->data = MemNewTable(sizeof(PtrT), newSize);
    self->size = 0;
  } else {
    self->data = TableResize(self->data, newSize);
  }

  if (newSize > self->reserved)
    ClearRange(self, self->reserved, newSize);

  self->reserved = newSize;
}

#define MIN_SIZE 16

static inline size_t NearestPow2(size_t num) {
  size_t i = 1;

  while (num > i)
    i += i;

  return max(i, MIN_SIZE);
}

/*
 * @brief Check if array can accomodate extra elements and resize if needed.
 */
static void MaybeGrow(PtrArrayT *self, size_t count) {
  if (self->size + count <= self->reserved) {
    self->size += count;
  } else {
    PtrArrayResize(self, NearestPow2(self->size));
  }
}

/*
 * @brief Check if array doesn't have too much of an extra space and shrink.
 */
static void MaybeShrink(PtrArrayT *self, size_t count) {
  self->size -= count;
}

/*
 * Sorting functions.
 */

void PtrArrayInsertionSort(PtrArrayT *self, ssize_t begin, ssize_t end) {
  CompareFuncT cmp = self->compareFunc;

  begin = CheckIndex(self, begin);
  end = CheckIndex(self, end);

  ASSERT(cmp, "Compare function not set!");
  ASSERT(begin < end, "Invalid range of elements specified [%d..%d]!",
         begin, end);

  {
    PtrT *data = self->data;
    size_t toInsert = begin + 1;

    while (toInsert <= end) {
      size_t index = begin;

      while ((index < toInsert) &&
             (cmp(data[index], data[toInsert]) <= 0))
        index++;

      if (index < toInsert) {
        PtrT tmp = data[toInsert];
        MoveRange(self, index + 1, index, toInsert - 1);
        data[index] = tmp;
      }

      toInsert++;
    }
  }
}

size_t PtrArrayPartition(PtrArrayT *self,
                         size_t left, size_t right, PtrT pivot)
{
  CompareFuncT cmp = self->compareFunc;

  ASSERT(cmp, "Compare function not set!");
  ASSERT(left < right, "Invalid range of elements specified [%d..%d]!",
         (int)left, (int)right);

  {
    PtrT *data = self->data;
    size_t partition = left;

    while (left < right) {
      while ((cmp(data[left], pivot) < 0) && (left < right)) {
        left++;
        partition++;
      }

      while ((cmp(pivot, data[right]) < 0) && (left < right))
        right--;

      {
        PtrT tmp;
      
        tmp = data[left];
        data[left] = data[right];
        data[right] = tmp;
      }

      left++;
      right--;
      partition++;
    }

    return partition;
  }
}

static void QuickSort(PtrArrayT *self asm("a0"),
                      size_t left asm("d0"), size_t right asm("d1"))
{
  while (left < right) {
    PtrT pivot = PtrArrayGet(self, (left + right) / 2);
    size_t i = PtrArrayPartition(self, left, right, pivot);

    if (i - left <= right - i) {
      QuickSort(self, left, i);
      left = i + 1;
    } else {
      QuickSort(self, i + 1, right);
      right = i;
    }
  }
}

void PtrArrayQuickSort(PtrArrayT *self, ssize_t begin, ssize_t end) {
  begin = CheckIndex(self, begin);
  end = CheckIndex(self, end);

  ASSERT(self->compareFunc, "Compare function not set!");
  ASSERT(begin < end, "Invalid range of elements specified [%d..%d]!",
         begin, end);

  QuickSort(self, begin, end);
}
