/* blp-canvas.c
 *
 * Copyright (C) 2014 Christian Hergert <christian@hergert.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include <clutter-gtk/clutter-gtk.h>

#include "blp-canvas.h"

struct _BlpCanvasPrivate
{
   BlpDocument *document;
   GtkWidget   *embed;
};

enum
{
   PROP_0,
   PROP_DOCUMENT,
   LAST_PROP
};

G_DEFINE_TYPE_WITH_PRIVATE (BlpCanvas, blp_canvas, GTK_TYPE_BIN)

static GParamSpec *gParamSpecs [LAST_PROP];

static void
blp_canvas_load (BlpCanvas *canvas)
{
   BlpCanvasPrivate *priv;
   ClutterActor *stage;
   BlpItemView *view;
   BlpItem *item;

   g_return_if_fail (BLP_IS_CANVAS (canvas));

   priv = canvas->priv;

   if (!priv->document) {
      return;
   }

   stage = gtk_clutter_embed_get_stage (GTK_CLUTTER_EMBED (priv->embed));
   g_assert (stage);

   item = blp_document_get_root (priv->document);
   if (!item) {
      return;
   }

   view = blp_item_get_view (item);
   if (!view) {
      return;
   }

   clutter_actor_add_child (CLUTTER_ACTOR (stage), CLUTTER_ACTOR (view));
}

static void
blp_canvas_unload (BlpCanvas *canvas)
{
   g_return_if_fail (BLP_IS_CANVAS (canvas));

   if (!canvas->priv->document) {
      return;
   }
}

GtkWidget *
blp_canvas_new (void)
{
   return g_object_new (BLP_TYPE_CANVAS, NULL);
}

/**
 * blp_canvas_get_document:
 * @canvas: (in): A #BlpCanvas.
 *
 * Fetch the document for the canvas.
 *
 * Returns: (transfer none): A #BlpDocument.
 */
BlpDocument *
blp_canvas_get_document (BlpCanvas *canvas)
{
   g_return_val_if_fail (BLP_IS_CANVAS (canvas), NULL);

   return canvas->priv->document;
}

void
blp_canvas_set_document (BlpCanvas   *canvas,
                         BlpDocument *document)
{
   BlpCanvasPrivate *priv;

   g_return_if_fail (BLP_IS_CANVAS (canvas));

   priv = canvas->priv;

   if (document) {
      document = g_object_ref (document);
   }

   blp_canvas_unload (canvas);
   priv->document = document;
   blp_canvas_load (canvas);

   g_object_notify_by_pspec (G_OBJECT (canvas), gParamSpecs [PROP_DOCUMENT]);
}

static void
blp_canvas_finalize (GObject *object)
{
   G_OBJECT_CLASS (blp_canvas_parent_class)->finalize (object);
}

static void
blp_canvas_get_property (GObject    *object,
                         guint       prop_id,
                         GValue     *value,
                         GParamSpec *pspec)
{
   BlpCanvas *canvas = BLP_CANVAS (object);

   switch (prop_id) {
   case PROP_DOCUMENT:
      g_value_set_object (value, blp_canvas_get_document (canvas));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_canvas_set_property (GObject      *object,
                         guint         prop_id,
                         const GValue *value,
                         GParamSpec   *pspec)
{
   BlpCanvas *canvas = BLP_CANVAS (object);

   switch (prop_id) {
   case PROP_DOCUMENT:
      blp_canvas_set_document (canvas, g_value_get_object (value));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_canvas_class_init (BlpCanvasClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS (klass);
   object_class->finalize = blp_canvas_finalize;
   object_class->get_property = blp_canvas_get_property;
   object_class->set_property = blp_canvas_set_property;

   gParamSpecs [PROP_DOCUMENT] =
      g_param_spec_object ("document",
                           _("Document"),
                           _("The document being viewed or edited."),
                           BLP_TYPE_DOCUMENT,
                           (G_PARAM_READWRITE |
                            G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (object_class, PROP_DOCUMENT,
                                    gParamSpecs [PROP_DOCUMENT]);
}

static void
blp_canvas_init (BlpCanvas *canvas)
{
   BlpCanvasPrivate *priv;
   BlpDocument *doc;
   ClutterColor background_color;
   ClutterActor *actor;

   priv = canvas->priv = blp_canvas_get_instance_private (canvas);

   clutter_color_from_string (&background_color, "#ffffffff");

   priv->embed = g_object_new (GTK_CLUTTER_TYPE_EMBED,
                               "hexpand", TRUE,
                               "vexpand", TRUE,
                               "visible", TRUE,
                               NULL);
   actor = gtk_clutter_embed_get_stage (GTK_CLUTTER_EMBED (priv->embed));
   clutter_actor_set_background_color (CLUTTER_ACTOR (actor), &background_color);
   gtk_container_add (GTK_CONTAINER (canvas), priv->embed);

   doc = blp_document_new ();
   blp_canvas_set_document (canvas, doc);
   g_object_unref (doc);
}
