/* blp-document.c
 *
 * Copyright (C) 2014 Christian Hergert <christian@hergert.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>

#include "blp-document.h"
#include "blp-layer.h"

struct _BlpDocumentPrivate
{
   BlpItem *root;
};

enum
{
   PROP_0,
   PROP_ROOT,
   LAST_PROP
};

G_DEFINE_TYPE_WITH_PRIVATE (BlpDocument, blp_document, GOM_TYPE_RESOURCE)

static GParamSpec *gParamSpecs [LAST_PROP];

BlpDocument *
blp_document_new (void)
{
   return g_object_new (BLP_TYPE_DOCUMENT, NULL);
}

BlpItem *
blp_document_get_root (BlpDocument *document)
{
   g_return_val_if_fail (BLP_IS_DOCUMENT (document), NULL);

   return document->priv->root;
}

void
blp_document_set_root (BlpDocument *document,
                       BlpItem     *root)
{
   BlpDocumentPrivate *priv;

   g_return_if_fail (BLP_IS_DOCUMENT (document));

   priv = document->priv;

   if (root) {
      root = g_object_ref (root);
   }

   g_clear_object (&priv->root);
   priv->root = root;

   if (root) {
      _blp_item_set_document (root, document);
   }

   g_object_notify_by_pspec (G_OBJECT (document), gParamSpecs [PROP_ROOT]);
}

static void
blp_document_finalize (GObject *object)
{
   BlpDocumentPrivate *priv;

   priv = BLP_DOCUMENT (object)->priv;

   g_clear_object (&priv->root);

   G_OBJECT_CLASS (blp_document_parent_class)->finalize (object);
}

static void
blp_document_get_property (GObject    *object,
                           guint       prop_id,
                           GValue     *value,
                           GParamSpec *pspec)
{
   BlpDocument *document = BLP_DOCUMENT (object);

   switch (prop_id) {
   case PROP_ROOT:
      g_value_set_object (value, blp_document_get_root (document));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_document_set_property (GObject      *object,
                           guint         prop_id,
                           const GValue *value,
                           GParamSpec   *pspec)
{
   BlpDocument *document = BLP_DOCUMENT (object);

   switch (prop_id) {
   case PROP_ROOT:
      blp_document_set_root (document, g_value_get_object (value));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_document_class_init (BlpDocumentClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS (klass);
   object_class->finalize = blp_document_finalize;
   object_class->get_property = blp_document_get_property;
   object_class->set_property = blp_document_set_property;

   gParamSpecs [PROP_ROOT] =
      g_param_spec_object ("root",
                           _("Root"),
                           _("The root item of the document."),
                           BLP_TYPE_ITEM,
                           (G_PARAM_READWRITE |
                            G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (object_class, PROP_ROOT,
                                    gParamSpecs [PROP_ROOT]);
}

static void
blp_document_init (BlpDocument *document)
{
   BlpItem *root;

   document->priv = blp_document_get_instance_private (document);

   root = g_object_new (BLP_TYPE_LAYER, NULL);
   blp_document_set_root (document, root);
   g_object_unref (root);
}
