/* blp-item.c
 *
 * Copyright (C) 2014 Christian Hergert <christian@hergert.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>

#include "blp-document.h"
#include "blp-item.h"

struct _BlpItemPrivate
{
   BlpItem     *parent;
   BlpItemView *view;
   BlpDocument *document;
   GSList      *children;
};

enum
{
   PROP_0,
   PROP_VIEW,
   LAST_PROP
};

G_DEFINE_TYPE_WITH_PRIVATE (BlpItem, blp_item, GOM_TYPE_RESOURCE)

static GParamSpec *gParamSpecs [LAST_PROP];

BlpItemView *
blp_item_get_view (BlpItem *item)
{
   g_return_val_if_fail (BLP_IS_ITEM (item), NULL);

   return item->priv->view;
}

void
blp_item_append_child (BlpItem *item,
                       BlpItem *child)
{
   BlpItemPrivate *priv;

   g_return_if_fail (BLP_IS_ITEM (item));
   g_return_if_fail (BLP_IS_ITEM (child));

   priv = item->priv;

   priv->children = g_slist_append (priv->children, g_object_ref (child));

   clutter_actor_add_child (CLUTTER_ACTOR (item->priv->view),
                            CLUTTER_ACTOR (child->priv->view));
}

void
_blp_item_set_document (BlpItem  *item,
                        gpointer  document)
{
   BlpItemPrivate *priv;

   g_return_if_fail (BLP_IS_ITEM (item));
   g_return_if_fail (!document || BLP_IS_DOCUMENT (document));

   priv = item->priv;

   if (priv->document) {
      g_object_remove_weak_pointer (G_OBJECT (priv->document),
                                    (gpointer *)&priv->document);
      priv->document = NULL;
   }

   if (document) {
      priv->document = document;
      g_object_add_weak_pointer (G_OBJECT (document),
                                 (gpointer *)&priv->document);
   }
}

static void
blp_item_finalize (GObject *object)
{
   BlpItemPrivate *priv = BLP_ITEM (object)->priv;

   g_slist_foreach (priv->children, (GFunc)g_object_unref, NULL);
   g_slist_free (priv->children);

   G_OBJECT_CLASS (blp_item_parent_class)->finalize (object);
}

static void
blp_item_get_property (GObject    *object,
                       guint       prop_id,
                       GValue     *value,
                       GParamSpec *pspec)
{
   BlpItem *item = BLP_ITEM (object);

   switch (prop_id) {
   case PROP_VIEW:
      g_value_set_object (value, blp_item_get_view (item));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_item_set_property (GObject      *object,
                       guint         prop_id,
                       const GValue *value,
                       GParamSpec   *pspec)
{
#if 0
   BlpItem *item = BLP_ITEM (object);
#endif

   switch (prop_id) {
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_item_class_init (BlpItemClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS (klass);
   object_class->finalize = blp_item_finalize;
   object_class->get_property = blp_item_get_property;
   object_class->set_property = blp_item_set_property;

   gParamSpecs [PROP_VIEW] =
      g_param_spec_object ("view",
                           _("View"),
                           _("The view for the item."),
                           BLP_TYPE_ITEM_VIEW,
                           (G_PARAM_READABLE |
                            G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (object_class, PROP_VIEW,
                                    gParamSpecs [PROP_VIEW]);
}

static void
blp_item_init (BlpItem *item)
{
   item->priv = blp_item_get_instance_private (item);

   item->priv->view = g_object_new (CLUTTER_TYPE_ACTOR,
                                    "x-expand", TRUE,
                                    "y-expand", TRUE,
                                    NULL);
}
