/* blp-pixbuf-item.c
 *
 * Copyright (C) 2014 Christian Hergert <christian@hergert.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <glib/gi18n.h>

#include "blp-pixbuf-item.h"
#include "blp-util.h"

struct _BlpPixbufItemPrivate
{
   GFile *file;
};

enum
{
   PROP_0,
   PROP_FILE,
   LAST_PROP
};

G_DEFINE_TYPE_WITH_PRIVATE (BlpPixbufItem, blp_pixbuf_item, BLP_TYPE_ITEM)

static GParamSpec *gParamSpecs [LAST_PROP];

BlpItem *
blp_pixbuf_item_new (GFile *file)
{
   return g_object_new (BLP_TYPE_PIXBUF_ITEM,
                        "file", file,
                        NULL);
}

GFile *
blp_pixbuf_item_get_file (BlpPixbufItem *item)
{
   g_return_val_if_fail (BLP_IS_PIXBUF_ITEM (item), NULL);

   return item->priv->file;
}

void
blp_pixbuf_item_set_file (BlpPixbufItem *item,
                          GFile         *file)
{
   ClutterContent *content;
   GInputStream *stream;
   BlpItemView *view;
   GdkPixbuf *pixbuf;

   g_return_if_fail (BLP_IS_PIXBUF_ITEM (item));
   g_return_if_fail (!file || G_IS_FILE (file));

   if (file) {
      file = g_object_ref (file);
   }

   g_clear_object (&item->priv->file);
   item->priv->file = file;

   stream = (GInputStream *)g_file_read (file, NULL, NULL);
   if (stream) {
      pixbuf = gdk_pixbuf_new_from_stream (stream, NULL, NULL);
      if (pixbuf) {
         view = blp_item_get_view (BLP_ITEM (item));
         content = blp_clutter_image_new_from_pixbuf (pixbuf);
         clutter_actor_set_content (CLUTTER_ACTOR (view), content);
         clutter_actor_set_size (CLUTTER_ACTOR (view),
                                 gdk_pixbuf_get_width (pixbuf),
                                 gdk_pixbuf_get_height (pixbuf));
         g_clear_object (&pixbuf);
      }
      g_clear_object (&stream);
   }
}

static void
blp_pixbuf_item_finalize (GObject *object)
{
   BlpPixbufItemPrivate *priv;

   priv = BLP_PIXBUF_ITEM (object)->priv;

   g_clear_object (&priv->file);

   G_OBJECT_CLASS (blp_pixbuf_item_parent_class)->finalize (object);
}

static void
blp_pixbuf_item_get_property (GObject    *object,
                              guint       prop_id,
                              GValue     *value,
                              GParamSpec *pspec)
{
   BlpPixbufItem *item = BLP_PIXBUF_ITEM (object);

   switch (prop_id) {
   case PROP_FILE:
      g_value_set_object (value, blp_pixbuf_item_get_file (item));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_pixbuf_item_set_property (GObject      *object,
                              guint         prop_id,
                              const GValue *value,
                              GParamSpec   *pspec)
{
   BlpPixbufItem *item = BLP_PIXBUF_ITEM (object);

   switch (prop_id) {
   case PROP_FILE:
      blp_pixbuf_item_set_file (item, g_value_get_object (value));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_pixbuf_item_class_init (BlpPixbufItemClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS (klass);
   object_class->finalize = blp_pixbuf_item_finalize;
   object_class->get_property = blp_pixbuf_item_get_property;
   object_class->set_property = blp_pixbuf_item_set_property;

   gParamSpecs [PROP_FILE] =
      g_param_spec_object ("file",
                           _("File"),
                           _("The file containing the image."),
                           G_TYPE_FILE,
                           (G_PARAM_READWRITE |
                            G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (object_class, PROP_FILE,
                                    gParamSpecs [PROP_FILE]);
}

static void
blp_pixbuf_item_init (BlpPixbufItem *item)
{
   item->priv = blp_pixbuf_item_get_instance_private (item);
}
