/* blp-canvas.c
 *
 * Copyright (C) 2014 Christian Hergert <christian@hergert.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include <math.h>

#include "blp-canvas.h"
#include "blp-util.h"

struct _BlpCanvasPrivate
{
   GFile       *background_file;
   BlpDocument *document;
};

enum
{
   PROP_0,
   PROP_BACKGROUND_FILE,
   PROP_DOCUMENT,
   LAST_PROP
};

G_DEFINE_TYPE_WITH_PRIVATE (BlpCanvas, blp_canvas, GTK_CLUTTER_TYPE_EMBED)

static GParamSpec *gParamSpecs [LAST_PROP];

GtkWidget *
blp_canvas_new (void)
{
   return g_object_new (BLP_TYPE_CANVAS, NULL);
}

GFile *
blp_canvas_get_background_file (BlpCanvas *canvas)
{
   g_return_val_if_fail (BLP_IS_CANVAS (canvas), NULL);

   return canvas->priv->background_file;
}

void
blp_canvas_set_background_file (BlpCanvas *canvas,
                                GFile     *background_file)
{
   ClutterContent *content = NULL;
   ClutterActor *stage;

   g_return_if_fail (BLP_IS_CANVAS (canvas));
   g_return_if_fail (!background_file || G_IS_FILE (background_file));

   if (background_file) {
      background_file = g_object_ref (background_file);
      content = blp_clutter_image_new_from_file (background_file);
   }

   if (content) {
      stage = gtk_clutter_embed_get_stage (GTK_CLUTTER_EMBED (canvas));
      clutter_actor_set_content (stage, content);
      clutter_actor_set_content_repeat (stage, CLUTTER_REPEAT_BOTH);
      g_object_unref (content);
   }

   g_clear_object (&canvas->priv->background_file);
   canvas->priv->background_file = background_file;
   g_object_notify_by_pspec (G_OBJECT (canvas), gParamSpecs [PROP_BACKGROUND_FILE]);
}

BlpDocument *
blp_canvas_get_document (BlpCanvas *canvas)
{
   g_return_val_if_fail (BLP_IS_CANVAS (canvas), NULL);

   return canvas->priv->document;
}

void
blp_canvas_set_document (BlpCanvas   *canvas,
                         BlpDocument *document)
{
   BlpCanvasPrivate *priv;
   ClutterActor *stage;
   BlpItem *root;

   g_return_if_fail (BLP_IS_CANVAS (canvas));

   priv = canvas->priv;

   if (document == priv->document) {
      return;
   }

   stage = gtk_clutter_embed_get_stage (GTK_CLUTTER_EMBED (canvas));

   if (priv->document) {
      if ((root = blp_document_get_root (document))) {
         clutter_actor_remove_child (stage, CLUTTER_ACTOR (root));
      }
   }

   g_clear_object (&priv->document);
   priv->document = g_object_ref (document);
   g_object_notify_by_pspec (G_OBJECT (canvas), gParamSpecs [PROP_DOCUMENT]);

   if (document) {
      if ((root = blp_document_get_root (document))) {
         clutter_actor_add_child (stage, CLUTTER_ACTOR (root));
      }
   }
}

static void
blp_canvas_finalize (GObject *object)
{
   BlpCanvasPrivate *priv;

   priv = BLP_CANVAS (object)->priv;

   g_clear_object (&priv->document);

   G_OBJECT_CLASS (blp_canvas_parent_class)->finalize (object);
}

static void
blp_canvas_get_property (GObject    *object,
                         guint       prop_id,
                         GValue     *value,
                         GParamSpec *pspec)
{
   BlpCanvas *canvas = BLP_CANVAS (object);

   switch (prop_id) {
   case PROP_BACKGROUND_FILE:
      g_value_set_object (value, blp_canvas_get_background_file (canvas));
      break;
   case PROP_DOCUMENT:
      g_value_set_object (value, blp_canvas_get_document (canvas));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_canvas_set_property (GObject      *object,
                         guint         prop_id,
                         const GValue *value,
                         GParamSpec   *pspec)
{
   BlpCanvas *canvas = BLP_CANVAS (object);

   switch (prop_id) {
   case PROP_BACKGROUND_FILE:
      blp_canvas_set_background_file (canvas, g_value_get_object (value));
      break;
   case PROP_DOCUMENT:
      blp_canvas_set_document (canvas, g_value_get_object (value));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_canvas_class_init (BlpCanvasClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS (klass);
   object_class->finalize = blp_canvas_finalize;
   object_class->get_property = blp_canvas_get_property;
   object_class->set_property = blp_canvas_set_property;

   gParamSpecs [PROP_BACKGROUND_FILE] =
      g_param_spec_object ("background-file",
                           _("Background File"),
                           _("The background for the canvas."),
                           G_TYPE_FILE,
                           (G_PARAM_READWRITE |
                            G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (object_class, PROP_BACKGROUND_FILE,
                                    gParamSpecs [PROP_BACKGROUND_FILE]);

   gParamSpecs [PROP_DOCUMENT] =
      g_param_spec_object ("document",
                           _("Document"),
                           _("The document for the canvas."),
                           BLP_TYPE_DOCUMENT,
                           (G_PARAM_READWRITE |
                            G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (object_class, PROP_DOCUMENT,
                                    gParamSpecs [PROP_DOCUMENT]);
}

static void
blp_canvas_init (BlpCanvas *canvas)
{
   ClutterActor *stage;

   canvas->priv = blp_canvas_get_instance_private (canvas);

   stage = gtk_clutter_embed_get_stage (GTK_CLUTTER_EMBED (canvas));
   g_object_set_data (G_OBJECT (stage), "embed", canvas);
}
