/* blp-item-pixbuf.c
 *
 * Copyright (C) 2014 Christian Hergert <christian@hergert.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <glib/gi18n.h>

#include "blp-item-pixbuf.h"

struct _BlpItemPixbufPrivate
{
   GFile *file;
};

enum
{
   PROP_0,
   PROP_FILE,
   LAST_PROP
};

G_DEFINE_TYPE_WITH_PRIVATE (BlpItemPixbuf, blp_item_pixbuf, BLP_TYPE_ITEM)

static GParamSpec *gParamSpecs [LAST_PROP];

BlpItem *
blp_item_pixbuf_new_from_path (const gchar *path)
{
   BlpItem *ret;
   GFile *file;

   file = g_file_new_for_path (path);
   ret = g_object_new (BLP_TYPE_ITEM_PIXBUF,
                       "file", file,
                       NULL);
   g_clear_object (&file);

   return ret;
}

/**
 * blp_item_pixbuf_set_file:
 * @pixbuf: (in): A #BlpItemPixbuf.
 * @file: (in) (allow-none): A #GFile or NULL.
 *
 * Set the backing file for the #BlpItemPixbuf.
 */
void
blp_item_pixbuf_set_file (BlpItemPixbuf *item,
                          GFile         *file)
{
   GFileInputStream *stream;
   ClutterContent *content;
   GdkPixbuf *pixbuf;
   gboolean r;
   GError *error = NULL;

   g_return_if_fail (BLP_IS_ITEM_PIXBUF (item));
   g_return_if_fail (!file || G_IS_FILE (file));

   if (file) {
      file = g_object_ref (file);
   }

   g_clear_object (&item->priv->file);
   item->priv->file = file;

   if (file) {
      stream = g_file_read (file, NULL, NULL);
      if (stream) {
         pixbuf = gdk_pixbuf_new_from_stream (G_INPUT_STREAM (stream),
                                              NULL, &error);
         if (pixbuf) {
            content = clutter_image_new ();
            r = clutter_image_set_data (CLUTTER_IMAGE (content),
                                        gdk_pixbuf_get_pixels (pixbuf),
                                        gdk_pixbuf_get_has_alpha (pixbuf) ?
                                           COGL_PIXEL_FORMAT_RGBA_8888 :
                                           COGL_PIXEL_FORMAT_RGB_888,
                                        gdk_pixbuf_get_width (pixbuf),
                                        gdk_pixbuf_get_height (pixbuf),
                                        gdk_pixbuf_get_rowstride (pixbuf),
                                        &error);
            if (!r) {
               g_warning ("%s", error->message);
               g_clear_error (&error);
            } else {
               clutter_actor_set_content (CLUTTER_ACTOR (item), content);
               clutter_actor_set_size (CLUTTER_ACTOR (item),
                                       gdk_pixbuf_get_width (pixbuf),
                                       gdk_pixbuf_get_height (pixbuf));
            }
            g_clear_object (&content);
            g_clear_object (&pixbuf);
         } else {
            g_warning ("%s", error->message);
         }
         g_clear_object (&stream);
      }
   }

   g_object_notify_by_pspec (G_OBJECT (item), gParamSpecs [PROP_FILE]);
}

/**
 * blp_item_pixbuf_get_file:
 * @pixbuf: (in): A #BlpItemPixbuf.
 *
 * Return the backing file for the pixbuf item, or NULL.
 *
 * Returns: (transfer none): A #GFile or NULL.
 */
GFile *
blp_item_pixbuf_get_file (BlpItemPixbuf *pixbuf)
{
   g_return_val_if_fail (BLP_IS_ITEM_PIXBUF (pixbuf), NULL);

   return pixbuf->priv->file;
}

static void
blp_item_pixbuf_finalize (GObject *object)
{
   BlpItemPixbufPrivate *priv;

   priv = BLP_ITEM_PIXBUF (object)->priv;

   g_clear_object (&priv->file);

   G_OBJECT_CLASS (blp_item_pixbuf_parent_class)->finalize (object);
}

static void
blp_item_pixbuf_get_property (GObject    *object,
                              guint       prop_id,
                              GValue     *value,
                              GParamSpec *pspec)
{
   BlpItemPixbuf *pixbuf = BLP_ITEM_PIXBUF (object);

   switch (prop_id) {
   case PROP_FILE:
      g_value_set_object (value, blp_item_pixbuf_get_file (pixbuf));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_item_pixbuf_set_property (GObject      *object,
                              guint         prop_id,
                              const GValue *value,
                              GParamSpec   *pspec)
{
   BlpItemPixbuf *pixbuf = BLP_ITEM_PIXBUF (object);

   switch (prop_id) {
   case PROP_FILE:
      blp_item_pixbuf_set_file (pixbuf, g_value_get_object (value));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_item_pixbuf_class_init (BlpItemPixbufClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS (klass);
   object_class->finalize = blp_item_pixbuf_finalize;
   object_class->get_property = blp_item_pixbuf_get_property;
   object_class->set_property = blp_item_pixbuf_set_property;

   gParamSpecs [PROP_FILE] =
      g_param_spec_object ("file",
                           _("File"),
                           _("A GFile to load."),
                           G_TYPE_FILE,
                           (G_PARAM_READWRITE |
                            G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (object_class, PROP_FILE,
                                    gParamSpecs [PROP_FILE]);
}

static void
blp_item_pixbuf_init (BlpItemPixbuf *pixbuf)
{
   pixbuf->priv = blp_item_pixbuf_get_instance_private (pixbuf);

   clutter_actor_set_reactive (CLUTTER_ACTOR (pixbuf), TRUE);
}
