/* blp-item-plaque.c
 *
 * Copyright (C) 2014 Christian Hergert <christian@hergert.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <gdk/gdk.h>
#include <glib/gi18n.h>

#include "blp-item-plaque.h"

#define lambda(return_type, function_body) \
   ((GCallback)({ \
      return_type __fn__ function_body __fn__; \
   }))

struct _BlpItemPlaquePrivate
{
   ClutterContent *canvas;
};

enum
{
   PROP_0,
   LAST_PROP
};

G_DEFINE_TYPE_WITH_PRIVATE (BlpItemPlaque, blp_item_plaque, BLP_TYPE_ITEM)

static GParamSpec *gParamSpecs [LAST_PROP];

BlpItem *
blp_item_plaque_new (void)
{
   return g_object_new (BLP_TYPE_ITEM_PLAQUE, NULL);
}

static void
rounded_rectangle (cairo_t *cr,
                   gint     x,
                   gint     y,
                   gint     width,
                   gint     height,
                   gint     x_radius,
                   gint     y_radius)
{
  gint x1, x2;
  gint y1, y2;
  gint xr1, xr2;
  gint yr1, yr2;

  x1 = x;
  x2 = x1 + width;
  y1 = y;
  y2 = y1 + height;

  x_radius = MIN (x_radius, width / 2.0);
  y_radius = MIN (y_radius, width / 2.0);

  xr1 = x_radius;
  xr2 = x_radius / 2.0;
  yr1 = y_radius;
  yr2 = y_radius / 2.0;

  cairo_move_to (cr, x1 + xr1, y1);
  cairo_line_to (cr, x2 - xr1, y1);
  cairo_curve_to (cr, x2 - xr2, y1, x2, y1 + yr2, x2, y1 + yr1);
  cairo_line_to (cr, x2, y2 - yr1);
  cairo_curve_to (cr, x2, y2 - yr2, x2 - xr2, y2, x2 - xr1, y2);
  cairo_line_to (cr, x1 + xr1, y2);
  cairo_curve_to (cr, x1 + xr2, y2, x1, y2 - yr2, x1, y2 - yr1);
  cairo_line_to (cr, x1, y1 + yr1);
  cairo_curve_to (cr, x1, y1 + yr2, x1 + xr2, y1, x1 + xr1, y1);
  cairo_close_path (cr);
}

static void
set_color (cairo_t *cr,
           const char *str,
           gdouble alpha)
{
   GdkRGBA rgba;

   if (gdk_rgba_parse (&rgba, str)) {
      rgba.alpha = alpha;
      gdk_cairo_set_source_rgba (cr, &rgba);
   }
}

const gchar *colors[3][3] = {
   { "#eeeeec", "#2e3436", "#ffffff" },
   { "#babdb6", "#2e3436", "#ffffff" },
   { "#729fcf", "#204a87", "#ffffff" },
};
const gdouble alphas[3][3] = {
   { .3, .3, .5 },
   { .5, .3, .5 },
   { .5, .3, .8 }
};

enum
{
   STATE_NORMAL,
   STATE_PRELIGHT,
   STATE_SELECTED,
};

static void
draw (cairo_t *cr,
      int x,
      int y,
      int width,
      int height,
      int state)
{
   cairo_save (cr);

   rounded_rectangle (cr, x, y, width, height, 9, 9);
   set_color (cr, colors [state][0], alphas [state][0]);
   cairo_fill (cr);

   x += 5;
   y += 5;
   width -= 10;
   height -= 10;
   rounded_rectangle (cr, x, y, width, height, 5, 5);
   set_color (cr, colors [state][1], alphas [state][1]);
   cairo_fill (cr);

   x += 2;
   y += 2;
   width -= 4;
   height -= 4;
   rounded_rectangle (cr, x, y, width, height, 3, 3);
   set_color (cr, colors [state][2], alphas [state][2]);
   cairo_fill (cr);

   cairo_restore (cr);
}

static void
blp_item_plaque_allocation_changed (ClutterActor           *actor,
                                    ClutterActorBox        *box,
                                    ClutterAllocationFlags  flags,
                                    gpointer                user_data)
{
   BlpItemPlaquePrivate *priv;
   gfloat width;
   gfloat height;

   g_return_if_fail (CLUTTER_IS_ACTOR (actor));
   g_return_if_fail (box);

   priv = BLP_ITEM_PLAQUE (actor)->priv;

   width = clutter_actor_get_width (actor);
   height = clutter_actor_get_height (actor);

   clutter_canvas_set_size (CLUTTER_CANVAS (priv->canvas), width, height);
   clutter_content_invalidate (priv->canvas);
}

static gboolean
blp_item_plaque_draw (ClutterCanvas *canvas,
                      cairo_t       *cr,
                      gint           width,
                      gint           height,
                      gpointer       user_data)
{
   draw (cr, 0, 0, width, height, STATE_SELECTED);

   return FALSE;
}

static void
blp_item_plaque_finalize (GObject *object)
{
   BlpItemPlaquePrivate *priv;

   priv = BLP_ITEM_PLAQUE (object)->priv;

   G_OBJECT_CLASS (blp_item_plaque_parent_class)->finalize (object);
}

static void
blp_item_plaque_get_property (GObject    *object,
                              guint       prop_id,
                              GValue     *value,
                              GParamSpec *pspec)
{
   BlpItemPlaque *plaque = BLP_ITEM_PLAQUE (object);

   switch (prop_id) {
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_item_plaque_set_property (GObject      *object,
                              guint         prop_id,
                              const GValue *value,
                              GParamSpec   *pspec)
{
   BlpItemPlaque *plaque = BLP_ITEM_PLAQUE (object);

   switch (prop_id) {
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
}

static void
blp_item_plaque_class_init (BlpItemPlaqueClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS (klass);
   object_class->finalize = blp_item_plaque_finalize;
   object_class->get_property = blp_item_plaque_get_property;
   object_class->set_property = blp_item_plaque_set_property;
}

static void
blp_item_plaque_init (BlpItemPlaque *plaque)
{
   plaque->priv = blp_item_plaque_get_instance_private (plaque);

   plaque->priv->canvas = clutter_canvas_new ();
   clutter_actor_set_content (CLUTTER_ACTOR (plaque),
                              plaque->priv->canvas);

   g_signal_connect (plaque->priv->canvas,
                     "draw",
                     G_CALLBACK (blp_item_plaque_draw),
                     NULL);

   g_signal_connect (plaque,
                     "allocation-changed",
                     G_CALLBACK (blp_item_plaque_allocation_changed),
                     NULL);

#if 0
   g_object_set (plaque, "reactive", TRUE, NULL);
   g_signal_connect (plaque, "button-press-event",
                     lambda (gboolean, (ClutterActor *a,
                                        ClutterEvent *e,
                                        gpointer p) {
         clutter_actor_set_pivot_point (a, 0.5, 0.5);
         clutter_actor_set_easing_duration (a, 300);
         clutter_actor_set_easing_mode (a, CLUTTER_EASE_IN_OUT_QUAD);
         clutter_actor_set_scale (a, 1.25, 1.25);
         return FALSE;
      }),
      NULL);
#endif
}
