/* mail-actions.c
 *
 * Copyright (C) 2010 Christian Hergert <chris@dronelabs.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <gdk/gdkkeysyms.h>

#include "mail-actions.h"

typedef struct
{
	GType                  widget_type;
	GType                  action_type;
	MailActionFunc         action_func;
	GtkActionEntry        *action_entries;
	GtkToggleActionEntry  *toggle_action_entries;
	guint                  action_n_entries;
} MailActionFactory;

static GArray *factories = NULL;

/**
 * mail_actions_init:
 *
 * Initialize actions subsystem.
 *
 * Returns: None.
 * Side effects: None.
 */
void
mail_actions_init (void)
{
	factories = g_array_new(FALSE, FALSE, sizeof(MailActionFactory));
}

/**
 * mail_actions_shutdown:
 *
 * Shutdown the actions subsystem and free resources.
 *
 * Returns: None.
 * Side effects: None.
 */
void
mail_actions_shutdown (void)
{
	g_array_free(factories, TRUE);
	factories = NULL;
}

/**
 * mail_actions_factory_create:
 * @factory: (in): A #MailActionFactory.
 * @widget: (in): A #GtkWidget.
 * @action_group: (in): A #GtkActionGroup.
 *
 * Creates one or more instances of a GtkAction using the factory.
 * The factories items are added to @action_group.
 *
 * Returns: None.
 * Side effects: None.
 */
void
mail_action_factory_create (MailActionFactory *factory,
                            GtkWidget *widget,
                            GtkActionGroup *action_group)
{
	GtkAction *action;

	g_return_if_fail(factory != NULL);
	g_return_if_fail(factory->action_type ||
	                 factory->action_func ||
	                 factory->action_entries ||
	                 factory->toggle_action_entries);
	g_return_if_fail(GTK_IS_WIDGET(widget));
	g_return_if_fail(GTK_IS_ACTION_GROUP(action_group));

	if (factory->action_type) {
		action = g_object_new(factory->action_type, NULL);
		gtk_action_group_add_action(action_group, action);
		g_object_unref(action);
	} else if (factory->action_func) {
		action = factory->action_func(widget);
		gtk_action_group_add_action(action_group, action);
		g_object_unref(action);
	} else if (factory->action_entries) {
		gtk_action_group_add_actions(action_group,
		                             factory->action_entries,
		                             factory->action_n_entries,
		                             widget);
	} else if (factory->toggle_action_entries) {
		gtk_action_group_add_toggle_actions(action_group,
		                                    factory->toggle_action_entries,
		                                    factory->action_n_entries,
		                                    widget);
	} else {
		g_assert_not_reached();
	}
}

/**
 * mail_actions_load:
 * @widget: (in): A #GtkWidget.
 * @accel_group: (in): A #GtkAccelGroup.
 * @action_group: (in): A #GtkActionGroup.
 *
 * Loads actions that where registered to the widgets GType using
 * mail_actions_register() or similar.
 *
 * If widget is a #GtkWindow, then @accel_group is attached to the
 * window as well as the resulting actions.
 *
 * Returns: None.
 * Side effects: None.
 */
void
mail_actions_load (GtkWidget *widget,
                   GtkAccelGroup *accel_group,
                   GtkActionGroup *action_group)
{
	MailActionFactory *factory;
	GType widget_type;
	GList *list;
	GList *iter;
	gint i;

	g_return_if_fail(GTK_IS_WIDGET(widget));
	g_return_if_fail(!accel_group || GTK_IS_ACTION_GROUP(action_group));
	g_return_if_fail(GTK_IS_ACTION_GROUP(action_group));
	g_return_if_fail(factories != NULL);

	widget_type = G_TYPE_FROM_INSTANCE(widget);

	for (i = 0; i < factories->len; i++) {
		factory = &g_array_index(factories, MailActionFactory, i);
		mail_action_factory_create(factory, widget, action_group);
	}

	if (accel_group) {
		if (g_type_is_a(widget_type, GTK_TYPE_WINDOW)) {
			list = gtk_action_group_list_actions(action_group);
			for (iter = list; iter; iter = iter->next) {
				gtk_action_set_accel_group(iter->data, accel_group);
			}
			g_list_free(list);
			gtk_window_add_accel_group(GTK_WINDOW(widget), accel_group);
		}
	}
}

/**
 * mail_actions_register:
 * @widget_type: (in): A #GType.
 * @action_type: (in): A #GType.
 *
 * Registers the GType of a GtkAction to be instantiated and added to
 * widgets of @widget_type.
 *
 * Returns: None.
 * Side effects: None.
 */
void
mail_actions_register (GType widget_type,
                       GType action_type)
{
	MailActionFactory factory = { 0 };

	g_return_if_fail(factories != NULL);

	factory.widget_type = widget_type;
	factory.action_type = action_type;
	g_array_append_val(factories, factory);
}

/**
 * mail_actions_register_func:
 * @widget_type: (in): A #GType.
 * @func: (in): A function to create a #GtkAction.
 *
 * Registers a callback to be executed to create an action for the given
 * widget type.
 *
 * Returns: None.
 * Side effects: None.
 */
void
mail_actions_register_func (GType widget_type,
                            MailActionFunc func)
{
	MailActionFactory factory = { 0 };

	g_return_if_fail(factories != NULL);

	factory.widget_type = widget_type;
	factory.action_func = func;
	g_array_append_val(factories, factory);
}

/**
 * mail_actions_register_entries:
 * @widget_type: (in): A #GType.
 * @entries: (array length=n_entries): An array of #GtkActionEntry.
 * @n_entries: (in): Number of @entries.
 *
 * Registers a series of #GtkActionEntry structures to create for
 * instances of @widget_type.
 *
 * Returns: None.
 * Side effects: None.
 */
void
mail_actions_register_entries (GType widget_type,
                               GtkActionEntry *entries,
                               guint n_entries)
{
	MailActionFactory factory = { 0 };

	g_return_if_fail(factories != NULL);
	g_return_if_fail(entries != NULL);

	factory.widget_type = widget_type;
	factory.action_entries = entries;
	factory.action_n_entries = n_entries;
	g_array_append_val(factories, factory);
}

/**
 * mail_actions_register_toggle_entries:
 * @widget_type: (in): A #GType.
 * @entries: (array length=n_entries): An array of #GtkToggleActionEntry.
 * @n_entries: (in): Number of elements in @entries.
 *
 * Registers a series of #GtkToggleActionEntry structures to create for
 * instances of @widget_type.
 *
 * Returns: None.
 * Side effects: None.
 */
void
mail_actions_register_toggle_entries (GType widget_type,
                                      GtkToggleActionEntry *entries,
                                      guint n_entries)
{
	MailActionFactory factory = { 0 };

	g_return_if_fail(factories != NULL);
	g_return_if_fail(entries != NULL);

	factory.widget_type = widget_type;
	factory.toggle_action_entries = entries;
	factory.action_n_entries = n_entries;
	g_array_append_val(factories, factory);
}
