/* mail-log.c
 *
 * Copyright (C) 2010 Christian Hergert <chris@dronelabs.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <time.h>

#include "mail-log.h"

G_LOCK_DEFINE(channels);

static GPtrArray *channels = NULL;

static const gchar *
mail_log_level_str (GLogLevelFlags log_level)
{
#define CASE_LOG_LEVEL(_l) case G_LOG_LEVEL_##_l: return #_l
	switch ((gint)log_level) {
	CASE_LOG_LEVEL(INFO);
	CASE_LOG_LEVEL(DEBUG);
	CASE_LOG_LEVEL(WARNING);
	CASE_LOG_LEVEL(MESSAGE);
	CASE_LOG_LEVEL(CRITICAL);
	CASE_LOG_LEVEL(ERROR);
	CASE_LOG_LEVEL(TRACE);
	default:
		return "UNKNOWN";
	}
#undef CASE_LOG_LEVEL
}

static void
mail_log_handler_func (const gchar *log_domain,
                       GLogLevelFlags log_level,
                       const gchar *message,
                       gpointer user_data)
{
	GIOChannel *channel;
	GTimeVal tv;
	gchar *format;
	gchar time_[64];
	struct tm tm;
	gint i;

	g_get_current_time(&tv);
	tm = *localtime(&tv.tv_sec);
	strftime(time_, sizeof(time_), "%Y/%m/%d %H:%M:%S", &tm);
	format = g_strdup_printf("%s  %s: %16s: %8s: %s\n",
	                         time_, g_get_host_name(), log_domain,
	                         mail_log_level_str(log_level),
	                         message);

	G_LOCK(channels);
	for (i = 0; i < channels->len; i++) {
		channel = g_ptr_array_index(channels, i);
		g_io_channel_write_chars(channel, format, -1, NULL, NULL);
		g_io_channel_flush(channel, NULL);
	}
	G_UNLOCK(channels);

	g_free(format);
}

/**
 * mail_log_init:
 *
 * Initialize the logging system and setup logging callback methods.
 *
 * Returns: None.
 * Side effects: GLib logging is redirected to our handler.
 */
void
mail_log_init (void)
{
	GIOChannel *channel;

	G_LOCK(channels);
	g_assert(!channels);
	channels = g_ptr_array_new();
	channel = g_io_channel_unix_new(0);
	g_ptr_array_add(channels, channel);
	G_UNLOCK(channels);

	g_log_set_default_handler(mail_log_handler_func, NULL);
}

/**
 * mail_log_shutdown:
 *
 * Shuts down our logging subsystem and reverts back to glibs logging
 * handler.
 *
 * Returns: None.
 * Side effects: None.
 */
void
mail_log_shutdown (void)
{
	g_log_set_default_handler(g_log_default_handler, NULL);

	G_LOCK(channels);
	g_ptr_array_foreach(channels, (GFunc)g_io_channel_unref, NULL);
	g_ptr_array_free(channels, TRUE);
	channels = NULL;
	G_UNLOCK(channels);
}

/**
 * mail_log_add_filename:
 * @filename: (in): Filename to receive log entries.
 *
 * Adds the file found at @filename to the list of observers for log entries.
 * The file will be opened in append mode and entries written as they are
 * received.
 *
 * Returns: None.
 * Side effects: The file is opened in append mode.
 */
void
mail_log_add_filename (const gchar *filename)
{
	GIOChannel *channel;
	GError *error = NULL;

	if (!(channel = g_io_channel_new_file(filename, "a", &error))) {
		g_critical("Failed to add logfile \"%s\": %s",
		           filename, error->message);
		g_error_free(error);
		return;
	}

	G_LOCK(channels);
	g_ptr_array_add(channels, channel);
	G_UNLOCK(channels);

	g_message("Now logging to file: %s", filename);
}
