/* mail-runtime.c
 *
 * Copyright (C) 2010 Christian Hergert <chris@dronelabs.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <gtk/gtk.h>

#include "mail-private.h"
#include "mail-runtime.h"

static gint exit_code = 0;

/**
 * mail_runtime_init:
 *
 * This is an internal function for initializing the process at beginning of
 * runtime. It will in tern initialize other subsystems.
 *
 * Returns: %TRUE if succcessful; otherwise %FALSE and the caller should exit.
 * Side effects: Everything.
 */
gboolean
mail_runtime_init (gint *argc,
                   gchar ***argv)
{
	/*
	 * Initialize logging subsystem.
	 */
	mail_log_init();

	/*
	 * Initialize preferences subsystem.
	 */
	if (!mail_prefs_init(argc, argv)) {
		return FALSE;
	}

	gtk_window_set_default_icon_name("emblem-mail");

	mail_actions_init();

	return TRUE;
}

/**
 * mail_runtime_shutdown:
 *
 * This is an internal function for gracefully shutting down services after
 * the MainLoop has quit. It makes it easier for us to determine where
 * memory leaks have occurred.
 *
 * Returns: None.
 * Side effects: Lingering resources are freed.
 */
void
mail_runtime_shutdown (void)
{
	mail_prefs_shutdown();
	mail_log_shutdown();
}

/**
 * mail_runtime_quit:
 *
 * This is functionally the same as calling mail_runtime_quit_with_code()
 * with an exit code of zero.
 *
 * Returns: None.
 * Side effects: The process is destroyed asynchronously.
 */
void
mail_runtime_quit (void)
{
	mail_runtime_quit_with_code(0);
}

/**
 * mail_runtime_quit_with_code:
 * @code: (in): The shell exit code.
 *
 * Requests that the program terminate using the exit code provided.
 * The MainLoop is stopped and therefore the process will terminate when
 * main() exits.
 *
 * Returns: None.
 * Side effects: The process is destroyed asynchronously.
 */
void
mail_runtime_quit_with_code (gint code)
{
	exit_code = code;
	gtk_main_quit();
}

/**
 * mail_runtime_run:
 *
 * Run the MailLoop for the application. If the application has an
 * unexpected failure, an exit code will be returned and the process
 * should exit immediately.
 *
 * Returns: Zero if successful; otherwise a shell exit code.
 * Side effects: Function blocks while running the program.
 */
gint
mail_runtime_run (void)
{
	gtk_main();
	return exit_code;
}
