/* mail-status-view.c
 *
 * Copyright (C) 2010 Christian Hergert <chris@dronelabs.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>

#include "mail-status-view.h"
#include "mail-task.h"

G_DEFINE_TYPE(MailStatusView, mail_status_view, GTK_TYPE_VBOX)

typedef struct
{
	GtkWidget *vbox;
	GtkWidget *title;
	GtkWidget *subtitle;
	GtkWidget *icon;
	GtkWidget *progress;
	GtkWidget *cancel;
} StatusItem;

struct _MailStatusViewPrivate
{
	StatusItem *status_items[MAIL_TASK_LAST];
};

static StatusItem *
status_item_new (const gchar *title,
                 const gchar *subtitle,
                 const gchar *icon_name)
{
	StatusItem *item;
	GtkWidget *image;
	GtkWidget *hbox;
	gchar *markup;

	item = g_new0(StatusItem, 1);

	item->vbox = g_object_new(GTK_TYPE_VBOX,
	                          "border-width", 6,
	                          "spacing", 3,
	                          "vexpand", FALSE,
	                          "visible", TRUE,
	                          NULL);

	markup =
		g_markup_printf_escaped(
				"<span size=\"smaller\" weight=\"bold\">%s</span>",
				title);
	item->title = g_object_new(GTK_TYPE_LABEL,
	                           "visible", TRUE,
	                           "xalign", 0.0f,
	                           "label", markup,
	                           "use-markup", TRUE,
	                           NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(item->vbox),
	                                  item->title,
	                                  "expand", FALSE,
	                                  NULL);
	g_free(markup);

	markup =
		g_markup_printf_escaped("<span size=\"smaller\">%s</span>",
		                        subtitle);
	item->subtitle = g_object_new(GTK_TYPE_LABEL,
	                              "visible", TRUE,
	                              "xalign", 0.0f,
	                              "label", markup,
	                              "use-markup", TRUE,
	                              NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(item->vbox),
	                                  item->subtitle,
	                                  "expand", FALSE,
	                                  NULL);
	g_free(markup);

	hbox = g_object_new(GTK_TYPE_HBOX,
	                    "spacing", 3,
	                    "visible", TRUE,
	                    NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(item->vbox), hbox,
	                                  "expand", FALSE,
	                                  NULL);

	item->icon = g_object_new(GTK_TYPE_IMAGE,
	                          "icon-name", icon_name,
	                          "icon-size", GTK_ICON_SIZE_MENU,
	                          "visible", TRUE,
	                          NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(hbox), item->icon,
	                                  "expand", FALSE,
	                                  NULL);

	item->progress = g_object_new(GTK_TYPE_PROGRESS_BAR,
	                              "visible", TRUE,
	                              "height-request", 16,
	                              NULL);
	gtk_container_add(GTK_CONTAINER(hbox), item->progress);

	image = g_object_new(GTK_TYPE_IMAGE,
	                     "icon-name", GTK_STOCK_CANCEL,
	                     "icon-size", GTK_ICON_SIZE_MENU,
	                     "visible", TRUE,
	                     NULL);
	item->cancel = g_object_new(GTK_TYPE_BUTTON,
	                            "child", image,
	                            "relief", GTK_RELIEF_NONE,
	                            "visible", TRUE,
	                            NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(hbox), item->cancel,
	                                  "expand", FALSE,
	                                  NULL);
	g_signal_connect_swapped(item->cancel, "clicked",
	                         G_CALLBACK(gtk_widget_hide), item->vbox);

	return item;
}
                                     
static void
mail_status_view_finalize (GObject *object)
{
	G_OBJECT_CLASS(mail_status_view_parent_class)->finalize(object);
}

static void
mail_status_view_class_init (MailStatusViewClass *klass)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS(klass);
	object_class->finalize = mail_status_view_finalize;
	g_type_class_add_private(object_class, sizeof(MailStatusViewPrivate));
}

static void
mail_status_view_init (MailStatusView *view)
{
	MailStatusViewPrivate *priv;

	priv = G_TYPE_INSTANCE_GET_PRIVATE(view, MAIL_TYPE_STATUS_VIEW,
	                                   MailStatusViewPrivate);
	view->priv = priv;

	g_object_set(view,
	             "vexpand", FALSE,
	             NULL);

	priv->status_items[MAIL_TASK_INDEXING] =
		status_item_new(_("Indexing"),
		                _("Indexing 137 of 211 items"),
		                GTK_STOCK_FIND);
	g_object_set(priv->status_items[MAIL_TASK_INDEXING]->progress,
	             "fraction", 0.6492f,
	             NULL);

	priv->status_items[MAIL_TASK_SYNCING] =
		status_item_new(_("Syncing"),
		                _("Syncing 1 of 10 items"),
		                "stock_mail-send-receive");
	g_object_set(priv->status_items[MAIL_TASK_SYNCING]->progress,
	             "fraction", 0.10f,
	             NULL);

	gtk_container_add(GTK_CONTAINER(view),
	                  priv->status_items[MAIL_TASK_INDEXING]->vbox);
	gtk_container_add(GTK_CONTAINER(view),
	                  priv->status_items[MAIL_TASK_SYNCING]->vbox);
}
