/* mail-task.c
 *
 * Copyright (C) 2010 Christian Hergert <chris@dronelabs.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mail-task.h"

G_DEFINE_TYPE(MailTask, mail_task, G_TYPE_INITIALLY_UNOWNED)

struct _MailTaskPrivate
{
	MailTaskCategory category;
	GError *error;
	gboolean finished;
	gdouble fraction;
};

enum
{
	PROP_0,
	PROP_CATEGORY,
	PROP_FINISHED,
	PROP_FRACTION,
	PROP_LAST
};

static GParamSpec *pspecs[PROP_LAST] = { 0 };

void
mail_task_set_category (MailTask *task,
                        MailTaskCategory category)
{
	MailTaskPrivate *priv;

	g_return_if_fail(MAIL_IS_TASK(task));

	priv = task->priv;
	priv->category = category;
}

MailTaskCategory
mail_task_get_category (MailTask *task)
{
	g_return_val_if_fail(MAIL_IS_TASK(task), 0);
	return task->priv->category;
}

void
mail_task_finish (MailTask *task)
{
	MailTaskPrivate *priv;

	g_return_if_fail(MAIL_IS_TASK(task));
	g_return_if_fail(!task->priv->finished);

	priv = task->priv;

	priv->finished = TRUE;
	priv->fraction = 1.0;
	g_object_notify_by_pspec(G_OBJECT(task), pspecs[PROP_FRACTION]);
	g_object_notify_by_pspec(G_OBJECT(task), pspecs[PROP_FINISHED]);
}

void
mail_task_finish_with_error (MailTask *task,
                             const GError *error)
{
	MailTaskPrivate *priv;

	g_return_if_fail(MAIL_IS_TASK(task));
	g_return_if_fail(error != NULL);

	priv = task->priv;

	priv->finished = TRUE;
	priv->fraction = 1.0;
	priv->error = g_error_copy(error);
	g_object_notify_by_pspec(G_OBJECT(task), pspecs[PROP_FRACTION]);
	g_object_notify_by_pspec(G_OBJECT(task), pspecs[PROP_FINISHED]);
}

void
mail_task_finish_take_error (MailTask *task,
                             GError *error)
{
	MailTaskPrivate *priv;

	g_return_if_fail(MAIL_IS_TASK(task));
	g_return_if_fail(error != NULL);

	priv = task->priv;

	priv->finished = TRUE;
	priv->fraction = 1.0;
	priv->error = error;
	g_object_notify_by_pspec(G_OBJECT(task), pspecs[PROP_FRACTION]);
	g_object_notify_by_pspec(G_OBJECT(task), pspecs[PROP_FINISHED]);
}

static void
mail_task_set_fraction (MailTask *task,
                        gdouble fraction)
{
	MailTaskPrivate *priv;

	g_return_if_fail(MAIL_IS_TASK(task));
	g_return_if_fail(fraction >= 0.0);
	g_return_if_fail(fraction <= 1.0);

	priv = task->priv;

	priv->fraction = fraction;
	g_object_notify_by_pspec(G_OBJECT(task), pspecs[PROP_FRACTION]);
}

static void
mail_task_finalize (GObject *object)
{
	G_OBJECT_CLASS(mail_task_parent_class)->finalize(object);
}

static void
mail_task_get_property (GObject *object,
                        guint prop_id,
                        GValue *value,
                        GParamSpec *pspec)
{
	MailTask *task = MAIL_TASK(object);

	switch (prop_id) {
	case PROP_CATEGORY:
		g_value_set_enum(value, task->priv->category);
		break;
	case PROP_FINISHED:
		g_value_set_boolean(value, task->priv->finished);
		break;
	case PROP_FRACTION:
		g_value_set_double(value, task->priv->fraction);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
	}
}

static void
mail_task_set_property (GObject *object,
                        guint prop_id,
                        const GValue *value,
                        GParamSpec *pspec)
{
	MailTask *task = MAIL_TASK(object);

	switch (prop_id) {
	case PROP_CATEGORY:
		mail_task_set_category(task, g_value_get_enum(value));
		break;
	case PROP_FRACTION:
		mail_task_set_fraction(task, g_value_get_double(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
	}
}

static void
mail_task_class_init (MailTaskClass *klass)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS(klass);
	object_class->finalize = mail_task_finalize;
	object_class->get_property = mail_task_get_property;
	object_class->set_property = mail_task_set_property;
	g_type_class_add_private(object_class, sizeof(MailTaskPrivate));

	pspecs[PROP_CATEGORY] = g_param_spec_enum("category",
	                                          "category",
	                                          "category",
	                                          MAIL_TYPE_TASK_CATEGORY,
	                                          MAIL_TASK_SYNCING,
	                                          G_PARAM_READWRITE);
	g_object_class_install_property(object_class, PROP_CATEGORY,
	                                pspecs[PROP_CATEGORY]);

	pspecs[PROP_FINISHED] = g_param_spec_boolean("finished",
	                                             "finished",
	                                             "finished",
	                                             FALSE,
	                                             G_PARAM_READABLE);
	g_object_class_install_property(object_class, PROP_FINISHED,
	                                pspecs[PROP_FINISHED]);

	pspecs[PROP_FRACTION] = g_param_spec_double("fraction",
	                                            "fraction",
	                                            "fraction",
	                                            0.0,
	                                            1.0,
	                                            0.0,
	                                            G_PARAM_READWRITE);
	g_object_class_install_property(object_class, PROP_FRACTION,
	                                pspecs[PROP_FRACTION]);
}

static void
mail_task_init (MailTask *task)
{
	MailTaskPrivate *priv;

	priv = G_TYPE_INSTANCE_GET_PRIVATE(task, MAIL_TYPE_TASK,
	                                   MailTaskPrivate);
	task->priv = priv;
}

GType
mail_task_category_get_type (void)
{
	static GType g_type = 0;
	static const GEnumValue values[] = {
		{ MAIL_TASK_SYNCING, "MAIL_TASK_SYNCING", "SYNCING" },
		{ MAIL_TASK_INDEXING, "MAIL_TASK_INDEXING", "INDEXING" },
		{ 0 }
	};

	if (G_UNLIKELY(!g_type)) {
		g_type = g_enum_register_static("MailTaskCategory", values);
	}

	return g_type;
}
