/* mail-window.c
 *
 * Copyright (C) 2010 Christian Hergert <chris@dronelabs.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "gtkcellrendererbubble.h"

#include "mail-actions.h"
#include "mail-enhanced-store.h"
#include "mail-message-view.h"
#include "mail-prefs-dialog.h"
#include "mail-search-tool-item.h"
#include "mail-status-view.h"
#include "mail-util.h"
#include "mail-window.h"
#include "mail-window-actions.h"
#include "mail-window-ui.h"

G_DEFINE_TYPE(MailWindow, mail_window, GTK_TYPE_WINDOW)

struct _MailWindowPrivate
{
	GtkActionGroup *actions;

	GtkWidget *folders_treeview;
	GtkWidget *search_toolitem;
	GtkWidget *status_view;
};

/**
 * mail_window_preferences_activate:
 * @action: (in): A #GtkAction.
 * @window: (in): A #MailWindow.
 *
 * Handles the "acitvate" signal on @action. The preferences dialog is
 * instantiated and displayed. The dialog is modal and therefore
 * interaction with the main window is blocked until the dialog is
 * closed.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
mail_window_preferences_activate (GtkAction *action,
                                  MailWindow *window)
{
	GtkWidget *dialog;

	g_return_if_fail(MAIL_IS_WINDOW(window));

	dialog = g_object_new(MAIL_TYPE_PREFS_DIALOG,
	                      "transient-for", window,
	                      NULL);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}

/**
 * mail_window_search_activate:
 * @action: (in): A #GtkAction.
 * @window: (in): A #MailWindow.
 *
 * Handle the "activate" signal for @action. The search entry is focused.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
mail_window_search_activate (GtkAction *action,
                             MailWindow *window)
{
	MailWindowPrivate *priv;
	GtkWidget *entry;

	g_return_if_fail(MAIL_IS_WINDOW(window));

	priv = window->priv;

	entry = mail_search_tool_item_get_entry(
		MAIL_SEARCH_TOOL_ITEM(priv->search_toolitem));
	gtk_widget_grab_focus(entry);
}

/**
 * mail_window_action_set:
 * @window: (in): A #MailWindow.
 *
 * Helper method similar to calling g_object_set() on a given action.
 * This method takes a @name parameter which is the name of a particular
 * action. The action is found and then g_object_set_valist() is called
 * using it.
 *
 * Returns: None.
 * Side effects: None.
 */
void
mail_window_action_set (MailWindow *window,
                        const gchar *name,
                        const gchar *first_property,
                        ...)
{
	MailWindowPrivate *priv;
	GtkAction *action;
	va_list args;

	g_return_if_fail(MAIL_IS_WINDOW(window));

	priv = window->priv;

	action = gtk_action_group_get_action(priv->actions, name);
	if (!action) {
		g_critical("Action %s could not be found", name);
		return;
	}

	va_start(args, first_property);
	g_object_set_valist(G_OBJECT(action), first_property, args);
	va_end(args);
}

static void
mail_window_finalize (GObject *object)
{
	G_OBJECT_CLASS(mail_window_parent_class)->finalize(object);
}

static void
mail_window_class_init (MailWindowClass *klass)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS(klass);
	object_class->finalize = mail_window_finalize;
	g_type_class_add_private(object_class, sizeof(MailWindowPrivate));

	mail_actions_register_entries(MAIL_TYPE_WINDOW,
	                              mail_window_action_entries,
	                              G_N_ELEMENTS(mail_window_action_entries));
}

static void
mail_window_init (MailWindow *window)
{
	MailWindowPrivate *priv;
	GtkTreeViewColumn *column;
	GtkCellRenderer *cell;
	GtkTreeStore *store;
	GtkWidget *b;
	GtkWidget *hbox;
	GtkWidget *hbox2;
	GtkWidget *hpaned;
	GtkWidget *l;
	GtkWidget *menubar;
	GtkWidget *message;
	GtkWidget *scroller;
	GtkWidget *toolbar;
	GtkWidget *treeview;
	GtkWidget *vbox;
	GtkWidget *vbox2;
	GtkWidget *vbox3;
	GtkWidget *view;
	GtkWidget *vpaned;

	priv = G_TYPE_INSTANCE_GET_PRIVATE(window, MAIL_TYPE_WINDOW,
	                                   MailWindowPrivate);
	window->priv = priv;

	g_object_set(window,
	             "default-width", 600,
	             "default-height", 400,
	             NULL);

	mail_util_load_ui(GTK_WIDGET(window), &priv->actions, mail_window_ui,
	                  "/menubar", &menubar,
	                  "/toolbar", &toolbar,
	                  NULL);

	gtk_style_context_add_class(gtk_widget_get_style_context(toolbar),
	                            GTK_STYLE_CLASS_PRIMARY_TOOLBAR);

	mail_window_action_set(window, "reply",
	                       "sensitive", FALSE,
	                       NULL);
	mail_window_action_set(window, "reply-all",
	                       "sensitive", FALSE,
	                       NULL);
	mail_window_action_set(window, "forward",
	                       "sensitive", FALSE,
	                       NULL);
	mail_window_action_set(window, "delete",
	                       "sensitive", FALSE,
	                       NULL);

	vbox = g_object_new(GTK_TYPE_VBOX,
	                    "visible", TRUE,
	                    NULL);
	gtk_container_add(GTK_CONTAINER(window), vbox);

	gtk_container_add_with_properties(GTK_CONTAINER(vbox), menubar,
	                                  "expand", FALSE,
	                                  NULL);

	hbox = g_object_new(GTK_TYPE_HBOX,
	                    "visible", TRUE,
	                    NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(vbox), hbox,
	                                  "expand", FALSE,
	                                  NULL);

	gtk_container_add_with_properties(GTK_CONTAINER(hbox), toolbar,
	                                  "expand", TRUE,
	                                  "pack-type", GTK_PACK_START,
	                                  NULL);

	toolbar = g_object_new(GTK_TYPE_TOOLBAR,
	                       "visible", TRUE,
	                       "width-request", 200,
	                       NULL);
	gtk_style_context_add_class(gtk_widget_get_style_context(toolbar),
	                            GTK_STYLE_CLASS_PRIMARY_TOOLBAR);
	gtk_container_add_with_properties(GTK_CONTAINER(hbox), toolbar,
	                                  "expand", FALSE,
	                                  "pack-type", GTK_PACK_END,
	                                  NULL);

	priv->search_toolitem = g_object_new(MAIL_TYPE_SEARCH_TOOL_ITEM,
	                                     "visible", TRUE,
	                                     NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(toolbar),
	                                  priv->search_toolitem,
	                                  "expand", TRUE,
	                                  NULL);

	hpaned = g_object_new(GTK_TYPE_HPANED,
	                      "position", 200,
	                      "visible", TRUE,
	                      NULL);
	gtk_container_add(GTK_CONTAINER(vbox), hpaned);

	vbox2 = g_object_new(GTK_TYPE_VBOX,
	                     "visible", TRUE,
	                     NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(hpaned), vbox2,
	                                  "shrink", FALSE,
	                                  NULL);

	scroller = g_object_new(GTK_TYPE_SCROLLED_WINDOW,
	                        "hscrollbar-policy", GTK_POLICY_NEVER,
	                        "shadow-type", GTK_SHADOW_NONE,
	                        "visible", TRUE,
	                        "vscrollbar-policy", GTK_POLICY_AUTOMATIC,
	                        NULL);
	gtk_container_add(GTK_CONTAINER(vbox2), scroller);

	store = mail_enhanced_store_new(NULL);

	priv->folders_treeview = g_object_new(GTK_TYPE_TREE_VIEW,
	                                      "headers-visible", FALSE,
	                                      "model", store,
	                                      "visible", TRUE,
	                                      NULL);
	gtk_container_add(GTK_CONTAINER(scroller), priv->folders_treeview);

	column = gtk_tree_view_column_new();
	gtk_tree_view_append_column(GTK_TREE_VIEW(priv->folders_treeview), column);

	cell = gtk_cell_renderer_pixbuf_new();
	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(column), cell, FALSE);
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(column), cell, "icon-name",
	                              MAIL_ENHANCED_STORE_COLUMN_ICON_NAME);

	cell = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(column), cell, TRUE);
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(column), cell, "text",
	                              MAIL_ENHANCED_STORE_COLUMN_TITLE);

#if 0
	cell = gtk_cell_renderer_bubble_new();
	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(column), cell, FALSE);
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(column), cell, "text",
	                              MAIL_ENHANCED_STORE_COLUMN_COUNT);
#endif

	priv->status_view = g_object_new(MAIL_TYPE_STATUS_VIEW,
	                                 "visible", TRUE,
	                                 NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(vbox2), priv->status_view,
	                                  "expand", FALSE,
	                                  NULL);

	vbox3 = g_object_new(GTK_TYPE_VBOX,
	                     "visible", TRUE,
	                     NULL);
	gtk_container_add(GTK_CONTAINER(hpaned), vbox3);

	message = g_object_new(GTK_TYPE_INFO_BAR,
	                       "visible", TRUE,
	                       "message-type", GTK_MESSAGE_INFO,
	                       NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(vbox3), message,
	                                  "expand", FALSE,
	                                  NULL);

	hbox2 = g_object_new(GTK_TYPE_HBOX,
	                     "spacing", 6,
	                     "visible", TRUE,
	                     NULL);
	gtk_container_add(
			GTK_CONTAINER(
				gtk_info_bar_get_content_area(GTK_INFO_BAR(message))),
			hbox2);

	b = g_object_new(GTK_TYPE_BUTTON,
	                 "label", _("Bugzilla"),
	                 "relief", GTK_RELIEF_NONE,
	                 "visible", TRUE,
	                 NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(hbox2), b,
	                                  "expand", FALSE,
	                                  NULL);

	l = g_object_new(GTK_TYPE_LABEL,
	                 "label", "<b>Need acknowledgement:</b> 10\n"
	                          "<b>Total:</b> 20",
	                 "use-markup", TRUE,
	                 "visible", TRUE,
	                 NULL);
	gtk_container_add_with_properties(GTK_CONTAINER(hbox2), l,
	                                  "expand", FALSE,
	                                  "pack-type", GTK_PACK_END,
	                                  NULL);

	vpaned = g_object_new(GTK_TYPE_VPANED,
	                      "position", 200,
	                      "visible", TRUE,
	                      NULL);
	gtk_container_add(GTK_CONTAINER(vbox3), vpaned);

	scroller = g_object_new(GTK_TYPE_SCROLLED_WINDOW,
	                        "hscrollbar-policy", GTK_POLICY_NEVER,
	                        "shadow-type", GTK_SHADOW_NONE,
	                        "visible", TRUE,
	                        "vscrollbar-policy", GTK_POLICY_AUTOMATIC,
	                        NULL);
	gtk_container_add(GTK_CONTAINER(vpaned), scroller);

	treeview = g_object_new(GTK_TYPE_TREE_VIEW,
	                        "headers-visible", TRUE,
	                        "visible", TRUE,
	                        NULL);
	gtk_container_add(GTK_CONTAINER(scroller), treeview);

	scroller = g_object_new(GTK_TYPE_SCROLLED_WINDOW,
	                        "hscrollbar-policy", GTK_POLICY_NEVER,
	                        "shadow-type", GTK_SHADOW_NONE,
	                        "vscrollbar-policy", GTK_POLICY_AUTOMATIC,
	                        "visible", TRUE,
	                        NULL);
	gtk_container_add(GTK_CONTAINER(vpaned), scroller);

	view = g_object_new(MAIL_TYPE_MESSAGE_VIEW,
	                    "visible", TRUE,
	                    NULL);
	gtk_container_add(GTK_CONTAINER(scroller), view);

	gtk_window_set_focus(GTK_WINDOW(window), priv->folders_treeview);
}
