/* news-application.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include <gom/gom.h>
#include <unistd.h>

#ifdef __linux__
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/utsname.h>
#endif

#include "news-application.h"
#include "news-client.h"
#include "news-debug.h"
#include "news-window.h"

G_DEFINE_TYPE(NewsApplication, news_application, GTK_TYPE_APPLICATION)

struct _NewsApplicationPrivate
{
   GtkActionGroup *actions;
   GomRepository  *repository;
   NewsClient     *client;
};

enum
{
   PROP_0,
   PROP_CLIENT,
   LAST_PROP
};

static GParamSpec *gParamSpecs[LAST_PROP];

/**
 * news_application_get_actions:
 * @application: (in): A #NewsApplication.
 *
 * Retrieve the GtkAction<!-- -->s that are global to the application.
 *
 * Returns: A #GtkActionGroup.
 * Side effects: None.
 */
GtkActionGroup *
news_application_get_actions (NewsApplication *application)
{
   g_return_val_if_fail(NEWS_IS_APPLICATION(application), NULL);
   return application->priv->actions;
}

/**
 * news_application_get_default:
 *
 * Retrieves the singleton instance for #NewsApplication. This resulting
 * object should not be freed.
 *
 * Returns: A #NewsApplication.
 * Side effects: None.
 */
NewsApplication *
news_application_get_default (void)
{
   static gsize initialized = FALSE;
   static NewsApplication *application = NULL;

   if (g_once_init_enter(&initialized)) {
      application = g_object_new(NEWS_TYPE_APPLICATION,
                           "application-id", "org.gnome.SimplyNews",
                           "flags", G_APPLICATION_FLAGS_NONE,
                           NULL);
      g_once_init_leave(&initialized, TRUE);
   }

   return application;
}

/**
 * news_application_get_adapter:
 * @application: (in): A #NewsApplication.
 *
 * Retrieves the adapter for the application.
 *
 * Returns: A #GomRepository if successful; otherwise %NULL.
 */
GomRepository *
news_application_get_adapter (NewsApplication *application)
{
   GomRepository *repository;
   gboolean ret;
   GError *error = NULL;
   gchar *sqlite_dir;
   gchar *sqlite_path;

   g_return_val_if_fail(NEWS_IS_APPLICATION(application), NULL);

   if (application->priv->adapter) {
      return application->priv->adapter;
   }

   /*
    * Create a new adapter for sqlite storage.
    */
   adapter = g_object_new(GOM_TYPE_ADAPTER_SQLITE, NULL);
   g_assert(adapter);

   /*
    * Determine the path that sqlite should use on disk.
    */
   sqlite_path = g_build_filename(g_get_user_data_dir(), "simply-news",
                           "news.db", NULL);

   /*
    * If the directory containing the path does not exist, create it.
    */
   sqlite_dir = g_path_get_dirname(sqlite_path);
   if (!g_file_test(sqlite_dir, G_FILE_TEST_IS_DIR)) {
      g_mkdir_with_parents(sqlite_dir, 0700);
   }
   g_free(sqlite_dir);

   /*
    * Open the sqlite database using our target path.
    */
   g_message("Loading database from %s.", sqlite_path);
   ret = gom_adapter_sqlite_load_from_file(GOM_ADAPTER_SQLITE(adapter),
                                 sqlite_path, &error);
   g_free(sqlite_path);

   /*
    * If we could not open the database, cleanup and fail.
    */
   if (!ret) {
      g_critical("%s", error->message);
      g_error_free(error);
      g_object_unref(adapter);
      return NULL;
   }

   return application->priv->adapter = adapter;
}

/**
 * news_application_activate:
 * @application: (in): A #NewsApplication.
 *
 * Handle activation of the process. Start the client. Show the window.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_application_activate (GApplication *application)
{
   NewsApplicationPrivate *priv;
   NewsApplication *news = (NewsApplication *)application;
   GomAdapter *adapter;
   GtkWindow *window;
   GError *error = NULL;
   gchar *dirpath;
   gchar *path;

   g_return_if_fail(NEWS_IS_APPLICATION(news));

   priv = news->priv;

   if (priv->client) {
      /*
       * FIXME: Show the application window?
       */
      return;
   }

   if (!(adapter = news_application_get_adapter(news))) {
      window = g_object_new(GTK_TYPE_MESSAGE_DIALOG,
                       "buttons", GTK_BUTTONS_CLOSE,
                       "message-type", GTK_MESSAGE_ERROR,
                       "secondary-text", _("There seems to be a problem "
                                      "accessing the database where"
                                      " <b>Simply News</b> stores y"
                                      "our articles. Startup cannot"
                                      " continue."),
                       "secondary-use-markup", TRUE,
                       "text", _("Cannot load Simply News database."),
                       NULL);
      gtk_dialog_run(GTK_DIALOG(window));
      gtk_widget_destroy(GTK_WIDGET(window));
      return;
   }

   /*
    * Create our NewsClient, using our sqlite database for storage.
    */
   priv->client = g_object_new(NEWS_TYPE_CLIENT,
                        "adapter", adapter,
                        NULL);

   /*
    * Load sources from the source list xml.
    */
   path = g_build_filename(g_get_user_config_dir(),
                     "simply-news",
                     "sources.xml",
                     NULL);
   dirpath = g_path_get_dirname(path);
   if (!g_file_test(dirpath, G_FILE_TEST_IS_DIR)) {
      g_mkdir_with_parents(dirpath, 0700);
   }
   g_free(dirpath);
   if (!news_client_load_from_file(priv->client, path, &error)) {
      g_message("Sources file \"%s\" does not exist", path);
   }
   g_free(path);

   window = g_object_new(NEWS_TYPE_WINDOW, NULL);
   gtk_application_add_window(GTK_APPLICATION(application), window);
   gtk_window_present(window);
}

NewsClient *
news_application_get_client (NewsApplication *application)
{
   g_return_val_if_fail(NEWS_IS_APPLICATION(application), NULL);
   return application->priv->client;
}

static inline gint
news_log_get_thread (void)
{
#if __linux__
   return (gint)syscall(SYS_gettid);
#else
   return getpid();
#endif /* __linux__ */
}

static inline const gchar *
news_log_level_str (GLogLevelFlags log_level) /* IN */
{
#define CASE_LEVEL_STR(_l) case G_LOG_LEVEL_##_l: return #_l
   switch ((long)log_level) {
   CASE_LEVEL_STR(ERROR);
   CASE_LEVEL_STR(CRITICAL);
   CASE_LEVEL_STR(WARNING);
   CASE_LEVEL_STR(MESSAGE);
   CASE_LEVEL_STR(INFO);
   CASE_LEVEL_STR(DEBUG);
   CASE_LEVEL_STR(TRACE);
   default:
      return "UNKNOWN";
   }
#undef CASE_LEVEL_STR
}

static void
news_application_log_func (const gchar    *log_domain,
                           GLogLevelFlags  log_level,
                           const gchar    *message,
                           gpointer        user_data)
{
   struct timespec ts;
   struct tm tt;
   time_t t;
   const gchar *level;
   gchar ftime[32];
   gchar *buffer;

   level = news_log_level_str(log_level);
   clock_gettime(CLOCK_REALTIME, &ts);
   t = (time_t)ts.tv_sec;
   tt = *localtime(&t);
   strftime(ftime, sizeof(ftime), "%Y/%m/%d %H:%M:%S", &tt);
   buffer = g_strdup_printf("%s.%04ld %s: %12s[%d]: %8s: %s\n",
                            ftime, ts.tv_nsec / 100000,
                            g_get_host_name(), log_domain,
                            news_log_get_thread(), level, message);
   //write(STDOUT_FILENO, buffer, strlen(buffer));
   g_free(buffer);
}

/**
 * news_application_finalize:
 * @object: (in): A #NewsApplication.
 *
 * Finalizer for a #NewsApplication instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_application_finalize (GObject *object)
{
   NewsApplicationPrivate *priv = NEWS_APPLICATION(object)->priv;

   g_clear_object(&priv->actions);
   g_clear_object(&priv->adapter);
   g_clear_object(&priv->client);

   G_OBJECT_CLASS(news_application_parent_class)->finalize(object);
}

/**
 * news_application_get_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (out): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Get a given #GObject property.
 */
static void
news_application_get_property (GObject    *object,
                               guint       prop_id,
                               GValue     *value,
                               GParamSpec *pspec)
{
   NewsApplication *application = NEWS_APPLICATION(object);

   switch (prop_id) {
   case PROP_CLIENT:
      g_value_set_object(value, application->priv->client);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_application_class_init:
 * @klass: (in): A #NewsApplicationClass.
 *
 * Initializes the #NewsApplicationClass and prepares the vtable.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_application_class_init (NewsApplicationClass *klass)
{
   GObjectClass *object_class;
   GApplicationClass *application_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_application_finalize;
   object_class->get_property = news_application_get_property;
   g_type_class_add_private(object_class, sizeof(NewsApplicationPrivate));

   application_class = G_APPLICATION_CLASS(klass);
   application_class->activate = news_application_activate;

   gParamSpecs[PROP_CLIENT] =
      g_param_spec_object("client",
                          _("Client"),
                          _("The news client."),
                          NEWS_TYPE_CLIENT,
                          G_PARAM_READABLE);
   g_object_class_install_property(object_class, PROP_CLIENT,
                                   gParamSpecs[PROP_CLIENT]);

   g_log_set_default_handler(news_application_log_func, NULL);
}

/**
 * news_application_init:
 * @: (in): A #NewsApplication.
 *
 * Initializes the newly created #NewsApplication instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_application_init (NewsApplication *application)
{
   NewsApplicationPrivate *priv;
   const GtkActionEntry global_actions[] = {
      { "quit", GTK_STOCK_QUIT, NULL, NULL, NULL, gtk_main_quit },
      { "about", GTK_STOCK_ABOUT },
   };

   application->priv = priv =
      G_TYPE_INSTANCE_GET_PRIVATE(application,
                                  NEWS_TYPE_APPLICATION,
                                  NewsApplicationPrivate);

   priv->actions = gtk_action_group_new("global");
   gtk_action_group_add_actions(priv->actions, global_actions,
                         G_N_ELEMENTS(global_actions),
                         application);
}
