/* news-client.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <gdk/gdk.h>
#include <glib/gi18n.h>
#include <gom/gom.h>

#include "news-debug.h"
#include "news-category.h"
#include "news-client.h"
#include "news-item.h"
#include "news-link.h"
#include "news-parser.h"
#include "news-person.h"
#include "news-source.h"
#include "news-source-file.h"
#include "news-source-soup.h"
#include "news-status.h"
#include "news-sync-task.h"
#include "xml-reader.h"
#include "xml-writer.h"

G_DEFINE_TYPE(NewsClient, news_client, G_TYPE_OBJECT)

struct _NewsClientPrivate
{
   GomAdapter    *adapter;        /* The database for saving objects. */
   GHashTable    *sources;        /* Hashtable of source by id. */
   GQueue        *todo;           /* Tasks not yet started. */
   guint          n_active;       /* Number of tasks active */
   GStaticRWLock  tasks_lock;     /* RW lock for accessing tasks */
   GList         *tasks;          /* List of active tasks */
   NewsStatus     status;         /* Status of the client */
   GMutex        *notify_mutex;   /* Mutex for notify handler */
   guint          notify_handler; /* status property notify handler */
};

enum
{
   PROP_0,
   PROP_ADAPTER,
   PROP_STATUS,
   LAST_PROP
};

enum
{
   SOURCE_ADDED,
   LAST_SIGNAL
};

static GParamSpec *gParamSpecs[LAST_PROP];
static guint       gSignals[LAST_SIGNAL];

#define MAX_ACTIVE 20

static gboolean
news_client_status_notify_cb (gpointer data)
{
   NewsClientPrivate *priv;
   NewsClient *client = data;

   g_return_val_if_fail(NEWS_IS_CLIENT(client), FALSE);

   priv = client->priv;

   g_atomic_int_set(&priv->notify_handler, 0);
   g_object_notify_by_pspec(G_OBJECT(client), gParamSpecs[PROP_STATUS]);

   return FALSE;
}

static void
news_client_status_notify (NewsClient *client)
{
   NewsClientPrivate *priv;

   g_return_if_fail(NEWS_IS_CLIENT(client));

   priv = client->priv;

   g_mutex_lock(priv->notify_mutex);

   if (!priv->notify_handler) {
      priv->notify_handler =
         g_timeout_add(50, news_client_status_notify_cb, client);
   }

   g_mutex_unlock(priv->notify_mutex);
}

/**
 * news_client_get_sources:
 * @client: (in): A #NewsClient.
 *
 * Retrieves a newly allocated #GList containing the #NewsSource controlled
 * by @client.
 *
 * Returns: A newly allocated #GList that should be freed with g_list_free().
 */
GList *
news_client_get_sources (NewsClient *client)
{
   g_return_val_if_fail(NEWS_IS_CLIENT(client), NULL);
   return g_hash_table_get_values(client->priv->sources);
}

void
news_client_cancel_tasks (NewsClient *client)
{
   NewsClientPrivate *priv;
   NewsSyncTask *task;
   GList *tasks;
   GList *iter;

   g_return_if_fail(NEWS_IS_CLIENT(client));

   priv = client->priv;

   g_static_rw_lock_writer_lock(&priv->tasks_lock);
   tasks = g_list_copy(priv->tasks);
   g_list_foreach(tasks, (GFunc)g_object_ref, NULL);
   g_static_rw_lock_writer_unlock(&priv->tasks_lock);

   while ((task = g_queue_pop_head(priv->todo))) {
      g_object_unref(task);
   }

   for (iter = tasks; iter; iter = iter->next) {
      news_sync_task_cancel(iter->data);
      g_object_unref(iter->data);
   }
   g_list_free(tasks);
}

static void
news_client_sync_task_fraction_notify (NewsClient   *client,
                                       GParamSpec   *pspec,
                                       NewsSyncTask *task)
{
   g_return_if_fail(NEWS_IS_CLIENT(client));
   g_return_if_fail(NEWS_IS_SYNC_TASK(task));
   news_client_status_notify(client);
}

static void
news_client_sync_task_cb (GObject       *object,
                          GAsyncResult  *result,
                          gpointer       user_data)
{
   NewsClientPrivate *priv;
   NewsSyncTask *task = (NewsSyncTask *)object;
   NewsClient *client = user_data;
   NewsSource *source;
   GError *error = NULL;

   g_return_if_fail(NEWS_IS_CLIENT(client));
   g_return_if_fail(G_IS_ASYNC_RESULT(result));

   priv = client->priv;

   g_object_get(task, "source", &source, NULL);
   if (!news_sync_task_run_finish(task, result, &error)) {
#if 0
      g_printerr("Sync failure: %s\n", error->message);
#endif
      news_source_take_error(source, error);
   } else {
      news_source_emit_refreshed(source);
   }
   g_object_unref(source);

   g_static_rw_lock_writer_lock(&priv->tasks_lock);
   priv->tasks = g_list_remove(priv->tasks, task);
   g_static_rw_lock_writer_unlock(&priv->tasks_lock);

   if ((task = g_queue_pop_head(priv->todo))) {
      g_static_rw_lock_writer_lock(&priv->tasks_lock);
      priv->tasks = g_list_prepend(priv->tasks, task);
      g_static_rw_lock_writer_unlock(&priv->tasks_lock);
      news_sync_task_run_async(task, NULL,
                               news_client_sync_task_cb,
                               g_object_ref(client));
      g_object_unref(task);
   } else {
      priv->n_active--;
   }

   news_client_status_notify(client);
}

/**
 * news_client_source_refresh:
 * @client: (in): A #NewsClient.
 *
 * Refresh @source by fetching the updated content.
 *
 * Returns: None.
 */
static void
news_client_source_refresh (NewsClient *client,
                            NewsSource *source)
{
   NewsClientPrivate *priv;
   NewsSyncTask *task;

   ENTRY;

   g_assert(NEWS_IS_CLIENT(client));
   g_assert(NEWS_IS_SOURCE(source));

   priv = client->priv;

   task = g_object_new(NEWS_TYPE_SYNC_TASK,
                       "adapter", priv->adapter,
                       "source", source,
                       NULL);
   g_signal_connect_swapped(task, "notify::fraction",
                            G_CALLBACK(news_client_sync_task_fraction_notify),
                            client);

   if (priv->n_active > MAX_ACTIVE) {
      g_queue_push_head(priv->todo, task);
   } else {
      priv->n_active++;
      g_static_rw_lock_writer_lock(&priv->tasks_lock);
      priv->tasks = g_list_prepend(priv->tasks, task);
      g_static_rw_lock_writer_unlock(&priv->tasks_lock);
      news_sync_task_run_async(task, NULL,
                               news_client_sync_task_cb,
                               g_object_ref(client));
      g_object_unref(task);
   }

   EXIT;
}

/**
 * news_client_refresh_all:
 * @client: (in): A #NewsClient.
 *
 * Refreshes all sources attached to the client.
 *
 * Returns: None.
 */
void
news_client_refresh_all (NewsClient *client)
{
   NewsClientPrivate *priv;
   GHashTableIter iter;
   NewsSource *source;

   g_return_if_fail(NEWS_IS_CLIENT(client));

   priv = client->priv;

   g_hash_table_iter_init(&iter, priv->sources);
   while (g_hash_table_iter_next(&iter, NULL, (gpointer *)&source)) {
      g_assert(NEWS_IS_SOURCE(source));
      news_client_source_refresh(client, source);
   }
}

/**
 * news_client_add_source:
 * @client: (in): A #NewsClient.
 * @source: (in): A #NewsSource.
 *
 * Adds @source to the list of sources that are monitored by @client.
 * The source will be regularly refreshed based on the frequency
 * specified in @source.
 */
void
news_client_add_source (NewsClient *client,
                        NewsSource *source)
{
   NewsClientPrivate *priv;
   const gchar *id;

   g_return_if_fail(NEWS_IS_CLIENT(client));
   g_return_if_fail(NEWS_IS_SOURCE(source));

   priv = client->priv;

   id = news_source_get_id(source);
   if (g_hash_table_lookup(priv->sources, id)) {
      g_critical("Client already contains source %s.", id);
      return;
   }
   g_hash_table_insert(priv->sources,
                       g_strdup(id),
                       g_object_ref_sink(source));
   g_signal_connect_swapped(source, "refresh",
                            G_CALLBACK(news_client_source_refresh),
                            client);

   g_signal_emit(client, gSignals[SOURCE_ADDED], 0, source);
}

gboolean
news_client_load_from_file (NewsClient   *client,
                     const gchar  *filename,
                     GError      **error)
{
   XmlReader *reader;

   g_return_val_if_fail(NEWS_IS_CLIENT(client), FALSE);

   reader = xml_reader_new();

   if (!xml_reader_load_from_file(reader, filename)) {
      g_set_error(error, NEWS_CLIENT_ERROR,
               NEWS_CLIENT_ERROR_FILE_NOT_FOUND,
               _("The file \"%s\" could not be found."), filename);
      g_object_unref(reader);
      return FALSE;
   }

   xml_reader_read(reader);

   if (!xml_reader_is_a_local(reader, "sources")) {
      goto failure;
   }

   while (xml_reader_read(reader)) {
      if (xml_reader_get_node_type(reader) == 1 &&
         xml_reader_is_a_local(reader, "source")) {
         gchar *id;
         gchar *type;
         gchar *uri;
         GType gtype;
         NewsSource *source;

         id = xml_reader_get_attribute(reader, "id");
         type = xml_reader_get_attribute(reader, "type");
         uri = xml_reader_get_attribute(reader, "uri");

         gtype = g_type_from_name(type);
         if (g_type_is_a(gtype, NEWS_TYPE_SOURCE)) {
            source = g_object_new(gtype,
                             "id", id,
                             "uri", uri,
                             NULL);
            news_client_add_source(client, source);
         }

         g_free(id);
         g_free(uri);
         g_free(type);
      }
   }

   g_object_unref(reader);

   return TRUE;

failure:
   g_set_error(error, NEWS_CLIENT_ERROR, NEWS_CLIENT_ERROR_BAD_FORMAT,
            _("The format of file \"%s\" is incorrect."), filename);
   return FALSE;
}

gboolean
news_client_save_to_file (NewsClient   *client,
                          const gchar  *filename,
                          GError      **error)
{
   NewsClientPrivate *priv;
   XmlWriter *writer;
   GList *iter;
   GList *sources;
   gchar *dirpath;

   g_return_val_if_fail(NEWS_IS_CLIENT(client), FALSE);
   g_return_val_if_fail(filename != NULL, FALSE);

   priv = client->priv;

   dirpath = g_path_get_dirname(filename);
   if (!g_file_test(dirpath, G_FILE_TEST_IS_DIR)) {
      if (-1 == g_mkdir_with_parents(dirpath, 0700)) {
         g_set_error(error, NEWS_CLIENT_ERROR,
                     NEWS_CLIENT_ERROR_FILE_NOT_FOUND,
                     _("Failed to create directory \"%s\""),
                     dirpath);
         g_free(dirpath);
         return FALSE;
      }
   }
   g_free(dirpath);

   writer = xml_writer_new();
   if (!xml_writer_set_file(writer, filename, error)) {
      g_object_unref(writer);
      return FALSE;
   }

   xml_writer_write_start_element(writer, "sources");

   sources = g_hash_table_get_values(priv->sources);

   for (iter = sources; iter; iter = iter->next) {
      gchar *id;
      gchar *uri;
      const gchar *type;

      type = g_type_name(G_TYPE_FROM_INSTANCE(iter->data));
      g_object_get(iter->data,
                   "id", &id,
                   "uri", &uri,
                   NULL);

      xml_writer_write_start_element(writer, "source");
      xml_writer_write_attribute_string(writer, "type", type);
      xml_writer_write_attribute_string(writer, "id", id);
      xml_writer_write_attribute_string(writer, "uri", uri);
      xml_writer_write_end_element(writer);

      g_free(id);
      g_free(uri);
   }

   xml_writer_write_end_element(writer);
   xml_writer_flush(writer);

   g_object_unref(writer);
   g_list_free(sources);

   return TRUE;
}

/**
 * news_client_get_default:
 *
 * Retrieves the default instance of NewsClient. The instance will be created
 * if needed.
 *
 * Returns: A #NewsClient. It should not be unrefed.
 */
NewsClient *
news_client_get_default (void)
{
   static NewsClient *client = NULL;
   static gsize initialized = FALSE;

   if (g_once_init_enter(&initialized)) {
      client = g_object_new(NEWS_TYPE_CLIENT, NULL);
      g_once_init_leave(&initialized, TRUE);
   }

   return client;
}

static const NewsStatus *
news_client_get_status (NewsClient *client)
{
   NewsClientPrivate *priv;
   gdouble fraction;
   GList *iter;
   gint active = 0;
   gint qlen;

   g_return_val_if_fail(NEWS_IS_CLIENT(client), NULL);

   priv = client->priv;

   g_static_rw_lock_reader_lock(&priv->tasks_lock);

   priv->status.fraction = 0;
   qlen = g_queue_get_length(priv->todo);

   for (iter = priv->tasks; iter; iter = iter->next) {
      active++;
      g_object_get(iter->data, "fraction", &fraction, NULL);
      priv->status.fraction += fraction;
   }

   priv->status.total = MAX(priv->status.total, active + qlen);
   priv->status.completed = priv->status.total - active - qlen;
   priv->status.fraction += priv->status.completed;

   if (priv->status.total) {
      priv->status.fraction /= (gdouble)priv->status.total;
   } else {
      priv->status.fraction = 1.0;
   }

   g_static_rw_lock_reader_unlock(&priv->tasks_lock);

   return &priv->status;
}

/**
 * news_client_finalize:
 * @object: (in): A #NewsClient.
 *
 * Finalizer for a #NewsClient instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 */
static void
news_client_finalize (GObject *object)
{
   NewsClientPrivate *priv = NEWS_CLIENT(object)->priv;

   ENTRY;

   g_clear_object(&priv->adapter);
   g_hash_table_destroy(priv->sources);
   g_static_rw_lock_free(&priv->tasks_lock);
   g_queue_free(priv->todo);
   g_mutex_free(priv->notify_mutex);

   G_OBJECT_CLASS(news_client_parent_class)->finalize(object);

   EXIT;
}

/**
 * news_client_get_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (out): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Get a given #GObject property.
 */
static void
news_client_get_property (GObject    *object,
                          guint       prop_id,
                          GValue     *value,
                          GParamSpec *pspec)
{
   NewsClient *client = NEWS_CLIENT(object);

   switch (prop_id) {
   case PROP_ADAPTER:
      g_value_set_object(value, client->priv->adapter);
      break;
   case PROP_STATUS:
      g_value_set_boxed(value, news_client_get_status(client));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_client_set_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (in): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Set a given #GObject property.
 */
static void
news_client_set_property (GObject      *object,
                          guint         prop_id,
                          const GValue *value,
                          GParamSpec   *pspec)
{
   NewsClient *client = NEWS_CLIENT(object);

   switch (prop_id) {
   case PROP_ADAPTER:
      g_clear_object(&client->priv->adapter);
      client->priv->adapter = g_value_dup_object(value);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_client_class_init:
 * @klass: (in): A #NewsClientClass.
 *
 * Initializes the #NewsClientClass and prepares the vtable.
 *
 * Returns: None.
 */
static void
news_client_class_init (NewsClientClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_client_finalize;
   object_class->get_property = news_client_get_property;
   object_class->set_property = news_client_set_property;
   g_type_class_add_private(object_class, sizeof(NewsClientPrivate));

   gParamSpecs[PROP_ADAPTER] =
      g_param_spec_object("adapter",
                          _("Adapter"),
                          _("The gom storage adapter."),
                          GOM_TYPE_ADAPTER,
                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
   g_object_class_install_property(object_class, PROP_ADAPTER,
                                   gParamSpecs[PROP_ADAPTER]);

   gParamSpecs[PROP_STATUS] =
      g_param_spec_boxed("status",
                         _("Status"),
                         _("News client pipeline status."),
                         NEWS_TYPE_STATUS,
                         G_PARAM_READABLE);
   g_object_class_install_property(object_class, PROP_STATUS,
                                   gParamSpecs[PROP_STATUS]);

   gSignals[SOURCE_ADDED] = g_signal_new("source-added",
                                NEWS_TYPE_CLIENT,
                                G_SIGNAL_RUN_FIRST,
                                0,
                                NULL,
                                NULL,
                                g_cclosure_marshal_VOID__OBJECT,
                                G_TYPE_NONE,
                                1,
                                NEWS_TYPE_SOURCE);
}

/**
 * news_client_init:
 * @: (in): A #NewsClient.
 *
 * Initializes the newly created #NewsClient instance.
 *
 * Returns: None.
 */
static void
news_client_init (NewsClient *client)
{
   client->priv =
      G_TYPE_INSTANCE_GET_PRIVATE(client,
                                  NEWS_TYPE_CLIENT,
                                  NewsClientPrivate);

   client->priv->sources =
      g_hash_table_new_full(g_str_hash, g_str_equal,
                            g_free, g_object_unref);
   client->priv->todo = g_queue_new();

   g_static_rw_lock_init(&client->priv->tasks_lock);
   client->priv->notify_mutex = g_mutex_new();

   /*
    * Type registration hack.
    */
   g_debug("Registering source type %s.", g_type_name(NEWS_TYPE_SOURCE_SOUP));
   g_debug("Registering source type %s.", g_type_name(NEWS_TYPE_SOURCE_FILE));
}

GQuark
news_client_error_quark (void)
{
   return g_quark_from_static_string("news-client-error-quark");
}
