/* news-feed.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>

#include "news-category.h"
#include "news-debug.h"
#include "news-feed.h"
#include "news-item.h"
#include "news-link.h"
#include "news-person.h"

G_DEFINE_TYPE(NewsFeed, news_feed, GOM_TYPE_RESOURCE)

/**
 * news_feed_is_newer:
 * @feed: (in): A #NewsFeed.
 * @other: (in): A #NewsFeed.
 *
 * Checks to see if @other is newer than @feed based on the
 * timing information specified within the feed.
 *
 * Returns: %TRUE if @feed is newer than @other.
 * Side effects: None.
 */
gboolean
news_feed_is_newer (NewsFeed *feed,
                    NewsFeed *other)
{
   GDateTime *feed_last_updated;
   GDateTime *other_last_updated;
   gboolean ret = FALSE;

   g_return_val_if_fail(NEWS_IS_FEED(feed), FALSE);
   g_return_val_if_fail(NEWS_IS_FEED(other), FALSE);

   g_object_get(feed,
                "last-updated-at", &feed_last_updated,
             NULL);
   g_object_get(other,
                "last-updated-at", &other_last_updated,
             NULL);

   if (feed_last_updated && other_last_updated) {
      ret = (g_date_time_compare(feed_last_updated,
                                 other_last_updated) > 0);
   } else if (!other_last_updated) {
      ret = TRUE;
   }

   if (feed_last_updated) {
      g_date_time_unref(feed_last_updated);
   }

   if (other_last_updated) {
      g_date_time_unref(other_last_updated);
   }

   return ret;
}

/**
 * news_feed_finalize:
 * @object: (in): A #NewsFeed.
 *
 * Finalize the #NewsFeed instance and free resources.
 *
 * Returns: None.
 */
static void
news_feed_finalize (GObject *object)
{
   ENTRY;
   G_OBJECT_CLASS(news_feed_parent_class)->finalize(object);
   EXIT;
}

/**
 * news_feed_class_init:
 * @klass: (in): A #NewsFeedClass.
 *
 * Initializes the #NewsFeedClass and prepares the vtable.
 *
 * Returns: None.
 */
static void
news_feed_class_init (NewsFeedClass *klass)
{
   GObjectClass *object_class;
   GomResourceClass *resource_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_feed_finalize;
   object_class->get_property = gom_resource_get_property;
   object_class->set_property = gom_resource_set_property;

   resource_class = GOM_RESOURCE_CLASS(klass);
   gom_resource_class_table(resource_class, "feeds");
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_int64("id",
                                                          _("Id"),
                                                          _("Feed Identifier"),
                                                          0,
                                                          G_MAXINT64,
                                                          0,
                                                          G_PARAM_READWRITE),
                                       "key", TRUE,
                                       "serial", TRUE,
                                       NULL);
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_string("base-uri",
                                                           _("Base URI"),
                                                           _("The base uri of the feed."),
                                                           NULL,
                                                           G_PARAM_READWRITE),
                                       NULL);
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_string("copyright",
                                                           _("Copyright"),
                                                           _("The copyright of the feed."),
                                                           NULL,
                                                           G_PARAM_READWRITE),
                                       NULL);
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_string("description",
                                                           _("Description"),
                                                           _("The description of the feed."),
                                                           NULL,
                                                           G_PARAM_READWRITE),
                                       NULL);
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_string("generator",
                                                           _("Generator"),
                                                           _("The generator of the feed."),
                                                           NULL,
                                                           G_PARAM_READWRITE),
                                       NULL);
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_string("feed-id",
                                                           _("Feed Id"),
                                                           _("The feeds upstream identifier"),
                                                           NULL,
                                                           G_PARAM_READWRITE),
                                       NULL);
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_string("image-url",
                                                           _("Image URL"),
                                                           _("The Image URL of the feed."),
                                                           NULL,
                                                           G_PARAM_READWRITE),
                                       NULL);
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_string("language",
                                                           _("Language"),
                                                           _("The language of the feed."),
                                                           NULL,
                                                           G_PARAM_READWRITE),
                                       NULL);
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_boxed("last-updated-at",
                                                           _("Last Updated At"),
                                                           _("The last update of the feed."),
                                                           G_TYPE_DATE_TIME,
                                                           G_PARAM_READWRITE),
                                       NULL);
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_string("source-id",
                                                           _("Source"),
                                                           _("The source which retrieved this feed."),
                                                           NULL,
                                                           G_PARAM_READWRITE),
                                       NULL);
   gom_resource_class_install_property(resource_class,
                                       g_param_spec_string("title",
                                                           _("Title"),
                                                           _("The title of the feed."),
                                                           NULL,
                                                           G_PARAM_READWRITE),
                                       NULL);
   gom_resource_class_has_many(resource_class,
                               "authors",
                               _("Authors"),
                               _("The feeds authors."),
                               NEWS_TYPE_PERSON,
                               "many-to-many", TRUE,
                               NULL);
   gom_resource_class_has_many(resource_class,
                               "categories",
                               _("Categories"),
                               _("The feeds categories."),
                               NEWS_TYPE_CATEGORY,
                               "many-to-many", TRUE,
                               NULL);
   gom_resource_class_has_many(resource_class,
                               "contributors",
                               _("Contributors"),
                               _("The feeds contributors."),
                               NEWS_TYPE_PERSON,
                               "many-to-many", TRUE,
                               NULL);
   gom_resource_class_has_many(resource_class,
                               "links",
                               _("Links"),
                               _("The feeds links."),
                               NEWS_TYPE_LINK,
                               "many-to-many", TRUE,
                               NULL);
   gom_resource_class_has_many(resource_class,
                               "items",
                               _("Items"),
                               _("The feeds items."),
                               NEWS_TYPE_ITEM,
                               NULL);
}

/**
 * news_feed_init:
 * @: (in): A #NewsFeed.
 *
 * Initializes the newly created #NewsFeed instance.
 *
 * Returns: None.
 */
static void
news_feed_init (NewsFeed *feed)
{
   ENTRY;
   EXIT;
}
