/* news-format-atom.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>

#include "news-category.h"
#include "news-content-kind.h"
#include "news-debug.h"
#include "news-format-atom.h"
#include "news-item.h"
#include "news-link.h"
#include "news-person.h"
#include "xml-reader.h"

G_DEFINE_TYPE(NewsFormatAtom, news_format_atom, NEWS_TYPE_FORMAT)

#define XMLNS1998 "http://www.w3.org/XML/1998/namespace"
#define ATOMNS    "http://www.w3.org/2005/Atom"
#define RETURN_ERROR                                       \
    G_STMT_START {                                         \
        g_set_error(error, XML_READER_ERROR,               \
                    XML_READER_ERROR_INVALID,              \
                    _("The atom parser does not know how " \
                      "to read this document"));           \
        RETURN(FALSE);                                     \
    } G_STMT_END

struct _NewsFormatAtomPrivate
{
   GError *error;
};

static NewsPerson *
news_format_atom_read_person (NewsFormatAtom *atom,
                              XmlReader      *reader,
                              GHashTable     *persons)
{
   GomAdapter *adapter;
   NewsPerson *person;
   gboolean valid = FALSE;
   gchar *str;

   ENTRY;

   g_assert(NEWS_IS_FORMAT_ATOM(atom));
   g_assert(XML_IS_READER(reader));

   adapter = news_format_get_adapter(NEWS_FORMAT(atom));
   person = g_object_new(NEWS_TYPE_PERSON,
                         "adapter", adapter,
                         "is-new", TRUE,
                         NULL);

   if (!xml_reader_is_empty_element(reader)) {
      xml_reader_read(reader);

      for (;
           xml_reader_get_node_type(reader) != XML_READER_TYPE_END_ELEMENT;
           xml_reader_read_to_next(reader)) {
         if (atom->priv->error) {
            valid = FALSE;
            goto failure;
         }
         if (xml_reader_get_node_type(reader) == XML_READER_TYPE_ELEMENT &&
             xml_reader_is_namespace(reader, ATOMNS)) {
            if (xml_reader_is_a_local(reader, "name")) {
               str = xml_reader_read_string(reader);
               valid = valid || (str && str[0]);
               g_object_set(person, "name", str, NULL);
               g_free(str);
            } else if (xml_reader_is_a_local(reader, "uri")) {
               str = xml_reader_read_string(reader);
               g_object_set(person, "uri", str, NULL);
               g_free(str);
            } else if (xml_reader_is_a_local(reader, "email")) {
               str = xml_reader_read_string(reader);
               valid = valid || (str && str[0]);
               g_object_set(person, "email", str, NULL);
               g_free(str);
            }
         }
      }
   }

failure:
   if (!valid) {
      g_object_unref(person);
      RETURN(NULL);
   }

   {
      GomResource *resource;
      gchar *email;
      gchar *name;
      gchar *key;

      g_object_get(person,
                   "name", &name,
                "email", &email,
                NULL);
      key = g_strdup_printf("%s|%s", name, email);
      if ((resource = g_hash_table_lookup(persons, key))) {
         g_object_unref(person);
         person = g_object_ref(resource);
      } else {
         g_hash_table_insert(persons, g_strdup(key),
                             g_object_ref(person));
      }
      g_free(name);
      g_free(email);
      g_free(key);
   }

   RETURN(person);
}

static NewsCategory *
news_format_atom_read_category (NewsFormatAtom *atom,
                                XmlReader      *reader,
                                GHashTable     *categories)
{
   NewsCategory *category = NULL;
   GomAdapter *adapter;
   const gchar *str;
   gboolean valid = FALSE;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_FORMAT_ATOM(atom), NULL);
   g_return_val_if_fail(XML_IS_READER(reader), NULL);
   g_return_val_if_fail(categories != NULL, NULL);

   adapter = news_format_get_adapter(NEWS_FORMAT(atom));
   g_assert(adapter);

   category = g_object_new(NEWS_TYPE_CATEGORY,
                           "adapter", adapter,
                           "is-new", TRUE,
                           NULL);

   if (xml_reader_move_to_first_attribute(reader)) {
      do {
         if (xml_reader_is_a_local(reader, "term")) {
            str = xml_reader_get_value(reader);
            valid = valid || (str && str[0]);
            g_object_set(category, "name", str, NULL);
         } else if (xml_reader_is_a_local(reader, "schema")) {
            str = xml_reader_get_value(reader);
            g_object_set(category, "schema", str, NULL);
         } else if (xml_reader_is_a_local(reader, "type")) {
            str = xml_reader_get_value(reader);
            g_object_set(category, "label", str, NULL);
         }
      } while (xml_reader_move_to_next_attribute(reader));
   }

   xml_reader_read_end_element(reader);

   if (!valid) {
      g_object_unref(category);
      RETURN(NULL);
   }

   {
      GomResource *resource;
      gchar *key;

      g_object_get(category, "name", &key, NULL);
      g_assert(key != NULL);

      if ((resource = g_hash_table_lookup(categories, key))) {
         g_object_unref(category);
         category = g_object_ref(resource);
      } else {
         g_hash_table_insert(categories, g_strdup(key),
                             g_object_ref(category));
      }

      g_free(key);
   }

   RETURN(category);
}

static NewsLink *
news_format_atom_read_link (NewsFormatAtom *atom,
                            XmlReader      *reader,
                            GHashTable     *links)
{
   GomAdapter *adapter;
   NewsLink *link_ = NULL;
   const char *str;
   gboolean valid = FALSE;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_FORMAT_ATOM(atom), NULL);
   g_return_val_if_fail(XML_IS_READER(reader), NULL);
   g_return_val_if_fail(links != NULL, NULL);

   adapter = news_format_get_adapter(NEWS_FORMAT(atom));
   g_assert(adapter);

   link_ = g_object_new(NEWS_TYPE_LINK,
                        "adapter", adapter,
                        "is-new", TRUE,
                        NULL);

   if (xml_reader_move_to_first_attribute(reader)) {
      do {
         if (xml_reader_is_a_local(reader, "href")) {
            str = xml_reader_get_value(reader);
            valid = valid || (str && str[0]);
            g_object_set(link_, "href", str, NULL);
         } else if (xml_reader_is_a_local(reader, "rel")) {
            str = xml_reader_get_value(reader);
            g_object_set(link_, "relation", str, NULL);
         } else if (xml_reader_is_a_local(reader, "type")) {
            str = xml_reader_get_value(reader);
            g_object_set(link_, "media-type", str, NULL);
         } else if (xml_reader_is_a_local(reader, "title")) {
            str = xml_reader_get_value(reader);
            g_object_set(link_, "title", str, NULL);
         } else if (xml_reader_is_a_local(reader, "length")) {
            gint64 length;

            if ((str = xml_reader_get_value(reader))) {
               length = g_ascii_strtoll(str, NULL, 0);
               g_object_set(link_, "length", length, NULL);
            }
         }
      } while (xml_reader_move_to_next_attribute(reader));
   }

   xml_reader_read_end_element(reader);

   {
      GomResource *resource;
      gchar *key;

      g_object_get(link_, "href", &key, NULL);
      g_assert(key != NULL);

      if ((resource = g_hash_table_lookup(links, key))) {
         g_object_unref(link_);
         link_ = g_object_ref(resource);
      } else {
         g_hash_table_insert(links, g_strdup(key),
                             g_object_ref(link_));
      }

      g_free(key);
   }

   RETURN(link_);
}

static gboolean
news_format_atom_read_item (NewsFormatAtom *atom,
                            XmlReader      *reader,
                            NewsItem       *item,
                            GHashTable     *persons,
                            GHashTable     *categories,
                            GHashTable     *links)
{
   gchar *str;

   xml_reader_read(reader);

   for (;
        xml_reader_get_node_type(reader) != XML_READER_TYPE_END_ELEMENT;
        xml_reader_read_to_next(reader)) {

      if (atom->priv->error) {
         return FALSE;
      }

      if (xml_reader_is_a_local(reader, "author")) {
         GomCollection *collection;
         NewsPerson *person;

         g_object_get(item, "authors", &collection, NULL);
         if ((person = news_format_atom_read_person(atom, reader, persons))) {
            gom_collection_add_resource(collection, GOM_RESOURCE(person));
            g_object_unref(person);
         }
         g_object_unref(collection);
      } else if (xml_reader_is_a_local(reader, "contributor")) {
         GomCollection *collection;
         NewsPerson *person;

         g_object_get(item, "contributors", &collection, NULL);
         if ((person = news_format_atom_read_person(atom, reader, persons))) {
            gom_collection_add_resource(collection, GOM_RESOURCE(person));
            g_object_unref(person);
         }
         g_object_unref(collection);
      } else if (xml_reader_is_a_local(reader, "category")) {
         GomCollection *collection;
         NewsCategory *category;

         g_object_get(item, "categories", &collection, NULL);
         if ((category = news_format_atom_read_category(atom, reader, categories))) {
            gom_collection_add_resource(collection, GOM_RESOURCE(category));
            g_object_unref(category);
         }
         g_object_unref(collection);
      } else if (xml_reader_is_a_local(reader, "link")) {
         GomCollection *collection;
         NewsLink *link_;

         g_object_get(item, "links", &collection, NULL);
         if ((link_ = news_format_atom_read_link(atom, reader, links))) {
            gom_collection_add_resource(collection, GOM_RESOURCE(link_));
            g_object_unref(link_);
         }
         g_object_unref(collection);
      } else if (xml_reader_is_a_local(reader, "id")) {
         str = xml_reader_read_string(reader);
         g_object_set(item, "item-id", str, NULL);
         g_free(str);
      } else if (xml_reader_is_a_local(reader, "rights")) {
         str = xml_reader_read_string(reader);
         g_object_set(item, "copyright", str, NULL);
         g_free(str);
      } else if (xml_reader_is_a_local(reader, "summary")) {
         str = xml_reader_read_string(reader);
         g_object_set(item, "summary", str, NULL);
         g_free(str);
      } else if (xml_reader_is_a_local(reader, "title")) {
         str = xml_reader_read_string(reader);
         g_object_set(item, "title", str, NULL);
         g_free(str);
      } else if (xml_reader_is_a_local(reader, "updated")) {
         GDateTime *dt;
         GTimeVal tv;

         if ((str = xml_reader_read_string(reader))) {
            if (g_time_val_from_iso8601(str, &tv)) {
               if ((dt = g_date_time_new_from_timeval_utc(&tv))) {
                  g_object_set(item,
                               "last-updated-at", dt,
                               NULL);
                  g_date_time_unref(dt);
               }
            }
            g_free(str);
         }
      } else if (xml_reader_is_a_local(reader, "source")) {
         /*
          * TODO:
          */
      } else if (xml_reader_is_a_local(reader, "published")) {
         GDateTime *dt;
         GTimeVal tv;

         if ((str = xml_reader_read_string(reader))) {
            if (g_time_val_from_iso8601(str, &tv)) {
               if ((dt = g_date_time_new_from_timeval_utc(&tv))) {
                  g_object_set(item,
                               "published-at", dt,
                               NULL);
                  g_date_time_unref(dt);
               }
            }
            g_free(str);
         }
      } else if (xml_reader_is_a_local(reader, "content")) {
         NewsContentKind kind;

         str = xml_reader_get_attribute(reader, "type");
         if (news_content_kind_parse(&kind, str)) {
            g_free(str);
            str = xml_reader_read_inner_xml(reader);
            g_object_set(item,
                         "content", str,
                         "content-type", kind,
                         NULL);
         }
         g_free(str);
      }
   }

   xml_reader_read_end_element(reader);

   return TRUE;
}

/**
 * news_format_atom_read_from_xml:
 * @atom: (in): A #NewsFormatAtom.
 * @reader: (in): An #XmlReader containing the feed.
 * @feed: (in): A #NewsFeed to populate with content from @reader.
 * @error: (out): A location for a #GError, or %NULL.
 *
 * Parses the XML found in @reader and stores the result into @feed.
 * If successful; %TRUE is returned. Otherwise %FALSE.
 *
 * Note that there is a side-effect of @feed being modified even if
 * the function is unsuccessful.
 *
 * Returns: None.
 */
static gboolean
news_format_atom_read_from_xml (NewsFormatAtom  *atom,
                                XmlReader       *reader,
                                NewsFeed        *feed,
                                GError         **error)
{
   GHashTable *categories = NULL;
   GHashTable *persons = NULL;
   GHashTable *links = NULL;
   GomAdapter *adapter;
   gchar *str;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_FORMAT_ATOM(atom), FALSE);
   g_return_val_if_fail(XML_IS_READER(reader), FALSE);
   g_return_val_if_fail(NEWS_IS_FEED(feed), FALSE);

   adapter = news_format_get_adapter(NEWS_FORMAT(atom));
   g_assert(adapter);

skip:
   if (!xml_reader_read_to_next(reader) ||
       !xml_reader_is_a(reader, "feed")) {
      if (xml_reader_is_a(reader, "xml-stylesheet") ||
          xml_reader_is_a(reader, "#comment")) {
         goto skip;
      }
#if 0
      {
         gchar *uri = NULL;
         NewsSource *source = NULL;
         g_object_get(atom, "source", &source, NULL);
         g_assert(NEWS_IS_SOURCE(source));
         g_object_get(source, "uri", &uri, NULL);
         g_assert(uri);
         g_printerr("URI=%s\n", uri);
         g_free(uri);
         g_clear_object(&source);
      }
#endif
      RETURN_ERROR;
   }

   if (xml_reader_move_to_first_attribute(reader)) {
      do {
         if (xml_reader_is_a_local(reader, "lang") &&
             xml_reader_is_namespace(reader, XMLNS1998)) {
            str = xml_reader_get_attribute(reader, "lang");
            g_object_set(feed, "language", str, NULL);
            g_free(str);
         }
      } while (xml_reader_move_to_next_attribute(reader));
   }

   persons = g_hash_table_new_full(g_str_hash, g_str_equal,
                                   g_free, g_object_unref);
   categories = g_hash_table_new_full(g_str_hash, g_str_equal,
                                      g_free, g_object_unref);
   links = g_hash_table_new_full(g_str_hash, g_str_equal,
                                 g_free, g_object_unref);

   xml_reader_read(reader);

   for (xml_reader_read_to_next(reader);
        xml_reader_get_node_type(reader) != XML_READER_TYPE_END_ELEMENT;
        xml_reader_read_to_next(reader)) {

      if ((xml_reader_get_node_type(reader) != XML_READER_TYPE_ELEMENT) || atom->priv->error) {
         g_hash_table_destroy(categories);
         g_hash_table_destroy(persons);
         g_hash_table_destroy(links);
         RETURN_ERROR;
      }

      if (xml_reader_is_namespace(reader, ATOMNS)) {
         if (xml_reader_is_a_local(reader, "author")) {
            GomCollection *authors;
            NewsPerson *person;

            g_object_get(feed, "authors", &authors, NULL);
            if ((person = news_format_atom_read_person(atom, reader, persons))) {
               gom_collection_add_resource(authors, GOM_RESOURCE(person));
               g_object_unref(person);
            }
            g_object_unref(authors);
         } else if (xml_reader_is_a_local(reader, "category")) {
            GomCollection *collection;
            NewsCategory *category;

            g_object_get(feed, "categories", &collection, NULL);
            if ((category = news_format_atom_read_category(atom, reader, categories))) {
               gom_collection_add_resource(collection, GOM_RESOURCE(category));
               g_object_unref(category);
            }
            g_object_unref(collection);
         } else if (xml_reader_is_a_local(reader, "contributor")) {
            GomCollection *contributors;
            NewsPerson *person;

            g_object_get(feed, "contributors", &contributors, NULL);
            if ((person = news_format_atom_read_person(atom, reader, persons))) {
               gom_collection_add_resource(contributors, GOM_RESOURCE(person));
               g_object_unref(person);
            }
            g_object_unref(contributors);
         } else if (xml_reader_is_a_local(reader, "generator")) {
            str = xml_reader_read_string(reader);
            g_object_set(feed, "generator", str, NULL);
            g_free(str);
         } else if (xml_reader_is_a_local(reader, "id")) {
            str = xml_reader_read_string(reader);
            g_object_set(feed, "feed-id", str, NULL);
            g_free(str);
         } else if (xml_reader_is_a_local(reader, "link")) {
            GomCollection *collection;
            NewsLink *link_;

            g_object_get(feed, "links", &collection, NULL);
            if ((link_ = news_format_atom_read_link(atom, reader, links))) {
               gom_collection_add_resource(collection, GOM_RESOURCE(link_));
               g_object_unref(link_);
            }
            g_object_unref(collection);
         } else if (xml_reader_is_a_local(reader, "logo")) {
            str = xml_reader_read_string(reader);
            g_object_set(feed, "image-url", str, NULL);
            g_free(str);
         } else if (xml_reader_is_a_local(reader, "rights")) {
            str = xml_reader_read_string(reader);
            g_object_set(feed, "copyright", str, NULL);
            g_free(str);
         } else if (xml_reader_is_a_local(reader, "subtitle")) {
            str = xml_reader_read_string(reader);
            g_object_set(feed, "description", str, NULL);
            g_free(str);
         } else if (xml_reader_is_a_local(reader, "title")) {
            str = xml_reader_read_string(reader);
            g_object_set(feed, "title", str, NULL);
            g_free(str);
         } else if (xml_reader_is_a_local(reader, "updated")) {
            GDateTime *dt;
            GTimeVal tv;

            if ((str = xml_reader_read_string(reader))) {
               if (g_time_val_from_iso8601(str, &tv)) {
                  if ((dt = g_date_time_new_from_timeval_utc(&tv))) {
                     g_object_set(feed,
                                  "last-updated-at", dt,
                                  NULL);
                     g_date_time_unref(dt);
                  }
               }
               g_free(str);
            }
         } else if (xml_reader_is_a_local(reader, "entry")) {
            GomCollection *collection;
            NewsItem *item;

            g_object_get(feed, "items", &collection, NULL);
            item = g_object_new(NEWS_TYPE_ITEM,
                           "adapter", adapter,
                           "is-new", TRUE,
                           NULL);
            if (news_format_atom_read_item(atom, reader, item, persons, categories, links)) {
               gom_collection_add_resource(collection,
                                           GOM_RESOURCE(item));
            }
            g_object_unref(item);
            g_object_unref(collection);
         }
      }
   }

   g_hash_table_destroy(categories);
   g_hash_table_destroy(links);
   g_hash_table_destroy(persons);

   RETURN(TRUE);
}

static void
news_format_atom_on_error (XmlReader      *reader,
                           const char     *message,
                           NewsFormatAtom *atom)
{
   g_return_if_fail(XML_IS_READER(reader));
   g_return_if_fail(NEWS_IS_FORMAT_ATOM(atom));

   if (!atom->priv->error) {
      g_set_error(&atom->priv->error, NEWS_FORMAT_ATOM_ERROR,
                  NEWS_FORMAT_ATOM_ERROR_INVALID_XML,
                  _("Invalid XML detected: %s"), message);
   }
}

/**
 * news_format_atom_read:
 * @atom: (in): A #NewsFormatAtom.
 *
 * Synchronous implementation to parse a feed from @stream.
 *
 * Returns: A #NewsFeed if successful; otherwise %NULL.
 */
static NewsFeed *
news_format_atom_read (NewsFormat    *format,
                       GInputStream  *stream,
                       GError       **error)
{
   NewsFormatAtom *atom = (NewsFormatAtom *)format;
   NewsSource *source;
   GomAdapter *adapter;
   XmlReader *reader;
   NewsFeed *feed;
   gchar *source_id;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_FORMAT_ATOM(atom), NULL);

   if (!(source = news_format_get_source(format))) {
      g_assert_not_reached();
      RETURN(NULL);
   }

   if (!(adapter = news_format_get_adapter(format))) {
      g_assert_not_reached();
      RETURN(NULL);
   }

   reader = xml_reader_new();
   if (!xml_reader_load_from_stream(reader, stream, error)) {
      g_object_unref(reader);
      RETURN(NULL);
   }

   g_signal_connect(reader, "error",
                    G_CALLBACK(news_format_atom_on_error),
                    format);

   g_object_get(source, "id", &source_id, NULL);
   feed = g_object_new(NEWS_TYPE_FEED,
                       "adapter", adapter,
                       "is-new", TRUE,
                       "source-id", source_id,
                       NULL);
   if (!news_format_atom_read_from_xml(atom, reader, feed, error)) {
      g_clear_object(&feed);
   } else if (atom->priv->error) {
      g_clear_object(&feed);
      if (error) {
         *error = g_error_copy(atom->priv->error);
      }
   }

   g_free(source_id);
   g_clear_object(&reader);

   RETURN(feed);
}

/**
 * news_format_atom_finalize:
 * @object: (in): A #NewsFormatAtom.
 *
 * Finalize the given GObject and free resources.
 *
 * Returns: None.
 */
static void
news_format_atom_finalize (GObject *object)
{
   NewsFormatAtomPrivate *priv = NEWS_FORMAT_ATOM(object)->priv;

   ENTRY;

   g_clear_error(&priv->error);

   G_OBJECT_CLASS(news_format_atom_parent_class)->finalize(object);

   EXIT;
}

/**
 * news_format_atom_class_init:
 * @klass: (in): A #NewsFormatAtomClass.
 *
 * Initialize the #NewsFormatAtomClass class.
 *
 * Returns: None.
 */
static void
news_format_atom_class_init (NewsFormatAtomClass *klass)
{
   NewsFormatClass *format_class;
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_format_atom_finalize;
   g_type_class_add_private(object_class, sizeof(NewsFormatAtomPrivate));

   format_class = NEWS_FORMAT_CLASS(klass);
   format_class->read = news_format_atom_read;
}

/**
 * news_format_atom_init:
 * @atom: (in): A #NewsFormatAtom.
 *
 * Initialize the #NewsFormatAtom instance.
 *
 * Returns: None.
 */
static void
news_format_atom_init (NewsFormatAtom *atom)
{
   ENTRY;
   atom->priv =
      G_TYPE_INSTANCE_GET_PRIVATE(atom,
                                  NEWS_TYPE_FORMAT_ATOM,
                                  NewsFormatAtomPrivate);
   EXIT;
}

GQuark
news_format_atom_error_quark (void)
{
   return g_quark_from_static_string("news-format-atom-error-quark");
}
