/* news-format-rss.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>

#include "news-category.h"
#include "news-content-kind.h"
#include "news-debug.h"
#include "news-item.h"
#include "news-format-rss.h"
#include "news-link.h"
#include "news-person.h"
#include "xml-reader.h"

G_DEFINE_TYPE(NewsFormatRss, news_format_rss, NEWS_TYPE_FORMAT)

#define RETURN_ERROR                                       \
    G_STMT_START {                                         \
        g_set_error(error, XML_READER_ERROR,               \
                    XML_READER_ERROR_INVALID,              \
                    _("The rss parser does not know how "  \
                      "to read this document"));           \
        RETURN(FALSE);                                     \
    } G_STMT_END

struct _NewsFormatRssPrivate
{
   GError *error;
};

static NewsPerson *
news_format_rss_read_person (NewsFormatRss *rss,
                             XmlReader      *reader,
                             GHashTable     *persons)
{
   GomAdapter *adapter;
   NewsPerson *person;
   gboolean valid = FALSE;
   gchar *str;

   ENTRY;

   g_assert(NEWS_IS_FORMAT_RSS(rss));
   g_assert(XML_IS_READER(reader));

   adapter = news_format_get_adapter(NEWS_FORMAT(rss));
   person = g_object_new(NEWS_TYPE_PERSON,
                         "adapter", adapter,
                         "is-new", TRUE,
                         NULL);

   if (!xml_reader_is_empty_element(reader)) {
      xml_reader_read(reader);

      for (;
           xml_reader_get_node_type(reader) != XML_READER_TYPE_END_ELEMENT;
           xml_reader_read_to_next(reader)) {
         if (rss->priv->error) {
            break;
         }
         if (xml_reader_get_node_type(reader) == XML_READER_TYPE_ELEMENT) {
            if (xml_reader_is_a_local(reader, "email")) {
               str = xml_reader_read_string(reader);
               valid = valid || (str && str[0]);
               g_object_set(person, "email", str, NULL);
               g_free(str);
            }
         }
      }
   }

   if (!valid) {
      g_object_unref(person);
      RETURN(NULL);
   }

   {
      GomResource *resource;
      gchar *email;
      gchar *name;
      gchar *key;

      g_object_get(person,
                   "name", &name,
                "email", &email,
                NULL);
      key = g_strdup_printf("%s|%s", name, email);
      if ((resource = g_hash_table_lookup(persons, key))) {
         g_object_unref(person);
         person = g_object_ref(resource);
      } else {
         g_hash_table_insert(persons, g_strdup(key),
                             g_object_ref(person));
      }
      g_free(name);
      g_free(email);
      g_free(key);
   }

   RETURN(person);
}

static NewsLink *
news_format_rss_read_link (NewsFormatRss *rss,
                           XmlReader      *reader,
                           GHashTable     *links)
{
   GomAdapter *adapter;
   NewsLink *link_ = NULL;
   gchar *str;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_FORMAT_RSS(rss), NULL);
   g_return_val_if_fail(XML_IS_READER(reader), NULL);
   g_return_val_if_fail(links != NULL, NULL);

   adapter = news_format_get_adapter(NEWS_FORMAT(rss));
   g_assert(adapter);

   str = xml_reader_read_string(reader);
   if (!str || !str[0]) {
      g_free(str);
      RETURN(NULL);
   }
   link_ = g_object_new(NEWS_TYPE_LINK,
                        "adapter", adapter,
                        "is-new", TRUE,
                   "href", str,
                        NULL);
   g_free(str);

   {
      GomResource *resource;
      gchar *key;

      g_object_get(link_, "href", &key, NULL);
      g_assert(key != NULL);

      if ((resource = g_hash_table_lookup(links, key))) {
         g_object_unref(link_);
         link_ = g_object_ref(resource);
      } else {
         g_hash_table_insert(links, g_strdup(key),
                             g_object_ref(link_));
      }

      g_free(key);
   }

   RETURN(link_);
}

static NewsCategory *
news_format_rss_read_category (NewsFormatRss *rss,
                               XmlReader      *reader,
                               GHashTable     *categories)
{
   NewsCategory *category = NULL;
   const gchar *str;
   GomAdapter *adapter;
   gchar *name;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_FORMAT_RSS(rss), NULL);
   g_return_val_if_fail(XML_IS_READER(reader), NULL);
   g_return_val_if_fail(categories != NULL, NULL);

   adapter = news_format_get_adapter(NEWS_FORMAT(rss));
   g_assert(adapter);

   category = g_object_new(NEWS_TYPE_CATEGORY,
                           "adapter", adapter,
                           "is-new", TRUE,
                           NULL);

   if (xml_reader_move_to_first_attribute(reader)) {
      do {
         if (xml_reader_is_a_local(reader, "domain")) {
            str = xml_reader_get_value(reader);
            g_object_set(category, "schema", str, NULL);
         }
      } while (xml_reader_move_to_next_attribute(reader));
   }

   if (!(name = xml_reader_read_inner_xml(reader))) {
      g_object_unref(category);
      RETURN(NULL);
   }

   g_object_set(category, "name", name, "label", name, NULL);
   g_free(name);

   {
      GomResource *resource;
      gchar *key;

      g_object_get(category, "name", &key, NULL);
      g_assert(key != NULL);

      if ((resource = g_hash_table_lookup(categories, key))) {
         g_object_unref(category);
         category = g_object_ref(resource);
      } else {
         g_hash_table_insert(categories, g_strdup(key),
                             g_object_ref(category));
      }

      g_free(key);
   }

   RETURN(category);
}

static gboolean
news_format_rss_read_item (NewsFormatRss *rss,
                           XmlReader     *reader,
                           NewsItem      *item,
                           GHashTable    *persons,
                           GHashTable    *categories,
                           GHashTable    *links)
{
   gchar *str;

   xml_reader_read(reader);

   for (;
        xml_reader_get_node_type(reader) != XML_READER_TYPE_END_ELEMENT;
        xml_reader_read_to_next(reader)) {
      if (rss->priv->error) {
         break;
      }

      if (xml_reader_is_a_local(reader, "author")) {
         GomCollection *collection;
         NewsPerson *person;

         g_object_get(item, "authors", &collection, NULL);
         if ((person = news_format_rss_read_person(rss, reader, persons))) {
            gom_collection_add_resource(collection, GOM_RESOURCE(person));
            g_object_unref(person);
         }
         g_object_unref(collection);
      } else if (xml_reader_is_a_local(reader, "category")) {
         GomCollection *collection;
         NewsCategory *category;

         g_object_get(item, "categories", &collection, NULL);
         if ((category = news_format_rss_read_category(rss, reader, categories))) {
            gom_collection_add_resource(collection, GOM_RESOURCE(category));
            g_object_unref(category);
         }
         g_object_unref(collection);
      } else if (xml_reader_is_a_local(reader, "link")) {
         GomCollection *collection;
         NewsLink *link_;

         g_object_get(item, "links", &collection, NULL);
         if ((link_ = news_format_rss_read_link(rss, reader, links))) {
            gom_collection_add_resource(collection, GOM_RESOURCE(link_));
            g_object_unref(link_);
         }
         g_object_unref(collection);
      } else if (xml_reader_is_a_local(reader, "guid")) {
         str = xml_reader_read_string(reader);
         g_object_set(item, "item-id", str, NULL);
         g_free(str);
      } else if (xml_reader_is_a_local(reader, "description")) {
         str = xml_reader_read_inner_xml(reader);
         g_object_set(item, "summary", str, NULL);
         g_free(str);
         xml_reader_read_end_element(reader);
      } else if (xml_reader_is_a_local(reader, "title")) {
         str = xml_reader_read_string(reader);
         g_object_set(item, "title", str, NULL);
         g_free(str);
      } else if (xml_reader_is_a_local(reader, "pubDate")) {
         GDateTime *dt;
         GTimeVal tv;

         if ((str = xml_reader_read_string(reader))) {
            if (g_time_val_from_iso8601(str, &tv)) {
               if ((dt = g_date_time_new_from_timeval_utc(&tv))) {
                  g_object_set(item,
                               "last-updated-at", dt,
                               NULL);
                  g_date_time_unref(dt);
               }
            }
            g_free(str);
         }
      } else if (xml_reader_is_a_local(reader, "source")) {
         /*
          * TODO:
          */
      }
   }

   xml_reader_read_end_element(reader);

   return TRUE;
}

static gboolean
news_format_rss_read_from_xml (NewsFormatRss  *rss,
                        XmlReader      *reader,
                        NewsFeed       *feed,
                        GError        **error)
{
   GHashTable *categories = NULL;
   GHashTable *persons = NULL;
   GHashTable *links = NULL;
   GomAdapter *adapter;
   gchar *str;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_FORMAT_RSS(rss), FALSE);
   g_return_val_if_fail(XML_IS_READER(reader), FALSE);
   g_return_val_if_fail(NEWS_IS_FEED(feed), FALSE);

   adapter = news_format_get_adapter(NEWS_FORMAT(rss));
   g_assert(adapter);

skip:
   if (!xml_reader_read_to_next(reader) ||
       !xml_reader_is_a(reader, "rss")) {
      if (xml_reader_is_a(reader, "xml-stylesheet") ||
          xml_reader_is_a(reader, "#comment")) {
         goto skip;
      }
      RETURN_ERROR;
   }

   if (!xml_reader_is_a_local(reader, "rss")) {
      RETURN_ERROR;
   }

   if ((str = xml_reader_get_attribute(reader, "version"))) {
      if (!!g_strcmp0(str, "2.0")) {
         g_set_error(error, XML_READER_ERROR, XML_READER_ERROR_INVALID,
                  _("RSS Version \"%s\" is not supported."), str);
         g_free(str);
         RETURN(FALSE);
      }
      g_free(str);
   }

   if (!xml_reader_read_start_element(reader, "channel")) {
      RETURN_ERROR;
   }

   persons = g_hash_table_new_full(g_str_hash, g_str_equal,
                                   g_free, g_object_unref);
   categories = g_hash_table_new_full(g_str_hash, g_str_equal,
                                      g_free, g_object_unref);
   links = g_hash_table_new_full(g_str_hash, g_str_equal,
                                 g_free, g_object_unref);

   xml_reader_read(reader);

   for (xml_reader_read_to_next(reader);
        xml_reader_get_node_type(reader) != XML_READER_TYPE_END_ELEMENT;
        xml_reader_read_to_next(reader)) {

      if ((xml_reader_get_node_type(reader) != XML_READER_TYPE_ELEMENT) || rss->priv->error) {
         g_hash_table_destroy(categories);
         g_hash_table_destroy(persons);
         g_hash_table_destroy(links);
         RETURN_ERROR;
      }

      if (xml_reader_is_a_local(reader, "title")) {
         str = xml_reader_read_string(reader);
         g_object_set(feed, "title", str, NULL);
         g_free(str);
      } else if (xml_reader_is_a_local(reader, "link")) {
         GomCollection *collection;
         NewsLink *link_;

         g_object_get(feed, "links", &collection, NULL);
         if ((link_ = news_format_rss_read_link(rss, reader, links))) {
            gom_collection_add_resource(collection, GOM_RESOURCE(link_));
            g_object_unref(link_);
         }
         g_object_unref(collection);
      } else if (xml_reader_is_a_local(reader, "description")) {
         str = xml_reader_read_string(reader);
         g_object_set(feed, "description", str, NULL);
         g_free(str);
         xml_reader_read_end_element(reader);
      } else if (xml_reader_is_a_local(reader, "language")) {
         str = xml_reader_read_string(reader);
         g_object_set(feed, "language", str, NULL);
         g_free(str);
      } else if (xml_reader_is_a_local(reader, "copyright")) {
         str = xml_reader_read_string(reader);
         g_object_set(feed, "copyright", str, NULL);
         g_free(str);
      } else if (xml_reader_is_a_local(reader, "managingEditor")) {
         GomCollection *collection;
         NewsPerson *person;

         g_object_get(feed, "authors", &collection, NULL);
         if ((person = news_format_rss_read_person(rss, reader, persons))) {
            gom_collection_add_resource(collection, GOM_RESOURCE(person));
            g_object_unref(person);
         }
         g_object_unref(collection);
      } else if (xml_reader_is_a_local(reader, "pubDate") ||
                 xml_reader_is_a_local(reader, "lastBuildDate")) {
         GDateTime *dt;
         GTimeVal tv;

         if ((str = xml_reader_read_string(reader))) {
            if (g_time_val_from_iso8601(str, &tv)) {
               if ((dt = g_date_time_new_from_timeval_utc(&tv))) {
                  g_object_set(feed,
                            "last-updated-at", dt,
                            NULL);
                  g_date_time_unref(dt);
               }
            }
            g_free(str);
         }
      } else if (xml_reader_is_a_local(reader, "category")) {
         GomCollection *collection;
         NewsCategory *category;

         g_object_get(feed, "categories", &collection, NULL);
         if ((category = news_format_rss_read_category(rss, reader, categories))) {
            gom_collection_add_resource(collection, GOM_RESOURCE(category));
            g_object_unref(category);
         }
         g_object_unref(collection);
      } else if (xml_reader_is_a_local(reader, "generator")) {
         str = xml_reader_read_string(reader);
         g_object_set(feed, "generator", str, NULL);
         g_free(str);
      } else if (xml_reader_is_a_local(reader, "item")) {
         GomCollection *collection;
         NewsItem *item;

         g_object_get(feed, "items", &collection, NULL);
         item = g_object_new(NEWS_TYPE_ITEM,
                        "adapter", adapter,
                        "is-new", TRUE,
                        NULL);
         if (news_format_rss_read_item(rss, reader, item, persons, categories, links)) {
            gom_collection_add_resource(collection,
                                 GOM_RESOURCE(item));
         }
         g_object_unref(item);
         g_object_unref(collection);
      }
   }

   g_hash_table_destroy(categories);
   g_hash_table_destroy(links);
   g_hash_table_destroy(persons);

   RETURN(TRUE);
}

static void
news_format_rss_on_error (XmlReader     *reader,
                          const gchar   *message,
                          NewsFormatRss *rss)
{
   g_return_if_fail(NEWS_IS_FORMAT_RSS(rss));
   g_return_if_fail(XML_IS_READER(reader));

   if (!rss->priv->error) {
      g_set_error(&rss->priv->error, NEWS_FORMAT_RSS_ERROR,
                  NEWS_FORMAT_RSS_ERROR_INVALID_XML,
                  _("Invalid XML detected: %s"), message);
   }
}

static NewsFeed *
news_format_rss_read (NewsFormat    *format,
                      GInputStream  *stream,
                      GError       **error)
{
   NewsFormatRssPrivate *priv;
   NewsFormatRss *rss = (NewsFormatRss *)format;
   GomAdapter *adapter;
   NewsSource *source;
   XmlReader *reader;
   NewsFeed *feed;
   gchar *source_id;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_FORMAT_RSS(rss), NULL);

   priv = rss->priv;

   if (!(source = news_format_get_source(format))) {
      g_assert_not_reached();
      RETURN(NULL);
   }

   if (!(adapter = news_format_get_adapter(format))) {
      g_assert_not_reached();
      RETURN(NULL);
   }

   reader = xml_reader_new();
   if (!xml_reader_load_from_stream(reader, stream, error)) {
      g_object_unref(reader);
      RETURN(NULL);
   }

   g_signal_connect(reader, "error",
                    G_CALLBACK(news_format_rss_on_error),
                    format);

   g_object_get(source, "id", &source_id, NULL);
   feed = g_object_new(NEWS_TYPE_FEED,
                       "adapter", adapter,
                       "is-new", TRUE,
                       "source-id", source_id,
                       NULL);
   if (!news_format_rss_read_from_xml(rss, reader, feed, error)) {
      g_clear_object(&feed);
   } else if (priv->error) {
      g_clear_object(&feed);
      if (error) {
         *error = g_error_copy(priv->error);
      }
   }

   g_free(source_id);
   g_object_unref(reader);

   RETURN(feed);
}

/**
 * news_format_rss_finalize:
 * @object: (in): A #NewsFormatRss.
 *
 * Finalizer for a #NewsFormatRss instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_format_rss_finalize (GObject *object)
{
   NewsFormatRssPrivate *priv = NEWS_FORMAT_RSS(object)->priv;

   ENTRY;

   g_clear_error(&priv->error);

   G_OBJECT_CLASS(news_format_rss_parent_class)->finalize(object);

   EXIT;
}

/**
 * news_format_rss_class_init:
 * @klass: (in): A #NewsFormatRssClass.
 *
 * Initializes the #NewsFormatRssClass and prepares the vtable.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_format_rss_class_init (NewsFormatRssClass *klass)
{
   NewsFormatClass *format_class;
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_format_rss_finalize;
   g_type_class_add_private(object_class, sizeof(NewsFormatRssPrivate));

   format_class = NEWS_FORMAT_CLASS(klass);
   format_class->read = news_format_rss_read;
}

/**
 * news_format_rss_init:
 * @: (in): A #NewsFormatRss.
 *
 * Initializes the newly created #NewsFormatRss instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_format_rss_init (NewsFormatRss *rss)
{
   ENTRY;
   rss->priv =
      G_TYPE_INSTANCE_GET_PRIVATE(rss,
                                  NEWS_TYPE_FORMAT_RSS,
                                  NewsFormatRssPrivate);
   EXIT;
}

GQuark
news_format_rss_error_quark (void)
{
   return g_quark_from_static_string("news-format-rss-error-quark");
}
