/* news-format.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include <gom/gom.h>
#include <sys/sysinfo.h>

#include "news-debug.h"
#include "news-format.h"

struct _NewsFormatPrivate
{
   GomAdapter *adapter;
   NewsSource *source;

   NewsFeed *feed;
   GInputStream *stream;
   GAsyncReadyCallback callback;
   gpointer user_data;
   GError *error;
};

enum
{
   PROP_0,
   PROP_ADAPTER,
   PROP_SOURCE,
   LAST_PROP
};

G_DEFINE_TYPE(NewsFormat, news_format, G_TYPE_OBJECT)

static GThreadPool *gThreadPool;
static GParamSpec  *gParamSpecs[LAST_PROP];

/**
 * news_format_get_adapter:
 * @format: (in): A #NewsFormat.
 *
 * Retrieves the #GomAdapter to used for generated feed items.
 *
 * Returns: A #GomAdapter if it has been set.
 */
GomAdapter *
news_format_get_adapter (NewsFormat *format)
{
   g_return_val_if_fail(NEWS_IS_FORMAT(format), NULL);
   return format->priv->adapter;
}

/**
 * news_format_read_thread_func:
 * @data: (in): A #NewsFormat.
 * @user_data: (in): unused.
 *
 * A threadpool callback function to process a non-async format reader.
 * The item is executed,
 *
 * Returns: None.
 */
static void
news_format_read_thread_func (gpointer data,
                              gpointer user_data)
{
   GSimpleAsyncResult *simple;
   NewsFormatPrivate *priv;
   NewsFormat *format = data;

   ENTRY;

   g_return_if_fail(NEWS_IS_FORMAT(format));

   priv = format->priv;

   priv->feed = NEWS_FORMAT_GET_CLASS(format)->read(format,
                                                    priv->stream,
                                                    &priv->error);
   g_clear_object(&priv->stream);
   simple = g_simple_async_result_new(G_OBJECT(format),
                                      priv->callback,
                                      priv->user_data,
                                      news_format_read_thread_func);
   g_simple_async_result_complete_in_idle(simple);
   g_object_unref(simple);
   g_object_unref(format);

   EXIT;
}

/**
 * news_format_read_async:
 * @format: (in): A #NewsFormat.
 * @stream: (in): A #GInputStream.
 * @cancellable: (in) (allow-none): A #GCancellable or %NULL.
 * @callback: (in): A #GAsyncReadyCallback to execute when completed.
 * @user_data: (in): user data for @callback.
 *
 * Asynchronously reads the feed by parsing @stream. If the format
 * implementation is not asynchronously, the operation will be completed
 * from within a thread.
 *
 * It is the responsibility of @callback to call news_format_read_finish().
 * Not doing so is considered a programming error.
 *
 * Returns: None.
 */
void
news_format_read_async (NewsFormat          *format,
                        GInputStream        *stream,
                        GCancellable        *cancellable,
                        GAsyncReadyCallback  callback,
                        gpointer             user_data)
{
   ENTRY;

   g_return_if_fail(NEWS_IS_FORMAT(format));
   g_return_if_fail(G_IS_INPUT_STREAM(stream));
   g_return_if_fail(!cancellable || G_IS_CANCELLABLE(cancellable));
   g_return_if_fail(callback != NULL);

   g_assert(NEWS_FORMAT_GET_CLASS(format)->read_async);

   NEWS_FORMAT_GET_CLASS(format)->read_async(format, stream, cancellable,
                                             callback, user_data);

   EXIT;
}

/**
 * news_format_read_finish:
 * @format: (in): A #NewsFormat.
 * @result: (in): A #GAsyncResult.
 * @error: (out): A location for a #GError, or %NULL.
 *
 * Completes an asynchronous request to news_format_read_async(). A newly
 * created #NewsFeed will be returned if successful; otherwise %NULL will
 * be returned and @error set.
 *
 * Returns: A #NewsFeed if successful; otherwise %NULL.
 */
NewsFeed *
news_format_read_finish (NewsFormat    *format,
                         GAsyncResult  *result,
                         GError       **error)
{
   NewsFeed *feed;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_FORMAT(format), NULL);
   g_return_val_if_fail(G_IS_ASYNC_RESULT(result), NULL);

   g_assert(NEWS_FORMAT_GET_CLASS(format)->read_finish);
   feed = NEWS_FORMAT_GET_CLASS(format)->read_finish(format, result, error);
   RETURN(feed);
}

static void
news_format_real_read_async (NewsFormat          *format,
                             GInputStream        *stream,
                             GCancellable        *cancellable,
                             GAsyncReadyCallback  callback,
                             gpointer             user_data)
{
   NewsFormatPrivate *priv;

   ENTRY;

   g_assert(NEWS_IS_FORMAT(format));
   g_assert(G_IS_INPUT_STREAM(stream));
   g_assert(!cancellable || G_IS_CANCELLABLE(cancellable));
   g_assert(callback);
   g_assert(NEWS_FORMAT_GET_CLASS(format)->read);

   priv = format->priv;

   priv->stream = g_object_ref(stream);
   priv->callback = callback;
   priv->user_data = user_data;
   g_thread_pool_push(gThreadPool, g_object_ref(format), NULL);

   EXIT;
}

static NewsFeed *
news_format_real_read_finish (NewsFormat    *format,
                       GAsyncResult  *result,
                       GError       **error)
{
   NewsFormatPrivate *priv;
   NewsFeed *feed = NULL;

   g_assert(NEWS_IS_FORMAT(format));
   g_assert(G_IS_ASYNC_RESULT(result));
   g_assert(NEWS_FORMAT_GET_CLASS(format)->read);

   priv = format->priv;

   if (priv->feed) {
      feed = priv->feed;
      priv->feed = NULL;
   } else {
      if (!priv->error) {
         g_set_error(error, NEWS_FORMAT_ERROR, NEWS_FORMAT_ERROR_UNKNOWN,
                     _("An unknown error ocurred while parsing feed."));
      } else {
         g_set_error(error, priv->error->domain, priv->error->code, "%s",
                     priv->error->message);
      }
   }

   RETURN(feed);
}

/**
 * news_format_get_source:
 * @format: (in): A #NewsFormat.
 *
 * Retrieves the source for wich the format is parsing.
 *
 * Returns: A #NewsSource.
 */
NewsSource *
news_format_get_source (NewsFormat *format)
{
   g_return_val_if_fail(NEWS_IS_FORMAT(format), NULL);
   return format->priv->source;
}

GQuark
news_format_error_quark (void)
{
   return g_quark_from_static_string("news-format-error-quark");
}

/**
 * news_format_finalize:
 * @object: (in): A #NewsFormat.
 *
 * Finalizer for a #NewsFormat instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 */
static void
news_format_finalize (GObject *object)
{
   NewsFormatPrivate *priv = NEWS_FORMAT(object)->priv;

   ENTRY;

   g_clear_object(&priv->adapter);
   g_clear_object(&priv->feed);
   g_clear_object(&priv->source);
   g_clear_object(&priv->stream);
   g_clear_error(&priv->error);

   G_OBJECT_CLASS(news_format_parent_class)->finalize(object);

   EXIT;
}

/**
 * news_format_get_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (out): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Get a given #GObject property.
 */
static void
news_format_get_property (GObject    *object,
                          guint       prop_id,
                          GValue     *value,
                          GParamSpec *pspec)
{
   NewsFormat *format = NEWS_FORMAT(object);

   switch (prop_id) {
   case PROP_ADAPTER:
      g_value_set_object(value, format->priv->adapter);
      break;
   case PROP_SOURCE:
      g_value_set_object(value, format->priv->source);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_format_set_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (in): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Set a given #GObject property.
 */
static void
news_format_set_property (GObject      *object,
                          guint         prop_id,
                          const GValue *value,
                          GParamSpec   *pspec)
{
   NewsFormat *format = NEWS_FORMAT(object);

   switch (prop_id) {
   case PROP_ADAPTER:
      g_clear_object(&format->priv->adapter);
      format->priv->adapter = g_value_dup_object(value);
      g_object_notify_by_pspec(object, pspec);
      break;
   case PROP_SOURCE:
      g_clear_object(&format->priv->source);
      format->priv->source = g_value_dup_object(value);
      g_object_notify_by_pspec(object, pspec);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_format_class_init:
 * @klass: (in): A #NewsFormatClass.
 *
 * Initializes the #NewsFormatClass and prepares the vtable.
 *
 * Returns: None.
 */
static void
news_format_class_init (NewsFormatClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_format_finalize;
   object_class->get_property = news_format_get_property;
   object_class->set_property = news_format_set_property;
   g_type_class_add_private(object_class, sizeof(NewsFormatPrivate));

   klass->read_async = news_format_real_read_async;
   klass->read_finish = news_format_real_read_finish;

   gParamSpecs[PROP_ADAPTER] =
      g_param_spec_object("adapter",
                          _("Adapter"),
                          _("The Gom Adapter for storage."),
                          GOM_TYPE_ADAPTER,
                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
   g_object_class_install_property(object_class, PROP_ADAPTER,
                                   gParamSpecs[PROP_ADAPTER]);

   gParamSpecs[PROP_SOURCE] =
      g_param_spec_object("source",
                          _("Source"),
                          _("The source."),
                          NEWS_TYPE_SOURCE,
                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
   g_object_class_install_property(object_class, PROP_SOURCE,
                                   gParamSpecs[PROP_SOURCE]);

   gThreadPool = g_thread_pool_new(news_format_read_thread_func,
                                   NULL, get_nprocs(), FALSE, NULL);
}

/**
 * news_format_init:
 * @: (in): A #NewsFormat.
 *
 * Initializes the newly created #NewsFormat instance.
 *
 * Returns: None.
 */
static void
news_format_init (NewsFormat *format)
{
   ENTRY;
   format->priv =
      G_TYPE_INSTANCE_GET_PRIVATE(format,
                                  NEWS_TYPE_FORMAT,
                                  NewsFormatPrivate);
   EXIT;
}
