/* news-item-view.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include <webkit/webkit.h>

#include "news-item.h"
#include "news-item-view.h"

struct _NewsItemViewPrivate
{
   NewsItem *item;

   GtkWidget *scroller;
   GtkWidget *webview;
};

enum
{
   PROP_0,
   PROP_ITEM,
   LAST_PROP
};

G_DEFINE_TYPE(NewsItemView, news_item_view, GTK_TYPE_ALIGNMENT)

static GParamSpec *gParamSpecs[LAST_PROP];

static void
news_item_view_set_item (NewsItemView *view,
                   NewsItem     *item)
{
   NewsItemViewPrivate *priv;
   gchar *base_uri;
   gchar *content;
   gchar *html;
   gchar *summary;
   gchar *title;

   g_return_if_fail(NEWS_IS_ITEM_VIEW(view));
   g_return_if_fail(!item || NEWS_IS_ITEM(item));

   priv = view->priv;

   if (item) {
      g_object_get(item,
                "base-uri", &base_uri,
                "content", &content,
                "summary", &summary,
                "title", &title,
                NULL);
      html = g_strdup_printf("<h3>%s</h3>%s", title,
                        content ? content : summary);
      webkit_web_view_load_html_string(WEBKIT_WEB_VIEW(priv->webview),
                               html, base_uri);
      g_free(base_uri);
      g_free(content);
      g_free(html);
      g_free(summary);
      g_free(title);
   }
}

/**
 * news_item_view_finalize:
 * @object: (in): A #NewsItemView.
 *
 * Finalizer for a #NewsItemView instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_item_view_finalize (GObject *object)
{
   NewsItemViewPrivate *priv = NEWS_ITEM_VIEW(object)->priv;

   g_clear_object(&priv->item);

   G_OBJECT_CLASS(news_item_view_parent_class)->finalize(object);
}

/**
 * news_item_view_get_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (out): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Get a given #GObject property.
 */
static void
news_item_view_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
   NewsItemView *view = NEWS_ITEM_VIEW(object);

   switch (prop_id) {
   case PROP_ITEM:
      g_value_set_object(value, view->priv->item);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_item_view_set_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (in): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Set a given #GObject property.
 */
static void
news_item_view_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
   NewsItemView *view = NEWS_ITEM_VIEW(object);

   switch (prop_id) {
   case PROP_ITEM:
      news_item_view_set_item(view, g_value_get_object(value));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_item_view_class_init:
 * @klass: (in): A #NewsItemViewClass.
 *
 * Initializes the #NewsItemViewClass and prepares the vtable.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_item_view_class_init (NewsItemViewClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_item_view_finalize;
   object_class->get_property = news_item_view_get_property;
   object_class->set_property = news_item_view_set_property;
   g_type_class_add_private(object_class, sizeof(NewsItemViewPrivate));

   gParamSpecs[PROP_ITEM] =
      g_param_spec_object("item",
                          _("Item"),
                          _("The news item to view."),
                          NEWS_TYPE_ITEM,
                          G_PARAM_READWRITE);
   g_object_class_install_property(object_class, PROP_ITEM,
                                   gParamSpecs[PROP_ITEM]);
}

/**
 * news_item_view_init:
 * @: (in): A #NewsItemView.
 *
 * Initializes the newly created #NewsItemView instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_item_view_init (NewsItemView *view)
{
   NewsItemViewPrivate *priv;

   view->priv = priv =
      G_TYPE_INSTANCE_GET_PRIVATE(view,
                                  NEWS_TYPE_ITEM_VIEW,
                                  NewsItemViewPrivate);

   priv->scroller = g_object_new(GTK_TYPE_SCROLLED_WINDOW,
                          "shadow-type", GTK_SHADOW_NONE,
                          "visible", TRUE,
                          NULL);
   gtk_container_add(GTK_CONTAINER(view), priv->scroller);

   priv->webview = g_object_new(WEBKIT_TYPE_WEB_VIEW,
                         "visible", TRUE,
                         NULL);
   gtk_container_add(GTK_CONTAINER(priv->scroller), priv->webview);
}
