/* news-item.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>

#include "news-category.h"
#include "news-debug.h"
#include "news-content-kind.h"
#include "news-item.h"
#include "news-link.h"
#include "news-person.h"

G_DEFINE_TYPE(NewsItem, news_item, GOM_TYPE_RESOURCE)

static void
news_item_finalize (GObject *object)
{
   ENTRY;
   G_OBJECT_CLASS(news_item_parent_class)->finalize(object);
   EXIT;
}

/**
 * news_item_class_init:
 * @klass: (in): A #NewsItemClass.
 *
 * Initializes the #NewsItemClass and prepares the vtable.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_item_class_init (NewsItemClass *klass)
{
   GomResourceClass *resource_class;
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_item_finalize;
   object_class->get_property = gom_resource_get_property;
   object_class->set_property = gom_resource_set_property;

   resource_class = GOM_RESOURCE_CLASS(klass);
   gom_resource_class_table(resource_class, "items");
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_int64("id",
                        _("Id"),
                        _("The item identifier."),
                        0,
                        G_MAXINT64,
                        0,
                        G_PARAM_READWRITE),
         "key", TRUE,
         "serial", TRUE,
         NULL);
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_string("base-uri",
                        _("Base URI"),
                        _("The base uri of the item."),
                        NULL,
                        G_PARAM_READWRITE),
         NULL);
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_string("content",
                        _("Content"),
                        _("The content of the item."),
                        NULL,
                        G_PARAM_READWRITE),
         NULL);
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_enum("content-type",
                       _("Content Type"),
                       _("The kind of content."),
                       NEWS_TYPE_CONTENT_KIND,
                       NEWS_CONTENT_TEXT,
                       G_PARAM_READWRITE),
         NULL);
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_string("copyright",
                        _("Copyright"),
                        _("The copyright of the item."),
                        NULL,
                        G_PARAM_READWRITE),
         NULL);
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_string("item-id",
                        _("Item Id"),
                        _("The unique upstream id for the item."),
                        NULL,
                        G_PARAM_READWRITE),
         NULL);
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_boxed("last-updated-at",
                        _("Last Updated At"),
                        _("The time the item was last updated."),
                        G_TYPE_DATE_TIME,
                        G_PARAM_READWRITE),
         NULL);
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_boxed("published-at",
                        _("Published At"),
                        _("The time the item was published."),
                        G_TYPE_DATE_TIME,
                        G_PARAM_READWRITE),
         NULL);
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_boxed("read-at",
                        _("Read At"),
                        _("The time the item was read."),
                        G_TYPE_DATE_TIME,
                        G_PARAM_READWRITE),
         NULL);
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_string("summary",
                        _("Summary"),
                        _("The summary of the item."),
                        NULL,
                        G_PARAM_READWRITE),
         NULL);
   gom_resource_class_install_property(
         resource_class,
         g_param_spec_string("title",
                        _("Title"),
                        _("The title of the item."),
                        NULL,
                        G_PARAM_READWRITE),
         NULL);
   gom_resource_class_has_many(resource_class,
                               "authors",
                               _("Authors"),
                               _("The items authors."),
                               NEWS_TYPE_PERSON,
                               "many-to-many", TRUE,
                               NULL);
   gom_resource_class_has_many(resource_class,
                               "categories",
                               _("Categories"),
                               _("The items categories."),
                               NEWS_TYPE_CATEGORY,
                               "many-to-many", TRUE,
                               NULL);
   gom_resource_class_has_many(resource_class,
                               "contributors",
                               _("Contributors"),
                               _("The items contributors."),
                               NEWS_TYPE_PERSON,
                               "many-to-many", TRUE,
                               NULL);
   gom_resource_class_has_many(resource_class,
                               "links",
                               _("Links"),
                               _("The items links."),
                               NEWS_TYPE_LINK,
                               "many-to-many", TRUE,
                               NULL);
}

/**
 * news_item_init:
 * @: (in): A #NewsItem.
 *
 * Initializes the newly created #NewsItem instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_item_init (NewsItem *item)
{
   ENTRY;
   EXIT;
}
