/* news-parser.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _GNU_SOURCE

#include <glib/gi18n.h>
#include <gom/gom.h>
#include <string.h>

#include "news-debug.h"
#include "news-format.h"
#include "news-format-atom.h"
#include "news-format-rss.h"
#include "news-parser.h"
#include "news-source.h"

/**
 * SECTION:news-parser
 * @short_description: asynchronous stream parser
 * @title: News parser
 * @see_also: #NewsClient #NewsFeed
 * @include: news/news.h
 *
 * The #NewsParser is used to asynchronously parse a #GInputStream containing
 * a feed to a #NewsFeed. The parser can only be used once; after which, it
 * must be discarded.
 */

struct _NewsParserPrivate
{
   GomAdapter         *adapter;
   GCancellable       *cancellable;
   NewsFeed           *feed;
   GSimpleAsyncResult *simple;
   NewsSource         *source;
};

typedef struct
{
   NewsFormatDiscoverer func;
   gint                 priority;
} Discoverer;

enum
{
   PROP_0,
   PROP_ADAPTER,
   PROP_SOURCE,
   LAST_PROP
};

static GParamSpec *gParamSpecs[LAST_PROP];
static GArray     *gDiscoverers;

#ifndef BUFFER_FILL_SIZE
#define BUFFER_FILL_SIZE 1024
#endif

G_DEFINE_TYPE(NewsParser, news_parser, G_TYPE_OBJECT)

/**
 * news_parser_get_source:
 * @parser: (in): A #NewsParser.
 *
 * Retrieves the #NewsSource that is being parsed.
 *
 * Returns:A #NewsSource or %NULL.
 */
NewsSource *
news_parser_get_source (NewsParser *parser)
{
   g_return_val_if_fail(NEWS_IS_PARSER(parser), NULL);
   return parser->priv->source;
}

/**
 * news_parser_get_adapter:
 * @parser: (in): A #NewsParser.
 *
 * Retrieves the storage adapter to use when creating new objects.
 *
 * Returns: A #GomAdapter or %NULL.
 */
GomAdapter *
news_parser_get_adapter (NewsParser *parser)
{
   g_return_val_if_fail(NEWS_IS_PARSER(parser), NULL);
   return parser->priv->adapter;
}

/**
 * discoverer_sort:
 * @a: (in): A Discoverer.
 * @b: (in): A Discoverer.
 *
 * Compares two Discover structures by their priority.
 *
 * Returns: qsort() style return value.
 */
static gint
discoverer_sort (gconstpointer a,
                 gconstpointer b)
{
   const Discoverer *da = a;
   const Discoverer *db = b;

   return da->priority - db->priority;
}

/**
 * news_parser_register_format:
 * @parser: (in): A #NewsParser.
 * @priority: (in): The discoverer priority.
 *
 * Registers a format discoverer with the given priority. The first match
 * is used when discovering formats. Lowest priorities first.
 *
 * Returns: None.
 */
void
news_parser_register_format (NewsFormatDiscoverer discoverer,
                             gint                 priority)
{
   static gsize initialized = FALSE;
   Discoverer disc = { 0 };

   g_return_if_fail(discoverer != NULL);

   if (g_once_init_enter(&initialized)) {
      gDiscoverers = g_array_new(FALSE, FALSE, sizeof disc);
      g_once_init_leave(&initialized, TRUE);
   }

   disc.func = discoverer;
   disc.priority = priority;
   g_array_append_val(gDiscoverers, disc);
   g_array_sort(gDiscoverers, discoverer_sort);
}

/**
 * news_parser_discover_format:
 * @parser: (in): A #NewsParser.
 * @buffer: (in): A buffer to read from.
 * @length: (in): The length of @buffer.
 *
 * Tries to discover the feed format using the contents of @buffer.
 *
 * Returns: A #GType of a #NewsFormat or %G_TYPE_NONE.
 */
static GType
news_parser_discover_format (NewsParser   *parser,
                             const guint8 *buffer,
                             gsize         length)
{
   Discoverer *discover;
   GType type;
   gint i;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_PARSER(parser), G_TYPE_INVALID);
   g_return_val_if_fail(buffer != NULL, G_TYPE_INVALID);

   if (!length) {
      RETURN(G_TYPE_INVALID);
   }

   if (gDiscoverers) {
      for (i = 0; i < gDiscoverers->len; i++) {
         discover = &g_array_index(gDiscoverers, Discoverer, i);
         if ((type = discover->func(buffer, length))) {
            RETURN(type);
         }
      }
   }

   RETURN(G_TYPE_INVALID);
}

/**
 * news_parser_format_read_cb:
 * @object: (in): A #NewsFormat.
 * @result: (in): A #GAsyncResult.
 * @user_data: (in): A #NewsParser.
 *
 * Handles the asynchronous completion of a call to news_format_read_async().
 * If the reading of the buffer was successful; then the resulting feed
 * is stored for use by news_parser_parse_finish(). Otherwise, a proper
 * error is set.
 *
 * Returns: None.
 */
static void
news_parser_format_read_cb (GObject      *object,
                            GAsyncResult *result,
                            gpointer      user_data)
{
   NewsParserPrivate *priv;
   NewsParser *parser = user_data;
   NewsFormat *format = (NewsFormat *)object;
   GError *error = NULL;

   ENTRY;

   g_return_if_fail(NEWS_IS_FORMAT(format));
   g_return_if_fail(G_IS_ASYNC_RESULT(result));
   g_return_if_fail(NEWS_IS_PARSER(parser));

   priv = parser->priv;

   if (!(priv->feed = news_format_read_finish(format, result, &error))) {
      g_simple_async_result_take_error(priv->simple, error);
   }

   g_simple_async_result_complete_in_idle(priv->simple);
   g_object_unref(parser);

   EXIT;
}

/**
 * news_parser_buffered_fill_cb:
 * @object: (in): A #GBufferedInputStream.
 * @result: (in): A #GAsyncResult.
 * @user_data: (in): A #NewsParser.
 *
 * Completes an asynchronous request to fill the buffer of a
 * #GBufferedInputStream. If the buffer was successfully filled, then
 * the feed format will be discovered and the stream parsed.
 *
 * Returns: None.
 */
static void
news_parser_buffered_fill_cb (GObject      *object,
                              GAsyncResult *result,
                              gpointer      user_data)
{
   GBufferedInputStream *buffered = (GBufferedInputStream *)object;
   NewsParserPrivate *priv;
   NewsParser *parser = user_data;
   NewsFormat *format;
   guint8 buffer[BUFFER_FILL_SIZE];
   gsize length = 0;
   GType format_type = G_TYPE_INVALID;

   ENTRY;

   g_assert(G_IS_BUFFERED_INPUT_STREAM(buffered));
   g_assert(G_IS_ASYNC_RESULT(result));
   g_assert(NEWS_IS_PARSER(parser));

   priv = parser->priv;

   length = g_buffered_input_stream_peek(buffered, buffer, 0, sizeof buffer);

   if (length <= 0) {
      g_simple_async_result_set_error(priv->simple, NEWS_PARSER_ERROR,
                                      NEWS_PARSER_ERROR_NO_DATA,
                                      _("No data was available to parse."));
      g_simple_async_result_complete_in_idle(priv->simple);
   } else {
      format_type = news_parser_discover_format(parser, buffer, length);
      if (format_type) {
         format = g_object_new(format_type,
                               "adapter", priv->adapter,
                          "source", priv->source,
                               NULL);
         news_format_read_async(format, G_INPUT_STREAM(buffered),
                                priv->cancellable,
                                news_parser_format_read_cb,
                                g_object_ref(parser));
         g_object_unref(format);
      } else {
         g_simple_async_result_set_error(priv->simple, NEWS_PARSER_ERROR,
                                         NEWS_PARSER_ERROR_NO_FORMAT,
                                         _("Failed to discover format."));
         g_simple_async_result_complete_in_idle(priv->simple);
#if 0
         {
            gchar *uri;
            g_object_get(priv->source, "uri", &uri, NULL);
            g_printerr("%s\n", uri);
            g_free(uri);
         }
#endif
      }
   }

   g_object_unref(parser);

   EXIT;
}

/**
 * news_parser_parse_async:
 * @parser: (in): A #NewsParser.
 * @stream: (in): A #GInputStream.
 * @cancellable: (in): A #GCancellable to to cancel the task.
 * @callback: (in): A callback to perform when the operation completes.
 * @user_data: (in): User data for @callback.
 *
 * Requests that @parser parse the feed contained in @stream asynchronously.
 * When the operation completes, @callback will be executed from the
 * main loop. @callback is responsible for calling news_parser_parse_finish()
 * to retrieve the result.
 *
 * Returns: None.
 */
void
news_parser_parse_async (NewsParser          *parser,
                         GInputStream        *stream,
                         GCancellable        *cancellable,
                         GAsyncReadyCallback  callback,
                         gpointer             user_data)
{
   NewsParserPrivate *priv;
   GInputStream *buffered;

   ENTRY;

   g_return_if_fail(NEWS_IS_PARSER(parser));
   g_return_if_fail(G_IS_INPUT_STREAM(stream));
   g_return_if_fail(!cancellable || G_IS_CANCELLABLE(cancellable));
   g_return_if_fail(callback != NULL);
   g_return_if_fail(parser->priv->simple == NULL);

   priv = parser->priv;

   priv->simple = g_simple_async_result_new(G_OBJECT(parser),
                                            callback, user_data,
                                            news_parser_parse_async);

   if (cancellable) {
      priv->cancellable = g_object_ref(cancellable);
   }

   buffered = g_buffered_input_stream_new_sized(stream, 1024);
   g_buffered_input_stream_fill_async(G_BUFFERED_INPUT_STREAM(buffered),
                                      BUFFER_FILL_SIZE,
                                      G_PRIORITY_DEFAULT,
                                      priv->cancellable,
                                      news_parser_buffered_fill_cb,
                                      g_object_ref(parser));
   g_object_unref(buffered);

   EXIT;
}

/**
 * news_parser_parse_finish:
 * @parser: (in): A #NewsParser.
 *
 * Completes an asynchronous request to parse a #NewsFeed from a
 * #GInputStream. If the feed was successfully parsed, then a
 * #NewsFeed will be returned; othrewise %NULL and @error is set.
 *
 * Returns: None.
 */
NewsFeed *
news_parser_parse_finish (NewsParser    *parser,
                          GAsyncResult  *result,
                          GError       **error)
{
   NewsParserPrivate *priv;
   NewsFeed *feed = NULL;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_PARSER(parser), NULL);
   g_return_val_if_fail(G_IS_SIMPLE_ASYNC_RESULT(result), NULL);

   priv = parser->priv;

   if (!priv->feed) {
      g_simple_async_result_propagate_error(priv->simple, error);
   } else {
      feed = g_object_ref(priv->feed);
   }

   g_clear_object(&priv->simple);

   RETURN(feed);
}

/**
 * news_parser_discover_atom:
 * @buffer: (in): The input buffer.
 * @length: (in): The length of @buffer.
 *
 * Tries to determine if @buffer contains an atom feed.
 *
 * Returns: A #GType or %G_TYPE_INVALID.
 */
static GType
news_parser_discover_atom (const guint8 *buffer,
                           gsize         length)
{
   ENTRY;
   /*
    * TODO: Make this better.
    */
#define ATOM_NEEDLE "<feed>"
#define ATOM_NEEDLE_ALT "<feed "
#define ATOM_NEEDLE_ALT2 "<feed\n"
   if (buffer &&
       (length > 5) &&
       !strncmp((gchar *)buffer, "<?xml", 5) &&
       (memmem(buffer, length, ATOM_NEEDLE, sizeof(ATOM_NEEDLE) - 1) ||
        memmem(buffer, length, ATOM_NEEDLE_ALT, sizeof(ATOM_NEEDLE_ALT) - 1) ||
        memmem(buffer, length, ATOM_NEEDLE_ALT2, sizeof(ATOM_NEEDLE_ALT2) - 1))) {
      RETURN(NEWS_TYPE_FORMAT_ATOM);
   }
#undef ATOM_NEEDLE
#undef ATOM_NEEDLE_ALT
#undef ATOM_NEEDLE_ALT2
   RETURN(G_TYPE_INVALID);
}


/**
 * news_parser_discover_rss:
 * @buffer: (in): The input buffer.
 * @length: (in): The length of @buffer.
 *
 * Tries to determine if @buffer contains an rss feed.
 *
 * Returns: A #GType or %G_TYPE_INVALID.
 */
static GType
news_parser_discover_rss (const guint8 *buffer,
                          gsize         length)
{
   ENTRY;
   /*
    * TODO: Make this better.
    */
#define RSS_NEEDLE "<rss"
   if (buffer &&
       (length > 4) &&
       memmem(buffer, length, RSS_NEEDLE, sizeof(RSS_NEEDLE) - 1)) {
      RETURN(NEWS_TYPE_FORMAT_RSS);
   }
#undef RSS_NEEDLE
   RETURN(G_TYPE_INVALID);
}

/**
 * news_parser_error_quark:
 *
 * Retrieves the error quark for #NewsParserError.
 *
 * Returns: A #GQuark.
 */
GQuark
news_parser_error_quark (void)
{
   return g_quark_from_static_string("news-parser-error-quark");
}

/**
 * news_parser_finalize:
 * @object: (in): A #NewsParser.
 *
 * Finalizer for a #NewsParser instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 */
static void
news_parser_finalize (GObject *object)
{
   NewsParserPrivate *priv = NEWS_PARSER(object)->priv;

   ENTRY;

   g_clear_object(&priv->adapter);
   g_clear_object(&priv->cancellable);
   g_clear_object(&priv->feed);
   g_clear_object(&priv->simple);
   g_clear_object(&priv->source);

   G_OBJECT_CLASS(news_parser_parent_class)->finalize(object);

   EXIT;
}

/**
 * news_parser_get_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (out): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Get a given #GObject property.
 */
static void
news_parser_get_property (GObject    *object,
                          guint       prop_id,
                          GValue     *value,
                          GParamSpec *pspec)
{
   NewsParser *parser = NEWS_PARSER(object);

   switch (prop_id) {
   case PROP_ADAPTER:
      g_value_set_object(value, parser->priv->adapter);
      break;
   case PROP_SOURCE:
      g_value_set_object(value, parser->priv->source);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_parser_set_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (in): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Set a given #GObject property.
 */
static void
news_parser_set_property (GObject      *object,
                          guint         prop_id,
                          const GValue *value,
                          GParamSpec   *pspec)
{
   NewsParser *parser = NEWS_PARSER(object);

   switch (prop_id) {
   case PROP_ADAPTER:
      g_clear_object(&parser->priv->adapter);
      parser->priv->adapter = g_value_dup_object(value);
      g_object_notify_by_pspec(object, pspec);
      break;
   case PROP_SOURCE:
      g_clear_object(&parser->priv->source);
      parser->priv->source = g_value_dup_object(value);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_parser_class_init:
 * @klass: (in): A #NewsParserClass.
 *
 * Initializes the #NewsParserClass and prepares the vtable.
 *
 * Returns: None.
 */
static void
news_parser_class_init (NewsParserClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_parser_finalize;
   object_class->get_property = news_parser_get_property;
   object_class->set_property = news_parser_set_property;
   g_type_class_add_private(object_class, sizeof(NewsParserPrivate));

   gParamSpecs[PROP_ADAPTER] =
      g_param_spec_object("adapter",
                          _("Adapter"),
                          _("The storage adapter."),
                          GOM_TYPE_ADAPTER,
                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
   g_object_class_install_property(object_class, PROP_ADAPTER,
                                   gParamSpecs[PROP_ADAPTER]);

   gParamSpecs[PROP_SOURCE] =
      g_param_spec_object("source",
                          _("Source"),
                          _("The source which is being parsed."),
                          NEWS_TYPE_SOURCE,
                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
   g_object_class_install_property(object_class, PROP_SOURCE,
                                   gParamSpecs[PROP_SOURCE]);

   /*
    * Register built in formats.
    */
   news_parser_register_format(news_parser_discover_atom,
                               NEWS_TYPE_FORMAT_ATOM);
   news_parser_register_format(news_parser_discover_rss,
                               NEWS_TYPE_FORMAT_RSS);
}

/**
 * news_parser_init:
 * @: (in): A #NewsParser.
 *
 * Initializes the newly created #NewsParser instance.
 *
 * Returns: None.
 */
static void
news_parser_init (NewsParser *parser)
{
   ENTRY;
   parser->priv =
      G_TYPE_INSTANCE_GET_PRIVATE(parser,
                                  NEWS_TYPE_PARSER,
                                  NewsParserPrivate);
   EXIT;
}
