/* news-source-file.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>

#include "news-source-file.h"

struct _NewsSourceFilePrivate
{
   GFile *file;
};

enum
{
   PROP_0,
   PROP_FILE,
   PROP_URI,
   LAST_PROP
};

G_DEFINE_TYPE(NewsSourceFile, news_source_file, NEWS_TYPE_SOURCE)

static GParamSpec *gParamSpecs[LAST_PROP];

static void
news_source_file_fetch_cb (GObject      *object,
                           GAsyncResult *result,
                           gpointer      user_data)
{
   GSimpleAsyncResult *simple = user_data;
   GFileInputStream *stream;
   GError *error = NULL;
   GFile *file = (GFile *)object;

   g_return_if_fail(G_IS_FILE(file));
   g_return_if_fail(G_IS_SIMPLE_ASYNC_RESULT(simple));

   if (!(stream = g_file_read_finish(file, result, &error))) {
      g_simple_async_result_take_error(simple, error);
   }

   g_simple_async_result_set_op_res_gpointer(simple, stream, NULL);
   g_simple_async_result_complete_in_idle(simple);
   g_object_unref(simple);
}

static void
news_source_file_fetch_async (NewsSource          *source,
                              GCancellable        *cancellable,
                              GAsyncReadyCallback  callback,
                              gpointer             user_data)
{
   NewsSourceFilePrivate *priv;
   GSimpleAsyncResult *simple;
   NewsSourceFile *file = (NewsSourceFile *)source;

   g_return_if_fail(NEWS_IS_SOURCE_FILE(file));
   g_return_if_fail(!cancellable || G_IS_CANCELLABLE(cancellable));
   g_return_if_fail(callback != NULL);

   priv = file->priv;

   simple = g_simple_async_result_new(G_OBJECT(source), callback, user_data,
                                      news_source_file_fetch_async);
   if (!priv->file) {
      g_simple_async_result_complete_in_idle(simple);
      g_object_unref(simple);
      return;
   }

   g_file_read_async(priv->file, G_PRIORITY_DEFAULT, cancellable,
                     news_source_file_fetch_cb, simple);
}

static GInputStream *
news_source_file_fetch_finish (NewsSource    *source,
                               GAsyncResult  *result,
                               GError       **error)
{
   GSimpleAsyncResult *simple;
   GInputStream *stream;

   g_return_val_if_fail(NEWS_IS_SOURCE_FILE(source), NULL);
   g_return_val_if_fail(G_IS_SIMPLE_ASYNC_RESULT(result), NULL);

   simple = G_SIMPLE_ASYNC_RESULT(result);

   if (!(stream = g_simple_async_result_get_op_res_gpointer(simple))) {
      g_simple_async_result_propagate_error(simple, error);
   }

   return stream;
}

/**
 * news_source_file_finalize:
 * @object: (in): A #NewsSourceFile.
 *
 * Finalizer for a #NewsSourceFile instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 */
static void
news_source_file_finalize (GObject *object)
{
   NewsSourceFilePrivate *priv = NEWS_SOURCE_FILE(object)->priv;

   g_clear_object(&priv->file);

   G_OBJECT_CLASS(news_source_file_parent_class)->finalize(object);
}

/**
 * news_source_file_get_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (out): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Get a given #GObject property.
 */
static void
news_source_file_get_property (GObject    *object,
                               guint       prop_id,
                               GValue     *value,
                               GParamSpec *pspec)
{
   NewsSourceFile *file = NEWS_SOURCE_FILE(object);

   switch (prop_id) {
   case PROP_FILE:
      g_value_set_object(value, file->priv->file);
      break;
   case PROP_URI:
      if (file->priv->file) {
         g_value_take_string(value, g_file_get_uri(file->priv->file));
      }
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_source_file_set_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (in): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Set a given #GObject property.
 */
static void
news_source_file_set_property (GObject      *object,
                               guint         prop_id,
                               const GValue *value,
                               GParamSpec   *pspec)
{
   NewsSourceFile *file = NEWS_SOURCE_FILE(object);

   switch (prop_id) {
   case PROP_FILE:
      g_clear_object(&file->priv->file);
      file->priv->file = g_value_dup_object(value);
      g_object_notify_by_pspec(object, pspec);
      break;
   case PROP_URI: {
      GFile *gfile;

      if ((gfile = g_file_new_for_uri(g_value_get_string(value)))) {
         g_object_set(object, "file", gfile, NULL);
         g_object_unref(gfile);
      }
      break;
   }
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_source_file_class_init:
 * @klass: (in): A #NewsSourceFileClass.
 *
 * Initializes the #NewsSourceFileClass and prepares the vtable.
 *
 * Returns: None.
 */
static void
news_source_file_class_init (NewsSourceFileClass *klass)
{
   GObjectClass *object_class;
   NewsSourceClass *source_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_source_file_finalize;
   object_class->get_property = news_source_file_get_property;
   object_class->set_property = news_source_file_set_property;
   g_type_class_add_private(object_class, sizeof(NewsSourceFilePrivate));

   source_class = NEWS_SOURCE_CLASS(klass);
   source_class->fetch_async = news_source_file_fetch_async;
   source_class->fetch_finish = news_source_file_fetch_finish;

   gParamSpecs[PROP_FILE] =
      g_param_spec_object("file",
                          _("File"),
                          _("The file to retrieve."),
                          G_TYPE_FILE,
                          G_PARAM_READWRITE);
   g_object_class_install_property(object_class, PROP_FILE,
                                   gParamSpecs[PROP_FILE]);

   gParamSpecs[PROP_URI] =
      g_param_spec_string("uri",
                          _("URI"),
                          _("The URI of the feed."),
                          NULL,
                          G_PARAM_READWRITE);
   g_object_class_install_property(object_class, PROP_URI,
                                   gParamSpecs[PROP_URI]);
}

/**
 * news_source_file_init:
 * @: (in): A #NewsSourceFile.
 *
 * Initializes the newly created #NewsSourceFile instance.
 *
 * Returns: None.
 */
static void
news_source_file_init (NewsSourceFile *file)
{
   file->priv =
      G_TYPE_INSTANCE_GET_PRIVATE(file,
                                  NEWS_TYPE_SOURCE_FILE,
                                  NewsSourceFilePrivate);
}
