/* news-source-soup.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include <libsoup/soup.h>
#include <string.h>

#include "news-debug.h"
#include "news-source-soup.h"

G_DEFINE_TYPE(NewsSourceSoup, news_source_soup, NEWS_TYPE_SOURCE)

struct _NewsSourceSoupPrivate
{
   SoupSession *session;
   gchar *uri;
};

enum
{
   PROP_0,
   PROP_URI,
   LAST_PROP
};

static GParamSpec *gParamSpecs[LAST_PROP];

/*
 * TODO: We need to make this use chunked reading so that we don't copy
 *       the buffer (nor need it fully in memory ever).
 */

static void
message_cb (SoupSession *session,
            SoupMessage *message,
            gpointer     user_data)
{
   GSimpleAsyncResult *simple = (GSimpleAsyncResult *)user_data;
   GInputStream *memory;
   gpointer data;
   gsize len;

   ENTRY;

   g_assert(SOUP_IS_SESSION(session));
   g_assert(SOUP_IS_MESSAGE(message));
   g_assert(G_IS_SIMPLE_ASYNC_RESULT(simple));

   if (SOUP_STATUS_IS_SUCCESSFUL(message->status_code)) {
      len = message->response_body->length;
      data = g_malloc(len);
      memcpy(data, message->response_body->data, len);
      memory = g_memory_input_stream_new_from_data(data, len, g_free);
      g_simple_async_result_set_op_res_gpointer(simple, memory, NULL);
   }

   g_simple_async_result_complete_in_idle(simple);
   g_object_unref(simple);

   EXIT;
}

static void
news_source_soup_fetch_async (NewsSource          *source,
                              GCancellable        *cancellable,
                              GAsyncReadyCallback  callback,
                              gpointer             user_data)
{
   NewsSourceSoup *soup = (NewsSourceSoup *)source;
   GSimpleAsyncResult *simple;
   NewsSourceSoupPrivate *priv;
   SoupMessage *message;

   ENTRY;

   g_return_if_fail(NEWS_IS_SOURCE_SOUP(soup));

   priv = soup->priv;

   simple = g_simple_async_result_new(G_OBJECT(source), callback, user_data,
                                      news_source_soup_fetch_async);
   message = soup_message_new("GET", priv->uri);
   soup_session_queue_message(priv->session, message, message_cb, simple);

   EXIT;
}

static GInputStream *
news_source_soup_fetch_finish (NewsSource    *source,
                               GAsyncResult  *result,
                               GError       **error)
{
   NewsSourceSoup *soup = (NewsSourceSoup *)source;
   GSimpleAsyncResult *simple = (GSimpleAsyncResult *)result;
   GInputStream *stream;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_SOURCE_SOUP(soup), FALSE);
   g_return_val_if_fail(G_IS_SIMPLE_ASYNC_RESULT(simple), FALSE);

   stream = g_simple_async_result_get_op_res_gpointer(simple);
   g_assert(!stream || G_IS_INPUT_STREAM(stream));

   if (!stream) {
      g_simple_async_result_propagate_error(simple, error);
   }

   RETURN(stream);
}

static void
news_source_soup_set_uri (NewsSourceSoup *soup,
                          const gchar    *uri)
{
   NewsSourceSoupPrivate *priv;

   g_return_if_fail(NEWS_IS_SOURCE_SOUP(soup));

   priv = soup->priv;

   g_free(priv->uri);
   priv->uri = g_strdup(uri);
   g_object_notify_by_pspec(G_OBJECT(soup), gParamSpecs[PROP_URI]);
}

static void
news_source_soup_finalize (GObject *object)
{
   NewsSourceSoupPrivate *priv = NEWS_SOURCE_SOUP(object)->priv;

   ENTRY;

   g_free(priv->uri);
   g_clear_object(&priv->session);

   G_OBJECT_CLASS(news_source_soup_parent_class)->finalize(object);

   EXIT;
}

static void
news_source_soup_get_property (GObject    *object,
                               guint       prop_id,
                               GValue     *value,
                               GParamSpec *pspec)
{
   NewsSourceSoup *soup = NEWS_SOURCE_SOUP(object);

   switch (prop_id) {
   case PROP_URI:
      g_value_set_string(value, soup->priv->uri);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

static void
news_source_soup_set_property (GObject      *object,
                               guint         prop_id,
                               const GValue *value,
                               GParamSpec   *pspec)
{
   NewsSourceSoup *soup = NEWS_SOURCE_SOUP(object);

   switch (prop_id) {
   case PROP_URI:
      news_source_soup_set_uri(soup, g_value_get_string(value));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

static void
news_source_soup_class_init (NewsSourceSoupClass *klass)
{
   GObjectClass *object_class;
   NewsSourceClass *source_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_source_soup_finalize;
   object_class->get_property = news_source_soup_get_property;
   object_class->set_property = news_source_soup_set_property;
   g_type_class_add_private(object_class, sizeof(NewsSourceSoupPrivate));

   source_class = NEWS_SOURCE_CLASS(klass);
   source_class->fetch_async = news_source_soup_fetch_async;
   source_class->fetch_finish = news_source_soup_fetch_finish;

   gParamSpecs[PROP_URI] =
      g_param_spec_string("uri",
                          _("URI"),
                          _("The URI of the source."),
                          NULL,
                          G_PARAM_READWRITE);
   g_object_class_install_property(object_class, PROP_URI,
                                   gParamSpecs[PROP_URI]);
}

static void
news_source_soup_init (NewsSourceSoup *soup)
{
   soup->priv =
      G_TYPE_INSTANCE_GET_PRIVATE(soup,
                                  NEWS_TYPE_SOURCE_SOUP,
                                  NewsSourceSoupPrivate);
   soup->priv->session = soup_session_async_new();
}
