/* news-source.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <gdk/gdk.h>
#include <glib/gi18n.h>

#include "news-debug.h"
#include "news-source.h"

struct _NewsSourcePrivate
{
   gchar *id;
   GList *errors;
};

enum
{
   PROP_0,
   PROP_HAS_ERROR,
   PROP_ID,
   LAST_PROP
};

enum
{
   REFRESH,
   REFRESHED,
   LAST_SIGNAL
};

static GParamSpec *gParamSpecs[LAST_PROP];
static guint       gSignals[LAST_SIGNAL];

G_DEFINE_TYPE(NewsSource, news_source, G_TYPE_INITIALLY_UNOWNED)

/**
 * news_source_get_id:
 * @source: (in): A #NewsSource.
 *
 * Retrieves the unique id for a particular source.
 *
 * Returns: A string which should not be modified or freed.
 */
const gchar *
news_source_get_id (NewsSource *source)
{
   g_return_val_if_fail(NEWS_IS_SOURCE(source), NULL);
   return source->priv->id;
}

/**
 * news_source_add_error:
 * @source: (in): A #NewsSource.
 * @error: (in): A #GError.
 *
 * Adds a #GError to the list of errors on the source. These errors
 * can be used to display any problems that may have occurred to the
 * user.
 *
 * Returns: None.
 */
void
news_source_add_error (NewsSource   *source,
                       const GError *error)
{
   NewsSourcePrivate *priv;

   g_return_if_fail(NEWS_IS_SOURCE(source));
   g_return_if_fail(error != NULL);

   priv = source->priv;

   priv->errors = g_list_append(priv->errors, g_error_copy(error));
   g_object_notify_by_pspec(G_OBJECT(source), gParamSpecs[PROP_HAS_ERROR]);
}

/**
 * news_source_take_error:
 * @source: (in): A #NewsSource.
 * @error: (in): A #GError.
 *
 * Like news_source_add_error() except that @error is not copied. This is
 * handy when the caller doesn't need to manipulate the #GError for any
 * reason. Allowing to save an additional copy of the error and free.
 *
 * Returns: None.
 */
void
news_source_take_error (NewsSource *source,
                        GError     *error)
{
   NewsSourcePrivate *priv;

   g_return_if_fail(NEWS_IS_SOURCE(source));
   g_return_if_fail(error != NULL);

   priv = source->priv;

   priv->errors = g_list_append(priv->errors, error);
   g_object_notify_by_pspec(G_OBJECT(source), gParamSpecs[PROP_HAS_ERROR]);
}

/**
 * news_source_clear_errors:
 * @source: (in): A #NewsSource.
 *
 * Clears the errors associated with a #NewsSource.
 *
 * Returns: None.
 */
void
news_source_clear_errors (NewsSource *source)
{
   NewsSourcePrivate *priv;
   GList *errors;

   g_return_if_fail(NEWS_IS_SOURCE(source));

   priv = source->priv;

   errors = priv->errors;
   priv->errors = NULL;

   g_list_foreach(errors, (GFunc)g_error_free, NULL);
   g_list_free(errors);

   g_object_notify_by_pspec(G_OBJECT(source), gParamSpecs[PROP_HAS_ERROR]);
}

/**
 * news_source_get_errors:
 * @source: (in): A #NewsSource.
 *
 * Retrieves a newly allocated list containing a private copy of the #GError
 * contained in the source. To free the list of errors, make sure to free
 * the #GError as well.
 *
 * [[
 * g_list_foreach(errors, (GFunc)g_error_free, NULL);
 * g_list_free(errors);
 * ]]
 *
 * Returns: A newly allocated list containing newly allocated #GError.
 */
GList *
news_source_get_errors (NewsSource *source)
{
   NewsSourcePrivate *priv;
   GList *errors = NULL;
   GList *iter;

   g_return_val_if_fail(NEWS_IS_SOURCE(source), NULL);

   priv = source->priv;

   errors = g_list_copy(priv->errors);
   for (iter = errors; iter; iter = iter->next) {
      iter->data = g_error_copy(iter->data);
   }
   return errors;
}

/**
 * news_source_get_has_error:
 * @source: (in): A #NewsSource.
 *
 * Checks to see if the source has errors. The #GList containing the
 * individual errors can be retrieved with news_source_get_errors().
 *
 * Returns: %TRUE if the source has errors; otherwise %FALSE.
 */
static gboolean
news_source_get_has_error (NewsSource *source)
{
   g_return_val_if_fail(NEWS_IS_SOURCE(source), FALSE);
   return !!source->priv->errors;
}

/**
 * news_source_refresh:
 * @source: (in): A #NewsSource.
 *
 * Emits the refresh signal. The owning NewsClient will refresh the
 * source.
 *
 * Returns: None.
 * Side effects: None.
 */
void
news_source_refresh (NewsSource *source)
{
   g_return_if_fail(NEWS_IS_SOURCE(source));
   g_signal_emit(source, gSignals[REFRESH], 0);
}

static gboolean
emit_refreshed (gpointer data)
{
   g_signal_emit(data, gSignals[REFRESHED], 0);
   g_object_unref(data);
   return FALSE;
}

/**
 * news_source_emit_refreshed:
 * @source: (in): A #NewsSource.
 *
 * Emits the "refreshed" signal for the #NewsSource.
 *
 * Returns: None.
 * Side effects: None.
 */
void
news_source_emit_refreshed (NewsSource *source)
{
   GMainContext *context;

   g_return_if_fail(NEWS_IS_SOURCE(source));

   context = g_main_context_default();
   if (!g_main_context_acquire(context)) {
      g_timeout_add(0, emit_refreshed, g_object_ref(source));
      return;
   }
   g_signal_emit(source, gSignals[REFRESHED], 0);
   g_main_context_release(context);
}

/**
 * news_source_fetch_async:
 * @source: (in): A #NewsSource.
 * @cancellable: (in): A #GCancellable.
 * @callback: (in): A #GAsyncReadyCallback to execute upon completion.
 * @user_data: (in): User data for @callback.
 *
 * Asynchronously requests that @source fetch the updated contents of the
 * feed. Once the content stream has been acquired, @callback will be
 * executed. It is the responsibility of @callback to execute
 * news_source_fetch_finish(). Not doing so is consered a programming error.
 *
 * Returns: None.
 */
void
news_source_fetch_async (NewsSource          *source,
                         GCancellable        *cancellable,
                         GAsyncReadyCallback  callback,
                         gpointer             user_data)
{
   ENTRY;

   g_return_if_fail(NEWS_IS_SOURCE(source));
   g_return_if_fail(!cancellable || G_IS_CANCELLABLE(cancellable));
   g_return_if_fail(callback != NULL);

   if (NEWS_SOURCE_GET_CLASS(source)->fetch_async) {
      NEWS_SOURCE_GET_CLASS(source)->
         fetch_async(source, cancellable, callback, user_data);
   }

   EXIT;
}

/**
 * news_source_fetch_finish:
 * @source: (in): A #NewsSource.
 * @result: (in): A #GAsyncResult.
 * @error: (out): A location for a #GError, or %NULL.
 *
 * Handles the completion of an asynchronous request to fetch a sources
 * updated content.
 *
 * Returns: A #GInputStream if successful; otherwise %NULL and @error is set.
 */
GInputStream *
news_source_fetch_finish (NewsSource    *source,
                          GAsyncResult  *result,
                          GError       **error)
{
   GInputStream *ret = NULL;

   ENTRY;

   g_return_val_if_fail(NEWS_IS_SOURCE(source), NULL);
   g_return_val_if_fail(G_IS_ASYNC_RESULT(result), NULL);

   if (NEWS_SOURCE_GET_CLASS(source)->fetch_finish) {
      ret = NEWS_SOURCE_GET_CLASS(source)->
         fetch_finish(source, result, error);
   }

   RETURN(ret);
}

/**
 * news_source_finalize:
 * @object: (in): A #NewsSource.
 *
 * Finalizer for a #NewsSource instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 */
static void
news_source_finalize (GObject *object)
{
   NewsSourcePrivate *priv = NEWS_SOURCE(object)->priv;

   ENTRY;

   g_free(priv->id);

   g_list_foreach(priv->errors, (GFunc)g_error_free, NULL);
   g_list_free(priv->errors);

   G_OBJECT_CLASS(news_source_parent_class)->finalize(object);

   EXIT;
}

/**
 * news_source_get_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (out): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Get a given #GObject property.
 */
static void
news_source_get_property (GObject    *object,
                          guint       prop_id,
                          GValue     *value,
                          GParamSpec *pspec)
{
   NewsSource *source = NEWS_SOURCE(object);

   switch (prop_id) {
   case PROP_ID:
      g_value_set_string(value, source->priv->id);
      break;
   case PROP_HAS_ERROR:
      g_value_set_boolean(value, news_source_get_has_error(source));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_source_set_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (in): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Set a given #GObject property.
 */
static void
news_source_set_property (GObject      *object,
                          guint         prop_id,
                          const GValue *value,
                          GParamSpec   *pspec)
{
   NewsSource *source = NEWS_SOURCE(object);

   switch (prop_id) {
   case PROP_ID:
      g_free(source->priv->id);
      source->priv->id = g_value_dup_string(value);
      g_object_notify_by_pspec(object, pspec);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_source_class_init:
 * @klass: (in): A #NewsSourceClass.
 *
 * Initializes the #NewsSourceClass and prepares the vtable.
 *
 * Returns: None.
 */
static void
news_source_class_init (NewsSourceClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_source_finalize;
   object_class->get_property = news_source_get_property;
   object_class->set_property = news_source_set_property;
   g_type_class_add_private(object_class, sizeof(NewsSourcePrivate));

   gParamSpecs[PROP_HAS_ERROR] =
      g_param_spec_boolean("has-error",
                           _("Has Error"),
                           _("If the source has errored."),
                           FALSE,
                           G_PARAM_READABLE);
   g_object_class_install_property(object_class, PROP_HAS_ERROR,
                                   gParamSpecs[PROP_HAS_ERROR]);

   gParamSpecs[PROP_ID] =
      g_param_spec_string("id",
                          _("ID"),
                          _("The source identifier."),
                          NULL,
                          G_PARAM_READWRITE);
   g_object_class_install_property(object_class, PROP_ID,
                                   gParamSpecs[PROP_ID]);

   gSignals[REFRESH] = g_signal_new("refresh",
                                    NEWS_TYPE_SOURCE,
                                    G_SIGNAL_RUN_LAST,
                                    0,
                                    NULL,
                                    NULL,
                                    g_cclosure_marshal_VOID__VOID,
                                    G_TYPE_NONE,
                                    0);

   gSignals[REFRESHED] = g_signal_new("refreshed",
                                      NEWS_TYPE_SOURCE,
                                      G_SIGNAL_RUN_LAST,
                                      G_STRUCT_OFFSET(NewsSourceClass, refreshed),
                                      NULL,
                                      NULL,
                                      g_cclosure_marshal_VOID__VOID,
                                      G_TYPE_NONE,
                                      0);
}

/**
 * news_source_init:
 * @: (in): A #NewsSource.
 *
 * Initializes the newly created #NewsSource instance.
 *
 * Returns: None.
 */
static void
news_source_init (NewsSource *source)
{
   source->priv =
      G_TYPE_INSTANCE_GET_PRIVATE(source,
                                  NEWS_TYPE_SOURCE,
                                  NewsSourcePrivate);
}
