/* news-status-view.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>

#include "news-status-view.h"

G_DEFINE_TYPE(NewsStatusView, news_status_view, NEWS_TYPE_ANIM_BIN)

struct _NewsStatusViewPrivate
{
   GtkWidget *image;
   GtkWidget *progress;
   GtkWidget *subtitle;
   GtkWidget *title;
};

enum
{
   PROP_0,
   PROP_FRACTION,
   PROP_ICON_NAME,
   PROP_SUBTITLE,
   PROP_TITLE,
   LAST_PROP
};

enum
{
   CANCEL,
   LAST_SIGNAL
};

static GParamSpec *gParamSpecs[LAST_PROP];
static guint       gSignals[LAST_SIGNAL];

static void
news_status_view_cancel_clicked (NewsStatusView *view,
                                 GtkWidget      *button)
{
   g_return_if_fail(NEWS_IS_STATUS_VIEW(view));
   g_signal_emit(view, gSignals[CANCEL], 0);
}

/**
 * news_status_view_finalize:
 * @object: (in): A #NewsStatusView.
 *
 * Finalizer for a #NewsStatusView instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_status_view_finalize (GObject *object)
{
   G_OBJECT_CLASS(news_status_view_parent_class)->finalize(object);
}

/**
 * news_status_view_set_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (in): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Set a given #GObject property.
 */
static void
news_status_view_set_property (GObject      *object,
                               guint         prop_id,
                               const GValue *value,
                               GParamSpec   *pspec)
{
   NewsStatusView *view = NEWS_STATUS_VIEW(object);

   switch (prop_id) {
   case PROP_FRACTION: {
      gdouble fraction = g_value_get_double(value);
      gchar *text = g_markup_printf_escaped(_("%d%%"), (gint)(fraction * 100));
      g_object_set(view->priv->progress,
                   "fraction", g_value_get_double(value),
                   "show-text", TRUE,
                   "text", text,
                   NULL);
      g_free(text);
      break;
   }
   case PROP_ICON_NAME: {
      const gchar *icon_name = g_value_get_string(value);
      g_object_set(view->priv->image,
                   "icon-name", icon_name,
                   "visible", !!icon_name,
                   NULL);
      break;
   }
   case PROP_SUBTITLE:
      g_object_set_property(G_OBJECT(view->priv->subtitle),
                            "label", value);
      break;
   case PROP_TITLE:
      g_object_set_property(G_OBJECT(view->priv->title),
                            "label", value);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_status_view_class_init:
 * @klass: (in): A #NewsStatusViewClass.
 *
 * Initializes the #NewsStatusViewClass and prepares the vtable.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_status_view_class_init (NewsStatusViewClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_status_view_finalize;
   object_class->set_property = news_status_view_set_property;
   g_type_class_add_private(object_class, sizeof(NewsStatusViewPrivate));

   gParamSpecs[PROP_ICON_NAME] =
      g_param_spec_string("icon-name",
                          _("Icon Name"),
                          _("The icon to display."),
                          NULL,
                          G_PARAM_WRITABLE);
   g_object_class_install_property(object_class, PROP_ICON_NAME,
                                   gParamSpecs[PROP_ICON_NAME]);

   gParamSpecs[PROP_TITLE] =
      g_param_spec_string("title",
                          _("Title"),
                          _("The status title."),
                          NULL,
                          G_PARAM_WRITABLE);
   g_object_class_install_property(object_class, PROP_TITLE,
                                   gParamSpecs[PROP_TITLE]);

   gParamSpecs[PROP_SUBTITLE] =
      g_param_spec_string("subtitle",
                          _("Subtitle"),
                          _("The status subtitle."),
                          NULL,
                          G_PARAM_WRITABLE);
   g_object_class_install_property(object_class, PROP_SUBTITLE,
                                   gParamSpecs[PROP_SUBTITLE]);

   gParamSpecs[PROP_FRACTION] =
      g_param_spec_double("fraction",
                          _("Fraction"),
                          _("The fraction complete."),
                          0.0,
                     1.0,
                     0.0,
                          G_PARAM_WRITABLE);
   g_object_class_install_property(object_class, PROP_FRACTION,
                                   gParamSpecs[PROP_FRACTION]);

   gSignals[CANCEL] = g_signal_new("cancel",
                                   NEWS_TYPE_STATUS_VIEW,
                           G_SIGNAL_RUN_FIRST,
                           0,
                           NULL,
                           NULL,
                           g_cclosure_marshal_VOID__VOID,
                           G_TYPE_NONE,
                           0);
}

/**
 * news_status_view_init:
 * @: (in): A #NewsStatusView.
 *
 * Initializes the newly created #NewsStatusView instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_status_view_init (NewsStatusView *view)
{
   NewsStatusViewPrivate *priv;
   PangoAttrList *attrs;
   PangoAttribute *attr;
   GtkWidget *cancel;
   GtkWidget *hbox;
   GtkWidget *image;
   GtkWidget *vbox;

   view->priv = priv =
      G_TYPE_INSTANCE_GET_PRIVATE(view,
                                  NEWS_TYPE_STATUS_VIEW,
                                  NewsStatusViewPrivate);

   vbox = g_object_new(GTK_TYPE_VBOX,
                       "visible", TRUE,
                  "border-width", 3,
                       NULL);
   gtk_container_add(GTK_CONTAINER(view), vbox);

   attrs = pango_attr_list_new();
   attr = pango_attr_scale_new(PANGO_SCALE_SMALL);
   pango_attr_list_insert(attrs, attr);
   attr = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
   pango_attr_list_insert(attrs, attr);
   priv->title = g_object_new(GTK_TYPE_LABEL,
                        "attributes", attrs,
                              "label", "XXX",
                              "visible", TRUE,
                              "xalign", 0.0f,
                              NULL);
   gtk_container_add_with_properties(GTK_CONTAINER(vbox), priv->title,
                                     "expand", FALSE,
                                     NULL);

   attrs = pango_attr_list_new();
   attr = pango_attr_scale_new(PANGO_SCALE_SMALL);
   pango_attr_list_insert(attrs, attr);
   priv->subtitle = g_object_new(GTK_TYPE_LABEL,
                          "attributes", attrs,
                                 "label", "XXX",
                                 "visible", TRUE,
                                 "xalign", 0.0f,
                                 NULL);
   gtk_container_add_with_properties(GTK_CONTAINER(vbox), priv->subtitle,
                                     "expand", FALSE,
                                     NULL);

   hbox = g_object_new(GTK_TYPE_HBOX,
                       "spacing", 3,
                       "visible", TRUE,
                  NULL);
   gtk_container_add_with_properties(GTK_CONTAINER(vbox), hbox,
                                     "expand", FALSE,
                                     NULL);

   priv->image = g_object_new(GTK_TYPE_IMAGE,
                              "icon-size", GTK_ICON_SIZE_MENU,
                              "visible", FALSE,
                              "xpad", 6,
                              NULL);
   gtk_container_add_with_properties(GTK_CONTAINER(hbox), priv->image,
                                     "expand", FALSE,
                                     NULL);

   priv->progress = g_object_new(GTK_TYPE_PROGRESS_BAR,
                                 "visible", TRUE,
                                 NULL);
   gtk_container_add_with_properties(GTK_CONTAINER(hbox), priv->progress,
                                     "expand", TRUE,
                                     NULL);

   image = g_object_new(GTK_TYPE_IMAGE,
                        "icon-name", GTK_STOCK_CANCEL,
                        "icon-size", GTK_ICON_SIZE_MENU,
                        "visible", TRUE,
                   NULL);
   cancel = g_object_new(GTK_TYPE_BUTTON,
                         "child", image,
                         "relief", GTK_RELIEF_NONE,
                         "tooltip-text", _("Cancel current tasks."),
                         "visible", TRUE,
                         NULL);
   g_signal_connect_swapped(cancel, "clicked",
                            G_CALLBACK(news_status_view_cancel_clicked),
                            view);
   gtk_container_add_with_properties(GTK_CONTAINER(hbox), cancel,
                                     "expand", FALSE,
                                     NULL);
}
