/* news-sync-task.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include <gom/gom.h>

#include "news-category.h"
#include "news-debug.h"
#include "news-feed.h"
#include "news-item.h"
#include "news-link.h"
#include "news-parser.h"
#include "news-person.h"
#include "news-source.h"
#include "news-sync-task.h"

G_DEFINE_TYPE(NewsSyncTask, news_sync_task, G_TYPE_OBJECT)

struct _NewsSyncTaskPrivate
{
   GomAdapter   *adapter;
   GCancellable *cancellable;
   NewsFeed     *feed;
   gdouble       fraction;
   NewsSource   *source;
   guint         saved;
   guint         total;
};

enum
{
   PROP_0,
   PROP_ADAPTER,
   PROP_FEED,
   PROP_FRACTION,
   PROP_SOURCE,
   LAST_PROP
};

static GParamSpec *gParamSpecs[LAST_PROP];

static void
news_sync_task_reduce_persons (NewsSyncTask    *task,
                               GomCollection *persons)
{
   GomCondition *cond = NULL;
   GomCondition *cond1 = NULL;
   GomCondition *cond2 = NULL;
   GomResource *resource;
   GomProperty *email_prop = NULL;
   GomProperty *name_prop = NULL;
   NewsPerson *person;
   GPtrArray *array;
   gchar *name;
   gchar *email;
   gint i;

   g_assert(NEWS_IS_SYNC_TASK(task));
   g_assert(GOM_IS_COLLECTION(persons));

   if ((array = gom_collection_get_additions(persons))) {
      for (i = 0; i < array->len; i++) {
         resource = g_ptr_array_index(array, i);

         if (!email_prop) {
            email_prop = gom_resource_find_property(resource, "email");
            name_prop = gom_resource_find_property(resource, "name");
         }

         g_object_get(resource,
                      "email", &email,
                      "name", &name,
                      NULL);

         if (email) {
            cond1 = gom_condition_equal(email_prop, email);
         }
         if (name) {
            cond2 = gom_condition_equal(name_prop, name);
         }

         if (cond1 && cond2) {
            cond = gom_condition_or(cond1, cond2);
            gom_condition_unref(cond1);
            gom_condition_unref(cond2);
         } else if (cond1) {
            cond = cond1;
         } else if (cond2) {
            cond = cond2;
         }

         if (cond) {
            person = gom_resource_find_first(NEWS_TYPE_PERSON,
                                             task->priv->adapter,
                                             cond, NULL);
            if (person) {
               gom_resource_merge(resource, GOM_RESOURCE(person));
               g_object_set(resource, "is-new", FALSE, NULL);
               g_clear_object(&person);
            }
            gom_condition_unref(cond);
         }

         g_free(email);
         g_free(name);
      }
   }
}

static void
news_sync_task_reduce_links (NewsSyncTask    *task,
                             GomCollection *links)
{
   GomCondition *cond = NULL;
   GomResource *resource;
   GomProperty *href_prop = NULL;
   GPtrArray *array;
   NewsLink *link_;
   gchar *href;
   gint i;

   g_assert(NEWS_IS_SYNC_TASK(task));
   g_assert(GOM_IS_COLLECTION(links));

   if ((array = gom_collection_get_additions(links))) {
      for (i = 0; i < array->len; i++) {
         resource = g_ptr_array_index(array, i);

         if (!href_prop) {
            href_prop = gom_resource_find_property(resource, "href");
            g_assert(href_prop);
         }

         g_object_get(resource, "href", &href, NULL);

         if (href) {
            cond = gom_condition_equal(href_prop, href);
         }

         if (cond) {
            link_ = gom_resource_find_first(NEWS_TYPE_LINK,
                                            task->priv->adapter,
                                            cond, NULL);
            if (link_) {
               gom_resource_merge(resource, GOM_RESOURCE(link_));
               g_object_set(resource, "is-new", FALSE, NULL);
               g_clear_object(&link_);
            }
            gom_condition_unref(cond);
         }

         g_free(href);
      }
   }
}

static void
news_sync_task_reduce_categories (NewsSyncTask    *task,
                                  GomCollection *categories)
{
   NewsCategory *category;
   GomCondition *cond = NULL;
   GomResource *resource;
   GomProperty *name_prop = NULL;
   GPtrArray *array;
   gchar *name;
   gint i;

   g_assert(NEWS_IS_SYNC_TASK(task));
   g_assert(GOM_IS_COLLECTION(categories));

   if ((array = gom_collection_get_additions(categories))) {
      for (i = 0; i < array->len; i++) {
         resource = g_ptr_array_index(array, i);

         if (!name_prop) {
            name_prop = gom_resource_find_property(resource, "name");
            g_assert(name_prop);
         }

         g_object_get(resource, "name", &name, NULL);

         if (name) {
            cond = gom_condition_equal(name_prop, name);
         }

         if (cond) {
            category = gom_resource_find_first(NEWS_TYPE_CATEGORY,
                                               task->priv->adapter,
                                               cond, NULL);
            if (category) {
               gom_resource_merge(resource, GOM_RESOURCE(category));
               g_object_set(resource, "is-new", FALSE, NULL);
               g_clear_object(&category);
            }
            gom_condition_unref(cond);
         }

         g_free(name);
      }
   }
}

static void
news_sync_task_reduce_items (NewsSyncTask    *task,
                             GomCollection *items,
                             guint         *count)
{
   GomCondition *cond;
   GomResource *resource;
   GomProperty *item_id_prop = NULL;
   GPtrArray *array;
   NewsItem *item;
   gchar *item_id;
   gint i;

   ENTRY;

   g_assert(NEWS_IS_SYNC_TASK(task));
   g_assert(GOM_IS_COLLECTION(items));
   g_assert(count != NULL);

   if ((array = gom_collection_get_additions(items))) {
      *count += array->len;
      for (i = 0; i < array->len; i++) {
         resource = g_ptr_array_index(array, i);

         if (!item_id_prop) {
            item_id_prop = gom_resource_find_property(resource, "item-id");
            g_assert(item_id_prop);
         }

         g_object_get(resource, "item-id", &item_id, NULL);

         if (item_id) {
            cond = gom_condition_equal(item_id_prop, item_id);
            item = gom_resource_find_first(NEWS_TYPE_ITEM,
                                           task->priv->adapter,
                                           cond, NULL);
            if (item) {
               gom_resource_merge(resource, GOM_RESOURCE(item));
               g_object_set(resource, "is-new", FALSE, NULL);
               g_clear_object(&item);
            }
            gom_condition_unref(cond);
            g_free(item_id);
         }

         {
            //GomCollection *authors;
            //GomCollection *categories;
            //GomCollection *contributors;
            GomCollection *links;

            g_object_get(resource,
                         //"authors", &authors,
                         //"categories", &categories,
                         //"contributors", &contributors,
                         "links", &links,
                         NULL);

            //g_assert(authors);
            //g_assert(categories);
            //g_assert(contributors);
            g_assert(links);

            //news_sync_task_reduce_persons(task, authors, count);
            //news_sync_task_reduce_persons(task, contributors, count);
            //news_sync_task_reduce_categories(task, categories, count);
            news_sync_task_reduce_links(task, links);

            //g_clear_object(&authors);
            //g_clear_object(&categories);
            //g_clear_object(&contributors);
            g_clear_object(&links);
         }
      }
   }

   EXIT;
}

static void
news_sync_task_reduce (NewsSyncTask *task)
{
   GomCollection *authors;
   GomCollection *categories;
   GomCollection *contributors;
   GomCollection *items;
   GomCollection *links;
   guint count = 0;

   g_assert(NEWS_IS_SYNC_TASK(task));
   g_assert(NEWS_IS_FEED(task->priv->feed));

   g_object_get(task->priv->feed,
                "authors", &authors,
                "categories", &categories,
                "contributors", &contributors,
             "items", &items,
                "links", &links,
                NULL);

   news_sync_task_reduce_persons(task, authors);
   news_sync_task_reduce_persons(task, contributors);
   news_sync_task_reduce_links(task, links);
   news_sync_task_reduce_categories(task, categories);
   news_sync_task_reduce_items(task, items, &count);

   g_clear_object(&authors);
   g_clear_object(&categories);
   g_clear_object(&contributors);
   g_clear_object(&items);
   g_clear_object(&links);
}

void
news_sync_task_cancel (NewsSyncTask *task)
{
   g_return_if_fail(NEWS_IS_SYNC_TASK(task));

   if (task->priv->cancellable) {
      g_cancellable_cancel(task->priv->cancellable);
   }

   task->priv->saved = 0;
   task->priv->total = 0;
   task->priv->fraction = 1.0;
   g_object_notify_by_pspec(G_OBJECT(task), gParamSpecs[PROP_FRACTION]);
}

static void
news_sync_task_resource_saved (GomResource  *resource,
                               NewsSyncTask *task)
{
   NewsSyncTaskPrivate *priv;
   gdouble fraction;

   g_return_if_fail(GOM_IS_RESOURCE(resource));
   g_return_if_fail(NEWS_IS_SYNC_TASK(task));

   priv = task->priv;

   priv->saved++;

   if (priv->total && (priv->saved != priv->total)) {
      fraction = priv->saved / (gdouble)priv->total;
      fraction *= 0.333;
      priv->fraction = 0.666 + fraction;
   } else {
      priv->fraction = 1.0;
   }

   g_object_notify_by_pspec(G_OBJECT(task), gParamSpecs[PROP_FRACTION]);
}

static void
news_sync_task_prepare_save (NewsSyncTask *task)
{
   NewsSyncTaskPrivate *priv;
   GomCollection *items;
   GPtrArray *additions;
   gint i;

   g_assert(NEWS_IS_SYNC_TASK(task));
   g_assert(task->priv->feed);

   priv = task->priv;

   /*
    * FIXME: Re-enable once this is threaded.
    */
   if (0)
   news_sync_task_reduce(task);

   g_object_get(priv->feed, "items", &items, NULL);
   additions = gom_collection_get_additions(items);
   priv->total = 1;

   if (additions) {
      priv->total += additions->len;
      for (i = 0; i < additions->len; i++) {
         g_signal_connect(g_ptr_array_index(additions, i), "saved",
                          G_CALLBACK(news_sync_task_resource_saved),
                          task);
      }
   }

   g_signal_connect(priv->feed, "saved",
                    G_CALLBACK(news_sync_task_resource_saved),
                    task);

   g_object_unref(items);
}

static void
news_sync_task_save_cb (GObject      *object,
                        GAsyncResult *result,
                        gpointer      user_data)
{
   NewsSyncTaskPrivate *priv;
   GSimpleAsyncResult *simple = user_data;
   NewsSyncTask *task;
   GomResource *resource = (GomResource *)object;
   GError *error = NULL;

   g_assert(NEWS_IS_FEED(resource));
   g_assert(G_IS_SIMPLE_ASYNC_RESULT(simple));
   task = (NewsSyncTask *)g_async_result_get_source_object(G_ASYNC_RESULT(simple));
   g_assert(NEWS_IS_SYNC_TASK(task));

   priv = task->priv;

   if (!gom_resource_save_finish(resource, result, &error)) {
      g_simple_async_result_take_error(simple, error);
      gom_adapter_sqlite_rollback(GOM_ADAPTER_SQLITE(priv->adapter));
   } else if (!gom_adapter_sqlite_commit(GOM_ADAPTER_SQLITE(priv->adapter),
                                         &error)) {
      g_simple_async_result_take_error(simple, error);
      gom_adapter_sqlite_rollback(GOM_ADAPTER_SQLITE(priv->adapter));
   }

   g_simple_async_result_complete_in_idle(simple);
   g_object_unref(simple);
}

static void
news_sync_task_parse_cb (GObject      *object,
                         GAsyncResult *result,
                         gpointer      user_data)
{
   NewsSyncTaskPrivate *priv;
   GSimpleAsyncResult *simple = user_data;
   NewsSyncTask *task;
   NewsParser *parser = (NewsParser *)object;
   GError *error = NULL;

   g_assert(NEWS_IS_PARSER(parser));
   g_assert(G_IS_SIMPLE_ASYNC_RESULT(simple));
   task = (NewsSyncTask *)g_async_result_get_source_object(G_ASYNC_RESULT(simple));
   g_assert(NEWS_IS_SYNC_TASK(task));

   priv = task->priv;

   priv->fraction = 0.666;
   g_object_notify_by_pspec(G_OBJECT(task), gParamSpecs[PROP_FRACTION]);

   if (!(priv->feed = news_parser_parse_finish(parser, result, &error))) {
      g_simple_async_result_take_error(simple, error);
      g_simple_async_result_complete_in_idle(simple);
      g_object_unref(simple);
      return;
   }

   if (g_cancellable_is_cancelled(priv->cancellable)) {
      g_simple_async_result_set_error(simple, NEWS_SYNC_TASK_ERROR,
                                      NEWS_SYNC_TASK_ERROR_CANCELLED,
                                      _("The task was cancelled."));
      g_simple_async_result_complete_in_idle(simple);
      g_object_unref(simple);
      return;
   }

   news_sync_task_prepare_save(task);

   gom_adapter_sqlite_begin(GOM_ADAPTER_SQLITE(priv->adapter));
   gom_resource_save_async(GOM_RESOURCE(priv->feed), priv->cancellable,
                           news_sync_task_save_cb, simple);
}

static void
news_sync_task_fetch_cb (GObject      *object,
                         GAsyncResult *result,
                         gpointer      user_data)
{
   NewsSyncTaskPrivate *priv;
   GSimpleAsyncResult *simple = (GSimpleAsyncResult *)user_data;
   GInputStream *stream;
   NewsSyncTask *task;
   NewsSource *source = (NewsSource *)object;
   NewsParser *parser;
   GError *error = NULL;

   g_assert(NEWS_IS_SOURCE(source));
   g_assert(G_IS_SIMPLE_ASYNC_RESULT(simple));
   task = (NewsSyncTask *)g_async_result_get_source_object(G_ASYNC_RESULT(simple));
   g_assert(NEWS_IS_SYNC_TASK(task));

   priv = task->priv;

   priv->fraction = 0.333;
   g_object_notify_by_pspec(G_OBJECT(task), gParamSpecs[PROP_FRACTION]);

   if (!(stream = news_source_fetch_finish(source, result, &error))) {
      g_simple_async_result_take_error(simple, error);
      g_simple_async_result_complete_in_idle(simple);
      g_object_unref(simple);
      return;
   }

   if (g_cancellable_is_cancelled(priv->cancellable)) {
      g_simple_async_result_set_error(simple, NEWS_SYNC_TASK_ERROR,
                                      NEWS_SYNC_TASK_ERROR_CANCELLED,
                                      _("The task was cancelled."));
      g_simple_async_result_complete_in_idle(simple);
      g_object_unref(simple);
      return;
   }

   parser = g_object_new(NEWS_TYPE_PARSER,
                         "adapter", priv->adapter,
                         "source", priv->source,
                         NULL);
   news_parser_parse_async(parser, stream, priv->cancellable,
                           news_sync_task_parse_cb, simple);
   g_object_unref(parser);
}

void
news_sync_task_run_async (NewsSyncTask        *task,
                          GCancellable        *cancellable,
                          GAsyncReadyCallback  callback,
                          gpointer             user_data)
{
   NewsSyncTaskPrivate *priv;
   GSimpleAsyncResult *simple;

   g_return_if_fail(NEWS_IS_SYNC_TASK(task));
   g_return_if_fail(!cancellable || G_IS_CANCELLABLE(cancellable));
   g_return_if_fail(callback != NULL);
   g_return_if_fail(task->priv->feed == NULL);

   priv = task->priv;

   simple = g_simple_async_result_new(G_OBJECT(task), callback, user_data,
                                      news_sync_task_run_async);

   if (!priv->adapter || !priv->source) {
      g_simple_async_result_set_error(simple, NEWS_SYNC_TASK_ERROR,
                                      NEWS_SYNC_TASK_ERROR_INVALID,
                                      _("Invalid task state"));
      g_simple_async_result_complete_in_idle(simple);
      g_object_unref(simple);
      return;
   }

   priv->cancellable = cancellable ? g_object_ref(cancellable)
                                   : g_cancellable_new();
   priv->fraction = 0.1;
   news_source_fetch_async(priv->source, cancellable,
                           news_sync_task_fetch_cb, simple);
}

gboolean
news_sync_task_run_finish (NewsSyncTask  *task,
                           GAsyncResult  *result,
                           GError       **error)
{
   NewsSyncTaskPrivate *priv;
   GSimpleAsyncResult *simple = (GSimpleAsyncResult *)result;
   gboolean ret;

   g_return_val_if_fail(NEWS_IS_SYNC_TASK(task), FALSE);
   g_return_val_if_fail(G_IS_SIMPLE_ASYNC_RESULT(simple), FALSE);

   priv = task->priv;

   if (!(ret = !!priv->feed)) {
      g_simple_async_result_propagate_error(simple, error);
   }

   return ret;
}

/**
 * news_sync_task_finalize:
 * @object: (in): A #NewsSyncTask.
 *
 * Finalizer for a #NewsSyncTask instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_sync_task_finalize (GObject *object)
{
   NewsSyncTaskPrivate *priv = NEWS_SYNC_TASK(object)->priv;

   g_clear_object(&priv->adapter);
   g_clear_object(&priv->cancellable);
   g_clear_object(&priv->feed);
   g_clear_object(&priv->source);

   G_OBJECT_CLASS(news_sync_task_parent_class)->finalize(object);
}

/**
 * news_sync_task_get_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (out): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Get a given #GObject property.
 */
static void
news_sync_task_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
   NewsSyncTask *task = NEWS_SYNC_TASK(object);

   switch (prop_id) {
   case PROP_ADAPTER:
      g_value_set_object(value, task->priv->adapter);
      break;
   case PROP_FEED:
      g_value_set_object(value, task->priv->feed);
      break;
   case PROP_FRACTION:
      g_value_set_double(value, task->priv->fraction);
      break;
   case PROP_SOURCE:
      g_value_set_object(value, task->priv->source);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_sync_task_set_property:
 * @object: (in): A #GObject.
 * @prop_id: (in): The property identifier.
 * @value: (in): The given property.
 * @pspec: (in): A #ParamSpec.
 *
 * Set a given #GObject property.
 */
static void
news_sync_task_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
   NewsSyncTask *task = NEWS_SYNC_TASK(object);

   switch (prop_id) {
   case PROP_ADAPTER:
      g_clear_object(&task->priv->adapter);
      task->priv->adapter = g_value_dup_object(value);
      break;
   case PROP_FEED:
      g_clear_object(&task->priv->feed);
      task->priv->feed = g_value_dup_object(value);
      break;
   case PROP_SOURCE:
      g_clear_object(&task->priv->source);
      task->priv->source = g_value_dup_object(value);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

/**
 * news_sync_task_class_init:
 * @klass: (in): A #NewsSyncTaskClass.
 *
 * Initializes the #NewsSyncTaskClass and prepares the vtable.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_sync_task_class_init (NewsSyncTaskClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_sync_task_finalize;
   object_class->get_property = news_sync_task_get_property;
   object_class->set_property = news_sync_task_set_property;
   g_type_class_add_private(object_class, sizeof(NewsSyncTaskPrivate));

   gParamSpecs[PROP_ADAPTER] =
      g_param_spec_object("adapter",
                          _("Adapter"),
                          _("Storage adapter."),
                          GOM_TYPE_ADAPTER,
                          G_PARAM_READWRITE);
   g_object_class_install_property(object_class, PROP_ADAPTER,
                                   gParamSpecs[PROP_ADAPTER]);

   gParamSpecs[PROP_FEED] =
      g_param_spec_object("feed",
                          _("Feed"),
                          _("The generated feed."),
                          NEWS_TYPE_FEED,
                          G_PARAM_READWRITE);
   g_object_class_install_property(object_class, PROP_FEED,
                                   gParamSpecs[PROP_FEED]);

   gParamSpecs[PROP_SOURCE] =
      g_param_spec_object("source",
                          _("Source"),
                          _("The source to fetch."),
                          NEWS_TYPE_SOURCE,
                          G_PARAM_READWRITE);
   g_object_class_install_property(object_class, PROP_SOURCE,
                                   gParamSpecs[PROP_SOURCE]);

   gParamSpecs[PROP_FRACTION] =
      g_param_spec_double("fraction",
                          _("Fraction"),
                          _("The fraction completed of the task."),
                          0.0,
                          1.0,
                          0.0,
                          G_PARAM_READABLE);
   g_object_class_install_property(object_class, PROP_FRACTION,
                                   gParamSpecs[PROP_FRACTION]);
}

/**
 * news_sync_task_init:
 * @: (in): A #NewsSyncTask.
 *
 * Initializes the newly created #NewsSyncTask instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_sync_task_init (NewsSyncTask *task)
{
   task->priv =
      G_TYPE_INSTANCE_GET_PRIVATE(task,
                                  NEWS_TYPE_SYNC_TASK,
                                  NewsSyncTaskPrivate);
}

GQuark
news_sync_task_error_quark (void)
{
   return g_quark_from_static_string("news-sync-task-error-quark");
}
