/* news-window.c
 *
 * Copyright (C) 2011 Christian Hergert <chris@dronelabs.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>

#include "news-application.h"
#include "news-client.h"
#include "news-item.h"
#include "news-item-view.h"
#include "news-source.h"
#include "news-source-soup.h"
#include "news-status.h"
#include "news-status-view.h"
#include "news-window.h"
#include "news-window-ui.h"

struct _NewsWindowPrivate
{
   GtkWidget *menubar;
   GtkWidget *toolbar;
   GtkWidget *container;
   GtkWidget *hpaned;
   GtkWidget *item_scroller;
   GtkWidget *item_treeview;
   GtkWidget *item_view;
   GtkWidget *src_scroller;
   GtkWidget *src_treeview;
   GtkWidget *src_vbox;
   GtkWidget *status;
   GtkWidget *status_box;
   GtkWidget *vpaned;

   GtkListStore *items;
   GtkListStore *sources;
};

typedef enum
{
   NEWS_WINDOW_NORMAL,
} NewsWindowLayout;

G_DEFINE_TYPE(NewsWindow, news_window, GTK_TYPE_WINDOW)

#if 0
static void
news_window_reload_items(NewsWindow *window)
{
   NewsWindowPrivate *priv;
   GomCollection *items;
   GtkTreeIter iter;
   GomAdapter *adapter;
   NewsItem *item;
   GError *error = NULL;
   gint64 length;
   gint64 i;

   g_return_if_fail(NEWS_IS_WINDOW(window));

   priv = window->priv;

   adapter = news_application_get_adapter(NEWS_APPLICATION_DEFAULT);
   items = gom_resource_find(NEWS_TYPE_ITEM, adapter, NULL, NULL);
   g_assert_no_error(error);
   g_assert(items);

   g_object_set(priv->item_treeview, "model", NULL, NULL);
   gtk_list_store_clear(priv->items);
   length = MIN(100, gom_collection_count(items));
   for (i = 0; i < length; i++) {
      item = gom_collection_get_nth(items, i);
      gtk_list_store_append(priv->items, &iter);
      gtk_list_store_set(priv->items, &iter, 0, item, -1);
      g_object_unref(item);
   }
   g_object_set(priv->item_treeview, "model", priv->items, NULL);
}
#endif

static void
news_window_source_refreshed (NewsWindow *window,
                       NewsSource *source)
{
   GError *error;
   GList *errors;
   GList *iter;

   //g_debug("Refreshed source");
   //news_window_reload_items(window);

   if ((errors = news_source_get_errors(source))) {
      for (iter = errors; iter; iter = iter->next) {
         error = iter->data;
         g_critical("%s", error->message);
         g_error_free(error);
      }
      g_list_free(errors);
   }
}

static void
news_window_source_added (NewsWindow *window,
                    NewsSource *source,
                    NewsClient *client)
{
   NewsWindowPrivate *priv;
   GtkTreeIter iter;

   g_return_if_fail(NEWS_IS_WINDOW(window));
   g_return_if_fail(NEWS_IS_SOURCE(source));
   g_return_if_fail(NEWS_IS_CLIENT(client));

   priv = window->priv;

   gtk_list_store_append(priv->sources, &iter);
   gtk_list_store_set(priv->sources, &iter, 0, source, -1);

   g_signal_connect_swapped(source, "refreshed",
                        G_CALLBACK(news_window_source_refreshed),
                        window);
}

static void
news_window_unparent (GtkWidget *widget)
{
   GtkWidget *parent;

   if ((parent = gtk_widget_get_parent(widget))) {
      gtk_container_remove(GTK_CONTAINER(parent), widget);
   }
}

static void
news_window_set_layout(NewsWindow       *window,
                  NewsWindowLayout  layout)
{
   NewsWindowPrivate *priv;

   g_return_if_fail(NEWS_IS_WINDOW(window));

   priv = window->priv;

   g_object_ref(priv->hpaned);
   g_object_ref(priv->vpaned);
   g_object_ref(priv->src_vbox);
   g_object_ref(priv->item_scroller);
   g_object_ref(priv->item_view);

   news_window_unparent(priv->src_vbox);
   news_window_unparent(priv->item_scroller);
   news_window_unparent(priv->item_view);
   news_window_unparent(priv->hpaned);
   news_window_unparent(priv->vpaned);

   switch (layout) {
   case NEWS_WINDOW_NORMAL:
      gtk_container_add_with_properties(GTK_CONTAINER(priv->hpaned),
                                        priv->src_vbox,
                                        "shrink", FALSE,
                                        NULL);
      gtk_container_add(GTK_CONTAINER(priv->hpaned),
                    priv->vpaned);
      gtk_container_add(GTK_CONTAINER(priv->vpaned),
                    priv->item_scroller);
      gtk_container_add(GTK_CONTAINER(priv->vpaned),
                    priv->item_view);
      gtk_container_add(GTK_CONTAINER(priv->container),
                    priv->hpaned);
      g_object_set(priv->hpaned, "position", 250, NULL);
      g_object_set(priv->vpaned, "position", 200, NULL);
      break;
   default:
      g_assert_not_reached();
   }

   g_object_unref(priv->hpaned);
   g_object_unref(priv->vpaned);
   g_object_unref(priv->src_vbox);
   g_object_unref(priv->item_scroller);
   g_object_unref(priv->item_view);
}

static void
item_title_func (GtkTreeViewColumn *column,
             GtkCellRenderer   *cell,
             GtkTreeModel      *model,
             GtkTreeIter       *iter,
             gpointer           user_data)
{
   NewsItem *item;
   GValue value = { 0 };

   gtk_tree_model_get(model, iter, 0, &item, -1);

   if (item) {
      g_value_init(&value, G_TYPE_STRING);
      g_object_get_property(G_OBJECT(item), "title", &value);
      g_object_set_property(G_OBJECT(cell), "text", &value);
      g_value_unset(&value);
      g_object_unref(item);
   }
}

static void
source_title_func (GtkTreeViewColumn *column,
               GtkCellRenderer   *cell,
               GtkTreeModel      *model,
               GtkTreeIter       *iter,
               gpointer           user_data)
{
   NewsItem *source;
   GValue value = { 0 };

   gtk_tree_model_get(model, iter, 0, &source, -1);

   if (source) {
      g_value_init(&value, G_TYPE_STRING);
      g_object_get_property(G_OBJECT(source), "id", &value);
      g_object_set_property(G_OBJECT(cell), "text", &value);
      g_value_unset(&value);
      g_object_unref(source);
   }
}

static void
news_window_item_selected (NewsWindow       *window,
                     GtkTreeSelection *selection)
{
   NewsWindowPrivate *priv;
   GtkTreeModel *model;
   GtkTreeIter iter;
   NewsItem *item;

   g_return_if_fail(NEWS_IS_WINDOW(window));
   g_return_if_fail(GTK_IS_TREE_SELECTION(selection));

   priv = window->priv;

   if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
      gtk_tree_model_get(model, &iter, 0, &item, -1);
      if (item) {
         g_object_set(priv->item_view, "item", item, NULL);
      }
   }
}

static void
text_received_cb (GtkClipboard *clipboard,
              const gchar *text,
              gpointer user_data)
{
   GtkEntry *entry = user_data;
   gchar *copy;

   g_return_if_fail(GTK_IS_ENTRY(entry));

   if (text && (!strncmp(text, "http:", 5) || !strncmp(text, "https:", 6))) {
      copy = g_strdup(text);
      g_strdelimit(copy, "\n", '\0');
      gtk_entry_set_text(entry, copy);
      g_free(copy);
   }

   g_object_unref(entry);
}

static void
news_window_add_activate (GtkAction *action,
                    NewsWindow *window)
{
   GtkClipboard *clipboard;
   const gchar *text;
   NewsClient *client;
   NewsSource *source;
   GtkWidget *dialog;
   GtkWidget *parent;
   GtkWidget *entry;

   g_return_if_fail(GTK_IS_ACTION(action));
   g_return_if_fail(NEWS_IS_WINDOW(window));

   dialog = g_object_new(GTK_TYPE_MESSAGE_DIALOG,
                    "buttons", GTK_BUTTONS_OK_CANCEL,
                    "message-type", GTK_MESSAGE_QUESTION,
                    "secondary-text", _("Please enter the URL of the "
                                   "subscription."),
                    "text", _("Add Subscription"),
                    "transient-for", window,
                    NULL);
   entry = g_object_new(GTK_TYPE_ENTRY,
                   "activates-default", TRUE,
                   "can-default", TRUE,
                   "text", "http://",
                   "visible", TRUE,
                   "width-request", 250,
                   NULL);
   gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
   clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
   gtk_clipboard_request_text(clipboard, text_received_cb,
                        g_object_ref(entry));
   parent = gtk_message_dialog_get_message_area(GTK_MESSAGE_DIALOG(dialog));
   gtk_container_add_with_properties(GTK_CONTAINER(parent), entry,
                                     "expand", FALSE,
                                     NULL);
   if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
      text = gtk_entry_get_text(GTK_ENTRY(entry));
      source = g_object_new(NEWS_TYPE_SOURCE_SOUP,
                       "id", text,
                       "uri", text,
                       NULL);
      client = news_application_get_client(NEWS_APPLICATION_DEFAULT);
      news_client_add_source(client, source);
      news_source_refresh(source);

      {
         gchar *path;
         GError *error = NULL;

         path = g_build_filename(g_get_user_config_dir(),
                                 "simply-news", "sources.xml", NULL);
         if (!news_client_save_to_file(client, path, &error)) {
            g_critical("%s", error->message);
            g_error_free(error);
         }
         g_free(path);
      }
   }
   gtk_widget_destroy(dialog);
}

static gboolean
show_status (gpointer data)
{
   news_client_refresh_all(data);
   return FALSE;
}

static void
news_window_refresh_activate (GtkAction  *action,
                       NewsWindow *window)
{
   NewsClient *client;

   g_return_if_fail(GTK_IS_ACTION(action));
   g_return_if_fail(NEWS_IS_WINDOW(window));

   gtk_widget_show(window->priv->status);
   client = news_application_get_client(NEWS_APPLICATION_DEFAULT);
   g_timeout_add(500, show_status, client);
}

static gboolean
hide_widget (gpointer data)
{
   gtk_widget_hide(data);
   return FALSE;
}

static void
status_changed (NewsClient *client,
            GParamSpec *pspec,
            GtkWidget  *view)
{
   NewsStatus *status;
   gchar *str;

   g_return_if_fail(NEWS_IS_CLIENT(client));
   g_return_if_fail(NEWS_IS_STATUS_VIEW(view));

   g_object_get(client, "status", &status, NULL);
   g_object_set(view, "fraction", status->fraction, NULL);
   str = g_strdup_printf(_("Downloading %lu of %lu subscriptions."),
                         status->completed, status->total);
   g_object_set(view,
                "subtitle", str,
                "visible", TRUE,
                NULL);
   g_free(str);

   if (status->fraction >= 1.0) {
      g_timeout_add_seconds(2, hide_widget, g_object_ref(view));
   }

   news_status_free(status);
}

static void
status_cancel (GtkWidget  *status,
               NewsWindow *window)
{
   NewsClient *client;

   g_return_if_fail(NEWS_IS_WINDOW(window));

   client = news_application_get_client(NEWS_APPLICATION_DEFAULT);
   news_client_cancel_tasks(client);
}

/**
 * news_window_finalize:
 * @object: (in): A #NewsWindow.
 *
 * Finalizer for a #NewsWindow instance.  Frees any resources held by
 * the instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_window_finalize (GObject *object)
{
   NewsWindowPrivate *priv = NEWS_WINDOW(object)->priv;

   g_clear_object(&priv->items);
   g_clear_object(&priv->sources);

   G_OBJECT_CLASS(news_window_parent_class)->finalize(object);
}

/**
 * news_window_class_init:
 * @klass: (in): A #NewsWindowClass.
 *
 * Initializes the #NewsWindowClass and prepares the vtable.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_window_class_init (NewsWindowClass *klass)
{
   GObjectClass *object_class;

   object_class = G_OBJECT_CLASS(klass);
   object_class->finalize = news_window_finalize;
   g_type_class_add_private(object_class, sizeof(NewsWindowPrivate));
}

/**
 * news_window_init:
 * @: (in): A #NewsWindow.
 *
 * Initializes the newly created #NewsWindow instance.
 *
 * Returns: None.
 * Side effects: None.
 */
static void
news_window_init (NewsWindow *window)
{
   NewsWindowPrivate *priv;
   GtkTreeViewColumn *column;
   GtkTreeSelection *sel;
   GtkCellRenderer *cell;
   GtkActionGroup *group;
   GtkUIManager *ui;
   NewsClient *client;
   GtkWidget *vbox;
   GError *error = NULL;
   GList *list;

   static const GtkActionEntry menu_actions[] = {
      { "file", GTK_STOCK_FILE },
      { "add", GTK_STOCK_ADD, N_("Add Subscription"), "<Control>n", NULL,
        G_CALLBACK(news_window_add_activate) },
      { "refresh", "mail-send-receive", N_("Get News"), "F5",
        N_("Retrieve your news subscriptions."),
        G_CALLBACK(news_window_refresh_activate) },
      { "edit", GTK_STOCK_EDIT },
      { "help", GTK_STOCK_HELP },
   };

   window->priv = priv =
      G_TYPE_INSTANCE_GET_PRIVATE(window,
                                  NEWS_TYPE_WINDOW,
                                  NewsWindowPrivate);

   g_object_set(window,
             "title", _("Simply News"),
             "default-width", 640,
             "default-height", 480,
             NULL);

   ui = gtk_ui_manager_new();
   group = news_application_get_actions(NEWS_APPLICATION_DEFAULT);
   gtk_ui_manager_insert_action_group(ui, group, 0);
   group = gtk_action_group_new("menu-actions");
   gtk_action_group_add_actions(group, menu_actions,
                                G_N_ELEMENTS(menu_actions),
                                window);
   g_object_set(gtk_action_group_get_action(group, "refresh"),
                "is-important", TRUE,
                NULL);
   gtk_ui_manager_insert_action_group(ui, group, 0);
   gtk_ui_manager_add_ui_from_string(ui, gWindowUi, -1, &error);
   g_assert_no_error(error);

   priv->menubar = gtk_ui_manager_get_widget(ui, "/menubar");
   priv->toolbar = gtk_ui_manager_get_widget(ui, "/toolbar");
   gtk_widget_hide(priv->menubar);
   g_assert(priv->menubar);
   g_assert(priv->toolbar);

   vbox = g_object_new(GTK_TYPE_VBOX,
                  "visible", TRUE,
                  NULL);
   gtk_container_add(GTK_CONTAINER(window), vbox);
   gtk_container_add_with_properties(GTK_CONTAINER(vbox), priv->menubar,
                                     "expand", FALSE,
                                     NULL);

   gtk_style_context_add_class(gtk_widget_get_style_context(priv->toolbar),
                        GTK_STYLE_CLASS_PRIMARY_TOOLBAR);
   gtk_container_add_with_properties(GTK_CONTAINER(vbox), priv->toolbar,
                                     "expand", FALSE,
                                     NULL);

   priv->container = g_object_new(GTK_TYPE_ALIGNMENT,
                           "visible", TRUE,
                           NULL);
   gtk_container_add(GTK_CONTAINER(vbox), priv->container);

   priv->hpaned = g_object_new(GTK_TYPE_HPANED,
                        "visible", TRUE,
                        NULL);
   priv->vpaned = g_object_new(GTK_TYPE_VPANED,
                        "visible", TRUE,
                        NULL);

   priv->src_vbox = g_object_new(GTK_TYPE_VBOX,
                                 "visible", TRUE,
                          NULL);
   priv->src_scroller = g_object_new(GTK_TYPE_SCROLLED_WINDOW,
                             "shadow-type", GTK_SHADOW_NONE,
                             "visible", TRUE,
                             NULL);
   priv->src_treeview = g_object_new(GTK_TYPE_TREE_VIEW,
                             "headers-visible", FALSE,
                             "visible", TRUE,
                             NULL);
   gtk_container_add(GTK_CONTAINER(priv->src_vbox), priv->src_scroller);
   gtk_container_add(GTK_CONTAINER(priv->src_scroller),
                 priv->src_treeview);

   column = g_object_new(GTK_TYPE_TREE_VIEW_COLUMN,
                    "title", _("Title"),
                    NULL);
   cell = g_object_new(GTK_TYPE_CELL_RENDERER_TEXT,
                  "ellipsize", PANGO_ELLIPSIZE_END,
                  NULL);
   gtk_tree_view_column_pack_start(column, cell, TRUE);
   gtk_tree_view_column_set_cell_data_func(column, cell, source_title_func,
                                 window, NULL);
   gtk_tree_view_append_column(GTK_TREE_VIEW(priv->src_treeview), column);

   priv->sources = gtk_list_store_new(1, NEWS_TYPE_SOURCE);
   g_object_set(priv->src_treeview, "model", priv->sources, NULL);

   priv->item_scroller = g_object_new(GTK_TYPE_SCROLLED_WINDOW,
                              "shadow-type", GTK_SHADOW_NONE,
                              "visible", TRUE,
                              NULL);
   priv->item_treeview = g_object_new(GTK_TYPE_TREE_VIEW,
                              "headers-visible", TRUE,
                              "visible", TRUE,
                              NULL);
   sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(priv->item_treeview));
   g_signal_connect_swapped(sel, "changed",
                        G_CALLBACK(news_window_item_selected),
                        window);
   gtk_container_add(GTK_CONTAINER(priv->item_scroller),
                 priv->item_treeview);

   column = g_object_new(GTK_TYPE_TREE_VIEW_COLUMN,
                    "title", _("Title"),
                    NULL);
   cell = gtk_cell_renderer_text_new();
   gtk_tree_view_column_pack_start(column, cell, TRUE);
   gtk_tree_view_column_set_cell_data_func(column, cell, item_title_func,
                                 window, NULL);
   gtk_tree_view_append_column(GTK_TREE_VIEW(priv->item_treeview), column);

   priv->items = gtk_list_store_new(1, NEWS_TYPE_ITEM);
   g_object_set(priv->item_treeview, "model", priv->items, NULL);

   priv->item_view = g_object_new(NEWS_TYPE_ITEM_VIEW,
                                  "visible", TRUE,
                                  NULL);

   priv->status_box = g_object_new(GTK_TYPE_VBOX,
                           "spacing", 3,
                           "visible", TRUE,
                           "vexpand", FALSE,
                           NULL);
   gtk_container_add_with_properties(GTK_CONTAINER(priv->src_vbox),
                             priv->status_box,
                                     "expand", FALSE,
                                     NULL);

   priv->status = g_object_new(NEWS_TYPE_STATUS_VIEW,
                               "fraction", 0.0,
                               "icon-name", GTK_STOCK_NETWORK,
                               "subtitle", "",
                               "title", _("Synchronizing"),
                               "visible", FALSE,
                               NULL);
   g_signal_connect(priv->status, "cancel",
                    G_CALLBACK(status_cancel), window);
   gtk_container_add(GTK_CONTAINER(priv->status_box), priv->status);

   news_window_set_layout(window, NEWS_WINDOW_NORMAL);

   client = news_application_get_client(NEWS_APPLICATION_DEFAULT);
   g_signal_connect_swapped(client, "source-added",
                            G_CALLBACK(news_window_source_added),
                            window);
   g_signal_connect(client, "notify::status",
                    G_CALLBACK(status_changed), priv->status);

   for (list = news_client_get_sources(client); list; list = list->next) {
      news_window_source_added(window, list->data, client);
   }
}
