/* xml-reader.c
 *
 * Copyright (C) 2009  Christian Hergert  <chris@dronelabs.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * Author:
 *   Christian Hergert  <chris@dronelabs.com>
 */

/* After using Emmanuele Bassi's XmlWriter, I realized I wanted a
 * few more features and to be based on the xmlTextReader API. This
 * is the result of my quick hack to do so.
 *
 * You might wonder why i simply wrapped xmlTextReader from libxml2
 * as it is already a .NET like API. It's simple, I wanted a GObject
 * that was easily bindable for my plugins to use as well as my
 * core library in C.
 */

#include <string.h>
#include <libxml/xmlreader.h>

#include "xml-reader.h"

#define XML_TO_CHAR(s)  ((char *) (s))
#define CHAR_TO_XML(s)  ((unsigned char *) (s))
#define RETURN_STRDUP_AND_XMLFREE(stmt) \
  G_STMT_START {                        \
    guchar *x;                          \
    gchar *y;                           \
    x = stmt;                           \
    y = g_strdup((char *)x);            \
    xmlFree(x);                         \
    return y;                           \
  } G_STMT_END

struct _XmlReaderPrivate
{
  xmlTextReaderPtr xml;
  GInputStream *stream;
  gchar *cur_name;
  gchar *encoding;
  gchar *uri;
};

enum
{
   PROP_0,
   PROP_ENCODING,
   PROP_URI,
};

enum
{
   ERROR,
   LAST_SIGNAL
};

G_DEFINE_TYPE (XmlReader, xml_reader, G_TYPE_OBJECT);

static guint gSignals[LAST_SIGNAL];

#define XML_NODE_TYPE_ELEMENT        1
#define XML_NODE_TYPE_END_ELEMENT   15
#define XML_NODE_TYPE_ATTRIBUTE      2

static void
xml_reader_set_encoding (XmlReader   *reader,
                         const gchar *encoding)
{
   g_return_if_fail (XML_IS_READER (reader));
   g_free (reader->priv->encoding);
   reader->priv->encoding = g_strdup (encoding);
}

static void
xml_reader_set_uri (XmlReader   *reader,
                    const gchar *uri)
{
   g_return_if_fail (XML_IS_READER (reader));
   g_free (reader->priv->uri);
   reader->priv->uri = g_strdup (uri);
}

static void
xml_reader_clear (XmlReader *reader)
{
   XmlReaderPrivate *priv;

   g_return_if_fail(XML_IS_READER(reader));

   priv = reader->priv;

   g_free (priv->cur_name);
   priv->cur_name = NULL;

   if (priv->xml) {
      xmlTextReaderClose(priv->xml);
      xmlFreeTextReader(priv->xml);
      priv->xml = NULL;
   }

   if (priv->stream) {
      g_object_unref(priv->stream);
      priv->stream = NULL;
   }
}

static void
xml_reader_finalize (GObject *object)
{
   XmlReaderPrivate *priv;
   XmlReader *reader = (XmlReader *)object;

   priv = reader->priv;

   xml_reader_clear (reader);

   g_free (priv->encoding);
   priv->encoding = NULL;

   g_free (priv->uri);
   priv->uri = NULL;

   G_OBJECT_CLASS (xml_reader_parent_class)->finalize (object);
}

static void
xml_reader_get_property (GObject    *object,
                         guint       prop_id,
                         GValue     *value,
                         GParamSpec *pspec)
{
   XmlReader *reader = (XmlReader *)object;

   switch (prop_id) {
   case PROP_ENCODING:
      g_value_set_string (value, reader->priv->encoding);
      break;
   case PROP_URI:
      g_value_set_string (value, reader->priv->uri);
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

static void
xml_reader_set_property (GObject      *object,
                         guint         prop_id,
                         const GValue *value,
                         GParamSpec   *pspec)
{
   XmlReader *reader = (XmlReader *)object;

   switch (prop_id) {
   case PROP_ENCODING:
      xml_reader_set_encoding (reader, g_value_get_string (value));
      break;
   case PROP_URI:
      xml_reader_set_uri (reader, g_value_get_string (value));
      break;
   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
   }
}

static void
xml_reader_class_init (XmlReaderClass *klass)
{
  GObjectClass *object_class;

  object_class = G_OBJECT_CLASS (klass);
  object_class->finalize = xml_reader_finalize;
  object_class->get_property = xml_reader_get_property;
  object_class->set_property = xml_reader_set_property;
  g_type_class_add_private (object_class, sizeof (XmlReaderPrivate));

  g_object_class_install_property(object_class,
                                  PROP_ENCODING,
                                  g_param_spec_string("encoding",
                                                      "encoding",
                                                      "encoding",
                                                      NULL,
                                                      G_PARAM_READWRITE));

  g_object_class_install_property(object_class,
                                  PROP_URI,
                                  g_param_spec_string("uri",
                                                      "uri",
                                                      "uri",
                                                      NULL,
                                                      G_PARAM_READWRITE));

  gSignals[ERROR] = g_signal_new("error",
                                 XML_TYPE_READER,
                                 G_SIGNAL_RUN_FIRST,
                                 0,
                                 NULL,
                                 NULL,
                                 g_cclosure_marshal_VOID__STRING,
                                 G_TYPE_NONE,
                                 1,
                                 G_TYPE_STRING);
}

static void
xml_reader_init (XmlReader *reader)
{
  reader->priv = G_TYPE_INSTANCE_GET_PRIVATE (reader,
                                              XML_TYPE_READER,
                                              XmlReaderPrivate);
}

XmlReader*
xml_reader_new (void)
{
  return g_object_new (XML_TYPE_READER, NULL);
}

static void
xml_reader_error_cb (void *arg,
                     const char *msg,
                     xmlParserSeverities severity,
                     xmlTextReaderLocatorPtr locator)
{
   g_signal_emit(arg, gSignals[ERROR], 0, msg);
}

gboolean
xml_reader_load_from_file (XmlReader   *reader,
                           const gchar *filename)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  xml_reader_clear (reader);

  if ((priv->xml = xmlNewTextReaderFilename (filename))) {
    xmlTextReaderSetErrorHandler (priv->xml, xml_reader_error_cb, reader);
  }

  return (priv->xml != NULL);
}

gboolean
xml_reader_load_from_data (XmlReader   *reader,
                           const gchar *data,
                           gsize        length,
                           const gchar *uri,
                           const gchar *encoding)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  xml_reader_clear (reader);

  if (length == -1)
    length = strlen (data);

  priv->xml = xmlReaderForMemory (data, length, uri, encoding, 0);
  xmlTextReaderSetErrorHandler (priv->xml, xml_reader_error_cb, reader);

  return (priv->xml != NULL);
}

static int
xml_reader_io_read_cb (void *context,
                       char *buffer,
                       int   len)
{
  GInputStream *stream = (GInputStream *)context;
  g_return_val_if_fail (G_IS_INPUT_STREAM(stream), -1);
  return g_input_stream_read (stream, buffer, len, NULL, NULL);
}

static int
xml_reader_io_close_cb (void *context)
{
  GInputStream *stream = (GInputStream *)context;
  g_return_val_if_fail (G_IS_INPUT_STREAM(stream), -1);
  return g_input_stream_close (stream, NULL, NULL) ? 0 : -1;
}

gboolean
xml_reader_load_from_stream (XmlReader     *reader,
                             GInputStream  *stream,
                             GError       **error)
{
   XmlReaderPrivate *priv;

   g_return_val_if_fail (XML_IS_READER(reader), FALSE);

   priv = reader->priv;

   xml_reader_clear (reader);

   priv->xml = xmlReaderForIO(xml_reader_io_read_cb,
                              xml_reader_io_close_cb,
                              stream,
                              priv->uri,
                              priv->encoding,
                              XML_PARSE_RECOVER | XML_PARSE_NOBLANKS | XML_PARSE_COMPACT);

   if (!priv->xml) {
      g_set_error(error, XML_READER_ERROR, XML_READER_ERROR_INVALID,
                  "Could not parse XML from stream");
      return FALSE;
   }

   xmlTextReaderSetErrorHandler (priv->xml, xml_reader_error_cb, reader);

   return TRUE;
}

G_CONST_RETURN gchar *
xml_reader_get_value (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), NULL);

  priv = reader->priv;

  g_return_val_if_fail (priv->xml != NULL, NULL);

  return XML_TO_CHAR (xmlTextReaderConstValue (priv->xml));
}

G_CONST_RETURN gchar *
xml_reader_get_name (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), NULL);
  g_return_val_if_fail (reader->priv->xml != NULL, NULL);

  priv = reader->priv;

  return XML_TO_CHAR (xmlTextReaderConstName (priv->xml));
}

gchar *
xml_reader_read_string (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), NULL);

  priv = reader->priv;

  g_return_val_if_fail (priv->xml != NULL, NULL);

  RETURN_STRDUP_AND_XMLFREE (xmlTextReaderReadString (priv->xml));
}

gchar *
xml_reader_get_attribute (XmlReader   *reader,
                          const gchar *name)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), NULL);
  g_return_val_if_fail (reader->priv->xml != NULL, NULL);

  priv = reader->priv;

  RETURN_STRDUP_AND_XMLFREE (xmlTextReaderGetAttribute (priv->xml, CHAR_TO_XML (name)));
}

static gboolean
read_to_type_and_name (XmlReader   *reader,
                       gint         type,
                       const gchar *name)
{
  XmlReaderPrivate *priv;
  gboolean success = FALSE;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  g_return_val_if_fail (priv->xml != NULL, FALSE);

  while (xmlTextReaderRead (priv->xml) == 1)
    {
      if (xmlTextReaderNodeType (priv->xml) == type)
        {
          if (g_str_equal (XML_TO_CHAR (xmlTextReaderConstName (priv->xml)), name))
            {
              success = TRUE;
              break;
            }
        }
    }

  return success;
}

gboolean
xml_reader_read_start_element (XmlReader   *reader,
                               const gchar *name)
{
  XmlReaderPrivate *priv;
  gboolean success;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  success = read_to_type_and_name (reader, XML_NODE_TYPE_ELEMENT, name);

  if (success)
    {
      g_free (priv->cur_name);
      priv->cur_name = g_strdup (name);
    }

  return success;
}

gboolean
xml_reader_read_end_element (XmlReader *reader)
{
  XmlReaderPrivate *priv;
  gboolean success = FALSE;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  if (priv->cur_name)
    success = read_to_type_and_name (reader, XML_NODE_TYPE_END_ELEMENT, priv->cur_name);

  return success;
}

gchar *
xml_reader_read_inner_xml (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  RETURN_STRDUP_AND_XMLFREE (xmlTextReaderReadInnerXml (priv->xml));
}

gchar*
xml_reader_read_outer_xml (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  RETURN_STRDUP_AND_XMLFREE (xmlTextReaderReadOuterXml (priv->xml));
}

gboolean
xml_reader_read_to_next (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  return (xmlTextReaderNext (priv->xml) == 1);
}

gboolean
xml_reader_read (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  return (xmlTextReaderRead (priv->xml) == 1);
}

gboolean
xml_reader_read_to_next_sibling (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  xmlTextReaderMoveToElement (priv->xml);

  return (xmlTextReaderNextSibling (priv->xml) == 1);
}

gint
xml_reader_get_depth (XmlReader *reader)
{
   XmlReaderPrivate *priv;

   g_return_val_if_fail (XML_IS_READER(reader), -1);

   priv = reader->priv;

   return xmlTextReaderDepth (priv->xml);
}

void
xml_reader_move_up_to_depth (XmlReader *reader,
                             gint       depth)
{
   g_return_if_fail(XML_IS_READER(reader));

   while (xml_reader_get_depth(reader) > depth) {
      xml_reader_read_end_element(reader);
   }
}

xmlReaderTypes
xml_reader_get_node_type (XmlReader *reader)
{
   XmlReaderPrivate *priv;

   g_return_val_if_fail (XML_IS_READER (reader), 0);

   priv = reader->priv;

   return xmlTextReaderNodeType (priv->xml);
}

gboolean
xml_reader_is_empty_element (XmlReader *reader)
{
   XmlReaderPrivate *priv;

   g_return_val_if_fail (XML_IS_READER (reader), FALSE);

   priv = reader->priv;

   return xmlTextReaderIsEmptyElement (priv->xml);
}

gboolean
xml_reader_is_a (XmlReader   *reader,
                 const gchar *name)
{
   return !g_strcmp0 (xml_reader_get_name (reader), name);
}

gboolean
xml_reader_is_a_local (XmlReader   *reader,
                       const gchar *local_name)
{
   return !g_strcmp0 (xml_reader_get_local_name (reader), local_name);
}

gboolean
xml_reader_is_namespace (XmlReader   *reader,
                         const gchar *ns)
{
   XmlReaderPrivate *priv;

   g_return_val_if_fail (XML_IS_READER (reader), FALSE);

   priv = reader->priv;

   return !g_strcmp0 (XML_TO_CHAR(xmlTextReaderConstNamespaceUri (priv->xml)), ns);
}

G_CONST_RETURN gchar *
xml_reader_get_local_name (XmlReader *reader)
{
   XmlReaderPrivate *priv;

   g_return_val_if_fail(XML_IS_READER (reader), NULL);

   priv = reader->priv;

   return XML_TO_CHAR (xmlTextReaderConstLocalName (priv->xml));
}

gboolean
xml_reader_move_to_element (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  return (xmlTextReaderMoveToElement (priv->xml) == 1);
}

gboolean
xml_reader_move_to_attribute (XmlReader   *reader,
                              const gchar *name)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  return (xmlTextReaderMoveToAttribute (priv->xml, CHAR_TO_XML (name)) == 1);
}

gboolean
xml_reader_move_to_first_attribute (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  return (xmlTextReaderMoveToFirstAttribute (priv->xml) == 1);
}

gboolean
xml_reader_move_to_next_attribute (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  return (xmlTextReaderMoveToNextAttribute (priv->xml) == 1);
}

gboolean
xml_reader_move_to_nth_attribute (XmlReader *reader,
                                  gint       nth)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  return (xmlTextReaderMoveToAttributeNo (priv->xml, nth) == 1);
}

gint
xml_reader_count_attributes (XmlReader *reader)
{
  XmlReaderPrivate *priv;

  g_return_val_if_fail (XML_IS_READER (reader), FALSE);

  priv = reader->priv;

  return xmlTextReaderAttributeCount (priv->xml);
}

GQuark
xml_reader_error_quark (void)
{
  return g_quark_from_static_string("xml-reader-error-quark");
}
