// File: svd.c
// Author: Doug VK
// Professor: Vladimir Rokhlin
// Description: Block Jacobi SVD of a matrix
#include <math.h>
#include <malloc.h>

void swap(double* U, double* V, double* A, int old_index, int new_index, int size);
void jacobi(double* A, int size, double* S, double* U_store, double* V_store);
void mult_matrix(double* left, double* right, double* result, int size);
void compute_S(double* A, double* U, double* V, double* S, int size);
void findMax(int* row_max, int* col_max, double* A, int size);
void copy_matrix(double* to_copy, double* copy, int size);
void print_matrix(double* A, char* string, int size);
double* val(double* A, int row, int col, int size);
int comp(const double* a, const double* b);
void transpose(double* A, int size);
void identity(double* A, int size);

//run with: ./SVD [dimension]
main(int argc, char* argv[])
{
	int size = atoi(argv[1]);

    double* S = (double*) malloc(size*sizeof(double));
    double* A = (double*) malloc(size*size*sizeof(double));
    double* U = (double*) malloc(size*size*sizeof(double));
    double* V = (double*) malloc(size*size*sizeof(double));

	int i,j;
	for(i=0;i<size;i++) {
		for(j=0;j<size;j++) {
            *val(A,i,j,size) = i + j;
		}
	}

    jacobi(A, size, S, U, V);
}

double* val(double* A, int row, int col, int size)
{
    return A + size*row + col;
}

//uses block jacobi svd to diagonalize symmetric matrix A
void jacobi(double* A, int size, double* S, double* U_store, double* V_store)
{
    double* U    = (double*) malloc(size*size*sizeof(double));
    double* V    = (double*) malloc(size*size*sizeof(double));
    double* temp    = (double*) malloc(size*size*sizeof(double));
    identity(U_store,size);
    identity(V_store,size);


    // find max of array, set epsilon
    double epsilon = 1e-9;
    int j, row_max, col_max;
    findMax(&row_max,&col_max,A,size);

    for(j=0;j<100000000;j++)
    {
        //break if no significant pivot remains
        if(fabs(*val(A,row_max,col_max,size)) + fabs(*val(A,col_max,row_max,size)) < epsilon)
            break;

        //a == top left pivot, b == top right pivot, c == bottom left pivot, d == bottom right pivot
        double a = *val(A,row_max,row_max,size), b = *val(A,row_max,col_max,size),
               c = *val(A,col_max,row_max,size), d = *val(A,col_max,col_max,size);

        double beta = (d - a) / (2.0*b);
        double sign = copysign(1.0, beta);
        double tangent = sign / (fabs(beta) + sqrt(beta*beta + 1.0));
        double cosine = 1.0 / sqrt(tangent*tangent + 1.0);
        double sine = cosine * tangent;
        double tau = sine / (1.0 + cosine);

        identity(U,size);
        identity(V,size);
        *val(U,row_max,row_max,size) = sine;
        *val(U,row_max,col_max,size) = cosine;
        *val(U,col_max,row_max,size) = -cosine;
        *val(U,col_max,col_max,size) = sine;
        *val(V,row_max,row_max,size) = sine;
        *val(V,row_max,col_max,size) = -cosine;
        *val(V,col_max,row_max,size) = cosine;
        *val(V,col_max,col_max,size) = sine;

        //save the matrices to the U & V stores
        mult_matrix(U,U_store,temp,size);
        copy_matrix(temp,U_store,size);
        mult_matrix(V_store,V,temp,size);
        copy_matrix(temp,V_store,size);

        //diagonalize
        mult_matrix(U,A,temp,size);
        mult_matrix(temp,V,A,size);

        findMax(&row_max,&col_max,A,size);
    }
    transpose(V_store,size);
    transpose(U_store,size);
    compute_S(A, U_store, V_store, S, size);
    

    free(U);
    free(V);
    free(temp);
}

//simple double comparison func
int comp(const double* a, const double* b)
{
    if(fabs(*a) == fabs(*b))
        return 0;
    if(fabs(*a) < fabs(*b))
        return 1;
    if(fabs(*a) > fabs(*b))
        return -1;
    return 0;
}

//grab the eigenvalues (on diagonal), sort, print
void compute_S(double* A, double* U, double* V, double* S, int size)
{
    double eigv[size];
    int i,j,last_index = size;
    for(i=0;i<size;i++)
        eigv[i] = *val(A,i,i,size);
    qsort(eigv,size,sizeof(double),comp);

    //only print if >= epsilon
    double epsilon = 1e-13;
    for(i=size-1;i>0;i--) {
        if(fabs(eigv[i]) < epsilon) last_index = i;
    }

    //assign vals to S
    for (i = 0; i < last_index; i++) {
        S[i] = eigv[i];
    }

    for (i = 0; i < last_index; i++) {
        double search_val = eigv[i];
        for (j = 0; j < size; j++) {
            if(search_val == *val(A,j,j,size) && j != i)
                swap(U,V,A,j,i,size);
        }
    }
}

void swap(double* U, double* V, double* A, int old_index, int new_index, int size)
{
    double temp;
    int i;

    //swap cols of U
    for (i = 0; i < size; i++) {
        temp = *val(U,i,new_index,size);
        *val(U,i,new_index,size) = *val(U,i,old_index,size);
        *val(U,i,old_index,size) = temp;
    }

    //swap rows of V
    for (i = 0; i < size; i++) {
        temp = *val(V,new_index,i,size);
        *val(V,new_index,i,size) = *val(V,old_index,i,size);
        *val(V,old_index,i,size) = temp;
    }

    temp = *val(A,new_index,new_index,size);
    *val(A,new_index,new_index,size) = *val(A,old_index,old_index,size);
    *val(A,old_index,old_index,size) = temp;
}

void transpose(double* A, int size)
{
    int i,j;
    double temp;
    for(i=0;i<size;i++) {
        for(j=i+1;j<size;j++) {
            temp = *val(A,i,j,size);
            *val(A,i,j,size) = *val(A,j,i,size);
            *val(A,j,i,size) = temp;
        }
    }
}

void mult_matrix(double* left, double* right, double* result, int size)
{
    int i,j,k;
    for(i=0;i<size;i++) {
        for(j=0;j<size;j++) {
            *val(result,i,j,size) = 0;
            for(k=0;k<size;k++) {
                *val(result,i,j,size) += (*val(left,i,k,size))*(*val(right,k,j,size));
            }
        }
    }
}

void print_matrix(double* A, char* string, int size)
{
    int i;
    cprin2(string, A, size);
    for(i=1;i<size;i++)
        cprin2("", val(A,i,0,size), size);
}

//find max value in array
void findMax(int* row_max, int* col_max, double* A, int size)
{
    double max_val = 0;
    int i,j;
    for(i=0;i<size;i++) {
        for(j=0;j<size;j++) {
            //make sure pivot is not on diagonal
            if(max_val < fabs(*val(A,i,j,size)) && i != j) {
                max_val = fabs(*val(A,i,j,size));
                *row_max = i;
                *col_max = j;
            }
        }
    }

    //make sure [row_max][col_max] is the top-rightmost pivot
    if(*row_max > *col_max)
    {
        int temp = *row_max;
        *row_max = *col_max;
        *col_max = temp;
    }
}

//create identity matrix
void identity(double* A, int size)
{
    int i,j;
    for(i=0;i<size;i++) {
        for(j=0;j<size;j++) {
            if(i==j) *val(A,i,j,size) = 1.0;
            if(i!=j) *val(A,i,j,size) = 0.0;
        }
    }
}

void copy_matrix(double* to_copy, double* copy, int size)
{
    int i,j;
    for(i=0;i<size;i++) {
        for(j=0;j<size;j++) {
            *val(copy,i,j,size) = *val(to_copy,i,j,size);
        }
    }
}
