// File: descent.c
// Author: Doug VK
// Professor: Vladimir Rokhlin
// Description: Steepest descent of f(x) == ||Ax-y||^2
#include <stdio.h>
#include <malloc.h>
#include <math.h>

void dumb_solve(double* A, double* y, int size, double eps, int numit, double* x, int* niter, double* discreps);
void evaluate_gradient(double* grad_f, double* x, double* evaluated_grad, int size);
double set_x(double* evaluated_grad, double* A, double* x, double* y, int size);
void mult_matrix(double* left, double* right, double* result, int size);
void print_matrix(double* A, char* string, int m, int n);
double* val(double* A, int row, int col, int col_size);
double magnitude(double* vector, int size);

int main(int argc, const char *argv[])
{
    int size = atoi(argv[1]);
    int numit = atoi(argv[2]);
    double* A = (double*) malloc(size*size*sizeof(double));
    double* y = (double*) malloc(size*sizeof(double));
    double* x = (double*) malloc(size*sizeof(double));
    double* discreps = (double*) malloc(10000*sizeof(double));
    int niter;

    int i,j;
    for (i = 0; i < size; i++) {
        for (j = 0; j < size; j++) {
            if(i == j) *val(A,i,j,size) = 1.0 / (float)((i+1)*(i+1));
        }
    }

    for (i = 0; i < size; i++) {
        y[i] = 1;
    }
    cprini("stdout","output_3");

    print_matrix(A,"A = ",size,size);

    dumb_solve(A,y,size,1e-6,numit,x,&niter,discreps);

    print_matrix(x,"x = ", size, 1);
    print_matrix(discreps,"discreps = ", niter, 1);
    printf("Num Iterations: %i\n",niter);

    free(A);
    free(y);
    free(x);
    free(discreps);
    return 0;
}

void dumb_solve(double* A, double* y, int size, double eps, int numit, double* x, int* niter, double* discreps)
{

    //initialize variables to hold coefficients of the gradient
    //of f(x). result just helper variable
    double* evaluated_grad = (double*) malloc(size*sizeof(double)); 
    double* result = (double*) malloc(size*sizeof(double));

    //initialize f to be f(x) w/ column indices [x1,...,xn,const]
    double* f = (double*) malloc(size*(size+1)*sizeof(double));

    int i,j;
    for (i = 0; i < size; i++) {
        for (j = 0; j < size; j++) {
            *val(f,i,j,size+1) = *val(A,i,j,size);
        }
    }
    
    //initialize const vals
    for (i = 0; i < size; i++)
        *val(f,i,size,size+1) = -y[i];
    print_matrix(f, "F = ",size,size+1);

    //calculate the gradient of f
    double* grad_f = (double*) malloc(size*(size+1)*sizeof(double));
    int k;
    for (i = 0; i < size; i++) {
        for (j = 0; j < size+1; j++) {
            *val(grad_f,i,j,size+1) = 0;
            for (k = 0; k < size; k++) {
                *val(grad_f,i,j,size+1) += size*(*val(f,k,i,size+1))*(*val(f,k,j,size+1));
            }
        }
    }
    
    //set x to -y to get ||Ax-y|| for first discrep val
    for (i = 0; i < size; i++) {
        x[i] = -y[i];
    }
    discreps[0] = magnitude(x,size);

    //set x to initial guess
    for (i = 0; i < size; i++) {
        x[i] = 0;
    }

    //iterate until reach numit or eps
    //starts at 1 because we already calculated x = 0
    for (i = 1; i < numit; i++) {
        //get gradient evaluated at x
        evaluate_gradient(grad_f,x,evaluated_grad,size);

        //set_x to the next best guess based on gradient
        discreps[i] = set_x(evaluated_grad, A, x, y, size);

        //find error and normalize
        mult_matrix(A,x,result,size);
        if((discreps[i] / magnitude(result,size)) < eps) break;
    }

    //set niter and realloc discreps down to the right size
    *niter = i+1;
    if(*niter > numit) *niter = numit;

    free(f);
    free(grad_f);
    free(evaluated_grad);
    free(result);
}

double set_x(double* evaluated_grad, double* A, double* x, double* y, int size)
{
    //helper arrays
    double* guess = (double*) malloc(size*sizeof(double));
    double* old_guess = (double*) malloc(size*sizeof(double));
    double* result = (double*) malloc(size*sizeof(double));

    //initialize memory and step values
    double step_size, scalar_grad = 0.0;
    double old_val, new_val;
    int i,j;

    //keep decreasing step size as you approach minimum
    for (step_size = 1; step_size > 1e-15; step_size /= 10.0) {
        old_val = 1e9;
        new_val = 1e8;

        //iterate until error begins to increase 
        for (i = 0; old_val > new_val; i++) {
            //save state and minimum guess from last iteration
            old_val = new_val;
            for (j = 0; j < size && i > 0; j++) {
                old_guess[j] = guess[j];
            }

            //try next minimum guess
            for (j = 0; j < size; j++) {
                guess[j] = x[j] - ((scalar_grad+(step_size*i))*evaluated_grad[j]);
            }

            //Ax
            mult_matrix(A,guess,result,size);

            //Ax - y
            for (j = 0; j < size; j++) {
                result[j] -= y[j];
            }

            //||Ax-y||^2
            new_val = pow(magnitude(result, size),2);
        }

        //save scalar multiple that minimized on previous step
        //need to go 3 steps back in case missed minimization
        if(i-3 > 0) scalar_grad += step_size*(i-3);
    }

    //set x to the best guess value
    for (i = 0; i < size; i++) {
        x[i] -= scalar_grad*evaluated_grad[i];
    }

    //return ||Ax-y|| (discrepancy)
    mult_matrix(A,x,result,size);
    for (i = 0; i < size; i++) {
        result[i] -= y[i];
    }
    return magnitude(result,size);

    free(guess);
    free(old_guess);
    free(result);
}

//evaluate gradient at x
void evaluate_gradient(double* grad_f, double* x, double* evaluated_grad, int size)
{
    int i,j;
    for (i = 0; i < size; i++) {
        //set to const values then add other values
        evaluated_grad[i] = *val(grad_f,i,size,size+1);
        for (j = 0; j < size; j++) {
            evaluated_grad[i] += *val(grad_f,i,j,size+1)*x[j];
        }
    }
}

//returns pointer to A[row][col]
double* val(double* A, int row, int col, int col_size)
{
    return A + col_size*row + col;
}

//print matrix A(m,n)
void print_matrix(double* A, char* string, int m, int n)
{
    int i;
    cprin2(string, A, n);
    for(i=1;i<m;i++)
        cprin2("", val(A,i,0,n), n);
}

//returns ||vector||
double magnitude(double* vector, int size)
{
    double result = 0;
    int i;
    for (i = 0; i < size; i++) {
        result += vector[i]*vector[i];
    }
    return sqrt(result);
}

//multiplies only A(n,n)*x(n,1) type
void mult_matrix(double* left, double* right, double* result, int size)
{
    int i,j,k;
    for(i=0;i<size;i++) {
        for(j=0;j<1;j++) {
            result[i] = 0;
            for(k=0;k<size;k++) {
                result[i] += (*val(left,i,k,size))*right[k];
            }
        }
    }
}
