/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sv/cdb-private.h"
#include "sv/cdb.h"

/* Match the data buffer to that in the cdb file at the specified position.
 * Performs the matching operation in chunks so the entire piece of the file
 * doesn't need to be loaded all at once. */
static inline int match(cdb_t *cdb, void *data, size_t len, size_t pos) {
  uint8_t buf[32];
  ssize_t n;
  
  while (len > 0) {
    n = sizeof(buf);
    if (n > len)
      n = len;
    if (cdb_read(buf, n, cdb, pos) == -1)
      return -1;
    if (memcmp(buf, data, n))
      return 0;
    pos += n;
    data += n;
    len -= n;
  }
  
  return 1;
}

/**
 * Find the next entry in the cdb file.
 *
 * @param cdb  cdb
 * @param key  key
 * @param klen key length
 *
 * @return  1 record found
 * @return  0 no record found
 * @return -1 read error
 */

int_fast8_t cdb_find_next(cdb_t *cdb, void *key, uint32_t klen) {
  uint32_t buf[2];
  uint32_t hash;
  uint32_t len;
  uint32_t pos;
  
  if (!cdb->loop) {
    /* Hash the key and use it to find the index associated with it. */
    hash = _cdb_hash(key, klen);
    if (cdb_read(buf, sizeof(buf), cdb, (hash << 3) & 2047) == -1)
      return -1;
      
    /* Unpack the index's hash length and position. If the length field is
     * empty it means there's no hash table. */
    cdb->hlen = be32toh(buf[1]);
    if (!cdb->hlen)
      return 0;
    cdb->hpos = be32toh(buf[0]);
    
    /* Save the key's hash; then divide it by 256, modulo the hash table's
     * length and that's the slot in the hash table to use. */
    cdb->khash = hash;
    hash >>= 8;
    hash %= cdb->hlen;
    hash <<= 3;
    cdb->kpos = cdb->hpos + hash;
  }
  
  while (cdb->loop < cdb->hlen) {
    /* Read the hash table's entry. */
    if (cdb_read(buf, sizeof(buf), cdb, cdb->kpos) == -1)
      return -1;
    pos = be32toh(buf[1]);
    if (!pos)
      return 0;
    hash = be32toh(buf[0]);
    
    /* Increment the loop counter and the position in the hash table.  If at
     * the last hash table entry then wrap around to the first entry. */
    cdb->loop += 1;
    cdb->kpos += 8;
    if (cdb->kpos == cdb->hpos + (cdb->hlen << 3))
      cdb->kpos = cdb->hpos;
      
    /* Check the key's hash against that in the hash table and then the key
     * itself to the records key. */
    if (hash == cdb->khash) {
      if (cdb_read(buf, sizeof(buf), cdb, pos) == -1)
        return -1;
      len = be32toh(buf[0]);
      if (len == klen) {
        switch (match(cdb, key, len, pos + 8)) {
          case -1:
            return -1;
          case 1:
            cdb->dlen = be32toh(buf[1]);
            cdb->dpos = pos + len + 8;
            return 1;
        }
      }
    }
  }
  
  return 0;
}
