/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sv/cdb-private.h"
#include "sv/cdbmake-private.h"
#include "sv/cdbmake.h"

static void hplist_free(struct cdbmake_hplist *head) {
  struct cdbmake_hplist *next;
  
  while (head) {
    next = head->next;
    free(head);
    head = next;
  }
}

/* Determine which hash table is the largest and return it's size. */
static inline uint32_t lht(uint32_t *htlen) {
  uint32_t i;
  uint32_t n = 0;
  for (i = 0; i < 256; i++) {
    if (n < htlen[i])
      n = htlen[i];
  }
  return n;
}

/**
 * Finish creating the cdb file by creating the hash tables and updating the
 * index.
 *
 * @param cm cdbmake
 *
 * @return  0 success
 * @return -1 failure
 */

int_fast8_t cdbmake_finish(cdbmake_t *cm) {
  uint32_t buf[2];
  struct cdbmake_hplist *hplist;
  struct cdbmake_hp *hp;
  struct cdbmake_hp *ht = 0;
  uint32_t htlen[256];
  uint32_t i;
  uint32_t index[512];
  uint32_t len;
  uint32_t pos;
  struct cdbmake_hp *split = 0;
  uint32_t start[256];
  uint32_t total;
  uint32_t u;
  
  /* Zero out the hash table sizes. */
  for (i = 0; i < 256; ++i)
    htlen[i] = 0;
    
  /* Walk the hplist's and determine the number of records in each hash table
   * as well as the total number of records overall. */
  total = 0;
  for (hplist = cm->head; hplist; hplist = hplist->next) {
    i = hplist->num;
    while (i--) {
      total++;
      htlen[hplist->hp[i].hash & 255]++;
    }
  }
  
  /* Here the start array is set to the last position for each of the hash
   * tables. This is because the array is filled in reverse to preserve the
   * order the records were added to the file. */
  pos = 0;
  for (i = 0; i < 256; i++) {
    pos += htlen[i];
    start[i] = pos;
  }
  
  /* Allocate enough space to hold the largest hash table which will be reused
   * in creating each of the individual ones. */
  ht = (struct cdbmake_hp *)malloc(lht(htlen) * 2 * sizeof(struct cdbmake_hp));
  if (!ht)
    return -1;
    
  /* Allocate enough space to hold every hp which will be used in splitting
   * them into 256 groups. */
  split = (struct cdbmake_hp *)malloc(total * sizeof(struct cdbmake_hp));
  if (!split)
    goto ERROR;
    
  /* Start walking the hplist and splitting them up. Each of the entries in the
   * array are added in reverse order. After this point each entry in the
   * start array really is the start. */
  for (hplist = cm->head; hplist; hplist = hplist->next) {
    i = hplist->num;
    while (i--)
      split[--start[255 & hplist->hp[i].hash]] = hplist->hp[i];
  }
  
  for (i = 0; i < 256; ++i) {
    len = htlen[i] * 2;
    
    /* Fill in the index with the position and size of the hash table. */
    index[(i * 2)] = htobe32(cm->pos);
    index[(i * 2) + 1] = htobe32(len);
    
    /* Clear enough of the hash table array to hold all the elements in this
     * particular one. */
    for (u = 0; u < len; ++u)
      ht[u].hash = ht[u].pos = 0;
      
    /* Start walking the hp entries in the split array and putting them in the
     * correct position in the hash table. */
    hp = split + start[i];
    for (u = 0; u < htlen[i]; u++) {
      pos = (hp->hash >> 8) % len;
      while (ht[pos].pos)
        if (++pos == len)
          pos = 0;
      ht[pos] = *hp++;
    }
    
    /* Write the hash table out to the cdb file. */
    for (u = 0; u < len; ++u) {
      buf[0] = htobe32(ht[u].hash);
      buf[1] = htobe32(ht[u].pos);
      
      if (write(cm->fd, buf, sizeof(buf)) == -1)
        goto ERROR;
      if (_cdbmake_update_pos(cm, 8) == -1)
        goto ERROR;
    }
  }
  
  /* Free the allocated memory used for splitting up the hash/position entries
   * and the hash table array. */
  free(split);
  free(ht);
  hplist_free(cm->head);
  
  /* Flush any data left in the buffer before seeking to the beginning of the
   * file and writing the index. */
  if (lseek(cm->fd, 0, SEEK_SET) == -1)
    return -1;
  if (write(cm->fd, index, sizeof(index)) == -1)
    return -1;
  return 0;
  
ERROR:
  if (ht)
    free(ht);
  if (split)
    free(split);
  hplist_free(cm->head);
  return -1;
}
