/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sv/common-private.h"
#include "sv/base.h"

const int8_t base32_charset[32] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";
const int8_t base32hex_charset[32] = "0123456789ABCDEFGHIJKLMNOPQRSTUV";
const int8_t base32pad = '=';

/**
 * Encode input buffer to base32 and store in output buffer.  It is the callers
 * responsibility to ensure the target buffer is large enough to hold the
 * entire encoded value.
 *
 * @param dst     output buffer
 * @param src     input buffer
 * @param len     input buffer length
 * @param charset character set
 *
 * @return Number of bytes written to output buffer.
 */

ssize_t _base32_encode(void *dst, void *src, size_t len, const int8_t *charset) {
  uint8_t a, b, c, d, e;
  unsigned int i;
  uint8_t *pd = dst;
  uint8_t *ps = src;
  
  if (len == 0)
    return 0;
    
  /* Read five bytes at a time from the input buffer and convert them to base32
   * and put them in the output buffer. */
  for (i = 0; i < len; i += 5) {
    a = ps[i];
    b = (i + 1) < len ? ps[i + 1] : 0;
    c = (i + 2) < len ? ps[i + 2] : 0;
    d = (i + 3) < len ? ps[i + 3] : 0;
    e = (i + 4) < len ? ps[i + 4] : 0;
    
    *pd++ = charset[(a >> 3)];
    *pd++ = charset[((a & 7) << 2) | (b >> 6)];
    *pd++ = (len <= i + 1) ? base32pad : charset[(b & 62) >> 1];
    *pd++ = (len <= i + 1) ? base32pad : charset[((b & 1) << 4) | (c >> 4)];
    *pd++ = (len <= i + 2) ? base32pad : charset[((c & 15) << 1) | (d >> 7)];
    *pd++ = (len <= i + 3) ? base32pad : charset[((d & 124) >> 2)];
    *pd++ = (len <= i + 3) ? base32pad : charset[((d & 3) << 3) | (e >> 5)];
    *pd++ = (len <= i + 4) ? base32pad : charset[(e & 31)];
  }
  
  return pd - (uint8_t *)dst;
}
