/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sv/common-private.h"
#include "sv/base.h"

const int8_t base64_lookup[128] = {
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
  52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
  -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
  -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1,
};

const int8_t base64url_lookup[128] = {
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1,
  52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
  -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, 63,
  -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1,
};

/**
 * Decode input buffer from base64 and store in output buffer.  It is the
 * callers responsibility to ensure the target buffer is large enough to hold
 * the entire decoded value.
 *
 * @param dst    output buffer
 * @param src    input buffer
 * @param len    input buffer length
 * @param lookup lookup table
 *
 * @return Number of bytes written to output buffer, or -1 on failure
 */

ssize_t _base64_decode(void *dst, void *src, size_t len, const int8_t *lookup) {
  int b = 0;
  char ch;
  uint32_t chunk = 0;
  int end = 0;
  uint8_t *p = dst;
  int pos = 0;
  const char *s = src;
  const char *srcend = src + len;
  uint8_t *start = dst;
  
  /* Walk through the entire input buffer one character at a time. */
  while (s < srcend) {
    ch = *s++;
    
    /* Skip white space. */
    if (ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r')
      continue;
      
    if (ch == '=') {
      if (!end) {
        if (pos == 2)
          end = 1;
        else if (pos == 3)
          end = 2;
        else
          return -1;
      }
      b = 0;
    } else {
      b = -1;
      if ((ch > 0) && (ch < 127))
        b = lookup[ch];
      if (b < 0)
        return -1;
    }
    
    /* Save the decoded character to the buffer by shifting it six bits and
     * adding the value to it.  Once three decoded characters are processed
     * they are written to the output buffer and control variables reset. */
    chunk = (chunk << 6) + b;
    if (++pos == 4) {
      *p++ = (chunk >> 16) & 255;
      if (end == 0 || end > 1)
        *p++ = (chunk >> 8) & 255;
      if (end == 0 || end > 2)
        *p++ = chunk & 255;
      chunk = 0;
      pos = 0;
    }
  }
  
  if (pos != 0)
    return -1;
    
  return p - start;
}
