/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sv/common-private.h"
#include "sv/cache.h"

static char *miss(cache_t *cache) {
  cache->stats.miss++;
  return 0;
}

/**
 * Get cache entry.
 *
 * @param cache cache
 * @param key key
 * @param klen length of key
 * @param dlen pointer to store length of data
 * @param ttl pointer to store ttl remaining
 * @return pointer to data
 */

void *cache_get(cache_t *cache, void *key, size_t klen, size_t *dlen, uint32_t *ttl) {
  cache_header_t *header;
  unsigned int loop = 0;
  uint32_t nextpos;
  uint64_t now = time(0);
  uint32_t pos;
  uint32_t prevpos;
  
  cache->stats.get++;
  
  if (!cache->data)
    return 0;
  if (klen > CACHE_MAX_KEY_SIZE)
    return miss(cache);
    
  prevpos = cache_key_hash(cache, key, klen);
  pos = _cache_get(cache, prevpos);
  
  while (pos) {
    header = cache->data + pos;
    
    /* If the keys match and TTL is valid set ttl, dlen, and return a
     * poniter to data. */
    if (klen == header->klen) {
      if (!memcmp(key, cache_entry_get_key(header), klen)) {
        if (now > header->expire)
          return miss(cache);
          
        *ttl = header->expire - now;
        *dlen = header->dlen;
        return cache_entry_get_data(header);
      }
    }
    
    /* Protect against hash flooding. */
    if (loop++ > 64)
      return miss(cache);
      
    nextpos = prevpos ^ header->link;
    prevpos = pos;
    pos = nextpos;
  }
  
  return miss(cache);
}
