/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sv/common-private.h"
#include "sv/cache.h"

/**
 * Set cache entry.
 *
 * @param cache cache
 * @param key key
 * @param klen length of key
 * @param data data
 * @param dlen length of data
 * @param ttl time to live
 *
 * @return -1 error occurred
 */

int cache_set(cache_t *cache, void *key, size_t klen, void *data, size_t dlen, uint32_t ttl) {
  unsigned int entrylen;
  uint64_t expire;
  cache_header_t *header;
  unsigned int keyhash;
  uint32_t pos;
  
  cache->stats.set++;
  
  if ((!cache->data) ||
      (klen > CACHE_MAX_KEY_SIZE) ||
      (dlen > CACHE_MAX_DATA_SIZE) ||
      (!ttl))
    return 0;
    
  if (ttl > CACHE_MAX_TTL)
    ttl = CACHE_MAX_TTL;
    
  entrylen = CACHE_HEADER_SIZE + klen + dlen;
  
  /* Start freeing up space by nuking the oldest elements until enough space is
   * reclaimed. */
  while (cache->writer + entrylen > cache->oldest) {
    /* If oldest equals unused then it means we're at end of the cache, so
     * unused will now point to the last position in the cache that was too
     * small to hold the entry, while oldest and writer point to the
     * beginning. */
    if (cache->oldest == cache->unused) {
      if (cache->writer <= cache->hash_size)
        return -1;
      cache->unused = cache->writer;
      cache->oldest = cache->hash_size;
      cache->writer = cache->hash_size;
    }
    
    header = cache->data + cache->oldest;
    pos = header->link;
    _cache_set(cache, pos, _cache_get(cache, pos) ^ cache->oldest);
    
    /* Update the oldest offset to the next cache entry. */
    cache->oldest += cache_entry_size(header);
    if (cache->oldest > cache->unused)
      return -1;
      
    /* If the oldest offset is the same as the unused offset then we're at the
     * last entry in the cache, so update oldest and unused offsets to point to
     * the very end and reclaim the unused space. */
    if (cache->oldest == cache->unused) {
      cache->unused = cache->size;
      cache->oldest = cache->size;
    }
  }
  
  keyhash = cache_key_hash(cache, key, klen);
  expire = time(0) + ttl;
  
  pos = _cache_get(cache, keyhash);
  if (pos)
    _cache_set(cache, pos, _cache_get(cache, pos) ^ keyhash ^ cache->writer);
    
  /* Link header to writer and start filling in the data. */
  header = cache->data + cache->writer;
  header->link = pos ^ keyhash;
  header->klen = klen;
  header->dlen = dlen;
  header->expire = expire;
  cache_entry_set_key(header, key);
  cache_entry_set_data(header, data);
  
  /* Update the hash table entry and update writer offset. */
  _cache_set(cache, keyhash, cache->writer);
  cache->writer += entrylen;
  cache->stats.motion += entrylen;
  
  return 0;
}
