/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef SV_ARRAY_H
#define SV_ARRAY_H

#include <stddef.h>
#include <unistd.h>

/**
 * @file array.h
 * @brief Array Routines
 * @defgroup sv_array Array Routines
 * @{
 */

/** Array. */
typedef struct array {
  void   *data;  /**< dynamically allocated data */
  size_t  dsize; /**< elements allocated in the data buffer */
  size_t  esize; /**< element size */
  size_t  size;  /**< elements in the array */
} array_t;

/** array initialization */
#define ARRAY_INIT(esize)    {(0),(0),(esize),(0)}

/** array is allocated */
#define array_allocated(x)   (((x)->data != 0) && ((x)->dsize != 0))

/** elements available in the array */
#define array_available(x)   (array_failed((x)) ? 0 : ((x)->dsize - (x)->size))

/** bytes used in the array */
#define array_bytes(x)       (array_failed((x)) ? 0 : ((x)->esize * (x)->size))

/** array is failed */
#define array_failed(x)      (((x)->data == 0) && ((x)->dsize != 0))

/** first element in the array */
#define array_start(x)       (array_failed((x)) ? 0 : ((x)->data))

/** number of elements in the array */
#define array_size(x)        (array_failed((x)) ? 0 : ((x)->size))

/** truncate an array */
#define array_trunc(x)       (array_failed((x)) ? array_reset((x)) : ((x)->size = 0))

/** array is unallocated */
#define array_unallocated(x) (((x)->data == 0) && ((x)->dsize == 0))

void     array_append(array_t *, void *, size_t);
void     array_append_null(array_t *);
void     array_concat(array_t *, array_t *);
void     array_fail(array_t *);
void    *array_get(array_t *, size_t);
void     array_init(array_t *, size_t);
ssize_t  array_read_fd(array_t *, int, size_t);
ssize_t  array_read_fn(array_t *, char *, size_t);
void     array_reset(array_t *);
void     array_resize(array_t *, size_t);
void    *array_set(array_t *, size_t, void *);
ssize_t  array_write_fd(int, array_t *);
ssize_t  array_write_fn(char *, array_t *);

/** @} */

#endif
