/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef SV_BASE_H
#define SV_BASE_H

#include <stddef.h>
#include <stdint.h>
#include <unistd.h>

/**
 * @file base.h
 * @brief Base Routines
 * @defgroup sv_base Base Routines
 * @{
 */

extern const int8_t base32_charset[32];    /**< base32 character set */
extern const int8_t base32_lookup[128];    /**< base32 lookup table */
extern const int8_t base32hex_charset[32]; /**< base32hex character set */
extern const int8_t base32hex_lookup[128]; /**< base32hex lookup table */

extern const int8_t base64_charset[64];    /**< base64 character set */
extern const int8_t base64_lookup[128];    /**< base64 lookup table */
extern const int8_t base64url_charset[64]; /**< base64url character set */
extern const int8_t base64url_lookup[128]; /**< base64url lookup table */

/** base16 encode length based on input size */
#define base16_encode_size(n) ((n) * 2)

/** base16 decode length based on input size */
#define base16_decode_size(n) ((n) / 2)

/** base32 encode length based on input size */
#define base32_encode_size(n) (((n) + 6 - (((n) + 6) % 5)) / 5 * 8)

/** base32 decode length based on input size */
#define base32_decode_size(n) (((n) * 3) >> 2)

/** base64 encode length based on input */
#define base64_encode_size(n) (((n) + 2 - (((n) + 2) % 3)) / 3 * 4)

/** base64 decode length based on encoded input */
#define base64_decode_size(n) (((n) * 3) >> 2)

/** decode input from base32 */
#define base32_decode(d,s,l)    _base32_decode(d,s,l,base32_lookup)

/** encode input to base32 */
#define base32_encode(d,s,l)    _base32_encode(d,s,l,base32_charset)

/** decode input from base32hex */
#define base32hex_decode(d,s,l) _base32_decode(d,s,l,base32hex_lookup)

/** encode input to base32hex */
#define base32hex_encode(d,s,l) _base32_encode(d,s,l,base32hex_charset)

/** decode input from base64 */
#define base64_decode(d,s,l)    _base64_decode(d,s,l,base64_lookup)

/** encode input to base64 */
#define base64_encode(d,s,l)    _base64_encode(d,s,l,base64_charset)

/** decode input from base64url */
#define base64url_decode(d,s,l) _base64_decode(d,s,l,base64url_lookup)

/** encode input to base64url */
#define base64url_encode(d,s,l) _base64_encode(d,s,l,base64url_charset)

ssize_t _base32_decode(void *, void *, size_t, const int8_t *);
ssize_t _base32_encode(void *, void *, size_t, const int8_t *);
ssize_t _base64_decode(void *, void *, size_t, const int8_t *);
ssize_t _base64_encode(void *, void *, size_t, const int8_t *);
ssize_t base16_decode(void *, void *, size_t);
ssize_t base16_encode(void *, void *, size_t);

/** @} */

#endif
