/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef SV_BIO_H
#define SV_BIO_H

#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include "sv/array.h"
#include "sv/bio.h"

/**
 * @file bio.h
 * @brief Buffered I/O Routines
 * @defgroup sv_bio Buffered I/O Routines
 * @{
 */

/** buffered i/o type */
typedef struct bio {
  ssize_t (*op)(); /**< operation */
  int      fd;     /**< file descriptor */
  void    *data;   /**< data buffer */
  ssize_t  size;   /**< data buffer size */
  ssize_t  pos;    /**< data buffer position */
} bio_t;

extern bio_t *bio_0;          /**< stdin bio */
extern uint8_t bio_0_data[8192]; /**< stdin buffer */
extern bio_t *bio_1;          /**< stdout bio */
extern uint8_t bio_1_data[8192]; /**< stdout buffer */
extern bio_t *bio_2;          /**< stderr bio */
extern uint8_t bio_2_data[1024]; /**< stderr buffer */

/** bio initialization */
#define BIO_INIT(op,fd,data,size) { (op), (fd), (data), (size), 0 }

/** pointer to the current position in the bio's data buffer */
#define BIO_PEEK(b) ((b)->data + (b)->size)

/** update position in the bio's data buffer */
#define BIO_SEEK(b,len) (((b)->pos -= (len)),((b)->size += (len)))

/** get character from bio's data buffer */
#define BIO_GETC(c,b) \
  ( ((b)->pos > 0) \
    ? ( *(c) = ((char *)(b)->data)[(b)->size], BIO_SEEK((b),1), 1 ) \
    : bio_get((c),1,(b)) \
  )

/** put character into bio's data buffer */
#define BIO_PUTC(b,c) \
  ( ((b)->size != (b)->pos) \
    ? (((char *)(b)->data)[(b)->pos++] = (c), 0 ) \
    : bio_put((b),&(c),1) \
  )

/** put string into bio's data buffer */
#define bio_put_str(b,s) bio_put(b, s, strlen(s))

ssize_t bio_feed(bio_t *);
ssize_t bio_flush(bio_t *);
ssize_t bio_get(void *, ssize_t, bio_t *);
ssize_t bio_get_line(array_t *, bio_t *);
ssize_t bio_get_fill(void *, ssize_t, bio_t *);
void    bio_init(bio_t *, ssize_t (*)(), int, void *, size_t);
ssize_t bio_put(bio_t *, const void *, ssize_t);
ssize_t bio_put_fill(bio_t *, const void *, ssize_t);
ssize_t bio_stream(bio_t *out, bio_t *in, ssize_t len);

/** @} */

#endif
