/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef SV_CACHE_H
#define SV_CACHE_H

#include <stddef.h>
#include <stdint.h>
#include <string.h>

/**
 * @file cache.h
 * @brief Cache Routines
 * @defgroup sv_cache Cache Routines
 * @{
 */

#define CACHE_HEADER_SIZE   (sizeof(cache_header_t))  /**< size of cache header */
#define CACHE_INIT          {(0),(0),(0),(0),(0),(0)} /**< cache initialization */

#define CACHE_MAX_DATA_SIZE 1048576    /**< max data size */
#define CACHE_MAX_KEY_SIZE  1024       /**< max key size */
#define CACHE_MAX_SIZE      2147483648 /**< max cache size */
#define CACHE_MAX_TTL       604800     /**< max ttl value */
#define CACHE_MIN_SIZE      128        /**< min cache size */

/** cache statistics */
typedef struct cache_stats_t {
  uint64_t get;    /**< number of cache_gets performed on the cache */
  uint64_t miss;   /**< number of cache misses */
  uint64_t motion; /**< number of bytes written to the cache */
  uint64_t set;    /**< number of cache_sets performed on the cache */
} cache_stats_t;

/** cache */
typedef struct cache_t {
  void          *data;      /**< dynamically allocated data */
  size_t         size;      /**< size of data */
  size_t         hash_size; /**< size of the hash index */
  size_t         writer;    /**< current position to write the next entry */
  size_t         oldest;    /**< oldest entry in the cache */
  size_t         unused;    /**< last entry in the cache which is unused */
  cache_stats_t  stats;     /**< statistics */
} cache_t;

/** cache element header */
typedef struct cache_header_t {
  uint32_t link;   /**< link to next element header */
  uint32_t klen;   /**< size of key */
  uint32_t dlen;   /**< size of data */
  uint64_t expire; /**< expiration */
} cache_header_t;

/** get value at cache offset */
#define _cache_get(c,o)           (*(uint32_t *)((c)->data + (o)))

/** set value at cache offset */
#define _cache_set(c,o,v)         (*(uint32_t *)((c)->data + (o)) = (v))

/** total cache entry size */
#define cache_entry_size(h)       (CACHE_HEADER_SIZE + (h)->klen + (h)->dlen)

/** get pointer to entry data */
#define cache_entry_get_data(h)   ((char *)(h) + CACHE_HEADER_SIZE + (h)->klen)

/** get pointer to entry key */
#define cache_entry_get_key(h)    ((char *)(h) + CACHE_HEADER_SIZE)

/** set cache entry data */
#define cache_entry_set_data(h,d) memcpy((char *)(h) + CACHE_HEADER_SIZE + (h)->klen, d, (h)->dlen)

/** set cache entry key */
#define cache_entry_set_key(h,k)  memcpy((char *)(h) + CACHE_HEADER_SIZE, k, (h)->klen)

void          cache_free(cache_t *);
void         *cache_get(cache_t *, void *, size_t, size_t *, uint32_t *);
int           cache_init(cache_t *, size_t);
unsigned int  cache_key_hash(cache_t *, void *, size_t);
int           cache_set(cache_t *, void *, size_t, void *, size_t, uint32_t);

/** @} */

#endif
