/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sv/time-private.h"
#include "sv/caltime.h"

/**
 * Scan a string in caltime format into a caltime struct.
 *
 * @param ct caltime struct
 * @param s  string
 *
 * @return number of characters scanned
 * @retval 0 invalid caldate string
 */

unsigned int caltime_scan(caltime_t *ct, char *s) {
  char *t = s;
  unsigned long z;
  unsigned long c;
  int sign;
  
  /* The first part of a caltime string is the date so use caldate_scan() to
   * parse the beginning of the string and update the pointer to the caltime
   * component. */
  t += caldate_scan(&ct->date, t);
  
  /* Skip over spaces and tabs if they exist. */
  while ((*t == ' ') || (*t == '\t') || (*t == 'T'))
    ++t;
    
  /* Scan the hours component. */
  z = 0;
  while ((c = (unsigned char) (*t - '0')) <= 9) {
    z = z * 10 + c;
    ++t;
  }
  ct->hour = z;
  
  /* Skip the colon if it exists or return if it doesn't. */
  if (*t++ != ':')
    return 0;
    
  /* Scan the minute component. */
  z = 0;
  while ((c = (unsigned char) (*t - '0')) <= 9) {
    z = z * 10 + c;
    ++t;
  }
  ct->minute = z;
  
  /* If a colon doesn't exist the seconds aren't specified so set it to zero,
   * otherwise scan the seconds component. */
  if (*t != ':')
    ct->second = 0;
  else {
    ++t;
    z = 0;
    while ((c = (unsigned char) (*t - '0')) <= 9) {
      z = z * 10 + c;
      ++t;
    }
    ct->second = z;
  }
  
  /* Skip over spaces and tabs if they exist. */
  while ((*t == ' ') || (*t == '\t'))
    ++t;
    
  /* Scan the timezone offset sign component or return 0 if it doesn't exist. */
  switch (*t) {
    case '+':
      sign = 1;
      break;
    case '-':
      sign = -1;
      break;
    default:
      return 0;
  }
  t++;
  
  c = (unsigned char) (*t++ - '0');
  if (c > 9)
    return 0;
  z = c;
  
  c = (unsigned char) (*t++ - '0');
  if (c > 9)
    return 0;
  z = z * 10 + c;
  
  c = (unsigned char) (*t++ - '0');
  if (c > 9)
    return 0;
  z = z * 6 + c;
  
  c = (unsigned char) (*t++ - '0');
  if (c > 9)
    return 0;
  z = z * 10 + c;
  
  ct->offset = z * sign;
  
  return t - s;
}
