/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef SV_H
#define SV_H

#include <unistd.h>
#include <sv/taia.h>

/**
 * @file sv.h
 * @brief SV Routines
 * @defgroup sv SV Routines
 * @{
 */

/* Custom errno values */
#define ESVSTATDOWN      -2   /**< failed to stat ./down */
#define ESVNOTRUN        -3   /**< supervise not running */
#define ESVOPENCONTROL   -4   /**< failed to open supervise/control */
#define ESVOPENOK        -5   /**< failed to open supervise/ok */
#define ESVOPENSTAT      -6   /**< failed to open supervise/status */
#define ESVOPENSTATNEW   -7   /**< failed to open supervise/status.new */
#define ESVREADSTAT      -8   /**< failed to read supervise/status */
#define ESVINVALSTAT     -9   /**< invalid status format */
#define ESVWRITECONTROL  -10  /**< failed to write to supervise/control */
#define ESVWRITESTAT     -11  /**< failed to write supervise/status */
#define ESVWRITESTATNEW  -12  /**< failed to write supervise/status.new */

/* Customer errno messages */
#define ESVSTATDOWN_MSG      "unable to stat down"
#define ESVNOTRUN_MSG        "supervise not running"
#define ESVOPENCONTROL_MSG   "unable to open supervise/control"
#define ESVOPENOK_MSG        "unable to open supervise/ok"
#define ESVOPENSTAT_MSG      "unable to open supervise/status"
#define ESVOPENSTATNEW_MSG   "unable to open supervise/status.new"
#define ESVREADSTAT_MSG      "unable to read supervise/status"
#define ESVINVALSTAT_MSG     "invalid supervise/status"
#define ESVWRITECONTROL_MSG  "unable to write supervise/control"
#define ESVWRITESTAT_MSG     "unable to write supervise/status"
#define ESVWRITESTATNEW_MSG  "unable to write supervise/status.new"

#define SUPERVISE_PATH            "supervise"          /**< supervise path */
#define SUPERVISE_CONTROL_PATH    "supervise/control"  /**< supervise/control path */
#define SUPERVISE_LOCK_PATH       "supervise/lock"     /**< supervise/lock path */
#define SUPERVISE_OK_PATH         "supervise/ok"       /**< supervise/ok path */
#define SUPERVISE_STATUS_NEW_EXT  ".new"               /**< status.new extension */
#define SUPERVISE_STATUS_PATH     "supervise/status"   /**< supervise/status path */
#define SUPERVISE_STATUS_SIZE     18                   /**< supervise/status size */

#define SVCONTROL_DOWN     'd'  /**< stop service */
#define SVCONTROL_EXIT     'x'  /**< exit supervise */
#define SVCONTROL_ONCE     'o'  /**< start service once */
#define SVCONTROL_UP       'u'  /**< start service */
#define SVCONTROL_SIGALRM  'a'  /**< send sigalrm signal */
#define SVCONTROL_SIGCONT  'c'  /**< send sigcont signal */
#define SVCONTROL_SIGHUP   'h'  /**< send sighup signal */
#define SVCONTROL_SIGINT   'i'  /**< send sigint signal */
#define SVCONTROL_SIGKILL  'k'  /**< send sigkill signal */
#define SVCONTROL_SIGSTOP  'p'  /**< send sigstop signal */
#define SVCONTROL_SIGTERM  't'  /**< send sigterm signal */
#define SVCONTROL_SIGQUIT  'q'  /**< send sigquit signal */
#define SVCONTROL_SIGUSR1  '1'  /**< send sigusr1 signal */
#define SVCONTROL_SIGUSR2  '2'  /**< send sigusr2 signal */

/** svstat contains information about a supervise service */
typedef struct sv_stat {
  int         autostart; /**< set if supervise autostarts the service */
  struct taia timestamp; /**< timestamp of last update */
  pid_t       pid;       /**< set if supervise is running a service */
  int         paused;    /**< set if child process has been paused */
  int         mode;      /**< supervise mode.  0, 'u', or 'd' */
} sv_stat_t;

int         sv_control(char *, char *, int);
const char *sv_error_str(int);
int         sv_restart(char *, int);
int         sv_restart_log(char *, int);
int         sv_start(char *);
int         sv_start_log(char *);
int         sv_stat(sv_stat_t *, char *);
int         sv_stat_load(sv_stat_t *, char *);
int         sv_stat_save(char *, sv_stat_t *);
int         sv_stop(char *, int);
int         sv_stop_log(char *, int);

/** @} */

#endif
