/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sv-private.h"
#include "sv.h"

/* Reset the array, fill it with the path arguments, and a null terminator. */
static int array_path(array_t *sa, char *path1, char *path2) {
  array_reset(sa);
  array_append(sa, path1, strlen(path1));
  array_append(sa, "/", 1);
  array_append(sa, path2, strlen(path2));
  array_append_null(sa);
  return array_failed(sa);
}

/**
 * Returns information about a supervised service.
 *
 * @param svst svstat
 * @param path service path
 *
 * @retval 0  success
 * @retval -1 failure
 * @retval ESVSTATDOWN error stat()ing down.
 * @retval ESVNOTRUN supervise is not running.
 * @retval ESVOPENOK error opening supervise/ok
 * @retval ESVINVALSTAT invalid supervise/status
 * @retval ESVOPENSTAT error opening supervise/status
 * @retval ESVREADSTAT error reading supervise/status
 */

int sv_stat(sv_stat_t *svst, char *path) {
  array_t full_path = ARRAY_INIT(1);
  int ok_fd;
  struct stat st;
  int status_fd;
  int32_t tmp;
  
  /* Check if 'down' exist indicating that the service isn't started when
   * supervise starts up. */
  if (array_path(&full_path, path, "down")) {
    array_reset(&full_path);
    errno = ENOMEM;
    return -1;
  }
  
  if (stat(array_start(&full_path), &st) == -1) {
    if (errno != ENOENT) {
      array_reset(&full_path);
      return ESVSTATDOWN;
    }
    svst->autostart = 1;
  } else
    svst->autostart = 0;
    
  /* Attempt to open the supervise/ok fifo to see if supervise is actually
   * running. */
  if (array_path(&full_path, path, SUPERVISE_OK_PATH)) {
    array_reset(&full_path);
    errno = ENOMEM;
    return -1;
  }
  ok_fd = open_write(array_start(&full_path));
  if (ok_fd == -1) {
    array_reset(&full_path);
    if (errno == ENODEV)
      return ESVNOTRUN;
    return ESVOPENOK;
  }
  close(ok_fd);
  
  /* Load the contents of the supervise/status file. */
  if (array_path(&full_path, path, SUPERVISE_STATUS_PATH)) {
    array_reset(&full_path);
    errno = ENOMEM;
    return -1;
  }
  if (sv_stat_load(svst, array_start(&full_path)) == -1) {
    array_reset(&full_path);
    return -1;
  }
  
  array_reset(&full_path);
  return 0;
}
