/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "sv-private.h"
#include "sv.h"

/**
 * Attempts to stop a supervised service and blocks until either the timeout
 * is reached, in which case it sends a final 'kill' message before returning,
 * or until supervise registers the service as being down.
 *
 * @param path    service path
 * @param timeout timeout in seconds
 *
 * @retval 0  success
 * @retval -1 failure
 *
 * @remarks if timeout is 0 then svstop blocks forever until the service stops.
 */

int sv_stop(char *path, int timeout) {
  struct tai expire;
  struct tai now;
  int r;
  sv_stat_t svst;
  
  tai_now(&now);
  tai_init(&expire, timeout);
  tai_add(&expire, &expire, &now);
  
  /* Stat the supervise path and return with an error if it failed. */
  r = sv_stat(&svst, path);
  if (r)
    return r;
    
  /* If supervise is currently monitoring a service first instruct it to
   * bring the service down and probe it every second until either the timeout
   * is reached or supervise reports its down.  If the timeout has passed and
   * the service is still running then instruct supervise to send a SIGKILL. */
  if (svst.pid) {
    sv_control(path, "d", 1);
    
    do {
      sleep(1);
      r = sv_stat(&svst, path);
      if (r)
        return r;
      tai_now(&now);
    } while ((svst.pid) && ((!timeout) || (tai_less(&now, &expire))));
    
    if (svst.pid)
      sv_control(path, "k", 1);
  }
  
  return 0;
}
