/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sv/pathexec.h>
#include <sv/scan.h>
#include <sv/sgetopt.h>
#include <sv/strerr.h>
#include <sys/resource.h>
#include <sys/time.h>

#define FATAL "softlimit: fatal: "

void usage() {
  strerr_die1x(100,
               "\n"
               "softlimit: usage: softlimit [ options ] command \n"
               "\n"
               "  -a AS, VMEM   Maximum total memory (bytes)\n"
               "  -c CORE       Maximum core file size (bytes)\n"
               "  -d DATA       Maximum size of data segment (bytes)\n"
               "  -f FSIZE      Maximum file size that can be created (bytes)\n"
               "  -l MEMLOCK    Maximum size of locked memory (bytes)\n"
               "  -o NOFILE     Maximum number of open files for this process\n"
               "  -p NPROC      Maximum number of simultaneous processes for user\n"
               "  -r RSS        Maximum resident set size (bytes)\n"
               "  -s STACK      Maximum stack size in bytes (bytes)\n"
               "  -t CPU        Maximum process CPU time (seconds)\n"
               "\n"
               "  -m AS, DATA, MEMLOCK, STACK, VMEM\n"
              );
}

void set_limit(int resource, char *arg) {
  struct rlimit r;
  unsigned long u;
  
  if (getrlimit(resource, &r) == -1)
    strerr_die2sys(111, FATAL, "getrlimit failed: ");
    
  if (!strcmp(arg, "="))
    r.rlim_cur = r.rlim_max;
  else {
    if (arg[scan_ulong(&u, arg)])
      usage();
    r.rlim_cur = u;
    if (r.rlim_cur > r.rlim_max)
      r.rlim_cur = r.rlim_max;
  }
  
  if (setrlimit(resource, &r) == -1)
    strerr_die2sys(111, FATAL, "setrlimit failed: ");
}

int main(int argc, char **argv, char **envp) {
  int opt;
  
  if (!*argv || !*(argv + 1))
    usage();
    
  while ((opt = sgetopt(argc, argv, "a:c:d:f:l:m:o:p:r:s:t:")) != -1)
    switch (opt) {
      case '?':
        usage();
      case 'a':
#ifdef RLIMIT_AS
        set_limit(RLIMIT_AS, soptarg);
#endif
#ifdef RLIMIT_VMEM
        set_limit(RLIMIT_VMEM, soptarg);
#endif
        break;
      case 'c':
#ifdef RLIMIT_CORE
        set_limit(RLIMIT_CORE, soptarg);
#endif
        break;
      case 'd':
#ifdef RLIMIT_DATA
        set_limit(RLIMIT_DATA, soptarg);
#endif
        break;
      case 'f':
#ifdef RLIMIT_FSIZE
        set_limit(RLIMIT_FSIZE, soptarg);
#endif
        break;
      case 'l':
#ifdef RLIMIT_MEMLOCK
        set_limit(RLIMIT_MEMLOCK, soptarg);
#endif
        break;
      case 'm':
#ifdef RLIMIT_DATA
        set_limit(RLIMIT_DATA, soptarg);
#endif
#ifdef RLIMIT_STACK
        set_limit(RLIMIT_STACK, soptarg);
#endif
#ifdef RLIMIT_MEMLOCK
        set_limit(RLIMIT_MEMLOCK, soptarg);
#endif
#ifdef RLIMIT_VMEM
        set_limit(RLIMIT_VMEM, soptarg);
#endif
#ifdef RLIMIT_AS
        set_limit(RLIMIT_AS, soptarg);
#endif
        break;
      case 'o':
#ifdef RLIMIT_NOFILE
        set_limit(RLIMIT_NOFILE, soptarg);
#endif
#ifdef RLIMIT_OFILE
        set_limit(RLIMIT_OFILE, soptarg);
#endif
        break;
      case 'p':
#ifdef RLIMIT_NPROC
        set_limit(RLIMIT_NPROC, soptarg);
#endif
        break;
      case 'r':
#ifdef RLIMIT_RSS
        set_limit(RLIMIT_RSS, soptarg);
#endif
        break;
      case 's':
#ifdef RLIMIT_STACK
        set_limit(RLIMIT_STACK, soptarg);
#endif
        break;
      case 't':
#ifdef RLIMIT_CPU
        set_limit(RLIMIT_CPU, soptarg);
#endif
        break;
    }
    
  argv += soptind;
  if (!*argv)
    usage();
    
  pathexec_run(*argv, argv, envp);
  strerr_die4sys(111, FATAL, "unable to run '", *argv, "': ");
}
