/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <errno.h>
#include <pwd.h>
#include <unistd.h>
#include <sv/array.h>
#include <sv/bio.h>
#include <sv/byte.h>
#include <sv/file.h>
#include <sv/str.h>
#include <sys/stat.h>
#include "supervise.h"

/**
 * Load the specified supervise.conf into the supervise_config struct.
 * 
 * @param config supervise config
 * @param path   path to supervise.conf
  
 * @retval -1 failure
 * @retval 0  success
 */

int config_load(struct supervise_config *config, char *path) {
  array_t line = ARRAY_INIT(1);
  char bio_fd_data[512];
  bio_t bio_fd;
  int fd;
  unsigned long mode;
  int pos;
  struct passwd *pwd;
  struct stat st;
  
  if (stat(path, &st) == -1) {
    if (errno == ENOENT)
      return 0;
    else
      return -1;
  }
  
  fd = open_read(path);
  if (fd == -1)
    return -1;
  
  bio_init(&bio_fd, read, fd, bio_fd_data, sizeof(bio_fd_data));
  
  /* Read each line of the configuration file and apply it's setting to the
   * supervise_config struct. */
  while (bio_get_line(&line, &bio_fd) > 0) {
    char *data = array_start(&line);
    int data_len = array_bytes(&line);
    array_t key = STRARRAY_INIT;
    array_t value = STRARRAY_INIT;
    
    /* If a '#' is found, set it to a null byte to effectively end the string
     * and ignore anything after it. */
    data[byte_chr(data, data_len, '#')] = 0;
    data_len = str_len(data);
    
    /* If the line doesn't contain a 'key=value' then print a warning and
     * continue on. */
    pos = str_chr(data, '=');
    if (pos == str_len(data)) {
      strerr_warn4x(WARNING, "invalid config line: '", data, "'");
      goto CONTINUE;
    }
    
    /* Load everything up to the '=' into the key, increment pos to skip over
     * the equal sign, and trim the spaces off of key. */
    array_append(&key, data, pos++);
    array_trim(&key);
    array_append_null(&key);
    
    /* Load the rest of the string into the value and trim it. */
    array_append_str(&value, data + pos);
    array_trim(&value);
    array_append_null(&value);
    
    /* Compare the key to each of the supported settings and handle setting
     * the supervise_config struct accordingly.  If it's an unknown setting 
     * print a warning to let the user know. */
    if (str_equal(array_start(&key), "control.group")) {
      pwd = getpwnam(array_start(&value));
      if (!pwd) {
        strerr_warn4x(WARNING, "invalid control.group: '", array_start(&value), "'");
        goto CONTINUE;
      }
      config->control_gid = pwd->pw_gid;
    }
    else if (str_equal(array_start(&key), "control.mode")) {
      char *value_str = array_start(&value);
      if (value_str[scan_octal(value_str, &mode)] != 0) {
        strerr_warn4x(WARNING, "invalid control.mode: '", value_str, "'");
        goto CONTINUE;
      }
      config->control_mode = mode;
    }
    else if (str_equal(array_start(&key), "ok.group")) {
      pwd = getpwnam(array_start(&value));
      if (!pwd) {
        strerr_warn4x(WARNING, "invalid ok.group: '", array_start(&value), "'");
        goto CONTINUE;
      }
      config->ok_gid = pwd->pw_gid;
    }
    else if (str_equal(array_start(&key), "ok.mode")) {
      char *value_str = array_start(&value);
      if (value_str[scan_octal(value_str, &mode)] != 0) {
        strerr_warn4x(WARNING, "invalid ok.mode: '", value_str, "'");
        goto CONTINUE;
      }
      config->ok_mode = mode;
    }
    else if (str_equal(array_start(&key), "status.group")) {
      pwd = getpwnam(array_start(&value));
      if (!pwd) {
        strerr_warn4x(WARNING, "invalid status.group: '", array_start(&value), "'");
        goto CONTINUE;
      }
      config->status_gid = pwd->pw_gid;
    }
    else if (str_equal(array_start(&key), "status.mode")) {
      char *value_str = array_start(&value);
      if (value_str[scan_octal(value_str, &mode)] != 0) {
        strerr_warn4x(WARNING, "invalid ok.mode: '", value_str, "'");
        goto CONTINUE;
      }
      config->status_mode = mode;
    }
    else {
      strerr_warn4x(WARNING, "unknown config option: '", array_start(&key), "'");
    }
    
CONTINUE:
    array_reset(&line);
  }
  
  return 0;
}
