/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <errno.h>
#include <sv.h>
#include <unistd.h>
#include <sv/file.h>
#include <sv/sig.h>
#include <sv/strerr.h>
#include <sys/stat.h>
#include "supervise.h"

struct supervise_config config;
struct supervise_status status;

void usage() {
  strerr_die1x(100,
               "\n"
               "supervise: usage: supervise path\n"
              );
}

int main(int argc, char **argv) {
  struct stat st;
  
  config_init(&config);
  status_init(&status);
  
  if (!*argv || !*++argv)
    usage();
  config.service_path = *argv;
  
  /* Jump into the service directory so all operations are relative to it. */
  if (chdir(config.service_path) == -1)
    strerr_die4sys(111, FATAL, "unable to chdir to ", config.service_path,
                   ": ");
                   
  /* Create the notify pipe used to signal supervise when the child process
   * terminates. */
  if (pipe(status.pipe) == -1)
    strerr_die4sys(111, FATAL, "unable to create pipe for ",
                   config.service_path, ": ");
                   
  close_on_exec(status.pipe[0]);
  close_on_exec(status.pipe[1]);
  blocking_disable(status.pipe[0]);
  blocking_disable(status.pipe[1]);
  
  /* When the child process terminates, trigger the main loop which will handle
   * restarting the service. */
  sig_block(SIGCHLD);
  sig_catch(SIGCHLD, trigger);
  
  /* Flag the service as being down if [service]/down exists. */
  if (stat("down", &st) != -1)
    status.flag_down = 1;
  else if (errno != ENOENT)
    strerr_die4sys(111, FATAL, "unable to stat ", config.service_path,
                   "/down: ");
                   
  /* Attempt to aquire the supervise lock and exit with an error message if it
   * fails. */
  mkdir(SUPERVISE_PATH, 0700);
  status.fd_lock = open_append(SUPERVISE_LOCK_PATH);
  if ((status.fd_lock == -1) || (flock_exnb(status.fd_lock) == -1))
    strerr_die6sys(111, FATAL, "unable to acquire ", config.service_path, "/",
                   SUPERVISE_LOCK_PATH, ": ");
  close_on_exec(status.fd_lock);
  
  /* Create the control fifo and exit with an error message if it fails. */
  mkfifo(SUPERVISE_CONTROL_PATH, 0600);
  status.fd_control = open_read(SUPERVISE_CONTROL_PATH);
  if (status.fd_control == -1)
    strerr_die6sys(111, FATAL, "unable to read ", config.service_path, "/",
                   SUPERVISE_CONTROL_PATH, ": ");
  close_on_exec(status.fd_control);
  
  /* Open the control fifo for writing. This is needed to prevent reads from
   * failing on the control fd in the main_loop. */
  status.fd_control_write = open_write(SUPERVISE_CONTROL_PATH);
  if (status.fd_control_write == -1)
    strerr_die6sys(111, FATAL, "unable to write ", config.service_path, "/",
                   SUPERVISE_CONTROL_PATH, ": ");
  close_on_exec(status.fd_control_write);
  
  update_status();
  
  /* Create the ok fifo which allows other processes to test if supervise is
   * actually running for a given service. */
  mkfifo(SUPERVISE_OK_PATH, 0600);
  status.fd_ok = open_read(SUPERVISE_OK_PATH);
  if (status.fd_ok == -1)
    strerr_die6sys(111, FATAL, "unable to read ", config.service_path, "/",
                   SUPERVISE_OK_PATH, ": ");
  close_on_exec(status.fd_ok);
  
  if (!status.flag_want || !status.flag_down)
    start();
    
  main_loop();
  
  update_status();
  _exit(0);
}
