/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <errno.h>
#include <poll.h>
#include <unistd.h>
#include <sv/sig.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "supervise.h"

void main_loop() {
  char ch;
  struct pollfd x[2];
  int r;
  int wstat;
  
  update_status();
  
  for (;;) {
    /* Return from the main loop when flagged to exit and there is no child
     * process still running. */
    if (status.flag_exit && !status.pid)
      return;
      
    /* Wait for either a command received on the control fifo, child process
     * terminating, or the timeout to be reached. */
    sig_unblock(SIGCHLD);
    x[0].fd = status.pipe[0];
    x[0].events = POLLIN;
    x[0].revents = 0;
    x[1].fd = status.fd_control;
    x[1].events = POLLIN;
    x[1].revents = 0;
    poll(x, 2, 3600000);
    sig_block(SIGCHLD);
    
    /* Discard any data that's been written to the pipe, it's only purpose is
     * to break out of poll(). */
    while (read(status.pipe[0], &ch, 1) == 1);
    
    /* Process the child process terminating, update the supervise status file
     * and exit if requested to. */
    for (;;) {
      r = waitpid(-1, &wstat, WNOHANG);
      if (!r)
        break;
      if ((r == -1) && (errno != EINTR))
        break;
      if (r == status.pid) {
        status.pid = 0;
        update_status();
        if (status.flag_exit)
          return;
        if (status.flag_want && !status.flag_down)
          start();
        break;
      }
    }
    
    /* Handle commands written to the control fifo. */
    if (read(status.fd_control, &ch, 1) == 1)
      switch (ch) {
        case SVCONTROL_DOWN:
          status.flag_want = 1;
          status.flag_down = 1;
          if (status.pid) {
            kill(status.pid, SIGTERM);
            kill(status.pid, SIGCONT);
            status.flag_paused = 0;
          }
          update_status();
          break;
        case SVCONTROL_UP:
          status.flag_want = 1;
          status.flag_down = 0;
          update_status();
          if (!status.pid)
            start();
          break;
        case SVCONTROL_ONCE:
          status.flag_want = 0;
          update_status();
          if (!status.pid)
            start();
          break;
        case SVCONTROL_SIGALRM:
          if (status.pid)
            kill(status.pid, SIGALRM);
          break;
        case SVCONTROL_SIGHUP:
          if (status.pid)
            kill(status.pid, SIGHUP);
          break;
        case SVCONTROL_SIGKILL:
          if (status.pid)
            kill(status.pid, SIGKILL);
          break;
        case SVCONTROL_SIGTERM:
          if (status.pid)
            kill(status.pid, SIGTERM);
          break;
        case SVCONTROL_SIGINT:
          if (status.pid)
            kill(status.pid, SIGINT);
          break;
        case SVCONTROL_SIGQUIT:
          if (status.pid)
            kill(status.pid, SIGQUIT);
          break;
        case SVCONTROL_SIGUSR1:
          if (status.pid)
            kill(status.pid, SIGUSR1);
          break;
        case SVCONTROL_SIGUSR2:
          if (status.pid)
            kill(status.pid, SIGUSR2);
          break;
        case SVCONTROL_SIGSTOP:
          status.flag_paused = 1;
          update_status();
          if (status.pid)
            kill(status.pid, SIGSTOP);
          break;
        case SVCONTROL_SIGCONT:
          status.flag_paused = 0;
          update_status();
          if (status.pid)
            kill(status.pid, SIGCONT);
          break;
        case SVCONTROL_EXIT:
          status.flag_exit = 1;
          update_status();
          break;
      }
  }
}
