/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <errno.h>
#include <sv.h>
#include <unistd.h>
#include <sv/bio.h>
#include <sv/scan.h>
#include <sv/sgetopt.h>
#include <sv/strerr.h>

#define DEFAULT_RESTART     1
#define DEFAULT_RESTART_LOG 0
#define DEFAULT_TIMEOUT     0
#define DEFAULT_VERBOSITY   1
#define FATAL               "svrestart: fatal: "

#define stdout_log(s)      if(config.verbosity >= 2) \
                             bio_put_str(bio_1, s); \
                             bio_flush(bio_1);
#define stdout_log3(a,b,c) if(config.verbosity >= 2) { \
                             bio_put_str(bio_1, a); \
                             bio_put_str(bio_1, b); \
                             bio_put_str(bio_1, c); \
                             bio_flush(bio_1); }
#define stdout_log4(a,b,c,d) if(config.verbosity >= 2) { \
                               bio_put_str(bio_1, a); \
                               bio_put_str(bio_1, b); \
                               bio_put_str(bio_1, c); \
                               bio_put_str(bio_1, d); \
                               bio_flush(bio_1); }

struct svstart_config {
  int            restart;     /**< restart service */
  int            restart_log; /**< restart log service */
  unsigned long  timeout;     /**< timeout in seconds */
  char          *timeout_str; /**< timeout string */
  int verbosity;              /**< verbosity level */
} config;

void usage() {
  strerr_die1x(100,
               "\n"
               "svrestart: usage: svrestart [ options ] path [ ... pathN ]\n"
               "\n"
               "  -h  help\n"
               "  -L  only restart log\n"
               "  -l  restart log\n"
               "  -Q  print errors     (default)\n"
               "  -q  quiet\n"
               "  -t  timeout          (default: 0)\n"
               "  -v  verbose\n"
               "\n"
               "  timeout: number of seconds to wait before sending 'kill' message.  if 0 then\n"
               "           svstop blocks forever until the service stops.\n"
               "  path:    path to the service directory.\n"
              );
}

void config_init() {
  config.restart     = DEFAULT_RESTART;
  config.restart_log = DEFAULT_RESTART_LOG;
  config.timeout     = DEFAULT_TIMEOUT;
  config.verbosity   = DEFAULT_VERBOSITY;
}

int main(int argc, char **argv, char **envp) {
  int opt;
  char *path;
  int path_len;
  char *path_log;
  int path_log_size;
  sv_stat_t svst;
  
  config_init();
  
  while ((opt = sgetopt(argc, argv, "LlQqt:v")) != -1) {
    switch (opt) {
      case 'h':
        usage();
      case 'L':
        config.restart = 0;
        config.restart_log = 1;
        break;
      case 'l':
        config.restart_log = 1;
        break;
      case 'Q':
        config.verbosity = 1;
        break;
      case 'q':
        config.verbosity = 0;
        break;
      case 't':
        config.timeout_str = soptarg;
        if (soptarg[scan_ulong(&config.timeout, soptarg)])
          strerr_die3x(111, FATAL, "invalid timeout ", soptarg);
        break;
      case 'v':
        config.verbosity = 2;
        break;
      default:
        usage();
    }
  }
  argv += soptind;
  
  if (!*argv)
    usage();
    
  /* Disable buffer_2 so nothing gets outputted. */
  if (!config.verbosity)
    bio_2->fd = -1;
    
  while ((path = *argv++)) {
    if (config.restart) {
      stdout_log3("Stopping ", path, " ... ");
      if (sv_stop(path, config.timeout)) {
        stdout_log("failed\n");
        strerr_die2x(100, FATAL, sv_error_str(errno));
      }
      stdout_log("done\n");
    }
    
    if (config.restart_log) {
      stdout_log4("Stopping ", path, (path[strlen(path) - 1] == '/' ? "" : "/"),
                  "log ... ");
      if (sv_stop_log(path, config.timeout)) {
        stdout_log("failed\n");
        strerr_die2x(100, FATAL, sv_error_str(errno));
      }
      stdout_log("done\n");
      
      stdout_log4("Starting ", path, (path[strlen(path) - 1] == '/' ? "" : "/"),
                  "log ... ");
      if (sv_start_log(path)) {
        stdout_log("failed\n");
        strerr_die2x(100, FATAL, sv_error_str(errno));
      }
      stdout_log("done\n");
    }
    
    if (config.restart) {
      stdout_log3("Starting ", path, " ... ");
      if (sv_start(path)) {
        stdout_log("failed\n");
        strerr_die2x(100, FATAL, sv_error_str(errno));
      }
      stdout_log("done\n");
    }
  }
  
  _exit(0);
}
