/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <dirent.h>
#include <stdlib.h>
#include <unistd.h>
#include <sv/array.h>
#include <sv/scan.h>
#include <sv/strerr.h>
#include <sys/stat.h>
#include "svscan.h"

typedef struct node {
  struct node *left;
  struct node *right;
  int          priority;
  array_t      name;
} node_t;

/**
 * Current priority being started.  This is used to add a delay in service
 * startups.
 */

static int current_priority = 0;

/**
 * Walk through the tree and free resources allocated to each node.
 */

static void free_all(node_t *node) {
  node_t *right;
  
  if (!node)
    return;
  right = node->right;
  
  if (node->left)
    free_all(node->left);
  array_reset(&node->name);
  free(node);
  if (right)
    free_all(right);
}

/**
 * Add a node to the tree based on the priority and name.
 */

static void node_add(node_t **tree, node_t *node) {
  if (!(*tree)) {
    *tree = node;
    return;
  }
  
  if (node->priority == (*tree)->priority) {
    if (strcmp(array_start(&node->name), array_start(&(*tree)->name)) < 0)
      node_add(&(*tree)->left, node);
    else if (strcmp(array_start(&node->name), array_start(&(*tree)->name)) > 0)
      node_add(&(*tree)->right, node);
  } else if (node->priority < (*tree)->priority)
    node_add(&(*tree)->left, node);
  else if (node->priority > (*tree)->priority)
    node_add(&(*tree)->right, node);
}

/**
 * Walk through the tree and start each of the services.
 */

static void start_all(node_t *node) {
  if (!node)
    return;
    
  if (node->left)
    start_all(node->left);
  if (node->priority > current_priority)
    sleep(1);
  current_priority = node->priority;
  start(array_start(&node->name));
  if (node->right)
    start_all(node->right);
}

/**
 * Walk through the main service directory and determine the priority of each
 * service and start them in the desired order.
 */

void first_start() {
  DIR *dir;
  struct dirent *entry;
  node_t *node;
  array_t path = ARRAY_INIT(1);
  unsigned long priority;
  array_t priority_arr = ARRAY_INIT(1);
  struct stat st;
  node_t *tree = 0;
  
  dir = opendir(".");
  if (!dir)
    direrror();
    
  /* Loop through each of the directory entries looking for the priority file
   * and add a new node to the tree with the service information. */
  for (;;) {
    priority = DEFAULT_PRIORITY;
    
    array_trunc(&path);
    array_trunc(&priority_arr);
    
    entry = readdir(dir);
    if (!entry)
      break;
    if (entry->d_name[0] == '.')
      continue;
      
    array_append(&path, entry->d_name, strlen(entry->d_name));
    array_append(&path, "/priority", strlen("/priority"));
    array_append_null(&path);
    if (stat(array_start(&path), &st) == 0) {
      array_read_fn(&priority_arr, array_start(&path), 8);
      scan_ulong(&priority, array_start(&priority_arr));
    }
    
    node = (node_t *)malloc(sizeof(node_t));
    array_init(&node->name, 1);
    node->left = 0;
    node->right = 0;
    node->priority = priority;
    array_append(&node->name, entry->d_name, strlen(entry->d_name));
    array_append_null(&node->name);
    node_add(&tree, node);
  }
  closedir(dir);
  
  start_all(tree);
  free_all(tree);
}
