/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <dirent.h>
#include <errno.h>
#include <unistd.h>
#include <sv/scan.h>
#include <sv/sgetopt.h>
#include <sv/strerr.h>
#include <sys/wait.h>
#include "svscan.h"

struct svscan_config config;
struct service services[SERVICES_MAX];
int services_count = 0;

static void usage() {
  strerr_die1x(100,
               "\n"
               "svscan: usage: svscan [ options ] dir\n"
               "\n"
               "  -i  check interval  (default: 5)\n"
              );
}

static void config_init() {
  config.interval = 5;
  config.interval_str = 0;
}

int main(int argc, char **argv) {
  DIR *dir;
  char *dirname;
  struct dirent *entry;
  int i;
  int opt;
  int r;
  int wstat;
  
  config_init();
  
  while ((opt = sgetopt(argc, argv, "i:")) != -1) {
    switch (opt) {
      case 'i':
        config.interval_str = soptarg;
        if (soptarg[scan_ulong(&config.interval, soptarg)])
          strerr_die3x(111, FATAL, "invalid interval ", soptarg);
        break;
      default:
        usage();
    }
  }
  argv += soptind;
  
  if (!*argv)
    usage();
  dirname = *argv;
  
  if (chdir(dirname) == -1)
    strerr_die4sys(111, FATAL, "unable to chdir to ", dirname, ": ");
    
  /* When svscan starts up it reads the 'priority' file within the service
   * directory and attempts to start the services in the correct order. */
  first_start();
  
  for (;;) {
    /* Check to see if any child processes (services / log services) have
     * terminated and if so set it's pid to 0 in the services array. They'll
     * get cleaned up later on. */
    for (;;) {
      r = waitpid(-1, &wstat, WNOHANG);
      if (!r)
        break;
      if (r == -1) {
        if (errno == EINTR)
          continue;
        break;
      }
      
      for (i = 0; i < services_count; ++i) {
        if (services[i].pid == r) {
          services[i].pid = 0;
          break;
        }
        if (services[i].pidlog == r) {
          services[i].pidlog = 0;
          break;
        }
      }
    }
    
    /* Attempt to open the main services directory and if it fails for any
     * reason then warn of the error and continue on, otherwise process it. */
    dir = opendir(".");
    if (!dir) {
      direrror();
    } else {
      /* Read each entry in the directory and blindly pass it to the start
       * method which will decide if the entry is a service directory, handle
       * starting the new service, or restarting a failed service. */
      for (;;) {
        errno = 0;
        entry = readdir(dir);
        if (!entry)
          break;
        start(entry->d_name);
      }
      if (errno)
        direrror();
        
      /* Now that the directory has been scanned, be nice and close it. */
      closedir(dir);
    }
    
    /* Loop through the services array and perform any cleanup operations that
     * are required and remove recently terminated processes from the list. */
    i = 0;
    while (i < services_count) {
      if (!services[i].pid && !services[i].pidlog) {
        if (services[i].flaglog) {
          close(services[i].pi[0]);
          close(services[i].pi[1]);
          services[i].flaglog = 0;
        }
        services[i] = services[--services_count];
        continue;
      }
      ++i;
    }
    
    sleep(config.interval);
  }
}
