/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <errno.h>
#include <string.h>
#include <sv/file.h>
#include <sv/pathexec.h>
#include <sv/strerr.h>
#include <sys/stat.h>
#include "svscan.h"

void start(char *fn) {
  unsigned int fnlen;
  char fnlog[260];
  struct stat st;
  int child;
  int i;
  char *args[3];
  
  /* Ignore hidden paths. */
  if (fn[0] == '.')
    return;
    
  /* Only process directories. */
  if (stat(fn, &st) == -1) {
    strerr_warn4sys(WARNING, "unable to stat ", fn, ": ");
    return;
  }
  if ((st.st_mode & S_IFMT) != S_IFDIR)
    return;
    
  /* Scan through the services array to see if there's already an entry for
   * the service in it and setting i to its array element. */
  for (i = 0; i < services_count; ++i)
    if (services[i].ino == st.st_ino)
      if (services[i].dev == st.st_dev)
        break;
        
  /* For new service we handle updating the services the array with the new
   * service entry and for fail services we skip this and go straight to
   * starting it again. */
  if (i == services_count) {
    if (services_count >= SERVICES_MAX) {
      strerr_warn4x(WARNING, "unable to start ", fn, ": running too many services");
      return;
    }
    
    services[i].ino = st.st_ino;
    services[i].dev = st.st_dev;
    services[i].pid = 0;
    services[i].pidlog = 0;
    services[i].flaglog = 0;
    
    /* Check to see if the service has a log service associated with it and
     * flag the entry for it. */
    fnlen = strlen(fn);
    if (fnlen <= 255) {
      memcpy(fnlog, fn, fnlen);
      memcpy(fnlog + fnlen, "/log", 5);
      if (stat(fnlog, &st) == 0)
        services[i].flaglog = 1;
      else {
        if (errno != ENOENT) {
          strerr_warn4sys(WARNING, "unable to stat ", fn, "/log: ");
          return;
        }
      }
    }
    
    /* If configured with a log service create the neccessary pipes. */
    if (services[i].flaglog) {
      if (pipe(services[i].pi) == -1) {
        strerr_warn4sys(WARNING, "unable to create pipe for ", fn, ": ");
        return;
      }
      close_on_exec(services[i].pi[0]);
      close_on_exec(services[i].pi[1]);
    }
    ++services_count;
  }
  
  /* New services and failed services alike will both have their service entry
   * pid set to zero, so check for it and start the service if needed. */
  if (!services[i].pid) {
    switch (child = fork()) {
      case -1:
        strerr_warn4sys(WARNING, "unable to fork for ", fn, ": ");
        return;
      case 0:
        /* Services that have a log service assocated with them need to have
         * their stdout configured to write to its log pipe. */
        if (services[i].flaglog)
          if (fd_move(1, services[i].pi[1]) == -1)
            strerr_die4sys(111, WARNING, "unable to set up descriptors for ", fn, ": ");
        args[0] = "supervise";
        args[1] = fn;
        args[2] = 0;
        pathexec_run(*args, args, environ);
        strerr_die4sys(111, WARNING, "unable to start supervise ", fn, ": ");
      default:
        services[i].pid = child;
    }
  }
  
  /* If the service is configured to have a log service and it's not running,
   * start it. */
  if (services[i].flaglog && !services[i].pidlog) {
    switch (child = fork()) {
      case -1:
        strerr_warn4sys(WARNING, "unable to fork for ", fn, "/log: ");
        return;
      case 0:
        /* Log services need to have their stdin configured to read from the
         * service pipe. */
        if (fd_move(0, services[i].pi[0]) == -1)
          strerr_die4sys(111, WARNING, "unable to set up descriptors for ", fn, "/log: ");
        if (chdir(fn) == -1)
          strerr_die4sys(111, WARNING, "unable to switch to ", fn, ": ");
        args[0] = "supervise";
        args[1] = "log";
        args[2] = 0;
        pathexec_run(*args, args, environ);
        strerr_die4sys(111, WARNING, "unable to start supervise ", fn, "/log: ");
      default:
        services[i].pidlog = child;
    }
  }
}
