/*
 * Copyright (c) 2011-2015 David Shane Holden. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <errno.h>
#include <sv.h>
#include <unistd.h>
#include <sv/bio.h>
#include <sv/sgetopt.h>
#include <sv/strerr.h>

#define DEFAULT_START     1
#define DEFAULT_START_LOG 0
#define DEFAULT_VERBOSITY 1
#define FATAL             "svstart: fatal: "

#define stdout_log(s)      if(config.verbosity >= 2) \
                             bio_put_str(bio_1, s); \
                             bio_flush(bio_1);
#define stdout_log3(a,b,c) if(config.verbosity >= 2) { \
                             bio_put_str(bio_1, a); \
                             bio_put_str(bio_1, b); \
                             bio_put_str(bio_1, c); \
                             bio_flush(bio_1); }
#define stdout_log4(a,b,c,d) if(config.verbosity >= 2) { \
                               bio_put_str(bio_1, a); \
                               bio_put_str(bio_1, b); \
                               bio_put_str(bio_1, c); \
                               bio_put_str(bio_1, d); \
                               bio_flush(bio_1); }

struct svstart_config {
  int start;     /**< start service */
  int start_log; /**< start log service */
  int verbosity; /**< verbosity level */
} config;

void usage() {
  strerr_die1x(100,
               "\n"
               "svstart: usage: svstart [ options ] path [ ... pathN ]\n"
               "\n"
               "  -h  help\n"
               "  -L  only start log\n"
               "  -l  start log\n"
               "  -Q  print errors   (default)\n"
               "  -q  quiet\n"
               "  -v  verbose\n"
               "\n"
               "  path: path to the service directory.\n"
              );
}

void config_init() {
  config.start     = DEFAULT_START;
  config.start_log = DEFAULT_START_LOG;
  config.verbosity = DEFAULT_VERBOSITY;
}

int main(int argc, char **argv, char **envp) {
  int opt;
  char *path;
  int path_len;
  char *path_log;
  int path_log_size;
  int r;
  sv_stat_t svst;
  
  config_init();
  
  while ((opt = sgetopt(argc, argv, "LlQqv")) != -1) {
    switch (opt) {
      case 'h':
        usage();
      case 'L':
        config.start = 0;
        config.start_log = 1;
        break;
      case 'l':
        config.start_log = 1;
        break;
      case 'Q':
        config.verbosity = 1;
        break;
      case 'q':
        config.verbosity = 0;
        break;
      case 'v':
        config.verbosity = 2;
        break;
      default:
        usage();
    }
  }
  argv += soptind;
  
  if (!*argv)
    usage();
    
  /* Disable buffer_2 so nothing gets outputted. */
  if (!config.verbosity)
    bio_2->fd = -1;
    
  while ((path = *argv++)) {
    if (config.start_log) {
      stdout_log4("Starting ", path, (path[strlen(path) - 1] == '/' ? "" : "/"),
                  "log ... ");
                  
      r = sv_start_log(path);
      if (r) {
        stdout_log("failed\n");
        strerr_die2x(100, FATAL, sv_error_str(r));
      }
      stdout_log("done\n");
    }
    
    if (config.start) {
      stdout_log3("Starting ", path, " ... ");
      r = sv_start(path);
      if (r) {
        stdout_log("failed\n");
        strerr_die2x(100, FATAL, sv_error_str(r));
      }
      stdout_log("done\n");
    }
  }
  
  _exit(0);
}
