/* -*- coding: utf-8; tab-width: 2; indent-tabs-mode: t -*- */
/*
 *   Copyright 2013 Daniel Trebbien
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

#pragma once
#ifndef VERLIBC_STRING_H
#define VERLIBC_STRING_H 1
#include <stddef.h>

#ifdef VERIFY
#define VCC_ONLY_NEW_SYNTAX
#include <vcc.h>

#define __restrict
#else
#define _(...)
#endif

#ifdef __cplusplus
extern "C" {
#endif

_(pure) size_t strlen(const char *str _(ghost size_t nul_lim))
	_(requires \thread_local_array(str, nul_lim + 1))
	_(requires \exists size_t i; i <= nul_lim && str[i] == '\0')
	_(reads \array_range(str, nul_lim + 1))
	_(ensures \result <= nul_lim)
	_(ensures str[\result] == '\0' && \forall size_t j; j < \result ==> str[j] != '\0')
	_(ensures \forall size_t j; j <= \result && (\forall size_t i; i < j ==> str[i] != '\0') ==> \result == j + strlen(str + j, nul_lim - j))
	;

/**
 * \brief Copies the first \p n bytes of the object pointed to by \p sv to the
 *   first \p n bytes of the object pointed to by \p dv.
 *
 * The memory regions [\p sv, \p sv + \p n) and [\p dv, \p dv + \p n) can not overlap.
 *
 * \returns \p dv
 * \see memmove
 * \see 7.21.2.1 The \c memcpy function
 */
void * memcpy(void *__restrict dv, const void *__restrict sv, size_t n)
	_(requires \thread_local_array((const char *) sv, n))
	_(requires \arrays_disjoint((const char *) sv, n, (char *) dv, n))
	_(writes \array_range((char *) dv, n))
	_(ensures \result == dv)
	_(ensures \forall size_t j; j < n ==> ((char *) dv)[j] == ((const char *) sv)[j])
	;

/**
 * \brief Copies the first \p n bytes of the object pointed to by \p sv to the
 *   first \p n bytes of the object pointed to by \p dv.
 *
 * Copying takes place as if the first \p n bytes of the object pointed to by
 * \p sv are first copied into a thread-local temporary buffer of \p n bytes, and then
 * the temporary buffer is copied to the first \p n bytes of the object pointed
 * to by \p dv. This allows the memory regions [\p sv, \p sv + \p n) and
 * [\p dv, \p dv + \p n) to overlap.
 *
 * \returns \p dv
 * \see memcpy
 * \see 7.21.2.2 The \c memmove function
 */
void * memmove(void *dv, const void *sv, size_t n)
	_(requires \thread_local_array((const char *) sv, n))
	_(writes \array_range((char *) dv, n))
	_(ensures \result == dv)
	_(ensures \forall size_t j; j < n ==> ((char *) dv)[j] == \old(((const char *) sv)[j]))
	_(ensures \arrays_disjoint((const char *) sv, n, (char *) dv, n) ==> (\forall size_t j; j < n ==> \unchanged(((const char *) sv)[j])))
	;

/**
 * \brief Copies the string pointed to by \p s (including the NUL terminator)
 *   to the array pointed to by \p d.
 *
 * The memory regions [\p s, <code>s + strlen(s)</code>] and
 * [\p d, <code>d + strlen(s)</code>] can not overlap.
 *
 * \returns \p d
 * \see strncpy
 * \see 7.21.2.3 The \c strcpy function
 */
char * strcpy(char *__restrict d, const char *__restrict s,
		_(ghost size_t s_nul_pos))
	_(requires \thread_local_array(s, s_nul_pos + 1))
	_(requires s[s_nul_pos] == '\0')
	_(requires \arrays_disjoint(s, s_nul_pos + 1, d, s_nul_pos + 1))
	_(writes \array_range(d, s_nul_pos + 1))
	_(ensures \result == d)
	_(ensures \forall size_t j; j <= strlen(s, s_nul_pos) ==> d[j] == s[j])
	_(ensures \forall size_t j; strlen(s, s_nul_pos) < j && j < s_nul_pos + 1 ==> \unchanged(d[j]))
	;

/**
 * \brief Copies at most the first \p n non-NUL characters of the string pointed to by
 *   \p s to the array pointed to by \p d.
 *
 * If the string pointed to by \p s is shorter than \p n characters, then NUL
 * characters are appended to the copy in the array pointed to by \p d until
 * exactly \p n characters have been written.
 *
 * The memory regions [\p d, <code>d + n</code>) and [\p s, <code>s + strlen(s)</code>]
 * can not overlap.
 *
 * If the string pointed to by \p s has \p n or more characters, then \p d is not
 * NUL-terminated.
 *
 * \returns \p d
 * \see strcpy
 * \see 7.21.2.4 The \c strncpy function
 */
char * strncpy(char *__restrict d, const char *__restrict s, size_t n
		_(ghost size_t s_nul_pos))
	_(requires \thread_local_array(s, s_nul_pos + 1))
	_(requires s[s_nul_pos] == '\0')
	_(requires \arrays_disjoint(d, n, s, s_nul_pos + 1))
	_(writes \array_range(d, n))
	_(ensures \result == d)
	_(ensures \forall size_t j; j < strlen(s, s_nul_pos) && j < n ==> d[j] == s[j])
	_(ensures \forall size_t j; strlen(s, s_nul_pos) <= j && j < n ==> d[j] == '\0')
	;

/**
 * \brief Appends a copy of the string pointed to by \p str to the end of the
 *   string pointed to by \p b.
 *
 * The initial character of \p str, if any, overwrites the NUL terminator at the
 * end of \p b.
 *
 * The memory regions [\p b, <code>b + strlen(b) + strlen(str)</code>] and
 * [\p str, <code>str + strlen(str)</code>] can not overlap.
 *
 * \returns \p b
 * \see strncat
 * \see 7.21.3.1 The \c strcat function
 */
char * strcat(char *__restrict b, const char *__restrict str
		_(ghost size_t b_nul_pos) _(ghost size_t str_nul_pos))
	_(requires b_nul_pos + str_nul_pos <= ((size_t ) -1))
	_(requires \thread_local_array(str, str_nul_pos + 1))
	_(requires str[str_nul_pos] == '\0')
	_(requires b[b_nul_pos] == '\0')
	_(requires \arrays_disjoint(b, b_nul_pos + str_nul_pos + 1, str, str_nul_pos + 1))
	_(writes \array_range(b, b_nul_pos + str_nul_pos + 1))
	_(ensures \result == b)
	_(ensures \forall size_t j; j < \old(strlen(b, b_nul_pos)) ==> \unchanged(b[j]))
	_(ensures \forall size_t j; \old(strlen(b, b_nul_pos)) <= j && j <= \old(strlen(b, b_nul_pos)) + strlen(str, str_nul_pos) ==> b[j] == str[j - \old(strlen(b, b_nul_pos))])
	_(ensures \forall size_t j; \old(strlen(b, b_nul_pos)) + strlen(str, str_nul_pos) < j && j < b_nul_pos + str_nul_pos + 1 ==> \unchanged(b[j]))
	_(ensures strlen(b, b_nul_pos + str_nul_pos) == \old(strlen(b, b_nul_pos)) + strlen(str, str_nul_pos))
	;

/**
 * \brief Appends a copy of at most the first \p n non-NUL characters of the
 *   string pointed to by \p str to the end of the string pointed to by \p b.
 *
 * The initial character of \p str, if any, overwrites the NUL terminator at the
 * end of \p b.
 *
 * A terminating NUL character is always appended to \p b. Thus, there must be
 * space in \p b for <code>strlen(b) + n + 1</code> characters.
 *
 * The memory regions [\p b, <code>b + strlen(b) + MIN(n, strlen(str)) + 1</code>) and
 * [\p str, <code>str + MIN(n, strlen(str))</code>] can not overlap.
 *
 * \returns \p b
 * \see strcat
 * \see 7.21.3.2 The \c strncat function
 */
char * strncat(char *__restrict b, const char *__restrict str, size_t n
		_(ghost size_t b_nul_pos))
	_(requires b_nul_pos + n + 1 <= ((size_t) -1))
	_(requires \thread_local_array(str, n))
	_(requires b[b_nul_pos] == '\0')
	_(requires \arrays_disjoint(b, b_nul_pos + n + 1, str, n))
	_(writes \array_range(b, b_nul_pos + n + 1))
	_(ensures \result == b)
	_(ensures \forall size_t j; j < \old(strlen(b, b_nul_pos)) ==> \unchanged(b[j]))
	_(ensures \forall size_t k; k < n && (\forall size_t j; j < k ==> str[j] != '\0') ==> (\forall size_t j; \old(strlen(b, b_nul_pos)) <= j && j <= \old(strlen(b, b_nul_pos)) + k ==> b[j] == str[j - \old(strlen(b, b_nul_pos))])) /* k is a number less than n such that each of the first k characters of the string pointed to by str is non-NUL. For each such k, \forall size_t j; \old(strlen(b, b_nul_pos)) <= j && j <= \old(strlen(b, b_nul_pos)) + k ==> b[j] == str[j - \old(strlen(b, b_nul_pos))], or the first k + 1 bytes of str were appended to the end of b. */
	_(ensures (\forall size_t j; j < n ==> str[j] != '\0') ==> b[\old(strlen(b, b_nul_pos)) + n] == '\0' && (\forall size_t j; \old(strlen(b, b_nul_pos)) <= j && j < \old(strlen(b, b_nul_pos)) + n ==> b[j] == str[j - \old(strlen(b, b_nul_pos))]))
	_(ensures \forall size_t str_nul_pos; str_nul_pos < n && str[str_nul_pos] == '\0' ==> (\forall size_t j; \old(strlen(b, b_nul_pos)) + str_nul_pos < j && j < b_nul_pos + n + 1 ==> \unchanged(b[j])))
	_(ensures \forall size_t j; j < n && str[j] == '\0' && (\forall size_t k; k < j ==> str[k] != '\0') ==> strlen(\result, \old(strlen(b, b_nul_pos)) + j) == \old(strlen(b, b_nul_pos)) + j)
	_(ensures (\forall size_t j; j < n ==> str[j] != '\0') ==> strlen(\result, \old(strlen(b, b_nul_pos)) + n) == \old(strlen(b, b_nul_pos)) + n)
	;

//memcmp
//strcmp
//strcoll
//strncmp
//strxfrm

/**
 * \brief Locates the first occurrence of \p c (converted to an <code>unsigned char</code> object)
 *   in the initial \p n characters (each interpreted as an <code>unsigned char</code> object) of the
 *   object pointed to by \p bv.
 *
 * \returns \c NULL if \p c does not occur in the first \p n characters of the
 *   object pointed to by \p bv;  otherwise, a pointer to the first occurrence.
 * \see 7.21.5.1 The \c memchr function
 */
_(pure) void * memchr(const void *bv, int c, size_t n)
	_(requires \forall size_t i; i < n ==> ((unsigned char *) bv) + i != 0) // TODO http://vcc.codeplex.com/discussions/433395
	_(requires \thread_local_array((const unsigned char *) bv, n))
	_(requires 0 <= c && c <= ((unsigned char) -1))
	_(reads \universe()) // http://vcc.codeplex.com/discussions/252343#post591593
	_(ensures ((unsigned char *) \result) == 0 <==> (\forall size_t j; j < n ==> ((const unsigned char *) bv)[j] != ((unsigned char) c)))
	_(ensures ((unsigned char *) \result) != 0 ==> ((\forall size_t j; j < ((size_t) (((unsigned char *) \result) - ((const unsigned char *) bv))) ==> ((const unsigned char *) bv)[j] != ((unsigned char) c)) && *((unsigned char *) \result) == ((unsigned char) c)))
	_(ensures ((unsigned char *) \result) != 0 ==> (\forall size_t k; k < n && (\forall size_t j; j < k ==> ((const unsigned char *) bv)[j] != ((unsigned char) c)) && ((const unsigned char *) bv)[k] == ((unsigned char) c) ==> ((unsigned char *) \result) == ((const unsigned char *) bv) + k))
	;

/**
 * \brief Locates the first occurrence of \p c (converted to a <code>char</code> object)
 *   in the string pointed to by \p str.
 *
 * The terminating NUL character of \p str is considered part of the string.
 *
 * \returns \c NULL if \p c does not occur in the string pointed to by \p str
 *   and is not <code>'\0'</code>;  otherwise, a pointer to the first occurrence.
 * \see strrchr
 * \see strpbrk
 * \see 7.21.5.2 The \c strchr function
 */
_(pure) char * strchr(const char *str, int c _(ghost size_t nul_pos))
	_(requires \forall size_t i; i < nul_pos + 1 ==> str + i != 0) // TODO http://vcc.codeplex.com/discussions/433395
	_(requires \thread_local_array(str, nul_pos + 1))
	_(requires str[nul_pos] == '\0')
	_(requires -128 <= c && c <= 127) // http://vcc.codeplex.com/discussions/252475
	_(reads \universe()) // http://vcc.codeplex.com/discussions/252343#post591593
	_(ensures \result == 0 <==> (\forall size_t j; j <= strlen(str, nul_pos) ==> str[j] != (char) c))
	_(ensures \result != 0 ==> (\forall size_t j; j < ((size_t) (\result - str)) ==> str[j] != (char) c) && *\result == (char) c)
	_(ensures \result != 0 ==> (\forall size_t k; k <= strlen(str, nul_pos) && (\forall size_t j; j < k ==> str[j] != (char) c) && str[k] == (char) c ==> \result == str + k))
	;

/**
 * \brief Locates the first occurrence of one of the <code>char</code> objects in the
 *   string pointed to by \p search_chars in the string pointed to by \p str.
 *
 * The terminating NUL characters of \p str and \p search_chars are included in the search.
 * Thus if <code>str[result] == '\0'</code>, then \c result is equal to the length of
 * \p str.
 *
 * \returns The index of the first character in the string pointed to by \p str
 *   that is in the string pointed to by \p search_chars.
 * \see strpbrk
 * \see 7.21.5.3 The \c strcspn function
 */
_(pure) size_t strcspn(const char *str, const char *search_chars
		_(ghost size_t str_nul_pos) _(ghost size_t search_chars_nul_pos))
	_(requires \thread_local_array(str, str_nul_pos + 1))
	_(requires \thread_local_array(search_chars, search_chars_nul_pos + 1))
	_(requires str[str_nul_pos] == '\0')
	_(requires search_chars[search_chars_nul_pos] == '\0')
	_(reads \universe()) // http://vcc.codeplex.com/discussions/252343#post591593
	_(ensures \result <= strlen(str, str_nul_pos))
	_(ensures \exists size_t j; j <= strlen(search_chars, search_chars_nul_pos) && str[\result] == search_chars[j])
	_(ensures str[\result] == '\0' ==> \result == strlen(str, str_nul_pos))
	_(ensures \forall size_t l; l <= strlen(str, str_nul_pos) && (\forall size_t k, j; k < l && j <= strlen(search_chars, search_chars_nul_pos) ==> str[k] != search_chars[j]) && (\exists size_t j; j < strlen(search_chars, search_chars_nul_pos) && str[l] == search_chars[j]) ==> \result == l)
	;

/**
 * \brief Locates the first occurrence of one of the <code>char</code> objects in the
 *   string pointed to by \p search_chars in the string pointed to by \p str.
 *
 * \returns \c NULL if none of the characters in the string pointed to by \p search_chars
 *   are characters of the string pointed to by \p str;  otherwise, a pointer to the first character
 *   in the string pointed to by \p str that is in the string pointed to by \p search_chars.
 * \see strchr
 * \see strcspn
 * \see 7.21.5.4 The \c strpbrk function
 */
_(pure) char * strpbrk(const char *str, const char *search_chars
		_(ghost size_t str_nul_pos) _(ghost size_t search_chars_nul_pos))
	_(requires \forall size_t i; i < str_nul_pos + 1 ==> str + i != 0) // TODO http://vcc.codeplex.com/discussions/433395
	_(requires \thread_local_array(str, str_nul_pos + 1))
	_(requires \thread_local_array(search_chars, search_chars_nul_pos + 1))
	_(requires str[str_nul_pos] == '\0')
	_(requires search_chars[search_chars_nul_pos] == '\0')
	_(reads \universe()) // http://vcc.codeplex.com/discussions/252343#post591593
	_(ensures \result == 0 <==> (\forall size_t k, j; k < strlen(str, str_nul_pos) && j < strlen(search_chars, search_chars_nul_pos) ==> str[k] != search_chars[j]))
	_(ensures \result != 0 ==> (\forall size_t k, j; k < ((size_t) (\result - str)) && j < strlen(search_chars, search_chars_nul_pos) ==> str[k] != search_chars[j]) && (\exists size_t j; j < strlen(search_chars, search_chars_nul_pos) && *\result == search_chars[j]))
	_(ensures \result != 0 ==> (\forall size_t l; l < strlen(str, str_nul_pos) && (\forall size_t k, j; k < l && j < strlen(search_chars, search_chars_nul_pos) ==> str[k] != search_chars[j]) && (\exists size_t j; j < strlen(search_chars, search_chars_nul_pos) && str[l] == search_chars[j]) ==> \result == str + l))
	;

/**
 * \brief Locates the last occurrence of \p c (converted to a <code>char</code> object)
 *   in the string pointed to by \p str.
 *
 * The terminating NUL character of \p str is considered part of the string.
 *
 * \returns \c NULL if \p c does not occur in the string pointed to by \p str
 *   and is not <code>'\0'</code>;  otherwise, a pointer to the last occurrence.
 * \see strchr
 * \see 7.21.5.5 The \c strrchr function
 */
_(pure) char * strrchr(const char *str, int c _(ghost size_t nul_pos))
	_(requires \forall size_t i; i < nul_pos + 1 ==> str + i != 0) // TODO http://vcc.codeplex.com/discussions/433395
	_(requires \thread_local_array(str, nul_pos + 1))
	_(requires str[nul_pos] == '\0')
	_(requires -128 <= c && c <= 127) // http://vcc.codeplex.com/discussions/252475
	_(reads \universe()) // http://vcc.codeplex.com/discussions/252343#post591593
	_(ensures \result == 0 <==> (\forall size_t j; j <= strlen(str, nul_pos) ==> str[j] != (char) c))
	_(ensures \result != 0 ==> (*\result == '\0' ==> ((size_t) (\result - str)) == strlen(str, nul_pos)))
	_(ensures \result != 0 ==> (\forall size_t j; ((size_t) (\result - str)) < j && j <= strlen(str, nul_pos) ==> str[j] != (char) c) && *\result == (char) c)
	_(ensures \result != 0 ==> (\forall size_t k; k <= strlen(str, nul_pos) && (\forall size_t j; k < j && j <= strlen(str, nul_pos) ==> str[j] != (char) c) && str[k] == (char) c ==> \result == str + k))
	;

//strspn
//strstr
//strtok

/**
 * \brief Copies \p c \p n times to the object pointed to by \p dv.
 *
 * Copies the value \p c converted to an <code>unsigned char</code> to the first
 * \p n characters of the object pointed to by \p dv.
 *
 * \returns \p dv
 * \see 7.21.6.1 The \c memset function
 */
void * memset(void *dv, int c, size_t n)
	_(requires 0 <= c && c <= ((unsigned char) -1))
	_(writes \array_range((unsigned char *) dv, n))
	_(maintains \mutable_array((unsigned char *) dv, n))
	_(ensures \result == dv)
	_(ensures \forall size_t j; j < n ==> ((unsigned char *) dv)[j] == (unsigned char) c)
	;

//strerror

/**
 * \fn strlen
 * \brief Computes the length of the string pointed to by \p str.
 *
 * \returns The number of characters that precede the terminating NUL character.
 * \see 7.21.6.3 The \c strlen function
 */

/* =============================== Extensions =============================== */

/**
 * \brief Copies up to <code>dest_len - 1</code> bytes from the string pointed to by \p src
 *   to the buffer pointed to by \p dest.
 *
 * Copies the string pointed to by \p src to the output buffer pointed to by \p dest, which has
 * space for \p dest_len bytes. If \p dest_len is 0, then the function simply returns
 * <code>strlen(src)</code>. Otherwise, at most <code>dest_len - 1</code> characters are
 * copied and \p dest is always NUL-terminated.
 *
 * \returns the length of the string pointed to by \p src. If the return value is greater than
 *   or equal to \p dest_len, then truncation occurred.
 */
size_t strlcpy(char *__restrict dest, const char *__restrict src, size_t dest_len _(ghost size_t src_nul_lim))
	_(requires \thread_local_array(src, src_nul_lim + 1))
	_(requires \exists size_t i; i <= src_nul_lim && src[i] == '\0')
	_(requires \arrays_disjoint(dest, dest_len, src, src_nul_lim + 1))
	_(writes \array_range(dest, dest_len))
	_(ensures \result == strlen(src, src_nul_lim))
	_(ensures dest_len > 0 ==> (\forall size_t j; j < \result && j < dest_len - 1 ==> dest[j] == \old(src[j])))
	_(ensures dest_len > 0 && dest_len > \result ==> dest[\result] == '\0')
	_(ensures dest_len > 0 && \result >= dest_len ==> dest[dest_len - 1] == '\0')
	_(ensures \forall size_t j; j > \result && j < dest_len ==> \unchanged(dest[j]))
	;

#ifdef __cplusplus
}
#endif

#endif
