/* -*- coding: utf-8; tab-width: 2; indent-tabs-mode: t -*- */
/*
 *   Copyright 2013 Daniel Trebbien
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

#include <verlibc/string.h>

size_t strlcpy(char *__restrict dest, const char *__restrict src, size_t dest_len _(ghost size_t src_nul_lim))
{
	if (dest_len == 0)
	{
		return strlen(src _(ghost src_nul_lim));
	}
	else
	{
		char *const dest_str_end = dest + dest_len - 1;
		char *__restrict d = dest;
		const char *__restrict s = src;
		_(ghost size_t i = 0;)

		while (d != dest_str_end)
			_(invariant d == dest + i)
			_(invariant s == src + i)
			_(invariant i < dest_len)
			_(invariant i <= src_nul_lim)
			_(invariant \forall size_t j; j < i ==> src[j] != '\0')
			_(invariant \forall size_t j; j < i ==> dest[j] == \old(src[j]))
			_(invariant \forall size_t j; j >= i ==> \unchanged(dest[j]))
		{
			if ((*d++ = *s++) == '\0')
			{
				_(assert *(s - 1) == '\0')
				_(assert *(d - 1) == '\0')
				return (d - 1 - dest);
			}
			_(ghost ++i;)
		}

		_(assert i == dest_len - 1)
		*d = '\0';
		_(assert i <= strlen(src, src_nul_lim))
		_(assert \forall size_t j; j < i ==> src[j] != '\0')
		_(assert \exists size_t j; j >= i && j <= src_nul_lim && src[j] == '\0' && (j - i) >= 0 && (j - i) <= (src_nul_lim - i) && s[j - i] == '\0')
		return (dest_len - 1) + strlen(s _(ghost src_nul_lim - i));
	}
}

#ifdef VERIFY
static void strlcpy_test1()
{
	char buf[10];
	size_t res;

	/* No room for the NUL terminator */
	buf[0] = 'a';
	const char test_src0[5] = { 't', 'e', 's', 't', '\0' };
	res = strlcpy(buf, test_src0, 0 _(ghost 4));
	_(assert res == 4);
	_(assert buf[0] == 'a');

	/* Only room for the NUL terminator and `*src' is a NUL character */
	buf[1] = buf[0] = 'a';
	const char test_src1[1] = { '\0' };
	res = strlcpy(buf, test_src1, 1 _(ghost 0));
	_(assert res == 0);
	_(assert buf[0] == '\0');
	_(assert buf[1] == 'a');

	/* `*src' is a NUL character and `dest' has space for characters other than the NUL terminator. */
	buf[1] = buf[0] = 'a';
	res = strlcpy(buf, test_src1, 10 _(ghost 0));
	_(assert res == 0);
	_(assert buf[0] == '\0');
	_(assert buf[1] == 'a');

	/* Only room for the NUL terminator and `*src' is not a NUL character */
	buf[1] = buf[0] = 'a';
	res = strlcpy(buf, test_src0, 1 _(ghost 4));
	_(assert res == 4);
	_(assert buf[0] == '\0');
	_(assert buf[1] == 'a');

	/* Only room to copy one character from `src' */
	buf[2] = buf[1] = buf[0] = 'a';
	res = strlcpy(buf, test_src0, 2 _(ghost 4));
	_(assert res == 4);
	_(assert buf[0] == 't');
	_(assert buf[1] == '\0');
	_(assert buf[2] == 'a');

	/* Only room to copy three characters from `src' */
	buf[4] = buf[3] = buf[2] = buf[1] = buf[0] = 'a';
	res = strlcpy(buf, test_src0, 4 _(ghost 4));
	_(assert res == 4);
	_(assert buf[0] == 't');
	_(assert buf[1] == 'e');
	_(assert buf[2] == 's');
	_(assert buf[3] == '\0');
	_(assert buf[4] == 'a');

	/* `dest' is just big enough to copy `src' and the NUL terminator. */
	buf[5] = buf[4] = buf[3] = buf[2] = buf[1] = buf[0] = 'a';
	res = strlcpy(buf, test_src0, 5 _(ghost 4));
	_(assert res == 4);
	_(assert buf[0] == 't');
	_(assert buf[1] == 'e');
	_(assert buf[2] == 's');
	_(assert buf[3] == 't');
	_(assert buf[4] == '\0');
	_(assert buf[5] == 'a');

	/* `dest' has space for more than `strlen(src) + 1' characters. */
	buf[5] = buf[4] = buf[3] = buf[2] = buf[1] = buf[0] = 'a';
	res = strlcpy(buf, test_src0, 10 _(ghost 4));
	_(assert res == 4);
	_(assert buf[0] == 't');
	_(assert buf[1] == 'e');
	_(assert buf[2] == 's');
	_(assert buf[3] == 't');
	_(assert buf[4] == '\0');
	_(assert buf[5] == 'a');
}
#endif
