/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

//
// ====================================================================
//
// Lattice Semiconductor Corporation
// 5555 NE Moore Court
// Hillsboro, OR 97124
// U.S.A
//
// TEL: 1-800-Lattice (USA and Canada)
// 408-826-6000 (other locations)
//
// web: http://www.latticesemi.com/
// email: techsupport@latticesemi.com
//
// ====================================================================
// File Details
// ====================================================================
// Project     : ISP 8 bit Microcontroller
// File        : isp8asm.c
// Title       : Assembler for Lattice 8 bit microcontroller
//
// Description : This program is the assembler for Lattice 8 bit
//       Microcontroller. Input is an assembler file output
//       EBR initialization file.
//
// Additional info. :
//
// ====================================================================
// Revision History
// ====================================================================
// Ver  | Author(s)        | Mod. Date| Changes Made
// --------------------------------------------------------------------
// 1.0  | Sandeep Dutta    | 04/27/05 | Initial Creation


// dwelch modified to use compiler instruction format

#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

/* ==================================================== */
/*          defines             */
/* ==================================================== */
#define COMMENT_CHAR    '#'
#define MAX_LINE_SIZE   1024
#define MAX_TOKEN_SIZE  100
#define MAX_REG     32
#define GETC() (*lbp ? *lbp++ : *lbp)
#define UNGETC()  lbp--
#define ABS(x) (x < 0 ? -x : x)

/* ==================================================== */
/*      types & data structures         */
/* ==================================================== */
typedef unsigned int uint32;

enum otype {
    OHEX = 0,
    OBIN
};

enum optype {
    OP_NONE,  OP_REGD,
    OP_REGB,  OP_LABEL,
    OP_CONST, OP_PORT
};

enum codeflags {
    CODE_ISDATA = 1,
    CODE_ISLDEF = CODE_ISDATA*2
};

typedef struct instr_tab {
    char   *opname;
    int opcode;
    int nops;
    int optypes[2];
} itab ;


typedef enum dirs {
    D_ORG, D_INCLUDE, D_EQU
} dirs;

typedef struct code {
    itab   *itabp;
    uint32  opcode;
    uint32  rd, rb, konst;
    uint32  flags;
    struct  stab   *label_ref;
    struct  stab   *label_def;
    struct  code   *next;
    int     loc_counter;
    int line_no;
    uint32  port;
} code;

typedef enum sym_types {
    SYM_LABEL = 1,
    SYM_EQU   = SYM_LABEL*2
} stypes;

typedef struct stab {
    char        *name;
    stypes       stype;
    int      value;
    int      loc_counter;
    code        *codep;
    struct stab *next;
} stab;


typedef struct directs {
    char *dname;
    void (*dparse_func)();
} directs;


/* ==================================================== */
/*      function forward declarations       */
/* ==================================================== */
static uint32 parse_const();
static void   dparse_equ();
static void   dparse_org();
static void   dparse_data();

/* ==================================================== */
/*          global variables        */
/* ==================================================== */
static code *croot = NULL;
static code *ccurr = NULL;
static stab *sroot = NULL;
static uint32 loc_counter = 0;
static FILE *infile = NULL;
static FILE *ofile = NULL;
static FILE *lfile = NULL;
static FILE *eifile = NULL;
static FILE *eofile = NULL;
static char *ifilename = NULL;
static char *ofilename = NULL;
static char *lfilename = NULL;
static int  line_no = 0;
static char line_buffer[MAX_LINE_SIZE];
static char *lbp;
static char token[MAX_TOKEN_SIZE];
static int omode = 0 ;
static int lgen  = 0 ;
static int prom_size = 512;
static int *vmstack;
static int  vmsp = 0;
static directs dirtab_array[] = {
    {"org", dparse_org},
    {"equ", dparse_equ},
    {"data",dparse_data}
};

static uint32 ocode_addr;

#define DIRTAB_SIZE sizeof(dirtab_array)/sizeof(directs)

static itab itab_array [] = {
    {"sub",     ((0x0 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"subi",    ((0x0 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_CONST}},
    {"subc",    ((0x1 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"subic",   ((0x1 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_CONST}},
    {"add",     ((0x2 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"addi",    ((0x2 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_CONST}},
    {"addc",    ((0x3 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"addic",   ((0x3 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_CONST}},
    {"nop",     ((0x4 << 14)|(0 << 13)|0), 0, {OP_NONE, OP_NONE}},
    {"mov",     ((0x4 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"movi",    ((0x4 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_CONST}},
    {"and",     ((0x5 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"andi",    ((0x5 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_CONST}},
    {"or",      ((0x6 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"ori",     ((0x6 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_CONST}},
    {"xor",     ((0x7 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"xori",    ((0x7 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_CONST}},
    {"cmp",     ((0x8 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"cmpi",    ((0x8 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_CONST}},
    {"test",    ((0x9 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"testi",   ((0x9 << 14)|(0 << 13)|0), 2, {OP_REGD, OP_CONST}},
    {"ror",     ((0xa << 14)|(0 << 13)|0), 2, {OP_REGD, OP_REGB}},
    {"rorc",    ((0xa << 14)|(0 << 13)|2), 2, {OP_REGD, OP_REGB}},
    {"rol",     ((0xa << 14)|(0 << 13)|1), 2, {OP_REGD, OP_REGB}},
    {"rolc",    ((0xa << 14)|(0 << 13)|3), 2, {OP_REGD, OP_REGB}},
    {"clrc",    ((0xb << 14)|(0 << 13)|0), 0, {OP_NONE, OP_NONE}},
    {"setc",    ((0xb << 14)|(0 << 13)|1), 0, {OP_NONE, OP_NONE}},
    {"clrz",    ((0xb << 14)|(0 << 13)|2), 0, {OP_NONE, OP_NONE}},
    {"setz",    ((0xb << 14)|(0 << 13)|3), 0, {OP_NONE, OP_NONE}},
    {"clri",    ((0xb << 14)|(0 << 13)|4), 0, {OP_NONE, OP_NONE}},
    {"seti",    ((0xb << 14)|(0 << 13)|5), 0, {OP_NONE, OP_NONE}},

    {"exporti", ((0xb << 14)|(1 << 13)|2), 2, {OP_REGD, OP_REGB}},
    {"importi", ((0xb << 14)|(1 << 13)|3), 2, {OP_REGD, OP_REGB}},
    {"sspi",    ((0xb << 14)|(1 << 13)|6), 2, {OP_REGD, OP_REGB}},
    {"lspi",    ((0xb << 14)|(1 << 13)|7), 2, {OP_REGD, OP_REGB}},
    {"export",  ((0xb << 14)|(1 << 13)|0), 2, {OP_REGD, OP_PORT}},
    {"halt",    ((0xb << 14)|(1 << 13)|(0x1F<<3)), 0, {OP_NONE, OP_NONE}},
    {"import",  ((0xb << 14)|(1 << 13)|1), 2, {OP_REGD, OP_PORT}},
    {"ssp",     ((0xb << 14)|(1 << 13)|4), 2, {OP_REGD, OP_PORT}},
    {"lsp",     ((0xb << 14)|(1 << 13)|5), 2, {OP_REGD, OP_PORT}},

    //{"bz",    ((0xc << 14)|(0 << 10)|0), 1, {OP_LABEL,OP_NONE}},
    //{"bnz",   ((0xc << 14)|(1 << 10)|0), 1, {OP_LABEL,OP_NONE}},
    //{"bc",    ((0xc << 14)|(2 << 10)|0), 1, {OP_LABEL,OP_NONE}},
    //{"bnc",   ((0xc << 14)|(3 << 10)|0), 1, {OP_LABEL,OP_NONE}},
    //{"b",     ((0xc << 14)|(4 << 10)|0), 1, {OP_LABEL,OP_NONE}},
    {"bz",      ((0xc << 14)|(0 << 12)|0), 1, {OP_LABEL,OP_NONE}},
    {"bnz",     ((0xc << 14)|(1 << 12)|0), 1, {OP_LABEL,OP_NONE}},
    {"bc",      ((0xc << 14)|(2 << 12)|0), 1, {OP_LABEL,OP_NONE}},
    {"bnc",     ((0xc << 14)|(3 << 12)|0), 1, {OP_LABEL,OP_NONE}},
    //{"callz", ((0xd << 14)|(0 << 10)|0), 1, {OP_LABEL,OP_NONE}},
    //{"callnz",((0xd << 14)|(1 << 10)|0), 1, {OP_LABEL,OP_NONE}},
    //{"callc", ((0xd << 14)|(2 << 10)|0), 1, {OP_LABEL,OP_NONE}},
    //{"callnc",((0xd << 14)|(3 << 10)|0), 1, {OP_LABEL,OP_NONE}},
    //{"call",  ((0xd << 14)|(4 << 10)|0), 1, {OP_LABEL,OP_NONE}},
    {"callz",   ((0xd << 14)|(0 << 12)|0), 1, {OP_LABEL,OP_NONE}},
    {"callnz",  ((0xd << 14)|(1 << 12)|0), 1, {OP_LABEL,OP_NONE}},
    {"callc",   ((0xd << 14)|(2 << 12)|0), 1, {OP_LABEL,OP_NONE}},
    {"callnc",  ((0xd << 14)|(3 << 12)|0), 1, {OP_LABEL,OP_NONE}},
    {"call",    ((0xe << 14)|(0 << 12)|0), 1, {OP_LABEL,OP_NONE}},
    //{"ret",   ((0xe << 14)|(0 << 10)|0), 0, {OP_NONE, OP_NONE}},
    //{"iret",  ((0xe << 14)|(4 << 10)|0), 0, {OP_NONE, OP_NONE}},
    {"ret",     ((0xe << 14)|(1 << 12)|0), 0, {OP_NONE, OP_NONE}},
    {"iret",    ((0xe << 14)|(2 << 12)|0), 0, {OP_NONE, OP_NONE}},
    {"b",       ((0xe << 14)|(3 << 12)|0), 1, {OP_LABEL,OP_NONE}}
    //{"importi", ((0xf << 14)|(0 << 10)|3), 2, {OP_REGD, OP_REGB}},
    //{"inpi",    ((0xf << 14)|(0 << 10)|3), 2, {OP_REGD, OP_REGB}},
    //{"exporti", ((0xf << 14)|(0 << 10)|2), 2, {OP_REGD, OP_REGB}},
    //{"outpi",   ((0xf << 14)|(0 << 10)|2), 2, {OP_REGD, OP_REGB}},
    //{"import",  ((0xf << 14)|(0 << 10)|1), 2, {OP_REGD, OP_PORT}},
    //{"inp",     ((0xf << 14)|(0 << 10)|1), 2, {OP_REGD, OP_PORT}},
    //{"export",  ((0xf << 14)|(0 << 10)|0), 2, {OP_REGD, OP_PORT}},
    //{"outp",    ((0xf << 14)|(0 << 10)|0), 2, {OP_REGD, OP_PORT}},
    //{"lspi",    ((0xf << 14)|(0 << 10)|7), 2, {OP_REGD, OP_REGB}},
    //{"sspi",    ((0xf << 14)|(0 << 10)|6), 2, {OP_REGD, OP_REGB}},
    //{"lsp",     ((0xf << 14)|(0 << 10)|5), 2, {OP_REGD, OP_PORT}},
    //{"ssp",     ((0xf << 14)|(0 << 10)|4), 2, {OP_REGD, OP_PORT}},
    //{"nop",     ((0x4 << 14)|(0 << 10)|0), 0, {OP_NONE, OP_NONE}}
};

#define ITAB_SIZE sizeof(itab_array)/sizeof(itab)

#if !defined(stricmp)
/* case independent string compare */
int stricmp(char *s1, char *s2)
{
    while (*s1 && *s2) {
        if (toupper(*s1) != toupper(*s2)) break;
        s1++; s2++;
    }
    return *s1 - *s2;
}
#endif

/* find a symbol a given type & name */
static stab *sym_find(char *name, stypes stype)
{
    stab *curr = sroot;

    while (curr) {
        if (!strcmp(curr->name,name) &&
            curr->stype == stype) {
            return curr;
        }
        curr = curr->next;
    }
    return NULL;
}

/* add a symbol into the table */
static stab *sym_insert(char *name, int loc, stypes stype)
{
    stab *snew = (stab *) malloc(sizeof(stab));

    snew->name = (char *) strdup(name);
    snew->loc_counter = loc;
    snew->codep = NULL;
    snew->stype = stype;
    snew->next  = NULL;
    if (!sroot) {
        sroot = snew;
    } else {
        snew->next = sroot;
        sroot = snew;
    }
    return snew;
}

static code *code_insert(itab *itabp, uint32 opcode,
             uint32 rd,   uint32 rb,
             uint32 konst,stab  *label_r,
             stab  *label_d, int port, uint32 flags)
{
    code *cnew = (code *) malloc(sizeof(code));

    cnew->itabp  = itabp;
    cnew->opcode = opcode;
    cnew->rd     = rd;
    cnew->rb     = rb;
    cnew->konst  = konst;
    cnew->label_ref  = label_r;
    cnew->label_def  = label_d;
    cnew->loc_counter= (flags & CODE_ISLDEF ? loc_counter : loc_counter++);
    cnew->port   = port;
    cnew->line_no = line_no;
    cnew->next   = NULL;
    cnew->flags  = flags;

    if (croot == NULL) {
        croot = cnew;
        ccurr = croot;
    } else {
        ccurr->next = cnew;
        ccurr = cnew;
    }
    return cnew;
}

static itab *itab_find(char *iname)
{
    int i = 0 ;

    for (i = 0 ; i < ITAB_SIZE ; i++ ) {
        if (!stricmp(iname,itab_array[i].opname))
            return &itab_array[i];
    }
    return NULL;
}

void error_exit(char *err_txt,...)
{
    va_list ap;

    va_start(ap,err_txt);
    fprintf(stderr,"%s (%d) ",ifilename,line_no);
    vfprintf(stderr,err_txt,ap);
    fprintf(stderr,"\n\n");
    va_end(ap);

    exit(0);
}

static int get_token(char *buff, char *delims)
{
    char nc = GETC();
    do {
        *buff++ = nc;
        nc = GETC();
    } while (!isspace(nc) && !strchr(delims,nc) && nc);
    *buff = '\0';
    return nc;
}

static int skip_space()
{
    char nc ;

    do {
        nc = GETC();
    } while (isspace(nc) && nc && nc != '\n');
    UNGETC();
    return nc;
}

static char *get_line()
{
    if (feof(infile)) return NULL;
    lbp = line_buffer;
    line_no++;
    return fgets(line_buffer,MAX_LINE_SIZE,infile);
}

/* parse .data directive format is
   .data <constant> */
static void dparse_data()
{
    char nc = skip_space();
    uint32 val;
    code *codep;

    val = parse_const();
    if (val & ~0x3FFFF) {
        error_exit(".data value cannot exceed 0x3ffff, value is 0x%X",val);
    }
    codep = code_insert( NULL, val, 0, 0, 0, NULL, NULL, 0, CODE_ISDATA);
}

/* parse .equ directive. the format is
         .equ <name>,<value> */
static void dparse_equ()
{
    char nc = skip_space();
    int value ;
    stab *sym;

    nc = get_token(token,",");
    UNGETC();
    nc = skip_space();
    if (nc != ',') {
        error_exit("invalid syntax for equate. expecting ','");
    }
    GETC(); /* skip the , */
    skip_space();
    value = parse_const();
    if ((sym = sym_find(token,SYM_EQU))) {
        error_exit("redifinition of .equ symbol %s",token);
    }
    sym = sym_insert(token,-1,SYM_EQU);
    sym->value = value;
}

/* parse .org directive : followed by a number */
static void dparse_org()
{
    char nc = skip_space();
    uint32 new_loc = parse_const();

    if (new_loc < loc_counter) {
        error_exit(".org overlays previous code; curr pc %d , new pc %d",
               loc_counter, new_loc);
    }
    loc_counter = new_loc;
}

/* parse assembler directives */
static void parse_direct()
{
    int i;
    get_token(token,"\n,");

    for (i = 0; i < DIRTAB_SIZE; i++) {
        if (!stricmp(dirtab_array[i].dname,token)) {
            (*dirtab_array[i].dparse_func)();
            return ;
        }
    }
    error_exit("unrecognized directive %s",token);
}

/* parse label definition */
static stab *parse_label_def(char *label)
{
    stab *stabp;

    if (!(stabp = sym_find(label,SYM_LABEL))) {
        stabp = sym_insert(label, loc_counter, SYM_LABEL);
    } else if (stabp->loc_counter != -1) {
        error_exit("multiple definition of label %s",label);
    }
    stabp->loc_counter = loc_counter;
    code_insert( NULL, 0, 0, 0, 0, NULL, stabp, 0, CODE_ISLDEF);
    return stabp;
}

/* parse label reference */
static stab *parse_label_ref()
{
    char nc = skip_space();
    stab *stabp;

    get_token(token,"\n");

    /* if !already exists then insert it*/
    if (!(stabp = sym_find(token, SYM_LABEL))) {
        stabp = sym_insert(token, -1, SYM_LABEL);
    }
    return stabp;
}

/* parse a number */
static uint32 parse_num()
{
    char nc = GETC();
    int sign = 0;
    int base = 10;
    int rval = 0;

    if (nc == '+') nc = GETC();
    if (nc == '-') {
        sign = 1;
        nc = GETC();
    }
    if (nc == '0') {
        /* octal */
        base = 8;
        nc = GETC();
        if (toupper(nc) == 'X') {
            /* nope .. hex */
            base = 16;
            nc = GETC();
        }
    }
    while (isdigit(nc) || isxdigit(toupper(nc))) {
        rval *= base;
        switch (base) {
        case 8:
        case 10:
            rval += (nc - '0');
            break;
        case 16:
            if (toupper(nc) >= 'A') {
                rval += (10 + (toupper(nc) - 'A'));
            } else {
                rval += nc - '0';
            }
            break;
        }
        nc = GETC();
    } while (isdigit(nc) || isxdigit(nc));

    if (nc != -1) {
        UNGETC();
    }
    if (sign) rval = -rval;
    return rval;
}

/* parse a character constant of the form '<char>' */
static uint32 parse_char()
{
    char nc = GETC();
    assert(nc == '\'');

    nc = GETC();
    /* escape character */
    if (nc == '\\') {
        nc = GETC();
        switch (nc) {
        case 'n':
            nc = '\n';
            break;
        case 'r':
            nc = '\r';
            break;
        case 't':
            nc = '\t';
            break;
        default:
            break;
        }
    }
    if (GETC() != '\'') {
        error_exit("invalis character constant");
    }
    return nc;
}

/* parse constant */
static uint32 parse_const()
{
    char nc;
    char token[MAX_TOKEN_SIZE];
    stab *esym;

    nc = GETC();
    UNGETC();
    if (isdigit(nc) || nc == '+' || nc == '-') {
        return parse_num();
    }
    if (nc == '\'') {
        return parse_char();
    }
    if (nc == '$') {
        GETC(); /* skip the $ */
        return loc_counter;
    }
    /* must be an equate; lookit up */
    get_token(token,"\n");
    if (!(esym = sym_find(token,SYM_EQU))) {
        error_exit("expected constant or equated symbol, found %s",
               token);
    }
    return esym->value;
}

/* parse registers */
static uint32 parse_reg(int maxreg)
{
    char nc;
    uint32 regno;

    nc = skip_space();
    if (toupper(nc) != 'R') {
        error_exit("excpected register");
    }
    GETC(); /* skip r */
    regno = parse_num();
    if (regno >= maxreg) {
        error_exit("invalid register number :"
               " expected reg number < R%d",maxreg-1);
    }
    return regno;
}

/* parse instructions */
static int parse_instr()
{
    char nc;
    itab *itabp;
    int  ops = 0;
    int fopcode = 0 ;
    int  rd = 0 , rb = 0 ;
    uint32  konst = 0, port = 0;
    stab *label = NULL;

    nc = skip_space();
    if (!nc) return 0;
    nc = get_token(token,"\n:");
    if (!isspace(nc)) {
        UNGETC();
    }

    /* if next character is a ':' then this is
       a label definition */
    nc = skip_space();
    if (nc == ':') {
        parse_label_def(token);
        return 1;
    } else {
        UNGETC();
    }

    if ((itabp = itab_find(token)) == NULL) {
        error_exit("unrecognized instruction mnemonic '%s'",token);
        return 0;
    }

    /* found instruction entry */
    for (ops = 0 ; ops < itabp->nops ; ops++) {
        switch (itabp->optypes[ops]) {
        case OP_REGD:
            rd = parse_reg(MAX_REG);
            break;
        case OP_REGB:
            rb = parse_reg(MAX_REG);
            break;
        case OP_CONST:
            konst = parse_const();
            fopcode |= 1 << 13;
            break;
        case OP_LABEL:
            label = parse_label_ref();
            //fopcode |= 1 << 13;
            break;
        case OP_PORT:
            port = parse_const();
            break;
        }
        /* if not last then separator must be present */
        if (ops  != (itabp->nops - 1)) {
            nc = skip_space();
            if (nc != ',') {
                error_exit("expected ',' after first operand");
            }
            GETC();
            nc = skip_space();
        }
    }
    fopcode |= itabp->opcode;
    fopcode |= (rd & 0x1f) << 8;
    fopcode |= (rb & 0x1f) << 3;
    fopcode |= (konst & 0xff);
    fopcode |= (port & 0x01f) << 3;

    code_insert(itabp, fopcode, rd, rb, konst, label, NULL, port, 0);
    return 1;
}

/* go thru the code and resolve
   all label references */
static void resolve_labels()
{
    code *codep = croot;

    while (codep) {
        if (codep->label_ref) {
            int diff;
            if (codep->label_ref->loc_counter == -1) {
                error_exit("undefined label %s",
                       codep->label_ref->name);
            }
            diff = (codep->label_ref->loc_counter - codep->loc_counter);
            //if ((codep->opcode == 0X0003B000) || (codep->opcode == 0X0003A000)) //if unconditional branch or call
            {
                if (ABS(diff) >= 2048) {
                    line_no = codep->line_no;
                    error_exit("branch out of range; must be"
                           "within 2048 instructions");
                }
                codep->opcode |= ((codep->label_ref->loc_counter -
                           codep->loc_counter) & 0xFFF);
            }
            //else {
                //if (ABS(diff) >= 512) {
                    //line_no = codep->line_no;
                    //error_exit("branch out of range; must be"
                           //"within 512 instructions");
                //}
                //codep->opcode |= ((codep->label_ref->loc_counter -
                           //codep->loc_counter) & 0x3FF);
            //}
        }
        codep = codep->next;
    }
}

static void print18bits(uint32 val, FILE *of)
{
    int i;

    for (i = 0 ; i < 18; i++) {
        fprintf(of,"%c", (val & (1 << 17)) ? '1' : '0');
        val <<= 1;
    }
}

static void print_listing(code *codep)
{
    int i;

    fprintf(lfile,"0x%04X\t",codep->loc_counter);
    if (codep->flags & CODE_ISLDEF) {
        fprintf(lfile,"\t%s:\n",codep->label_def->name);
        return ;
    }
    fprintf(lfile,"0x%05X\t",codep->opcode);
    print18bits(codep->opcode,lfile);
    if (codep->flags & CODE_ISDATA) {
        fprintf(lfile,"\t.data\t%d\n",codep->opcode);
        return ;
    }
    fprintf(lfile,"\t%s\t",codep->itabp->opname);

    for (i = 0 ; i < codep->itabp->nops; i++) {
        switch (codep->itabp->optypes[i]) {
        case OP_REGD:
            fprintf(lfile,"R%02d",codep->rd);
            break;
        case OP_REGB:
            fprintf(lfile,"R%02d",codep->rb);
            break;
        case OP_CONST:
            fprintf(lfile,"0x%02X",codep->konst & 0xff);
            break;
        case OP_LABEL:
            fprintf(lfile,"%s",codep->label_ref->name);
            break;
        case OP_PORT:
            fprintf(lfile,"#%d",codep->port);
            break;
        }
        /* if not last then separator must be present */
        if (i  != (codep->itabp->nops - 1)) {
            fprintf(lfile,",");
        }
    }
    fprintf(lfile,"\n");
}

static void print_code(int opcode)
{
    if (omode == OBIN) {
        print18bits(opcode,ofile);
        fprintf(ofile,"\n");
    } else if (omode == OHEX)
    {
        unsigned int checksum;

        //;llaaaattdddddd
        checksum=0x03;
        //not going to worry about programs bigger than 0xFFFF bytes
        checksum+=(ocode_addr>>8)&0xFF;
        checksum+=(ocode_addr>>0)&0xFF;
        checksum+=0x00;
        checksum+=(opcode>>16)&0xFF;
        checksum+=(opcode>> 8)&0xFF;
        checksum+=(opcode>> 0)&0xFF;
        checksum=(-checksum)&0xFF;
        fprintf(ofile,":03%04X00%06X%02X\n",ocode_addr&0xFFFF,opcode,checksum);
        ocode_addr+=3;
    } else {
        // push it into the stack , printed later
        // in output_code()
        vmstack[vmsp++] = opcode;
    }
}

static void output_code()
{
    code *codep ;
    int last_lc = 0;

    ocode_addr=0;

    if (lfile) {
        fprintf(lfile,"Loc\tOpcode\tOpcode\t\n");
        fprintf(lfile,"Counter\t(Hex)\t(Bin)\n");
    }
    /* allocate space for vhdl/verilog init */
    vmstack = (int *) malloc((sizeof(int))*prom_size);
    for (codep = croot; codep; codep = codep->next) {
        if (lfile) {
            print_listing(codep);
        }
        if (codep->flags & CODE_ISLDEF) continue;
        if ((codep->loc_counter - last_lc) > 1) {
            /* fill with zero */
            while ((codep->loc_counter - 1) > last_lc) {
                last_lc++;
                print_code(0);
            }
        }
        print_code(codep->opcode);
        last_lc = codep->loc_counter;
    }
    /* fill the rest of program rom with zeros */
//    for (;last_lc < prom_size - 1; last_lc++) {
  //      print_code(0);
//    }

}

static void parse_line()
{
    char nc;

    nc = skip_space();
    if (nc == COMMENT_CHAR) return ;
    if (nc == '\n') return;
    GETC();
    /* must be something recognizable.
       instruction or directive */
    if (nc == '.') {
        parse_direct();
    } else {
        UNGETC();
        parse_instr();
    }
}

static void print_usage(char *pname)
{
    printf("%s Usage : [options] input-file-name\n", pname);
    printf("options:\n");
    printf("\t-o output filename\n");
    printf("\t-s program rom size (default 512)\n");
    printf("\t-l generate assembler listing <filename>.lst\n");
    printf("\t-vx <output file in hex>(default)\n");
    printf("\t-vb <output file in binary\n");
    printf("\t-? <this message>\n");
}

int main(int argc, char *argv[])
{
    int i;

    if (argc < 1 ) {
        print_usage(argv[0]);
        exit(-1);
    }

    for (i = 1 ; i < argc; i++) {
        if (argv[i][0] == '-') {
            switch(argv[i][1]) {
            case 'o':
                i++;
                ofilename = strdup(argv[i]);
                break;
            case 'l':
                lgen = 1;
                break;
            case 's':
                i++;
                sscanf(argv[i],"%d",&prom_size);
                break;
            case 'v':
                switch (argv[i][2]) {
                case 'x':
                    omode = OHEX;
                    break;
                case 'b':
                    omode = OBIN;
                    break;
                default:
                    fprintf(stderr,
                        "invalid output option\n"
                        "\t-vx - hex, -vb binary\n");
                }
                break;
            case 'e':
                switch (argv[i][2]) {
                case 'i':
                    i++;
                    if (!(eifile=fopen(argv[i],"r"))) {
                        fprintf(stderr,"cannot open input edif file\n");
                        exit(-1);
                    }
                    break;
                case 'o':
                    i++;
                    if (!(eofile=fopen(argv[i],"w"))) {
                        fprintf(stderr,"cannot open output edif file\n");
                        exit(-1);
                    }
                    break;
                }
                break;
            case '?':
                print_usage(argv[0]);
                exit(0);
                break;
            default:
                fprintf(stderr,"invalid option %s\n",argv[i]);
                print_usage(argv[0]);
                break;
            }
        } else {
            /* assume it is the input file */
            ifilename = (char *) strdup(argv[i]);
            if (!(infile = fopen(ifilename,"r"))) {
                error_exit("unable to open file %s\n",ifilename);
            }
        }
    }
    if ((eifile && !eifile) || (!eifile && eofile)) {
        fprintf(stderr,"both input and output edif files must be provided\n");
        exit(-1);
    }
    if (!ifilename) {
        fprintf(stderr,"input file name missing\n");
        print_usage(argv[0]);
        exit(-1);
    }

    if (ofilename) {
        if (!(ofile=fopen(ofilename,"w"))) {
            error_exit("unable to open output file %s",ofilename);
        }
    } else {
        ofile = stdout;
    }
    if (lgen) {
        char *lfe;
        lfilename = (char *) malloc(strlen(ifilename)+10);
        strcpy(lfilename,ifilename);
        lfe = lfilename + strlen(lfilename);
        while (*lfe != '.' && lfe > lfilename) lfe--;
        if (lfe == lfilename) {
            strcat(lfilename,".lst");
        } else {
            strcpy(lfe,".lst");
        }
        if (!(lfile=fopen(lfilename,"w"))) {
            error_exit("unable to open listing file %s",lfilename);
        }
    } else {
        lfile = NULL;
    }

    while (get_line()) {
        parse_line();
    }
    resolve_labels();
    output_code();

    if (lfile) fclose(lfile);
}
