// compile: gcc generate-packets-prob.c
// execute: ./a.out [-h num_hops] [-n packetsize] [-p probability] [-r] [-s samples]
// summary:
// generates s permutations of length n for each "hop"; packets are generated based
// on the sending of a single ordered packet, and are permuted based on the
// sending probability p of the network; s (sample) controls how many times this
// sending process is simulated; using -i initializes the packet to the proper ordering
// each time.  The STDOUT may be captured and measured using measure-order.c:
//
// ./gen-packets-prob.x -h 4 -p 0.9 | ./measure-order.x 
//
// Output format:
//
// numNodes(or hops) numPackets numSuccessfulRuns(or data sets) 
// foreach S numSuccessfulRuns
//   foreach N numHodes
//     nodeNumber: orderPacketsReceived
//


#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

// defaults
#define _NUM_PACKETS   5
#define _PROBABILITY   0.5
#define _SAMPLES       100000 
#define _USE_SRAND     0      // use current usec and srand to seed rand()
#define _NUM_HOPS      1      // initialize new packet for each sample, or use last one?

// so we can use cmdline flags
int    NUM_PACKETS=_NUM_PACKETS,SAMPLES=_SAMPLES,USE_SRAND=_USE_SRAND,NUM_HOPS=_NUM_HOPS;
double PROBABILITY   = _PROBABILITY;

// dumps array to stdout
void print_array( int* array, int size )
{ int i;
 for (i = 0; i < size; i++)
     printf("%d ", array[i]);
 printf("\n");
}

// provides high resolution time in microseconds for srand
int get_usec ( void )
{ struct timeval tv;
 struct timezone tz;
 struct tm *tm;
 gettimeofday(&tv, &tz);
 return tv.tv_usec;
}

// globals - mins are initialized using INT_MAX
static int *permutation, send_attempts, hopCount;


// generates permutation based on probability, given some initial sequence
int messagePermutation(int *inputPermutation)
{ int i, *temp, tempPosition=0, numItemsInTemp=NUM_PACKETS, permPosition=0, attempts=0;
 double prob;
 temp = (int *)malloc(sizeof(int) * NUM_PACKETS);
 for (i=0; i<NUM_PACKETS; i++) temp[i] = inputPermutation[i];
 while (permPosition < NUM_PACKETS)
 {     //printf("temp: "); print_array(temp, numItemsInTemp);
       for (i=0; i < numItemsInTemp; i++)
       {   prob = 1.0*rand()/RAND_MAX;
           if (prob < PROBABILITY) permutation[permPosition++] = temp[i];
           else temp[tempPosition++] = temp[i];
           attempts++;
       }
       //printf("permutation: "); print_array(permutation, permPosition);
       numItemsInTemp = tempPosition;
       tempPosition = 0;      
 }
 //printf("permutation: "); print_array(permutation, NUM_PACKETS);
 return attempts;
}

// takes argument processing out of main
void processArgs(int argc, char** argv)
{ int i;
  extern char *optarg;
  extern int optind, optopt;
  /* command line flag stuff */
  while ((i = getopt(argc, argv, "gh:n:p:s:rv?")) != -1)
  {   switch(i)
      {case 'h':
          NUM_HOPS = atoi(optarg);
	  break;
       case 'n':
          NUM_PACKETS = atoi(optarg);
          break;
       case 'p':
          PROBABILITY = atof(optarg);
          break;
       case 'r':
          USE_SRAND = 1;
	  break;
       case 's':
          SAMPLES = atoi(optarg);
          break;
       /* handle bad arguments */
       case ':': /* n, p, or s without operand */
          exit (EXIT_FAILURE);
          break;
       case '?':
          printf("./a.out [-h num_hops] [-n packetsize] [-p probability] [-r] [-s samples] [-[g]v[v[v]]]\n");
          exit (EXIT_FAILURE);
          break;
      }
  }
}

int main(int argc, char** argv)
{ int i,j;
 double prob;
 processArgs(argc, argv); //took argument processing out of main()
  
 permutation = (int*)malloc(sizeof(int)*NUM_PACKETS);
 printf("%d %d %d // numNodes(or hops) numPackets numSuccessfulRuns(or data sets)\n",NUM_HOPS,NUM_PACKETS,SAMPLES);
 
 if (1 == USE_SRAND)     // seed rand() with current time in micro-seconds if -r is used
     srand(get_usec()); 

 // initialize permutation array - happens once all the time regardless
 for (i=0; i<NUM_PACKETS; i++) permutation[i] = i;
 
 // iterate over the specified number of samples*hops so each hop gets the SAMPLES number of samples
 hopCount = 0;
 for (j=1; j <= SAMPLES*NUM_HOPS; j++)
 { // get sequence, get number of attempts it took to get message there
   send_attempts = messagePermutation(permutation); 
   
   // output
   printf("%d: ",hopCount);
   print_array(permutation,NUM_PACKETS);

   hopCount++;
   // reinitialize to identity every NUM_HOPS times
   if (NUM_HOPS <= hopCount )
   {  for (i=0; i<NUM_PACKETS; i++) 
          permutation[i] = i;  
      hopCount = 0;
   } 
 }
 return EXIT_SUCCESS;
}
