// compile: gcc -lm measure-ordering-all.c
// execute: ./a.out [-h num_hops] [-n packetsize] [-p probability] [-r] [-s samples] [-[g]v[v[v]]]
// summary:
// computes measures for s permutations of length n; packets are generated based
// on the sending of a single ordered packet, and are permuted based on the
// sending probability p of the network; s (sample) controls how many times this
// sending process is simulated; using -i initializes the packet to the proper ordering
// each time

// todo:
// review, refine code
// have -i accept a number that says re-init packet every N times in sample pop S?
// see about tweaking output to make plotting easier; write script to iterate and plot

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <sys/time.h>

// defaults
#define _DISORDER_BASE 2.0
#define _NUM_PACKETS   5
#define _PROBABILITY   0.5
#define _SAMPLES       100000 
#define _VERBOSE       0
#define _USE_SRAND     0    // use current usec and srand to seed rand()
#define _NUM_HOPS    1    // initialize new packet for each sample, or use last one?
#define _FOR_GRAPHING  0    // formats all output for graphing purposes

// so we can use cmdline flags
double DISORDER_BASE = _DISORDER_BASE;
int    NUM_PACKETS   = _NUM_PACKETS;
double PROBABILITY   = _PROBABILITY;
int    SAMPLES       = _SAMPLES;
int    VERBOSE       = _VERBOSE;
int    USE_SRAND     = _USE_SRAND;
int    NUM_HOPS      = _NUM_HOPS ;
int    FOR_GRAPHING  = _FOR_GRAPHING;

// dumps array to stdout
void print_array( int* array, int size )
{ int i;
 for (i = 0; i < size; i++)
     printf("%d ", array[i]);
 printf("\n");
}

// provides high resolution time in microseconds for srand
int get_usec ( void )
{ struct timeval tv;
 struct timezone tz;
 struct tm *tm;
 gettimeofday(&tv, &tz);
 return tv.tv_usec;
}

// globals - mins are initialized using INT_MAX
static int *permutation, meth2, meth2_min=INT_MAX, meth2_max, meth2_sum, send_attempts, 
           **m2_histogram, M2_MAX, hopCount;
static double meth1, meth1_min = 1.0*INT_MAX, meth1_max, meth1_sum;

// sum of base^offset
double calc_disorder_1 ( int* permutation )
{ int i;
 double disorder = 0.0;
 for (i = 0; i < NUM_PACKETS; i++)
     disorder += pow(DISORDER_BASE,abs(i - permutation[i]));
 return disorder;
}

// sum of 'larger' packets arriving before i
int calc_disorder_2 ( int* permutation )
{ int i, j, disorder = 0;
 for (i = 0; i < NUM_PACKETS; i++)
 { for (j = i - 1; j >= 0; j--)
   { if ( permutation[j] > permutation[i] )
          disorder++;
   }
 }
 return disorder;
}

// manages measurements and stats for method 1
void makeMeasurements1 (int *permutation) 
{  meth1 = calc_disorder_1 ( permutation );
   meth1_sum += meth1;
   if (meth1 > meth1_max)
     meth1_max = meth1;
   if (meth1 < meth1_min)
     meth1_min = meth1;
}

// manages measurements and stats for method 1
void makeMeasurements2 (int *permutation) 
{  meth2 = calc_disorder_2 ( permutation );
   meth2_sum += meth2;
   if (meth2 > meth2_max)
     meth2_max = meth2;
   if (meth2 < meth2_min)
     meth2_min = meth2;
   // store meth2 results in m2_histogram
   m2_histogram[hopCount][meth2]++;
}

// handles outputting of stats for each permutation
void showIntermediateStats (int send_attempts)
{   // verbosity control
   if (2 == VERBOSE || 3 == VERBOSE)             // -vv and -vvv
   {   printf("(hop %d)",hopCount);
       print_array( permutation, NUM_PACKETS ); }

   if (1 == VERBOSE || 3 == VERBOSE)             // -v  and -vvv
       if (0 == FOR_GRAPHING)
           printf("packets: %d   prob: %2.3f   attempts: %d   disorder1: %10.3f   disorder2: %5d \n",
               NUM_PACKETS,PROBABILITY,send_attempts,meth1,meth2);
       else
           printf("%d\t%f\t%d\t%f\t%d\n",NUM_PACKETS,PROBABILITY,send_attempts,meth1,meth2);
}

// handles outputting of stats at the end (average, min, max, etc)
void showFinalStats ( void )
{  int i,j,total_samples = 1;
   double meth1_avg, meth2_avg;
    
   // averages 
   meth1_avg = meth1_sum/SAMPLES;
   meth2_avg = meth2_sum/SAMPLES;
   
   // output every time for human to read or for plotting
   if (0 == FOR_GRAPHING)
     printf("Prob: %4.2f; %d packets; samples: %d in %d hops; METH1: avg = %5.3f; min = %5.3f; max = %5.3f; METH2: avg = %5.3f; min = %d; max = %d\n",
         PROBABILITY,NUM_PACKETS,SAMPLES,NUM_HOPS,meth1_avg,meth1_min,meth1_max,meth2_avg,meth2_min,meth2_max);
   else
     printf("%f\t%d\t%f\t%f\t%f\t%f\t%d\t%d\n",PROBABILITY,SAMPLES,meth1_avg,meth1_min,meth1_max,meth2_avg,meth2_min,meth2_max);

   // output m2_histogram   
   for (i=0;i<NUM_HOPS;i++)
   { printf("Hop %d\nM2   Count:\n",i);
     for (j=0;j<=M2_MAX;j++)
       printf("%3d   %d\n",j,m2_histogram[i][j]); // put in prob distribution 
   }
}

// generates permutation based on probability, given some initial sequence
int messagePermutation(int *inputPermutation)
{ int i, *temp, tempPosition=0, numItemsInTemp=NUM_PACKETS, permPosition=0, attempts=0;
 double prob;
 temp = (int *)malloc(sizeof(int) * NUM_PACKETS);
 for (i=0; i<NUM_PACKETS; i++) temp[i] = inputPermutation[i];
 while (permPosition < NUM_PACKETS)
 {     //printf("temp: "); print_array(temp, numItemsInTemp);
       for (i=0; i < numItemsInTemp; i++)
       {   prob = 1.0*rand()/RAND_MAX;
           if (prob < PROBABILITY) permutation[permPosition++] = temp[i];
           else temp[tempPosition++] = temp[i];
           attempts++;
       }
       //printf("permutation: "); print_array(permutation, permPosition);
       numItemsInTemp = tempPosition;
       tempPosition = 0;      
 }
 //printf("permutation: "); print_array(permutation, NUM_PACKETS);
 return attempts;
}

// takes argument processing out of main
void processArgs(int argc, char** argv)
{ int i;
  extern char *optarg;
  extern int optind, optopt;
  /* command line flag stuff */
  while ((i = getopt(argc, argv, "gh:n:p:s:rv?")) != -1)
  {   switch(i)
      {case 'g':
          FOR_GRAPHING++;
          break;
       case 'h':
           NUM_HOPS = atoi(optarg);
	   break;
       case 'n':
          NUM_PACKETS = atoi(optarg);
          break;
       case 'p':
          PROBABILITY = atof(optarg);
          break;
       case 'r':
          USE_SRAND = 1;
	  break;
       case 's':
          SAMPLES = atoi(optarg);
          break;
       case 'v':
          VERBOSE++;
          break;
       /* handle bad arguments */
       case ':': /* n, p, or s without operand */
          exit (EXIT_FAILURE);
          break;
       case '?':
          printf("./a.out [-h num_hops] [-n packetsize] [-p probability] [-r] [-s samples] [-[g]v[v[v]]]\n");
          exit (EXIT_FAILURE);
          break;
      }
  }
}

int main(int argc, char** argv)
{ int i,j, send_attempts;
 double prob;
 processArgs(argc, argv); //took argument processing out of main()
  
 // 2d array to track histgram dims: NUM_HOPS x M2_MAX
 M2_MAX = NUM_PACKETS*(NUM_PACKETS-1)/2;
 m2_histogram = (int**)malloc(sizeof(int*)*NUM_HOPS);   // init array of int ptrs
 for (i=0;i<NUM_HOPS;i++)
 { m2_histogram[i] = (int*)malloc(sizeof(int)*M2_MAX);  // init array of ints
   for (j=0;j<M2_MAX;j++)
       m2_histogram[i][j] = 0;                          // init int vals
 }
 permutation = (int*)malloc(sizeof(int)*NUM_PACKETS);

 if (1 == USE_SRAND)     // seed rand() with current time in micro-seconds if -r is used
     srand(get_usec()); 

 // initialize permutation array - happens once all the time regardless
 for (i=0; i<NUM_PACKETS; i++) permutation[i] = i;
 
 // iterate over the specified number of samples
 hopCount = 0;
 for (j=1; j <= SAMPLES; j++)
 { // get sequence, get number of attempts it took to get message there
   send_attempts = messagePermutation(permutation); 

   // not sure if this is the best way to do this, but it works
   makeMeasurements1(permutation);
   makeMeasurements2(permutation); 

   // intermediate stats
   showIntermediateStats(send_attempts);
   
   hopCount++;
   // reinitialize to identity every NUM_HOPS times
   if (NUM_HOPS <= hopCount )
   {  for (i=0; i<NUM_PACKETS; i++) 
          permutation[i] = i;  
      hopCount = 0;
   } 
 }
  showFinalStats();
 return EXIT_SUCCESS;
}
