/* Copyright (c) 2011-2012 Florian Albrechtskirchinger. All rights reserved.
 *
 * This file is part of jdp-incubator.
 *
 * jdp-incubator is licensed under the Simplified BSD License (the "License").
 * You should have received a copy of the License along with jdp-incubator.
 * If not, see <http://www.opensource.org/licenses/bsd-license.php>.
 */

#include <stdlib.h>

#include <util/delay.h>

#include <abl/isp.h>
#include <abl/spi.h>

#define ISP_MAX_TRIES					8

static uint8_t isp_enable_programming(isp_session_t const * const session);
static uint8_t isp_issue_command(uint8_t const * const command);

static isp_return_code_t isp_write_flash_paged_using_delay(isp_session_t const * const session, uint8_t const num_pages, uint8_t const page_size_words, double const wait_delay, prog_uint8_t const *buffer, uint32_t const buffer_size, isp_command_callback_t const command_callback, isp_progress_callback_t const progress_callback);
static isp_return_code_t isp_write_flash_paged_using_polling(isp_session_t const * const session, uint8_t const num_pages, uint8_t const page_size_words, prog_uint8_t const *buffer, uint32_t const buffer_size, isp_command_callback_t const command_callback, isp_progress_callback_t const progress_callback);
static isp_return_code_t isp_verify_flash(isp_session_t const * const session, prog_uint8_t const *buffer, uint16_t const buffer_size_words, isp_command_callback_t const command_callback, isp_progress_callback_t const progress_callback);

static uint8_t isp_enable_programming(isp_session_t const * const session) {
	uint8_t command[4];
	uint8_t success = 0;

	session->command_callback(session, ISP_COMMAND_ENABLE_PROGRAMMING, command, 0, 0, 0);

	for(uint8_t i = 0; !success && i < ISP_MAX_TRIES; ++i) {
		// Reset target
		*session->reset_port |= session->reset_mask;
		_delay_ms(1.0);
		*session->reset_port &= ~session->reset_mask;
		_delay_ms(session->reset_delay / 1000.0);

		// Transmit the command; we expect the target to start echoing back after the second byte
		(void)spi_transmit(command[0]);
		(void)spi_transmit(command[1]);
		success  = (spi_transmit(command[2]) == command[1]);
		success &= (spi_transmit(command[3]) == command[2]);
	}

	return success;
}

static uint8_t isp_issue_command(uint8_t const * const command) {
	(void)spi_transmit(command[0]);
	(void)spi_transmit(command[1]);
	(void)spi_transmit(command[2]);
	return spi_transmit(command[3]);
}

isp_session_t *isp_session_init(isp_session_t *self) {
	if(!self) {
		self = malloc(sizeof(isp_session_t));
		self->flags = OBJECT_MODEL_FLAGS_DYNAMIC;
	}

	self->state = ISP_STATE_IDLE;

	return self;
}

void isp_session_free(isp_session_t *self) {
	if(self->flags & OBJECT_MODEL_FLAGS_DYNAMIC)
		free(self);
}

isp_return_code_t isp_begin(isp_session_t * const session, uint8_t volatile * const reset_port, uint8_t const reset_mask, uint16_t const reset_delay, isp_target_t const * const target, isp_options_t const options, isp_command_callback_t const command_callback) {
	uint8_t command[4];
	uint8_t success = 0;

	if(session->state != ISP_STATE_IDLE)
		return ISP_RETURN_CODE_INVALID_OPERATION;

	session->reset_port = reset_port;
	session->reset_mask = reset_mask;
	session->reset_delay = reset_delay;
	session->target = target;
	session->options = options;
	session->command_callback = command_callback;

	if(!isp_enable_programming(session))
		return ISP_RETURN_CODE_ENABLE_PROGRAMMING_FAILED;

	session->state = ISP_STATE_PROGRAMMING;

	return ISP_RETURN_CODE_SUCCESS;
}

isp_return_code_t isp_end(isp_session_t * const session) {
	if(session->state != ISP_STATE_PROGRAMMING)
		return ISP_RETURN_CODE_INVALID_OPERATION;

	session->state = ISP_STATE_IDLE;
	*session->reset_port |= session->reset_mask;

	return ISP_RETURN_CODE_SUCCESS;
}

isp_return_code_t isp_chip_erase(isp_session_t const * const session) {
	uint8_t command[4];
	isp_return_code_t return_code;

	return_code = session->command_callback(session, ISP_COMMAND_CHIP_ERASE, command, 0, 0, 0);

	if(return_code == ISP_RETURN_CODE_SUCCESS) {
		(void)isp_issue_command(command);

		_delay_ms(session->target->chip_erase_wait_delay / 1000.0);

		// isp_enable_programming toggles the reset line to end chip erase and re-enters programming mode
		if(!isp_enable_programming(session))
			return ISP_RETURN_CODE_ENABLE_PROGRAMMING_FAILED;
	}

	return return_code;
}

isp_return_code_t isp_read_memory(isp_session_t const * const session, isp_memory_type_t const memory_type, uint8_t const *buffer, uint32_t const buffer_size, isp_progress_callback_t const progress_callback) {
	return ISP_RETURN_CODE_NOT_SUPPORTED;
}

isp_return_code_t isp_write_memory(isp_session_t const * const session, isp_memory_type_t const memory_type, prog_uint8_t const *buffer, uint32_t const buffer_size, isp_progress_callback_t const progress_callback) {
	isp_memory_t const *memory = NULL;

	for(uint8_t i = 0; i < session->target->num_memories; ++i)
		if(session->target->memories[i].type == memory_type) {
			memory = &session->target->memories[i];
			break;
		}

	if(!memory)
		return ISP_RETURN_CODE_NOT_SUPPORTED;

	switch(memory->type) {
		case ISP_MEMORY_TYPE_FLASH:
			switch(memory->structure) {
				case ISP_MEMORY_STRUCTURE_PAGED:
					if(session->options & ISP_OPTION_USE_POLLING)
						return isp_write_flash_paged_using_polling(session, memory->size.paged.num_pages, memory->size.paged.page_size >> 1, buffer, buffer_size, session->command_callback, progress_callback);
					else
						return isp_write_flash_paged_using_delay(session, memory->size.paged.num_pages, memory->size.paged.page_size >> 1, memory->wait_delay / 1000.0, buffer, buffer_size, session->command_callback, progress_callback);
					break;
				case ISP_MEMORY_STRUCTURE_NON_PAGED:
					//return isp_write_flash_non_paged(session, memory, buffer, buffer_size, progress_callback);
					break;
			}
			break;
	}

	return ISP_RETURN_CODE_NOT_SUPPORTED;
}

isp_return_code_t isp_verify_memory(isp_session_t const * const session, isp_memory_type_t const memory_type, prog_uint8_t const *buffer, uint32_t const buffer_size, isp_progress_callback_t const progress_callback) {
	switch(memory_type) {
		case ISP_MEMORY_TYPE_FLASH:
			return isp_verify_flash(session, buffer, buffer_size >> 1, session->command_callback, progress_callback);
			break;
	}

	return ISP_RETURN_CODE_NOT_SUPPORTED;
}

isp_return_code_t isp_write_flash_paged_using_delay(isp_session_t const * const session, uint8_t const num_pages, uint8_t const page_size_words, double const wait_delay, prog_uint8_t const *buffer, uint32_t const buffer_size, isp_command_callback_t const command_callback, isp_progress_callback_t const progress_callback) {
	uint32_t remaining = buffer_size;
	uint8_t data;
	uint8_t command[4];

	progress_callback(session, 0, buffer_size);

	for(uint8_t page = 0; remaining && page < num_pages; ++page) {
		for(uint8_t word = 0; remaining && word < page_size_words; ++word, remaining -= 2) {
			command_callback(session, ISP_COMMAND_LOAD_PROGRAM_MEMORY_PAGE_LOW_BYTE, command, page, word, pgm_read_byte(buffer));
			isp_issue_command(command);
			++buffer;

			command_callback(session, ISP_COMMAND_LOAD_PROGRAM_MEMORY_PAGE_HIGH_BYTE, command, page, word, pgm_read_byte(buffer));
			isp_issue_command(command);
			++buffer;
		}

		command_callback(session, ISP_COMMAND_WRITE_PROGRAM_MEMORY_PAGE, command, page, 0, 0);
		isp_issue_command(command);

		_delay_ms(wait_delay);

		progress_callback(session, buffer_size - remaining, buffer_size);
	}

	return ISP_RETURN_CODE_SUCCESS;
}

isp_return_code_t isp_write_flash_paged_using_polling(isp_session_t const * const session, uint8_t const num_pages, uint8_t const page_size_words, prog_uint8_t const *buffer, uint32_t const buffer_size, isp_command_callback_t const command_callback, isp_progress_callback_t const progress_callback) {
	uint32_t remaining = buffer_size;
	uint8_t data;
	uint8_t command[4];
	uint8_t poll_data;
	uint16_t poll_address;
	isp_command_t poll_command;

	progress_callback(session, 0, buffer_size);

	for(uint8_t page = 0; remaining && page < num_pages; ++page) {
		poll_data = 0xff;

		for(uint8_t word = 0; remaining && word < page_size_words; ++word, remaining -= 2) {
			data = pgm_read_byte(buffer);
			if(poll_data == 0xff && data != 0xff) {
				poll_data = data;
				poll_address = (page * page_size_words) | word;
				poll_command = ISP_COMMAND_READ_PROGRAM_MEMORY_LOW_BYTE;
			}
			command_callback(session, ISP_COMMAND_LOAD_PROGRAM_MEMORY_PAGE_LOW_BYTE, command, page, word, data);
			isp_issue_command(command);
			++buffer;

			data = pgm_read_byte(buffer);
			if(poll_data == 0xff && data != 0xff) {
				poll_data = data;
				poll_address = (page * page_size_words) | word;
				poll_command = ISP_COMMAND_READ_PROGRAM_MEMORY_HIGH_BYTE;
			}
			command_callback(session, ISP_COMMAND_LOAD_PROGRAM_MEMORY_PAGE_HIGH_BYTE, command, page, word, data);
			isp_issue_command(command);
			++buffer;
		}

		command_callback(session, ISP_COMMAND_WRITE_PROGRAM_MEMORY_PAGE, command, page, 0, 0);
		isp_issue_command(command);

		command_callback(session, poll_command, command, poll_address, 0, 0);
		while(poll_data != 0xff && isp_issue_command(command) != poll_data) ;

		progress_callback(session, buffer_size - remaining, buffer_size);
	}

	return ISP_RETURN_CODE_SUCCESS;
}

//isp_return_code_t isp_write_flash_non_paged(isp_session_t const * const session, isp_memory_t const * const memory, uint8_t const *buffer, uint32_t const buffer_size, isp_progress_callback_t const callback) {
//	callback(session, buffer_size, buffer_size);

//	return ISP_RETURN_CODE_NOT_IMPLEMENTED;
//}

isp_return_code_t isp_verify_flash(isp_session_t const * const session, prog_uint8_t const *buffer, uint16_t const buffer_size_words, isp_command_callback_t const command_callback, isp_progress_callback_t const progress_callback) {
	uint8_t command[4];
	uint32_t buffer_size = buffer_size_words << 1;
	uint32_t address = 0;

	progress_callback(session, address, buffer_size);

	for(uint16_t word = 0; word < buffer_size_words; ++word, address += 2) {
		command_callback(session, ISP_COMMAND_READ_PROGRAM_MEMORY_LOW_BYTE, command, word, 0, 0);
		if(isp_issue_command(command) != pgm_read_byte(buffer))
			return ISP_RETURN_CODE_OPERATION_FAILED;
		++buffer;

		command_callback(session, ISP_COMMAND_READ_PROGRAM_MEMORY_HIGH_BYTE, command, word, 0, 0);
		if(isp_issue_command(command) != pgm_read_byte(buffer))
			return ISP_RETURN_CODE_OPERATION_FAILED;
		++buffer;

		progress_callback(session, address, buffer_size);
	}

	return ISP_RETURN_CODE_SUCCESS;
}
