/* Copyright (c) 2011 Florian Albrechtskirchinger. All rights reserved.
 *
 * This file is part of jdp-incubator.
 *
 * jdp-incubator is licensed under the Simplified BSD License (the "License").
 * You should have received a copy of the License along with jdp-incubator.
 * If not, see <http://www.opensource.org/licenses/bsd-license.php>.
 */

#ifndef ABL_ISP_H_
#define ABL_ISP_H_

#include <avr/pgmspace.h>

#include <abl/gcc.h>
#include <abl/objectmodel.h>
#include <abl/types.h>

typedef enum {
	ISP_RETURN_CODE_SUCCESS				= 0,
	ISP_RETURN_CODE_INVALID_OPERATION,
	ISP_RETURN_CODE_NOT_SUPPORTED,
	ISP_RETURN_CODE_NOT_IMPLEMENTED,
	ISP_RETURN_CODE_ENABLE_PROGRAMMING_FAILED,
	ISP_RETURN_CODE_OPERATION_FAILED,
} GCC_PACKED isp_return_code_t;

typedef enum {
	ISP_STATE_IDLE,
	ISP_STATE_PROGRAMMING,
} GCC_PACKED isp_state_t;

typedef enum {
	ISP_COMMAND_ENABLE_PROGRAMMING,
	ISP_COMMAND_CHIP_ERASE,
	ISP_COMMAND_LOAD_PROGRAM_MEMORY_PAGE_LOW_BYTE,
	ISP_COMMAND_LOAD_PROGRAM_MEMORY_PAGE_HIGH_BYTE,
	ISP_COMMAND_WRITE_PROGRAM_MEMORY_PAGE,
	ISP_COMMAND_READ_PROGRAM_MEMORY_LOW_BYTE,
	ISP_COMMAND_READ_PROGRAM_MEMORY_HIGH_BYTE
} GCC_PACKED isp_command_t;

typedef enum {
	ISP_OPTION_NONE					= 0,
	ISP_OPTION_SAFE_MODE				= (1 << 0),
	ISP_OPTION_SAFE_MODE_RESTORE_FUSES		= (1 << 1) | ISP_OPTION_SAFE_MODE,
	ISP_OPTION_USE_POLLING				= (1 << 2),
} GCC_PACKED isp_options_t;

typedef enum {
	ISP_MEMORY_TYPE_FLASH,
	ISP_MEMORY_TYPE_EEPROM,
	ISP_MEMORY_TYPE_FUSE,
	ISP_MEMORY_TYPE_LFUSE,
	ISP_MEMORY_TYPE_HFUSE,
	ISP_MEMORY_TYPE_EFUSE,
	ISP_MEMORY_TYPE_LOCK,
	ISP_MEMORY_TYPE_SIGNATURE,
	ISP_MEMORY_TYPE_CALIBRATION,
} GCC_PACKED isp_memory_type_t;

typedef enum {
	ISP_MEMORY_STRUCTURE_PAGED			= 1,
	ISP_MEMORY_STRUCTURE_NON_PAGED			= 0,
} GCC_PACKED isp_memory_structure_t;

typedef struct isp_command_parameters isp_command_parameters_t;
typedef struct isp_memory isp_memory_t;
typedef struct isp_target isp_target_t;
typedef struct isp_session isp_session_t;

typedef isp_return_code_t (*isp_command_callback_t)(isp_session_t const * const session, isp_command_t const command, uint8_t buffer[4], uint16_t const address_or_page, uint16_t const word, uint8_t const data);
typedef void (*isp_progress_callback_t)(isp_session_t const * const session, uint32_t const address, uint32_t const size);

struct isp_memory {
	isp_memory_type_t const type;
	isp_memory_structure_t const structure;
	uint16_t const wait_delay;
	union {
		struct {
			uint16_t const num_pages;
			uint16_t const page_size;
		} paged;
		struct {
			uint16_t const size;
		} non_paged;
	} size;
};

struct isp_target {
	uint8_t const signature[3];
	uint16_t const chip_erase_wait_delay;
	uint8_t const num_memories;
	isp_memory_t const * const memories;
};

struct isp_session {
	uint8_t flags;
	isp_state_t state;
	uint8_t volatile * reset_port;
	uint8_t reset_mask;
	uint16_t reset_delay;
	isp_target_t const *target;
	isp_options_t options;
	isp_command_callback_t command_callback;
	void *user_data;
};

isp_session_t *isp_session_init(isp_session_t *self);
void isp_session_free(isp_session_t *self);

isp_return_code_t isp_begin(isp_session_t * const session, uint8_t volatile * const reset_port, uint8_t const reset_mask, uint16_t const reset_delay, isp_target_t const * const target, isp_options_t const options, isp_command_callback_t const command_callback);
isp_return_code_t isp_end(isp_session_t * const session);

isp_return_code_t isp_chip_erase(isp_session_t const *session);

isp_return_code_t isp_read_memory(isp_session_t const * const session, isp_memory_type_t const memory_type, uint8_t const *buffer, uint32_t const buffer_size, isp_progress_callback_t const progress_callback);
isp_return_code_t isp_write_memory(isp_session_t const * const session, isp_memory_type_t const memory_type, prog_uint8_t const *buffer, uint32_t const buffer_size, isp_progress_callback_t const progress_callback);
isp_return_code_t isp_verify_memory(isp_session_t const * const session, isp_memory_type_t const memory_type, prog_uint8_t const *buffer, uint32_t const buffer_size, isp_progress_callback_t const progress_callback);

#endif /* ABL_ISP_H_ */
