/* Copyright (c) 2012 Florian Albrechtskirchinger. All rights reserved.
 *
 * This file is part of jdp-incubator.
 *
 * jdp-incubator is licensed under the Simplified BSD License (the "License").
 * You should have received a copy of the License along with jdp-incubator.
 * If not, see <http://www.opensource.org/licenses/bsd-license.php>.
 */

#include <abl/spi.h>

#include <nibo/cepro/copro.h>

static isp_target_t const nibo_isp_target = {
	.signature = { 0x1e, 0x93, 0x0a, },
	.chip_erase_wait_delay = 9000,
	.num_memories = 1,
	.memories = (const isp_memory_t[]) {
		{
			.type = ISP_MEMORY_TYPE_FLASH,
			.structure = ISP_MEMORY_STRUCTURE_PAGED,
			.wait_delay = 4500,
			.size.paged = {
				.num_pages = 128,
				.page_size = 64,
			},
		},
	},
};

static uint8_t nibo_copro_initialized = 0;

static isp_return_code_t nibo_isp_command_callback(isp_session_t const * const session, isp_command_t const command, uint8_t buffer[4], uint16_t const page_or_address, uint16_t const word, uint8_t const data);

uint8_t nibo_copro_init(void) {
	if(!nibo_copro_initialized) {
		NIBO_COPRO_RESET_DDR |= NIBO_COPRO_RESET_MASK;
		NIBO_COPRO_RESET_PORT |= NIBO_COPRO_RESET_MASK;

		nibo_copro_initialized = 1;
	}
}

isp_return_code_t nibo_copro_program_flash(prog_uint8_t const *buffer, uint16_t const buffer_size, isp_progress_callback_t const progress_callback) {
	isp_session_t session;
	isp_return_code_t return_code;

	// TODO Save/restore SPI and GPIO functions
	spi_configure_and_enable2(SPI_ROLE_MASTER, SPI_DATA_ORDER_MSB_FIRST, SPI_MODE_0, SPI_CLOCK_RATE_FOSC_BY_128, SPI_INTERRUPT_DISABLED, SPI_GPIO_FUNCTION_MASTER_SS_DRIVEN);

	isp_session_init(&session);

	if((return_code = isp_begin(&session, &NIBO_COPRO_RESET_PORT, NIBO_COPRO_RESET_MASK, 25000, &nibo_isp_target, ISP_OPTION_USE_POLLING, &nibo_isp_command_callback)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	if((return_code = isp_chip_erase(&session)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	if((return_code = isp_write_memory(&session, ISP_MEMORY_TYPE_FLASH, buffer, buffer_size, progress_callback)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	if((return_code = isp_end(&session)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	return ISP_RETURN_CODE_SUCCESS;
}

isp_return_code_t nibo_copro_verify_flash(prog_uint8_t const *buffer, uint16_t const buffer_size, isp_progress_callback_t const progress_callback) {
	isp_session_t session;
	isp_return_code_t return_code;

	spi_configure_and_enable2(SPI_ROLE_MASTER, SPI_DATA_ORDER_MSB_FIRST, SPI_MODE_0, SPI_CLOCK_RATE_FOSC_BY_128, SPI_INTERRUPT_DISABLED, SPI_GPIO_FUNCTION_MASTER_SS_DRIVEN);

	isp_session_init(&session);

	if((return_code = isp_begin(&session, &NIBO_COPRO_RESET_PORT, NIBO_COPRO_RESET_MASK, 25000, &nibo_isp_target, ISP_OPTION_USE_POLLING, &nibo_isp_command_callback)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	if((return_code = isp_verify_memory(&session, ISP_MEMORY_TYPE_FLASH, buffer, buffer_size, progress_callback)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	if((return_code = isp_end(&session)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	return ISP_RETURN_CODE_SUCCESS;
}

isp_return_code_t nibo_copro_program_and_verify_flash(prog_uint8_t const *buffer, uint16_t const buffer_size, isp_progress_callback_t const write_progress_callback, isp_progress_callback_t const verify_progress_callback) {
	isp_session_t session;
	isp_return_code_t return_code;

	spi_configure_and_enable2(SPI_ROLE_MASTER, SPI_DATA_ORDER_MSB_FIRST, SPI_MODE_0, SPI_CLOCK_RATE_FOSC_BY_128, SPI_INTERRUPT_DISABLED, SPI_GPIO_FUNCTION_MASTER_SS_DRIVEN);

	isp_session_init(&session);

	if((return_code = isp_begin(&session, &NIBO_COPRO_RESET_PORT, NIBO_COPRO_RESET_MASK, 25000, &nibo_isp_target, ISP_OPTION_USE_POLLING, &nibo_isp_command_callback)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	if((return_code = isp_chip_erase(&session)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	if((return_code = isp_write_memory(&session, ISP_MEMORY_TYPE_FLASH, buffer, buffer_size, write_progress_callback)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	if((return_code = isp_verify_memory(&session, ISP_MEMORY_TYPE_FLASH, buffer, buffer_size, verify_progress_callback)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	if((return_code = isp_end(&session)) != ISP_RETURN_CODE_SUCCESS)
		return return_code;

	return ISP_RETURN_CODE_SUCCESS;
}

isp_return_code_t nibo_isp_command_callback(isp_session_t const * const session, isp_command_t const command, uint8_t buffer[4], uint16_t const page_or_address, uint16_t const word, uint8_t const data) {
	switch(command) {
		case ISP_COMMAND_ENABLE_PROGRAMMING:
			buffer[0] = 0xac;
			buffer[1] = 0x53;
			buffer[2] = 0x00;
			buffer[3] = 0x00;
			break;
		case ISP_COMMAND_CHIP_ERASE:
			buffer[0] = 0xac;
			buffer[1] = 0x80;
			buffer[2] = 0x00;
			buffer[3] = 0x00;
			break;
		case ISP_COMMAND_LOAD_PROGRAM_MEMORY_PAGE_LOW_BYTE:
			buffer[0] = 0x40;
			buffer[1] = 0x00;
			buffer[2] = word;
			buffer[3] = data;
			break;
		case ISP_COMMAND_LOAD_PROGRAM_MEMORY_PAGE_HIGH_BYTE:
			buffer[0] = 0x48;
			buffer[1] = 0x00;
			buffer[2] = word;
			buffer[3] = data;
			break;
		case ISP_COMMAND_READ_PROGRAM_MEMORY_LOW_BYTE:
			buffer[0] = 0x20;
			buffer[1] = page_or_address >> 8;
			buffer[2] = page_or_address;
			buffer[3] = 0x00;
			break;
		case ISP_COMMAND_READ_PROGRAM_MEMORY_HIGH_BYTE:
			buffer[0] = 0x28;
			buffer[1] = page_or_address >> 8;
			buffer[2] = page_or_address;
			buffer[3] = 0x00;
			break;
		case ISP_COMMAND_WRITE_PROGRAM_MEMORY_PAGE:
			buffer[0] = 0x4c;
			buffer[1] = page_or_address >> 3;
			buffer[2] = page_or_address << 5;
			buffer[3] = 0x00;
			break;
		default:
			return ISP_RETURN_CODE_NOT_SUPPORTED;
	}

	return ISP_RETURN_CODE_SUCCESS;
}
