/*
 * Copyright 2011 Florian Albrechtskirchinger. All rights reserved.
 *
 * This file is part of libat32uc3.
 *
 * libat32uc3 is licensed under the Simplified BSD License
 * (the "License").
 *
 * A copy of the License should be included in the distribution.
 * Alternatively, you may obtain a copy of the License at
 * http://www.opensource.org/licenses/bsd-license.php
 *
 * Author(s): Florian Albrechtskirchinger <falbrechtskirchinger@gmail.com>
 */

#include <at32uc3/gpio.h>

uint32_t gpio_port_read_register(void volatile *port, uint32_t const offset) {
	uint32_t volatile * const p = (uint32_t volatile *)(port + offset);

	return *p;
}

void gpio_port_write_register(void volatile *port, uint32_t const offset, uint32_t const value) {
	uint32_t volatile * const p = (uint32_t volatile *)(port + offset);

	*p = value;
}

void gpio_port_reset(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_GPERS, mask);
	gpio_port_write_register(port, GPIO_REG_PMR0C, mask);
	gpio_port_write_register(port, GPIO_REG_PMR1C, mask);
	gpio_port_write_register(port, GPIO_REG_ODERC, mask);
	gpio_port_write_register(port, GPIO_REG_OVRC, mask);
	gpio_port_write_register(port, GPIO_REG_PUERC, mask);
	gpio_port_write_register(port, GPIO_REG_ODMERC, mask);
	gpio_port_write_register(port, GPIO_REG_IERC, mask);
	gpio_port_write_register(port, GPIO_REG_IMR0C, mask);
	gpio_port_write_register(port, GPIO_REG_IMR1C, mask);
	gpio_port_write_register(port, GPIO_REG_GFERS, mask);
}

void gpio_port_enable(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_GPERS, mask);
}

void gpio_port_disable(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_GPERC, mask);
}

void gpio_port_select_input(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_ODERC, mask);
}

void gpio_port_select_output(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_ODERS, mask);
}

void gpio_port_enable_pullup(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_PUERS, mask);
}

void gpio_port_disable_pullup(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_PUERC, mask);
}

void gpio_port_enable_open_drain_mode(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_ODMERS, mask);
}

void gpio_port_disable_open_drain_mode(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_ODMERC, mask);
}

void gpio_port_enable_glitch_filter(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_GFERS, mask);
}

void gpio_port_disable_glitch_filter(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_GFERC, mask);
}

void gpio_port_enable_interrupts(void volatile *port, uint32_t const mask, gpio_interrupt_mode_t const mode) {
	gpio_port_write_register(port, GPIO_REG_IERS, mask);

	switch(mode) {
		case GPIO_INTERRUPT_MODE_ANY_EDGE:
			gpio_port_write_register(port, GPIO_REG_PMR0C, mask);
			gpio_port_write_register(port, GPIO_REG_PMR1C, mask);
			break;
		case GPIO_INTERRUPT_MODE_RISING_EDGE:
			gpio_port_write_register(port, GPIO_REG_PMR0S, mask);
			gpio_port_write_register(port, GPIO_REG_PMR1C, mask);
			break;
		case GPIO_INTERRUPT_MODE_FALLING_EDGE:
			gpio_port_write_register(port, GPIO_REG_PMR0C, mask);
			gpio_port_write_register(port, GPIO_REG_PMR1S, mask);
			break;
	}
}

void gpio_port_disable_interrupts(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_IERC, mask);
}

void gpio_port_enable_input(void volatile *port, uint32_t const mask, gpio_input_options_t const options, gpio_interrupt_mode_t const mode) {
	gpio_port_select_input(port, mask);

	if(options & GPIO_INPUT_OPTION_PULLUP)
		gpio_port_enable_pullup(port, mask);
	if(options & GPIO_INPUT_OPTION_OPEN_DRAIN_MODE)
		gpio_port_enable_open_drain_mode(port, mask);
	if(options & GPIO_INPUT_OPTION_GLITCH_FILTER)
		gpio_port_enable_open_drain_mode(port, mask);
	if(options & GPIO_INPUT_OPTION_INTERRUPT)
		gpio_port_enable_interrupts(port, mask, mode);

	gpio_port_enable(port, mask);
}

void gpio_port_enable_output(void volatile *port, uint32_t const mask) {
	gpio_port_select_output(port, mask);
	gpio_port_enable(port, mask);
}

uint32_t gpio_port_read_pin_values(void volatile *port) {
	return gpio_port_read_register(port, GPIO_REG_PVR);
}

uint32_t gpio_port_read_output_value(void volatile *port) {
	return gpio_port_read_register(port, GPIO_REG_OVR);
}

void gpio_port_write_output_value(void volatile *port, uint32_t const value) {
	gpio_port_write_register(port, GPIO_REG_OVR, value);
}

void gpio_port_set_output_value(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_OVRS, mask);
}

void gpio_port_clear_output_value(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_OVRC, mask);
}

void gpio_port_toggle_output_value(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_OVRT, mask);
}

uint32_t gpio_port_read_interrupt_flags(void volatile *port) {
	return gpio_port_read_register(port, GPIO_REG_IFR);
}

void gpio_port_deassert_interrupts(void volatile *port, uint32_t const mask) {
	gpio_port_write_register(port, GPIO_REG_IFRC, mask);
}

void gpio_port_select_function(void volatile *port, uint32_t const mask, gpio_function_t const function) {
	switch(function) {
		case GPIO_FUNCTION_A:
			gpio_port_write_register(port, GPIO_REG_PMR0C, mask);
			gpio_port_write_register(port, GPIO_REG_PMR1C, mask);
			break;
		case GPIO_FUNCTION_B:
			gpio_port_write_register(port, GPIO_REG_PMR0S, mask);
			gpio_port_write_register(port, GPIO_REG_PMR1C, mask);
			break;
		case GPIO_FUNCTION_C:
			gpio_port_write_register(port, GPIO_REG_PMR0C, mask);
			gpio_port_write_register(port, GPIO_REG_PMR1S, mask);
			break;
	}
}

void gpio_port_enable_function(void volatile *port, uint32_t const mask, gpio_function_t const function) {
	gpio_port_select_function(port, mask, function);
	gpio_port_disable(port, mask);
}

gpio_map_t *gpio_map_init(gpio_map_t *self) {
	*self = (gpio_map_t) {
		.ports = { 0, 0, 0, 0, },
		.masks = { 0, 0, 0, 0, },
	};

	return self;
}

gpio_functions_map_t *gpio_functions_map_init(gpio_functions_map_t *self) {
	*self = (gpio_functions_map_t) {
		.ports = { 0, 0, 0, 0, },
		.masks = {
			{ 0, 0, 0, },
			{ 0, 0, 0, },
			{ 0, 0, 0, },
			{ 0, 0, 0, },
		},
	};

	return self;
}

void gpio_map_add_pin(gpio_map_t * const map, uint8_t const pin_nb) {
	uint8_t const port_nb = (pin_nb >> 5);
	uint32_t const mask = (1 << (pin_nb & 0x1f));

	map->ports[port_nb] = GPIO_PORT(port_nb);
	map->masks[port_nb] |= mask;
}

void gpio_map_add_pins(gpio_map_t * const map, uint32_t const n, uint8_t const *pin_nbs) {
	for(uint32_t i = n; i; --i, ++pin_nbs)
		gpio_map_add_pin(map, *pin_nbs);
}

void gpio_functions_map_add_pin_function(gpio_functions_map_t * const map, gpio_pin_function_t const pin_function) {
	uint8_t const port_nb = (pin_function.pin_nb >> 5);
	uint32_t const mask = (1 << (pin_function.pin_nb & 0x1f));

	map->ports[port_nb] = GPIO_PORT(port_nb);
	map->masks[port_nb][pin_function.function] |= mask;
}

void gpio_functions_map_add_pin_functions(gpio_functions_map_t * const map, uint32_t const n, gpio_pin_function_t const *pin_functions) {
	for(uint32_t i = n; i; --i, ++pin_functions)
		gpio_functions_map_add_pin_function(map, *pin_functions);
}

void gpio_map_reset(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_reset(port, mask);
	}
}

void gpio_map_enable(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_enable(port, mask);
	}
}

void gpio_map_disable(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_disable(port, mask);
	}
}

void gpio_map_select_input(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_select_input(port, mask);
	}
}

void gpio_map_select_output(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_select_output(port, mask);
	}
}

void gpio_map_enable_pullup(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_enable_pullup(port, mask);
	}
}

void gpio_map_disable_pullup(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_disable_pullup(port, mask);
	}
}

void gpio_map_enable_open_drain_mode(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_enable_open_drain_mode(port, mask);
	}
}

void gpio_map_disable_open_drain_mode(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_disable_open_drain_mode(port, mask);
	}
}

void gpio_map_enable_glitch_filter(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_enable_glitch_filter(port, mask);
	}
}

void gpio_map_disable_glitch_filter(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_disable_glitch_filter(port, mask);
	}
}

void gpio_map_enable_interrupts(gpio_map_t const * const map, gpio_interrupt_mode_t const mode) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_enable_interrupts(port, mask, mode);
	}
}

void gpio_map_disable_interrupts(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_disable_interrupts(port, mask);
	}
}

void gpio_map_enable_input(gpio_map_t const * const map, gpio_input_options_t const options, gpio_interrupt_mode_t const mode) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_enable_input(port, mask, options, mode);
	}
}

void gpio_map_enable_output(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_enable_output(port, mask);
	}
}

void gpio_map_set_output_value(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_set_output_value(port, mask);
	}
}

void gpio_map_clear_output_value(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_clear_output_value(port, mask);
	}
}

void gpio_map_toggle_output_value(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_toggle_output_value(port, mask);
	}
}

void gpio_map_deassert_interrupts(gpio_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;

	GPIO_MAP_FOREACH(map, i, port, mask) {
		gpio_port_deassert_interrupts(port, mask);
	}
}

void gpio_functions_map_select_functions(gpio_functions_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;
	gpio_function_t function;

	GPIO_FUNCTIONS_MAP_FOREACH(map, i, port, function, mask)
		gpio_port_select_function(port, mask, function);
}

void gpio_functions_map_enable_functions(gpio_functions_map_t const * const map) {
	uint8_t i;
	void volatile *port;
	uint32_t mask;
	gpio_function_t function;

	GPIO_FUNCTIONS_MAP_FOREACH(map, i, port, function, mask)
		gpio_port_enable_function(port, mask, function);
}
