/*
 * Copyright 2011 Florian Albrechtskirchinger. All rights reserved.
 *
 * This file is part of libat32uc3.
 *
 * libat32uc3 is licensed under the Simplified BSD License
 * (the "License").
 *
 * A copy of the License should be included in the distribution.
 * Alternatively, you may obtain a copy of the License at
 * http://www.opensource.org/licenses/bsd-license.php
 *
 * Author(s): Florian Albrechtskirchinger <falbrechtskirchinger@gmail.com>
 */

#include <at32uc3/pm.h>

uint32_t pm_read_register(uint32_t const offset) {
	uint32_t volatile * const p = (uint32_t volatile * const)(PM_ADDRESS + offset);

	return *p;
}

void pm_write_register(uint32_t const offset, uint32_t const value) {
	uint32_t volatile * const p = (uint32_t volatile * const)(PM_ADDRESS + offset);

	*p = value;
}

void pm_write_register_field(uint32_t const offset, uint32_t const field_mask, uint32_t const value) {
	uint32_t const tmp = pm_read_register(offset);

	pm_write_register(offset, (tmp & ~field_mask) | value);
}

void pm_enable_oscillator(pm_oscillator_t const oscillator) {
	pm_write_register_field(PM_REG_MCCTRL, PM_REG_MCCTRL_OSC0EN_MASK | PM_REG_MCCTRL_OSC1EN_MASK, oscillator);
}

void pm_disable_oscillator(pm_oscillator_t const oscillator) {
	pm_write_register_field(PM_REG_MCCTRL, oscillator, 0);
}

void pm_enable_pll(pm_pll_t const pll) {
	pm_write_register_field(pll, PM_REG_PLLX_PLLEN_MASK, PM_REG_PLLX_PLLEN_MASK);
}

void pm_disable_pll(pm_pll_t const pll) {
	pm_write_register_field(pll, PM_REG_PLLX_PLLEN_MASK, 0);
}

void pm_configure_pll(pm_pll_t const pll, pm_pll_vco_freqeuncy_range_t const frequency_range, pm_pll_output_divider_t const output_divider, pm_pll_wide_bandwidth_mode_t const wide_bandwidth_mode, uint8_t const multiplier, uint8_t const divider, uint8_t const lock_delay) {
	pm_write_register(pll, frequency_range | output_divider | wide_bandwidth_mode | (multiplier << PM_REG_PLLX_PLLMUL_OFFSET) | (divider << PM_REG_PLLX_PLLDIV_OFFSET) | (lock_delay << PM_REG_PLLX_PLLCOUNT_OFFSET));
}

void pm_configure_and_enable_pll(pm_pll_t const pll, pm_pll_vco_freqeuncy_range_t const frequency_range, pm_pll_output_divider_t const output_divider, pm_pll_wide_bandwidth_mode_t const wide_bandwidth_mode, uint8_t const multiplier, uint8_t const divider, uint8_t const lock_delay) {
	pm_write_register(pll, PM_REG_PLLX_PLLEN_MASK | frequency_range | output_divider | wide_bandwidth_mode | (multiplier << PM_REG_PLLX_PLLMUL_OFFSET) | (divider << PM_REG_PLLX_PLLDIV_OFFSET) | (lock_delay << PM_REG_PLLX_PLLCOUNT_OFFSET));
}

void pm_enable_osc0(void) {
	uint32_t const value = pm_read_register(PM_REG_MCCTRL);

	pm_write_register(PM_REG_MCCTRL, value | PM_REG_MCCTRL_OSC0EN_MASK);
}

void pm_disable_osc0(void) {
	uint32_t const value = pm_read_register(PM_REG_MCCTRL);

	pm_write_register(PM_REG_MCCTRL, value & ~PM_REG_MCCTRL_OSC0EN_MASK);
}

void pm_enable_osc1(void) {
	uint32_t const value = pm_read_register(PM_REG_MCCTRL);

	pm_write_register(PM_REG_MCCTRL, value | PM_REG_MCCTRL_OSC1EN_MASK);
}

void pm_disable_osc1(void) {
	uint32_t const value = pm_read_register(PM_REG_MCCTRL);

	pm_write_register(PM_REG_MCCTRL, value & ~PM_REG_MCCTRL_OSC1EN_MASK);
}

void pm_switch_main_clock(pm_main_clock_t clock) {
	uint32_t const value = pm_read_register(PM_REG_MCCTRL);

	pm_write_register(PM_REG_MCCTRL,
			(value & ~PM_REG_MCCTRL_MCSEL_MASK) | clock);
}

void pm_wait(pm_power_oscillator_status_t status) {
	while(!(pm_read_register(PM_REG_POSCSR) & status)) ;
}

void pm_generic_clock_enable(pm_generic_clock_t const generic_clock) {
	pm_write_register_field(generic_clock, PM_REG_GCCTRL_CEN_MASK, PM_REG_GCCTRL_CEN_MASK);
}

void pm_generic_clock_disable(pm_generic_clock_t const generic_clock) {
	pm_write_register_field(generic_clock, PM_REG_GCCTRL_CEN_MASK, 0);
}

void pm_generic_clock_configure(pm_generic_clock_t const generic_clock, pm_generic_clock_source_t const generic_clock_source, pm_generic_clock_divider_t const divider, uint32_t const divider_value) {
	pm_write_register_field(generic_clock, PM_REG_GCCTRL_OSCSEL_MASK | PM_REG_GCCTRL_PLLSEL_MASK | PM_REG_GCCTRL_DIVEN_MASK | PM_REG_GCCTRL_DIV_MASK, generic_clock_source | divider | ((divider_value << PM_REG_GCCTRL_DIV_OFFSET) & PM_REG_GCCTRL_DIV_MASK));
}
