/*
 * Copyright 2011 Florian Albrechtskirchinger. All rights reserved.
 *
 * This file is part of libat32uc3.
 *
 * libat32uc3 is licensed under the Simplified BSD License
 * (the "License").
 *
 * A copy of the License should be included in the distribution.
 * Alternatively, you may obtain a copy of the License at
 * http://www.opensource.org/licenses/bsd-license.php
 *
 * Author(s): Florian Albrechtskirchinger <falbrechtskirchinger@gmail.com>
 */

#include <at32uc3/pwm.h>

uint32_t pwm_read_register(uint32_t const offset) {
	uint32_t volatile const * const p = (uint32_t volatile *)(PWM_ADDRESS + offset);

	return *p;
}

void pwm_write_register(uint32_t const offset, uint32_t const value) {
	uint32_t volatile * const p = (uint32_t volatile *)(PWM_ADDRESS + offset);

	*p = value;
}

void pwm_configure_clock_a(pwm_clock_t const prea, uint8_t const diva) {
	uint32_t const value = pwm_read_register(PWM_REG_MR);

	pwm_write_register(PWM_REG_MR,
			(value & ~(PWM_REG_MR_PREA_MASK | PWM_REG_MR_DIVA_MASK)) |
			(prea << PWM_REG_MR_PREA_OFFSET) |
			(diva << PWM_REG_MR_DIVA_OFFSET));
}

void pwm_configure_clock_b(pwm_clock_t const preb, uint8_t const divb) {
	uint32_t const value = pwm_read_register(PWM_REG_MR);

	pwm_write_register(PWM_REG_MR,
			(value & ~(PWM_REG_MR_PREB_MASK | PWM_REG_MR_DIVB_MASK)) |
			(preb << PWM_REG_MR_PREB_OFFSET) |
			(divb << PWM_REG_MR_DIVB_OFFSET));
}

void pwm_configure_clocks(pwm_clock_t const prea, uint8_t const diva, pwm_clock_t const preb, uint8_t const divb) {
	pwm_write_register(PWM_REG_MR,
			(prea << PWM_REG_MR_PREA_OFFSET) |
			(diva << PWM_REG_MR_DIVA_OFFSET) |
			(preb << PWM_REG_MR_PREB_OFFSET) |
			(divb << PWM_REG_MR_DIVB_OFFSET));
}

void pwm_enable_channel(uint8_t const channel_no) {
	pwm_write_register(PWM_REG_ENA, 1 << channel_no);
}

void pwm_disable_channel(uint8_t const channel_no) {
	pwm_write_register(PWM_REG_DIS, 1 << channel_no);
}

uint32_t pwm_channel_read_register(void volatile *channel, uint32_t const offset) {
	uint32_t volatile * const p = (uint32_t volatile *)(channel + offset);

	return *p;
}

void pwm_channel_write_register(void volatile *channel, uint32_t const offset, uint32_t const value) {
	uint32_t volatile * const p = (uint32_t volatile *)(channel + offset);

	*p = value;
}

void pwm_channel_write_register_field(void volatile *channel, uint32_t const offset, uint32_t const field_mask, uint32_t const value) {
	uint32_t const tmp = pwm_channel_read_register(channel, offset);

	pwm_channel_write_register(channel, offset, (tmp & ~field_mask) | value);
}

void pwm_channel_set_clock(void volatile *channel, pwm_channel_clock_t const clock) {
	pwm_channel_write_register_field(channel, PWM_CHANNEL_REG_CMR, PWM_CHANNEL_REG_CMR_CPRE_MASK, clock);
}

void pwm_channel_set_alignment(void volatile *channel, pwm_channel_alignment_t const alignment) {
	pwm_channel_write_register_field(channel, PWM_CHANNEL_REG_CMR, PWM_CHANNEL_REG_CMR_CALG_MASK, alignment);
}

void pwm_channel_set_polarity(void volatile *channel, pwm_channel_polarity_t const polarity) {
	pwm_channel_write_register_field(channel, PWM_CHANNEL_REG_CMR, PWM_CHANNEL_REG_CMR_CPOL_MASK, polarity);
}

void pwm_channel_set_update_mode(void volatile *channel, pwm_channel_update_mode_t const update_mode) {
	pwm_channel_write_register_field(channel, PWM_CHANNEL_REG_CMR, PWM_CHANNEL_REG_CMR_CPD_MASK, update_mode);
}

void pwm_channel_set_period(void volatile *channel, uint32_t const period) {
	pwm_channel_write_register(channel, PWM_CHANNEL_REG_CPRD, period);
}

void pwm_channel_set_duty_cycle(void volatile *channel, uint32_t const duty_cycle) {
	pwm_channel_write_register(channel, PWM_CHANNEL_REG_CDTY, duty_cycle);
}

void pwm_channel_configure(void volatile *channel, pwm_channel_clock_t const clock, pwm_channel_alignment_t const alignment, pwm_channel_polarity_t const polarity, pwm_channel_update_mode_t const update_mode) {
	pwm_channel_write_register(channel, PWM_CHANNEL_REG_CMR, clock | alignment | polarity | update_mode);
}

void pwm_channel_configure2(void volatile *channel, pwm_channel_clock_t const clock, pwm_channel_alignment_t const alignment, pwm_channel_polarity_t const polarity, pwm_channel_update_mode_t const update_mode, uint32_t const period, uint32_t const duty_cycle) {
	pwm_channel_configure(channel, clock, alignment, polarity, update_mode);
	pwm_channel_set_period(channel, period);
	pwm_channel_set_duty_cycle(channel, duty_cycle);
}

void pwm_channel_update(void volatile *channel, uint32_t value) {
	pwm_channel_write_register(channel, PWM_CHANNEL_REG_CUPD, value);
}

