/*
 * Copyright 2011 Florian Albrechtskirchinger. All rights reserved.
 *
 * This file is part of libat32uc3.
 *
 * libat32uc3 is licensed under the Simplified BSD License
 * (the "License").
 *
 * A copy of the License should be included in the distribution.
 * Alternatively, you may obtain a copy of the License at
 * http://www.opensource.org/licenses/bsd-license.php
 *
 * Author(s): Florian Albrechtskirchinger <falbrechtskirchinger@gmail.com>
 */

#include <at32uc3/usart.h>

static void usart_reset(void volatile *usart);
static usart_oversampling_t usart_set_sync_baud_rate(void volatile *usart, uint32_t const baud, uint32_t const clock_freq);
static usart_oversampling_t usart_set_async_baud_rate(void volatile *usart, uint32_t const baud, uint32_t const clock_freq);
static usart_oversampling_t usart_set_async_fp_baud_rate(void volatile *usart, uint32_t const baud, uint32_t const clock_freq);

void usart_reset(void volatile *usart) {
	usart_write_register(usart, USART_REG_MR, 0);
	usart_write_register(usart, USART_REG_RTOR, 0);
	usart_write_register(usart, USART_REG_TTGR, 0);
	usart_write_register(usart, USART_REG_CR,
		USART_REG_CR_RSTRX_MASK |
		USART_REG_CR_RSTTX_MASK |
		USART_REG_CR_RSTSTA_MASK |
		USART_REG_CR_RSTIT_MASK |
		USART_REG_CR_RSTNACK_MASK |
		USART_REG_CR_RTSDIS_MASK);
}

usart_oversampling_t usart_set_sync_baud_rate(void volatile *usart, uint32_t const baud, uint32_t const clock_freq) {
	uint32_t cd = clock_freq / baud;

	usart_write_register(usart, USART_REG_BRGR,
		(cd << USART_REG_BRGR_CD_OFFSET));

	return 0;
}

usart_oversampling_t usart_set_async_baud_rate(void volatile *usart, uint32_t const baud, uint32_t const clock_freq) {
	uint32_t over = (clock_freq >= (baud << 4)) ? 16 : 8;
	uint32_t cd = over * clock_freq / baud;

	usart_write_register(usart, USART_REG_BRGR,
		(cd << USART_REG_BRGR_CD_OFFSET));

	return (over == 16) ? USART_OVERSAMPLING_X16 : USART_OVERSAMPLING_X8;
}

usart_oversampling_t usart_set_async_fp_baud_rate(void volatile *usart, uint32_t const baud, uint32_t const clock_freq) {
	uint32_t over = (clock_freq >= (baud << 4)) ? 16 : 8;
	uint32_t cd_fp = ((1 << 3) * clock_freq + (over * baud) / 2) / (over * baud);
	uint32_t cd = cd_fp >> 3;
	uint32_t fp = cd_fp & ((1 << 3) - 1);

	usart_write_register(usart, USART_REG_BRGR,
		(cd << USART_REG_BRGR_CD_OFFSET) |
		(fp << USART_REG_BRGR_FP_OFFSET));

	return (over == 16) ? USART_OVERSAMPLING_X16 : USART_OVERSAMPLING_X8;
}


uint32_t usart_read_register(void volatile *usart, uint32_t const offset) {
	uint32_t const volatile * const p = (uint32_t *)(usart + offset);

	return *p;
}

void usart_write_register(void volatile *usart, uint32_t const offset, uint32_t const value) {
	uint32_t volatile * const p = (uint32_t *)(usart + offset);

	*p = value;
}

void usart_write_register_field(void volatile *usart, uint32_t const offset, uint32_t const field_mask, uint32_t const value) {
	uint32_t const tmp = usart_read_register(usart, offset);

	usart_write_register(usart, offset, (tmp & ~field_mask) | value);
}

void usart_configure_rs232(void volatile *usart, usart_rs232_mode_t const mode, usart_channel_mode_t const channel_mode, usart_bit_order_t const bit_order, uint8_t const char_length, usart_stopbits_t const stopbits, usart_parity_t const parity, uint32_t const baud, uint32_t const pba_freq) {
	usart_oversampling_t oversampling;
	uint32_t stopbits_mask;
	uint32_t char_length_mask;

	usart_reset(usart);

	if(mode & USART_REG_MR_SYNC_MASK)
		oversampling = usart_set_sync_baud_rate(usart, baud, pba_freq);
	else if(mode == USART_RS232_MODE_NORMAL_ASYNC)
		oversampling = usart_set_async_fp_baud_rate(usart, baud, pba_freq);
	else if(mode == USART_RS232_MODE_HARDWARE_HANDSHAKING_ASYNC)
		oversampling = usart_set_async_baud_rate(usart, baud, pba_freq);

	if(stopbits > 2) {
		stopbits_mask = USART_STOPBITS_2 << USART_REG_MR_NBSTOP_OFFSET;
		usart_write_register(usart, USART_REG_TTGR, stopbits - 2);
	}
	else
		stopbits_mask = stopbits << USART_REG_MR_NBSTOP_OFFSET;

	if(char_length > 8)
		char_length_mask = USART_REG_MR_MODE9_MASK;
	else
		char_length_mask = (char_length - 5) << USART_REG_MR_CHRL_OFFSET;

	usart_write_register(usart, USART_REG_MR,
		mode | USART_CLOCK_MCK | char_length_mask |
		parity | stopbits_mask | channel_mode | oversampling);
}

void usart_enable_rx(void volatile *usart) {
	usart_write_register(usart, USART_REG_CR, USART_REG_CR_RXEN_MASK);
}

void usart_disable_rx(void volatile *usart) {
	usart_write_register(usart, USART_REG_CR, USART_REG_CR_RXDIS_MASK);
}

void usart_enable_tx(void volatile *usart) {
	usart_write_register(usart, USART_REG_CR, USART_REG_CR_TXEN_MASK);
}

void usart_disable_tx(void volatile *usart) {
	usart_write_register(usart, USART_REG_CR, USART_REG_CR_TXDIS_MASK);
}

void usart_enable(void volatile *usart) {
	usart_write_register(usart, USART_REG_CR,
		USART_REG_CR_RXEN_MASK | USART_REG_CR_TXEN_MASK);
}

void usart_disable(void volatile *usart) {
	usart_write_register(usart, USART_REG_CR,
		USART_REG_CR_RXDIS_MASK | USART_REG_CR_TXDIS_MASK);
}

void usart_reset_status(void volatile *usart) {
	usart_write_register(usart, USART_REG_CR, USART_REG_CR_RSTSTA_MASK);
}

usart_rc_t usart_try_read_char(void volatile *usart, char *c) {
	uint32_t csr = usart_read_register(usart, USART_REG_CSR);

	if(csr & (USART_REG_CSR_OVRE_MASK | USART_REG_CSR_FRAME_MASK | USART_REG_CSR_PARE_MASK))
		return USART_RC_RX_ERROR;

	if(csr & USART_REG_CSR_RXRDY_MASK) {
		*c = (usart_read_register(usart, USART_REG_RHR) & USART_REG_RHR_RXCHR_MASK) >> USART_REG_RHR_RXCHR_OFFSET;
		return USART_RC_SUCCESS;
	}
	else
		return USART_RC_RX_EMPTY;
}

usart_rc_t usart_read_char(void volatile *usart, char *c) {
	usart_rc_t rc;

	while((rc = usart_try_read_char(usart, c)) == USART_RC_RX_EMPTY) ;

	return rc;
}

void usart_write_char(void volatile *usart, char const c) {
	for(uint32_t csr = usart_read_register(usart, USART_REG_CSR); !(csr & USART_REG_CSR_TXRDY_MASK); csr = usart_read_register(usart, USART_REG_CSR)) ;

	usart_write_register(usart, USART_REG_THR, c);
}

void usart_write_char_auto_cr(void volatile *usart, char const c) {
	if(c == '\n')
		usart_write_char(usart, '\r');
	usart_write_char(usart, c);
}

void usart_write_string(void volatile *usart, char const *str) {
	while (*str)
		usart_write_char_auto_cr(usart, *(str++));
}
