/*
 * -----------------------------------------------------------------------------
 * file:        oratest.c
 * purpose:     tests SQL query of a Oracle database with libsqlora
 * author:      04/23/2006 Frank4DD
 *
 * gcc -O3 -Wall -g -I/home/oracle/include -Wall   -c -o oratest.o oratest.c
 *                   ^^^^^^^^^^^^^^^^^^^^^
 *                   location of sqlora.h and Oracle instantclient headers
 *
 * gcc -L/home/oracle/lib -L/home/oracle/lib32 -lclntsh -lsqlora8 oratest.o -o oratest
 *      ^^^^^^^^^^^^^^^^^  ^^^^^^^^^^^^^^^^^^^
 *      location of libsqlora  location of instantclient libs
 *
 * -----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <sqlora.h>
#include <edacs.h>

/**********************************************************/
/* adjust these parameters to get connectivity to Oracle  */
/* The Defines are equivalent to their counterpart in     */
/* edacs.h (removing the MY_), so when this program works */
/* the variables in edacs.h can be updated and the main   */
/* application will equally work.                         */
/**********************************************************/
#define MY_ORACLEHOME_ENV "ORACLE_HOME=/var/edacsdb/oracle"
#define MY_TNS_STRING "system/test@work1"
/**********************************************************/
/*      No need to adjust anything below this line        */
/**********************************************************/

#define error_exit(handle, msg)  do_error_exit(handle, __FILE__, __LINE__, msg)
void do_error_exit(sqlo_db_handle_t dbh, const char * file,
                                 int line, const char * msg);

/* already defined in edacs.h: */
/* static int _abort_flag = 0; */

/* functions */
void sigint_handler(void) {
  printf("Catched SIGINT\n");
  _abort_flag++;
}

int col_count(sqlo_db_handle_t dbh, char * table_name) {
  int stat;
  if ( 0 > (stat = sqlo_count(dbh, "USER_TAB_COLUMNS", "TABLE_NAME",
                              table_name, NULL)))
    error_exit(dbh, "sqlo_count");
  return stat;
}

/* print error and exit with failure */
void do_error_exit(sqlo_db_handle_t dbh, const char * file,
                                 int line, const char * msg) {
  printf("%s (line: %d):\n%s: %s\n", file, line, msg, sqlo_geterror(dbh));
  sqlo_rollback(dbh);
  sqlo_finish(dbh);
  exit(EXIT_FAILURE);
}

int cgiMain() {

  sqlo_db_handle_t dbh;		/* database handle */
  int stat;                     /* status of sqlo calls */
  sqlo_stmt_handle_t sth1;      /* statement handle 1 */
  sqlo_stmt_handle_t sth2;      /* statement handle 2 */
  sqlo_stmt_handle_t sth3;      /* statement handle 3 */
  char server_version[1024];	/* string for returned server version */
  char sqlquery_str[1024];	/* SQL query string */
  int handle;                   /* handle of the interrupt handler */
  const char ** v;              /* values */
  const char ** n;              /* column names */
  CONST int *nl;                /* column name lengths */
  CONST unsigned short *vl;     /* value lengths */
  int nc;                       /* number of columns */
  unsigned int i, j;		/* loop variable, why is it always i? */

  char sql_insert_str[4096] = "INSERT INTO edacsadmin.EDACS_TEMPLOG (action, router, username, action_date, bytes_in, bytes_out, ip_addr, tty, protocol, packets_in, packets_out, task_id, service, ip_or_phone) VALUES ('Start', '192.168.89.11', 'kschlieg', TO_DATE('2008-08-15 01:57:35', 'YYYY-MM-DD HH24:MI:SS'), 740120, 17082256, '192.168.91.1', '1048','PPP', 10665, 25044, 'D100005D', 'Framed-User', '192.168.42.175')";

  char sql_drop_str[4096] = "DELETE edacsadmin.EDACS_TEMPLOG WHERE router='192.168.89.11' and  username='kschlieg' and TO_CHAR(action_date, 'YYYY-MM-DD HH24:MI:SS')='2008-08-15 01:57:35'";

  /***********************************************************************/
  /* ORACLE_HOME is needed for OCI8 to find tnsnames.ora and other stuff */
  /***********************************************************************/
  putenv(MY_ORACLEHOME_ENV);

  /* initialize the connection */
  if (SQLO_SUCCESS != sqlo_init(SQLO_OFF, 1, 100)) {
    printf("Failed to init libsqlora8\n");
    return EXIT_FAILURE;
  }

  /* register the interrupt handler */
  sqlo_register_int_handler(&handle, sigint_handler);

  /* login to the database */
  if (SQLO_SUCCESS != sqlo_connect(&dbh, MY_TNS_STRING)) {
    printf("Cannot login with %s\n", MY_TNS_STRING);
    return EXIT_FAILURE;
  }
  RETURN_ON_ABORT; /* finish if SIGINT was catched */

  if (SQLO_SUCCESS != sqlo_server_version(dbh, server_version,
                                        sizeof(server_version))) {
    printf("Failed to get the server version: %s\n", sqlo_geterror(dbh));
    return EXIT_FAILURE;
  }
  RETURN_ON_ABORT; /* finish if SIGINT was catched */

  printf("Connected to:\n%s\n\n", server_version);

  /* enable autocommit, each statement is committet as a single transaction */
  stat = sqlo_set_autocommit(dbh, 1);

  /* parse the statement */

  if (0>(sth1 = sqlo_prepare(dbh, sql_insert_str)))
    error_exit(dbh, "sqlo_prepare");

  printf("-------------------------------------------------------------------\n");
  printf("Insert table: %s\n", sql_insert_str);

  /* execute the statement */
  stat = sqlo_execute(sth1, 1);
  printf("Finished inserting with status %d.\n", stat);
  if(stat <0) printf("Finished inserting with error %s.\n", sqlo_geterror(dbh));

  if ( SQLO_SUCCESS != sqlo_close(sth1))
    error_exit(dbh, "sqlo_close");
  printf("-------------------------------------------------------------------\n\n");

  snprintf(sqlquery_str, sizeof(sqlquery_str), "SELECT * FROM edacsadmin.EDACS_TEMPLOG");

  sth2 = SQLO_STH_INIT;
  if ( 0 > (sqlo_open2(&sth2, dbh, sqlquery_str, 0, NULL)))
    error_exit(dbh, "sqlo_open");

  printf("-------------------------------------------------------------------\n");
  printf("Select from table: %s\n", sqlquery_str);
  /* get the output column names */
  n = sqlo_ocol_names(sth2, &nc);

  /* get the output column name lengths */
  nl = sqlo_ocol_name_lens(sth2, NULL);

  printf("number of output columns: %d \n\n", nc);

  /* print the table column header(s) */
  for (i = 0; i < nc; ++i) printf("%-*s ", nl[i], n[i]);
  printf("\n");

  for (i = 0; i < nc; ++i) {
    for (j = 0; j < nl[i]; ++j) putchar('-');
    putchar('+');
  }
  putchar('\n');

  /* fetch the data */
  while ( SQLO_SUCCESS == (stat = (sqlo_fetch(sth2, 1)))) {

    /* get one record */
    v = sqlo_values(sth2, NULL, 1);

    /* get the length of the data items */
    vl = sqlo_value_lens(sth2, NULL);

    /* print the column values */
    for (i = 0; i < nc; ++i)
      printf("%-*s ", (vl[i] > nl[i] ? vl[i] : nl[i]), v[i]);
    printf("\n");
  }

  if (0 > stat)
    error_exit(dbh, "sqlo_fetch");

  if ( SQLO_SUCCESS != sqlo_close(sth2))
    error_exit(dbh, "sqlo_close");
  printf("-------------------------------------------------------------------\n\n");

  /* parse the 3 statement */

  if (0>(sth3 = sqlo_prepare(dbh, sql_drop_str)))
    error_exit(dbh, "sqlo_prepare");

  printf("-------------------------------------------------------------------\n");
  printf("Delete rows in table: %s\n", sql_drop_str);

  /* execute the statement */
  stat = sqlo_execute(sth3, 1);
  printf("Finished dropping with status %d.\n", stat);
  if(stat <0) printf("Finished dropping with error %s.\n", sqlo_geterror(dbh));

  if ( SQLO_SUCCESS != sqlo_close(sth3))
    error_exit(dbh, "sqlo_close");
  printf("-------------------------------------------------------------------\n");

return 1;
}

/******************************************************************************
 * fm@susie102:/home/devel/edacs-1.1/demo> ./oratest2
 * Connected to:
 * Oracle Database 10g Enterprise Edition Release 10.1.0.2.0 - Production
 * With the Partitioning, OLAP and Data Mining options
 * 
 * -------------------------------------------------------------------
 * Insert table: INSERT INTO edacsadmin.EDACS_TEMPLOG (action, router, 
 * username, action_date, bytes_in, bytes_out, ip_addr, tty, protocol,
 * packets_in, packets_out, task_id, service, ip_or_phone) VALUES ('Start',
 * '192.168.89.11', 'kschlieg', TO_DATE('2008-08-15 01:57:35',
 * 'YYYY-MM-DD HH24:MI:SS'), 740120, 17082256, '192.168.91.1', '1048','PPP',
 * 10665, 25044, 'D100005D', 'Framed-User', '192.168.42.175')
 * Finished inserting with status 0.
 * -------------------------------------------------------------------
 * 
 * -------------------------------------------------------------------
 * Select from table: SELECT * FROM edacsadmin.EDACS_TEMPLOG
 * number of output columns: 14
 * 
 * ACTION ROUTER USERNAME ACTION_DATE BYTES_IN BYTES_OUT IP_ADDR TTY PROTOCOL
 * PACKETS_IN PACKETS_OUT TASK_ID SERVICE IP_OR_PHONE
 * ------+------+--------+-----------+--------+---------+-------+---+--------+
 * ----------+-----------+-------+-------+-----------+
 * Start  192.168.89.11 kschlieg 15-AUG-08   740120   17082256  192.168.91.1
 * 1048 PPP      10665      25044       D100005D Framed-User 192.168.42.175
 * -------------------------------------------------------------------
 * 
 * -------------------------------------------------------------------
 * Delete rows in table: DELETE edacsadmin.EDACS_TEMPLOG WHERE
 * router='192.168.89.11' and  username='kschlieg' and TO_CHAR(action_date,
 * 'YYYY-MM-DD HH24:MI:SS')='2008-08-15 01:57:35'
 * Finished dropping with status 0.
 * -------------------------------------------------------------------
 ******************************************************************************/
