/* ------------------------------------------------------------ *
 * file:        archive_file.c                                  *
 * purpose:     archive raw data files into archive directory   *
 *              schema based on the loader date, compressed     *
 * author:      08/23/2006 Frank4DD                             *
 * remarks:     It is not important that files are archived in  *
 *              sorted order, but logs will be more consistent. *
 * ------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
//#include "getline.h"
#include "../edacs.h"

/* ------------------------------------------------------------ *
 * archive_file(): archive the raw .rad data files              *
 * ------------------------------------------------------------ */
int archive_file() {
  int radcount = 0;
  int i        = 0;
  struct dirent **radfile_list;

  radcount = scandir(RAW_DIR, &radfile_list, rad_select, alphasort);

  if(radcount == 0) {
    fprintf(logfile, "[archive_file] !Error!: No .rad file found in %s.\n", RAW_DIR);
    return 0;
  }

  /* --------------------------------------------------------- *
   * Cycle through the ordered .rad files found in RAW_DIR     *
   * --------------------------------------------------------- */
  i = 0;
  while(i<radcount) {
   struct stat radentrybuf;
   char radentrypath[256]  = "";

    /* --------------------------------------------------------- *
     * generate the full path and file name of the .rad file     *
     * --------------------------------------------------------- */
    snprintf(radentrypath, sizeof(radentrypath),
        "%s/%s", RAW_DIR, radfile_list[i]->d_name);

    /* --------------------------------------------------------- *
     * debug: print the name of the .rad raw file                *
     * --------------------------------------------------------- */
    if(DEBUG >0) printf("[archive_file] RAD data file %d: %s\n", i, radentrypath);

    /* --------------------------------------------------------- *
     * try to stat the .rad file                                 *
     * --------------------------------------------------------- */
    if (stat(radentrypath, &radentrybuf) == -1) {
      snprintf(errormsg, sizeof(errormsg),
             "[parse_radius] !Error!: Cannot stat %s.\n", radentrypath);
      ldr_error(errormsg);
      continue;
    }

    /* --------------------------------------------------------- *
     * if the .rad file is empty, delete it                      *
     * --------------------------------------------------------- */
    if( (radentrybuf.st_size) == 0) {
      fprintf(logfile, "[parse_radius] Warning: Empty file %s.\n", radentrypath);
      if(unlink(radentrypath) != 0) {
        snprintf(errormsg, sizeof(errormsg),
             "[parse_radius] !Error!: Cannot unlink %s.\n", radentrypath);
        ldr_error(errormsg);
      }
      else
        fprintf(logfile, "[parse_radius] Warning: Erased file %s.\n", radentrypath);

      continue;
    }

    /* ------------------------------------------------------- *
     * prepare the archive directory based on the filename     *
     * date extension schema                                   *
     * ------------------------------------------------------- */
    char arch_tstampbuf[11]   = "";
    char arch_year[5]         = "";
    char arch_month[3]        = "";
    char arch_day[3]          = "";
    char arch_year_path[256]  = "";
    char arch_month_path[256] = "";
    char arch_day_path[256]   = "";
    char radarchivepath[256]  = "";

    strncpy(arch_tstampbuf, strchr(radfile_list[i]->d_name, '-') + 1, sizeof(arch_tstampbuf));
    arch_tstampbuf[10] = '\0';
    strncpy(arch_year, arch_tstampbuf, sizeof(arch_year));
    arch_year[4] = '\0';
    strncpy(arch_month, strchr(arch_tstampbuf, '.') + 1, sizeof(arch_month));
    arch_month[2] = '\0';
    strncpy(arch_day, strrchr(arch_tstampbuf, '.') + 1, sizeof(arch_day));
    arch_day[2] = '\0';

    if(DEBUG >2) printf("[archive_file] RAD archive tstamp: %s year: %s month: %s day: %s\n",
                                               arch_tstampbuf, arch_year, arch_month, arch_day);

    /* ------------------------------------------------------- *
     * Generate the archive path names                         *
     * ------------------------------------------------------- */
    snprintf(arch_year_path, sizeof(arch_year_path),
        "%s/%s", ARCHIVE_DIR, arch_year);
    snprintf(arch_month_path, sizeof(arch_month_path),
        "%s/%s", arch_year_path, arch_month);
    snprintf(arch_day_path, sizeof(arch_day_path),
        "%s/%s", arch_month_path, arch_day);

    if(DEBUG >2) printf("[archive_file] RAD archive path: %s\n", arch_day_path);

    /* ------------------------------------------------------- *
     * Create the archive directories if they do not exist     *
     * ------------------------------------------------------- */
    if(opendir(arch_day_path) == NULL && errno == ENOENT) {

      if(opendir(arch_month_path) == NULL && errno == ENOENT) {

        if(opendir(arch_year_path) == NULL && errno == ENOENT)  {
          if(mkdir(arch_year_path, 00755) != 0) {
            snprintf(errormsg, sizeof(errormsg),
               "[archive_file] !Error!: Cannot create %s.\n", arch_year_path);
            ldr_error(errormsg);
          }
          fprintf(logfile, "[archive_file] Created archive directory: %s/.\n",
                                                                     arch_year);
        }

        if(mkdir(arch_month_path, 00755) != 0) {
          snprintf(errormsg, sizeof(errormsg),
             "[archive_file] !Error!: Cannot create %s.\n", arch_month_path);
          ldr_error(errormsg);
        }
        fprintf(logfile, "[archive_file] Created archive directory: %s/%s/.\n",
                                                          arch_year, arch_month);
      }

      if(mkdir(arch_day_path, 00755) != 0) {
            snprintf(errormsg, sizeof(errormsg),
               "[archive_file] !Error!: Cannot create %s.\n", arch_day_path);
            ldr_error(errormsg);
      }
      fprintf(logfile, "[archive_file] Created archive directory: %s/%s/%s/.\n",
                                                arch_year, arch_month, arch_day);
    }

    /* ------------------------------------------------- *
     * generate the full path of the archive file        *
     * ------------------------------------------------- */
    snprintf(radarchivepath, sizeof(radarchivepath),
        "%s/%s", arch_day_path, radfile_list[i]->d_name);

    if(DEBUG >2) printf("[archive_file] archive file: %s\n", radarchivepath);

    /* ------------------------------------------------------- *
     * Archive the radius accounting file, move from RAW_DIR   *
     * to ARCHIVE_DIR/<year>/<month>/<day>/                    *
     * ------------------------------------------------------- */

    if(link(radentrypath, radarchivepath) != 0) {
      snprintf(errormsg, sizeof(errormsg),
             "[archive_file] !Error!: Cannot move %s to %s.\n", radentrypath, radarchivepath);
      ldr_error(errormsg);
      i++;
      continue;
    }

    /* ------------------------------------------------- *
     * delete the original data file in RAW_DIR          *
     * ------------------------------------------------- */
    if(unlink(radentrypath) != 0) {
      snprintf(errormsg, sizeof(errormsg),
               "[archive_file] !Error!: Cannot unlink %s.\n", radentrypath);
      ldr_error(errormsg);
      i++;
      continue;
    }

    /* ------------------------------------------------- *
     * compress the archived data file in ARCHIVE_DIR... *
     * ------------------------------------------------- */
    char systemcall[256] = "";
    snprintf(systemcall, sizeof(systemcall),
        "%s %s", GZIP_BIN, radarchivepath);

    if(system(systemcall) == -1) {
      snprintf(errormsg, sizeof(errormsg),
               "[archive_file] !Error!: Cannot compress %s.\n", radarchivepath);
      ldr_error(errormsg);
      i++;
      continue;
    }

    fprintf(logfile, "[archive_file] %s: archived radius file\n", radfile_list[i]->d_name);

    /* update the total number of archived radius records */
    rad_archivedfiles++;
    i++;
  }

  if(rad_archivedfiles > 0) {
    if(DEBUG >0) printf("[archive_file] Archived RAD data files: %d\n", rad_archivedfiles);
  }
  else {
    fprintf(logfile, "[archive_file] !Error!: No radius files archived.\n");
  }

  return rad_archivedfiles;
}
