/* ---------------------------------------------------------------------- *
 * file:        edacs_loader.c                                            *
 * purpose:     parse, load and archive radius/tacacs+ accounting files   *
 *              This is the main program to be called                     *
 * author:      08/18/2006 Frank4DD                                       *
 * usage:       run via crontab hourly or every 30mins, depending on load *
 * ---------------------------------------------------------------------- */

#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>  /* FOR EXIT() */
#include "../edacs.h"

/* ---------------------------------------------------------------------- *
 * handle_error_ldr(): the error handler for errors in the edacs-loader   *
 * ---------------------------------------------------------------------- */
void handle_error_ldr(const char *file, int lineno, const char *msg) {

  fprintf(logfile, "%s. %s - file %s line %d.\n", msg, strerror(errno), file, lineno);
  if(DEBUG >0) perror(msg);
}

/* ---------------------------------------------------------------------- *
 * handle_error_dbs(): is the error handler for errors with the database  *
 * -----------------------------------------------------------------------*/
#ifdef MYSQL_DB
  void handle_error_dbs(MYSQL *dbh, const char * file, int lineno, const char * msg) {
    fprintf(logfile, "%s file %s line:%d\nMySQL  Error: %u %s\n", msg, file,
                                  lineno, mysql_errno(dbh), mysql_error(dbh));
  mysql_rollback(dbh);
  }
#endif

#ifdef ORACLE_DB
  void handle_error_dbs(sqlo_db_handle_t dbh, const char * file,
                                 int lineno, const char * msg) {
    fprintf(logfile, "%s file %s line:%d\nOracle Error: %s\n", msg, file,
                                                   lineno, sqlo_geterror(dbh));
    sqlo_rollback(dbh);
  }
#endif

/* ---------------------------------------------------------------------- *
 * rad_select(): runs through the RAW_DIR directory and selects all files *
 * identified by the .rad.gz extension.                                   *
 * ---------------------------------------------------------------------- */
int rad_select(const struct dirent *entry) {
  char *lastdot = NULL;         /* file extension separator   */

  /* --------------------------------------------------------------------- *
   * check for "." and ".." directory entries to sort them out             *
   * --------------------------------------------------------------------- */
  if(entry->d_name[0]=='.') return 0;

  /* --------------------------------------------------------------------- *
   * single out the .rad files in RAW_DIR to process                       *
   * --------------------------------------------------------------------- */
  lastdot = strrchr(entry->d_name, '.');
  if(strcmp(lastdot, ".rad") != 0) return 0;

  return 1;
}

/* ----------------------------------------------------------------------- *
 * ************** This is the main routine for the edacs loader ********** *
 * ------------------------------------------------------------------------*/
int main() {
  int ret = 0;
  time_t now;
  struct tm *nowtmptr;
  char logfilepath[256]  = "";

  /* --------------------------------------------------------------------- *
   * generate the timestamp for the global ldr_tstamp value. It will be    *
   * used as identifier in all filenames generated during this loader run. *
   * --------------------------------------------------------------------- */
  now = time(NULL);
  nowtmptr = localtime(&now);
  strftime(ldr_tstring, sizeof(ldr_tstring), "%Y.%m.%d.%H%M", nowtmptr);

  /* --------------------------------------------------------------------- *
   * generate the full path of the logfile                                 *
   * --------------------------------------------------------------------- */
  snprintf(logfilepath, sizeof(logfilepath),
      "%s/edacs-%s.log", LOG_DIR, ldr_tstring);

  /* --------------------------------------------------------------------- *
   * debug: print the name of the logfile                                  *
   * --------------------------------------------------------------------- */
  if(DEBUG >0) printf("[edacs_loader] open logfile: %s.\n", logfilepath);

  /* --------------------------------------------------------------------- *
   * open the new logfile for writing                                      *
   * --------------------------------------------------------------------- */
  if (! (logfile = fopen(logfilepath, "w")))
     printf("[edacs_loader] Cannot open logfile: %s.\n", logfilepath);

  fprintf(logfile, "[edacs_loader] started new edacs run at: %s", ctime(&now));

  /* ---------------------------------------------------------------------- *
   * split detail files in RADACCT_DIR subdirs                              *
   * ---------------------------------------------------------------------- */
  ret = split_radius();

  if(ret > 0) { 
  fprintf(logfile, "[edacs_loader] Finished [split_radius]. processed Radius files: %d\n", ret);
  }
  else {
    fprintf(logfile, "[edacs_loader] Failed to get data, check for hand-fed logs in raw.\n");
  }

  /* ---------------------------------------------------------------------- *
   * Verify basic DB structures and tables before continuing to load data   *
   * ---------------------------------------------------------------------- */
  ret = verify_dbase();

  if(ret != 0) {
    fprintf(logfile, "[edacs_loader] Finished [verify_dbase]. Database is not OK.\n");
    return -1;
  }
  fprintf(logfile, "[edacs_loader] Finished [verify_dbase]. database is OK.\n");

  /* ---------------------------------------------------------------------- *
   * parse_radius() parses the raw radius files in RAW_DIR and writes       * 
   * the results to GOOD_DIR. It returns the number of processed files.     * 
   * ---------------------------------------------------------------------- */
  ret = parse_radius();
  fprintf(logfile, "[edacs_loader] Finished [parse_radius]. processed records: %d in %d files.\n", radius_rawrecords, ret);

  /* ---------------------------------------------------------------------- *
   * load files in GOOD_DIR into the database, delete them after successful *
   * loading and archive the raw files from  RAW_DIR into ARCHIVE_DIR.      *
   * ---------------------------------------------------------------------- */
  ret = insert_dbase();
  fprintf(logfile, "[edacs_loader] Finished [insert_dbase]. loaded statements: %d in %d files.\n", sql_loadedrecords, ret);

  if(sql_loadedrecords < radius_rawrecords)
    fprintf(logfile, "[edacs_loader] missmatch between raw radius records [%d] and loaded sql records [%d]!\n", radius_rawrecords, sql_loadedrecords);

  if(ret > 0) {
    /* ------------------------------------------------------------------- *
     * Archive the original accounting data files                          *
     * ------------------------------------------------------------------- */
    ret = archive_file();
    fprintf(logfile, "[edacs_loader] Finished [archive_file]. archived files: %d\n", ret);
  }
  else {
    fprintf(logfile, "[edacs_loader] !Error!: Could not load a .sql file successfully.\n");
  }

  /* -------------------------------------------------------------------- *
   * Clean up orphaned records in edacs_templog table;                    *
   * but only if we run on current split data and not bulk data uploads   *
   * -------------------------------------------------------------------- */
  if (rad_detailfiles != 0) {
    ret = erase_orphan();
    fprintf(logfile, "[edacs_loader] Finished [erase_orphan]. Deleted records: %d\n", ret);
  } else {
    fprintf(logfile, "[edacs_loader] Skipping [erase_orphan]. No current split.\n");
  }

  /* -------------------------------------------------------------------- *
   * Clean up temporary generated .sql statement files                    *
   * -------------------------------------------------------------------- */
  ret = erase_sqltmp();
  fprintf(logfile, "[edacs_loader] Finished [erase_sqltmp]. Deleted files: %d\n", ret);

  /* -------------------------------------------------------------------- *
   * Clean up expired log files from previous runs of edacsdb             *
   * -------------------------------------------------------------------- */
  ret = erase_ldrlog();
  fprintf(logfile, "[edacs_loader] Finished [erase_ldrlog]. Deleted files: %d\n", ret);

  /* --------------------------------------------------------------------- *
   * Clean up expired sql error files from previous runs of edacsdb        *
   * --------------------------------------------------------------------- */
  ret = erase_sqlerr();
  fprintf(logfile, "[edacs_loader] Finished [erase_sqlerr]. Deleted files: %d\n", ret);

  /* --------------------------------------------------------------------- *
   * generate the ending timestamp                                         *
   * --------------------------------------------------------------------- */
  now = time(NULL);
  fprintf(logfile, "[edacs_loader] Finished edacs run at: %s", ctime(&now));

  if(fclose(logfile) != 0)
     printf("[edacs_loader] Cannot close logfile: %s.\n", logfilepath);

  return 1;
}
