/* ------------------------------------------------------------------------ *
 * file:        erase_ldrlog.c                                              *
 * purpose:     deletes expired edacs loader log files in the log directory *
 * author:      08/23/2006 Frank4DD                                         *
 * remarks:     housekeeping, remove logs older then x days, see edacs.h    *
 * ------------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
//#include "getline.h"
#include "../edacs.h"

/* ----------------------------------------------------------- *
 * log_select runs through the LOG_DIR directory and selects   *
 * all files identified by .log extension older then x days.   *
 * ----------------------------------------------------------- */
int log_select(const struct dirent *entry) {
  char *lastdot = NULL;         /* file extension separator    */
  char logfilepath[256] = "";   /* full path & name of logfile */
  struct stat logfilebuf;       /* logfile stat structure      */
  int expireseconds = 0;
  time_t exp;

  /* --------------------------------------------------------- *
   * check for "." and ".." directory entries to sort them out *
   * --------------------------------------------------------- */
  if(entry->d_name[0]=='.') return 0;

  /* --------------------------------------------------------- *
   * single out the .log files in LOG_DIR to process           *
   * --------------------------------------------------------- */
  lastdot = strrchr(entry->d_name, '.');
  if(strcmp(lastdot, ".log") != 0) return 0;

  /* --------------------------------------------------------- *
   * generate the full path and file name of the log files     *
   * --------------------------------------------------------- */
  snprintf(logfilepath, sizeof(logfilepath),
                     "%s/%s", LOG_DIR, entry->d_name);

  /* --------------------------------------------------------- *
   * try to stat the log file                                  *
   * --------------------------------------------------------- */
  if (stat(logfilepath, &logfilebuf) == -1) {
    snprintf(errormsg, sizeof(errormsg),
             "[erase_ldrlog] !Error!: Cannot stat %s.\n", logfilepath);
    ldr_error(errormsg);
    return 0;
  }

  /* --------------------------------------------------------- *
   * Calculate the expiration time                             *
   * --------------------------------------------------------- */
  expireseconds = KEEP_LOGS * 86400;
  exp = time(NULL) - expireseconds;
  if(DEBUG >2) printf("Delete logs older then: %s", ctime(&exp));

  /* --------------------------------------------------------- *
   *  process only logfiles older then KEEP_DAYS               *
   * --------------------------------------------------------- */
  if( (logfilebuf.st_mtime) >= exp) return 0;

  if(DEBUG >1) printf("Delete log: %s.\n", entry->d_name);
  return 1;
} /* end of log_select() */

/* ------------------------------------------------------------ *
 * erase_ldrlog(): delete expired loader log files files        *
 * ------------------------------------------------------------ */
int erase_ldrlog() {
  int logcount = 0;
  int i        = 0;
  struct dirent **logfile_list;

  /* ------------------------------------------------------------ *
   * Reset the global values                                      *
   * ------------------------------------------------------------ */
  log_erasedfiles = 0;

  /* ------------------------------------------------------------ *
   * Select and order the expired .log files found in LOG_DIR     *
   * ------------------------------------------------------------ */
  logcount = scandir(LOG_DIR, &logfile_list, log_select, alphasort);

  if(logcount == 0) {
    fprintf(logfile, "[erase_ldrlog] No log file expired in %s.\n", LOG_DIR);
    return logcount;
  }

  /* ----------------------------------------------------------- *
   * Cycle through the ordered list of expired .log files found  *
   * ----------------------------------------------------------- */
  i = 0;
  while(i<logcount) {
   char logentrypath[256] = "";

    /* --------------------------------------------------------- *
     * generate the full path and file name of the .log file     *
     * --------------------------------------------------------- */
    snprintf(logentrypath, sizeof(logentrypath),
        "%s/%s", LOG_DIR, logfile_list[i]->d_name);

    /* --------------------------------------------------------- *
     * debug: print the name of the expired .log file            *
     * --------------------------------------------------------- */
    if(DEBUG >0) printf("[erase_ldrlog] Expired log file %d: %s.\n", i, logentrypath);

    /* --------------------------------------------------------- *
     * delete the expired .log file                              *
     * --------------------------------------------------------- */
    if(unlink(logentrypath) != 0) {
      snprintf(errormsg, sizeof(errormsg),
           "[erase_ldrlog] !Error!: Cannot unlink %s.\n", logentrypath);
      ldr_error(errormsg);
    }
    else
      fprintf(logfile, "[erase_ldrlog] %s: erased expired log file\n", logfile_list[i]->d_name);

    log_erasedfiles++;
    i++;
  }

  if(log_erasedfiles > 0) {
    if(DEBUG >0) printf("[erase_ldrlog] Deleted expired log files: %d\n", log_erasedfiles);
  }
  else {
    fprintf(logfile, "[erase_ldrlog] !Error!: No expired log file deleted.\n");
  }

  return log_erasedfiles;
}
