/* ------------------------------------------------------------------------ *
 * file:        erase_sqlerr.c                                              *
 * purpose:     deletes the SQL error files in the edacsdb/error directory  *
 * author:      08/23/2006 Frank4DD                                         *
 * remarks: housekeeping, remove error files older then x days, see edacs.h *
 * ------------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
//#include "getline.h"
#include "../edacs.h"

/* ----------------------------------------------------------- *
 * err_select runs through the ERROR_DIR directory and selects * 
 * the expired .err files which are older then x days.         *
 * ----------------------------------------------------------- */
int err_select(const struct dirent *entry) {
  char *lastdot = NULL;         /* file extension separator    */
  char errfilepath[256] = "";   /* full path & .err file name  */
  struct stat errfilebuf;       /* error file stat structure   */
  int expireseconds = 0;
  time_t exp;

  /* --------------------------------------------------------- *
   * check for "." and ".." directory entries to sort them out *
   * --------------------------------------------------------- */
  if(entry->d_name[0]=='.') return 0;

  /* --------------------------------------------------------- *
   * single out the .err files in ERROR_DIR to process         *
   * --------------------------------------------------------- */
  lastdot = strrchr(entry->d_name, '.');
  if(strcmp(lastdot, ".err") != 0) return 0;

  /* --------------------------------------------------------- *
   * generate the full path and file name of the error files   *
   * --------------------------------------------------------- */
  snprintf(errfilepath, sizeof(errfilepath),
                     "%s/%s", ERROR_DIR, entry->d_name);

  /* --------------------------------------------------------- *
   * try to stat the error file                                *
   * --------------------------------------------------------- */
  if (stat(errfilepath, &errfilebuf) == -1) {
    snprintf(errormsg, sizeof(errormsg),
             "[erase_sqlerr] !Error!: Cannot stat %s.\n", errfilepath);
    ldr_error(errormsg);
    return 0;
  }

  /* ---------------------------------------------------------- *
   * Calculate the expiration time                              *
   * ---------------------------------------------------------- */
  expireseconds = KEEP_ERRS * 86400;
  exp = time(NULL) - expireseconds;
  if(DEBUG >2) printf("Delete SQL error files older then: %s", ctime(&exp));

  /* ---------------------------------------------------------- *
   *  process only .err files older then KEEP_DAYS              *
   * ---------------------------------------------------------- */
  if( (errfilebuf.st_mtime) >= exp) return 0;

  if(DEBUG >1) printf("Delete SQL error file: %s.\n", entry->d_name);

  return 1;
}

/* ------------------------------------------------------------ *
 * erase_sqlerr deletes the sql loader files to clean up.       *
 * ------------------------------------------------------------ */
int erase_sqlerr() {
  int errcount = 0;
  int i        = 0;
  struct dirent **errfile_list;

  /* ------------------------------------------------------------ *
   * Reset the global values                                      *
   * ------------------------------------------------------------ */
  err_erasedfiles = 0;

  /* ------------------------------------------------------------ *
   * Select and order the .sql files found in ERROR_DIR           *
   * ------------------------------------------------------------ */
  errcount = scandir(ERROR_DIR, &errfile_list, err_select, alphasort);

  if(errcount == 0) {
    fprintf(logfile, "[erase_sqlerr] No expired .err file found in %s.\n", ERROR_DIR);
    return errcount;
  }

  /* ----------------------------------------------------------- *
   * Cycle through the ordered .sql files found in GOOD_DIR      *
   * ----------------------------------------------------------- */
  i = 0;
  while(i<errcount) {
   char errentrypath[256] = "";

    /* --------------------------------------------------------- *
     * generate the full path and file name of the .sql file     *
     * --------------------------------------------------------- */
    snprintf(errentrypath, sizeof(errentrypath),
        "%s/%s", ERROR_DIR, errfile_list[i]->d_name);

    /* --------------------------------------------------------- *
     * debug: print the name of the sql error file               *
     * --------------------------------------------------------- */
    if(DEBUG >0) printf("[erase_sqlerr] SQL error file %d: %s.\n", i, errentrypath);

    /* --------------------------------------------------------- *
     * delete the .err file                                      *
     * --------------------------------------------------------- */
    if(unlink(errentrypath) != 0) {
      snprintf(errormsg, sizeof(errormsg),
           "[erase_sqlerr] !Error!: Cannot unlink %s.\n", errentrypath);
      ldr_error(errormsg);
    }
    else
      fprintf(logfile, "[erase_sqlerr] %s: erased SQL error file\n", errfile_list[i]->d_name);

    err_erasedfiles++;
    i++;
  }

  if(err_erasedfiles > 0) {
    fprintf(logfile, "[erase_sqlerr] Deleted SQL error files: %d\n", err_erasedfiles);
  }
  else {
    fprintf(logfile, "[erase_sqlerr] No expired error files deleted.\n");
  }

  return err_erasedfiles;
}
