/* ------------------------------------------------------------------------ *
 * file:        parse_radius.c                                              *
 * purpose:     parse the radius accounting files in edacsdb/raw directory  *
 *              save into statement files .sql in the edacsdb/sql directory *
 * author:      08/18/2006 Frank4DD                                         *
 * ------------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
//#include "getline.h"
#include "../edacs.h"

/* ------------------------------------------------------------------- *
 * parse_radius picks up the .rad radius files in RAW_DIR and parses   *
 * them into SQL values that are written into a .sql file in GOOD_DIR  *
 * ------------------------------------------------------------------- */
int parse_radius() {
  struct dirent **rawfile_list;
  int i                  = 0;      /* helper variable in loop  */
  int radcount           = 0;      /* count of .rad files      */

  /* --------------------------------------------------------- *
   * Reset the global counters                                 *
   * --------------------------------------------------------- */
  radius_rawrecords      = 0;     /* Number of raw records     */
  rad_parsedfiles        = 0;     /* Number of parsed files    */

  /* --------------------------------------------------------- *
   * Select and order the .sql files found in GOOD_DIR         *
   * --------------------------------------------------------- */
  radcount = scandir(RAW_DIR, &rawfile_list, rad_select, alphasort);

  if(radcount == 0) {
    fprintf(logfile, "[parse_radius] !Error!: No .rad file found in %s.\n", RAW_DIR);
    return radcount;
  }

  /* --------------------------------------------------------- *
   * Cycle through the ordered .rad files found in RAW_DIR     *
   * --------------------------------------------------------- */
  while(i<radcount) {

    char sqldatafile[80]   = "";
    char sqldatapath[256]  = "";
    char radentrypath[256] = "";
    FILE *sqlfiledesc      = NULL;
    FILE *radfiledesc      = NULL;
    char radacctkey[256]   = "";
    char radacctvalue[256] = "";
    char *linebuf          = NULL;
    time_t timestamp	   = 0;
    size_t len             = 0;
    size_t splitoffset     = 0;
    ssize_t read           = 0;
    int pair_counter       = 0;
    struct stat radentrybuf;

    /* --------------------------------------------------------- * 
     * generate the full path and file name of the .rad file     * 
     * --------------------------------------------------------- */
    snprintf(radentrypath, sizeof(radentrypath),
        "%s/%s", RAW_DIR, rawfile_list[i]->d_name);

    /* --------------------------------------------------------- * 
     * debug: print the name of the radius source file           * 
     * --------------------------------------------------------- */
    if(DEBUG >0) printf("[parse_radius] RAD data file %d: %s.\n", i, radentrypath);

    /* --------------------------------------------------------- * 
     * try to stat the .rad file                                 *
     * --------------------------------------------------------- */
    if (stat(radentrypath, &radentrybuf) == -1) {
      snprintf(errormsg, sizeof(errormsg), 
             "[parse_radius] !Error!: Cannot stat %s.\n", radentrypath);
      ldr_error(errormsg);
      i++;
      continue;
    }

    /* --------------------------------------------------------- * 
     * if the .rad file is empty, delete it                      *
     * --------------------------------------------------------- */
    if( (radentrybuf.st_size) == 0) {
      fprintf(logfile, "[parse_radius] Warning: Empty file %s.\n", radentrypath);
      if(unlink(radentrypath) != 0) {
        snprintf(errormsg, sizeof(errormsg), 
               "[parse_radius] !Error!: Cannot unlink %s.\n", radentrypath);
        ldr_error(errormsg);
      }
      else
        fprintf(logfile, "[parse_radius] Warning: Erased file %s.\n", radentrypath);

      i++;
      continue;
    }

    /* ------------------------------------------------------------ * 
     * open the radius accounting file, if not able skip processing *
     * ------------------------------------------------------------ */
    radfiledesc = fopen(radentrypath, "r");
    if(radfiledesc == NULL) {
      snprintf(errormsg, sizeof(errormsg), 
               "[parse_radius] !Error!: Cannot open %s.\n", radentrypath);
      ldr_error(errormsg);
      i++;
      continue;
    }

    /* ------------------------------------------------- * 
     * prepare the opening of the sql loader data file   * 
     * ------------------------------------------------- */

    /* ------------------------------------------------- * 
     * generate the file name of the sql data file       *
     * by using the raw file name and changing the       *
     * extension from .dat to .sql                       *
     * ------------------------------------------------- */
    strncpy(sqldatafile, rawfile_list[i]->d_name, sizeof(sqldatafile));
    strncpy(strstr(sqldatafile, ".rad"), ".sql", 4);

    /* ------------------------------------------------- * 
     * generate the full path of the sql data file       *
     * ------------------------------------------------- */
    snprintf(sqldatapath, sizeof(sqldatapath),
        "%s/%s", GOOD_DIR, sqldatafile);

    /* ------------------------------------------------- * 
     * debug: print the name of the sql data file        * 
     * ------------------------------------------------- */
    if(DEBUG >0) printf("[parse_radius] SQL data file %d: %s.\n", i, sqldatapath);

    /* ------------------------------- * 
     * open the sql loader data file   * 
     * ------------------------------- */
    sqlfiledesc = fopen(sqldatapath, "w");
    if(sqlfiledesc == NULL) {
      snprintf(errormsg, sizeof(errormsg), 
               "[parse_radius] !Error!: Cannot open %s.\n", sqldatapath);
      ldr_error(errormsg);
      i++;
      continue;
    }

    /* --------------------------------------------------- * 
     * cycle through each line of the .rad accounting file * 
     * --------------------------------------------------- */
    struct radius_pair new_radius_pair = {"","","","",0,0,"none",
                                         "","n/a",0,0,"","","unknown"};

    while ((read = getline(&linebuf, &len, radfiledesc)) != -1) {

      /* --------------------------------------------------------- * 
       * ignore the beginning lines with the radius acct timestamp * 
       * --------------------------------------------------------- */
      if(strchr(linebuf, '\t') == linebuf) {

        /* --------------------------------------------------- * 
         * remove the newline character at the end  of linebuf * 
         * --------------------------------------------------- */
        if((linebuf[strlen(linebuf) -1] = '\n'))
        linebuf[strlen(linebuf) -1] = '\0';
  
        /* --------------------------------------------------- * 
         * debug: show the line read from the radacct file     *
         * --------------------------------------------------- */
        if(DEBUG >1) printf("[parse_radius] the line is: >%s<\n", linebuf+1);
  
        /* --------------------------------------- *
         * split the line into key and value pairs *
         * --------------------------------------- */
        if(strstr(linebuf, " = ") != NULL) {
  
          /* ---------------------------- * 
           * get the radacct key entries  * 
           * ---------------------------- */

          /* --------------------------------------------------- *
           * calculate the offset of the pair separator in bytes *
           * --------------------------------------------------- */
          splitoffset = (strstr(linebuf, " = ") - linebuf) -1;
  
          /* --------------------------------------------------- *
           * use +1 as startaddress to remove the tabs and       *
           * splitoffset -1 as the number of bytes to copy       *
           * --------------------------------------------------- */
          strncpy(radacctkey, linebuf+1, splitoffset);
  
          /* ------------------------------------------------------------- *
           * strncpy does not add the terminating '\0', we must do it here *
           * ------------------------------------------------------------- */
          radacctkey[splitoffset] = '\0';
  
          /* --------------------------------------------------- * 
           * get the radacct value entries                       *
           * --------------------------------------------------- */
          strncpy(radacctvalue, strstr(linebuf, " = ") + 3,
                                               sizeof(radacctvalue));
  
          /* --------------------------------------------------------- * 
           * debug: write all keys, values and key termination offsets *
           * --------------------------------------------------------- */
          if(DEBUG >2) {
            printf("[parse_radius] key length [%d] key >%s< value >%s<\n",
                    (int) splitoffset, radacctkey, radacctvalue);
          } /* close debug */

        } /* close if line contains the pair separator " = " */
      } /* close if line is not just a timestamp */
      else {

      /* --------------------------------------------------------------- * 
       * if the line is a newline we are finished with the radius record * 
       * --------------------------------------------------------------- */
      if(strchr(linebuf, '\n') == linebuf) {

        if(DEBUG >1) printf( "==> '%s', '%s', '%s', %s, %d, %d, '%s', '%s', '%s', %d, %d, '%s', '%s', '%s'\n",
            new_radius_pair.action,
            new_radius_pair.router,
            new_radius_pair.username,
            new_radius_pair.action_date,
            new_radius_pair.bytes_in,
            new_radius_pair.bytes_out,
            new_radius_pair.ip_addr,
            new_radius_pair.tty,
            new_radius_pair.protocol,
            new_radius_pair.packets_in,
            new_radius_pair.packets_out,
            new_radius_pair.task_id,
            new_radius_pair.service,
            new_radius_pair.ip_or_phone);

         fprintf(sqlfiledesc, "'%s', '%s', '%s', %s, %d, %d, '%s', '%s', '%s', %d, %d, '%s', '%s', '%s'\n",
            new_radius_pair.action,
            new_radius_pair.router,
            new_radius_pair.username,
            new_radius_pair.action_date,
            new_radius_pair.bytes_in,
            new_radius_pair.bytes_out,
            new_radius_pair.ip_addr,
            new_radius_pair.tty,
            new_radius_pair.protocol,
            new_radius_pair.packets_in,
            new_radius_pair.packets_out,
            new_radius_pair.task_id,
            new_radius_pair.service,
            new_radius_pair.ip_or_phone);

          /* ------------------------------------------------------------ * 
           * increase radius record counter of SQL statements written out *
           * ------------------------------------------------------------ */
          pair_counter++;
        } /* close if line is a newline, indicating the end of a record */
      } /* close else line is not a '\t' pair line */

      /* -------------------------------------------------------------- * 
       * if the key is Acct-Status-Type and the value is Interim-Update * 
       * remove the 'Interim-' part from the value string               * 
       * -------------------------------------------------------------- */
      if(strcmp(radacctkey, "Acct-Status-Type") == 0)
        if(strcmp(radacctvalue, "Interim-Update") == 0) {
          strncpy(radacctvalue, strstr(radacctvalue, "Update"),
                                                    sizeof(radacctvalue));
          if(DEBUG >2) printf("[parse_radius] [prefilter] adjusted Interim-Update to: %s.\n", radacctvalue);
        }

      /* ------------------------------ * 
       * remove the "" from some values * 
       * ------------------------------ */
      if(radacctvalue[0] == '"')
        if(radacctvalue[strlen(radacctvalue)-1] == '"') {
          /* ------------------------------ * 
           * remove the last '"'            *
           * ------------------------------ */
          radacctvalue[strlen(radacctvalue)-1] = '\0';

          /* ------------------------------ * 
           * remove the first '"'           *
           * ------------------------------ */
          strncpy(radacctvalue, radacctvalue+1, sizeof(radacctvalue));
          if(DEBUG >2) printf("[parse_radius] [prefilter] Removed \"\" from value: %s.\n", radacctvalue);
        }

      /* ------------------------------------------------------ * 
       * if "Calling-Station-Id" is empty because its an analog * 
       * call, set value to "unknown"                           * 
       * ------------------------------------------------------ */
      if(strcmp(radacctkey, "Calling-Station-Id") == 0 &&
        strcmp(radacctvalue, "") == 0) {
        strncpy(radacctvalue, "unknown", sizeof(radacctvalue));
        if(DEBUG >2) printf("[parse_radius] [prefilter] Changed Calling-Station-Id to: %s.\n", radacctvalue);
      }

      /* --------------------------- *
       * Fill the radius_pair struct *
       * --------------------------- */
      if(strcmp(radacctkey, "Acct-Status-Type") == 0)
        strncpy(new_radius_pair.action, radacctvalue,
                      sizeof(new_radius_pair.action));

      if(strcmp(radacctkey, "NAS-IP-Address") == 0)
        strncpy(new_radius_pair.router, radacctvalue,
                        sizeof(new_radius_pair.router));

      if(strcmp(radacctkey, "User-Name") == 0)
        strncpy(new_radius_pair.username, radacctvalue,
                        sizeof(new_radius_pair.username));

      if(strcmp(radacctkey, "Timestamp") == 0) {
        timestamp = (time_t) atoi(radacctvalue);
        strftime(new_radius_pair.action_date,
                 sizeof(new_radius_pair.action_date),
#ifdef ORACLE_DB
                 "TO_DATE('%Y-%m-%d %H:%M:%S', 'YYYY-MM-DD HH24:MI:SS')",
#endif
#ifdef MYSQL_DB
                 "DATE_FORMAT('%Y-%m-%d %H:%M:%S', '%%Y-%%m-%%d %%H:%%i:%%s')",
#endif
                 gmtime(&timestamp));
        if(DEBUG >2) printf("[parse_radius] [timestamp] Set timestamp to: %d.\n", atoi(radacctvalue));
        if(DEBUG >2) printf("[parse_radius] [datestring] Set datestring to: %s.\n", new_radius_pair.action_date);
      }

      if(strcmp(radacctkey, "Acct-Input-Octets") == 0)
        new_radius_pair.bytes_in = atoi(radacctvalue);

      if(strcmp(radacctkey, "Acct-Output-Octets") == 0)
        new_radius_pair.bytes_out = atoi(radacctvalue);

      if(strcmp(radacctkey, "Framed-IP-Address") == 0)
        strncpy(new_radius_pair.ip_addr, radacctvalue,
                        sizeof(new_radius_pair.ip_addr));

      if(strcmp(radacctkey, "NAS-Port") == 0)
        strncpy(new_radius_pair.tty, radacctvalue,
                        sizeof(new_radius_pair.tty));

      if(strcmp(radacctkey, "Framed-Protocol") == 0)
        strncpy(new_radius_pair.protocol, radacctvalue,
                        sizeof(new_radius_pair.protocol));

      if(strcmp(radacctkey, "Acct-Input-Packets") == 0)
        new_radius_pair.packets_in = atoi(radacctvalue);
  
      if(strcmp(radacctkey, "Acct-Output-Packets") == 0)
        new_radius_pair.packets_out = atoi(radacctvalue);
  
      if(strcmp(radacctkey, "Acct-Session-Id") == 0)
        strncpy(new_radius_pair.task_id, radacctvalue,
                        sizeof(new_radius_pair.task_id));

      if(strcmp(radacctkey, "Service-Type") == 0)
        strncpy(new_radius_pair.service, radacctvalue,
                        sizeof(new_radius_pair.service));

      if(strcmp(radacctkey, "Calling-Station-Id") == 0)
        strncpy(new_radius_pair.ip_or_phone, radacctvalue,
                        sizeof(new_radius_pair.ip_or_phone));

      /* ----------------------------------------------------- *
       * if authorization is local then set username = "local" *
       * ----------------------------------------------------- */
      if((strcmp(radacctkey, "Acct-Authentic") == 0 &&
          strcmp(radacctvalue, "Local") == 0) &&
          strcmp(new_radius_pair.username, "") == 0) {
        strncpy(new_radius_pair.username, "local",
                        sizeof(new_radius_pair.username));
        if(DEBUG >2) printf("[parse_radius] [endfilter] Set username to: %s.\n", new_radius_pair.username);
      }
    } /* close while reading lines from file */

    /* --------------------------------------------------------- *
     * close the radius accounting file and sql loader data file *
     * --------------------------------------------------------- */
    if(fclose(sqlfiledesc) != 0) {
      snprintf(errormsg, sizeof(errormsg), 
               "[parse_radius] !Error!: Cannot close %s.\n", sqldatapath);
      ldr_error(errormsg);
      i++;
      continue;
    }

    if(fclose(radfiledesc) != 0) {
      snprintf(errormsg, sizeof(errormsg), 
               "[parse_radius] !Error!: Cannot close %s.\n", radentrypath);
      ldr_error(errormsg);
      i++;
      continue;
    }
  
    rad_parsedfiles++;
    fprintf(logfile, "[parse_radius] %s: processed radius records: %d.\n",
                                     rawfile_list[i]->d_name, pair_counter);
 
    /* update the total number of processed radius records */
    radius_rawrecords = radius_rawrecords + pair_counter;
    i++;
  }/* close while loop processing the .rad files in RAW_DIR */

  return rad_parsedfiles;
}
