/* -------------------------------------------------------------------------- *
 * file:         service-list.c                                               *
 * purpose:      display the content of the edacs_service table               *
 *		 select service from edacs_service order by service;          *
 * ---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <cgic.h>
#include "../edacs.h"

int cgiMain() {
#ifdef MYSQL_DB
  static MYSQL *dbh;              /* database connect handle */
  static MYSQL_RES *result;       /* database query results  */
  static MYSQL_ROW values;        /* record data values */
  unsigned int colcount    =0;    /* number of returned columns */
  int server_version;             /* returned server version */
#endif
#ifdef ORACLE_DB
  sqlo_db_handle_t dbh;           /* database handle */
  sqlo_stmt_handle_t sth1;        /* statement handle 1 */
  char server_version[1024]="";   /* string for returned server version */
  int stat                 =0;    /* status of sqlo calls */
  int handle               =0;    /* handle of the interrupt handler */
  const char ** values = NULL;    /* record data values */
#endif
  char err_str[2048]       ="";   /* use for combined error string */
  char sqlquery_str[1024]  ="";   /* SQL query string */
  int rowcount             =0;    /* number of returned rows */
  div_t oddline_calc;             /* calculates even/odd row color */

  static char title[] = "List of Services";
  _abort_flag = 0;

#ifdef ORACLE_DB
  /* ------------------------------------------------------------------- * 
   * ORACLE_HOME is needed for OCI8 to find tnsnames.ora                 *
   * ------------------------------------------------------------------- */
  putenv(WEB_ORACLE_ENV);

  /* initialize the connection */
  if (SQLO_SUCCESS != sqlo_init(SQLO_OFF, 1, 100))
    cgi_error("Error: Failed to init libsqlora8.");

  /* register the interrupt handler */
  sqlo_register_int_handler(&handle, sigint_handler);

  /* login to the database */
  if (SQLO_SUCCESS != sqlo_connect(&dbh, WEB_TNS_STRING))
    cgi_error("Error: Cannot connect to database.");
  RETURN_ON_ABORT; /* finish if SIGINT was catched */

  if (SQLO_SUCCESS != sqlo_server_version(dbh, server_version,
                                        sizeof(server_version)))
    cgi_error(sqlo_geterror(dbh));
  RETURN_ON_ABORT; /* finish if SIGINT was catched */

  /* enable autocommit, each statement is commited as a single transaction */
  stat = sqlo_set_autocommit(dbh, 1);

  /* define the SQL query */
  snprintf(sqlquery_str, sizeof(sqlquery_str), "SELECT * FROM %s.EDACS_SERVICE ORDER BY SERVICE", EDACSADMIN);

  /* initialize the statement handle */
  sth1 = SQLO_STH_INIT;

  /* opens a cursor for the query statement */
  if ( 0 > (sqlo_open2(&sth1, dbh, sqlquery_str, 0, NULL)))
    cgi_error(sqlo_geterror(dbh));
  RETURN_ON_ABORT; /* finish if SIGINT was catched */

  /* get the output column names */
  //if (SQLO_SUCCESS != sqlo_ocol_names2(sth1, &colcount, &colnames))
  //  cgi_error("Error getting the DB columns with sqlo_ocol_names2()");
  //RETURN_ON_ABORT; /* finish if SIGINT was catched */
#endif
#ifdef MYSQL_DB
  /* initialize the connection */
  dbh = mysql_init(NULL);
  if(dbh == NULL) cgi_error("Error:  Failed to init MySQL DB.");

  /* login to the database */
  if (mysql_real_connect(dbh, MYSQLIP, EDACSADMIN, ADMIN_PASS, DB_NAME, DB_PORT, NULL, 0) == 0)
    cgi_error("Error: Cannot connect to database.");

  /* Get the database version */
  server_version = mysql_get_server_version(dbh);

  /* create the SQL query string */
  snprintf(sqlquery_str, sizeof(sqlquery_str), "SELECT * FROM edacs_service ORDER BY service");

  /* Prepare and execute the SQL statement */
  if(mysql_query(dbh, sqlquery_str) != 0) {
    if(DEBUG == 0) cgi_error(mysql_error(dbh));
    else snprintf(err_str, sizeof(err_str), "DB error %s\n\nQuery string %s",
             mysql_error(dbh), sqlquery_str);
    cgi_error(err_str);
  }
 /* get query results set */
  result = mysql_store_result(dbh);
  if (result == NULL) {
    snprintf(err_str, sizeof(err_str), "No results for query: %s\n", sqlquery_str);
    cgi_error( err_str);
  }

  colcount = mysql_num_fields(result);
#endif

/* -------------------------------------------------------------------------- *
 * start the html output                                                      *
 * ---------------------------------------------------------------------------*/

   pagehead(title);
   fprintf(cgiOut, "<div id=\"content\">\n");

   fprintf(cgiOut, "<table class=\"inner\" width=\"100%%\">\n");
   fprintf(cgiOut, "<tr>\n");
   fprintf(cgiOut, "<th class=\"inner\" width=40>#</th>\n");
   fprintf(cgiOut, "<th class=\"inner\">\n");
   fprintf(cgiOut, "List of Services");
   fprintf(cgiOut, "</th>\n");
   fprintf(cgiOut, "</tr>\n");

   /* fetch the data */
#ifdef ORACLE_DB
   while ( SQLO_SUCCESS == (stat = (sqlo_fetch(sth1, 1)))) {
     /* get one record */
     values = sqlo_values(sth1, NULL, 1);
#endif
#ifdef MYSQL_DB
   while((values = mysql_fetch_row(result)) != NULL) {
#endif
     rowcount++;

     /* check for even/odd rows */
     oddline_calc = div(rowcount, 2);
     if(oddline_calc.rem) fprintf(cgiOut, "<tr class=\"odd\">\n");
     else fprintf(cgiOut, "<tr class=\"even\">\n");

     /* column 0 = row counter */
     fprintf(cgiOut, "<td>%d</td>\n", rowcount);

     /* column 1 = service name */
     fprintf(cgiOut, "<td>%s</td>\n", values[0]);
     fprintf(cgiOut, "</tr>\n");
   }
#ifdef ORACLE_DB
  if (SQLO_SUCCESS != sqlo_close(sth1))
    cgi_error("Error Closing the SQL statment handle.");
  RETURN_ON_ABORT; /* finish if SIGINT was catched */
#endif
#ifdef MYSQL_DB
   mysql_close(dbh);
#endif
   fprintf(cgiOut, "<tr>\n");
   fprintf(cgiOut, "<th class=\"inner\" colspan=2>");
   fprintf(cgiOut, "&nbsp;");
   fprintf(cgiOut, "</th>\n");
   fprintf(cgiOut, "</tr>\n");
   fprintf(cgiOut, "</table>\n");

   fprintf(cgiOut, "<h3>Additional Information</h3>\n");
   fprintf(cgiOut, "<hr>\n");
   fprintf(cgiOut, "<p>");
   fprintf(cgiOut, "Service Attributes are defined in <a href=\"http://www.ietf.org/rfc/rfc2865.txt\">");
   fprintf(cgiOut, "RFC 2865</a>:");
   fprintf(cgiOut, "</p>");
   fprintf(cgiOut, "<ul>");
   fprintf(cgiOut, "<li>Framed-User is a indicator for a PPP connection used with modem or ISDN dial-in.");
   fprintf(cgiOut, "<li>Login-User indicates a telnet connection to the router by the network staff");
   fprintf(cgiOut, "<li>NAS-Prompt-User indicates the user authentication to the router before a session.");
   fprintf(cgiOut, "</ul>");

   pageside();
   pagefoot();
   return(0);
}
