#include <gidx.h>
#include <gidx-rd.h>
#include <gidx-query.h>
#include <gidx-format.h>
#include <gidx-str-format.h>
#define _THIS_IS_LIBGIDX_RD 1
#include "gidx-rd-core.h"

#if 1
#define DDEBUG DEBUG
#else
#define DDEBUG(x...) do{}while(0);
#endif

struct str_cond_radix {
	const struct radix_dedge	*cr_root;
	unsigned int			cr_num_edges;
	const uint8_t			*cr_strtab;
	const uint8_t			*cr_strtab_end;
	struct ro_vec			cr_str;
};

struct str_cond_ordered {
};

static uint32_t radix_edges_index(const struct radix_dedge *e)
{
	return (e->re_edges_hi << 16) | e->re_edges_idx;
}

static uint32_t radix_strtab_ofs(const struct radix_dedge *e)
{
	return (e->re_strtab_hi << 16) | e->re_strtab_ofs;
}

static int string_is_resident(const struct radix_dedge *e)
{
	return e->re_strtab_len < 4;
}

static int do_radix_execute(struct str_cond_radix *rc,
				const struct radix_dedge *re,
				unsigned int num_edges,
				struct ro_vec *str)
{
	unsigned int i;

	while(num_edges) {
		uint32_t edges_idx, strtab_ofs;
		uint8_t res[3];
		const uint8_t *ptr;
		struct ro_vec match, suff;
		int cmp;

		i = (num_edges / 2);

		edges_idx = radix_edges_index(re + i);
		strtab_ofs = radix_strtab_ofs(re + i);
		if ( string_is_resident(re + i) ) {
			res[0] = strtab_ofs & 0xff;
			res[1] = (strtab_ofs >> 8) & 0xff;
			res[2] = (strtab_ofs >> 16) & 0xff;
			ptr = res;
		}else{
			ptr = rc->cr_strtab + strtab_ofs;
		}

		suff = match = *str;
		match.v_len = re[i].re_strtab_len;
		suff.v_ptr += match.v_len;
		suff.v_len -= match.v_len;

		cmp = memcmp(match.v_ptr, ptr, match.v_len);
		DDEBUG("'%.*s' vs '%.*s' (idx[%lu]: %u @ 0x%.4x) = %d",
			(int)match.v_len,
			match.v_ptr,
			(int)re[i].re_strtab_len,
			ptr,
			&re[i] - rc->cr_root,
			re[i].re_strtab_len,
			strtab_ofs, cmp);

		if ( cmp < 0 ) {
			num_edges = i;
		}else if ( cmp > 0 ) {
			re = re + (i + 1);
			num_edges = num_edges - (i + 1);
		}else{
			if ( !suff.v_len ) {
				DDEBUG("DONE");
				return 1;
			}
			DDEBUG("RECURSE");
			return do_radix_execute(rc, rc->cr_root + edges_idx,
					re->re_num_edges, &suff);
		}
	}

	return 0;
}

static int radix_execute(void *priv)
{
	struct str_cond_radix *rc = priv;
	struct ro_vec match = rc->cr_str;

	DEBUG("matching %.*s", (int)rc->cr_str.v_len, rc->cr_str.v_ptr);

	if ( do_radix_execute(rc, rc->cr_root, 1, &match) ) {
		DEBUG("Found it");
	}

	return 1;
}

static const struct query_cond_ops radix_ops = {
	.execute = radix_execute,
	.dtor = free,
};


static int radix_query(gidx_query_t q, gidx_idx_t idx, struct ro_vec *str)
{
	struct str_cond_radix *rc;
	const uint8_t *radix_ptr, *strtab_ptr;
	size_t radix_sz, strtab_sz;

	radix_ptr = _gidx_query_get_chunk(q, idx,
						RADIX_CHUNK_EDGES,
						&radix_sz);
	strtab_ptr = _gidx_query_get_chunk(q, idx,
						RADIX_CHUNK_STRTAB,
						&strtab_sz);
	if ( NULL == radix_ptr || NULL == strtab_ptr )
		return 0;

	rc = calloc(1, sizeof(*rc));
	if ( NULL == rc )
		return 0;

	if ( radix_sz && strtab_sz ) {
		assert(0 == (radix_sz % sizeof(struct radix_dedge)));
		rc->cr_root = (struct radix_dedge *)radix_ptr;
		rc->cr_num_edges = radix_sz / sizeof(struct radix_dedge);
		rc->cr_strtab = strtab_ptr;
		rc->cr_strtab_end = strtab_ptr + strtab_sz;
	} /* else every field is NULL */

	rc->cr_str = *str;

	if ( !_gidx_query_cond_new(q, idx, &radix_ops, rc) ) {
		free(rc);
		return 0;
	}

	return 1;
}

int gidx_query_str(gidx_query_t q, gidx_idx_t idx, struct ro_vec *str)
{
	switch(idx->i_format) {
	case STR_FORMAT_ORDERED:
		ERR("STR_FORMAT_ORDERED not implemented");
		break;
	case STR_FORMAT_RADIX:
		return radix_query(q, idx, str);
	default:
		ERR("unknown format: 0x%.2x", idx->i_format);
	}
	return 0;
}

int gidx_query_strp(gidx_query_t q, gidx_idx_t idx, struct ro_vec *str)
{
	if ( idx->i_type != GIDX_TYPE_STRING ) {
		ERR("Expected string index, got %s", gidx_type_str(idx->i_type));
		return 0;
	}

	switch(idx->i_format) {
	case STR_FORMAT_ORDERED:
		/* FIXME: could be a LEX ordering option for this, but it's
		 * a bad idea...
		 */
		ERR("prefix query not possible with STR_FORMAT_ORDERED");
		break;
	case STR_FORMAT_RADIX:
		ERR("prefix queries not implemented");
		break;
	default:
		ERR("unknown format: 0x%.2x", idx->i_format);
	}
	return 0;
}
