#include <gidx.h>
#include <gidx-wr.h>
#include <gang.h>
#include <gidx-format.h>
#define _THIS_IS_LIBGIDX_WR 1
#include "gidx-wr-core.h"
#define _THIS_IS_GIDX_STR_WR 1
#include "gidx-str-wr.h"

static struct str_val *get_node(struct str_wr_priv *strp, struct ro_vec *val)
{
	struct rb_node *n, *parent, **p;
	struct str_val *node;
	uint8_t *strbuf;

	DEBUG("searching for node val=%.*s", (int)val->v_len, val->v_ptr);

	for(p = &strp->p_vals.rbt_root, n = parent = NULL; *p; ) {
		struct str_val *v;
		int ret;

		parent = *p;

		v = (struct str_val *)parent;
		ret = vcmp(val, &v->n_str);
		if ( ret < 0 ) {
			p = &(*p)->rb_child[CHILD_LEFT];
		}else if ( ret > 0 ) {
			p = &(*p)->rb_child[CHILD_RIGHT];
		}else{
			DEBUG("found node for val=%.*s, %u records",
				(int)v->n_str.v_len,
				v->n_str.v_ptr,
				v->n_num_oid);
			return v;
		}
	}

	node = hgang_alloc(strp->p_nodes);
	if ( NULL == node )
		return NULL;

	node->n_oids = hgang_new(sizeof(gidx_oid_t), 128);
	if ( node->n_oids == NULL ) {
		ERR("hgang_new: %s", os_err());
		hgang_return(strp->p_nodes, node);
		return NULL;
	}

	strbuf = gang_alloc(strp->p_strtab, val->v_len);
	if ( NULL == strbuf ) {
		ERR("gang_alloc: %s", os_err());
		hgang_free(node->n_oids);
		hgang_return(strp->p_nodes, node);
		return NULL;
	}

	memcpy(strbuf, val->v_ptr, val->v_len);

	node->n_str.v_ptr = strbuf;
	node->n_str.v_len = val->v_len;
	node->n_num_oid = 0;
	memset(&node->n_rbt, 0, sizeof(node->n_rbt));
	node->n_rbt.rb_parent = parent;
	*p = &node->n_rbt;
	rbtree_insert_rebalance(&strp->p_vals, &node->n_rbt);

	strp->p_num_vals++;

	DEBUG("inserted node for val=%.*s",
		(int)node->n_str.v_len, node->n_str.v_ptr);
	return node;
}

static int node_oid_id_insert(struct str_val *node, gidx_oid_t oid_id)
{
	gidx_oid_t *poid;

	poid = hgang_alloc(node->n_oids);
	if ( NULL == poid ) {
		ERR("hgang_alloc: %s", os_err());
		return 0;
	}

	*poid = oid_id;
	node->n_num_oid++;

	DEBUG("append oid_id=%u to node val=%.*s", oid_id,
		(int)node->n_str.v_len, node->n_str.v_ptr);
	return 1;
}

static int str_wr_insert(void *priv, gidx_oid_t oid_id,
			void *rec, void *field,
			union schema_get_val get_val)
{
	struct str_wr_priv *strp = priv;
	struct str_val *node;
	struct ro_vec vec;

	if ( !(*get_val.su_get_str)(rec, field, &vec) )
		return 0;

	node = get_node(strp, &vec);
	if ( node == NULL )
		return 0;

	if ( !node_oid_id_insert(node, oid_id) )
		return 0;

	DEBUG("priv=%p rec=%p val=%.*s", priv, rec, (int)vec.v_len, vec.v_ptr);
	return 1;
}

static void *str_wr_init(gidx_mode_t mode, unsigned int options)
{
	struct str_wr_priv *strp;

	strp = calloc(1, sizeof(*strp));
	if ( NULL == strp )
		return NULL;

	switch(mode) {
	case GIDX_PRIMARY_KEY:
	case GIDX_SECONDARY_KEY:
		strp->p_format = STR_FORMAT_RADIX;
		break;
	case GIDX_NONE_KEY:
		strp->p_format = STR_FORMAT_ORDERED;
		break;
	default:
		abort();
	}

	strp->p_nodes = hgang_new(sizeof(struct str_val), 256);
	if ( NULL == strp->p_nodes ) {
		ERR("hgang_new: %s", os_err());
		free(strp);
		return NULL;
	}

	strp->p_strtab = gang_new(512, 0);
	if ( NULL == strp->p_strtab ) {
		ERR("gang_new: %s", os_err());
		hgang_free(strp->p_nodes);
		free(strp);
		return NULL;
	}

	strp->p_mode = mode;
	DEBUG("priv=%p", strp);
	return strp;
}

static void do_free(struct rb_node *n)
{
	struct str_val *node;

	if ( NULL == n )
		return;

	if ( n->rb_child[CHILD_LEFT] )
		do_free(n->rb_child[CHILD_LEFT]);

	node = (struct str_val *)n;
	DEBUG("node %.*s, %u oids",
		(int)node->n_str.v_len, node->n_str.v_ptr, node->n_num_oid);
	hgang_free(node->n_oids);

	if ( n->rb_child[CHILD_RIGHT] )
		do_free(n->rb_child[CHILD_RIGHT]);
}

static gidx_oid_t str_wr_num_vals(void *priv)
{
	struct str_wr_priv *strp = priv;
	return strp->p_num_vals;
}

static void str_wr_fini(void *priv)
{
	struct str_wr_priv *strp = priv;

	DEBUG("priv=%p", priv);
	do_free(strp->p_vals.rbt_root);
	hgang_free(strp->p_nodes);
	gang_free(strp->p_strtab);
	if ( strp->p_prepped ) {
		switch(strp->p_format) {
		case STR_FORMAT_RADIX:
			_str_radix_free(strp);
			break;
		case STR_FORMAT_ORDERED:
			_str_ordered_free(strp);
			break;
		default:
			assert(0);
		}
	}
	free(priv);
}

static int do_write_lookup(struct rb_node *n, skey_ctx_t pk)
{
	struct str_val *node;

	if ( NULL == n )
		return 1;

	if ( n->rb_child[CHILD_LEFT] )
		if ( !do_write_lookup(n->rb_child[CHILD_LEFT], pk) )
			return 0;

	node = (struct str_val *)n;
	if ( !skey_tbl_entry(pk, node->n_num_oid) )
		return 0;

	if ( n->rb_child[CHILD_RIGHT] )
		if ( !do_write_lookup(n->rb_child[CHILD_RIGHT], pk) )
			return 0;
	
	return 1;
}

static int str_wr_skey_tbl(void *priv, skey_ctx_t pk)
{
	struct str_wr_priv *strp = priv;
	return do_write_lookup(strp->p_vals.rbt_root, pk);
}

static int do_write_skey(struct rb_node *n, skey_ctx_t pk)
{
	struct str_val *node;

	if ( NULL == n )
		return 1;

	if ( n->rb_child[CHILD_LEFT] )
		if ( !do_write_skey(n->rb_child[CHILD_LEFT], pk) )
			return 0;

	node = (struct str_val *)n;
	skey_keyblock(pk, node->n_oids);

	if ( n->rb_child[CHILD_RIGHT] )
		if ( !do_write_skey(n->rb_child[CHILD_RIGHT], pk) )
			return 0;

	return 1;
}

static int str_wr_skey_blks(void *priv, skey_ctx_t pk)
{
	struct str_wr_priv *strp = priv;
	return do_write_skey(strp->p_vals.rbt_root, pk);
}

static void do_write_oid(struct rb_node *n, gidx_oid_t *idx,
						gidx_oid_t *blk,
						gidx_oid_t num_rec)
{
	struct str_val *node;

	if ( NULL == n )
		return;

	if ( n->rb_child[CHILD_LEFT] )
		do_write_oid(n->rb_child[CHILD_LEFT], idx, blk, num_rec);

	node = (struct str_val *)n;
	DEBUG("oid: val[%u] = %.*s", *idx,
		(int)node->n_str.v_len, node->n_str.v_ptr);
	oid_from_hgang(*idx, node->n_oids, blk, num_rec);
	(*idx)++;

	if ( n->rb_child[CHILD_RIGHT] )
		do_write_oid(n->rb_child[CHILD_RIGHT], idx, blk, num_rec);
}

static int str_wr_oid_blk(void *priv, gidx_oid_t *blk, gidx_oid_t num_rec)
{
	struct str_wr_priv *strp= priv;
	gidx_oid_t idx = 0;
	assert(strp->p_format == STR_FORMAT_ORDERED);
	do_write_oid(strp->p_vals.rbt_root, &idx, blk, num_rec);
	return 1;
}

static uint16_t str_wr_num_chunks(void *priv)
{
	struct str_wr_priv *strp = priv;
	switch(strp->p_format) {
	case STR_FORMAT_RADIX:
		return _str_radix_num_chunks(strp);
	case STR_FORMAT_ORDERED:
		return _str_ordered_num_chunks(strp);
	default:
		abort();
	}
	return 0;
}

static off_t str_wr_chunk_size(void *priv, unsigned int chunk_id)
{
	struct str_wr_priv *strp = priv;
	switch(strp->p_format) {
	case STR_FORMAT_RADIX:
		return _str_radix_chunk_size(strp, chunk_id);
	case STR_FORMAT_ORDERED:
		return _str_ordered_chunk_size(strp, chunk_id);
	default:
		abort();
	}
	return 0;
}

static int str_wr_chunk(void *priv, unsigned int chunk_id, fobuf_t buf)
{
	struct str_wr_priv *strp = priv;
	switch(strp->p_format) {
	case STR_FORMAT_RADIX:
		return _str_radix_chunk(strp, chunk_id, buf);
	case STR_FORMAT_ORDERED:
		return _str_ordered_chunk(strp, chunk_id, buf);
	default:
		abort();
		return 0;
	}
	return 1;
}

static int str_wr_prep(void *priv, uint8_t *format)
{
	struct str_wr_priv *strp = priv;

	switch(strp->p_format) {
	case STR_FORMAT_RADIX:
		if ( !_str_radix_prep(strp) )
			return 0;
		break;
	case STR_FORMAT_ORDERED:
		if ( !_str_ordered_prep(strp) )
			return 0;
		break;
	default:
		abort();
	}

	(*format) = strp->p_format;
	return 1;
}

struct _gidx_type_wr _gidx_str_type = {
	.t_name = "varchar",
	.t_wr_init = str_wr_init,
	.t_wr_fini = str_wr_fini,
	.t_wr_insert = str_wr_insert,
	.t_wr_prep = str_wr_prep,
	.t_wr_num_chunks = str_wr_num_chunks,
	.t_wr_chunk_size = str_wr_chunk_size,
	.t_wr_chunk = str_wr_chunk,
	.t_wr_num_vals = str_wr_num_vals,
	.t_wr_skey_tbl = str_wr_skey_tbl,
	.t_wr_skey_blks = str_wr_skey_blks,
	.t_wr_oid_blk = str_wr_oid_blk,
};
