#include <gidx.h>
#include <gidx-wr.h>
#include <gidx-format.h>
#include <gidx-uint-format.h>
#define _THIS_IS_LIBGIDX_WR 1
#include "gidx-wr-core.h"

#include <rbtree.h>

struct u32_val {
	struct rb_node 		n_rbt;
	uint32_t 		n_val;
	gidx_oid_t		n_num_oid;
	hgang_t			n_oids;
};

struct u32_wr_priv {
	struct rbtree		p_vals;
	gidx_oid_t		p_num_vals;
	hgang_t			p_nodes;
	gidx_mode_t		p_mode;
};

static struct u32_val *get_node(struct u32_wr_priv *u32p, uint32_t val)
{
	struct rb_node *n, *parent, **p;
	struct u32_val *node;

	DEBUG("searching for node val=%u", val);

	for(p = &u32p->p_vals.rbt_root, n = parent = NULL; *p; ) {
		struct u32_val *v;

		parent = *p;

		v = (struct u32_val *)parent;
		if ( val < v->n_val ) {
			p = &(*p)->rb_child[CHILD_LEFT];
		}else if ( val > v->n_val ) {
			p = &(*p)->rb_child[CHILD_RIGHT];
		}else{
			DEBUG("found node for val=%u, %u records",
				v->n_val, v->n_num_oid);
			return v;
		}
	}

	node = hgang_alloc(u32p->p_nodes);
	if ( NULL == node )
		return NULL;

	node->n_val = val;
	node->n_num_oid = 0;
	node->n_oids = hgang_new(sizeof(gidx_oid_t), 128);
	if ( node->n_oids == NULL ) {
		hgang_return(u32p->p_nodes, node);
		return NULL;
	}

	u32p->p_num_vals++;

	memset(&node->n_rbt, 0, sizeof(node->n_rbt));
	node->n_rbt.rb_parent = parent;
	*p = &node->n_rbt;
	rbtree_insert_rebalance(&u32p->p_vals, &node->n_rbt);

	DEBUG("inserted node for val=%u", node->n_val);
	return node;
}

static int node_oid_id_insert(struct u32_val *node, gidx_oid_t oid_id)
{
	gidx_oid_t *poid;

	poid = hgang_alloc(node->n_oids);
	if ( NULL == poid ) {
		ERR("hgang_alloc: %s", os_err());
		return 0;
	}

	*poid = oid_id;
	node->n_num_oid++;

	DEBUG("append oid_id=%u to node val=%u", oid_id, node->n_val);
	return 1;
}

static int u32_wr_insert(void *priv, gidx_oid_t oid_id,
			void *rec, void *field,
			union schema_get_val get_val)
{
	struct u32_wr_priv *u32p = priv;
	struct u32_val *node;
	uint32_t val;

	if ( !(*get_val.su_get_uint)(rec, field, &val) )
		return 0;

	node = get_node(u32p, val);
	if ( node == NULL )
		return 0;

	if ( !node_oid_id_insert(node, oid_id) )
		return 0;

	DEBUG("priv=%p rec=%p val=%u", priv, rec, val);
	return 1;
}

static void *u32_wr_init(gidx_mode_t mode, unsigned int options)
{
	struct u32_wr_priv *u32p;

	assert(mode == GIDX_SECONDARY_KEY ||
		mode == GIDX_PRIMARY_KEY ||
		mode == GIDX_NONE_KEY);
	u32p = calloc(1, sizeof(*u32p));
	if ( NULL == u32p )
		return NULL;

	u32p->p_nodes = hgang_new(sizeof(struct u32_val), 256);
	if ( NULL == u32p->p_nodes ) {
		free(u32p);
		return NULL;
	}

	u32p->p_mode = mode;
	DEBUG("priv=%p", u32p);
	return u32p;
}

static void do_free(struct rb_node *n)
{
	struct u32_val *node;

	if ( NULL == n )
		return;

	if ( n->rb_child[CHILD_LEFT] )
		do_free(n->rb_child[CHILD_LEFT]);

	node = (struct u32_val *)n;
	DEBUG("node %u, %u oids", node->n_val, node->n_num_oid);
	hgang_free(node->n_oids);

	if ( n->rb_child[CHILD_RIGHT] )
		do_free(n->rb_child[CHILD_RIGHT]);
}

static void u32_wr_fini(void *priv)
{
	struct u32_wr_priv *u32p = priv;

	DEBUG("priv=%p", priv);
	do_free(u32p->p_vals.rbt_root);
	hgang_free(u32p->p_nodes);
	free(priv);
}

static uint16_t u32_wr_num_chunks(void *priv)
{
	return UINT_NUM_CHUNKS;
}

static off_t u32_wr_chunk_size(void *priv, unsigned int chunk_id)
{
	struct u32_wr_priv *u32p = priv;

	switch(chunk_id) {
	case UINT_CHUNK_VALS:
		return u32p->p_num_vals * sizeof(uint32_t);
	default:
		ERR("unknown chunk id %u", chunk_id);
		return 0;
	}
}

static int do_write_val(struct rb_node *n, fobuf_t buf)
{
	struct u32_val *node;

	if ( NULL == n )
		return 1;

	if ( n->rb_child[CHILD_LEFT] )
		if ( !do_write_val(n->rb_child[CHILD_LEFT], buf) )
			return 0;

	node = (struct u32_val *)n;
	if ( !fobuf_write(buf, &node->n_val, sizeof(node->n_val)) )
		return 0;

	if ( n->rb_child[CHILD_RIGHT] )
		if ( !do_write_val(n->rb_child[CHILD_RIGHT], buf) )
			return 0;
	
	return 1;
}

static int write_vals(struct u32_wr_priv *u32p, fobuf_t buf)
{
	return do_write_val(u32p->p_vals.rbt_root, buf);
}

static int do_write_lookup(struct rb_node *n, skey_ctx_t pk)
{
	struct u32_val *node;

	if ( NULL == n )
		return 1;

	if ( n->rb_child[CHILD_LEFT] )
		if ( !do_write_lookup(n->rb_child[CHILD_LEFT], pk) )
			return 0;

	node = (struct u32_val *)n;
	if ( !skey_tbl_entry(pk, node->n_num_oid) )
		return 0;

	if ( n->rb_child[CHILD_RIGHT] )
		if ( !do_write_lookup(n->rb_child[CHILD_RIGHT], pk) )
			return 0;
	
	return 1;
}

static int u32_wr_skey_tbl(void *priv, skey_ctx_t pk)
{
	struct u32_wr_priv *u32p = priv;
	return do_write_lookup(u32p->p_vals.rbt_root, pk);
}

static int do_write_skey(struct rb_node *n, skey_ctx_t pk)
{
	struct u32_val *node;

	if ( NULL == n )
		return 1;

	if ( n->rb_child[CHILD_LEFT] )
		if ( !do_write_skey(n->rb_child[CHILD_LEFT], pk) )
			return 0;

	node = (struct u32_val *)n;
	skey_keyblock(pk, node->n_oids);

	if ( n->rb_child[CHILD_RIGHT] )
		if ( !do_write_skey(n->rb_child[CHILD_RIGHT], pk) )
			return 0;

	return 1;
}

static int u32_wr_skey_blks(void *priv, skey_ctx_t pk)
{
	struct u32_wr_priv *u32p = priv;
	return do_write_skey(u32p->p_vals.rbt_root, pk);
}

static void do_write_oid(struct rb_node *n, gidx_oid_t *idx,
						gidx_oid_t *blk,
						gidx_oid_t num_rec)
{
	struct u32_val *node;

	if ( NULL == n )
		return;

	if ( n->rb_child[CHILD_LEFT] )
		do_write_oid(n->rb_child[CHILD_LEFT], idx, blk, num_rec);

	node = (struct u32_val *)n;
	DEBUG("oid: val[%u] = %u", *idx, node->n_val);
	oid_from_hgang(*idx, node->n_oids, blk, num_rec);
	(*idx)++;

	if ( n->rb_child[CHILD_RIGHT] )
		do_write_oid(n->rb_child[CHILD_RIGHT], idx, blk, num_rec);
}

static int u32_wr_oid_blk(void *priv, gidx_oid_t *blk, gidx_oid_t num_rec)
{
	struct u32_wr_priv *u32p = priv;
	gidx_oid_t idx = 0;
	do_write_oid(u32p->p_vals.rbt_root, &idx, blk, num_rec);
	return 1;
}

static gidx_oid_t u32_wr_num_vals(void *priv)
{
	struct u32_wr_priv *u32p = priv;
	return u32p->p_num_vals;
}

static int u32_wr_chunk(void *priv, unsigned int chunk_id, fobuf_t buf)
{
	struct u32_wr_priv *u32p = priv;

	switch(chunk_id) {
	case UINT_CHUNK_VALS:
		return write_vals(u32p, buf);
	default:
		ERR("unknown chunk id %u", chunk_id);
		return 0;
	}
}

struct _gidx_type_wr _gidx_u32_type = {
	.t_name = "uint32_t",
	.t_wr_init = u32_wr_init,
	.t_wr_fini = u32_wr_fini,
	.t_wr_insert = u32_wr_insert,
	.t_wr_num_chunks = u32_wr_num_chunks,
	.t_wr_chunk_size = u32_wr_chunk_size,
	.t_wr_chunk = u32_wr_chunk,
	.t_wr_num_vals = u32_wr_num_vals,
	.t_wr_skey_tbl = u32_wr_skey_tbl,
	.t_wr_oid_blk = u32_wr_oid_blk,
	.t_wr_skey_blks = u32_wr_skey_blks,
};

