/*
 ============================================================================
 Name        : hev-impathy-auth-factory.c
 Author      : Heiher <admin@heiher.info>
 Version     : 0.0.1
 Copyright   : Copyright (C) 2011 everyone.
 Description : 
 ============================================================================
 */

#include <telepathy-glib/dbus.h>
#include <telepathy-glib/util.h>
#include <telepathy-glib/interfaces.h>
#include <telepathy-glib/handle-channels-context.h>
#include <telepathy-glib/channel-dispatch-operation.h>

#include "hev-impathy-auth-factory.h"
#include "hev-impathy-utils.h"
#include "hev-impathy-server-tls-handler.h"
#include "hev-impathy-server-sasl-handler.h"

enum
{
	NEW_SERVER_TLS_HANDLER,
	NEW_SERVER_SASL_HANDLER,
	LAST_SIGNAL
};

static guint hev_impathy_auth_factory_signals[LAST_SIGNAL] = { 0 };

static HevImpathyAuthFactory *auth_factory_singleton = NULL;

typedef struct _HandlerContextData HandlerContextData;

struct _HandlerContextData
{
	TpHandleChannelsContext *context;
	HevImpathyAuthFactory *self;
};

typedef struct _ObserveChannelsData ObserveChannelsData;

struct _ObserveChannelsData
{
	HevImpathyAuthFactory *self;
	TpObserveChannelsContext *context;
	TpChannelDispatchOperation *dispatch_operation;
	TpAccount *account;
	TpChannel *channel;
};

#define HEV_IMPATHY_AUTH_FACTORY_GET_PRIVATE(obj)	(G_TYPE_INSTANCE_GET_PRIVATE((obj), HEV_TYPE_IMPATHY_AUTH_FACTORY, HevImpathyAuthFactoryPrivate))

typedef struct _HevImpathyAuthFactoryPrivate HevImpathyAuthFactoryPrivate;

struct _HevImpathyAuthFactoryPrivate
{
	/* Keep a ref here so the auth client doesn't have to mess with
	 * refs. It will be cleared when the channel (and so the handler)
	 * gets invalidated.
	 *
	 * The channel path of the handler's channel (borrowed gchar *) ->
	 * reffed (EmpathyServerSASLHandler *)
	 */
	GHashTable *sasl_handlers;
};

G_DEFINE_TYPE(HevImpathyAuthFactory, hev_impathy_auth_factory, TP_TYPE_BASE_CLIENT);

static HandlerContextData * handler_context_data_new(HevImpathyAuthFactory *self,
			TpHandleChannelsContext *context)
{
	HandlerContextData *data = NULL;

	data = g_slice_new0(HandlerContextData);
	data->self = g_object_ref(self);

	if(NULL != context)
	  data->context = g_object_ref(context);

	return data;
}

static void handler_context_data_free(HandlerContextData *data)
{
	tp_clear_object(&data->self);
	tp_clear_object(&data->context);

	g_slice_free(HandlerContextData, data);
}

static void observe_channels_data_free(ObserveChannelsData *data)
{
	g_object_unref(data->context);
	g_object_unref(data->account);
	g_object_unref(data->channel);
	g_object_unref(data->dispatch_operation);
	g_slice_free(ObserveChannelsData, data);
}

static void hev_impathy_auth_factory_dispose(GObject * obj)
{
	HevImpathyAuthFactory * self = HEV_IMPATHY_AUTH_FACTORY(obj);
	HevImpathyAuthFactoryPrivate * priv = HEV_IMPATHY_AUTH_FACTORY_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(priv->sasl_handlers)
	{
		g_hash_table_unref(priv->sasl_handlers);
		priv->sasl_handlers = NULL;
	}

	G_OBJECT_CLASS(hev_impathy_auth_factory_parent_class)->dispose(obj);
}

static void hev_impathy_auth_factory_finalize(GObject * obj)
{
	HevImpathyAuthFactory * self = HEV_IMPATHY_AUTH_FACTORY(obj);
	HevImpathyAuthFactoryPrivate * priv = HEV_IMPATHY_AUTH_FACTORY_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	G_OBJECT_CLASS(hev_impathy_auth_factory_parent_class)->finalize(obj);
}

static GObject * hev_impathy_auth_factory_constructor(GType type, guint n, GObjectConstructParam * param)
{
	GObject *obj = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(NULL != auth_factory_singleton)
	{
		obj = g_object_ref(auth_factory_singleton);
	}
	else
	{
		obj = G_OBJECT_CLASS(hev_impathy_auth_factory_parent_class)->constructor
			(type, n, param);

		auth_factory_singleton = HEV_IMPATHY_AUTH_FACTORY(obj);

		g_object_add_weak_pointer(obj, (gpointer*)&auth_factory_singleton);
	}

	return obj;
}

static void hev_impathy_auth_factory_constructed(GObject * obj)
{
	HevImpathyAuthFactory * self = HEV_IMPATHY_AUTH_FACTORY(obj);
	HevImpathyAuthFactoryPrivate * priv = HEV_IMPATHY_AUTH_FACTORY_GET_PRIVATE(self);
	TpBaseClient *client = TP_BASE_CLIENT(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	G_OBJECT_CLASS(hev_impathy_auth_factory_parent_class)->constructed(obj);

	tp_base_client_set_handler_bypass_approval(client, FALSE);

	/* Handle ServerTLSConnection and ServerAuthentication channels */
	tp_base_client_take_handler_filter(client, tp_asv_new(
					/* ChannelType */
					TP_PROP_CHANNEL_CHANNEL_TYPE, G_TYPE_STRING,
					TP_IFACE_CHANNEL_TYPE_SERVER_TLS_CONNECTION,
					/* AuthenticationMethod */
					TP_PROP_CHANNEL_TARGET_HANDLE_TYPE, G_TYPE_UINT,
					TP_HANDLE_TYPE_NONE, NULL));

	tp_base_client_take_handler_filter(client, tp_asv_new(
					/* ChannelType */
					TP_PROP_CHANNEL_CHANNEL_TYPE, G_TYPE_STRING,
					TP_IFACE_CHANNEL_TYPE_SERVER_AUTHENTICATION,
					/* AuthenticationMethod */
					TP_PROP_CHANNEL_TYPE_SERVER_AUTHENTICATION_AUTHENTICATION_METHOD,
					G_TYPE_STRING, TP_IFACE_CHANNEL_INTERFACE_SASL_AUTHENTICATION,
					NULL));

	/* We are also an observer so that we can see new auth channels
	 * popping up and if we have the password already saved to one
	 * account where an auth channel has just appeared we can cal
	 * Claim() on the CDO so the approver won't get it, which makes
	 * sense. */

	/* Observe ServerAuthentication channels */
	tp_base_client_take_observer_filter(client, tp_asv_new(
					/* ChannelType */
					TP_PROP_CHANNEL_CHANNEL_TYPE, G_TYPE_STRING,
					TP_IFACE_CHANNEL_TYPE_SERVER_AUTHENTICATION,
					/* AuthenticationMethod */
					TP_PROP_CHANNEL_TYPE_SERVER_AUTHENTICATION_AUTHENTICATION_METHOD,
					G_TYPE_STRING, TP_IFACE_CHANNEL_INTERFACE_SASL_AUTHENTICATION,
					NULL));

	tp_base_client_set_observer_delay_approvers(client, TRUE);
}

static gboolean common_checks(HevImpathyAuthFactory *self,
			GList *channels, gboolean observe,
			GError **error)
{
	HevImpathyAuthFactoryPrivate *priv =
		HEV_IMPATHY_AUTH_FACTORY_GET_PRIVATE(self);
	TpChannel *channel = NULL;
	GHashTable *props = NULL;
	const gchar * const *available_mechanisms = NULL;
	const GError *dbus_error = NULL;
	GObject *handler = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	/* there can't be more than one ServerTLSConnection or
	 * ServerAuthentication channels at the same time, for the same
	 * connection/account.
	 */
	if(1 != g_list_length(channels))
	{
		g_set_error(error, TP_ERROR, TP_ERROR_INVALID_ARGUMENT,
					"Can't %s more than one ServerTLSConnection or ServerAuthentication "
					"channel for the same connection.", observe ? "observe" : "handle");

		return FALSE;
	}

	channel = channels->data;

	if(TP_IFACE_QUARK_CHANNEL_TYPE_SERVER_AUTHENTICATION!=
				tp_channel_get_channel_type_id(channel))
	{
		/* If we are observing we care only about ServerAuthentication channels.
		 * If we are handling we care about ServerAuthentication and
		 * ServerTLSConnection channels. */

		if(observe || (TP_IFACE_QUARK_CHANNEL_TYPE_SERVER_TLS_CONNECTION!=
						tp_channel_get_channel_type_id(channel)))
		{
			g_set_error(error, TP_ERROR, TP_ERROR_INVALID_ARGUMENT,
						"Can only %s ServerTLSConnection or ServerAuthentication channels, "
						"this was a %s channel", observe ? "observe" : "handle",
						tp_channel_get_channel_type(channel));

			return FALSE;
		}
	}

	handler = g_hash_table_lookup(priv->sasl_handlers,
				tp_proxy_get_object_path(channel));

	if(TP_IFACE_QUARK_CHANNEL_TYPE_SERVER_AUTHENTICATION==
				(tp_channel_get_channel_type_id(channel)) &&
				(NULL!=handler) && !observe)
	{
		g_set_error(error, TP_ERROR, TP_ERROR_INVALID_ARGUMENT,
					"We are already handling this channel: %s",
					tp_proxy_get_object_path (channel));

		return FALSE;
	}

	props = tp_channel_borrow_immutable_properties(channel);
	available_mechanisms = tp_asv_get_boxed (props,
				TP_PROP_CHANNEL_INTERFACE_SASL_AUTHENTICATION_AVAILABLE_MECHANISMS,
				G_TYPE_STRV);

	if((TP_IFACE_QUARK_CHANNEL_TYPE_SERVER_AUTHENTICATION==
				tp_channel_get_channel_type_id (channel)) &&
				(!tp_strv_contains(available_mechanisms, "X-TELEPATHY-PASSWORD")))
    {
		g_set_error_literal(error, TP_ERROR, TP_ERROR_INVALID_ARGUMENT,
					"Only the X-TELEPATHY-PASSWORD SASL mechanism is supported");

		return FALSE;
    }

	dbus_error = tp_proxy_get_invalidated (channel);

	if(NULL != dbus_error)
    {
		*error = g_error_copy (dbus_error);
		return FALSE;
    }

	return TRUE;
}

static void server_tls_handler_ready_handler(GObject *source,
			GAsyncResult *res, gpointer user_data)
{
	GObject *handler = NULL;
	GError *error = NULL;
	HandlerContextData *data = user_data;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	handler = hev_impathy_server_tls_handler_new_finish(res,
				&error);

	if(NULL != error)
	{
		tp_handle_channels_context_fail(data->context, error);
		g_error_free(error);
	}
	else
	{
		tp_handle_channels_context_accept(data->context);
		g_signal_emit(data->self,
					hev_impathy_auth_factory_signals[NEW_SERVER_TLS_HANDLER],
					0, handler);
		g_object_unref(handler);
	}

	handler_context_data_free(data);
}

static void sasl_handler_invalidated_handler(HevImpathyServerSASLHandler *handler,
			gpointer user_data)
{
	HevImpathyAuthFactory *self = HEV_IMPATHY_AUTH_FACTORY(user_data);
	HevImpathyAuthFactoryPrivate *priv =
		HEV_IMPATHY_AUTH_FACTORY_GET_PRIVATE(self);
	TpChannel *channel = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	channel = hev_impathy_server_sasl_handler_get_channel(handler);
	g_return_if_fail(TP_IS_CHANNEL(channel));

	g_hash_table_remove(priv->sasl_handlers,
				tp_proxy_get_object_path(channel));
}

static void server_sasl_handler_ready_handler(GObject *source,
			GAsyncResult *res, gpointer user_data)
{
	HevImpathyAuthFactoryPrivate *priv = NULL;
	GError *error = NULL;
	HandlerContextData *data = user_data;
	GObject *handler = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	priv = HEV_IMPATHY_AUTH_FACTORY_GET_PRIVATE(data->self);
	handler = hev_impathy_server_sasl_handler_new_finish(res,
				&error);

	if(NULL != error)
	{
		if(NULL != data->context)
		  tp_handle_channels_context_fail(data->context,
					  error);
		g_error_free(error);
	}
	else
	{
		TpChannel *channel = NULL;

		if(NULL != data->context)
		  tp_handle_channels_context_accept(data->context);

		channel = hev_impathy_server_sasl_handler_get_channel(
					HEV_IMPATHY_SERVER_SASL_HANDLER(handler));
		g_assert(NULL!=channel);

		/* Pass the ref to the hash table */
		g_hash_table_insert(priv->sasl_handlers,
					(gpointer)tp_proxy_get_object_path(channel),
					handler);

		tp_g_signal_connect_object(handler, "invalidated",
					G_CALLBACK(sasl_handler_invalidated_handler),
					data->self, 0);

		g_signal_emit(data->self,
					hev_impathy_auth_factory_signals[NEW_SERVER_SASL_HANDLER],
					0, handler);
	}

	handler_context_data_free(data);
}

static void hev_impathy_auth_factory_handle_channels(TpBaseClient *client,
			TpAccount *account, TpConnection *connection,
			GList *channels, GList *requests_satisfied,
			gint64 user_action_time, TpHandleChannelsContext *context)
{
	TpChannel *channel = NULL;
	GError *error = NULL;
	HevImpathyAuthFactory *self = HEV_IMPATHY_AUTH_FACTORY(client);
	HandlerContextData *data = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(!common_checks(self, channels, FALSE, &error))
	{
		tp_handle_channels_context_fail(context, error);
		g_clear_error(&error);
		return;
	}

	/* The common checks above have checked this is fine. */
	channel = channels->data;

	data = handler_context_data_new(self, context);
	tp_handle_channels_context_delay(context);

	/* create a handler */
	if(TP_IFACE_QUARK_CHANNEL_TYPE_SERVER_TLS_CONNECTION==
				tp_channel_get_channel_type_id(channel))
	{
		hev_impathy_server_tls_handler_new_async(channel,
					server_tls_handler_ready_handler, data);
	}
	else if(TP_IFACE_QUARK_CHANNEL_TYPE_SERVER_AUTHENTICATION==
				tp_channel_get_channel_type_id(channel))
	{
		hev_impathy_server_sasl_handler_new_async(account,
					channel, server_sasl_handler_ready_handler,
					data);
	}
}

static void claim_handler(GObject *source,
			GAsyncResult *res, gpointer user_data)
{
	ObserveChannelsData *data = user_data;
	GError *error = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(!tp_channel_dispatch_operation_claim_with_finish(
					TP_CHANNEL_DISPATCH_OPERATION(source),
					res, &error))
	{
		g_clear_error(&error);
	}
	else
	{
		HandlerContextData *h_data = NULL;

		h_data = handler_context_data_new(data->self, NULL);

		hev_impathy_server_sasl_handler_new_async(TP_ACCOUNT(data->account),
					data->channel, server_sasl_handler_ready_handler, h_data);
	}

	observe_channels_data_free(data);
}

static void get_password_handler(GObject *source,
			GAsyncResult *res, gpointer user_data)
{
	ObserveChannelsData *data = user_data;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(NULL==hev_impathy_keyring_get_account_password_finish(TP_ACCOUNT(source),
					res, NULL))
	{
		/* We don't actually mind if this fails, just let the approver
		 * go ahead and take the channel. */

		tp_observe_channels_context_accept(data->context);
		observe_channels_data_free(data);
	}
	else
	{
		tp_channel_dispatch_operation_claim_with_async(data->dispatch_operation,
					TP_BASE_CLIENT(data->self), claim_handler, data);

		tp_observe_channels_context_accept(data->context);
	}
}

static void hev_impathy_auth_factory_observe_channels(TpBaseClient *client,
			TpAccount *account, TpConnection *connection,
			GList *channels, TpChannelDispatchOperation *dispatch_operation,
			GList *requests, TpObserveChannelsContext *context)
{
	HevImpathyAuthFactory *self = HEV_IMPATHY_AUTH_FACTORY(client);
	TpChannel *channel = NULL;
	GError *error = NULL;
	ObserveChannelsData *data = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(!common_checks(self, channels, TRUE, &error))
	{
		tp_observe_channels_context_fail(context, error);
		g_clear_error(&error);
		return;
	}

	/* We're now sure this is a server auth channel using the SASL auth
	 * type and X-TELEPATHY-PASSWORD is available. Great. */

	channel = channels->data;

	data = g_slice_new0(ObserveChannelsData);
	data->self = self;
	data->context = g_object_ref(context);
	data->dispatch_operation = g_object_ref(dispatch_operation);
	data->account = g_object_ref(account);
	data->channel = g_object_ref(channel);

	hev_impathy_keyring_get_account_password_async(account,
				get_password_handler, data);

	tp_observe_channels_context_delay(context);
}

static void hev_impathy_auth_factory_class_init(HevImpathyAuthFactoryClass * klass)
{
	GObjectClass * obj_class = G_OBJECT_CLASS(klass);
	TpBaseClientClass *base_client_class = TP_BASE_CLIENT_CLASS(klass);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	obj_class->constructor = hev_impathy_auth_factory_constructor;
	obj_class->constructed = hev_impathy_auth_factory_constructed;
	obj_class->dispose = hev_impathy_auth_factory_dispose;
	obj_class->finalize = hev_impathy_auth_factory_finalize;

	base_client_class->handle_channels =
		hev_impathy_auth_factory_handle_channels;
	base_client_class->observe_channels =
		hev_impathy_auth_factory_observe_channels;

	hev_impathy_auth_factory_signals[NEW_SERVER_TLS_HANDLER] =
		g_signal_new("new-server-tls-handler",
					G_TYPE_FROM_CLASS(klass),
					G_SIGNAL_RUN_LAST, 0,
					NULL, NULL,
					g_cclosure_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					HEV_TYPE_IMPATHY_SERVER_TLS_HANDLER);
	
	hev_impathy_auth_factory_signals[NEW_SERVER_SASL_HANDLER] =
		g_signal_new("new-server-sasl-handler",
					G_TYPE_FROM_CLASS(klass),
					G_SIGNAL_RUN_LAST, 0,
					NULL, NULL,
					g_cclosure_marshal_VOID__OBJECT,
					G_TYPE_NONE, 1,
					HEV_TYPE_IMPATHY_SERVER_SASL_HANDLER);

	g_type_class_add_private(klass, sizeof(HevImpathyAuthFactoryPrivate));
}

static void hev_impathy_auth_factory_init(HevImpathyAuthFactory * self)
{
	HevImpathyAuthFactoryPrivate * priv = HEV_IMPATHY_AUTH_FACTORY_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	priv->sasl_handlers = g_hash_table_new_full(g_str_hash,
				g_str_equal, NULL, g_object_unref);
}

GObject * hev_impathy_auth_factory_dup_singleton(void)
{
	GObject *obj = NULL;
	TpDBusDaemon *dbus_daemon = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	dbus_daemon = tp_dbus_daemon_dup(NULL);
	obj = g_object_new(HEV_TYPE_IMPATHY_AUTH_FACTORY,
				"dbus-daemon", dbus_daemon,
				"name", "Hev.Impathy.Auth",
				NULL);
	g_object_unref(dbus_daemon);

	return obj;
}

gboolean hev_impathy_auth_factory_register(HevImpathyAuthFactory *self,
			GError **error)
{
	return tp_base_client_register(TP_BASE_CLIENT(self), error);
}

