/*
 ============================================================================
 Name        : hev-impathy-server-sasl-handler.c
 Author      : Heiher <admin@heiher.info>
 Version     : 0.0.1
 Copyright   : Copyright (C) 2011 everyone.
 Description : 
 ============================================================================
 */

#include <string.h>
#include <telepathy-glib/util.h>
#include <telepathy-glib/interfaces.h>

#include "hev-impathy-server-sasl-handler.h"
#include "hev-impathy-utils.h"

enum
{
	PROP_CHANNEL = 1,
	PROP_ACCOUNT,
	N_PROPERTIES
};

enum
{
	INVALIDATED,
	LAST_SIGNAL
};

static guint hev_impathy_signals[LAST_SIGNAL] = { 0 };

static const gchar *sasl_statuses[] =
{
	"not started",
	"in progress",
	"server succeeded",
	"client accepted",
	"succeeded",
	"server failed",
	"client failed"
};

#define HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(obj)	(G_TYPE_INSTANCE_GET_PRIVATE((obj), HEV_TYPE_IMPATHY_SERVER_SASL_HANDLER, HevImpathyServerSASLHandlerPrivate))

typedef struct _HevImpathyServerSASLHandlerPrivate HevImpathyServerSASLHandlerPrivate;

struct _HevImpathyServerSASLHandlerPrivate
{
	TpChannel *channel;
	TpAccount *account;

	GSimpleAsyncResult *result;

	gchar *password;
	gboolean save_password;

	GSimpleAsyncResult *async_init_res;
};

static void hev_async_initable_iface_init(GAsyncInitableIface *iface);

G_DEFINE_TYPE_WITH_CODE(HevImpathyServerSASLHandler, hev_impathy_server_sasl_handler, G_TYPE_OBJECT,
			G_IMPLEMENT_INTERFACE(G_TYPE_ASYNC_INITABLE, hev_async_initable_iface_init));

static void hev_impathy_server_sasl_handler_dispose(GObject * obj)
{
	HevImpathyServerSASLHandler * self = HEV_IMPATHY_SERVER_SASL_HANDLER(obj);
	HevImpathyServerSASLHandlerPrivate * priv = HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(priv->channel)
	{
		tp_clear_object(&priv->channel);
		priv->channel = NULL;
	}

	if(priv->account)
	{
		tp_clear_object(&priv->account);
		priv->account = NULL;
	}

	G_OBJECT_CLASS(hev_impathy_server_sasl_handler_parent_class)->dispose(obj);
}

static void hev_impathy_server_sasl_handler_finalize(GObject * obj)
{
	HevImpathyServerSASLHandler * self = HEV_IMPATHY_SERVER_SASL_HANDLER(obj);
	HevImpathyServerSASLHandlerPrivate * priv = HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(priv->password)
	{
		tp_clear_pointer(&priv->password, g_free);
		priv->password = NULL;
	}

	G_OBJECT_CLASS(hev_impathy_server_sasl_handler_parent_class)->finalize(obj);
}

static GObject * hev_impathy_server_sasl_handler_constructor(GType type, guint n, GObjectConstructParam * param)
{
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);
	return G_OBJECT_CLASS(hev_impathy_server_sasl_handler_parent_class)->constructor(type, n, param);
}

static void hev_impathy_server_sasl_handler_set_password_handler(GObject *source,
			GAsyncResult *res, gpointer user_data)
{
	GError *error = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);
	
	if(!hev_impathy_keyring_set_account_password_finish(TP_ACCOUNT(source),
					res, &error))
	{
		g_clear_error(&error);
	}
}

static void sasl_status_changed_handler(TpChannel *channel,
			guint status, const gchar *error,
			GHashTable *details, gpointer user_data,
			GObject *weak_object)
{
	HevImpathyServerSASLHandler *self =
		HEV_IMPATHY_SERVER_SASL_HANDLER(weak_object);
	HevImpathyServerSASLHandlerPrivate *priv =
		HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(G_N_ELEMENTS(sasl_statuses) <= status)
	  return;

	if(TP_SASL_STATUS_SERVER_SUCCEEDED == status)
	{
		if(priv->save_password)
		{
			hev_impathy_keyring_set_account_password_async(priv->account,
						priv->password, hev_impathy_server_sasl_handler_set_password_handler,
						NULL);
		}

		tp_cli_channel_interface_sasl_authentication_call_accept_sasl(
					priv->channel, -1, NULL, NULL, NULL, NULL);
	}
	else if(TP_SASL_STATUS_SUCCEEDED == status)
	{
		tp_cli_channel_call_close(priv->channel, -1, NULL, NULL,
					NULL, NULL);
	}
}

static void channel_invalidated_handler(TpProxy *proxy,
			guint domain, gint code, gchar *message,
			HevImpathyServerSASLHandler *self)
{
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_signal_emit(self, hev_impathy_signals[INVALIDATED], 0);
}

static void hev_impathy_server_sasl_handler_constructed(GObject * obj)
{
	HevImpathyServerSASLHandler *self = HEV_IMPATHY_SERVER_SASL_HANDLER(obj);
	HevImpathyServerSASLHandlerPrivate *priv =
		HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);
	GError *error = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	G_OBJECT_CLASS(hev_impathy_server_sasl_handler_parent_class)->constructed(obj);

	tp_cli_channel_interface_sasl_authentication_connect_to_sasl_status_changed(
				priv->channel, sasl_status_changed_handler, NULL, NULL, obj, &error);
	if(NULL != error)
	{
		g_clear_error(&error);
		return;
	}

	tp_g_signal_connect_object(priv->channel, "invalidated",
				G_CALLBACK(channel_invalidated_handler), obj, 0);
}

static gboolean hev_impathy_server_sasl_handler_give_password(gpointer user_data)
{
	HevImpathyServerSASLHandler *self = HEV_IMPATHY_SERVER_SASL_HANDLER(user_data);
	HevImpathyServerSASLHandlerPrivate *priv =
		HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	hev_impathy_server_sasl_handler_provide_password(self,
				priv->password, FALSE);

	return FALSE;
}

static void hev_impathy_server_sasl_handler_get_password_async_handler(GObject *source,
			GAsyncResult *res, gpointer user_data)
{
	HevImpathyServerSASLHandler *self = HEV_IMPATHY_SERVER_SASL_HANDLER(user_data);
	HevImpathyServerSASLHandlerPrivate *priv =
		HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);
	const gchar *password = NULL;
	GError *error = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	password = hev_impathy_keyring_get_account_password_finish(TP_ACCOUNT(source),
				res, &error);
	if(NULL != password)
	{
		priv->password = g_strdup(password);

		/* Do this in an idle so the async result will get there
		 * first. */
		g_idle_add(hev_impathy_server_sasl_handler_give_password, user_data);
	}

	g_simple_async_result_complete(priv->async_init_res);
	tp_clear_object(&priv->async_init_res);
}

static void hev_impathy_server_sasl_handler_init_async(GAsyncInitable *initable,
			gint io_priority, GCancellable *cancellable,
			GAsyncReadyCallback callback, gpointer user_data)
{
	HevImpathyServerSASLHandler *self = HEV_IMPATHY_SERVER_SASL_HANDLER(initable);
	HevImpathyServerSASLHandlerPrivate *priv =
		HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(TP_IS_ACCOUNT(priv->account));

	priv->async_init_res = g_simple_async_result_new(G_OBJECT(self),
				callback, user_data,
				hev_impathy_server_sasl_handler_new_async);

	hev_impathy_keyring_get_account_password_async(priv->account,
				hev_impathy_server_sasl_handler_get_password_async_handler,
				self);
}

static gboolean hev_impathy_server_sasl_handler_init_finish(GAsyncInitable *initable,
			GAsyncResult *res, GError **error)
{
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(g_simple_async_result_propagate_error(G_SIMPLE_ASYNC_RESULT(res),
					error))
	  return FALSE;

	return TRUE;
}

static void hev_async_initable_iface_init(GAsyncInitableIface *iface)
{
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	iface->init_async = hev_impathy_server_sasl_handler_init_async;
	iface->init_finish = hev_impathy_server_sasl_handler_init_finish;
}

static void hev_impathy_server_sasl_handler_get_property(GObject *object,
			guint property_id, GValue *value, GParamSpec *pspec)
{
	HevImpathyServerSASLHandler *self = HEV_IMPATHY_SERVER_SASL_HANDLER(object);
	HevImpathyServerSASLHandlerPrivate *priv =
		HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	switch(property_id)
	{
	case PROP_CHANNEL:
		g_value_set_object(value, priv->channel);
		break;
	case PROP_ACCOUNT:
		g_value_set_object(value, priv->account);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object,
					property_id, pspec);
		break;
	}
}

static void hev_impathy_server_sasl_handler_set_property(GObject *object,
			guint property_id, const GValue *value, GParamSpec *pspec)
{
	HevImpathyServerSASLHandler *self = HEV_IMPATHY_SERVER_SASL_HANDLER(object);
	HevImpathyServerSASLHandlerPrivate *priv =
		HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	switch(property_id)
	{
	case PROP_CHANNEL:
		priv->channel = g_value_dup_object(value);
		break;
	case PROP_ACCOUNT:
		priv->account = g_value_dup_object(value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object,
					property_id, pspec);
		break;
	}
}

static void hev_impathy_server_sasl_handler_class_init(HevImpathyServerSASLHandlerClass * klass)
{
	GObjectClass * obj_class = G_OBJECT_CLASS(klass);
	GParamSpec *pspec = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	obj_class->constructor = hev_impathy_server_sasl_handler_constructor;
	obj_class->constructed = hev_impathy_server_sasl_handler_constructed;
	obj_class->dispose = hev_impathy_server_sasl_handler_dispose;
	obj_class->finalize = hev_impathy_server_sasl_handler_finalize;
	obj_class->get_property = hev_impathy_server_sasl_handler_get_property;
	obj_class->set_property = hev_impathy_server_sasl_handler_set_property;

	pspec = g_param_spec_object("channel", "The TpChannel",
				"The TpChannel this handler is supposed to handle.",
				TP_TYPE_CHANNEL, G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY |
				G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(obj_class, PROP_CHANNEL, pspec);

	pspec = g_param_spec_object("account", "The TpAccount",
				"The TpAccount this channel belongs to.",
				TP_TYPE_ACCOUNT, G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY |
				G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(obj_class, PROP_ACCOUNT, pspec);

	hev_impathy_signals[INVALIDATED] = g_signal_new("invalidated",
				G_TYPE_FROM_CLASS(klass),
				G_SIGNAL_RUN_LAST, 0,
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0);

	g_type_class_add_private(klass, sizeof(HevImpathyServerSASLHandlerPrivate));
}

static void hev_impathy_server_sasl_handler_init(HevImpathyServerSASLHandler * self)
{
	HevImpathyServerSASLHandlerPrivate * priv = HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);
}

void hev_impathy_server_sasl_handler_new_async(
			TpAccount *account, TpChannel *channel,
			GAsyncReadyCallback callback,
			gpointer user_data)
{
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(TP_IS_ACCOUNT(account));
	g_return_if_fail(TP_IS_CHANNEL(channel));
	g_return_if_fail(NULL!=callback);

	g_async_initable_new_async(HEV_TYPE_IMPATHY_SERVER_SASL_HANDLER,
				G_PRIORITY_DEFAULT, NULL, callback, user_data,
				"account", account, "channel", channel,
				NULL);
}

GObject * hev_impathy_server_sasl_handler_new_finish(
			GAsyncResult *res, GError **error)
{
	GObject *object = NULL, *source_object = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	source_object = g_async_result_get_source_object(res);

	object = g_async_initable_new_finish(G_ASYNC_INITABLE(source_object),
				res, error);
	g_object_unref(source_object);

	return object;
}

static void start_mechanism_with_data_handler(TpChannel *channel,
			const GError *error, gpointer user_data,
			GObject *weak_object)
{
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);
}

void hev_impathy_server_sasl_handler_provide_password(
			HevImpathyServerSASLHandler *self,
			const gchar *password, gboolean remember)
{
	HevImpathyServerSASLHandlerPrivate * priv = NULL;
	GArray *array = NULL;
	gboolean may_save_response = FALSE,
			 may_save_response_valid = FALSE;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(HEV_IS_IMPATHY_SERVER_SASL_HANDLER(self));

	priv = HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	array = g_array_sized_new(TRUE, FALSE,
				sizeof(gchar), strlen(password));

	g_array_append_vals(array, password, strlen(password));

	tp_cli_channel_interface_sasl_authentication_call_start_mechanism_with_data(
				priv->channel, -1, "X-TELEPATHY-PASSWORD", array,
				start_mechanism_with_data_handler, NULL, NULL, G_OBJECT(self));

	g_array_unref(array);

	/* determine if we are permitted to save the password locally */
	may_save_response = tp_asv_get_boolean(
				tp_channel_borrow_immutable_properties(priv->channel),
				TP_PROP_CHANNEL_INTERFACE_SASL_AUTHENTICATION_MAY_SAVE_RESPONSE,
				&may_save_response_valid);

	if(!may_save_response_valid)
	{
		may_save_response = TRUE;
	}

	if(remember)
	{
		if(may_save_response)
		{
			g_free(priv->password);

			/* We'll save the password if we manage to connect */
			priv->password = g_strdup(password);
			priv->save_password = TRUE;
		}
#if 0 /* TODO */
		else if(tp_proxy_has_interface_by_id(priv->channel,
						EMP_IFACE_QUARK_CHANNEL_INTERFACE_CREDENTIALS_STORAGE))
		{
		}
		else
		{
		}
#endif
	}

	if(!may_save_response)
	{
		/* delete any password present, it shouldn't be there */
		hev_impathy_keyring_delete_account_password_async(priv->account,
					NULL, NULL);
	}

#if 0 /* TODO */
	/* Additionally, if we implement Ch.I.CredentialsStorage, inform that
	 * whether we want to remember the password. */
	if(tp_proxy_has_interface_by_id(priv->channel,
					EMP_IFACE_QUARK_CHANNEL_INTERFACE_CREDENTIALS_STORAGE))
	{
		emp_cli_channel_interface_credentials_storage_call_store_credentials(
					TP_PROXY(priv->channel), -1, remember, NULL, NULL, NULL, NULL);
	}
#endif
}

void hev_impathy_server_sasl_handler_cancel(
			HevImpathyServerSASLHandler *self)
{
	HevImpathyServerSASLHandlerPrivate * priv = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(HEV_IS_IMPATHY_SERVER_SASL_HANDLER(self));

	priv = HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	tp_cli_channel_interface_sasl_authentication_call_abort_sasl(
				priv->channel, -1, TP_SASL_ABORT_REASON_USER_ABORT,
				"User cancelled the authentication",
				NULL, NULL, NULL, NULL);
}

TpAccount * hev_impathy_server_sasl_handler_get_account(
			HevImpathyServerSASLHandler *self)
{
	HevImpathyServerSASLHandlerPrivate * priv = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(HEV_IS_IMPATHY_SERVER_SASL_HANDLER(self));

	priv = HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	return priv->account;
}

TpChannel * hev_impathy_server_sasl_handler_get_channel(
			HevImpathyServerSASLHandler *self)
{
	HevImpathyServerSASLHandlerPrivate * priv = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(HEV_IS_IMPATHY_SERVER_SASL_HANDLER(self));

	priv = HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	return priv->channel;
}

gboolean hev_impathy_server_sasl_handler_has_password(
			HevImpathyServerSASLHandler *self)
{
	HevImpathyServerSASLHandlerPrivate * priv = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(HEV_IS_IMPATHY_SERVER_SASL_HANDLER(self));

	priv = HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	return (NULL != priv->password);
}

gboolean hev_impathy_server_sasl_handler_can_save_response_somewhere(
			HevImpathyServerSASLHandler *self)
{
	HevImpathyServerSASLHandlerPrivate * priv = NULL;
	gboolean may_save_response = FALSE,
			 may_save_response_valid = FALSE;
	gboolean has_storage_iface = FALSE;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(HEV_IS_IMPATHY_SERVER_SASL_HANDLER(self));

	priv = HEV_IMPATHY_SERVER_SASL_HANDLER_GET_PRIVATE(self);

	/* determine if we are permitted to save the password locally */
	may_save_response = tp_asv_get_boolean(
				tp_channel_borrow_immutable_properties(priv->channel),
				TP_PROP_CHANNEL_INTERFACE_SASL_AUTHENTICATION_MAY_SAVE_RESPONSE,
				&may_save_response_valid);

	if(!may_save_response_valid)
	{
		may_save_response = TRUE;
	}

#if 0 /* TODO */
	has_storage_iface = tp_proxy_has_interface_by_id(priv->channel,
				EMP_IFACE_QUARK_CHANNEL_INTERFACE_CREDENTIALS_STORAGE);
#endif

	return may_save_response || has_storage_iface;

}

