/*
 ============================================================================
 Name        : hev-impathy-tls-verifier.c
 Author      : Heiher <admin@heiher.info>
 Version     : 0.0.1
 Copyright   : Copyright (C) 2011 everyone.
 Description : 
 ============================================================================
 */

#include <gnutls/gnutls.h>
#include <gnutls/x509.h>
#include <telepathy-glib/util.h>

#include "hev-impathy-tls-verifier.h"
#include "hev-impathy-tls-certificate.h"

enum
{
	PROP_TLS_CERTIFICATE = 1,
	PROP_HOSTNAME,
	N_PROPERTIES
};

#define HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(obj)	(G_TYPE_INSTANCE_GET_PRIVATE((obj), HEV_TYPE_IMPATHY_TLS_VERIFIER, HevImpathyTLSVerifierPrivate))

typedef struct _HevImpathyTLSVerifierPrivate HevImpathyTLSVerifierPrivate;

struct _HevImpathyTLSVerifierPrivate
{
	GPtrArray *cert_chain;
	GPtrArray *trusted_ca_list;

	GObject *certificate;
	gchar *hostname;

	GSimpleAsyncResult *verify_result;
	GHashTable *details;
};

G_DEFINE_TYPE(HevImpathyTLSVerifier, hev_impathy_tls_verifier, G_TYPE_OBJECT);

static const gchar *system_ca_paths[] =
{
	"/etc/ssl/certs/ca-certificates.crt",
	NULL
};

static void hev_impathy_tls_verifier_dispose(GObject * obj)
{
	HevImpathyTLSVerifier * self = HEV_IMPATHY_TLS_VERIFIER(obj);
	HevImpathyTLSVerifierPrivate * priv = HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(priv->certificate)
	{
		tp_clear_object(&priv->certificate);
		priv->certificate = NULL;
	}

	G_OBJECT_CLASS(hev_impathy_tls_verifier_parent_class)->dispose(obj);
}

static void hev_impathy_tls_verifier_finalize(GObject * obj)
{
	HevImpathyTLSVerifier * self = HEV_IMPATHY_TLS_VERIFIER(obj);
	HevImpathyTLSVerifierPrivate * priv = HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(priv->cert_chain)
	{
		tp_clear_pointer(&priv->cert_chain, g_ptr_array_unref);
		priv->cert_chain = NULL;
	}

	if(priv->trusted_ca_list)
	{
		tp_clear_pointer(&priv->trusted_ca_list, g_ptr_array_unref);
		priv->trusted_ca_list = NULL;
	}

	if(priv->details)
	{
		tp_clear_boxed(G_TYPE_HASH_TABLE, &priv->details);
		priv->details = NULL;
	}

	if(priv->hostname)
	{
		g_free(priv->hostname);
		priv->hostname = NULL;
	}

	G_OBJECT_CLASS(hev_impathy_tls_verifier_parent_class)->finalize(obj);
}

static GObject * hev_impathy_tls_verifier_constructor(GType type, guint n, GObjectConstructParam * param)
{
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);
	return G_OBJECT_CLASS(hev_impathy_tls_verifier_parent_class)->constructor(type, n, param);
}

static void build_gnutls_cert_list(HevImpathyTLSVerifier *self)
{
	HevImpathyTLSVerifierPrivate *priv =
		HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);
	guint num_certs = 0;
	guint idx = 0;
	GPtrArray *certificate_data = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_object_get(priv->certificate, "cert-data",
				&certificate_data, NULL);
	num_certs = certificate_data->len;

	priv->cert_chain = g_ptr_array_new_with_free_func(
				(GDestroyNotify)gnutls_x509_crt_deinit);
	for(idx=0; idx<num_certs; idx++)
	{
		gnutls_x509_crt_t cert = 0;
		GArray *one_cert = NULL;
		gnutls_datum_t datum = { NULL, 0 };

		one_cert = g_ptr_array_index(certificate_data, idx);
		datum.data = (guchar *)one_cert->data;
		datum.size = one_cert->len;

		gnutls_x509_crt_init(&cert);
		gnutls_x509_crt_import(cert, &datum, GNUTLS_X509_FMT_DER);

		g_ptr_array_add(priv->cert_chain, cert);
	}
}

static void hev_impathy_tls_verifier_constructed(GObject * obj)
{
	HevImpathyTLSVerifier *self = HEV_IMPATHY_TLS_VERIFIER(obj);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	G_OBJECT_CLASS(hev_impathy_tls_verifier_parent_class)->constructed(obj);

	build_gnutls_cert_list(self);
}

static void hev_impathy_tls_verifier_get_property(GObject *object,
			guint property_id, GValue *value, GParamSpec *pspec)
{
	HevImpathyTLSVerifier *self = HEV_IMPATHY_TLS_VERIFIER(object);
	HevImpathyTLSVerifierPrivate *priv =
		HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	switch(property_id)
	{
	case PROP_TLS_CERTIFICATE:
		g_value_set_object(value, priv->certificate);
		break;
	case PROP_HOSTNAME:
		g_value_set_string(value, priv->hostname);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object,
					property_id, pspec);
		break;
	}
}

static void hev_impathy_tls_verifier_set_property(GObject *object,
			guint property_id, const GValue *value, GParamSpec *pspec)
{
	HevImpathyTLSVerifier *self = HEV_IMPATHY_TLS_VERIFIER(object);
	HevImpathyTLSVerifierPrivate *priv =
		HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	switch(property_id)
	{
	case PROP_TLS_CERTIFICATE:
		priv->certificate = g_value_dup_object(value);
		break;
	case PROP_HOSTNAME:
		priv->hostname = g_value_dup_string(value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object,
					property_id, pspec);
		break;
	}
}

static void hev_impathy_tls_verifier_class_init(HevImpathyTLSVerifierClass * klass)
{
	GObjectClass * obj_class = G_OBJECT_CLASS(klass);
	GParamSpec *pspec = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	obj_class->constructor = hev_impathy_tls_verifier_constructor;
	obj_class->constructed = hev_impathy_tls_verifier_constructed;
	obj_class->dispose = hev_impathy_tls_verifier_dispose;
	obj_class->finalize = hev_impathy_tls_verifier_finalize;
	obj_class->get_property = hev_impathy_tls_verifier_get_property;
	obj_class->set_property = hev_impathy_tls_verifier_set_property;

	pspec = g_param_spec_object("certificate", "The HevImpathyTLSCertificate",
				"The HevImpathyTLSCertificate to be verified.",
				HEV_TYPE_IMPATHY_TLS_CERTIFICATE,
				G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE |
				G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(obj_class, PROP_TLS_CERTIFICATE, pspec);

	pspec = g_param_spec_string("hostname", "The hostname",
				"The hostname which is certified by the certificate.",
				NULL, G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE |
				G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(obj_class, PROP_HOSTNAME, pspec);

	g_type_class_add_private(klass, sizeof(HevImpathyTLSVerifierPrivate));
}

static void hev_impathy_tls_verifier_init(HevImpathyTLSVerifier * self)
{
	HevImpathyTLSVerifierPrivate * priv = HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	priv->details = tp_asv_new(NULL, NULL);
}

GObject * hev_impathy_tls_verifier_new(GObject *certificate,
			const gchar *hostname)
{
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_val_if_fail(HEV_IS_IMPATHY_TLS_CERTIFICATE(certificate),
				NULL);
	g_return_val_if_fail(NULL!=hostname, NULL);

	return g_object_new(HEV_TYPE_IMPATHY_TLS_VERIFIER,
				"certificate", certificate,
				"hostname", hostname,
				NULL);
}

static gchar * hev_impathy_get_x509_certificate_hostname(gnutls_x509_crt_t cert)
{
	gchar dns_name[256] = { 0 };
	gsize dns_name_size = 0;
	gint idx = 0;
	gint res = 0;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	/* this snippet is taken from GnuTLS.
	 * see gnutls/lib/x509/rfc2818_hostname.c
	 */
	for(idx=0; 0<=res; idx++)
	{
		dns_name_size = sizeof(dns_name);
		res = gnutls_x509_crt_get_subject_alt_name(cert,
					idx, dns_name, &dns_name_size, NULL);

		if((GNUTLS_SAN_DNSNAME==res) ||
					(GNUTLS_SAN_IPADDRESS==res))
		  return g_strndup(dns_name, dns_name_size);
	}

	dns_name_size = sizeof(dns_name);
	res = gnutls_x509_crt_get_dn_by_oid(cert, GNUTLS_OID_X520_COMMON_NAME,
				0, 0, dns_name, &dns_name_size);

	if(0 <= res)
	  return g_strndup(dns_name, dns_name_size);

	return NULL;
}

static void abort_verification(HevImpathyTLSVerifier *self,
			TpTLSCertificateRejectReason reason)
{
	HevImpathyTLSVerifierPrivate *priv =
		HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_simple_async_result_set_error(priv->verify_result,
				G_IO_ERROR, reason,
				"TLS verification failed with reason %u",
				reason);
	g_simple_async_result_complete_in_idle(priv->verify_result);

	tp_clear_object(&priv->verify_result);
}

static void complete_verification(HevImpathyTLSVerifier *self)
{
	HevImpathyTLSVerifierPrivate *priv =
		HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_simple_async_result_complete_in_idle(priv->verify_result);

	tp_clear_object(&priv->verify_result);
}

static gboolean verification_output_to_reason(gint res,
			guint verify_output, TpTLSCertificateRejectReason *reason)
{
	gboolean retval = TRUE;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_val_if_fail(NULL!=reason, FALSE);

	if(GNUTLS_E_SUCCESS != res)
	{
		retval = FALSE;

		/* the certificate is not structurally valid */
		switch (res)
        {
		case GNUTLS_E_INSUFFICIENT_CREDENTIALS:
			*reason = TP_TLS_CERTIFICATE_REJECT_REASON_UNTRUSTED;
			break;
		case GNUTLS_E_CONSTRAINT_ERROR:
			*reason = TP_TLS_CERTIFICATE_REJECT_REASON_LIMIT_EXCEEDED;
			break;
		default:
			*reason = TP_TLS_CERTIFICATE_REJECT_REASON_UNKNOWN;
			break;
		}
		
		goto out;
	}

	/* the certificate is structurally valid, check for other errors. */
	if (verify_output & GNUTLS_CERT_INVALID)
	{
		retval = FALSE;

		if (verify_output & GNUTLS_CERT_SIGNER_NOT_FOUND)
			*reason = TP_TLS_CERTIFICATE_REJECT_REASON_SELF_SIGNED;
		else if (verify_output & GNUTLS_CERT_SIGNER_NOT_CA)
			*reason = TP_TLS_CERTIFICATE_REJECT_REASON_UNTRUSTED;
		else if (verify_output & GNUTLS_CERT_INSECURE_ALGORITHM)
			*reason = TP_TLS_CERTIFICATE_REJECT_REASON_INSECURE;
		else if (verify_output & GNUTLS_CERT_NOT_ACTIVATED)
			*reason = TP_TLS_CERTIFICATE_REJECT_REASON_NOT_ACTIVATED;
		else if (verify_output & GNUTLS_CERT_EXPIRED)
			*reason = TP_TLS_CERTIFICATE_REJECT_REASON_EXPIRED;
		else
			*reason = TP_TLS_CERTIFICATE_REJECT_REASON_UNKNOWN;

		goto out;
	}

out:
	return retval;
}

static gboolean verify_certificate(HevImpathyTLSVerifier *self,
			gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer,
			TpTLSCertificateRejectReason *reason)
{
	guint verify_output = 0;
	gint res = 0;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	res = gnutls_x509_crt_verify(cert, &issuer, 1, 0, &verify_output);

	return verification_output_to_reason(res, verify_output, reason);
}

static gnutls_x509_crt_t * ptr_array_to_x509_crt_list(GPtrArray *chain)
{
	gnutls_x509_crt_t *retval = NULL;
	gint idx = 0;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	retval = g_malloc0(sizeof(gnutls_x509_crt_t) * chain->len);

	for(idx=0; idx<(gint)chain->len; idx++)
	  retval[idx] = g_ptr_array_index(chain, idx);

	return retval;
}

static gboolean verify_last_certificate(HevImpathyTLSVerifier *self,
			gnutls_x509_crt_t cert, TpTLSCertificateRejectReason *reason)
{
	guint verify_output = 0;
	gint res = 0;
	gnutls_x509_crt_t *trusted_ca_list = NULL;
	HevImpathyTLSVerifierPrivate *priv =
		HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(0 < priv->trusted_ca_list->len)
	{
		trusted_ca_list = ptr_array_to_x509_crt_list(priv->trusted_ca_list);
		res = gnutls_x509_crt_verify(cert, trusted_ca_list,
					priv->trusted_ca_list->len, 0, &verify_output);

		g_free(trusted_ca_list);
	}
	else
	{
		/* check it against itself to see if it's structurally valid */
		res = gnutls_x509_crt_verify(cert, &cert, 1, 0, &verify_output);

		/* if it's valid, return the SelfSigned error, so that we can add it
		* later to our trusted CAs whitelist.
		*/
		if(GNUTLS_E_SUCCESS == res)
		{
			*reason = TP_TLS_CERTIFICATE_REJECT_REASON_SELF_SIGNED;
			return FALSE;
		}
	}

	return verification_output_to_reason (res, verify_output, reason);
}

static void real_start_verification(HevImpathyTLSVerifier *self)
{
	HevImpathyTLSVerifierPrivate *priv =
		HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);
	gnutls_x509_crt_t first_cert = 0, last_cert = 0;
	gint idx = 0;
	gboolean res = FALSE;
	gint num_certs = 0;
	TpTLSCertificateRejectReason reason =
		TP_TLS_CERTIFICATE_REJECT_REASON_UNKNOWN;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	/* check if the certificate matches the hostname first. */
	first_cert = g_ptr_array_index(priv->cert_chain, 0);
	if(0 == gnutls_x509_crt_check_hostname(first_cert, priv->hostname))
	{
		gchar *certified_hostname = NULL;

		reason = TP_TLS_CERTIFICATE_REJECT_REASON_HOSTNAME_MISMATCH;
		certified_hostname =
			(gchar *)hev_impathy_get_x509_certificate_hostname(first_cert);
		tp_asv_set_string(priv->details,
					"expected-hostname", priv->hostname);
		tp_asv_set_string(priv->details,
					"certificate-hostname", certified_hostname);

		g_free(certified_hostname);
		
		goto out;
	}

	num_certs = priv->cert_chain->len;

	if(0 < priv->trusted_ca_list->len)
	{
		/* if the last certificate is self-signed, and we have a list of
		* trusted CAs, ignore it, as we want to check the chain against our
		* trusted CAs list first.
		* if we have only one certificate in the chain, don't ignore it though,
		* as it's the CA certificate itself.
		*/
		last_cert = g_ptr_array_index(priv->cert_chain, num_certs - 1);

		if((0<gnutls_x509_crt_check_issuer(last_cert, last_cert)) &&
					(1<num_certs))
		  num_certs --;
	}

	for (idx = 1; idx < num_certs; idx++)
	{
		res = verify_certificate(self,
					g_ptr_array_index(priv->cert_chain, idx-1),
		g_ptr_array_index(priv->cert_chain, idx), &reason);

		if(!res)
		{
			abort_verification(self, reason);
			return;
		}
	}

	res = verify_last_certificate(self,
				g_ptr_array_index(priv->cert_chain,
					num_certs - 1), &reason);

out:
	if(!res)
	{
		abort_verification(self, reason);
		return;
	}

	complete_verification(self);
}

static gboolean start_verification(gpointer user_data)
{
	HevImpathyTLSVerifier *self = HEV_IMPATHY_TLS_VERIFIER(user_data);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	real_start_verification(self);

	return FALSE;
}

static gint get_number_and_type_of_certificates(gnutls_datum_t *datum,
			gnutls_x509_crt_fmt_t *format)
{
	gnutls_x509_crt_t fake = 0;
	guint retval = 1;
	gint res = 0;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	res = gnutls_x509_crt_list_import(&fake, &retval, datum,
				GNUTLS_X509_FMT_PEM, GNUTLS_X509_CRT_LIST_IMPORT_FAIL_IF_EXCEED);

	if(GNUTLS_E_SHORT_MEMORY_BUFFER==res || 0<res)
	{
		*format = GNUTLS_X509_FMT_PEM;
		return retval;
	}

	/* Try DER */
	res = gnutls_x509_crt_list_import(&fake, &retval, datum,
				GNUTLS_X509_FMT_DER, 0);

	if(0 < res)
	{
		*format = GNUTLS_X509_FMT_DER;
		return retval;
	}

	return res;
}

static gboolean build_gnutls_ca_and_crl_lists(GIOSchedulerJob *job,
			GCancellable *cancellable, gpointer user_data)
{
	HevImpathyTLSVerifier *self = HEV_IMPATHY_TLS_VERIFIER(user_data);
	HevImpathyTLSVerifierPrivate *priv =
		HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);
	gint idx = 0;
	gchar *user_certs_dir = NULL;
	GDir *dir = NULL;
	GError *error = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	priv->trusted_ca_list = g_ptr_array_new_with_free_func(
				(GDestroyNotify)gnutls_x509_crt_deinit);

	for(idx=0; idx<(gint)G_N_ELEMENTS(system_ca_paths)-1; idx++)
	{
		const gchar *path = NULL;
		gchar *contents = NULL;
		gsize length = 0;
		gint res = 0, n_certs = 0;
		gnutls_x509_crt_t *cert_list = NULL;
		gnutls_datum_t datum = { NULL, 0 };
		gnutls_x509_crt_fmt_t format = 0;

		path = system_ca_paths[idx];
		g_file_get_contents(path, &contents, &length, &error);

		if(NULL != error)
		{
			g_clear_error(&error);
			continue;
		}

		datum.data = (guchar *)contents;
		datum.size = length;
		n_certs = get_number_and_type_of_certificates(&datum, &format);

		if(0 > n_certs)
		{
			g_free(contents);
			continue;
		}

		cert_list = g_malloc0(sizeof(gnutls_x509_crt_t) * n_certs);
		res = gnutls_x509_crt_list_import(cert_list, (guint *)&n_certs,
					&datum, format, 0);

		if(0 > res)
		{
			g_free(contents);
			continue;
		}

		for(idx=0; idx<n_certs; idx++)
		  g_ptr_array_add(priv->trusted_ca_list, cert_list[idx]);

		g_free(contents);
		g_free(cert_list);
	}

	/* user certs */
	user_certs_dir = g_build_filename(g_get_user_config_dir(),
				"telepathy", "certs", NULL);
	dir = g_dir_open(user_certs_dir, 0, &error);

	if(NULL == error)
	{
		const gchar *cert_name = NULL;

		while(NULL != (cert_name=g_dir_read_name(dir)))
		{
			gchar *contents = NULL, *cert_path = NULL;
			gsize length = 0;
			gint res = 0;
			gnutls_datum_t datum = { NULL, 0 };
			gnutls_x509_crt_t cert = 0;

			cert_path = g_build_filename(user_certs_dir, cert_name, NULL);

			g_file_get_contents(cert_path, &contents, &length, &error);
			if(NULL != error)
			{
				g_clear_error(&error);
				g_free(cert_path);
				continue;
			}

			datum.data = (guchar *)contents;
			datum.size = length;

			gnutls_x509_crt_init(&cert);
			res = gnutls_x509_crt_import(cert, &datum,
						GNUTLS_X509_FMT_PEM);
			if(GNUTLS_E_SUCCESS == res)
			  g_ptr_array_add(priv->trusted_ca_list, cert);

			g_free(contents);
			g_free(cert_path);
		}

		g_dir_close(dir);
	}
	else
	{
		g_error_free(error);
	}

	g_free(user_certs_dir);

	g_io_scheduler_job_send_to_mainloop_async(job,
				start_verification, self, NULL);

	return FALSE;
}

void hev_impathy_tls_verifier_verify_async(HevImpathyTLSVerifier *self,
			GAsyncReadyCallback callback, gpointer user_data)
{
	HevImpathyTLSVerifierPrivate *priv = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(HEV_IS_IMPATHY_TLS_VERIFIER(self));

	priv = HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);

	g_return_if_fail(NULL==priv->verify_result);

	priv->verify_result = g_simple_async_result_new(G_OBJECT(self),
				callback, user_data, NULL);
	g_io_scheduler_push_job(build_gnutls_ca_and_crl_lists,
				self, NULL, G_PRIORITY_DEFAULT, NULL);
}

gboolean hev_impathy_tls_verifier_verify_finish(HevImpathyTLSVerifier *self,
			GAsyncResult *res, TpTLSCertificateRejectReason *reason,
			GHashTable **details, GError **error)
{
	HevImpathyTLSVerifierPrivate *priv = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_val_if_fail(HEV_IS_IMPATHY_TLS_VERIFIER(self), FALSE);

	priv = HEV_IMPATHY_TLS_VERIFIER_GET_PRIVATE(self);

	if(g_simple_async_result_propagate_error(G_SIMPLE_ASYNC_RESULT(res),
					error))
	{
		if(NULL != reason)
		  *reason = (*error)->code;

		if(NULL != details)
		{
			*details = tp_asv_new(NULL, NULL);
			tp_g_hash_table_update(*details, priv->details,
						(GBoxedCopyFunc)g_strdup,
						(GBoxedCopyFunc)tp_g_value_slice_dup);
		}

		return FALSE;
	}

	if(NULL != reason)
	  *reason = TP_TLS_CERTIFICATE_REJECT_REASON_UNKNOWN;

	return TRUE;
}

