/*
 ============================================================================
 Name        : hev-impathy-utils.c
 Author      : Heiher <admin@heiher.info>
 Version     : 0.0.1
 Copyright   : Copyright (C) 2011 everyone.
 Description : 
 ============================================================================
 */

#include <string.h>
#include <gnome-keyring.h>
#include <telepathy-glib/account.h>
#include <telepathy-glib/channel.h>

#include "hev-impathy-utils.h"

static GnomeKeyringPasswordSchema account_keyring_schema =
{
	GNOME_KEYRING_ITEM_GENERIC_SECRET,
	{
		{ "account-id", GNOME_KEYRING_ATTRIBUTE_TYPE_STRING },
		{ "param-name", GNOME_KEYRING_ATTRIBUTE_TYPE_STRING },
		{ NULL }
	}
};

static void store_password_handler(GnomeKeyringResult result, gpointer user_data)
{
	GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT(user_data);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(GNOME_KEYRING_RESULT_OK != result)
	{
		GError *error = g_error_new_literal(TP_ERROR,
					TP_ERROR_DOES_NOT_EXIST,
					gnome_keyring_result_to_message(result));
		g_simple_async_result_set_from_error(simple, error);
		g_clear_error(&error);
	}

	g_simple_async_result_complete(simple);
	g_object_unref(simple);
}

void hev_impathy_keyring_set_account_password_async(TpAccount *account,
			const gchar *password, GAsyncReadyCallback callback, gpointer user_data)
{
	GSimpleAsyncResult *simple = NULL;
	const gchar *account_id = NULL;
	gchar *name = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(TP_IS_ACCOUNT(account));
	g_return_if_fail(NULL!=password);

	simple = g_simple_async_result_new(G_OBJECT(account), callback, user_data,
				hev_impathy_keyring_set_account_password_async);

	account_id = tp_proxy_get_object_path(account) +
		strlen(TP_ACCOUNT_OBJECT_PATH_BASE);

	name = g_strdup_printf("IM account password for %s (%s)",
				tp_account_get_display_name(account), account_id);

	gnome_keyring_store_password(&account_keyring_schema, NULL,
				name, password, store_password_handler, simple, NULL,
				"account-id", account_id, "param-name", "password",
				NULL);

	g_free(name);
}

gboolean hev_impathy_keyring_set_account_password_finish(TpAccount *account,
			GAsyncResult *res, GError **error)
{
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);
	
	if(g_simple_async_result_propagate_error(G_SIMPLE_ASYNC_RESULT(res),
					error))
	  return FALSE;

	return g_simple_async_result_is_valid(res, G_OBJECT(account),
				hev_impathy_keyring_set_account_password_async);
}

static void find_items_handler(GnomeKeyringResult result,
			GList *list, gpointer user_data)
{
	GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT(user_data);
	GnomeKeyringFound *found = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(GNOME_KEYRING_RESULT_OK != result)
	{
		GError *error = g_error_new_literal(TP_ERROR,
					TP_ERROR_DOES_NOT_EXIST,
					gnome_keyring_result_to_message(result));
		g_simple_async_result_set_from_error(simple, error);
		g_clear_error(&error);
		goto out;
	}

	if(NULL == list)
	{
		g_simple_async_result_set_error(simple, TP_ERROR,
					TP_ERROR_DOES_NOT_EXIST,
					"Password not found");
		goto out;
	}

	/* Get the first password returned. Ideally we should use the latest
	 * modified or something but we don't have this information from
	 * gnome-keyring atm. */
	found = list->data;

	g_simple_async_result_set_op_res_gpointer(simple,
				g_strdup(found->secret), g_free);
out:
	g_simple_async_result_complete(simple);
	g_object_unref(simple);
}

void hev_impathy_keyring_get_account_password_async(TpAccount *account,
			GAsyncReadyCallback callback, gpointer user_data)
{
	GSimpleAsyncResult *simple = NULL;
	GnomeKeyringAttributeList *match = NULL;
	const gchar *account_id = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(TP_IS_ACCOUNT(account));
	g_return_if_fail(NULL!=callback);

	simple = g_simple_async_result_new(G_OBJECT(account),
				callback, user_data,
				hev_impathy_keyring_get_account_password_async);

	account_id = tp_proxy_get_object_path(account) +
		strlen(TP_ACCOUNT_OBJECT_PATH_BASE);

	match = gnome_keyring_attribute_list_new();
	gnome_keyring_attribute_list_append_string(match, "account-id",
				account_id);
	gnome_keyring_attribute_list_append_string(match, "param-name",
				"password");

	gnome_keyring_find_items(GNOME_KEYRING_ITEM_GENERIC_SECRET,
				match, find_items_handler, simple, NULL);

	gnome_keyring_attribute_list_free(match);
}

const gchar * hev_impathy_keyring_get_account_password_finish(TpAccount *account,
			GAsyncResult *res, GError **error)
{
	GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT(res);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(g_simple_async_result_propagate_error(simple, error))
	  return NULL;
	
	g_return_val_if_fail(g_simple_async_result_is_valid(res,
					G_OBJECT(account),
					hev_impathy_keyring_get_account_password_async),
				NULL);
	return (const gchar *)g_simple_async_result_get_op_res_gpointer(simple);
}

static void item_delete_handler(GnomeKeyringResult result,
			gpointer user_data)
{
	GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT(user_data);

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(GNOME_KEYRING_RESULT_OK != result)
	{
		GError *error = g_error_new_literal(TP_ERROR,
					TP_ERROR_DOES_NOT_EXIST,
		gnome_keyring_result_to_message(result));
		g_simple_async_result_set_from_error(simple, error);
		g_clear_error(&error);
	}

	g_simple_async_result_complete(simple);
	g_object_unref(simple);
}

static void find_item_to_delete_handler(GnomeKeyringResult result,
			GList *list, gpointer user_data)
{
	GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT(user_data);
	GnomeKeyringFound *found = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if((GNOME_KEYRING_RESULT_OK!=result) ||
				(1!=g_list_length(list)))
	{
		GError *error = NULL;
		
		error = g_error_new_literal(TP_ERROR,
					TP_ERROR_DOES_NOT_EXIST,
					gnome_keyring_result_to_message(result));
		g_simple_async_result_set_from_error(simple, error);
		g_clear_error(&error);

		g_simple_async_result_complete(simple);
		g_object_unref(simple);

		return;
	}

	found = list->data;

	gnome_keyring_item_delete(NULL, found->item_id,
				item_delete_handler, simple, NULL);
}

void hev_impathy_keyring_delete_account_password_async(TpAccount *account,
			GAsyncReadyCallback callback, gpointer user_data)
{
	GSimpleAsyncResult *simple = NULL;
	GnomeKeyringAttributeList *match = NULL;
	const gchar *account_id = NULL;

	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	g_return_if_fail(TP_IS_ACCOUNT(account));

	simple = g_simple_async_result_new(G_OBJECT(account),
				callback, user_data,
				hev_impathy_keyring_delete_account_password_async);

	account_id = tp_proxy_get_object_path(account) +
		strlen(TP_ACCOUNT_OBJECT_PATH_BASE);

	match = gnome_keyring_attribute_list_new();
	gnome_keyring_attribute_list_append_string(match, "account-id",
				account_id);
	gnome_keyring_attribute_list_append_string(match, "param-name",
				"password");

	gnome_keyring_find_items(GNOME_KEYRING_ITEM_GENERIC_SECRET,
				match, find_item_to_delete_handler, simple, NULL);

	gnome_keyring_attribute_list_free(match);
}

gboolean hev_impathy_keyring_delete_account_password_finish(TpAccount *account,
			GAsyncResult *res, GError **error)
{
	g_debug("%s:%d[%s]", __FILE__, __LINE__, __FUNCTION__);

	if(g_simple_async_result_propagate_error(G_SIMPLE_ASYNC_RESULT(res),
					error))
	  return FALSE;

	g_return_val_if_fail(g_simple_async_result_is_valid(res,
					G_OBJECT(account),
					hev_impathy_keyring_delete_account_password_async),
				FALSE);

	return TRUE;
}

