/*
Copyright (c) 2012, Ilya Arefiev <arefiev.id@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.
 * Neither the name of the author nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LOG_H
#define LOG_H

#include <common.h>

/** log verbose level.
 *
 * When specific log level is set, messages with this log
 * level and all following would be logged. Otherwise they
 * would be ignored
 */
typedef enum {
	LOGLEVEL_DEBUG,					/**< most verbose level */
	LOGLEVEL_LOG,					/**< normal log level */
	LOGLEVEL_WARNING,				/**< just warnings or notice */
	LOGLEVEL_ERROR,					/**< serious error occured */

	/* -------------- */
	LOGLEVEL_COUNT					/**< last element. actual log levels should be defined above */
} LOGLEVEL;

/** structure to keep log file and level together.
 *
 * just for convinience
 */
typedef struct LOGOBJ_STRUCT {
	FILE* logfile;					/**< destination log file */
	LOGLEVEL loglevel;				/**< defined log verbose level */
} LOGOBJ;

/** global log object
 *
 * just for convinience
 */
LOGOBJ logobj;

bool l_init(const char*, const LOGLEVEL);
bool l_deinit();

bool l_printf(const LOGLEVEL, const char*, ...);

/** alias for #l_printf
 *
 * calls #l_printf with passed log log level and format prepended
 * with __FILE__, __LINE__ and __func__ macroses
 *
 * @param level log level (@see #LOGLEVEL)
 * @param fmt log message format
 * @param ... undefined number of variables for logging
 *
 * @return same as #l_printf
 */
#define L_PRINTF(level, fmt, ...)\
	l_printf( level, "%s:%d %s: " fmt , \
	__FILE__, __LINE__, __func__, __VA_ARGS__ )

/** alias for #L_PRINTF macros
 *
 * calls #L_PRINTF with #LOGLEVEL_DEBUG
 */
#define L_DEBUG(fmt, ...)\
	L_PRINTF( LOGLEVEL_DEBUG, fmt, __VA_ARGS__ )
/** alias for #L_PRINTF macros
 *
 * calls #L_PRINTF with #LOGLEVEL_LOG
 */
#define L_LOG(fmt, ...)\
	L_PRINTF( LOGLEVEL_LOG, fmt, __VA_ARGS__ )
/** alias for #L_PRINTF
 *
 * calls #L_PRINTF with #LOGLEVEL_WARNING
 */
#define L_WARN(fmt, ...)\
	L_PRINTF( LOGLEVEL_WARNING, fmt, __VA_ARGS__ )
/** alias for #L_PRINTF
 *
 * calls #L_PRINTF with #LOGLEVEL_ERROR
 */
#define L_ERROR(fmt, ...)\
	L_PRINTF( LOGLEVEL_ERROR, fmt, __VA_ARGS__ )

#endif
