/*
Copyright (c) 2012, Ilya Arefiev <arefiev.id@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.
 * Neither the name of the author nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "data.h"

#include "common.h"
#include "utils.h"
#include "log.h"

/** initializes #LIST_ROOT.
 *
 * currently only zeroes passed pointer to the #LIST_ROOT
 * so it is not necessary zero it before the call
 *
 * @param _root pointer to the root node
 *
 * @return currently #true in case of non-null pointer \n
 *			#false otherwise
 */
bool d_list_init(LIST_ROOT* _root)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	memset( _root, 0, sizeof(LIST_ROOT) );

	return true;
}

/** deinitializes #LIST_ROOT.
 *
 * removes all nodes in the list (by calling #d_list_remove_head)
 * and zeroes passed pointer to the #LIST_ROOT
 *
 * logs warning if #d_list_remove_head was not successfully done
 *
 * @param _root pointer to the root node
 *
 * @return currently #true in case of non-null pointer \n
 *			#false otherwise
 */
bool d_list_deinit(LIST_ROOT* _root)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	bool clr = true;
	while ( _root->count )
	{
		clr &= d_list_remove_head( _root );
	}

	if ( !clr )
	{
		L_WARN( "memleaks are possible\n", "" );
	}

	memset( _root, 0, sizeof(_root) );

	return true;
}

/** inserts data into the list.
 *
 * searches the position in the list,
 * allocates memory for the new node and data (according passed _size),
 * links previous node with new one,
 * links new node with the next node,
 * updates LIST_ROOT#head and LIST_ROOT#tail pointers if necessary
 *
 * if passed _position is greater than list count (LIST_ROOT#count)
 * than tail-insert performed. \n
 * if passed _position is lesser than list count (LIST_ROOT#count)
 * than normal insert performed. \n
 * if passed _position is negative than head-insert performed.
 *
 * if list already contains 2 nodes and insert into _position=1
 * requested, then current tail will be replaced with the new node,
 * current tail will be new tail.
 *
 * @param _root root node of the list
 * @param _position position where to insert new node (zero-based)
 * @param _data data to insert into new node
 * @param _size size of data to be inserted
 *
 * @return #true if insert was successful \n
 *			#false otherwise (see logs for details in this case)
 */
bool d_list_insert(LIST_ROOT* _root, const int _position, const void* _data, const size_t _size)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x %d\n", _root, _position ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	int position = _position;

	if ( position < 0 )
	{
		L_WARN( "_position is NEGATIVE(%d). insert HEAD\n",
			position
			);

		position = 0;
	}
	else if ( position > _root->count )
	{
		L_WARN( "_position(%d) is GREATER than count(%d). insert tail\n",
			position, _root->count
			);

		position = _root->count;
	}

	LIST_NODE* current = _root->head;
	LIST_NODE* prev = current;

	/* L_DEBUG( "current = 0x%08x; position = %d\n", */
	/* 	current, position */
	/* 	); */

	if ( position != _root->count )	// insert somewhere inside list
	{
		while ( current && position )
		{
			prev = current;
			current = current->next;

			position--;

			/* L_DEBUG( "in-cycle: current = 0x%08x; prev = 0x%08x; pos = %d\n", */
			/* 	current, prev, position */
			/* 	); */
		}
	}
	else
	{
		current = NULL;
		prev = _root->tail;
		position = 0;

		/* L_DEBUG( "insert new tail. skip cycle and goto end of the list\n", "" ); */
	}

	if ( position )
	{
		/* L_ERROR( "broken list? prev 0x%08x; current 0x%08x; position %d\n", */
		/* 	prev, current, position */
		/* 	); */

		return false;
	}

	LIST_NODE* node = (LIST_NODE*)calloc( 1, sizeof(LIST_NODE) );
	if ( !node )
	{
		L_ERROR( "unable to allocate node\n", "" );
		return false;
	}

	void* data = calloc( 1, _size );
	if ( !data )
	{
		L_ERROR( "unable to allocate data\n", "" );
		return false;
	}

	/* L_DEBUG( "new node 0x%08x with data 0x%08x\n", */
	/* 	node, data */
	/* 	); */

	memcpy( data, _data, _size );
	node->data = data;
	node->sz = _size;
	node->next = current;
	if ( prev == current )	// in case of new head
	{
		/* L_DEBUG( "new head inserted\n", "" ); */

		_root->head = node;
	}
	else
	{
		/* L_DEBUG( "new node inserted\n", "" ); */

		prev->next = node;
	}

	/* L_DEBUG( "p(0x%08x)->0x%08x; n(0x%08x)->0x%08x; c(0x%08x)->0x%08x\n", */
	/* 	prev, ( prev ? prev->next : NULL ), */
	/* 	node, node->next, */
	/* 	current, ( current ? current->next : NULL ) */
	/* 	); */

	if ( _position == _root->count )	// in case of new tail
	{
		/* L_DEBUG( "new tail inserted\n", "" ); */

		_root->tail = node;
	}

	/* L_DEBUG( "node(0x%08x)->0x%08x\n", */
	/* 	node, node->next */
	/* 	); */

	_root->count++;
	/* L_DEBUG( "new count %d\n", _root->count ); */
	
	return true;
}

/** inserts new head.
 *
 * performs call to #d_list_insert with _position=0
 *
 * @param _root root node of the list
 * @param _data data to insert into new node
 * @param _size size of data to be inserted
 * 
 * @return #true if insert was successful \n
 *			#false otherwise (see logs for details in this case)
 *
 * @see d_list_insert
 */
bool d_list_insert_head(LIST_ROOT* _root, const void* _data, const size_t _size)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}
	
	return d_list_insert(_root, 0, _data, _size);
}

/** inserts new tail.
 *
 * performs call to #d_list_insert with _position = LIST_ROOT#count
 *
 * @param _root root node of the list
 * @param _data data to insert into new node
 * @param _size size of data to be inserted
 * 
 * @return #true if insert was successful \n
 *			#false otherwise (see logs for details in this case)
 *
 * @see d_list_insert
 */
bool d_list_insert_tail(LIST_ROOT* _root, const void* _data, const size_t _size)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	return d_list_insert(_root, _root->count, _data, _size);
}

/** removes node of the list.
 *
 * searches the position in the list,
 * deallocates memory with data,
 * deallocates memory with node,
 * links previous node with next
 * updates LIST_ROOT#head and LIST_ROOT#tail pointers if necessary
 *
 * if passed _position is greater than list count (LIST_ROOT#count)
 * or negative than nothing removed and #false returned.
 *
 * @param _root root node of the list
 * @param _position number of node to remove (zero-based)
 *
 * @return #true if removed successfully \n
 *			#false otherwise (see logs for details in this case)
 */
bool d_list_remove(LIST_ROOT* _root, const int _position)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x %d\n", _root, _position ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	if ( !(_root->count) )
	{
		L_WARN( "nothing to remove. list is empty\n", "" );

		return true;
	}

	if ( _position < 0 || _position >= _root->count )
	{
		L_ERROR( "incorrect _position: %d. list contains %d\n",
			_position, _root->count
			);

		return false;
	}

	int position = _position;

	LIST_NODE* current = _root->head;
	LIST_NODE* prev = current;

	/* L_DEBUG( "current = 0x%08x; position = %d\n", */
	/* 	current, position */
	/* 	); */

	if ( _position )
	{
		while ( current && position )
		{
			if ( current->next )
			{
				prev = current;
				current = current->next;
			}
			else
			{
				break;
			}

			position--;

			/* L_DEBUG( "in-cycle: current = 0x%08x; prev = 0x%08x; pos = %d\n", */
			/* 	current, prev, position */
			/* 	); */
		}

		if ( position )
		{
			L_ERROR( "broken list? prev 0x%08x; current 0x%08x; position %d\n",
				prev, current, position
				);

			return false;
		}

		prev->next = current->next;

		if ( !(prev->next) )	// in case of new tail
		{
			_root->tail = prev;

			/* L_DEBUG( "new tail 0x%08x\n", _root->tail ); */
		}
	}
	else	// in case of rm head
	{
		_root->head = current->next;

		if ( !_root->head )
		{
			_root->tail = NULL;

			/* L_DEBUG( "tail = head = NULL\n", "" ); */
		}
	
		/* L_DEBUG( "new head 0x%08x\n", _root->head ); */
	}

	if ( current )
	{
		if ( current->data )
		{
			free( current->data );
		}
		free( current );

		_root->count--;
		/* L_DEBUG( "new count %d\n", _root->count ); */
	}

	return true;
}

/** removes head of the list.
 *
 * performs call of #d_list_remove with _position=0
 *
 * @param _root root node of the list
 *
 * @return #true if removed successfully \n
 *			#false otherwise (see logs for details in this case)
 *
 * @see d_list_remove
 */
bool d_list_remove_head(LIST_ROOT* _root)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	return d_list_remove(_root, 0);
}

/** removes tail of the list.
 *
 * performs call of #d_list_remove with _position = LIST_ROOT#count
 *
 * @param _root root node of the list
 *
 * @return #true if removed successfully \n
 *			#false otherwise (see logs for details in this case)
 *
 * @see d_list_remove
 */
bool d_list_remove_tail(LIST_ROOT* _root)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	return d_list_remove(_root, _root->count-1);
}

/** get data in the node with position.
 *
 * searches the position in the list,
 * copies data in the node into passed buffer
 * if there is enough space.
 *
 * if passed position is greater than LIST_ROOT#count or negative
 * then nothing is copied and #false returned. \n
 * if passed buffer size is lesser than requested data
 * then nothing copied, error logged and #false returned. \n
 * if passed buffer size is greater then requested data
 * then data is copied and warning raised.
 *
 * NOTE: passed buffer filled with NULLs before copying requested data.
 *
 * @param _root root node of the list
 * @param _position number of node which data to get (zero-based)
 * @param _data pointer to the buffer in which to copy node's data
 * @param _size size of _data buffer
 *
 * @return #true if data in the node fits into passed _data \n
 *			#false otherwise or any argument is bad
 */
bool d_list_get(LIST_ROOT* _root, const int _position, void* _data, size_t _size)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x %d\n", _root, _position ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	if ( !_root->count )
	{
		L_WARN( "nothing to get. list is empty\n", "" );

		return false;
	}

	if ( _position < 0 || _position >= _root->count )
	{
		L_ERROR( "incorrect _position: %d. list size: %d\n",
			_position, _root->count
			);

		return false;
	}

	if ( !_data )
	{
		L_ERROR( "_data == NULL\n", "" );

		return false;
	}

	int position = _position;

	/* TODO: rewrite using head and tail pointers */
	LIST_NODE* current = _root->head;
	LIST_NODE* prev = current;

	/* L_DEBUG( "current = 0x%08x; pos = %d\n", */
	/* 	current, position */
	/* 	); */

	while ( current && position )
	{
		if ( current->next )
		{
			prev = current;
			current = current->next;
		}
		else
		{
			break;
		}

		position--;

		/* L_DEBUG( "in-cycle: current = 0x%08x; prev = 0x%08x; pos = %d\n", */
		/* 	current, prev, position */
		/* 	); */
	}

	if ( !current )
	{
		L_ERROR( "broken list? prev 0x%08x; current 0x%08x; position %d\n",
			prev, current, position
			);

		return false;
	}
	else
	{
		if ( current->sz > _size )
		{
			L_ERROR( "not enough space(%d) for data(%d) passed\n",
				_size, current->sz
				);
			
			return false;
		}
		else if ( current->sz < _size )
		{
			L_WARN( "current->sz(%d) lesser than _size(%d)\n",
				current->sz, _size
				);
		}

		memset( _data, 0, _size);
		memcpy( _data, current->data, current->sz );
	}

	return true;
}

/** get data in the head of the list.
 *
 * performs call of #d_list_get with position = 0
 *
 * @param _root root node of the list
 * @param _data pointer to the destination buffer
 * @param _size size of destination buffer
 *
 * @return #true if data in the node fits into passed _data \n
 *			#false otherwise or any argument is bad
 *
 * @see d_list_get
 */
bool d_list_get_head(LIST_ROOT* _root, void* _data, size_t _size)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	return d_list_get(_root, 0, _data, _size);
}

/** get data in the tail of the list.
 *
 * performs call of #d_list_get with position = LIST_ROOT#count -1
 *
 * @param _root root node of the list
 * @param _data pointer to the destination buffer
 * @param _size size of destination buffer
 *
 * @return #true if data in the node fits into passed _data \n
 *			#false otherwise or any argument is bad
 *
 * @see d_list_get
 */
bool d_list_get_tail(LIST_ROOT* _root, void* _data, size_t _size)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	return d_list_get(_root, _root->count-1, _data, _size);
}

/** initializes #QUEUE_ROOT.
 *
 * calls #d_list_init
 *
 * @param _root root node of the queue
 *
 * @return #true if initialized successfully \n
 *			#false otherwise
 *
 *	@see d_list_init
 */
bool d_queue_init(QUEUE_ROOT* _root)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	return d_list_init((LIST_ROOT*)_root);
}

/** deinitializes #QUEUE_ROOT.
 *
 * calls #d_list_deinit
 *
 * @param _root root node of the queue
 *
 * @return #true if deinitialized successfully \n
 *			#false otherwise
 *
 * @see d_list_deinit
 */
bool d_queue_deinit(QUEUE_ROOT* _root)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	return d_list_deinit((LIST_ROOT*)_root);
}

/** pushes data into the tail of the queue.
 *
 * calls #d_list_insert_tail
 *
 * @param _root root node of the queue
 * @param _data data to push
 * @param _size size of data to push
 *
 * @return #true if pushed successfully \n
 *			#false otherwise
 *
 * @see d_list_insert_tail
 */
bool d_queue_push(QUEUE_ROOT* _root, const void* _data, const size_t _size)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	return d_list_insert_tail((LIST_ROOT*)_root, _data, _size);
}

/** copies data from the head of the queue into passed buffer.
 *
 * if data from the head of the queue copied successfully
 * removes head of the queue. \n
 * if head of the queue was not removed successfully
 * logs warning.
 *
 * @param _root root node of the queue
 * @param _data buffer in which head's data would be copied
 * @param _size passed buffer size
 *
 * @return #true if successfully poped \n
 *			#false otherwise
 *
 *	@see d_list_get_head
 *	@see d_list_remove_head
 */
bool d_queue_pop(QUEUE_ROOT* _root, void* _data, size_t _size)
{
/* #ifdef _DEBUG */
/* 	L_DEBUG( "0x%08x\n", _root ); */
/* #endif */
	
	if ( !_root )
	{
		L_ERROR( "_root == NULL\n", "" );

		return false;
	}

	if ( d_list_get_head((LIST_ROOT*)_root, _data, _size) )
	{
		if ( !d_list_remove_head((LIST_ROOT*)_root) )
		{
			// pointer moved anyway
			// just let developers know that here is memleaks
			L_WARN( "unable to remove QUEUE_HEAD\n", "" );
		}

		return true;
	}

	return false;
}
