/*
Copyright (c) 2012, Ilya Arefiev <arefiev.id@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.
 * Neither the name of the author nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "datetime.h"

#include "common.h"

/* NOTE: we can not use l_printf here due to l_printf uses this functions to get timestamps */

/** fills _buf with passed _timestamp in text representation.
 *
 * destination _buf filling depends on passed _format.
 * formats (with minimum sizes of passed _buf) are: \n
 * <pre>
 * "YYYY-mm-dd HH:MM:SS.ssssss[ +ZZZZ]\0" \n
 * DT_FORMAT#DT_FORMAT_ALL		4+1+2+1+2 +1+ 2+1+2+1+2 +1+6[+1+5] +1 =	27 [33] \n
 * DT_FORMAT#DT_FORMAT_TIME		2+1+2+1+2 +1+6[+1+5] +1 =				16 [22] \n
 * DT_FORMAT#DT_FORMAT_DATE		4+1+2+1+2 [+1+5] +1 =					11 [17] \n
 * </pre>
 * 
 * NOTE: zeroes _buf before filling with date.
 *
 * @param _buf destination buffer
 * @param _bufsz destination buffer size
 * @param _timestamp source time structure
 * @param _local flag for local or GMT time
 * @param _format format of _buf filling
 *
 * @return #true if _buf filled successfully \n
 *			#false otherwise
 */
bool dt_datetime_sprintf(char* _buf, const size_t _bufsz, const struct timeval* _timestamp, const bool _local, const DT_FORMAT _format)
{
/* #ifdef _DEBUG */
/* 	printf("== %s:%d %s (0x%08x, %d, 0x%08x(%ld, %ld), %C, %d)\n", */
/* 	     __FILE__, __LINE__, __func__, */
/* 		 (unsigned int)_buf, _bufsz, */
/* 		 (unsigned int)_timestamp, _timestamp->tv_sec, _timestamp->tv_usec, */
/* 		 ( _local ? 'T' : 'F' ), _format */
/* 	     ); */
/* #endif */

	if ( !_buf )
	{
		fprintf( stderr, "-- _buf == NULL\n" );

		return false;
	}

	if ( _format < 0 || _format >= DT_FORMAT_COUNT )
	{
		fprintf( stderr, "-- incorrect _format(%d)\n",
			_format );

		return false;
	}

	/* "YYYY-mm-dd HH:MM:SS.ssssss[ +ZZZZ]\0" */
	/* date + time	4+1+2+1+2 +1+ 2+1+2+1+2 +1+6[+1+5] +1 =	27 [33] */
	/* timeOnly		2+1+2+1+2 +1+6[+1+5] +1 =				16 [22] */
	/* dateOnly		4+1+2+1+2 [+1+5] +1 =					11 [17] */
	if ( _bufsz <= 0 || (
		( _format == DT_FORMAT_ALL && _bufsz < 27 ) ||
		( _format == DT_FORMAT_ALL && _bufsz < 33 && _local ) ||
		( _format == DT_FORMAT_TIME && _bufsz < 16 ) ||
		( _format == DT_FORMAT_TIME && _bufsz < 22 && _local ) ||
		( _format == DT_FORMAT_DATE && _bufsz < 11 ) ||
		( _format == DT_FORMAT_DATE && _bufsz < 17 && _local )
		) )
	{
		fprintf( stderr, "-- incorrect _bufsz(%d)\n",
			_bufsz );

		return false;
	}

	if ( !_timestamp )
	{
		fprintf( stderr, "-- _timestamp == NULL\n" );

		return false;
	}

	struct tm *ptime;
	if ( _local )
	{
		ptime = localtime( &(_timestamp->tv_sec) );
	}
	else
	{
		ptime = gmtime( &(_timestamp->tv_sec) );
	}

	memset( _buf, 0, sizeof(_buf) );

	if ( _timestamp )
	{
		size_t ret;

		switch ( _format )
		{
			case DT_FORMAT_ALL:
			{
				ret = strftime( _buf, BUFSZ, "%Y-%m-%d %H:%M:%S", ptime );
				break;
			}
			case DT_FORMAT_TIME:
			{
				ret = strftime( _buf, BUFSZ, "%H:%M:%S", ptime );
				break;
			}
			case DT_FORMAT_DATE:
			{
				ret = strftime( _buf, BUFSZ, "%Y-%m-%d", ptime );
				break;
			}
			default:
			{
				/* should not ever happen */
				fprintf( stderr, "-- something went wrong. _format = %d\n",
					_format );

				return false;
			}
		};

		if ( ret > 0 )
		{
			if ( ( _format == DT_FORMAT_ALL ) ||
				( _format == DT_FORMAT_TIME )
				)
			{
				snprintf( _buf+ret, _bufsz-ret, ".%06ld", _timestamp->tv_usec );
			}

			if ( _local )
			{
				ret = strftime( _buf + strlen(_buf), BUFSZ-strlen(_buf), " %z", ptime );
			}
			else
			{
				return true;
			}

			if ( ret > 0 )
			{
				return true;
			}
			else
			{
				fprintf( stderr, "-- strftime error %d: %s\n",
					errno, strerror(errno) );
			}
		}
		else
		{
			fprintf( stderr, "-- strftime error %d: %s\n",
				errno, strerror(errno) );
		}
	}
	else
	{
		fprintf( stderr, "-- localtime error %d: %s\n",
			errno, strerror(errno) );
	}

	return false;
}

/** fills _buf with date in passed _timestamp.
 *
 * calls #dt_datetime_sprintf with _format = DT_FORMAT#DT_FORMAT_DATE
 *
 * @param _buf destination buffer
 * @param _bufsz destination buffer size
 * @param _tv source time structure
 * @param _local flag for local or GMT time
 *
 * @return #true if _buf filled successfully \n
 *			#false otherwise
 *
 * @see dt_datetime_sprintf
 */
bool dt_date_sprintf(char* _buf, const size_t _bufsz, const struct timeval* _tv, const bool _local)
{
/* #ifdef _DEBUG */
/* 	printf("== %s:%d %s (0x%08x, %d, 0x%08x(%ld, %ld), %C)\n", */
/* 	     __FILE__, __LINE__, __func__, */
/* 		 (unsigned int)_buf, _bufsz, */
/* 		 (unsigned int)_tv, _tv->tv_sec, _tv->tv_usec, */
/* 		 ( _local ? 'T' : 'F' ) */
/* 	     ); */
/* #endif */

	if ( !_buf || _bufsz <= 0 )
	{
		/* NOTE: dt_datetime_sprintf will make more checks for us */
		return false;
	}

	return dt_datetime_sprintf( _buf, _bufsz, _tv, _local, DT_FORMAT_DATE );
}

/** fills _buf with time in passed _timestamp.
 *
 * calls #dt_datetime_sprintf with _format = DT_FORMAT#DT_FORMAT_TIME
 *
 * @param _buf destination buffer
 * @param _bufsz destination buffer size
 * @param _tv source time structure
 * @param _local flag for local or GMT time
 *
 * @return #true if _buf filled successfully \n
 *			#false otherwise
 *
 * @see dt_datetime_sprintf
 */
bool dt_time_sprintf(char* _buf, const size_t _bufsz, const struct timeval* _tv, const bool _local)
{
/* #ifdef _DEBUG */
/* 	printf("== %s:%d %s (0x%08x, %d, 0x%08x(%ld, %ld), %C)\n", */
/* 	     __FILE__, __LINE__, __func__, */
/* 		 (unsigned int)_buf, _bufsz, */
/* 		 (unsigned int)_tv, _tv->tv_sec, _tv->tv_usec, */
/* 		 ( _local ? 'T' : 'F' ) */
/* 	     ); */
/* #endif */

	if ( !_buf || _bufsz <= 0 )
	{
		/* NOTE: dt_datetime_sprintf will make more checks for us */
		return false;
	}

	return dt_datetime_sprintf( _buf, _bufsz, _tv, _local, DT_FORMAT_TIME );
}

/** fills _buf with current date and time in text representation.
 *
 * retrieves current time into tv structre and
 * calls #dt_datetime_sprintf with tv structure.
 *
 * @param _buf destination buffer
 * @param _bufsz destination buffer size
 * @param _local flag for local or GMT time
 * @param _format format of _buf filling
 *
 * @return #true if _buf filled successfully \n
 *			#false otherwise
 *
 * @see dt_datetime_sprintf
 */
bool dt_datetime_sprintf_current(char* _buf, const size_t _bufsz, const bool _local, const DT_FORMAT _format)
{
/* #ifdef _DEBUG */
/* 	printf("== %s:%d %s (0x%08x, %d, %C, %d)\n", */
/* 	     __FILE__, __LINE__, __func__, */
/* 		 (unsigned int)_buf, _bufsz, */
/* 		 ( _local ? 'T' : 'F' ), _format */
/* 	     ); */
/* #endif */

	if ( !_buf || _bufsz <= 0 )
	{
		/* NOTE: dt_datetime_sprintf will make more checks for us */
		return false;
	}

	struct timeval tv;
	struct timezone tz;
	memset( &tv, 0, sizeof(tv) );
	memset( &tz, 0, sizeof(tz) );

	if ( gettimeofday( &tv, &tz ) !=0 )
	{
		fprintf( stderr, "-- error while gettimeofday %d: %s\n",
			errno, strerror(errno) );

		return false;
	}
	
	return dt_datetime_sprintf( _buf, _bufsz, &tv, _local, _format );
}

/** fills _buf with current date in text representation.
 *
 * calls #dt_datetime_sprintf_current with _format = DT_FORMAT#DT_FORMAT_DATE
 *
 * @param _buf destination buffer
 * @param _bufsz destination buffer size
 * @param _local flag for local or GMT time
 *
 * @return #true if _buf filled successfully \n
 *			#false otherwise
 *
 * @see dt_datetime_sprintf_current
 */
bool dt_date_sprintf_current(char* _buf, const size_t _bufsz, const bool _local)
{
/* #ifdef _DEBUG */
/* 	printf("== %s:%d %s (0x%08x, %d, %C)\n", */
/* 	     __FILE__, __LINE__, __func__, */
/* 		 (unsigned int)_buf, _bufsz, */
/* 		 ( _local ? 'T' : 'F' ) */
/* 	     ); */
/* #endif */

	if ( !_buf || _bufsz <= 0 )
	{
		/* NOTE: dt_datetime_sprintf will make more checks for us */
		return false;
	}

	return dt_datetime_sprintf_current( _buf, _bufsz, _local, DT_FORMAT_DATE );
}

/** fills _buf with current time in text representation.
 *
 * calls #dt_datetime_sprintf_current with _format = DT_FORMAT#DT_FORMAT_TIME
 *
 * @param _buf destination buffer
 * @param _bufsz destination buffer size
 * @param _local flag for local or GMT time
 *
 * @return #true if _buf filled successfully \n
 *			#false otherwise
 *
 * @see dt_datetime_sprintf_current
 */
bool dt_time_sprintf_current(char* _buf, const size_t _bufsz, const bool _local)
{
/* #ifdef _DEBUG */
/* 	printf("== %s:%d %s (0x%08x, %d, %C)\n", */
/* 	     __FILE__, __LINE__, __func__, */
/* 		 (unsigned int)_buf, _bufsz, */
/* 		 ( _local ? 'T' : 'F' ) */
/* 	     ); */
/* #endif */

	if ( !_buf || _bufsz <= 0 )
	{
		/* NOTE: dt_datetime_sprintf will make more checks for us */
		return false;
	}

	return dt_datetime_sprintf_current( _buf, _bufsz, _local, DT_FORMAT_TIME );
}

