/*
Copyright (c) 2012, Ilya Arefiev <arefiev.id@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.
 * Neither the name of the author nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "log.h"
#include "common.h"
#include "utils.h"
#include "datetime.h"

/** string representation of #LOGLEVEL
 *
 * used in log file.
 */
char* LOGLEVEL_STRING[] = {
	"debug",
	"log  ",
	"WARN ",
	"ERROR",

	/* last element */
	"     "
};

/** initializes global #logobj.
 *
 * zeroes #logobj, opens log file with passed name in _filename,
 * sets maximum allowed #LOGLEVEL to write into log file and
 * inserts splitter in the end of file.
 *
 * NOTE: should be called once (since zeroes #logobj)
 * in start time and before any logging functionality use.
 * also #l_deinit should be called before since no logging
 * functionality needed anymore.
 *
 * @param _filename
 * @param _level
 *
 * @return #true if initialized successfully \n
 *			#false otherwise
 */
bool l_init(const char* _filename, const LOGLEVEL _level)
{
/* #ifdef _DEBUG */
/* 	printf("== %s:%d %s (0x%08x(%s), %d)\n", */
/* 	     __FILE__, __LINE__, __func__, */
/* 		 (unsigned int)_filename, _filename, _level */
/* 	     ); */
/* #endif */

	if ( !_filename )
	{
		fprintf( stderr, "-- _filename = NULL\n" );

		return false;
	}

	if ( _level < 0 || _level > LOGLEVEL_COUNT )
	{
		fprintf( stderr, "-- _level is out of range: %d (max %d)\n",
		_level, LOGLEVEL_COUNT );

		return false;
	}

	memset( &logobj, 0, sizeof(logobj) );

	logobj.logfile = fopen( _filename, "a" );
	if ( logobj.logfile == NULL )
	{
		fprintf( stderr, "-- unable to open logfile(%s) %d: %s\n",
			_filename, errno, strerror(errno) );

		return false;
	}

	if ( fseek( logobj.logfile, 0, SEEK_END ) == -1 )
	{
		fprintf( stderr, "-- error while fseek to the end of the logfile %d: %s\n",
			errno, strerror(errno) );
	}

	long logpos = ftell( logobj.logfile );
	if ( logpos != 0 )
	{
		fprintf( logobj.logfile, "\n" );
	}

	char tbuf[BUFSZ];
	memset( tbuf, 0, sizeof(tbuf) );

	if ( dt_date_sprintf_current( tbuf, sizeof(tbuf), false ) )
	{
		fprintf( logobj.logfile, "---- new entry: %s ----\n", tbuf );
	}

	logobj.loglevel = _level;

	return true;
}

/** deinitializes global #logobj.
 *
 * flushes and closes log file and
 * zeroes #logobj if closed successfully.
 *
 * @return #true if log file closed successfully \n
 *			#false otherwise
 */
bool l_deinit()
{
/* #ifdef _DEBUG */
/* 	printf("== %s:%d %s\n", */
/* 	     __FILE__, __LINE__, __func__ */
/* 	     ); */
/* #endif */

	if ( logobj.logfile == NULL )
	{
		fprintf( stderr, "-- logobj contains invalid logfile\n" );

		return false;
	}

	if ( fflush(logobj.logfile) == EOF )
	{
		fprintf( stderr, "-- error while fflush logfile %d: %s\n",
			errno, strerror(errno) );
	}

	if ( fclose(logobj.logfile) == EOF )
	{
		fprintf( stderr, "-- error while closing logfile %d: %s\n",
			errno, strerror(errno) );

		return false;
	}

	memset( &logobj, 0, sizeof(logobj) );

	return true;
}

/** writes entry into log file or stdout.
 *
 * checks allowed #LOGLEVEL (LOGOBJ#loglevel),
 * determines whether write message into file
 * or just into stdout, retrieves current GMT date,
 * writes formated message into log file or stdout.
 *
 * NOTE: write log message into stdout also in case of
 * debug version was built
 *
 * @param _level log message level
 * @param _fmt log message format
 * @param ... undefined number of parametres for message format
 *
 * @return currently #true since prints into stdout too in debug version
 */
bool l_printf(const LOGLEVEL _level, const char* _fmt, ...)
{
/* #ifdef _DEBUG */
/* 	printf("== %s:%d %s (%d, 0x%08x(%s))\n", */
/* 	     __FILE__, __LINE__, __func__, */
/* 		 _level, (unsigned int)_fmt, _fmt */
/* 	     ); */
/* #endif */

	bool tofile = true;

	if ( logobj.logfile == NULL )
	{
		tofile = false;
	}

	if ( _level >= logobj.loglevel )
	{
		/* FIXME: redundant gettimeofday? */
		struct timeval tv;
		struct timezone tz;
		memset( &tv, 0, sizeof(tv) );
		memset( &tz, 0, sizeof(tz) );

		char tbuf[BUFSZ];
		memset( tbuf, 0, sizeof(tbuf) );

		if ( gettimeofday( &tv, &tz ) == 0 )
		{
#ifdef _DEBUG
			if ( !
#endif
				dt_time_sprintf_current( tbuf, sizeof(tbuf), false )
#ifdef _RELEASE
				;
#else
				)
			{
				fprintf( stderr, "-- unable to get current time\n" );
			}
#endif
		}
#ifdef _DEBUG
		else
		{
			fprintf( stderr, "-- gettimeofday error %d: %s\n",
				errno, strerror(errno) );
		}
#endif

		// 1 for ' ' char betw. date && msg type
		if ( sizeof(tbuf) >= 1 + strlen(tbuf) + strlen( LOGLEVEL_STRING[LOGLEVEL_COUNT] ) )
		{
			snprintf( tbuf+strlen(tbuf), sizeof(tbuf)-(1+strlen(tbuf)), " %s", LOGLEVEL_STRING[_level] );
		}

		if ( tofile )
		{
			fprintf( logobj.logfile, "%s: ", tbuf );

#ifdef _DEBUG
			printf( "%s: ", tbuf );
#endif
		}
		else
		{
			printf( "%s: ", tbuf );
		}

		va_list arg;

		va_start( arg, _fmt );
		if ( tofile )
		{
			vfprintf( logobj.logfile, _fmt, arg );
#ifdef _DEBUG
			vprintf( _fmt, arg );
#endif
		}
		else
		{
			vprintf( _fmt, arg );
		}
		va_end( arg );

		if ( tofile )
		{
			if ( fflush( logobj.logfile ) == EOF )
			{
				fprintf( stderr, "-- error while fflush %d: %s\n",
					errno, strerror(errno) );
			}
		}
	}
#ifdef _DEBUG
	else
	{
		fprintf( stderr, "-- _level(%d) < required minimum(%d)\n",
			_level, logobj.loglevel );
	}
#endif

	return true;
}
