/*
Copyright (c) 2012, Ilya Arefiev <arefiev.id@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.
 * Neither the name of the author nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "net.h"
#include "common.h"
#include "log.h"
#include "utils.h"
#include "datetime.h"

/** initializes #NET_NODE.
 *
 * initializes #NET_NODE with IPv4 broadcast UDP socket.
 *
 * NOTE: zeroes _node right after parameters checking.
 *
 * @param _node structure to initialize
 * @param _port port number on which socket would be bind
 *
 * @return #true if initialized successfully \n
 *			#false otherwise
 */
bool net_init(NET_NODE* _node, const int _port)
{
#ifdef _DEBUG
	L_DEBUG( "(0x%08x, %d)\n",
		_node, _port
		);
#endif
	
	if ( !_node )
	{
		L_ERROR( "_node = NULL\n","" );

		return false;
	}

	if ( _port < 1024 || _port > 65535 )
	{
		L_ERROR( "incorrect port: %d\n",
			_port
			);

		return false;
	}

	memset( (void*)_node, 0, sizeof(_node) );
	
	_node->sock = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
	if ( _node->sock < 0 )
	{
		L_ERROR( "create socket error %d: %s\n",
			errno, strerror(errno)
			);

		return false;
	}

	int on = 1;
	if ( setsockopt( _node->sock, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on) ) < 0 )
	{
		L_ERROR( "setsockopt error %d: %s\n",
			errno, strerror(errno)
			);

		return false;
	}

	_node->sadr.sin_family = AF_INET;
	_node->sadr.sin_addr.s_addr = INADDR_BROADCAST;
	_node->sadr.sin_port = htons(_port);

	L_LOG( "create broadcast socket on port %d\n",
		_port
		);

	return true;
}

/** deinitilizes #NET_NODE.
 *
 * closes socket and zeroes _node structure.
 *
 * @param _node structure to deinitialize
 *
 * @return #true if socket close was successfully done \n
 *			#false otherwise
 */
bool net_deinit(const NET_NODE* _node)
{
#ifdef _DEBUG
	L_DEBUG( "(0x%08x)\n",
		_node
		);
#endif
	
	if ( !_node )
	{
		L_ERROR( "_node == NULL\n","" );

		return false;
	}

	if ( close( _node->sock ) < 0 )
	{
		L_ERROR( "socket close error %d: %s\n",
			errno, strerror(errno)
			);

		return false;
	}
	memset( (void*)_node, 0, sizeof(_node) );

	return true;
}

/** sends multicast message.
 *
 * calls #net_unicast with _dst = INADDR_BROADCAST and _dport from NET_NODE#sadr
 *
 * @param _node from which socket send a message
 * @param _msg message to send
 * @param _msgsz size of message to send
 *
 * @return #true if successfully sent \n
 *			#false otherwise
 *
 * @see net_unicast
 */
bool net_multicast(const NET_NODE* _node, const char* _msg, const size_t _msgsz)
{
#ifdef _DEBUG
	L_DEBUG( "(0x%08x, 0x%08x, %d)\n",
		_node, _msg, _msgsz
		);
#endif

	if ( !_node )
	{
		L_ERROR( "_node == NULL\n","" );

		return false;
	}

	struct in_addr bcast_addr;
	memset( &bcast_addr, 0, sizeof(bcast_addr) );
	bcast_addr.s_addr = INADDR_BROADCAST;

	/* using port to which socket was inited */
	/* TODO: is it correct? */
	return net_unicast( _node, &bcast_addr, ntohs(_node->sadr.sin_port), _msg, _msgsz);
}

/** sends unicast message.
 *
 * sends unicast message to the destination address and port
 * passed as parameters.
 *
 * NOTE: copies NET_NODE#sadr into local value, replaces
 * address and port with passed _dst and _dport and executes sendto()
 *
 * @param _node from which socket send a message
 * @param _dst destination address
 * @param _dport destination port (host byte order)
 * @param _msg message to send
 * @param _msgsz size of message to send
 *
 * @return #true if message was successfully sent \n
 *			#false otherwise
 */
bool net_unicast(const NET_NODE* _node, const struct in_addr* _dst, const int _dport, const char* _msg, const size_t _msgsz)
{
#ifdef _DEBUG
	L_DEBUG( "(0x%08x, 0x%08x, %d, 0x%08x, %d)\n",
		_node, _dst, _dport, _msg, _msgsz
		);

	u_hexout( (const unsigned char*)_msg, _msgsz );
#endif

	if ( !_node )
	{
		L_ERROR( "_node == NULL\n","" );

		return false;
	}
	
	if ( !_dst )
	{
		L_ERROR( "_dst == NULL\n","" );

		return false;
	}

	if ( _dport <=0 || _dport > 65535 )
	{
		L_ERROR( "incorrect destination port: %d\n",
			_dport
			);

		return false;
	}

	if ( !_msg )
	{
		L_ERROR( "_msg == NULL\n","" );

		return false;
	}

	if ( _msgsz <= 0 )
	{
		L_WARN( "incorrect _msgsz: %d\n",
			_msgsz
			);

		return false;
	}

	char dst_ips[BUFSZ];
	memset( dst_ips, 0, sizeof(dst_ips) );
	if ( !inet_ntop( AF_INET, _dst, dst_ips, BUFSZ ) )
	{
		sprintf( dst_ips, "(NULL)" );
	}

	L_DEBUG( "sendto %s:%d attempt\n",
		dst_ips, _dport
		);

	struct sockaddr_in dst;
	memset( &dst, 0, sizeof(dst) );

	memcpy( &dst, &(_node->sadr), sizeof(dst) );
	/* dst.sin_addr = _dst; */
	memcpy( &(dst.sin_addr), _dst, sizeof(_dst) );
	dst.sin_port = htons( _dport );

	size_t sent = sendto( _node->sock, _msg, _msgsz, 0,
						(struct sockaddr*)&dst, sizeof(dst) );
	
	if ( sent < 0 || sent != _msgsz )
	{
		L_ERROR( "sendto error %d: %s\n",
			errno, strerror(errno)
			);

		return false;
	}

	L_LOG( "sendto %s:%d success\n",
		dst_ips, _dport
		);

	return true;
}

