/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining 
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom 
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included 
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#ifndef HiME_ARCH_TYPES_H
#define HiME_ARCH_TYPES_H

#include "hime-config.h"
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#ifdef __cplusplus
extern "C"
{
#endif

/*
 * HiME type suffix notation:
 *
 * C types:
 * b		boolean				hime_bool_t
 * ub		unsigned byte			uint8_t
 * sb		signed byte			int8_t
 * us		unsigned short			uint16_t
 * ss		(signed) short			int16_t
 * i		32-bit int			int32_t
 * u		unsigned 32-bit int		uint32_t
 * ul		unsigned 64-bit int (long)	uint64_t
 * sl		(signed) 64-bit int (long)	int64_t
 * f		32-bit float			float
 * d		64-bit double float		double
 * c		8-bit char			char
 * wc		wide 16-bit char, wchar_t	wchar_t
 * s		UTF-8 string			char*
 * ws		UTF-16 string			wchar_t*
 * v						void*
 *
 * v + one of C scalar types: C array of that type:
 * 	vub	uint8_t[]
 * 	vc	char[]
 *
 * HiME types:
 * _str		string				hime_str_t*
 * _vec		vector				hime_vector_t*
 */

#define HiME_ARRAY_LENGTH(T) sizeof(T) / sizeof(T[0])

/* Types of guaranteed size */
typedef int8_t		hime_int8_t;
typedef uint8_t		hime_uint8_t;
typedef int16_t		hime_int16_t;
typedef uint16_t	hime_uint16_t;
typedef int32_t		hime_int32_t;
typedef uint32_t	hime_uint32_t;
typedef int64_t		hime_int64_t;
typedef uint64_t	hime_uint64_t;

typedef hime_int64_t	hime_lpos_t;
typedef hime_uint64_t	hime_lsize_t;
typedef hime_int64_t	hime_ltime_t;

typedef enum {
	HiME_False	= 0,
	HiME_True	= 1
} hime_bool_t;

#define HiME_MAX(x, y)   (((x) > (y)) ? (x) : (y))
#define HiME_MIN(x, y)   (((x) < (y)) ? (x) : (y))

#if defined(__LP64__)
    #define HiME_64BIT
#elif defined(_WIN64)
    #define HiME_64BIT
#else
    #define HiME_32BIT
#endif


/* Swap bytes in 16-bit word. */
#define HiME_SWAP_ENDIAN_16(A) \
		(hime_uint16_t)((((A) & 0xff00) >> 8) | (((A) & 0x00ff) << 8))

/* Swap bytes in 32-bit word. */
#define HiME_SWAP_ENDIAN_32(A) \
		(hime_uint32_t)( \
		(((A) & 0xff000000) >> 24) | \
		(((A) & 0x00ff0000) >>  8) | \
		(((A) & 0x0000ff00) <<  8) | \
		(((A) & 0x000000ff) << 24) )


/* Make big/litte-endian 16/32-bit words in platform independent way. */
#ifdef __BIG_ENDIAN__
	#define HiME_LITTLE_ENDIAN_16(A)	HiME_SWAP_ENDIAN_16(A)
	#define HiME_LITTLE_ENDIAN_32(A)	HiME_SWAP_ENDIAN_32(A)
	#define HiME_BIG_ENDIAN_16(A)		(hime_uint16_t)(A)
	#define HiME_BIG_ENDIAN_32(A)		(hime_uint32_t)(A)
#else
	#define HiME_LITTLE_ENDIAN_16(A)	(hime_uint16_t)(A)
	#define HiME_LITTLE_ENDIAN_32(A)	(hime_uint32_t)(A)
	#define HiME_BIG_ENDIAN_16(A)		HiME_SWAP_ENDIAN_16(A)
	#define HiME_BIG_ENDIAN_32(A)		HiME_SWAP_ENDIAN_32(A)
#endif

#ifdef __cplusplus
}
#endif

#endif /* HiME_ARCH_TYPES_H */
