/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#include "hime-win32.h"
#include "hime-arch-thread.h"
#include "hime-arch-memory.h"
#include "hime-win32-errors.h"

typedef struct
{
    void* arg;
    uint32_t (*proc)(void*);
} HiME_Win32ThreadData;


static DWORD WINAPI threadProc(LPVOID lpParameter)
{
    HiME_Win32ThreadData* p;
    void* arg;
    uint32_t (*proc)(void*);
    DWORD val;

    p = (HiME_Win32ThreadData*)lpParameter;
    arg = p->arg;
    proc = p->proc;
    HiME_SystemFree(p);
    val = (uint32_t)proc(arg);
    return val;
}


void HiME_CreateThread(HiME_Thread** t, uint32_t (*proc)(void*), void *arg)
{
    HiME_Win32ThreadData* dat;
    HANDLE h;

    HiME_ASSERT(t);

    *t = NULL;
    
    dat = (HiME_Win32ThreadData*)HiME_SystemAlloc(sizeof(HiME_Win32ThreadData));
    if (!dat)
        return;
    dat->arg = arg;
    dat->proc = proc;
    h = CreateThread(NULL, 0, threadProc, dat, CREATE_SUSPENDED, NULL);
    if (h)
    {
        *t = (HiME_Thread*)h;
        ResumeThread(h);
    }
}


void HiME_DeleteThread(HiME_Thread** t)
{
    HANDLE h;

    HiME_ASSERT(t);

    h = (HANDLE)*t;
    CloseHandle(h);
    *t = NULL;
}


void HiME_ExitThread(uint32_t val)
{
    ExitThread((DWORD)val);
}


uint32_t HiME_JoinThread(HiME_Thread* t)
{
    HANDLE h = (HANDLE)t;
    DWORD dw;

    WaitForSingleObject(h, INFINITE);
    GetExitCodeThread(h, &dw);

    return (uint32_t)dw;
}


HiME_ThreadID HiME_GetThreadId(HiME_Thread *t)
{
    int i = (HiME_ThreadID)GetThreadId((HANDLE)t);
    return i;
}


HiME_ThreadID HiME_GetCurrentThreadId()
{
    return (HiME_ThreadID)GetCurrentThreadId();
}


void HiME_SetThreadPriority(HiME_Thread *t, HiME_ThreadPriority pr)
{
    static int win32pr[] =
    {
        THREAD_PRIORITY_IDLE,
        THREAD_PRIORITY_LOWEST,
        THREAD_PRIORITY_NORMAL,
        THREAD_PRIORITY_HIGHEST
    };
            
    HANDLE h = (HANDLE)t;
    SetThreadPriority(h, win32pr[pr]);
}


void HiME_Sleep(uint32_t ms)
{
    Sleep(ms);
}


HiME_Mutex* HiME_CreateMutex()
{
    CRITICAL_SECTION* cs;
    
    cs = (CRITICAL_SECTION*)HiME_SystemAlloc(sizeof(CRITICAL_SECTION));
    InitializeCriticalSection(cs);
    return (HiME_Mutex*)cs;
}


void HiME_DeleteMutex(HiME_Mutex** mutex)
{
    CRITICAL_SECTION* cs;
    
    HiME_ASSERT(mutex);
    
    cs = (CRITICAL_SECTION*)*mutex;
    DeleteCriticalSection(cs);
    HiME_SystemFree(cs);
    *mutex = NULL;
}


void HiME_LockMutex(HiME_Mutex* mutex)
{
    CRITICAL_SECTION* cs = (CRITICAL_SECTION*)mutex;
    EnterCriticalSection(cs);
}


void HiME_UnlockMutex(HiME_Mutex* mutex)
{
    CRITICAL_SECTION* cs = (CRITICAL_SECTION*)mutex;
    LeaveCriticalSection(cs);
}


HiME_Semaphore* HiME_CreateSemaphore()
{
    HANDLE h = CreateSemaphore(NULL, 0, 1000, NULL);
    return (HiME_Semaphore*)h;
}


void HiME_DeleteSemaphore(HiME_Semaphore** sem)
{
    HANDLE h;

    HiME_ASSERT(sem);

    h = (HANDLE*)*sem;
    CloseHandle(h);
    *sem = NULL;
}


void HiME_SemaphoreDown(HiME_Semaphore* sem)
{
    HANDLE h = (HANDLE*)sem;
    WaitForSingleObject(h, INFINITE);
}


void HiME_SemaphoreUp(HiME_Semaphore* sem)
{
    HANDLE h = (HANDLE*)sem;
    ReleaseSemaphore(h, 1, NULL);
}
