/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#ifndef HiME_ARRAY_H
#define HiME_ARRAY_H

#include "hime-memory.h"
#include "hime-type-traits.h"
#include "hime-algorithm.h"
#include "hime-stats.h"
#include "hime-new.h"

namespace HiME
{

///
template<class T, class U = typename TypeTraitsMap<T>::Traits>
class Array
{
    HiME_DEFINE_NEW_OPERATOR();
public:

    typedef T ValueType;
    typedef U TypeTraits;
    typedef Array<ValueType, TypeTraits> SelfType;

    //-------------------------------------------------------------------------
    // Construction//assignment
    //-------------------------------------------------------------------------

    /// Constructor.
    Array();

    ///
    explicit Array(const SelfType& arr);

    ///
    SelfType& operator=(const SelfType& arr);

    ///
    void assign(const SelfType& arr);

    ///
    ~Array();
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Allocation routines.
    //-------------------------------------------------------------------------

    ///
    void clear();

    ///
    void reserve(size_t n);

    ///
    void trim();

    ///
    void resize(size_t n);

    ///
    /// Equivalent to (resize(size() + n))
    void grow(size_t n);

    /// Returns true if array size is zero.
    /// Equivalent to (size() == 0)
    bool empty() const;

    ///
    size_t size() const;

    ///
    size_t byteSize() const;

    /// Returns amount of elements which can be added to the array
    /// without additional memory allocation.
    size_t capacity() const;

    //
    // Data access routines.
    //

    ///
    ValueType* data();
    const ValueType* data() const;

    ///
    ValueType& operator[](size_t i);
    const ValueType& operator[](size_t i) const;

    ///
    ValueType& front();
    const ValueType& front() const;

    ///
    ValueType& back();
    const ValueType& back() const;
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Data modification routines.
    //-------------------------------------------------------------------------

    ///
    void replace(size_t pos, size_t old_num,
        const ValueType* src_arr, size_t new_num);

    ///
    void replace(size_t pos, size_t old_num,
        const SelfType& src, size_t src_pos = 0, size_t new_num = NPOS);

    ///
    void fill(size_t pos, size_t old_num,
        const ValueType& src,  size_t new_num);

    ///
    void insert(size_t pos, const ValueType& src);

    ///
    void append(const ValueType& t);

    ///
    SelfType& operator+=(const ValueType& t);

    ///
    void append(const SelfType& t, size_t src_pos = 0, size_t src_n = NPOS);

    ///
    SelfType& operator+=(const SelfType& t);

    ///
    void erase(size_t pos, size_t cnt);

    ///
    void pop();
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Data search routines.
    //-------------------------------------------------------------------------

    /// Finds a value.
    size_t find(const ValueType& val) const;

    /// Finds a value starting from pos and using n elements.
    size_t findFrom(size_t pos, size_t n, const ValueType& val) const;

    /// Finds a value sequence.
    size_t findArray(const SelfType& s, size_t src_pos = 0,
        size_t src_n = NPOS) const;

    /// Finds a value sequence starting from pos and using n elements.
    size_t findArrayFrom(size_t pos, size_t n, const SelfType& s,
        size_t src_pos, size_t src_n) const;

    /// Finds a value sequence.
    size_t findArray(const ValueType* s, size_t src_n) const;

    /// Finds a value sequence starting from pos and using n elements.
    size_t findArrayFrom(size_t pos, size_t n, const ValueType* s,
        size_t src_n) const;

    ///
    size_t rfind(const ValueType& val) const;
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Data substution routines.
    //-------------------------------------------------------------------------

    ///
    size_t subst(const ValueType& val, const ValueType& to);

    ///
    size_t substFrom(size_t pos, size_t n,
        const ValueType& val, const ValueType& to);

    ///
    size_t substArrayFrom(size_t pos, size_t n,
        const SelfType& sample, size_t sample_pos, size_t sample_n,
        const SelfType& to, size_t to_pos, size_t to_n);

    ///
    size_t substArrayFrom(size_t pos, size_t n,
        const ValueType* sample, size_t sample_n,
        const ValueType* to, size_t to_n);
    //-------------------------------------------------------------------------

protected:

    void makeSpace(size_t pos, size_t old_num, size_t new_num);

    ValueType*  m_data;
    size_t      m_size;
    size_t      m_capacity;
};

} // namespace HiME

#include "bits/hime-array.inl"

#endif // HiME_ARRAY_H
