/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining 
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom 
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included 
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#include "hime-memory.h"
#include "private/hime-private-mutex.h"
#include "hime-lock.h"
#include "hime-stats.h"
#include "hime-dlmalloc.h"
#include "hime-new.h"
#include "hime-arch-stdio.h"

using namespace HiME;

namespace
{
struct MemoryLogEntry
{
    void*           ptr;
    size_t          sz;
    const char*     file;
    int             line;
    size_t          next;
};
} // unnamed namespace

static struct MemoryLogEntry* g_memLog = NULL;
static PrivateMutex* g_logMutex = NULL;
static size_t g_memLogSize = 0;
static size_t g_nextFree = 0;
#define MEM_LOG_GROW 4
#define MEM_FRONT_GUARD 0x11223344
#define MEM_BACK_GUARD 0xaabbccdd
static size_t g_maxAllocSize = 0;

static mspace g_mspace;

void HiME_InitializeMemory()
{
    static char mem[sizeof(PrivateMutex)];
    g_logMutex = new (mem) PrivateMutex;

    g_mspace = create_mspace(0, 1);
}

void HiME_ShutdownMemory()
{
    destroy_mspace(g_mspace);

    g_logMutex->~PrivateMutex();
    HiME_SystemFree(g_memLog);
}

static void insertEntry(void* ptr, size_t sz, const char* file, int line)
{
    size_t old = g_nextFree;
    size_t i, ns;

    if (g_nextFree == g_memLogSize)
    {
        ns = g_memLogSize + MEM_LOG_GROW;

        g_memLog = (MemoryLogEntry*)HiME_SystemRealloc(g_memLog, ns * sizeof(g_memLog[0]));
        HiME_ASSUME(g_memLog);

        for (i = g_memLogSize; i < ns; ++i)
        {
            g_memLog[i].next = i + 1;
            g_memLog[i].sz = (size_t)-1;
        }
        g_memLogSize = ns;
    }

    g_nextFree = g_memLog[g_nextFree].next;
    g_memLog[old].ptr = ptr;
    g_memLog[old].sz = sz;
    g_memLog[old].file = file;
    g_memLog[old].line = line;
    g_memLog[old].next = (size_t)-1;
}


static void removeEntry(void* ptr, size_t sz)
{
    size_t i;

    for (i = 0; i < g_memLogSize; ++i)
    {
        if (g_memLog[i].sz != (size_t)-1 && g_memLog[i].ptr == ptr)
        {
            HiME_ASSERT(sz == (size_t)-1 || g_memLog[i].sz == sz);
            g_memLog[i].next = g_nextFree;
            g_memLog[i].sz = (size_t)-1;
            g_nextFree = i;
            return;
        }
    }
    HiME_ASSUME(0);
}


void* HiME_DebugAlloc(size_t sz, const char* file, int line)
{
    Lock<PrivateMutex> lock(*g_logMutex);

    void* ptr;
    char* t;

    if (g_maxAllocSize && sz > g_maxAllocSize)
        return NULL;

    ptr = mspace_malloc(g_mspace, sz + sizeof(size_t) * 3);
    if (!ptr)
        return NULL;

    Stats::IncrementAmount(Stats::Allocs, (int)sz);

    *(size_t*)ptr = MEM_FRONT_GUARD;
    t = (char*)ptr;
    t += sizeof(size_t);
    *(size_t*)t = sz;
    t += sizeof(size_t);
    ptr = t;
    t += sz;
    *(size_t*)t = MEM_BACK_GUARD;

    insertEntry(ptr, sz, file, line);

    return ptr;
}


void* HiME_DebugRealloc(void* p, size_t sz, size_t new_sz,
    const char* file, int line)
{
    Lock<PrivateMutex> lock(*g_logMutex);

    void* ptr;
    char* t = NULL;

    if (g_maxAllocSize && new_sz > g_maxAllocSize)
        return NULL;

    if (p)
    {
        Stats::DecrementAmount(Stats::Allocs, (int)sz);

        t = (char*)p;
        t -= sizeof(size_t);
        HiME_ASSERT(sz == (size_t)-1 || *(size_t*)(t) == sz);
        HiME_ASSERT(*(size_t*)(t + *(size_t*)t + sizeof(size_t)) == MEM_BACK_GUARD);
        t -= sizeof(size_t);
        HiME_ASSERT(*(size_t*)(t) == MEM_FRONT_GUARD);

        removeEntry(p, sz);
    }

    ptr = mspace_realloc(g_mspace, t, new_sz + sizeof(size_t) * 3);

    if (ptr)
    {
        Stats::IncrementAmount(Stats::Allocs, (int)new_sz);

        *(size_t*)ptr = MEM_FRONT_GUARD;
        t = (char*)ptr;
        t += sizeof(size_t);
        *(size_t*)t = new_sz;
        t += sizeof(size_t);
        ptr = t;
        t += new_sz;
        *(size_t*)t = MEM_BACK_GUARD;

        insertEntry(ptr, new_sz, file, line);
    }

    return ptr;
}


void HiME_DebugFree(void* p, size_t sz, const char* file, int line)
{
    Lock<PrivateMutex> lock(*g_logMutex);

    char* t;

    if (p)
    {
        Stats::DecrementAmount(Stats::Allocs, (int)sz);

        t = (char*)p;
        t -= sizeof(size_t);
        HiME_ASSERT(sz == (size_t)-1 || *(size_t*)(t) == sz);
        HiME_ASSERT(*(size_t*)(t + *(size_t*)t + sizeof(size_t)) == MEM_BACK_GUARD);
        t -= sizeof(size_t);
        HiME_ASSERT(*(size_t*)(t) == MEM_FRONT_GUARD);

        removeEntry(p, sz);
        //HiME_SystemFree(t);
        mspace_free(g_mspace, t);
    }
}


void* HiME_DebugVirtualAlloc(size_t sz, const char *file, int line)
{
    Lock<PrivateMutex> lock(*g_logMutex);

    void* ptr;

    if (g_maxAllocSize && sz > g_maxAllocSize)
        return NULL;

    ptr = HiME_SystemVirtualAlloc(sz);
    if (ptr)
    {
        Stats::IncrementAmount(Stats::VirtualAllocs, (int)sz);
        insertEntry(ptr, sz, file, line);
    }

    return ptr;
}


void HiME_DebugVirtualFree(void* p, size_t sz)
{
    Lock<PrivateMutex> lock(*g_logMutex);

    if (p)
    {
        Stats::DecrementAmount(Stats::VirtualAllocs, (int)sz);
        removeEntry(p, sz);
        HiME_SystemVirtualFree(p, sz);
    }
}


void HiME_ReportLeaks()
{
    Lock<PrivateMutex> lock(*g_logMutex);

    size_t i, k, n;
    char buff[21];
    char* s;
    char fname[256];

    for (i = 0; i < g_memLogSize; ++i)
    {
        if (g_memLog[i].sz != (size_t)-1)
        {
            n = g_memLog[i].sz < 20 ? g_memLog[i].sz : 20;
            s = (char*)g_memLog[i].ptr;
            for (k = 0; k < n; ++k)
            {
                if (*s >= 32 && *s < 127)
                {
                    buff[k] = *s;
                }
                else
                {
                    buff[k] = '?';
                }
                ++s;
            }
            buff[k] = 0;

            HiME_snprintf(fname, HiME_ARRAY_LENGTH(fname), "%s(%d): ", g_memLog[i].file, g_memLog[i].line);
            HiME_Trace("% -60s", fname);
#ifdef HiME_64BIT
    #define HiME_ADDR_FMT "0x%016x"
#else
    #define HiME_ADDR_FMT "0x%08x"
#endif
            HiME_Trace(HiME_ADDR_FMT" % 8d bytes: <%s>\n", g_memLog[i].ptr, g_memLog[i].sz, buff);
        }
    }
}

char* HiME_DebugStrDup(const char *s, const char *file, int lines)
{
    size_t sz;
    char* dst;

    if (!s)
        return NULL;

    sz = strlen(s) + 1;
    dst = (char*)HiME_DebugAlloc(sz, file, lines);
    if (!dst)
        return NULL;

    strcpy(dst, s);
    return dst;
}


void HiME_DebugSetMaxAllocSize(size_t sz)
{
    g_maxAllocSize = sz;
}
