/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining 
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom 
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included 
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#include "hime-stats.h"
#include "private/hime-private-mutex.h"
#include "hime-lock.h"
#include "hime-new.h"

namespace HiME
{
namespace Stats
{

static PrivateMutex* g_statsMutex;

static Counters g_stats[Stats_Max] =
{
    { "allocs" },
    { "vallocs" },
    { "threads" },
    { "mutexes" },
    { "mutex locks" },
    { "mutex unlocks" },
    { "semaphores" },
    { "semaphore ups" },
    { "semaphore downs" },
    { "jobs run" },
    { "strings" },
    { "vectors" },
    { "file handles" },
    { "bytes read" },
    { "bytes written" }
};


void InitializeStats()
{
    static char t[sizeof(PrivateMutex)];
    g_statsMutex = new (t) PrivateMutex;
}


void ShutdownStats()
{
    g_statsMutex->~PrivateMutex();
}


const Counters* GetStatistics(StatID id)
{
    HiME_ASSERT(id < Stats_Max);

    return &g_stats[id];
}


void IncrementCount(StatID id)
{
    Lock<PrivateMutex> lock(*g_statsMutex);

    HiME_ASSERT(id < Stats_Max);
    ++g_stats[id].count;
    ++g_stats[id].countPerFrame;
    ++g_stats[id].total;
}

void DecrementCount(StatID id)
{
    Lock<PrivateMutex> lock(*g_statsMutex);

    HiME_ASSERT(id < Stats_Max);
    --g_stats[id].count;
}


void IncrementAmount(StatID id, int64_t n)
{
    Lock<PrivateMutex> lock(*g_statsMutex);

    HiME_ASSERT(id < Stats_Max);
    ++g_stats[id].count;
    ++g_stats[id].countPerFrame;
    ++g_stats[id].total;
    g_stats[id].amount += n;
    g_stats[id].amountPerFrame += n;
    g_stats[id].totalAmount += n;
}


void DecrementAmount(StatID id, int64_t n)
{
    Lock<PrivateMutex> lock(*g_statsMutex);

    --g_stats[id].count;
    g_stats[id].amount -= n;
}


void Report()
{
    Lock<PrivateMutex> lock(*g_statsMutex);

    size_t i;

    HiME_Trace("HiME statistics:\n");
    HiME_Trace("----------------------------------------------------------------------------------------------------------\n");
    HiME_Trace("%-16s % 14s % 14s % 14s % 14s % 14s % 14s\n", "name",
        "count", "count/frame", "total",
        "amount", "amount/frame", "total amount");
    HiME_Trace("----------------------------------------------------------------------------------------------------------\n");
    for(i = 0; i < Stats_Max; ++i)
    {
        HiME_Trace("%-16s % 14d % 14d % 14d % 14d % 14d % 14d\n", g_stats[i].name,
            (int)g_stats[i].count, (int)g_stats[i].countPerFrame, (int)g_stats[i].total,
            (int)g_stats[i].amount, (int)g_stats[i].amountPerFrame,
            (int)g_stats[i].totalAmount);
    }
    HiME_Trace("----------------------------------------------------------------------------------------------------------\n");
}

} // namespace Stats
} // namespace HiME
