/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#ifndef HiME_STRING_H
#define HiME_STRING_H

#include "hime-array.h"

namespace HiME
{

template<class T>
class BaseString : protected Array<T>
{
    HiME_DEFINE_NEW_OPERATOR();
public:

    typedef T ValueType;
    typedef Array<T> BaseType;
    typedef BaseString<T> SelfType;

    //-------------------------------------------------------------------------
    // Construction//assignment
    //-------------------------------------------------------------------------

    ///
    BaseString();

    ///
    BaseString(const SelfType& s);

    ///
    BaseString(size_t n, T c);

    ///
    BaseString(const T* s);

    ///
    SelfType& operator=(const SelfType& s);

    ///
    SelfType& operator=(const T* s);

    ///
    void assign(const SelfType& s);

    ///
    void assign(size_t n, T c);

    /**
     * Assigning NULL will make a string with NULL data.
     * Assigning empty non-null string will make a string with
     * an empty non-NULL data.
     *
     * @param s     A source zero-terminated character array.
     */
    void assign(const T* s);

    ///
    ~BaseString();
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    //
    //-------------------------------------------------------------------------
    using BaseType::clear;
    using BaseType::empty;
    using BaseType::size;
    using BaseType::byteSize;
    using BaseType::capacity;
    ///
    void reserve(size_t n);

    ///
    void trim();

    ///
    void resize(size_t n);

    ///
    void grow(size_t n);
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Data access routines.
    //-------------------------------------------------------------------------
    using BaseType::data;
    using BaseType::operator[];
    using BaseType::front;
    using BaseType::back;

    ///
    T* c_str();

    ///
    const T* c_str() const;
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Data modification routines.
    //-------------------------------------------------------------------------
    using BaseType::pop;

    ///
    void replace(size_t pos, size_t old_num, const T* str, size_t new_num);

    ///
    void replace(size_t pos, size_t old_num, const T* str);

    ///
    void fill(size_t pos, size_t old_num, const T& src, size_t new_num);

    ///
    void insert(size_t pos, T c);

    ///
    void append(const SelfType& s, size_t src_pos = 0, size_t src_n = NPOS);

    ///
    void append(const T* s);

    ///
    void append(const T* s, size_t sz);

    ///
    void append(T c);

    ///
    SelfType& operator+=(const SelfType& s);

    ///
    SelfType& operator+=(const T* s);

    ///
    SelfType& operator+=(T c);

    ///
    SelfType operator+(const SelfType& b) const;

    ///
    SelfType operator+(const T* str) const;

    ///
    void erase(size_t pos, size_t cnt);
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Data search routines.
    //-------------------------------------------------------------------------

    ///
    size_t find(const T& val) const;

    ///
    size_t findFrom(size_t pos, size_t n, const T& val) const;

    ///
    size_t find(const SelfType& s, size_t src_pos = 0,
        size_t src_n = NPOS) const;

    ///
    size_t findFrom(size_t pos, size_t n, const SelfType& s,
        size_t src_pos = 0, size_t src_n = NPOS) const;

    ///
    size_t find(const T* s, size_t src_n = NPOS) const;

    ///
    size_t findFrom(size_t pos, size_t n, const T* s,
            size_t src_n = NPOS) const;

    ///
    size_t rfind(const T& val) const;
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Data substution routines.
    //-------------------------------------------------------------------------

    ///
    size_t subst(const T& from, const T& to);

    ///
    size_t substFrom(size_t pos, size_t n, const T& from, const T& to);

    ///
    size_t subst(const SelfType& from, const SelfType& to);

    ///
    size_t subst(const T* from, const T* to);

    ///
    size_t subst(const T* from, size_t nfrom, const T* to, size_t nto);

    ///
    size_t substFrom(size_t pos, size_t n, const SelfType& from,
        const SelfType& to);

    ///
    size_t substFrom(size_t pos, size_t n, const T* from, const T* to);

    ///
    size_t substFrom(size_t pos, size_t n, const T* from, size_t nfrom,
        const T* to, size_t nto);
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Data comparison routines.
    //-------------------------------------------------------------------------

    ///
    bool operator==(const SelfType& s) const;

    ///
    bool operator!=(const SelfType& s) const;

    ///
    bool operator==(const T* s) const;

    ///
    bool operator!=(const T* s) const;

    ///
    int compare(const T* s) const;

    ///
    static int compare(const SelfType& s1, const SelfType& s2);

    ///
    static int compare(const T* s1, const T* s2);
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    //
    //-------------------------------------------------------------------------

    /// Returns length in characters of a given zero-terminated string,
    /// without the final zero character.
    static size_t length(const T* s);
    //-------------------------------------------------------------------------

protected:

    using BaseType::m_data;
    using BaseType::m_size;
    using BaseType::m_capacity;

    void makestr()
    {
        if (m_data)
        {
            if (m_size == m_capacity)
                BaseType::reserve(m_capacity + 1);
            m_data[m_size] = (T)'\0';
        }
    }

    bool isValid() const
    {
        return (m_data == NULL || m_data[m_size] == (T)'\0');
    }
};

typedef BaseString<char> String;
typedef BaseString<wchar_t> WideString;

} // namespace HiME

#include "bits/hime-string.inl"

#endif // HiME_STRING_H
