/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#ifndef HiME_TYPE_TRAITS_H
#define HiME_TYPE_TRAITS_H

#include "hime-memory.h"
#include "hime-exceptions.h"
#include "hime-no-copy.h"
#include <new>

namespace HiME
{

///
template<class T>
class ClassTypeTraits : public NoCreate
{
public:
    static T* alloc(size_t n)
    {
        T* t = (T*)HiME_Alloc(n * sizeof(T));
        if (!t)
            throw std::bad_alloc();
        return t;
    }

    static void dealloc(T* p, size_t n)
    {
        HiME_Free(p, n * sizeof(T));
    }

    static void destroy(T* p, size_t n)
    {
        while (n--)
        {
            p->~T();
            ++p;
        }
    }

    static void copy(T* dst, const T* src_arr, size_t n)
    {
        if (src_arr)
        {
            while (n--)
            {
                new (dst++) T(*src_arr++);
            }
        }
        else
        {
            while (n--)
            {
                new (dst++) T;
            }
        }
    }

    static void fill(T* dst, const T& src, size_t n)
    {
        while (n--)
        {
            new (dst++) T(src);
        }
    }

    static void move(T* dst, T* src, size_t n)
    {
        HiME_ASSERT((dst && src) || n == 0);

        if (n != 0)
        {
            if (dst + n <= src || src + n <= dst )
            {
                memcpy(dst, src, n * sizeof(T));
            }
            else
            {
                memmove(dst, src, n * sizeof(T));
            }
        }
    }
};



///
template<class T>
class PlainTypeTraits : public NoCreate
{
public:
    static T* alloc(size_t n)
    {
        T* t = (T*)HiME_Alloc(n * sizeof(T));
        if (!t)
            throw std::bad_alloc();
        return t;
    }

    static void dealloc(T* p, size_t n)
    {
        HiME_Free(p, n * sizeof(T));
    }

    static void destroy(T* p, size_t n)
    {
    }

    static void copy(T* dst, const T* src, size_t n)
    {
        HiME_ASSERT((dst && src) || n == 0);
        if (n != 0)
        {
            HiME_ASSERT(dst + n <= src || src + n <= dst);
            memcpy(dst, src, n * sizeof(T));
        }
    }

    static void fill(T* dst, const T& src, size_t n)
    {
        while (n--)
        {
            memcpy(dst, &src, sizeof(T));
            ++dst;
        }
    }

    static void move(T* dst, const T* src, size_t n)
    {
        HiME_ASSERT((dst && src) || n == 0);

        if (n != 0)
        {
            if (dst + n <= src || src + n <= dst )
            {
                memcpy(dst, src, n * sizeof(T));
            }
            else
            {
                memmove(dst, src, n * sizeof(T));
            }
        }
    }
};



//
template<class T>
struct TypeTraitsMap : public NoCreate
{
    typedef ClassTypeTraits<T>      Traits;
};

template<> struct TypeTraitsMap<bool>       { typedef PlainTypeTraits<bool>         Traits; };
template<> struct TypeTraitsMap<char>       { typedef PlainTypeTraits<char>         Traits; };
template<> struct TypeTraitsMap<wchar_t>    { typedef PlainTypeTraits<wchar_t>      Traits; };
template<> struct TypeTraitsMap<uint8_t>    { typedef PlainTypeTraits<uint8_t>      Traits; };
#ifndef _CHAR_IS_SIGNED
template<> struct TypeTraitsMap<int8_t>     { typedef PlainTypeTraits<int8_t>       Traits; };
#endif
template<> struct TypeTraitsMap<uint16_t>   { typedef PlainTypeTraits<uint16_t>     Traits; };
template<> struct TypeTraitsMap<int16_t>    { typedef PlainTypeTraits<int16_t>      Traits; };
template<> struct TypeTraitsMap<uint32_t>   { typedef PlainTypeTraits<uint32_t>     Traits; };
template<> struct TypeTraitsMap<int32_t>    { typedef PlainTypeTraits<int32_t>      Traits; };
template<> struct TypeTraitsMap<uint64_t>   { typedef PlainTypeTraits<uint64_t>     Traits; };
template<> struct TypeTraitsMap<int64_t>    { typedef PlainTypeTraits<int64_t>      Traits; };
template<> struct TypeTraitsMap<float>      { typedef PlainTypeTraits<float>        Traits; };
template<> struct TypeTraitsMap<double>     { typedef PlainTypeTraits<double>       Traits; };

} // namespace HiME

#endif // HiME_TYPE_TRAITS_H
