/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#include "hime-vfs-posix.h"
#include "hime-io-exceptions.h"
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <math.h>

namespace HiME
{
namespace VFS
{

static void throw_errno(const String& path)
{
    switch (errno)
    {
    case EACCES:
        throw AccessDenied(path.c_str());

    case ENOENT:
        throw FileNotFound(path.c_str());

    default:
        throw IOException(path.c_str());
    }
}


static void convertTimespec(hime_timespec_t* dst, const struct timespec* src)
{
    dst->sec = src->tv_sec;
    dst->nano = src->tv_nsec;
}


//
//
//
PosixDriver::PosixDriver()
{
}


PosixDriver::~PosixDriver()
{
}


Node* PosixDriver::mount(const FilePath& path, const String& nativePath)
{
    FileName name = path.fileName();
    PosixNode* node = new PosixNode(name, nativePath);
    return node;
}




//
//
//
PosixNode::PosixNode(const FileName& name, const String& path)
    : m_name(name), m_path(path)
{
}


Node* PosixNode::open(const FileName& child)
{
    String cn = m_path;
    cn += "/";
    cn += child.str();

    struct stat st;
    if (::stat(cn.c_str(), &st) != 0)
        throw_errno(cn);

    PosixNode* node = new PosixNode(child, cn);
    return node;
}


void PosixNode::create(const FileName& child, uint32_t flags)
{
}


void PosixNode::remove(const FileName& child)
{
}


void PosixNode::rename(const FileName& old_name, const FileName& new_name)
{
}


Iterator* PosixNode::children()
{
    return new PosixIterator(m_path);
}


void PosixNode::stat(IO::Stat& dst)
{
    struct stat st;
    if (::stat(m_path.c_str(), &st) != 0)
        throw_errno(m_path.c_str());

    dst.size = st.st_size;
    convertTimespec(&dst.ctime, &st.st_ctim);
    convertTimespec(&dst.mtime, &st.st_mtim);

    dst.flags = 0;
    if (S_ISDIR(st.st_mode))
        dst.flags |= IO::NodeDirectory;
    else if (S_ISLNK(st.st_mode))
        dst.flags |= IO::NodeLink;
    else
        dst.flags |= IO::NodeDataFile;
}


const FileName& PosixNode::name()
{
    return m_name;
}


size_t PosixNode::streams()
{
    return 1;
}


FileStream* PosixNode::stream(int i, IO::Access access)
{
    PosixStream* s = new PosixStream(m_path, access);
    return s;
}




//
//
//
PosixIterator::PosixIterator(const String& path)
{
    m_path = path;
    m_more = false;

    m_dir = opendir(path.c_str());
    if (!m_dir)
    {
        if (errno == ENOTDIR)
            return;

        throw_errno(path.c_str());
    }
}


PosixIterator::~PosixIterator()
{
    if (m_dir)
    {
        closedir((DIR*)m_dir);
    }
}


void PosixIterator::current(FileName& dst)
{
    dst = m_current;
}


bool PosixIterator::more()
{
    return m_more;
}


bool PosixIterator::next()
{
    if (!m_dir)
        return false;

    struct dirent* entry;
    entry = readdir((DIR*)m_dir);
    if (entry)
    {
        if (entry->d_name[0] == '.')
        {
            if (entry->d_name[1] == '\0')
                return next();
            if (entry->d_name[1] == '.' && entry->d_name[2] == '\0')
                return next();
        }
        m_current = entry->d_name;
        m_more = true;
    }
    else
    {
        m_more = false;
    }
    return m_more;
}


void PosixIterator::reset()
{
    rewinddir((DIR*)m_dir);
}




//
//
//
PosixStream::PosixStream(const String& path, IO::Access access)
    : m_path(path)
{
    const char* as;
    switch (access)
    {
    case IO::AccessRead:
        as = "r";
        break;

    case IO::AccessWrite:
        as = "a";
        break;

    default:
        as = "r+";
        break;
    }

    m_file = fopen(path.c_str(), as);
    if (!m_file)
    {
        throw_errno(path);
    }

    Stats::IncrementCount(Stats::FileHandles);
}


PosixStream::~PosixStream()
{
    if (m_file)
    {
        Stats::DecrementCount(Stats::FileHandles);
        fclose((FILE*)m_file);
    }
}


hime_lsize_t PosixStream::read(void* dst, hime_lsize_t bytes)
{
    if (!m_file)
        throw InvalidObject("PosixStream::read");

    hime_lsize_t ret = fread(dst, bytes, 1, (FILE*)m_file);
    ret *= bytes;

    Stats::IncrementAmount(Stats::BytesRead, (int64_t)ret);

    return bytes;
}


hime_lsize_t PosixStream::write(const void* src, hime_lsize_t bytes)
{
    if (!m_file)
        throw InvalidObject("PosixStream::size");

    hime_lsize_t ret = fwrite(src, bytes, 1, (FILE*)m_file);
    ret *= bytes;

    Stats::IncrementAmount(Stats::BytesWritten, (int64_t)ret);

    return bytes;
}


hime_lsize_t PosixStream::seek(hime_lpos_t offset, IO::SeekOrigin org)
{
    static int whence[] = { SEEK_SET, SEEK_CUR, SEEK_END };

    if (fseeko((FILE*)m_file, offset, whence[org]) != 0)
        throw_errno(m_path.c_str());

    return tell();
}


hime_lsize_t PosixStream::tell()
{
    off_t pos = ftello((FILE*)m_file);
    if (pos == -1)
        throw_errno(m_path.c_str());
    return pos;
}


hime_lsize_t PosixStream::size()
{
    struct stat st;
    int fd = fileno((FILE*)m_file);
    if (fstat(fd, &st) != 0)
        throw_errno(m_path.c_str());
    return st.st_size;
}

} // namspace VFS
} // namspace HiME
