/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#include "hime-win32.h"
#include "hime-win32-errors.h"
#include "hime-vfs-win32.h"
#include "hime-io-exceptions.h"
#include "hime-utf8.h"
#include "hime-win32-errors.h"
#include "hime-stats.h"

namespace HiME
{
namespace VFS
{

Win32Driver::Win32Driver()
{
}


Win32Driver::~Win32Driver()
{
}


Node* Win32Driver::mount(const FilePath& path, const String& nativePath)
{
    if (nativePath.find('?') != NPOS ||
        nativePath.find('*') != NPOS ||
        nativePath.size() == 0)
    {
        throw InvalidFilePath(nativePath.c_str());
    }

    FileName name = path.fileName();
    WideString wpath;
    UTF8ToUCS2(wpath, nativePath);
    wpath.subst('/', '\\');
    while (wpath.back() == '\\')
        wpath.pop();
    Win32Node* node = new Win32Node(name, wpath);
    return node;
}




Win32Node::Win32Node(const FileName& name, const WideString& path)
    : m_name(name), m_path(path)
{
}


Node* Win32Node::open(const FileName& child)
{
    WideString wpath = m_path;
    wpath += '\\';
    UTF8ToUCS2(wpath, child.str());

    if (GetFileAttributesW(wpath.c_str()) == INVALID_FILE_ATTRIBUTES)
        Win32ThrowLastError("%s", child.c_str());

    Win32Node* node = new Win32Node(child, wpath);
    return node;
}


void Win32Node::create(const FileName& child, uint32_t flags)
{
}


void Win32Node::remove(const FileName& child)
{
}


void Win32Node::rename(const FileName& old_name, const FileName& new_name)
{
}


Iterator* Win32Node::children()
{
    return new Win32Iterator(m_path);
}


static inline void filetimeToUint64(hime_timespec_t& dst, const FILETIME& src)
{
    uint64_t t;
    t = src.dwHighDateTime;
    t <<= 32;
    t |= src.dwLowDateTime;
    t -= 116444736000000000;        // shift the Epoch from 1601 to 1970
    dst.sec = t / 10000000;
    dst.nano = 100 * (t - dst.sec * 10000000);
}


void Win32Node::stat(IO::Stat& dst)
{
    WIN32_FILE_ATTRIBUTE_DATA attrs = { 0 };
    GetFileAttributesExW(m_path.c_str(), GetFileExInfoStandard, &attrs);

    filetimeToUint64(dst.ctime, attrs.ftCreationTime);
    filetimeToUint64(dst.mtime, attrs.ftLastWriteTime);

    dst.size = ((uint64_t)attrs.nFileSizeHigh << 32) | attrs.nFileSizeLow;
    dst.flags = 0;

    if (attrs.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
        dst.flags |= IO::NodeDirectory;
    else
        dst.flags |= IO::NodeDataFile;

    if (attrs.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN)
        dst.flags |= IO::Hidden;

    if (attrs.dwFileAttributes & FILE_ATTRIBUTE_READONLY)
        dst.flags |= IO::ReadOnly;
}


const FileName& Win32Node::name()
{
    return m_name;
}


size_t Win32Node::streams()
{
    return 1;
}


FileStream* Win32Node::stream(int i, IO::Access access)
{
    return new Win32Stream(m_path, access);
}




Win32Iterator::Win32Iterator(const WideString& path)
    : m_path(path),
    m_hFind(INVALID_HANDLE_VALUE),
    m_more(false)
{
    m_path += L"\\*";
}


Win32Iterator::~Win32Iterator()
{
    if (m_hFind != INVALID_HANDLE_VALUE)
    {
        FindClose(m_hFind);
    }
}


void Win32Iterator::current(FileName& dst)
{
    dst= m_current;
}


bool Win32Iterator::more()
{
    return m_more;
}


bool Win32Iterator::next()
{
    WIN32_FIND_DATAW fdata;

    m_more = false;

    if (m_hFind == INVALID_HANDLE_VALUE)
    {
        m_hFind = FindFirstFileW(m_path.c_str(), &fdata);
        if (m_hFind == INVALID_HANDLE_VALUE)
            return false;
    }
    else
    {
        if (!FindNextFileW(m_hFind, &fdata))
            return false;
    }

    if (fdata.cFileName[0] == '.')
    {
        if (fdata.cFileName[1] == '\0')
            return next();
        if (fdata.cFileName[1] == '.' && fdata.cFileName[2] == '\0')
            return next();
    }

    String s;
    UCS2ToUTF8(s, fdata.cFileName);
    m_current = s;

    m_more = true;
    return true;
}


void Win32Iterator::reset()
{
    if (m_hFind != INVALID_HANDLE_VALUE)
    {
        FindClose(m_hFind);
        m_hFind = INVALID_HANDLE_VALUE;
    }
}




Win32Stream::Win32Stream(const WideString& path, IO::Access access)
    : m_path(path)
{
    DWORD wac = 0;
    if (access & IO::AccessRead)
        wac |= GENERIC_READ;
    if (access & IO::AccessWrite)
        wac |= GENERIC_WRITE;

    m_hFile = CreateFile(path.c_str(), wac, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, 0);
    if (m_hFile == INVALID_HANDLE_VALUE)
        Win32ThrowLastError("Win32Stream::Win32Stream '%S'", path.c_str());

    Stats::IncrementCount(Stats::FileHandles);
}


Win32Stream::~Win32Stream()
{
    if (m_hFile != INVALID_HANDLE_VALUE)
    {
        Stats::DecrementCount(Stats::FileHandles);
        CloseHandle(m_hFile);
    }
}


hime_lsize_t Win32Stream::read(void* dst, hime_lsize_t bytes)
{
    if (m_hFile == INVALID_HANDLE_VALUE)
        throw InvalidObject("Win32Stream::read");

    DWORD read;
    if (!ReadFile(m_hFile, dst, (DWORD)bytes, &read, NULL))
        Win32ThrowLastError("Win32Stream::read '%S'", m_path.c_str());

    Stats::IncrementAmount(Stats::BytesRead, (int64_t)read);

    return (hime_lsize_t)read;
}


hime_lsize_t Win32Stream::write(const void* src, hime_lsize_t bytes)
{
    if (m_hFile == INVALID_HANDLE_VALUE)
        throw InvalidObject("Win32Stream::write");

    DWORD written;
    if (!WriteFile(m_hFile, src, (DWORD)bytes, &written, NULL))
        Win32ThrowLastError("NativeStream::write");

    Stats::IncrementAmount(Stats::BytesWritten, (int64_t)written);

    return (hime_lsize_t)written;
}


hime_lsize_t Win32Stream::seek(hime_lpos_t offset, IO::SeekOrigin org)
{
    static DWORD worg[] =
    {
        FILE_BEGIN,
        FILE_CURRENT,
        FILE_END
    };

    if (m_hFile == INVALID_HANDLE_VALUE)
        throw InvalidObject("Win32Stream::seek");

    LARGE_INTEGER li;
    li.QuadPart = offset;
    li.LowPart = SetFilePointer(m_hFile, li.LowPart, &li.HighPart, worg[org]);
    if (li.LowPart == INVALID_SET_FILE_POINTER)
        Win32ThrowLastError("Win32Stream::seek");

    return (hime_lsize_t)li.QuadPart;
}


hime_lsize_t Win32Stream::tell()
{
    return seek(0, IO::SeekCurrent);
}


hime_lsize_t Win32Stream::size()
{
    if (m_hFile == INVALID_HANDLE_VALUE)
        throw InvalidObject("Win32Stream::size");
    ULARGE_INTEGER li;
    li.LowPart = GetFileSize(m_hFile, &li.HighPart);
    if (li.LowPart == INVALID_FILE_SIZE)
        Win32ThrowLastError("Win32Stream::seek");
    return (hime_lsize_t)li.QuadPart;
}

} // namspace VFS
} // namspace HiME
