/**************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining 
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom 
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included 
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#include "hime-core.h"
#include "hime-scale.h"
#include "hime-surface.h"

namespace HiME
{
namespace imageAux
{

#define READ_PIXEL(ADDR)  (sfmt == Surface::Format_ARGB_8888 ? HiME_READ_32_FROM_32(ADDR) : (sfmt == Surface::Format_RGB_888 ? HiME_READ_32_FROM_24(ADDR) : HiME_READ_32_FROM_8(ADDR)))

// This a modified version of Leptonica (http://www.leptonica.com) scaleColorAreaMapLow (scalelow.c) function.
// Modifications:
// - alpha channel support
// - reading / writng to HiME surface Surface::Format
// - usage of HiME data types
void scaleColorAreaMapLow(Surface& dst, const Surface& src)
{
    uint8_t* datad = dst.bits();
    Surface::Format dfmt = dst.format();
    int wd = dst.width();
    int hd = dst.height();
    int dstride = dst.stride();
    const uint8_t* datas = src.bits();
    Surface::Format sfmt = src.format();
    int ws = src.width();
    int hs = src.height();
    int sstride = src.stride();

    HiME_ASSERT(dfmt == Surface::Format_ARGB_8888);

    int    i, j, k, m, wm2, hm2;
    int    area00, area10, area01, area11, areal, arear, areat, areab;
    int    xu, yu;  // UL corner in src image, to 1/16 of a pixel
    int    xl, yl;  // LR corner in src image, to 1/16 of a pixel
    int    xup, yup, xuf, yuf;  // UL src pixel: integer and fraction
    int    xlp, ylp, xlf, ylf;  // LR src pixel: integer and fraction
    int    delx, dely, area;
    int    v00r, v00g, v00b, v00a;  // contrib. from UL src pixel
    int    v01r, v01g, v01b, v01a;  // contrib. from LL src pixel
    int    v10r, v10g, v10b, v10a;  // contrib from UR src pixel
    int    v11r, v11g, v11b, v11a;  // contrib from LR src pixel
    int    vinr, ving, vinb, vina;  // contrib from all full interior src pixels
    int    vmidr, vmidg, vmidb, vmida;  // contrib from side parts
    int    rval, gval, bval, aval;
    uint32_t   pixel00, pixel10, pixel01, pixel11, pixel;
    uint8_t* lined;
    const uint8_t* lines;
    float  scx, scy;

    int dbpp = Surface::bytesPerPixel(dfmt);
    int sbpp = Surface::bytesPerPixel(sfmt);

    // (scx, scy) are scaling factors that are applied to the
    // dest coords to get the corresponding src coords.
    // We need them because we iterate over dest pixels
    // and must find the corresponding set of src pixels.
    scx = 16.0f * (float)ws / (float)wd;
    scy = 16.0f * (float)hs / (float)hd;

    wm2 = ws - 2;
    hm2 = hs - 2;

     // Iterate over the destination pixels
    for (i = 0; i < hd; i++)
    {
        yu = (int)(scy * i);
        yl = (int)(scy * (i + 1.0));
        yup = yu >> 4;
        yuf = yu & 0x0f;
        ylp = yl >> 4;
        ylf = yl & 0x0f;
        dely = ylp - yup;
        lined = datad + i * dstride;
        lines = datas + yup * sstride;
        for (j = 0; j < wd; j++)
        {
            xu = (int)(scx * j);
            xl = (int)(scx * (j + 1.0));
            xup = xu >> 4;
            xuf = xu & 0x0f;
            xlp = xl >> 4;
            xlf = xl & 0x0f;
            delx = xlp - xup;

            uint8_t* dptr = lined + j * dbpp;

            // If near the edge, just use a src pixel value
            if (xlp > wm2 || ylp > hm2)
            {
                *(uint32_t*)(dptr) = READ_PIXEL(lines + xup * sbpp);
                continue;
            }

            // Area summed over, in subpixels.  This varies
            // due to the quantization, so we can't simply take
            // the area to be a constant: area = scx * scy.
            area = ((16 - xuf) + 16 * (delx - 1) + xlf) *
                   ((16 - yuf) + 16 * (dely - 1) + ylf);

            // Do area map summation
            pixel00 = READ_PIXEL(lines + xup * sbpp);
            pixel10 = READ_PIXEL(lines + xlp * sbpp);
            pixel01 = READ_PIXEL(lines + dely * sstride +  xup * sbpp);
            pixel11 = READ_PIXEL(lines + dely * sstride +  xlp * sbpp);
            area00 = (16 - xuf) * (16 - yuf);
            area10 = xlf * (16 - yuf);
            area01 = (16 - xuf) * ylf;
            area11 = xlf * ylf;

            v00r = area00 * ((pixel00 >> Surface::RedShift) & 0xff);
            v00g = area00 * ((pixel00 >> Surface::GreenShift) & 0xff);
            v00b = area00 * ((pixel00 >> Surface::BlueShift) & 0xff);
            v00a = area00 * ((pixel00 >> Surface::AlphaShift) & 0xff);

            v10r = area10 * ((pixel10 >> Surface::RedShift) & 0xff);
            v10g = area10 * ((pixel10 >> Surface::GreenShift) & 0xff);
            v10b = area10 * ((pixel10 >> Surface::BlueShift) & 0xff);
            v10a = area10 * ((pixel10 >> Surface::AlphaShift) & 0xff);

            v01r = area01 * ((pixel01 >> Surface::RedShift) & 0xff);
            v01g = area01 * ((pixel01 >> Surface::GreenShift) & 0xff);
            v01b = area01 * ((pixel01 >> Surface::BlueShift) & 0xff);
            v01a = area01 * ((pixel01 >> Surface::AlphaShift) & 0xff);

            v11r = area11 * ((pixel11 >> Surface::RedShift) & 0xff);
            v11g = area11 * ((pixel11 >> Surface::GreenShift) & 0xff);
            v11b = area11 * ((pixel11 >> Surface::BlueShift) & 0xff);
            v11a = area11 * ((pixel11 >> Surface::AlphaShift) & 0xff);

            vinr = ving = vinb = vina = 0;
            for (k = 1; k < dely; k++)
            {   // for full src pixels
                for (m = 1; m < delx; m++)
                {
                    pixel = READ_PIXEL(lines + k * sstride + (xup + m) * sbpp);
                    vinr += 256 * ((pixel >> Surface::RedShift) & 0xff);
                    ving += 256 * ((pixel >> Surface::GreenShift) & 0xff);
                    vinb += 256 * ((pixel >> Surface::BlueShift) & 0xff);
                    vina += 256 * ((pixel >> Surface::AlphaShift) & 0xff);
                }
            }
            vmidr = vmidg = vmidb = vmida = 0;
            areal = (16 - xuf) * 16;
            arear = xlf * 16;
            areat = 16 * (16 - yuf);
            areab = 16 * ylf;
            for (k = 1; k < dely; k++)
            {   // for left side
                pixel = READ_PIXEL(lines + k * sstride + xup * sbpp);
                vmidr += areal * ((pixel >> Surface::RedShift) & 0xff);
                vmidg += areal * ((pixel >> Surface::GreenShift) & 0xff);
                vmidb += areal * ((pixel >> Surface::BlueShift) & 0xff);
                vmida += areal * ((pixel >> Surface::AlphaShift) & 0xff);
            }
            for (k = 1; k < dely; k++)
            {   // for right side
                pixel = READ_PIXEL(lines + k * sstride + xlp * sbpp);
                vmidr += arear * ((pixel >> Surface::RedShift) & 0xff);
                vmidg += arear * ((pixel >> Surface::GreenShift) & 0xff);
                vmidb += arear * ((pixel >> Surface::BlueShift) & 0xff);
                vmida += arear * ((pixel >> Surface::AlphaShift) & 0xff);
            }
            for (m = 1; m < delx; m++)
            {   // for top side
                pixel = READ_PIXEL(lines + xup * sbpp + m * sbpp);
                vmidr += areat * ((pixel >> Surface::RedShift) & 0xff);
                vmidg += areat * ((pixel >> Surface::GreenShift) & 0xff);
                vmidb += areat * ((pixel >> Surface::BlueShift) & 0xff);
                vmida += areat * ((pixel >> Surface::AlphaShift) & 0xff);
            }
            for (m = 1; m < delx; m++)
            {   // for bottom side
                pixel = READ_PIXEL(lines + dely * sstride + xup * sbpp + m * sbpp);
                vmidr += areab * ((pixel >> Surface::RedShift) & 0xff);
                vmidg += areab * ((pixel >> Surface::GreenShift) & 0xff);
                vmidb += areab * ((pixel >> Surface::BlueShift) & 0xff);
                vmida += areab * ((pixel >> Surface::AlphaShift) & 0xff);
            }

            // Sum all the contributions
            rval = (v00r + v01r + v10r + v11r + vinr + vmidr + 128) / area;
            gval = (v00g + v01g + v10g + v11g + ving + vmidg + 128) / area;
            bval = (v00b + v01b + v10b + v11b + vinb + vmidb + 128) / area;
            aval = (v00a + v01a + v10a + v11a + vina + vmida + 128) / area;

            dptr[0] = bval & 0xff;
            dptr[1] = gval & 0xff;
            dptr[2] = rval & 0xff;
            dptr[3] = aval & 0xff;
        }
    }
}

} // namespace imageAux
} // namespace HiME
