/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining 
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom 
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included 
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#include "hime-scale.h"
#include "hime-surface.h"

namespace HiME
{
namespace imageAux
{

#define READ_QUAD(READ_F, SBPP) \
    pixels1 = READ_F(lines + xp * SBPP); \
    if (xp > wm2 || yp > hm2) \
    { \
        if (yp > hm2 && xp <= wm2) \
        {  /* pixels near bottom */ \
            pixels2 = READ_F(lines + (xp + 1) * SBPP); \
            pixels3 = pixels1; \
            pixels4 = pixels2; \
        } \
        else if (xp > wm2 && yp <= hm2) \
        {   /* pixels near right side */ \
            pixels2 = pixels1; \
            pixels3 = READ_F(lines + sstride + xp * SBPP); \
            pixels4 = pixels3; \
        } \
        else \
        {  /* pixels at LR corner */ \
            pixels4 = pixels3 = pixels2 = pixels1; \
        } \
    } \
    else \
    { \
        pixels2 = READ_F(lines + (xp + 1) * SBPP); \
        pixels3 = READ_F(lines + sstride + xp * SBPP); \
        pixels4 = READ_F(lines + sstride + (xp + 1) * SBPP); \
    }


// This a modified version of Leptonica (http://www.leptonica.com) scaleColorLILow (scalelow.c) function.
// Modifications:
// - alpha channel support
// - reading / writng to HiME surface format
// - usage of HiME data types
void scaleLow(Surface& dst, const Surface& src)
{
    uint8_t* datad = dst.bits();
    Surface::Format dfmt = dst.format();
    int wd = dst.width();
    int hd = dst.height();
    int dstride = dst.stride();
    const uint8_t* datas = src.bits();
    Surface::Format sfmt = src.format();
    int ws = src.width();
    int hs = src.height();
    int sstride = src.stride();

    HiME_ASSERT(dfmt == Surface::Format_ARGB_8888);

    int i, j, wm2, hm2;
    int xpm, ypm;  /* location in src image, to 1/16 of a pixel */
    int xp, yp, xf, yf;  /* src pixel and pixel fraction coordinates */
    int v00r, v01r, v10r, v11r;
    int v00g, v01g, v10g, v11g;
    int v00b, v01b, v10b, v11b;
    int v00a, v01a, v10a, v11a;
    int area00, area01, area10, area11;
    uint32_t pixels1 = 0, pixels2 = 0, pixels3 = 0, pixels4 = 0;
    const uint8_t* lines;
    uint8_t* lined;
    float scx, scy;
    uint32_t rr, gg, bb, aa;

    /* (scx, scy) are scaling factors that are applied to the
    * dest coords to get the corresponding src coords.
    * We need them because we iterate over dest pixels
    * and must find the corresponding set of src pixels. */
    scx = 16.0f * (float)ws / (float)wd;
    scy = 16.0f * (float)hs / (float)hd;

    wm2 = ws - 2;
    hm2 = hs - 2;

    int dbpp = Surface::bytesPerPixel(dfmt);

    /* Iterate over the destination pixels */
    for (i = 0; i < hd; i++)
    {
        ypm = (int32_t)(scy * (float)i);
        yp = ypm >> 4;
        yf = ypm & 0x0f;
        lined = datad + i * dstride;
        lines = datas + yp * sstride;
        for (j = 0; j < wd; j++)
        {
            xpm = (int32_t)(scx * (float)j);
            xp = xpm >> 4;
            xf = xpm & 0x0f;

            /* Do bilinear interpolation.  This is a simple
            * generalization of the calculation in scaleGrayLILow().
            * Without this, we could simply subsample:
            *     *(lined + j) = *(lines + xp);
            * which is faster but gives lousy results!  */
            switch (sfmt)
            {
            case Surface::Format_ARGB_8888:
                    READ_QUAD(HiME_READ_32_FROM_32, 4)
                    break;

            case Surface::Format_RGB_888:
                    READ_QUAD(HiME_READ_32_FROM_24, 3)
                    break;

            case Surface::Format_A_8:
                    READ_QUAD(HiME_READ_32_FROM_8, 1)
                    break;

            default:
                HiME_ASSERT(0);
                break;
            }

            area00 = (16 - xf) * (16 - yf);
            area10 = xf * (16 - yf);
            area01 = (16 - xf) * yf;
            area11 = xf * yf;
            v00r = area00 * ((pixels1 >> Surface::RedShift) & 0xff);
            v00g = area00 * ((pixels1 >> Surface::GreenShift) & 0xff);
            v00b = area00 * ((pixels1 >> Surface::BlueShift) & 0xff);

            v10r = area10 * ((pixels2 >> Surface::RedShift) & 0xff);
            v10g = area10 * ((pixels2 >> Surface::GreenShift) & 0xff);
            v10b = area10 * ((pixels2 >> Surface::BlueShift) & 0xff);

            v01r = area01 * ((pixels3 >> Surface::RedShift) & 0xff);
            v01g = area01 * ((pixels3 >> Surface::GreenShift) & 0xff);
            v01b = area01 * ((pixels3 >> Surface::BlueShift) & 0xff);

            v11r = area11 * ((pixels4 >> Surface::RedShift) & 0xff);
            v11g = area11 * ((pixels4 >> Surface::GreenShift) & 0xff);
            v11b = area11 * ((pixels4 >> Surface::BlueShift) & 0xff);

            uint8_t* dptr = lined + j * dbpp;

            rr = v00r + v10r + v01r + v11r + 128;
            gg = v00g + v10g + v01g + v11g + 128;
            bb = v00b + v10b + v01b + v11b + 128;

            dptr[0] = (bb >> 8) & 0xff;
            dptr[1] = (gg >> 8) & 0xff;
            dptr[2] = (rr >> 8) & 0xff;

            //if (dfmt == Format_ARGB_8888)
            {
                v00a = area00 * ((pixels1 >> Surface::AlphaShift) & 0xff);
                v10a = area10 * ((pixels2 >> Surface::AlphaShift) & 0xff);
                v01a = area01 * ((pixels3 >> Surface::AlphaShift) & 0xff);
                v11a = area11 * ((pixels4 >> Surface::AlphaShift) & 0xff);
                aa = v00a + v10a + v01a + v11a + 128;
                dptr[3] = (aa >> 8) & 0xff;
            }
        }
    }
}

} // namespace imageAux
} // namespace HiME
