/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#include "hime-surface.h"
#include "hime-memory.h"
#include "hime-scale.h"

namespace HiME
{

Surface::Surface() :
    m_fmt(Format_Invalid),
    m_bits(0),
    m_owns(false)
{
    destroy();
}


Surface::Surface(Format fmt, int w, int h) :
    m_fmt(Format_Invalid),
    m_bits(0),
    m_owns(false)
{
    create(fmt, w, h);
}


Surface::Surface(Format fmt, int w, int h, int stride, uint8_t* bits) :
    m_fmt(Format_Invalid),
    m_bits(0),
    m_owns(false)
{
    create(fmt, w, h, stride, bits);
}


Surface::~Surface()
{
    destroy();
}


void Surface::create(Format fmt, int w, int h)
{
    HiME_ASSERT(fmt > Format_Invalid && fmt < Format_Max);
    HiME_ASSERT(w >= 0);
    HiME_ASSERT(h >= 0);

    destroy();
    int stride = w * bytesPerPixel(fmt);
    if (stride % 4)
        stride = (stride & ~0x3) + 4;
    size_t sz = stride * h;
    m_bits = (uint8_t*)HiME_Alloc(sz);
    memset(m_bits, 0, sz);
    m_width = w;
    m_height = h;
    m_stride = stride;
    m_fmt = fmt;
    m_owns = true;
}


void Surface::create(Format fmt, int w, int h, int stride, uint8_t* bits)
{
    HiME_ASSERT(fmt > Format_Invalid && fmt < Format_Max);
    HiME_ASSERT(w >= 0);
    HiME_ASSERT(h >= 0);

    destroy();
    m_fmt = fmt;
    m_width = w;
    m_height = h;
    m_stride = stride;
    m_bits = bits;
    m_owns = false;
}


void Surface::destroy()
{
    m_fmt  = Format_Invalid;
    if (m_owns)
        HiME_Free(m_bits, m_stride * m_height);
    m_bits = 0;
    m_width = 0;
    m_height = 0;
    m_stride = 0;
}


void Surface::copyBits(const uint8_t* src, int stride)
{
    HiME_ASSERT(stride == m_stride);
    memcpy(m_bits, src, stride * m_height);
}


Surface* Surface::getScaled(int dst_wi, int dst_hi, Filter filter) const
{
    Surface* surf = new Surface(Format_ARGB_8888, dst_wi, dst_hi);
    getScaled(*surf, filter);
    return surf;
}


void Surface::getScaled(Surface& surf, Filter filter) const
{
    imageAux::scale(surf, *this);
}


Surface* Surface::getChannel(ColorChannel ch) const
{
    Surface* surf = new Surface(Format_A_8, m_width, m_height);
    getChannel(*surf, ch);
    return surf;
}


void Surface::getChannel(Surface& surf, ColorChannel ch) const
{
    HiME_ASSERT(surf.m_width == m_width);
    HiME_ASSERT(surf.m_height == m_height);
    HiME_ASSERT(surf.m_fmt == Format_A_8);

    switch (m_fmt)
    {
    case Format_A_8:
        memcpy(surf.m_bits, m_bits, m_stride);
        break;

    case Format_ARGB_8888:
        for (int i = 0; i < m_height; ++i)
        {
            uint8_t* dst = surf.line(i);
            const uint8_t* src = line(i);
            for (int k = 0; k < m_width; ++k)
            {
                *dst++ = src[ch];
                src += 4;
            }
        }
        break;

    case Format_RGB_888:
        HiME_ASSERT(ch != Surface::AlphaChannel);
        for (int i = 0; i < m_height; ++i)
        {
            uint8_t* dst = surf.line(i);
            const uint8_t* src = line(i);
            for (int k = 0; k < m_width; ++k)
            {
                *dst++ = src[ch];
                src += 3;
            }
        }
        break;

    default:
        HiME_ASSERT(0);
        break;
    }
}


void Surface::combine(Surface& surf, Surface& r, Surface& g, Surface& b)
{
    HiME_ASSERT(surf.m_width == r.m_width);
    HiME_ASSERT(surf.m_width == g.m_width);
    HiME_ASSERT(surf.m_width == b.m_width);
    HiME_ASSERT(surf.m_height == r.m_height);
    HiME_ASSERT(surf.m_height == g.m_height);
    HiME_ASSERT(surf.m_height == b.m_height);
    HiME_ASSERT(surf.m_fmt == Format_RGB_888);
    HiME_ASSERT(r.m_fmt == Format_A_8);
    HiME_ASSERT(g.m_fmt == Format_A_8);
    HiME_ASSERT(b.m_fmt == Format_A_8);

    for (int i = 0; i < surf.m_height; ++i)
    {
        uint8_t* dst = surf.line(i);
        const uint8_t* sr = r.line(i);
        const uint8_t* sg = g.line(i);
        const uint8_t* sb = b.line(i);
        for (int k = 0; k < surf.m_width; ++k)
        {
            *dst++ = *sr++;
            *dst++ = *sg++;
            *dst++ = *sb++;
        }
    }
}


void Surface::combine(Surface& surf, Surface& r, Surface& g, Surface& b, Surface& a)
{
    HiME_ASSERT(surf.m_width == r.m_width);
    HiME_ASSERT(surf.m_width == g.m_width);
    HiME_ASSERT(surf.m_width == b.m_width);
    HiME_ASSERT(surf.m_width == a.m_width);
    HiME_ASSERT(surf.m_height == r.m_height);
    HiME_ASSERT(surf.m_height == g.m_height);
    HiME_ASSERT(surf.m_height == b.m_height);
    HiME_ASSERT(surf.m_height == a.m_height);
    HiME_ASSERT(surf.m_fmt == Format_ARGB_8888);
    HiME_ASSERT(r.m_fmt == Format_A_8);
    HiME_ASSERT(g.m_fmt == Format_A_8);
    HiME_ASSERT(b.m_fmt == Format_A_8);
    HiME_ASSERT(a.m_fmt == Format_A_8);

    for (int i = 0; i < surf.m_height; ++i)
    {
        uint8_t* dst = surf.line(i);
        const uint8_t* sr = r.line(i);
        const uint8_t* sg = g.line(i);
        const uint8_t* sb = b.line(i);
        const uint8_t* sa = a.line(i);
        for (int k = 0; k < surf.m_width; ++k)
        {
            *dst++ = *sb++;
            *dst++ = *sg++;
            *dst++ = *sr++;
            *dst++ = *sa++;
        }
    }
}


size_t Surface::bytesPerPixel(Format f)
{
    switch (f)
    {
    case Format_ARGB_8888:      return 4;
    case Format_RGB_888:        return 3;
    case Format_A_8:            return 1;
    default:
        HiME_ASSERT(0);
        return (size_t)-1;
    }
}

} // namespace HiME
