/****************************************************************************
Copyright (c) 2011 kona4kona (kona4kona@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

#ifndef HiME_SURFACE_H
#define HiME_SURFACE_H

#include "hime-no-copy.h"
#include "hime-geometry.h"
#include "hime-new.h"

namespace HiME
{

class Surface : public NoCopy
{
    HiME_DEFINE_NEW_OPERATOR();
public:

    enum Format
    {
        Format_Invalid = -1,
        Format_ARGB_8888 = 0,        // AA-RR-GG-BB
        Format_RGB_888 = 1,          // RR-GG-BB
        Format_A_8 = 2,              // AA
        Format_Max
    };

    enum ColorChannelShift
    {
        AlphaShift = 24,
        RedShift = 16,
        GreenShift = 8,
        BlueShift = 0
    };

    enum ColorChannel
    {
        RedChannel = 2,
        GreenChannel = 1,
        BlueChannel = 0,
        AlphaChannel = 3
    };

    enum Filter
    {
        Filter_Fast,
        Filter_Best
    };

    enum ScaleMode
    {
        ScaleMode_FitViewport,
        ScaleMode_FitViewportIfBigger,
    };

    Surface();
    Surface(Format fmt, int w, int h);
    Surface(Format fmt, int w, int h, int stride, uint8_t* bits);
    virtual ~Surface();

    void create(Format fmt, int w, int h);
    void create(Format fmt, int w, int h, int stride, uint8_t* bits);
    void destroy();

    void copyBits(const uint8_t* src, int sourceStride);

    inline uint8_t* bits();
    inline const uint8_t* bits() const;
    inline uint8_t* line(int y);
    inline const uint8_t* line(int y) const;
    inline int width() const;
    inline int height() const;
    inline int stride() const;
    inline size_t size() const;
    inline Format format() const;
    inline bool isValid() const;

    Surface* getScaled(int dst_w, int dst_h, Filter filter) const;
    void getScaled(Surface& dst, Filter filter) const;

    Surface* getChannel(ColorChannel ch) const;
    void getChannel(Surface& surf, ColorChannel ch) const;

    static void combine(Surface& dst, Surface& r, Surface& g, Surface& b);
    static void combine(Surface& dst, Surface& r, Surface& g, Surface& b, Surface& a);
    static size_t bytesPerPixel(Format f);

protected:
    Format m_fmt;
    uint8_t* m_bits;
    bool m_owns;
    int m_width;
    int m_height;
    int m_stride;
};


//
//
//
uint8_t* Surface::bits()
{
    return m_bits;
}


const uint8_t* Surface::bits() const
{
    return m_bits;
}


uint8_t* Surface::line(int y)
{
    HiME_ASSERT(y >= 0 && y < height());
    uint8_t* p = m_bits + y * m_stride;
    return p;
}


const uint8_t* Surface::line(int y) const
{
    HiME_ASSERT(y >= 0 && y < height());
    const uint8_t* p = m_bits + y * m_stride;
    return p;
}


int Surface::width() const
{
    return m_width;
}


int Surface::height() const
{
    return m_height;
}


int Surface::stride() const
{
    return m_stride;
}


size_t Surface::size() const
{
    return m_stride * m_height;
}


Surface::Format Surface::format() const
{
    return m_fmt;
}


bool Surface::isValid() const
{
    return m_fmt != Format_Invalid;
}

} // namespace HiME

#endif // HiME_SURFACE_H
 
