#include "./env.h"
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "./sql.h"
#include "./complr.h"
#include "./judge.h"
#include "./struct.h"
#include "./log.h"

static char buffer[_BUFFSZ_];

void kill_sig(int sig) {
#ifdef _VERBOSE_
	log_write("process killed");
#endif
	exit(0);
}

void exit_commit(void) {
	sql_final();
}

void exit_logout(void) {
	log_final();
}

void exit_unlock(void) {
	int i;
	unlink(_LK_PATH_);
	for(i=0;i<_THREAD_;i++) {
		sprintf(buffer,"rm -Rf ./p%d",i);
		system(buffer);
	}
}

int check_only(int arg) {
	int fd;
	if((fd = open(_LK_PATH_,O_RDWR|O_CREAT,_F_MODE_)) < 0) {
		log_error("[E_MN01]:Open Lock File Error");
		return _FAILURE_;
	} else {
		int sz = read(fd,buffer,20);
		if(sz < 0) {
			log_error("[E_MN02]:Lock File Reading Error");
			return _FAILURE_;
		} else if(sz > 0) {
#ifdef _VERBOSE_
			log_write("process existed or not cleaned");
#endif
			if(arg & CLEN_PID) {
				ftruncate(fd,0);
				close(fd);
				fd = open(_LK_PATH_,O_WRONLY);
			} else {
				log_error("[E_MN03]:Process Already Running");
				return _FAILURE_;
			}
		} else {
			/* succeed */
			return fd;
		}
	}
}

void core_run(int r_id,int sub_id) {
	struct problem *pro_p;
	struct submit *sub_p;
	sprintf(buffer,"./p%d/",r_id);
	if(0 > chdir(buffer)) {
		return;
	}
	do {
		if(_SUCCEED_ != sql_fetch_submit(&sub_p,sub_id)) {
			break;
		}
		if(_SUCCEED_ != sql_fetch_problem(&pro_p,sub_p->prob_id)) {
			break;
		}
		int ret;
		if((ret = cc_make(sub_p))!=_SUCCEED_) {
			if(ret == _EMPTYRE_) {
				sql_free_submit(sub_p);
				sql_free_problem(pro_p);
				return;
			} else {
				break;
			}
		}
		if(_SUCCEED_ != action_run(sub_p,prob_p)) {
			break;
		}
		if(_SUCCEED_ != sql_update_submit(sub_p->r_ptr,sub_id)) {
			break;
		}
	} while(0);
	sql_free_submit(sub_p);
	sql_free_problem(pro_p);
#ifdef _VERBOSE_
	log_write("core run failed");
#endif
}

void dead_judger(void) {
	int sleep_t = 1,i,j,res;
	pid_t thread[_THREAD_];
	int submit[_THREAD_];
	for(i=0;i<_THREAD_;i++) {
		thread[i] = -1; /* not used */
	}
	for(;;) {
		if(sql_new_submit(submit,_THREAD_) > 0) {
			sleep_t = 1;
			int nfound = 1;
			while(nfound) {
				for(i=0;i<_THREAD_;i++) {
					if(submit[i] >= 0) {
						for(j=0;j<_THREAD_;j++) {
							if(thread[j] < 0) {
								nfound = 0;
								if((thread[j] = fork()) < 0) {
									log_error("[E_MN09]:Fork Error");
								} else if(thread[j] == 0) {
									core_run(j,submit[i]);
								}
							}
						}
					}
				}
				if(nfound) {
					int stat;
					waitpid(thread[0],&stat,0);
				}
			}
		} else {
			sleep(sleep_t);
			if(sleep_t < 32) {
				sleep_t *= 2; /* there will be a better solution */
			}
		}
	}
}

int init_all(arg) {
	if(log_init(_LG_PATH_,arg)!=_SUCCEED_) {
		return _FAILURE_;
	}
	atexit(exit_logout);
	if(sql_init(_DB_PATH_)!=_SUCCEED_) {
		return _FAILURE_;
	}
	atexit(exit_commit);
	if(check_only(arg)!=_SUCCEED_) {
		return _FAILURE_;
	}
	if(signal(SIGINT,kill_sig) < 0) {
		log_error("[E_MN05]:Failed to Bind SIGINT");
		return _FAILURE_;
	}
	if(signal(SIGQUIT,kill_sig) < 0) {
		log_error("[E_MN05]:Failed to Bind SIGQUIT");
		return _FAILURE_;
	}
	if(signal(SIGHUP,SIG_IGN) < 0) {
		log_error("[E_MN05]:Failed to Ignore SIGHUP");
		return _FAILURE_;
	}
	return _SUCCEED_;
}

void mk_folder(void) {
	int i;
	for(i=0;i<_THREAD_;i++) {
		sprintf(buffer,"./p%d",i);
		int sig = S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH;
		mkdir(buffer,sig);
	}
}

int parse(int ac,char **av) {
	int i,arg = 0;
	for(i=0;i<ac;i++) {
		char ch = *(*(group + i) + 1);
		if(ch == 'l') {
			arg |= CLEN_LOG;
		} else if(ch == 'p') {
			arg |= CLEN_PID;
		} else {
			log_error("[E_MN04]:Unknown Argument");
		}
	}
	return arg;
}

int make_daemon(int fd) {
	pid_t pid;
	if((pid = fork()) < 0) {
		log_error("[E_MN06]:Fork #1 Error");
		return _FAILURE_;
	} else if(pid == 0) { /* child */
		if((pid = fork()) < 0) {
			log_error("[E_MN07]:Fork #2 Error");
		} else if(pid == 0) {
			/* orphan created */
			sprintf(buffer,"%d\n",(int)getpid());
			write(fd,buffer,strlen(buffer));
			close(fd);
			mk_folder();
			atexit(exit_unlock);
#ifdef _VERBOSE_
			log_write("working in daemon");
#endif
			setpgid(0,0);
			setsid();
			dead_judger();
		} else {
			exit(0);
		}
	} else {
		if(waitpid(pid,NULL,0) != pid) {
			log_error("[E_MN08]:Wait Error");
		}
		exit(0);
	}
}

int main(int argc,char **argv) {
	int arg = parse(argc,argv);
	int fd;
	if((fd = init_all(arg))==_FAILURE_) {
		exit(1);
	}
	if(make_daemon(fd)==_FAILURE_) {
		exit(1);
	}
	log_error("[E_MN00]:System Error");
	exit(1);
	return 0;
}