#include "./env.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <memory.h>
#include <unistd.h>
#include <sqlite3.h>
#include "./sql.h"
#include "./log.h"
#include "./struct.h"

static sqlite3 *db = NULL;
static sqlite3_stmt *stmt = NULL;
static char *err_msg = NULL;
static char *buffer;
static char *err_buf;

int sql_final(void) {
	if(stmt != NULL) {
		if(SQLITE_OK != sqlite3_finalize(stmt)) {
			log_error("[E_DB07]:Finalize Error");
			return _FAILURE_;
		}
	}
	if(buffer) {
		free(buffer);
	}
	if(err_buf) {
		free(err_buf);
	}
	if(SQLITE_OK != sqlite3_close(db)) {
		log_error("[E_DB07]:Finalize Error");
		return _FAILURE_;
	}
#ifdef _VERBOSE_
	log_write("database closed");
#endif
	return _SUCCEED_;
}

int sql_init(const char *path) {
	if(SQLITE_OK != sqlite3_open(path,&db)) {
		log_error("[E_DB01]:Database Open Error");
		return _FAILURE_;
	}
	if((!(err_buf = malloc(sizeof(char) * _BUFFSZ_)))||!((buffer = malloc(sizeof(char) * _BUFFSZ_)))) {
		log_error("[E_DB02]:Buffer Alloc Error");
		return _FAILURE_;
	}
#ifdef _VERBOSE_
	log_write("database initialized");
#endif
	return _SUCCEED_;
}

int sql_new_submit(int *ary,int sz) {
	if(stmt) {
		memset(ary,-1,sizeof(int) * sz);
		int idx = 0;
		/* the space used to hold strings is freed automatically */
		while(SQLITE_ROW == sqlite3_step(stmt)) {
			const char *re_t = sqlite3_column_text(stmt,0);
			if(NULL == re_t) {
				log_error("[E_DB06]:SQL Text Error");
				return _FAILURE_;
			} else {
				ary[idx++] = atoi(re_t);
			}
		}
#ifdef _VERBOSE_
		sprintf(err_buf,"%d submits fetched",idx);
		log_write(err_buf);
#endif
		if(SQLITE_OK != sqlite3_reset(stmt)) {
			log_error("[E_DB05]:SQL Reset Error");
			return _FAILURE_;
		}
		return idx;
	} else {
		sprintf(buffer,"SELECT sub_id FROM submit WHERE result = %d ORDER BY result ASC LIMIT %d",_PENDIN_,sz);
		if(SQLITE_OK == sqlite3_prepare(db,buffer,-1,&stmt,NULL)) {
			return sql_new_submit(ary,sz);
		} else {
			log_error("[E_DB04]:SQL Prepare Error");
			return _FAILURE_;
		}
	}
}

int sql_update_submit(const struct result *rtp,int prob_id) {
	if(rtp->err_msg) {
		int length = strlen(rtp->err_msg);
		if(length > _BUFFSZ_-30) {
			char *buff = malloc(sizeof(char) * (length + _BUFFSZ_));
			int idx = (int)sprintf(buff,"UPDATE submit SET result=%d,mmr_u=-1,time_u=-1,jg_time=%s,err_msg=\"",rtp->result,_TF_SQL_);
			int i;
			for(i=0;i<length;i++) {
				if(rtp->err_msg[i] == '"') {
					buff[idx++] = '+';
				}
				buff[idx++] = rtp->err_msg[i];
				if(idx >= length + _BUFFSZ_ - 50) {
					sprintf(err_buf,"[E_DB20]:Too Many Quotes # %d",prob_id);
					log_error(err_buf);
					break;
				}
			}
			sprintf(buff+idx,"\" WHERE sub_id=%d",prob_id);
			if(SQLITE_OK != sqlite3_exec(db,buff,NULL,NULL,&err_msg)) {
				sqlite3_free(err_msg);
				sprintf(err_buf,"[E_DB19]:SQL Exec Error # %d",prob_id);
				log_error(err_buf);
				free(buff);
				return _FAILURE_;
			}
			free(buff);
			return _SUCCEED_;
		} else {
			int idx = (int)sprintf(buffer,"UPDATE submit SET result=%d,mmr_u=-1,time_u=-1,jg_time=%s,err_msg=\"",rtp->result,_TF_SQL_);
			int i;
			for(i=0;i<length;i++) {
				if(rtp->err_msg[i] == '"') {
					buffer[idx++] = '"';
				}
				buffer[idx++] = rtp->err_msg[i];
				if(idx >= _BUFFSZ_ - 30) {
					sprintf(err_buf,"[E_DB21]:Too Many Quotes # %d",prob_id);
					log_error(err_buf);
					break;
				}
			}
			sprintf(buffer+idx,"\" WHERE sub_id=%d;",prob_id);
		}
	} else {
		sprintf(buffer,"UPDATE submit SET result=%d,mmr_u=%d,time_u=%d,jg_time=%s WHERE sub_id=%d;",rtp->result,rtp->mmr_u,rtp->time_u,_TF_SQL_,prob_id);
	}
	if(SQLITE_OK != sqlite3_exec(db,buffer,NULL,NULL,&err_msg)) {
		sqlite3_free(err_msg);
		sprintf(err_buf,"[E_DB18]:SQL Exec Error # %d",prob_id);
		log_error(err_buf);
		return _FAILURE_;
	}
	return _SUCCEED_;
}

int sql_fetch_problem(struct problem **pbp,int prob_id) {
	sqlite3_stmt *st_buf;
	*pbp = malloc(sizeof(struct problem));
	sprintf(buffer,"SELECT std_in,std_ot,mmr_l,time_l FROM problem WHERE prob_id=%d",prob_id);
	if(SQLITE_OK == sqlite3_prepare(db,buffer,-1,&st_buf,NULL)) {
		if(sqlite3_step(st_buf) != SQLITE_ROW) {
			sprintf(err_buf,"[E_DB13]:SQL Step Error # %d",prob_id);
			log_error(err_buf);
			return _FAILURE_;
		} else {
			if(4 != sqlite3_column_count(st_buf)) {
				sprintf(err_buf,"[E_DB14]:SQL Query Error # %d",prob_id);
				log_error(err_buf);
				return _FAILURE_;
			}
			struct problem *ptr = *pbp;
			ptr->prob_id = prob_id;
			const char *bz_buf = sqlite3_column_text(st_buf,0);
			if(bz_buf) {
				int sz = sizeof(char) * (strlen(bz_buf)+1);
				ptr->std_in = malloc(sz);
				memcpy(ptr->std_in,bz_buf,sz);
			} else {
				ptr->std_in = NULL;
			}
			bz_buf = sqlite3_column_text(st_buf,1);
			int sz = sizeof(char) * (strlen(bz_buf)+1);
			ptr->std_ot = malloc(sz);
			memcpy(ptr->std_ot,bz_buf,sz);
			ptr->mmr_l = atoi(sqlite3_column_text(st_buf,2));
			ptr->time_l = atoi(sqlite3_column_text(st_buf,3));
			/* negative means no limit */
		}
		if(sqlite3_step(st_buf) == SQLITE_ROW) {
			sprintf(err_buf,"[E_DB15]:Two Problem With Same ID # %d",prob_id);
			log_error(err_buf);
		}
		if(SQLITE_OK != sqlite3_finalize(st_buf)) {
			log_error("[E_DB16]:SQL Finalize Error");
		}
	} else {
		log_error("[E_DB17]:SQL Prepare Error");
		return _FAILURE_;
	}
	return _SUCCEED_;
}

int sql_fetch_submit(struct submit **sbp,int sub_id) {
	sqlite3_stmt *st_buf;
	*sbp = malloc(sizeof(struct submit));
	sprintf(buffer,"SELECT prob_id,code,lang,result FROM submit WHERE sub_id=%d",sub_id);
	if(SQLITE_OK == sqlite3_prepare(db,buffer,-1,&st_buf,NULL)) {
		if(sqlite3_step(st_buf) != SQLITE_ROW) {
			sprintf(err_buf,"[E_DB09]:SQL Step Error # %d",sub_id);
			log_error(err_buf);
			return _FAILURE_;
		} else {
			if(3 != sqlite3_column_count(st_buf)) {
				sprintf(err_buf,"[E_DB12]:SQL Query Error # %d",sub_id);
				log_error(err_buf);
				return _FAILURE_;
			}
			struct submit *ptr = *sbp;
			ptr->sub_id = sub_id;
			ptr->prob_id = atoi(sqlite3_column_text(st_buf,0));
			const char *co_buf = sqlite3_column_text(st_buf,1);
			int sz = sizeof(char) * (strlen(co_buf)+1);
			ptr->code = malloc(sz);
			memcpy(ptr->code,co_buf,sz);
			ptr->lang = atoi(sqlite3_column_text(st_buf,2));
			int result = atoi(sqlite3_column_text(st_buf,3));
			if(result != _PENDIN_) {
				if(sql_free_submit(ptr) == _FAILURE_) {
					log_error("[E_DB03]:Free Submit Error");
				}
				*sbp = NULL;
				return _EMPTYRE_;
			}
			ptr->r_ptr = NULL;
		}
		if(sqlite3_step(st_buf) == SQLITE_ROW) {
			sprintf(err_buf,"[E_DB10]:Two Submit With Same ID # %d",sub_id);
			log_error(err_buf);
		}
		if(SQLITE_OK != sqlite3_finalize(st_buf)) {
			log_error("[E_DB11]:SQL Finalize Error");
		}
	} else {
		log_error("[E_DB08]:SQL Prepare Error");
		return _SUCCEED_;
	}
	return _SUCCEED_;
}