/*******************************************************************************
 * Copyright (c) 2011, Jean-David Gadina - www.xs-labs.com
 * Distributed under the Boost Software License, Version 1.0.
 * 
 * Boost Software License - Version 1.0 - August 17th, 2003
 * 
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 * 
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ******************************************************************************/

/* $Id$ */

/*!
 * @file        egpack.c
 * @copyright   (c) 2011 - Jean-David Gadina - www.xs-labs.com
 * @abstract    Program's main functions
 */

/* Local includes */
#include "eos-skl/eos-skl.h"
#include "egpack.h"

/*!
 * @function    main
 * @abstract    Program's entry point
 * @param       argc    The number of CLI arguments
 * @param       argv    A array with the CLI arguments
 * @result      The program's exit status
 */
int main( int argc, char * argv[] )
{
    FILE          * fp;
    egpack_cli_args args;
    egpack_status   status;
    char            destination[ EGPK_FILENAME_MAX ];
    
    /* Process the CLI arguments */
    egpack_get_cli_args( argc, argv, &args );
    
    /* Checks if the CLI arguments are valid */
    if( args.version == true )
    {
        /* Prints the version number and exit */
        egpack_print_version( argv[ 0 ] );
        
        return EXIT_SUCCESS;
    }
    else if
    (
             args.help    == true
        ||   args.source  == NULL
        || ( args.archive == true  && args.unarchive == true )
        || ( args.archive == false && args.unarchive == false )
    )
    {
        /* Invalid arguments - Prints the help dialog and exit */
        egpack_help( argv[ 0 ] );
        
        return EXIT_SUCCESS;
    }
    
    /* Turns on the debugging mode if needed */
    if( args.debug == true )
    {
        libdebug_enable();
        DEBUG( "Debug mode is now enabled" );
    }
    
    DEBUG
    (
        "Command line arguments:\n"
        "          - Archive:     %s\n"
        "          - Un-archive:  %s\n"
        "          - Version:     %s\n"
        "          - Help:        %s\n"
        "          - Debug:       %s\n"
        "          - Source:      %s",
        ( args.archive   == true ) ? "yes"       : "no",
        ( args.unarchive == true ) ? "yes"       : "no",
        ( args.version   == true ) ? "yes"       : "no",
        ( args.help      == true ) ? "yes"       : "no",
        ( args.debug     == true ) ? "yes"       : "no",
        ( args.source    != NULL ) ? args.source : "N/A"
    );
    
    /* Gets the destination filename */
    memset( destination, 0, EGPK_FILENAME_MAX );
    egpack_get_destination_filename( args.source, destination, ( args.archive == true ) ? true : false );
    DEBUG( "Destination filename: %s", destination );
    
    /* Checks if the source must be archived or un-archived */
    if( args.archive == true )
    {
        /* Opens a file pointer for the destination file */
        if( NULL == ( fp = fopen( destination, "wb" ) ) )
        {
            ERROR( "cannot open a file handle: %s", destination );
        }
        
        DEBUG( "Opening the file handle" );
        DEBUG( "Beginning the archive process for '%s'", args.source );
        
        /* Archive the source */
        status = egpack_archive( args.source, fp );
        
        DEBUG( "Closing the file handle" );
        fclose( fp );
        
        /* Checks the return status */
        if( status != EGPACK_OK )
        {
            /* Error - Removes the destination file */
            DEBUG( "Removing destination file" );
            remove( destination );
            printf( "Error: %s.\n", egpack_err_str( status ) );
            
            return EXIT_FAILURE;
        }
    }
    else if( args.unarchive == true )
    {
        /* Opens a file pointer for the source file */
        if( NULL == ( fp = fopen( args.source, "rb" ) ) )
        {
            ERROR( "cannot open a file handle: %s", destination );
        }
        
        DEBUG( "Opening the file handle" );
        DEBUG( "Beginning the un-archive process for '%s'", args.source );
        
        /* Un-archive the source */
        status = egpack_unarchive( fp, destination );
        
        DEBUG( "Closing the file handle" );
        fclose( fp );
        
        /* Checks the return status */
        if( status != EGPACK_OK )
        {
            printf( "Error: %s.\n", egpack_err_str( status ) );
            
            return EXIT_FAILURE;
        }
    }
    else
    {
        egpack_help( argv[ 0 ] );
    }
    
    return EXIT_SUCCESS;
}
