#include <sys/stat.h>

#include <assert.h>
#include <errno.h>
#include <getopt.h>
#include <libgen.h>
#include <limits.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "andrew_bench.h"
#include "device_ops.h"
#include "ioprio.h"
#include "opt_parse.h"

enum scheduler {
	SCHED_ANTI,
	SCHED_CFQ,
	SCHED_NOOP
};

const char *prog;

static int		isdir(const char *);
static int		set_priority(enum io_priority, uint8_t);
static void		usage();

static void
usage()
{
	printf(
"Usage: %s [OPTIONS...]\n"
"\n"
"GLOBAL OPTIONS:\n"
"  -d DIR               [REQUIRED] Destination directory (a subdirectory will\n"
"			be created within DIR).\n"
"  -h, --help           Show this message and exit.\n"
"  -m MAKE_ARGS... --   [REQUIRED] The `make' program and arguments to run.\n"
"  -n INT               The number of trials to run [default=1].\n"
"  --prio PRIO          I/O priority [default=4].\n"
"  -o PATH_PFX          [REQUIRED] Data file output prefix.\n"
"  -s DIR               [REQUIRED] Read-only source directory tree.\n"
"ARGUMENT FORMATS:\n"
"  PRIO  ^(IDLE|NORM[0-7]|RT[0-7])$     See ionice(1) for explanation.\n"
	    , prog);
}

/* -1: error
 * 0: no
 * 1: yes */
static int
isdir(const char *path)
{
	struct stat st;
	if (lstat(path, &st) == -1)
		return errno == ENOENT ? 0 : -1;
	return S_ISDIR(st.st_mode);
}

static char **
alloc_args(char **argv)
{
	char	**args;
	size_t	count = 0;

	if (!strcmp(optarg, "--")) {
		fprintf(stderr, "%s: -m requires at least one argument\n",
		    prog);
		return 0;
	}
	for (;;) {
		if (!strcmp(argv[optind + count], "--"))
			break;
		count++;
	}
	count++;

	if (!(args = malloc(sizeof(char *) * (count + 1)))) {
		fprintf(stderr, "%s: out of memory\n", prog);
		optind += count;
		return 0;
	}

	/* terminate */
	args[count] = 0;

	for (size_t i = 0; i < count; i++) {
		char	*s;
		size_t	len;

		s = i ? argv[optind + i - 1] : optarg;
		len = strlen(s);
		if (!(args[i] = malloc(len + 1))) {
			fprintf(stderr, "%s: out of memory\n", prog);
			i--;
			do {
				free(args[i]);
			} while (i--);
			free(args);

			optind += count;
			return 0;
		}
		memcpy(args[i], s, len + 1);
	}

	optind += count;
	return args;
}

static int
set_priority(enum io_priority cls, uint8_t lvl)
{
	int class = cls == PRIO_IDLE ? IOPRIO_CLASS_IDLE :
	    cls == PRIO_NORM ? IOPRIO_CLASS_BE : IOPRIO_CLASS_RT;
	int prio = IOPRIO_PRIO_VALUE(class, lvl);

	if (ioprio_set(IOPRIO_WHO_PROCESS, getpid(), prio) == -1) {
		fprintf(stderr, "%s: ioprio_set failed: %s\n",
		    prog, strerror(errno));
		return -1;
	}
	return 0;
}

int
main(int argc, char **argv)
{
	/* '+1' is due to realpath()'s vagueness */
	char			datapfx[PATH_MAX+1];
	char			dst[PATH_MAX+1];
	char			src[PATH_MAX+1];
	char			tmp[PATH_MAX+1];
	char			**args = 0;
	char			*base;
	int			status;
	unsigned		trials = 1;
	/* default CFQ priority */
	enum io_priority	prio_cls = PRIO_NORM;
	uint8_t			prio_lvl = 4;

	prog = *argv;
	*datapfx = '\0';
	*dst = '\0';
	*src = '\0';

	/* for all long options, getopt_long() will return a positive
	 * non-ascii value */
	enum { OPT_PRIO=0x80 };
	struct option longopts[] = {
	    {	"help", 0, 0, 'h' },
	    {	"prio", 1, 0, OPT_PRIO },
	    {	0, 0, 0, 0 }
	};

	if (argc == 1) {
		usage();
		status = 1;
		goto end;
	}

	for (;;) {
		int opt = getopt_long(argc, argv, "+d:hm:n:o:s:", longopts,
		    0);
		if (opt == -1) {
			if (argv[optind]) {
				fprintf(stderr,
				    "%s: stray argument on command line\n",
				    prog);
				status = 1;
				goto end;
			}
			break;
		} else if (opt == 'h') {
			usage();
			status = 0;
			goto end;
		}

		switch (opt) {
		case 'd': /* [d]estination */
			if (!realpath(optarg, dst)) {
				fprintf(stderr,
				    "%s: bad argument for -d: %s\n",
				    prog, strerror(errno));
				status = 1;
				goto end;
			}
			break;
		case 'm': /* [m]ake */
			if (!(args = alloc_args(argv))) {
				status = 1;
				goto end;
			}
			break;
		case 'n': /* [n]umber */
		{
			long	val;
			char	*end;

			errno = 0;
			val = strtol(optarg, &end, 10);
			if (errno || *end != '\0' || val < 1) {
				fprintf(stderr, "%s: bad argument for -n\n",
				    prog);
				status = 1;
				goto end;
			}
			trials = val;
			break;
		}
		case 'o': /* [o]utput */
			if (!realpath(optarg, datapfx)) {
				if (errno != ENOENT) {
					fprintf(stderr,
					    "%s: bad argument for -o: %s\n",
					    prog, strerror(errno));
					status = 1;
					goto end;
				}
			}
			break;
		case OPT_PRIO:
		{
			enum io_priority	cls = 0;
			uint8_t			lvl = 0;

			if (parse_prio(optarg, &cls, &lvl) == -1) {
				fprintf(stderr,
				    "%s: bad argument for --prio\n", prog);
				status = 1;
				goto end;
			}

			prio_cls = cls;
			prio_lvl = lvl;
			break;
		}
		case 's': /* [s]ource */
			if (!realpath(optarg, src)) {
				fprintf(stderr,
				    "%s: bad argument for -s: %s\n",
				    prog, strerror(errno));
				status = 1;
				goto end;
			}
			break;
		case ':': /* missing[:] an argument */
		case '?': /* what does this option mean[?] */
			status = 1;
			goto end;
		default:
			assert(0);
		}
	}

	/* check needed arguments */

	if (!*args || !*datapfx || !*dst || !*src) {
		const char *opt = 0;
		if (!*dst) opt = "-d";
		else if (!*args) opt = "-m";
		else if (!*datapfx) opt = "-o";
		else if (!*src) opt = "-s";
		fprintf(stderr, "%s: missing mandatory option %s\n",
		    prog, opt);
		status = 1;
		goto end;
	}

	/* validate directories */

	strcpy(tmp, datapfx);
	base = dirname(tmp);
	switch (isdir(base)) {
	case -1:
		fprintf(stderr, "%s: stat failed for `%s': %s\n",
		    prog, base, strerror(errno));
		status = 1;
		goto end;
	case 0:
		fprintf(stderr, "%s: not a directory: %s\n", prog, base);
		status = 1;
		goto end;
	default:;
	}

	switch (isdir(dst)) {
	case -1:
		fprintf(stderr, "%s: stat failed for `%s': %s\n",
		    prog, dst, strerror(errno));
		status = 1;
		goto end;
	case 0:
		fprintf(stderr, "%s: not a directory: %s\n", prog, dst);
		status = 1;
		goto end;
	default:;
	}

	switch (isdir(src)) {
	case -1:
		fprintf(stderr, "%s: stat failed for `%s': %s\n",
		    prog, src, strerror(errno));
		status = 1;
		goto end;
	case 0:
		fprintf(stderr, "%s: not a directory: %s\n", prog, src);
		status = 1;
		goto end;
	default:;
	}

	if (set_priority(prio_cls, prio_lvl) == -1) {
		status = -1;
		goto end;
	}

	/* go go go */
	if (andrew_bench(src, dst, datapfx, args, trials) == -1) {
		status = 1;
		goto end;
	}

	status = 0;
end:
	if (args) {
		for (size_t i = 0; args[i]; i++)
			free(args[i]);
		free(args);
	}
	return status;
}
