#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "andrew_bench.h"
#include "clock.h"
#include "device_ops.h"
#include "signames.h"

#define ANDREW_STAGES 5

extern char	*prog;

static int	andy_make_dir(const char *, const char *, uint64_t *);
static int	andy_copy(const char *, const char *, uint64_t *);
static int	andy_scan_dir(const char *, const char *, uint64_t *);
static int	andy_read(const char *, const char *, uint64_t *);
static int	andy_make(const char *, const char *, char *const *,
		    uint64_t *);

static int	recur_make_dir(char[PATH_MAX + 1], char[PATH_MAX + 1]);
static int	recur_copy(char[PATH_MAX + 1], char[PATH_MAX + 1]);
static int	recur_scan_dir(char[PATH_MAX + 1]);
static int	recur_read(char[PATH_MAX + 1]);

static int	copy(int, int, blksize_t);
static int	make(const char *, char *const *);
static int	read_all(int, blksize_t);

static int	andrew_bench_single(const char *, const char *, char *const *,
		    uint64_t[ANDREW_STAGES]);
static int	dump_stats(FILE *, unsigned, uint64_t *);

static int
andy_make_dir(const char *src, const char *dst, uint64_t *time)
{
	char		srcbuf[PATH_MAX+1];
	char		dstbuf[PATH_MAX+1];
	struct stat	st;
	uint64_t	start;
	const char	*base;
	size_t		len;
	size_t		len_base;

	srcbuf[PATH_MAX] = '\0';
	strncpy(srcbuf, src, PATH_MAX+1);
	assert(!srcbuf[PATH_MAX]);

	base = strrchr(srcbuf, '/');
	if (!base) base = src;
	else base++;

	len = strlen(dst);
	len_base = strlen(base);
	assert(len + len_base + 1 <= PATH_MAX);

	memcpy(dstbuf, dst, len);
	dstbuf[len] = '/';
	memcpy(dstbuf + len + 1, base, len_base + 1);

	sync();
	start = clock_monotonic();

	/* get original mode */
	if (lstat(srcbuf, &st) == -1) {
		fprintf(stderr, "%s: stat on `%s' failed: %s\n",
		    prog, srcbuf, strerror(errno));
		return -1;
	}
	/* create with same mode, plus writing */
	if (mkdir(dstbuf, (st.st_mode & ~S_IFMT) | S_IWUSR) == -1) {
		fprintf(stderr, "%s: mkdir failed for `%s': %s\n",
		    prog, dstbuf, strerror(errno));
		return -1;
	}
	/* descend */
	if (recur_make_dir(srcbuf, dstbuf))
		return -1;

	sync();
	*time = clock_monotonic() - start;

	return 0;
}

static int
andy_copy(const char *src, const char *dst, uint64_t *time)
{
	char		srcbuf[PATH_MAX+1];
	char		dstbuf[PATH_MAX+1];
	uint64_t	start;
	const char	*base;
	size_t		len;
	size_t		len_base;

	srcbuf[PATH_MAX] = '\0';
	strncpy(srcbuf, src, PATH_MAX+1);
	assert(!srcbuf[PATH_MAX]);

	base = strrchr(srcbuf, '/');
	if (!base) base = src;
	else base++;

	len = strlen(dst);
	len_base = strlen(base);
	assert(len + len_base + 1 <= PATH_MAX);

	memcpy(dstbuf, dst, len);
	dstbuf[len] = '/';
	memcpy(dstbuf + len + 1, base, len_base + 1);

	sync();
	start = clock_monotonic();

	/* descend */
	if (recur_copy(srcbuf, dstbuf))
		return -1;

	sync();
	*time = clock_monotonic() - start;

	return 0;
}

static int
andy_scan_dir(const char *src, const char *dst, uint64_t *time)
{
	char		dstbuf[PATH_MAX+1];
	uint64_t	start;
	const char	*base;
	size_t		len;
	size_t		len_base;

	base = strrchr(src, '/');
	if (!base) base = src;
	else base++;

	len = strlen(dst);
	len_base = strlen(base);
	assert(len + len_base + 1 <= PATH_MAX);

	memcpy(dstbuf, dst, len);
	dstbuf[len] = '/';
	memcpy(dstbuf + len + 1, base, len_base + 1);

	sync();
	start = clock_monotonic();

	/* descend */
	if (recur_scan_dir(dstbuf))
		return -1;

	sync();
	*time = clock_monotonic() - start;

	return 0;
}

static int
andy_read(const char *src, const char *dst, uint64_t *time)
{
	char		dstbuf[PATH_MAX+1];
	uint64_t	start;
	const char	*base;
	size_t		len;
	size_t		len_base;

	base = strrchr(src, '/');
	if (!base) base = src;
	else base++;

	len = strlen(dst);
	len_base = strlen(base);
	assert(len + len_base + 1 <= PATH_MAX);

	memcpy(dstbuf, dst, len);
	dstbuf[len] = '/';
	memcpy(dstbuf + len + 1, base, len_base + 1);

	sync();
	start = clock_monotonic();

	/* descend */
	if (recur_read(dstbuf))
		return -1;

	sync();
	*time = clock_monotonic() - start;

	return 0;
}

static int
andy_make(const char *src, const char *dst, char *const *args, uint64_t *time)
{
	char		dstbuf[PATH_MAX+1];
	uint64_t	start;
	const char	*base;
	size_t		len;
	size_t		len_base;

	base = strrchr(src, '/');
	if (!base) base = src;
	else base++;

	len = strlen(dst);
	len_base = strlen(base);
	assert(len + len_base + 1 <= PATH_MAX);

	memcpy(dstbuf, dst, len);
	dstbuf[len] = '/';
	memcpy(dstbuf + len + 1, base, len_base + 1);

	sync();
	start = clock_monotonic();

	/* descend */
	if (make(dstbuf, args))
		return -1;

	sync();
	*time = clock_monotonic() - start;

	return 0;
}

static int
recur_make_dir(char src[PATH_MAX+1], char dst[PATH_MAX+1])
{
	struct stat	st;
	DIR		*dir = 0;
	struct dirent	*entry;
	size_t		len_dst = 0;
	size_t		len_nam;
	size_t		len_src = 0;
	int		status;

	if (!(dir = opendir(src))) {
		fprintf(stderr, "%s: opendir for `%s' failed: %s\n",
		    prog, src, strerror(errno));
		status = -1;
		goto end;
	}

	len_dst = strlen(dst);
	len_src = strlen(src);
	assert(len_src + 1 <= PATH_MAX);
	if (len_dst + 1 > PATH_MAX) {
		fprintf(stderr, "%s: pathname `%s/' too long\n", prog, dst);
		status = -1;
		goto end;
	}

	dst[len_dst++] = '/';
	dst[len_dst] = '\0';
	src[len_src++] = '/';
	src[len_src] = '\0';

	while ((entry = readdir(dir))) {
		if (!strcmp(entry->d_name, ".") ||
		    !strcmp(entry->d_name, "..")) continue;

		switch (entry->d_type) {
		case DT_DIR:
		case DT_UNKNOWN:
			len_nam = strlen(entry->d_name);

			/* append to src */
			assert(len_nam + len_src <= PATH_MAX);
			memcpy(src + len_src, entry->d_name, len_nam + 1);

			/* get mode */
			if (lstat(src, &st) == -1) {
				fprintf(stderr,
				    "%s: stat failed for `%s': %s\n",
				    prog, src, strerror(errno));
				status = -1;
				goto end;
			}

			/* in case of DT_UNKNOWN, check mode */
			if (S_ISDIR(st.st_mode));
			else if (S_ISREG(st.st_mode) || S_ISLNK(st.st_mode)) {
				src[len_src] = '\0';
				break;
			} else {
				fprintf(stderr,
				    "%s: unsupported file type: %s\n",
				    prog, src);
				status = -1;
				goto end;
			}

			/* append to dst */
			if (len_nam + len_dst > PATH_MAX) {
				fprintf(stderr,
				    "%s: pathname `%s%s' too long\n",
				    prog, dst, entry->d_name);
				status = -1;
				goto end;
			}
			memcpy(dst + len_dst, entry->d_name, len_nam + 1);

			/* create with same mode */
			if (mkdir(dst, st.st_mode & ~S_IFMT) == -1) {
				fprintf(stderr, "%s: mkdir failed for `%s': %s\n",
				    prog, dst, strerror(errno));
				status = -1;
				goto end;
			}

			/* descend */
			if (recur_make_dir(src, dst)) {
				status = -1;
				goto end;
			}

			src[len_src] = '\0';
			dst[len_dst] = '\0';
			break;
		case DT_REG:
		case DT_LNK:
			break;
		default:
			fprintf(stderr, "%s: unsupported file type: %s\n",
			    prog, src);
			status = -1;
			goto end;
		}
	}

	status = 0;
end:
	if (len_dst) dst[--len_dst] = '\0';
	if (len_src) src[--len_src] = '\0';
	if (dir) closedir(dir);
	return status;
}

static int
recur_copy(char src[PATH_MAX+1], char dst[PATH_MAX+1])
{
	char		path[PATH_MAX+1];
	struct stat	st;
	DIR		*dir = 0;
	struct dirent	*entry;
	size_t		len_dst = 0;
	size_t		len_nam;
	size_t		len_src = 0;
	int		status;

	if (!(dir = opendir(src))) {
		fprintf(stderr, "%s: opendir for `%s' failed: %s\n",
		    prog, src, strerror(errno));
		status = -1;
		goto end;
	}

	len_dst = strlen(dst);
	len_src = strlen(src);
	assert(len_src + 1 <= PATH_MAX);
	assert(len_dst + 1 <= PATH_MAX);

	dst[len_dst++] = '/';
	dst[len_dst] = '\0';
	src[len_src++] = '/';
	src[len_src] = '\0';

	while ((entry = readdir(dir))) {
		bool	reg;
		bool	dir;

		if (!strcmp(entry->d_name, ".") ||
		    !strcmp(entry->d_name, "..")) continue;

		len_nam = strlen(entry->d_name);

		/* append to src */
		assert(len_nam + len_src <= PATH_MAX);
		memcpy(src + len_src, entry->d_name, len_nam + 1);

		/* append to dst */
		assert(len_nam + len_dst <= PATH_MAX);
		memcpy(dst + len_dst, entry->d_name, len_nam + 1);

		if (entry->d_type != DT_DIR) {
			/* get mode */
			if (lstat(src, &st) == -1) {
				fprintf(stderr,
				    "%s: stat failed for `%s': %s\n",
				    prog, src, strerror(errno));
				status = -1;
				goto end;
			}

			assert(S_ISREG(st.st_mode) ||
			    S_ISDIR(st.st_mode) ||
			    S_ISLNK(st.st_mode));

			reg = S_ISREG(st.st_mode);
			dir = S_ISDIR(st.st_mode);
		} else {
			reg = false;
			dir = true;
		}

		/* S_ISDIR() check in case of DT_UNKNOWN */
		if (dir) {
			/* descend */
			if (recur_copy(src, dst)) {
				status = -1;
				goto end;
			}

			src[len_src] = '\0';
			dst[len_dst] = '\0';
		} else if (reg) {
			int dfd;
			int sfd;

			if ((sfd = open(src, O_RDONLY)) == -1) {
				fprintf(stderr,
				    "%s: open failed for `%s': %s\n",
				    prog, src, strerror(errno));
				status = -1;
				goto end;
			}

			/* create with same mode, plus writing */
			if ((dfd = creat(dst, (st.st_mode & ~S_IFMT) |
			    S_IWUSR)) == -1) {
				fprintf(stderr,
				    "%s: creat failed for `%s': %s\n",
				    prog, dst, strerror(errno));
				close(sfd);
				status = -1;
				goto end;
			}

			/* do it */
			if (copy(sfd, dfd, st.st_blksize) == -1) {
				fprintf(stderr, "%s: failed to copy `%s'\n",
				    prog, src);
				close(sfd);
				close(dfd);
				status = -1;
				goto end;
			}

			close(sfd);
			close(dfd);
		} else {
			ssize_t	bytes;
			if ((bytes = readlink(src, path, PATH_MAX)) == -1) {
				fprintf(stderr,
				    "%s: readlink failed for `%s': %s\n",
				    prog, src, strerror(errno));
				status = -1;
				goto end;
			}
			path[bytes] = '\0';
			if (symlink(path, dst) == -1) {
				fprintf(stderr,
				    "%s: symlink failed for `%s': %s\n",
				    prog, src, strerror(errno));
				status = -1;
				goto end;
			}
		}
	}

	status = 0;
end:
	if (len_dst) dst[--len_dst] = '\0';
	if (len_src) src[--len_src] = '\0';
	if (dir) closedir(dir);
	return status;
}

static int
recur_scan_dir(char dst[PATH_MAX + 1])
{
	struct stat	st;
	DIR		*dir = 0;
	struct dirent	*entry;
	size_t		len_dst = 0;
	size_t		len_nam;
	int		status;

	if (!(dir = opendir(dst))) {
		fprintf(stderr, "%s: opendir for `%s' failed: %s\n",
		    prog, dst, strerror(errno));
		status = -1;
		goto end;
	}

	len_dst = strlen(dst);
	assert(len_dst + 1 <= PATH_MAX);
	dst[len_dst++] = '/';
	dst[len_dst] = '\0';

	while ((entry = readdir(dir))) {
		if (!strcmp(entry->d_name, ".") ||
		    !strcmp(entry->d_name, "..")) continue;

		len_nam = strlen(entry->d_name);

		/* append to dst */
		assert(len_nam + len_dst <= PATH_MAX);
		memcpy(dst + len_dst, entry->d_name, len_nam + 1);

		if (lstat(dst, &st) == -1) {
			fprintf(stderr,
			    "%s: stat failed for `%s': %s\n",
			    prog, dst, strerror(errno));
			status = -1;
			goto end;
		}

		if (S_ISDIR(st.st_mode)) {
			/* descend */
			if (recur_scan_dir(dst)) {
				status = -1;
				goto end;
			}

			dst[len_dst] = '\0';
		}
	}

	status = 0;
end:
	if (len_dst) dst[--len_dst] = '\0';
	if (dir) closedir(dir);
	return status;
}

static int
recur_read(char dst[PATH_MAX + 1])
{
	char		path[PATH_MAX+1];
	struct stat	st;
	DIR		*dir = 0;
	struct dirent	*entry;
	size_t		len_dst = 0;
	size_t		len_nam;
	int		status;

	if (!(dir = opendir(dst))) {
		fprintf(stderr, "%s: opendir for `%s' failed: %s\n",
		    prog, dst, strerror(errno));
		status = -1;
		goto end;
	}

	len_dst = strlen(dst);
	assert(len_dst + 1 <= PATH_MAX);
	dst[len_dst++] = '/';
	dst[len_dst] = '\0';

	while ((entry = readdir(dir))) {
		bool	dir;
		bool	have_stat = false;
		bool	reg;

		if (!strcmp(entry->d_name, ".") ||
		    !strcmp(entry->d_name, "..")) continue;

		len_nam = strlen(entry->d_name);

		/* append to dst */
		assert(len_nam + len_dst <= PATH_MAX);
		memcpy(dst + len_dst, entry->d_name, len_nam + 1);

		if (entry->d_type == DT_REG) {
			dir = false;
			reg = true;
		} else if (entry->d_type == DT_DIR) {
			dir = true;
			reg = false;
		} else if (entry->d_type == DT_LNK) {
			dir = false;
			reg = false;
		} else {
			if (lstat(dst, &st) == -1) {
				fprintf(stderr,
				    "%s: stat failed for `%s': %s\n",
				    prog, dst, strerror(errno));
				status = -1;
				goto end;
			}

			assert(S_ISREG(st.st_mode) ||
			    S_ISLNK(st.st_mode) ||
			    S_ISDIR(st.st_mode));
			dir = S_ISDIR(st.st_mode);
			reg = S_ISREG(st.st_mode);
			have_stat = true;
		}

		if (reg) {
			int fd;

			if ((fd = open(dst, O_RDONLY)) == -1) {
				fprintf(stderr,
				    "%s: open failed for `%s': %s\n",
				    prog, dst, strerror(errno));
				status = -1;
				goto end;
			}

			if (!have_stat && fstat(fd, &st) == -1) {
				fprintf(stderr,
				    "%s: stat failed for `%s': %s\n",
				    prog, dst, strerror(errno));
				close(fd);
				status = -1;
				goto end;
			}

			if (read_all(fd, st.st_blksize) == -1) {
				fprintf(stderr,
				    "%s: failed to read all of `%s'\n",
				    prog, dst);
				close(fd);
				status = -1;
				goto end;
			}

			close(fd);
		} else if (dir) {
			/* descend */
			if (recur_read(dst)) {
				status = -1;
				goto end;
			}

			dst[len_dst] = '\0';
		} else {
			if (readlink(dst, path, PATH_MAX) == -1) {
				fprintf(stderr,
				    "%s: readlink failed for `%s': %s\n",
				    prog, dst, strerror(errno));
				status = -1;
				goto end;
			}
		}
	}

	status = 0;
end:
	if (len_dst) dst[--len_dst] = '\0';
	if (dir) closedir(dir);
	return status;
}

static int
copy(int src, int dst, blksize_t szblk)
{
	uint8_t	buf[szblk];
	ssize_t bytes;
	size_t	size;
	size_t	written;
	bool	eof = false;

	for (;;) {
		/* read full block (partial if last block) */
		size = 0;
		do {
			if ((bytes = read(src, buf + size,
			    sizeof(buf) - size)) == -1) {
				if (errno == EINTR) continue;
				fprintf(stderr, "%s: read failed: %s\n",
				    prog, strerror(errno));
				return -1;
			}
			if (!bytes) {
				eof = true;
				break;
			}
			size += bytes;
		} while (size < sizeof(buf));

		if (!size) return 0;

		/* write all that was read */
		written = 0;
		do {
			if ((bytes = write(dst, buf + written,
			    size - written)) == -1) {
				if (errno == EINTR) continue;
				fprintf(stderr, "%s: write failed: %s\n",
				    prog, strerror(errno));
				return -1;
			}
			written += bytes;
		} while (written < size);

		if (eof) return 0;
	}
}

static int
make(const char *dst, char *const *args)
{
	char	cwd[PATH_MAX+1];
	int	child_status;
	int	status;
	pid_t	pid = 0;
	bool	changed = false;

	if (!getcwd(cwd, PATH_MAX+1)) {
		fprintf(stderr, "%s: getcwd failed: %s\n",
		    prog, strerror(errno));
		return -1;
	}
	if (chdir(dst) == -1) {
		fprintf(stderr, "%s: chdir failed for `%s': %s\n",
		    prog, dst, strerror(errno));
		return -1;
	}
	changed = true;

	if ((pid = fork()) == -1) {
		fprintf(stderr, "%s: chdir failed for `%s': %s\n",
		    prog, dst, strerror(errno));
		return -1;
	} else if (!pid) {
		int	devnull;
		int	savederr;
		if ((devnull = open("/dev/null", O_WRONLY)) == -1) {
			fprintf(stderr, "%s: failed to open /dev/null: %s\n",
			    prog, strerror(errno));
			exit(0xff);
		}
		if ((savederr = dup(STDERR_FILENO)) == -1) {
			fprintf(stderr, "%s: failed to open /dev/null: %s\n",
			    prog, strerror(errno));
			exit(0xff);
		}
		if (dup2(devnull, STDOUT_FILENO) == -1 ||
		    dup2(devnull, STDERR_FILENO) == -1) {
			fprintf(stderr,
			    "%s: redirect stdout to /dev/null failed: %s\n",
			    prog, strerror(errno));
			exit(0xff);
		}
		execvp(*args, args);

		long saved_errno = errno;
		if (dup2(savederr, STDERR_FILENO) == -1) {
			fprintf(stderr, "shit\n");
			exit(0xff);
		}
		fprintf(stderr, "%s: exec failed: %s\n",
		    prog, strerror(saved_errno));
		exit(0xff);
	}

	for (;;) {
		if (waitpid(pid, &child_status, 0) == -1) {
			if (errno == EINTR) continue;
			fprintf(stderr,
			    "%s: waitpid failed for `make' process: %s\n",
			    prog, strerror(errno));
			status = -1;
			goto end;
		}
		break;
	}
	pid = 0;

	if (WIFEXITED(child_status)) {
		if (WEXITSTATUS(child_status)) {
			if (WEXITSTATUS(child_status) == 0xff) {
				status = -1;
				goto end;
			}
			fprintf(stderr, "%s: child exited with %d\n",
			    prog, WEXITSTATUS(child_status));
		}
	} else if (WIFSIGNALED(child_status)) {
		fprintf(stderr, "%s: child terminated with signal %s (%d)\n",
		    prog, signame(WTERMSIG(child_status)),
		    WTERMSIG(child_status));
		status = -1;
		goto end;
	}

	status = 0;
end:
	if (pid && kill(pid, SIGKILL) == -1) {
		fprintf(stderr, "%s: kill failed for `make' process: %s\n",
		    prog, strerror(errno));
		status = -1;
	}
	if (changed && chdir(cwd) == -1) {
		fprintf(stderr, "%s: failed to chdir back: %s\n",
		    prog, strerror(errno));
		status = -1;
	}
	return status;
}

static int
read_all(int fd, blksize_t szblk)
{
	uint8_t	buf[szblk];
	ssize_t bytes;
	size_t	size;

	for (;;) {
		/* read full block (partial if last block) */
		size = 0;
		do {
			if ((bytes = read(fd, buf + size,
			    sizeof(buf) - size)) == -1) {
				if (errno == EINTR) continue;
				fprintf(stderr, "%s: read failed: %s\n",
				    prog, strerror(errno));
				return -1;
			}
			if (!bytes) break;
			size += bytes;
		} while (size < sizeof(buf));

		if (!size) return 0;
	}
}

/* removes ($dst . basename($src)) to use Perl syntax */
static int
rm_rf(const char *src, const char *dst)
{
	char		dstbuf[PATH_MAX+1];
	const char	*base;
	size_t		len;
	size_t		len_base;
	int		child_status;
	int		status;
	pid_t		pid = 0;

	base = strrchr(src, '/');
	if (!base) base = src;
	else base++;

	len = strlen(dst);
	len_base = strlen(base);
	assert(len + len_base + 1 <= PATH_MAX);

	memcpy(dstbuf, dst, len);
	dstbuf[len] = '/';
	memcpy(dstbuf + len + 1, base, len_base + 1);

	if ((pid = fork()) == -1) {
		fprintf(stderr, "%s: fork failed: %s\n",
		    prog, strerror(errno));
		status = -1;
		goto end;
	} else if (!pid) {
		execlp("rm", "rm", "-rf", dstbuf, 0);
		fprintf(stderr, "%s: exec failed for `rm': %s\n",
		    prog, strerror(errno));
		exit(0xff);
	}

	for (;;) {
		if (waitpid(pid, &child_status, 0) == -1) {
			if (errno == EINTR) continue;
			fprintf(stderr,
			    "%s: wait failed on `rm' process: %s\n",
			    prog, strerror(errno));
			status = -1;
			goto end;
		}
		break;
	}
	pid = 0;

	if (WIFEXITED(child_status)) {
		switch (WEXITSTATUS(child_status)) {
		case 0:
			break;
		case 0xff:
			status = -1;
			goto end;
		default:
			fprintf(stderr, "%s: `rm' exited with %d\n",
			    prog, WEXITSTATUS(child_status));
			status = -1;
			goto end;
		}
	} else if (WIFSIGNALED(child_status)) {
		fprintf(stderr, "%s: `rm' terminated with signal %s (%d)\n",
		    prog, signame(WTERMSIG(child_status)),
		    WTERMSIG(child_status));
		status = -1;
		goto end;
	}

	status = 0;
end:
	if (pid && kill(pid, SIGKILL) == -1) {
		fprintf(stderr, "%s: kill failed for `rm' process: %s\n",
		    prog, strerror(errno));
		status = -1;
	}
	return status;
}

static int
andrew_bench_single(const char *src_dir, const char *dst_dir,
    char *const *args, uint64_t out_times[ANDREW_STAGES])
{
	uint64_t	times[ANDREW_STAGES];
	int		status;
	bool		started = false;

	printf(">> Andrew-MakeDir...\n");
	if (drop_cache() == -1) {
		status = -1;
		goto end;
	}
	started = true;
	if (andy_make_dir(src_dir, dst_dir, times + 0)) {
		fprintf(stderr, "%s: Andrew-MakeDir failed\n", prog);
		status = -1;
		goto end;
	}
	printf("== %.3f s\n", times[0] / 1000000000.0f);

	printf(">> Andrew-Copy...\n");
	if (drop_cache() == -1) {
		status = -1;
		goto end;
	}
	if (andy_copy(src_dir, dst_dir, times + 1)) {
		fprintf(stderr, "%s: Andrew-Copy failed\n", prog);
		status = -1;
		goto end;
	}
	printf("== %.3f s\n", times[1] / 1000000000.0f);

	printf(">> Andrew-ScanDir...\n");
	if (drop_cache() == -1) {
		status = -1;
		goto end;
	}
	if (andy_scan_dir(src_dir, dst_dir, times + 2)) {
		fprintf(stderr, "%s: Andrew-ScanDir failed\n", prog);
		status = -1;
		goto end;
	}
	printf("== %.3f s\n", times[2] / 1000000000.0f);

	printf(">> Andrew-ReadAll...\n");
	if (drop_cache() == -1) {
		status = -1;
		goto end;
	}
	if (andy_read(src_dir, dst_dir, times + 3)) {
		fprintf(stderr, "%s: Andrew-ReadAll failed\n", prog);
		status = -1;
		goto end;
	}
	printf("== %.3f s\n", times[3] / 1000000000.0f);

	printf(">> Andrew-Make...\n");
	if (drop_cache() == -1) {
		status = -1;
		goto end;
	}
	if (andy_make(src_dir, dst_dir, args, times + 4)) {
		fprintf(stderr, "%s: Andrew-Make failed\n", prog);
		status = -1;
		goto end;
	}
	printf("== %.3f s\n", times[4] / 1000000000.0f);

	memcpy(out_times, times, sizeof(uint64_t) * ANDREW_STAGES);
	status = 0;
end:
	if (started) {
		printf(">> Cleaning...\n");
		if (rm_rf(src_dir, dst_dir) == -1)
			status = -1;
	}
	return status;
}

static int
dump_stats(FILE *fp, unsigned trials, uint64_t *times)
{
	bool		write_error;

	printf("> Dumping stats...\n");
	write_error = false;
	for (uint8_t i = 0; i < ANDREW_STAGES; i++) {
		double	mean;
		double	sd;
		double	sqsum = 0;
		double	sum = 0;

		if (fprintf(fp, "TIMES%hhu\t", i) < 0) {
			write_error = true;
			break;
		}
		for (unsigned j = 0; j < trials; j++) {
			size_t k = i + j * ANDREW_STAGES;
			sum += times[k];
			/* use floating point arithmetic to avoid overflow */
			sqsum += (double)times[k] * times[k];
			if (fprintf(fp, "%llu%c",
			    (long long unsigned)times[k],
			    j == trials-1 ? '\n' : ' ') < 0) {
				write_error = true;
				break;
			}
		}
		if (write_error) break;

		mean = sum / trials;
		sd = sqrt(sqsum / trials - mean * mean);
		if (fprintf(fp, "MEAN%hhu\t%llu\nDEV%hhu\t%llu\n",
		    i, (long long unsigned)(mean + .5),
		    i, (long long unsigned)(sd + .5)) < 0) {
			write_error = true;
			break;
		}
	}
	if (write_error) {
		fprintf(stderr, "%s: error writing data file: %s\n",
		    prog, strerror(errno));
		return -1;
	}
	return 0;
}

int
andrew_bench(const char *src_dir, const char *dst_dir, const char *datapfx,
    char *const *args, unsigned trials)
{
	char		buf_a[PATH_MAX];
	char		buf_c[PATH_MAX];
	char		device[16];
	/* a 2d row-major matrix (m x n) where m=trials, n=ANDREW_STAGES */
	uint64_t	*times_a = 0;
	uint64_t	*times_c = 0;
	FILE		*fp_a = 0;
	FILE		*fp_c = 0;
	int		retval;
	int		status;
	bool		using_anti = false;

	static const char *nice_names[] = { "Anticipatory", "CFQ" };
	static const char *names[] = { "anticipatory", "cfq" };
	uint64_t	*times[2];

	assert(trials);

	/* XXX assume that both paths are on the same device */
	if (get_scheduler_device(src_dir, device) == -1) {
		status = -1;
		goto end;
	}

	if (!(times_a = calloc(trials, ANDREW_STAGES * sizeof(uint64_t))) ||
	    !(times_c = calloc(trials, ANDREW_STAGES * sizeof(uint64_t)))) {
		fprintf(stderr, "%s: out of memory\n", prog);
		status = -1;
		goto end;
	}

	retval = snprintf(buf_a, PATH_MAX, "%s-anti.dat", datapfx);
	assert(0 < retval && retval < PATH_MAX);
	retval = snprintf(buf_c, PATH_MAX, "%s-cfq.dat", datapfx);
	assert(0 < retval && retval < PATH_MAX);

	if (!(fp_a = fopen(buf_a, "w"))) {
		fprintf(stderr, "%s: open failed for datafile `%s': %s\n",
		    prog, buf_a, strerror(errno));
		status = -1;
		goto end;
	}
	if (!(fp_c = fopen(buf_c, "w"))) {
		fprintf(stderr, "%s: open failed for datafile `%s': %s\n",
		    prog, buf_c, strerror(errno));
		status = -1;
		goto end;
	}

	/* run all interleaved trials; I have discovered that the order
	 * in which the tests are run (all cfq, all anti) causes the second
	 * to perform worse (or was it better?) */
	times[0] = times_a;
	times[1] = times_c;
	for (unsigned i = 0; i < trials; i++) {
		for (uint8_t j = 0; j < 2; j++) {
			if (set_scheduler(device, names[j]) == -1) {
				status = -1;
				goto end;
			}
			using_anti = !j;
			printf("> Trial %d (%s)\n", i, nice_names[j]);
			if (andrew_bench_single(src_dir, dst_dir, args,
			    times[j] + ANDREW_STAGES * i) == -1) {
				status = -1;
				goto end;
			}
		}
	}

	if (dump_stats(fp_a, trials, times_a) == -1) {
		fprintf(stderr, "%s: failed to write `%s'", prog, buf_a);
		status = -1;
		goto end;
	}
	if (dump_stats(fp_c, trials, times_c) == -1) {
		fprintf(stderr, "%s: failed to write `%s'", prog, buf_c);
		status = -1;
		goto end;
	}

	status = 0;
end:
	/* revert to anticipatory */
	if (!using_anti && set_scheduler(device, "anticipatory") == -1)
			status = -1;
	if (fp_c) {
		fclose(fp_c);
		if (status) unlink(buf_c);
	}
	if (fp_a) {
		fclose(fp_a);
		if (status) unlink(buf_a);
	}
	if (times_c) free(times_c);
	if (times_a) free(times_a);
	return status;
}
