#include <sys/stat.h>
#include <sys/wait.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "device_ops.h"
#include "proc_spec.h"
#include "signames.h"

extern char	*prog;
const char	*BLOCK_DIR = "/sys/block";

static int	qsort_strcmp(const void *, const void *);

/* longer than drop_cache.c; returns 0/-1; prints error */
int
drop_cache()
{
	pid_t	pid;
	int	status;

	if ((pid = fork()) == -1) {
		fprintf(stderr, "%s: fork failed: %s\n",
		    prog, strerror(errno));
		return -1;
	} else if (!pid) {
		execl("drop_cache", "./drop_cache", (const char *)0);
		execlp("drop_cache", "drop_cache", (const char *)0);
		fprintf(stderr, "%s: failed to exec drop_cache: %s\n",
		    prog, strerror(errno));
		exit(0xff);
	}

	while (waitpid(-1, &status, 0) == -1)
		assert(errno == EINTR);
	if (WIFEXITED(status)) {
		switch (WEXITSTATUS(status)) {
		case 0:
			/* completed successfully */
			break;
		case 0xff:
			return -1;
		default:
			fprintf(stderr, "%s: drop_cache exited with %d\n",
			    prog, WEXITSTATUS(status));
			return -1;
		}
	} else {
		int signum = WTERMSIG(status);
		fprintf(stderr, "%s: drop_cache terminated with %s (%d)\n",
		    prog, signame(signum), signum);
		return -1;
	}

	return 0;
}

/* returns 0/-1; prints error */
int
set_readahead_path(const char *path_dev, uint16_t readahead)
{
	char	arg[6];
	pid_t	pid;
	int	status;

	/* XXX temporarily disabled */
	return 0;

	sprintf(arg, "%hu", readahead);

	if ((pid = fork()) == -1) {
		fprintf(stderr, "%s: fork failed: %s\n",
		    prog, strerror(errno));
		return -1;
	} else if (!pid) {
		execl("set_readahead", "./set_readahead",
		    path_dev, arg, (const char *)0);
		execlp("set_readahead", "set_readahead",
		    path_dev, arg, (const char *)0);
		fprintf(stderr, "%s: failed to exec set_readahead: %s\n",
		    prog, strerror(errno));
		exit(0xff);
	}

	waitpid(pid, &status, 0);
	if (WIFEXITED(status)) {
		switch (WEXITSTATUS(status)) {
		case 0:
			/* completed successfully */
			break;
		case 0xff:
			return -1;
		default:
			fprintf(stderr, "%s: set_readahead exited with %d\n",
			    prog, WEXITSTATUS(status));
			return -1;
		}
	} else {
		fprintf(stderr, "%s: set_readahead terminated with signal %d\n",
		    prog, WTERMSIG(status));
		return -1;
	}

	return 0;
}

/* trace up the tree until dev_id(parent_dir(n)) != dev_id(n),
 * n copied to 'out' */
int
path_to_mountpoint(const char *path, char out[PATH_MAX])
{
	char		buf[PATH_MAX];
	char		buf_next[PATH_MAX];
	struct stat	st;
	char		*end;
	dev_t		dev_id;
	dev_t		dev_id_next;

	if (!realpath(path, buf)) {
		fprintf(stderr, "%s: realpath failed for `%s': %s\n",
		    prog, path, strerror(errno));
		return -1;
	}

	/* parent directory */
	strcpy(buf_next, buf);
	end = strrchr(buf_next, '/');
	*end = '\0';

	if (stat(buf, &st) == -1) {
		fprintf(stderr, "%s: stat failed: %s\n",
		    prog, strerror(errno));
		return -1;
	}
	dev_id = st.st_dev;

	for (;;) {
		/* root directory */
		if (!*buf_next)
			strcpy(buf_next, "/");

		if (stat(buf_next, &st) == -1) {
			fprintf(stderr, "%s: stat failed: %s\n",
			    prog, strerror(errno));
			return -1;
		}
		dev_id_next = st.st_dev;

		if (dev_id != dev_id_next)
			break;

		/* go up directory */
		dev_id = dev_id_next;
		*strrchr(buf, '/') = '\0';
		*strrchr(buf_next, '/') = '\0';
	}

	strcpy(out, buf);
	return 0;
}

/* this could be one very short awk script; parse /etc/mtab */
int
mountpoint_to_device(const char *mntpoint, char buf[PATH_MAX])
{
	FILE *file;
	char *line;
	/* 88 is the minimum size needed on my /etc/mtab, this should be good
	 * enough; if not, likely there would only be one realloc */
	size_t sz = 128;

	if (!(line = malloc(sz))) {
		fprintf(stderr, "%s: out of memory\n", prog);
		return -1;
	}

	if (!(file = fopen("/etc/mtab", "r"))) {
		fprintf(stderr, "%s: fopen failed: %s\n",
		    prog, strerror(errno));
		free(line);
		return -1;
	}

	for (;;) {
		size_t	off = 0;
		char	*saveptr;
		char	*dev;
		char	*point;

		/* read line, set line=0 if EOF/error */
		for (;;) {
			size_t len;
			char *newbuf;

			if (!fgets(line + off, sz - off, file)) {
				free(line);
				if (feof(file))
					fprintf(stderr, "%s: mount point not "
					    "in /etc/mtab\n",
					    prog);
				else
					fprintf(stderr, "%s: error reading "
					    "/etc/mtab: %s\n",
					    prog, strerror(errno));
				fclose(file);
				return -1;
			}

			len = strlen(line);
			if (len < sz - 1 || line[sz - 1] == '\n')
				break;

			/* double buffer size */
			sz <<= 1;
			if (!(newbuf = realloc(line, sz))) {
				fprintf(stderr, "%s: out of memory\n",
				    prog);
				fclose(file);
				free(line);
				return -1;
			}
			line = newbuf;
			off = len;
		}

		saveptr = 0;
		if (!(dev = strtok_r(line, " ", &saveptr))) {
			fprintf(stderr, "%s: bad /etc/mtab format\n", prog);
			fclose(file);
			free(line);
			return -1;
		}
		if (!(point = strtok_r(0, " ", &saveptr))) {
			fprintf(stderr, "%s: bad /etc/mtab format\n", prog);
			fclose(file);
			free(line);
			return -1;
		}
		if (!strcmp(point, mntpoint)) {
			strcpy(buf, dev);
			fclose(file);
			free(line);
			return 0;
		}
	}
}

static int
qsort_strcmp(const void *a, const void *b)
{
	const char **a_ = (const char **)a;
	const char **b_ = (const char **)b;
	return strcmp(*a_, *b_);
}

int
set_all_readahead(const struct proc_spec *specs, uint16_t count, uint16_t val)
{
	const char	**devs = 0;
	int		status;

	if (!(devs = malloc(count * sizeof(char *)))) {
		fprintf(stderr, "%s: out of memory\n", prog);
		status = -1;
		goto end;
	}

	for (uint16_t i = 0; i < count; i++)
		devs[i] = (const char *)specs[i].path_device;

	qsort(devs, count, sizeof(char *), qsort_strcmp);
	for (uint16_t i = 0; i < count; i++) {
		/* skip duplicates */
		if (i && !strcmp(devs[i], devs[i-1])) continue;

		if (set_readahead_path(devs[i], val) == -1) {
			status = -1;
			goto end;
		}
	}

	status = 0;
end:
	if (devs) free(devs);
	return status;
}

int
get_scheduler_device(const char *path, char device[16])
{
	struct stat	st;
	char		buf0[PATH_MAX];
	char		buf1[PATH_MAX];
	char		*begin;
	size_t		len;

	if (path_to_mountpoint(path, buf0) == -1 ||
	    mountpoint_to_device(buf0, buf0) == -1)
		return -1;

	begin = strrchr(buf0, '/');
	assert(begin);
	begin++;

	/* copy characters following '/' up to but not including the first
	 * digit or null */
	for (len = 0; len < 16 && begin[len] && !isdigit(begin[len]); len++);
	assert(len < 16);
	begin[len] = '\0';

	/* ensure it actually exists; no overflow possible */
	if (sprintf(buf1, "%s/%s", BLOCK_DIR, begin) < 0) assert(0);
	if (stat(buf1, &st) == -1) {
		fprintf(stderr, "%s: found a device not in %s\n",
		    prog, BLOCK_DIR);
		return -1;
	}
	memcpy(device, begin, len + 1);

	return 0;
}

int
set_scheduler(const char *device, const char *scheduler)
{
	pid_t	pid;
	int	status;

	if ((pid = fork()) == -1) {
		fprintf(stderr, "%s: fork failed: %s\n",
		    prog, strerror(errno));
		return -1;
	} else if (!pid) {
		execl("set_sched", "./set_sched", device, scheduler,
		    (const char *)0);
		execlp("set_sched", "set_sched", device, scheduler,
		    (const char *)0);
		fprintf(stderr, "%s: failed to exec set_sched: %s\n",
		    prog, strerror(errno));
		exit(0xff);
	}

	waitpid(pid, &status, 0);
	if (WIFEXITED(status)) {
		switch (WEXITSTATUS(status)) {
		case 0:		break;
		case 0xff:	return -1;
		default:
			fprintf(stderr, "%s: set_sched exited with %d\n",
			    prog, WEXITSTATUS(status));
			return -1;
		}
	} else {
		fprintf(stderr, "%s: set_sched terminated with signal %d\n",
		    prog, WTERMSIG(status));
		return -1;
	}

	return 0;
}
