#include <sys/types.h>
#include <sys/stat.h>

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "disk_stats.h"

extern const char *prog;

static inline const char *
		basename(const char *);


static inline const char *
basename(const char *path)
{
	const char *t = strrchr(path, '/');
	return t ? t+1 : path;
}

int
find_device_stat(const char *device, char path[PATH_MAX])
{
	struct stat	st;
	int		len_buf;

	device = basename(device);

	len_buf = snprintf(path, PATH_MAX, "/sys/block/%s", device);
	assert(len_buf >= 0);
	if (len_buf >= PATH_MAX) {
		fprintf(stderr, "%s: string overflow avoided\n", prog);
		return -1;
	}

	/* strip off trailing digits */
	while (isdigit(path[len_buf-1]))
		path[--len_buf] = '\0';

	/* don't worry about following link */
	if (lstat(path, &st) == -1) {
		if (errno == ENOENT)
			fprintf(stderr, "%s: could not find sysfs directory "
			    "entry `%s'\n", prog, path);
		else
			fprintf(stderr, "%s: stat failed for sysfs "
			    "directory: %s\n", prog, strerror(errno));
		return -1;
	}

	/* pretty sure I don't want /sys/block/sda/sda1/stat */

	/* 5: len('/' + 'stat') */
	if (len_buf + 5 >= PATH_MAX) {
		fprintf(stderr, "%s: string overflow avoided\n",
		    prog);
		return -1;
	}

	/* append "/stat" */
	memcpy(path + len_buf, "/stat", 6);

	if (stat(path, &st) == -1) {
		fprintf(stderr, "%s: stat failed for disk "
		    "stat file `%s': %s\n",
		    prog, path, strerror(errno));
		return -1;
	}

	return 0;
}

int
read_disk_stats(const char *caller, const char *statpath,
    struct disk_stats *stats)
{
	const size_t	COUNT = 11;

	struct disk_stats temp_stats;
	struct stat	st;
	uint32_t	*members[COUNT];
	char		*buf = 0;
	FILE		*fp = 0;
	char		*saveptr;
	char		*strtok_arg;
	char		*tok;
	int		status;

	if (!(fp = fopen(statpath, "r"))) {
		fprintf(stderr, "%s: fopen failed for `%s': %s\n",
		    caller, statpath, strerror(errno));
		status = -1;
		goto end;
	}

	if (fstat(fileno(fp), &st) == -1) {
		fprintf(stderr, "%s: fstat failed for `%s': %s\n",
		    caller, statpath, strerror(errno));
		status = -1;
		goto end;
	}

	if (!(buf = malloc(st.st_size + 1))) {
		fprintf(stderr, "%s: out of memory\n", caller);
		status = -1;
		goto end;
	}

	if (!fgets(buf, st.st_size + 1, fp)) {
		fprintf(stderr, "%s: read error for `%s': %s\n",
		    caller, statpath, strerror(errno));
		status = -1;
		goto end;
	}
	fclose(fp);
	fp = 0;

	members[ 0] = &temp_stats.reads_completed;
	members[ 1] = &temp_stats.reads_merged;
	members[ 2] = &temp_stats.sectors_read;
	members[ 3] = &temp_stats.milli_reading;
	members[ 4] = &temp_stats.writes_completed;
	members[ 5] = &temp_stats.writes_merged;
	members[ 6] = &temp_stats.sectors_written;
	members[ 7] = &temp_stats.milli_writing;
	members[ 8] = &temp_stats.pending_io;
	members[ 9] = &temp_stats.milli_io;
	members[10] = &temp_stats.milli_io_weighted;

	saveptr = 0;
	strtok_arg = buf;
	for (size_t i = 0; i < COUNT; i++) {
		long long	ll;
		char		*endptr;

		if (!(tok = strtok_r(strtok_arg, " \t\n", &saveptr))) {
			fprintf(stderr, "%s: bad stat format\n", caller);
			status = -1;
			goto end;
		}
		strtok_arg = 0;

		errno = 0;
		endptr = 0;
		ll = strtoll(tok, &endptr, 10);
		if (errno) {
			fprintf(stderr, "%s: bad number `%s': %s\n",
			    caller, tok, strerror(errno));
			status = -1;
			goto end;
		}
		if (*endptr || ll & ~0xffffffff) {
			fprintf(stderr, "%s: bad number `%s'\n",
			    caller, tok);
			status = -1;
			goto end;
		}

		*members[i] = (uint32_t)ll;
	}

	*stats = temp_stats;
	status = 0;
end:
	if (buf) free(buf);
	if (fp) fclose(fp);
	return status;
}
