#include <sys/types.h>
#include <sys/stat.h>

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <time.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "clock.h"
#include "device_ops.h"
#include "mgmt.h"
#include "opt_parse.h"
#include "proc_spec.h"
#include "workload.h"


/* this happens to be the default when I call ioctl with BLKRAGET */
const int	DEFAULT_READAHEAD = 256;


const char *prog;

static inline int
		timespec_cmp(const struct timespec *, const struct timespec *);
static inline bool
		timespec_zero(const struct timespec *);
static void	usage();

static void
usage()
{
	printf(
"Usage: %s -n INT [OPTIONS...]\n"
"       %s --spins TIME\n"
"\n"
"GLOBAL OPTIONS:\n"
"  -a                Apply per-process options to all processes.\n"
"  -h, --help        Show this message and exit.\n"
"  -n INT            [REQUIRED] Run test with INT processes.\n"
"  -o PATH           [REQUIRED] Prefix for data files.\n"
"  -t INT            Apply per-process options to only process INT.\n"
"                    INT in [0, num_procs).\n"
"  --duration TIME   Run tests for the duration TIME [default=10s].\n"
"  --spins TIME      Just calculate and print the number of spins needed to\n"
"                    cause a delay this long.  This can later be fed into\n"
"                    the --gap option.\n"
"PER-PROCESS OPTIONS:\n"
"  -b SIZE           Block size of the operations [default=4kB].\n"
"  -f PATH           [REQUIRED] File to read/write to. `Should' not be shared.\n"
"  -w WORK [OPT...]  See `WORKLOAD TYPES' below [default=rseq]\n"
"  --end TIME        Time (after synchronous start) for process to end\n"
"                    [default=0].\n"
"  --gap TIME        Gap between operations [default=0].\n"
"  --mmap            Map the file into memory with mmap.\n"
"  --prio PRIO       I/O priority [default=4].\n"
"  --rapid INT       Make INT requests in between gaps [default=1].\n"
"  --start TIME      Time (after synchronous start) for process to start\n"
"                    [default=0].\n"
"WORKLOAD TYPES:\n"
"  rseq              Sequential reads.\n"
"  wseq              Sequential writes.\n"
"  rrand INT         Reads at random positions. The maximum distance between\n"
"                    reads is INT blocks. If 0, reads are completely random.\n"
"  wrand INT         (See rrand.)\n"
"  rrev              Read blocks in reverse.\n"
"  wrev              Write blocks in reverse.\n"
"ARGUMENT FORMATS:\n"
"  PRIO  ^(IDLE|NORM[0-7]|RT[0-7])$     See ionice(1) for explanation.\n"
"  SIZE  ^[0-9]+(|k|M)B?$               If no units, bytes assumed.\n"
"  TIME  ^FLOAT(|us|ms|s|m|min|h|hr|loops)$   If no units, seconds assumed.\n"
"                                       'loops' unit valid only for --gap.\n"
	    , prog, prog);
}

static inline bool
timespec_zero(const struct timespec *a)
{
	return !a->tv_sec && !a->tv_nsec;
}

static inline int
timespec_cmp(const struct timespec *a, const struct timespec *b)
{
	return
	    a->tv_sec < b->tv_sec ? -1 :
	    a->tv_sec > b->tv_sec ? 1 :
	    a->tv_nsec < b->tv_nsec ? -1 :
	    a->tv_nsec > b->tv_nsec ? 1 : 0;
}

/* open the cmd file, returning the fd or NULL; prints error */
static int
write_cmdfile(const char *dataprefix, char **argv)
{
	char		*buf = 0;
	char		*join;
	size_t		len;
	FILE		*fp = 0;
	int		fd = -1;
	int		retval;
	int		status;

	len = strlen(dataprefix);
	if (dataprefix[len-1] == '/')
		/* "$PREFIX/cmd" (but the '/' is implicit) */
		join = "";
	else {
		/* "$PREFIX-cmd" */
		join = "-";
		len++;
	}
	/* enough room for 'dataprefix', 'join', "cmd" */
	if (!(buf = malloc(len + 4))) {
		fprintf(stderr, "%s: out of memory\n", prog);
		status = -1;
		goto end;
	}

	retval = snprintf(buf, len + 4, "%s%scmd", dataprefix, join);
	assert(retval < len + 4);

	/* open for truncated writing */
	fd = open(buf, O_WRONLY | O_CREAT | O_TRUNC,
	    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		fprintf(stderr, "%s: open failed to create cmd file: %s\n",
		    prog, strerror(errno));
		status = -1;
		goto end;
	}

	if (!(fp = fdopen(fd, "w"))) {
		fprintf(stderr, "%s: fdopen failed for cmd file: %s\n",
		    prog, strerror(errno));
		status = -1;
		goto end;
	}
	fd = -1;

	do {
		char suffix = argv[1] ? ' ' : '\n';
		if (fprintf(fp, "%s%c", *argv, suffix) < 0) {
			fprintf(stderr, "%s: fprintf error: %s\n",
			    prog, strerror(errno));
			status = -1;
			goto end;
		}
	} while (*++argv);

	status = 0;
end:
	if (fp) fclose(fp);
	if (fd != -1) close(fd);
	if ((fp || fd != -1) && status) unlink(buf);
	if (buf) free(buf);
	return status;
}

int
main(int argc, char **argv)
{
	struct timespec		duration;
	struct proc_spec	*specs = 0;
	struct proc_spec	*which = 0;
	int			status;
	uint16_t		num_procs = -1;
	bool			all = false;
	bool			first;
	bool			readahead = true;

	prog = *argv;
	duration.tv_sec = 10;
	duration.tv_nsec = 0;

	/* for all long options, getopt_long() will return a positive
	 * non-ascii value */
	enum { OPT_DURATION=0x80, OPT_END, OPT_GAP, OPT_MMAP, OPT_PRIO,
	    OPT_RAPID, OPT_SPINS, OPT_START };
	struct option longopts[] = {
	    {	"duration", 1, 0, OPT_DURATION },
	    {	"end", 1, 0, OPT_END },
	    {	"gap", 1, 0, OPT_GAP },
	    {	"help", 0, 0, 'h' },
	    {   "mmap", 0, 0, OPT_MMAP },
	    {	"prio", 1, 0, OPT_PRIO },
	    {	"rapid", 1, 0, OPT_RAPID },
	    {   "spins", 1, 0, OPT_SPINS },
	    {	"start", 1, 0, OPT_START },
	    {	0, 0, 0, 0 }
	};

	first = true;
	for (;;) {
		/* it isn't standardized, but glibc and the BSDs all
		 * have it */
		int opt = getopt_long(argc, argv, "+:ahn:o:t:b:f:w:",
		    longopts, 0);
		if (opt == -1) {
			if (argv[optind]) {
				fprintf(stderr,
				    "%s: stray argument on command line\n",
				    prog);
				status = 1;
				goto end;
			}
			break;
		}

		if (opt == 'h') {
			usage();
			status = 0;
			goto end;
		} else if (opt == OPT_SPINS) {
			struct timespec	    tv;
			uint64_t	    spins;
			if (parse_time(optarg, &tv) != 0) {
				fprintf(stderr,
				    "%s: bad argument for --spins\n",
				    prog);
				status = 1;
				goto end;
			}

			spins = required_spins(&tv);
			printf("%llu\n", (long long unsigned)spins);
			status = 0;
			goto end;
		}

		/* ignore [:?]; they are handled later */
		if (first) {
			switch (opt) {
			case 'n':
			case ':':
			case '?':
				break;
			default:
				fprintf(stderr,
				    "%s: first argument must be -n\n", prog);
				status = 1;
				goto end;
			}
			first = false;
		}

		/* -a or -t# must precede process-specific options */
		if (!(which || all)) {
			switch (opt) {
			case 'b':
			case 'f':
			case 'w':
			case OPT_END:
			case OPT_GAP:
			case OPT_MMAP:
			case OPT_PRIO:
			case OPT_RAPID:
			case OPT_START:
				fprintf(stderr,
				    "%s: select a process before using "
				    "a process-specific option\n", prog);
				status = 1;
				goto end;
			}
		}

		switch (opt) {

		/* global options */

		case 'a': /* apply options to [a]ll processes */
			which = 0;
			all = true;
			break;
		case OPT_DURATION:
			if (parse_time(optarg, &duration) != 0) {
				fprintf(stderr,
				    "%s: bad argument for --duration\n",
				    prog);
				status = 1;
				goto end;
			}
			break;
		case 'n': /* [n]umber processes */
		{
			long	val;
			char	*end;

			if (specs) {
				fprintf(stderr,
				    "%s: -n may only be specified once\n",
				    prog);
				status = 1;
				goto end;
			}

			errno = 0;
			val = strtol(optarg, &end, 10);
			if (errno || *end != '\0' ||
			    val < 1 || val > 0xffff) {
				fprintf(stderr, "%s: bad argument for -n\n",
				    prog);
				status = 1;
				goto end;
			}
			num_procs = val;

			/* allocate, construct */
			specs = malloc(sizeof(struct proc_spec) * num_procs);
			if (!specs) {
				fprintf(stderr, "%s: out of memory\n", prog);
				status = 1;
				goto end;
			}
			for (uint16_t id = 0; id < num_procs; id++)
				proc_spec_init(specs + id, id);

			break;
		}
		case 'o': /* [o]utput */
		{
			for (uint16_t id = 0; id < num_procs; id++)
				if (proc_spec_setdataprefix(
				    specs + id, optarg) == -1) {
					fprintf(stderr, "%s: out of memory\n",
					    prog);
					status = 1;
					goto end;
				}
			break;
		}
		case 't': /* proc ([t]hread) to apply options to */
		{
			long	val;
			char	*end;
			errno = 0;
			val = strtol(optarg, &end, 10);
			if (errno || *end != '\0' ||
			    val < 0 || val >= num_procs) {
				fprintf(stderr, "%s: bad argument for -t\n",
				    prog);
				status = 1;
				goto end;
			}

			which = specs + val;
			all = false;
			break;
		}

		/* per-process options */

		case 'b': /* [b]lock size */
		{
			size_t sz = 0;

			if (parse_size(optarg, &sz) == -1) {
				fprintf(stderr, "%s: bad argument for -b\n",
				    prog);
				status = 1;
				goto end;
			}

			if (all)
				for (uint16_t id = 0; id < num_procs; id++)
					specs[id].sz = sz;
			else
				which->sz = sz;
			break;
		}
		case OPT_END:
		{
			struct timespec	tv;

			if (parse_time(optarg, &tv) != 0) {
				fprintf(stderr,
				    "%s: bad argument for --end\n", prog);
				status = 1;
				goto end;
			}

			if (all)
				for (uint16_t id = 0; id < num_procs; id++)
					specs[id].finish = tv;
			else
				which->finish = tv;
			break;
		}
		case 'f': /* test [f]ile */
		{
			if (all && num_procs > 1)
				printf("You are using a common test file; "
				    "this is probably a bad idea.\n");
			if (all) {
				for (uint16_t id = 0; id < num_procs; id++)
					if (proc_spec_settestpath(
					    specs + id, optarg) == -1) {
						fprintf(stderr,
						    "%s: out of memory\n",
						    prog);
						status = 1;
						goto end;
					}
			} else if (proc_spec_settestpath(
			    which, optarg) == -1) {
				fprintf(stderr, "%s: out of memory\n", prog);
				status = 1;
				goto end;
			}
			break;
		}
		case OPT_GAP:
		{
			struct timespec	tv;
			uint64_t	spins;
			int		retval;

			if ((retval = parse_time(optarg, &tv)) == -1) {
				fprintf(stderr,
				    "%s: bad argument for --gap\n", prog);
				status = 1;
				goto end;
			}

			if (retval)
				spins = (uint64_t)tv.tv_sec * 1000000000ULL +
				    tv.tv_nsec;
			else
				spins = required_spins(&tv);

			if (all)
				for (uint16_t id = 0; id < num_procs; id++)
					specs[id].spins = spins;
			else
				which->spins = spins;
			break;
		}
		case OPT_MMAP:
			if (all)
				for (uint16_t id = 0; id < num_procs; ++id)
					specs[id].mmap = true;
			else
				which->mmap = true;
			break;
		case OPT_PRIO:
		{
			enum io_priority	cls = 0;
			uint8_t			lvl = 0;

			if (parse_prio(optarg, &cls, &lvl) == -1) {
				fprintf(stderr,
				    "%s: bad argument for --prio\n", prog);
				status = 1;
				goto end;
			}

			if (all)
				for (uint16_t id = 0; id < num_procs; ++id) {
					specs[id].prio_cls = cls;
					specs[id].prio_lvl = lvl;
				}
			else {
				which->prio_cls = cls;
				which->prio_lvl = lvl;
			}
			break;
		}
		case OPT_RAPID:
		{
			long	val;
			char	*end;

			errno = 0;
			val = strtol(optarg, &end, 10);
			if (errno || *end != '\0' ||
			    val < 1 || val > 0xffff) {
				fprintf(stderr,
				    "%s: bad argument for --rapid\n", prog);
				status = 1;
				goto end;
			}

			if (all)
				for (uint16_t id = 0; id < num_procs; ++id)
					specs[id].rapid = val;
			else
				which->rapid = 1;

			break;
		}
		case OPT_START:
		{
			struct timespec	tv;

			if (parse_time(optarg, &tv) != 0) {
				fprintf(stderr,
				    "%s: bad argument for --start\n", prog);
				status = 1;
				goto end;
			}

			if (all)
				for (uint16_t id = 0; id < num_procs; id++)
					specs[id].start = tv;
			else
				which->start = tv;
			break;
		}
		case 'w': /* [w]orkload */
		{
			/* XXX avoid unfounded warning */
			struct workload	wl = { 0, 0 };

			if (parse_workload(argv, &wl) == -1) {
				status = 1;
				goto end;
			}

			if (all)
				for (uint16_t id = 0; id < num_procs; id++)
					specs[id].wl = wl;
			else
				which->wl = wl;
			break;
		}
		case ':': /* missing[:] an argument */
		case '?': /* what does this option mean[?] */
			status = 1;
			goto end;
		default:
			assert(0);
		}
	}

	if (!specs) {
		usage();
		status = 1;
		goto end;
	}

	if (!specs[0].dataprefix) {
		fprintf(stderr, "%s: missing mandatory option -o\n", prog);
		status = 1;
		goto end;
	}

	/* for each process, resolve start and end times using 'duration' */
	for (uint16_t id = 0; id < num_procs; id++) {
		if (timespec_zero(&specs[id].finish) ||
		    timespec_cmp(&specs[id].finish, &duration) > 0)
			specs[id].finish = duration;

		if (timespec_cmp(&specs[id].start, &specs[id].finish) >= 0) {
			fprintf(stderr,
			    "%s: process %hu has a non-positive runtime\n",
			    prog, id);
			status = 1;
			goto end;
		}

		if (!specs[id].testpath) {
			fprintf(stderr,
			    "%s: process %hu needs a test file\n",
			    prog, id);
			status = 1;
			goto end;
		}
	}

	if (write_cmdfile(specs[0].dataprefix, argv) == -1) {
		status = 1;
		goto end;
	}

	if (drop_cache() == -1) {
		status = 1;
		goto end;
	}

	if (set_all_readahead(specs, num_procs, 0) == -1) {
		status = 1;
		goto end;
	}
	readahead = false;

	if (run_processes(specs, num_procs) == -1) {
		status = 1;
		goto end;
	}

	status = 0;
end:
	if (!readahead)
		set_all_readahead(specs, num_procs, 256);
	if (specs) {
		for (uint16_t id = 0; id < num_procs; id++)
			proc_spec_destroy(specs + id);
		free(specs);
	}

	return status;
}
