#include <sys/types.h>
#include <linux/fs.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "clock.h"
#include "device_ops.h"
#include "disk_stats.h"
#include "signames.h"

/* egg + wrapper = eggroll ... yes, I am a genius */

const char	*prog;

static void
usage()
{
	printf(
"usage: %s DEVICE PROG [ARGS...]\n"
"\n"
"Collect timing and disk utilization statistics for a command.\n"
"DEVICE   The device to collect statistics of (e.g. sda0)\n"
	    , prog);
}

static char *
basename(const char *path)
{
	char *s = strrchr(path, '/');
	return s ? s : (char *)path;
}

static int16_t
sector_size(const char *device)
{
	char	buf[PATH_MAX];
	int	fd;
	int	retval;
	int	sz;

	retval = snprintf(buf, PATH_MAX, "/dev/%s", basename(device));
	assert(retval > 0 && retval < PATH_MAX);

	if ((fd = open(buf, O_RDONLY | O_NONBLOCK)) == -1)
		return -1;

	sz = 0;
	if (ioctl(fd, BLKSSZGET, &sz) == -1) {
		long	save = errno;
		close(fd);
		errno = save;
		return -1;
	}

	close(fd);
	return sz;
}

static int
measured_exec(const char *device, char *const *args)
{
	struct disk_stats	stats0;
	struct disk_stats	stats1;
	char			stat_path[PATH_MAX];
	uint64_t		time0;
	uint64_t		time1;
	uint32_t		bytes;
	uint32_t		milli_io;
	float			milli_tot;
	float			seconds;
	int			status;
	pid_t			pid;
	uint16_t		sector_sz;

	if (find_device_stat(device, stat_path) == -1)
		return -1;

	if ((sector_sz = sector_size(device)) == -1) {
		fprintf(stderr, "%s: failed to determine sector size: %s\n",
		    prog, strerror(errno));
		return -1;
	}

	if (drop_cache() == -1)
		return -1;

	/* collect opening data */
	if (read_disk_stats(prog, stat_path, &stats0) == -1)
		return -1;
	time0 = clock_monotonic();

	/* start child */
	if ((pid = fork()) == -1) {
	fprintf(stderr, "%s: fork failed: %s\n",
		    prog, strerror(errno));
		return -1;
	} else if (!pid) {
		execvp(*args, args);
		fprintf(stderr, "%s: exec failed: %s\n",
		    prog, strerror(errno));
		exit(0xff);
	}

	/* wait for child */
	status = 0;
	for (;;) {
		if (waitpid(pid, &status, 0) == -1) {
			if (errno == EINTR) continue;
			fprintf(stderr, "%s: wait failed: %s\n",
			    prog, strerror(errno));
			return -1;
		}
		break;
	}

	/* collect closing data */
	if (read_disk_stats(prog, stat_path, &stats1) == -1)
		return -1;
	time1 = clock_monotonic();

	/* check for child error */
	if (WIFSIGNALED(status)) {
		fprintf(stderr, "%s: child terminated by signal %s\n",
		    prog, signame(WTERMSIG(status)));
		return -1;
	} else if (WEXITSTATUS(status)) {
		if (WEXITSTATUS(status) != 0xff)
			fprintf(stderr, "%s: child exited with %d\n",
			    prog, WEXITSTATUS(status));
		return -1;
	}

	/* process data */
	bytes = ((stats1.sectors_read + stats1.sectors_written) -
	    (stats0.sectors_read + stats0.sectors_written)) * sector_sz;

	milli_io = stats1.milli_io - stats0.milli_io;
	milli_tot = (time1 - time0) / 1000000.0f;
	seconds = (time1 - time0) / 1000000000.0f;

	/* print data */
	printf("RATE_KBPS\t%.10e\n", bytes / (seconds * 1024));
	printf("RATE_MBPS\t%.10e\n", bytes / (seconds * 1048576));
	printf("TIME_S\t\t%.10e\n", seconds);
	printf("UTIL\t\t%.10e\n", milli_io / milli_tot);

	return 0;
}

int main(int argc, char **argv)
{
	prog = *argv;

	if (argc > 1 && (!strcmp(argv[1], "-h") ||
	    !strcmp(argv[1], "--help"))) {
		usage();
		return 0;
	} else if (argc < 3) {
		usage();
		return 1;
	}

	if (measured_exec(argv[1], argv + 2) == -1)
		return 1;
	return 0;
}
