#include <errno.h>
#include <getopt.h>
#include <pcre.h>
#include <stdio.h>

#include "opt_parse.h"

extern char	*prog;

static int	check_missing_argument(char **);

/* parse size, return 0 or -1 */
int
parse_size(const char *s, size_t *sz)
{
	int		ovector[9];
	const char	*re_err;
	int		re_errpos;
	pcre		*regex = 0;
	int		stringcount;
	const char	**subs = 0;

	long		val;
	int		retval;
	int		status;

	if (!(regex = pcre_compile("^ (\\d+) \\h* (|k|M) B? $",
	    PCRE_EXTENDED, &re_err, &re_errpos, 0))) {
		fprintf(stderr,
		    "%s: size regex compilation failed at %d: %s\n",
		    prog, re_errpos, re_err);
		return 1;
	}

	stringcount = pcre_exec(regex, 0, s, strlen(s), 0, 0, ovector, 9);
	if (stringcount == PCRE_ERROR_NOMATCH) {
		status = -1;
		goto end;
	} else if (stringcount < 0) {
		fprintf(stderr, "%s: pcre parse error: %d\n",
		    prog, stringcount);
		status = -1;
		goto end;
	}

	retval = pcre_get_substring_list(s, ovector, stringcount, &subs);
	if (retval < 0) {
		fprintf(stderr,
		    "%s: pcre failed to get substrings for size string: %d\n",
		    prog, retval);
		status = -1;
		goto end;
	}

	errno = 0;
	val = strtol(subs[1], 0, 10);
	if (errno)
		return -1;

	if (*subs[2] == '\0')
		*sz = val;
	else if (!strcmp(subs[2], "k")) {
		size_t sz_ = (size_t)val << 10;
		if (sz_ >> 10 != val) { /* overflow check */
			status = -1;
			goto end;
		}
		*sz = sz_;
	} else /* if (!strcmp(subs[2], "M")) */ {
		size_t sz_ = (size_t)val << 20; /* overflow check */
		if (sz_ >> 20 != val) {
			status = -1;
			goto end;
		}
		*sz = sz_;
	}

	status = 0;
end:
	if (subs) pcre_free_substring_list(subs);
	if (regex) pcre_free(regex);
	return status;
}

/* parse /^ \d+ (|us|ms|s|m|min|h|hr|loops) $/x
 *
 * \retval 1	'tv' has time in terms of loops
 * \retval 0	'tv' has time in terms of seconds
 * \retval -1	Error */
int
parse_time(const char *s, struct timespec *tv)
{
	int		ovector[9];
	const char	*re_err;
	int		re_errpos;
	pcre		*regex = 0;
	int		stringcount;
	const char	**subs = 0;

	uint64_t	nano;
	double		val;
	int		retval;
	int		status;

	if (!(regex = pcre_compile(
	    "^ ( (?:\\d*\\.\\d+ | \\d+\\.?) (?:e[+-]\\d+)? ) \\h* "
	    "(|us|ms|s|m|min|h|hr|loops) $",
	    PCRE_EXTENDED, &re_err, &re_errpos, 0))) {
		fprintf(stderr,
		    "%s: time regex compilation failed at %d: %s\n",
		    prog, re_errpos, re_err);
		return 1;
	}

	stringcount = pcre_exec(regex, 0, s, strlen(s), 0, 0, ovector, 9);
	if (stringcount == PCRE_ERROR_NOMATCH) {
		status = -1;
		goto end;
	} else if (stringcount < 0) {
		fprintf(stderr, "%s: pcre parse error: %d\n",
		    prog, stringcount);
		status = -1;
		goto end;
	}

	retval = pcre_get_substring_list(s, ovector, stringcount, &subs);
	if (retval < 0) {
		fprintf(stderr,
		    "%s: pcre failed to get substrings for time string: %d\n",
		    prog, retval);
		status = -1;
		goto end;
	}

	errno = 0;
	val = strtod(subs[1], 0);
	if (errno) {
		status = -1;
		goto end;
	}

	status = 0;
	if (*subs[2] == '\0' || strcmp(subs[2], "s") == 0)
		nano = val * 1000000000ULL;
	else if (strcmp(subs[2], "ms") == 0)
		nano = val * 1000000;
	else if (strcmp(subs[2], "us") == 0)
		nano = val * 1000;
	else if (strcmp(subs[2], "m") == 0 || strcmp(subs[2], "min") == 0)
		nano = val * 60 * 1000000000ULL;
	else if (strcmp(subs[2], "h") == 0 || strcmp(subs[2], "hr") == 0)
		nano = val * 3600 * 1000000000ULL;
	else /*if (strcmp(subs[2], "loops") == 0)*/ {
		status = 1;
		nano = val;
	}

	tv->tv_sec = nano / 1000000000ULL;
	tv->tv_nsec = nano % 1000000000ULL;
end:
	if (subs) pcre_free_substring_list(subs);
	if (regex) pcre_free(regex);
	return status;
}

/* parse /^( RT[0-7] | IDLE | [0-7] )$/x, return 0 or -1 */
int
parse_prio(const char *s, enum io_priority *out_prio_cls,
    uint8_t *out_prio_lvl)
{
	int		ovector[9];
	const char	*re_err;
	int		re_errpos;
	pcre		*regex = 0;
	int		stringcount;
	const char	**subs = 0;

	int		retval;
	int		status;

	if (!(regex = pcre_compile(
	    "^(?: (RT) \\h* ([0-7]) | (IDLE) | ([0-7]) )$",
	    PCRE_EXTENDED, &re_err, &re_errpos, 0))) {
		fprintf(stderr,
		    "%s: prio regex compilation failed at %d: %s\n",
		    prog, re_errpos, re_err);
		return 1;
	}

	stringcount = pcre_exec(regex, 0, s, strlen(s), 0, 0, ovector, 9);
	if (stringcount == PCRE_ERROR_NOMATCH) {
		status = -1;
		goto end;
	} else if (stringcount < 0) {
		fprintf(stderr, "%s: pcre parse error: %d\n",
		    prog, stringcount);
		status = -1;
		goto end;
	}

	retval = pcre_get_substring_list(s, ovector, stringcount, &subs);
	if (retval < 0) {
		fprintf(stderr,
		    "%s: pcre failed to get string for prio string: %d\n",
		    prog, retval);
		status = -1;
		goto end;
	}

	if (!strcmp(subs[1], "RT")) {
		*out_prio_cls = PRIO_RT;
		*out_prio_lvl = strtol(subs[2], 0, 10);
	} else if (!strcmp(subs[1], "IDLE")) {
		*out_prio_cls = PRIO_IDLE;
		*out_prio_lvl = 0;
	} else {
		*out_prio_cls = PRIO_NORM;
		*out_prio_lvl = strtol(subs[1], 0, 10);
	}

	status = 0;
end:
	if (subs) pcre_free_substring_list(subs);
	if (regex) pcre_free(regex);
	return status;
}

int
parse_workload(char **argv, struct workload *wl)
{
	enum workload_type	type;
	/* unnecessary initialization */
	size_t			dist = 0;

	/* note: argv[optind] contains optarg */
	if (strcmp(optarg, "rseq") == 0)
		type = WL_RSEQ;
	else if (strcmp(optarg, "wseq") == 0)
		type = WL_WSEQ;
	else if (strcmp(optarg, "rrand") == 0)
		type = WL_RRAND;
	else if (strcmp(optarg, "wrand") == 0)
		type = WL_WRAND;
	else if (strcmp(optarg, "rrev") == 0)
		type = WL_RREV;
	else if (strcmp(optarg, "wrev") == 0)
		type = WL_WREV;
	else {
		fprintf(stderr, "%s: bad workload type: %s\n", prog, optarg);
		return -1;
	}

	switch (type) {
	case WL_RRAND:
	case WL_WRAND:
	{
		long	val;
		char	*end;

		if (check_missing_argument(argv) == -1) return -1;

		errno = 0;
		val = strtol(argv[optind], &end, 10);
		if (errno || *end != '\0' || val < 0) {
			fprintf(stderr, "%s: bad argument for %s: %s\n",
			    prog, optarg, argv[optind]);
			return -1;
		}
		dist = val;
		optind++;
	}
	case WL_RSEQ:
	case WL_WSEQ:
	case WL_RREV:
	case WL_WREV:
		break;
	}

	/* write into the wl structure */
	wl->type = type;
	switch (type) {
	case WL_RSEQ:
	case WL_WSEQ:
	case WL_RREV:
	case WL_WREV:
		wl->dist = 0;
		break;
	case WL_RRAND:
	case WL_WRAND:
		wl->dist = dist;
	}
	return 0;
}

/* if there are no remaining arguments, prints a message (relying on global
 * variables) and returns -1 */
static int
check_missing_argument(char **argv)
{
	if (!argv[optind]) {
		fprintf(stderr, "%s: missing argument for %s\n",
		    prog, optarg);
		return -1;
	}
	return 0;
}
