#include <assert.h>

#include "clock.h"
#include "device_ops.h"
#include "disk_stats.h"
#include "proc_spec.h"

extern const char	*prog;

/* 100M spins; on 2.66GHz, takes ~0.301 seconds */
/* time to execute spin_for(SPINS): timed_spins */
static const uint32_t	SPINS = 100000000;
static uint64_t		timed_spins = 0; /* (nanoseconds) */

static void	init_timed_spins();


int
proc_spec_settestpath(struct proc_spec *spec, const char *path)
{
	char	buf[PATH_MAX];
	size_t	len;
	char	*device = 0;
	char	*mountpoint = 0;
	char	*stat = 0;
	char	*testpath = 0;
	int	status;

	/* copy path */
	len = strlen(path);
	testpath = malloc(len + 1);
	if (!testpath) {
		status = -1;
		goto end;
	}
	memcpy(testpath, path, len + 1);

	/* determine, copy mountpoint */
	if (path_to_mountpoint(path, buf) == -1) {
		status = -1;
		goto end;
	}
	len = strlen(buf);
	mountpoint = malloc(len + 1);
	if (!mountpoint) {
		status = -1;
		goto end;
	}
	memcpy(mountpoint, buf, len + 1);

	/* determine, copy device */
	if (mountpoint_to_device(mountpoint, buf) == -1) {
		status = -1;
		goto end;
	}
	len = strlen(buf);
	device = malloc(len + 1);
	if (!device) {
		status = -1;
		goto end;
	}
	memcpy(device, buf, len + 1);

	/* determine, copy stat */
	if (find_device_stat(device, buf) == -1) {
		status = -1;
		goto end;
	}
	len = strlen(buf);
	stat = malloc(len + 1);
	if (!stat) {
		status = -1;
		goto end;
	}
	memcpy(stat, buf, len + 1);

	/* move pointers into struct */
	if (spec->testpath)
		free(spec->testpath);
	if (spec->path_mountpoint)
		free(spec->path_mountpoint);
	if (spec->path_device)
		free(spec->path_device);
	if (spec->path_stat)
		free(spec->path_stat);
	spec->testpath = testpath;
	spec->path_mountpoint = mountpoint;
	spec->path_device = device;
	spec->path_stat = stat;

	status = 0;
end:
	if (status) {
		if (stat) free(stat);
		if (device) free(device);
		if (mountpoint) free(mountpoint);
		if (testpath) free(testpath);
	}
	return 0;
}

int
proc_spec_setdataprefix(struct proc_spec *spec, const char *path)
{
	size_t	len;
	char	*n;

	len = strlen(path);
	n = malloc(len + 1);
	if (!n)
		return -1;
	memcpy(n, path, len + 1);

	if (spec->dataprefix)
		free(spec->dataprefix);
	spec->dataprefix = n;
	return 0;
}

/* initialize timed_spins */
static void
init_timed_spins()
{
	uint64_t	start;

	start = clock_monotonic();
	spin_for(SPINS);
	/* timespec's members are always signed, so no cast errors here */
	timed_spins = clock_monotonic() - start;
}

uint64_t
required_spins(const struct timespec *tv)
{
	const unsigned REFINEMENTS = 2;
	long double translated;
	long double required;

	if (!timed_spins) init_timed_spins();

	translated = tv->tv_sec * 1000000000ULL + tv->tv_nsec;
	if (!translated) return 0;
	/* SPINS * (translated / (timed_spins / 1000)) */
	required = SPINS * translated / timed_spins;
	assert(required > 0);

	for (unsigned i = 0; i < REFINEMENTS; i++) {
		uint64_t period;
		uint64_t start;

		start = clock_monotonic();
		spin_for((uint64_t)required);
		period = clock_monotonic() - start;

		required = required * translated / period;
		assert(required > 0);
	}

	return (uint64_t)required;
}

void
spin_for(uint64_t spins)
{
	volatile static uint64_t t;
	t = 0;
	for (uint64_t i = 0; i < spins; i++)
		t *= i;
}
