#include <sys/types.h>
#include <sys/stat.h>

#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <limits.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pcre.h>

const char *CPU_DIR = "/sys/devices/system/cpu";
const char *CPU_GOVERNOR_FORMAT =
    "/sys/devices/system/cpu/cpu%d/cpufreq/scaling_governor";
const char *CPU_GOVERNORS_FORMAT =
    "/sys/devices/system/cpu/cpu%d/cpufreq/scaling_available_governors";

const char *prog;

void usage()
{
	printf(
"usage: %s [GOVERNOR]\n"
"Manipulate the CPU governor.\n"
"With no arguments, prints available governors and current settings.\n",
	    prog);
}

/* read in a file, copy all up to first \n into alloc'd string and return */
char *
alloc_read(const char *path)
{
	char	buf[160];
	char	*alloc;
	size_t	bytes;
	FILE	*fp;
	size_t	len;

	/* open first for reading */
	if (!(fp = fopen(path, "r"))) {
		fprintf(stderr, "%s: fopen(read) failed for `%s': %s\n",
		    prog, path, strerror(errno));
		return 0;
	}

	/* read current value */
	bytes = fread(buf, 1, sizeof(buf)-1, fp);
	if (bytes == sizeof(buf)-1 && !feof(fp)) {
		fprintf(stderr, "%s: file `%s' too long\n", prog, path);
		fclose(fp);
		return 0;
	}
	if (ferror(fp)) {
		fprintf(stderr,
		    "%s: fread failed for `%s': %s\n",
		    prog, path, strerror(errno));
		fclose(fp);
		return 0;
	}
	fclose(fp);
	buf[bytes] = '\0';
	len = bytes;
	for (size_t i = 0; i < len; i++)
		if (buf[i] == '\n') {
			buf[i] = '\0';
			len = i;
			break;
		}

	/* copy current setting to new buffer */
	if (!(alloc = malloc(len + 1))) {
		perror(prog);
		return 0;
	}
	memcpy(alloc, buf, len + 1);
	return alloc;
}

/* return the number of cpu\d+ directories in /sys/devices/system/cpu */
int
num_cpus()
{
	DIR		*dir;
	struct dirent	*dirent;
	const char	*errstr;
	pcre		*regex;
	int		count;
	int		errpos;

	if (!(regex = pcre_compile("^ cpu \\d+ $",
	    PCRE_EXTENDED, &errstr, &errpos, 0))) {
		fprintf(stderr,
		    "%s: size regex compilation failed at %d: %s\n",
		    prog, errpos, errstr);
		return -1;
	}

	dir = opendir(CPU_DIR);
	if (!dir) {
		fprintf(stderr, "%s: opendir failed for `%s': %s\n",
		    prog, CPU_DIR, strerror(errno));
		pcre_free(regex);
		return -1;
	}

	count = 0;
	while ((dirent = readdir(dir))) {
		int	retval;

		retval = pcre_exec(regex, 0, dirent->d_name,
		    strlen(dirent->d_name), 0, 0, 0, 0);
		if (retval == PCRE_ERROR_NOMATCH);
		else if (retval < 0) {
			fprintf(stderr, "%s: pcre parse error: %d\n",
			    prog, retval);
			closedir(dir);
			pcre_free(regex);
			return -1;
		} else
			count++;
	}
	assert(count);

	closedir(dir);
	pcre_free(regex);
	return count;
}

bool
cpu_available(uint8_t cpu)
{
	struct stat	st;
	char		pathbuf[PATH_MAX];
	int		len;

	len = snprintf(pathbuf, PATH_MAX, CPU_GOVERNOR_FORMAT, cpu);
	assert(len > 0 && len < PATH_MAX);
	return stat(pathbuf, &st) == 0;
}

/* print the current setting much like set_sched;
 * i.e. 'cpuN: notused [used] notused' */
int
print_current(uint8_t cpus)
{
	char	pathbuf[PATH_MAX];
	char	*all = 0;
	char	*current = 0;
	int	status;
	bool	all_unavail = true;

	for (uint8_t i = 0; i < cpus; i++) {
		char		*saveptr;
		char		*start;
		char		*tok;
		int		len;

		if (!cpu_available(i)) {
			printf("cpu%d: not available\n", i);
			continue;
		}
		all_unavail = false;

		len = snprintf(pathbuf, PATH_MAX, CPU_GOVERNOR_FORMAT, i);
		assert(len > 0 && len < PATH_MAX);
		if (!(current = alloc_read(pathbuf))) {
			status = -1;
			goto end;
		}

		len = snprintf(pathbuf, PATH_MAX, CPU_GOVERNORS_FORMAT, i);
		assert(len > 0 && len < PATH_MAX);
		if (!(all = alloc_read(pathbuf))) {
			status = -1;
			goto end;
		}

		printf("cpu%d:", i);
		start = all;
		while ((tok = strtok_r(start, " \t\n", &saveptr))) {
			if (!strcmp(tok, current))
				printf(" [%s]", tok);
			else
				printf(" %s", tok);

			start = 0;
		}
		printf("\n");

		free(all);
		all = 0;
		free(current);
		current = 0;
	}

	if (all_unavail) {
		fprintf(stderr, "%s: no CPUs available\n", prog);
		status = 1;
		goto end;
	}

	status = 0;
end:
	if (all) free(all);
	if (current) free(current);
	return status;
}

/* set governor for all CPUs */
int
set_governor(const char *governor, int cpus)
{
	char	pathbuf[PATH_MAX];
	FILE	*fp = 0;
	int	len;
	char	*new = 0;
	char	*next = 0;
	char	*old = 0;
	int	status;
	bool	all_unavail;

	len = strlen(governor);
	if (!(next = malloc(len + 2))) {
		perror(prog);
		status = -1;
		goto end;
	}
	memcpy(next, governor, len);
	next[len] = '\n';
	next[len + 1] = '\0';

	all_unavail = true;
	for (uint8_t i = 0; i < cpus; i++) {
		int	retval;

		if (!cpu_available(i)) {
			printf("cpu%d: not available\n", i);
			continue;
		}
		all_unavail = false;

		retval = snprintf(pathbuf, PATH_MAX, CPU_GOVERNOR_FORMAT, i);
		assert(retval > 0 && retval < PATH_MAX);
		if (!(old = alloc_read(pathbuf))) {
			status = -1;
			goto end;
		}

		if (!(fp = fopen(pathbuf, "w"))) {
			fprintf(stderr, "%s: open failed for `%s': %s\n",
			    prog, pathbuf, strerror(errno));
			status = -1;
			goto end;
		}
		if (fputs(next, fp) < 0) {
			fprintf(stderr, "%s: write failed for `%s': %s\n",
			    prog, pathbuf, strerror(errno));
			status = -1;
			goto end;
		}
		fclose(fp);
		fp = 0;

		if (!(new = alloc_read(pathbuf))) {
			status = -1;
			goto end;
		}

		/* if governor,new are different up to len bytes, or if
		 * the new value is longer */
		if (strncmp(governor, new, len) || new[len]) {
			fprintf(stderr, "%s: change for cpu%hhu failed\n",
			    prog, i);
			status = 1;
			goto end;
		}

		printf("cpu%hhu: %s => %s\n", i, old, new);

		free(new);
		new = 0;
		free(old);
		old = 0;
	}

	if (all_unavail) {
		fprintf(stderr, "%s: no CPUs available\n", prog);
		status = 1;
		goto end;
	}

	status = 0;
end:
	if (new) free(new);
	if (fp) fclose(fp);
	if (old) free(old);
	if (next) free(next);
	return status;
}

int main(int argc, char **argv)
{
	int		cpus;

	prog = *argv;

	if (argc > 3) {
		usage();
		return 1;
	}
	if (argc == 2 && (!strcmp(argv[1], "-h") ||
	    !strcmp(argv[1], "--help"))) {
		usage();
		return 0;
	}

	cpus = num_cpus();
	if (cpus < 0)
		return 1;

	if (argc < 2) {
		if (print_current(cpus) == -1)
			return 1;
		printf("\n");
		usage();
		return 0;
	}

	if (set_governor(argv[1], cpus) == -1)
		return -1;
	else
		return 0;
}
