#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

const char *BLOCK_DIR = "/sys/block";
const char *SCHEDULER_PATH_FORMAT = "/sys/block/%s/queue/scheduler";

const char *prog;

void usage()
{
	printf(
"usage: %s BLK_DEV [SCHEDULER]\n"
"\n"
"BLK_DEV - a device in %s\n"
"If SCHEDULER is not supplied, available schedulers and the current setting\n"
"are displayed.\n",
	    prog, BLOCK_DIR);
}

/* read in a file, copy all up to first \n into alloc'd string and return */
char *
alloc_read(const char *path)
{
	char	buf[160];
	char	*alloc;
	size_t	bytes;
	FILE	*fp;
	size_t	len;

	/* open first for reading */
	if (!(fp = fopen(path, "r"))) {
		fprintf(stderr, "%s: fopen(read) failed: %s\n",
		    prog, strerror(errno));
		return 0;
	}

	/* read current value */
	bytes = fread(buf, 1, sizeof(buf)-1, fp);
	if (bytes == sizeof(buf)-1 && !feof(fp)) {
		fprintf(stderr, "%s: I/O scheduler file too long\n", prog);
		fclose(fp);
		return 0;
	}
	if (ferror(fp)) {
		fprintf(stderr,
		    "%s: fread failed for I/O scheduler file: %s\n",
		    prog, strerror(errno));
		fclose(fp);
		return 0;
	}
	fclose(fp);
	buf[bytes] = '\0';
	len = bytes;

	/* strip first newline */
	for (size_t i = 0; i < len; i++)
		if (buf[i] == '\n') {
			buf[i] = '\0';
			len = i;
			break;
		}

	/* copy current setting to new buffer */
	if (!(alloc = malloc(len + 1))) {
		perror(prog);
		return 0;
	}
	memcpy(alloc, buf, len + 1);
	return alloc;
}

/* find text inside first '[' and following ']', move to beginning of buffer,
 * null-terminate */
void
transform_to_current(char *str)
{
	char	*begin;
	char	*end;

	begin = strchr(str, '[');
	if (!begin) return;
	end = strchr(++begin, ']');
	if (!end) return;

	memcpy(str, begin, end - begin);
	str[end - begin] = '\0';
}

int main(int argc, char **argv)
{
	/* far more than enough room */
	char		buf[160];
	const char	*dev;
	FILE		*file = 0;
	size_t		len;
	char		*new = 0;
	char		*old = 0;
	char		*pathbuf = 0;
	int		retval;
	const char	*scheduler;
	int		status;

	prog = *argv;

	if (argc < 2 || argc > 3) {
		usage();
		return 1;
	}

	dev = argv[1];
	scheduler = argc > 2 ? argv[2] : 0;

	/* no funny business */
	if (strchr(dev, '/') || !strcmp(dev, ".") || !strcmp(dev, "..")) {
		fprintf(stderr, "%s: invalid BLK_DEV\n", prog);
		status = 1;
		goto end;
	}

	/* one more byte than needed */
	len = strlen(SCHEDULER_PATH_FORMAT) + strlen(dev);
	if (!(pathbuf = malloc(len + 1))) {
		perror(prog);
		status = 1;
		goto end;
	}
	retval = snprintf(pathbuf, len + 1, SCHEDULER_PATH_FORMAT, dev);
	assert(retval <= len);

	if (!(old = alloc_read(pathbuf))) {
		status = 1;
		goto end;
	}

	if (scheduler)
		transform_to_current(old);
	else {
		printf("%s\n", old);
		status = 0;
		goto end;
	}

	/* prepare to write */
	retval = snprintf(buf, sizeof(buf), "%s\n", scheduler);
	assert(retval >= 0);
	if (retval >= sizeof(buf)) {
		fprintf(stderr, "%s: scheduler name too long; trying to pull "
		    "a fast one, eh?\n", prog);
		status = 1;
		goto end;
	}
	len = retval;

	/* reopen the file for truncated writing */
	if (!(file = fopen(pathbuf, "w"))) {
		fprintf(stderr, "%s: fopen(write) failed: %s\n",
		    prog, strerror(errno));
		status = 1;
		goto end;
	}

	if (fwrite(buf, len, 1, file) != 1) {
		fprintf(stderr, "%s: fwrite failed: %s\n",
		    prog, strerror(errno));
		status = 1;
		goto end;
	}
	fclose(file);
	file = 0;

	if (!(new = alloc_read(pathbuf))) {
		status = 1;
		goto end;
	}
	transform_to_current(new);

	if (strcmp(scheduler, new)) {
		fprintf(stderr, "%s: change failed\n", prog);
		status = 1;
		goto end;
	}

	printf("%s: %s => %s\n", dev, old, new);

	status = 0;
end:
	if (new) free(new);
	if (old) free(old);
	if (file) fclose(file);
	if (pathbuf) free(pathbuf);
	return status;
}
