#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/stat.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "clock.h"
#include "disk_stats.h"
#include "ioprio.h"
#include "proc_spec.h"
#include "proc_stats.h"
#include "worker.h"
#include "workload.h"

/* looks like: "./egg(12)" */
static char	*name = 0;

static int	dump_stats(const struct proc_spec *, FILE *,
		    const struct proc_stats *);
static int	init_name(const struct proc_spec *);
static void	init_path_datafile(const char *, uint16_t, char[PATH_MAX]);
static FILE	*open_datafile(const char *);
static int	open_testfile(const char *, enum workload_type, bool mmap);
static int	set_priority(const struct proc_spec *);
static void	*subworker_start(void *);
static int	sem_wait(int proc_id, int semid);
static void	wakeup_handler();

static int
init_name(const struct proc_spec *spec)
{
	extern const char	*prog;

	/* largest id = 65535, 5 digits */
	size_t	len = strlen(prog) + 5 + 2;
	int	ret;

	if (!(name = malloc(len + 1))) {
		fprintf(stderr, "%s(%hu): out of memory\n", prog, spec->id);
		return -1;
	}
	ret = snprintf(name, len + 1, "%s(%hu)", prog, spec->id);
	assert(ret > 0 && ret <= len);
	return 0;
}

static int
set_priority(const struct proc_spec *spec)
{
	int class =
	    spec->prio_cls == PRIO_IDLE ? IOPRIO_CLASS_IDLE :
	    spec->prio_cls == PRIO_NORM ? IOPRIO_CLASS_BE : IOPRIO_CLASS_RT;
	int prio = IOPRIO_PRIO_VALUE(class, spec->prio_lvl);

	if (ioprio_set(IOPRIO_WHO_PROCESS, getpid(), prio) == -1) {
		fprintf(stderr, "%s: ioprio_set failed: %s\n",
		    name, strerror(errno));
		return -1;
	}
	return 0;
}

static int
sem_wait(int proc_id, int semid)
{
	struct sembuf	op;

	op.sem_num = 0;
	op.sem_op = -1;
	op.sem_flg = 0;
	if (semop(semid, &op, 1) == -1) {
		fprintf(stderr, "%s(%hu): failed to wait on semaphore: %s\n",
		    name, proc_id, strerror(errno));
		return -1;
	}
	return 0;
}

/* open the test file, returning the fd or -1; prints error */
static int
open_testfile(const char *path, enum workload_type type, bool mmap)
{
	int		fd;
	/* avoid unfounded warning with init*/
	int		flags = 0;

	switch (type) {
	case WL_RSEQ:
	case WL_RRAND:
	case WL_RREV:
		flags = O_RDONLY;
		break;
	case WL_WSEQ:
	case WL_WRAND:
	case WL_WREV:
		/* mmap() needs to be able to read the file to load it into
		 * memory */
		flags = mmap ? O_RDWR : O_WRONLY;
	}

	fd = open(path, flags, S_IRUSR | S_IWUSR);
	if (fd == -1) {
		fprintf(stderr, "%s: open failed: %s\n",
		    name, strerror(errno));
		return -1;
	}

	return fd;
}

static void
init_path_datafile(const char *dataprefix, uint16_t id, char path[PATH_MAX])
{
	const char	*join;
	size_t		len;
	int		retval;

	len = strlen(dataprefix);
	if (dataprefix[len-1] == '/')
		/* PREFIX='alpha/', ID=3 yield 'alpha/003' */
		join = "";
	else
		/* PREFIX='alpha',  ID=3 yield 'alpha-003' */
		join = "-";

	retval = snprintf(path, PATH_MAX, "%s%s%03hu.raw",
	    dataprefix, join, id);
	assert(retval > 0 && retval < PATH_MAX);
}

/* open the data file, returning the fp or NULL; prints error */
static FILE *
open_datafile(const char *path)
{
	FILE		*fp;
	int		fd;

	/* open for truncated writing */
	fd = open(path, O_WRONLY | O_CREAT | O_TRUNC,
	    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		fprintf(stderr, "%s: open failed to create data file: %s\n",
		    name, strerror(errno));
		return 0;
	}

	if (!(fp = fdopen(fd, "w"))) {
		fprintf(stderr, "%s: fdopen failed for data file: %s\n",
		    name, strerror(errno));
		close(fd);
		return 0;
	}
	return fp;
}

static int
dump_stats(const struct proc_spec *spec, FILE *fp,
    const struct proc_stats *stats)
{
	off_t		bytes = 0;
	/* absolute position in 'stats' */
	uint32_t	k;
	uint32_t	milli_io;
	float		milli_tot;
	bool		done = false;
	bool		first = true;

	k = 0;
	for (const struct ps_node *n = stats->head; n; n = n->next) {
		for (uint16_t i = 0; i < n->count; i++) {
			uint64_t t = n->data[i];
			assert(!done);

			/* I just hate printf's size specifiers; you can
			 * only specify you need far more than you need and
			 * cast up */

			if (first) {
				fprintf(fp, "TICK0\t%llu\n",
				    (long long unsigned)t);
				fprintf(fp, "TIME0\t%llu\n",
				    (long long unsigned)stats->time_start);
				first = false;
			} else if (!n->next && i == n->count-1) {
				fprintf(fp, "TICKN\t%llu\n",
				    (long long unsigned)t);
				fprintf(fp, "TIMEN\t%llu\n",
				    (long long unsigned)stats->time_end);
				done = true;
			} else {
				const char *prepost = k % 2 ? "PRE" : "POST";
				fprintf(fp, "%s\t%lld\t%llu\n",
				    prepost, (long long)bytes,
				    (long long unsigned)t);

				/* XXX assumption that all I/O ops didn't go
				 * short */
				if (k % 2 == 1)
					/* increment after printing PRE */
					bytes += spec->sz;
			}
			k++;
		}
	}

	milli_io = stats->disk_end.milli_io - stats->disk_start.milli_io;
	milli_tot = (stats->time_end - stats->time_start) / 1000000.0;
	fprintf(fp, "UTIL\t%.10e\n", milli_io / milli_tot);
	fprintf(fp, "CPUTIME\t%.10e\n", stats->cpu_time / 1000000000.0);

	return 0;
}

enum subworker_state {
	/* in order */

	/* parent initializes */
	CHILD_PREPARE,

	/* child sets with lock, signals condition, waits on condition */
	CHILD_PREWAIT,

	/* parent sets with lock, signals condition */
	CHILD_RUN,

	/* parent sets */
	CHILD_STOP
};

/* argument to subworker_start() thread */
struct subworker_arg {
	const struct proc_spec	*spec;
	struct proc_stats	stats;
	pthread_t		parent;

	pthread_mutex_t		lock;
	pthread_cond_t		state_changed;

	/* subworker_state,status protected by lock,state_changed */

	enum subworker_state	subworker_state;
	/* return value of thread */
	int			status;
};
static void *
subworker_start(void *arg_)
{
	/* arguments */
	struct subworker_arg	*arg = (struct subworker_arg *)arg_;
	const struct proc_spec	*spec = arg->spec;
	struct proc_stats	*stats = &arg->stats;
	const struct workload	*wl = &arg->spec->wl;

	/* these get used inside the loop, so I'll cut out the pointer
	 * dereferences; they are all constant */
	size_t			dist = wl->dist;
	size_t			szblk = arg->spec->sz;
	uint16_t		rapid = arg->spec->rapid;
	enum workload_type	type = wl->type;

	/* data */
	off_t		endpos;
	off_t		pos;
	uint64_t	cpu_end;
	uint64_t	cpu_start;
	uint8_t		*hole = 0;
	int		status;
	int		testfd = -1;
	uint8_t		*testmap = 0;
	uint16_t	rapid_remaining;

	if ((testfd = open_testfile(arg->spec->testpath, type,
	    arg->spec->mmap)) == -1) {
		status = -1;
		goto end;
	}

	/* source/dest for writes/reads */
	if (!(hole = malloc(szblk))) {
		fprintf(stderr, "%s: out of memory\n", name);
		status = -1;
		goto end;
	}

	/* get file size */
	pos = 0;
	if ((endpos = lseek(testfd, 0, SEEK_END)) == -1) {
		fprintf(stderr, "%s: seek failed: %s\n",
		    name, strerror(errno));
		status = -1;
		goto end;
	}
	if (lseek(testfd, 0, SEEK_SET) == -1) {
		fprintf(stderr, "%s: seek failed: %s\n",
		    name, strerror(errno));
		status = -1;
		goto end;
	}

	if (endpos < szblk) {
		fprintf(stderr, "%s: test file smaller than block size\n"
		    "this will not do at all\n", name);
		status = -1;
		goto end;
	}

	/* endpos is the working file size, and should be at most the real
	 * file size; it is decreased until it is at a size so that all
	 * read/write operations are on a full block */
	if (arg->spec->mmap) {
		/* avoid unfounded warning with init */
		int page_size;
		int protect = 0;

		page_size = sysconf(_SC_PAGE_SIZE);
		if (endpos < page_size) {
			fprintf(stderr,
			    "%s: test file smaller than page size (%d)\n"
			    "this will not do at all\n",
			    name, page_size);
			status = -1;
			goto end;
		}

		if (endpos != (size_t)endpos)
			/* endpos = SIZET_MAX;
			 * signed => unsigned conversion has completely
			 * defined behavior */
			endpos = (size_t) - 1;

		/* make endpos a multiple of both page and block size */
		endpos -= endpos % page_size;
		endpos -= endpos % szblk;
		switch (type) {
		case WL_RSEQ:
		case WL_RRAND:
		case WL_RREV:
			protect = PROT_READ;
			break;
		case WL_WSEQ:
		case WL_WRAND:
		case WL_WREV:
			protect = PROT_WRITE;
		}
		if ((testmap = mmap(0, endpos,
		    protect, MAP_PRIVATE, testfd, 0)) == (void *)-1) {
			fprintf(stderr, "%s: mmap failed: %s\n",
			    name, strerror(errno));
			status = -1;
			goto end;
		}
	} else
		endpos -= endpos % szblk;

	switch (type) {
	case WL_RREV:
	case WL_WREV:
		pos = endpos;
	default:;
	}

	/* wait for monitor thread to give the go-ahead */
	pthread_mutex_lock(&arg->lock);
	arg->subworker_state = CHILD_PREWAIT;
	do {
		pthread_cond_wait(&arg->state_changed, &arg->lock);
	} while (arg->subworker_state != CHILD_RUN);
	pthread_mutex_unlock(&arg->lock);

	if (proc_stats_push(stats, clock_ticks())) {
		fprintf(stderr, "%s: out of memory\n", name);
		status = -1;
		goto end;
	}
	stats->time_start = clock_monotonic();
	cpu_start = clock_cpu_thread();

	if (read_disk_stats(name, spec->path_stat, &stats->disk_start) ==
	    -1) {
		status = -1;
		goto end;
	}

	rapid_remaining = rapid;
	for (;;) {
		/* init 'bytes', avoid unfounded warning */
		ssize_t	bytes = 0;
		bool	need_seek = false;

		/* determine seek */
		switch (type) {
		case WL_RSEQ:
		case WL_WSEQ:
			if (pos + szblk > endpos) {
				pos = 0;
				need_seek = !testmap;
			}
			break;
		case WL_RRAND:
		case WL_WRAND:
		{
			off_t	block = pos / szblk;
			off_t	last = endpos / szblk - 1;
			off_t	min;
			off_t	max;
			if (dist) {
				min = block - dist;
				max = block + dist;
				if (max > last) max = last;
				if (min < 0) min = 0;
			} else {
				min = 0;
				max = last;
			}
			pos = (min + rand() % (max - min + 1)) * szblk;

			need_seek = !testmap;
			break;
		}
		case WL_RREV:
		case WL_WREV:
			pos -= 2 * szblk;
			if (pos < 0) pos = endpos;
			need_seek = !testmap;
		}

		/* perform seek */
		if (need_seek && lseek(testfd, pos, SEEK_SET) == -1) {
			fprintf(stderr, "%s: seek error: %s\n",
			    name, strerror(errno));
			status = -1;
			goto end;
		}

		/* odd position: pre-operation time */
		if (proc_stats_push(stats, clock_ticks())) {
			fprintf(stderr, "%s: out of memory\n", name);
			status = -1;
			goto end;
		}

		/* perform read/write */
		switch (type) {
		case WL_RSEQ:
		case WL_RRAND:
		case WL_RREV:
			if (testmap) {
				memcpy(hole, testmap + pos, szblk);
				bytes = szblk;
			} else
				bytes = read(testfd, hole, szblk);
			break;
		case WL_WSEQ:
		case WL_WRAND:
		case WL_WREV:
			if (testmap) {
				memcpy(testmap + pos, hole, szblk);
				bytes = szblk;
			} else
				bytes = write(testfd, hole, szblk);
		}
		if (bytes == -1) {
			if (errno != EINTR) {
				fprintf(stderr, "%s: read/write error: %s\n",
				    name, strerror(errno));
				status = -1;
				goto end;
			}
		} else
			pos += bytes;

		if (bytes < szblk)
			printf("%s: short read; this will skew results "
			    "if it occurs too often.\n", name);

		/* even position: post-operation time */
		if (proc_stats_push(stats, clock_ticks())) {
			fprintf(stderr, "%s: out of memory\n", name);
			status = -1;
			goto end;
		}

		if (arg->subworker_state == CHILD_STOP) break;

		/* for the interested: this spins for at *least* the amount of
		 * time specified on the command line; usually, it spins for
		 * twice as long (it is uncanny), which I really cannot
		 * explain switches; one can only hope that during
		 * calibration, it is accurate, but it seems that is the
		 * case */
		if (!--rapid_remaining) {
			spin_for(spec->spins);
			rapid_remaining = rapid;
		}
	}

	if (proc_stats_push(stats, clock_ticks())) {
		fprintf(stderr, "%s: out of memory\n", name);
		status = -1;
		goto end;
	}
	stats->time_end = clock_monotonic();
	cpu_end = clock_cpu_thread();
	stats->cpu_time = cpu_end - cpu_start;

	if (read_disk_stats(name, spec->path_stat, &stats->disk_end) == -1) {
		status = -1;
		goto end;
	}

	status = 0;
end:
	if (testmap) munmap(testmap, endpos);
	if (hole) free(hole);
	if (testfd != -1) close(testfd);

	/* set return status, possibly alert parent */
	if (status) {
		pthread_mutex_lock(&arg->lock);
		arg->status = status;
		pthread_cond_signal(&arg->state_changed);
		pthread_mutex_unlock(&arg->lock);

		/* wake up possible nanosleep */
		pthread_kill(arg->parent, SIGUSR1);
	} else
		/* control arrived here naturally (no goto);
		 * parent not waiting */
		arg->status = status;

	return arg;
}

static void
wakeup_handler()
{}

int
worker_begin(const struct proc_spec *spec, int semid)
{
	char			path_datafile[PATH_MAX];
	struct subworker_arg	arg;
	struct timespec		remainder;
	struct timespec		runtime;
	struct sigaction	wakeup_action;
	pthread_t		child;
	FILE			*outfp = 0;
	void			*voidp;
	int			err;
	int			status;
	bool			have_child = false;

	/* times have different meanings, depending on position i:
	 * i=0: zero -- each process gets a different zero; in
	 *     post-processing, the minimum is the accepted zero
	 * i=-1: final
	 * i odd: pre-op
	 * i positive-even: post-op
	 */
	proc_stats_init(&arg.stats);

	/* init these first so they should always be destroyed */
	pthread_mutex_init(&arg.lock, 0);
	pthread_cond_init(&arg.state_changed, 0);

	if (init_name(spec) == -1) {
		status = -1;
		goto end;
	}

	/* set signal handler so nanosleep can be woken up */
	wakeup_action.sa_handler = wakeup_handler;
	sigemptyset(&wakeup_action.sa_mask);
	wakeup_action.sa_flags = 0;
	if (sigaction(SIGUSR1, &wakeup_action, 0) == -1) {
		fprintf(stderr, "%s: sigaction failed: %s\n",
		    name, strerror(errno));
		status = -1;
		goto end;
	}

	init_path_datafile(spec->dataprefix, spec->id, path_datafile);
	if (!(outfp = open_datafile(path_datafile))) {
		status = -1;
		goto end;
	}

	/* start subworker thread; this should get to the lock wait
	 * before the process gets through the semaphore */
	arg.spec = spec;
	arg.parent = pthread_self();
	arg.subworker_state = CHILD_PREPARE;
	arg.status = 0;
	if ((err = pthread_create(&child, 0, subworker_start, &arg))) {
		fprintf(stderr, "%s: failed to create child thread: %s\n",
		    name, strerror(err));
		status = -1;
		goto end;
	}
	have_child = true;

	if (set_priority(spec) == -1) {
		status = -1;
		goto end;
	}

	runtime.tv_sec = spec->finish.tv_sec - spec->start.tv_sec;
	runtime.tv_nsec = spec->finish.tv_nsec - spec->start.tv_nsec;
	if (runtime.tv_nsec < 0) {
		runtime.tv_sec--;
		runtime.tv_nsec += 1000000000;
	}

	/* wait for child to be ready (or exit) */
	pthread_mutex_lock(&arg.lock);
	for (;;) {
		if (arg.status) {
			/* child ended early with error */
			status = -1;
			pthread_mutex_unlock(&arg.lock);
			goto end;
		}
		if (arg.subworker_state == CHILD_PREWAIT) break;
		pthread_cond_wait(&arg.state_changed, &arg.lock);
	}

	/* wait for multi-worker semaphore to change */
	if (sem_wait(spec->id, semid) == -1) {
		status = -1;
		goto end;
	}

	/* delay start */
	if (spec->start.tv_nsec || spec->start.tv_sec) {
		remainder = spec->start;
		for (;;) {
			if (nanosleep(&remainder, &remainder) == -1) {
				if (errno == EINTR) {
					/* no lock needed, the child would
					 * have set before signalling */
					if (arg.status) {
						status = -1;
						goto end;
					}
					continue;
				}
				fprintf(stderr, "%s: nanosleep failed: %s\n",
				    name, strerror(errno));
				status = -1;
				goto end;
			}
			break;
		}
	}

	/* go go go */
	printf("%s: starting\n", name);

	arg.subworker_state = CHILD_RUN;
	pthread_cond_signal(&arg.state_changed);
	pthread_mutex_unlock(&arg.lock);

	/* after 'runtime' time has passed, tell thread to stop */
	remainder = runtime;
	for (;;) {
		if (nanosleep(&remainder, &remainder) == -1) {
			if (errno == EINTR) {
				/* no lock needed, the child would
				 * have set before signalling */
				if (arg.status) {
					status = -1;
					goto end;
				}
				continue;
			}
			fprintf(stderr, "%s: nanosleep failed: %s\n",
			    name, strerror(errno));
			status = -1;
			goto end;
		}
		break;
	}

	/* tell child to stop and wait */
	arg.subworker_state = CHILD_STOP;

	/* wait for child */
	pthread_join(child, &voidp);
	if (arg.status) {
		fprintf(stderr, "%s: child returned with error\n", name);
		status = -1;
		goto end;
	}

	/* signal parent process that finished, and wait for it to allow
	 * procession */
	if (sem_wait(spec->id, semid) == -1) {
		status = -1;
		goto end;
	}

	if (dump_stats(spec, outfp, &arg.stats) == -1) {
		status = -1;
		goto end;
	}

	status = 0;
end:
	if (have_child) pthread_kill(child, SIGKILL);
	if (outfp) {
		fflush(outfp);
		fsync(fileno(outfp));
		fclose(outfp);
		if (status)
			unlink(path_datafile);
	}
	if (name) free(name);
	pthread_cond_destroy(&arg.state_changed);
	pthread_mutex_destroy(&arg.lock);
	proc_stats_destroy(&arg.stats);
	return status;
}
