/*
 * craneweb.h -- modern micro web framework for C.
 * (C) 2011-2014 Francesco Romani <fromani at gmail dot com>
 * ZLIB licensed.
 */

/** \file craneweb.h
    \brief The craneweb public interface.
*/

#ifndef CRANEWEB_H
#define CRANEWEB_H

#include <stdarg.h>

/*** implementation limits ***********************************************/

/** \enum the craneweb implementation limits.

    TL;DR: tune at your own risk.

    It is generally feasible to ramp up those limits but you must be
    prepared and carefully test your build. While craneweb should scale
    gently (and not blowup), the performance and memory cost could be
    not linear.
*/
enum {
    CRW_MAX_ROUTE_ARGS = 16,      /**< max number of :tags per route */
    CRW_MAX_HANDLER_ROUTES = 16,  /**< max number of aliases per handler */
    CRW_MAX_REQUEST_HEADERS = 64  /**< max number of HTTP headers exported
                                       into a CRW_Request. The remainder
                                       is silently dropped.
                                   */
};
/* TODO: doxygen anchor */

/*** versioning **********************************************************/

/**< \fn CRW_version_number
     \brief returns the craneweb version as split identifiers.

     Use this function for machine-friendly consumption.
     TODO: explain versioning rules.

     \param major pointer to an integer to hold the craneweb major version 
     \param major pointer to an integer to hold the craneweb minor version 
     \param major pointer to an integer to hold the craneweb micro version 
     \param major pointer to an integer to hold the craneweb patch version 
     \return nothing.

     \see CRW_version_string
*/
void CRW_version_number(int *major, int *minor, int *micro, int *patch);

/**< \fn CRW_version_string
     \brief returns the craneweb version as string.

     Use this function for human-friendly consumption.

     \return a constant pointer to the craneweb version string identifier.
             the return pointer shall NOT be free()d.

     \see CRW_version_number
*/
const char *CRW_version_string(void);

/*** logger **************************************************************/

/** \enum CRW_LogLevel
    \brief the log message severity levels
*/
typedef enum {
    CRW_LOG_CRITICAL = 0, /**< this MUST be the first and it is
                               the most important. -- PANIC!       */
    CRW_LOG_ERROR,        /**< you'll need to see this             */
    CRW_LOG_WARNING,      /**< you'd better to see this            */
    CRW_LOG_INFO,         /**< informative messages (for tuning)   */
    CRW_LOG_DEBUG,        /**< debug messages (for devs)           */
} CRW_LogLevel;


/** \var typedef CRW_LogHandler
    \brief logging callback function.

    This callback is invoked by the craneweb runtime whenever is needed
    to log a message.

    \param userdata a pointer given at the callback registration
           time. Fully opaque for craneweb.
    \param tag string identifying the craneweb module/subsystem.
    \param level the severity of the message.
    \param fmt printf-like format string for the message.
    \param args va_list of the arguments to complete the format string.
    \return 0 on success, <0 otherwise.

    \see CRW_LogLevel
*/
typedef int (*CRW_LogHandler)(void *userdata,
                              CRW_LogLevel level, const char *tag,
                              const char *fmt, va_list args);

/** \fn CRW_logger_console
    \brief default logging callback, emitting messages to the console.

    This is the default cranweb logging function. It logs messages
    on the console (stderr).

    \see CRW_LogHandler
*/
int CRW_logger_console(void *userdata, CRW_LogLevel level, const char *tag,
                       const char *fmt, va_list args);

/*** instance (1) ********************************************************/

/** \enum CRW_ServerType
    \brief the embedded server supported by craneweb
*/
typedef enum crwservertype_ {
    CRW_SERVER_NONE = -1,   /**< none (you should'nt see this) */
    CRW_SERVER_DEFAULT = 0, /**< implementation default */
    CRW_SERVER_MONGOOSE     /**< mongoose */
} CRW_ServerType;

/** \var typedef CRW_Instance
    \brief A CRW_Instance represent a craneweb application.

    You'll need to instantiate one and just one, and this as to live as long
    as you want to use the cranweb.
    A CRW_Instance has to be fully opaque to the caller.
*/
typedef struct crwinstance_ CRW_Instance;

/*** request *************************************************************/

/** \var typedef CRW_Request
    \brief A CRW_Request abstracts an HTTP request to your application.

    Accessing the fields a CRW_Request allow the application to understand
    what to do and how properly respond to the request itself.

    Both high-level and low-level (raw HTTP fileds) accessors are provided.

    The CRW_Request are provided, already filled and set up, by the
    craneweb runtime to your application. There is (should) no need
    to build or to free or to modify them on the application.
    Thereof, all the accessor function are getters only.
*/
typedef struct crwrequest_ CRW_Request;

/** \enum CRW_RequestMethod
    \brief the HTTP method of the request.
*/
typedef enum crwrequestmethod_ {
    CRW_REQUEST_METHOD_UNSUPPORTED = -1, /**< invalid data marker */
    CRW_REQUEST_METHOD_UNKNOWN = 0,      /**< on error (runtime) */
    CRW_REQUEST_METHOD_GET,              /**< GET */
    CRW_REQUEST_METHOD_HEAD,             /**< HEAD */
    CRW_REQUEST_METHOD_POST,             /**< POST */
    CRW_REQUEST_METHOD_PUT,              /**< PUT */
    CRW_REQUEST_METHOD_DELETE            /**< DELETE */
} CRW_RequestMethod;

/** \fn CRW_request_get_method
    \brief get the HTTP method of a given request

    \param req CRW_Request to be examined.
    \return the CRW_RequestMethod of the given request.

    \see CRW_RequestMethod
*/
CRW_RequestMethod CRW_request_get_method(const CRW_Request *req);

/** \fn CRW_request_count_headers
    \brief tell how many attached HTTP headers are avalaible for
           the given request.

    In the common case, this function returns the number of H
    headers attached to the HTTP request wrapped to craneweb.
    So, the real number of headers passed on the wire.

    However, given the implementation limits of craneweb, on
    certain (rare) cases, only the first N can be provided.
    the remainder of the headers are silently discarded.

    \param req CRW_Request to be examined.
    \return <0 on error, the number of avalaible headers on success.
*/
int CRW_request_count_headers(const CRW_Request *req);

/** \fn CRW_request_get_header_by_idx
    \brief access the header data.

    Access the HTTP header data (in a readonly way) using an index.
    The HTTP headers are indexed from 0 (zero) to N, [0,N) where
    N is the upper bound is provided by CRW_request_count_headers.

    The header data is returned preserving the case.

    \param req CRW_Request to be examined.
    \param idx index of the header data requested.
    \param[out] header const pointer to be set to the idx-th header key.
    \param[out] value const pointer to be set to the idx-th header value.
    \return 0 on success, <0 on error.

    \see CRW_request_count_headers
*/
int CRW_request_get_header_by_idx(const CRW_Request *req, int idx,
                                  const char **header, const char **value);

/** \fn CRW_request_get_header_value
    \brief access the header value given its name.

    Access the HTTP header data (in a readonly way) using the header name.
    The name search is done in a case INsenstive way.
    The header data is returned preserving the case.

    \param req CRW_Request to be examined.
    \param header the header name (key) to be searched.
    \return NULL on error, the value of the header key on success.

    \see CRW_request_get_header_by_idx
    \see CRW_request_count_headers
*/
const char *CRW_request_get_header_value(const CRW_Request *req, const char *header);

/** \fn CRW_request_is_xhr
    \brief checks if the request is made by the client using
           XmlHttpRequest.

    CAUTION: proper detection required that the client code honores the
    custom http header. This function can give wrong result if the client
    is non-compliant.

    \param req CRW_Request to be examined.
    \return !0 if the request is detected made using XmlHttpRequest,
            0 otherwise.
*/
int CRW_request_is_xhr(const CRW_Request *req);

/**\fn CRW_request_count_query_items
   \brief tell how many key/value item pairs are avalaible from
          the parsed GET query string.

   Access the GET parameters (aka URL arguments) count.

   \param req CRW_Request to be examined.
   \return <0 on error, the number of avalaible pairs on success.

   \see CRW_RequestMethod
   \see CRW_request_get_method
*/
int CRW_request_count_query_items(const CRW_Request *req);

/** \fn CRW_request_get_query_item_by_name
    \brief access the GET query item value given its name.

    Access the GET parameter (in a readonly way) using the item name.
    The name search is done in a case INsenstive way.
    The item data is returned preserving the case.

    \param req CRW_Request to be examined.
    \param item the parameter name name (key) to be searched.
    \return NULL on error, the value of the parameter name on success.

    \see CRW_request_count_query_items
*/
const char *CRW_request_get_query_item_by_name(const CRW_Request *req,
                                               const char *name);

/**\fn CRW_request_count_query_items
   \brief tell how many key/value item pairs are avalaible from
          the parsed POST/PUT query body.

   Access the count of the form values parsed from an url-encoded
   POST or PUT request body.

   \param req CRW_Request to be examined.
   \return <0 on error, the number of avalaible pairs on success.

   \see CRW_RequestMethod
   \see CRW_request_get_method
*/
int CRW_request_count_form_items(const CRW_Request *req);

/** \fn CRW_request_get_form_item_by_name
    \brief access the POST/PUT query item value given its name.

    Access the POST/PUT parameter (in a readonly way) using the item name.
    The name search is done in a case INsenstive way.
    The item data is returned preserving the case.

    \param req CRW_Request to be examined.
    \param item the parameter name name (key) to be searched.
    \return NULL on error, the value of the parameter name on success.

    \see CRW_request_count_form_items
*/
const char *CRW_request_get_form_item_by_name(const CRW_Request *req,
                                              const char *name);

/**\fn CRW_request_get_uri
   \brief get the raw request URI, including the schema, up to
   the '&' of the URL.

   Provides a constant reference to the request URI.

   \param req CRW_Request to be examined.
   \return a constant pointer to the URI on success, NULL on error.
*/
const char *CRW_request_get_uri(const CRW_Request *req);

/* \fn CRW_request_get_query_string
   \brief get the raw query part of the URL (everything in between ? and #)
          as a string.

   Provides a constant reference to the request raw query_string.

   \param req CRW_Request to be examined.
   \return a constant pointer to the query string on success, NULL on error.
*/
const char *CRW_request_get_query_string(const CRW_Request *req);

/** \fn CRW_request_get_content_length
    \brief provide the expected request body length, in bytes.

    CAUTION: proper detection required that the client code honores the
    http header `Content-Length'. This function can give wrong result if the client
    is non-compliant. If the header is missing, the function will return
    an error value, buy this by no means imply there is no body in the
    HTTP request.

    \param req CRW_Request to be examined.
    \return the content lenght in bytes if the information is avalaible,
            <0 otherwise.
*/
int CRW_request_get_content_length(const CRW_Request *req);

/*** response ************************************************************/

/** \var typedef CRW_Response
    \brief A CRW_Response abstracts an HTTP response from your application.

    Using a CRW_Response the client code can programmatically build the
    HTTP response for a given request.
    The CRW_Response just provides raw access, meaning then it handles just
    raw binary strings.

    A CRW_Response provides distinct access to the response headers,
    treated as strings using the usual http
    key:value\r\n
    convention, and to the response body, treated as opaque binary string.

    Both headers and body can be accessed only in append mode.

    The application needs to allocate a new CRW_Response (fill it)
    and pass it to the craneweb runtime. The craneweb runtime will send
    back to the requested and automatically release it.

    Givin back a NULL response means that a critical internal error happended
    somewhere inside.
*/
typedef struct crwresponse_ CRW_Response;

/** \fn CRW_response_new
    \brief allocate a new CRW_Response out of a CRW_Instance.

    \param inst a CRW_Instance reference.
    \return a new empty CRW_Response reference on success,
            NULL on error.
*/
CRW_Response *CRW_response_new(CRW_Instance *inst);

/** \fn CRW_response_del
    \brief release a CRW_Response and all related resources.

    \param res CRW_Response to be released.
*/
void CRW_response_del(CRW_Response *res);

/** FIXME */
int CRW_response_status_code(CRW_Response *res, int status_code);

/** \fn CRW_response_add_header
    \brief adds an HTTP header to a CRW_Response.

    \param res the CRW_Response to be augmented.
    \param name name of the HTTP header to add.
    \value value value of the HTTP header to add.
    \return 0 on success, <0 on error.
*/
int CRW_response_add_header(CRW_Response *res,
                            const char *name, const char *value);

/** \fn CRW_response_add_body
    \brief add a chunk of data to the response body

    adds a NULL-terminated character string to the response body.

    \param res the CRW_Response to be augmented.
    \param chunk chunk of NULL-terminated data to be added.
    \return 0 on success, <0 on error.
*/
int CRW_response_add_body(CRW_Response *res, const char *chunk);

/** \fn CRW_response_load_file
    \brief load a file from the storage and append it verbatim in the
           response body.

    adds the content from a given file on the storage to the current
    response body, by appending the content on the existing data.

    \param res the CRW_Response to be augmented.
    \param path path on storage of the file to be load.
    \return 0 on success, <0 on error.
*/
int CRW_response_load_file(CRW_Response *res, const char *path);

/** \fn CRW_response_send_file
    \brief send directly a file from the storage as response body.

    load and send a given file on the storage as a full response
    body. The content of the given file fully replaces any existing
    response body data.
    The send is made as efficiently (as in space and speed) as possible.

    Use this functio if you want to use static files as response objects.

    \param res the CRW_Response to be augmented.
    \param path path on storage of the file to be sent.
    \return 0 on success, <0 on error.
*/
int CRW_response_send_file(CRW_Response *res, const char *path);


/*** template ************************************************************/

/** \var typedef CRW_Context
    \brief A CRW_Context holds all the actual values to be rendered
    in a template.

    The CRW_Context allows the client code to fill the template with the
    given values for the variables. You need to provide all the variable
    you want to fill before to feed the template with the context.

    Values not required by variables in the template will be silently
    discarded.
*/
typedef struct crwcontext_ CRW_Context;

/** \fn CRW_context_new
    \brief allocate a new CRW_Context out of a CRW_Instance.

    \param inst a CRW_Instance reference.
    \return a new empty CRW_Context reference on success,
            NULL on error.
*/
CRW_Context *CRW_context_new(CRW_Instance *inst);

/** \fn CRW_context_del
    \brief release a CRW_Context and all related resources.

    \param res CRW_Context to be released.
*/
void CRW_context_del(CRW_Context *ctx);

/** \fn CRW_context_set
    \brief set a variable in the given context with the given value.

    \param ctx a CRW_Context to be filled.
    \param key the variable name to be set.
    \param val the value of the variable to be set.
    \return 0 on success, <0 on error.
*/
int CRW_context_set(CRW_Context *ctx, const char *key, const char *val);

/** \fn CRW_context_setm
    \brief set a group of variables in the given context.

    Like CRW_context_set but a NULL-terminated array of key and values.
    The variables are set until the first NULL (either in the key or
    val) is found. The remaining values are silently discarded.

    \param ctx a CRW_Context to be filled.
    \param key the NULL-terminated array of variable names to be set.
    \param val the NULL-terminated array of corresponding variable values.
    \return 0 on success, <0 on the FIRST error.

    \see CRW_context_set
*/
int CRW_context_setm(CRW_Context *ctx, const char **key, const char **val);


/** \var typedef CRW_Template
    \brief A CRW_Template represents a (NG)template.

    A Template requires to be filled using a CRW_Context before to be
    rendered (= produce the actual result ready to be sent to the client).

    A CRW_Template needs to be loaded from a file on disk.
*/
typedef struct crwtemplate_ CRW_Template;

/**\fn CRW_template_new_from_file
   \brief loads a (NG)Template from a disk file and returns a
   CRW_Template instance representing it.

   If the template couldn't be loaded for any reason, an error
   will be CRW_log()ged.

   Template files are been searched for loading
   1. in the template_root directory (see CRW_Config)
   2. in the document_root (if and only if template_root is NULL)
   3. in the current directory.

   \param inst a CRW_Instance reference.
   \param filename the template name (NOT the full path).
   \return a new empty CRW_Context reference on success,
           NULL on error.
   \see CRW_Config.
*/
CRW_Template *CRW_template_new_from_file(CRW_Instance *inst,
                                         const char *filename);
/** \fn CRW_template_del
    \brief release a CRW_Template and all related resources.

    \param res CRW_Template to be released.
*/
void CRW_template_del(CRW_Template *tmpl);

/** \fn CRW_template_render
    \brief render a CRW_Template using the variables provided by
    a CRW_Context.

    the output of the rendering is a C-string returned using
    the `text' parameter. The caller SHALL NOT free the pointer
    once it has done.

    The pointer is valid up to
    1. a new call of CRW_template_render is made
    2. CRW_template_cleanup is called
    3. CRW_template_del is called

    \param tmpl a CRW_Template instance to be rendered.
    \param ctx a CRW_Context to be used to fill the template.
    \param text pointer to be set with the rendered result.
    \return 0 on success, <0 on error.
    \see CRW_template_cleanup
*/
int CRW_template_render(CRW_Template *tmpl,
                        CRW_Context *ctx, char **text);

/** \fn CRW_template_cleanup
    \brief voids the previous rendering, if any.

    \param tmpl the template to be cleaned up.
    \return 0 on success, <0 on error.
    \see CRW_template_render
*/
int CRW_template_cleanup(CRW_Template *tmpl);

/** \fn CRW_template_render_to_response
    \brief shortcut (FIXME WRITEME)
*/
int CRW_template_render_to_response(CRW_Instance *inst,
                                    const char *filename, CRW_Context *ctx,
                                    CRW_Response *res);

/*** route ***************************************************************/

/** \var typedef CRW_RouteArgs
    \brief A CRW_RouteArgs provides access to the values of the route
           parameters (aka tags).

    When using a dynamic route (/something/:with/some/:tags)
    the handling code needs to access to the actual values of the parameters.

    The tag value is always returned preserving the case.

    The handling code is fed with an opaque reference of CRW_RouteArgs
    which can be used to access (in a read-only way) this data.
*/
typedef struct crwrouteargs_ CRW_RouteArgs;

/** \fn CRW_route_args_count
    \brief counts the real number of tags bound to a dynamic route.

    This is usually equal to the number of expected tags. Otherwise
    something very wrong has happened.

    \param args the CRW_RouteArgs instance to be accessed.
    \return <0 on error, the number of actual tags otherwise.
*/
int CRW_route_args_count(const CRW_RouteArgs *args);

/** \fn CRW_route_args_get_by_idx
    \brief access a given arg by index.

    Access the route tags data (in a readonly way) using an index.
    The tag data are indexed from 0 (zero) to N, [0,N) where
    N is the upper bound is provided by CRW_route_args_count.

    \param args the CRW_RouteArgs instance to be accessed.
    \return NULL on error, a const read-only pointer to the tag data
            on success. Client code MUST NOT free() this pointer.
*/
const char *CRW_route_args_get_by_idx(const CRW_RouteArgs *args, int idx);

/** \fn CRW_route_args_get_by_tag
    \brief access a given arg by name.

    Access the route tags data (in a readonly way) using by finding it
    by name. The search is performed in a case SENSITIVE way.
    Differently from CRW_route_args_get_by_idx this search can fail.

    \param args the CRW_RouteArgs instance to be accessed.
    \return NULL on error OR if the given tag is not found.
            Otherwise, a const read-only pointer to the tag data
            on success. Client code MUST NOT free() this pointer.
*/
const char *CRW_route_args_get_by_tag(const CRW_RouteArgs *args,
                                      const char *tag);

/*** handler *************************************************************/

/** \var typedef CRW_Handler
    \brief A CRW_Handler is on charge to respond on a request directed to
           a particular route.

    The CRW_Handler is the workhorse of a craneweb application.
    Handlers are invoked by the craneweb runtime to provide a response
    after a request is received.

    Handlers are found by selecting the first which matches a given route.
    Routes of course can by static or dynamic (aka /with/:tags/within).
    Every handler must be attached to a given route, so it cannot exists
    without a corresponding route.
    An handler, however, can respond to multiple routes.

    An handler embody a client-provided callback which executes the
    actual work to craft the response.
*/
typedef struct crwhandler_ CRW_Handler;

/** \var typedef CRW_HandlerCallback
    \brief handler worker callback function.

    This callback is invoked by the craneweb runtime whenever the owner
    handler matches a route and thus is being called to respond to a
    request.

    CAUTION: the callback CAN be (depending on the server used)
             in a different thread respect of the one which calls
             CRW_run. Use appropriate locking.

    CAUTION: the CRW_RouteArgs object is fully managed by the craneweb
             runtime.
    CAUTION: the CRW_Request object is fully managed by the craneweb
             runtime.
    CAUTION: is responsability of the callback to create the CRW_Response
             object. However, it is freed by the craneweb runtime.

    \param inst the CRW_Instance on which this callback is operating.
    \param args CRW_RouteArgs const reference, to access the dynamic
           route arguments, if any.
    \param req CRW_Request const reference, representing the request
           being served.
    \param userdata opaque pointer to the callback userdata.
    \return NULL on error, a new valid CRW_Response (usually filled)
            on success.

    \see CRW_Handler
*/
typedef CRW_Response *(*CRW_HandlerCallback)(CRW_Instance *inst,
                                             const CRW_RouteArgs *args,
                                             const CRW_Request *req,
                                             void *userdata);

/** \fn CRW_handler_new
    \brief build a new CRW_Handler.

    Builds and readies a new CRW_Handler, ready to be attached
    to the managing CRW_Instance.

    CAUTION: the client must manage on its own the userdata lifecycle.

    \param inst the CRW_Instance on which the handler will operate.
    \param route string representation of the route which this
           handler will respond.
    \param callback the CRW_HandlerCallback which implements this handler.
    \param userdata opaque pointer which will be passed to every
           callback invocation.
    \return a new CRW_Handler instance on success,
            NULL on error.

    \see CRW_handler_del
*/
CRW_Handler *CRW_handler_new(CRW_Instance *inst,
                             const char *route,
                             CRW_HandlerCallback callback,
                             void *userdata);

/** \fn CRW_handler_del
    \brief release a CRW_Handler instance and all related resources.

    \param handler the handler to be released.

    \see CRW_handler_new
*/
void CRW_handler_del(CRW_Handler *handler);

/** \fn CRW_handler_add_route
    \brief attach a new route to an handler

    attach a new route which the given handler responds to.

    CAUTION: the handler is invoked on the first match among
             all the avalaible routes.

    \param handler the handler to be augmented.
    \param route string representation of the route which this
           handler will also respond.
    \return 0 on success,
            <0 on error.

    \see CRW_handler_new
*/
int CRW_handler_add_route(CRW_Handler *handler, const char *route);


/*** instance (2) ********************************************************/

/** \fn CRW_instance_new
    \brief builds a new CRW_Instance emboding an application.

    Builds a new CRW_Instance ready to be used to serve requests.

    \param server the type of server to use.
    \return a valid CRW_Instance reference on success,
            NULL on error.

    \see CRW_ServerType;
*/
CRW_Instance *CRW_instance_new(CRW_ServerType server);

/** \fn CRW_instance_del
    \brief release a CRW_Instance handle and all ITS resources.

    You need to explicitely del any craneweb object you explicitely
    created (e.g. CRW_Handler-s) because they WILL NOT be automatically
    collected when you destroy a CRW_Instance.

    \param inst the instance to be released.

    \see CRW_instance_new
*/
void CRW_instance_del(CRW_Instance *inst);

/** \fn CRW_instance_set_logger
    \brief replace the instance logging callback.

    \param inst the instance to be changed.
    \param logger the new logger callback.
    \return 0 on success,
            <0 on error.

    \see CRW_LogHandler
*/
int CRW_instance_set_logger(CRW_Instance *inst, CRW_LogHandler logger);

/** \fn CRW_instance_add_handler
    \brief attach an handler to an instance.

    By default CRW_Instance-s does not have any active handlers, so
    will silently and diligently ignore any incoming request.

    You'll need to add the handler you create to the serving
    CRW_Instance to get them running on requests.

    \param inst the instance to be augmented.
    \param handler reference to a CRW_Handler to be attached.
    \return 0 on success,
            <0 on error.

    \see CRW_Handler
*/
int CRW_instance_add_handler(CRW_Instance *inst, CRW_Handler *handler);

/*** runtime(!) **********************************************************/

/** \struct CRW_Config
    \brief runtime configuration of a CRW_Instance.

    Export the tunables of the CRW_Instance.

    Client code needs to properly fill it for running a CRW_Instance.
*/
typedef struct crwconfig_ CRW_Config;
struct crwconfig_ {
    const char *host;           /**< host IP to listen on */
    int port;                   /**< listening port */
    const char *document_root;  /**< document root (static files)*/
    const char *template_root;  /**< template directory.
                                     if NULL defaults to document_root. */
};

/**< \fn CRW_run
     \brief runs a CRW_Instance, allowing it to serve requests.

     This function blocks the caller until the CRW_Instance runs.

     \param instance the CRW_Instance to be run.
     \param cfg CRW_Config describing how to run the given instance.
     \return 0 on success, <0 on error.

     \see CRW_Instance
     \see CRW_Config
*/
int CRW_run(CRW_Instance *instance, const CRW_Config *cfg);

/**< \fn CRW_start
     \brief starts a CRW_Instance, allowing it to serve requests.

     This function DO NOT block the caller, allowing it to continue
     with the operations.
     PLEASE NOTE: the return value will only represent the status
     of the *starting* action. Further errors will not be reported.

     \param instance the CRW_Instance to be started.
     \param cfg CRW_Config describing how to start the given instance.
     \return 0 on success, <0 on error.

     \see CRW_run
     \see CRW_Instance
     \see CRW_Config
*/
int CRW_start(CRW_Instance *instance, const CRW_Config *cfg);

/**< \fn CRW_is_running
     \brief checks if a given instance is currently running.

     This function return a truth value representing the current
     running status of the instance.

     \param instance the CRW_Instance to be checked.
     \return !0 if running, 0 otherwise.

     \see CRW_run
     \see CRW_start
     \see CRW_Instance
*/
int CRW_is_running(CRW_Instance *instance);

/*** that's all folks! ***************************************************/

#endif /* CRANEWEB_H */

/* vim: set ts=4 sw=4 et */
/* EOF!*/

