/** @mainpage frei0rx - a minimalistic plugin API for video effects - eXtended
 *
 * @section sec_intro Introduction
 *
 * This is frei0rx - a minimalistic plugin API for video effects. eXtended.
 * frei0rX attempts to bridge the gap from frei0r to the video processing needs
 * by preserving the minimalistic style of the former. This is done by adding
 * the bare minimum of characteristics, most notably the planar format and the
 * audio handling.
 *
 * The utilities and helper routine are intentionally left out to keep the extension
 * lightweight as much as possible.
 *
 * @section sec_overview Overview
 *
 * The main need addressed by this extension is the necessity to cope with planar
 * image data and audio data. In order to reduce as much as is possible clumsy
 * casts, a minimal `frame` data structure is added for practical purposes.
 *
 * The following functions replaces the vanilla frei0r ones in order to cope with
 * the new data
 *
 * - \ref f0rx_construct
 * - \ref f0rx_get_plugin_info
 * - \ref f0rx_set_params_values
 * - \ref f0rx_is_ready
 * - \ref f0rx_update
 *
 * For the same reason the plugin info data structure (\ref f0rx_plugin_info)
 * is augmented with new fields.
 * The plugin developer must now specify
 * - which content to handle, video or audio (\ref CONTENT_TYPE),
 * - which sample model, if dealing with audio (\ref SAMPLE_MODEL), and
 * - which channel layout (\ref CHANNEL_LAYOUT) your plugin will support.
 *
 * A pair of new functions are added for practicality:
 *
 * - \ref f0rx_is_ready
 * - \ref f0rx_set_params_values
 *
 */

/** \file
 * \brief This file defines the frei0rX api, version 1.2.
 *
 * A conforming plugin must implement and export all functions declared in
 * this header.
 *
 * A conforming application must accept only those plugins which use
 * allowed values for the described fields.
 */

#ifndef INCLUDED_FREI0RX_H
#define INCLUDED_FREI0RX_H


#include <stdint.h>

#include "frei0r.h"

//---------------------------------------------------------------------------

/** \addtogroup COLOR_MODEL Color Models */
/*@{*/
/**
 * The image data is stored in the YUV420P (aka I420) format.
 * First all the Y data, then all the U data, finally all the V data.
 * Padding is allowed both intra-frame (aka per-line by leveraging the pitch
 * field) and inter-frame. See the \ref f0rx_frame for the details and all
 * the constraints.
 */
#define F0RX_COLOR_MODEL_PLANAR_YUV420P 8
/*@}*/


// TODO: sort out the layout constraint for the audio data
/** \addtogroup SAMPLE_MODEL Sample Models
 * List of supported color models.
 */
/*@{*/
/** Each sample is stored in one signed byte.
 */
#define F0RX_SAMPLE_MODEL_S8 0

/** Each sample is stored in one unsigned byte.
 */
#define F0RX_SAMPLE_MODEL_U8 1

/** Each sample is stored in two signed bytes, in little-endian format.
 */
#define F0RX_SAMPLE_MODEL_S16LE	2

/** Each sample is stored in two unsigned bytes, in little-endian format.
 */
#define F0RX_SAMPLE_MODEL_U16LE	3

/** Each sample is stored in two signed bytes, in big-endian format.
 */
#define F0RX_SAMPLE_MODEL_S16BE	4

/** Each sample is stored in two unsigned bytes, in big-endian format.
 */
#define F0RX_SAMPLE_MODEL_U16BE	5
/*@}*/

/** \addtogroup CHANNEL_LAYOUT Channel Layouts */
/*@{*/
/**
 * WRITEME
 */
// TODO
/*@}*/

/** \addtogroup CONTENT_TYPE Content Types */
/*@{*/
/** The content managed by the plugin is video (image) data.
 */
#define F0RX_CONTENT_TYPE_VIDEO 0

/** The content managed by the plugin is audio (sample) data.
 */
#define F0RX_CONTENT_TYPE_AUDIO 1
/*@}*/

/**
 * The f0rx_plugin_info_t structure...
 * WRITEME
 */
typedef struct f0rx_plugin_info
{
  f0r_plugin_info_t base; /**< base data. Not-overriden
                               fields are still valid. */
  int content_type; /**< the content managed by the plugin.
                         see \ref CONTENT_TYPE. */
  int sample_model; /**< if content is audio, the expected
                         sample model. See \ref SAMPLE_MODEL. */
  int channel_layout; /**< if content is audio, the expected
                           channel layout. See \ref CHANNEL_LAYOUT. */
} f0rx_plugin_info_t;


/**
 * Is called once after init. The plugin has to fill in the values in info.
 * Replaces the f0r_get_plugin_info call.
 * A conformant plugin must return the same subset of information
 * with both the extended and the vanilla version.
 *
 * \param info Pointer to an info struct allocated by the application.
 */
void f0rx_get_plugin_info(f0rx_plugin_info_t* info);

//---------------------------------------------------------------------------


/**
 * Constructor for effect instances. The plugin returns a pointer to
 * its internal instance structure.
 *
 * The Format parameter specifies the data format the host application
 * requests to the plugin. A conformant plugin can support more than
 * a data format (e.g. different colorspaces for image data).
 * If the plugin can perform an internal optimal conversion, it must
 * accept a Format different from the one specified by the f0rx_plugin_info.
 * Otherwise, the plugin must return 0 (error) and the data conversion
 * duty is left to the host application.
 *
 * If the plugin content type is video, then param1 and param2 shall be
 * interpreted as frame width and frame height. In this case,
 * the resolution must be an integer multiple of 8,
 * must be in the range [0, 2048] in both dimensions.
 * The plugin must set the values for all parameters in this function.
 *
 * If the plugin content type is audio, then param1 and param2 have to be
 * interpreted as sample rate and channel count. In this case,
 * the sample rate must be an integer in the range [8000, 48000]
 * The channel count must be an integer in the range [2, 6].
 *
 * NOTE: a conformant plugin MUST return 0 (error) if the parameters
 * are out of the above bonds.
 *
 * \param format Requested format for this instance.
 * \param params Array of exactly num_params values. If a parameter
 *        doesn't need to be set, just pass NULL.
 *        pass NULL as array if no parameter has to be set.
 * \param value1 Dependent of the content type of the plugin.
 * \param value2 Dependent of the content type of the plugin.
 * \returns 0 on failure or a pointer != 0 on success
 *
 * \see f0r_destruct
 */
f0r_instance_t f0rx_construct(int format, f0r_param_t *params,
                              unsigned int value1, unsigned int value2);


/**
 * Set all the given plugin parameters in one batch. The plugin can
 * choose to set all those parameters independently (so, like a
 * series of call to f0r_set_param_value) or atomically dependently
 * of the implementation.
 *
 * The client code should issue this call whenever it is known in
 * advance that more than one parameter, or some related parameter,
 * have to be set together.
 *
 * \param instance the effect instance.
 * \param param array of exactly num_params values. If a parameter
 *        doesn't need to be set, just pass NULL.
 * \return how many parameters are been succesfully set.
 * \see f0r_set_param_value
 * \see f0r_plugin_info
 */
int f0rx_set_params_values(f0r_instance_t instance,
                           f0r_param_t *params);

/**
 * Verifies if the given instance is properly configured,
 * meaning that any *update* call will produce a properly processed frame
 * or not.
 * This function can (and usually should) called any time any number of time
 * between two f0r_set_param calls.
 *
 * \param instance the effect instance.
 * \param param_index index of the parameter to check. Use -1 to mean "all".
 * \return >0 if the requested configuration is correct, 0 otherwise.
 */
int f0rx_is_ready(f0r_instance_t instance, int param_index);

//---------------------------------------------------------------------------

/**
 * Much like the vanilla fri0r f0r_update, but with \ref f0rx_frame
 * parameters. See the documentation of \ref f0rx_frame for the details
 * and the constraints of the frame structure.
 *
 * \param instance the effect instance.
 * TODO DOC
 *
 */
int f0rx_update(f0r_instance_t instance,
                double intime,
                const uint32_t *inframe1,
                const uint32_t *inframe2,
                const uint32_t *inframe3,
                double *outtime,
                uint32_t *outframe);

//---------------------------------------------------------------------------

#endif

