/*  denoise3d.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) 2003 Daniel Moreno <comac@comac.darktech.org>
 *  Converted for use in transcode by Tilmann Bitterberg <transcode@tibit.org>
 *  Converted hqdn3d -> denoise3d and also heavily optimised for transcode
 *     by Erik Slagter <erik@oldconomy.org> (GPL) 2003
 *   
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "frei0rx.h"

/*
    set tabstop=4 for best layout

    Changelog

    1.0.3   EMS first public version
    1.0.4   EMS added YUV422 support
    1.0.5   EMS added RGB support
                large cleanup
                added arbitrary layout support
                denoising U&V (colour) planes now actually works
    1.0.6   EMS fixed annoying typo
*/

#define MAX_PLANES              3

#define DEFAULT_LUMA_SPATIAL    4.0
#define DEFAULT_CHROMA_SPATIAL  3.0
#define DEFAULT_LUMA_TEMPORAL   6.0
#define DEFAULT_CHROMA_TEMPORAL 4.0


#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

enum {
    RGB_BYTES             = 3, /* how much storage a pixel take? */
    RGBA_BYTES            = 4, /* ditto */
    PIXEL_BLACK_THRESHOLD = 24,
    PLUGIN_PARAM_NUM      = 4
};


typedef enum { dn3d_yuv420p, dn3d_yuv422, dn3d_rgb } dn3d_fmt_t;
typedef enum { dn3d_planar, dn3d_packed } dn3d_basic_layout_t;
typedef enum { dn3d_luma, dn3d_chroma, dn3d_disabled } dn3d_plane_type_t;

typedef enum
{
    dn3d_off_y420,  dn3d_off_u420,  dn3d_off_v420,
    dn3d_off_y422,  dn3d_off_u422,  dn3d_off_v422,
    dn3d_off_r,     dn3d_off_g,     dn3d_off_b
} dn3d_offset_t;

typedef struct
{
    dn3d_plane_type_t   plane_type;
    dn3d_offset_t       offset;
    int                 skip;
    int                 scale_x;
    int                 scale_y;
} dn3d_single_layout_t;

typedef struct
{
    int                     pix_fmt;
    dn3d_fmt_t              fmt;
    dn3d_basic_layout_t     layout_type;
    dn3d_single_layout_t    layout[MAX_PLANES];
} dn3d_layout_t;

typedef struct
{
    dn3d_layout_t   layout_data;

    struct
    {
        double luma_spatial;
        double chroma_spatial;
        double luma_temporal;
        double chroma_temporal;
    } parameter;

    int             coefficients[4][512];
    unsigned char * lineant;
    unsigned char * previous;
    int             prefilter;
    int             enable_luma;
    int             enable_chroma;

} dn3d_private_data_t;

static const dn3d_layout_t dn3d_layout[] =
{
//  { CODEC_YUV,    dn3d_yuv420p, dn3d_planar, {{ dn3d_luma, dn3d_off_y420,  1, 1, 1 }, { dn3d_chroma, dn3d_off_u420,  1, 2, 2 }, { dn3d_chroma, dn3d_off_v420,  1, 2, 2 }}},
//  { CODEC_YUV422, dn3d_yuv422,  dn3d_planar, {{ dn3d_luma, dn3d_off_y422,  1, 1, 1 }, { dn3d_chroma, dn3d_off_u422,  1, 2, 1 }, { dn3d_chroma, dn3d_off_v422,  1, 2, 1 }}},
    { F0R_COLOR_MODEL_RGBA8888, dn3d_rgb, dn3d_packed, {{ dn3d_luma, dn3d_off_r, 4, 1, 1 }, { dn3d_luma, dn3d_off_g, 4, 1, 1 }, { dn3d_luma, dn3d_off_b, 4, 1, 1 }}}
};

typedef struct denoise3dinstance_ {
    int             width;
    int             height;

	dn3d_layout_t	layout_data;

	struct
	{
		double luma_spatial;
		double chroma_spatial;
		double luma_temporal;
		double chroma_temporal;
	} parameter;

	int				coefficients[4][512];
	unsigned char *	lineant;
	unsigned char * previous;
	int				prefilter;
	int				enable_luma;
	int				enable_chroma;
} denoise3d_instance_t;


#define LowPass(prev, curr, coef) (curr + coef[prev - curr])

static inline int ABS(int i)
{
    return(((i) >= 0) ? i : (0 - i));
}

static void deNoise(unsigned char * frame, unsigned char * frameprev, unsigned char * lineant,
                    int w, int h,
                    int * horizontal, int * vertical, int * temporal,
                    int offset, int skip)
{
    int x, y;
    unsigned char pixelant;
    unsigned char * lineantptr = lineant;

    horizontal += 256;
    vertical   += 256;
    temporal   += 256;

    frame       += offset;
    frameprev   += offset;

    // First pixel has no left nor top neighbour, only previous frame

    *lineantptr = pixelant = *frame;
    *frame = *frameprev = LowPass(*(frameprev), *lineantptr, temporal);
    frame += skip;
    frameprev += skip;
    lineantptr++;

    // First line has no top neighbour, only left one for each pixel and last frame

    for(x = 1; x < w; x++)
    {
        pixelant = LowPass(pixelant, *frame,  horizontal);
        *lineantptr = pixelant;
        *frame = *frameprev = LowPass(*(frameprev), *lineantptr, temporal);
        frame += skip;
        frameprev += skip;
        lineantptr++;
    }

    for (y = 1; y < h; y++)
    {
        lineantptr = lineant;

        // First pixel on each line doesn't have previous pixel

        pixelant = *frame;
        *lineantptr = LowPass(*lineantptr, pixelant, vertical);
        *frame = *frameprev = LowPass(*(frameprev), *lineantptr, temporal);
        frame += skip;
        frameprev += skip;
        lineantptr++;

        for (x = 1; x < w; x++)
        {
            // The rest is normal

            pixelant = LowPass(pixelant, *frame,  horizontal);
            *lineantptr = LowPass(*lineantptr, pixelant, vertical);
            *frame = *frameprev = LowPass(*(frameprev), *lineantptr, temporal);
            // *frame ^= 255; // debug
            frame += skip;
            frameprev += skip;
            lineantptr++;
        }
    }
}

static void PrecalcCoefs(int * ct, double dist25)
{
    int i;
    double gamma, simil, c;

    gamma = log(0.25) / log(1.0 - dist25 / 255.0);

    for(i = -256; i <= 255; i++)
    {
        simil = 1.0 - (double)ABS(i) / 255.0;
        c = pow(simil, gamma) * (double)i;
        ct[256 + i] = (int)((c < 0) ? (c - 0.5) : (c + 0.5));
    }
}


// TODO

int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "denoise3d";
    info->author = "Daniel Moreno, A'rpi";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"High speed 3D Denoiser."
" This filter aims to reduce image noise producing"
" smooth images and making still images really still"
" (This should enhance compressibility)."
"";
}

void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Luma Spatial";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Spatial Luma Strength.";
        break;
      case 1:
        info->name = "Chroma Spatial";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Spatial Chroma Strength.";
        break;
      case 2:
        info->name = "Luma Temporal";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Temporal Luma Strength.";
        break;
      case 3:
        info->name = "Chroma Temporal";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Temporal Chroma Strength.";
        break;
      default:
        break; // XXX
    }
    return;
}

static void denoise3d_setup(denoise3d_instance_t *pd)
{
    int plane_index = 0;

    if((pd->parameter.luma_spatial < 0) || (pd->parameter.luma_temporal < 0))
        pd->enable_luma = 0;
    else
    {
            pd->enable_luma = 1;

        if(pd->parameter.luma_spatial == 0)
        {
            if(pd->parameter.luma_temporal == 0)
            {
                pd->parameter.luma_spatial  = DEFAULT_LUMA_SPATIAL;
                pd->parameter.luma_temporal = DEFAULT_LUMA_TEMPORAL;
            }
            else
            {
                pd->parameter.luma_spatial = pd->parameter.luma_temporal * 3 / 2;
            }
        }
        else
        {
            if(pd->parameter.luma_temporal == 0)
            {
                pd->parameter.luma_temporal = pd->parameter.luma_spatial * 2 / 3;
            }
        }
    }

    if((pd->parameter.chroma_spatial < 0) || (pd->parameter.chroma_temporal < 0))
        pd->enable_chroma = 0;
    else
    {
        pd->enable_chroma = 1;

        if(pd->parameter.chroma_spatial == 0)
        {
            if(pd->parameter.chroma_temporal == 0)
            {
                pd->parameter.chroma_spatial    = DEFAULT_CHROMA_SPATIAL;
                pd->parameter.chroma_temporal   = DEFAULT_CHROMA_TEMPORAL;
            }
            else
            {
                pd->parameter.chroma_spatial = pd->parameter.chroma_temporal * 3 / 2;
            }
        }
        else
        {
            if(pd->parameter.chroma_temporal == 0)
            {
                pd->parameter.chroma_temporal = pd->parameter.chroma_spatial * 2 / 3;
            }
        }
    }

    fprintf(stderr, "[%s] luma enabled: %s, chroma enabled: %s",
            "denoise3d",
            pd->enable_luma ? "yes" : "no",
            pd->enable_chroma ? "yes" : "no");

    for(plane_index = 0; plane_index < MAX_PLANES; plane_index++)
    {
        if((pd->layout_data.layout[plane_index].plane_type == dn3d_luma) && !pd->enable_luma)
            pd->layout_data.layout[plane_index].plane_type = dn3d_disabled;

        if((pd->layout_data.layout[plane_index].plane_type == dn3d_chroma) && !pd->enable_chroma)
            pd->layout_data.layout[plane_index].plane_type = dn3d_disabled;
    }

    PrecalcCoefs(pd->coefficients[0], pd->parameter.luma_spatial);
    PrecalcCoefs(pd->coefficients[1], pd->parameter.luma_temporal);
    PrecalcCoefs(pd->coefficients[2], pd->parameter.chroma_spatial);
    PrecalcCoefs(pd->coefficients[3], pd->parameter.chroma_temporal);

    return;
}

void f0r_set_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    denoise3d_instance_t* inst = (denoise3d_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Luma Spatial */
        inst->parameter.luma_spatial = *((f0r_param_double*)param);
        break;
      case 1: /* Chroma Spatial */
        inst->parameter.chroma_spatial = *((f0r_param_double*)param);
        break;
      case 2: /* Luma Temporal */
        inst->parameter.luma_temporal = *((f0r_param_double*)param);
        break;
      case 3: /* Chroma Temporal */
        inst->parameter.chroma_temporal = *((f0r_param_double*)param);
        break;
      default:
        break; // XXX
    }
    denoise3d_setup(inst);
    return;
}

void f0r_get_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    denoise3d_instance_t* inst = (denoise3d_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Luma Spatial */
        *((f0r_param_double*)param) = (f0r_param_double)inst->parameter.luma_spatial;
        break;
      case 1: /* Chroma Spatial */
        *((f0r_param_double*)param) = (f0r_param_double)inst->parameter.chroma_spatial;
        break;
      case 2: /* Luma Temporal */
        *((f0r_param_double*)param) = (f0r_param_double)inst->parameter.luma_temporal;
        break;
      case 3: /* Chroma Temporal */
        *((f0r_param_double*)param) = (f0r_param_double)inst->parameter.chroma_temporal;
        break;
      default:
        break; // XXX
    }
    return;
}

static int denoise3d_init(denoise3d_instance_t *pd, int pix_fmt,
                          unsigned int width, unsigned int height)
{
    int format_index, found = 0;
    const dn3d_layout_t *lp = NULL;
    size_t size;

    pd->parameter.luma_spatial    = DEFAULT_LUMA_SPATIAL;
    pd->parameter.luma_temporal   = DEFAULT_LUMA_TEMPORAL;
    pd->parameter.chroma_spatial  = DEFAULT_CHROMA_SPATIAL;
    pd->parameter.chroma_temporal = DEFAULT_CHROMA_TEMPORAL;

    denoise3d_setup(pd);

    for(format_index = 0, found = 0; format_index < (sizeof(dn3d_layout) / sizeof(*dn3d_layout)); format_index++)
    {
        if(pix_fmt == dn3d_layout[format_index].pix_fmt)
        {
            found = 1;
            break;
        }
    }
    if (!found) {
        return -1;
    }

    lp = &dn3d_layout[format_index];
    pd->layout_data = *lp;

    size = width * MAX_PLANES * 2;

    pd->lineant = calloc(1, size);
    if(pd->lineant == NULL)
    {
        return -1;
    }

    size *= height * 2;

    pd->previous = calloc(1, size);
    if(pd->previous == NULL)
    {
        return -1;
    }
    return 0;
}

void f0r_destruct(f0r_instance_t instance)
{
    denoise3d_instance_t* inst = (denoise3d_instance_t*)instance;
    RETURN_IF_NULL(inst);
    if(inst->previous)
    {
        free(inst->previous);
        inst->previous = NULL;
    }

    if(inst->lineant)
    {
        free(inst->lineant);
        inst->lineant = NULL;
    }

    free(instance);
    return;
}


f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    denoise3d_instance_t* inst = calloc(1, sizeof(*inst));
    if (inst) {
        int pix_fmt = F0R_COLOR_MODEL_RGBA8888;
        int err = denoise3d_init(inst, pix_fmt, width, height);
        if (err) {
            f0r_destruct(inst);
            inst = NULL;
        }
    }
    return (f0r_instance_t)inst;
}




// FIXME
void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    denoise3d_instance_t* inst = (denoise3d_instance_t*)instance;
    RETURN_IF_NULL(inst);

    int plane_index, coef[2];
    int offset = 0;
    const dn3d_single_layout_t *lp = NULL;

    // FIXME
    memcpy(outframe, inframe1, inst->width * inst->height * RGBA_BYTES);

    for(plane_index = 0; plane_index < MAX_PLANES; plane_index++)
    {
        lp = &inst->layout_data.layout[plane_index];

        if(lp->plane_type != dn3d_disabled)
        {
            coef[0] = (lp->plane_type == dn3d_luma) ? 0 : 2;
            coef[1] = coef[0] + 1;

            // FIXME
            switch(lp->offset)
            {
                case(dn3d_off_r):       offset = 0; break;
                case(dn3d_off_g):       offset = 1; break;
                case(dn3d_off_b):       offset = 2; break;

                case(dn3d_off_y420):    offset = inst->width * inst->height * 0 / 4; break;
                case(dn3d_off_u420):    offset = inst->width * inst->height * 4 / 4; break;
                case(dn3d_off_v420):    offset = inst->width * inst->height * 5 / 4; break;

                case(dn3d_off_y422):    offset = inst->width * inst->height * 0 / 2; break;
                case(dn3d_off_u422):    offset = inst->width * inst->height * 2 / 2; break;
                case(dn3d_off_v422):    offset = inst->width * inst->height * 3 / 2; break;
            }

            deNoise((unsigned char*)outframe,       // frame
                    inst->previous,                 // previous (saved) frame
                    inst->lineant,                  // line buffer
                    inst->width / lp->scale_x,      // width (pixels)
                    inst->height / lp->scale_y,     // height (pixels) // debug
                    inst->coefficients[coef[0]],    // horizontal (spatial) strength
                    inst->coefficients[coef[0]],    // vertical (spatial) strength
                    inst->coefficients[coef[1]],    // temporal strength
                    offset,                         // offset in bytes of first relevant pixel in frame
                    lp->skip                        // skip this amount of bytes between two pixels
            );
        }
    }
    return;
}


void f0r_update(f0r_instance_t instance, double time,
                const uint32_t* inframe, uint32_t* outframe)
{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}

/*
 * vim: expandtab shiftwidth=4:
 */

