/*  detectclipping.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Tilmann Bitterberg - June 2002
 *    Based on Code from mplayers cropdetect by A'rpi
 *    Updated by Antonio Beamud Montero (Microgenesis S.A.) - Jan 2009
 *   
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "frei0r.h"


typedef struct detectclippinginstance_ {
    /* configurable */
    unsigned int start;
    unsigned int end;
    unsigned int step;
    int limit;
    char* log_path;
    /* internal */
    int left, top, right, bottom;
    int width, height;
    int stride;
    int frame_num;
    int boolstep;
} detectclipping_instance_t;

#define RETURN_IF_NULL(INST) do { \
    if (!(INST)) { \
        return; \
    } \
} while (0)

enum {
    RGB_BYTES             = 3, /* how much storage a pixel take? */
    RGBA_BYTES            = 4, /* ditto */
    PIXEL_BLACK_THRESHOLD = 24
};


static void update_step_data(detectclipping_instance_t* inst)
{
    RETURN_IF_NULL(inst);
    if (inst->start % inst->step == 0) {
        inst->boolstep = 0;
    } else {
        inst->boolstep = 1;
    }
    return;
}

enum {
    MACHINE_FRIENDLY = 0,
    HUMAN_FRIENDLY   = 1
};

static int report_stats(detectclipping_instance_t* inst, FILE* out,
                        int verbose)
{
    int top    =                 (inst->top    + 1) & (~1);
    int left   =                 (inst->left   + 1) & (~1);
    int bottom = inst->height - ((inst->bottom + 1) & (~1));
    int right  = inst->width  - ((inst->right  + 1) & (~1));
    if (verbose) { /* human friendlier */
        fprintf(out,
                "detectclipping valid area: X: %i..%i Y: %i..%i  -> %i,%i,%i,%i",
                inst->left, inst->right, inst->top, inst->bottom,
                top, left, bottom, right);
    } else { /* machine friendlier */
        fprintf(out, "%i %i %i %i %i\n",
                inst->frame_num, top, left, bottom, right);
    }
    fflush(out);
    return 0;
}

/* len is in pixels */
static int extimate_black_level(const uint8_t* src, int stride,
                                int len, int bpp)
{
    int total = 0;
    int div = len * bpp;
    while (--len >= 0) {
        /* we don't account the A channel (of course) */
        total += src[0] + src[1] + src[2];
        src += stride;
    }
    total /= div;
    return total;
}


int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "detectclipping";
    info->author = "Tilmann Bitterberg, A'rpi, A. Beamud";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  5;
    info->explanation = ""
"Detect black regions on top, bottom, left and right of an image."
" It is suggested that the filter is run for around 100 frames."
" It will print its detected parameters every frame. If you"
" don't notice any change in the printout for a while, the filter"
" probably won't find any other values."
" The filter converges, meaning it will learn."
"";
}

void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Start";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The first frame to consider.";
        break;
      case 1:
        info->name = "End";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The last frame to consider.";
        break;
      case 2:
        info->name = "Step";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Apply the check every `Step' frames.";
        break;
      case 3:
        info->name = "Limit";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The sum of a line must be below this"
                            " limit to be considered black.";
        break;
      case 4:
        info->name = "Log path";
        info->type = F0R_PARAM_STRING;
        info->explanation = "File to save a detailed statistics.";
        break;
      default:
        break; // XXX
    }
    return;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    detectclipping_instance_t* inst = calloc(1, sizeof(*inst));
    if (inst) {
        inst->end = (unsigned int)-1;
        inst->step = 1;
        inst->limit = PIXEL_BLACK_THRESHOLD;
        inst->log_path = NULL;
        inst->width = width;
        inst->height = height;
        inst->left = width;
        inst->top = height;
        inst->stride = width * RGBA_BYTES;
    }
    return (f0r_instance_t)inst;
}

void f0r_destruct(f0r_instance_t instance)
{
    detectclipping_instance_t* inst = (detectclipping_instance_t*)instance;
    RETURN_IF_NULL(inst);
    if (inst->frame_num > 0) {
        report_stats(inst, stderr, HUMAN_FRIENDLY);
        if (inst->log_path) {
            FILE* out = fopen(inst->log_path, "wt+");
            if (out) {
                report_stats(inst, out, MACHINE_FRIENDLY);
                fclose(out); /* XXX what about failures? */
            } /* XXX else? */
            free(inst->log_path);
        }
        free(instance);
    }
    return;
}

void f0r_set_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    detectclipping_instance_t* inst = (detectclipping_instance_t*)instance;
    f0r_param_string path;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Start */
        inst->start = (unsigned int) *((f0r_param_double*)param);
        update_step_data(inst);
        break;
      case 1: /* End */
        inst->end = (unsigned int) *((f0r_param_double*)param);
        break;
      case 2: /* Step */
        inst->step = (unsigned int) *((f0r_param_double*)param);
        update_step_data(inst);
        break;
      case 3: /* Limit */
        inst->limit = (int) *((f0r_param_double*)param);
        break;
      case 4: /* Log path */
        path = *((f0r_param_string*)param);
        if (strcmp(path, inst->log_path) != 0) {
            free(inst->log_path);
            inst->log_path = strdup(path);
        }
        break;
      default:
        break; // XXX
    }
    return;
}

void f0r_get_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    detectclipping_instance_t* inst = (detectclipping_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Start */
        *((f0r_param_double*)param) = (f0r_param_double)inst->start;
        break;
      case 1: /* End */
        *((f0r_param_double*)param) = (f0r_param_double)inst->end;
        break;
      case 2: /* Step */
        *((f0r_param_double*)param) = (f0r_param_double)inst->step;
        break;
      case 3: /* Limit */
        *((f0r_param_double*)param) = (f0r_param_double)inst->limit;
        break;
      case 4: /* Log path */
        *((f0r_param_string*)param) = inst->log_path;
        break;
      default:
        break; // XXX
    }
    return;
}

static void detectclipping_frame(detectclipping_instance_t* inst,
                                 const uint8_t* frame)
{
    int j;
    /* Is that "unrolling", lazyness or unneeded work?
     * Or something completely different? */
    for (j = 0; j < inst->top; j++) {
        const uint8_t* ptr = frame + inst->stride * j; /* next row */
        int b = extimate_black_level(ptr, RGBA_BYTES, inst->width, RGB_BYTES);
        if (b > inst->limit) {
            inst->top = j;
            break;
        }
    }
    for (j = inst->height - 1; j > inst->bottom; j--) {
        const uint8_t* ptr = frame + inst->stride * j; /* ditto */
        int b = extimate_black_level(ptr, RGBA_BYTES, inst->width, RGB_BYTES);
        if (b > inst->limit) {
            inst->bottom = j;
            break;
        }
    }
    for (j = 0; j < inst->left; j++) {
        const uint8_t* ptr = frame + RGBA_BYTES * j; /* next column*/
        int b = extimate_black_level(ptr, inst->stride, inst->height, RGB_BYTES);
        if (b > inst->limit) {
            inst->left = j;
            break;
        }
    }
    for (j = inst->width - 1; j > inst->right; j--) {
        const uint8_t* ptr = frame + RGBA_BYTES * j; /* ditto */
        int b = extimate_black_level(ptr, inst->stride, inst->height, RGB_BYTES);
        if (b > inst->limit) {
            inst->right = j;
            break;
        }
    }
    return;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    detectclipping_instance_t* inst = (detectclipping_instance_t*)instance;
    RETURN_IF_NULL(inst);
    /* first, let's provide a valid output */
    memcpy(outframe, inframe1, inst->width * inst->height * RGBA_BYTES);
    /* then the actual processing */
    if (inst->start <= inst->frame_num && inst->frame_num <= inst->end
     && inst->frame_num % inst->step == inst->boolstep) {
        detectclipping_frame(inst, (const uint8_t*)inframe1);
    }
    inst->frame_num++;
    return;
}

void f0r_update(f0r_instance_t instance, double time,
                const uint32_t* inframe, uint32_t* outframe)
{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}


/*
 * vim: expandtab shiftwidth=4:
 */

