/* dnr.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Gerhard Monzel - November 2001
 *
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "frei0r.h"
#include "frei0r_math.h"


enum {
    RGBA_BYTES = 4,
    LUT_SIZE   = 256,
    DEFAULT_LT = 10,
    DEFAULT_LL = 4,
    DEFAULT_CT = 16,
    DEFAULT_CL = 8,
    DEFAULT_SC = 30,
};


typedef struct dnrinstance_ {
    int width;
    int height;
   
    int is_first_frame;
    int pPartial;
    int pThreshold;
    int pThreshold2;
    int pPixellock;
    int pPixellock2;
    int pScene;

    int isYUV;
    uint8_t *lastframe;
    uint8_t *origframe;
    int gu_ofs, bv_ofs;

    uint8_t *src_data;
    uint8_t *undo_data;
    long src_h, src_w;
    int img_size;
    int hist_size;
    int pitch;
    int undo;
    int line_size_c;
    int line_size_l;

    unsigned char lookup[LUT_SIZE][LUT_SIZE];
    unsigned char *lockhistory;
} dnr_instance_t;

#define RETURN_IF_NULL(INST) do { \
    if (!(INST)) { \
        return; \
    } \
} while (0)

/* ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "dnr";
    info->author = "Gerhard Monzel";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  5;
    info->explanation = ""
"This filter performs dynamic noise reduction on video frames."
"";
}

void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "lt";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Threshold to blend red. [1, 128]"; /* /luma */
        break;
      case 1:
        info->name = "ll";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Threshold to lock red. [1, 128]"; /* /luma */
        break;
      case 2:
        info->name = "ct";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Threshold to blend green+blue. [1, 128]"; /* /chroma */
        break;
      case 3:
        info->name = "cl";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Threshold to lock green+blue. [1, 128]"; /* /chroma */
        break;
      case 4:
        info->name = "sc";
        info->type = F0R_PARAM_STRING;
        info->explanation = "Percentage of picture differrence"
                            " (scene change). [1, 90 ]";
        break;
      default:
        break; // XXX
    }
    return;
}

static void init_lut(dnr_instance_t *inst)
{
    double low1, low2;
    double high1, high2;
    int a, b, dif1, dif2;
    // setup a biased thresholding difference matrix
    // this is an expensive operation we only want to to once
    for (a = 0; a < LUT_SIZE; a++) {
        for (b = 0; b < LUT_SIZE; b++) {
            // instead of scaling linearly
            // we scale according to the following formulas
            // N = 256
            // val1 = N * (x / N) ^ .9
            // and
            // val2 = N * (x / N) ^ (1/.9)
            // and we choose the maximum distance between two points
            // based on these two scales
            low1 = a;
            low2 = b;
            low1 = low1 / LUT_SIZE;
            low1 = LUT_SIZE * pow(low1, .9);
            low2 = low2 / LUT_SIZE;
            low2 = LUT_SIZE * pow(low2, .9);
            // the low scale should make all values larger
            // and the high scale should make all values smaller
            high1 = a;
            high2 = b;
            high1 = high1 / LUT_SIZE;
            high2 = high2 / LUT_SIZE;
            high1 = LUT_SIZE * pow(high1, 1.0 / .9);
            high2 = LUT_SIZE * pow(high2, 1.0 / .9);
            dif1 = abs((int) (low1 - low2));
//            if (dif1 < 0)
//                dif1 *= -1;
            dif2 = abs((int) (high1 - high2));
//            if (dif2 < 0)
//                dif2 *= -1;
            inst->lookup[a][b] = MAX(dif1, dif2);
        }
    }
    return;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    dnr_instance_t* inst = calloc(1, sizeof(*inst));
    if (inst) {
        //-- PARAMETERS --
        inst->pThreshold = DEFAULT_LT;  // threshold to blend luma/red (default 10)
        inst->pPixellock = DEFAULT_LL;  // threshold to lock luma/red (default 4)
        inst->pThreshold2 = DEFAULT_CT; // threshold to blend croma/green+blue (default 16)
        inst->pPixellock2 = DEFAULT_CL; // threshold to lock croma/green+blue (default 8)
        inst->pScene = DEFAULT_SC;  // percentage of picture difference
        // to interpret as a new scene (default 30%)
        inst->pPartial = 0;         // operating mode [0,1] (default 0)
        //----------------
        inst->isYUV = 0;
        inst->src_h = height;
        inst->src_w = width;
        inst->is_first_frame = 1;
        inst->lastframe = calloc(1, inst->src_h * inst->src_w * RGBA_BYTES);
        inst->origframe = calloc(1, inst->src_h * inst->src_w * RGBA_BYTES);
        inst->lockhistory = calloc(1, inst->src_h * inst->src_w);
        inst->hist_size = inst->src_h * inst->src_w;

        if (inst->isYUV) {
            inst->gu_ofs = inst->hist_size;
            inst->bv_ofs = inst->gu_ofs + (inst->src_h / 2) * (inst->src_w / 2);
            inst->img_size = inst->bv_ofs + (inst->src_h / 2) * (inst->src_w / 2);
            inst->pitch = 1;

            inst->line_size_c = (inst->src_w >> 1);
            inst->line_size_l = inst->src_w;
        } else {
            inst->img_size = inst->hist_size * RGBA_BYTES;
            inst->gu_ofs = 1;
            inst->bv_ofs = 2;
            inst->pitch = RGBA_BYTES;

            inst->line_size_c = inst->src_w * RGBA_BYTES;
            inst->line_size_l = inst->src_w * RGBA_BYTES;
        }
        if (!inst->lastframe || !inst->origframe || !inst->lockhistory) {
            f0r_destruct(inst);
            inst = NULL;
        } else {
            init_lut(inst);
        }
    }
    return (f0r_instance_t)inst;
}

void f0r_destruct(f0r_instance_t instance)
{
    dnr_instance_t* inst = (dnr_instance_t*)instance;
    RETURN_IF_NULL(inst);
    free(inst->lastframe);
    free(inst->origframe);
    free(inst->lockhistory);
    free(instance);
    return;
}

void f0r_set_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    dnr_instance_t* inst = (dnr_instance_t*)instance;
    int value = 0;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* lt */
        value = (int) *((f0r_param_double*)param);
        inst->pThreshold = CLAMP(value, 1, 128);
        break;
      case 1: /* ll */
        value = (int) *((f0r_param_double*)param);
        inst->pPixellock = CLAMP(value, 1, 128);
        break;
      case 2: /* ct */
        value = (int) *((f0r_param_double*)param);
        inst->pThreshold2 = CLAMP(value, 1, 128);
        break;
      case 3: /* cl */
        value = (int) *((f0r_param_double*)param);
        inst->pPixellock2 = CLAMP(value, 1, 128);
        break;
      case 4: /* sc */
        value = (int) *((f0r_param_double*)param);
        inst->pScene = CLAMP(value, 1, 90);
        break;
      default:
        break; // XXX
    }
    return;
}

void f0r_get_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    dnr_instance_t* inst = (dnr_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* lt */
        *((f0r_param_double*)param) = (f0r_param_double)inst->pThreshold;
        break;
      case 1: /* ll */
        *((f0r_param_double*)param) = (f0r_param_double)inst->pPixellock;
        break;
      case 2: /* ct */
        *((f0r_param_double*)param) = (f0r_param_double)inst->pThreshold2;
        break;
      case 3: /* cl */
        *((f0r_param_double*)param) = (f0r_param_double)inst->pPixellock2;
        break;
      case 4: /* sc */
        *((f0r_param_double*)param) = (f0r_param_double)inst->pScene;
        break;
      default:
        break; // XXX
    }
    return;
}

static int dnr_run(dnr_instance_t *inst, uint8_t *data)
{
    uint8_t *RY1, *RY2, *RY3, *GU1, *GU2, *GU3, *BV1, *BV2, *BV3;
    int rl, rc, w, h, update_needed, totpixels;
    int threshRY, threshGU = 0, threshBV = 0;
    int ry1, ry2, gu1 = 0, gu2 = 0, bv1 = 0, bv2 = 0;
    long totlocks = 0;
    uint8_t *lockhistory = inst->lockhistory;

    //-- get data into account --
    inst->src_data = data;

    //-- if we are dealing with the first --
    //-- frame, just make a copy.         --
    if (inst->is_first_frame) {
        ac_memcpy(inst->lastframe, inst->src_data, inst->img_size);
        inst->undo_data = inst->lastframe;
        inst->is_first_frame = 0;

        return 0;
    }

    //-- make sure to preserve the existing frame --
    //-- in case this is a scene change           --
    ac_memcpy(inst->origframe, inst->src_data, inst->img_size);

    if (inst->isYUV) {
        RY1 = inst->src_data;
        GU1 = RY1 + inst->gu_ofs;
        BV1 = RY1 + inst->bv_ofs;   
   
        RY2 = inst->lastframe;
        GU2 = RY2 + inst->gu_ofs;
        BV2 = RY2 + inst->bv_ofs;  

        RY3 = inst->src_data;    
        GU3 = RY3 + inst->gu_ofs;
        BV3 = RY3 + inst->bv_ofs;  
    } else {
        BV1 = inst->src_data;
        GU1 = BV1 + inst->gu_ofs;
        RY1 = BV1 + inst->bv_ofs;   
   
        BV2 = inst->lastframe;
        GU2 = BV2 + inst->gu_ofs;
        RY2 = BV2 + inst->bv_ofs;  

        BV3 = inst->src_data;    
        GU3 = BV3 + inst->gu_ofs;
        RY3 = BV3 + inst->bv_ofs;
    }

    h = inst->src_h;
    do {
        w = inst->src_w;
        rl = rc = 0;
        do {
            update_needed = 1;

            //-- on every row get (luma) actual/locked pixels --
            //-- -> calculate thresold (biased diff.)         --
            //--------------------------------------------------
            ry1 = RY1[rl];
            ry2 = RY2[rl];
            threshRY = inst->lookup[ry1][ry2];

            //-- in YUV-Mode on every even row  (RGB every --
            //-- row) get (chroma) actual/locked pixels    --
            //-- -> calculate thresold (biased diff.)      --
            //-----------------------------------------------
            if (!inst->isYUV || !(rl & 0x01)) {
                gu1 = GU1[rc];
                bv1 = BV1[rc];
                gu2 = GU2[rc];
                bv2 = BV2[rc];
                threshGU = inst->lookup[gu1][gu2];
                threshBV = inst->lookup[bv1][bv2];
            }

            //-- PARTIAL --
            //-------------
            if (inst->pPartial) {
                // we're doing a full pixel lock since we're --
                // under all thresholds in a couple of time  --
                //---------------------------------------------
                if ((threshRY < inst->pPixellock)
                    && (threshGU < inst->pPixellock2)
                    && (threshBV < inst->pPixellock2)) {

                    //-- if we've locked more than 30 times at --
                    //-- this point, let's refresh the pixel.  --
                    if (*lockhistory > 30) {
                        *lockhistory = 0;

                        ry1 = (ry1 + ry2) / 2;
                        gu1 = (gu1 + gu2) / 2;
                        bv1 = (bv1 + bv2) / 2;
                    }
                    else {
                        *lockhistory = *lockhistory + 1;

                        //-- take locked pixels --
                        ry1 = ry2;
                        gu1 = gu2;
                        bv1 = bv2;
                    }
                } else if ((threshRY < inst->pPixellock)
                         && (threshGU < inst->pThreshold2)
                         && (threshBV < inst->pThreshold2)) {
                    //-- If the luma is within pixellock, and the chroma is within   --
                    //-- blend, lets blend the chroma and lock the luma.
                    //-----------------------------------------------------------------
                    *lockhistory = 0;

                    ry1 = ry2;
                    gu1 = (gu1 + gu2) / 2;
                    bv1 = (bv1 + bv2) / 2;
                } else if ((threshRY < inst->pThreshold)
                         && (threshGU < inst->pThreshold2)
                         && (threshBV < inst->pThreshold2)) {
                    //-- We are above pixellock in luma and chroma, but     --
                    //-- below the blend thresholds in both, so let's blend --
                    //--------------------------------------------------------
                    *lockhistory = 0;

                    ry1 = (ry1 + ry2) / 2;
                    gu1 = (gu1 + gu2) / 2;
                    bv1 = (bv1 + bv2) / 2;
                } else {
                    //-- if we are above all thresholds, --
                    //-- just leave the output untouched --
                    //-------------------------------------
                    *lockhistory = 0;
                    update_needed = 0;
                    totlocks++;
                }
            }
            //-- nonPARTIAL --
            //----------------
            else {
                //-- beneath pixellock so lets keep   --
                //-- the existing pixel (most likely) --
                //--------------------------------------
                if ((threshRY < inst->pPixellock)
                    && (threshGU < inst->pPixellock2)
                    && (threshBV < inst->pPixellock2)) {
                    // if we've locked more than 30 times at this point,
                    // let's refresh the pixel
                    if (*lockhistory > 30) {
                        *lockhistory = 0;

                        ry1 = (ry1 + ry2) / 2;
                        gu1 = (gu1 + gu2) / 2;
                        bv1 = (bv1 + bv2) / 2;
                    } else {
                        *lockhistory = *lockhistory + 1;

                        ry1 = ry2;
                        gu1 = gu2;
                        bv1 = bv2;
                    }
                } else if ((threshRY < inst->pThreshold) &&
                         (threshGU < inst->pThreshold2) &&
                         (threshBV < inst->pThreshold2)) {
                    //-- we are above pixellock, but below the --
                    //-- blend threshold so we want to blend   --
                    //-------------------------------------------
                    *lockhistory = 0;

                    ry1 = (ry1 + ry2) / 2;
                    gu1 = (gu1 + gu2) / 2;
                    bv1 = (bv1 + bv2) / 2;
                } else {
                    //-- it's beyond the thresholds, just leave it alone --
                    //-----------------------------------------------------
                    *lockhistory = 0;
                    update_needed = 0;
                    totlocks++;
                }
            }

            //-- set destination --
            //---------------------
            if (update_needed) {
                RY3[rl] = ry1;
                GU3[rc] = gu1;
                BV3[rc] = bv1;
            }

            //-- refresh locked pixels --
            //---------------------------
            if (*lockhistory == 0) {
                RY2[rl] = ry1;
                GU2[rc] = gu1;
                BV2[rc] = bv1;
            }

            lockhistory++;

            rl += inst->pitch;
            rc = (inst->isYUV) ? (rl >> 1) : rl;

        } while (--w);

        //-- next line ... --
        RY1 += inst->line_size_l;
        RY2 += inst->line_size_l;
        RY3 += inst->line_size_l;

        //-- ... in YUV-Mode for chromas, only on even luma-lines --
        if (!inst->isYUV || !(h & 0x01)) {
            GU1 += inst->line_size_c;
            BV1 += inst->line_size_c;

            GU2 += inst->line_size_c;
            BV2 += inst->line_size_c;

            GU3 += inst->line_size_c;
            BV3 += inst->line_size_c;
        }

    } while (--h);

    totpixels = inst->src_h * inst->src_w;
    totpixels *= inst->pScene;
    totpixels /= 100;

    // If more than the specified percent of pixels have exceeded all thresholds
    // then we restore the saved frame.  (this doesn't happen very often
    // hopefully)  We also set the pixellock history to 0 for all frames

    if (totlocks > totpixels) {
        uint8_t *ptmp = inst->lastframe;

        inst->lastframe = inst->origframe;
        inst->undo_data = inst->lastframe;
        inst->origframe = ptmp;
        inst->undo = 1;

        memset(inst->lockhistory, 0, inst->hist_size);
    } else {
        inst->undo_data = inst->src_data;
        inst->undo = 0;
    }

    if (inst->undo) {
        ac_memcpy(data, inst->undo_data, inst->img_size);
        inst->undo = 0;
    }
    return 0;
}


static void f0r_dnr(dnr_instance_t *inst,
                    const uint8_t *src, uint8_t *dst)
{
    /* FIXME: I'm horrofied at myself. May Knuth forgive me. */
    /* FIXME: I wrote this too much times. */
    ac_memcpy(dst, src, inst->img_size);
    dnr_run(inst, dst);
    return;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    dnr_instance_t* inst = (dnr_instance_t*)instance;
    RETURN_IF_NULL(inst);
    f0r_dnr(inst, (const uint8_t*)inframe1, (uint8_t*)outframe);
    return;
}

void f0r_update(f0r_instance_t instance, double time,
                const uint32_t* inframe, uint32_t* outframe)
{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}


/*
 * vim: expandtab shiftwidth=4:
 */

