/*  facemask.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Julien Tierny <julien.tierny@wanadoo.fr> - October 2004
 *  Copyright (C) Thomas Oestreich - June 2001
 *   
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "frei0r.h"


typedef struct facemaskinstance_ {
    int       xpos;
    int       ypos;
    int       xresolution;
    int       yresolution;
    int       xdim;
    int       ydim;
    int       width;
    int       height;
} facemask_instance_t;

#define RETURN_IF_NULL(INST) do { \
    if (!(INST)) { \
        return; \
    } \
} while (0)

enum {
    RGB_BYTES  = 3, /* how much storage a pixel take? */
    RGBA_BYTES = 4, /* ditto */
};

#if 0
static int check_parameters(facemask_instance_t *inst)
{

    /* First, we check if the face-zone is contained in the picture */
    if ((inst->xpos + inst->xdim) > inst->width){
        /* "Face zone is larger than the picture!" */
        return -1;
    }
    if ((inst->ypos + inst->ydim) > inst->height){
        /* "Face zone is taller than the picture!" */
        return -1;
    }

    /* Then, we check the resolution */
    if ((inst->ydim % inst->yresolution) != 0) {
        /* "Incorrect Y resolution!" */
        return -1;
    }
    if ((inst->xdim % inst->xresolution) != 0) {
        /* "Incorrect X resolution! */
        return -1;
    }
    return 0;
}
#endif

static int average_neighbourhood(int x, int y, int w, int h,
                                 int width,
                                 const uint8_t *src, uint8_t *dst)
{
    uint32_t red=0, green=0, blue=0;
    int      i=0,j=0,s=RGBA_BYTES;/* assume srcstride=dststride */

    for (j=y; j<=y+h; j++){
        for (i=s*(x + width*(j-1)); i<s*(x + w + (j-1)*width); i+=s){
            red     += (int) src[i  ];
            green   += (int) src[i+1];
            blue    += (int) src[i+2];
        }
    }

    red     /= ((w+1)*h);
    green   /= ((w+1)*h);
    blue    /= ((w+1)*h);

    for (j=y; j<y+h; j++) {
        for (i=s*(x + width*(j-1)); i<s*(x + w + (j-1)*width); i+=s) {
            dst[i  ]     = (char)red;
            dst[i+1]     = (char)green;
            dst[i+2]     = (char)blue;
        }
    }
    return 0;
}

static int draw_mask(int x, int y, int w, int h, int W, int H,
                      int width, const uint8_t *src, uint8_t *dst)
{
    int i = 0,j = 0;
    for (j=y; j<=y+H; j+=h) {
        for (i=x; i<=x+W; i+=w) {
            average_neighbourhood(i, j, w, h, width, src, dst);
        }
    }
    return 0;
}


int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "facemask";
    info->author = "Julien Tierny";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  6;
    info->explanation = "Mask people faces in video interviews."
" Warning:"
" You have to calibrate by your own the mask dimensions"
" and positions so as it fits to your video sample."
" You also have to choose a resolution that is multiple"
" of the mask dimensions."
"";
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Xpos";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Position of the upper left"
                            " corner of the mask (x).";
        break;
      case 1:
        info->name = "Ypos";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Position of the upper left"
                            " corner of the mask (y).";
        break;
      case 2:
        info->name = "Xresolution";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Resolution of the mask (width).";
        break;
      case 3:
        info->name = "Yresolution";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Resolution of the mask (height).";
        break;
      case 4:
        info->name = "Xdim";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Width of the mask (= n*xresolution).";
        break;
      case 5:
        info->name = "Ydim";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Height of the mask (= m*yresolution).";
      default:
        break; // XXX
    }
    return;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    facemask_instance_t* inst = calloc(1, sizeof(*inst));
    if (inst) {
        inst->xpos           = 0;
        inst->ypos           = 0;
        inst->xresolution    = 1;
        inst->yresolution    = 1;
        inst->xdim           = 1;
        inst->ydim           = 1;
        inst->width          = width;
        inst->height         = height;
    }
    return (f0r_instance_t)inst;
}

void f0r_destruct(f0r_instance_t instance)
{
    facemask_instance_t* inst = (facemask_instance_t*)instance;
    RETURN_IF_NULL(inst);
    /* nothing to do! */
    free(instance);
    return;
}

void f0r_set_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    facemask_instance_t* inst = (facemask_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Xpos */
        inst->xpos = (int) *((f0r_param_double*)param);
        break;
      case 1: /* Ypos */
        inst->ypos = (int) *((f0r_param_double*)param);
        break;
      case 2: /* Xresolution */
        inst->xresolution = (int) *((f0r_param_double*)param);
        break;
      case 3: /* Yresolution */
        inst->yresolution = (int) *((f0r_param_double*)param);
        break;
      case 4: /* Xdim */
        inst->xdim = (int) *((f0r_param_double*)param);
        break;
      case 5: /* Ydim */
        inst->ydim = (int) *((f0r_param_double*)param);
        break;
      default:
        break; // XXX
    }
/*  check_parameters(inst); TODO */
    return;
}

void f0r_get_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    facemask_instance_t* inst = (facemask_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Xpos */
        *((f0r_param_double*)param) = (f0r_param_double)inst->xpos;
        break;
      case 1: /* Ypos */
        *((f0r_param_double*)param) = (f0r_param_double)inst->ypos;
        break;
      case 2: /* Xresolution */
        *((f0r_param_double*)param) = (f0r_param_double)inst->xresolution;
        break;
      case 3: /* Yresolution */
        *((f0r_param_double*)param) = (f0r_param_double)inst->yresolution;
        break;
      case 4: /* Xdim */
        *((f0r_param_double*)param) = (f0r_param_double)inst->xdim;
        break;
      case 5: /* Ydim */
        *((f0r_param_double*)param) = (f0r_param_double)inst->ydim;
        break;
      default:
        break; // XXX
    }
    return;
}


void f0r_update2(f0r_instance_t instance,
		 double time,
		 const uint32_t* inframe1,
		 const uint32_t* inframe2,
		 const uint32_t* inframe3,
		 uint32_t* outframe)
{
    facemask_instance_t* inst = (facemask_instance_t*)instance;
    RETURN_IF_NULL(inst);
    draw_mask(inst->xpos, inst->ypos,
              inst->xresolution, inst->yresolution,
              inst->xdim, inst->ydim,
              inst->width,
              (const uint8_t *)inframe1,
              (uint8_t *)outframe);
    return;
}

void f0r_update(f0r_instance_t instance, double time,
                const uint32_t* inframe, uint32_t* outframe)
{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}


/*
 * vim: expandtab shiftwidth=4:
 */

