/*  mask.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Thomas Oestreich, Chad Page - February, March 2002
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "frei0r.h"

enum {
    PLUGIN_PARAM_NUM = 4
};

typedef struct maskinstance_ mask_instance_t;

typedef void (*process_mask_t)(mask_instance_t *inst, uint8_t *buffer, int begin, int end);

struct maskinstance_ {
    int width;
    int height; 
 
    int lc;
    int rc;
    int tc;
    int bc;

    process_mask_t xmask_left;
    process_mask_t xmask_right;
    process_mask_t ymask_top;
    process_mask_t ymask_bottom;
};

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

/* ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

enum {
    RGB_BYTES             = 3, /* how much storage a pixel take? */
    RGBA_BYTES            = 4, /* ditto */
    PIXEL_BLACK_THRESHOLD = 24
};

/*************************************************************************/

static void ymask_rgb(mask_instance_t *inst, uint8_t *buffer, int top, int bottom)
{
    int i;

    for (i = top; i <= bottom; i++) {
        memset(&buffer[i * inst->width * RGBA_BYTES], 0, inst->width * RGBA_BYTES);
    }
}

static void xmask_rgb(mask_instance_t *inst, uint8_t *buffer, int left, int right)
{
    int y;
    unsigned char *ptr;

    for (y = 0; y < inst->height; y++) {
        ptr = &buffer[(y * inst->width * RGBA_BYTES) + (left * RGBA_BYTES)];
        memset(ptr, 0, (right - left) * RGBA_BYTES);
    }
}

static void mask_null(mask_instance_t *inst, uint8_t *buffer, int begin, int end)
{
    return; /* nothing to do here */
}  

/*************************************************************************/

int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "mask";
    info->author = "Thomas Oestreich, Chad Page";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"Filter through a rectangular Mask."
" This filter applies an rectangular mask to the video."
" Everything outside the mask is set to black."
"";
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Left";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "X value of Upper left corner of the box.";
        break;
      case 1:
        info->name = "Top";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Y value of Upper left corner of the box.";
        break;
      case 2:
        info->name = "Right";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "X value of Lower right corner of the box.";
        break;
      case 3: 
        info->name = "Bottom";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Y value of Lower right corner of the box.";
        break;
      default:
        break; // XXX
    }
    return;
}

static void mask_init(mask_instance_t *inst,
                      unsigned int width, unsigned int height)
{
    inst->width        = width;
    inst->height       = height;
    inst->lc           = 0;
    inst->tc           = 0;
    inst->rc           = inst->width;
    inst->bc           = inst->height;
    inst->xmask_left   = xmask_rgb;
    inst->xmask_right  = xmask_rgb;
    inst->ymask_top    = ymask_rgb;
    inst->ymask_bottom = ymask_rgb;
    return;
}


f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    mask_instance_t* inst = calloc(1, sizeof(*inst));
    if (inst) {
        mask_init(inst, width, height);
    }
    return (f0r_instance_t)inst;
}

void f0r_destruct(f0r_instance_t instance)
{
    mask_instance_t* inst = (mask_instance_t*)instance;
    RETURN_IF_NULL(inst);
    /* nothing more to do */
    free(instance);
    return;
}

void f0r_set_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    mask_instance_t* inst = (mask_instance_t*)instance;
    int value = 0;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Left */
        value = (int) *((f0r_param_double*)param);
        inst->lc = value;
        if (inst->lc > 2) {
            inst->xmask_left = xmask_rgb;
        } else {
            inst->xmask_left = mask_null;
        }
        break;
      case 1: /* Top */
        value = (int) *((f0r_param_double*)param);
        inst->tc = value;
        if (inst->tc > 2) {
            inst->ymask_top = ymask_rgb;
        } else {
            inst->ymask_top = mask_null;
        }
        break;
      case 2: /* Right */
        value = (int) *((f0r_param_double*)param);
        inst->rc = value;
        if ((inst->width - inst->rc) > 1) {
            inst->xmask_right = xmask_rgb;
        } else {
            inst->xmask_right = mask_null;
        }
        break;
      case 3: /* Bottom */
        value = (int) *((f0r_param_double*)param);
        inst->bc = value;
        if ((inst->height - inst->bc) > 1) {
            inst->ymask_bottom = ymask_rgb;
        } else {
            inst->ymask_bottom = mask_null;
        }
        break;
      default:
        break; // XXX
    }
    return;
}

void f0r_get_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    mask_instance_t* inst = (mask_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Left */
        *((f0r_param_double*)param) = (f0r_param_double)inst->lc;
        break;
      case 1: /* Top */
        *((f0r_param_double*)param) = (f0r_param_double)inst->tc;
        break;
      case 2: /* Right */
        *((f0r_param_double*)param) = (f0r_param_double)inst->rc;
        break;
      case 3: /* Bottom */
        *((f0r_param_double*)param) = (f0r_param_double)inst->bc;
        break;
      default:
        break; // XXX
    }
    return;
}


void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    mask_instance_t* inst = (mask_instance_t*)instance;
    uint8_t *out = (uint8_t *)outframe;
    size_t img_size = 0;
    RETURN_IF_NULL(inst);
    img_size = inst->width * inst->height * RGBA_BYTES;
    ac_memcpy(outframe, inframe1, img_size);
    inst->ymask_top(inst,    out, 0,        inst->tc - 1);
    inst->ymask_bottom(inst, out, inst->bc, inst->height - 1);
    inst->xmask_left(inst,   out, 0,        inst->lc - 1);
    inst->xmask_right(inst,  out, inst->rc, inst->width - 1);
    return;
}


void f0r_update(f0r_instance_t instance, 
                double time, const uint32_t* inframe, uint32_t* outframe)
{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}

/*
 * vim: expandtab shiftwidth=4:
 */

