/*  msharpen.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 * Copyright (C) 1999-2000 Donal A. Graft
 *   modified 2003 by William Hawkins for use with transcode
 *
 * MSharpen Filter for VirtualDub -- performs sharpening
 * limited to edge areas of the frame.
 * Copyright (C) 1999-2000 Donald A. Graft
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author can be contacted at:
 * Donald Graft
 * neuron2@home.com.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "frei0r.h"
#include "frei0r_math.h"


typedef struct msharpeninstance_ {
    /* parameters */
    int strength;
    int threshold;
    int mask;
    int highq;
    /* private */
    uint8_t *blur;
    uint8_t *work;
    int width;
    int height;
} msharpen_instance_t;

#define RETURN_IF_NULL(INST) do { \
    if (!(INST)) { \
        return; \
    } \
} while (0)

enum {
    DEFAULT_STRENGTH = 100, /* A little bird told me this was a good value */
    DEFAULT_THRESHOLD = 10,
    BGRA_BYTES        = 4
};

/* XXX ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "msharpen";
    info->author = "Donald Graft, William Hawkins";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_BGRA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  4;
    info->explanation = ""
"This plugin implements an unusual concept in spatial sharpening."
" Although designed specifically for anime, it also works well with"
" normal video. The filter is very effective at sharpening important"
"edges without amplifying noise."
"";
}

void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Strenght";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = ""
"This is the strength of the sharpening to be applied to the edge"
" detail areas. It is applied only to the edge detail areas as"
" determined by the 'threshold' parameter. Strength 255 is the"
" strongest sharpening. 0 is the weakest."
"";
        break;
      case 1:
        info->name = "Threshold";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = ""
"This parameter determines what is detected as edge detail and"
" thus sharpened. To see what edge detail areas will be sharpened,"
" use the 'mask' parameter. The accepted range is [0, 255]"
"";
        break;
      case 2:
        info->name = "Mask";
        info->type = F0R_PARAM_BOOL;
        info->explanation = ""
"When set to true, the areas to be sharpened are shown in white"
" against a black background. Use this to set the level of detail to"
" be sharpened. This function also makes a basic edge detection filter."
"";
        break;
      case 3:
        info->name = "HighQ";
        info->type = F0R_PARAM_BOOL;
        info->explanation = ""
"This parameter lets you tradeoff speed for quality of detail"
" detection. Set it to true for the best detail detection."
" Set it to false for maximum speed."
"";
        break;
      default:
        break; // XXX
    }
    return;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    msharpen_instance_t* inst = calloc(1, sizeof(*inst));
    if (inst) {
        inst->width = width;
        inst->height = height;
        inst->strength = DEFAULT_STRENGTH;
        inst->threshold = DEFAULT_THRESHOLD;
        inst->mask = 0;
        inst->highq = 1;
        /* fetch memory */
        inst->blur = calloc(1, BGRA_BYTES * width * height); // XXX
        inst->work = calloc(1, BGRA_BYTES * width * height); // XXX
        if (!inst->blur || !inst->work) {
            f0r_destruct(inst);
            inst = NULL;
            // XXX
        }
    }
    return (f0r_instance_t)inst;
}

void f0r_destruct(f0r_instance_t instance)
{
    msharpen_instance_t* inst = (msharpen_instance_t*)instance;
    RETURN_IF_NULL(inst);
    free(inst->blur);
    free(inst->work);
    free(instance);
    return;
}

void f0r_set_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    msharpen_instance_t* inst = (msharpen_instance_t*)instance;
    int32_t value;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Strength */
        value = (int32_t) *((f0r_param_double*)param);
        inst->strength = CLAMP0255(value);
        break;
      case 1: /* Threshold */
        value = (int32_t) *((f0r_param_double*)param);
        inst->threshold = CLAMP0255(value);
        break;
      case 2: /* Mask */
        inst->mask = (f0r_param_bool) *((f0r_param_double*)param);
      case 3: /* HighQ */
        inst->highq = (f0r_param_bool) *((f0r_param_double*)param);
        break;
      default:
        break; // XXX
    }
    return;
}

void f0r_get_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    msharpen_instance_t* inst = (msharpen_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Strength */
        *((f0r_param_double*)param) = (f0r_param_double)inst->strength;
        break;
      case 1: /* Threshold */
        *((f0r_param_double*)param) = (f0r_param_double)inst->threshold;
        break;
      case 2: /* Mask */
        *((f0r_param_bool*)param) = (f0r_param_bool)inst->mask;
        break;
      case 3: /* HighQ */
        *((f0r_param_bool*)param) = (f0r_param_bool)inst->highq;
        break;
      default:
        break; // XXX
    }
    return;
}

static void f0r_msharpen(msharpen_instance_t *inst,
                         const uint8_t* src, uint8_t* dst)
{
    const int  width  = inst->width;
    const int  height = inst->height;
    const long pitch  = inst->width * BGRA_BYTES;
    const int  dstpitch = inst->width * BGRA_BYTES;
    int        bwidth = width * BGRA_BYTES;
    const uint8_t* srcpp = NULL;
    const uint8_t* srcp = NULL;
    const uint8_t *srcpn = NULL;
    uint8_t *workp, *blurp, *blurpn, *dstp;
    int r1, r2, r3, r4, g1, g2, g3, g4, b1, b2, b3, b4;
    int x, y, max, strength, invstrength, threshold;

    threshold   = inst->threshold;
    strength    = inst->strength;
    invstrength = 255 - strength;

    /* Blur the source image prior to detail detection. Separate
       dimensions for speed. */
    /* Vertical. */
    srcpp = src;
    srcp  = srcpp + pitch;
    srcpn = srcp + pitch;
    workp = inst->work + bwidth;
    for (y = 1; y < height - 1; y++) {
        for (x = 0; x < bwidth; x++) {
            workp[x] = (srcpp[x] + srcp[x] + srcpn[x]) / 3;
        }
        srcpp += pitch;
        srcp  += pitch;
        srcpn += pitch;
        workp += bwidth;
    }

    /* Horizontal. */
    workp  = inst->work;
    blurp  = inst->blur;
    for (y = 0; y < height; y++) {
        for (x = 4; x < bwidth - 4; x++) {
            blurp[x] = (workp[x-4] + workp[x] + workp[x+4]) / 3;
        }
        workp += bwidth;
        blurp += bwidth;
    }

    /* Fix up blur frame borders. */
    srcp  = src;
    blurp = inst->blur;
    ac_memcpy(blurp, srcp, bwidth);
    ac_memcpy(blurp + (height-1)*bwidth, srcp + (height-1)*pitch, bwidth);
    for (y = 0; y < height; y++) {
        *((uint32_t*)(&blurp[0])) = *((uint32_t*)(&srcp[0]));
        *((uint32_t*)(&blurp[bwidth-4])) = *((uint32_t*)(&srcp[bwidth-4]));
        srcp  += pitch;
        blurp += bwidth;
    }

    /* Diagonal detail detection. */
    blurp  = inst->blur;
    blurpn = blurp + bwidth;
    workp  = inst->work;
    for (y = 0; y < height - 1; y++) {
        b1 = blurp[0];
        g1 = blurp[1];
        r1 = blurp[2];
        b3 = blurpn[0];
        g3 = blurpn[1];
        r3 = blurpn[2];
        for (x = 0; x < bwidth - 4; x+=4) {
            b2 = blurp[x+4];
            g2 = blurp[x+5];
            r2 = blurp[x+6];
            b4 = blurpn[x+4];
            g4 = blurpn[x+5];
            r4 = blurpn[x+6];
            if ((abs(b1 - b4) >= threshold) || (abs(g1 - g4) >= threshold) || (abs(r1 - r4) >= threshold)
             || (abs(b2 - b3) >= threshold) || (abs(g2 - g3) >= threshold) || (abs(g2 - g3) >= threshold)) {
                *((uint32_t*)(&workp[x])) = 0xffffffff;
            } else {
                *((uint32_t*)(&workp[x])) = 0x0;
            }
            b1 = b2;
            b3 = b4;
            g1 = g2;
            g3 = g4;
            r1 = r2;
            r3 = r4;
        }
        workp  += bwidth;
        blurp  += bwidth;
        blurpn += bwidth;
        (void)r3; /* I haven't yet figured out if this variable can be drop. */
    }

    if (inst->highq) {
        /* Vertical detail detection. */
        for (x = 0; x < bwidth; x += 4) {
            blurp  = inst->blur;
            blurpn = blurp + bwidth;
            workp  = inst->work;
            b1 = blurp[x];
            g1 = blurp[x+1];
            r1 = blurp[x+2];
            for (y = 0; y < height - 1; y++) {
                b2 = blurpn[x];
                g2 = blurpn[x+1];
                r2 = blurpn[x+2];
                if (abs(b1 - b2) >= threshold || abs(g1 - g2) >= threshold || abs(r1 - r2) >= threshold) {
                    *((uint32_t*)(&workp[x])) = 0xffffffff;
                }
                b1 = b2;
                g1 = g2;
                r1 = r2;
                workp += bwidth;
                blurp += bwidth;
                blurpn += bwidth;
            }
        }

        /* Horizontal detail detection. */
        blurp = inst->blur;
        workp = inst->work;
        for (y = 0; y < height; y++) {
            b1 = blurp[0];
            g1 = blurp[1];
            r1 = blurp[2];
            for (x = 0; x < bwidth - 4; x += 4) {
                b2 = blurp[x+4];
                g2 = blurp[x+5];
                r2 = blurp[x+6];
                if (abs(b1 - b2) >= threshold || abs(g1 - g2) >= threshold || abs(r1 - r2) >= threshold) {
                    *((uint32_t*)(&workp[x])) = 0xffffffff;
                }
                b1 = b2;
                g1 = g2;
                r1 = r2;
            }
            workp += bwidth;
            blurp += bwidth;
        }
    }

    /* Fix up detail map borders. */
    memset(inst->work + (height-1)*bwidth, 0, bwidth);
    workp = inst->work;
    for (y = 0; y < height; y++) {
        *((uint32_t*)(&workp[bwidth-4])) = 0;
        workp += bwidth;
    }

    if (inst->mask) {
        workp = inst->work;
        dstp  = dst;
        for (y = 0; y < height; y++) {
            for (x = 0; x < bwidth; x++) {
                dstp[x] = workp[x];
            }
            workp += bwidth;
            dstp = dstp + dstpitch;
        }
        return;
    }

    /* Fix up output frame borders. */
    srcp = src;
    dstp = dst;
    ac_memcpy(dstp, srcp, bwidth);
    ac_memcpy(dstp + (height-1)*pitch, srcp + (height-1)*pitch, bwidth);
    for (y = 0; y < height; y++) {
        *((uint32_t*)(&dstp[0])) = *((uint32_t*)(&srcp[0]));
        *((uint32_t*)(&dstp[bwidth-4])) = *((uint32_t*)(&srcp[bwidth-4]));
        srcp += pitch;
        dstp += pitch;
    }

    /* Now sharpen the edge areas and we're done! */
    srcp = src + pitch;
    dstp = dst + pitch;
    workp = inst->work + bwidth;
    blurp = inst->blur + bwidth;
    for (y = 1; y < height - 1; y++) {
        for (x = 4; x < bwidth - 4; x+=4) {
            int xplus1 = x + 1, xplus2 = x + 2;

            if (workp[x]) {
                b4 = (4*(int)srcp[x] - 3*blurp[x]);
                g4 = (4*(int)srcp[x+1] - 3*blurp[x+1]);
                r4 = (4*(int)srcp[x+2] - 3*blurp[x+2]);

                if (b4 < 0) b4 = 0;
                if (g4 < 0) g4 = 0;
                if (r4 < 0) r4 = 0;
                max = b4;
                if (g4 > max) max = g4;
                if (r4 > max) max = r4;
                if (max > 255) {
                    b4 = (b4 * 255) / max;
                    g4 = (g4 * 255) / max;
                    r4 = (r4 * 255) / max;
                }
                dstp[x]      = (strength * b4 + invstrength * srcp[x])      >> 8;
                dstp[xplus1] = (strength * g4 + invstrength * srcp[xplus1]) >> 8;
                dstp[xplus2] = (strength * r4 + invstrength * srcp[xplus2]) >> 8;
            } else {
                dstp[x]   = srcp[x];
                dstp[xplus1] = srcp[xplus1];
                dstp[xplus2] = srcp[xplus2];
            }
        }
        srcp  += pitch;
        dstp  += pitch;
        workp += bwidth;
        blurp += bwidth;
    }
    return;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    msharpen_instance_t* inst = (msharpen_instance_t*)instance;
    RETURN_IF_NULL(inst);
    f0r_msharpen(inst, (const uint8_t*)inframe1, (uint8_t*)outframe);
    return;
}

void f0r_update(f0r_instance_t instance, double time,
                const uint32_t* inframe, uint32_t* outframe)
{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}


/*
 * vim: expandtab shiftwidth=4:
 */

