/*  whitebalance.c
 *  (C) 2012-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  White Balance Filter - correct images with a broken white balance
 *  (typically, images from a dv camcorder with an unset white balance or
 *  wrongly forced to indoor or outdoor)
 *
 *  Copyright (C) 2003 Guillaume Cottenceau <gc at mandrakesoft.com>
 *
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <math.h>

#include "frei0rx.h"

typedef struct whitebalanceinstance_ {
    int level;
    uint8_t red_filter[256];
    uint8_t blue_filter[256];
    unsigned int start;
    unsigned int end;
    unsigned int framenum;
    unsigned int width;
    unsigned int height;
} whitebalance_instance_t;

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

enum {
    RGB_BYTES           = 3,  /* how much storage a pixel take? */
    RGBA_BYTES          = 4,  /* ditto */
    DEFAULT_LEVEL       = 40, /* magic number inherited
                                 from the original code */
    PLUGIN_PARAM_NUM    = 3
};

// TODO

static void build_maps(whitebalance_instance_t *inst)
{
    int i = 0;
    double factor = 1 + ((double)abs(inst->level))/100;
    if (inst->level < 0) {
        factor = 1/factor;
    }
    /* preprocess filters for performance */
    for (i = 0; i < 256; i++) {
        inst->red_filter[i]  = pow(((double) i)/255, 1/factor) * 255;
        inst->blue_filter[i] = pow(((double) i)/255, factor)   * 255;
    }
    return;
}



int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "whitebalance";
    info->author = "Guillaume Cottenceau";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"White Balance Filter - correct images with a broken white balance"
" (typically, images from a dv camcorder with an unset white balance or"
" wrongly forced to indoor or outdoor)."
"";
}

void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Start";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The first frame to consider.";
        break;
      case 1:
        info->name = "End";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The last frame to consider.";
        break;
      case 2:
        info->name = "Level";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Level of blue-to-yellow white balance"
                            " shifting(can be negative).";
        break;
      default:
        break; // XXX
    }
    return;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    whitebalance_instance_t* inst = calloc(1, sizeof(*inst));
    if (inst) {
        inst->width = width;
        inst->height = height;
        inst->level = DEFAULT_LEVEL;
        inst->end = -1;
        build_maps(inst);
    }
    return (f0r_instance_t)inst;
}

void f0r_destruct(f0r_instance_t instance)
{
    whitebalance_instance_t* inst = (whitebalance_instance_t*)instance;
    RETURN_IF_NULL(inst);
    /* nothing to do here */
    free(instance);
    return;
}

void f0r_set_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    whitebalance_instance_t* inst = (whitebalance_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Start */
        inst->start = (unsigned int) *((f0r_param_double*)param);
        break;
      case 1: /* End */
        inst->end = (unsigned int) *((f0r_param_double*)param);
        break;
      case 2: /* Level */
        inst->level = (int) *((f0r_param_double*)param);
        build_maps(inst);
        break;
      default:
        break; // XXX
    }
    return;
}

void f0r_get_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    whitebalance_instance_t* inst = (whitebalance_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Start */
        *((f0r_param_double*)param) = (f0r_param_double)inst->start;
        break;
      case 1: /* End */
        *((f0r_param_double*)param) = (f0r_param_double)inst->end;
        break;
      case 2: /* Level */
        *((f0r_param_double*)param) = (f0r_param_double)inst->level;
        break;
      default:
        break; // XXX
    }
    return;
}


void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    whitebalance_instance_t* inst = (whitebalance_instance_t*)instance;
    const uint8_t* in  = (const uint8_t*)inframe1;
    uint8_t*       out = (uint8_t*      )outframe;
    RETURN_IF_NULL(inst);
    if (inst->start > inst->framenum && inst->framenum <= inst->end) {
        int x, y;
        for (y = 0; y < inst->height; y++) {
            int stride = inst->width * RGBA_BYTES;
            const uint8_t* line = in + (y * stride);
            for (x = 0; x < stride; x += RGBA_BYTES) {
                /* we modify red and blue curves to enhance/reduce mostly mediums */
                out[x  ] = inst->red_filter[ line[x  ]];
                out[x+2] = inst->blue_filter[line[x+2]];
            }
        }
    }
    inst->framenum++;
    return;
}


void f0r_update(f0r_instance_t instance, double time,
                const uint32_t* inframe, uint32_t* outframe)
{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}

/*
 * vim: expandtab shiftwidth=4:
 */

