/*  xsharpen.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) 1999-2000 Donald A. Graft
 *    modified 2002 by Tilmann Bitterberg for use with transcode
 *    modified 2007 by Branko Kokanovic <branko.kokanovic at gmail dot com> to use NMS
 *  This file is part of transcode, a video stream processing tool
 *
 *  Xsharpen Filter for VirtualDub -- sharpen by mapping pixels
 *  to the closest of window max or min.
 *   
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  The author can be contacted at:
 *  Donald Graft
 *  http://sauron.mordor.net/dgraft/
 *  neuron2@home.com.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "frei0r.h"
#include "frei0r_math.h"


/* vdub compat (doubtfl) */
typedef uint32_t Pixel;
typedef uint32_t Pixel32;
typedef uint8_t  Pixel8;
typedef int      PixCoord;
typedef int      PixDim;
typedef int      PixOffset;


typedef struct xsharpeninstance_ {
    int             strength;
    int             strengthInv;
    int             threshold;
    int             srcPitch;
    int             dstPitch;
    unsigned int    width;
    unsigned int    height;
    uint8_t         *lumaMap;
    /* TODO: verify if it is better, performance-wise, to embed
       the luma value near the current pixel, as the original
       filter_xsharpen does.
     */
} xsharpen_instance_t;

/* XXX ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

enum {
    BGR_BYTES             = 3,   /* how much storage a pixel take? */
    BGRA_BYTES            = 4,   /* ditto */
    PIXEL_BLACK_THRESHOLD = 24,
    DEFAULT_STRENGTH      = 200, /* 255 is too much */
    DEFAULT_THRESHOLD     = 255
};

#define UPDATE_BGR_RANGE(p, luma) do { \
    if (luma > lumamax) { \
        lumamax = luma; \
        max = p; \
    } \
    if (luma < lumamin) { \
        lumamin = luma; \
        min = p; \
    } \
} while (0)

#define SEARCH_BGR_LUMA(PTR, X) do { \
    p = ((Pixel32 *)(PTR))[(X)-1]; \
    luma = p >> 24; \
    UPDATE_BGR_RANGE(p, luma); \
    \
    p = ((Pixel32 *)(PTR))[(X)]; \
    luma = p >> 24; \
    UPDATE_BGR_RANGE(p, luma); \
    \
    p = ((Pixel32 *)(PTR))[(X)+1]; \
    luma = p >> 24; \
    UPDATE_BGR_RANGE(p, luma); \
} while (0)


static int xsharpen_rgb_frame(xsharpen_instance_t *inst,
                              const Pixel32 *src_buf, Pixel32 *dst_buf)
{
    const PixDim    width  = inst->width;
    const PixDim    height = inst->height;
    const Pixel32   *src = NULL;
    Pixel32         *dst = NULL;
    int             x, y;
    int             r, g, b, R, G, B;
    Pixel32         p, min=1000, max=-1;
    int             luma, lumac, lumamax, lumamin, mindiff, maxdiff;
    const int       srcpitch = inst->width*sizeof(Pixel32);
    const int       dstpitch = inst->width*sizeof(Pixel32);

    /* First copy through the four border lines. */
    src = src_buf;
    dst = dst_buf;
    for (x = 0; x < width; x++){
        dst[x] = src[x];
    }
    src = (Pixel *)((char *)src_buf + (height - 1) * srcpitch);
    dst = (Pixel *)((char *)dst_buf + (height - 1) * dstpitch);
    for (x = 0; x < width; x++){
        dst[x] = src[x];
    }
    src = src_buf;
    dst = dst_buf;
    for (y = 0; y < height; y++){
        dst[0] = src[0];
        dst[width-1] = src[width-1];
        src = (Pixel *)((char *)src + srcpitch);
        dst = (Pixel *)((char *)dst + dstpitch);
    }

    /* Now calculate and store the pixel luminances for the remaining pixels. */
    src = src_buf;
    for (y = 0; y < height; y++){
        for (x = 0; x < width; x++){
            r = (src[x] >> 16) & 0xff;
            g = (src[x] >> 8) & 0xff;
            b = src[x] & 0xff;
            luma = (55 * r + 182 * g + 19 * b) >> 8;
            inst->lumaMap[y * width + x] = luma; // FIXME
        }
        src = (Pixel *)((char *)src + srcpitch);
    }

    /* Finally run the 3x3 rank-order sharpening kernel over the pixels. */
    src = (Pixel *)((char *)src_buf + srcpitch);
    dst = (Pixel *)((char *)dst_buf + dstpitch);
    for (y = 1; y < height - 1; y++){
        for (x = 1; x < width - 1; x++){
            /* Find the brightest and dimmest pixels in the 3x3 window
            surrounding the current pixel. */
            lumamax = -1;
            lumamin = 1000;

            SEARCH_BGR_LUMA(((char *)src - srcpitch), x);

            p = src[x-1];
            luma = inst->lumaMap[y * width + x-1];
            UPDATE_BGR_RANGE(p, luma);

            p = src[x];
            lumac = luma = inst->lumaMap[y * width + x];
            UPDATE_BGR_RANGE(p, luma);

            p = src[x+1];
            luma = inst->lumaMap[y * width + x+1];
            UPDATE_BGR_RANGE(p, luma);

            SEARCH_BGR_LUMA(((char *)src + srcpitch), x);

            /* Determine whether the current pixel is closer to the
               brightest or the dimmest pixel. Then compare the current
               pixel to that closest pixel. If the difference is within
               threshold, map the current pixel to the closest pixel;
               otherwise pass it through. */
            p = -1;
            if (inst->strength != 0){
                mindiff = lumac - lumamin;
                maxdiff = lumamax - lumac;
                if (mindiff > maxdiff) {
                    if (maxdiff < inst->threshold) {
                        p = max;
                    }
                } else {
                    if (mindiff < inst->threshold) {
                        p = min;
                    }
                }
            }
            if (p == -1){
                dst[x] = src[x];
            }else{
                R = (src[x] >> 16) & 0xff;
                G = (src[x] >> 8 ) & 0xff;
                B =  src[x]        & 0xff;
                r = (p >> 16) & 0xff;
                g = (p >> 8 ) & 0xff;
                b =  p        & 0xff;
                r = (inst->strength * r + inst->strengthInv * R) / 255;
                g = (inst->strength * g + inst->strengthInv * G) / 255;
                b = (inst->strength * b + inst->strengthInv * B) / 255;
                dst[x] = (r << 16) | (g << 8) | b;
            }
        }
        src = (Pixel *)((char *)src + srcpitch);
        dst = (Pixel *)((char *)dst + dstpitch);
    }

    return 0;
}

#undef UPDATE_BGR_RANGE


// TODO

int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "xsharpen";
    info->author = "Donald Graft, Tilmann Bitterberg";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_BGRA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  2;
    info->explanation = ""
"VirtualDub's XSharpen Filter."
" This filter performs a subtle but useful sharpening effect."
" The result is a sharpening effect that not only avoids amplifying"
" noise, but also tends to reduce it. A welcome side effect is that"
" files processed with this filter tend to compress to smaller files."
"";
}

void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Strength";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = ""
        "When this value is 255, mapped pixels are not blended with the"
        " original pixel values, so a full-strength effect is obtained."
        " As the value is reduced, each mapped pixel is blended with more of the"
        " original pixel. At a value of 0, the original pixels are passed"
        " through and there is no sharpening effect."
        "";
        break;
      case 1:
        info->name = "Threshold";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = ""
        "This value determines how close a pixel must be to the brightest or"
        " dimmest pixel to be mapped. If a pixel is more than threshold away"
        " from the brightest or dimmest pixel, it is not mapped.  Thus, as"
        " the threshold is reduced, pixels in the mid range start to be"
        " spared.";
        break;
      default:
        break; // XXX
    }
    return;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    xsharpen_instance_t* inst = calloc(1, sizeof(*inst));
    if (inst) {
        inst->width          = width;
        inst->height         = height;
        inst->strength       = DEFAULT_STRENGTH;
        inst->strengthInv    = 255 - inst->strength;
        inst->threshold      = DEFAULT_THRESHOLD;
        inst->srcPitch       = 0;
        inst->dstPitch       = 0;
        inst->lumaMap        = malloc(width * height);
        if (!inst->lumaMap) {
            free(inst);
            inst = NULL;
        }
    }
    return (f0r_instance_t)inst;
}

void f0r_destruct(f0r_instance_t instance)
{
    xsharpen_instance_t* inst = (xsharpen_instance_t*)instance;
    RETURN_IF_NULL(inst);
    free(inst->lumaMap);
    free(instance);
    return;
}

void f0r_set_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    xsharpen_instance_t* inst = (xsharpen_instance_t*)instance;
    int value;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Start */
        value = (int) *((f0r_param_double*)param);
        inst->strength      = CLAMP0255(value);
        inst->strengthInv   = 255 - inst->strength;
        break;
      case 1: /* Threshold */
        value = (int) *((f0r_param_double*)param);
        inst->threshold = CLAMP0255(value);
        break;
      default:
        break; // XXX
    }
    return;
}

void f0r_get_param_value(f0r_instance_t instance,
                         f0r_param_t param, int param_index)
{
    xsharpen_instance_t* inst = (xsharpen_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Strength */
        *((f0r_param_double*)param) = (f0r_param_double)inst->strength;
        break;
      case 1: /* Threshold */
        *((f0r_param_double*)param) = (f0r_param_double)inst->threshold;
        break;
      default:
        break; // XXX
    }
    return;
}


void f0r_update2(f0r_instance_t instance, double time,
                const uint32_t* inframe1,
                const uint32_t* inframe2,
                const uint32_t* inframe3,
                uint32_t* outframe)
{
    xsharpen_instance_t* inst = (xsharpen_instance_t*)instance;
    RETURN_IF_NULL(inst);
    xsharpen_rgb_frame(inst, inframe1, outframe);
    return;
}


void f0r_update(f0r_instance_t instance, double time,
                const uint32_t* inframe, uint32_t* outframe)
{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}

/*
 * vim: expandtab shiftwidth=4:
 */

#if 0
/* intentionally left in for future usage/extension */

#define UPDATE_YUV_RANGE(luma) do { \
    if (luma > lumamax) { \
        lumamax = luma; \
    } \
    if (luma < lumamin) { \
        lumamin = luma; \
    } \
} while (0)


#define SEARCH_YUV_LUMA(PTR, X) do { \
    luma = (PTR)[(X)-1] & 0xff; \
    UPDATE_YUV_RANGE(luma); \
    luma = (PTR)[(X)]   & 0xff; \
    UPDATE_YUV_RANGE(luma); \
    luma = (PTR)[(X)+1] & 0xff; \
    UPDATE_YUV_RANGE(luma); \
} while (0)


static int xsharpen_yuv_frame(xsharpen_instance_t *inst,
                              const uint8_t *src, uint8_t *dst)
{
    const PixDim       width = inst->width;
    const PixDim       height = inst->height;
    int                x, y;
    int                luma = 0, lumac = 0, lumamax, lumamin;
    int                p, mindiff, maxdiff;
    const int          srcpitch = inst->width;
    const int          dstpitch = inst->width;

    /* First copy through the four border lines. */
    ac_memcpy(dst, src, width);

    /* last */
    src = src_buf+srcpitch*(height-1);
    dst = inst->dst_buf+dstpitch*(height-1);
    ac_memcpy(dst, src, width);

    /* copy Cb and Cr */
    ac_memcpy(inst->dst_buf+dstpitch*height,
              src_buf+srcpitch*height,
              width*height>>1);

    src = src_buf;
    dst = inst->dst_buf;
    for (y = 0; y < height; y++){
        *dst = *src;
        *(dst+width-1) = *(src+width-1);
        dst += dstpitch;
        src += srcpitch;
    }

    src = src_buf + srcpitch;
    dst = inst->dst_buf + dstpitch;

    /* Finally run the 3x3 rank-order sharpening kernel over the pixels. */
    for (y = 1; y < height - 1; y++){
        for (x = 1; x < width - 1; x++){
            /* Find the brightest and dimmest pixels in the 3x3 window
               surrounding the current pixel. */
            lumamax = -1000;
            lumamin = 1000;

            SEARCH_YUV_LUMA((src - srcpitch), x);
            SEARCH_YUV_LUMA( src,             x);
            SEARCH_YUV_LUMA((src + srcpitch), x);

            /* Determine whether the current pixel is closer to the
               brightest or the dimmest pixel. Then compare the current
               pixel to that closest pixel. If the difference is within
               threshold, map the current pixel to the closest pixel;
               otherwise pass it through. */

            p = -1;
            if (inst->strength != 0){
                mindiff = lumac   - lumamin;
                maxdiff = lumamax - lumac;
                if (mindiff > maxdiff){
                    if (maxdiff < inst->threshold)
                    p = lumamax & 0xff;
                } else {
                    if (mindiff < inst->threshold)
                    p = lumamin & 0xff;
                }
            }
            if (p == -1) {
                dst[x] = src[x];
            } else {
                int t;
                lumac = src[x] & 0xff;
                t = ((inst->strength*p + inst->strengthInv*lumac)/255) & 0xff;
                t = TC_CLAMP(t, 16, 240);
                dst[x] = t & 0xff;
            }
        }
        src += srcpitch;
        dst += dstpitch;
    }

    ac_memcpy(frame->video_buf, inst->dst_buf, width*height*3/2);
    return 0;
}

#endif

