/*  deshake.c
 *  (C) 2011-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Georg Martius - November 2011
 *   georg dot martius at web dot de 
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "frei0r.h"
#include "frei0rx_paramext.h"

#include "memutils_compat.h"

#include "libdeshake.h"


#define MOD_NAME "deshake"

enum {
    PLUGIN_PARAM_NUM = 13,
    RGBA_BYTES       = 4
};


typedef struct deshakeinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
    MotionDetect md;
    TransformData td;
    SlidingAvgTrans avg;

    PixelFormat pixfmt;

    char* result;
    char* fpath;
    FILE* f;

    int (*transformWorker)(TransformData* td, Transform t);
} deshake_instance_t;


#define RETURN_IF_NULL(INST) do { \
    if (!(INST)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)


/* XXX ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

// FIXME
static int vs_framesize(int format, int width, int height)
{
    int size = -1;
    switch (format) {
      case F0R_COLOR_MODEL_RGBA8888: //fallthrough
      case F0R_COLOR_MODEL_BGRA8888:
        size = width * height * RGBA_BYTES;
        break;
    }
    return size;
}

static PixelFormat vs_format_from_f0r(int format)
{
    PixelFormat pixfmt = PF_NONE;
    switch (format) {
      case F0R_COLOR_MODEL_RGBA8888:
        pixfmt = PF_RGBA;
        break;
      case F0R_COLOR_MODEL_BGRA8888:
        pixfmt = PF_BGRA;
        break;
    }
    return pixfmt;
}

static int transformNull(TransformData* td, Transform t)
{
    return 0;
}



int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}


void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "deshake";
    info->author = "Georg Martius";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"Deshakes a video clip by extracting relative transformations"
" of subsequent frames, and transforms the high-frequency away."
" This is a single pass version, less powerful than the filters"
" stabilize and transform.\n"
" It also generates a file with relative transform information"
" to be used by the transform filter separately."
"";
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Smoothing";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The number of frames*2 + 1 used for lowpass"
                            " filtering. used for stabilizing (def: 10).";
        break;
      case 1:
        info->name = "Shakiness";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "How shaky is the video and how quick is the"
                            " camera?"
                            " 1: little (fast) 10: very strong/quick (slow) (def: 4).";
        break;
      case 2:
        info->name = "Accuracy";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Accuracy of detection process (>=shakiness)"
                            " 1: low (fast) 15: high (slow) (def: 4).";
        break;
      case 3:
        info->name = "Stepsize";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Stepsize of search process, region around minimum"
                            " is scanned with 1 pixel resolution (def: 6).";
        break;
      case 4:
        info->name = "Algo";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "0: brute force (translation only);"
                            " 1: small measurement fields (def).";
        break;
      case 5:
        info->name = "MinContrast";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "below this contrast a field is discarded (0-1)"
                            " (def: 0.3).";
        break;
      case 6:
        info->name = "Result";
        info->type = F0R_PARAM_STRING;
        info->explanation = "path to the file used to write the transforms"
                            " (def:deshakedata.stab).";
        break;
      case 7:
        info->name = "MaxShift";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "maximal number of pixels to translate image"
                            " (def: -1 no limit).";
        break;
      case 8:
        info->name = "MaxAngle";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "maximal angle in rad to rotate image"
                            " (def: -1 no limit).";
        break;
      case 9:
        info->name = "Crop";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "0: keep border (def), 1: black background.";
        break;
      case 10:
        info->name = "Zoom";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "percentage to zoom >0: zoom in, <0 zoom out"
                            " (def: 2).";
        break;
      case 11:
        info->name = "OptZoom";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "0: nothing, 1: determine optimal zoom (def).";
        break;
      case 12:
        info->name = "InterPol";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Type of interpolation:"
                            " 0: no interpolation,"
                            " 1: linear (horizontal),"
                            " 2: bi-linear (def),"
                            " 3: bi-cubic.";
        break;
      default:
        break; // XXX
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    deshake_instance_t* inst = (deshake_instance_t*)instance;
    return &inst->attrs[param_index];
}

static int deshake_disable(deshake_instance_t *inst)
{
    inst->transformWorker = transformNull;
    return 0;
}

static int deshake_enable(deshake_instance_t *inst)
{
    if (inst->pixfmt == PF_RGBA) {
        inst->transformWorker = transformRGB;
    } else if (inst->pixfmt == PF_YUV) {
        inst->transformWorker = transformYUV;
    } else {
        fprintf(stderr, "[%s] invalid Pixel Format 0x%X\n",
                MOD_NAME, inst->pixfmt);
        return -1;
    }
    return 0;
}



static int vs_deshake_init(deshake_instance_t *sd,
                           int format,
                           int width, int height,
                           int verbose)
{
    int ret = DS_OK;
    MotionDetect* md = &(sd->md);
    TransformData* td = &(sd->td);
    DSFrameInfo fi_src;
    DSFrameInfo fi_dst;
    
    sd->pixfmt = vs_format_from_f0r(format);

    if (!deshake_enable(sd)) {
        // already logged
        return -1;
    }
  
    md->show     = 0;
    td->relative = 1;
    td->invert   = 0;
    td->verbose  = verbose;

    sd->avg.initialized = 0;

    // init MotionDetect part
    fi_src.width     = width;
    fi_src.height    = height;
    fi_src.strive    = width;
    fi_src.framesize = vs_framesize(format, width, height);
    fi_src.pFormat   = sd->pixfmt;
    ret = initMotionDetect(md, &fi_src, MOD_NAME);
    if (ret != DS_OK) {
        fprintf(stderr, "[%s] initialization of Motion Detection failed\n",
                MOD_NAME);
        return -1;
    }
    
    // init trasform part
    /* Todo: in case we can scale the images, calc new size later */
    fi_dst.width     = width;
    fi_dst.height    = height;
    fi_dst.strive    =	width;
    fi_dst.framesize = vs_framesize(format, width, height);

    ret = initTransformData(td, &fi_src, &fi_dst, MOD_NAME);
    if (ret != DS_OK) {
        fprintf(stderr, "[%s] initialization of Transform failed\n",
                MOD_NAME);
        return -1;
    }

    return 0;
}


static void reconfigure_vidstab(f0r_instance_t instance, int param_index)
{
    int ret = DS_OK;
    deshake_instance_t* inst = (deshake_instance_t*)instance;
    MotionDetect* md = NULL;
    TransformData* td = NULL;

    RETURN_IF_NULL(inst);

    md = &(inst->md);
    td = &(inst->td);

    if (param_index >= 1 && param_index <= 5) {
        ret = configureMotionDetect(md);
        if (ret != DS_OK) {
            fprintf(stderr, "[%s] configuration of Motion Detection failed\n",
                    MOD_NAME);
            deshake_disable(inst);
            return;
        }
    } else if ((param_index >= 7 && param_index <= 12) || param_index == 0) {
        ret = configureTransformData(td);
        if (ret != DS_OK) {
            fprintf(stderr, "[%s] configuration of Transform failed\n",
                    MOD_NAME);
            deshake_disable(inst);
            return;
        }
    } else if (param_index == 6) {
        inst->f = fopen(inst->result, "w");
        if (inst->f == NULL) {
            fprintf(stderr, "[%s] cannot open result file [%s] (skipped)\n",
                    MOD_NAME, inst->result);
            // not fatal!
            return;
        }    
    }
    deshake_enable(inst);
    return;
}


f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    deshake_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        int err = 0;
        int verbose = 2;
        int format = F0R_COLOR_MODEL_RGBA8888;
        err = vs_deshake_init(inst, format, width, height, verbose);
        if (!err) {
            MotionDetect* md = &(inst->md);
            TransformData* td = &(inst->td);

            f0rx_bind_param_attr_remap_int(inst, 1,  &md->shakiness);
            f0rx_bind_param_attr_remap_int(inst, 2,  &md->accuracy);
            f0rx_bind_param_attr_remap_int(inst, 3,  &md->stepSize);
            f0rx_bind_param_attr_remap_int(inst, 4,  &md->algo);
            f0rx_bind_param_attr(          inst, 5,  &md->contrastThreshold);

            f0rx_bind_param_attr_remap_int(inst, 0,  &td->smoothing);
            f0rx_bind_param_attr_remap_int(inst, 7,  &td->maxShift);
            f0rx_bind_param_attr          (inst, 8,  &td->maxAngle);
            f0rx_bind_param_attr_remap_int(inst, 9,  &td->crop);
            f0rx_bind_param_attr          (inst, 10, &td->zoom);
            f0rx_bind_param_attr_remap_int(inst, 11, &td->optZoom);
            f0rx_bind_param_attr_remap_int(inst, 12, &td->interpolType);

            f0rx_bind_param_attr(          inst, 6,  &inst->result);
            
            f0rx_hook_param_attr_post_set(inst, F0RX_PARAM_ALL,
                                          reconfigure_vidstab);
        } else {
            tc_free(inst);
            inst = NULL;
        }
    }
    return (f0r_instance_t)inst;
}


void f0r_destruct(f0r_instance_t instance)
{
    deshake_instance_t* inst = (deshake_instance_t*)instance;
    RETURN_IF_NULL(inst);
    if (inst->f) {
        fclose(inst->f);
        inst->f = NULL;
    }

    cleanupMotionDetection(&inst->md);
    if (inst->result) {
        tc_free(inst->result);
        inst->result = NULL;
    }

    cleanupTransformData(&inst->td);
 
    tc_free(instance);
    return;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    int ret = DS_OK;
    unsigned char *inframe = (unsigned char *)inframe1; // FIXME
    unsigned char *exframe = (unsigned char *)outframe;
    deshake_instance_t* inst = (deshake_instance_t*)instance;
    MotionDetect* md = NULL;
    TransformData* td = NULL;
    Transform motion;
    Transform worker;

    RETURN_IF_NULL(inst);

    md = &(inst->md);
    td = &(inst->td);
    
    ret = motionDetection(md, &motion, inframe);
    if (ret == DS_OK) {
        if (inst->f) {
            fprintf(inst->f, "%i %6.4lf %6.4lf %8.5lf %6.4lf %i\n",
                    md->frameNum, motion.x, motion.y, motion.alpha, motion.zoom, 
    	            motion.extra);
        }
    } else {
        fprintf(stderr, "[%s] Motion Detection failed\n", MOD_NAME);
        return;
    }
    
    transformPrepare(td, inframe, exframe);
    worker = lowPassTransforms(td, &inst->avg, &motion);
    inst->transformWorker(td, worker);
    transformFinish(td);
    return;
}

void f0r_update(f0r_instance_t instance, double time,
                const uint32_t* inframe,
                uint32_t* outframe)
{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}


/*
 * vim: expandtab shiftwidth=4:
 */

