/*  stabilize.c
 *  (C) 2011-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Georg Martius - November 2011
 *   georg dot martius at web dot de 
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "frei0r.h"
#include "frei0rx_paramext.h"

#include "memutils_compat.h"

#include "libdeshake.h"


#define MOD_NAME "stabilize"

enum {
    PLUGIN_PARAM_NUM = 7,
    RGBA_BYTES       = 4
};


typedef struct stabilizeinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
    MotionDetect md;

    PixelFormat pixfmt;

    char* result;
    char* fpath;
    FILE* f;
} stabilize_instance_t;


#define RETURN_IF_NULL(INST) do { \
    if (!(INST)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)


/* XXX ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

// FIXME
static int vs_framesize(int format, int width, int height)
{
    int size = -1;
    switch (format) {
      case F0R_COLOR_MODEL_RGBA8888: //fallthrough
      case F0R_COLOR_MODEL_BGRA8888:
        size = width * height * RGBA_BYTES;
        break;
    }
    return size;
}

static PixelFormat vs_format_from_f0r(int format)
{
    PixelFormat pixfmt = PF_NONE;
    switch (format) {
      case F0R_COLOR_MODEL_RGBA8888:
        pixfmt = PF_RGBA;
        break;
      case F0R_COLOR_MODEL_BGRA8888:
        pixfmt = PF_BGRA;
        break;
    }
    return pixfmt;
}



int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}


void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "stabilize";
    info->author = "Georg Martius";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"Generates a file with relative transform information."
" (translation, rotation) about subsequent frames."
"";
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Result";
        info->type = F0R_PARAM_STRING;
        info->explanation = "path to the file used to write the transforms"
                            " (def:deshakedata.stab).";
        break;
      case 1:
        info->name = "Shakiness";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "How shaky is the video and how quick is the"
                            " camera?"
                            " 1: little (fast) 10: very strong/quick (slow) (def: 4).";
        break;
      case 2:
        info->name = "Accuracy";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Accuracy of detection process (>=shakiness)"
                            " 1: low (fast) 15: high (slow) (def: 4).";
        break;
      case 3:
        info->name = "Stepsize";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Stepsize of search process, region around minimum"
                            " is scanned with 1 pixel resolution (def: 6).";
        break;
      case 4:
        info->name = "Algo";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "0: brute force (translation only);"
                            " 1: small measurement fields (def).";
        break;
      case 5:
        info->name = "MinContrast";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "below this contrast a field is discarded (0-1)"
                            " (def: 0.3).";
        break;
      case 6:
        info->name = "Show";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "0: draw nothing (def); 1,2: show fields and transforms"
                            " in the resulting frames.";
        break;
      default:
        break; // XXX
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    stabilize_instance_t* inst = (stabilize_instance_t*)instance;
    return &inst->attrs[param_index];
}

static int vs_stabilize_write_header(stabilize_instance_t *sd)
{
    MotionDetect* md = &(sd->md);
    // write parameters as comments to file
    fprintf(sd->f, "#      accuracy = %d\n", md->accuracy);
    fprintf(sd->f, "#     shakiness = %d\n", md->shakiness);
    fprintf(sd->f, "#      stepsize = %d\n", md->stepSize);
    fprintf(sd->f, "#          algo = %d\n", md->algo);
    fprintf(sd->f, "#   mincontrast = %f\n", md->contrastThreshold);
    fprintf(sd->f, "#        result = %s\n", sd->result);
    // write header line
    fprintf(sd->f, "# Transforms\n#C FrameNr x y alpha zoom extra\n");
    return 0;
}

static int vs_stabilize_init(stabilize_instance_t *sd,
                               int format,
                               int width, int height,
                               int verbose)
{
    int ret = DS_OK;
    MotionDetect* md = &(sd->md);
    DSFrameInfo fi;
    
    sd->pixfmt = vs_format_from_f0r(format);

    if (sd->pixfmt == PF_NONE) {
        // TODO: log it out
        return -1;
    }
  
    // init MotionDetect part
    fi.width     = width;
    fi.height    = height;
    fi.strive    = width;
    fi.framesize = vs_framesize(format, width, height);
    fi.pFormat   = sd->pixfmt;
    ret = initMotionDetect(md, &fi, MOD_NAME);
    if (ret != DS_OK) {
        fprintf(stderr, "[%s] initialization of Motion Detection failed\n",
                MOD_NAME);
        return -1;
    }
    
    return 0;
}


static void reconfigure_vidstab(f0r_instance_t instance, int param_index)
{
    int ret = DS_OK;
    MotionDetect* md = NULL;
    stabilize_instance_t* inst = (stabilize_instance_t*)instance;

    RETURN_IF_NULL(inst);

    md = &(inst->md);

    if (param_index >= 1 && param_index <= 6) {
        ret = configureMotionDetect(md);
        if (ret != DS_OK) {
            fprintf(stderr, "[%s] configuration of Motion Detection failed\n",
                    MOD_NAME);
            // FIXME
            return;
        }
    } else if (param_index == 0) {
        inst->f = fopen(inst->result, "w");
        if (inst->f == NULL) {
            fprintf(stderr, "[%s] cannot open result file [%s] (skipped)\n",
                    MOD_NAME, inst->result);
            // not fatal! really? FIXME
            return;
        }
        vs_stabilize_write_header(inst);
    }
    return;
}


f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    stabilize_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        int err = 0;
        int verbose = 2;
        int format = F0R_COLOR_MODEL_RGBA8888;
        err = vs_stabilize_init(inst, format, width, height, verbose);
        if (!err) {
            MotionDetect* md = &(inst->md);

            f0rx_bind_param_attr(          inst, 0,  &inst->result);

            f0rx_bind_param_attr_remap_int(inst, 1,  &md->shakiness);
            f0rx_bind_param_attr_remap_int(inst, 2,  &md->accuracy);
            f0rx_bind_param_attr_remap_int(inst, 3,  &md->stepSize);
            f0rx_bind_param_attr_remap_int(inst, 4,  &md->algo);
            f0rx_bind_param_attr(          inst, 5,  &md->contrastThreshold);
            f0rx_bind_param_attr_remap_int(inst, 6,  &md->show);
            
            f0rx_hook_param_attr_post_set(inst, F0RX_PARAM_ALL,
                                          reconfigure_vidstab);
        } else {
            tc_free(inst);
            inst = NULL;
        }
    }
    return (f0r_instance_t)inst;
}


void f0r_destruct(f0r_instance_t instance)
{
    stabilize_instance_t* inst = (stabilize_instance_t*)instance;
    RETURN_IF_NULL(inst);
    if (inst->f) {
        fclose(inst->f);
        inst->f = NULL;
    }

    cleanupMotionDetection(&inst->md);
    if (inst->result) {
        tc_free(inst->result);
        inst->result = NULL;
    }

    tc_free(instance);
    return;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    int ret = DS_OK;
    unsigned char *inframe = (unsigned char *)inframe1; // FIXME
    Transform motion;
    MotionDetect* md = NULL;
    stabilize_instance_t* inst = (stabilize_instance_t*)instance;

    RETURN_IF_NULL(inst);

    md = &(inst->md);

    ac_memcpy(outframe, inframe, md->fi.framesize);

    ret = motionDetection(md, &motion, inframe);
    if (ret == DS_OK) {
        if (inst->f) {
            fprintf(inst->f, "%i %6.4lf %6.4lf %8.5lf %6.4lf %i\n",
                    md->frameNum, motion.x, motion.y, motion.alpha, motion.zoom, 
    	            motion.extra);
        }
    } else {
        fprintf(stderr, "[%s] Motion Detection failed\n", MOD_NAME);
        return;
    }
    
    return;
}


void f0r_update(f0r_instance_t instance, double time,
                const uint32_t* inframe,
                uint32_t* outframe)

{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}



/*
 * vim: expandtab shiftwidth=4:
 */

