/*  transform.c
 *  (C) 2011-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Georg Martius - November 2011
 *   georg dot martius at web dot de 
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "frei0r.h"
#include "frei0rx_paramext.h"

#include "memutils_compat.h"

#include "libdeshake.h"


#define MOD_NAME "deshake"

enum {
    PLUGIN_PARAM_NUM = 10,
    BUF_LINE         = 256,
    RGBA_BYTES       = 4
};


// TODO: sharpen using boxblur

typedef struct transforminstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
    TransformData td;

    PixelFormat pixfmt;
    Transformations trans; // transformations

    char *input;

    int (*transformWorker)(TransformData* td, Transform t);
} transform_instance_t;


#define RETURN_IF_NULL(INST) do { \
    if (!(INST)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)


/* XXX ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

// FIXME
static int vs_framesize(int format, int width, int height)
{
    int size = -1;
    switch (format) {
      case F0R_COLOR_MODEL_RGBA8888: //fallthrough
      case F0R_COLOR_MODEL_BGRA8888:
        size = width * height * RGBA_BYTES;
        break;
    }
    return size;
}

static PixelFormat vs_format_from_f0r(int format)
{
    PixelFormat pixfmt = PF_NONE;
    switch (format) {
      case F0R_COLOR_MODEL_RGBA8888:
        pixfmt = PF_RGBA;
        break;
      case F0R_COLOR_MODEL_BGRA8888:
        pixfmt = PF_BGRA;
        break;
    }
    return pixfmt;
}

static int transformNull(TransformData* td, Transform t)
{
    return 0;
}



int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}


void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "transform";
    info->author = "Georg Martius";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"Ttransforms each frame according to transformations"
" given in an input file (e.g. translation, rotate)"
" see also the stabilize filter."
"";
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Input";
        info->type = F0R_PARAM_STRING;
        info->explanation = "path to the file used to read the transforms"
                            " (def:deshakedata.stab).";
        break;
      case 1:
        info->name = "MaxShift";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "maximal number of pixels to translate image"
                            " (def: -1 no limit).";
        break;
      case 2:
        info->name = "MaxAngle";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "maximal angle in rad to rotate image"
                            " (def: -1 no limit).";
        break;
      case 3:
        info->name = "Smoothing";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The number of frames*2 + 1 used for lowpass"
                            " filtering. used for stabilizing (def: 10).";
        break;
      case 4:
        info->name = "Crop";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "0: keep border (def), 1: black background.";
        break;
      case 5:
        info->name = "Invert";
        info->type = F0R_PARAM_BOOL;
        info->explanation = "WRITEME"; // FIXME
        break;
      case 6:
        info->name = "Relative";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "WRITEME"; // FIXME
        break;
      case 7:
        info->name = "Zoom";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "percentage to zoom >0: zoom in, <0 zoom out"
                            " (def: 2).";
        break;
      case 8:
        info->name = "OptZoom";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "0: nothing, 1: determine optimal zoom (def).";
        break;
      case 9:
        info->name = "InterPol";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Type of interpolation:"
                            " 0: no interpolation,"
                            " 1: linear (horizontal),"
                            " 2: bi-linear (def),"
                            " 3: bi-cubic.";
        break;
      default:
        break; // XXX
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    transform_instance_t* inst = (transform_instance_t*)instance;
    return &inst->attrs[param_index];
}

static int transform_disable(transform_instance_t *inst)
{
    inst->transformWorker = transformNull;
    return 0;
}

static int transform_enable(transform_instance_t *inst)
{
    if (inst->pixfmt == PF_RGBA) {
        inst->transformWorker = transformRGB;
    } else if (inst->pixfmt == PF_YUV) {
        inst->transformWorker = transformYUV;
    } else {
        fprintf(stderr, "[%s] invalid Pixel Format 0x%X\n",
                MOD_NAME, inst->pixfmt);
        return -1;
    }
    return 0;
}



static int vs_transform_init(transform_instance_t *sd,
                             int format,
                             int width, int height,
                             int verbose)
{
    int ret = DS_OK;
    TransformData* td = &(sd->td);
    DSFrameInfo fi_src;
    DSFrameInfo fi_dst;
    
    sd->pixfmt = vs_format_from_f0r(format);

    if (!transform_enable(sd)) {
        // already logged
        return -1;
    }
  
    td->verbose  = verbose;

    fi_src.width     = width;
    fi_src.height    = height;
    fi_src.strive    = width;
    fi_src.pFormat   = sd->pixfmt;
    fi_src.framesize = vs_framesize(format, width, height);
    
    /* Todo: in case we can scale the images, calc new size later */
    fi_dst.width     = width;
    fi_dst.height    = height;
    fi_dst.strive    = width;
    fi_dst.framesize = vs_framesize(format, width, height);

    initTransformations(&sd->trans);
    
    ret = initTransformData(td, &fi_src, &fi_dst, MOD_NAME);
    if (ret != DS_OK) {
        fprintf(stderr, "[%s] initialization of Transform failed\n",
                MOD_NAME);
        return -1;
    }

    return 0;
}


static void reconfigure_vidstab(f0r_instance_t instance, int param_index)
{
    int ret = DS_OK;
    TransformData* td = NULL;
    transform_instance_t* inst = (transform_instance_t*)instance;

    RETURN_IF_NULL(inst);

    td = &(inst->td);

    if (param_index >= 1 && param_index <= 10) {
        ret = configureTransformData(td);
        if (ret != DS_OK) {
            fprintf(stderr, "[%s] configuration of Transform failed\n",
                    MOD_NAME);
            transform_disable(inst);
            return;
        }
    } else if (param_index == 0) {
        FILE *f = fopen(inst->input, "r");
        if (f == NULL) {
            fprintf(stderr, "[%s] cannot open input file [%s] (skipped)\n",
                    MOD_NAME, inst->input);
            return;
        }
        
        cleanupTransformations(&inst->trans);
        if (!readTransforms(td, f, &inst->trans)) { /* read input file */
            fprintf(stderr, "[%s] error parsing input file %s!\n",
                    MOD_NAME, inst->input);
            transform_disable(inst);
        } else if (preprocessTransforms(td, &inst->trans)!= DS_OK ) {
            fprintf(stderr, "[%s] error while preprocessing transforms!\n",
                    MOD_NAME);
            transform_disable(inst);
        }

        fclose(f);

    }
    transform_enable(inst);
    return;
}


f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    transform_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        int err = 0;
        int verbose = 2;
        int format = F0R_COLOR_MODEL_RGBA8888;
        err = vs_transform_init(inst, format, width, height, verbose);
        if (!err) {
            TransformData* td = &(inst->td);

            f0rx_bind_param_attr(          inst, 0, &inst->input);

            f0rx_bind_param_attr_remap_int(inst, 1, &td->maxShift);
            f0rx_bind_param_attr          (inst, 2, &td->maxAngle);
            f0rx_bind_param_attr_remap_int(inst, 3, &td->smoothing);
            f0rx_bind_param_attr_remap_int(inst, 4, &td->crop);
            f0rx_bind_param_attr_remap_int(inst, 5, &td->invert);
            f0rx_bind_param_attr_remap_int(inst, 6, &td->relative);
            f0rx_bind_param_attr          (inst, 7, &td->zoom);
            f0rx_bind_param_attr_remap_int(inst, 8, &td->optZoom);
            f0rx_bind_param_attr_remap_int(inst, 9, &td->interpolType);

            f0rx_hook_param_attr_post_set(inst, F0RX_PARAM_ALL,
                                          reconfigure_vidstab);
        } else {
            tc_free(inst);
            inst = NULL;
        }
    }
    return (f0r_instance_t)inst;
}


void f0r_destruct(f0r_instance_t instance)
{
    transform_instance_t* inst = (transform_instance_t*)instance;

    RETURN_IF_NULL(inst);

    cleanupTransformations(&inst->trans);
    cleanupTransformData(&inst->td);
 
    tc_free(instance);
    return;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    unsigned char *inframe = (unsigned char *)inframe1; // FIXME
    unsigned char *exframe = (unsigned char *)outframe;
    transform_instance_t* inst = (transform_instance_t*)instance;
    Transform worker;

    RETURN_IF_NULL(inst);
    
    transformPrepare(&inst->td, inframe, exframe);
    worker = getNextTransform(&inst->td, &inst->trans);
    inst->transformWorker(&inst->td, worker);
    transformFinish(&inst->td);
    return;
}

void f0r_update(f0r_instance_t instance, double time,
                const uint32_t* inframe,
                uint32_t* outframe)
{
    f0r_update2(instance, time, inframe, NULL, NULL, outframe);
    return;
}


/*
 * vim: expandtab shiftwidth=4:
 */

