/*  acopychannel.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) William H Wittig - May 2003
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "memutils.h"

#include "frei0rx.h"
#include "frei0rx_paramext.h"
#include "frei0rx_compat.h"

enum {
    PLUGIN_PARAM_NUM = 2
};

typedef struct acopychannelinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];

    int src;
    int dst;

    unsigned int sample_rate;
    unsigned int channel_count;
} acopychannel_instance_t;

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)

/* ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

enum {
    S16LE_MAX         = 0x7FFF,
    S16LE_BYTES       = 2,
    SILENCE_MAX_VALUE = 0 /* XXX */
};


int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "acopychannel";
    info->author = "William H Wittig";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = 0;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
" Copy one audio channel to another channel;"
" Channel order: 0 left, 1 right" /* FIXME */
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_AUDIO;
    info->sample_model = F0RX_SAMPLE_MODEL_S16LE;
    info->channel_layout = 0; /* TODO */
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Source";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Source channel for the copy.";
        break;
      case 1:
        info->name = "Destination";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Destionation channel for the copy.";
        break;
      default:
        break; /* XXX */
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    acopychannel_instance_t* inst = (acopychannel_instance_t*)instance;
    return &inst->attrs[param_index];
}

f0r_instance_t f0rx_construct(int format, f0r_param_t *params,
                              unsigned int value1, unsigned int value2)
{
    acopychannel_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        inst->src           = 0;
        inst->dst           = 1;
        inst->sample_rate   = value1;
        inst->channel_count = value2;
        f0rx_bind_param_attr_remap_int(inst, 0, &inst->src);
        f0rx_bind_param_attr_remap_int(inst, 1, &inst->dst);
    }
    return (f0r_instance_t)inst;
}

int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    int ready = 0;
    acopychannel_instance_t* inst = (acopychannel_instance_t*)instance;
    if (inst) {
        /* only stereo is supported yet */
        ready = (inst->src >= 0 && inst->dst >= 0
              && inst->src != inst->dst
              && inst->src <= 1 && inst->dst <= 1);
    }
    return ready;
}

int f0rx_set_params_values(f0r_instance_t instance,
                           f0r_param_t *params)
{
    return 0;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0RX_SAMPLE_MODEL_S16LE, NULL, width, height);
}

void f0r_destruct(f0r_instance_t instance)
{
    acopychannel_instance_t* inst = (acopychannel_instance_t*)instance;
    RETURN_IF_NULL(inst);
    /* nothing to do here */
    free(instance);
    return;
}

static int copy_channel(int16_t *out, const int16_t *in, int len, int src, int dst)
{
    int i;
    /* Implicitly assumes even number of samples (e.g. l,r pairs) */
    for (i = 0; i < len; i += 2) {
        if (src == 0) {
            out[i  ] = in[i];
            out[i+1] = in[i];
        } else {
            out[i]   = in[i+1];
            out[i+1] = in[i+1];
        }
    }
    return 0;
}

int f0rx_update(f0r_instance_t instance,
                double intime,
                const uint32_t *inframe1,
                const uint32_t *inframe2,
                const uint32_t *inframe3,
                double *outtime,
                uint32_t *outframe)
{
    acopychannel_instance_t* inst = (acopychannel_instance_t*)instance;
    RETURN_ERROR_IF_NULL(inst);
    if (f0rx_is_ready(instance, F0RX_PARAM_ALL)) {
        int channel_size = S16LE_BYTES * inst->sample_rate;
        copy_channel((int16_t *)outframe, (const int16_t *)inframe1,
                     channel_size, inst->src, inst->dst);
    }
    return 0;
}

/*
 * vim: expandtab shiftwidth=4:
 */

