/*  adetectsilence.c
 *  (C) 2011-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Tilmann Bitterberg - July 2003
 *
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "memutils.h"

#include "frei0rx.h"
#include "frei0rx_compat.h"
#include "frei0rx_paramext.h"


#define MOD_NAME "adetectsilence"

enum {
    SILENCE_FRAMES = 4,
    MAX_TRACKS     = 128
};

struct segment {
    int begin;
    int end;
};

typedef struct adetectsilenceinstance_ {
    f0rx_param_attr_t attrs[2]; /* XXX */
    int channel_size; /* samples */

    int frame_num;
    int zeros;
    int next;
    struct segment tracks[MAX_TRACKS];

    int silence_frames;
    char* log_path;

    unsigned int sample_rate;
    unsigned int channel_count;
} adetectsilence_instance_t;

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)

/* ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

enum {
    S16LE_MAX         = 0x7FFF,
    S16LE_BYTES       = 2,
    SILENCE_MAX_VALUE = 0 /* XXX */
};

enum {
    MACHINE_FRIENDLY = 0,
    HUMAN_FRIENDLY   = 1
};

static int report_stats(adetectsilence_instance_t* inst, FILE* out, int verbose)
{
    int i = 0;
    for (i = 0; i < inst->next; i++) {
        if (verbose) {
            fprintf(out, "%02i silence in frames [%i, %i]\n",
                    i,
                    inst->tracks[i].begin,
                    inst->tracks[i].end);
        } else {
            fprintf(out, "%i %i\n",
                    inst->tracks[i].begin,
                    inst->tracks[i].end);
        }
    }
    return 0;
}

static int adetectsilence_frame(adetectsilence_instance_t *inst,
                                const int16_t *samples)
{
    double p = 0.0;
    int i, sum;

    for (i = 0; i < inst->channel_size; i++) {
        p += fabs((double)(*samples++)/(S16LE_MAX * 1.0)); 
    }
    sum = (int)p;
    
    /* Is this frame silence? */
    if (sum == 0){
        inst->zeros++;
    }
    /*
     * former silence, now no more silence
     */
    if (inst->zeros >= inst->silence_frames && sum > 0) {
        inst->tracks[inst->next].begin = inst->frame_num - inst->zeros;
        inst->tracks[inst->next].end   = inst->frame_num - 1;
        inst->next++;
        if (inst->next > MAX_TRACKS) {
            /* if you see this, start to worry. */
            fprintf(stderr, "[%s] Cannot save more tracks\n",
                    MOD_NAME);
        }
        inst->zeros = 0;
    }
    return 0; /* Not rocket science, but it's not easy as it seems. */
}


static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    adetectsilence_instance_t* inst = (adetectsilence_instance_t*)instance;
    return &inst->attrs[param_index];
}

int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}


void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "adetectsilence";
    info->author = "Tilmann Bitterberg";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = 0;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  2;
    info->explanation = ""
"This filter detect silence intervals in audio track. It can just"
" print out to screen the position and duration of audio silence"
" intervals, or, assuming the audio track is a soundtrack or something"
" like that, it can generate a tcmp3cut command line to cut the track"
" in tracks."
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_AUDIO;
    info->sample_model = F0RX_SAMPLE_MODEL_S16LE;
    info->channel_layout = 0; /* TODO */
}



void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Silence Frames";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = ""
"The threshold used internally by filter to decide if"
" silence interval is a song transition or not."
" The higher is this value, the longer should silence"
" interval be."
"";
        break;
      case 1:
        info->name = "Log Path";
        info->type = F0R_PARAM_STRING;
        info->explanation = "File to save a detailed statistics.";
        break;
      default:
        break; /* XXX */
    }
    return;
}

int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    int ready = 1;
    /* FIXME: implement the silence_frames sanity check. */
    return ready;
}

int f0rx_set_params_values(f0r_instance_t instance,
                           f0r_param_t *params)
{
    return 0;
}

f0r_instance_t f0rx_construct(int format, f0r_param_t *params,
                              unsigned int value1, unsigned int value2)
{
    adetectsilence_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        int i = 0;
        inst->log_path      = NULL;
        inst->sample_rate   = value1;
        inst->channel_count = value2;
        inst->silence_frames= SILENCE_FRAMES;
        inst->channel_size  = (inst->sample_rate * S16LE_BYTES);
        for (i = 0; i < MAX_TRACKS; i++) {
            inst->tracks[i].begin = -1;
            inst->tracks[i].end = -1;
        }
        f0rx_bind_param_attr_remap_int(inst, 0, &inst->silence_frames);
        f0rx_bind_param_attr(inst, 1, &inst->log_path);
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0RX_SAMPLE_MODEL_S16LE, NULL, width, height);
}


void f0r_destruct(f0r_instance_t instance)
{
    adetectsilence_instance_t* inst = (adetectsilence_instance_t*)instance;
    RETURN_IF_NULL(inst);
    if (inst->frame_num > 0) {
        report_stats(inst, stderr, HUMAN_FRIENDLY);
        if (inst->log_path) {
            FILE* out = fopen(inst->log_path, "wt+");
            if (out) {
                report_stats(inst, out, MACHINE_FRIENDLY);
                fclose(out); /* XXX what about failures? */
            } /* XXX else? */
            free(inst->log_path);
        }
    }
    free(instance);
    return;
}

int f0rx_update(f0r_instance_t instance,
                double intime,
                const uint32_t *inframe1,
                const uint32_t *inframe2,
                const uint32_t *inframe3,
                double *outtime,
                uint32_t *outframe)
{
    adetectsilence_instance_t* inst = (adetectsilence_instance_t*)instance;
    int channel_size = 0;
    RETURN_ERROR_IF_NULL(inst);
    channel_size = S16LE_BYTES * inst->sample_rate;
    /* first, let's provide a valid output */
    memcpy(outframe, inframe1,
           channel_size * inst->channel_count);
    /* then the actual processing */
    /* FIXME: this is the original algo. I'm not sold, though. */
    adetectsilence_frame(inst, (const int16_t *)inframe1);
    inst->frame_num++;
    return 0;
}

/*
 * vim: expandtab shiftwidth=4:
 */

