/*  lowpass.c
 *  (C) 2012-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Based on `filt'-code by Markus Wandel
 *    http://wandel.ca/homepage/audiohacks.html
 *  Copyright (C) Tilmann Bitterberg
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include "memutils.h"

#include "frei0rx.h"
#include "frei0rx_compat.h"
#include "frei0rx_paramext.h"


enum {
    DEFAULT_TAPS        = 30,
    PLUGIN_PARAM_NUM    = 1
};

typedef struct lowpassinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
    unsigned int sample_rate;
    unsigned int channel_count;
    int16_t *array_l, *array_r;
    int highpass;               /* sign flag */
    int taps;
    int p;                      /* pointer */
    int ready;                  /* flag */
} lowpass_instance_t;

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)

/*************************************************************************/

static int alloc_buffers(lowpass_instance_t *inst)
{
    int err = 0;
    inst->array_r = tc_zalloc(inst->taps * sizeof(int16_t));
    inst->array_l = tc_zalloc(inst->taps * sizeof(int16_t));

    if (inst->array_r && inst->array_l) {
        inst->ready = 1;
    } else {
        inst->ready = 0;
        err = -1;
    }
    return err;
}

static void free_buffers(lowpass_instance_t *inst)
{
    if (inst->array_r) {
        tc_free(inst->array_r);
        inst->array_r = NULL; 
    }
    if (inst->array_l) {
        tc_free(inst->array_l);
        inst->array_l = NULL;
    }
    return;
}

static void post_set_param(f0r_instance_t instance, int param_index)
{
    lowpass_instance_t* inst = (lowpass_instance_t*)instance;
    RETURN_IF_NULL(inst);
    /* no switch: we got just one parameter. */
    if (inst->taps < 0) {
        inst->taps = -inst->taps;
        inst->highpass = -1;
    } else {
        inst->highpass = +1;
    }
    free_buffers(inst);
    alloc_buffers(inst);
    return;
}


/*************************************************************************/

int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}


void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "lowpass";
    info->author = "Tilmann Bitterberg";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = 0;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"High and low pass filter."
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_AUDIO;
    info->sample_model = F0RX_SAMPLE_MODEL_S16LE;
    info->channel_layout = 0; /* TODO */
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Taps";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Filter strength (may be negative).";
        break;
      default:
        break; /* XXX */
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    lowpass_instance_t* inst = (lowpass_instance_t*)instance;
    return &inst->attrs[param_index];
}

int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    int ready = 0;
    lowpass_instance_t* inst = (lowpass_instance_t*)instance;
    if (inst) {
        ready = inst->ready;
    }
    return ready;
}

int f0rx_set_params_values(f0r_instance_t instance,
                           f0r_param_t *params)
{
    return 0;
}

f0r_instance_t f0rx_construct(int format, f0r_param_t *params,
                              unsigned int value1, unsigned int value2)
{
    lowpass_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        inst->sample_rate   = value1;
        inst->channel_count = value2;
        inst->taps          = DEFAULT_TAPS;
        inst->highpass      = +1;
        inst->p             = 0;

        alloc_buffers(inst);

        f0rx_bind_param_attr_remap_int(inst, 0, &inst->taps);
        f0rx_hook_param_attr_post_set(inst, 0, post_set_param);
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0RX_SAMPLE_MODEL_S16LE, NULL, width, height);
}


void f0r_destruct(f0r_instance_t instance)
{
    lowpass_instance_t* inst = (lowpass_instance_t*)instance;
    RETURN_IF_NULL(inst);
    free_buffers(inst);
    tc_free(instance);
    return;
}

int f0rx_update(f0r_instance_t instance,
                double intime,
                const uint32_t *inframe1,
                const uint32_t *inframe2,
                const uint32_t *inframe3,
                double *outtime,
                uint32_t *outframe)
{
    lowpass_instance_t* inst = (lowpass_instance_t*)instance;
    const int16_t *s = (const int16_t *)inframe1;
    int16_t *d = (int16_t *)outframe;
    int i, j, al = 0, ar = 0;
    int audio_frame_size = 0;
    RETURN_ERROR_IF_NULL(inst);

    if (!inst->ready) {
        return 0;
    }

    if (inst->channel_count == 1) {
        for (i = 0; i < audio_frame_size / 2; i++) {
            inst->array_r[inst->p] = s[i];
            for (j = 0; j < inst->taps; j++) {
                ar += inst->array_r[j];
	        }
	        inst->p = (inst->p+1) % inst->taps;
            ar /= inst->taps;
		    d[i] = ar * inst->highpass;
	    }
    } else {
        for (i = 0; i < audio_frame_size / 2; i++) {
            inst->array_l[inst->p] = s[i+0];
	        inst->array_r[inst->p] = s[i+1];
            for (j = 0; j < inst->taps; j++) {
                al += inst->array_l[j];
                ar += inst->array_r[j];
	        }
	        inst->p = (inst->p+1) % inst->taps;
            al /= inst->taps;
            ar /= inst->taps;
		    d[i+0] = al * inst->highpass;
		    d[i+1] = ar * inst->highpass;
	    }
    }

    return 0;
}

/*
 * vim: expandtab shiftwidth=4:
 */

