/*  anormalize.c
 *  (C) 2012-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) pl <p_l@gmx.fr> 2002 and beyond...
 *                Tilmann Bitterberg - June 2002 ported to transcode
 *
 *  Sources: some ideas from volnorm plugin for xmms
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "memutils.h" 

#include "frei0rx.h"
#include "frei0rx_compat.h"
#include "frei0rx_paramext.h"

enum {
    PLUGIN_PARAM_NUM = 3
};

/* Values for AVG (algo):
 * 1: uses a 1 value memory and coefficients new=a*old+b*cur (with a+b=1)
 *
 * 2: uses several samples to smooth the variations (standard weighted mean
 *    on past samples)
 *
 * Limitations:
 *  - only AFMT_S16_LE supported
 */

// basic parameter
// mul is the value by which the samples are scaled
// and has to be in [MUL_MIN, MUL_MAX]
#define MUL_INIT 1.0
#define MUL_MIN 0.1
#define MUL_MAX 5.0

#define MIN_SAMPLE_SIZE 32000

// FIXME: extract common factors
// Some limits
#define MIN_S16 -32768
#define MAX_S16  32767

// "Ideal" level
#define MID_S16 (MAX_S16 * 0.25)

// Silence level
// FIXME: should be relative to the level of the samples
#define SIL_S16 (MAX_S16 * 0.01)

#define DEFAULT_SMOOTH 0.06

#define NSAMPLES 128

struct mem_t {
    double avg;		// average level of the sample
    int32_t len;	// sample size (weight)
};

typedef struct anormalizeinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
	int format;
	double mul;
	double SMOOTH_MUL;
	double SMOOTH_LASTAVG;
	double lastavg;
	int idx;
	struct mem_t mem[NSAMPLES];
	int AVG;
    unsigned int sample_rate;
    unsigned int channel_count;
} anormalize_instance_t;

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)


#define CLAMP(x,m,M) do { \
    if ((x)<(m)) (x) = (m); \
    else if ((x)>(M)) (x) = (M); \
} while(0)

static void post_set_param(f0r_instance_t instance, int param_index)
{
    anormalize_instance_t* inst = (anormalize_instance_t*)instance;
    RETURN_IF_NULL(inst);
    if (param_index == 2) {
        CLAMP(inst->AVG, 1, 2);
    }
    return;
}

/*************************************************************************/


int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}


void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "anormalize";
    info->author = "pl, Tilmann Bitterberg";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = 0;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"Audio volume normalizer"
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_AUDIO;
    info->sample_model = F0RX_SAMPLE_MODEL_S16LE;
    info->channel_layout = 0; /* TODO */
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Smooth";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Smoothing factor.";
        break;
      case 1:
        info->name = "Smooth Last";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Smoothing factor for the last sample.";
        break;
      case 2:
        info->name = "Algo";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Smoothing Algorhythm to use."
" 1: uses a 1 value memory and coefficients new=a*old+b*cur (with a+b=1)."
" 2: uses several samples to smooth the variations (standard weighted mean"
" on past samples).";
        break;
      default:
        break; /* XXX */
    }
    return;
}

int f0rx_set_params_values(f0r_instance_t instance,
                           f0r_param_t *params)
{
    return 0;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    anormalize_instance_t* inst = (anormalize_instance_t*)instance;
    return &inst->attrs[param_index];
}


f0r_instance_t f0rx_construct(int format, f0r_param_t *params,
                              unsigned int value1, unsigned int value2)
{
    anormalize_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        inst->sample_rate    = value1;
        inst->channel_count  = value2;
        inst->format         = 1; /* XXX bogus */
        inst->mul            = MUL_INIT;
        inst->lastavg        = MID_S16;
        inst->idx            = 0;
        inst->SMOOTH_MUL     = DEFAULT_SMOOTH;
        inst->SMOOTH_LASTAVG = DEFAULT_SMOOTH;
        inst->AVG            = 1;

//        memset(&inst->mem, 0, sizeof(inst->mem));

        f0rx_bind_param_attr(inst,           0, &inst->SMOOTH_MUL);
        f0rx_bind_param_attr(inst,           1, &inst->SMOOTH_LASTAVG);
        f0rx_bind_param_attr_remap_int(inst, 2, &inst->AVG);
        f0rx_hook_param_attr_post_set(inst,  2, post_set_param);
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0RX_SAMPLE_MODEL_S16LE, NULL, width, height);
}


void f0r_destruct(f0r_instance_t instance)
{
    anormalize_instance_t* inst = (anormalize_instance_t*)instance;
    RETURN_IF_NULL(inst);
    /* nothing to do here */
    tc_free(instance);
    return;
}

int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    int ready = 1;
    /* FIXME: implement the parameters sanity check. */
    return ready;
}

int f0rx_update(f0r_instance_t instance,
                double intime,
                const uint32_t *inframe1,
                const uint32_t *inframe2,
                const uint32_t *inframe3,
                double *outtime,
                uint32_t *outframe)
{
    const int16_t* src = (const int16_t *)inframe1;
    int16_t* dst = (int16_t *)outframe;
    int32_t i, tmp, len, totallen;
    double curavg, newavg, avg, neededmul;
    anormalize_instance_t* inst = (anormalize_instance_t*)instance;
    RETURN_ERROR_IF_NULL(inst);
    len = inst->sample_rate * inst->channel_count;
    // Evaluate current samples average level
    curavg = 0.0;
    for (i = 0; i < len ; ++i) {
        tmp = src[i];
        curavg += tmp * tmp;
    }
    curavg = sqrt(curavg / (double) len);

    // Evaluate an adequate 'mul' coefficient based on previous state, current
    // samples level, etc
    if (inst->AVG == 1) {
    	if (curavg > SIL_S16) {
	        neededmul = MID_S16 / ( curavg * inst->mul);
	        inst->mul = (1.0 - inst->SMOOTH_MUL) * inst->mul + inst->SMOOTH_MUL * neededmul;
    	    // Clamp the mul coefficient
	        CLAMP(inst->mul, MUL_MIN, MUL_MAX);
	    }
    } else if (inst->AVG == 2) {
    	avg = 0.0;
	    totallen = 0;

    	for (i = 0; i < NSAMPLES; ++i) {
	        avg += inst->mem[i].avg * (double) inst->mem[i].len;
	        totallen += inst->mem[i].len;
    	}

	    if (totallen > MIN_SAMPLE_SIZE) {
	        avg /= (double) totallen;
    	    if (avg >= SIL_S16) {
	        	inst->mul = (double) MID_S16 / avg;
		        CLAMP(inst->mul, MUL_MIN, MUL_MAX);
    	    }
    	}
    }

    // Scale & clamp the samples
    for (i = 0; i < len ; ++i) {
        tmp = inst->mul * src[i];
        CLAMP(tmp, MIN_S16, MAX_S16);
        dst[i] = tmp;
    }

    // Evaluation of newavg (not 100% accurate because of values clamping)
    newavg = inst->mul * curavg;

    // Stores computed values for future smoothing
    if (inst->AVG == 1) {
    	inst->lastavg = (1.0 - inst->SMOOTH_LASTAVG) * inst->lastavg
                      + inst->SMOOTH_LASTAVG * newavg;
    } else if (inst->AVG == 2) {
	    inst->mem[inst->idx].len = len;
    	inst->mem[inst->idx].avg = newavg;
	    inst->idx = (inst->idx + 1) % NSAMPLES;
    }

    return 0;
}

/*
 * vim: expandtab shiftwidth=4:
 */

