/*  astat.c
 *  (C) 2011-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Thomas Oestreich - June 2001
 *
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "memutils.h"

#include "frei0rx.h"
#include "frei0rx_paramext.h"
#include "frei0rx_compat.h"

enum {
    PLUGIN_PARAM_NUM = 2
};

typedef struct astatinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
    int32_t min;
    int32_t max;
    int frame_num;

    int silence_limit;
    char* log_path;

    unsigned int sample_rate;
    unsigned int channel_count;
} astat_instance_t;

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)

/* ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

enum {
    S16LE_MAX         = 0x7FFF,
    S16LE_BYTES       = 2,
    SILENCE_MAX_VALUE = 0 /* XXX */
};

enum {
    MACHINE_FRIENDLY = 0,
    HUMAN_FRIENDLY   = 1
};

static int report_stats(astat_instance_t* inst, FILE* out, int verbose)
{
    int ret = 0;
    double fmin = -((double) inst->min)/S16LE_MAX;
    double fmax =  ((double) inst->max)/S16LE_MAX;
    if (inst->min >= inst->silence_limit
     && inst->max <= inst->silence_limit) {
        if (verbose) {
            fprintf(out, "audio track seems only silence\n");
        } else {
            fprintf(out, "min=%.3f max=%.3f scale=1.0\n",
                         -fmin, fmax);
        }
    } else if (inst->min == 0 || inst->max == 0) {
        if (verbose) {
            fprintf(out, "bad minimum/maximum value,"
                         " unable to find scale value\n");
        } else {
            fprintf(out, "min=0.0 max=0.0 scale=0.0\n");
        }
        ret = -1;
    } else {
        double vol = (fmin < fmax) ? 1./fmax : 1./fmin;

        if (verbose) {
            fprintf(out, "(min=%.3f/max=%.3f), "
                         "normalize volume with scale=%.3f\n",
                         -fmin, fmax, vol);
        } else {
            fprintf(out, "min=%.3f max=%.3f scale=%.3f\n",
                         -fmin, fmax, vol);
        }
    }
    return ret;
}

static void set_range(astat_instance_t *inst, int v)
{
    if (v > inst->max) {
        inst->max = v;
    } else if (v < inst->min) {
        inst->min = v;
    }
    return;
}

static int astat_frame(astat_instance_t *inst,
                       const int16_t *channel_data,
                       int channel_size)
{
    const int16_t *s = channel_data;
    int n;
    for (n = 0; n < channel_size; n++) {
        set_range(inst, (int) (*s));
        s++;
    }
    return 0;
}


int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

int f0rx_set_params_values(f0r_instance_t instance,
                           f0r_param_t *params)
{
    return 0; /* unsupported */
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "astat";
    info->author = "Thomas Oestreich";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = 0;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"Audio statistics filter plugin."
" This filter scan audio track and compute optimal rescale value."
" It can also detect if the audio track is silence only."
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_AUDIO;
    info->sample_model = F0RX_SAMPLE_MODEL_S16LE;
    info->channel_layout = 0; /* TODO */
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Silence Limit";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The maximum audio amplitude of silence values.";
        break;
      case 1:
        info->name = "Log Path";
        info->type = F0R_PARAM_STRING;
        info->explanation = "File to save a detailed statistics.";
        break;
      default:
        break; /* XXX */
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    astat_instance_t* inst = (astat_instance_t*)instance;
    return &inst->attrs[param_index];
}


f0r_instance_t f0rx_construct(int format, f0r_param_t *params,
                              unsigned int value1, unsigned int value2)
{
    astat_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        inst->min           = 0;
        inst->max           = 0;
        inst->log_path      = NULL;
        inst->silence_limit = SILENCE_MAX_VALUE;
        inst->sample_rate   = value1;
        inst->channel_count = value2;
        f0rx_bind_param_attr_remap_int(inst, 0, &inst->silence_limit);
        f0rx_bind_param_attr(inst, 1, &inst->log_path);
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0RX_SAMPLE_MODEL_S16LE, NULL, width, height);
}


void f0r_destruct(f0r_instance_t instance)
{
    astat_instance_t* inst = (astat_instance_t*)instance;
    RETURN_IF_NULL(inst);
    if (inst->frame_num > 0) {
        report_stats(inst, stderr, HUMAN_FRIENDLY);
        if (inst->log_path) {
            FILE* out = fopen(inst->log_path, "wt+");
            if (out) {
                report_stats(inst, out, MACHINE_FRIENDLY);
                fclose(out); /* XXX what about failures? */
            } /* XXX else? */
            free(inst->log_path);
        }
    }
    free(instance);
    return;
}

int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    return 1; /* easy, uh? */
}

int f0rx_update(f0r_instance_t instance,
                double intime,
                const uint32_t *inframe1,
                const uint32_t *inframe2,
                const uint32_t *inframe3,
                double *outtime,
                uint32_t *outframe)
{
    astat_instance_t* inst = (astat_instance_t*)instance;
    int channel_size;
    RETURN_ERROR_IF_NULL(inst);
    channel_size = S16LE_BYTES * inst->sample_rate;
    /* first, let's provide a valid output */
    ac_memcpy(outframe, inframe1,
              channel_size * inst->channel_count);
    /* then the actual processing */
    inst->frame_num++;
    /* FIXME: this is the original algo. I'm not sold, though. */
    return astat_frame(inst, (const int16_t *)inframe1, channel_size);
}


/*
 * vim: expandtab shiftwidth=4:
 */

