/*  barrel.c
 *  (C) 2011-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  filter_barrel.c -- barrel filter: applies or removes barrel distortion
 *  Copyright (C) 2009-2010 Andrew Church; based on filter_invert.c, with
 *      concept taken from GIMP lens distortion filter
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "memutils.h"

#include "frei0rx.h"
#include "frei0rx_frame.h"
#include "frei0rx_paramext.h"
#include "frei0rx_compat.h"


enum {
    PLUGIN_PARAM_NUM = 7,
    BLACK_Y          = 16,
    BLACK_UV         = 128
};

typedef struct distortion_map_entry_ {
    int16_t x, y;
    uint16_t weight[3][3];  // Weight of [y-1..y+1][x-1..x+1] out of 0x8000
} distortion_map_entry_t;


typedef struct barrelinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
    int width, height;              // Frame width/height (for reference)
    double order2, order4;          // Order-2 and order-4 filter strength
    int cx, cy;                     // Center of distortion
    unsigned int start, end, step;  // Filter application range and step
    unsigned int framenum;
    distortion_map_entry_t *map_y;  // Distortion map (Y plane)
    distortion_map_entry_t *map_uv; // Distortion map (U/V planes)
} barrel_instance_t;

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)



/**
 * gen_distortion_map:  Generates a distortion map for the given filter
 * parameters.
 *
 * Parameters:
 *      width: Frame width, in pixels.
 *     height: Frame height, in pixels.
 *     cx, cy: Pixel coordinates of center of distortion effect.
 *     order2: Strength of second-order distortion.
 *     order4: Strength of fourth-order distortion.
 * Return value:
 *     Newly-allocated distortion map (NULL on error).
 */

static distortion_map_entry_t *gen_distortion_map(int width, int height,
                                                  double cx, double cy,
                                                  double order2, double order4)
{
    distortion_map_entry_t *map;
    const double r_scale_sq = 4.0 / (width*width + height*height);
    int x, y, index;

    map = tc_malloc(sizeof(*map) * (width * height));
    if (!map) {
        return NULL;
    }

    for (index = 0, y = 0; y < height; y++) {
        for (x = 0; x < width; x++, index++) {
            const double dx = (x+0.5) - cx, dy = (y+0.5) - cy;
            const double r_sq = (dx*dx + dy*dy) * r_scale_sq;
            const double mult = 1 + (order2 * r_sq)
                                  + (order4 * r_sq * r_sq);
            const double srcx = cx + (mult * dx);
            const double srcy = cy + (mult * dy);
            map[index].x = (int16_t)floor(srcx);
            map[index].y = (int16_t)floor(srcy);
            /* Compute raw weights */
            double weight[3][3], total_weight = 0;
            int xx, yy;
            for (yy = -1; yy <= 1; yy++) {
                const double weight_dy = (map[index].y + yy + 0.5) - srcy;
                for (xx = -1; xx <= 1; xx++) {
                    const double weight_dx = (map[index].x + xx + 0.5) - srcx;
                    const double dist =
                        sqrt(weight_dx*weight_dx + weight_dy*weight_dy);
                    /* Simple cubic (FIXME: just a quick hack; better ideas?)*/
                    weight[yy+1][xx+1] = dist>=1 ? 0 :
                        (3.0 + (dist * dist * (-7.0 + (dist * 4.0)))) / 3.0;
                    total_weight += weight[yy+1][xx+1];
                }
            }
            /* Rescale to a total of 0x8000 */
            uint16_t int_total = 0;
            for (yy = -1; yy <= 1; yy++) {
                for (xx = -1; xx <= 1; xx++) {
                    map[index].weight[yy+1][xx+1] = (uint16_t)floor(
                        ((weight[yy+1][xx+1] / total_weight) * 0x8000) + 0.5
                    );
                    int_total += map[index].weight[yy+1][xx+1];
                }
            }
            if (int_total != 0x8000) {  // Tweak for rounding error
                map[index].weight[1][1] += 0x8000 - (int)int_total;
            }
        }
    }

    return map;
}


/**
 * filter_plane:  Apply the barrel distortion filter to a single plane of a
 * video frame.
 *
 * Parameters:
 *        src: Pointer to source data.
 *       dest: Pointer to destination buffer.
 *        map: Pointer to distortion map.
 *     defval: Value for out-of-frame pixels.
 *      width: Width of plane, in pixels.
 *     height: Height of plane, in pixels.
 * Return value: None.
 */

static void filter_plane(const uint8_t *src, uint8_t *dest,
                         const distortion_map_entry_t *map, uint8_t defval,
                         int width, int height)
{
    int x, y, index;

    if (!src || !dest) { // can't happen
        return;
    }

    for (index = 0, y = 0; y < height; y++) {
        for (x = 0; x < width; x++, index++) {
            uint32_t pixel_total = 0;
            int xx, yy;
            for (yy = -1; yy <= 1; yy++) {
                const int srcy = map[index].y + yy;
                for (xx = -1; xx <= 1; xx++) {
                    const int srcx = map[index].x + xx;
                    uint32_t pixel;
                    if (srcx < 0 || srcx >= width
                     || srcy < 0 || srcy >= height
                    ) {
                        pixel = defval;
                    } else {
                        pixel = src[srcy * width + srcx];
                    }
                    pixel_total += pixel * map[index].weight[yy+1][xx+1];
                }
            }
            dest[index] = pixel_total >> 15;
        }
    }
}

static void barrel_defaults(barrel_instance_t *inst)
{
    /* Set defaults */
    inst->order2 = 0;
    inst->order4 = 0;
    inst->cx = inst->width / 2;
    inst->cy = inst->height / 2;
    inst->start = 0;
    inst->end = (unsigned int)-1;
    inst->step = 1;
}

static void post_set_param(f0r_instance_t instance, int param_index)
{
    barrel_instance_t* inst = (barrel_instance_t*)instance;
    RETURN_IF_NULL(inst);
    if (param_index >=0 && param_index < 4) { // FIXME magic number
        /* Compute distortion map based on settings */
        tc_free(inst->map_y);
        inst->map_y = gen_distortion_map(inst->width, inst->height, inst->cx, inst->cy,
                                         inst->order2, inst->order4);
        tc_free(inst->map_uv);
        inst->map_uv = gen_distortion_map(inst->width/2, inst->height/2,
                                          (double)inst->cx/2, (double)inst->cy/2,
                                          inst->order2, inst->order4);
        if (!inst->map_y || !inst->map_uv) {
            fprintf(stderr, "[%s] Not enough memory for %dx%d distortion map\n",
                    __FILE__, inst->width, inst->height);
        }
    }
    return;
}


int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    int ready = 0;
    barrel_instance_t* inst = (barrel_instance_t*)instance;
    if (inst->map_y && inst->map_uv) {
        ready = 1;
    }
    return ready;
}


int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "barrel";
    info->author = "Andrew Church";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0RX_COLOR_MODEL_PLANAR_YUV420P;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"Apply/remove barrel distortion.\n"
" Apply or remove barrel distortion, such as that produced by a"
" wide-angle camera lens.  Positive values for \"order2\" or \"order4\""
" apply barrel distortion, while negative values remove barrel"
" distortion (or, conversely, apply pincushion distortion)."
" Filter is applied before basic transformations (zoom, etc)."
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_VIDEO;
    info->sample_model = 0;
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Order2";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Strength of order-2 distortion.";
        break;
      case 1:
        info->name = "Order4";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Strength of order-4 distortion.";
        break;
      case 2:
        info->name = "Center X";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Center of distortion (X).";
        break;
      case 3:
        info->name = "Center X";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Center of distortion (Y).";
        break;
      case 4:
        info->name = "Start";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The first frame to consider.";
        break;
      case 5:
        info->name = "End";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The last frame to consider.";
        break;
      case 6:
        info->name = "Step";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Apply the check every `Step' frames.";
        break;
      default:
        break; /* XXX */
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    barrel_instance_t* inst = (barrel_instance_t*)instance;
    return &inst->attrs[param_index];
}


f0r_instance_t f0rx_construct(int format,
                              unsigned int param1, unsigned int param2)
{
    barrel_instance_t* inst = NULL;
    if (format == F0RX_COLOR_MODEL_PLANAR_YUV420P) {
        inst = tc_zalloc(sizeof(*inst));
        if (inst) {
            inst->width = param1;
            inst->height = param2;
            inst->framenum = 0;

            barrel_defaults(inst);

            f0rx_bind_param_attr(inst,           0, &inst->order2);
            f0rx_bind_param_attr(inst,           1, &inst->order4);
            f0rx_bind_param_attr_remap_int(inst, 2, &inst->cx);
            f0rx_bind_param_attr_remap_int(inst, 3, &inst->cy);
            f0rx_bind_param_attr_remap_int(inst, 4, &inst->start);
            f0rx_bind_param_attr_remap_int(inst, 5, &inst->end);
            f0rx_bind_param_attr_remap_int(inst, 6, &inst->step);
            f0rx_hook_param_attr_post_set(inst, F0RX_PARAM_ALL, post_set_param);
        }
    } else {
        fprintf(stderr, "[%s] Unsupported format [0x%X]\n", __FILE__, format);
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0RX_COLOR_MODEL_PLANAR_YUV420P, width, height);
}


void f0r_destruct(f0r_instance_t instance)
{
    barrel_instance_t* inst = (barrel_instance_t*)instance;
    RETURN_IF_NULL(inst);
    tc_free(inst->map_y);
    tc_free(inst->map_uv);
    free(instance);
    return;
}

int f0rx_update2(f0r_instance_t instance,
                 const f0rx_frame_t* in1,
                 const f0rx_frame_t* in2,
                 const f0rx_frame_t* in3,
                 f0rx_frame_t* out)
{
    barrel_instance_t* inst = (barrel_instance_t*)instance;
    RETURN_ERROR_IF_NULL(inst);
    if (f0rx_is_ready(instance, -1)
     && (inst->framenum >= inst->start && inst->framenum <= inst->end)
     && (inst->framenum - inst->start) % inst->step == 0
    ) {
        filter_plane(in1->planes[0], out->planes[0], inst->map_y,  BLACK_Y,
                     inst->width,   inst->height);
        filter_plane(in1->planes[1], out->planes[1], inst->map_uv, BLACK_UV,
                     inst->width/2, inst->height/2);
        filter_plane(in1->planes[2], out->planes[2], inst->map_uv, BLACK_UV,
                     inst->width/2, inst->height/2);
    }
    inst->framenum++;
    return 0;
}

/*
 * vim: expandtab shiftwidth=4:
 */

