/*  compare.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Antonio Beamud Montero <antonio.beamud@linkend.com>
 *  Copyright (C) Microgenesis S.A.
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

/* TODO: this still applies?
 * Note: because of ImageMagick bogosity, this must be included first, so
 * we can undefine the PACKAGE_* symbols it splats into our namespace
#include <magick/api.h>
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION
*/

#include "memutils.h"
#include "strutils.h"
#include "gmlib.h"

#include "frei0rx.h"
#include "frei0rx_param.h"
#include "frei0rx_compat.h"


#define DELTA_COLOR 45.0

/* FIXME: I found the whole algorythm debatable.
 * I'm pretty sure we can do better.
 */

enum {
    PLUGIN_PARAM_NUM = 3
};

typedef struct pixelsmask_ {
    uint32_t row;
    uint32_t col;
    uint8_t r;
    uint8_t g;
    uint8_t b;
} pixels_mask_t;


typedef struct compareinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];

    FILE *results;

    char *pattern_name;
    char *results_name;
    double delta;

    TCMagickContext magick;
    pixels_mask_t *pixel_mask;
    int pixel_count;

    unsigned int frames;

    int width;
    int height;
    int size;

    int ready;
} compare_instance_t;

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)


/* ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

/*************************************************************************/

enum {
    RGB_BYTES   = 3, /* how much storage a pixel take? */
    RGBA_BYTES  = 4, /* ditto */
};


static int compare_load_pixels(compare_instance_t *inst, Image *pattern)
{
    unsigned int t, r, index, j = 0;
    PixelPacket *pixel_packet = GetImagePixels(pattern, 0, 0,
                                                pattern->columns,
                                                pattern->rows);
    if (!pixel_packet) {
        return -1;
    }

    inst->pixel_count = 0;
    for (t = 0; t < pattern->rows; t++) {
        for (r = 0; r < pattern->columns; r++) {
            index = t * pattern->columns + r;
            if (pixel_packet[index].opacity == 0){
                inst->pixel_count++;
            }
        }
    }

    tc_free(inst->pixel_mask);
    inst->pixel_mask = tc_zalloc(inst->pixel_count * sizeof(pixels_mask_t));
    if (!inst->pixel_mask) {
        return -1;
    }

    for (t = 0; t < pattern->rows; t++) {
        for (r = 0; r < pattern->columns; r++) {
            index = t * pattern->columns + r;
            if (pixel_packet[index].opacity == 0) {
                pixels_mask_t *temp = &inst->pixel_mask[j];
                temp->row = t;
                temp->col = r;
                temp->r = (uint8_t)ScaleQuantumToChar(pixel_packet[index].red);
                temp->g = (uint8_t)ScaleQuantumToChar(pixel_packet[index].green);
                temp->b = (uint8_t)ScaleQuantumToChar(pixel_packet[index].blue);
                j++;
            }
        }
    }
    return 0;
}

static void compare_load_pattern(f0r_instance_t instance, int unused)
{
    compare_instance_t* inst = (compare_instance_t*)instance;
    ExceptionInfo *exc = &inst->magick.exception_info;
    Image *pattern = NULL;
    Image *resized = NULL;
    Image *original = NULL;
 
    strlcpy(inst->magick.image_info->filename, inst->pattern_name, MaxTextExtent);
    original = ReadImage(inst->magick.image_info, exc);
    if (original == NULL) {
        MagickError(exc->severity, exc->reason, exc->description);
        return;
    }
    resized = ResizeImage(original,
                          inst->width,
                          inst->height,
                          GaussianFilter,
                          1,
                          exc);
    if (resized == NULL) {
        MagickError(exc->severity, exc->reason, exc->description);
        return;
    }
    pattern = FlipImage(resized, exc); /* XXX: why? */
    if (pattern == NULL) {
        MagickError(exc->severity, exc->reason, exc->description);
        return;
    }
    compare_load_pixels(inst, pattern);
    return;
}

static int f0rx_try_to_open(compare_instance_t *inst)
{
    if (inst->frames == 0) {
        inst->results = fopen(inst->results_name, "wt");
        if (inst->results) {
            inst->ready = 1;
        } else {
            perror("could not open file for writing");
            inst->ready = 0;
        }
    }
    return 0;
}

static int f0rx_try_to_close(compare_instance_t *inst)
{
    int ret = 0;
    if (inst->results) {
        ret = fclose(inst->results);
    }
    return ret;
}

/*************************************************************************/


int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    int ready = 0;
    compare_instance_t* inst = (compare_instance_t*)instance;
    if (inst) {
        ready = inst->ready;
    }
    return ready;
}

int f0r_init()
{
    /* intialization deferred into f0r_construct */
    return 1;
}

void f0r_deinit()
{
    /* deintialization deferred into f0r_denstruct */
    return;
}


void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "compare";
    info->author = "Antonio Beamud";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"Generate a file in with information about the times,"
" frame, etc the pattern defined in the image"
" parameter is observed."
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_VIDEO;
    info->sample_model = 0;
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Pattern";
        info->type = F0R_PARAM_STRING;
        info->explanation = "The path to the file used like pattern.";
        break;
      case 1:
        info->name = "Results";
        info->type = F0R_PARAM_STRING;
        info->explanation = "The path to the file used to write the results.";
        break;
      case 2:
        info->name = "Delta";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Delta error allowed.";
        break;
      default:
        break; // XXX
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    compare_instance_t* inst = (compare_instance_t*)instance;
    return &inst->attrs[param_index];
}

static int f0rx_compare_init(compare_instance_t *inst,
                             unsigned int param1, unsigned int param2)
{
    int err = 0;
    
    inst->delta = DELTA_COLOR;
    inst->frames = 0;
    inst->pixel_mask = NULL;
    inst->width = param1;
    inst->height = param2;

    err = tc_magick_init(&inst->magick, GMLIB_QUALITY_DEFAULT);
    if (!err) {
        compare_load_pattern(inst, 0);
    }
    return err;
}

f0r_instance_t f0rx_construct(int format,
                              unsigned int param1, unsigned int param2)
{
    compare_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        int err = f0rx_compare_init(inst, param1, param2);
        if (!err) {
            f0rx_bind_param_attr(inst, 0, &inst->pattern_name);
            f0rx_bind_param_attr(inst, 1, &inst->results_name);
            f0rx_bind_param_attr(inst, 2, &inst->delta);

            f0rx_hook_param_attr_post_set(inst, 0, compare_load_pattern);
        } else {
            f0r_destruct(inst);
            inst = NULL;
        }
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0R_COLOR_MODEL_RGBA8888, width, height);
}


void f0r_destruct(f0r_instance_t instance)
{
    compare_instance_t* inst = (compare_instance_t*)instance;
    RETURN_IF_NULL(inst);
    tc_magick_fini(&inst->magick);
    f0rx_try_to_close(inst);
    free(inst->pattern_name);
    free(inst->results_name);
    free(instance);
    return;
}

static int f0rx_compare(compare_instance_t *inst, const uint8_t *src)
{
    f0rx_try_to_open(inst);
    if (inst->ready && inst->pixel_mask != NULL) {
        const char *res = "n";
        int r, g, b, c = 0, j = 0;
        double avg_dr = 0.0, avg_dg = 0.0, avg_db = 0.0;
        double sr = 0.0;
        double sg = 0.0;
        double sb = 0.0;

        double width_long = inst->width * RGBA_BYTES;
        for (j = 0; j < inst->pixel_count; j++) {
            pixels_mask_t *item = &inst->pixel_mask[j];
            r = item->row * width_long + item->col * RGBA_BYTES;
            g = item->row * width_long + item->col * RGBA_BYTES + 1;
            b = item->row * width_long + item->col * RGBA_BYTES + 2;
            /* diff between points */
            /* Interchange RGB values if necesary */
            sr = sr + (double)abs(src[r] - item->r);
            sg = sg + (double)abs(src[g] - item->g);
            sb = sb + (double)abs(src[b] - item->b);
            c++;
        }

        avg_dr = sr/(double)c;
        avg_dg = sg/(double)c;
        avg_db = sb/(double)c;

        if ((avg_dr < inst->delta) && (avg_dg < inst->delta) && (avg_db < inst->delta)) {
            res = "1";
        }
        if (inst->results){
            fprintf(inst->results, "%s", res);
            fflush(inst->results);
        }
    }
    return 0;
}

int f0rx_update2(f0r_instance_t instance,
                 const f0rx_frame_t* in1,
                 const f0rx_frame_t* in2,
                 const f0rx_frame_t* in3,
                 f0rx_frame_t* out)
{
    compare_instance_t* inst = (compare_instance_t*)instance;
    RETURN_ERROR_IF_NULL(inst);
    ac_memcpy(out->planes[0], in1->planes[0], out->pitches[0]);
    f0rx_compare(inst, in1->planes[0]);
    inst->frames++;
    return 0;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    f0rx_frame_t in;
    f0rx_frame_t out;
    compare_instance_t* inst = (compare_instance_t*)instance;
    RETURN_IF_NULL(inst);
    memset(&in, 0, sizeof(in));
    memset(&out, 0, sizeof(out));
    in.planes[0] = (void *)inframe1;
    in.pitches[0] = RGBA_BYTES * inst->width * inst->height;
    in.time.d = time;
    out.planes[0] = (void *)outframe;
    out.pitches[0] = RGBA_BYTES * inst->width * inst->height;
    out.time.d = time;
    f0rx_update2(instance, &in, NULL, NULL, &out);
    return;

}

/*
 * vim: expandtab shiftwidth=4:
 */

