/*  hqdn3d.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) 2003 Daniel Moreno <comac@comac.darktech.org>
 *   Converted for use in transcode by Tilmann Bitterberg <transcode@tibit.org>
 *   
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "memutils.h"

#include "frei0rx.h"
#include "frei0rx_frame.h"
#include "frei0rx_compat.h"
#include "frei0rx_param.h"

enum {
    PLUGIN_PARAM_NUM = 4
};

typedef struct hqdn3dinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
    int Coefs[4][512*16];
    uint32_t *Line;
    uint16_t *Frame[3];
    unsigned int Width;
    unsigned int Height;
    /* configurable */
    double LumaSpatial;
    double LumaTemporal;
    double ChromaSpatial;
    double ChromaTemporal;
} hqdn3d_instance_t;

enum {
    RGBA_BYTES = 4
};

#define RETURN_IF_NULL(INST) do { \
    if (!(INST)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)


/* meh */
#define PARAM1_DEFAULT 4.0
#define PARAM2_DEFAULT 3.0
#define PARAM3_DEFAULT 6.0

static void PrecalcCoefs(int *Ct, double Dist25)
{
    int i;
    double Gamma, Simil, C;

    Gamma = log(0.25) / log(1.0 - Dist25/255.0 - 0.00001);

    for (i = -256*16; i < 256*16; i++) {
        Simil = 1.0 - (i>0?i:-i) / (16*255.0);
        C = pow(Simil, Gamma) * 65536.0 * (double)i / 16.0;
        Ct[16*256+i] = (C<0) ? (C-0.5) : (C+0.5);
    }
    return;
}


static inline uint32_t LowPassMul(uint32_t PrevMul, uint32_t CurrMul, int* Coef)
{
//    int dMul= (PrevMul&0xFFFFFF)-(CurrMul&0xFFFFFF);
    int dMul= PrevMul-CurrMul;
    int d=((dMul+0x10007FF)/(65536/16));
    return CurrMul + Coef[d];
}

static void deNoise(const uint8_t *Frame,
                    uint8_t *FrameDest,
                    uint32_t *LineAnt, /* (width bytes) */
                    uint16_t **FrameAntPtr,
                    int W, int H, int sStride, int dStride,
                    int *Horizontal, int *Vertical, int *Temporal)
{
    int X, Y;
    int sLineOffs = 0, dLineOffs = 0;
    uint32_t PixelAnt;
    int PixelDst;
    uint16_t* FrameAnt=(*FrameAntPtr);

    if(!FrameAnt){
        (*FrameAntPtr) = FrameAnt = tc_bufalloc(W * H * sizeof(uint16_t));
        // FIXME
        for (Y = 0; Y < H; Y++){
            const uint8_t* src=Frame+Y*sStride;
            uint16_t* dst=&FrameAnt[Y*W];
            for (X = 0; X < W; X++)
                dst[X]=src[X]<<8;
        }
    }

    /* First pixel has no left nor top neightbour. Only previous frame */
    LineAnt[0] = PixelAnt = Frame[0]<<16;
    PixelDst = LowPassMul(FrameAnt[0]<<8, PixelAnt, Temporal);
    FrameAnt[0] = ((PixelDst+0x1000007F)/256);
    FrameDest[0]= ((PixelDst+0x10007FFF)/65536);

    /* Fist line has no top neightbour. Only left one for each pixel and
     * last frame */
    for (X = 1; X < W; X++) {
        LineAnt[X] = PixelAnt = LowPassMul(PixelAnt, Frame[X]<<16, Horizontal);
        PixelDst = LowPassMul(FrameAnt[X]<<8, PixelAnt, Temporal);
        FrameAnt[X] = ((PixelDst+0x1000007F)/256);
        FrameDest[X]= ((PixelDst+0x10007FFF)/65536);
    }

    for (Y = 1; Y < H; Y++){
        uint32_t PixelAnt;
        uint16_t* LinePrev=&FrameAnt[Y*W];
        sLineOffs += sStride, dLineOffs += dStride;
        /* First pixel on each line doesn't have previous pixel */
        PixelAnt = Frame[sLineOffs]<<16;
        LineAnt[0] = LowPassMul(LineAnt[0], PixelAnt, Vertical);
        PixelDst = LowPassMul(LinePrev[0]<<8, LineAnt[0], Temporal);
        LinePrev[0] = ((PixelDst+0x1000007F)/256);
        FrameDest[dLineOffs]= ((PixelDst+0x10007FFF)/65536);

        for (X = 1; X < W; X++){
            int PixelDst;
            /* The rest are normal */
            PixelAnt = LowPassMul(PixelAnt, Frame[sLineOffs+X]<<16, Horizontal);
            LineAnt[X] = LowPassMul(LineAnt[X], PixelAnt, Vertical);
            PixelDst = LowPassMul(LinePrev[X]<<8, LineAnt[X], Temporal);
            LinePrev[X] = ((PixelDst+0x1000007F)/256);
            FrameDest[dLineOffs+X]= ((PixelDst+0x10007FFF)/65536);
        }
    }
    return;
}


static void post_set_param(f0r_instance_t instance, int param_index)
{
    hqdn3d_instance_t* inst = (hqdn3d_instance_t*)instance;
    RETURN_IF_NULL(inst);
    switch (param_index) {
      case 0: /* Luma spatial */
        inst->LumaTemporal = PARAM3_DEFAULT * inst->LumaSpatial / PARAM1_DEFAULT;
        inst->ChromaSpatial = PARAM2_DEFAULT * inst->LumaSpatial / PARAM1_DEFAULT;
        inst->ChromaTemporal = inst->LumaTemporal * inst->ChromaSpatial / inst->LumaSpatial;
        break;
      case 1: /* Chroma spatial */
        inst->ChromaTemporal = inst->LumaTemporal * inst->ChromaSpatial / inst->LumaSpatial;
        break;
      case 2: /* Luma temporal */
        inst->ChromaTemporal = inst->LumaTemporal * inst->ChromaSpatial / inst->LumaSpatial;
        break;
      default:
        break;
    }
    /* Any given sunday. Every damn time. */
    PrecalcCoefs(inst->Coefs[0], inst->LumaSpatial);
    PrecalcCoefs(inst->Coefs[1], inst->LumaTemporal);
    PrecalcCoefs(inst->Coefs[2], inst->ChromaSpatial);
    PrecalcCoefs(inst->Coefs[3], inst->ChromaTemporal);
    return;
}

int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    return 1; /* pretty easy, uh?*/
}

int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "hqdn3d";
    info->author = "Daniel Moreno, A'rpi";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0RX_COLOR_MODEL_PLANAR_YUV420P;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"This filter aims to reduce image noise producing"
" smooth images and making still images really still"
" (This should enhance compressibility)."
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_VIDEO;
    info->sample_model = 0;
    info->channel_layout = 0;
}

void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Luma Spatial";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Luma spatial strenght.";
        break;
      case 1:
        info->name = "Chroma Spatial";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Chroma spatial strenght.";
        break;
      case 2:
        info->name = "Luma Temporal";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Luma temporal strenght.";
        break;
      case 3:
        info->name = "Chroma Temporal";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Chroma temporal strenght.";
        break;
      default:
        break; // XXX
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    hqdn3d_instance_t* inst = (hqdn3d_instance_t*)instance;
    return &inst->attrs[param_index];
}


f0r_instance_t f0rx_construct(int format,
                              unsigned int param1, unsigned int param2)
{
    hqdn3d_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        inst->LumaSpatial = PARAM1_DEFAULT;
        inst->ChromaSpatial = PARAM2_DEFAULT;
        inst->LumaTemporal = PARAM3_DEFAULT;
        inst->ChromaTemporal = inst->LumaTemporal * inst->ChromaSpatial / inst->LumaSpatial;
        inst->Line = tc_zalloc(2 * param1);
        inst->Width = param1;
        inst->Height = param2;
        if (!inst->Line) {
            f0r_destruct(inst);
            inst = NULL;
        } else {
            f0rx_bind_param_attr(inst, 0, &inst->LumaSpatial);
            f0rx_bind_param_attr(inst, 1, &inst->ChromaSpatial);
            f0rx_bind_param_attr(inst, 2, &inst->LumaTemporal);
            f0rx_bind_param_attr(inst, 3, &inst->ChromaTemporal);
            f0rx_hook_param_attr_post_set(inst, F0RX_PARAM_ALL, post_set_param);
        }
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0RX_COLOR_MODEL_PLANAR_YUV420P, width, height);
}

void f0r_destruct(f0r_instance_t instance)
{
    hqdn3d_instance_t* inst = (hqdn3d_instance_t*)instance;
    RETURN_IF_NULL(inst);
    tc_free(inst->Line);
    tc_buffree(inst->Frame[0]);
    tc_buffree(inst->Frame[1]);
    tc_buffree(inst->Frame[2]);
    tc_free(instance);
    return;
}

int f0rx_update2(f0r_instance_t instance,
                 const f0rx_frame_t* in1,
                 const f0rx_frame_t* in2,
                 const f0rx_frame_t* in3,
                 f0rx_frame_t* out)
{
    hqdn3d_instance_t* inst = (hqdn3d_instance_t*)instance;
    RETURN_ERROR_IF_NULL(inst);
    deNoise(in1->planes[0], out->planes[0],
            inst->Line, &inst->Frame[0],
            inst->Width, inst->Height,
            in1->pitches[0], out->pitches[0],
            inst->Coefs[0], inst->Coefs[0], inst->Coefs[1]);
    deNoise(in1->planes[1], out->planes[1],
            inst->Line, &inst->Frame[1],
            inst->Width >> 1, inst->Height >> 1, // FIXME
            in1->pitches[1], out->pitches[1],
            inst->Coefs[2], inst->Coefs[2], inst->Coefs[3]);
    deNoise(in1->planes[2], out->planes[2],
            inst->Line, &inst->Frame[2],
            inst->Width >> 1, inst->Height >> 1, // FIXME
            in1->pitches[2], out->pitches[2],
            inst->Coefs[2], inst->Coefs[2], inst->Coefs[3]);
    return 0;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    f0rx_frame_t in;
    f0rx_frame_t out;
    hqdn3d_instance_t* inst = (hqdn3d_instance_t*)instance;
    RETURN_IF_NULL(inst);
    f0rx_init_frame_yuv420p(&in, (void *)inframe1, time, // XXX
                            inst->Width, inst->Height);
    f0rx_init_frame_yuv420p(&out, outframe, time,
                            inst->Width, inst->Height);
    f0rx_update2(instance, &in, NULL, NULL, &out);
    return;
}

/*
 * vim: expandtab shiftwidth=4:
 */

