/*  levels.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) 2004 Bryan Mayland <bmayland@leoninedev.com>
 *  For use in transcode by Tilmann Bitterberg <transcode@tibit.org>
 *  
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* this is a pet peeve of mine and kinda like a `hello world' for transcode
 * related plugins. You very likely ain't gonna need that. -- Francesco
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "memutils.h"

#include "frei0rx.h"
#include "frei0rx_frame.h"
#include "frei0rx_compat.h"
#include "frei0rx_paramext.h"


enum {
    DEFAULT_IN_BLACK  = 0,
    DEFAULT_IN_WHITE  = 255,
    DEFAULT_OUT_BLACK = 0,
    DEFAULT_OUT_WHITE = 255,

    MAP_SIZE =          256,

    PLUGIN_PARAM_NUM =  5
};

typedef struct levelsinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
    int in_black;
    int in_white;
    int out_black;
    int out_white;
    float in_gamma;
    uint8_t lumamap[MAP_SIZE];
    unsigned int width;
    unsigned int height;
} levels_instance_t;

#define RETURN_IF_NULL(INST) do { \
    if (!(INST)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)


// XXX
#define DEFAULT_IN_GAMMA   1.0

static void build_map(uint8_t *map, int inlow, int inhigh,
                      float ingamma, int outlow, int outhigh)
{
    int i;

    for (i = 0; i < MAP_SIZE; i++) {
        if (i <= inlow) {
            map[i] = outlow;
        } else if (i >= inhigh) {
            map[i] = outhigh;
        } else {
            float f = (float)(i - inlow) / (inhigh - inlow);
            map[i] = pow(f, 1/ingamma) * (outhigh - outlow) + outlow; // XXX
        }
    }
    return;
}


static void post_set_param(f0r_instance_t instance, int param_index)
{
    levels_instance_t* inst = (levels_instance_t*)instance;
    build_map(inst->lumamap, inst->in_black, inst->in_white,
              inst->in_gamma, inst->out_black, inst->out_white);
}


int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}


void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "levels";
    info->author = "Bryan Mayland";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0RX_COLOR_MODEL_PLANAR_YUV420P;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"Luminosity level scaler."
" Scales luminosity values in the source image, similar to"
" VirtualDub's 'levels' filter.  This is useful to scale ITU-R601"
" video (which limits luma to 16-235) back to the full 0-255 range."
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_VIDEO;
    info->sample_model = 0;
    info->channel_layout = 0;
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Input black";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Luma range of input: black.";
        break;
      case 1:
        info->name = "Input white";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Luma range of input: white.";
        break;
      case 2:
        info->name = "Gamma";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Gamma ramp to apply to input luma.";
        break;
      case 3:
        info->name = "Output black";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Luma range of output: black.";
        break;
      case 4:
        info->name = "Output white";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Luma range of output: white.";
        break;
      default:
        break; // XXX
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    levels_instance_t* inst = (levels_instance_t*)instance;
    return &inst->attrs[param_index];
}


f0r_instance_t f0rx_construct(int format,
                              unsigned int param1, unsigned int param2)
{
    levels_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        inst->in_black     = DEFAULT_IN_BLACK;
        inst->in_white     = DEFAULT_IN_WHITE;
        inst->in_gamma     = DEFAULT_IN_GAMMA;
        inst->out_black    = DEFAULT_OUT_BLACK;
        inst->out_white    = DEFAULT_OUT_WHITE;
        inst->width        = param1;
        inst->height       = param2;
        build_map(inst->lumamap, inst->in_black, inst->in_white,
                  inst->in_gamma, inst->out_black, inst->out_white);
        f0rx_bind_param_attr_remap_int(inst, 0, &inst->in_black);
        f0rx_bind_param_attr_remap_int(inst, 1, &inst->in_white);
        f0rx_bind_param_attr_remap_int(inst, 2, &inst->in_gamma);
        f0rx_bind_param_attr_remap_int(inst, 3, &inst->out_black);
        f0rx_bind_param_attr_remap_int(inst, 4, &inst->out_white);
        f0rx_hook_param_attr_post_set(inst, F0RX_PARAM_ALL, post_set_param);
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0RX_COLOR_MODEL_PLANAR_YUV420P, width, height);
}

void f0r_destruct(f0r_instance_t instance)
{
    levels_instance_t* inst = (levels_instance_t*)instance;
    RETURN_IF_NULL(inst);
    tc_free(instance);
    return;
}

int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    int ready = 1;
    /* FIXME: either complain about wrong parameters or autocorrect them. */
    return ready;
}

int f0rx_update2(f0r_instance_t instance,
                 const f0rx_frame_t* in1,
                 const f0rx_frame_t* in2,
                 const f0rx_frame_t* in3,
                 f0rx_frame_t* out)
{
    int i = 0, j = 0;
    const uint8_t *src = NULL;
    uint8_t *dst = NULL;
    levels_instance_t* inst = (levels_instance_t*)instance;
    RETURN_ERROR_IF_NULL(inst);
    for (j = 0; j < inst->height; j++) {
        src = ((const uint8_t *)in1->planes[0]) + j * in1->pitches[0];
        dst =         (uint8_t*)out->planes[0]  + j * out->pitches[0];
        for (i = 0; i < inst->width; i++) {
            dst[i] = inst->lumamap[src[i]];
        }
    }
    return 0;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    f0rx_frame_t in;
    f0rx_frame_t out;
    levels_instance_t* inst = (levels_instance_t*)instance;
    RETURN_IF_NULL(inst);
    f0rx_init_frame_yuv420p(&in, (void *)inframe1, time, // XXX
                            inst->width, inst->height);
    f0rx_init_frame_yuv420p(&out, outframe, time,
                            inst->width, inst->height);
    f0rx_update2(instance, &in, NULL, NULL, &out);
    return;
}

/*
 * vim: expandtab shiftwidth=4:
 */

