/*  logo.c
 *  (C) 2011 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Tilmann Bitterberg - April 2002
 *  filter uinstates, enhancements and cleanup:
 *  Copyright (C) Sebastian Kun <seb at sarolta dot com> - March 2006
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * WARNING: the YUV support is intentionally dropped for the first
 * version of the ported filter.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "memutils.h"
#include "strutils.h"
#include "gmlib.h"

#include "frei0rx.h"
#include "frei0rx_paramext.h"
#include "frei0rx_compat.h"

/* XXX */
/* Add workaround for deprecated ScaleCharToQuantum() function */
#undef ScaleCharToQuantum
#if QuantumDepth == 8
# define ScaleCharToQuantum(x) ((x))
#elif QuantumDepth == 16
# define ScaleCharToQuantum(x) ((x)*257)
#elif QuantumDepth == 32
# define ScaleCharToQuantum(x) ((x)*16843009)
#elif QuantumDepth == 64
# define ScaleCharToQuantum(x) ((x)*71777214294589695ULL)
#else
# define ScaleCharToQuantum(x) ((x))
#warning "QuantumDepth undefined"
#endif

#define MOD_NAME            "logo"
#define DEFAULT_LOGO_FILE   "logo.png"
#define MAX_UINT8_VAL       ((uint8_t)(-1))

// basic parameter
enum POS { NONE, TOP_LEFT, TOP_RIGHT, BOT_LEFT, BOT_RIGHT, CENTER };

enum {
    RGB_BYTES        = 3, /* how much storage a pixel take? */
    RGBA_BYTES       = 4, /* ditto */
};

enum {
    PLUGIN_PARAM_NUM = 10
};


typedef struct logoinstance_ logo_instance_t;
typedef struct workitem_ work_item_t;

typedef int (*RenderLogoFn)(logo_instance_t *inst,
                            const work_item_t *W, f0rx_frame_t *frame);

struct workitem_ {
    PixelPacket *pixels;
    int         do_fade;
    float       fade_coeff;
};

typedef struct logoinstance_ {
    /* public */
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];  /* XXX */
    unsigned int framenum;
    unsigned int width;
    unsigned int height;
    char         *logo_file;
    int          logo_loaded;
    int          posx;           /* X offset in video               */
    int          posy;           /* Y offset in video               */
    enum POS     pos;            /* predifined position             */
    int          flip;           /* bool if to flip image           */
    int          rgbswap;        /* bool if swap colors             */
    unsigned int start, end;     /* ranges                          */
    unsigned int fadein;         /* No. of frames to fade in        */
    unsigned int fadeout;        /* No. of frames to fade out       */

    /* private */
    unsigned int nr_of_images;   /* animated: number of images      */
    unsigned int cur_seq;        /* animated: current image         */
    int          rgb_offset;

    /* These used to be static (per-module), but are now per-instance. */
    TCMagickContext magick;
    Image       *images;

    /* Coefficients used for transparency calculations. Pre-generating these
     * in a lookup table provides a small speed boost.
     */
    float img_coeff_lookup[MAX_UINT8_VAL + 1];
    float vid_coeff_lookup[MAX_UINT8_VAL + 1];

    RenderLogoFn  render;
} logo_instance_t;


#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)

#define RETURN_IF_NOT_OK(INST, RET) do { \
    if ((RET) != 0) { \
        (INST)->logo_loaded = 1; \
        return; \
    } \
} while (0)


/* ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove


/*************************************************************************/

static int flogo_compute_position(logo_instance_t *inst)
{
    switch (inst->pos) {
      case NONE: /* 0 */
        break;
      case TOP_LEFT:
        inst->posx = 0;
        inst->posy = inst->rgb_offset;
        break;
      case TOP_RIGHT:
        inst->posx = inst->width  - inst->magick.image->columns;
        break;
      case BOT_LEFT:
        inst->posy = inst->height - inst->magick.image->rows 
                 - inst->rgb_offset;
        break;
      case BOT_RIGHT:
        inst->posx = inst->width  - inst->magick.image->columns;
        inst->posy = inst->height - inst->magick.image->rows
                 - inst->rgb_offset;
        break;
      case CENTER:
        inst->posx = (inst->width - inst->magick.image->columns)/2;
        inst->posy = (inst->height- inst->magick.image->rows)/2;
        /* align to not cause color disruption */
        if (inst->posx & 1)
            inst->posx++;
        if (inst->posy & 1)
            inst->posy++;
        break;
    }

    if (inst->posy < 0 || inst->posx < 0
     || (inst->posx + inst->magick.image->columns) > inst->width
     || (inst->posy + inst->magick.image->rows)    > inst->height) {
        fprintf(stderr, "[%s] invalid position\n", MOD_NAME);
        return -1;
    }
    return 0;
}

static int flogo_calc_coeff(logo_instance_t *inst)
{
    /* Set up image/video coefficient lookup tables */
    int i;
    float maxrgbval = (float)MaxRGB; // from GraphicsMagick

    for (i = 0; i <= MAX_UINT8_VAL; i++) {
        float x = (float)ScaleCharToQuantum(i);
        /* Alternatively:
         *  img_coeff = (maxrgbval - x) / maxrgbval;
         *  vid_coeff = x / maxrgbval;
         */
        inst->img_coeff_lookup[i] = 1.0 - (x / maxrgbval);
        inst->vid_coeff_lookup[i] = 1.0 - inst->img_coeff_lookup[i];
    }
    return 0;
}

static void set_fade(work_item_t *W, int id, const logo_instance_t *inst)
{
    if (id - inst->start < inst->fadein) {
        // fading-in
        W->fade_coeff = (float)(inst->start - id + inst->fadein) / (float)(inst->fadein);
        W->do_fade    = 1;
    } else if (inst->end - id < inst->fadeout) {
        // fading-out
        W->fade_coeff = (float)(id - inst->end + inst->fadeout) / (float)(inst->fadeout);
        W->do_fade    = 1;
    } else {
        /* enforce defaults */
        W->fade_coeff = 0.0;
        W->do_fade    = 0;
    }
}

static int load_images(logo_instance_t *inst)
{
    Image         *timg;
    Image         *nimg;

    inst->images = GetFirstImageInList(inst->magick.image);
    nimg = NewImageList();

    while (inst->images != (Image *)NULL) {
        if (inst->flip) {
            timg = FlipImage(inst->images, 
                             &inst->magick.exception_info);
            if (timg == NULL) {
                CatchException(&inst->magick.exception_info);
                return -1;
            }
            AppendImageToList(&nimg, timg);
        }

        inst->images = GetNextImageInList(inst->images);
        inst->nr_of_images++;
    }

    // check for memleaks;
    //DestroyImageList(image);
    if (inst->flip) {
        /* DANGEROUS!!! */
        inst->magick.image = nimg;
    }

    /* for running through image sequence */
    /* DANGEROUS!!! */
    inst->images = inst->magick.image;

    return 0;
}

static int sanity_check(logo_instance_t *inst,
                        const char *logo_file)
{
    if (inst->magick.image->columns > inst->width
     || inst->magick.image->rows    > inst->height) {
        fprintf(stderr, "[%s] \"%s\" is too large\n",
                MOD_NAME, logo_file);
        return -1;
    }
    return 0;
}

static int setup_logo_rgb(logo_instance_t *inst)
{
    /* for RGB format is origin bottom left */
    /* for RGB, rgbswap is done in the frame routine */
    inst->rgb_offset = inst->height - inst->magick.image->rows;
    inst->posy       = inst->rgb_offset   - inst->posy;

    return 0;
}

static void logo_setup(f0r_instance_t instance, int unused)
{
    logo_instance_t* inst = (logo_instance_t*)instance;
    int ret = 0;

    ret = flogo_compute_position(inst);
    RETURN_IF_NOT_OK(inst, ret);

    setup_logo_rgb(inst);

    inst->logo_loaded = 1;
    return;
}

static void logo_load_file(f0r_instance_t instance, int unused)
{
    logo_instance_t* inst = (logo_instance_t*)instance;
    int ret = 0;

    ret = tc_magick_filein(&inst->magick, inst->logo_file);
    RETURN_IF_NOT_OK(inst, ret);

    ret = sanity_check(inst, inst->logo_file);
    RETURN_IF_NOT_OK(inst, ret);

    ret = load_images(inst);
    RETURN_IF_NOT_OK(inst, ret);

    logo_setup(instance, unused);
}

static int render_logo_rgb(logo_instance_t *inst,
                             const work_item_t *W, f0rx_frame_t *frame)
{
    int row, col;
    unsigned long r_off = 0, g_off = 1, b_off = 2;
    float img_coeff, vid_coeff;
    /* Note: GraphicsMagick defines opacity = 0 as fully visible, and
     * opacity = MaxRGB as fully transparent.
     */
    Quantum opacity;
    uint8_t *video_buf = NULL;
    PixelPacket *pixels = GetImagePixels(inst->images, 0, 0,
                                         inst->images->columns,
                                         inst->images->rows);

    if (inst->rgbswap) {
        r_off = 2;
        b_off = 0;
    }

    for (row = 0; row < inst->magick.image->rows; row++) {
        video_buf = (uint8_t*)frame->planes[0] + RGBA_BYTES * ((row + inst->posy) * inst->width + inst->posx);

        for (col = 0; col < inst->magick.image->columns; col++) {
            opacity = pixels->opacity;

            if (W->do_fade)
                opacity += (Quantum)((MaxRGB - opacity) * W->fade_coeff);

            if (opacity == 0) {
                *(video_buf + r_off) = ScaleQuantumToChar(pixels->red);
                *(video_buf + g_off) = ScaleQuantumToChar(pixels->green);
                *(video_buf + b_off) = ScaleQuantumToChar(pixels->blue);
            } else if (opacity < MaxRGB) {
                uint8_t opacity_byte = ScaleQuantumToChar(opacity);
                img_coeff = inst->img_coeff_lookup[opacity_byte];
                vid_coeff = inst->vid_coeff_lookup[opacity_byte];

                *(video_buf + r_off) = (uint8_t)((*(video_buf + r_off)) * vid_coeff)
                                        + (uint8_t)(ScaleQuantumToChar(pixels->red)   * img_coeff);
                *(video_buf + g_off) = (uint8_t)((*(video_buf + g_off)) * vid_coeff)
                                        + (uint8_t)(ScaleQuantumToChar(pixels->green) * img_coeff);
                *(video_buf + b_off) = (uint8_t)((*(video_buf + b_off)) * vid_coeff)
                                        + (uint8_t)(ScaleQuantumToChar(pixels->blue)  * img_coeff);
            }

            video_buf += RGBA_BYTES;
            pixels++;
        }
    }
    return 0;
}

/*************************************************************************/

int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    int ready = 0;
    logo_instance_t* inst = (logo_instance_t*)instance;
    if (inst) {
        ready = inst->logo_loaded; /* FIXME: that's it? */
    }
    return ready;
}


int f0r_init()
{
    /* intialization deferred into f0r_construct */
    return 1;
}

void f0r_deinit()
{
    /* deintialization deferred into f0r_destruct */
    return;
}

void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "logo";
    info->author = "Tilmann Bitterberg, Sebastian Kun";;
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0R_COLOR_MODEL_RGBA8888;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"This filter renders an user specified image into the video."
" Any image format GraphicsMagick can read is accepted."
" Transparent images are also supported."
" Image origin is at the very top left."
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_VIDEO;
    info->sample_model = 0;
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "File";
        info->type = F0R_PARAM_STRING;
        info->explanation = "The logo image filename.";
        break;
      case 1:
        info->name = "Position X";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The X position of the logo.";
        break;
      case 2:
        info->name = "Position Y";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The Y position of the logo.";
        break;
      case 3:
        info->name = "Position Definition";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "The logo position definition"
                            " (0=None, 1=TopL, 2=TopR, 3=BotL, 4=BotR, 5=Center).";
      case 4:
        info->name = "Start";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "First frame in range to render logo in.";
        break;
      case 5:
        info->name = "End";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Last frame in range to render logo in.";
        break;
      case 6:
        info->name = "Fade In";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "First frame to fade in the logo.";
        break;
      case 7:
        info->name = "Fade Out";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Last frame to fade out the logo.";
        break;
      case 8:
        info->name = "Flip";
        info->type = F0R_PARAM_BOOL;
        info->explanation = "Mirror image.";
        break;
      case 9:
        info->name = "RGB Swap";
        info->type = F0R_PARAM_BOOL;
        info->explanation = "Swap G/B colors.";
        break;
      default:
        break; // XXX
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    logo_instance_t* inst = (logo_instance_t*)instance;
    return &inst->attrs[param_index];
}

static int f0rx_logo_init(logo_instance_t *inst,
                          unsigned int param1, unsigned int param2)
{
    int err = 0;

    inst->end       = (unsigned int)-1;
    inst->render    = render_logo_rgb;
    inst->logo_file = tc_strdup(DEFAULT_LOGO_FILE);

    flogo_calc_coeff(inst);
    
    err = tc_magick_init(&inst->magick, GMLIB_QUALITY_DEFAULT);
    if (!err) {
        logo_load_file(inst, 0);
    }
    return err;
}

f0r_instance_t f0rx_construct(int format,
                              unsigned int param1, unsigned int param2)
{
    logo_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        int err = f0rx_logo_init(inst, param1, param2);
        if (!err) {
            f0rx_bind_param_attr(inst, 0, &inst->logo_file);
            f0rx_bind_param_attr_remap_int(inst,  1, &inst->posx);
            f0rx_bind_param_attr_remap_int(inst,  2, &inst->posy);
            f0rx_bind_param_attr_remap_int(inst,  3, &inst->pos);
            f0rx_bind_param_attr_remap_int(inst,  4, &inst->start);
            f0rx_bind_param_attr_remap_int(inst,  5, &inst->end);
            f0rx_bind_param_attr_remap_int(inst,  6, &inst->fadein);
            f0rx_bind_param_attr_remap_int(inst,  7, &inst->fadeout);
            f0rx_bind_param_attr_remap_int(inst,  8, &inst->flip);
            f0rx_bind_param_attr_remap_int(inst,  9, &inst->rgbswap);

            f0rx_hook_param_attr_post_set(inst, 0, logo_load_file);
            f0rx_hook_param_attr_post_set(inst, 1, logo_setup);
            f0rx_hook_param_attr_post_set(inst, 2, logo_setup);
            f0rx_hook_param_attr_post_set(inst, 3, logo_setup);
            f0rx_hook_param_attr_post_set(inst, 8, logo_load_file);
        } else {
            f0r_destruct(inst);
            inst = NULL;
        }
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0R_COLOR_MODEL_RGBA8888, width, height);
}


void f0r_destruct(f0r_instance_t instance)
{
    logo_instance_t* inst = (logo_instance_t*)instance;
    RETURN_IF_NULL(inst);
    tc_magick_fini(&inst->magick);
    free(inst->logo_file);
    free(instance);
    return;
}

int f0rx_update2(f0r_instance_t instance,
                 const f0rx_frame_t* in1,
                 const f0rx_frame_t* in2,
                 const f0rx_frame_t* in3,
                 f0rx_frame_t* out)
{
    int ret = 0;
    logo_instance_t* inst = (logo_instance_t*)instance;
    RETURN_ERROR_IF_NULL(inst);
    ac_memcpy(out->planes[0], in1->planes[0], out->pitches[0]);
    if (inst->framenum >= inst->start && inst->framenum <= inst->end) {
        work_item_t W = { NULL, 0, 0.0 };
        set_fade(&W, inst->framenum, inst);
        W.pixels = GetImagePixels(inst->images, 0, 0,
                                  inst->images->columns,
                                  inst->images->rows);
        ret = inst->render(inst, &W, out);
    }
    inst->framenum++;
    return ret;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    f0rx_frame_t in;
    f0rx_frame_t out;
    logo_instance_t* inst = (logo_instance_t*)instance;
    RETURN_IF_NULL(inst);
    memset(&in, 0, sizeof(in));
    memset(&out, 0, sizeof(out));
    in.planes[0] = (void *)inframe1;
    in.pitches[0] = RGBA_BYTES * inst->width * inst->height;
    in.time.d = time;
    out.planes[0] = (void *)outframe;
    out.pitches[0] = RGBA_BYTES * inst->width * inst->height;
    out.time.d = time;
    f0rx_update2(instance, &in, NULL, NULL, &out);
    return;
}

/*
 * vim: expandtab shiftwidth=4:
 */

