/*  smooth.c
 *  (C) 2012-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) Chad Page - October 2002
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "memutils.h"

#include "frei0rx.h"
#include "frei0rx_compat.h"
#include "frei0rx_paramext.h"
#define ENABLE_FREI0RX_FRAME_ALLOC 1
#include "frei0rx_frame.h"

enum {
    PLUGIN_PARAM_NUM = 4
};


typedef struct smoothinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
    int cdiff;
    int ldiff;
    int range;
    double strength;
    f0rx_frame_t buf;
} smooth_instance_t;

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)



/* ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove

/*************************************************************************/
/* core functions                                                        */
/*************************************************************************/

#define PTR(P) ((uint8_t*)(P))

static int smooth_yuv(smooth_instance_t *inst,
                      const f0rx_frame_t* in, f0rx_frame_t* out)
{
    int x, y, pu, cpu, cdiff;
    int xa, ya, oval, ldiff;
    float dist, ratio, nval;

    /* FIXME: here is assumed width == pitch (aka stride) */
    /* FIXME: here is assumed the planes are layed out consecutivally */
    /* First pass - horizontal */
    for (y = 0; y < inst->buf.P2.height; y++) {
        for (x = 0; x < inst->buf.P1.width; x++) {
            pu = ((y * inst->buf.P1.width) / 2) + (x / 2);
            oval = PTR(in->planes[0])[x + (y * inst->buf.P1.width)]; /* FIXME */
            nval = ((float)oval);
            for (xa = x - inst->range;
                 (xa <= (x + inst->range)) && (xa < inst->buf.P1.width);
                 xa++) {
                if (xa < 0)
                    xa = 0;
                if (xa == x)
                    xa++;
                cpu = ((y * inst->buf.P1.width) / 2) + (xa / 2);
                cdiff  = abs(PTR(in->planes[2])[pu] - PTR(in->planes[2])[cpu]);
                       + abs(PTR(in->planes[1])[pu] - PTR(in->planes[1])[cpu]);

                /* If color difference not too great, average the pixel according to distance */
                ldiff = abs(PTR(in->planes[0])[xa + (y * inst->buf.P1.width)] - oval);
                if ((cdiff < inst->cdiff) && (ldiff < inst->ldiff)) {
                    dist = abs(xa - x);
                    ratio = inst->strength / dist;
                    nval = nval * (1 - ratio);
                    nval += ((float)PTR(in->planes[0])[xa + (y * inst->buf.P1.width)]) * ratio;
                }
            }
            PTR(inst->buf.planes[0])[x + (y * inst->buf.P1.width)] = (uint8_t)(nval + 0.5);
        }
    }

    /* Second pass - vertical lines */
    for (y = 0; y < inst->buf.P2.height; y++) {
        for (x = 0; x < inst->buf.P1.width; x++) {
            pu = ((y * inst->buf.P1.width) / 2) + (x / 2);
            oval = PTR(inst->buf.planes[0])[x + (y * inst->buf.P1.width)]; /* FIXME */
            nval = ((float)oval);
            for (ya = y - inst->range;
                 (ya <= (y + inst->range)) && (ya < inst->buf.P2.height);
                 ya++) {
                if (ya < 0)
                    ya = 0;
                if (ya == y)
                    ya++;
                cpu = ((ya * inst->buf.P1.width) / 2) + (x / 2);
                cdiff = abs(PTR(inst->buf.planes[2])[pu] - PTR(inst->buf.planes[2])[cpu]);
                      + abs(PTR(inst->buf.planes[1])[pu] - PTR(inst->buf.planes[1])[cpu]);

                /* If color difference not too great, average the pixel according to distance */
                ldiff = abs(PTR(inst->buf.planes[0])[x + (ya * inst->buf.P1.width)] - oval);
                if ((cdiff < inst->cdiff) && (ldiff < inst->ldiff)) {
                    dist = abs(ya - y);
                    ratio = inst->strength / dist;
                    nval = nval * (1 - ratio);
                    nval += ((float)PTR(inst->buf.planes[0])[x + (ya * inst->buf.P1.width)]) * ratio;
                }
            }
            PTR(out->planes[0])[x + (y * inst->buf.P1.width)] = (uint8_t)(nval + 0.5);
        }
    }
    return 0;
}


/*************************************************************************/
/* the frei0r API                                                        */
/*************************************************************************/

int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    return 1; /* parameters autocorrection. */
}


int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}


void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "smooth";
    info->author = "Chad Page";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0RX_COLOR_MODEL_PLANAR_YUV420P;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"(single-frame) smoothing plugin"
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_VIDEO;
    info->sample_model = 0;
}

void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Strength";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Blending factor.";
        break;
      case 1:
        info->name = "Chroma Difference";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Max difference in chroma values.";
        break;
      case 2:
        info->name = "Luma Difference";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Max difference in luma values.";
        break;
      case 3:
        info->name = "Range";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Search range in pixels.";
        break;
      default:
        break; // XXX
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    smooth_instance_t* inst = (smooth_instance_t*)instance;
    return &inst->attrs[param_index];
}

static void f0rx_instance_defaults(smooth_instance_t* inst)
{
    if (inst) {
        inst->strength = 0.25; /* Blending factor.  Do not exceed 2 ever */
        inst->cdiff    = 6;    /* Max difference in UV values */
        inst->ldiff    = 8;    /* Max difference in Y value */
        inst->range    = 4;    /* Search range */
    }
    return;
}


static void post_set_param(f0r_instance_t instance, int param_index)
{
    smooth_instance_t* inst = (smooth_instance_t*)instance;
    if (inst->strength > 0.9) {
        inst->strength = 0.9;
    }
    return;
}

f0r_instance_t f0rx_construct(int format,
                              unsigned int param1, unsigned int param2)
{
    smooth_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        int err = f0rx_alloc_frame_yuv420p(&inst->buf, param1, param2);
        if (!err) {
            f0rx_instance_defaults(inst);
            f0rx_bind_param_attr_remap_int(inst, 0, &inst->strength);
            f0rx_bind_param_attr_remap_int(inst, 1, &inst->cdiff);
            f0rx_bind_param_attr_remap_int(inst, 2, &inst->ldiff);
            f0rx_bind_param_attr_remap_int(inst, 3, &inst->range);
            f0rx_hook_param_attr_post_set(inst,  0, post_set_param);
        } else {
            /* FIXME: log it */
            free(inst);
            inst = NULL;
        }
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0RX_COLOR_MODEL_PLANAR_YUV420P, width, height);
}


void f0r_destruct(f0r_instance_t instance)
{
    smooth_instance_t* inst = (smooth_instance_t*)instance;
    RETURN_IF_NULL(inst);
    f0rx_dispose_frame_yuv420p(&inst->buf);
    tc_free(instance);
    return;
}

int f0rx_update2(f0r_instance_t instance,
                 const f0rx_frame_t* in1,
                 const f0rx_frame_t* in2,
                 const f0rx_frame_t* in3,
                 f0rx_frame_t* out)
{
    smooth_instance_t* inst = (smooth_instance_t*)instance;
    RETURN_ERROR_IF_NULL(inst);
    return smooth_yuv(inst, in1, out);
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    f0rx_frame_t in;
    f0rx_frame_t out;
    smooth_instance_t* inst = (smooth_instance_t*)instance;
    RETURN_IF_NULL(inst);
    f0rx_init_frame_yuv420p(&in, (void *)inframe1, time, // XXX
                            inst->buf.P1.width, inst->buf.P2.height);
    f0rx_init_frame_yuv420p(&out, outframe, time,
                            inst->buf.P1.width, inst->buf.P2.height);
    f0rx_update2(instance, &in, NULL, NULL, &out);
    return;
}

/*
 * vim: expandtab shiftwidth=4:
 */

