/*  unsharp.c
 *  (C) 2012-2014 Francesco Romani (fromani a gmail d com)
 *  this file is a frei0r(x) plugin based on
 *
 *  Copyright (C) 2002 Remi Guyomarch <rguyom@pobox.com>
 * 
 *  This file is part of transcode, a video stream processing tool
 *
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "memutils.h"

#include "frei0rx.h"
#include "frei0rx_frame.h"
#include "frei0rx_paramext.h"
#include "frei0rx_compat.h"


#define TC_MAX(a, b)		(((a) > (b)) ?(a) :(b))
#define TC_MIN(a, b)		(((a) < (b)) ?(a) :(b))
/* clamp x between a and b */
#define TC_CLAMP(x, a, b)	TC_MIN(TC_MAX((a), (x)), (b))

/* ugly hack until I fix out some aclib related issues. */
#define ac_memcpy memmove


enum {
    MIN_MATRIX_SIZE  = 3,
    MAX_MATRIX_SIZE  = 63,
    PLUGIN_PARAM_NUM = 6
};

typedef struct FilterParam {
    int msizeX, msizeY;
    double amount;
    uint32_t *SC[MAX_MATRIX_SIZE-1];
} FilterParam;


typedef struct unsharpinstance_ {
    f0rx_param_attr_t attrs[PLUGIN_PARAM_NUM];
    FilterParam lumaParam;
    FilterParam chromaParam;
    unsigned int width;
    unsigned int height;
} unsharp_instance_t;

#define RETURN_IF_NULL(inst) do { \
    if (!(inst)) { \
        return; \
    } \
} while (0)

#define RETURN_ERROR_IF_NULL(INST) do { \
    if (!(INST)) { \
        return -1; \
    } \
} while (0)


enum {
    RGB_BYTES             = 3, /* how much storage a pixel take? */
    RGBA_BYTES            = 4, /* ditto */
    PIXEL_BLACK_THRESHOLD = 24
};

/* This code is based on :

An Efficient algorithm for Gaussian blur using finite-state machines
Frederick M. Waltz and John W. V. Miller

SPIE Conf. on Machine Vision Systems for Inspection and Metrology VII
Originally published Boston, Nov 98

*/

static void unsharp( uint8_t *dst, uint8_t *src, int dstStride, int srcStride, int width, int height, FilterParam *fp ) {

    uint32_t **SC = fp->SC;
    uint32_t SR[MAX_MATRIX_SIZE-1], Tmp1, Tmp2;
    uint8_t* src2 = src; // avoid gcc warning

    int32_t res;
    int x, y, z;
    int amount = fp->amount * 65536.0;
    int stepsX = fp->msizeX/2;
    int stepsY = fp->msizeY/2;
    int scalebits = (stepsX+stepsY)*2;
    int32_t halfscale = 1 << ((stepsX+stepsY)*2-1);

    if( !fp->amount ) {
	if( src == dst )
	    return;
	if( dstStride == srcStride )
	    ac_memcpy( dst, src, srcStride*height );
	else
	    for( y=0; y<height; y++, dst+=dstStride, src+=srcStride )
		ac_memcpy( dst, src, width );
	return;
    }

    for( y=0; y<2*stepsY; y++ )
	memset( SC[y], 0, sizeof(SC[y][0]) * (width+2*stepsX) );

    for( y=-stepsY; y<height+stepsY; y++ ) {
	if( y < height ) src2 = src;
	memset( SR, 0, sizeof(SR[0]) * (2*stepsX-1) );
	for( x=-stepsX; x<width+stepsX; x++ ) {
	    Tmp1 = x<=0 ? src2[0] : x>=width ? src2[width-1] : src2[x];
	    for( z=0; z<stepsX*2; z+=2 ) {
		Tmp2 = SR[z+0] + Tmp1; SR[z+0] = Tmp1;
		Tmp1 = SR[z+1] + Tmp2; SR[z+1] = Tmp2;
	    }
	    for( z=0; z<stepsY*2; z+=2 ) {
		Tmp2 = SC[z+0][x+stepsX] + Tmp1; SC[z+0][x+stepsX] = Tmp1;
		Tmp1 = SC[z+1][x+stepsX] + Tmp2; SC[z+1][x+stepsX] = Tmp2;
	    }
	    if( x>=stepsX && y>=stepsY ) {
		uint8_t* srx = src - stepsY*srcStride + x - stepsX;
		uint8_t* dsx = dst - stepsY*dstStride + x - stepsX;

		res = (int32_t)*srx + ( ( ( (int32_t)*srx - (int32_t)((Tmp1+halfscale) >> scalebits) ) * amount ) >> 16 );
		*dsx = res>255 ? 255 : res<0 ? 0 : (uint8_t)res;
	    }
	}
	if( y >= 0 ) {
	    dst += dstStride;
	    src += srcStride;
	}
    }
}

static int alloc_buffers(FilterParam *fp, int width)
{
    int err = 0, z = 0;
    int stepsX = fp->msizeX / 2;
    int stepsY = fp->msizeY / 2;
    memset(fp->SC, 0, sizeof(fp->SC));
    for (z = 0; !err && z < 2*stepsY; z++) {
        fp->SC[z] = tc_bufalloc(sizeof(*(fp->SC[z])) * (width + 2 * stepsX));
        if (fp->SC[z] == NULL) {
            err = -1;
        }
    }
    return err;
}

static void free_buffers(FilterParam *fp)
{
    int z = 0;
    for (z = 0; z < sizeof(fp->SC)/sizeof(fp->SC[0]); z++) {
        if (fp->SC[z]) {
            tc_buffree(fp->SC[z]);
	        fp->SC[z] = NULL;
        }
    }
    return;
}

static void post_set_param(f0r_instance_t instance, int param_index)
{
    unsharp_instance_t* inst = (unsharp_instance_t*)instance;
    RETURN_IF_NULL(inst);
    if (param_index == 2 || param_index == 3) { /* Luma Matrix */
        inst->lumaParam.msizeX   = 1 | TC_CLAMP(inst->lumaParam.msizeX, MIN_MATRIX_SIZE, MAX_MATRIX_SIZE);
        inst->lumaParam.msizeY   = 1 | TC_CLAMP(inst->lumaParam.msizeY, MIN_MATRIX_SIZE, MAX_MATRIX_SIZE);
        free_buffers(&inst->lumaParam);
        alloc_buffers(&inst->lumaParam, inst->width);
    }
    if (param_index == 4 || param_index == 5) {
        inst->chromaParam.msizeX = 1 | TC_CLAMP(inst->chromaParam.msizeX, MIN_MATRIX_SIZE, MAX_MATRIX_SIZE);
        inst->chromaParam.msizeY = 1 | TC_CLAMP(inst->chromaParam.msizeY, MIN_MATRIX_SIZE, MAX_MATRIX_SIZE);
        free_buffers(&inst->chromaParam);
        alloc_buffers(&inst->chromaParam, inst->width);
    }
    return;
}



/*************************************************************************/

int f0r_init()
{
    /* no initialization required */
    return 1;
}

void f0r_deinit()
{
    /* no initialization required */
    return;
}


void f0r_get_plugin_info(f0r_plugin_info_t* info)
{
    info->name = "unsharp";
    info->author = "Remi Guyomarch";
    info->plugin_type = F0R_PLUGIN_TYPE_FILTER;
    info->color_model = F0RX_COLOR_MODEL_PLANAR_YUV420P;
    info->frei0r_version = FREI0R_MAJOR_VERSION;
    info->major_version = 0;
    info->minor_version = 1;
    info->num_params =  PLUGIN_PARAM_NUM;
    info->explanation = ""
"This filter blurs or sharpens an image depending on"
" the sign of \"amount\". You can either set amount for"
" both luma and chroma or you can set it individually"
" (recommended). A positive value for amount will sharpen"
" the image, a negative value will blur it. A sane range"
" for amount is -1.5 to 1.5."
" The matrix sizes must be odd and define the"
" range/strength of the effect. Sensible ranges are 3x3"
" to 7x7."
" It sometimes makes sense to sharpen the sharpen the"
" luma and to blur the chroma."
"";
}

void f0rx_get_plugin_info(f0rx_plugin_info_t* info)
{
    f0r_get_plugin_info(&(info->base));
    info->content_type = F0RX_CONTENT_TYPE_VIDEO;
    info->sample_model = 0;
}


void f0r_get_param_info(f0r_param_info_t* info, int param_index)
{
    switch (param_index) {
      case 0:
        info->name = "Luma";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Luma (un)sharpness amount.";
        break;
      case 1:
        info->name = "Chroma";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Chroma (un)sharpness amount.";
        break;
      case 2:
        info->name = "X Luma Matrix";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Luma search matrix size: X dimension.";
        break;
      case 3:
        info->name = "Y Luma Matrix";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Luma search matrix size: Y dimension.";
        break;
      case 4:
        info->name = "X Chroma Matrix";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Chroma search matrix size: X dimension.";
        break;
      case 5:
        info->name = "Y Chroma Matrix";
        info->type = F0R_PARAM_DOUBLE;
        info->explanation = "Chroma search matrix size: Y dimension.";
        break;
      default:
        break; // XXX
    }
    return;
}

static f0rx_param_attr_t *f0rx_get_param_attr(f0r_instance_t instance,
                                              int param_index)
{
    unsharp_instance_t* inst = (unsharp_instance_t*)instance;
    return &inst->attrs[param_index];
}


f0r_instance_t f0rx_construct(int format,
                              unsigned int param1, unsigned int param2)
{
    unsharp_instance_t* inst = tc_zalloc(sizeof(*inst));
    if (inst) {
        inst->width = param1;
        inst->height = param2;
        f0rx_bind_param_attr(          inst, 0, &inst->lumaParam.amount);
        f0rx_bind_param_attr(          inst, 1, &inst->chromaParam.amount);
        f0rx_bind_param_attr_remap_int(inst, 2, &inst->lumaParam.msizeX);
        f0rx_bind_param_attr_remap_int(inst, 3, &inst->lumaParam.msizeY);
        f0rx_bind_param_attr_remap_int(inst, 4, &inst->chromaParam.msizeX);
        f0rx_bind_param_attr_remap_int(inst, 5, &inst->chromaParam.msizeY);

        f0rx_hook_param_attr_post_set(inst, F0RX_PARAM_ALL, post_set_param);
    }
    return (f0r_instance_t)inst;
}

f0r_instance_t f0r_construct(unsigned int width, unsigned int height)
{
    return f0rx_construct(F0RX_COLOR_MODEL_PLANAR_YUV420P, width, height);
}


void f0r_destruct(f0r_instance_t instance)
{
    unsharp_instance_t* inst = (unsharp_instance_t*)instance;
    RETURN_IF_NULL(inst);
    free_buffers(&inst->lumaParam);
    free_buffers(&inst->chromaParam);
    free(instance);
    return;
}

int f0rx_is_ready(f0r_instance_t instance, int param_index)
{
    int ready = 0;
    unsharp_instance_t* inst = (unsharp_instance_t*)instance;
    if (inst) {
        ready = (inst->lumaParam.msizeX || inst->chromaParam.msizeX);
        // FIXME buffers check
    }
    return ready;
}

int f0rx_update2(f0r_instance_t instance,
                 const f0rx_frame_t* in1,
                 const f0rx_frame_t* in2,
                 const f0rx_frame_t* in3,
                 f0rx_frame_t* out)
{
    unsharp_instance_t* inst = (unsharp_instance_t*)instance;
    RETURN_ERROR_IF_NULL(inst);
    if (f0rx_is_ready(instance, -1)) {
        unsharp(out->planes[0], in1->planes[0], out->P1.width, in1->P1.width,
                out->P1.width, out->P2.height,
                &inst->lumaParam);
        unsharp(out->planes[1], in1->planes[1], out->P1.width, in1->P1.width,
                out->P1.width, out->P2.height,
                &inst->chromaParam);
        unsharp(out->planes[2], in1->planes[2], out->P1.width, in1->P1.width,
                out->P1.width, out->P2.height,
                &inst->chromaParam);
    }
    return 0;
}

void f0r_update2(f0r_instance_t instance, double time,
                 const uint32_t* inframe1,
                 const uint32_t* inframe2,
                 const uint32_t* inframe3,
                 uint32_t* outframe)
{
    f0rx_frame_t in;
    f0rx_frame_t out;
    unsharp_instance_t* inst = (unsharp_instance_t*)instance;
    RETURN_IF_NULL(inst);
    f0rx_init_frame_yuv420p(&in, (void *)inframe1, time, // XXX
                            inst->width, inst->height);
    f0rx_init_frame_yuv420p(&out, outframe, time,
                            inst->width, inst->height);
    f0rx_update2(instance, &in, NULL, NULL, &out);
    return;
}


/*
 * vim: expandtab shiftwidth=4:
 */

