/*
 * inspect0r.c- tycho-plugins debug tool:
 *              provides meta data about a frei0r plugin.
 * (C) 2011-2014 Francesco Romani <fromani at gmail dot com>
 * license: ZLIB.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "logkit.h"

#include "l0ader.h"


#define EXE "inspect0r"

enum {
    STATUS_OK = 0,
    STATUS_BAD_PARAM,
    STATUS_NO_MODULE,
    STATUS_MODULE_ERROR,
    STATUS_BAD_MODULES,
    STATUS_MODULE_FAILED,
    STATUS_LOADER_ERROR
};

void version(void)
{
    fprintf(stderr, "%s (%s v%s) (C) 2011-2014 Francesco Romani\n",
            EXE, PACKAGE, VERSION);
}

static void usage(void)
{
    version();
    fprintf(stderr, "Usage: %s [options] -i name\n", EXE);
    fprintf(stderr, "    -i name       Plugin name information (like \'smooth\')\n");
    fprintf(stderr, "    -v vendor     Use 'vendor' as module vendor [none]\n");
    fprintf(stderr, "    -d verbosity  Verbosity mode [1]\n");
}

static const char *content_type_to_string(int content_type)
{
    const char *str = "unknown";
    switch (content_type) {
      case F0RX_CONTENT_TYPE_VIDEO:
        str = "video";
        break;
      case F0RX_CONTENT_TYPE_AUDIO:
        str = "audio";
        break;
      default:
        str = "unknown";
        break;
    }
    return str;
}


static const char *color_model_to_string(int color_model)
{
    const char *str = "unknown";
    switch (color_model) {
      case F0R_COLOR_MODEL_BGRA8888:
        str = "BGRA";
        break;
      case F0R_COLOR_MODEL_RGBA8888:
        str = "RGBA";
        break;
      case F0R_COLOR_MODEL_PACKED32:
        str = "PACKED32";
        break;
      case F0RX_COLOR_MODEL_PLANAR_YUV420P:
        str= "YUV420P";
        break;
      default:
        str = "unknown";
        break;
    }
    return str;
}

static const char *plugin_type_to_string(int plugin_type)
{
    const char *str = "unknown";
    switch (plugin_type) {
      case F0R_PLUGIN_TYPE_FILTER:
        str = "filter";
        break;
      case F0R_PLUGIN_TYPE_SOURCE:
        str = "source";
        break;
      case F0R_PLUGIN_TYPE_MIXER2:
        str = "two-in mixer";
        break;
      case F0R_PLUGIN_TYPE_MIXER3:
        str = "three-in mixer";
        break;
      default:
        str = "unknown";
        break;
    }
    return str;
}

static const char *param_type_to_string(int param_type)
{
    const char *str = "unknown";
    switch (param_type) {
      case F0R_PARAM_BOOL:
        str = "bool";
        break;
      case F0R_PARAM_DOUBLE:
        str = "double";
        break;
      case F0R_PARAM_COLOR:
        str = "color";
        break;
      case F0R_PARAM_POSITION:
        str = "position";
        break;
      case F0R_PARAM_STRING:
        str = "string";
        break;
      default:
        str = "unknown";
        break;
    }
    return str;
}

static void print_plugin_param(int idx, const f0r_param_info_t *param)
{
    printf("  #%02i (%s): %s\n", idx,
           param_type_to_string(param->type),
           param->name);
    printf("    %s\n", param->explanation);
}

static void print_plugin_meta_data(const f0rx_plugin_meta_data_t *meta_data)
{
    int j = 0;
    printf("plugin: %s\n", meta_data->info.base.name);
    if (meta_data->info.base.explanation) {
        printf(" description  : %s\n", meta_data->info.base.explanation);
    }
    printf(" author       : %s\n", meta_data->info.base.author);
    printf(" version      : %i.%i\n",
               meta_data->info.base.major_version,
               meta_data->info.base.minor_version);
    printf(" API version  : %i%s\n",
           meta_data->info.base.frei0r_version,
           (meta_data->extended) ?" (extended)" :"");
    printf(" plugin type  : %s\n",
           plugin_type_to_string(meta_data->info.base.plugin_type));
    printf(" content type : %s\n",
            content_type_to_string(meta_data->info.content_type));
    if (meta_data->info.content_type == F0RX_CONTENT_TYPE_VIDEO) {
        printf(" color model  : %s\n",
               color_model_to_string(meta_data->info.base.color_model));
    } else if (meta_data->info.content_type == F0RX_CONTENT_TYPE_AUDIO) {
        // TODO: sample_model
    }
    printf(" parameters   :\n");
    for (j =0; j < meta_data->info.base.num_params; j++) {
        print_plugin_param(j, &(meta_data->params[j]));
    }
    return;
}

#define VALIDATE_OPTARG(optarg) do { \
    if (!(optarg) || (optarg)[0] == '-') { \
        usage(); \
        return STATUS_BAD_PARAM; \
    } \
} while (0)


int main(int argc, char *argv[])
{
    int ch;
    int verbosity = CX_LOG_DEBUG;
    int error = 0;
    const char *modname = NULL;
    const char *modvendor = "";

    f0rx_plugin_meta_data_t meta_data;
    f0rx_loader_t loader = NULL;
    CX_LogContext *logger = NULL;

    if (argc == 1) {
        usage();
        return STATUS_BAD_PARAM;
    }

    while (1) {
        ch = getopt(argc, argv, "d:i:v:Vh?");
        if (ch == -1) {
            break;
        }

        switch (ch) {
          case 'd':
            VALIDATE_OPTARG(optarg);
            verbosity = atoi(optarg);
            break;
          case 'i':
            VALIDATE_OPTARG(optarg);
            modname = optarg;
            break;
          case 'v':
            VALIDATE_OPTARG(optarg);
            modvendor = optarg;
            break;
          case 'V':
            version();
            return STATUS_OK;
          case '?': /* fallthrough */
          case 'h': /* fallthrough */
          default:
            usage();
            return STATUS_OK;
        }
    }

    logger = CX_log_open_console((CX_LogLevel)verbosity, stderr); /* h4x0r l33t */

    if (!modname) {
        CX_log_trace(logger, CX_LOG_ERROR, EXE,
                     "Unspecified plugin name.");
        return STATUS_BAD_PARAM;
    }
    memset(&meta_data, 0, sizeof(meta_data));

    error = f0rx_loader_build(&loader, logger);
    if (error) {
        CX_log_trace(logger, CX_LOG_ERROR, EXE,
                     "Loader creation failed error=%s.",
                     f0rx_loader_strerror(error));
        return STATUS_LOADER_ERROR;
    }

    error = f0rx_loader_get_plugin_meta_data(loader,
                                             modvendor, modname, &meta_data);
    if (error) {
        CX_log_trace(logger, CX_LOG_ERROR, EXE,
               "Loader meta data fetching error=%s.",
               f0rx_loader_strerror(error));
        return STATUS_LOADER_ERROR;
    }
    print_plugin_meta_data(&meta_data);

    error = f0rx_loader_clean(&loader);
    if (error) {
        CX_log_trace(logger, CX_LOG_ERROR, EXE,
               "Loader destruction failed error=%s.",
               f0rx_loader_strerror(error));
        return STATUS_LOADER_ERROR;
    }

    CX_log_close(logger);

    return STATUS_OK;
}

/*
 * vim: expandtab shiftwidth=4:
 */

