/*
 * l0ader.c - simpl{e,istic} frei0r(x) plugin loader implementation.
 * (C) 2011-2014 Francesco Romani <fromani at gmail dot com>
 * license: ZLIB. Or GPLv2 if you wish.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include "arraykit.h"
#include "l0ader.h"

enum {
    LOADER_PATH_LEN     = 1024,
    LOADER_PATHS_NUM    = 8,
    LOADER_PLUGINS_NUM  = 32
};

#define F0RX_LOADER_TAG "l0ader"


#define RETURN_IF_REF_INVALID(PTR) do { \
    if (!(PTR)) { \
        return F0RX_LOADER_ERROR_REF_INVALID; \
    } \
} while (0)

#define WIRE_SYMBOL(loader, handle, dst, name, required) do { \
    const char *sym = "f0r_" # name ; \
    (dst)->name = dlsym(handle, sym); \
    if ((dst)->name == NULL) { \
        CX_log_trace(FL->log_context, CX_LOG_WARNING, F0RX_LOADER_TAG, \
                     "wiring error: %s", dlerror()); \
        if (required) { \
            return F0RX_LOADER_ERROR_PLUGIN_SYMBOL; \
        } \
    } \
} while (0)

#define WIRE_XSYMBOL(loader, handle, dst, name, required) do { \
    const char *sym = "f0rx_" # name ; \
    (dst)->x ## name = dlsym(handle, sym); \
    if ((dst)->x ## name == NULL) { \
        CX_log_trace(FL->log_context, CX_LOG_ERROR, F0RX_LOADER_TAG, \
                     "wiring error: %s", dlerror()); \
        if (required) { \
            return F0RX_LOADER_ERROR_PLUGIN_SYMBOL; \
        } \
    } \
} while (0)

/* FIXME: find a (CLEAN!) way to merge the two above */

struct f0rxplugin_ {
    const char *vendor;
    const char *name;
    int num_instances;
    void *handle;
    int (*init)(void);
    void (*deinit)(void);
    void (*get_plugin_info)(f0r_plugin_info_t* info);
    void (*get_param_info)(f0r_param_info_t* info, int param_index);
    f0r_instance_t (*construct)(unsigned int width, unsigned int height);
    void (*destruct)(f0r_instance_t instance);
    /* extensions */
    void (*xget_plugin_info)(f0rx_plugin_info_t* info);
    f0r_instance_t (*xconstruct)(int format, f0r_param_t *params,
                                 unsigned int value1, unsigned int value2);
};

typedef struct f0rxplugintemplate_ f0rx_plugin_template_t;
struct f0rxplugintemplate_ {
    struct f0rxplugin_ plugin;
    f0rx_plugin_meta_data_t meta_data;
    f0rx_element_t base;
};


struct f0rxloader_ {
    CX_VArray *plugin_paths;
    CX_VArray *plugin_templates;
    CX_LogContext *log_context;
};


static void free_item(void *ptr, void *unused);
static void free_template(void *item, void *user_data);
static int varray_foreach(CX_VArray *va,
                          void (*func)(void *, void *),
                          void *user_data);
static char *f0rx_strdup(const char *str);
static void f0rx_free(void **ptr);
static int f0rx_is_ready_stub(f0r_instance_t instance, int param_index);
static int f0rx_set_params_values_stub(f0r_instance_t instance,
                                       f0r_param_t *params);
static int f0rx_loader_mem_exausted(f0rx_loader_t FL, const char *where);
static int f0rx_loader_dlopen_lookup(f0rx_loader_t FL,
                                   const char *vendor,
                                   const char *name,
                                   void **handle);
static int f0rx_loader_dlclose(f0rx_loader_t FL, void **handle);
static int f0rx_loader_wire_template(f0rx_loader_t FL,
                                     f0rx_plugin_template_t *PT);
static int f0rx_loader_mark_template(f0rx_loader_t FL,
                                     f0rx_plugin_template_t *PT,
                                     const char *action);
static int f0rx_loader_init_template(f0rx_loader_t FL,
                                     f0rx_plugin_template_t *PT);
static int f0rx_loader_fini_template(f0rx_loader_t FL,
                                     f0rx_plugin_template_t *PT);
static int f0rx_loader_clean_template(f0rx_loader_t FL,
                                      f0rx_plugin_template_t **PT);
static int f0rx_loader_build_template(f0rx_loader_t FL,
                                     const char *vendor,
                                     const char *name,
                                     f0rx_plugin_template_t **PT);
static int f0rx_loader_find_template(f0rx_loader_t FL,
                                    const char *vendor,
                                    const char *name,
                                    f0rx_plugin_template_t **PT,
                                    int *idx);
static int f0rx_loader_new_template(f0rx_loader_t FL,
                                   const char *vendor,
                                   const char *name,
                                   f0rx_plugin_template_t **PT);
static int f0rx_loader_del_template(f0rx_loader_t FL,
                                   const char *vendor,
                                   const char *name);
static int f0rx_loader_add_plugin_path(f0rx_loader_t FL,
                                      const char *plugin_path);
static int f0rx_loader_setup_plugin_paths(f0rx_loader_t FL);
static int f0rx_loader_init(f0rx_loader_t FL, CX_LogContext *context);
static int f0rx_loader_fini(f0rx_loader_t FL);


static void free_item(void *ptr, void *user_data)
{
    f0rx_loader_t FL = user_data;
    CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                 "freed plugin path: %s", ptr);
    free(ptr);
}

/* adapter */
static void free_template(void *item, void *user_data)
{
    f0rx_plugin_template_t *PT = item;
    f0rx_loader_t FL = user_data;
    f0rx_loader_clean_template(FL, &PT);
}

static int varray_foreach(CX_VArray *va,
                          void (*func)(void *, void *),
                          void *user_data)
{
    int32_t n = CX_varray_length(va);
    int32_t j = 0;
    for (j = 0; j < n; j++) {
        void *ptr = NULL;
        CX_varray_get(va, j, &ptr);
        func(ptr, user_data);
    }
    return F0RX_LOADER_ERROR_NONE;
}

/* just to make sure strdup behaves correctly*/
static char *f0rx_strdup(const char *str)
{
    char *res = NULL;
    if (str) {
        res = strdup(str);
    }
    return res;
}

static void f0rx_free(void **ptr)
{
    free(*ptr);
    *ptr = NULL;
}

static int f0rx_is_ready_stub(f0r_instance_t instance, int param_index)
{
    return 0;
}

static int f0rx_set_params_values_stub(f0r_instance_t instance,
                                       f0r_param_t *params)
{
    return 0;
}


static int f0rx_loader_mem_exausted(f0rx_loader_t FL, const char *where)
{
    CX_log_trace(FL->log_context, CX_LOG_ERROR, F0RX_LOADER_TAG,
                 "%s: memory exausted", where);
    return F0RX_LOADER_ERROR_MEM_EXAUSTED;
}


static int f0rx_loader_dlopen_lookup(f0rx_loader_t FL,
                                   const char *vendor,
                                   const char *name,
                                   void **handle)
{
    int err = F0RX_LOADER_ERROR_PLUGIN_LOADING;
    int32_t n = CX_varray_length(FL->plugin_paths);
    int32_t j = 0;
    if (vendor == NULL) {
        vendor = "";
    }
    *handle = NULL; // XXX
    for (j = 0; !(*handle) && j < n; j++) {
        char *dir = NULL;
        char path[LOADER_PATH_LEN] = { '\0' };
        CX_varray_get(FL->plugin_paths, j, &dir);
        snprintf(path, sizeof(path), "%s%s%s/%s.so",
                 dir, (vendor[0]) ?"/" :"", vendor, name);
        *handle = dlopen(path, RTLD_NOW|RTLD_LOCAL);
        CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                     "Trying to load %s: %s%s%s",
                     path,
                     (*handle) ?"found" :"failed",
                     (*handle) ?"" :" ",
                     (*handle) ?"" :dlerror());
    }
    if (*handle) {
        err = F0RX_LOADER_ERROR_NONE;
    }
    return err;
}

static int f0rx_loader_dlclose(f0rx_loader_t FL, void **handle)
{
    int err = F0RX_LOADER_ERROR_NONE;
    if (*handle) {
        dlclose(*handle);
        if (!err) {
            *handle = NULL;
        } else {
            err = F0RX_LOADER_ERROR_PLUGIN_LOADING;
        }
    }
    return err;
}

static int f0rx_loader_wire_template(f0rx_loader_t FL,
                                     f0rx_plugin_template_t *PT)
{
    WIRE_SYMBOL(FL,  PT->plugin.handle, &PT->plugin, init,              1);
    WIRE_SYMBOL(FL,  PT->plugin.handle, &PT->plugin, deinit,            1);
    WIRE_SYMBOL(FL,  PT->plugin.handle, &PT->plugin, get_plugin_info,   1);
    WIRE_SYMBOL(FL,  PT->plugin.handle, &PT->plugin, get_param_info,    1);
    WIRE_SYMBOL(FL,  PT->plugin.handle, &PT->plugin, construct,         1);
    WIRE_SYMBOL(FL,  PT->plugin.handle, &PT->plugin, destruct,          1);
    WIRE_XSYMBOL(FL, PT->plugin.handle, &PT->plugin, get_plugin_info,   0);
    WIRE_XSYMBOL(FL, PT->plugin.handle, &PT->plugin, construct,         0);
    WIRE_XSYMBOL(FL, PT->plugin.handle, &PT->base,   is_ready,          0);
    WIRE_XSYMBOL(FL, PT->plugin.handle, &PT->base,   set_params_values, 0);
    WIRE_SYMBOL(FL,  PT->plugin.handle, &PT->base,   set_param_value,   1);
    WIRE_SYMBOL(FL,  PT->plugin.handle, &PT->base,   get_param_value,   1);
    WIRE_SYMBOL(FL,  PT->plugin.handle, &PT->base,   update,            1);
    WIRE_SYMBOL(FL,  PT->plugin.handle, &PT->base,   update2,           0);
    WIRE_XSYMBOL(FL, PT->plugin.handle, &PT->base,   update,            0);
    return F0RX_LOADER_ERROR_NONE;
}

static int f0rx_loader_mark_template(f0rx_loader_t FL,
                                     f0rx_plugin_template_t *PT,
                                     const char *action)
{
    CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                 "%s template %p for plugin %s%s%s",
                 action, PT, PT->plugin.vendor,
                 (PT->plugin.vendor[0]) ?"/" :"",
                 PT->plugin.name);
    return 0;
}

static int f0rx_loader_init_template(f0rx_loader_t FL,
                                     f0rx_plugin_template_t *PT)
{
    int err = f0rx_loader_wire_template(FL, PT);
    if (!err) {
        int ret = PT->plugin.init();
        if (ret < 0) {
            CX_log_trace(FL->log_context, CX_LOG_ERROR, F0RX_LOADER_TAG,
                         "initialization error=%i for template %p plugin %s%s%s",
                         ret, PT, PT->plugin.vendor,
                         (PT->plugin.vendor[0]) ?"/" :"",
                         PT->plugin.name);
            err = F0RX_LOADER_ERROR_PLUGIN_ERROR;
        } else {
            f0rx_loader_mark_template(FL, PT, "initializing");
            if (PT->plugin.xget_plugin_info
             && PT->plugin.xconstruct
             && PT->base.xupdate) { /* only those are mandatory */
                CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                             "plugin uses extended API");
                PT->meta_data.extended = 1;
            }
            if (PT->plugin.xget_plugin_info) {
                PT->plugin.xget_plugin_info(&PT->meta_data.info);
            } else {
                PT->plugin.get_plugin_info(&PT->meta_data.info.base);
                PT->meta_data.info.content_type = F0RX_CONTENT_TYPE_VIDEO;
                PT->meta_data.info.sample_model = 0;
            }
            PT->meta_data.params = calloc(PT->meta_data.info.base.num_params,
                                          sizeof(f0r_param_info_t));
            if (PT->meta_data.params) {
                int j = 0;
                for (j = 0; j < PT->meta_data.info.base.num_params; j++) {
                    PT->plugin.get_param_info(&(PT->meta_data.params[j]), j);
                }
            } else {
                err = f0rx_loader_mem_exausted(FL, __func__);
            }
        }
    }
    return err;
}

static int f0rx_loader_fini_template(f0rx_loader_t FL,
                                     f0rx_plugin_template_t *PT)
{
    f0rx_loader_mark_template(FL, PT, "finalizing");
    free(PT->meta_data.params);
    if (PT->plugin.deinit) {
        /* this function can be called also against
           an half-baked template. */
        /* TODO: would the null pattern better/cleaner/nicer? */
        PT->plugin.deinit();
    }
    return F0RX_LOADER_ERROR_NONE;
}

static int f0rx_loader_clean_template(f0rx_loader_t FL,
                                     f0rx_plugin_template_t **PT)
{
    int err = f0rx_loader_fini_template(FL, *PT);
    err = f0rx_loader_dlclose(FL, &(((*PT))->plugin.handle));
    f0rx_loader_mark_template(FL, *PT, "cleaning");
    if (!err) {
        free((void *)(*PT)->plugin.vendor);
        free((void *)(*PT)->plugin.name);
        f0rx_free((void **)PT);
    }
    return err;
}

static int f0rx_loader_build_template(f0rx_loader_t FL,
                                      const char *vendor,
                                      const char *name,
                                      f0rx_plugin_template_t **PT)
{
    int err = F0RX_LOADER_ERROR_NONE;
    (*PT) = calloc(1, sizeof(f0rx_plugin_template_t));
    if ((*PT)) {
        (*PT)->plugin.vendor = f0rx_strdup(vendor);
        (*PT)->plugin.name = f0rx_strdup(name);
        f0rx_loader_mark_template(FL, *PT, "looking up");
        err = f0rx_loader_dlopen_lookup(FL, vendor, name,
                                       &((*PT)->plugin.handle));
        f0rx_loader_mark_template(FL, *PT, "building");
        if (!err && (*PT)->plugin.vendor && (*PT)->plugin.name) {
            err = f0rx_loader_init_template(FL, *PT);
        } else {
            CX_log_trace(FL->log_context, CX_LOG_ERROR, F0RX_LOADER_TAG,
                         "error loading %s::%s %i",
                         vendor, name, err);
        }
        if (err) {
            f0rx_loader_clean_template(FL, PT);
        }
    }
    return err;
}

static int f0rx_loader_find_template(f0rx_loader_t FL,
                                    const char *vendor,
                                    const char *name,
                                    f0rx_plugin_template_t **PT,
                                    int *idx)
{
    int err = F0RX_LOADER_ERROR_REF_INVALID;
    int32_t n = CX_varray_length(FL->plugin_templates);
    int32_t j = 0;
    *PT = NULL; // meh
    for (j = 0; !(*PT) && j < n; j++) {
        f0rx_plugin_template_t *cur = NULL;
        CX_varray_get(FL->plugin_templates, j, &cur);
        if (!strcmp(cur->plugin.vendor, vendor)
         && !strcmp(cur->plugin.name, name)) {
            err = F0RX_LOADER_ERROR_NONE;
            *PT = cur;
            *idx = j;
        }
    }
    return err;
}

static int f0rx_loader_new_template(f0rx_loader_t FL,
                                   const char *vendor,
                                   const char *name,
                                   f0rx_plugin_template_t **PT)
{
    int j = 0;
    int err = f0rx_loader_find_template(FL, vendor, name, PT, &j);
    if (!err) {
        CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                     "found cached template %p @%i for %s%s%s",
                     *PT, j,
                     vendor, (vendor[0]) ?"/" :"", name);
    } else {
        CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                     "building new template for %s%s%s",
                     vendor, (vendor[0]) ?"/" :"", name);
        err = f0rx_loader_build_template(FL, vendor, name, PT);
        if (!err) {
            CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                         "caching new template for %s%s%s",
                         vendor, (vendor[0]) ?"/" :"", name);
            err = CX_varray_append(FL->plugin_templates, PT);
            if (!err) {
                err = F0RX_LOADER_ERROR_NONE; /* to be sure */
            } else {
                f0rx_loader_clean_template(FL, PT); /* cannot fail */
            }
        }
    }
    return err;
}

static int f0rx_loader_del_template(f0rx_loader_t FL,
                                   const char *vendor,
                                   const char *name)
{
    f0rx_plugin_template_t *PT = NULL;
    int j = 0;
    int err = f0rx_loader_find_template(FL, vendor, name, &PT, &j);
    if (!err) {
        err = f0rx_loader_clean_template(FL, &PT);
        if (!err) {
            CX_varray_remove(FL->plugin_templates, j);
        }
    }
    return err;
}

static int f0rx_loader_add_plugin_path(f0rx_loader_t FL,
                                      const char *plugin_path)
{
    int ret = CX_VARRAY_OK;
    int err = F0RX_LOADER_ERROR_MEM_EXAUSTED;
    RETURN_IF_REF_INVALID(FL);
    RETURN_IF_REF_INVALID(plugin_path);
    const char *path = strdup(plugin_path);
    if (path) {
        ret = CX_varray_append(FL->plugin_paths, &path);
        if (ret == CX_VARRAY_OK) {
            CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                            "added '%s' to the plugin path", path);
            err = F0RX_LOADER_ERROR_NONE;
        } else {
            err = f0rx_loader_mem_exausted(FL, __func__);
        }
    }
    return err;
}

static int f0rx_loader_setup_plugin_paths(f0rx_loader_t FL)
{
    int err = F0RX_LOADER_ERROR_NONE;
    char *env = f0rx_strdup(getenv("FREI0R_PATH"));
    if (env) {
        char *p = NULL;
        char *ptr = NULL;
        CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                     "setting plugin path from ENV");
        for (p = env; !err && (p = strtok_r(p, ":", &ptr)); p = NULL) {
            err = f0rx_loader_add_plugin_path(FL, p);
        }
        free(env);
    } else {
        const char *home = getenv("HOME");
        CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                     "setting plugin path statically");
        if (home) {
            char path[LOADER_PATH_LEN] = { '\0' };
            snprintf(path, sizeof(path), "%s/.frei0r-1/lib", home);
            err = f0rx_loader_add_plugin_path(FL, path);
        } else {
            CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                         "missing $HOME");
        }
        if (!err) {
            err = f0rx_loader_add_plugin_path(FL,
                                              "/usr/local/lib/frei0r-1");
        }
        if (!err) {
            err = f0rx_loader_add_plugin_path(FL,
                                              "/usr/lib/frei0r-1");
        }
    }
    return err;
}

static int f0rx_loader_init(f0rx_loader_t FL, CX_LogContext *log_context)
{
    int err = F0RX_LOADER_ERROR_NONE;
    FL->log_context = log_context;
    FL->plugin_paths = CX_varray_new(LOADER_PATHS_NUM,
                                     sizeof(const char *));
    FL->plugin_templates = CX_varray_new(LOADER_PLUGINS_NUM,
                                         sizeof(f0rx_plugin_template_t *));
    if (!FL->plugin_templates || ! FL->plugin_paths) {
        err = f0rx_loader_mem_exausted(FL, __func__);
    } else {
        err = f0rx_loader_setup_plugin_paths(FL);
    }
    return err;
}

static int f0rx_loader_fini(f0rx_loader_t FL)
{
    int err = F0RX_LOADER_ERROR_NONE;
    if (FL->plugin_paths) {
        varray_foreach(FL->plugin_paths, free_item, FL);
        CX_varray_del(FL->plugin_paths);
        CX_log_trace(FL->log_context, CX_LOG_DEBUG, F0RX_LOADER_TAG,
                     "freed plugin paths");
    }
    if (FL->plugin_templates) {
        varray_foreach(FL->plugin_templates, free_template, FL);
        CX_varray_del(FL->plugin_templates);
    }
    return err;
}

static int f0rx_loader_init_element(f0rx_element_t *element,
                                   f0rx_plugin_template_t *PT,
                                   uint32_t flags, int format, f0r_param_t *params,
                                   unsigned int value1, unsigned int value2)
{
    int err = F0RX_LOADER_ERROR_NONE;
    memcpy(element, &PT->base, sizeof(*element));
    element->plugin = &PT->plugin;
    element->id = PT->plugin.num_instances++;
    if (!PT->meta_data.extended && (flags & F0RX_LOADER_FLAG_STUBS)) {
        element->xset_params_values = f0rx_set_params_values_stub;
        element->xis_ready = f0rx_is_ready_stub;
    }
    if ((flags & F0RX_LOADER_FLAG_EXTENDED)
     && PT->plugin.xconstruct /* parachute */) {
        element->instance = PT->plugin.xconstruct(format, params, value1, value2);
    } else {
        element->instance = PT->plugin.construct(value1, value2);
    }
    if (!element->instance) {
        err = F0RX_LOADER_ERROR_PLUGIN_ERROR;
    }
    return err;
}

static int f0rx_loader_fini_element(f0rx_element_t *element)
{
    int err = F0RX_LOADER_ERROR_NONE;
    element->plugin->destruct(element->instance);
    element->plugin->num_instances--;
    return err;
}


int f0rx_loader_load_element(f0rx_loader_t loader,
                             const char *vendor, const char *name,
                             uint32_t flags, int format, f0r_param_t *params,
                             unsigned int value1, unsigned int value2,
                             f0rx_element_t **element)
{
    int err = F0RX_LOADER_ERROR_NONE;
    f0rx_plugin_template_t *pt = NULL;
    RETURN_IF_REF_INVALID(loader);
    RETURN_IF_REF_INVALID(vendor);
    RETURN_IF_REF_INVALID(name);
    RETURN_IF_REF_INVALID(element);
    err = f0rx_loader_new_template(loader, vendor, name, &pt);
    if (!err) {
        if (flags & F0RX_LOADER_FLAG_EXTENDED && !pt->meta_data.extended
         && !(flags & F0RX_LOADER_FLAG_FALLBACK)) {
            err = F0RX_LOADER_ERROR_PLUGIN_SYMBOL;
        } else {
            *element = calloc(1, sizeof(f0rx_element_t));
            if (*element) {
                err = f0rx_loader_init_element(*element, pt,
                                               flags, format, params,
                                               value1, value2);
            } else {
                err = f0rx_loader_mem_exausted(loader, __func__);
            }
        }
    }
    return err;
}


int f0rx_loader_unload_element(f0rx_loader_t loader, f0rx_element_t **element)
{
    f0rx_plugin_t plugin = NULL;
    int err = F0RX_LOADER_ERROR_NONE;
    RETURN_IF_REF_INVALID(loader);
    RETURN_IF_REF_INVALID(element);
    plugin = (*element)->plugin;
    err = f0rx_loader_fini_element(*element); /* cannot fail */
    f0rx_free((void **)element);
    if (!err && plugin->num_instances == 0) {
        err = f0rx_loader_del_template(loader,
                                      plugin->vendor, plugin->name);
    }
    return err;
}

int f0rx_loader_get_plugin_meta_data(f0rx_loader_t loader,
                                     const char *vendor, const char *name,
                                     f0rx_plugin_meta_data_t *meta_data)
{
    int err = F0RX_LOADER_ERROR_NONE;
    f0rx_plugin_template_t *pt = NULL;
    RETURN_IF_REF_INVALID(loader);
    RETURN_IF_REF_INVALID(vendor);
    RETURN_IF_REF_INVALID(name);
    RETURN_IF_REF_INVALID(meta_data);
    err = f0rx_loader_new_template(loader, vendor, name, &pt);
    if (!err) {
        memcpy(meta_data, &(pt->meta_data), sizeof(*meta_data));
    }
    return err;
}

int f0rx_loader_build(f0rx_loader_t *loader, CX_LogContext *log_context)
{
    int err = F0RX_LOADER_ERROR_NONE;
    RETURN_IF_REF_INVALID(loader);
    RETURN_IF_REF_INVALID(log_context);
    *loader = calloc(1, sizeof(struct f0rxloader_));
    if (*loader) {
        err = f0rx_loader_init(*loader, log_context);
        if (err) {
            f0rx_loader_clean(loader);
        }
    } else {
        err = F0RX_LOADER_ERROR_MEM_EXAUSTED;
    }
    return err;
}

int f0rx_loader_clean(f0rx_loader_t *loader)
{
    int err = F0RX_LOADER_ERROR_NONE;
    RETURN_IF_REF_INVALID(loader);
    f0rx_loader_fini(*loader);
    f0rx_free((void **)loader);
    return err;
}

const char *f0rx_loader_strerror(int error)
{
    static const char *error_strings[] = { /* TODO reread dsohowto.pdf */
        "no error",
        "invalid/NULL pointer",
        "memory exausted",
        "error loading a plugin",
        "plugin not found",
        "missing symbol in a plugin",
        "plugin internal error",
        "",  /* man-sao, the first defence */
        NULL /* wu-sao, the last defence */
    };
    int err_idx = (error > 0) ?error :-error;
    if (err_idx > F0RX_LOADER_ERROR_NONE
     || err_idx < F0RX_LOADER_ERROR_PLUGIN_ERROR) {
        return "invalid error code";
    }
    return error_strings[err_idx];
}

/*
 * vim: set et sw=4
 */

