/*
 * sprocess0r.c- simpl{e,istic} frei0rx stream processor engine.
 * (C) 2012-2014 Francesco Romani <fromani at gmail dot com>
 * license: ZLIB.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>

#include "arraykit.h"
#include "l0ader.h"


#define EXE "sprocess0r"

enum {
    DEFAULT_WIDTH  = 640,
    DEFAULT_HEIGHT = 480
};

enum {
    STATUS_INTERNAL_ERROR = -1,
    /* How's that working out for you? What? Being clever. */
    STATUS_OK = 0,
    STATUS_BAD_PARAM,
    STATUS_NO_MODULE,
    STATUS_MODULE_ERROR,
    STATUS_BAD_MODULES,
    STATUS_MODULE_FAILED,
    STATUS_LOADER_ERROR
};

typedef struct sp_item_conf sp_item_conf_t;
struct sp_item_conf {
    const char *vendor; /* soft ref */
    const char *name;
    const char *option_string;
};

typedef struct sp_item sp_item_t;
struct sp_item {
    f0rx_plugin_meta_data_t meta_data;
    f0rx_element_t *element;
};

typedef struct sp_pipe sp_pipe_t;
struct sp_pipe {
    int stream_index;
    AVCodec *dec; /* soft ref */
    AVCodecContext *dec_ctx;
    int64_t last_pts;
    CX_VArray *items;
};

typedef struct sp_options sp_options_t;
struct sp_options {
    int done;
    int verbosity;
    uint32_t flags;
    const char *mod_vendor;
    const char *input_file;
    const char *output_file;
};

typedef struct sp_context sp_context_t;
struct sp_context {
    f0rx_loader_t loader;

    AVFormatContext *fmt_ctx;

    sp_pipe_t video;
};

#define RETURN_ERR_IF_NULL(PTR) do { \
    if ((PTR) == NULL) { \
        return -1; \
    } \
} while (0)

/* looks like I'm not really so clever. */
#define RETURN_ERR_IF_ERR(ERR) do { \
    if ((ERR)) { \
        return (ERR); \
    } \
} while (0)

static sp_item_conf_t *sp_item_conf_new(const char *vendor,
                                        const char *name, const char *opts)
{
    sp_item_conf_t *ic = NULL;
    if (vendor && name && opts) {
        size_t name_len = strlen(name);
        size_t opts_len = strlen(opts);
        size_t len = sizeof(*ic) + name_len + 1 + opts_len + 1;
        char *pc = NULL;
        char *mem = calloc(1, len);
        if (mem) {
            ic = (sp_item_conf_t *)mem;
            ic->vendor = vendor; /* soft ref */
            pc = mem + sizeof(*ic);
            strncpy(pc, name, name_len + 1);
            ic->name = pc;
            pc += name_len + 1;
            strncpy(pc, opts, opts_len + 1);
            ic->option_string = pc;
        }
    }
    return ic;
}

static void sp_item_conf_del(sp_item_conf_t *ic)
{
    free(ic);
}

static int sp_pipe_init(sp_pipe_t *PP)
{
    int err = -1;
    if (PP) {
        memset(PP, 0, sizeof(*PP));
        PP->stream_index = -1;
        PP->last_pts = AV_NOPTS_VALUE;
        err = 0;
    }
    return err;
}

static int sp_pipe_cleanup(CX_VArray *items)
{
    int32_t i = 0, num = 0;
    RETURN_ERR_IF_NULL(items);

    num = CX_varray_length(items);
    for (i = 0; i < num; i++) {
    }
    return STATUS_OK;
}

static int sp_pipe_fini(sp_pipe_t *PP)
{
    int err = -1;
    if (PP) {
        if (PP->dec_ctx) {
            avcodec_close(PP->dec_ctx);
            PP->dec_ctx = NULL;
        }
        if (PP->items) {
            sp_pipe_cleanup(PP->items);
            CX_varray_del(PP->items);
            PP->items = NULL;
        }
        err = 0;
    }
    return err;
}

static int sp_context_open_output(sp_context_t *CTX, const sp_options_t *opts)
{
    int ret = STATUS_OK;

    RETURN_ERR_IF_NULL(CTX);
    RETURN_ERR_IF_NULL(opts);

    return ret;
}

static int sp_context_open_input(sp_context_t *CTX, const sp_options_t *opts)
{
    int ret = 0;

    RETURN_ERR_IF_NULL(CTX);
    RETURN_ERR_IF_NULL(opts);
   
    ret = avformat_open_input(&CTX->fmt_ctx, opts->input_file, NULL, NULL);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot open input file\n");
        return STATUS_INTERNAL_ERROR;
    }

    ret = avformat_find_stream_info(CTX->fmt_ctx, NULL);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot find stream information\n");
        return STATUS_INTERNAL_ERROR;
    }
    return STATUS_OK;
}

static int sp_context_open_pipe_video(sp_context_t *CTX)
{
    int ret = 0;

    RETURN_ERR_IF_NULL(CTX);

    sp_pipe_init(&CTX->video);

    /* select the video stream */
    ret = av_find_best_stream(CTX->fmt_ctx, AVMEDIA_TYPE_VIDEO, -1, -1,
                              &CTX->video.dec, 0);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot find a video stream in the input file\n");
        return STATUS_INTERNAL_ERROR;
    }
    CTX->video.stream_index = ret;
    CTX->video.dec_ctx = CTX->fmt_ctx->streams[CTX->video.stream_index]->codec;

    /* init the video decoder */
    ret = avcodec_open2(CTX->video.dec_ctx, CTX->video.dec, NULL);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot open video decoder\n");
        return STATUS_INTERNAL_ERROR;
    }

    return STATUS_OK;
}


// FIXME
static void params_from_meta_data(const f0rx_plugin_meta_data_t *meta_data,
                                  unsigned int *param1,
                                  unsigned int *param2)
{
    if (meta_data->info.content_type == F0RX_CONTENT_TYPE_VIDEO) {
        *param1 = DEFAULT_WIDTH;
        *param2 = DEFAULT_HEIGHT;
    } else {
        *param1 = 0; /* TODO */
        *param2 = 0; /* TODO */
    }
    return;
}



static int sp_context_load_plugin(sp_context_t *CTX,
                                  const sp_item_conf_t *ic, sp_item_t *item)
{
    int format = 0;
    unsigned int param1 = 0, param2 = 0;
    uint32_t flags = F0RX_LOADER_FLAG_EXTENDED|F0RX_LOADER_FLAG_FALLBACK;
    int err = 0;

    err = f0rx_loader_get_plugin_meta_data(CTX->loader,
                                           ic->vendor, ic->name,
                                           &item->meta_data);
    if (err) {
    }
    params_from_meta_data(&item->meta_data, &param1, &param2);
    format = item->meta_data.info.base.color_model;

    err = f0rx_loader_load_element(CTX->loader,
                                   ic->vendor, ic->name,
                                   flags, format, NULL,
                                   param1, param2,
                                   &item->element);

    return err;
}

static int sp_context_load_plugins_video(sp_context_t *CTX,
                                         const CX_VArray *plugins)
{
    int err = STATUS_OK;
    int32_t i = 0, num= 0;

    RETURN_ERR_IF_NULL(CTX);
    RETURN_ERR_IF_NULL(plugins);

    num = CX_varray_length(plugins);
    for (i = 0; !err && i < num; i++) {
        sp_item_t *item = calloc(1, sizeof(*item));
        if (item) {
            sp_item_conf_t *ic = NULL;

            err = CX_varray_get(plugins, i, &ic);

            err = sp_context_load_plugin(CTX, ic, item);
        } else {
            err = STATUS_INTERNAL_ERROR;
        }
    }
    
    return err;
}

static int sp_context_open(sp_context_t *CTX,
                           const sp_options_t *opts,
                           const CX_VArray* plugins_video)
{
    int err = STATUS_OK;

    err = f0rx_loader_build(&CTX->loader,
                            f0rx_logger_console, stderr,
                            F0RX_LOG_DEBUG); // TODO verbosity
    if (err) {
        fprintf(stderr, "%s: Loader creation failed error=%i.\n", EXE, err);
        return STATUS_LOADER_ERROR;
    }
 
    err = sp_context_open_input(CTX, opts);
    RETURN_ERR_IF_ERR(err);

    err = sp_context_open_pipe_video(CTX);
    RETURN_ERR_IF_ERR(err);

    err = sp_context_load_plugins_video(CTX, plugins_video);
    RETURN_ERR_IF_ERR(err);

    err = sp_context_open_output(CTX, opts);
    return err;
}


static int sp_context_close(sp_context_t *CTX)
{
    return -1;
}


static void version(void)
{
    fprintf(stderr, "%s (%s v%s) (C) 2012-2014 Francesco Romani\n",
            EXE, PACKAGE, VERSION);
}

static void usage(void)
{
    version();
    fprintf(stderr, "Usage: %s [options] -i name\n", EXE);
    fprintf(stderr, "    -v vendor     Use 'vendor' as module vendor\n");
    fprintf(stderr, "    -p name       Plugin name to load (like \'smooth\')\n");
    fprintf(stderr, "    -i file       Input file name\n");
    fprintf(stderr, "    -o file       Output file name\n");
    fprintf(stderr, "    -d verbosity  Verbosity mode [1]\n");
}

#define VALIDATE_OPTARG(optarg) do { \
    if (optarg[0] == '-') { \
        usage(); \
        return STATUS_BAD_PARAM; \
    } \
} while (0)

static int sp_options_init(sp_options_t *opts)
{
    RETURN_ERR_IF_NULL(opts);
    memset(opts, 0, sizeof(*opts));
    return 0;
}

#define OPTIONS_STRING "d:i:o:vVh?"

static int sp_options_generic_parse(sp_options_t *opts, int argc, char *argv[])
{
    int ch;

    opts->done = 0;
    while (!opts->done) {
        ch = getopt(argc, argv, OPTIONS_STRING);
        if (ch == -1) {
            break;
        }

        switch (ch) {
          case 'd':
            VALIDATE_OPTARG(optarg);
            opts->verbosity = atoi(optarg);
            break;
          case 'i':
            VALIDATE_OPTARG(optarg);
            opts->input_file = optarg;
            break;
          case 'o':
            VALIDATE_OPTARG(optarg);
            opts->output_file = optarg;
            break;
          case 'v':
            VALIDATE_OPTARG(optarg);
            opts->mod_vendor = optarg;
            break;
          case 'V':
            version();
            opts->done = 1;
            break;
          case '?': /* fallthrough */
          case 'h': /* fallthrough */
          default:
            usage();
            opts->done = 1;
            break;
        }
    }

    return STATUS_OK;
}

static int sp_options_plugins_parse(CX_VArray *plugins,
                                    const sp_options_t *options,
                                    int argc, char *argv[])
{
    int ch = 0, err = 0;

    while (!err) {
        ch = getopt(argc, argv, "p"OPTIONS_STRING);
        if (ch == -1) {
            break;
        }

        switch (ch) {
          case 'p': {
            VALIDATE_OPTARG(optarg);
            sp_item_conf_t *ic = sp_item_conf_new(options->mod_vendor,
                                                  optarg, "");
            err = CX_varray_append(plugins, &ic);
          } break;
          default:
            break;/* pass by swift and silent */
        }
    }

    return STATUS_OK;
}

static int sp_options_free_fini(CX_VArray *plugins)
{
    if (plugins) {
        int err = 0;
        int32_t i = 0, num = CX_varray_length(plugins);
        for (i = 0; !err && i < num; i++) {
            sp_item_conf_t *ic = NULL;
            err = CX_varray_get(plugins, i, &ic);
            sp_item_conf_del(ic);
        }
    }
    return STATUS_OK;
}


int main(int argc, char *argv[])
{
    int err = STATUS_OK;
    sp_options_t options;
    sp_context_t context;
    CX_VArray *plugins_video = NULL;

    if (argc == 1) {
        usage();
        return STATUS_BAD_PARAM;
    }

    sp_options_init(&options);
    err = sp_options_generic_parse(&options, argc, argv);
    if (err || options.done) {
        return err;
    }

    plugins_video = CX_varray_new(16/* FIXME */, sizeof(sp_item_conf_t *));
    if (!plugins_video) {
        /* FIXME */
        return STATUS_INTERNAL_ERROR;
    }
    err = sp_options_plugins_parse(plugins_video, &options, argc, argv);
    if (err || CX_varray_length(plugins_video) == 0) {
        usage();
        return STATUS_BAD_PARAM;
    }

    err = sp_context_open(&context, &options, plugins_video);
    RETURN_ERR_IF_ERR(err);


    return STATUS_OK;
}

/*
 * vim: expandtab shiftwidth=4:
 */

