/*
 * validat0rx.c- tycho-plugins debug tool:
 *               does basic validations against a frei0rx plugin.
 * (C) 2011-2014 Francesco Romani <fromani at gmail dot com>
 * license: ZLIB.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "logkit.h"

#include "l0ader.h"


#define EXE "validat0rx"

enum {
    STATUS_OK = 0,
    STATUS_BAD_PARAM,
    STATUS_NO_MODULE,
    STATUS_MODULE_ERROR,
    STATUS_BAD_MODULES,
    STATUS_MODULE_FAILED,
    STATUS_LOADER_ERROR
};

enum {
    DEFAULT_WIDTH  = 640,
    DEFAULT_HEIGHT = 480
};

void version(void)
{
    fprintf(stderr, "%s (%s v%s) (C) 2011-2014 Francesco Romani\n",
            EXE, PACKAGE, VERSION);
}

static void usage(void)
{
    version();
    fprintf(stderr, "Usage: %s [options] -i name\n", EXE);
    fprintf(stderr, "    -i name       Plugin name information (like \'smooth\')\n");
    fprintf(stderr, "    -v vendor     Use 'vendor' as module vendor\n");
    fprintf(stderr, "    -d verbosity  Verbosity mode [1]\n");
    fprintf(stderr, "    -x            Fallback to standard frei0r if needed [no]\n");
    fprintf(stderr, "    -c            Signal succesfull validation [quiet]\n");
}

#define VALIDATE_OPTARG(optarg) do { \
    if (!optarg || optarg[0] == '-') { \
        usage(); \
        return STATUS_BAD_PARAM; \
    } \
} while (0)

static void params_from_meta_data(const f0rx_plugin_meta_data_t *meta_data,
                                  unsigned int *param1,
                                  unsigned int *param2)
{
    if (meta_data->info.content_type == F0RX_CONTENT_TYPE_VIDEO) {
        *param1 = DEFAULT_WIDTH;
        *param2 = DEFAULT_HEIGHT;
    } else {
        *param1 = 0; /* TODO */
        *param2 = 0; /* TODO */
    }
    return;
}



int main(int argc, char *argv[])
{
    int ch;
    int ret = STATUS_OK;
    int quiet = 1;
    int verbosity = 0;
    int error = 0;
    int format = 0;
    const char *modname = NULL;
    const char *modvendor = "";
    unsigned int param1 = 0;
    unsigned int param2 = 0;
    uint32_t flags = F0RX_LOADER_FLAG_EXTENDED;

    f0rx_plugin_meta_data_t meta_data;
    f0rx_loader_t loader = NULL;
    f0rx_element_t *element = NULL;

    CX_LogContext *logger = NULL;

    if (argc == 1) {
        usage();
        return STATUS_BAD_PARAM;
    }

    while (1) {
        ch = getopt(argc, argv, "cd:i:v:xVh?");
        if (ch == -1) {
            break;
        }

        switch (ch) {
          case 'c': /* chatty */
            quiet = 0;
            break;
          case 'd':
            VALIDATE_OPTARG(optarg);
            verbosity = atoi(optarg);
            break;
          case 'i':
            VALIDATE_OPTARG(optarg);
            modname = optarg;
            break;
          case 'v':
            VALIDATE_OPTARG(optarg);
            modvendor = optarg;
            break;
          case 'x':
            flags |= F0RX_LOADER_FLAG_FALLBACK;
            break;
          case 'V':
            version();
            return STATUS_OK;
          case '?': /* fallthrough */
          case 'h': /* fallthrough */
          default:
            usage();
            return STATUS_OK;
        }
    }

    logger = CX_log_open_console((CX_LogLevel)verbosity, stderr); /* h4x0r l33t */

    if (!modname) {
        CX_log_trace(logger, CX_LOG_ERROR, EXE,
                     "Unspecified plugin name.");
        return STATUS_BAD_PARAM;
    }
    memset(&meta_data, 0, sizeof(meta_data));

    error = f0rx_loader_build(&loader, logger);
    if (error) {
        CX_log_trace(logger, CX_LOG_ERROR, EXE,
                     "Loader creation failed error=%s.",
                     f0rx_loader_strerror(error));
        return STATUS_LOADER_ERROR;
    }

    error = f0rx_loader_get_plugin_meta_data(loader,
                                             modvendor, modname, &meta_data);
    if (error) {
        CX_log_trace(logger, CX_LOG_ERROR, EXE,
                     "Loader meta data fetching error=%s.",
                     f0rx_loader_strerror(error));
        return STATUS_LOADER_ERROR;
    }
    params_from_meta_data(&meta_data, &param1, &param2);

    format = meta_data.info.base.color_model;
    error = f0rx_loader_load_element(loader,
                                     modvendor, modname,
                                     flags, format, NULL,
                                     param1, param2,
                                     &element);
    if (error) {
        if (!quiet) {
            printf("%s%s%s is NOT a valid frei0rx plugin.\n",
                   modvendor, (modvendor[0]) ?"/" :"", modname);
        }
        ret = STATUS_LOADER_ERROR;
    } else {
        if (!quiet) {
            printf("%s%s%s is a valid frei0rx plugin.\n",
                   modvendor, (modvendor[0]) ?"/" :"", modname);
        }

        error = f0rx_loader_unload_element(loader, &element);
        if (error) {
            CX_log_trace(logger, CX_LOG_ERROR, EXE,
                         "Loader element destruction failed error=%s.",
                         f0rx_loader_strerror(error));
        }
    }

    error = f0rx_loader_clean(&loader);
    if (error) {
        CX_log_trace(logger, CX_LOG_ERROR, EXE,
                     "Loader destruction failed error=%s.",
                     f0rx_loader_strerror(error));
        return STATUS_LOADER_ERROR;
    }

    return ret;
}

/*
 * vim: expandtab shiftwidth=4:
 */

