/* LIBARAN - Fast Multipole Method library
 * Copyright (C) 2006-2007 Pierre Gay
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "aran-config.h"

#include <stdlib.h>

#include <complex.h>

#include <math.h>

#include "aran/aran.h"
#include "aran/aransolver3d.h"
#include "aran/aranbinomial.h"
#include "aran/aranprofile.h"
#include "aran/aranprofiledb.h"

#include "glib/gprintf.h"

/* tree bbox size */
#define TR (1.)
/* circle radius */
#define R (0.999 * TR)



typedef struct _PointAccum PointAccum;

struct _PointAccum
{
  VsgVector3d vector;

  gdouble density;

  gcomplex128 accum;

  guint id;
};

static void point_accum_clear_accum (PointAccum *pa)
{
  pa->accum = 0.;
}

void p2p (PointAccum *one, PointAccum *other)
{
  if (one != other)
    {
      VsgVector3d tmp;
      gdouble inv_r;

      /* destination - source */
      vsg_vector3d_sub (&one->vector, &other->vector, &tmp);

      inv_r = 1. / vsg_vector3d_norm (&tmp);

      /* G(xd, xs) = 1./(xd-xs) * src_density */

      one->accum += inv_r * other->density;
      other->accum += inv_r * one->density;
    }
}

void p2m (PointAccum *particle, const VsgPRTree3dNodeInfo *dst_node,
          AranDevelopment3d *dst)
{
  aran_development3d_p2m(&particle->vector, particle->density, dst_node, dst);
}

void l2p (const VsgPRTree3dNodeInfo *devel_node, AranDevelopment3d *devel,
          PointAccum *particle)
{
  particle->accum += aran_development3d_local_evaluate (devel_node, devel,
							&particle->vector);
}

void p2l (PointAccum *particle, const VsgPRTree3dNodeInfo *dst_node,
          AranDevelopment3d *dst)
{
  aran_development3d_p2l (&particle->vector, particle->density, dst_node, dst);
}

void m2p (const VsgPRTree3dNodeInfo *devel_node, AranDevelopment3d *devel,
          PointAccum *particle)
{
  particle->accum += aran_development3d_multipole_evaluate (devel_node, devel,
                                                            &particle->vector);
}

static void _direct (PointAccum **points, guint np)
{
  guint i, j;

  for (i=0; i<np; i ++)
    {
      for (j=i+1; j<np; j ++)
        {
          p2p (points[i], points[j]);
        }
    }
}

static void _profile_operators (const gchar *db_filename)
{
  GKeyFile *profiles_file = g_key_file_new ();
  gchar *profiles_data;
  PointAccum p1 = {{0.1, 0.1, 0.1}, 0.1, 0., 0};
  PointAccum p2 = {{0.5, 0.5, 0.5}, 0.1, 0., 1};
  gdouble chisq, t;
  AranPoly1d *ap1d = aran_poly1d_new (2);
  gint maxbox = 100;
  const gint maxdeg = 20;
  const gint nsamples = maxdeg+1;
  gdouble abscissas[nsamples];
  gdouble samples[nsamples];
  const gchar *profiles_group;
  FILE *f = stdout;
  int i;

  for (i=0; i<nsamples; i ++)
    {
      gint deg = (maxdeg * i) / (nsamples-1);

      abscissas[i] = (gdouble) deg;
    }

  profiles_group = g_getenv ("ARAN_PROFILE_GROUP");
  if (profiles_group == NULL) profiles_group = ARAN_PROFILE_DB_DEFAULT_GROUP;

  if (db_filename != NULL)
    {
      g_key_file_load_from_file (profiles_file, db_filename,
                                 G_KEY_FILE_KEEP_COMMENTS |
                                 G_KEY_FILE_KEEP_TRANSLATIONS, NULL);
      f = fopen (db_filename, "w");
    }

  ap1d->degree = 0;
  t = aran_profile_nearfunc_3d ((AranParticle2ParticleFunc3d) p2p,
                                (AranParticleInitFunc3d) point_accum_clear_accum,
                                &p1, &p2, maxbox);
  ap1d->terms[0] = t / (maxbox * maxbox);
  aran_poly1d_write_key_file (ap1d, profiles_file, profiles_group, "p2p");


  ap1d->degree = 2;
  chisq =
    aran_poly1d_profile_p2m_3d_samples ((AranParticle2MultipoleFunc3d) p2m,
                                        (AranZeroFunc) aran_development3d_set_zero,
                                        &p1,
                                        (AranDevelopmentNewFunc)
                                        aran_development3d_new,
                                        (GDestroyNotify) aran_development3d_free,
                                        ap1d, nsamples, abscissas, samples);

  aran_profile_key_file_poly1d_write (profiles_file, profiles_group,
                                      "p2m", ap1d, chisq, nsamples,
                                      abscissas, samples);

  ap1d->degree = 2;
  t =
    aran_poly1d_profile_l2p_3d_samples ((AranLocal2ParticleFunc3d) l2p,
                                        NULL,
                                        &p1,
                                        (AranDevelopmentNewFunc) aran_development3d_new,
                                        (GDestroyNotify) aran_development3d_free,
                                        ap1d, nsamples, abscissas, samples);

  aran_profile_key_file_poly1d_write (profiles_file, profiles_group,
                                      "l2p", ap1d, chisq, nsamples,
                                      abscissas, samples);

  profiles_data = g_key_file_to_data (profiles_file, NULL, NULL);
  g_key_file_free (profiles_file);

  g_fprintf (f, "%s", profiles_data);
  g_free (profiles_data);

  if (db_filename != NULL) fclose (f);
}

static void _one_circle_distribution (PointAccum **points,
				      AranSolver3d *solver);
static void _random_distribution (PointAccum **points,
				  AranSolver3d *solver);

static void _grid_distribution (PointAccum **points,
                                AranSolver3d *solver);

static void _plummer_fill (PointAccum **points,
                           AranSolver3d *solver);

static gdouble err_lim = 1.E-3;
static guint np = 12;
static guint order = 20;
static gboolean check = TRUE;
static gboolean direct = FALSE;
static guint maxbox = 1;
static guint semifar_threshold = G_MAXUINT;
static gboolean verbose = FALSE;

static AranMultipole2MultipoleFunc3d m2m =
(AranMultipole2MultipoleFunc3d) aran_development3d_m2m;

static AranMultipole2LocalFunc3d m2l =
(AranMultipole2LocalFunc3d) aran_development3d_m2l;

static AranLocal2LocalFunc3d l2l =
(AranLocal2LocalFunc3d) aran_development3d_l2l;

static void (*_distribution) (PointAccum **, AranSolver3d *solver) =
_one_circle_distribution;


static
void parse_args (int argc, char **argv)
{
  int iarg = 1;
  char *arg;

  while (iarg < argc)
    {
      arg = argv[iarg];

      if (g_ascii_strcasecmp (arg, "-profile") == 0)
	{
	  iarg ++;

	  arg = (iarg<argc) ? argv[iarg] : NULL;

          _profile_operators (arg);

          exit (0);
	}
      else if (g_ascii_strcasecmp (arg, "-np") == 0)
	{
	  guint tmp = 0;
	  iarg ++;

	  arg = (iarg<argc) ? argv[iarg] : NULL;

	  if (sscanf (arg, "%u", &tmp) == 1)
	      np = tmp;
	  else
	    g_printerr ("Invalid particles number (-np %s)\n", arg);
	}
      else if (g_ascii_strcasecmp (arg, "-pr") == 0)
	{
	  guint tmp = 0;
	  iarg ++;

	  arg = (iarg<argc) ? argv[iarg] : NULL;

	  if (sscanf (arg, "%u", &tmp) == 1 && tmp > 0)
	      order = tmp;
	  else
	    g_printerr ("Invalid precision order value (-pr %s)\n", arg);
	}
      else if (g_ascii_strcasecmp (arg, "-s") == 0)
	{
	  guint tmp = 0;
	  iarg ++;

	  arg = (iarg<argc) ? argv[iarg] : NULL;

	  if (sscanf (arg, "%u", &tmp) == 1 && tmp > 0)
	      maxbox = tmp;
	  else
	    g_printerr ("Invalid maximum box size value (-s %s)\n", arg);
	}
      else if (g_ascii_strcasecmp (arg, "-err") == 0)
	{
	  gdouble tmp = 0;
	  iarg ++;

	  arg = (iarg<argc) ? argv[iarg] : NULL;

	  if (sscanf (arg, "%lf", &tmp) == 1 && tmp > 0.)
	      err_lim = tmp;
	  else
	    g_printerr ("Invalid error limit value (-err %s)\n", arg);
	}
      else if (g_ascii_strcasecmp (arg, "-semifar") == 0)
	{
	  guint tmp = 0;
	  iarg ++;

	  arg = (iarg<argc) ? argv[iarg] : NULL;

	  if (sscanf (arg, "%u", &tmp) == 1)
	      semifar_threshold = tmp;
	  else
	    g_printerr ("Invalid semifar threshold (-semifar %s)\n", arg);
	}
      else if (g_ascii_strcasecmp (arg, "-nocheck") == 0)
	{
	  check = FALSE;
	}

      else if (g_ascii_strcasecmp (arg, "-check") == 0)
	{
	  check = TRUE;
	}
      else if (g_ascii_strcasecmp (arg, "-direct") == 0)
	{
	  direct = TRUE;
	}
      else if (g_ascii_strcasecmp (arg, "-dist") == 0)
	{
	  iarg ++;

	  arg = (iarg<argc) ? argv[iarg] : NULL;

	  if (g_ascii_strcasecmp (arg, "circle") == 0)
	    {
	      _distribution = _one_circle_distribution;
	    }
	  else if (g_ascii_strcasecmp (arg, "random") == 0)
	    {
	      _distribution = _random_distribution;
	    }
	  else if (g_ascii_strcasecmp (arg, "grid") == 0)
	    {
	      _distribution = _grid_distribution;
	    }
	  else if (g_ascii_strcasecmp (arg, "plummer") == 0)
	    {
	      _distribution = _plummer_fill;
	    }
	  else 
	    {
	      g_printerr ("Invalid distribution name (-dist %s)\n", arg);
	    }
	}
      else if (g_ascii_strcasecmp (arg, "-translation") == 0)
	{
	  iarg ++;

	  arg = (iarg<argc) ? argv[iarg] : NULL;

	  if (g_ascii_strcasecmp (arg, "normal") == 0)
	    {
              m2m = (AranMultipole2MultipoleFunc3d) aran_development3d_m2m;

              m2l = (AranMultipole2LocalFunc3d) aran_development3d_m2l;

              l2l = (AranLocal2LocalFunc3d) aran_development3d_l2l;
	    }
	  else if (g_ascii_strcasecmp (arg, "kkylin") == 0)
	    {
              m2m =
                (AranMultipole2MultipoleFunc3d) aran_development3d_m2m_kkylin;

              m2l = (AranMultipole2LocalFunc3d) aran_development3d_m2l_kkylin;

              l2l = (AranLocal2LocalFunc3d) aran_development3d_l2l_kkylin;
	    }
	  else if (g_ascii_strcasecmp (arg, "rotate") == 0)
	    {
              m2m =
                (AranMultipole2MultipoleFunc3d) aran_development3d_m2m_rotate;

              m2l = (AranMultipole2LocalFunc3d) aran_development3d_m2l_rotate;

              l2l = (AranLocal2LocalFunc3d) aran_development3d_l2l_rotate;
	    }
	  else 
	    {
	      g_printerr ("Invalid translation name (-translation %s)\n", arg);
	    }
	}
      else if (g_ascii_strcasecmp (arg, "--verbose") == 0)
	{
	  verbose = TRUE;
	}
      else if (g_ascii_strcasecmp (arg, "--version") == 0)
	{
	  g_printerr ("%s version %s\n", argv[0], PACKAGE_VERSION);
	  exit (0);
	}
      else
	{
	  g_printerr ("Invalid argument \"%s\"\n", arg);
	}

      iarg ++;
    }
}

static void _one_circle_distribution (PointAccum **points,
				      AranSolver3d *solver)
{
  guint i;

  for (i=0; i<np; i++)
    {
      PointAccum *point = g_malloc0 (sizeof (PointAccum));

      point->vector.x = R * cos(2.*G_PI*i/np);
      point->vector.y = R * sin(2.*G_PI*i/np);
      point->vector.z = 0.0;
/*       point->vector.z = point->vector.x; */
      point->density = 1./np;
      point->accum = 0.;
      point->id = i;

      points[i] = point;

      if (!direct) aran_solver3d_insert_point (solver, point);
    }
}

static void _random_distribution (PointAccum **points,
				  AranSolver3d *solver)
{
  guint i, cptr = 0;
  gdouble tol = 1.e-2;

  aran_solver3d_set_tolerance (solver, tol);

  for (i=0; i<np; i++)
    {
      PointAccum *point = g_malloc0 (sizeof (PointAccum));
      PointAccum *found;

      point->vector.x = g_random_double_range (-R,R);
      point->vector.y = g_random_double_range (-R,R);
      point->vector.z = g_random_double_range (-R,R);
      point->density = g_random_double_range (0., 2./np);
      point->accum = 0.;
      point->id = cptr;

      if (direct)
        {
          points[cptr] = point;
          cptr ++;
        }
      else
        {
          found = aran_solver3d_find_point (solver, point);

          if (found != NULL)
            {
              found->density += point->density;

              g_free (point);
            }
          else
            {
              points[cptr] = point;
              aran_solver3d_insert_point (solver, point);
              cptr ++;
            }
        }
    }

  np = cptr;
}

static void _grid_distribution (PointAccum **points,
                                AranSolver3d *solver)
{
  guint i, j, k, size, cptr = 0;
  gdouble x, y, z, density;

  size = floor (pow (np, 1./3.));

  np = size*size*size;
  density = 1./np;

  for (i=0; i<size; i ++)
    {
      x = R * (2.*i/(size-1.)-1.);
      for (j=0; j<size; j ++)
        {
          y = R * (2.*j/(size-1.)-1.);
          for (k=0; k<size; k ++)
            {
              PointAccum *point = g_malloc0 (sizeof (PointAccum));

              z = R * (2.*k/(size-1.)-1.);

              point->vector.x = x;
              point->vector.y = y;
              point->vector.z = z;
              point->density = density;
              point->accum = 0.;
              point->id = cptr;

              if (!direct)
                aran_solver3d_insert_point (solver, point);

              points[cptr] = point;

              cptr ++;
            }
        }
    }
}

static guint32 _random_seed = 0;

static void _plummer_fill (PointAccum **points,
                           AranSolver3d *solver)
{
  gdouble rmax = 300.;
  guint i, real_np = 0;
  PointAccum lb, ub;
  PointAccum *point;
  GRand *rand = g_rand_new_with_seed (_random_seed);

  point = g_malloc0 (sizeof (PointAccum));

  ub.vector.x = rmax;
  ub.vector.y = rmax;
  ub.vector.z = rmax;

  lb.vector.x = - rmax;
  lb.vector.y = - rmax;
  lb.vector.z = - rmax;

  aran_solver3d_insert_point (solver, &ub);
  aran_solver3d_insert_point (solver, &lb);

  aran_solver3d_remove_point (solver, &lb);
  aran_solver3d_remove_point (solver, &ub);

  for (i=0; i<np; i++)
    {
      gdouble x1, x2, x3, r, tmp;

      x1 = g_rand_double_range (rand, 0., 1.);
      x2 = g_rand_double_range (rand, 0., 1.);
      x3 = g_rand_double_range (rand, 0., 1.);

      r = 1./ sqrt (pow (x1, -2./3.) -1.);

      point->vector.z = R * (1. - (x2+x2)) * r;

      tmp = sqrt (r*r - point->vector.z*point->vector.z);

      point->vector.x = R * tmp * cos (2.*G_PI * x3);
      point->vector.y = R * tmp * sin (2.*G_PI * x3);

      if (fabs (point->vector.x) > rmax) continue;
      if (fabs (point->vector.y) > rmax) continue;
      if (fabs (point->vector.z) > rmax) continue;

      point->density = 1./np;
      point->accum = 0.;
      point->id = real_np;

      aran_solver3d_insert_point (solver, point);
      points[real_np] = point;

      point = g_malloc0 (sizeof (PointAccum));
      real_np ++;

    }

  np = real_np;

  g_free (point);

  g_rand_free (rand);
}

int main (int argc, char **argv)
{
  VsgVector3d lbound = {-TR, -TR, -TR};
  VsgVector3d ubound = {TR, TR, TR};
  PointAccum **points;
  VsgPRTree3d *prtree;
  AranSolver3d *solver;
  int ret = 0;
  guint i;

  aran_init();

  vsg_packed_msg_trace_set_active (FALSE);

  parse_args (argc, argv);

  points = g_malloc0 (np * sizeof (PointAccum *));

  prtree =
    vsg_prtree3d_new_full (&lbound, &ubound,
			    (VsgPoint3dLocFunc) vsg_vector3d_vector3d_locfunc,
			    (VsgPoint3dDistFunc) vsg_vector3d_dist,
			    (VsgRegion3dLocFunc) NULL, maxbox);

  solver = aran_solver3d_new (prtree, ARAN_TYPE_DEVELOPMENT3D,
			      aran_development3d_new (0, order),
			      (AranZeroFunc) aran_development3d_set_zero);

  aran_solver3d_set_functions (solver,
                               (AranParticle2ParticleFunc3d) p2p,
                               (AranParticle2MultipoleFunc3d) p2m,
                               m2m,
                               m2l,
                               l2l,
                               (AranLocal2ParticleFunc3d)l2p);

  if (semifar_threshold < G_MAXUINT)
    {
      aran_solver3d_set_functions_full (solver,
                                        (AranParticle2ParticleFunc3d) p2p,
                                        (AranParticle2MultipoleFunc3d) p2m,
                                        m2m,
                                        m2l,
                                        l2l,
                                        (AranLocal2ParticleFunc3d) l2p,
                                        (AranParticle2LocalFunc3d) p2l,
                                        (AranMultipole2ParticleFunc3d) m2p,
                                        semifar_threshold);
    }

  _distribution (points, solver);


/*    g_printerr ("ok depth = %d size = %d\n", */
/*                aran_solver3d_depth (solver), */
/*                aran_solver3d_point_count (solver)); */

  if (direct) _direct (points, np);
  else aran_solver3d_solve (solver);

/*   vsg_prtree3d_write (prtree, stderr); */

  if (verbose)
    {
      glong zero_count, p2p_count, p2m_count, m2m_count;
      glong m2l_count, l2l_count, l2p_count, p2l_count, m2p_count;
      glong p2p_remote_count, m2l_remote_count;

      g_printerr ("particle count=%d\n\n", np);


      aran_solver3d_get_stats (solver, &zero_count,
                               &p2p_count, &p2m_count,
                               &m2m_count, &m2l_count,
                               &l2l_count, &l2p_count,
                               &p2l_count, &m2p_count,
                               &p2p_remote_count,
                               &m2l_remote_count);

      g_printerr ("zero count=%ld\n", zero_count);
      g_printerr ("p2p count=%ld\n", p2p_count);
      g_printerr ("p2p remote count=%ld\n", p2p_remote_count);
      g_printerr ("p2m count=%ld\n", p2m_count);
      g_printerr ("m2m count=%ld\n", m2m_count);
      g_printerr ("m2l count=%ld\n", m2l_count);
      g_printerr ("m2l remote count=%ld\n", m2l_remote_count);
      g_printerr ("l2l count=%ld\n", l2l_count);
      g_printerr ("l2p count=%ld\n", l2p_count);
      g_printerr ("p2l count=%ld\n", p2l_count);
      g_printerr ("m2p count=%ld\n", m2p_count);
    }

  aran_solver3d_free (solver);

  if (check)
    {
      for (i=0; i<np; i++)
	{
	  guint j;
	  gcomplex128 sum = 0.;
	  gcomplex128 err;

	  for (j=0; j<np; j++)
	    {
	      if (i != j)
		{
		  VsgVector3d tmp;
		  gdouble inv_r;

		  vsg_vector3d_sub (&points[i]->vector, &points[j]->vector,
				    &tmp);

		  inv_r = 1. / vsg_vector3d_norm (&tmp);
		  sum += inv_r * points[j]->density;
		}
	    }

	  err = (points[i]->accum - sum) /
	    MAX(cabs (points[i]->accum), cabs (sum));

	  if (cabs (err) > err_lim || !finite (cabs (err)))
	    {
	      g_printerr ("Error: pt%u (%e,%e): (%e,%e)!=(%e,%e) -> ",
			  i,
			  creal (err), cimag (err),
			  creal (points[i]->accum), cimag (points[i]->accum),
			  creal (sum), cimag (sum));
	      vsg_vector3d_write (&points[i]->vector, stderr);
	      g_printerr ("\n");
	      ret ++;
	    }
	}
    }

  for (i=0; i<np; i++)
    {
      g_free (points[i]);
    }

  g_free (points);

  return ret;
}

