/**
 *  @file main.c
 *  @brief aapng main source file.
 *
 *  @note Copyright Přemysl Janouch 2010. All right reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <locale.h>

#include <config.h>

#ifdef HAVE_GETOPT
	#include <unistd.h>
#else /* ! HAVE_GETOPT */
	#include "getopt.h"
#endif /* ! HAVE_GETOPT */

#include "waa.h"
#include "waautil.h"
#include "xalloc.h"


/** Free program resources. */
static void mainExit (void);
/** Program main loop. */
static int mainLoop (WaaContext *ctx, const char *file);

/** Stretch an image over the console area and render it. */
static void drawImage (WaaContext *ctx, WaaSurface *img);

/** Scale a rectangle into another rectangle.
 *  @param[in,out] rect  The rectangle we are scaling.
 *  @param[in] borders  The rectangle we're trying to fit @a rect into.
 */
static void scaleInto (WaaRect *restrict rect,
	const WaaRect *restrict borders);

/** Put values into a rectangle structure. */
#define SET_RECT(r, xx, yy, ww, hh) \
do { \
	(r).x = (xx); \
	(r).y = (yy); \
	(r).w = (ww); \
	(r).h = (hh); \
} while (0)


/** Console context. */
static WaaContext *g_ctx;


/** Version string for this application. */
const char *g_version_string = PACKAGE_NAME " " PACKAGE_VERSION "\n";

/** Usage string for this application. */
const char *g_usage_string =
	"Usage:\n"
	"  " PACKAGE_NAME " [OPTION..] file.png\n"
	"\n"
	"Options:\n"
	"  -h            Show this help.\n"
	"  -V            Show program version.\n"
	"  -p palette    Change the character palette. The palette may be\n"
	"                \"ascii\", \"block\" or your own palette string.\n"
	"                In the last case, each character denotes a single\n"
	"                entry in the palette, ranging from black to white.\n"
	"                The default palette is \"ascii\".\n"
	"\n";

int main (int argc, char *argv[])
{
	int ch;
	char *palette = NULL;

	setlocale(LC_ALL, "");

	/* Process options given to the program. */
	while ((ch = getopt(argc, argv, "hVp:")) != -1)
	{
		switch (ch)
		{
		case 'h':
			fputs(_(g_usage_string), stderr);
			exit(EXIT_SUCCESS);
		case 'V':
			fputs(g_version_string, stderr);
			exit(EXIT_SUCCESS);
		case 'p':
			palette = optarg;
			break;
		default:
			fputs(_(g_usage_string), stderr);
			exit(EXIT_FAILURE);
		}
	}
	if (argc - optind != 1)
	{
		fputs(_(g_usage_string), stderr);
		exit(EXIT_FAILURE);
	}

	/* Initialize the library. */
	g_ctx = waaInit();
	if (!g_ctx)
	{
		fprintf(stderr, _("WAA initialization failed.\n"));
		exit(EXIT_FAILURE);
	}

	/* Set the palette using which the image will be rendered. */
	if (!palette || !strcmp(palette, "ascii"))
		waaSetPalette(g_ctx, waaPaletteAscii);
	else if        (!strcmp(palette, "block"))
		waaSetPalette(g_ctx, waaPaletteBlock);
	else
	{
		wchar_t *wcs;
		size_t len;

		len = mbstowcs(NULL, palette, 0);
		if (len == (size_t) -1)
		{
			fputs(_("Failed to convert the supplied palette"
				" to a wide string.\n"), stderr);
			exit(EXIT_FAILURE);
		}

		wcs = xmalloc(sizeof(wchar_t) * ++len);
		len = mbstowcs(wcs, palette, len);
		if (len == (size_t) -1)
		{
			fputs(_("Failed to convert the supplied palette"
				" to a wide string.\n"), stderr);
			free(wcs);
			exit(EXIT_FAILURE);
		}

		waaSetPalette(g_ctx, wcs);
		free(wcs);
	}

	atexit(mainExit);
	if (mainLoop(g_ctx, argv[optind]))
		exit(EXIT_FAILURE);

	return 0;
}

static void mainExit (void)
{
	waaClose(g_ctx);
}

static void scaleInto (WaaRect *restrict rect,
	const WaaRect *restrict borders)
{
	double ratioBorders, ratioRect;

	ratioBorders = (double) borders->w / borders->h;
	ratioRect    = (double) rect->w    / rect->h;

	rect->x = borders->x;
	rect->y = borders->y;

	/* If the target area is wider, recompute the height. */
	if      (ratioBorders < ratioRect)
	{
		rect->h = (int) (borders->w / ratioRect);
		rect->w = borders->w;
	}
	/* If the source area is wider, recompute the width.  */
	else if (ratioBorders > ratioRect)
	{
		rect->w = (int) (borders->h * ratioRect);
		rect->h = borders->h;
	}
}

/** Approximate horizontal size of a character. */
#define W_COEFF 2
/** Approximate vertical size of a character. */
#define H_COEFF 3

static void drawImage (WaaContext *ctx, WaaSurface *img)
{
	WaaSurface *scaled;
	WaaRect imgRect, scaledRect, destRect;

	/* Console characters are tall. */
	SET_RECT(destRect, 0, 0, ctx->width * W_COEFF, ctx->height * H_COEFF);

	/* Stretch it over the whole console area. */
	SET_RECT(imgRect, 0, 0, img->width, img->height);
	scaledRect = imgRect;
	scaleInto(&scaledRect, &destRect);

	scaled = waaCreateSurface(WAA_FORMAT_RGB, scaledRect.w, scaledRect.h);
	waauBilinearBlit(img, &imgRect, scaled, &scaledRect);

	waaClear(ctx, 0, L' ');

	/* Render the image in the center of the console. */
	SET_RECT(destRect,
		(ctx->width  - scaledRect.w / W_COEFF) / 2,
		(ctx->height - scaledRect.h / H_COEFF) / 2,
		W_COEFF, H_COEFF);
	waaRender(ctx, scaled, &scaledRect, &destRect);

	waaFreeSurface(scaled);
}

static int mainLoop (WaaContext *ctx, const char *file)
{
	WaaSurface *img;

	img = waauLoadPNG(file);
	if (!img)
	{
		fprintf(stderr, _("Failed to load \"%s\".\n"), file);
		exit(EXIT_FAILURE);
	}

	drawImage(ctx, img);
	waaFlush(ctx);

	while (1)
	{
		int event;

		event = waaGetEvent(ctx, -1);

		if (event == WAA_RESIZE)
		{
			drawImage(ctx, img);
			waaFlush(ctx);
		}
		else if (event != WAA_NONE)
			break;
	}

	waaFreeSurface(img);
	return 0;
}

