/**
 *  @file waa.c
 *  @brief ASCII art library.
 *
 *  @note Copyright Přemysl Janouch 2010. All right reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <wchar.h>
#include <assert.h>

#include "config.h"

#include "xalloc.h"
#include "waa.h"


/** Convert RGB to gray. */
#define RGB_TO_G(r, g, b) ((r) * 0.299 + (g) * 0.587 + (b) * 0.114)

/** Get console dimensions and initialize the text and attribute buffers. */
static void initBuffers (WaaContext *ctx);


WaaSurface *waaCreateSurface (WaaPixelFormat fmt, int width, int height)
{
	WaaSurface *surface;

	assert(width > 0 && height > 0);

	surface = xmalloc(sizeof(WaaSurface));
	surface->format = fmt;
	surface->width = width;
	surface->height = height;
	surface->pitch = (width * waaGetBytesPerPixel(fmt) + 3) & ~3;
	surface->pixels = xmalloc(height * surface->pitch);

	surface->clipRect.x = 0;
	surface->clipRect.y = 0;
	surface->clipRect.w = width;
	surface->clipRect.h = height;

	return surface;
}

void waaFreeSurface (WaaSurface *surface)
{
	assert(surface != NULL);

	free(surface->pixels);
	free(surface);
}

void waaGetRGBA (uint32_t value, WaaPixelFormat fmt, uint8_t *restrict r,
	uint8_t *restrict g, uint8_t *restrict b, uint8_t *restrict a)
{
	assert(r && g && b && a);

	switch (fmt)
	{
	case WAA_FORMAT_GRAY:
		*r = *g = *b = value & 0xFF;
		*a = WAA_ALPHA_OPAQUE;
		break;
	case WAA_FORMAT_GRAY_ALPHA:
		*r = *g = *b = value & 0xFF;
		*a = (value >> 8) & 0xFF;
		break;
	case WAA_FORMAT_RGB:
		*a = WAA_ALPHA_OPAQUE;
		*b = (value >> 16) & 0xFF;
		*g = (value >> 8)  & 0xFF;
		*r =  value        & 0xFF;
		break;
	case WAA_FORMAT_RGB_ALPHA:
		*a = (value >> 24) & 0xFF;
		*b = (value >> 16) & 0xFF;
		*g = (value >> 8)  & 0xFF;
		*r =  value        & 0xFF;
		break;
	default:
		/* Should not be reached. */
		exit(EXIT_FAILURE);
	}
}

uint32_t waaMapRGBA (WaaPixelFormat fmt,
	uint8_t r, uint8_t g, uint8_t b, uint8_t a)
{
	switch (fmt)
	{
	case WAA_FORMAT_GRAY:
		return (uint8_t) RGB_TO_G(r, g, b);
	case WAA_FORMAT_GRAY_ALPHA:
		return (uint8_t) RGB_TO_G(r, g, b) | (a << 8);
	case WAA_FORMAT_RGB:
		return             (b << 16) | (g << 8) | r;
	case WAA_FORMAT_RGB_ALPHA:
		return (a << 24) | (b << 16) | (g << 8) | r;
	default:
		/* Should not be reached. */
		exit(EXIT_FAILURE);
	}
}

uint32_t waaGetPixel (WaaSurface *surface, int x, int y)
{
	/* Here p is the address to the pixel we want to retrieve. */
	uint8_t *p;

	assert(surface != NULL);

	p = (uint8_t *) surface->pixels + y * surface->pitch
		+ x * waaGetBytesPerPixel(surface->format);

	/* The pixel must be inside the clipping area. */
	if (x < surface->clipRect.x || y < surface->clipRect.y
		|| x >= surface->clipRect.x + surface->clipRect.w
		|| y >= surface->clipRect.y + surface->clipRect.h)
		return -1;

	assert(x >= 0 && y >= 0);

	switch (waaGetBytesPerPixel(surface->format))
	{
	case 1:
		return *p;
	case 2:
		return *(uint16_t *) p;
	case 3:
		if (WAA_BYTEORDER == WAA_BIG_ENDIAN)
			return p[0] << 16 | p[1] << 8 | p[2];
		else
			return p[0]       | p[1] << 8 | p[2] << 16;
	case 4:
		return *(uint32_t *) p;
	default:
		/* Shouldn't happen, but avoids warnings. */
		return -1;
	}
}

void waaPutPixel (WaaSurface *surface, int x, int y, uint32_t pixel)
{
	/* Here p is the address to the pixel we want to set. */
	uint8_t *p;

	assert(surface != NULL);

	p = (uint8_t *) surface->pixels + y * surface->pitch + x
		* waaGetBytesPerPixel(surface->format);

	/* The pixel must be inside the clipping area. */
	if (x < surface->clipRect.x || y < surface->clipRect.y
		|| x >= surface->clipRect.x + surface->clipRect.w
		|| y >= surface->clipRect.y + surface->clipRect.h)
		return;

	assert(x >= 0 && y >= 0);

	switch (waaGetBytesPerPixel(surface->format))
	{
	case 1:
		*p = pixel;
		break;
	case 2:
		*(uint16_t *) p = pixel;
		break;
	case 3:
		if (WAA_BYTEORDER == WAA_BIG_ENDIAN)
		{
			p[0] = (pixel >> 16) & 0xFF;
			p[1] = (pixel >> 8)  & 0xFF;
			p[2] =  pixel        & 0xFF;
		}
		else
		{
			p[0] =  pixel        & 0xFF;
			p[1] = (pixel >> 8)  & 0xFF;
			p[2] = (pixel >> 16) & 0xFF;
		}
		break;
	case 4:
		*(uint32_t *) p = pixel;
		break;
	}
}

void waaFillRect (WaaSurface *surface, const WaaRect *rect, uint32_t pixel)
{
	WaaRect r;
	int x, y;

	assert(surface != NULL);

	if (rect)
		r = *rect;
	else
	{
		r.x = 0;
		r.y = 0;
		r.w = surface->width;
		r.h = surface->height;
	}

	for (x = r.x; x < r.x + r.w; x++)
		for (y = r.y; y < r.y + r.h; y++)
			waaPutPixel(surface, x, y, pixel);
}

/* TODO: Options to e.g. disable handling of keyboard. */
WaaContext *waaInit (void)
{
	WaaContext *ctx;

	ctx = xmalloc(sizeof(WaaContext));
/* TODO: Resolve it in a smarter way. */
#ifdef WAA_BUILD_DRIVER_WIN32
	ctx->driver = &waaDriverWin32;
#elif defined (WAA_BUILD_DRIVER_UNIX)
	ctx->driver = &waaDriverUnix;
#else /* ! WAA_BUILD_DRIVER_UNIX */
	#error No driver for WAA available.
#endif /* ! WAA_BUILD_DRIVER_UNIX */

	if (ctx->driver->init(&ctx->params, &ctx->driverData))
	{
		free(ctx);
		return NULL;
	}

	initBuffers(ctx);
	ctx->palette = xmalloc(sizeof(wchar_t) * 256);
	waaSetPalette(ctx, NULL);

	return ctx;
}

static void initBuffers (WaaContext *ctx)
{
	size_t size;

	ctx->driver->getSize(ctx, &ctx->width, &ctx->height);
	size = ctx->width * ctx->height;
	ctx->text  = xmalloc(sizeof(wchar_t) * size);
	ctx->attrs = xmalloc(sizeof(int)     * size);

	waaClear(ctx, 0, L' ');
}

void waaClose (WaaContext *ctx)
{
	ctx->driver->uninit(ctx);
	free(ctx->text);
	free(ctx->attrs);
	free(ctx->palette);
	free(ctx);
}


const wchar_t *waaPaletteBlock = L" ·░▒█";
const wchar_t *waaPaletteAscii = L" .:;t%SX8@";

void waaSetPalette (WaaContext *restrict ctx,
	const wchar_t *restrict palette)
{
	size_t len;
	unsigned i;

	assert(ctx != NULL);

	if (!palette || !*palette)
		palette = L"?";

	len = wcslen(palette) - 1;
	for (i = 0; i < 256; i++)
		ctx->palette[i] = palette[(int) (i / 255.0 * len + 0.5)];
}


int waaRender (WaaContext *ctx, WaaSurface *surface,
	const WaaRect *rect, const WaaRect *params)
{
	int x, xc, xg, y, yc, yg;
	int wc, hc;

	assert(ctx != NULL && surface != NULL && rect != NULL && params != NULL);
	assert(params->w >= 0 && params->h >= 0);
	assert(params->x >= 0 && params->y >= 0);

	/* The width and height of the area where the surface will be rendered. */
	wc = rect->w / params->w;
	hc = rect->h / params->h;

	/* Check bounds and fix the area eventually. */
	if (params->x + wc >= ctx->width)
		wc = ctx->width  - params->x;
	if (params->y + hc >= ctx->height)
		hc = ctx->height - params->y;

	/* Iterate over the console area. */
	for     (xc = xg = 0; xc < wc; xc++, xg += params->w)
	{
		for (yc = yg = 0; yc < hc; yc++, yg += params->h)
		{
			int r = 0, g = 0, b = 0, n = 0;
			int index;

			/* Iterate over the surface. */
			for     (x = 0; x < params->w && xg + x < rect->w; x++)
			{
				for (y = 0; y < params->h && yg + y < rect->h; y++)
				{
					uint8_t rl, gl, bl, al;

					waaGetRGBA(waaGetPixel(surface, xg + x, yg + y),
						surface->format, &rl, &gl, &bl, &al);
					r += rl;
					g += gl;
					b += bl;
					n++;
				}
			}

			/* If wc and hc is computed correctly, this should not happen. */
			assert(n != 0);

			index = (params->x + xc) + (params->y + yc) * ctx->width;

			ctx->attrs[index] = 0;
			ctx->text [index] = ctx->palette
				[(int) (RGB_TO_G(r, g, b) / n + 0.5)];
		}
	}
	return 0;
}

int waaPrintf (WaaContext *restrict ctx, int x, int y, uint16_t attr,
	const wchar_t *restrict format, ...)
{
	va_list ap;
	size_t size;
	wchar_t *buffer;
	int r;

	size = ctx->width * (ctx->height - y) - x + 1;
	buffer = xmalloc(size * sizeof(wchar_t));

	va_start(ap, format);
	r = vswprintf(buffer, size, format, ap);
	va_end(ap);

	if (r > 0)
		waaPuts(ctx, x, y, attr, buffer);

	free(buffer);
	return r;
}

void waaPuts (WaaContext *restrict ctx, int x, int y, uint16_t attr,
	const wchar_t *restrict s)
{
	size_t len, size;
	unsigned start, i;

	len = wcslen(s);
	start = x + y * ctx->width;
	size = ctx->width * ctx->height;

	/* Check bounds, truncate. */
	if (start >= size)
		return;
	if (start + len >= size)
		len = size - start;

	wcsncpy(ctx->text + start, s, len);
	for (i = 0; i < len; i++)
		ctx->attrs[start + i] = attr;
}

void waaClear (WaaContext *ctx, uint16_t attr, wchar_t ch)
{
	size_t size;

	size = ctx->width * ctx->height;
	wmemset(ctx->text, ch, size);
	memset(ctx->attrs, 0, sizeof(int) * size);
}

void waaFlush (WaaContext *ctx)
{
	ctx->driver->flush(ctx);
}

int waaGetEvent (WaaContext *ctx, int milliseconds)
{
	int event;

	event = ctx->driver->getEvent(ctx, milliseconds);

	if (event == WAA_RESIZE)
	{
		free(ctx->text);
		free(ctx->attrs);
		initBuffers(ctx);
	}
	return event;
}
