/**
 *  @file waa.h
 *  @brief ASCII art library.
 *
 *  @note Copyright Přemysl Janouch 2010. All right reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __WAA_H__
#define __WAA_H__

/* === Generic definitions ================================================= */

/** The most significant byte is stored last. */
#define WAA_BIG_ENDIAN    1234
/** The most significant byte is stored first. */
#define WAA_LITTLE_ENDIAN 4321

/** The endianity of this current machine. */
#ifdef WORDS_BIGENDIAN
	#define WAA_BYTEORDER WAA_BIG_ENDIAN
#else /* ! WORDS_BIGENDIAN */
	#define WAA_BYTEORDER WAA_LITTLE_ENDIAN
#endif /* ! WORDS_BIGENDIAN */

/** A rectangle. */
typedef struct
{
	/** Coordinates for the top-left corner. */
	int x, y;
	/** Dimensions of the retangle. */
	int w, h;
}
WaaRect;

/* === Generic graphics ==================================================== */

#define WAA_RGB_BUILDCOLOR(r, g, b) \
	((r) | (g) << 8 | (b) << 16)
#define WAA_RGB_BUILDCOLOR_A(r, g, b, a) \
	((r) | (g) << 8 | (b) << 16 | (a) << 24)

#define WAA_RGB_GETR(x)  ((x)        & 0xFF)
#define WAA_RGB_GETG(x) (((x) >> 8)  & 0xFF)
#define WAA_RGB_GETB(x) (((x) >> 16) & 0xFF)
#define WAA_RGB_GETA(x) (((x) >> 24) & 0xFF)

#define WAA_RGB_RMASK 0x000000FF
#define WAA_RGB_GMASK 0x0000FF00
#define WAA_RGB_BMASK 0x00FF0000
#define WAA_RGB_AMASK 0xFF000000

#define WAA_ALPHA_TRANSPARENT 255
#define WAA_ALPHA_OPAQUE 0

/** How the pixels are stored. */
typedef enum
{
	/** Gray (8 bits). */
	WAA_FORMAT_GRAY,
	/** Gray with alpha (16 bits). */
	WAA_FORMAT_GRAY_ALPHA,
	/** BGR (24 bits). */
	WAA_FORMAT_RGB,
	/** ABGR (32 bits). */
	WAA_FORMAT_RGB_ALPHA
}
WaaPixelFormat;

/** Returns how many bytes define a pixel.
 *  @param[in] fmt  Pixel format.
 */
inline static size_t waaGetBytesPerPixel (WaaPixelFormat fmt)
{
	switch (fmt)
	{
	case WAA_FORMAT_GRAY:
		return 1;
	case WAA_FORMAT_GRAY_ALPHA:
		return 2;
	case WAA_FORMAT_RGB:
		return 3;
	case WAA_FORMAT_RGB_ALPHA:
		return 4;
	default:
		/* Should not be reached. */
		exit(EXIT_FAILURE);
	}
}

/** A surface one can paint on. */
typedef struct
{
	/** RAW surface pixel data. */
	void *pixels;
	/** How the pixels are stored. */
	WaaPixelFormat format;

	/** Dimensions of the surface. */
	int width, height;
	/** The width of a line in bytes. */
	size_t pitch;

	/** Clipping rectangle. */
	WaaRect clipRect;
}
WaaSurface;

/** Create a surface. The pixel data is initially uninitialized.
 *  @param[in] fmt  Pixel format for the surface.
 *  @param[in] width  Width of the surface.
 *  @param[in] height  Height of the surface.
 *  @return A new surface object.
 */
WaaSurface *waaCreateSurface (WaaPixelFormat fmt, int width, int height);

/** Free a surface and it's pixel buffer.
 *  @param[in] surface  The surface to be destroyed.
 */
void waaFreeSurface (WaaSurface *surface);

/** Decompose a pixel value.
 *  @param[in] value  A pixel value.
 *  @param[in] fmt  Pixel format of the value.
 *  @param[out] r  Where the red component will be stored.
 *  @param[out] g  Where the green component will be stored.
 *  @param[out] b  Where the blue component will be stored.
 *  @param[out] a  Where the alpha component will be stored.
 */
void waaGetRGBA (uint32_t value, WaaPixelFormat fmt, uint8_t *restrict r,
	uint8_t *restrict g, uint8_t *restrict b, uint8_t *restrict a);

/** Compose a pixel value.
 *  @param[in] fmt  Pixel format for the value to be returned.
 *  @param[in] r  The red component.
 *  @param[in] g  The green component.
 *  @param[in] b  The blue component.
 *  @param[in] a  The alpha component.
 *  @return A pixel value composed from all the single components.
 */
uint32_t waaMapRGBA (WaaPixelFormat fmt,
	uint8_t r, uint8_t g, uint8_t b, uint8_t a);

/** Get the raw value of a pixel.
 *  @param[in] surface  A surface object.
 *  @param[in] x  The X coordinate.
 *  @param[in] y  The Y coordinate.
 *  @return The pixel value at the specified coordinates on the surface.
 */
uint32_t waaGetPixel (WaaSurface *surface, int x, int y);

/** Set the raw value of a pixel.
 *  @param[in] surface  A surface object.
 *  @param[in] x  The X coordinate.
 *  @param[in] y  The Y coordinate.
 *  @param[in] pixel  The raw pixel value.
 */
void waaPutPixel (WaaSurface *surface, int x, int y, uint32_t pixel);

/** Fill the given rectangle with some color.
 *  @param[in] surface  A surface object.
 *  @param[in] rect  The area to be filled. NULL ptr. means the whole surface.
 *  @param[in] pixel  A raw pixel value.
 */
void waaFillRect (WaaSurface *surface, const WaaRect *rect, uint32_t pixel);

/* === pre-typedef the world =============================================== */

typedef struct WaaContext WaaContext;
typedef struct WaaDriver WaaDriver;
typedef struct WaaDriverParams WaaDriverParams;

/* === Console attributes ================================================== */

#define WAA_ATTR_BRIGHT (1 << 0)

/* === Driver interface ==================================================== */

/** Parameters for the driver. */
struct WaaDriverParams
{
	/** Mask of supported attributes. */
	uint16_t attributes;
};

/** Driver initialization. */
typedef int (*WaaDriverInit) (WaaDriverParams *params, void **driverData);
/** Driver deinitialization. */
typedef void (*WaaDriverUninit) (WaaContext *ctx);
/** Get available dimensions. */
typedef void (*WaaDriverGetSize) (WaaContext *ctx, int *w, int *h);
/** Flush the image. */
typedef void (*WaaDriverFlush) (WaaContext *ctx);
/** Check for pending events. */
typedef int (*WaaDriverGetEvent) (WaaContext *ctx, int milliseconds);

/** Interface for a driver. */
struct WaaDriver
{
	/** The name of this driver. */
	const char *name;
	/** Driver initialization. */
	WaaDriverInit init;
	/** Driver deinitialization. */
	WaaDriverUninit uninit;
	/** Get available dimensions. */
	WaaDriverGetSize getSize;
	/** Flush the image. */
	WaaDriverFlush flush;
	/** Check for pending events. */
	WaaDriverGetEvent getEvent;
};

#ifdef WAA_BUILD_DRIVER_WIN32
extern WaaDriver waaDriverWin32;
#endif /* WAA_BUILD_DRIVER_WIN32 */

#ifdef WAA_BUILD_DRIVER_UNIX
extern WaaDriver waaDriverUnix;
#endif /* WAA_BUILD_DRIVER_UNIX */

/* === Console context ===================================================== */

/** Console context. */
struct WaaContext
{
	/** The driver used by this context. */
	WaaDriver *driver;
	/** Driver parameters. */
	WaaDriverParams params;
	/** A two-dimensional array of characters. */
	wchar_t *text;
	/** A two-dimensional array of attributes for @a text. */
	uint16_t *attrs;
	/** Rendering palette (table of 256 items). */
	wchar_t *palette;
	/** The width of @a text and @a attrs. */
	int width;
	/** The height of @a text and @a attrs. */
	int height;
	/** Private data for the driver. */
	void *driverData;
};

/** Get a context.
 *  @return A new context object.
 */
WaaContext *waaInit (void);

/** Destroy the context.
 *  @param[in] ctx  The context object to be destroyed.
 */
void waaClose (WaaContext *ctx);


/** Default Unicode blocky palette. */
extern const wchar_t *waaPaletteBlock;

/** Default ASCII palette. */
extern const wchar_t *waaPaletteAscii;

/** Set the rendering palette. If the palette string is either NULL
 *  or 0 characters long, the function will set the palette to L"?".
 *  @param[in,out] ctx  A console context object.
 *  @param[in] palette  A wide string that forms the palette.
 */
void waaSetPalette (WaaContext *restrict ctx,
	const wchar_t *restrict palette);


/** Render a surface onto the context buffer.
 *  @param[in,out] ctx  A console context object.
 *  @param[in] surface  The surface to be rendered.
 *  @param[in] rect  The area on the surface to be rendered.
 *  @param[in] params  The @a x and @a y members represent placement
 *  	on the screen, the @a w and @a h members represent how many
 *  	horizontal and vertical pixels form a character.
 *  @return The number of characters written or -1 on error.
 */
int waaRender (WaaContext *ctx, WaaSurface *surface,
	const WaaRect *rect, const WaaRect *params);

/** Flush the context buffer and show it to the user.
 *  @param[in] ctx  A console context object.
 */
void waaFlush (WaaContext *ctx);

/** Print some formatted text onto the context buffer.
 *  @param[in,out] ctx  A console context object.
 *  @param[in] x  Horizontal placement on the screen.
 *  @param[in] y  Vertical placement on the screen.
 *  @param[in] attr  Text attributes.
 *  @param[in] format  A wide-character format string.
 */
int waaPrintf (WaaContext *restrict ctx, int x, int y, uint16_t attr,
	const wchar_t *restrict format, ...);

/** Print text onto the context buffer.
 *  @param[in,out] ctx  A console context object.
 *  @param[in] x  Horizontal placement on the screen.
 *  @param[in] y  Vertical placement on the screen.
 *  @param[in] attr  Text attributes.
 *  @param[in] s  The text to be displayed.
 */
void waaPuts (WaaContext *restrict ctx, int x, int y, uint16_t attr,
	const wchar_t *restrict s);

/** Fill the whole console buffer with the given character and attribute values.
 *  @param[in,out] ctx  A console context object.
 *  @param[in] attr  An attribute value.
 *  @param[in] ch  A wide character.
 */
void waaClear (WaaContext *ctx, uint16_t attr, wchar_t ch);

/* === Console events ====================================================== */

/** No event has occured. */
#define WAA_NONE   256

/** The console has been resized and the text
 *  and attribute buffers have been cleared.
 */
#define WAA_RESIZE 257

/** Check for pending events.
 *  @param[in] ctx  A console context object.
 *  @param[in] milliseconds  How many milliseconds to wait for an event.
 *  The value of -1 means wait forever. There is however no guarantee that
 *  the function will not return WAA_NONE.
 *  @return The above constants or < 256 on keyboard events (ASCII).
 */
int waaGetEvent (WaaContext *ctx, int milliseconds);


#endif /* __WAA_H__ */

