/**
 *  @file waaunix.c
 *  @brief ASCII art library: Unix driver.
 *
 *  @note Copyright Přemysl Janouch 2010. All right reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>

#include <unistd.h>
#include <signal.h>
#include <termios.h>
#include <poll.h>
#include <fcntl.h>

#include "config.h"

/* ioctl() is located in unistd.h on some systems. */
#ifdef HAVE_SYS_IOCTL_H
	#include <sys/ioctl.h>
#endif /* HAVE_SYS_IOCTL_H */

#include "xalloc.h"
#include "waa.h"


/** Unix driver initialization. */
static int unixInit (WaaDriverParams *params, void **driverData);
/** Unix driver deinitialization. */
static void unixUninit (WaaContext *ctx);
/** Unix driver terminal dimensions. */
static void unixGetSize (WaaContext *ctx, int *width, int *height);
/** Unix driver flushing. */
static void unixFlush (WaaContext *ctx);
/** Unix driver event checking. */
static int unixGetEvent (WaaContext *ctx, int milliseconds);


WaaDriver waaDriverUnix =
{
	"unix",
	unixInit,
	unixUninit,
	unixGetSize,
	unixFlush,
	unixGetEvent
};


static int unixInitialized = 0;
static int unixInstances = 0;
static struct termios unixPrevAttrs;

/* This is probably going to be 0 and 1,
 * but we'll resolve stdin and stdout anyway.
 */
static int unixIn, unixOut;

#define CSI "\033["

#define WRITE_BYTES(b, len) write(unixOut, (b), (len))
#define WRITE_STRING(s) write(unixOut, (s), strlen(s))

#ifdef SIGWINCH
	#define N_POLL_FDS 2
static int unixPipeFds[2];
static struct sigaction unixPrevHandler;

static void unixSigWinchHandler (int signo)
{
	static char ch = '\0';

	/* Flush the pipe first. */
	read (unixPipeFds[0], &ch, sizeof(char));

	/* Write a single byte into the pipe. */
	write(unixPipeFds[1], &ch, sizeof(char));
}
#else /* ! SIGWINCH */
	#define N_POLL_FDS 1
#endif /* ! SIGWINCH */

static struct pollfd unixFds[N_POLL_FDS];


static int unixInit (WaaDriverParams *params, void **driverData)
{
	fflush(stdout);

	if (!unixInitialized)
	{
		struct termios t;
#ifdef SIGWINCH
		struct sigaction sa;

		if (sigaction(SIGWINCH, NULL, &unixPrevHandler))
			return -1;
		if (pipe(unixPipeFds))
			return -1;
		if (fcntl(unixPipeFds[0], F_SETFL, O_NONBLOCK) ||
			fcntl(unixPipeFds[1], F_SETFL, O_NONBLOCK))
		{
			close(unixPipeFds[0]);
			close(unixPipeFds[1]);
			return -1;
		}

		/* If this signal interrupts the poll(), restart it. */
		sa.sa_flags = SA_RESTART;
		sa.sa_handler = unixSigWinchHandler;
		sigemptyset(&sa.sa_mask);
		if (sigaction(SIGWINCH, &sa, NULL))
			return -1;
#endif /* SIGWINCH */

		unixIn  = fileno(stdin);
		unixOut = fileno(stdout);

		if (tcgetattr(unixIn, &t))
		{
#ifdef SIGWINCH
			sigaction(SIGWINCH, &unixPrevHandler, NULL);
			close(unixPipeFds[0]);
			close(unixPipeFds[1]);
#endif /* SIGWINCH */
			return -1;
		}
		unixPrevAttrs = t;

		/* Don't wait for anything, don't echo anything. */
		t.c_lflag &= ~(ECHO | ICANON);
		t.c_cc[VTIME] = 0;
		t.c_cc[VMIN]  = 1;
		if (tcsetattr(unixIn, TCSANOW, &t))
		{
#ifdef SIGWINCH
			sigaction(SIGWINCH, &unixPrevHandler, NULL);
			close(unixPipeFds[0]);
			close(unixPipeFds[1]);
#endif /* SIGWINCH */
			return -1;
		}

		unixFds[0].fd = unixIn;
		unixFds[0].events = POLLIN;
#ifdef SIGWINCH
		unixFds[1].fd = unixPipeFds[0];
		unixFds[1].events = POLLIN;
#endif /* SIGWINCH */

		/* TODO: Terminal:
		 *  a. Use term.h if available? setupterm(), tgetstr(), tparm(), ...
		 *     opengroup.org/onlinepubs/007908799/xcurses/tigetstr.html
		 *     Write a macro that uses either terminfo or a given string.
		 *  b. We should use curses directly. Maybe let's do that.
		 */

		/* Hide the cursor. */
		WRITE_STRING(CSI "?25l");

		unixInitialized = 1;
	}

	params->attributes = WAA_ATTR_BRIGHT;
	*driverData = NULL;

	unixInstances++;
	return 0;
}

static void unixUninit (WaaContext *ctx)
{
	if (!unixInitialized)
		return;

	if (!--unixInstances)
	{
		/* Show the cursor. */
		WRITE_STRING(CSI "?25h");

#ifdef SIGWINCH
		if (sigaction(SIGWINCH, &unixPrevHandler, NULL))
		{
			fputs(_("Failed to restore signal handler.\n"), stderr);
			exit(EXIT_FAILURE);
		}

		close(unixPipeFds[0]);
		close(unixPipeFds[1]);
#endif /* SIGWINCH */

		if (tcsetattr(unixIn, TCSANOW, &unixPrevAttrs))
		{
			fputs(_("Failed to restore terminal flags.\n"), stderr);
			exit(EXIT_FAILURE);
		}

		unixInitialized = 0;
	}
}

static void unixGetSize (WaaContext *ctx, int *width, int *height)
{
	struct winsize ws;

/* XXX: If neither the first nor the second is defined,
 *      the "on error" block below gets executed.
 */
#ifdef TIOCGSIZE
	if (ioctl(unixOut, TIOCGSIZE,  &ws) == -1)
#elif defined (TIOCGWINSZ)
	if (ioctl(unixOut, TIOCGWINSZ, &ws) == -1)
#endif /* TIOCGWINSZ */
	{
		/* Assume the default terminal. */
		*width  = 80;
		*height = 25;
		return;
	}

	*width  = ws.ws_col;
	*height = ws.ws_row;
}

static void unixWriteChar (wchar_t c)
{
	char mb[MB_CUR_MAX];
	int len;

	/* TODO: Also check for \a, \033 etc. */
	if (c == L'\0')
		len = 0;
	else
		len = wctomb(mb, c);

	if (len < 1)
		WRITE_STRING("?");
	else
		WRITE_BYTES(mb, len);
}

static void unixFlush (WaaContext *ctx)
{
	uint16_t attr = 0, attrCur;
	int i;

	/* Get to the upper left corner and reset attributes. */
	WRITE_STRING(CSI "H");
	WRITE_STRING(CSI "0m");

	for (i = 0; i < ctx->width * ctx->height; i++)
	{
		attrCur = ctx->attrs[i];
		if (attrCur != attr)
		{
			if (attrCur & WAA_ATTR_BRIGHT)
				WRITE_STRING(CSI "1m");
			else
				WRITE_STRING(CSI "0m");
			attr = attrCur;
		}
		unixWriteChar(ctx->text[i]);
	}
	WRITE_STRING(CSI "0m");
}

static int unixGetEvent (WaaContext *ctx, int milliseconds)
{
	unsigned char ch;

	if (poll(unixFds, N_POLL_FDS, milliseconds) <= 0)
		return WAA_NONE;

#ifdef SIGWINCH
	if (unixFds[1].revents & POLLIN)
	{
		read(unixPipeFds[0], &ch, 1);
		return WAA_RESIZE;
	}
#endif /* SIGWINCH */

	/* FIXME: What about cursor keys, multibyte chars etc.? */
	if (unixFds[0].revents & POLLIN)
		if (read(unixIn, &ch, 1) == 1)
			return ch;

	return WAA_NONE;
}
