/**
 *  @file waautil.c
 *  @brief WAA helper functions.
 *
 *  @note Copyright Přemysl Janouch 2010. All right reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>
#include <stdio.h>

#include <png.h>
#include <setjmp.h>
#include <math.h>

#include "config.h"

#include "xalloc.h"
#include "waa.h"
#include "waautil.h"


void waauRectClip (WaaRect *restrict r1, const WaaRect *restrict r2)
{
	if (r1->x < r2->x)
	{
		r1->w -= r2->x - r1->x;
		r1->x = r2->x;
	}

	if (r1->w > r2->w)
		r1->w = r2->w;

	if (r1->y < r2->y)
	{
		r1->h -= r2->y - r1->y;
		r1->y = r2->y;
	}

	if (r1->h > r2->h)
		r1->h = r2->h;
}


void waauPutPixelAlpha (WaaSurface *surface, int x, int y, uint32_t pixel)
{
	uint8_t ro, go, bo, ao;
	float rorig, rnew;

	rnew = (float) WAA_RGB_GETA(pixel) / 255;
	rorig = 1.0 - rnew;

	waaGetRGBA(waaGetPixel(surface, x, y), surface->format,
		&ro, &go, &bo, &ao);
	waaPutPixel(surface, x, y, waaMapRGBA(surface->format,
		(int) (ro * rorig + WAA_RGB_GETR(pixel) * rnew),
		(int) (go * rorig + WAA_RGB_GETG(pixel) * rnew),
		(int) (bo * rorig + WAA_RGB_GETB(pixel) * rnew),
		WAA_ALPHA_OPAQUE));
}


void waauDrawRectangle (WaaSurface *surface,
	const WaaRect *rect, uint32_t pixel)
{
	int x, y;
	float ratio;

	if (WAA_RGB_GETA(pixel) != WAA_ALPHA_OPAQUE)
	{
		ratio = (float) WAA_RGB_GETA(pixel) / 255;

		for     (x = rect->x; x < rect->w + rect->x; x++)
			for (y = rect->y; y < rect->h + rect->y; y++)
				waaPutPixel(surface, x, y, waauRGBBlendColor(
					waaGetPixel(surface, x, y), pixel, ratio));
	}
	else
	{
		for     (x = rect->x; x < rect->w + rect->x; x++)
			for (y = rect->y; y < rect->h + rect->y; y++)
				waaPutPixel(surface, x, y, pixel);
	}
}


static void pngUserRead (png_structp pngp, png_bytep data, png_size_t length)
{
	if (fread(data, length, 1, png_get_io_ptr(pngp)) != 1)
		png_error(pngp, _("Reading from the input stream failed."));
}

WaaSurface *waauLoadPNG (const char *filename)
{
	FILE *fp = NULL;
	unsigned char hdr[8];
	png_structp pngp = NULL;
	png_infop infop = NULL, endp = NULL;
	png_bytep *rowPtrs;

	WaaSurface *res = NULL;
	png_bytep p;
	png_byte colorType;
	int width, height;
	int x, y;

	if (!(fp = fopen(filename, "rb")))
		goto loadpng_error;

	/* Check the signature. */
	if (fread(hdr, 8, 1, fp) != 1 || png_sig_cmp(hdr, 0, 8))
		goto loadpng_error;

	/* Create internal structures. */
	if (!(pngp = png_create_read_struct(PNG_LIBPNG_VER_STRING,
		NULL, NULL, NULL)))
		goto loadpng_error;
	if (!(infop = png_create_info_struct(pngp)))
		goto loadpng_error;
	if (!(endp = png_create_info_struct(pngp)))
		goto loadpng_error;

	/* Error handler. */
	if (setjmp(png_jmpbuf(pngp)))
		goto loadpng_error;

	/* Initialize IO. */
	/* XXX: If the libpng library is using the same fread() as we do,
	 *      png_init_io(pngp, fp); can be used instead of the following.
	 */
	png_set_read_fn(pngp, fp, pngUserRead);
	png_set_sig_bytes(pngp, 8);

	/* Read the image. */
	png_read_png(pngp, infop,
		PNG_TRANSFORM_EXPAND | PNG_TRANSFORM_STRIP_16, NULL);
	rowPtrs = png_get_rows(pngp, infop);

	width = png_get_image_width(pngp, infop);
	height = png_get_image_height(pngp, infop);

	colorType = png_get_color_type(pngp, infop);
	res = waaCreateSurface(WAA_FORMAT_RGB_ALPHA, width, height);

	if (colorType == PNG_COLOR_TYPE_RGB_ALPHA)
	{
		for (y = 0; y < height; y++)
			for (x = 0, p = rowPtrs[y]; x < width; x++, p += 4)
				((uint32_t *) ((char *) res->pixels + y * res->pitch))[x] =
					WAA_RGB_BUILDCOLOR_A(p[0], p[1], p[2], p[3]);
	}
	else
	{
		for (y = 0; y < height; y++)
			for (x = 0, p = rowPtrs[y]; x < width; x++, p += 3)
				((uint32_t *) ((char *) res->pixels + y * res->pitch))[x] =
					WAA_RGB_BUILDCOLOR_A(p[0], p[1], p[2], WAA_ALPHA_OPAQUE);
	}

	png_destroy_read_struct(&pngp, &infop, &endp);
	fclose(fp);
	return res;

loadpng_error:
	if (endp)
		png_destroy_read_struct(&pngp, &infop, &endp);
	else if (infop)
		png_destroy_read_struct(&pngp, &infop, NULL);
	else if (pngp)
		png_destroy_read_struct(&pngp, NULL, NULL);
	if (res)
		waaFreeSurface(res);
	if (fp)
		fclose(fp);
	return NULL;
}


/** Horizontal blit processing function prototype. The function should
 *  horizontally stretch one line into the buffer. */
typedef void (*bbHorizontalFun) (WaaSurface *, uint32_t *, int, int, int, int);

/** Linear interpolation function. */
static void bbHInterpolate (WaaSurface *source, uint32_t *buffer,
	int srcX, int srcY, int srcWidth, int tgtWidth);

/** Linear reduction function. */
static void bbHReduce (WaaSurface *source, uint32_t *buffer,
	int srcX, int srcY, int srcWidth, int tgtWidth);


/* TODO: Precompute ratioH and put the info into a structure.
 *       Probably use an inline function for this. */
static void bbHInterpolate (WaaSurface *source, uint32_t *buffer,
	int srcX, int srcY, int srcWidth, int tgtWidth)
{
	double ratioH;
	int x;

	ratioH = (double) (srcWidth - 1) / tgtWidth;
	for (x = 0; x < tgtWidth; x++)
	{
		double srcXX = srcX + x * ratioH;
		uint8_t r1, g1, b1, r2, g2, b2;
		uint8_t a;

		waaGetRGBA(waaGetPixel(source, (int) ceil(srcXX),  srcY),
			source->format, &r1, &g1, &b1, &a);
		waaGetRGBA(waaGetPixel(source, (int) floor(srcXX), srcY),
			source->format, &r2, &g2, &b2, &a);

		buffer[x] = waauRGBBlendColor(WAA_RGB_BUILDCOLOR(r1, g1, b1),
			WAA_RGB_BUILDCOLOR(r2, g2, b2), srcXX - floor(srcXX));
	}
}

static void bbHReduce (WaaSurface *source, uint32_t *buffer,
	int srcX, int srcY, int srcWidth, int tgtWidth)
{
	double ratioH;
	int r, g, b, n;
	uint8_t cr, cg, cb, ca;
	int x, myX, lastX;

	ratioH = (double) (srcWidth - 1) / tgtWidth;
	lastX = 0;
	r = g = b = n = 0;

	for (x = 0; x < srcWidth; x++)
	{
		waaGetRGBA(waaGetPixel(source, srcX + x, srcY), source->format,
			&cr, &cg, &cb, &ca);

		r += cr;
		g += cg;
		b += cb;
		n++;

		myX = (int) (x / ratioH + 0.5);
		if (myX != lastX)
		{
			buffer[lastX] = WAA_RGB_BUILDCOLOR(r / n, g / n, b / n);
			lastX = myX;
			r = g = b = n = 0;
		}
	}
}

/* This function is designed to be rather memory-effective (which is
 * not actually important these days) than fast. */
void waauBilinearBlit (WaaSurface *source, const WaaRect *srcRect,
	WaaSurface *target, const WaaRect *tgtRect)
{
	/* The buffer saves us from having to allocate image data for the whole
	 * horizontally stretched image. */
	bbHorizontalFun hfun;
	uint32_t *buffer;
	int bufLine = -1;
	unsigned int bufHeightMin, bufHeight, bufHeightMask;

	double ratioV;
	int x, y;

	/* TODO: Make an intersection of clipping rectangles and ^^. */

	/* The minimal height of vertical buffer. */
	bufHeightMin = ceil((double) srcRect->h / tgtRect->h) + 1;

	/* Find the nearest power of two that's at least as big,
	 * so we may replace the modulo operation with bitwise and. */
	bufHeight = 1;
	while (bufHeight < bufHeightMin)
		bufHeight <<= 1;
	bufHeightMask = bufHeight - 1;

	buffer = xmalloc(sizeof(uint32_t) * tgtRect->w * bufHeight);

	if (srcRect->w > tgtRect->w)
		hfun = bbHReduce;
	else
		hfun = bbHInterpolate;

	/* Vertical interpolation. */
	if (tgtRect->h > srcRect->h)
	{
		ratioV = (double) (srcRect->h - 1) / tgtRect->h;
		for (y = 0; y < tgtRect->h; y++)
		{
			double srcY = y * ratioV;
			int higherY = ceil(srcY);
			int lowerY = floor(srcY);

			while (bufLine < higherY)
			{
				bufLine++;
				hfun(source, buffer + tgtRect->w * (bufLine & bufHeightMask),
					srcRect->x, srcRect->y + bufLine, srcRect->w, tgtRect->w);
			}

			for (x = 0; x < tgtRect->w; x++)
			{
				/* XXX: The target surface must be RGB or RGBA. */
				waaPutPixel(target, tgtRect->x + x, tgtRect->y + y,
					waauRGBBlendColor(
					buffer[tgtRect->w * (higherY & bufHeightMask) + x],
					buffer[tgtRect->w * (lowerY  & bufHeightMask) + x],
					srcY - floor(srcY)));
			}
		}
	}
	/* Vertical height reduction. */
	else
	{
		int lowerY;
		int higherY = 0;

		ratioV = (double) srcRect->h / tgtRect->h;
		for (y = 0; y < tgtRect->h; y++)
		{
			lowerY = higherY;
			higherY = (y + 1) * ratioV;

			while (bufLine < higherY)
			{
				bufLine++;
				hfun(source, buffer + tgtRect->w * (bufLine & bufHeightMask),
					srcRect->x, srcRect->y + bufLine, srcRect->w, tgtRect->w);
			}

			for (x = 0; x < tgtRect->w; x++)
			{
				int r, g, b, n, srcY;
				uint32_t px;

				r = g = b = n = 0;

				for (srcY = lowerY; srcY < higherY; srcY++)
				{
					px = buffer[tgtRect->w * (srcY & bufHeightMask) + x];

					r += WAA_RGB_GETR(px);
					g += WAA_RGB_GETG(px);
					b += WAA_RGB_GETB(px);
					n++;
				}
				waaPutPixel(target, tgtRect->x + x, tgtRect->y + y,
					WAA_RGB_BUILDCOLOR(r / n, g / n, b / n));
			}
		}
	}

	free(buffer);
}

