/**
 *  @file waautil.h
 *  @brief WAA helper functions.
 *
 *  @note Copyright Přemysl Janouch 2010. All right reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __WAAUTIL_H__
#define __WAAUTIL_H__


/** Blend two colors in the specified ratio.
 *  @param[in] c1  The first color.
 *  @param[in] c2  The second color.
 *  @param[in] ratio  0.0 is the first color, 1.0 is the second.
 *  @return The blended color.
 */
inline static uint32_t waauRGBBlendColor (uint32_t c1, uint32_t c2, float ratio)
{
	float invtr = (1.0 - ratio);

	return WAA_RGB_BUILDCOLOR (
		(int) (WAA_RGB_GETR(c1) * ratio + WAA_RGB_GETR(c2) * invtr),
		(int) (WAA_RGB_GETG(c1) * ratio + WAA_RGB_GETG(c2) * invtr),
		(int) (WAA_RGB_GETB(c1) * ratio + WAA_RGB_GETB(c2) * invtr)
	);
}


/** This function clips @a r1 inside of @a r2.
 *  @param[in,out] r1  The recangle to clip.
 *  @param[in] r2  A rectangle that specifies the clipping area.
 */
void waauRectClip (WaaRect *restrict r1, const WaaRect *restrict r2);

/** Checks the relation between a rectangle and a point.
 *  @return Returns non-zero if the point given by the @a x and @a y
 *          coordinates is inside the given rectangle.
 */
inline static int waauPointInRect (const WaaRect *rect, int x, int y)
{
	return (x >= rect->x && y >= rect->y
		&& x < rect->x + rect->w && y < rect->y + rect->h);
}


/** Get the value of a pixel in a 32-bit surface.
 *  This function doesn't check the coordinates, therefore it is not safe.
 *  @return The value of the pixel or -1 on failure.
 */
static inline uint32_t waauGetPixel32Unsafe (WaaSurface *surface, int x, int y)
{
	return *(uint32_t *) ((uint8_t *) surface->pixels + y * surface->pitch
		+ x * waaGetBytesPerPixel(surface->format));
}

/** Draw a pixel. This alpha variant of waaPutPixel performs blending
 *  of the color that is on the surface with the given pixel value.
 */
void waauPutPixelAlpha (WaaSurface *surface, int x, int y, uint32_t pixel);


/** Draw a filled rectangle. */
void waauDrawRectangle (WaaSurface *surface,
	const WaaRect *rect, uint32_t pixel);


/** Load an image into a 32-bit surface.
 *  @param[in] filename  The filename of the image to be loaded.
 *  @return A surface object with the loaded image or NULL on failure.
 */
WaaSurface *waauLoadPNG (const char *filename);

/** Copy an area of the source surface onto the target surface.
 *  The source and target rects must not overlap the borders of their surfaces.
 */
void waauBilinearBlit (WaaSurface *source, const WaaRect *sourceRect,
	WaaSurface *target, const WaaRect *targetRect);


#endif /* __WAAUTIL_H__ */

