/**
 *  @file waawin32.c
 *  @brief ASCII art library: Win32 driver.
 *
 *  @note Copyright Přemysl Janouch 2010. All right reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#define UNICODE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <windows.h>

#include "config.h"

#include "xalloc.h"
#include "waa.h"


/** Win32 driver initialization. */
static int win32Init (WaaDriverParams *params, void **driverData);
/** Win32 driver deinitialization. */
static void win32Uninit (WaaContext *ctx);
/** Win32 driver terminal dimensions. */
static void win32GetSize (WaaContext *ctx, int *width, int *height);
/** Win32 driver flushing. */
static void win32Flush (WaaContext *ctx);
/** Win32 driver event checking. */
static int win32GetEvent (WaaContext *ctx, int milliseconds);


WaaDriver waaDriverWin32 =
{
	"win32",
	win32Init,
	win32Uninit,
	win32GetSize,
	win32Flush,
	win32GetEvent
};


static int win32Initialized = 0;
static int win32Instances = 0;

static int win32AllocatedConsole;
static DWORD win32PrevInputMode;
static HANDLE win32PrevOut;
static HANDLE win32In, win32Out;


static int win32Init (WaaDriverParams *params, void **driverData)
{
	fflush(stdout);

	if (!win32Initialized)
	{
		CONSOLE_CURSOR_INFO cci;
		CONSOLE_SCREEN_BUFFER_INFO csbi;
		COORD size;

		win32AllocatedConsole = AllocConsole();

		/* Get console input buffer. */
		win32In = CreateFile(TEXT("CONIN$"), GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (win32In == INVALID_HANDLE_VALUE)
			return -1;

		/* Set a new input mode. */
		GetConsoleMode(win32In, &win32PrevInputMode);
		if (!SetConsoleMode(win32In,
			ENABLE_WINDOW_INPUT | ENABLE_PROCESSED_INPUT))
			return -1;

		/* Get active console screen buffer. */
		win32PrevOut = CreateFile(TEXT("CONOUT$"), GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (win32PrevOut == INVALID_HANDLE_VALUE)
			return -1;

		/* Create our own screen buffer and set it active. */
		win32Out = CreateConsoleScreenBuffer(GENERIC_READ | GENERIC_WRITE,
			0, NULL, CONSOLE_TEXTMODE_BUFFER, NULL);
		if (win32Out == INVALID_HANDLE_VALUE)
			return -1;
		if (!SetConsoleActiveScreenBuffer(win32Out))
			return -1;

		/* Make screen buffer size equal to console window size. */
		/* If the user is smart enough to change the buffer size in the menu,
		 * he is sure smart enough to also resize the window manually. */
		if (!GetConsoleScreenBufferInfo(win32Out, &csbi))
			return -1;

		size.X = csbi.srWindow.Right - csbi.srWindow.Left + 1;
		size.Y = csbi.srWindow.Bottom - csbi.srWindow.Top + 1;
		if (!SetConsoleScreenBufferSize(win32Out, size))
			return -1;

		/* Turn off the cursor. */
		cci.dwSize = 100;
		cci.bVisible = FALSE;
		SetConsoleCursorInfo(win32Out, &cci);

		win32Initialized = 1;
	}

	params->attributes = WAA_ATTR_BRIGHT;
	*driverData = NULL;

	win32Instances++;
	return 0;
}

static void win32Uninit (WaaContext *ctx)
{
	if (!win32Initialized)
		return;

	if (!--win32Instances)
	{
		/* Activate the original screen buffer. */
		if (!SetConsoleActiveScreenBuffer(win32PrevOut))
		{
			fprintf(stderr, _("Failed to restore screen buffer.\n"));
			exit(EXIT_FAILURE);
		}
		CloseHandle(win32Out);
		CloseHandle(win32PrevOut);

		/* Set the input mode to previous values. */
		if (!SetConsoleMode(win32In, win32PrevInputMode))
		{
			fprintf(stderr, _("Failed to restore input mode.\n"));
			exit(EXIT_FAILURE);
		}
		CloseHandle(win32In);

		/* If we've allocated a console, free it. */
		if (win32AllocatedConsole)
			FreeConsole();

		win32Initialized = 0;
	}
}

static void win32GetSize (WaaContext *ctx, int *width, int *height)
{
	CONSOLE_SCREEN_BUFFER_INFO csbi;

	/* On failure this just does nothing. */
	if (!GetConsoleScreenBufferInfo(win32Out, &csbi))
		return;
	*width  = csbi.dwSize.X;
	*height = csbi.dwSize.Y;
}

static void win32Flush (WaaContext *ctx)
{
	uint16_t attr = 0, attrCur;
	WORD wattr = FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED;
	COORD pos;
	TCHAR tch;
	DWORD i, t;

	/* XXX: It would be better to write it all at once. */
	for (i = pos.Y = 0; pos.Y < ctx->height; pos.Y++)
	{
		for (pos.X = 0; pos.X < ctx->width;  pos.X++, i++)
		{
			tch     = ctx->text [i];
			attrCur = ctx->attrs[i];
			if (attrCur != attr)
			{
				if (attrCur & WAA_ATTR_BRIGHT)
					wattr |=  FOREGROUND_INTENSITY;
				else
					wattr &= ~FOREGROUND_INTENSITY;
				attr = attrCur;
			}
			WriteConsoleOutputCharacter(win32Out, &tch,   1, pos, &t);
			WriteConsoleOutputAttribute(win32Out, &wattr, 1, pos, &t);
		}
	}
}

static int win32GetEvent (WaaContext *ctx, int milliseconds)
{
	DWORD n;
	INPUT_RECORD event;

	/* On error, pretend there's nothing to read. */
	if (WaitForSingleObject(win32In,
		milliseconds < 0 ? INFINITE : milliseconds) != WAIT_OBJECT_0)
		return WAA_NONE;

	if (!ReadConsoleInput(win32In, &event, 1, &n) || !n)
		return WAA_NONE;

	/* If the event is not something we want to see, we recurse. */
	switch (event.EventType)
	{
	case KEY_EVENT:
		/* FIXME: What about cursor keys, multibyte chars etc.? */
		if (event.Event.KeyEvent.bKeyDown
			&& event.Event.KeyEvent.uChar.AsciiChar)
			return event.Event.KeyEvent.uChar.AsciiChar;
		else
			return win32GetEvent(ctx, 0);
	case WINDOW_BUFFER_SIZE_EVENT:
		return WAA_RESIZE;
	default:
		return win32GetEvent(ctx, 0);
	}
}

