/**
 *  @file macro-test.c
 *
 *  @note Copyright Premysl Janouch 2008, 2009, 2012. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */


#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <time.h>

#include <configure.h>

#ifdef HAVE_READLINE
	#include <readline/readline.h>
	#include <readline/history.h>
#endif

#include <xtnd/xtnd.h>
#include <xtndprivate.h>

#ifdef _WIN32
	/* Pwetty colors! */
	#include "waeprintf.h"
	#ifdef printf
		#undef printf
	#endif
	#define printf waeprintf
#endif


#define MACRO_REQUIRE(x) \
	if (ctx->stackSize < (x)) \
		return MACRO_ERR_STACK_UNDERFLOW;


static char *macroItoa (int i);
static size_t macroIntLen (int i);

static void recurseList (struct Item *list);
static void processMacro (const char *def);

static int subHelp (ExecContext ctx);
static int subTime (ExecContext ctx);
static int subRand (ExecContext ctx);
static int subDup (ExecContext ctx);
static int subPop (ExecContext ctx);
static int subDot (ExecContext ctx);
static int subEval (ExecContext ctx);
static int subDump (ExecContext ctx);
static int subSwap (ExecContext ctx);
static int subEqual (ExecContext ctx);
static int subIf (ExecContext ctx);
static int subConcat (ExecContext ctx);
static int subMap (ExecContext ctx);
static int subDip (ExecContext ctx);
static int compareLists (struct Item *l1, struct Item *l2);
static int compareItems (struct Item *i1, struct Item *i2);


const char helpString[] =
	"\x1b[1mSupported commands are:\x1b[0m\n"
	" .	print the top of the stack\n"
	" dump	print the entire stack\n"
	" swap	swap the two topmost items on the stack\n"
	" +	add two numbers or strings\n"
	" rand	push a random number\n"
	" dup	duplicate the top of the stack\n"
	" concat	join two lists or strings\n"
	" map	execute a program on each element of an array\n"
	" dip	\n"
	" pop	removes the topmost item\n"
	" i	evaluate the top of the stack\n"
	" =	push 1 if the two items are the same, 0 otherwise\n"
	" if	conditional statement\n\n";



/** Show help. */
static int subHelp (ExecContext ctx)
{
	printf ("%s", helpString);
	return 0;
}

/** Puts the time on the stack. */
static int subTime (ExecContext ctx)
{
	struct Item *i;

	i = macroAllocItem (ctx);
	i->type = MACRO_INT;
	i->data.num = (int) time (NULL);

	macroPush (ctx, i);
	return 0;
}

/** Convert a macro/list back to string.
 *  @param[in] list  A list of macro items.
 *  @param[in] one  Convert just the first item.
 */
static char *macroToString (struct Item *list, int one)
{
	char *buff, *p, *k;
	size_t bufflen = 512, size;
	int first = 1;

	buff = xmalloc (bufflen);
	*buff = '\0';

	while (list)
	{
		switch (list->type)
		{
		case MACRO_INT:
			size = macroIntLen (list->data.num) + 2;
			if (strlen (buff) + size > bufflen)
				buff = xrealloc (buff, bufflen += size + 512);

			if (!first)
				strcat (buff, " ");
			snprintf (buff + strlen (buff), size, "%d", list->data.num);
			break;
		case MACRO_SUB:
			size = strlen (list->data.str) + 2;
			if (strlen (buff) + size > bufflen)
				buff = xrealloc (buff, bufflen += size + 512);

			if (!first)
				strcat (buff, " ");
			strcat (buff, list->data.str);
			break;
		case MACRO_STRING:
			size = 4;
			for (p = list->data.str; *p; p++)
			{
				if (*p == '\\' || *p == '"')
					size += 2;
				else
					size++;
			}
			if (strlen (buff) + size > bufflen)
				buff = xrealloc (buff, bufflen += size + 512);

			k = buff + strlen (buff);
			if (!first)
				*k++ = ' ';

			*k++ = '"';
			for (p = list->data.str; *p; p++)
			{
				if (*p == '\\')
				{
					*k++ = '\\';
					*k++ = '\\';
				}
				else if (*p == '"')
				{
					*k++ = '\\';
					*k++ = '"';
				}
				else
					*k++ = *p;
			}
			*k++ = '"';
			*k = '\0';
			break;
		case MACRO_LIST:
			p = macroToString (list->data.list, 0);
			size = (p ? strlen (p) : 0) + 4;
			if (strlen (buff) + size > bufflen)
				buff = xrealloc (buff, bufflen += size + 512);

			if (!first)
				strcat (buff, " ");
			strcat (buff, "[");
			if (p)
			{
				strcat (buff, p);
				free (p);
			}
			strcat (buff, "]");
		default:
			break;
		}

		if (one)
			break;
		first = 0;
		list = list->next;
	}

	if (*buff)
		return buff;

	free (buff);
	return NULL;
}

/** Push a random number. */
static int subRand (ExecContext ctx)
{
	struct Item *i;

	i = macroAllocItem (ctx);
	i->type = MACRO_INT;
	i->next = NULL;
	i->data.num = rand ();

	macroPush (ctx, i);
	return 0;
}

/** Duplicate the top of the stack */
static int subDup (ExecContext ctx)
{
	struct Item *i;

	MACRO_REQUIRE (1);

	if (!(i = macroDupItem (ctx, macroAt (ctx, 0))))
		return MACRO_ERR_RUNTIME;

	macroPush (ctx, i);
	return 0;
}

/** Show the top of the stack. */
static int subDot (ExecContext ctx)
{
	struct Item *i;
	char *s;

	MACRO_REQUIRE (1);

	i = macroAt (ctx, 0);
	printf ("%s\n", (s = macroToString (i, 1)) ? s : "");
	if (s)
		free (s);
	return 0;
}

/** Evaluate the top of the stack. */
static int subEval (ExecContext ctx)
{
	struct Item *i;
	int error = 0;

	MACRO_REQUIRE (1);

	if (ctx->stackTop->type != MACRO_LIST)
		return MACRO_ERR_INCOMPATIBLE_ITEM;

	i = macroPop (ctx);
	error = macroExecInner (ctx, i->data.list);

	return error;
}

/** Show the entire stack. */
static int subDump (ExecContext ctx)
{
	puts ("Stack dump:");

	if (ctx->stackTop)
		recurseList (ctx->stackTop);
	else
		puts (" The stack is empty.");

	return 0;
}

/** Swap the two topmost items. */
static int subSwap (ExecContext ctx)
{
	struct Item *i1, *i2;

	MACRO_REQUIRE (2);

	i1 = macroPop (ctx);
	i2 = macroPop (ctx);

	macroPush (ctx, i1);
	macroPush (ctx, i2);

	return 0;
}

/** Concatentates the two topmost items. */
static int subConcat (ExecContext ctx)
{
	struct Item *i1, *i2, *iRes;

	MACRO_REQUIRE (2);

	i2 = macroPop (ctx);
	i1 = macroPop (ctx);

	if (i1->type != i2->type)
		return MACRO_ERR_INCOMPATIBLE_ITEM;

	switch (i1->type)
	{
		struct Item *i, *iList;
		char *s;

	case MACRO_LIST:
		iRes = macroDupItem (ctx, i1);
		iList = macroDupList (ctx, i2->data.list);

		i = iRes->data.list;
		while (i->next)
			i = i->next;
		i->next = iList;
		break;
	case MACRO_STRING:
		s = xmalloc (strlen (i1->data.str) + strlen (i2->data.str) + 1);
		*s = '\0';

		strcpy (s, i1->data.str);
		strcat (s, i2->data.str);

		iRes = macroAllocItem (ctx);
		iRes->type = MACRO_STRING;
		iRes->next = NULL;
		iRes->data.str = s;
		break;
	default:
		return MACRO_ERR_INCOMPATIBLE_ITEM;
	}

	macroPush (ctx, iRes);

	return 0;
}

/** Append an item to a list. */
int macroListAppend (struct Item *list, struct Item *item)
{
	if (!list)
		return MACRO_ERR_NULL_POINTER;
	if (list->type != MACRO_LIST)
		return MACRO_ERR_INCOMPATIBLE_ITEM;

	if (list->data.list)
	{
		list = list->data.list;
		while (list->next)
			list = list->next;
		list->next = item;
	}
	else
		list->data.list = item;

	return 0;
}

/** This one's weird. Should work if used properly, though.
 *  A [P] -> B
 */
static int subMap (ExecContext ctx)
{
	struct Item *prog, *agg, *i, *res, *next = NULL;

	MACRO_REQUIRE (2);

	if (ctx->stackTop->type != MACRO_LIST
		|| ctx->stackTop->next->type != MACRO_LIST)
		return MACRO_ERR_INCOMPATIBLE_ITEM;

	prog = macroPop (ctx);
	agg = macroPop (ctx);

	res = macroAllocItem (ctx);
	res->type = MACRO_LIST;
	res->next = NULL;

	/* Walk through the list. */
	i = agg->data.list;
	while (i)
	{
		struct Item *i2;
		int error;

		/* Push an item from the list and execute the cited program. */
		/* FIXME: Does it have to run in a separate context? */
		macroPush (ctx, macroDupItem (ctx, i));
		if ((error = macroExecInner (ctx, prog->data.list)))
			return error;

		/* Get the result of the program from the stack. */
		MACRO_REQUIRE (1);
		i2 = macroPop (ctx);

		/* Append it to the result. */
		if (!next)
			res->data.list = i2;
		else
			next->next = i2;
		next = i2;

		i = i->next;
	}

	macroPush (ctx, res);

	return 0;
}

/** Pops an item from the stack, executes the top of stack, pushes the item.
 *  A [B] -> ... A
 */
static int subDip (ExecContext ctx)
{
	struct Item *i, *list;

	MACRO_REQUIRE (2);

	if (ctx->stackTop->type != MACRO_LIST)
		return MACRO_ERR_INCOMPATIBLE_ITEM;

	list = macroPop (ctx);
	i = macroPop (ctx);

	macroExecInner (ctx, list->data.list);
	macroPush (ctx, i);

	return 0;
}

/** Test two items for equality. */
static int subEqual (ExecContext ctx)
{
	struct Item *i1, *i2, *res;

	MACRO_REQUIRE (2);

	i1 = macroPop (ctx);
	i2 = macroPop (ctx);

	res = macroAllocItem (ctx);
	res->next = NULL;
	res->type = MACRO_INT;

	res->data.num = compareItems (i1, i2);

	macroPush (ctx, res);
	return 0;
}

/** Delete the topmost item. */
static int subPop (ExecContext ctx)
{
	struct Item *i;

	if (!(i = macroPop (ctx)))
		return MACRO_ERR_STACK_UNDERFLOW;

	return 0;
}

/** Conditional statement.
 *  cond [then] [else] if
 */
static int subIf (ExecContext ctx)
{
	struct Item *iIf, *iThen, *iElse;

	MACRO_REQUIRE (3);

	iElse = macroPop (ctx);
	iThen = macroPop (ctx);
	iIf = macroPop (ctx);

	if (iIf->type != MACRO_INT)
		return MACRO_ERR_INCOMPATIBLE_ITEM;

	if (iIf->data.num)
	{
		if (iThen->type == MACRO_LIST)
			return macroExecInner (ctx, iThen->data.list);
		return macroExecInner (ctx, iThen);
	}
	else
	{
		if (iElse->type == MACRO_LIST)
			return macroExecInner (ctx, iElse->data.list);
		return macroExecInner (ctx, iElse);
	}
}

/** Compare two items. */
int compareItems (struct Item *i1, struct Item *i2)
{
	/* XXX: This should throw an error. */
	if (!i1 || !i2)
		return 0;

	/* With different types, always return 0. */
	if (i1->type != i2->type)
		return 0;

	switch (i1->type)
	{
	case MACRO_STRING:
	case MACRO_SUB:
		return strcmp (i1->data.str, i2->data.str) == 0;
	case MACRO_INT:
		return i1->data.num == i2->data.num;
	case MACRO_LIST:
		return compareLists (i1->data.list, i2->data.list);
	default:
		/* Error, should not happen. */
		return 0;
	}
}

/** Compare two lists. */
int compareLists (struct Item *l1, struct Item *l2)
{
	while (l1 && l2)
	{
		if (!compareItems (l1, l2))
			return 0;

		l1 = l1->next;
		l2 = l2->next;
	}
	return !l1 && !l2;
}

/** Add two numbers or strings. */
static int subAdd (ExecContext ctx)
{
	struct Item *i1, *i2, *res;

	MACRO_REQUIRE (2);

	/* The item on the top is the second operand. */
	i2 = macroPop (ctx);
	i1 = macroPop (ctx);

	if ((i1->type != MACRO_INT && i1->type != MACRO_STRING)
		|| (i2->type != MACRO_INT && i2->type != MACRO_STRING))
		return MACRO_ERR_INCOMPATIBLE_ITEM;

	res = macroAllocItem (ctx);
	res->next = NULL;

	if (i1->type == MACRO_STRING || i2->type == MACRO_STRING)
	{
		char *s, *b1 = NULL, *b2 = NULL, *s1, *s2;
		int bufflen;

		if (i1->type == MACRO_INT)
			b1 = macroItoa (i1->data.num);
		if (i2->type == MACRO_INT)
			b2 = macroItoa (i2->data.num);

		s1 = b1 ? b1 : i1->data.str;
		s2 = b2 ? b2 : i2->data.str;

		bufflen = strlen (s1) + strlen (s2) + 1;
		s = xmalloc (bufflen);
		snprintf (s, bufflen, "%s%s", s1, s2);

		if (b1)
			free (b1);
		if (b2)
			free (b2);

		res->type = MACRO_STRING;
		res->data.str = s;
	}
	else
	{
		res->type = MACRO_INT;
		res->data.num = i1->data.num + i2->data.num;
	}

	macroPush (ctx, res);

	return 0;
}

/** Converts an integer to string.
 *  The result points to allocated memory.
 */
static char *macroItoa (int i)
{
	char *a;
	size_t size;

	size = macroIntLen (i) + 1;
	a = xmalloc (size);
	snprintf (a, size, "%d", i);

	return a;
}

/** Compute the length of the integer if expressed as text. */
static size_t macroIntLen (int i)
{
	int s = 1;

	if (i < 0)
		s++;
	while (i /= 10)
		s++;

	return s;
}

static int subTest (ExecContext ctx)
{
	struct Item *i, *k, *x, *cnt;
	int j;

	MACRO_REQUIRE (1);

	cnt = macroPop (ctx);
	if (cnt->type != MACRO_INT)
		return MACRO_ERR_INCOMPATIBLE_ITEM;

	k = macroAllocItem (ctx);
	k->type = MACRO_LIST;
	k->data.list = NULL;
	for (j = 0; j < cnt->data.num; j++)
	{
		i = macroAllocItem (ctx);
		i->data.num = j;
		if (j % 2)
		{
			if (!k->data.list)
				x = k->data.list = i;
			else
				x = x->next = i;
		}
	}
	macroPush (ctx, k);

	return 0;
}


/** Program entry point. */
int main (int argc, char *argv[])
{
	static const char *prompt = "\n\x1b[33;1m> \x1b[0m";
#ifdef HAVE_READLINE
	char *s;
#else
	char buff[80];
#endif

	printf ("\x1b[33;1mWelcome to the testing interpreter!\x1b[0m\n\n");
	printf ("%s", helpString);
	printf ("\x1b[1mSome examples:\x1b[0m\n"
		" \"rand .\" -- print a random number\n"
		" \"12 34 + 46 = .\" -- prints 1 if 12 + 34 = 46\n"
		" \"[rand] dup i swap i + .\" -- add two random numbers\n\n"
		"\x1b[90mPress just [Enter] to exit\x1b[0m");

#ifdef HAVE_READLINE
	while ((s = readline (prompt)))
	{
		if (!strlen (s))
			break;
		processMacro (s);

		add_history (s);
		free (s);
	}
#else /* ! HAVE_READLINE */
	/* Windows XP and newer need no readline to be user-friendly. */
	printf (prompt);
	while (fgets (buff, sizeof buff, stdin))
	{
		char *nl;
		if ((nl = strpbrk (buff, "\r\n")))
			*nl = '\0';

		if (!strlen (buff))
			return 0;

		processMacro (buff);
		printf (prompt);
	}
#endif /* ! HAVE_READLINE */

	return 0;
}

/** Process the macro. */
static void processMacro (const char *def)
{
	static const struct MacroSubInfo subs[] =
	{
		{"help",    subHelp},
		{"time",    subTime},
		{".",       subDot},

		{"i",       subEval},
		{"swap",    subSwap},
		{"dup",     subDup},
		{"pop",     subPop},
		{"dip",     subDip},

		{"if",      subIf},
		{"concat",  subConcat},
		{"map",     subMap},

		{"+",       subAdd},
		{"=",       subEqual},
		{"rand",    subRand},
		/* -, abs, sign */

		{"dump",    subDump},
		{"t",       subTest},
		{NULL,      NULL}
	};

	Macro macro;
	int error;
	char errorInfo[80];

	if ((error = macroParse (def, &macro)))
		printf ("\x1b[31;1mParser error: %s\x1b[0m\n", macroErrorString (error));

	if (macro)
	{
		if ((error = macroExec (macro, subs, NULL,
			errorInfo, sizeof errorInfo)))
		{
			printf ("\x1b[31;1mRuntime error: %s\x1b[0m\n",
				*errorInfo ? errorInfo : macroErrorString (error));
		}
		macroFree (&macro);
	}
}

/** Recurse thru' the list and show the content.
 *  Not thread-safe nor anything.
 */
static void recurseList (struct Item *list)
{
	static int ind = 1;
	int i;

	for (; list; list = list->next)
	{
		for (i = ind; i; i--)
			putchar(' ');

		switch (list->type)
		{
			char *s;

		case MACRO_INT:
			printf ("INT: %d\n", list->data.num);
			break;
		case MACRO_STRING:
			printf ("STRING: %s\n", (s = macroToString(list, 1)));
			free (s);
			break;
		case MACRO_SUB:
			printf ("SUB: %s\n", list->data.str);
			break;
		case MACRO_LIST:
			printf ("LIST\n");
			ind++;
			recurseList(list->data.list);
			ind--;
		}
	}
}



