/**
 *  @file waeprintf.c
 *  @brief printf() with ANSI escape characters support for Windows
 *
 *  @note Copyright Premysl Janouch 2009, 2012. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdio.h>
#include <windows.h>
#include <stdarg.h>

#include "waeprintf.h"


/** Process an ANSI escape code.
 *  @param[in,out] csbi  Console information.
 *  @param[in] code  The escape code.
 *  @param[in] param  The parameter to process.
 *  @return A non-zero value if any change has been done.
 */
static int waeprocess (PCONSOLE_SCREEN_BUFFER_INFO csbi, char code, int param);


#define WAE_BGCLRMASK (BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE)

static int waeBgTable [] =
{
	/* 0 black   */	0,
	/* 1 red     */	BACKGROUND_RED,
	/* 2 green   */ BACKGROUND_GREEN,
	/* 3 yellow  */ BACKGROUND_RED | BACKGROUND_GREEN,
	/* 4 blue    */ BACKGROUND_BLUE,
	/* 5 magenta */ BACKGROUND_RED | BACKGROUND_BLUE,
	/* 6 cyan    */ BACKGROUND_BLUE | BACKGROUND_GREEN,
	/* 7 white   */ BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE,
	/* 8 undef.  */	0,
	/* 9 reset   */ 0
};

#define WAE_FGCLRMASK (FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE)

static int waeFgTable [] =
{
	/* 0 black   */	0,
	/* 1 red     */	FOREGROUND_RED,
	/* 2 green   */ FOREGROUND_GREEN,
	/* 3 yellow  */ FOREGROUND_RED | FOREGROUND_GREEN,
	/* 4 blue    */ FOREGROUND_BLUE,
	/* 5 magenta */ FOREGROUND_RED | FOREGROUND_BLUE,
	/* 6 cyan    */ FOREGROUND_BLUE | FOREGROUND_GREEN,
	/* 7 white   */ FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE,
	/* 8 undef.  */	0,
	/* 9 reset   */ FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE
};


int waeprintf (const char *fmt, ...)
{
	int result = 0;
	int strSize, param;
	char *strBuff, *s;
	va_list args;
	HANDLE hStdOut;
	CONSOLE_SCREEN_BUFFER_INFO csbi;

	hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	if (hStdOut == INVALID_HANDLE_VALUE
	 || !GetConsoleScreenBufferInfo(hStdOut, &csbi))
	{
		/* Attempt fallback, fixes MSYS. */
		va_start (args, fmt);
		result = vprintf (fmt, args);
		va_end (args);
		fflush (stdout);
		return result;
	}

	/* Find out how much memory to allocate. */
	va_start(args, fmt);
	strSize = vsnprintf(NULL, 0, fmt, args);
	va_end(args);

	if (!(strBuff = malloc(strSize + 1)))
		return -1;

	va_start(args, fmt);
	result = vsnprintf(strBuff, strSize + 1, fmt, args);
	va_end(args);

	for (s = strBuff; *s;)
	{
		char *p, code, changed;

		if (*s != '\x1b')
		{
			putchar(*s++);
			continue;
		}
		if (*++s != '[')
		{
			putchar(*s++);
			continue;
		}
		p = ++s;
		while (*s && (*s < 64 || *s > 126))
			s++;
		if (!*s)
			break;

		/* We only care for colors so far. */
		if ((code = *s++) != 'm')
			continue;
		for (param = changed = 0; p < s; p++)
		{
			if (*p >= '0' && *p <= '9')
				param = param * 10 + *p - '0';
			else
			{
				changed |= waeprocess(&csbi, code, param);
				param = 0;
			}
		}
		if (changed)
			SetConsoleTextAttribute(hStdOut, csbi.wAttributes);
	}

	free(strBuff);
	return result;
}

static int waeprocess (PCONSOLE_SCREEN_BUFFER_INFO csbi, char code, int param)
{
	if (param < 30)
	{
		switch (param)
		{
			case 0:
				/* The default is white on black. */
				csbi->wAttributes = WAE_FGCLRMASK;
				break;
			case 1:
				csbi->wAttributes |= FOREGROUND_INTENSITY;
				break;
			case 22:
				csbi->wAttributes &= ~FOREGROUND_INTENSITY;
				break;
/* It doesn't really matter if we don't support the LVB attributes. */
#ifdef COMMON_LVB_UNDERSCORE
			case 4:
			case 21:
				csbi->wAttributes |= COMMON_LVB_UNDERSCORE;
				break;
			case 24:
				csbi->wAttributes &= ~COMMON_LVB_UNDERSCORE;
				break;
#endif
#ifdef COMMON_LVB_REVERSE_VIDEO
			case 7:
				csbi->wAttributes |= COMMON_LVB_REVERSE_VIDEO;
				break;
			case 27:
				csbi->wAttributes &= ~COMMON_LVB_REVERSE_VIDEO;
				break;
#endif
			default:
				return 0;
		}
	}
	else if (param < 40)
		csbi->wAttributes = (csbi->wAttributes & ~WAE_FGCLRMASK)
			| waeFgTable[param - 30];
	else if (param < 50)
		csbi->wAttributes = (csbi->wAttributes & ~WAE_BGCLRMASK)
			| waeBgTable[param - 40];
	else if (param < 90)
		return 0;
	else if (param < 100)
		csbi->wAttributes = (csbi->wAttributes & ~WAE_FGCLRMASK)
			| waeFgTable[param - 90] | FOREGROUND_INTENSITY;
	else if (param < 110)
		csbi->wAttributes = (csbi->wAttributes & ~WAE_BGCLRMASK)
			| waeBgTable[param - 100] | BACKGROUND_INTENSITY;
	else
		return 0;
	return 1;
}
