/**
 *  @file testbtree.c
 *  @brief Test the B-tree implementation.
 *
 *  This file is in the public domain.
 *
 */

#include <stdlib.h>
#include <string.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


/* Well... I didn't like those warnings, I am sure about what I'm doing.*/
#define PTR_TO_INT(ptr) ((int) ((char *) ptr - (char *) NULL))
#define INT_TO_PTR(i) ((char *) NULL + (i))


static void fail (const char *msg)
{
	puts (msg);
	exit (1);
}

static int btKeyCmp (const void *first, const void *second)
{
	return (int) ((char *) first - (char *) second);
}

static void testBtInsert (BTree bt, const int *numbers, const int n_numbers)
{
	int i;

	for (i = 0; i < n_numbers; i++)
	{
		btInsert (bt, INT_TO_PTR (numbers[i]));
		if (PTR_TO_INT (btFind (bt, INT_TO_PTR (numbers[i]))) != numbers[i])
			fail ("Failed to find inserted key");
	}

	/* Well, that would be weird but anyway. */
	for (i = n_numbers; i--;)
	{
		if (PTR_TO_INT (btFind (bt, INT_TO_PTR (numbers[i]))) != numbers[i])
			fail ("Failed to find inserted key on second try");
	}

	if (btFind (bt, INT_TO_PTR (-1)) != NULL)
		fail ("Found something that's not even in the B-tree");
}

static void testBtDelete (BTree bt, const int *numbers, const int n_numbers)
{
	int i;

	for (i = 0; i < n_numbers; i++)
	{
		int exists, deleted;

		exists = !!btFind (bt, INT_TO_PTR (numbers[i]));
		deleted = !!btDelete (bt, INT_TO_PTR (numbers[i]));

		if (exists && !deleted)
			fail ("Could not delete key although present");
		else if (!exists && deleted)
			fail ("Deleted a key although not present");
		if (btFind (bt, INT_TO_PTR (numbers[i])))
			fail ("Found a key that's not supposed to be there");
	}
}

static int btIntPtrCmp (const void *a, const void *b)
{
	return *(int *) a - *(int *) b;
}

static void btIntPtrFree (void *item, void *extra)
{
	free (item);
}

static void btIntPtrWalkCB (void *item, void *extra)
{
	if (!btFind (extra, item))
		fail ("The walk callback has given us a wrong item pointer.");
}

/** Test the B-tree implementation. */
int testbtree (void)
{
	BTree bt;
	int i;

	int numbers[4096];
	const int n_numbers = sizeof (numbers) / sizeof (numbers[0]);

	/* First an increasing set of numbers. */
	for (i = 0; i < n_numbers; i++)
		numbers[i] = 1 + i;

	puts ("-- Trying an increasing set...");
	bt = btCreate (3, btKeyCmp, NULL, NULL);
	testBtInsert (bt, numbers, n_numbers);
	testBtDelete (bt, numbers, n_numbers);
	btFree (bt);

	/* Then a decreasing one with some repeating. */
	for (i = 0; i < n_numbers; i++)
		numbers[i] = 1 + (n_numbers - i - 1) % (n_numbers / 2);

	puts ("-- Trying a decreasing set with duplicates...");
	bt = btCreate (6, btKeyCmp, NULL, NULL);
	testBtInsert (bt, numbers, n_numbers);
	testBtDelete (bt, numbers, n_numbers);
	btFree (bt);

	/* Then proceed to something more or less random.
	 * It's almost certain that there will be duplicates as well.
	 */
	srand (0);

	for (i = 0; i < n_numbers; i++)
		numbers[i] = 1 + rand () % n_numbers;

	puts ("-- Trying a random set...");
	bt = btCreate (9, btKeyCmp, NULL, NULL);
	testBtInsert (bt, numbers, n_numbers);
	testBtDelete (bt, numbers, n_numbers);
	btFree (bt);

	puts ("-- Trying callbacks...");
	bt = btCreate (5, btIntPtrCmp, btIntPtrFree, NULL);

	for (i = 0; i < n_numbers; i++)
	{
		int *intPtr;
		intPtr = malloc (sizeof *intPtr);
		*intPtr = numbers[i];
		btInsert (bt, intPtr);
	}

	btWalk (bt, btIntPtrWalkCB, bt);
	btFree (bt);

	puts ("-- B-tree test successful");
	return 0;
}
