/**
 *  @file testexception.c
 *  @brief Test the implementation of exceptions.
 *
 *  This file is in the public domain.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


enum TestException
{
	TEST_A = 0x10,
	TEST_B
};

static void fail (const char *msg)
{
	puts (msg);
	exit (1);
}

static void unexpected (int code)
{
	printf ("Unexpected exception code %#x\n", code);
	exit (1);
}

static void throwCode (int code)
{
	printf ("-- Throwing code %#x\n", code);
	THROW (code);
}

static void testXSimple (void)
{
	int finally = 0;

	TRY
		throwCode (TEST_A);
		fail ("Exception not thrown");
	EXCEPT (TEST_A):
		break;
	EXCEPT_DEFAULT:
		unexpected (e.code);
	FINALLY:
		finally = 1;
	END_TRY

	if (!finally)
		fail ("FINALLY not executed");
}

static void testXNested (void)
{
	TRY
		testXSimple ();
	EXCEPT (TEST_A):
		break;
	EXCEPT_DEFAULT:
		unexpected (e.code);
	FINALLY:
		break;
	END_TRY
}

static void testXRethrow (void)
{
	TRY
		TRY
			throwCode (TEST_B);
			fail ("Exception not thrown [inner block]");
		EXCEPT_DEFAULT:
			RETHROW ();
			fail ("Exception not rethrown [inner block]");
		FINALLY:
			break;
		END_TRY

		fail ("Exception not thrown [outer block]");
	EXCEPT (TEST_B):
		break;
	EXCEPT_DEFAULT:
		unexpected (e.code);
	FINALLY:
		break;
	END_TRY
}

/** Test the implementation of exceptions. */
int testexception (void)
{
	puts ("-- Trying a simple exception...");
	testXSimple ();

	puts ("-- Trying a nested exception...");
	testXNested ();

	puts ("-- Trying rethrowing exceptions...");
	testXRethrow ();

	puts ("-- Exception test successful");
	return 0;
}
