/**
 *  @file testhmap.c
 *  @brief Test the hash map implementation.
 *
 *  This file is in the public domain.
 *
 */

#include <stdlib.h>
#include <string.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


static void fail (const char *msg)
{
	puts (msg);
	exit (1);
}

static unsigned hmHashStr (const void *key)
{
	return hmFNVHash (key, strlen ((const char *) key));
}

/** Test the hash map implementation. */
int testhmap (void)
{
	static const struct
	{
		const char *key;
		const char *value;
	}
	*iter, table[] =
	{
		{"aaa", "111"}, {"aab", "112"}, {"aac", "113"}, {"aad", "114"},
		{"baa", "211"}, {"bab", "212"}, {"bac", "213"}, {"bad", "214"},
		{"aae", "115"}, {"aaf", "116"}, {"aag", "117"}, {"aah", "118"},
		{"bae", "215"}, {"baf", "216"}, {"bag", "217"}, {"bah", "218"},
		{"abc", "def"}, {"cba", "fed"}, {"123", "456"}, {"xyz", "zyx"},
		{"foo", "bar"}, {"bar", "baz"}, {"ghi", "jkl"}, {"mno", "pqr"},
		{"stu", "vwx"}, {"val", "ues"}, {"idu", "nno"}, {"the", "end"},
		{NULL, NULL}
	};

	HMap m;

	puts ("-- Trying basic operations...");
	m = hmCreate (0, hmHashStr, (HmKeyCmpCB) strcmp, NULL, NULL);
	if (hmIsElement (m, "foo"))
		fail ("Key found even when not supposed to be present");

	hmInsert (m, (char *) "foo", (char *) "bar");
	hmInsert (m, (char *) "bar", (char *) "baz");

	if (!hmIsElement (m, "foo"))
		fail ("Key not found");
	if (strcmp (hmFind (m, "foo"), "bar"))
		fail ("Key has invalid value");
	if (!hmIsElement (m, "bar"))
		fail ("Key not found");
	if (strcmp (hmFind (m, "bar"), "baz"))
		fail ("Key has invalid value");

	if (!hmDelete (m, "foo"))
		fail ("Failed to delete key");
	if (hmIsElement (m, "foo"))
		fail ("Deleted key stays present");
	if (!hmIsElement (m, "bar"))
		fail ("Key not found after deleting another");

	hmInsert (m, (char *) "bar", (char *) "qux");

	if (!hmIsElement (m, "bar"))
		fail ("Key not found after reassigning");
	if (strcmp (hmFind (m, "bar"), "qux"))
		fail ("Invalid value after reassigning");

	hmClear (m);
	if (hmIsElement (m, "bar"))
		fail ("Key stays present after clearing");
	hmFree (m);

	puts ("-- Trying with many items...");
	m = hmCreate (2, hmHashStr, (HmKeyCmpCB) strcmp, free, free);
	for (iter = table; iter->key != NULL; iter++)
		hmInsert (m, xstrdup (iter->key), xstrdup (iter->value));

	for (iter = table; iter->key != NULL; iter++)
	{
		char *value;

		value = hmFind (m, iter->key);
		if (!value)
			fail ("Key not found");
		if (strcmp (value, iter->value))
			fail ("Key has wrong value");
		if (!hmDelete (m, iter->key))
			fail ("Key could not be deleted");
	}

	for (iter = table; iter->key != NULL; iter++)
		if (hmIsElement (m, iter->key))
			fail ("Deleted key stays present");
	hmFree (m);

	puts ("-- Hash map test successful");
	return 0;
}
