/**
 *  @file testsockets.c
 *  @brief Test sockets.
 *
 *  This file is in the public domain.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
	#include <windows.h>
#else /* ! _WIN32 */
	#include <pthread.h>
#endif /* ! _WIN32 */

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


#define TIMEOUT 100

#define SELF "localhost"
#define SELF_PORT 9876


static void fail (const char *msg)
{
	puts (msg);
	exit (1);
}

#define CHECK_RESULT(msg)                                                     \
	if (r == -1)                                                              \
	{                                                                         \
		if (socketsGetLastError () == SOCKETS_ETIMEDOUT)                      \
			puts ("Timeouted");                                               \
		fail (msg);                                                           \
	}

#ifdef _WIN32
static DWORD WINAPI serverThread (LPVOID ptr)
#else /* ! _WIN32 */
static void *serverThread (void *ptr)
#endif /* ! _WIN32 */
{
	static char buff[4096];
	int listen_sck, sck, r, received, sent;

	puts ("-- [Server] Waiting for connection");
	listen_sck = *(int *) ptr;
	sck = socketsTcpAcceptNB (listen_sck, -1);
	if (sck == -1)
		fail ("Error: Server failed accepting.");

	puts ("-- [Server] Receiving");
	received = 0;
	while ((r = socketsRecvNB (sck,
		buff + received, sizeof buff - received, TIMEOUT)) > 0)
		received += r;
	CHECK_RESULT ("Error: Server failed receiving.");

	puts ("-- [Server] Sending");
	sent = 0;
	while ((r = socketsSendNB (sck,
		buff + sent, received - sent, TIMEOUT)) > 0)
		sent += r;
	CHECK_RESULT ("Error: Server failed sending.");

	puts ("-- [Server] Finished");
	socketsClose (sck);
	socketsClose (listen_sck);
	return 0;
}

static void serverSpawn (int *listen_sck)
{
#ifdef _WIN32
	HANDLE id;
	if (!(id = CreateThread (NULL, 0, serverThread, listen_sck, 0, NULL)))
#else /* ! _WIN32 */
	pthread_t id;
	if (pthread_create (&id, NULL, serverThread, listen_sck))
#endif /* ! _WIN32 */
		fail ("Error: Failed to spawn a thread.");
}

int testsockets (void)
{
	const char data[] = "Something to send back to myself over TCP.";
	int srv_sck, cli_sck, r, received, sent;
	char buff[sizeof data];

	puts ("-- Initializing the library");
	socketsInit ();

	printf ("-- Trying to listen on TCP %s:%d\n", SELF, SELF_PORT);
	if ((srv_sck = socketsTcpListen (SELF, SELF_PORT, 10)) == -1)
		fail ("Error: Cannot listen");

	puts ("-- Spawning server thread");
	serverSpawn (&srv_sck);

	puts ("-- Connecting");
	if ((cli_sck = socketsTcpConnectNB (SELF, SELF_PORT, 1000)) == -1)
		fail ("Error: Cannot connect.");

	puts ("-- Sending");
	sent = 0;
	while ((r = socketsSendNB (cli_sck,
		data + sent, sizeof data - sent, TIMEOUT)) > 0)
		sent += r;
	CHECK_RESULT ("Error: Sending failed.");

	puts ("-- Shutting down the send end");
	socketsShutdown (cli_sck, SOCKETS_SHUT_SEND);

	puts ("-- Receiving");
	received = 0;
	while ((r = socketsRecvNB (cli_sck,
		buff + received, sizeof data - received, 1000)) > 0)
		received += r;
	CHECK_RESULT ("Error: Receiving failed.");

	if (received != sizeof data)
		fail ("Error: Not all data retrieved back.");
	if (memcmp (data, buff, sizeof data))
		fail ("Error: Corrupt data.");

	puts ("-- Finished");
	socketsClose (cli_sck);
	return 0;
}

