/**
 *  @file conf.c
 *  @brief Working with configuration files
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

/* ? Set strcmp in confCreate => set the compare function for sections, too. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


/** Specifies the format of line endings. DOS uses \\r\\n, UNIX uses \\n
 *  and MAC uses \\r to separate lines. */
enum ConfLineEnd {CONF_DOS, CONF_UNIX, CONF_MAC, CONF_EOF};

/** An array of line endings (can be indexed with ConfLineEnd). */
static const char *confLineEnds[] = {"\r\n", "\n", "\r", ""};


struct ConfLine
{
	/** Specifies the type of an item. */
	enum
	{
		/** Regular expression: /^[^#;].*$/
		 *  Defines an assignment with a name and a value.
		 *  The @a value can be NULL (no equal sign).
		 */
		CONF_ASSIGN,
		/** Regular expression: /^\\[.*\\]$/
		 *  Defines the beginning of a section. Used only at parse time.
		 *  Only @a name is used.
		 */
		CONF_SECTION,
		/** Any other value, like comments or blank lines.
		 *  Only @a value is used.
		 */
		CONF_OTHER
	}
	type;
	/** The name of an item. */
	char *name;
	/** The value of an item. */
	char *value;
	/** The format of line ending for this line. */
	enum ConfLineEnd lineEnd;
};

struct Conf
{
	/** Assignments and other lines that have no appropriate section
	 *  (ie. they occur before any section is started).
	 */
	DynArray nullSection;
	/** Associative array of sections. When a section is in the file more than
	 *  once, all parts of it are processed. When the file is saved again,
	 *  the parts will be merged.
	 */
	StrList sections;
	/** A callback used to compare keys. */
	int (*strcmp) (const char *, const char *);
	/** The global format of line endings. */
	enum ConfLineEnd lineEnd;
};


/* Local function declarations. */
static int confFreeDynArrayCB (const void *array, void *extra);
static void confFreeConfLineCB (const void *item, void *extra);

static struct ConfLine *confReadLine (FILE *restrict file,
	struct ConfLine *restrict out);
static DynArray confAskForSection (Conf restrict cfg,
	const char *restrict name, int create);


/** A callback for freeing of dynamic arrays. */
static int confFreeDynArrayCB (const void *array, void *extra)
{
	daFree (*(DynArray *) array);
	return 0;
}

/** A callback for freeing of configuration file lines. */
static void confFreeConfLineCB (const void *item, void *extra)
{
	struct ConfLine *line = (struct ConfLine *) item;

	if (line->name)
		free (line->name);
	if (line->value)
		free (line->value);
}


Conf confCreate (void)
{
	Conf result;
	SlDataInfo diSections = {sizeof (DynArray),
		confFreeDynArrayCB, NULL, NULL};

	result = xmalloc (sizeof (struct Conf));

#ifdef _WIN32
	result->lineEnd = CONF_DOS;
#else
	result->lineEnd = CONF_UNIX;
#endif
	/* Case sensitive by default. */
	result->strcmp = strcmp;

	if ((result->nullSection = daCreate (sizeof (struct ConfLine),
		confFreeConfLineCB, NULL)) == NULL)
	{
		free (result);
		return NULL;
	}
	if ((result->sections
		= slCreate (SL_DUP_IGNORE_NEW, strcmp, &diSections)) == NULL)
	{
		daFree (result->nullSection);
		free (result);
		return NULL;
	}

	return result;
}

void confDestroy (Conf cfg)
{
	daFree (cfg->nullSection);
	slFree (cfg->sections);
	free (cfg);
}

void confClear (Conf cfg)
{
	assert (cfg != NULL);

	daClear (cfg->nullSection);
	slClear (cfg->sections);
}

void confSetKeyCompare (Conf restrict cfg,
	int (*strcmp) (const char *, const char *))
{
	cfg->strcmp = strcmp;
}

/** Returns a pointer to the section with the specified name.
 *  If there is no such section and @a create is non-zero, it creates it.
 *  Note that NULL is a valid section, too.
 *  @param[in] cfg  A configuration object.
 *  @param[in] name  Section name.
 *  @param[in] create  Whether to create the section if it doesn't exist.
 *  @return A pointer to the section object on success, NULL on error.
 */
static DynArray confAskForSection (Conf restrict cfg,
	const char *restrict name, int create)
{
	DynArray res;
	int i;

	if (name == NULL)
		return cfg->nullSection;

	if ((i = slFind (cfg->sections, name, 0)) != -1)
		res = *(DynArray *) slGetData (cfg->sections, i);
	else if (create)
	{
		res = daCreate (sizeof (struct ConfLine), confFreeConfLineCB, NULL);
		if ((i = slAppend (cfg->sections, name)) == -1)
			return NULL;
		slSetData (cfg->sections, i, &res);
	}
	else
		return NULL;

	return res;
}

int confRead (Conf restrict cfg, const char *restrict filename)
{
	FILE *file;
	DynArray cursection;
	struct ConfLine cfgline;
	unsigned cnt[] = {0, 0, 0, 0};

	if ((file = fopen (filename, "rb")) == NULL)
		return -1;

	/* Clear everything. */
	daClear (cfg->nullSection);
	slClear (cfg->sections);

	cursection = cfg->nullSection;

	while (confReadLine (file, &cfgline))
	{
		switch (cfgline.type)
		{
		case CONF_ASSIGN:
		case CONF_OTHER:
			daAppend (cursection, &cfgline);
			break;
		case CONF_SECTION:
			cursection = confAskForSection (cfg, cfgline.name, 1);
			break;
		}
		cnt[cfgline.lineEnd]++;
	}

	/* Detect the prevailing line ending. */
	if (CONF_DOS[cnt] > CONF_UNIX[cnt])
		cfg->lineEnd = CONF_MAC[cnt] > CONF_DOS[cnt] ? CONF_MAC : CONF_DOS;
	else
		cfg->lineEnd = CONF_MAC[cnt] > CONF_UNIX[cnt] ? CONF_MAC : CONF_UNIX;

	fclose (file);
	return 0;
}

/** Read a line from the current position
 *  and set the position to the start of next line.
 *  @param[in] file  A file to read from.
 *  @param[out] out  An output structure.
 *  @return The value of @a out on success, NULL otherwise.
 */
static struct ConfLine *confReadLine (FILE *restrict file,
	struct ConfLine *restrict out)
{
	int ch;
	char *line, *p;
	size_t linelen;
	MemStream buff;

	if (feof (file))
		return NULL;

	if ((buff = msCreate (128)) == NULL)
		return NULL;

	out->lineEnd = CONF_EOF;
	while ((ch = getc (file)) != EOF)
	{
		if (ch == '\n')
		{
			out->lineEnd = CONF_UNIX;
			break;
		}
		else if (ch == '\r')
		{
			ch = getc (file);
			if (ch != '\n')
			{
				out->lineEnd = CONF_MAC;
				ungetc (ch, file);
			}
			else
				out->lineEnd = CONF_DOS;
			break;
		}
		msWriteInt8 (buff, ch);
	};
	msWriteInt8 (buff, '\0');
	line = (char *) msGetPtr (buff);

	linelen = strlen (line);
	if (line[0] == '#' || line[0] == ';' || line[0] == '\0')
	{
		out->type = CONF_OTHER;
		out->name = NULL;
		out->value = xstrdup (line);
	}
	else if (line[0] == '[' && line[linelen - 1] == ']')
	{
		line[linelen - 1] = '\0';
		out->type = CONF_SECTION;
		out->name = xstrdup (line + 1);
		out->value = NULL;
	}
	else if ((p = strchr (line, '=')) != NULL)
	{
		*p = '\0';
		out->type = CONF_ASSIGN;
		out->name = xstrdup (line);
		out->value = xstrdup (p + 1);
	}
	else
	{
		out->type = CONF_ASSIGN;
		out->name = xstrdup (line);
		out->value = NULL;
	}

	msFree (buff);

	return out;
}

int confWrite (Conf restrict cfg, const char *restrict filename)
{
	FILE *file;
	DynArray section;
	unsigned i, nsections;

	if ((file = fopen (filename, "wb")) == NULL)
		return -1;

	section = cfg->nullSection;
	nsections = slGetSize (cfg->sections);
	for (i = 0; ; i++)
	{
		unsigned k, arraylen;

		arraylen = daGetSize (section);
		for (k = 0; k < arraylen; k++)
		{
			struct ConfLine *cl;

			cl = daGet (section, k);
			switch (cl->type)
			{
			case CONF_ASSIGN:
				fputs (cl->name, file);
				if (cl->value)
				{
					fputc ('=', file);
					fputs (cl->value, file);
				}
				fputs (confLineEnds[cl->lineEnd], file);
				break;
			case CONF_OTHER:
				fputs (cl->value, file);
				fputs (confLineEnds[cl->lineEnd], file);
				break;
			default:
				break;
			}
		}

		if (i == nsections)
			break;

		fprintf (file, "[%s]%s", slGet (cfg->sections, i),
			confLineEnds[cfg->lineEnd]);
		section = *(DynArray *) slGetData (cfg->sections, i);
	}
	fclose (file);
	return 0;
}

StrList confGetSections (Conf cfg)
{
	/* TODO: Duplicate the list? */
	return cfg->sections;
}

const char *confGet (Conf restrict cfg,
	const char *restrict section, const char *restrict key)
{
	DynArray confsection;
	unsigned i, len;

	confsection = confAskForSection (cfg, section, 0);
	if (confsection == NULL)
		return NULL;

	len = daGetSize (confsection);
	for (i = 0; i < len; i++)
	{
		struct ConfLine *cl;

		cl = daGet (confsection, i);
		if (cl->type == CONF_ASSIGN && !cfg->strcmp (key, cl->name))
			return cl->value;
	}
	return NULL;
}

void confDeleteSection (Conf restrict cfg,
	const char *restrict section)
{
	if (section == NULL)
		daClear (cfg->nullSection);
	else
	{
		int i;

		if ((i = slFind (cfg->sections, section, 0)) != -1)
			slDelete (cfg->sections, i);
	}
}

void confSet (Conf restrict cfg,
	const char *restrict section, const char *restrict key,
	const char *restrict value)
{
	DynArray confsection;
	struct ConfLine newcl;
	unsigned i, len;

	confsection = confAskForSection (cfg, section, 1);

	len = daGetSize (confsection);
	for (i = 0; i < len; i++)
	{
		struct ConfLine *cl;

		cl = daGet (confsection, i);
		if (cl->type == CONF_ASSIGN && !cfg->strcmp (key, cl->name))
		{
			cl->value = value ? xstrdup (value) : NULL;
			return;
		}
	}

	/* No assignment found, make a new one on the end. */
	newcl.type = CONF_ASSIGN;
	newcl.name = xstrdup (key);
	newcl.value = value ? xstrdup (value) : NULL;
	newcl.lineEnd = cfg->lineEnd;
	daAppend (confsection, &newcl);
}

StrList confGetList (Conf restrict cfg, const char *restrict section)
{
	DynArray mysect;
	unsigned length, i;
	SlDataInfo di;
	StrList list;

	if ((mysect = confAskForSection (cfg, section, 0)) == NULL)
		return NULL;
	length = daGetSize (mysect);

	di.dataSize = sizeof (char *);
	di.dataFree = slFreeStringCB;
	di.dataClone = slCloneStringCB;
	di.cbExtra = NULL;

	if ((list = slCreate (SL_DUP_IGNORE_NEW, cfg->strcmp, &di)) == NULL)
		return NULL;

	for (i = 0; i < length; i++)
	{
		struct ConfLine *line;
		char *p;
		int index;

		line = daGet (mysect, i);
		if (line->type != CONF_ASSIGN)
			continue;

		index = slAppend (list, line->name);
		if (index != -1)
		{
			p = line->value ? xstrdup (line->value) : NULL;
			slSetData (list, index, &p);
		}
	}

	return list;
}

void confSetList (Conf restrict cfg,
	const char *restrict section, StrList restrict list)
{
	unsigned i, length;

	length = slGetSize (list);
	for (i = 0; i < length; i++)
	{
		confSet (cfg, section, slGet (list, i), *(char **) slGetData (list, i));
	}
}



int confGetInt (Conf restrict cfg, const char *restrict section,
	const char *restrict key, int *restrict value)
{
	const char *temp;
	char *ep;
	int number;

	if ((temp = confGet (cfg, section, key)) == NULL)
		return -1;

	number = strtol (temp, &ep, 10);
	if (ep == temp)
		return -1;

	*value = number;
	return 0;
}

void confSetInt (Conf restrict cfg, const char *restrict section,
	const char *restrict key, int value)
{
	char buffer[32];

	snprintf (buffer, sizeof (buffer), "%d", value);
	confSet (cfg, section, key, buffer);
}

