/**
 *  @file conf.h
 *  @brief Configuration files
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_CONF_H__
#define __XTND_CONF_H__

/**
 *  @defgroup conf Configuration files
 *  @ingroup util
 *
 *  This module allows you to work with configuration files that use a file
 *  format very similar to Windows INI files.
 *
 *  @{
 */

/** A configuration class. */
typedef struct Conf *Conf;


/** Create a configuration object. It has DOS line endings on Windows
 *  and UNIX line endings on other systems by default.
 *  @return A pointer to the object.
 */
Conf confCreate (void);

/** Destroy the configuration object.
 *  @param[in] cfg  The object to be destroyed.
 */
void confDestroy (Conf cfg);

/** Clear the configuration object.
 *  @param[in] cfg  The object to be cleared.
 */
void confClear (Conf cfg);

/** Set the compare function for keys. By default, keys are case-sensitive.
 *  Sections always use standard strcmp.
 *  @param[in] cfg  A configuration object.
 *  @param[in] strcmp  The new compare function.
 */
void confSetKeyCompare (Conf restrict cfg,
	int (*strcmp) (const char *, const char *));

/** Clear the object and read the specified file into it. Any line ending
 *  is accepted, which means it is impossible to include \\r or \\n in
 *  an assignment. The prevailing line ending format becomes the default
 *  line ending for new assignments.
 *  @param[out] cfg  A configuration object.
 *  @param[in] filename  The file to be read.
 *  @return 0 on success, -1 otherwise.
 */
int confRead (Conf restrict cfg, const char *restrict filename);

/** Write the data to a file, overwrite the contents.
 *  @param[in] cfg  A configuration object.
 *  @param[in] filename  The file to be written to.
 *  @return 0 on success, -1 otherwise.
 */
int confWrite (Conf restrict cfg, const char *restrict filename);

/** Gets data for a specific key. The first assignment for the key
 *  in the specified section is returned, the others are ignored.
 *  The returned value remains valid as long as the configuration object exists
 *  and the assignment doesn't change.
 *  @param[in] cfg  A configuration object.
 *  @param[in] section  The name of the section.
 *  @param[in] key  The name of the key.
 *  @return The string value of the key on success, NULL otherwise.
 */
const char *confGet (Conf restrict cfg,
	const char *restrict section, const char *restrict key);

/** Sets data for a specific key. Only the first assignment for the key
 *  in the specified section will be modified, the others are ignored.
 *  @param[in] cfg  A configuration object.
 *  @param[in] section  The name of the section.
 *  @param[in] key  The name of the key.
 *  @param[in] value  The string value to be written.
 */
void confSet (Conf restrict cfg,
	const char *restrict section, const char *restrict key,
	const char *restrict value);


/** Delete a section from the configuration object.
 *  @param[in] cfg  A configuration object.
 *  @param[in] section  The name of the section.
 */
void confDeleteSection (Conf restrict cfg,
	const char *restrict section);

/** Gets a list of sections. The returned list is internal and must not
 *  be modified. It remains valid as long as the configuration object exists.
 *  @param[in] cfg  A configuration object.
 *  @return A read-only string list object.
 */
StrList confGetSections (Conf cfg);


/** Gets a list of assignments in a section. This list uses the same strcmp
 *  function for its keys as the configuration object and contains only unique
 *  values.
 *  @param[in] cfg  A configuration object.
 *  @param[in] section  The name of the section.
 *  @return A string list of assignments (char *).
 */
StrList confGetList (Conf restrict cfg,
	const char *restrict section);

/** Sets a list of assignments in a section. If any key repeats in the list,
 *  the latest value will be used.
 *  @param[in] cfg  A configuration object.
 *  @param[in] section  The name of the section.
 *  @param[in] list  The string list with assignments (char *).
 */
void confSetList (Conf restrict cfg,
	const char *restrict section, StrList restrict list);

/** Get an integer value.
 *  @param[in] cfg  A configuration object.
 *  @param[in] section  The name of the section.
 *  @param[in] key  The name of the key.
 *  @param[out] value  Points to a buffer for the value.
 *  @return 0 if a value was read, -1 otherwise.
 */
int confGetInt (Conf restrict cfg, const char *restrict section,
	const char *restrict key, int *restrict value);

/** Set an integer value.
 *  @param[in] cfg  A configuration object.
 *  @param[in] section  The name of the section.
 *  @param[in] key  The name of the key.
 *  @param[in] value  An integer value to be assigned to the key.
 */
void confSetInt (Conf restrict cfg, const char *restrict section,
	const char *restrict key, int value);

/** @} */

#endif /* ! __XTND_CONF_H__ */

