/**
 *  @file csv.h
 *  @brief CSV files
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Přemysl Janouch 2010. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_CSV_H__
#define __XTND_CSV_H__

/**
 *  @defgroup csv CSV files
 *  @ingroup util
 *
 *  This module allows you to work with CSV files.
 *
 *  There's no support for using semicolons instead of commas,
 *  though it can be easily added.
 *
 *  @{
 */

/** A CSV file reader. */
typedef struct CsvReader *CsvReader;

/** Creates a CSV reader for an opened file stream. */
CsvReader csvReaderNew (FILE *stream);

/** What has been read. */
typedef enum
{
	/** A normal field. */
	CSV_FIELD,
	/** End of record. */
	CSV_EOR,
	/** End of file. */
	CSV_EOF,
	/** An error has occured. */
	CSV_ERROR
}
CsvReadStatus;

/** Read a field from the file.
 *  @param[in]  rdr    A reader object.
 *  @param[out] field  The field data, if the function returns CSV_FIELD.
 *                     You may pass NULL if you don't want it. It has to be
 *                     freed with the standard system free() function.
 */
CsvReadStatus csvRead (CsvReader restrict rdr, char **restrict field);

/** Destroy a CSV reader.
 *  @param[in] rdr  A reader object.
 */
void csvReaderDestroy (CsvReader rdr);


/** A CSV file writer. */
typedef struct CsvWriter *CsvWriter;

/** Creates a CSV reader for an opened file stream. */
CsvWriter csvWriterNew (FILE *stream);

/** Write a field or end of record.
 *  @param[in] wrt    A writer object.
 *  @param[in] field  The field to be written or end of record if NULL.
 *  @return Zero on success, non-zero otherwise.
 */
int csvWrite (CsvWriter restrict wrt, const char *restrict field);

/** Destroy a CSV writer.
 *  @param[in] wrt  A writer object.
 */
void csvWriterDestroy (CsvWriter wrt);

/** @} */

#endif /* ! __XTND_CSV_H__ */

