/**
 *  @file btree.c
 *  @brief A B-tree implementation
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009, 2011, 2012. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


struct BTree
{
	/** The minimum count of items in a node. */
	unsigned min;
	/** The maximum count of items in a node. */
	unsigned max;
	/** Callback function for comparing keys. */
	BtKeyCmpCB keycmp;
	/** Callback function for freeing keys. */
	BtWalkCB free;
	/** Specifies the extra parameter for BtFreeCB. */
	void *freeExtra;
	/** Points to the root node. */
	void **rootNode;
};

/*  A node is formed from an array of (order * 2 - 1) pointers.
 *  The node tree of a B-tree of order 3 could look like this:
 *
 *                           ___ _0_ ___ _1_ ___
 *                          |_P_|_#_|_P_|_#_|_P_|
 *    ______________________/ ________|         \___
 *   /__ _0_ ___ _1_ ___     /__ _0_ ___ _1_ ___    \___ _0_ ___ _1_ ___
 *  |_0_|_#_|_0_|_0_|_0_|   |_0_|_#_|_0_|_#_|_0_|   |_0_|_#_|_0_|_#_|_0_|
 *
 *  where 0 is a NULL pointer, # is non-NULL item data, P is a child pointer
 *  and the numbers above are key indexes.
 */

#define BT_NODE_ALLOC(bt) (xcalloc ((bt)->max * 2 + 1, sizeof (void *)))

#define BT_NODE_BEFORE(node, index) ((node)[2 * (index)])
#define BT_NODE_DATA(node, index)   ((node)[2 * (index) + 1])
#define BT_NODE_AFTER(node, index)  ((node)[2 * (index) + 2])

#define BT_NODE_IS_LEAF(node)       (!*(node))
#define BT_NODE_IS_FULL(bt, node)   ((node)[2 * ((bt)->max) - 1])

#define BT_NODE_SIMPLE_INSERT(node, i, size, key, nodeAfterKey) \
{ \
	/* Move what follows. */ \
	if ((size) - (i)) \
		memmove ((node) + (i) * 2 + 3, (node) + (i) * 2 + 1, \
			((size) - (i)) * 2 * sizeof (void *)); \
	BT_NODE_DATA ((node), (i)) = (key); \
	BT_NODE_AFTER ((node), (i)) = (nodeAfterKey); \
}


/** Recursively delete a node with its children. */
static void btClearNode (BTree bt, void **node);


BTree btCreate (unsigned order, BtKeyCmpCB cmp, BtWalkCB free, void *freeExtra)
{
	BTree bt;

	assert (order >= 3);
	assert (cmp != NULL);

	bt = xmalloc (sizeof (struct BTree));
	bt->max = order - 1;
	bt->min = bt->max / 2;
	bt->keycmp = cmp;
	bt->free = free;
	bt->freeExtra = freeExtra;
	bt->rootNode = NULL;

	return bt;
}

void btFree (BTree bt)
{
	assert (bt != NULL);

	btClear (bt);
	free (bt);
}

void btClear (BTree bt)
{
	assert (bt != NULL);

	if (bt->rootNode)
		btClearNode (bt, bt->rootNode);
	bt->rootNode = NULL;
}

static void btClearNode (BTree bt, void **node)
{
	unsigned i;

	if (bt->free)
		for (i = 0; i < bt->max && BT_NODE_DATA (node, i); i++)
			bt->free (BT_NODE_DATA (node, i), bt->freeExtra);
	for (i = 0; i <= bt->max && BT_NODE_BEFORE (node, i); i++)
		btClearNode (bt, BT_NODE_BEFORE (node, i));

	free (node);
}

void *btFind (BTree restrict bt, const void *restrict key)
{
	void **node, *el;
	unsigned i;
	int c;

	assert (bt != NULL);

	if (!key)
		return NULL;

	for (node = bt->rootNode, i = 0; node;)
	{
		el = BT_NODE_DATA (node, i);
		c = bt->keycmp (el, key);
		if (!c)
			return el;

		if (c < 0)
		{
			i++;
			if (i < bt->max && BT_NODE_DATA (node, i))
				continue;
		}
		node = BT_NODE_BEFORE (node, i);
		i = 0;
	}
	return NULL;
}

/** Compute the length of @a node. */
static inline unsigned btNodeSize (BTree restrict bt, void **restrict node)
{
	unsigned i, size = 0;
	for (i = 0; i < bt->max; i++)
		if (BT_NODE_DATA (node, i))
			size++;
	return size;
}

/** Insert into a full node and split it.
 *  @param[in] bt  A B-tree object.
 *  @param[in] node  The node the item is to be inserted into.
 *  @param[in] i  The index to insert at.
 *  @param[in] item  The item to be inserted.
 *  @param[in] nodeAfterItem  This node follows @a item.
 *  @param[out] upItem  The item that moves up.
 *  @param[out] upNode  This node follows @a upItem.
 */
static void btInsertSplit (BTree restrict bt, void **restrict node, int i,
	void *restrict item, void **restrict nodeAfterItem,
	void **restrict upItem, void **restrict upNode)
{
	void **newNode, **currentNodeAfter, *currentItem;
	int median, sourceIndex, k;

	assert (BT_NODE_IS_FULL (bt, node));

	*upNode = newNode = BT_NODE_ALLOC (bt);
	median = bt->max / 2;

	/* Move data into the new node. */
	for (k = bt->max; ; k--)
	{
		if (k == i)
		{
			currentItem = item;
			currentNodeAfter = nodeAfterItem;
		}
		else
		{
			sourceIndex = k > i ? k - 1 : k;

			currentItem = BT_NODE_DATA (node, sourceIndex);
			currentNodeAfter = BT_NODE_AFTER (node, sourceIndex);
			BT_NODE_DATA (node, sourceIndex) = NULL;
			BT_NODE_AFTER (node, sourceIndex) = NULL;
		}

		if (k == median)
		{
			*upItem = currentItem;
			BT_NODE_BEFORE (newNode, 0) = currentNodeAfter;
			break;
		}

		BT_NODE_DATA (newNode, k - 1 - median) = currentItem;
		BT_NODE_AFTER (newNode, k - 1 - median) = currentNodeAfter;
	}

	/* Since the original node is half-empty now, we do a simple insertion. */
	if (i < median)
		BT_NODE_SIMPLE_INSERT (node, i, median, item, nodeAfterItem);
}

/** The result of a call to btInsertInternal(). */
enum BtInsertInternalResult
{
	/** Item already found in the tree. */
	BT_INSERT_INTERNAL_FOUND,
	/** Item inserted, there was enough space in the node. */
	BT_INSERT_INTERNAL_INSERTED,
	/** Item inserted, node split into two, moving the median up. */
	BT_INSERT_INTERNAL_INSERTED_SPLIT
};

/** Internal B-tree insertion function.
 *  @param[in] bt  A B-tree object.
 *  @param[in] key  An item to be inserted.
 *  @param[in] node  The node the item is to be inserted into.
 *  @param[out] upItem  If split, this is the item that moves up.
 *  @param[out] upNode  If split, this is the new node after @a upItem.
 *  @return See @a BtInsertInternalResult. If the node is already full
 *  	and splits because of this, @a upItem and @a upNode are set.
 */
static enum BtInsertInternalResult btInsertInternal (BTree restrict bt,
	void *restrict item, void **node, void **upItem, void **upNode)
{
	void *nodeAfterItem = NULL;
	unsigned i, insertIndex, nItems;
	int c;

	if (!node)
	{
		*upItem = item;
		*upNode = NULL;
		return BT_INSERT_INTERNAL_INSERTED_SPLIT;
	}

	nItems = btNodeSize (bt, node);

	for (i = 0; ; i++)
	{
		c = bt->keycmp (BT_NODE_DATA (node, i), item);
		if (!c)
			return BT_INSERT_INTERNAL_FOUND;

		/* Insert, append or continue. */
		if (c > 0)
			insertIndex = i;
		else if (i + 1 == nItems)
			insertIndex = i + 1;
		else
			continue;

		if (!BT_NODE_IS_LEAF (node))
		{
			enum BtInsertInternalResult result;

			/* Recurse; if a key is promoted up, just fall through
			 * and insert it as if this node was a leaf node.
			 */
			result = btInsertInternal (bt, item,
				BT_NODE_BEFORE (node, insertIndex),
				(void **) &item, &nodeAfterItem);
			if (result != BT_INSERT_INTERNAL_INSERTED_SPLIT)
				return result;
		}

		/* Full nodes need splitting. */
		if (nItems == bt->max)
		{
			btInsertSplit (bt, node, insertIndex,
				item, nodeAfterItem, upItem, upNode);
			return BT_INSERT_INTERNAL_INSERTED_SPLIT;
		}

		BT_NODE_SIMPLE_INSERT (node, insertIndex, nItems, item, nodeAfterItem);
		return BT_INSERT_INTERNAL_INSERTED;
	}
}

void *btInsert (BTree restrict bt, void *restrict item)
{
	void *upItem, *upNode;
	void **newRoot;

	assert (bt != NULL);

	if (!item)
		return NULL;

	switch (btInsertInternal (bt, item, bt->rootNode, &upItem, &upNode))
	{
	case BT_INSERT_INTERNAL_FOUND:
	default:
		return NULL;
	case BT_INSERT_INTERNAL_INSERTED_SPLIT:
		newRoot = BT_NODE_ALLOC (bt);
		BT_NODE_BEFORE (newRoot, 0) = bt->rootNode;
		BT_NODE_DATA (newRoot, 0) = upItem;
		BT_NODE_AFTER (newRoot, 0) = upNode;
		bt->rootNode = newRoot;
	case BT_INSERT_INTERNAL_INSERTED:
		return item;
	}
}

enum BtDeleteInternalResult
{
	BT_DELETE_INTERNAL_OK,
	BT_DELETE_INTERNAL_DEFICIENT
};

/** Delete item on index @a i and its right child from @a node,
 *  which has @a size items total, and return if the node has become deficient.
 */
static inline enum BtDeleteInternalResult btDeleteSimple (BTree restrict bt,
	void **restrict node, unsigned i, unsigned size)
{
	if (--size - i)
		memmove (node + i * 2 + 1, node + i * 2 + 3,
			(size - i) * 2 * sizeof *node);
	BT_NODE_DATA (node, size) = NULL;
	BT_NODE_AFTER (node, size) = NULL;

	if (size < bt->min)
		return BT_DELETE_INTERNAL_DEFICIENT;
	return BT_DELETE_INTERNAL_OK;
}

/** Fix the deficient left child of item @a i in @a node, sized @a size. */
static enum BtDeleteInternalResult btRebalance (BTree restrict bt,
	void **restrict node, unsigned i, unsigned size)
{
	unsigned defSize, leftSize, rightSize;
	void **defNode, **leftSibling = NULL, **rightSibling = NULL;

	defNode = BT_NODE_BEFORE (node, i);
	defSize = btNodeSize (bt, defNode);

	if (i > 0)
	{
		leftSibling = BT_NODE_BEFORE (node, i - 1);
		leftSize = btNodeSize (bt, leftSibling);
		/* Try to steal an item from the left sibling. */
		if (leftSize > bt->min)
		{
			/* Make place for an item. */
			if (defSize)
				memmove (defNode + 2, defNode,
					(defSize * 2 + 1) * sizeof *defNode);
			/* Use the sibling's rightmost subtree as our leftmost. */
			leftSize--;
			BT_NODE_BEFORE (defNode, 0) = BT_NODE_AFTER (leftSibling, leftSize);
			/* And right shift the item via the parent node. */
			BT_NODE_DATA (defNode, 0) = BT_NODE_DATA (node, i - 1);
			BT_NODE_DATA (node, i - 1) = BT_NODE_DATA (leftSibling, leftSize);
			/* Delete the stolen item. */
			BT_NODE_DATA (leftSibling, leftSize) = NULL;
			BT_NODE_AFTER (leftSibling, leftSize) = NULL;
			return BT_DELETE_INTERNAL_OK;
		}
	}
	if (i < size)
	{
		rightSibling = BT_NODE_AFTER (node, i);
		rightSize = btNodeSize (bt, rightSibling);
		/* Try to steal an item from the right sibling. */
		if (rightSize > bt->min)
		{
			/* Use the sibling's leftmost subtree as our rightmost. */
			BT_NODE_AFTER (defNode, defSize) = BT_NODE_BEFORE (rightSibling, 0);
			/* And left shift the item via the parent node. */
			BT_NODE_DATA (defNode, defSize) = BT_NODE_DATA (node, i);
			BT_NODE_DATA (node, i) = BT_NODE_DATA (rightSibling, 0);
			/* Delete the stolen item. */
			rightSize--;
			memmove (rightSibling, rightSibling + 2,
				(rightSize * 2 + 1) * sizeof *defNode);
			BT_NODE_DATA (rightSibling, rightSize) = NULL;
			BT_NODE_AFTER (rightSibling, rightSize) = NULL;
			return BT_DELETE_INTERNAL_OK;
		}
	}

	/* We have to merge the node with one of the siblings. */
	if (leftSibling)
	{
		/* Move the separator and the deficient node to the sibling. */
		BT_NODE_DATA (leftSibling, leftSize++) = BT_NODE_DATA (node, i - 1);
		memmove (leftSibling + leftSize * 2, defNode,
			(defSize * 2 + 1) * sizeof *defNode);
		free (defNode);
		/* Remove the separator and the deficient node from the parent. */
		return btDeleteSimple (bt, node, i - 1, size);
	}
	if (rightSibling)
	{
		/* Move the separator and the sibling to the deficient node. */
		BT_NODE_DATA (defNode, defSize++) = BT_NODE_DATA (node, i);
		memmove (defNode + defSize * 2, rightSibling,
			(rightSize * 2 + 1) * sizeof *defNode);
		free (rightSibling);
		/* Remove the separator and the right sibling from the parent. */
		return btDeleteSimple (bt, node, i, size);
	}

	abort ();
}

/** Like btDeleteInternal(), but this removes the leftmost item in a subtree. */
static enum BtDeleteInternalResult btStealSuccessor (BTree restrict bt,
	void **restrict node, void **restrict successor)
{
	unsigned nItems;
	enum BtDeleteInternalResult result;

	assert (node != NULL);
	nItems = btNodeSize (bt, node);

	/* Delete the leftmost item in this subtree and return it. */
	if (BT_NODE_IS_LEAF (node))
	{
		*successor = BT_NODE_DATA (node, 0);
		return btDeleteSimple (bt, node, 0, nItems);
	}

	result = btStealSuccessor (bt, BT_NODE_BEFORE (node, 0), successor);
	if (result == BT_DELETE_INTERNAL_DEFICIENT)
		return btRebalance (bt, node, 0, nItems);
	return BT_DELETE_INTERNAL_OK;
}

/** Remove an item matching with @a key from the subtree rooted at @a node,
 *  store this item in @a outputItem and return whether the node has become
 *  deficient.
 */
static enum BtDeleteInternalResult btDeleteInternal (BTree restrict bt,
	void **restrict node, const void *restrict key, void **restrict outputItem)
{
	unsigned i, nItems;
	enum BtDeleteInternalResult result;
	int c;

	if (!node)
		return BT_DELETE_INTERNAL_OK;

	nItems = btNodeSize (bt, node);

	for (i = 0; ; i++)
	{
		c = bt->keycmp (BT_NODE_DATA (node, i), key);
		if (!c)
		{
			*outputItem = BT_NODE_DATA (node, i);
			if (BT_NODE_IS_LEAF (node))
				return btDeleteSimple (bt, node, i, nItems);

			/* If we have to remove from a non-leaf node and the node
			 * becomes deficient, we must replace it with its successor
			 * (or predecessor, which is not implemented here). */
			result = btStealSuccessor (bt, BT_NODE_AFTER (node, i),
				&BT_NODE_DATA (node, i));
			i++;
		}
		else if (c > 0)
			result = btDeleteInternal (bt, BT_NODE_BEFORE (node, i),
				key, outputItem);
		else if (i + 1 == nItems)
			result = btDeleteInternal (bt, BT_NODE_BEFORE (node, ++i),
				key, outputItem);
		else
			continue;

		if (result == BT_DELETE_INTERNAL_DEFICIENT)
			return btRebalance (bt, node, i, nItems);
		return BT_DELETE_INTERNAL_OK;
	}
}

void *btDelete (BTree restrict bt, const void *restrict key)
{
	void *item = NULL;

	assert (bt != NULL);

	if (!key)
		return NULL;

	/* If the root has no elements but one child, replace it with the child.
	 * Otherwise we don't care if it's deficient.
	 */
	if (btDeleteInternal (bt, bt->rootNode, key, &item)
		== BT_DELETE_INTERNAL_DEFICIENT && !BT_NODE_DATA (bt->rootNode, 0))
	{
		void **root;

		root = bt->rootNode;
		bt->rootNode = BT_NODE_BEFORE (bt->rootNode, 0);
		free (root);
	}

	return item;
}

static void btWalkInternal (BTree restrict bt, void **restrict node,
	BtWalkCB walk, void *restrict extra)
{
	unsigned i;

	if (!node)
		return;

	for (i = 0; i < bt->max && BT_NODE_DATA (node, i); i++)
	{
		btWalkInternal (bt, BT_NODE_BEFORE (node, i), walk, extra);
		walk (BT_NODE_DATA (node, i), extra);
	}
	btWalkInternal (bt, BT_NODE_BEFORE (node, i), walk, extra);
}

void btWalk (BTree restrict bt, BtWalkCB walk, void *restrict extra)
{
	assert (bt != NULL);
	assert (walk != NULL);

	btWalkInternal (bt, bt->rootNode, walk, extra);
}
