/**
 *  @file btree.h
 *  @brief B-tree
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009, 2011, 2012. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_BTREE_H__
#define __XTND_BTREE_H__

/**
 *  @defgroup btree B-tree
 *  @ingroup ds
 *
 *  This module implements a B-tree with no duplicates allowed.
 *
 *  @{
 */

/** The B-tree class. */
typedef struct BTree *BTree;

/** Defines a callback function for comparing two items.
 *  It should return > 0 if @a k1 is greater than @a k2,
 *  0 if they're equal and < 0 if @a k2 is greater than @a k1.
 */
typedef int (*BtKeyCmpCB) (const void *k1, const void *k2);

/** Defines a callback for going through items, given as the first argument
 *  of the function.  The second argument is user-defined.
 */
typedef void (*BtWalkCB) (void *p, void *extra);


/** Create a new B-tree object.
 *  @param[in] order  The order of the B-tree (one less than the maximum
 *                    count of children a node can have).
 *  @param[in] cmp  A callback function for comparing items.
 *  @param[in] free  Function for freeing items (can be NULL).
 *  @param[in] freeExtra  The extra parameter given to @a free.
 *  @return A handle to the object.
 *  @see btFree()
 */
BTree btCreate (unsigned order, BtKeyCmpCB cmp, BtWalkCB free, void *freeExtra);

/** Free a B-tree object.
 *  @param[in] bt  A B-tree object.
 *  @see btCreate()
 */
void btFree (BTree bt);

/** Clear a B-tree.
 *  @param[in] bt  A B-tree object.
 */
void btClear (BTree bt);

/** Try to find an item matching with the key and return it.
 *  @param[in] bt  A B-tree object.
 *  @param[in] key  Specifies the item to search for.
 *  @return The stored item if found, NULL otherwise.
 */
void *btFind (BTree restrict bt, const void *restrict key);

/** Insert an item into the B-tree.
 *  @param[in] bt  A B-tree object.
 *  @param[in] item  The item to insert.
 *  @return @a item if the item was inserted, NULL otherwise.
 */
void *btInsert (BTree restrict bt, void *restrict item);

/** Delete the item matching with the key and return it.
 *  @param[in] bt  A B-tree object.
 *  @param[in] key  Specifies the item to be deleted.
 *  @return The stored item if found and deleted, NULL otherwise.
 */
void *btDelete (BTree restrict bt, const void *restrict key);

/** Walk through all items in the tree in ascending order.
 *  @param[in] bt  A B-tree object.
 *  @param[in] walk  A user-defined walk callback.
 *  @param[in] extra  A user-defined parameter for @a walk.
 */
void btWalk (BTree restrict bt, BtWalkCB walk, void *restrict extra);

/** @} */

#endif /* ! __XTND_BTREE_H__ */

