/**
 *  @file dynarray.c
 *  @brief A dynamic array
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2007 - 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


struct DynArray
{
	/** An internal memory stream object. */
	MemStream ms;
	/** Specifies the size of one item. */
	size_t itemsize;
	/** Points to a function that will be called before an item will be
	 *  deleted or overriden.
	 */
	DaFreeCB free;
	/** Specifies the extra parameter for DaFreeCB. */
	void *freeExtra;
};


DynArray daCreate (size_t itemsize, DaFreeCB free, void *freeExtra)
{
	DynArray d;

	if (!itemsize)
		return NULL;

	d = xmalloc (sizeof (struct DynArray));

	/* TODO: Don't enforce a MemoryStream,
	 *       make a Stream interface.
	 */

	d->ms = msCreate (0);
	d->itemsize = itemsize;
	d->free = free;
	d->freeExtra = freeExtra;

	return d;
}

void daFree (DynArray d)
{
	daClear (d);
	msFree (d->ms);
	free (d);
}

void daClear (DynArray d)
{
	if (d->free)
	{
		int i, max;

		max = daGetSize (d);
		for (i = 0; i < max; i++)
			d->free (daGet (d, i), d->freeExtra);
	}
	msClear (d->ms);
}

unsigned int daGetSize (DynArray d)
{
	return msGetSize (d->ms) / d->itemsize;
}

void *daGet (DynArray d, unsigned int index)
{
	if (index >= daGetSize (d))
		return NULL;

	return (char *) msGetPtr (d->ms) + index * d->itemsize;
}

void daSet (DynArray d, unsigned int index, void *item)
{
	if (index >= daGetSize (d))
		return;

	if (d->free)
		d->free (daGet (d, index), d->freeExtra);

	msSetPos (d->ms, index * d->itemsize, 0);
	msWrite (d->ms, item, d->itemsize, 1);
}

void daDelete (DynArray d, unsigned int index)
{
	if (index >= daGetSize (d))
		return;

	if (d->free)
		d->free (daGet (d, index), d->freeExtra);

	msMove (d->ms, (index + 1) * d->itemsize, index * d->itemsize,
		(daGetSize (d) - index - 1) * d->itemsize, 1);
	msSetSize (d->ms, msGetSize (d->ms) - d->itemsize);
}

void daInsert (DynArray d, unsigned int index, void *item)
{
	if (index >= daGetSize (d))
		return;

	msSetSize (d->ms, msGetSize (d->ms) + d->itemsize);
	msMove (d->ms, index * d->itemsize, (index + 1) * d->itemsize,
		(daGetSize (d) - index - 1) * d->itemsize, 1);
	daSet (d, index, item);
}

int daAppend (DynArray d, void *item)
{
	int index;
	index = msGetSize (d->ms);

	msSetPos (d->ms, index, 0);
	msWrite (d->ms, item, d->itemsize, 1);
	return index / d->itemsize;
}
