/**
 *  @file dynarray.h
 *  @brief Dynamic array
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2007 - 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_DYNARRAY_H__
#define __XTND_DYNARRAY_H__

/**
 *  @defgroup dynarray Dynamic array
 *  @ingroup ds
 *
 *  This module implements a dynamic array.
 *
 *  @{
 */

/** The dynamic array class. The purpose of the dynamic array class is
 *  to manage dynamic arrays.
 */
typedef struct DynArray *DynArray;

/** Defines a function for freeing items in a dynamic array (ie.\ the item
 *  destructor). The function takes two arguments. The first is a pointer to
 *  the data which have to be freed. The second argument is unspecified.
 */
typedef void (*DaFreeCB) (const void *p, void *extra);


/** Create a new dynamic array object.
 *  @param[in] itemsize  The size of one item
 *  @param[in] free  Function for freeing items (can be NULL)
 *  @param[in] freeExtra  The extra parameter given to @a free.
 *  @return A pointer to the object.
 *  @see daFree()
 */
DynArray daCreate (size_t itemsize, DaFreeCB free, void *freeExtra);

/** Free a dynamic array.
 *  @param[in] d  The dynamic array object
 *  @see daCreate()
 */
void daFree (DynArray d);

/** Clear a dynamic array.
 *  @param[in,out] d  The dynamic array object
 */
void daClear (DynArray d);

/** Get size (ie.\ count of items) of a dynamic array.
 *  @param[in] d  The dynamic array object
 *  @return The size.
 */
unsigned int daGetSize (DynArray d);

/** Get an item from the array.
 *  @param[in] d  The dynamic array object
 *  @param[in] index  Item index
 *  @return A pointer to the item.
 *  @see daSet()
 */
void *daGet (DynArray d, unsigned int index);

/** Set an item from the array to @a item.
 *  @note If a function for freeing items was specified, it will be called
 *  for the item that is being overriden by this function.
 *  @param[in,out] d  The dynamic array object
 *  @param[in] index  Item index
 *  @param[in] item  New item data
 *  @see daGet()
 */
void daSet (DynArray d, unsigned int index, void *item);

/** Delete an item from a dynamic array.
 *  @param[in,out] d  The dynamic array object
 *  @param[in] index  Item index
 */
void daDelete (DynArray d, unsigned int index);

/** Insert an item to a dynamic array.
 *  @param[in,out] d  The dynamic array object
 *  @param[in] index  Item index
 *  @param[in] item  The new item data
 */
void daInsert (DynArray d, unsigned int index, void *item);

/** Append an item to the end of a dynamic array.
 *  @param[in] d  The dynamic array object
 *  @param[in] item  The new item data
 *  @return The index of the appended item.
 */
int daAppend (DynArray d, void *item);

/** @} */

#endif /* ! __XTND_DYNARRAY_H__ */

