/**
 *  @file hmap.c
 *  @brief A hash map implementation
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2012. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


#define HMAP_MAX_LOAD_FACTOR   0.8
#define HMAP_MIN_LOAD_FACTOR   0.2
#define HMAP_DEFAULT_SIZE     16


typedef struct HMapItem HMapItem;

struct HMapItem
{
	void *key;
	void *value;

	HMapItem *next;
};

struct HMap
{
	size_t      size;
	size_t      used;
	HMapItem  **items;

	HmHashCB    key_hash;
	HmKeyCmpCB  key_compare;
	HmFreeCB    key_free;
	HmFreeCB    value_free;
};


HMap
hmCreate (size_t n, HmHashCB keyH, HmKeyCmpCB keyC,
	HmFreeCB keyF, HmFreeCB valueF)
{
	HMap m;

	assert (keyH != NULL);
	assert (keyC != NULL);

	m = xcalloc (sizeof *m, 1);

	if (!n)
		n = HMAP_DEFAULT_SIZE;

	m->size = n;
	m->used = 0;
	m->items = xcalloc (sizeof *m->items, n);

	m->key_hash    = keyH;
	m->key_compare = keyC;
	m->key_free    = keyF;
	m->value_free  = valueF;
	return m;
}

void
hmClear (HMap m)
{
	unsigned i;
	HMapItem *item, *tmp;

	assert (m != NULL);

	for (i = 0; i < m->size; i++)
		for (item = m->items[i]; item; )
	{
		if (m->key_free)
			m->key_free   (item->key);
		if (m->value_free)
			m->value_free (item->value);

		tmp = item;
		item = item->next;
		free (tmp);
	}

	memset (m->items, 0, sizeof *m->items * m->size);
}

void
hmFree (HMap m)
{
	assert (m != NULL);

	hmClear (m);
	free (m->items);
	free (m);
}

#define HM_DEFINE_FIND(type, name, true_result, false_result)                \
type                                                                         \
name (HMap m, const void *key)                                               \
{                                                                            \
	HMapItem *item;                                                          \
	int c;                                                                   \
                                                                             \
	assert (m != NULL);                                                      \
                                                                             \
	item = m->items[m->key_hash (key) % m->size];                            \
	for (; item; item = item->next)                                          \
	{                                                                        \
		c = m->key_compare (key, item->key);                                 \
		if (!c)                                                              \
			return (true_result);                                            \
		if (c < 0)                                                           \
			/* Not present in the sorted list. */                            \
			break;                                                           \
	}                                                                        \
	return (false_result);                                                   \
}

HM_DEFINE_FIND (unsigned, hmIsElement, TRUE,        FALSE)
HM_DEFINE_FIND (void *,   hmFind,      item->value, NULL)

#define HM_FIND_FOREACH(key_, body)                                          \
	p = &m->items[m->key_hash ((key_)) % m->size];                           \
	for (; (item = *p); p = &item->next) {                                   \
		c = m->key_compare ((key_), item->key);                              \
		if (!c) {                                                            \
			body                                                             \
		} else if (c < 0)                                                    \
			/* Not present in the sorted list. */                            \
			break;                                                           \
	}

/** Simplified hmInsert(). */
static void
hmResizeInsert (HMap m, HMapItem *new_item)
{
	HMapItem **p, *item;
	int c;

	HM_FIND_FOREACH (new_item->key,
		assert (c != 0);
	)

	new_item->next = *p;
	*p = new_item;
}

/** Resize the map and reinsert entries. */
static void
hmResize (HMap m, size_t new_size)
{
	HMapItem **old_items, *item, *tmp;
	size_t old_size;

	old_items = m->items;
	old_size = m->size;

	m->items = xcalloc (new_size, sizeof *m->items);
	m->size = new_size;

	while (old_size--)
		for (item = old_items[old_size]; item; item = tmp)
	{
		tmp = item->next;
		hmResizeInsert (m, item);
	}

	free (old_items);
}

void
hmInsert (HMap m, void *key, void *value)
{
	HMapItem **p, *item, *new_item;
	int c;

	assert (m != NULL);

	HM_FIND_FOREACH (key,
		if (m->key_free)
			m->key_free   (key);
		if (m->value_free)
			m->value_free (item->value);

		item->value = value;
		return;
	)

	new_item = xmalloc (sizeof *new_item);
	new_item->key = key;
	new_item->value = value;
	new_item->next = *p;
	*p = new_item;

	if (++m->used > HMAP_MAX_LOAD_FACTOR * m->size)
		hmResize (m, m->size << 1);
}

unsigned
hmDelete (HMap m, const void *key)
{
	HMapItem **p, *item;
	int c;

	assert (m != NULL);

	HM_FIND_FOREACH (key,
		if (m->key_free)
			m->key_free   (item->key);
		if (m->value_free)
			m->value_free (item->value);

		*p = item->next;
		free (item);

		if (--m->used < HMAP_MIN_LOAD_FACTOR * m->size)
			hmResize (m, m->size >> 1);
		return TRUE;
	)

	return FALSE;
}

unsigned
hmFNVHash (const void *data, size_t size)
{
	unsigned const char *p = data;
	unsigned i, h = 2166136261;

	for (i = 0; i < size; i++)
		h = (h * 16777619) ^ p[i];

	return h;
}
