/**
 *  @file hmap.h
 *  @brief Hash map
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2012. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_HMAP_H__
#define __XTND_HMAP_H__

/**
 *  @defgroup hmap Hash map
 *  @ingroup ds
 *
 *  This module implements a hash map.
 *
 *  @{
 */

/** The hash map class. */
typedef struct HMap *HMap;

/** Defines a callback function for computing hashes. */
typedef unsigned (*HmHashCB) (const void *key);

/** Defines a callback function for comparing keys. */
typedef int (*HmKeyCmpCB) (const void *k1, const void *k2);

/** Defines a callback function for freeing data. */
typedef void (*HmFreeCB) (void *p);


/** Create a new hash map object.
 *  @param[in] n  Initial table size.  0 for a default value.
 *  @param[in] keyH  Key hashing function.
 *  @param[in] keyC  Key comparison function.
 *  @param[in] keyF  Key freeing function.  May be NULL.
 *  @param[in] valueF  Value freeing function.  May be NULL.
 *  @return A handle to the object.
 *  @see hmFree()
 */
HMap hmCreate (size_t n, HmHashCB keyH, HmKeyCmpCB keyC,
	HmFreeCB keyF, HmFreeCB valueF);

/** Clear the map of all contents.
 *  @param[in] m  A hash map object.
 */
void hmClear (HMap m);

/** Free a hash map object.
 *  @param[in] m  A hash map object.
 */
void hmFree (HMap m);

/** Look if a key is present in the map.
 *  @param[in] m  A hash map object.
 *  @param[in] key  The key to look for.
 *  @return TRUE if found, FALSE otherwise.
 */
unsigned hmIsElement (HMap m, const void *key);

/** Search for a key and return its associated value.
 *  @param[in] m  A hash map object.
 *  @param[in] key  The key to look for.
 *  @return The value associated to the key.  NULL if not found.
 */
void *hmFind (HMap m, const void *key);

/** Insert a key-value pair into the map.  The map assumes ownership of
 *  the pair and frees them when appropriate.
 *  @param[in] m  A hash map object.
 *  @param[in] key  The key.  May be NULL, too.
 *  @param[in] value  The value.  May be NULL, too.
 */
void hmInsert (HMap m, void *key, void *value);

/** Delete a key from the map.
 *  @param[in] m  A hash map object.
 *  @param[in] key  The key to be deleted from the map.
 *  @return TRUE if the key was found and deleted, FALSE otherwise.
 */
unsigned hmDelete (HMap m, const void *key);


/** Compute the FNV hash of supplied data.
 *  @param[in] data  A pointer to data.
 *  @param[in] size  Length of the data in bytes.
 *  @return A 32-bit FNV hash.
 */
unsigned hmFNVHash (const void *data, size_t size);

/** @} */

#endif /* ! __XTND_HMAP_H__ */
