/**
 *  @file memstream.c
 *  @brief A memory stream
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2007 - 2009, 2011. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


struct MemStream
{
	/** A pointer to the data of a memory stream object.
	 *  @see msGetPtr()
	 */
	void *ptr;
	/** How much memory was allocated for data. */
	size_t alloc;
	/** The actual size of data.
	 *  @see msGetSize(), msSetSize()
	 */
	size_t size;
	/** The position data will be read from or written to.
	 *  @see msGetPos(), msSetPos()
	 */
	unsigned int pos;
	/** Integer byte order. */
	MsByteOrder byteOrder;
};

#ifdef WORDS_BIGENDIAN
	#define MS_BYTE_ORDER_SWAP MS_BYTE_ORDER_LE
#else /* ! WORDS_BIGENDIAN */
	#define MS_BYTE_ORDER_SWAP MS_BYTE_ORDER_BE
#endif /* ! WORDS_BIGENDIAN */

#define MS_SWAP_16(i) ((((i) & 0xFF) << 8) | (((i) >> 8) & 0xFF))


/** The standard size of a memory page. */
static size_t msBlockSize = 4096;


MemStream msCreate (size_t size)
{
	MemStream ms;

	ms = xmalloc (sizeof (struct MemStream));
	ms->ptr = NULL;
	ms->alloc = ms->size = ms->pos = 0;
	msSetSize (ms, size);

	return ms;
}

void msFree (MemStream ms)
{
	assert (ms != NULL);

	if (ms->ptr)
		free (ms->ptr);
	free (ms);
}

void *msGetPtr (MemStream ms)
{
	assert (ms != NULL);
	return ms->ptr;
}

void *msRelease (MemStream ms)
{
	void *p;

	assert (ms != NULL);

	p = ms->ptr;
	ms->ptr = NULL;
	ms->alloc = ms->size = ms->pos = 0;
	return p;
}

void msClear (MemStream ms)
{
	assert (ms != NULL);

	if (ms->ptr)
		free (ms->ptr);
	ms->ptr = NULL;
	ms->alloc = ms->size = ms->pos = 0;
}

size_t msGetSize (MemStream ms)
{
	assert (ms != NULL);
	return ms->size;
}

void msSetSize (MemStream ms, size_t size)
{
	assert (ms != NULL);

	if (ms->alloc < size || (ms->alloc >> 1) > size)
	{
		size_t prevSize;

		prevSize = ms->size;
		ms->alloc = (size / msBlockSize + 1) * msBlockSize;
		ms->ptr = xrealloc (ms->ptr, ms->alloc);

		if (prevSize < size)
			memset ((char *) ms->ptr + prevSize, '\0', size - prevSize);
	}
	ms->size = size;
}

int msEOF (MemStream ms)
{
	assert (ms != NULL);
	return ms->pos >= ms->size;
}

unsigned int msGetPos (MemStream ms)
{
	assert (ms != NULL);
	return ms->pos;
}

void msSetPos (MemStream ms, unsigned int pos, int relative)
{
	assert (ms != NULL);

	if (relative)
		ms->pos += pos;
	else
		ms->pos = pos;
}

void msSetByteOrder (MemStream ms, MsByteOrder order)
{
	assert (ms != NULL);
	assert (order == MS_BYTE_ORDER_LE
		 || order == MS_BYTE_ORDER_BE
		 || order == MS_BYTE_ORDER_HE);

	ms->byteOrder = order;
}

MsByteOrder msGetByteOrder (MemStream ms)
{
	assert (ms != NULL);
	return ms->byteOrder;
}

int msReadInt8 (MemStream ms, int8_t *buffer)
{
	return msRead (ms, buffer, sizeof (int8_t),  1);
}

int msReadUint8 (MemStream ms, uint8_t *buffer)
{
	return msReadInt8 (ms, (int8_t *) buffer);
}

void msWriteInt8 (MemStream ms, int8_t number)
{
	msWrite (ms, &number, sizeof (int8_t), 1);
}

void msWriteUint8 (MemStream ms, uint8_t number)
{
	msWriteInt8 (ms, (int8_t) number);
}

int msReadInt16 (MemStream ms, int16_t *buffer)
{
	assert (ms != NULL);

	if (ms->byteOrder == MS_BYTE_ORDER_SWAP)
	{
		int16_t i;

		if (!msRead (ms, &i, sizeof (int16_t), 1))
			return 0;

		*buffer = MS_SWAP_16 (i);
		return 1;
	}
	return msRead (ms, buffer, sizeof (int16_t), 1);
}

int msReadUint16 (MemStream ms, uint16_t *buffer)
{
	return msReadInt16 (ms, (int16_t *) buffer);
}

void msWriteInt16 (MemStream ms, int16_t number)
{
	assert (ms != NULL);

	if (ms->byteOrder == MS_BYTE_ORDER_SWAP)
	{
		int16_t i;

		i = MS_SWAP_16 (number);
		msWrite (ms, &i, sizeof (int16_t), 1);
	}
	else
		msWrite (ms, &number, sizeof (int16_t), 1);
}

void msWriteUint16 (MemStream ms, uint16_t number)
{
	msWriteInt16 (ms, (int16_t) number);
}

int msReadStr (MemStream ms, char *str, int max)
{
	int read = 0;
	char c;

	assert (ms != NULL);

	while (ms->pos < ms->size && max--)
	{
		c = *((char *) ms->ptr + ms->pos++);
		*str++ = c;
		read++;

		if (c == '\0')
			break;
	}
	return read;
}

void msWriteStr (MemStream ms, const char *s)
{
	msWrite (ms, s, strlen (s) + 1, 1);
}

int msWrite (MemStream ms, const void *data, size_t size, int count)
{
	int written = 0;

	assert (ms != NULL);
	assert (data != NULL);

	if (ms->size < ms->pos + size * count)
		msSetSize (ms, ms->pos + size * count);

	while (count--)
	{
		memcpy ((char *) ms->ptr + ms->pos, data, size);
		ms->pos += size;
		data = (const char *) data + size;
		written++;
	}
	return written;
}

int msRead (MemStream ms, void *buffer, size_t size, int count)
{
	int read = 0;

	assert (ms != NULL);
	assert (buffer != NULL);

	while (count-- && (ms->size >= ms->pos + size))
	{
		memcpy (buffer, (char *) ms->ptr + ms->pos, size);
		ms->pos += size;
		buffer = (char *) buffer + size;
		read++;
	}
	return read;
}

void msMove (MemStream ms, unsigned int from, unsigned int to,
	size_t size, int justcopy)
{
	assert (ms != NULL);

	/* Boundary check. */
	if (from + size > ms->size || to + size > ms->size)
		return;

	if (justcopy)
		memmove ((char *) ms->ptr + to, (char *) ms->ptr + from, size);
	else
	{
		void *buff;
		int mvlen, mvfrom, mvto;

		if (from < to)
		{
			mvlen = to - from;
			mvfrom = from + size;
			mvto = from;
		}
		else
		{
			mvlen = from - to;
			mvfrom = to;
			mvto = to + size;
		}

		buff = xmalloc (mvlen);
		memcpy (buff, (char *) ms->ptr + mvfrom, mvlen);
		memmove ((char *) ms->ptr + to, (char *) ms->ptr + from, size);
		memcpy ((char *) ms->ptr + mvto, buff, mvlen);
		free (buff);
	}
}

