/**
 *  @file memstream.h
 *  @brief Memory stream
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2007 - 2009, 2011. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_MEMSTREAM_H__
#define __XTND_MEMSTREAM_H__

/**
 *  @defgroup memstream Memory stream
 *  @ingroup ds
 *
 *  This module implements a memory stream.
 *
 *  @{
 */

/** The memory stream class. The purpose of the memory stream class is
 *  to simplify the manipulation of data within the memory.
 */
typedef struct MemStream *MemStream;

/** The byte order in which numerical values are stored. */
typedef enum
{
	/** Host endian; the native endianity of this CPU. */
	MS_BYTE_ORDER_HE,
	/** Little endian; least significant byte first. */
	MS_BYTE_ORDER_LE,
	/** Big endian; most significant byte first. */
	MS_BYTE_ORDER_BE
}
MsByteOrder;

/** Creates a new memory stream object and initializes it.
 *  The initial size of data is set to @a size bytes.
 *  @param[in] size  The initial size
 *  @return A pointer to the object.
 *  @see msFree()
 */
MemStream msCreate (size_t size);

/** Deallocate the stream and its data.
 *  @param[in] ms  The memory stream object
 *  @see msCreate()
 */
void msFree (MemStream ms);

/** Set the memory stream byte order.
 *  @param[in,out] ms  The memory stream object.
 *  @param[in] order  The new byte order.
 *  @see msGetByteOrder()
 */
void msSetByteOrder (MemStream ms, MsByteOrder order);

/** Get the memory stream byte order.
 *  @param[in] ms  The memory stream object.
 *  @return The byte order value.
 *  @see msSetByteOrder()
 */
MsByteOrder msGetByteOrder (MemStream ms);

/** Read a signed 8-bit integer value from the memory stream.
 *  @param[in,out] ms  The memory stream object
 *  @param[out] buffer  Where the value is returned.
 *  @return 1 on success, 0 on end of stream.
 */
int msReadInt8 (MemStream ms, int8_t *buffer);

/** Read an unsigned 8-bit integer value from the memory stream.
 *  @param[in,out] ms  The memory stream object
 *  @param[out] buffer  Where the value is returned.
 *  @return 1 on success, 0 on end of stream.
 */
int msReadUint8 (MemStream ms, uint8_t *buffer);

/** Write a signed 8-bit integer value to the memory stream.
 *  @param[in,out] ms  The memory stream object
 *  @param[in] number  The integer value to write
 */
void msWriteInt8 (MemStream ms, int8_t number);

/** Write an unsigned 8-bit integer value to the memory stream.
 *  @param[in,out] ms  The memory stream object
 *  @param[in] number  The integer value to write
 */
void msWriteUint8 (MemStream ms, uint8_t number);

/** Read a signed 16-bit integer value from the memory stream.
 *  @param[in,out] ms  The memory stream object
 *  @param[out] buffer  Where the value is returned.
 *  @return 1 on success, 0 on end of stream.
 */
int msReadInt16 (MemStream ms, int16_t *buffer);

/** Read an unsigned 16-bit integer value from the memory stream.
 *  @param[in,out] ms  The memory stream object
 *  @param[out] buffer  Where the value is returned.
 *  @return 1 on success, 0 on end of stream.
 */
int msReadUint16 (MemStream ms, uint16_t *buffer);

/** Write a signed 16-bit integer value to the memory stream.
 *  @param[in,out] ms  The memory stream object
 *  @param[in] number  The integer value to write
 */
void msWriteInt16 (MemStream ms, int16_t number);

/** Write an unsigned 16-bit integer value to the memory stream.
 *  @param[in,out] ms  The memory stream object
 *  @param[in] number  The integer value to write
 */
void msWriteUint16 (MemStream ms, uint16_t number);

/** Read a string from the memory stream. Reads until '\\0' is reached
 *  or the buffer is full.
 *  @param[in,out] ms  The memory stream object.
 *  @param[out] str  A buffer for the string
 *  @param[in] max  Maximum length of the string
 *  @return Count of characters stored in the buffer.
 */
/* FIXME: Very stupid API, should return a string. */
int msReadStr (MemStream ms, char *str, int max);

/** Write a string to the memory stream, including the '\\0' character.
 *  @param[in,out] ms  The memory stream object
 *  @param[in] s  The string to be written
 */
void msWriteStr (MemStream ms, const char *s);

/** Read @a count elements, store them in @a buffer and advance the position.
 *  Does not try to read an element partially.
 *  @param[in] ms  The memory stream object
 *  @param[out] buffer  The output buffer
 *  @param[in] size  The size of an element
 *  @param[in] count  The count of elements to read
 *  @return The real count of elements stored in @a buffer.
 */
int msRead (MemStream ms, void *buffer, size_t size, int count);

/** Write @a count elements to the memory stream and advance the position.
 *  @param[in,out] ms  The memory stream object
 *  @param[in] data  A pointer to the data to write
 *  @param[in] size  The size of an element
 *  @param[in] count  The count of elements to write
 *  @return The real count of elements written to the memory stream.
 */
int msWrite (MemStream ms, const void *data, size_t size, int count);

/** Move @a size bytes from position @a from to position @a to.
 *  @param[in,out] ms  The memory stream object
 *  @param[in] from  The position to copy/move from
 *  @param[in] to  The position to copy/move to
 *  @param[in] size  How many bytes to copy/move
 *  @param[in] justcopy  If non-zero, just copy the data
 */
void msMove (MemStream ms, unsigned int from, unsigned int to,
	size_t size, int justcopy);

/** Gets the EOF state of the memory stream.
 *  @param[in] ms  The memory stream object
 *  @return 1 if there is no more data to read, 0 otherwise.
 */
int msEOF (MemStream ms);

/** Get the current position of the memory stream.
 *  @param[in] ms  The memory stream object
 *  @return The position.
 */
unsigned int msGetPos (MemStream ms);

/** Set the current position of the memory stream.
 *  @param[in,out] ms  The memory stream object
 *  @param[in] pos  The new position
 *  @param[in] relative  Whether the position is relative to the current one
 */
void msSetPos (MemStream ms, unsigned int pos, int relative);

/** Get the pointer to the data of the stream.
 *  @param[in] ms  The memory stream object
 *  @return The pointer (can be NULL if the size is zero).
 */
void *msGetPtr (MemStream ms);

/** Resizes the memory stream. The newly available space is filled with zeros.
 *  If the required size is bigger or much smaller (one half), then reallocates.
 *  @param[in,out] ms  The memory stream object
 *  @param[in] size  Initial size of the object
 *  @see daGetSize()
 */
void msSetSize (MemStream ms, size_t size);

/** Get the size of a memory stream object.
 *  @param[in] ms  The memory stream object
 *  @return The size in bytes.
 *  @see msSetSize()
 */
size_t msGetSize (MemStream ms);

/** Release data from the memory stream object and return the pointer to them.
 *  The pointer should be freed when not needed.
 *  @param[in,out] ms  The memory stream object
 */
void *msRelease (MemStream ms);

/** Clears the data of a memory stream object.
 *  @param[in,out] ms  The memory stream object
 */
void msClear (MemStream ms);

/** @} */

#endif /* ! __XTND_MEMSTREAM_H__ */

