/**
 *  @file ptrlist.h
 *  @brief Pointer list
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2007 - 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_PTRLIST_H__
#define __XTND_PTRLIST_H__

/**
 *  @defgroup ptrlist Pointer list
 *  @ingroup ds
 *
 *  This module implements a pointer list.
 *
 *  @{
 */

/** The pointer list class struct. The purpose of the pointer list class is
 *  to manage a list of pointers.
 */
typedef struct PtrList *PtrList;


/** Defines a function for freeing items in a pointer list (ie.\ the item
 *  destructor). The function takes one argument, which is the pointer
 *  to be freed. The prototype is the same as of @em free from stdlib.h.
 */
typedef void (*PL_FREE) (const void *p);


/** Create a new pointer list object.
 *  @param[in] free  Function for freeing items (can be NULL)
 *  @return A pointer to the object.
 *  @see plFree()
 */
PtrList plCreate (PL_FREE free);

/** Free a pointer list.
 *  @param[in] p  The pointer list object
 *  @see plCreate()
 */
void plFree (PtrList p);

/** Clear a pointer list.
 *  @param[in,out] p  The pointer list object
 */
void plClear (PtrList p);

/** Get size (ie.\ count of items) of a pointer list.
 *  @param[in] p  The pointer list object
 *  @return The size.
 */
int plGetSize (PtrList p);

/** Get an item from the pointer list.
 *  @param[in] p  The pointer list object
 *  @param[in] index  Item index
 *  @return The item.
 *  @see plSet()
 */
void *plGet (PtrList p, unsigned int index);

/** Set an item from the list to @a item.
 *  @note If a function for freeing items was specified, it will be called
 *  for the item that is being overriden by this function.
 *  @param[in,out] p  The pointer list object
 *  @param[in] index  Item index
 *  @param[in] item  New item data
 *  @see plGet()
 */
void plSet (PtrList p, unsigned int index, void *item);

/** Delete an item from a pointer list.
 *  @param[in,out] p  The pointer list object
 *  @param[in] index  Item index
 */
void plDelete (PtrList p, unsigned int index);

/** Insert an item to a pointer list.
 *  @param[in,out] p  The pointer list object
 *  @param[in] index  Item index
 *  @param[in] item  The new item data
 */
void plInsert (PtrList p, unsigned int index, void *item);

/** Add an item to the end of a pointer list.
 *  @param[in] p  The pointer list object
 *  @param[in] item  The new item data
 *  @return The index of the added item.
 */
int plAppend (PtrList p, void *item);

/** @} */

#endif /* ! __XTND_PTRLIST_H__ */

