/**
 *  @file queue.c
 *  @brief A queue
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>
#include <string.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


struct Queue
{
	/** Pointers to blocks. */
	void *sblock, *eblock;
	/** Offsets in blocks. */
	unsigned int soffset, eoffset;
	/** Actual and maximal length of the queue. */
	unsigned int len, maxlen;
	/** Sizes of blocks and items. */
	size_t blocksize, itemsize;
};

#define QU_NEXTBLOCKPTR(block, size) \
	(*(void **) (((char **) (block)) + (size) - sizeof (void *)))

Queue quCreate (size_t itemsize, unsigned int maxlen)
{
	Queue qu;

	qu = xmalloc (sizeof (struct Queue));

	/* Use at least 2k blocks, so we don't waste memory with pointers.
	 * May also minimalize cache misses.
	 */
	qu->itemsize = itemsize;
	qu->blocksize = sizeof (void *)
		+ (itemsize < 2048)
		? ((4096 - sizeof (Queue)) / itemsize) * itemsize
		: itemsize;

	qu->len = 0;
	qu->maxlen = maxlen;

	qu->sblock = qu->eblock = NULL;
	qu->soffset = qu->eoffset = 0;

	return qu;
}

void quFree (Queue qu)
{
	void *block, *tmp;

	block = qu->sblock;
	while (block)
	{
		tmp = block;
		block = QU_NEXTBLOCKPTR (block, qu->blocksize);
		free (tmp);
	}
	free (qu);
}

unsigned int quGetLength (Queue qu)
{
	return qu->len;
}

int quRead (Queue qu, void *buffer)
{
	/* Don't let it underflow. */
	if (!qu->len)
		return -1;

	if (buffer)
		memcpy (buffer, ((char *) qu->sblock) + qu->soffset, qu->itemsize);

	if (qu->soffset + qu->itemsize >= (qu->blocksize - sizeof (void *)))
	{
		void *tmp;

		tmp = qu->sblock;
		qu->sblock = QU_NEXTBLOCKPTR (qu->sblock, qu->blocksize);
		qu->soffset = 0;
		free (tmp);

		if (tmp == qu->eblock)
		{
			qu->eblock = NULL;
			qu->eoffset = 0;
		}
	}
	else
		qu->soffset += qu->itemsize;
	qu->len--;

	return 0;
}

int quWrite (Queue qu, const void *buffer)
{
	/* Don't let it overflow. */
	if (qu->maxlen && qu->len + 1 > qu->maxlen)
		return -1;

	if (!qu->eblock)
		qu->sblock = qu->eblock = xmalloc (qu->blocksize);
	/* Append a new block, if there's no space. */
	else if (qu->eoffset >= (qu->blocksize - sizeof (void *)))
	{
		void *newblock;

		newblock = xmalloc (qu->blocksize);
		QU_NEXTBLOCKPTR (qu->eblock, qu->blocksize) = newblock;
		qu->eblock = newblock;
		qu->eoffset = 0;
	}

	memcpy (((char *) qu->eblock) + qu->eoffset, buffer, qu->itemsize);

	qu->eoffset += qu->itemsize;
	qu->len++;

	return 0;
}

