/**
 *  @file strlist.c
 *  @brief A string list
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2007 - 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>
#include <string.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


struct StrList
{
	/** An internal dynamic array. */
	DynArray da;
	/** Defines whether duplicates are allowed. */
	SlDuplicates duplicates;
	/** Points to a function which compares strings. */
	SlStrCmpCB strcmp;
	/** Info about extra data. */
	SlDataInfo di;
};


static void slFreeWrapper (const void *p, void *extra)
{
	StrList sl;

	sl = (StrList) extra;
	free (*((void **) p));

	if (sl->di.dataSize && sl->di.dataFree)
		sl->di.dataFree ((char *) p + sizeof (char *), sl->di.cbExtra);
}

StrList slCreate (SlDuplicates dup, SlStrCmpCB strcmp, const SlDataInfo *di)
{
	StrList sl;

	sl = xmalloc (sizeof (struct StrList));
	sl->da = daCreate (sizeof (char *) + (di ? di->dataSize : 0),
		slFreeWrapper, sl);
	sl->duplicates = dup;
	sl->strcmp = strcmp;

	if (di)
		sl->di = *di;
	else
	{
		sl->di.dataSize = 0;
		sl->di.dataFree = NULL;
		sl->di.dataClone = NULL;
		sl->di.cbExtra = NULL;
	}
	return sl;
}

int slFreeStringCB (const void *p, void *extra)
{
    free (*((char **) p));
    return 0;
}

int slCloneStringCB (const void *in, void *out, void *extra)
{
	if (*(char **) in)
		return (*(char **) out = xstrdup (*(char **) in)) != NULL;
	else
		return 0;
}

void slFree (StrList sl)
{
	daFree (sl->da);
}

void slClear (StrList sl)
{
	daClear (sl->da);
}

unsigned int slGetSize (StrList sl)
{
	return daGetSize (sl->da);
}

const char *slGet (StrList sl, unsigned int index)
{
	const char **p;

	p = (const char **) daGet (sl->da, index);
	return p ? *p : NULL;
}

void slSet (StrList sl, unsigned int index, const char *s)
{
	char *dup;

	if (!s || index >= daGetSize (sl->da))
		return;

	if (sl->duplicates == SL_DUP_IGNORE_NEW)
	{
		if (slFind (sl, s, 0) != -1)
			return;
	}
	else if (sl->duplicates == SL_DUP_DELETE_OLD)
	{
		int i = 0;

		while ((i = slFind (sl, s, i)) >= 0)
		{
			if ((unsigned) i == index)
			{
				i++;
				continue;
			}

			slDelete (sl, i);
			if (index > (unsigned) i)
				index--;
		}
	}

	dup = xstrdup (s);
	if (sl->di.dataSize)
	{
		void *buff;

		buff = xcalloc (1, sizeof (char *) + sl->di.dataSize);
		*((char **) buff) = dup;
		daSet (sl->da, index, buff);
	}
	else
		daSet (sl->da, index, &dup);
}

void *slGetData (StrList sl, unsigned int index)
{
	if (!sl->di.dataSize || index >= slGetSize (sl))
		return NULL;
	return (char *) daGet (sl->da, index) + sizeof (char *);
}

void slSetData (StrList sl, unsigned int index, void *data)
{
	if (!sl->di.dataSize || index >= slGetSize (sl))
		return;
	memcpy ((char *) daGet (sl->da, index) + sizeof (char *),
		data, sl->di.dataSize);
}

void slDelete (StrList sl, unsigned int index)
{
	daDelete (sl->da, index);
}

void slInsert (StrList sl, unsigned int index, const char *s)
{
	char *dup;

	if (!s || index >= slGetSize (sl))
		return;

	if (sl->duplicates == SL_DUP_IGNORE_NEW)
	{
		if (slFind (sl, s, 0) != -1)
			return;
	}
	else if (sl->duplicates == SL_DUP_DELETE_OLD)
	{
		int i = 0;

		while ((i = slFind (sl, s, i)) >= 0)
		{
			slDelete (sl, i);
			if (index > (unsigned) i)
				index--;
		}
	}

	dup = xstrdup (s);
	if (sl->di.dataSize)
	{
		void *buff;

		buff = xcalloc (1, sizeof (char *) + sl->di.dataSize);
		*((char **) buff) = dup;
		daInsert (sl->da, index, buff);
	}
	else
		daInsert (sl->da, index, &dup);
}

int slAppend (StrList sl, const char *s)
{
	char *dup;

	if (!s)
		return -1;

	if (sl->duplicates == SL_DUP_IGNORE_NEW)
	{
		if (slFind (sl, s, 0) != -1)
			return -1;
	}
	else if (sl->duplicates == SL_DUP_DELETE_OLD)
	{
		int i = 0;

		while ((i = slFind (sl, s, i)) != -1)
			slDelete (sl, i);
	}

	dup = xstrdup (s);
	if (sl->di.dataSize)
	{
		void *buff;

		buff = xcalloc (1, sizeof (char *) + sl->di.dataSize);
		*((char **) buff) = dup;
		return daAppend (sl->da, buff);
	}
	else
		return daAppend (sl->da, &dup);
}

void slAdd (StrList sl, StrList app)
{
	int i, j, len;

	len = slGetSize (app);
	for (i = 0; i < len; i++)
	{
		if ((j = slAppend (sl, slGet (app, i))) != -1)
		{
			if (sl->di.dataClone)
				sl->di.dataClone (slGetData (app, i), slGetData (sl, j),
					sl->di.cbExtra);
			else
				slSetData (sl, j, slGetData (app, i));
		}
	}
}

int slFind (StrList sl, const char *s, int start)
{
	int i, size;

	if (start < 0)
		start = 0;

	size = slGetSize (sl);
	for (i = start; i < size; i++)
	{
		if (!sl->strcmp (s, slGet (sl, i)))
			return i;
	}
	return -1;
}

void slParseAdd (StrList sl, const char *s, const char *delimiter)
{
#if 0
	/* Alternative code without strtok. */
	char *b, *p, *q;
	int l;

	l = strlen (delimiter);
	p = b = xstrdup (s);
	while ((q = strstr (p, delimiter)))
	{
		*q = '\0';
		slAppend (sl, p);
		p = q + l;
	}
	free (b);
#endif

	char *b, *p, *strtokstate;

	b = xstrdup (s);
	p = strtok_r (b, delimiter, &strtokstate);
	while (p)
	{
		slAppend (sl, p);
		p = strtok_r (NULL, delimiter, &strtokstate);
	}
	free (b);
}

char *slToString (StrList sl, const char *delimiter)
{
	MemStream ms;
	char *res;
	int i, l;

	ms = msCreate (0);
	l = daGetSize (sl->da);
	for (i = 0; i < l; i++)
	{
		const char *s;

		s = *((const char **) daGet (sl->da, i));
		msWrite (ms, s, strlen (s), 1);
		if (i != l - 1)
            msWrite (ms, delimiter, strlen (delimiter), 1);
	}
	msWriteInt8 (ms, '\0');

	res = msRelease (ms);
	msFree (ms);
	return res;
}

#if 0
static int slSortCompare (const void *first, const void *second)
{
	/* Problem here -- We must use the strcmp function from StrList
	 *  Solutions:
	 *    1. Use some asm code (may not work)
	 *    2. A global variable with a lock (too complicated)
	 *    3. Use own sort function -- see qsort.c
	 */
	return strcmp (*((const char **) first), *((const char **) second));
}

void slSort (StrList sl)
{
	qsort (daGetPtr (sl->da), slGetSize (sl), sizeof (char *) + sl->dataSize,
		slSortCompare);
}
#endif
