/**
 *  @file strlist.h
 *  @brief String list
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2007 - 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_STRLIST_H__
#define __XTND_STRLIST_H__

/**
 *  @defgroup strlist String list
 *  @ingroup ds
 *
 *  This module implements a string list.
 *
 *  @{
 */

/** The string list class struct. The purpose of the string list class is
 *  to manage a list of C strings. It has a special ability to mantain data
 *  associated to each string in the list.
 */
typedef struct StrList *StrList;


/** Defines a function for comparing items in a string list. The prototype
 *  is the same as of @em strcmp in string.h.
 */
typedef int (*SlStrCmpCB) (const char *s1, const char *s2);

/** Defines a callback function for freeing extra data in a string list. */
typedef int (*SlDataFreeCB) (const void *p, void *extra);

/** Defines a callback function for freeing extra data in a string list. */
typedef int (*SlDataCloneCB) (const void *in, void *out, void *extra);

/** Defines whether duplicates are allowed. */
typedef enum
{
	/** Duplicates are fully accepted. */
	SL_DUP_ACCEPT,
	/** Trying to add already existing values will return with an error. */
	SL_DUP_IGNORE_NEW,
	/** Older duplicated values are deleted. */
	SL_DUP_DELETE_OLD
}
SlDuplicates;

/** Information about data assigned to items in a @em StrList. */
typedef struct
{
	/** The size of extra data associated to each string. */
	size_t dataSize;
	/** Callback for freeing extra data. */
	SlDataFreeCB dataFree;
	/** Callback for cloning extra data. If this callback is NULL,
	 *  the data will just be copied.
	 */
	SlDataCloneCB dataClone;
	/** An undefined pointer passed as a parameter to callbacks. */
	void *cbExtra;
}
SlDataInfo;


/** When the extra data is a string pointer,
 *  you can use this callback for freeing it.
 */
int slFreeStringCB (const void *p, void *extra);

/** When the extra data is a string pointer,
 *  you can use this callback for cloning it.
 */
int slCloneStringCB (const void *in, void *out, void *extra);

/** Create a new string list object.
 *  @param[in] dup  Defines how to deal with duplicate values.
 *  @param[in] strcmp  Function for comparing items.
 *  @param[in] di  Information about data assigned to each string (may be NULL).
 *  @return A pointer to the object.
 *  @see slFree()
 */
StrList slCreate (SlDuplicates dup, SlStrCmpCB strcmp, const SlDataInfo *di);

/** Free a string list.
 *  @param[in] sl  The string list object
 *  @see slCreate()
 */
void slFree (StrList sl);

/** Clear a string list.
 *  @param[in,out] sl  The string list object
 */
void slClear (StrList sl);

/** Get size (ie.\ count of items) of a string list.
 *  @param[in] sl  The string list object
 *  @return The size.
 */
unsigned int slGetSize (StrList sl);

/** Get an item from the string list.
 *  @param[in] sl  The string list object
 *  @param[in] index  Item index
 *  @return The item.
 *  @see slSet()
 */
const char *slGet (StrList sl, unsigned int index);

/** Set an item from the list to @a s.
 *  @note If a function for freeing items was specified, it will be called
 *  for the item that is being overriden by this function.
 *  @param[in,out] sl  The string list object
 *  @param[in] index  Item index
 *  @param[in] s  New item name
 *  @see slGet()
 */
void slSet (StrList sl, unsigned int index, const char *s);

/** Returns a pointer to the data associated to an item.
 *  @param[in] sl  The string list object
 *  @param[in] index  The index of the item
 *  @return The pointer to the associated data.
 */
void *slGetData (StrList sl, unsigned int index);

/** Sets the data associated to an item.
 *  @param[in] sl  The string list object
 *  @param[in] index  The index of the item
 *  @param[in] data  A pointer to the new data
 */
void slSetData (StrList sl, unsigned int index, void *data);

/** Delete an item from a string list.
 *  @param[in,out] sl  The string list object
 *  @param[in] index  Item index
 */
void slDelete (StrList sl, unsigned int index);

/** Insert an item to a string list.
 *  @param[in,out] sl  The string list object
 *  @param[in] index  Item index
 *  @param[in] s  The new item data
 */
void slInsert (StrList sl, unsigned int index, const char *s);

/** Append an item to the end of a string list.
 *  @param[in] sl  The string list object
 *  @param[in] s  The new item
 *  @return The index of the newly added item.
 */
int slAppend (StrList sl, const char *s);

/** Add a list to another one.  Copies the items including the extra data.
 *  @param[in,out] sl  A string list object.
 *  @param[in] app  The string list to be appended to @a sl.
 */
void slAdd (StrList sl, StrList app);

/** Find an item in the list and return its index.
 *  @param sl  The string list object
 *  @param s  The item being searched for
 *  @param start  Initial index from which to search
 *  @return -1 if not successful, the item index otherwise.
 */
int slFind (StrList sl, const char *s, int start);

/** Parse a string and add the items to a list.
 *  @param sl  The string list object
 *  @param s  The string
 *  @param delimiter  Every char in this string is an item separator.
 */
void slParseAdd (StrList sl, const char *s, const char *delimiter);

/** Return the list in one string.
 *  @param sl  The string list object
 *  @param delimiter  Item separator
 *  @return A string containing the whole list. The pointer should be freed
 *  when not needed.
 */
char *slToString (StrList sl, const char *delimiter);

/** @} */

#endif /* ! __XTND_STRLIST_H__ */

