/**
 *  @file exception.c
 *  @brief Exceptions hacked into C
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2011. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>
#include <assert.h>

#ifdef _WIN32
	#include <windows.h>
	#define TLSINDEX DWORD
	#define TLSGET TlsGetValue
	#define TLSSET TlsSetValue
#else
	#include <pthread.h>
	#define TLSINDEX pthread_key_t
	#define TLSGET pthread_getspecific
	#define TLSSET pthread_setspecific
#endif /* ! _WIN32 */

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


/** Thread-local storage index for the last exception frame. */
static TLSINDEX g_exceptionIndex;
/** Whether the module has been initialized. */
static int g_exceptionInitialized;


void
exceptionInit (void)
{
	if (g_exceptionInitialized)
		return;

#ifdef _WIN32
	g_exceptionIndex = TlsAlloc ();
	if (g_exceptionIndex == TLS_OUT_OF_INDEXES)
		abort ();
#else /* ! _WIN32 */
	if (pthread_key_create (&g_exceptionIndex, free))
		abort ();
#endif /* ! _WIN32 */

	g_exceptionInitialized = 1;
}

void
exceptionRaise (int code)
{
	ExceptionCtx e;

	assert (code != 0);

	if (!g_exceptionInitialized)
		exceptionInit ();

	/* Nowhere to jump. */
	if (!(e = TLSGET (g_exceptionIndex)))
		abort ();

	e->code = code;
	longjmp (e->jmp, code);
}

void
exceptionPushPoint (ExceptionCtx e)
{
	if (!g_exceptionInitialized)
		exceptionInit ();

	e->doneFinally = 0;
	e->rethrow = 0;

	/* Push the frame to the stack. */
	e->next = TLSGET (g_exceptionIndex);
	TLSSET (g_exceptionIndex, e);
}

void
exceptionPopPoint (ExceptionCtx e)
{
	if (!g_exceptionInitialized)
		exceptionInit ();

	assert (e == TLSGET (g_exceptionIndex));

	if (e->doneFinally)
	{
		/* Remove the frame from the stack. */
		TLSSET (g_exceptionIndex, e->next);

		if (e->rethrow)
			exceptionRaise (e->code);
	}
	else
	{
		e->doneFinally = 1;
		THROW (EXCEPTION_FINALLY);
	}
}

