/**
 *  @file exception.h
 *  @brief Exceptions hacked into C
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2011. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_EXCEPTION_H__
#define __XTND_EXCEPTION_H__

/**
 *  @defgroup exception Exceptions
 *  @ingroup util
 *
 *  This module implements simple exceptions using the setjmp() mechanism.
 *
 *  Each thread has its own exception stack just as expected. You must call
 *  exceptionInit() before using any of the module functions in a multithreaded
 *  environment, though, to prevent race conditions, since otherwise the module
 *  initializes itself at the first function call.
 *
 *  No part of libxtnd depends on exceptions yet, so you don't have to worry
 *  about their intialization if you don't use them. Nevertheless,
 *  the situation may change in future versions of the API.
 *
 *  Within the TRY block, the local exception context is available for reading
 *  as a local variable named @em e, though the only useful member of this
 *  structure is @em e.code.
 *
 *  Restrictions of this implementation are these:
 *
 *   - Exception codes must be positive integers, that is, greater than zero.
 *
 *   - Each TRY block must contain a FINALLY clause, which would otherwise
 *     be caught by EXCEPT_DEFAULT. The later is also required to be present
 *     for exceptions to propagate up. They don't do that automatically,
 *     you must RETHROW() them.
 *
 *   - Local register variables modified in the first part of a TRY block have
 *     to be declared as @em volatile, otherwise any changes to them would
 *     be forgotten (a feature of longjmp()).
 *
 *   - Don't use @em return in TRY blocks. It corrupts the exception stack.
 *     If needed to get past TRY_END, use the @em break; statement. You can't
 *     skip the execution of FINALLY.
 *
 *  Inspired by http://www.on-time.com/ddj0011.htm
 *
 *  @{
 */

/** Exception frame context. */
typedef struct ExceptionCtx *ExceptionCtx;

/** Exception frame context structure. */
struct ExceptionCtx
{
	/** This forms a linked list to create a stack. */
	ExceptionCtx next;

	/** Save point. */
	jmp_buf jmp;
	/** Whether the FINALLY block has been processed yet. */
	int doneFinally;
	/** The current state/exception code of the TRY block. */
	int code;
	/** Whether to rethrow the last exception after FINALLY. */
	int rethrow;
};

/** A special exception code for the FINALLY block. Do not use. */
#define EXCEPTION_FINALLY -1

/** Begin a TRY block. */
#define TRY \
	do { \
		struct ExceptionCtx e; \
		exceptionPushPoint (&e); \
		switch (setjmp (e.jmp)) { \
			case 0:
/** Catch the specified exception code. */
#define EXCEPT \
			case
/** Catch any exception we didn't match explicitly. */
#define EXCEPT_DEFAULT \
			default
/** This block is always executed in the end. */
#define FINALLY \
			case EXCEPTION_FINALLY
/** Finish a TRY block. */
#define END_TRY \
		} \
		exceptionPopPoint (&e); \
	} while (0);

/** Throw an exception. */
#define THROW(code) exceptionRaise (code)
/** Run the FINALLY block and propagate the exception up. */
#define RETHROW() \
	do { \
		e.rethrow = 1; \
		exceptionRaise (EXCEPTION_FINALLY); \
	} while (0)


/** Initialize the module. This function is not thread-safe. */
void exceptionInit (void);

/** Raise an exception. Do not use directly. */
void exceptionRaise (int code);

/** Push an exception point to the internal stack. Do not use directly. */
void exceptionPushPoint (ExceptionCtx e);

/** Pop an exception point from the internal stack, but first check for
 *  equality with @a e. Do not use directly.
 */
void exceptionPopPoint (ExceptionCtx e);

/** @} */

#endif /* ! __XTND_EXCEPTION_H__ */
