/**
 *  @file macro.c
 *  @brief Simple stack-based macro processor
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2007 - 2009, 2011. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>

#ifdef XTND_USE_OWN_MACROGC
	#include <xtnd/macrogc.h>
#else
	/* Boehm's libgc. It's strongly recommended to use the same allocator
	 * for strings in the application, otherwise extensive leaks will occur.
	 */
	#include <gc/gc.h>
#endif


/** Maximum count of items on the stack. */
#define MACRO_STACK_SIZE 4096


/* Local functions. */
static void macroSubListFree (const void *p, void *extra);
static int macroExecProcessItem (ExecContext ctx, struct Item *item);


const char *macroErrorString (int error)
{
	struct
	{
		int id;
		char *describ;
	}
	*err, errs[] =
	{
		{MACRO_ERR_SUCCESS,            N_("No error")},
		{MACRO_ERR_NULL_POINTER,       N_("Null pointer")},

		{MACRO_ERR_SYNTAX,             N_("Syntax error")},
		{MACRO_ERR_BAD_CHAR,           N_("Bad character")},
		{MACRO_ERR_TOO_NESTED,         N_("Too nested lists")},
		{MACRO_ERR_NOT_IN_LIST,        N_("Not in a list")},
		{MACRO_ERR_STRING_NO_END,      N_("String has no end")},
		{MACRO_ERR_LIST_NOT_CLOSED,    N_("List not closed")},
		{MACRO_ERR_INTEGER_OVERFLOW,   N_("Integer overflow")},

		{MACRO_ERR_RUNTIME,            N_("Runtime error")},
		{MACRO_ERR_INCOMPATIBLE_ITEM,  N_("Incompatible item")},
		{MACRO_ERR_NO_SUB,             N_("Subprogram does not exist")},
		{MACRO_ERR_STACK_UNDERFLOW,    N_("Stack underflow")},
		{-1, NULL}
	};

	for (err = errs; ~err->id; err++)
	{
		if (error == err->id)
			return _(err->describ);
	}
	return NULL;
}

static void macroSubListFree (const void *p, void *extra)
{
	if (p)
		free (((MacroSubInfo) p)->name);
}

MacroSubCB macroGetSub (ExecContext ctx, const char *name)
{
	int i, max;
	MacroSubInfo p;

	if (!ctx->subList)
		return NULL;

	max = daGetSize (ctx->subList);
	for (i = 0; i < max; i++)
	{
		p = (MacroSubInfo) daGet (ctx->subList, i);
		if (!strcmp (p->name, name))
			return p->sub;
	}

	return NULL;
}

void macroSetSub (ExecContext ctx, const char *name, MacroSubCB sub)
{
	int i, max;
	MacroSubInfo p;

	if (!name || !*name)
		return;

	if (!ctx->subList)
		ctx->subList = daCreate (sizeof (struct MacroSubInfo),
			macroSubListFree, NULL);

	/* Try to find the subprogram in the array. */
	max = daGetSize (ctx->subList);
	for (i = 0; i < max; i++)
	{
		p = (MacroSubInfo) daGet (ctx->subList, i);
		if (strcmp (p->name, name))
			continue;

		if (sub)
			p->sub = sub;
		else
			daDelete (ctx->subList, i);
		return;
	}

	if (sub)
	{
		struct MacroSubInfo info;

		info.name = xstrdup (name);
		info.sub = sub;
		daAppend (ctx->subList, &info);
	}
}


struct Item *macroAllocItem (ExecContext ctx)
{
#ifdef XTND_USE_OWN_MACROGC
	return macroGCAlloc (ctx->gd);
#else
	return GC_MALLOC (sizeof (struct Item));
#endif
}

struct Item *macroDupList (ExecContext ctx, struct Item *list)
{
	struct Item *res = NULL, *cur = NULL, *next;

	while (list)
	{
		if (!(next = macroDupItem (ctx, list)))
			return NULL;

		if (cur)
			cur->next = next;
		cur = next;
		if (!res)
			res = cur;

		list = list->next;
	}

	return res;
}

struct Item *macroDupItem (ExecContext ctx, struct Item *item)
{
	struct Item *res;

	if (!item)
		return NULL;

	/* Create a new item. */
	if (!(res = macroAllocItem (ctx)))
		return NULL;

	res->type = item->type;
	res->next = NULL;

	/* Copy the data. */
	switch (item->type)
	{
	case MACRO_INT:
		res->data.num = item->data.num;
		break;
	case MACRO_STRING:
	case MACRO_SUB:
		res->data.str = xstrdup (item->data.str);
		break;
	case MACRO_LIST:
		res->data.list = macroDupList (ctx, item->data.list);
		break;
	default:
		return NULL;
	}
	return res;
}


/* TODO: GLib-like error passing. */
int macroExec (Macro macro, const struct MacroSubInfo subs[], void *extra,
	char *errorInfo, size_t errorInfoLen)
{
	int error = MACRO_ERR_SUCCESS;
	ExecContext ctx;
	const struct MacroSubInfo *i;

#ifdef XTND_USE_OWN_MACROGC
	ctx = xmalloc (sizeof (struct ExecContext));
#else
	/* libgc doesn't search for pointers in malloc()ed memory.
	 * And, you know, we have the whole stack inside the ExecContext.
	 */
	ctx = GC_MALLOC_UNCOLLECTABLE (sizeof (struct ExecContext));
#endif
	ctx->subList = NULL;
	ctx->errorInfo = errorInfo;
	ctx->errorInfoLen = errorInfoLen;
	ctx->stackTop = NULL;
	ctx->stackSize = 0;
	ctx->extra = extra;

	/* Define the subs in @a subs. */
	if (subs)
		for (i = subs; i->name && i->sub; i++)
			macroSetSub (ctx, xstrdup (i->name), i->sub);

	/* The default is an empty string. */
	if (ctx->errorInfo && ctx->errorInfoLen)
		*errorInfo = '\0';

#ifdef XTND_USE_OWN_MACROGC
	if (!(ctx->gd = macroGCInit (free)))
		return MACRO_ERR_MALLOC_FAIL;
#endif

	while (macro)
	{
		error = macroExecProcessItem (ctx, macro);
		if (error)
			break;

#ifdef XTND_USE_OWN_MACROGC
		macroGCRun (ctx);
#endif

		macro = macro->next;
	}

#ifdef XTND_USE_OWN_MACROGC
	macroGCDestroy (ctx->gd);
	free (ctx);
#else
	GC_FREE (ctx);
#endif

	return error;
}

int macroExecInner (ExecContext ctx, struct Item *macro)
{
	int error;

	while (macro)
	{
		if ((error = macroExecProcessItem (ctx, macro)))
			return error;
		macro = macro->next;
	}
	return MACRO_ERR_SUCCESS;
}

static int macroExecProcessItem (ExecContext ctx, struct Item *item)
{
	int error = MACRO_ERR_SUCCESS;

	/* The item is a call to a subprogram. */
	if (item->type == MACRO_SUB)
	{
		MacroSubCB sub;

		if ((sub = macroGetSub (ctx, item->data.str)))
			/* Execute the subprogram. */
			/* TODO: Some useful info to ctx->errorInfo */
			error = sub (ctx);
		else
		{
			if (ctx->errorInfo)
				snprintf (ctx->errorInfo, ctx->errorInfoLen,
					_("Subprogram not found: %s"), item->data.str);

			error = MACRO_ERR_NO_SUB;
		}
	}
	/* Push the item to the stack. */
	else
	{
		struct Item *item2;

		item2 = macroDupItem (ctx, item);
		error = macroPush (ctx, item2);
	}
	return error;
}

struct Item *macroPop (ExecContext ctx)
{
	if (ctx->stackTop)
	{
		struct Item *res = ctx->stackTop;
		ctx->stackTop = res->next;
		ctx->stackSize--;
		res->next = NULL;
		return res;
	}
	else
		return NULL;
}

struct Item *macroAt (ExecContext ctx, unsigned int index)
{
	struct Item *item;

	for (item = ctx->stackTop; index; index--)
	{
		if (!(item = item->next))
			return NULL;
	}
	return item;
}

int macroPush (ExecContext ctx, struct Item *item)
{
	if (!item)
		return MACRO_ERR_NULL_POINTER;

	if (ctx->stackSize >= MACRO_STACK_SIZE)
		return MACRO_ERR_STACK_OVERFLOW;

	item->next = ctx->stackTop;
	ctx->stackTop = item;
	ctx->stackSize++;

	return 0;
}

