/**
 *  @file macro.h
 *  @brief Simple stack-based macro processor
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2007, 2008, 2009, 2011.
 *        All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_MACRO_H__
#define __XTND_MACRO_H__

/**
 *  @defgroup macro Simple stack-based macro processor
 *  @ingroup util
 *
 *  This module implements a simple stack-based scripting language.
 *
 *  @{
 */

/* FIXME: For fuck's sake, it leakes ALL string pointers during runtime!
 *
 *        For our own GC, ExecContext should contain a free() callback,
 *            so we know how to free strings allocated in subprocedures.
 *        For Boehm's libgc, this file should define a malloc() function,
 *            so subprocedures know how to allocate strings for us to free.
 *
 *        Solution? Add a generic macroAlloc(), macroStrdup(), ...
 *         - With own GC, it maps to malloc() and the GC calls free().
 *         - With libgc, it maps to GC_MALLOC() and frees automatically.
 *
 *        Also beware, the same macroDupItem() is used during both parsing
 *        and the runtime. It probably needs to be split in two.
 */

/** No error. */
#define MACRO_ERR_SUCCESS 0x0000
/** Function received a NULL pointer. */
#define MACRO_ERR_NULL_POINTER 0x0001
/** Memory allocation failed. */
#define MACRO_ERR_MALLOC_FAIL 0x0002

/** General syntax error. */
#define MACRO_ERR_SYNTAX 0x0100
/** Bad character. */
#define MACRO_ERR_BAD_CHAR 0x0101
/** List is too nested. */
#define MACRO_ERR_TOO_NESTED 0x0102
/** Tried to end a list when there was none open. */
#define MACRO_ERR_NOT_IN_LIST 0x0103
/** String has no end. */
#define MACRO_ERR_STRING_NO_END 0x0104
/** List has not been not closed. */
#define MACRO_ERR_LIST_NOT_CLOSED 0x0105
/** Integer too large. */
#define MACRO_ERR_INTEGER_OVERFLOW 0x0106

/** General runtime error. */
#define MACRO_ERR_RUNTIME 0x0200
/** Incompatible item on the stack. */
#define MACRO_ERR_INCOMPATIBLE_ITEM 0x0201
/** Stack underflow (tried to pop an empty stack). */
#define MACRO_ERR_STACK_UNDERFLOW 0x0202
/** Stack overflow (too many items on the stack). */
#define MACRO_ERR_STACK_OVERFLOW 0x0203
/** Unknown subprogram name. */
#define MACRO_ERR_NO_SUB 0x0204


/** Macro execution context. */
typedef struct ExecContext *ExecContext;

/** Subprocedure definition. */
typedef struct MacroSubInfo *MacroSubInfo;

/** Defines a preprocessed macro. */
typedef struct Item *Macro;


/** Macro data item. Never make attempt to call free() upon items returned
 *  from an execution context, since these are collected automatically.
 */
struct Item
{
	/** A pointer to the next item in the list. */
	struct Item *next;
	/** Specifies the type of an item. */
	enum ItemType {MACRO_INT, MACRO_STRING, MACRO_SUB, MACRO_LIST} type;
	/** Item's data. */
	union
	{
		char *str;
		int num;
		struct Item *list;
	} data;
};

/** Defines the prototype of functions processing macro subprograms. */
typedef int (*MacroSubCB) (ExecContext ctx);

/** Subprocedure definition structure. */
struct MacroSubInfo
{
	char *name;
	MacroSubCB sub;
};

/** Macro execution context structure. */
struct ExecContext
{
	/** A pointer to the top of stack. It is recommended to access the stack
	 *  indirectly with macroPop(), macroPush() and macroAt().
	 */
	struct Item *stackTop;
	/** Size of the stack. Don't forget to increment/decrement this value
	 *  if you're adding/removing items to/from the stack directly.
	 */
	unsigned int stackSize;
	/** A list of MacroSubInfo's defined in the context. */
	DynArray subList;
	/** Error information buffer. */
	char *errorInfo;
	/** Length of the errorInfo buffer. */
	size_t errorInfoLen;
	/** Unspecified. */
	void *extra;
	/** Used internally for garbage collection. */
	struct GCData *gd;
};


/** Translate an error code to string.
 *  @param[in] error  The error code
 *  @return The error code string.
 */
const char *macroErrorString (int error);

/** Parse a macro and return a preprocessed macro handle.
 *  @param[in] str  The macro to parse.
 *  @param[out] macro  The output handle.
 */
int macroParse (const char *str, Macro *macro);

/** Free a parsed macro.
 *  @param[in,out] macro  The macro to free.
 */
void macroFree (Macro *macro);

/** Execute the macro.
 *  @param[in] macro  The macro.
 *  @param[in] subs  A list of subprocedures to define in the environment
 *                   or NULL. The last item is {NULL, NULL}.
 *  @param[in] extra  Pointer to custom data.
 *  @param[out] errorInfo  Extended error info buffer.
 *  @param[in] errorInfoLen  Buffer length.
 *  @return  An error or MACRO_ERR_SUCCESS
 */
int macroExec (Macro macro, const struct MacroSubInfo subs[], void *extra,
	char *errorInfo, size_t errorInfoLen);

/** Execute the macro in an existing execution context.
 *  @param[in] ctx  The execution context.
 *  @param[in] macro  The macro.
 */
int macroExecInner (ExecContext ctx, struct Item *macro);

/** Pop an item from the stack.
 *  @param[in] ctx  The execution context.
 *  @return A pointer to the item.
 *  @see macroPush(), macroAt()
 */
struct Item *macroPop (ExecContext ctx);

/** Get the @em index-th item on the stack.
 *  @param[in] ctx  The execution context.
 *  @param[in] index  Index of the item.
 *  @return A pointer to the item.
 *  @see macroPop(), macroPush()
 */
struct Item *macroAt (ExecContext ctx, unsigned int index);

/** Push an item to the stack.
 *  @param[in] ctx  The execution context.
 *  @param[in] item  The item to push.
 *  @see macroPop(), macroAt()
 */
int macroPush (ExecContext ctx, struct Item *item);


/** Get the pointer for a subprogram.
 *  @param[in] ctx  The execution context.
 *  @param[in] name  The name of the subprogram.
 *  @return A pointer to the subprogram.
 *  @see macroSetSub()
 */
MacroSubCB macroGetSub (ExecContext ctx, const char *name);

/** Set or remove the pointer for a subprogram.
 *  @param[in] ctx  The execution context.
 *  @param[in] name  The name of the subprogram.
 *  @param[in] sub  The subprogram handler, NULL to remove.
 *  @see macroGetSub()
 */
void macroSetSub (ExecContext ctx, const char *name, MacroSubCB sub);


/** Allocate an item.
 *  @param[in] ctx  The execution context.
 *  @return  A pointer to the allocated item.
 */
struct Item *macroAllocItem (ExecContext ctx);

/** Duplicate an item.
 *  @param[in] ctx  The execution context.
 *  @param[in] item  The item.
 *  @see macroDupList()
 */
struct Item *macroDupItem (ExecContext ctx, struct Item *item);

/** Duplicate a list.
 *  @param[in] ctx  The execution context.
 *  @param[in] list  The list.
 *  @see macroDupItem()
 */
struct Item *macroDupList (ExecContext ctx, struct Item *list);

/** @} */

#endif /* ! __XTND_MACRO_H__ */

