/**
 *  @file macrogc.c
 *  @brief A crappy garbage collector for macros.
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>

#ifdef XTND_MACROGC_DETECTLEAKS
	#define GC_DEBUG
	#include <gc/gc.h>

	#define malloc(n) GC_MALLOC(n)
	#define calloc(m,n) GC_MALLOC((m)*(n))
	#define free(p) GC_FREE(p)
	#define realloc(p,n) GC_REALLOC((p),(n))
#else /* ! XTND_MACROGC_DETECTLEAKS */
	#define malloc(n) xmalloc (n)
	#define calloc(m,n) xcalloc (m,n)
	#define realloc(p,n) xrealloc ((p),(n))
#endif /* ! XTND_MACROGC_DETECTLEAKS */


/** Initial heap size of the primary heap (in number of items). */
const unsigned gcInitHeapSize = 2048;
/** Minimal heap size of the primary heap (in number of items). */
const unsigned gcMinHeapSize = 1024;
/** The size of secondary heaps. */
const unsigned gcSecHeapSize = 1024;


/* Shortcuts. */
typedef struct HeapItem  *HeapItem;
typedef struct Heap      *Heap;
typedef struct GCData    *GCData;

/** An item on the heap. */
struct HeapItem
{
	/** The actual data of the item. */
	struct Item item;
	/** The state of the item. Used while collecting. */
	enum {GC_WHITE, GC_BLACK} type;
};

/** A heap. */
struct Heap
{
	/** Pointer to an array of heap items. */
	HeapItem array;
	/** Size of the heap. */
	unsigned int size;
	/** Next free item index on the heap. */
	unsigned int nextFree;
};

/** Data for the garbage collector. */
struct GCData
{
	/** Index of the heap to be used for allocation.
	 *  Zero is for the primary heap.
	 */
	unsigned int hIndex;
	/** A function for freeing string data in items. */
	void (*stringFree) (void *ptr);

	/** The primary heap. */
	struct Heap priHeap;
	/** A list of secondary heaps. The zero'th is reserved. */
	DynArray secHeaps;
};


/** Marks items on the heap as black, i.e. in use. */
static void macroGCMarkBlack (struct Item *list, unsigned int *count);
/** Move a heap item to another location. */
static void macroGCMoveItem (ExecContext ctx, unsigned int startSecHeap,
	HeapItem from, HeapItem to);


#define GC_DESTROYITEM(i) \
	if ((i).type == MACRO_STRING || (i).type == MACRO_SUB) \
		gd->stringFree ((i).data.str);

#define GC_MOVEITEMPTR(i, l) \
	(i) = (struct Item *) ((char *) (i) + (l));

GCData macroGCInit (void (*stringFree) (void *))
{
	GCData gd;
	struct Heap t = {NULL, 0, 0};

#ifdef XTND_MACROGC_DETECTLEAKS
	GC_find_leak = 1;
#endif

	if (!(gd = malloc (sizeof (struct GCData))))
		return NULL;

	gd->hIndex = 0;
	gd->stringFree = stringFree;

	if (!(gd->priHeap.array = calloc (gcInitHeapSize, sizeof (struct HeapItem))))
	{
		free (gd);
		return NULL;
	}

	gd->priHeap.size = gcInitHeapSize;
	gd->priHeap.nextFree = 0;

	gd->secHeaps = daCreate (sizeof (struct Heap), NULL, NULL);

	/* Reserve an item. */
	daAppend (gd->secHeaps, &t);

	return gd;
}

void macroGCDestroy (GCData gd)
{
	int secHeapIndex = 1;
	Heap h;
	unsigned int i;

#ifndef NDEBUG
	puts ("GC: Destroying");
#endif

	h = &gd->priHeap;
	do
	{
		for (i = 0; i < h->nextFree; i++)
		{
			HeapItem hi;

			hi = h->array + i;
			GC_DESTROYITEM (hi->item);
		}
		free (h->array);
		h = daGet (gd->secHeaps, secHeapIndex++);
	}
	while (h);

	daFree (gd->secHeaps);
	free (gd);

#ifdef XTND_MACROGC_DETECTLEAKS
	GC_gcollect ();
#endif
}

#if 0
static void dumpPriHeap (GCData gd)
{
	unsigned i;

	puts ("-- Heap debug ------------------");
	for (i = 0; i < gd->priHeap.nextFree; i++)
	{
		HeapItem hi;

		hi = &gd->priHeap.array[i];
		printf (" HeapItem %d\n  Type: %s\n",
			i, hi->type ? "GC_BLACK" : "GC_WHITE");
		printf ("  Item address: %p\n", (void *) &hi->item);
		switch (hi->item.type)
		{
		case MACRO_INT:
			printf ("  INT: %d\n", hi->item.data.num);
			break;
		case MACRO_SUB:
			printf ("  SUB: %s (%p)\n", hi->item.data.str,
				(void *) hi->item.data.str);
			break;
		case MACRO_LIST:
			printf ("  LIST: %p\n", (void *) hi->item.data.list);
			break;
		case MACRO_STRING:
			break;
		}
		printf ("  Next: %p\n", (void *) hi->item.next);
	}
}
#endif

void macroGCRun (ExecContext ctx)
{
	unsigned int i, secHeapIter, nItems;
	Heap h;
	GCData gd;

	gd = ctx->gd;

	/* 1. Mark all black items as white. */
#ifndef NDEBUG
	puts ("GC: Marking all white");
#endif
	secHeapIter = 1;
	h = &gd->priHeap;
	do
	{
		for (i = 0; i < h->nextFree; i++)
			h->array[i].type = GC_WHITE;
		h = daGet (gd->secHeaps, secHeapIter++);
	}
	while (h);

	/* 2. Walk through the stack and mark items as black. */
#ifndef NDEBUG
	puts ("GC: Marking the stack black");
#endif
	nItems = 0;
	macroGCMarkBlack (ctx->stackTop, &nItems);

	/* 3. Compact and prepare the primary heap. */
#ifndef NDEBUG
	puts ("GC: Compacting the heap");
#endif
	/* TODO: Some condition on which to actually run the collector. */
	if (nItems > gd->priHeap.size)
	{
		/* a. Allocate a new primary heap. */
		HeapItem newHeapArray;
		size_t newHeapSize;

		if ((newHeapSize = nItems << 1) < gcMinHeapSize)
			newHeapSize = gcMinHeapSize;
		if (!(newHeapArray = malloc (newHeapSize * sizeof (struct HeapItem))))
			return;

		/* Use the reserved heap. */
		daSet (gd->secHeaps, 0, &gd->priHeap);
		h = daGet (gd->secHeaps, 0);
		secHeapIter = 0;

		gd->priHeap.size = newHeapSize;
		gd->priHeap.nextFree = 0;
		gd->priHeap.array = newHeapArray;
	}
	else
	{
		/* b. Just compact the existing primary heap. */
		int heapMove = 0;

		for (i = 0; i < gd->priHeap.nextFree; i++)
		{
			HeapItem hi;

			hi = &gd->priHeap.array[i];
			if (hi->type == GC_WHITE)
			{
				GC_DESTROYITEM (hi->item);
				heapMove++;
			}
			else if (heapMove)
				macroGCMoveItem (ctx, 1, hi, hi - heapMove);
		}
		gd->priHeap.nextFree -= heapMove;
		secHeapIter = 1;
	}

	/* 4. Move all other heaps into it. */
	for (; (h = daGet (gd->secHeaps, secHeapIter)); secHeapIter++)
	{
		while (h->nextFree--)
		{
			HeapItem hi;

			hi = &h->array[h->nextFree];
			if (hi->type == GC_WHITE)
			{
				GC_DESTROYITEM (hi->item);
			}
			else
			{
				macroGCMoveItem (ctx, secHeapIter, hi,
					&gd->priHeap.array[gd->priHeap.nextFree++]);
			}
		}
		free (h->array);
	}
	/* Delete all the secondary heaps (except the reserved one). */
	for (i = daGetSize (gd->secHeaps) - 1; i; i--)
		daDelete (gd->secHeaps, i);
	gd->hIndex = 0;

#ifndef NDEBUG
	puts ("GC: Garbage collected");
#endif
}

/* TODO: Optimize this as it takes the most of CPU time. */
/* a. Use handles instead of pointers and translate them into pointers
 *    using a translation table.
 *      - This will change the API [struct Item *macroXlate(Handle handle)].
 *      - More memory will be needed.
 *      + Moving an item will take O(1) time instead of O(n).
 *      + Might be possible to collect garbage in parallel by watching
 *        items for which pointers have been requested.
 *
 * Reference: http://www.youtube.com/watch?v=1nYDflSL0Mg
 *  * He uses void **handle instead of typedef void **Handle;
 *    which might be somewhat better
 *  * Pointers need to be locked before using
 *
 * # It might be possible to choose the method at compile time (however
 *   the consequent changes might be very extensive).
 */
static void macroGCMoveItem (ExecContext ctx, unsigned int startSecHeap,
	HeapItem from, HeapItem to)
{
	struct Item *ptr, *newPtr;
	unsigned int k = 0;
	Heap h;
	HeapItem hi;

	ptr = &from->item;
	newPtr = &to->item;

	from->type = GC_WHITE;

	h = &ctx->gd->priHeap;
	do
	{
		for (k = 0; k < h->nextFree; k++)
		{
			hi = &h->array[k];
			if (hi->type != GC_BLACK)
				continue;

			if (hi->item.data.list == ptr && hi->item.type == MACRO_LIST)
				hi->item.data.list = newPtr;
			if (hi->item.next == ptr)
				hi->item.next = newPtr;
		}
		h = daGet (ctx->gd->secHeaps, startSecHeap++);
	}
	while (h);

	if (ctx->stackTop == ptr)
		ctx->stackTop = newPtr;

	to->type = GC_BLACK;
	to->item = from->item;
}

static void macroGCMarkBlack (struct Item *list, unsigned int *count)
{
	while (list)
	{
		/* It should always be GC_WHITE. */
		assert (((HeapItem) list)->type == GC_WHITE);

		((HeapItem) list)->type = GC_BLACK;
		(*count)++;

		if (list->type == MACRO_LIST && list->data.list)
			macroGCMarkBlack (list->data.list, count);
		list = list->next;
	}
}

struct Item *macroGCAlloc (GCData gd)
{
	struct Item *result;
	Heap h;

	h = gd->hIndex ? daGet (gd->secHeaps, gd->hIndex) : &gd->priHeap;

	/* Make a new heap if there's no space. */
	if (h->size <= h->nextFree)
	{
		struct Heap newHeap;

		newHeap.size = gcSecHeapSize;
		newHeap.nextFree = 0;

		if ((newHeap.array = malloc (newHeap.size
			* sizeof (struct HeapItem))) == NULL)
			return NULL;

		if ((gd->hIndex = daAppend (gd->secHeaps, &newHeap)) == -1)
			return NULL;

		/* This heap is reserved. */
		assert (gd->hIndex != 0);

		h = daGet (gd->secHeaps, gd->hIndex);
	}

	result = &h->array[h->nextFree].item;
	result->next = NULL;
	result->type = MACRO_INT;
	h->array[h->nextFree].type = GC_BLACK;
	h->nextFree++;

	return result;
}

