/**
 *  @file macrogc.h
 *  @brief Crappy garbage collector for macros
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_MACROGC_H__
#define __XTND_MACROGC_H__

/**
 *  @defgroup macrogc Crappy garbage collector for macros
 *  @ingroup util
 *
 *  This module serves as a trivial garbage collector for macros.
 *
 *  This collector should work nicely but has two major disadvantages:
 *   -# You must "stop the world" before a collection may occur.
 *   -# It doesn't look for pointers on the program stack.
 *  Practically this means that it cannot release items while any macro
 *  item is being processed. Choosing a better collector is recommended.
 *
 *  With this garbage collector, all items on the macro stack
 *  are required to be allocated on runtime by macroAllocItem().
 *  You may use the macroDupList() and macroDupItem() functions
 *  to reallocate already existing items and lists.
 *
 *  @{
 */

/** Initialize the garbage collector. */
struct GCData *macroGCInit (void (*stringFree) (void *));

/** Destroy the garbage collector data. */
void macroGCDestroy (struct GCData *gd);

/** Collect garbage. The stack from the macro execution context is the only
 *  root the collector uses, so ensure that there are no pointers to items
 *  allocated by macroGCAlloc() in local or global variables or other
 *  parts of memory, otherwise they will become invalid.
 */
void macroGCRun (ExecContext ctx);

/** Allocate an item. */
struct Item *macroGCAlloc (struct GCData *gd);

/** @} */

#endif /* ! __XTND_MACROGC_H__ */

