/**
 *  @file macroparser.c
 *  @brief Simple stack-based macro processor -- parser
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2007 - 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


typedef struct ParseContext *ParseContext;

struct ParseContext
{
	int top, depthsize;
	struct Item **begin, **end;

	long num;

	size_t buffsize;
	char *buff, *buffptr;
};


/* Local functions. */
static void parseAppend (ParseContext ctx, enum ItemType type);
static void parseAppendItem (ParseContext ctx, struct Item *item);
static int parseFinishList (ParseContext ctx);

static void macroFreeItem (struct Item *item);
static void macroFreeList (struct Item *list);


/** Remove an item from memory.
 *  @param[in] item  The item
 *  @see macroFreeList()
 */
static void macroFreeItem (struct Item *item)
{
	if (!item)
		return;

	switch (item->type)
	{
		case MACRO_STRING:
		case MACRO_SUB:
			free (item->data.str);
			break;
		case MACRO_LIST:
			macroFreeList (item->data.list);
		default:
			break;
	}
	free (item);
}

/** Remove a whole list from memory.
 *  @param[in] list  The list
 *  @see macroFreeItem()
 */
static void macroFreeList (struct Item *list)
{
	struct Item *p;

	while (list)
	{
		p = list->next;
		macroFreeItem (list);
		list = p;
	}
}

void macroFree (Macro *macro)
{
	macroFreeList (*macro);
	*macro = NULL;
}

int macroParse (const char *str, Macro *macro)
{
	struct ParseContext ctx;
	struct Item *result = NULL;
	int error = MACRO_ERR_SUCCESS;
	enum {COND_BEGIN, COND_END, COND_STRING} condition;

	if (!macro)
		return MACRO_ERR_NULL_POINTER;

	ctx.num = 0;

	/* Allocate the buffer. */
	ctx.buffsize = 64;
	ctx.buffptr = ctx.buff = xmalloc (ctx.buffsize);

	/* Allocate the list stack. */
	ctx.depthsize = 8;
	ctx.begin = xmalloc (ctx.depthsize * sizeof (struct Item *));
	ctx.end = xmalloc (ctx.depthsize * sizeof (struct Item *));

	/* Initialize. */
	ctx.top = 0;
	ctx.begin[0] = NULL;
	ctx.end[0] = NULL;

	condition = COND_BEGIN;

	while (1)
	{
		const char *s;
		const char *marker;

		s = str;
/*!re2c
		re2c:indent:top = 1;
		re2c:yyfill:enable = 0;
		re2c:condenumprefix = "COND_";
		re2c:define:YYGETCONDITION = "condition";
		re2c:define:YYGETCONDITION:naked = 1;
		re2c:define:YYSETCONDITION = "condition = ";
		re2c:define:YYCTYPE = "unsigned char";
		re2c:define:YYCURSOR = "str";
		re2c:define:YYMARKER = "marker";

		<BEGIN,END> [\t\n\r ]+ => BEGIN
		{
			continue;
		}

		<BEGIN> "["
		{
			if (ctx.top + 1 >= ctx.depthsize)
			{
				error = MACRO_ERR_TOO_NESTED;
				break;
			}

			ctx.top++;
			ctx.begin[ctx.top] = NULL;
			ctx.end[ctx.top] = NULL;
			continue;
		}

		<BEGIN,END> "]" => END
		{
			if (parseFinishList (&ctx))
			{
				error = MACRO_ERR_NOT_IN_LIST;
				break;
			}
			continue;
		}

		<BEGIN> "-"? [0-9]+ => END
		{
			ctx.num = strtol (s, NULL, 10);
			parseAppend (&ctx, MACRO_INT);
			continue;
		}

		<BEGIN> [-/+*.=a-zA-Z] [-/+*.=a-zA-Z0-9]* => END
		{
			strncpy (ctx.buff, s, str - s);
			ctx.buff[str - s] = '\0';
			parseAppend (&ctx, MACRO_SUB);
			continue;
		}

		<BEGIN> '"' => STRING
		{
			ctx.buffptr = ctx.buff;
			continue;
		}

		<STRING> "\\" [^\000]
		{
			if (ctx.buffsize <= ctx.buffptr - ctx.buff + 1)
				ctx.buff = xrealloc (ctx.buff, ctx.buffsize += 64);

			switch (*(s + 1))
			{
				case 'r': *ctx.buffptr++ = '\r'; break;
				case 'n': *ctx.buffptr++ = '\n'; break;
				case 't': *ctx.buffptr++ = '\t'; break;
				case 'a': *ctx.buffptr++ = '\a'; break;
				case 'b': *ctx.buffptr++ = '\b'; break;
				case 'f': *ctx.buffptr++ = '\f'; break;
				default:  *ctx.buffptr++ = *(s + 1);
			}
			continue;
		}

		<STRING> '"' => END
		{
			*ctx.buffptr = '\0';
			parseAppend (&ctx, MACRO_STRING);
			continue;
		}

		<STRING> [\000]
		{
			error = MACRO_ERR_STRING_NO_END;
			break;
		}

		<STRING> [^]
		{
			if (ctx.buffsize <= ctx.buffptr - ctx.buff + 1)
				ctx.buff = xrealloc (ctx.buff, ctx.buffsize += 64);

			*ctx.buffptr++ = *s;
			continue;
		}

		<BEGIN,END> [\000]
		{
			if (ctx.top)
				error = MACRO_ERR_LIST_NOT_CLOSED;
			break;
		}

		<BEGIN,END> [^]
		{
			error = MACRO_ERR_BAD_CHAR;
			break;
		}
*/
	}

	if (error)
	{
		if (ctx.top)
			macroFreeList (ctx.begin[ctx.top]);

		/* It is in a consistent state, so we can free it normally. */
		macroFreeList (ctx.begin[0]);
	}
	else
		result = ctx.begin[0];

	free (ctx.buff);
	free (ctx.begin);
	free (ctx.end);

	*macro = result;

	/* MACRO_ERR_SUCCESS is the default error */
	return error;
}


/* Append the data. */
static void parseAppend (ParseContext ctx, enum ItemType type)
{
	struct Item *item;

	/* Make a new item. */
	item = xmalloc (sizeof (struct Item));
	item->next = NULL;
	item->type = type;

	switch (type)
	{
		case MACRO_INT:
			item->data.num = ctx->num;
			break;
		case MACRO_STRING:
		case MACRO_SUB:
			item->data.str = xstrdup (ctx->buff);
			break;
		default:
			/* Should not happen. */
			free (item);
			return;
	}

	parseAppendItem(ctx, item);
}

/* Append an item. */
static void parseAppendItem (ParseContext ctx, struct Item *item)
{
	if (ctx->end[ctx->top])
		ctx->end[ctx->top]->next = item;
	ctx->end[ctx->top] = item;
	if (!ctx->begin[ctx->top])
		ctx->begin[ctx->top] = item;
}


/* Finish processing a list. */
static int parseFinishList (ParseContext ctx)
{
	struct Item *item;

	/* No list to finish. */
	if (!ctx->top)
		return 1;

	/* Append a new item. */
	item = xmalloc (sizeof (struct Item));
	item->next = NULL;
	item->type = MACRO_LIST;
	item->data.list = ctx->begin[ctx->top--];
	parseAppendItem (ctx, item);

	return 0;
}


