/**
 *  @file ircproto.c
 *  @brief A class that handles IRC connections.
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

/* TODO: Pth compatibility? */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#ifdef _WIN32
	#include <windows.h>
#else /* ! _WIN32 */
	#include <pthread.h>
	#include <semaphore.h>
#endif /* ! _WIN32 */

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


/** The maximum count of parameters parsed in a message. */
#define IRC_MAXPARAMS 32


/** Handle for an IRC connection. */
struct IrcHandle
{
	/** The IRC connection socket. */
	int socket;
	/** Thread ID. */
#ifdef _WIN32
	HANDLE threadId;
#else /* ! _WIN32 */
	pthread_t threadId;
#endif /* ! _WIN32 */
	/** Callback for events. */
	IrcMessageCB callback;
	/** User-specified data. */
	void *userData;
	/** Termination request. */
	volatile int terminate;
};


/* Local functions. */
static void ircParseLine (char *line, IrcMessage msg);

#ifdef _WIN32
static DWORD WINAPI ircThread (LPVOID ptr);
#else /* ! _WIN32 */
static void *ircThread (void *ptr);
#endif /* ! _WIN32 */



int ircInit (const char *hostname, unsigned short port, int timeout,
	IrcMessageCB callback, void *userData, IrcHandle *ih)
{
	int mysocket;
	IrcHandle data;
#ifndef _WIN32
	int r;
	pthread_attr_t threadAttr;
#endif /* ! _WIN32 */

	if ((mysocket = socketsTcpConnectNB (hostname, port, timeout)) == -1)
		return IRC_ERR_CONNECT;

	data = xmalloc (sizeof (struct IrcHandle));
	data->socket = mysocket;
	data->callback = callback;
	data->userData = userData;
	data->terminate = 0;

#ifdef _WIN32
	/* TODO: CREATE_SUSPENDED, assign data->threadId, resume? */
	if (!(data->threadId = CreateThread (NULL, 0, ircThread, data, 0, NULL)))
	{
		free (data);
		socketsClose (mysocket);
		return IRC_ERR_THREAD;
	}
#else /* ! _WIN32 */
	pthread_attr_init (&threadAttr);
	pthread_attr_setdetachstate (&threadAttr, PTHREAD_CREATE_JOINABLE);
	r = pthread_create (&data->threadId, &threadAttr, ircThread, data);
	pthread_attr_destroy (&threadAttr);
	if (r)
	{
		free (data);
		socketsClose (mysocket);
		return IRC_ERR_THREAD;
	}
#endif /* ! _WIN32 */

	*ih = data;
	return 0;
}

int ircDestroy (IrcHandle ih)
{
	/* Send a request for thread termination. */
	ih->terminate = 1;
#ifdef _WIN32
	WaitForSingleObject (ih->threadId, INFINITE);
#else /* ! _WIN32 */
	pthread_join (ih->threadId, NULL);
#endif /* ! _WIN32 */
	free (ih);
	return 0;
}


#ifdef _WIN32
static DWORD WINAPI ircThread (LPVOID ptr)
#else /* ! _WIN32 */
static void *ircThread (void *ptr)
#endif /* ! _WIN32 */
{
	IrcHandle ih = (IrcHandle) ptr;
	char *buffer, *line, *params[IRC_MAXPARAMS], *p;
	int rcvd, offset = 0;

	buffer = xmalloc (513 * sizeof (char));
	line = xmalloc (513 * sizeof (char));

	/* Read data from the IRC server -- 2 seconds timeout. */
	while ((rcvd = socketsRecvNB (ih->socket, buffer + offset,
		513 - 1 - offset, 2000)) != 0)
	{
		if (ih->terminate)
		{
			socketsShutdown (ih->socket, SOCKETS_SHUT_BOTH);
			/* For the case we get here again. */
			ih->terminate = 0;
			continue;
		}
		if (rcvd == -1)
		{
			if (socketsGetLastError () == SOCKETS_ETIMEDOUT)
				continue;
			else
				break;
		}

		rcvd += offset;
		buffer[rcvd] = '\0';

		/* Process the lines. */
		offset = 0;
		while ((p = strstr (buffer + offset, "\r\n")) != NULL)
		{
			struct IrcMessage msg = {NULL, NULL, NULL, 0};

			memcpy (line, buffer + offset, p - buffer - offset);
			line[p - buffer - offset] = '\0';
			offset = p - buffer + 2;

			/* TODO: A callback event. */

			msg.params = params;
			ircParseLine (line, &msg);

			ih->callback (ih, IRC_EVENT_MESSAGE, &msg, ih->userData);
		}

		/* If data still remain, move them to the beginning. */
		if (offset < rcvd)
			memmove (buffer, buffer + offset, rcvd - offset);
		offset = rcvd - offset;
	}

	free (line);
	free (buffer);
	socketsClose (ih->socket);

	ih->callback (ih, IRC_EVENT_DISCONNECT, NULL, ih->userData);
	return 0;
}

/** Parses the line into @a msg. The IrcMessage struct must be initialized
 *  before calling this function. After return, the contents of @a line
 *  is undefined.
 */
static void ircParseLine (char *line, IrcMessage msg)
{
	char *q, *strtokstate;
	unsigned linelen;

	linelen = strlen (line);

	q = strtok_r (line, " ", &strtokstate);
	while (q != NULL)
	{
		if (q[0] == ':' && q == line)
			msg->prefix = q + 1;
		else if (msg->command == NULL)
			msg->command = q;
		else if (q[0] == ':')
		{
			/* Remove the trailing zero added by strtok(). */
			if (q - line + strlen (q) != linelen)
				q[strlen (q)] = ' ';
			msg->params[msg->paramsCount++] = q + 1;
			break;
		}
		else
		{
			msg->params[msg->paramsCount++] = q;
			/* Don't let it overflow. */
			if (msg->paramsCount + 1 == IRC_MAXPARAMS)
				break;
		}
		q = strtok_r (NULL, " ", &strtokstate);
	}
}


int ircSendLine (IrcHandle ih, const char *line)
{
	char *buff;
	size_t size;
	int r;

	/* TODO: A callback event. */

	size = strlen (line) + 3;
	buff = xmalloc (size);
	snprintf (buff, size, "%s\r\n", line);

	r = socketsSend (ih->socket, buff, size - 1);
	free (buff);
	return r == -1 ? -1 : 0;
}

int ircPrintf (IrcHandle ih, char *format, ...)
{
	va_list args;
	char buff[513];

	va_start (args, format);
	vsnprintf (buff, 511, format, args);
	va_end (args);

	return ircSendLine (ih, buff);
}

int ircPrintfEx (IrcHandle ih, char *prefix, int space, char *format, ...)
{
	va_list args;

	unsigned sent = 0;
	int msgsize, msgspace = 510 - strlen (prefix);
	char *msgbuff, *msgptr, *p, *partbuff, buff[513];

	if (space && space < msgspace)
		msgspace = space;

	/* Allocate the buffer. */
	va_start (args, format);
	msgsize = vsnprintf (NULL, 0, format, args);
	va_end (args);

	msgptr = msgbuff = xmalloc (msgsize + 1);
	partbuff = xmalloc (msgspace + 1);

	/* Fill the buffer. */
	va_start (args, format);
	vsnprintf (msgbuff, msgsize + 1, format, args);
	va_end (args);

	while (msgptr < msgbuff + msgsize)
	{
		if (strlen (msgptr) < msgspace)
		{
			strcpy (partbuff, msgptr);
			msgptr += strlen (msgptr);
		}
		else
		{
			/* If possible, split at the spaces. */
			p = msgptr + msgspace;
			while (--p != msgptr)
				if (isspace (*p))
					break;

			if (p != msgptr)
			{
				*p = '\0';
				strcpy (partbuff, msgptr);
				msgptr += strlen (msgptr) + 1;
			}
			else
			{
				partbuff[msgspace] = '\0';
				strncpy (partbuff, msgptr, msgspace);
				msgptr += msgspace;
			}
		}

		snprintf (buff, 511, "%s%s", prefix, partbuff);

		if (ircSendLine (ih, buff) == -1)
			break;
		sent++;
	}

	free (msgbuff);
	free (partbuff);
	return sent;
}


int ircToLower (int ch)
{
	switch (ch)
	{
	case '[':
		return '{';
	case ']':
		return '}';
	case '\\':
		return '|';
	case '~':
		return '^';
	default:
		return (ch >= 'A' && ch <= 'Z') ? ch ^ 32 : ch;
	}
}

int ircToUpper (int ch)
{
	switch (ch)
	{
	case '{':
		return '[';
	case '}':
		return ']';
	case '|':
		return '\\';
	case '^':
		return '~';
	default:
		return (ch >= 'a' && ch <= 'z') ? ch ^ 32 : ch;
	}
}

int ircStrCmp (const char *s1, const char *s2)
{
	char ch;

	while (*s1 || *s2)
	{
		if ((ch = (ircToLower (*s1++) - ircToLower (*s2++))))
			return ch < 0 ? -1 : 1;
	}

	return 0;
}
