/**
 *  @file ircproto.h
 *  @brief IRC connections
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_IRCPROTO_H__
#define __XTND_IRCPROTO_H__

/**
 *  @defgroup ircproto IRC connections
 *  @ingroup net
 *
 *  This module allows you to establish and work with a connection
 *  to an IRC server.
 *
 *  @{
 */

/** Cannot connect to the server. */
#define IRC_ERR_CONNECT 1
/** Cannot create a new thread. */
#define IRC_ERR_THREAD 2
/** An invalid handle has been passed. */
#define IRC_ERR_BADHANDLE 3

/** A message event. The @em msg parameter of a callback function
 *  will contain message information.
 */
#define IRC_EVENT_MESSAGE 1
/** Connection lost. The object will still exist, but functions
 *  that require a connection to the server will fail. ircDestroy()
 *  should be called from the parent thread upon receiving this event.
 */
#define IRC_EVENT_DISCONNECT 2


/** IRC protocol message. */
typedef struct IrcMessage *IrcMessage;

/** IRC protocol message structure. */
struct IrcMessage
{
	char *prefix;
	char *command;
	char **params;
	unsigned int paramsCount;
};

/** IRC server object handle. */
typedef struct IrcHandle *IrcHandle;

/** A callback for incoming IRC messages. */
typedef int (*IrcMessageCB) (IrcHandle ih, int event,
	IrcMessage msg, void *userData);


/** Try to establish a connection to an IRC server. All the class does
 *  is parsing and maintaining a connection in another thread.
 *  @param[in] hostname  A hostname of the IRC server to connect to.
 *  @param[in] port  The port to connect to.
 *  @param[in] timeout  Connection timeout in seconds.
 *  @param[in] callBack  A callback function for incoming data.
 *  @param[in] userData  An undefined pointer to be in @em ih.
 *  @param[out] ih  A handle to a new IRC object.
 *  @return 0 on success, an error otherwise.
 */
int ircInit (const char *hostname, unsigned short port, int timeout,
	IrcMessageCB callBack, void *userData, IrcHandle *ih);

/** End the worker thread, if it's still alive, and free resources.
 *  @param[in] ih  A handle to the IRC object.
 *  @return 0 on success, an error otherwise.
 */
int ircDestroy (IrcHandle ih);

/** Send a line to the server. */
int ircSendLine(IrcHandle ih, const char *line);

/** Send formated text to the server.
 *  @param[in] ih  IRC server connection handle
 *  @param[in] format  The formatting string
 *  @return 0 on success, -1 on error.
 */
int ircPrintf (IrcHandle ih, char *format, ...);

/** Extended version of ircPrintf.
 *  @param[in] ih  IRC server connection handle
 *  @param[in] prefix  A prefix to the messages
 *  @param[in] space  After how many chars to wrap the message
 *  @param[in] format  The formatting string
 *  @return The count of successfully sent messages.
 */
int ircPrintfEx (IrcHandle ih, char *prefix, int space, char *format, ...);


/** tolower for IRC. RFC 2812: []\~ = {}|^
 *  @param[in] ch  The character to be returned in lower-case
 *  @return The character in lower-case.
 */
int ircToLower (int ch);

/** toupper for IRC. RFC 2812: []\~ = {}|^
 *  @param[in] ch  The character to be returned in upper-case
 *  @return The character in upper-case.
 */
int ircToUpper (int ch);

/** IRC case insensitive strcmp
 *  @param[in] s1  The first string
 *  @param[in] s2  The second string
 */
int ircStrCmp (const char *s1, const char *s2);

/** @} */

#endif /* ! __XTND_IRCPROTO_H__ */

