/**
 *  @file sockets.c
 *  @brief A wrapper library for sockets.
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009, 2011. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

/* TODO: IPv6, Pth compatibility? */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#ifdef _WIN32
	#define _WIN32_WINNT 0x0501
	#include <winsock2.h>
	#include <windows.h>
	#include <ws2tcpip.h>

	#define SOCKETS_ERR WSAGetLastError()
	#define SOCKETS_ERRID(e) WSA##e
	#define SOCKETS_IFWIN32(a, b) a
	#define SOCKETS_CLOSE closesocket
#else /* ! _WIN32 */
	#include <netdb.h>
	#include <unistd.h>
	#include <sys/socket.h>
	#include <sys/un.h>
	#include <arpa/inet.h>
	#include <net/if.h>
	#include <netinet/in.h>
	#include <netinet/tcp.h>
	#include <errno.h>
	#include <fcntl.h>
	#include <pthread.h>
	#include <poll.h>

	#define SOCKETS_ERR errno
	#define SOCKETS_ERRID(e) e
	#define SOCKETS_IFWIN32(a, b) b
	#define SOCKETS_CLOSE close
#endif /* ! _WIN32 */

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


#if defined (_MSC_VER)
	#define DATA_ALIGN(decl, alignment) \
		__declspec (align (alignment)) decl
#elif defined (__GNUC__)
	#define DATA_ALIGN(decl, alignment) \
		decl __attribute__ ((aligned (alignment)))
#else
	#define DATA_ALIGN(decl, alignment)
#endif


/** Thread-local storage index for the last error. */
static SOCKETS_IFWIN32 (DWORD, pthread_key_t) g_socketsErrorIndex;
/** Whether the module has been initialized. */
static int g_socketsInitialized;

/** Wait for read availability. */
#define SOCKETS_SELECT_READ 1

/** Wait for write availability. */
#define SOCKETS_SELECT_WRITE 2


/* Local functions. */
static int socketsPoll (int sockfd, int event, int timeout);
static int socketsGetAddrInfo (const char *restrict nodename,
	unsigned short service, const struct addrinfo *restrict hints,
	struct addrinfo **restrict res);
static int socketsSetNonBlocking (int sockfd, int value);
static void socketsSetLastError (SocketsError error);


void socketsInit (void)
{
#ifdef _WIN32
	WSADATA WSAData;
#endif /* _WIN32 */

	if (g_socketsInitialized)
		return;

#ifdef _WIN32
	WSAStartup (MAKEWORD (2, 2), &WSAData);

	g_socketsErrorIndex = TlsAlloc ();
	if (g_socketsErrorIndex == TLS_OUT_OF_INDEXES)
		abort ();
#else /* ! _WIN32 */
	if (pthread_key_create (&g_socketsErrorIndex, free))
		abort ();
#endif /* ! _WIN32 */

	g_socketsInitialized = 1;
}

/** Sets the last error. */
static void socketsSetLastError (SocketsError error)
{
#ifdef _WIN32
	if (!g_socketsInitialized)
		socketsInit ();

	TlsSetValue (g_socketsErrorIndex, (LPVOID) (INT_PTR) error);
#else /* ! _WIN32 */
	int *container;

	if (!g_socketsInitialized)
		socketsInit ();

	container = pthread_getspecific (g_socketsErrorIndex);
	if (!container)
	{
		container = xmalloc (sizeof (int));
		pthread_setspecific (g_socketsErrorIndex, container);
	}
	*container = error;
#endif /* ! _WIN32 */
}

SocketsError socketsGetLastError(void)
{
#ifdef _WIN32
	if (!g_socketsInitialized)
		socketsInit ();

	return (int) (INT_PTR) TlsGetValue (g_socketsErrorIndex);
#else /* ! _WIN32 */
	int *container;

	if (!g_socketsInitialized)
		socketsInit ();

	container = pthread_getspecific (g_socketsErrorIndex);
	return container ? *container : 0;
#endif /* ! _WIN32 */
}

/** Sets a sockets as blocking or non-blocking.
 *  Errors: None
 *  @return 0 on success, -1 on failure
 */
static int socketsSetNonBlocking (int sockfd, int value)
{
#ifdef _WIN32
	unsigned long int i;

	i = !!value;
	return ioctlsocket (sockfd, FIONBIO, &i) ? -1 : 0;
#else /* ! _WIN32 */
	int i;

	if ((i = fcntl (sockfd, F_GETFL)) < 0)
		return -1;

	if (value)
		i |= O_NONBLOCK;
	else
		i &= ~O_NONBLOCK;

	return fcntl (sockfd, F_SETFL, i) == -1 ? -1 : 0;
#endif /* ! _WIN32 */
}

/** getaddrinfo() wrapper. */
static int socketsGetAddrInfo (const char *restrict nodename,
	unsigned short service, const struct addrinfo *restrict hints,
	struct addrinfo **restrict res)
{
	char servname[8];

	snprintf (servname, sizeof (servname), "%d", service);
	return getaddrinfo (nodename, servname, hints, res);
}

int socketsTcpConnectNB (const char *hostname, unsigned short port,
	int timeout)
{
	struct addrinfo hints = {0}, *res;
	int mysocket;

	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	if (socketsGetAddrInfo (hostname, port, &hints, &res))
	{
		socketsSetLastError (SOCKETS_EFAULT);
		return -1;
	}
	if ((mysocket = socket (res->ai_family, res->ai_socktype, 0)) == -1)
	{
		socketsSetLastError (SOCKETS_EFAULT);
		freeaddrinfo (res);
		return -1;
	}
	socketsSetNonBlocking (mysocket, 1);

	while (connect (mysocket, res->ai_addr, res->ai_addrlen) == -1)
	{
		int error;

		error = SOCKETS_ERR;
		if (error == SOCKETS_IFWIN32 (WSAEWOULDBLOCK, EINPROGRESS))
		{
			SOCKETS_IFWIN32 (int, socklen_t) size = sizeof (int);
			int status;

			switch (socketsPoll (mysocket, SOCKETS_SELECT_WRITE, timeout))
			{
			case -1:
				socketsSetLastError (SOCKETS_EFAULT);
				goto socketsTcpConnectNB_error;
			case 0:
				socketsSetLastError (timeout ? SOCKETS_ETIMEDOUT
					: SOCKETS_EWOULDBLOCK);
				goto socketsTcpConnectNB_error;
			default:
				/* Get the error state of the socket. */
				if (getsockopt (mysocket, SOL_SOCKET, SO_ERROR,
					SOCKETS_IFWIN32 ((char *) &status, &status), &size) == -1
					|| size != sizeof (status))
				{
					socketsSetLastError (SOCKETS_EFAULT);
					goto socketsTcpConnectNB_error;
				}
				else if (status == 0)
					goto socketsTcpConnectNB_success;
				error = status;
			}
		}

		switch (error)
		{
#ifndef _WIN32
		case EINTR:
			continue;
#endif /* ! _WIN32 */
		case SOCKETS_ERRID (ENOTSOCK):
		case SOCKETS_IFWIN32 (WSAEINVAL, EBADF):
			socketsSetLastError (SOCKETS_EINVAL);
			break;
		case SOCKETS_ERRID (ENETUNREACH):
			socketsSetLastError (SOCKETS_ENETUNREACH);
			break;
		default:
			socketsSetLastError (SOCKETS_EFAULT);
		}
		goto socketsTcpConnectNB_error;
	}

socketsTcpConnectNB_success:
	freeaddrinfo (res);
	return mysocket;

socketsTcpConnectNB_error:
	SOCKETS_CLOSE (mysocket);
	freeaddrinfo (res);
	return -1;
}

int socketsTcpListen (const char *hostname, unsigned short port, int length)
{
	struct addrinfo hints = {0}, *res;
	int mysocket, error = 0;

	hints.ai_flags = AI_PASSIVE;
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	if (socketsGetAddrInfo (hostname, port, &hints, &res))
	{
		socketsSetLastError (SOCKETS_EFAULT);
		return -1;
	}
	if ((mysocket = socket (res->ai_family, res->ai_socktype, 0)) == -1)
	{
		socketsSetLastError (SOCKETS_EFAULT);
		freeaddrinfo (res);
		return -1;
	}

	if (bind (mysocket, res->ai_addr, res->ai_addrlen))
	{
		switch (SOCKETS_ERR)
		{
		case SOCKETS_ERRID (EADDRINUSE):
			socketsSetLastError (SOCKETS_EADDRINUSE);
			break;
		default:
			socketsSetLastError (SOCKETS_EFAULT);
		}
		error = 1;
	}
	else if (listen (mysocket, length))
	{
		socketsSetLastError (SOCKETS_EFAULT);
		error = 1;
	}
	freeaddrinfo (res);

	if (error)
	{
		SOCKETS_CLOSE (mysocket);
		return -1;
	}
	else
	{
		socketsSetNonBlocking (mysocket, 1);
		return mysocket;
	}
}

int socketsTcpAcceptNB (int sockfd, int timeout)
{
	int conn;

	switch (socketsPoll (sockfd, SOCKETS_SELECT_READ, timeout))
	{
	case -1:
		return -1;
	case 0:
		socketsSetLastError (timeout ? SOCKETS_ETIMEDOUT
			: SOCKETS_EWOULDBLOCK);
		return -1;
	}

	while ((conn = accept (sockfd, NULL, NULL)) == -1)
	{
		switch (SOCKETS_ERR)
		{
		case SOCKETS_IFWIN32 (WSAECONNRESET, ECONNABORTED):
		case SOCKETS_ERRID (EWOULDBLOCK):
			socketsSetLastError (timeout
				? SOCKETS_ETIMEDOUT : SOCKETS_EWOULDBLOCK);
			break;
#ifndef _WIN32
		case EINTR:
			continue;
		case EBADF:
#endif /* ! _WIN32 */
		case SOCKETS_ERRID (ENOTSOCK):
		case SOCKETS_ERRID (EINVAL):
			socketsSetLastError (SOCKETS_EINVAL);
			break;
		default:
			socketsSetLastError (SOCKETS_EFAULT);
		}
		return -1;
	}
	return conn;
}

size_t socketsRecvNB (int sockfd, void *buf, size_t len, int timeout)
{
	int rcvd, r;

	while ((rcvd = recv (sockfd, buf, len, 0)) == -1)
	{
		switch (SOCKETS_ERR)
		{
#ifndef _WIN32
		case EINTR:
			continue;
#endif /* ! _WIN32 */
		case SOCKETS_ERRID (ENOTCONN):
			socketsSetLastError (SOCKETS_ENOTCONN);
			break;
		case SOCKETS_IFWIN32 (WSAEWOULDBLOCK, EAGAIN):
			if (timeout == 0)
			{
				socketsSetLastError (SOCKETS_EWOULDBLOCK);
				return -1;
			}

			r = socketsPoll (sockfd, SOCKETS_SELECT_READ, timeout);
			if (r > 0)
				continue;
			if (r == -1)
				socketsSetLastError (SOCKETS_EFAULT);
			else if (!r)
				socketsSetLastError (SOCKETS_ETIMEDOUT);
			break;
		default:
			socketsSetLastError (SOCKETS_EFAULT);
		}
		return -1;
	}
	return rcvd;
}

size_t socketsSendNB (int sockfd, const void *data, size_t len, int timeout)
{
	int sent, r;

	while ((sent = send (sockfd, data, len, 0)) == -1)
	{
		switch (SOCKETS_ERR)
		{
#ifndef _WIN32
		case EINTR:
			continue;
#endif /* ! _WIN32 */
		case SOCKETS_ERRID (ENOTCONN):
			socketsSetLastError (SOCKETS_ENOTCONN);
			break;
		case SOCKETS_IFWIN32 (WSAENOTSOCK, EBADF):
		case SOCKETS_ERRID (EINVAL):
			socketsSetLastError (SOCKETS_EINVAL);
			break;
		case SOCKETS_ERRID (EWOULDBLOCK):
			if (timeout == 0)
			{
				socketsSetLastError (SOCKETS_EWOULDBLOCK);
				return -1;
			}

			r = socketsPoll (sockfd, SOCKETS_SELECT_WRITE, timeout);
			if (r > 0)
				continue;
			if (r == -1)
				socketsSetLastError (SOCKETS_EFAULT);
			else if (!r)
				socketsSetLastError (SOCKETS_ETIMEDOUT);
			break;
		default:
			socketsSetLastError (SOCKETS_EFAULT);
		}
		return -1;
	}
	return sent;
}

/** select/poll() wrapper for a single socket.
 *  Errors: EINVAL, ETIMEDOUT, EFAULT
 *  @return 0 on success, -1 on failure
 */
static int socketsPoll (int sockfd, int event, int timeout)
{
	int r;
/* TODO: Use WSAEvent -> replace int with a Socket struct. */
#ifdef _WIN32
	struct timeval mytimeout, *mytimeptr = NULL;
	fd_set readfds, writefds, *readptr, *writeptr;

	if (timeout != -1)
	{
		mytimeout.tv_sec = timeout / 1000;
		mytimeout.tv_usec = timeout % 1000 * 1000;
		mytimeptr = &mytimeout;
	}

	if (event & SOCKETS_SELECT_READ)
	{
		FD_ZERO (&readfds);
		FD_SET (sockfd, &readfds);
		readptr = &readfds;
	}
	else
		readptr = NULL;

	if (event & SOCKETS_SELECT_WRITE)
	{
		FD_ZERO (&writefds);
		FD_SET (sockfd, &writefds);
		writeptr = &writefds;
	}
	else
		writeptr = NULL;

	if ((r = select (sockfd + 1, readptr, writeptr, NULL, mytimeptr)) == -1)
#else /* ! _WIN32 */
	struct pollfd pfd;

	pfd.fd = sockfd;
	pfd.events = 0;

	if (event & SOCKETS_SELECT_READ)
		pfd.events |= POLLIN;
	if (event & SOCKETS_SELECT_WRITE)
		pfd.events |= POLLOUT;

	socketsPoll_start:
	if ((r = poll (&pfd, 1, timeout)) == -1)
#endif /* ! _WIN32 */
	{
		switch (SOCKETS_ERR)
		{
#ifndef _WIN32
		case EINTR:
			goto socketsPoll_start;
#endif /* ! _WIN32 */
		case SOCKETS_IFWIN32 (WSAENOTSOCK, EBADF):
		case SOCKETS_ERRID (EINVAL):
			socketsSetLastError (SOCKETS_EINVAL);
			break;
		default:
			socketsSetLastError (SOCKETS_EFAULT);
		}
	}
	return r;
}

int socketsShutdown (int sockfd, SocketsShutHow how)
{
	int i;

	switch (how)
	{
	case SOCKETS_SHUT_SEND:
		i = SOCKETS_IFWIN32 (SD_SEND, SHUT_WR);
		break;
	case SOCKETS_SHUT_RECEIVE:
		i = SOCKETS_IFWIN32 (SD_RECEIVE, SHUT_RD);
		break;
	case SOCKETS_SHUT_BOTH:
		i = SOCKETS_IFWIN32 (SD_BOTH, SHUT_RDWR);
		break;
	default:
		socketsSetLastError (SOCKETS_EINVAL);
		return -1;
	}
	;
	if ((i = shutdown (sockfd, i)) == -1)
	{
		switch (SOCKETS_ERR)
		{
		case SOCKETS_ERRID (ENOTCONN):
			socketsSetLastError (SOCKETS_ENOTCONN);
			break;
		case SOCKETS_IFWIN32 (WSAENOTSOCK, EBADF):
			socketsSetLastError (SOCKETS_EINVAL);
			break;
		default:
			socketsSetLastError (SOCKETS_EFAULT);
		}
	}
	return i;
}

int socketsClose (int sockfd)
{
#ifndef _WIN32
	socketsClose_start:
#endif /* ! _WIN32 */
	if (SOCKETS_CLOSE (sockfd) == -1)
	{
		switch (SOCKETS_ERR)
		{
#ifndef _WIN32
		case EINTR:
			goto socketsClose_start;
#endif /* ! _WIN32 */
		case SOCKETS_IFWIN32 (WSAENOTSOCK, EBADF):
			socketsSetLastError (SOCKETS_EINVAL);
			break;
		default:
			socketsSetLastError (SOCKETS_EFAULT);
		}
		return -1;
	}
	return 0;
}


