/**
 *  @file sockets.h
 *  @brief Sockets wrapper
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009, 2012. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_SOCKETS_H__
#define __XTND_SOCKETS_H__

/**
 *  @defgroup sockets Sockets wrapper
 *  @ingroup net
 *
 *  This module wraps the socket API in a multiplatform way.
 *
 *  Each thread has its own last error information just as expected. You must
 *  call socketsInit() before using any of the module functions in
 *  a multithreaded environment, though, to prevent race conditions, since
 *  otherwise the module initializes itself at the first opportunity to do so.
 *
 *  @{
 */

/** The last error that happened on a socket. */
typedef enum
{
	SOCKETS_EFAULT = 1,    /**< General error. */
	SOCKETS_EINVAL,        /**< Invalid value/argument. */
	SOCKETS_ETIMEDOUT,     /**< Operation has timeouted. */
	SOCKETS_ENETUNREACH,   /**< Network is unreachable. */
	SOCKETS_ENOTCONN,      /**< Not connected. */
	SOCKETS_EWOULDBLOCK,   /**< Non-blocking. */
	SOCKETS_EADDRINUSE     /**< Address in use. */
}
SocketsError;

/** How to shutdown the communication. */
typedef enum
{
	SOCKETS_SHUT_SEND,     /**< Shutdown send operations. */
	SOCKETS_SHUT_RECEIVE,  /**< Shutdown receive operations. */
	SOCKETS_SHUT_BOTH      /**< Shutdown all operations. */
}
SocketsShutHow;


/* TODO: Replace integers with struct pointers.
 * + Try to find something interesting for non-Win32 to add here.
 */
#if 0
typedef struct
{
	int socket;
#ifdef _WIN32
	WSAEvent read;
	WSAEvent write;
#endif
}
*Socket;
#endif


/** Initialize the module. This function is not thread-safe. */
void socketsInit (void);

/** Tells what error has happened.
 *  @return The last error code.
 */
SocketsError socketsGetLastError (void);

/** Initiate a TCP connection using a non-blocking socket.
 *  @param[in] hostname  Destination hostname or address.
 *  @param[in] port  Destination port.
 *  @param[in] timeout  The time to wait in milliseconds.
 *  @return -1 on error, a socket otherwise.
 *  @par Errors:
 *  EFAULT, ETIMEDOUT, EWOULDBLOCK, EINVAL, ENETUNREACH, EFAULT
 */
int socketsTcpConnectNB (const char *hostname, unsigned short port,
	int timeout);

/** Listen for connections.
 *  @param[in] hostname  Where to listen (NULL for everywhere).
 *  @param[in] port  The port to listen at.
 *  @param[in] length  The length of connections queue.
 *  @return -1 on error, a socket otherwise.
 *  @par Errors:
 *  EFAULT, EADDRINUSE
 */
int socketsTcpListen (const char *hostname, unsigned short port, int length);

/** Wait for an incomming connection and accept it.
 *  @param[in] sockfd  A listening socket.
 *  @param[in] timeout  The time to wait in milliseconds.
 *                      A value of -1 means "wait forever".
 *  @return -1 on error, a socket otherwise.
 *  @par Errors:
 *  EFAULT, EINVAL, ETIMEOUT, EWOULDBLOCK
 */
int socketsTcpAcceptNB (int sockfd, int timeout);

/** Time-outed recv for non-blocking sockets.
 *  @param[in] sockfd  The socket to read data from.
 *  @param[out] buf  A buffer for incoming data.
 *  @param[in] len  Size of the buffer.
 *  @param[in] timeout  The time to wait in milliseconds.
 *                      A value of -1 means "wait forever".
 *  @return -1 on error, the number of bytes stored in the buffer otherwise.
 *  @par Errors:
 *  EFAULT, ENOTCONN, EWOULDBLOCK, ETIMEDOUT
 */
size_t socketsRecvNB (int sockfd, void *buf, size_t len, int timeout);

/** Similar to recv() but waits for non-blocking sockets.
 *  @par Errors:
 *  EFAULT, ENOTCONN, EWOULDBLOCK
 */
#define socketsRecv(s, b, l) socketsRecvNB (s, b, l, -1)

/** Time-outed send for non-blocking sockets. */
size_t socketsSendNB (int sockfd, const void *data, size_t len, int timeout);

/** Similar to normal send() but waits for non-blocking sockets. */
#define socketsSend(s, d, l) socketsSendNB (s, d, l, -1)

/** Shutdown a socket.
 *  @param[in] sockfd  The socket to shutdown.
 *  @param[in] how  How to shutdown the communication.
 *  @return -1 on error, 0 otherwise.
 *  @par Errors:
 *  ENOTCONN, EINVAL, EFAULT
 */
int socketsShutdown (int sockfd, SocketsShutHow how);

/** Close a socket.
 *  @return -1 on error, 0 otherwise.
 *  @par Errors:
 *  EINVAL, EFAULT
 */
int socketsClose (int sockfd);

/** @} */

#endif /* ! __XTND_SOCKETS_H__ */

