/**
 *  @file socks.c
 *  @brief SOCKS5 support
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Premysl Janouch 2009, 2011, 2012. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#ifdef _WIN32
	#include <winsock2.h>
	#include <ws2tcpip.h>
#else /* ! _WIN32 */
	#include <sys/socket.h>
	#include <arpa/inet.h>
	#include <netinet/in.h>
#endif /* ! _WIN32 */

#include <xtnd/xtnd.h>
#include <xtndprivate.h>


/* RFC: http://www.ietf.org/rfc/rfc1928.txt */

#if defined (HAVE_INET_NTOP) && defined (HAVE_INET_PTON)
/* Address types. */
#define SOCKS_ATYP_IPV4    0x01
#define SOCKS_ATYP_IPV6    0x04
#define SOCKS_ATYP_DOMAIN  0x03


/** Initialize. */
static int socksInit (int sockfd)
{
	/* Protocol version 5, no authentication. */
	uint8_t initMsg[] = {5, 1, 0x00};
	/* Protocol version, chosen authentication method. */
	uint8_t methodMsg[2];

	if (socketsSend (sockfd, initMsg,   sizeof initMsg)   != sizeof initMsg)
		return -1;
	if (socketsRecv (sockfd, methodMsg, sizeof methodMsg) != sizeof methodMsg)
		return -1;

	/* No acceptable methods? */
	if (methodMsg[1] == 0xFF)
		return -1;

	return 0;
}

/** Send a connection request. */
static int socksSendRequest (int sockfd, const char *host, int port)
{
	struct in_addr ia;
	struct in6_addr ia6;
	MemStream ms;
	int result = 0;

	assert (host != NULL);

	ms = msCreate (64);
	msSetByteOrder (ms, MS_BYTE_ORDER_BE);

	msWriteUint8 (ms, 5);      /* Protocol version 5. */
	msWriteUint8 (ms, 0x01);   /* Command: Connect. */
	msWriteUint8 (ms, 0x00);   /* Reserved. */

	/* Destination address, based on what's stored in the argument. */
	/* TODO: Don't depend on inet_pton(), have an inet_addr() fallback. */
	if      (inet_pton (AF_INET,  host, &ia)  == 1)
	{
		msWriteUint8 (ms, SOCKS_ATYP_IPV4);
		msWrite (ms, &ia,  sizeof (ia),  1);
	}
	else if (inet_pton (AF_INET6, host, &ia6) == 1)
	{
		msWriteUint8 (ms, SOCKS_ATYP_IPV6);
		msWrite (ms, &ia6, sizeof (ia6), 1);
	}
	else
	{
		size_t len;

		len = strlen (host);
		assert (len <= 255);

		msWriteUint8 (ms, SOCKS_ATYP_DOMAIN);
		msWriteUint8 (ms, len);
		msWrite (ms, host, len, 1);
	}

	msWriteUint16 (ms, port);  /* Destination port. */

	if (socketsSend (sockfd, msGetPtr (ms), msGetPos (ms)) != msGetPos (ms))
		result = -1;

	msFree (ms);
	return result;
}

/** Process reply messages.
 *  @return Socket descriptor for the reply on success, -1 otherwise.
 */
static int socksProcessReply (int sockfd)
{
	struct in_addr ia;
	struct in6_addr ia6;
	char *bound_host = NULL;
	uint16_t bound_port;

	MemStream ms;
	uint8_t reply, atype, domain_len, dummy;
	int result = -1;

	ms = msCreate (64);
	msSetByteOrder (ms, MS_BYTE_ORDER_BE);

	/* Read just the first four bytes first. */
	if (socketsRecv (sockfd, msGetPtr (ms), 4) != 4)
		goto socksProcessReply_error;

	msReadUint8 (ms, &dummy);  /* Protocol version. */
	msReadUint8 (ms, &reply);  /* Reply field. */

	/* Did we succeed? */
	if (reply != 0x00)
		goto socksProcessReply_error;

	msReadUint8 (ms, &dummy);  /* Reserved. */
	msReadUint8 (ms, &atype);  /* Address type. */

	/* Read the server bound address. */
	/* TODO: Don't depend on inet_ntop(), have a fallback. */
	switch (atype)
	{
	case SOCKS_ATYP_IPV4:
		if (socketsRecv (sockfd, &ia.s_addr,
			sizeof ia.s_addr) != sizeof ia.s_addr)
			goto socksProcessReply_error;

		bound_host = xmalloc (INET_ADDRSTRLEN);
		if (!inet_ntop (AF_INET, &ia, bound_host, INET_ADDRSTRLEN))
			goto socksProcessReply_error;
		break;
	case SOCKS_ATYP_IPV6:
		if (socketsRecv (sockfd, &ia6.s6_addr,
			sizeof ia6.s6_addr) != sizeof ia6.s6_addr)
			goto socksProcessReply_error;

		bound_host = xmalloc (INET6_ADDRSTRLEN);
		if (!inet_ntop (AF_INET, &ia, bound_host, INET6_ADDRSTRLEN))
			goto socksProcessReply_error;
		break;
	case SOCKS_ATYP_DOMAIN:
		if (socketsRecv (sockfd, &domain_len, 1) != 1)
			goto socksProcessReply_error;

		msSetPos (ms, 0, 0);
		msSetSize (ms, domain_len + 1);

		if (socketsRecv (sockfd, msGetPtr (ms), domain_len) != domain_len)
			goto socksProcessReply_error;

		msWriteUint8 (ms, '\0');
		bound_host = msRelease (ms);
		break;
	}

	/* Read the server bound port. */
	msSetPos (ms, 0, 0);
	msSetSize (ms, 2);

	if (socketsRecv (sockfd, msGetPtr (ms), 2) != 2)
		goto socksProcessReply_error;

	msReadUint16 (ms, &bound_port);

	/* Connect. */
	if (bound_port)
		/* XXX: The timeout value, see below. */
		result = socketsTcpConnectNB (bound_host, bound_port, 30000);
	else
		/* XXX: I don't know where it says the server may do this
		 *      (this isn't SOCKS4) but this seems to be the solution. */
		result = sockfd;

socksProcessReply_error:
	free (bound_host);
	msFree (ms);
	return result;
}

int socksTcpConnect (const char *sockshost, int socksport,
	const char *targethost, int targetport)
{
	int sockfd, result = -1;

	/* XXX: The timeout should be either some kind of a global default
	 *      or a parameter that can be set to a value with a default meaning.
	 */
	sockfd = socketsTcpConnectNB (sockshost, socksport, 30000);
	if (sockfd == -1)
		return -1;

	if (!socksInit (sockfd)
	 && !socksSendRequest (sockfd, targethost, targetport))
		result = socksProcessReply (sockfd);

	if (result != sockfd)
		socketsClose (sockfd);
	return result;
}
#else /* HAVE_INET_NTOP && HAVE_INET_PTON */
int socksTcpConnect (const char *sockshost, int socksport,
	const char *targethost, int targetport)
{
	/* Not available. */
	return -1;
}
#endif /* HAVE_INET_NTOP && HAVE_INET_PTON */

