/**
 *  @file xalloc.h
 *  @brief malloc() wrapper
 *
 *  @note This file is a part of libxtnd.
 *        Copyright Přemysl Janouch 2010. All rights reserved.
 *  @note See the file LICENSE for licensing information.
 *
 */

#ifndef __XTND_XALLOC_H__
#define __XTND_XALLOC_H__

/**
 *  @defgroup xalloc malloc() wrapper
 *  @ingroup util
 *
 *  This module wraps malloc() and realloc(), so that they abort execution
 *  of the program if they fail to allocate memory.
 *
 *  @{
 */

/** The default function to call on memory allocation failure. */
void xtndXallocDie (const char *fun) ATTRIBUTE_NORETURN;

/** A function that should terminate the program when memory allocation
 *  fails. It may also print an error message. This function defaults
 *  to xtndXallocDie().
 *  @param[in] fun  The name of function where the error has happened.
 */
extern void (*g_xtndDie) (const char *);

/** A malloc wrapper that terminates the program on error. */
static inline void *xmalloc (size_t size) ATTRIBUTE_MALLOC;

/** A calloc wrapper that terminates the program on error. */
static inline void *xcalloc (size_t n, size_t size) ATTRIBUTE_MALLOC;

/** A realloc wrapper that terminates the program on error. */
static inline void *xrealloc (void *object, size_t size) ATTRIBUTE_MALLOC;

/** Make a copy of a string, terminate the program on error. */
char *xstrdup (const char *s);


static inline void *xmalloc (size_t size)
{
	void *p;

	p = malloc (size);
	if (!p)
		g_xtndDie ("malloc");
	return p;
}

static inline void *xcalloc (size_t n, size_t size)
{
	void *p;

	p = calloc (n, size);
	if (!p)
		g_xtndDie ("calloc");
	return p;
}

static inline void *xrealloc (void *object, size_t size)
{
	void *p;

	p = realloc (object, size);
	if (!p)
		g_xtndDie ("realloc");
	return p;
}

/** @} */

#endif /* ! __XTND_XALLOC_H__ */

