/*
 * vim: tabstop=8 shiftwidth=2 noexpandtab softtabstop=2 cinoptions=>2,{2,:0,t0,(0,W4
 *
 * <preamble>
 * gswat - A graphical program debugger for Gnome
 * Copyright (C) 2006  Robert Bragg
 * </preamble>
 *
 * <license>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * </license>
 *
 */

#include <gx/gx-window.h>
#include <gx/gx-connection.h>
#include <gx/gx-protocol-error.h>

#include <string.h>

/* Macros and defines */
#define GX_WINDOW_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), GX_TYPE_WINDOW, GXWindowPrivate))

/* Enums/Typedefs */
/* add your signals here */
enum
{
  EVENT_SIGNAL,
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_WRAP_XID,
  PROP_DEPTH,
  PROP_PARENT,
  PROP_X,
  PROP_Y,
  PROP_WIDTH,
  PROP_HEIGHT,
  PROP_BORDER_WIDTH,
  PROP_CLASS,
  PROP_VISUAL,
  PROP_ATTRIBUTE_ITEMS,
};

struct _GXWindowPrivate
{
  /* NB: I think it will lead to trouble if we try and maintain these
   * properties as anything other than write once - construct only
   * properties.
   *
   * Like XCB we should not try and provide any caching of state, instead
   * we can leave that to higher levels.
   *
   * Also you can't implement such properties via autogenerated code,
   * so you could only hope for a limited amount of sugar.
   */
  GXWindow *parent_construct;
  gboolean wrap_construct;
  gint16 x_construct;
  gint16 y_construct;
  guint16 width_construct;
  guint16 height_construct;
  guint16 border_width_construct;
  guint16 class_construct;
  guint32 visual_construct;
  GXMaskValueItem *attribute_items_construct;

  gint8 depth_construct;
};


/* Function definitions */
static void gx_window_class_init (_GXWindowClass * klass);
static void gx_window_get_property (GObject * object,
				    guint id,
				    GValue * value, GParamSpec * pspec);
static void gx_window_set_property (GObject * object,
				    guint property_id,
				    const GValue * value, GParamSpec * pspec);
/* static void gx_window_mydoable_interface_init(gpointer interface,
   gpointer data); */
static void gx_window_init (GXWindow * self);
static GObject *
gx_window_constructor (GType type,
		       guint n_construct_params,
		       GObjectConstructParam *construct_params);
static void gx_window_finalize (GObject * self);


/* Variables */
static GXDrawableClass *parent_class = NULL;
static guint gx_window_signals[LAST_SIGNAL] = { 0 };

static GHashTable *xid_to_windows_map = NULL;

/* NB: We have to hand code gx_window_get_type because we have to
 * have to name the class as _GXWindowClass so that it does not
 * conflict with the auto-generated typedef enum {} GXWindowClass;
 */
GType
gx_window_get_type (void)
{
  static GType self_type = 0;

  if (!self_type)
    {
      static const GTypeInfo object_info = {
	  sizeof (_GXWindowClass), /* class structure size */
	  NULL,			/* base class initializer */
	  NULL,			/* base class finalizer */
	  (GClassInitFunc) gx_window_class_init, /* class initializer */
	  NULL,			/* class finalizer */
	  NULL,			/* class data */
	  sizeof (GXWindow),	/* instance structure size */
	  0,			/* preallocated instances */
	  (GInstanceInitFunc) gx_window_init, /* instance initializer */
	  NULL			/* function table */
      };

      /* add the type of your parent class here */
      self_type = g_type_register_static (GX_TYPE_DRAWABLE, /* parent GType */
					  "GXWindow",	/* type name */
					  &object_info,	/* type info */
					  0	/* flags */
      );
#if 0
      /* add interfaces here */
      static const GInterfaceInfo mydoable_info = {
	  (GInterfaceInitFunc) gx_window_mydoable_interface_init,
	  (GInterfaceFinalizeFunc) NULL,
	  NULL			/* interface data */
      };

      if (self_type != G_TYPE_INVALID)
	{
	  g_type_add_interface_static (self_type,
				       MY_Type_MYDOABLE, &mydoable_info);
	}
#endif
    }

  return self_type;
}

static void
gx_window_class_init (_GXWindowClass * klass)	/* Class Initialization */
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GParamSpec *new_param;

  parent_class = g_type_class_peek_parent (klass);

  gobject_class->finalize = gx_window_finalize;

  gobject_class->get_property = gx_window_get_property;
  gobject_class->set_property = gx_window_set_property;

  gobject_class->constructor = gx_window_constructor;

  /* FIXME - mutex */
  if (!xid_to_windows_map)
    {
      xid_to_windows_map = g_hash_table_new (g_direct_hash, g_direct_equal);
    }

  new_param = g_param_spec_boolean("wrap", /* name */
				   "Wrap",	/* nick name */
				   "Simply wrap an already existing xid",
				   FALSE, /* default */
				   G_PARAM_WRITABLE
				   | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_WRAP_XID, new_param);

  new_param = g_param_spec_uint ("depth", /* name */
			         "Depth", /* nick name */
			         "Window Depth", /* description */
			         0, /* minimum */
			         G_MAXUINT8, /* maximum */
			         0, /* default */
			         G_PARAM_WRITABLE
			         | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_DEPTH, new_param);

  new_param = g_param_spec_object("parent", /* name */
				  "Parent",	/* nick name */
				  "Parent window",	/* description */
				  GX_TYPE_WINDOW,	/* GType */
				  G_PARAM_WRITABLE	/* flags */
				  | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_PARENT, new_param);

  new_param = g_param_spec_int("x", /* name */
			       "X",	/* nick name */
			       "X Coordinate",	/* description */
			       G_MININT16,	/* minimum */
			       G_MAXINT16,	/* maximum */
			       0,	/* default */
			       G_PARAM_WRITABLE
			       | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_X, new_param);

  new_param = g_param_spec_int("y", /* name */
			       "Y",	/* nick name */
			       "Y Coordinate",	/* description */
			       G_MININT16,	/* minimum */
			       G_MAXINT16,	/* maximum */
			       0,		/* default */
			       G_PARAM_WRITABLE
			       | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_Y, new_param);

  new_param = g_param_spec_uint("width", /* name */
			       "Width",	/* nick name */
			       "Window width",	/* description */
			       1,	/* minimum */
			       G_MAXUINT16,	/* maximum */
			       1,	/* default */
			       G_PARAM_WRITABLE
			       | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_WIDTH, new_param);

  new_param = g_param_spec_uint ("height", /* name */
			         "Height", /* nick name */
			         "Window height", /* description */
			         1, /* minimum */
			         G_MAXUINT16, /* maximum */
			         1, /* default */
			         G_PARAM_WRITABLE
			         | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_HEIGHT, new_param);

  new_param = g_param_spec_uint ("border_width", /* name */
			         "Border Width", /* nick name */
			         "Width of the Window Border",
			         0, /* minimum */
			         G_MAXUINT16, /* maximum */
			         0, /* default */
			         G_PARAM_WRITABLE
			         | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_BORDER_WIDTH,
				  new_param);

  new_param = g_param_spec_uint ("class", /* name */
			         "Window Class", /* nick name */
			         "The window class",
			         0, /* minimum */
			         G_MAXUINT16, /* maximum */
			         0, /* default */
			         G_PARAM_WRITABLE
			         | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_CLASS, new_param);

  new_param = g_param_spec_uint ("visual", /* name */
			         "Window Visual", /* nick name */
			         "The window visual",
			         0, /* minimum */
			         G_MAXUINT32, /* maximum */
			         0, /* default */
			         G_PARAM_WRITABLE
			         | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_VISUAL, new_param);

  new_param = g_param_spec_pointer ("attribute_items", /* name */
			            "Attribute Mask Value Items",
			            "An array of attributes as "
				    "Mask Value Items",
				    G_PARAM_WRITABLE
				    | G_PARAM_CONSTRUCT_ONLY);
  g_object_class_install_property (gobject_class, PROP_ATTRIBUTE_ITEMS,
				   new_param);

  /* set up signals */
  klass->event = NULL;
  gx_window_signals[EVENT_SIGNAL] =
    g_signal_new ("event", /* name */
		  G_TYPE_FROM_CLASS (klass), /* interface GType */
		  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED, /* signal flags */
		  G_STRUCT_OFFSET (_GXWindowClass, event), NULL,
		  NULL,	/* accumulator data */
		  g_cclosure_marshal_VOID__POINTER, /* c marshaller */
		  G_TYPE_NONE, /* return type */
		  1, /* number of parameters */
		  G_TYPE_POINTER/* vararg, list of param types */
    );

  g_type_class_add_private (klass, sizeof (GXWindowPrivate));
}

static void
gx_window_get_property (GObject * object,
			guint id, GValue * value, GParamSpec * pspec)
{
  /* GXWindow* self = GX_WINDOW(object); */

  switch (id)
    {
#if 0 /* template code */
    case PROP_NAME:
      g_value_set_int (value, self->priv->property);
      break;
#endif
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
      break;
    }
}

static void
gx_window_set_property (GObject * object,
			guint property_id,
			const GValue * value, GParamSpec * pspec)
{
  GXWindow* self = GX_WINDOW(object);

  switch (property_id)
    {
    case PROP_WRAP_XID:
      self->priv->wrap_construct = g_value_get_boolean (value);
      break;
    case PROP_DEPTH:
      self->priv->depth_construct = g_value_get_uint (value);
      break;
    case PROP_PARENT:
      self->priv->parent_construct = g_value_get_object (value);
      break;
    case PROP_X:
      self->priv->x_construct = g_value_get_int (value);
      break;
    case PROP_Y:
      self->priv->y_construct = g_value_get_int (value);
      break;
    case PROP_WIDTH:
      self->priv->width_construct = g_value_get_uint (value);
      break;
    case PROP_HEIGHT:
      self->priv->height_construct = g_value_get_uint (value);
      break;
    case PROP_BORDER_WIDTH:
      self->priv->border_width_construct = g_value_get_uint (value);
      break;
    case PROP_CLASS:
      self->priv->class_construct = g_value_get_uint (value);
      break;
    case PROP_VISUAL:
      self->priv->visual_construct = g_value_get_uint (value);
      break;
    case PROP_ATTRIBUTE_ITEMS:
      self->priv->attribute_items_construct = g_value_get_pointer (value);
      break;
    default:
      g_warning ("gx_window_set_property on unknown property");
      return;
    }
}

#if 0 /* template code */
static void
gx_window_mydoable_interface_init (gpointer interface, gpointer data)
{
  MyDoableIface *mydoable = interface;
  g_assert (G_TYPE_FROM_INTERFACE (mydoable) == MY_Type_MYDOABLE);

  mydoable->method1 = gx_window_method1;
  mydoable->method2 = gx_window_method2;
}
#endif

static void
gx_window_init (GXWindow * self)
{
  self->priv = GX_WINDOW_GET_PRIVATE (self);
  /* populate object here */
}

static GObject *
gx_window_constructor (GType type,
		       guint n_construct_params,
		       GObjectConstructParam *construct_params)
{
  GObject *object;
  GXWindow *self;
  GXDrawable *drawable;
  GXConnection *connection;
  xcb_connection_t *xcb_connection;
  guint32 xid;
  gboolean is_wrapping_xid = FALSE;
  int i;

  /* If this window is wrapping an existing XID, then we first look
   * for an existing window object to return instead */

  for (i = 0; i < n_construct_params; i++)
    {
      GObjectConstructParam *construct_param = &construct_params[i];
      GParamSpec *pspec = construct_param->pspec;
      if (strcmp (g_param_spec_get_name (pspec), "wrap") == 0)
	{
	  is_wrapping_xid = TRUE;
	  break;
	}
    }

  if (is_wrapping_xid)
    {
      GXWindow *existing_window;

      for (i = 0; i < n_construct_params; i++)
	{
	  GObjectConstructParam *construct_param = &construct_params[i];
	  GParamSpec *pspec = construct_param->pspec;
	  if (strcmp (g_param_spec_get_name (pspec), "xid") == 0)
	    {
	      xid = g_value_get_uint (construct_param->value);
	      break;
	    }
	}
#warning "Test XID wrapping code!!!"
      existing_window = gx_window_find_from_xid (xid);
      if (existing_window)
	return G_OBJECT (existing_window);
    }

  object = G_OBJECT_CLASS (parent_class)->constructor (type,
						       n_construct_params,
						       construct_params);
  self = GX_WINDOW (object);
  drawable = GX_DRAWABLE (self);

  connection = gx_drawable_get_connection (drawable);
  g_assert (connection);
  xcb_connection = gx_connection_get_xcb_connection (connection);

  xid = drawable->xid;
  if (!drawable->xid)
    {
      g_assert (!self->priv->wrap_construct);
      drawable->xid = xcb_generate_id (xcb_connection);
    }

  if (!self->priv->wrap_construct)
    {
      guint32 value_list_len = 0;
      guint32 *value_list = NULL;
      guint32 value_mask = 0;

      if (self->priv->attribute_items_construct)
	{
	  value_list_len =
	    gx_mask_value_items_get_count (
		self->priv->attribute_items_construct);
	  value_list = alloca (value_list_len * 4);

	  gx_mask_value_items_get_list (
	      self->priv->attribute_items_construct,
	      &value_mask,
	      value_list);
	}

      xcb_create_window (
	 xcb_connection,
	 self->priv->depth_construct,
	 drawable->xid,
	 gx_drawable_get_xid (GX_DRAWABLE(self->priv->parent_construct)),
	 self->priv->x_construct,
	 self->priv->y_construct,
	 self->priv->width_construct,
	 self->priv->height_construct,
	 self->priv->border_width_construct,
	 self->priv->class_construct,
	 self->priv->visual_construct,
	 value_mask,
	 value_list);
    }

  g_object_unref (connection);

  g_hash_table_insert (xid_to_windows_map,
		       GUINT_TO_POINTER (drawable->xid),
		       object);
  return object;
}

/* Instantiation wrapper */
GXWindow *
gx_window_new (GXConnection *connection,
	       GXWindow *parent,
	       guint16 x,
	       guint16 y,
	       guint16 width,
	       guint16 height,
	       guint32 event_mask)
{
  GXMaskValueItem mask_value_items[] = {
    { GX_CW_EVENT_MASK, event_mask },
    {0}
  };
  return GX_WINDOW (g_object_new (gx_window_get_type (),
				  "connection", connection,
				  "parent", parent,
				  "x", x,
				  "y", y,
				  "width", width,
				  "height", height,
				  "attribute_items", mask_value_items,
				  NULL));
#if 0
  return GX_WINDOW (g_object_new (gx_window_get_type (),
				  "connection", connection,
				  "depth", 0, /* Copy from parent */
				  "parent", parent,
				  "x", x,
				  "y", y,
				  "width", width,
				  "height", height,
				  "border_width", 0,
				  "class", 0, /* Copy from parent */
				  "visual", 0, /* Copy from parent */
				  "attribute_items", NULL,
				  NULL));
#endif
}

/* Users could instead just use g_object_new, with full control over
 * the properties passed and thus full control over the windows
 * construction. */
#if 0
GXWindow *
gx_window_new_raw (GXConnection *connection,
		   guint8 depth,
		   GXWindow *parent,
		   guint16 x,
		   guint16 y,
		   guint16 width,
		   guint16 height,
		   guint16 border_width,
		   guint16 _class,
		   GXVisualID visual,
		   guint32 visual_mask,
		   const guint32 value_list)
{
  /* FIXME */
  g_assert (0);
}
#endif

/* Instance Destruction */
void
gx_window_finalize (GObject * object)
{
  /* GXWindow *self = GX_WINDOW(object); */

  /* destruct your object here */
  G_OBJECT_CLASS (parent_class)->finalize (object);
}

GXConnection *
gx_window_get_connection (GXWindow *self)
{
  return gx_drawable_get_connection (GX_DRAWABLE (self));
}

/* Currently only for internal use, this does a lookup for an existing
 * GXWindow that corresponds to the passed xid. */
GXWindow *
gx_window_find_from_xid (guint32 xid)
{
  /* FIXME - mutex */
  GXWindow *window = g_hash_table_lookup (xid_to_windows_map,
					  GUINT_TO_POINTER (xid));
  return window ? g_object_ref (window) : NULL;
}

