// Service methods for performing kinematics calculations

#include <math.h>
#include "kinematics.h"   

#define ARM_LENGTH 1.0
#define ARM_LENGTH_SQUARED (ARM_LENGTH*ARM_LENGTH)

// Kinematics as solved by Mathematica:
// 
// Simplify[Solve[{
//   p == q, // both arms are considered of equal length
//   x == p*Cos[a] + q*Cos[a + b], 
//   y == p*Sin[a] + q*Sin[a + b]}, 
//   {a, b}]][[6]]
//
// {b -> ArcCos[(-2 q^2 + x^2 + y^2)/(2 q^2)], 
//  a -> ArcCos[(q x (x^2 + y^2) - Sqrt[q^2 y^2 (4 q^2 - x^2 - y^2) (x^2 + y^2)])/
//              (2 q^2 (x^2 + y^2))]}

void calculate_rotations_for_coordinate(boom_state *state, float x, float y) {
  float x2 = x*x;
  float y2 = y*y;  
  state->b = acos((-2*ARM_LENGTH_SQUARED+x2+y2)/(2*ARM_LENGTH_SQUARED));
  state->a = acos((ARM_LENGTH*x2+y2-sqrt(ARM_LENGTH_SQUARED*y2*(4*ARM_LENGTH_SQUARED-x2-y2)*(x2+y2)))/
            (2*ARM_LENGTH_SQUARED*(x2+y2)));  
}

// Find the angle in radians of deviance from the positive y axis. negative angles to the left of y-axis, 
// positive to the right.
double theta(double x, double y)
{
  double theta = atan(x/fabs(y));
  if (y>0) {
    return(theta);
  } else {
    if (theta>0) 
    {
      return(M_PI-theta);
    } else {
      return(-M_PI-theta);
    }
  }
}

// Calculates the rotations for the three joints in order to position the camera at the desired cartesian 
// coordinate pointing in the provided direction.
void calculate_position_camera(boom_state *state, float x, float y, float direction) {
  // Calculate rotations for the first two joints
  calculate_rotations_for_coordinate(state, x,y);
  // Calculate rotation for last joint by compensating for the rotation of the preceding joints
  state->c = direction - state->a - state->b;  
}
 
// Calculates the rotations for the three joints in order to position the camera at the desired cartesian
// coordinate pointing towards the provided target point.
void calculate_target_camera(boom_state *state, float x, float y, float target_x, float target_y) {
  // Calculate sight angle from position to target
  float angle_to_target = theta(target_x-x, target_y-y);
  calculate_position_camera(state, x,y,angle_to_target);
}
