/* avltree - library for manipulation of binary trees.
 * Ported from libavl.
 *
 * Copyright (C) 2012 Varphone Wong <varphone@qq.com>
 * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2004 Free Software
 * Foundation, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gallocator.h"
#include "avltree.h"

/// Tree data structure.
struct avltree_table {
    struct avltree_node* root;              ///< Tree's root.
    avltree_comparison_t compare;           ///< Comparison function.
    void* param;                            ///< Extra argument to |compare|.
    struct gallocator* allocator;           ///< Memory allocator.
    size_t count;                           ///< Number of items in tree.
    unsigned long generation;               ///< Generation number.
};

/// An AVL tree node.
struct avltree_node {
    struct avltree_node* link[2];   ///< Subtrees.
    void* data;                     ///< Pointer to data.
    signed char balance;            ///< Balance factor.
};

/// AVL traverser structure.
struct avltree_traverser {
    struct avltree_table* table;                    ///< Tree being traversed.
    struct avltree_node* node;                      ///< Current node in tree.
    struct avltree_node* stack[AVLTREE_MAX_HEIGHT]; ///< All the nodes above |node|.
    size_t height;                                  ///< Number of nodes in |parent|.
    unsigned long generation;                       ///< Generation number.
    unsigned long flags;                            ///< Flags.
};

/// Table traverser functions.
void traverser_init(struct avltree_traverser*, struct avltree_table*);
void* traverser_first(struct avltree_traverser*);
void* traverser_last(struct avltree_traverser*);
void* traverser_find(struct avltree_traverser*, void*);
void* traverser_insert(struct avltree_traverser*, void*);
void* traverser_copy(struct avltree_traverser*, const struct avltree_traverser*);
void* traverser_next(struct avltree_traverser*);
void* traverser_prev(struct avltree_traverser*);
void* traverser_cur(struct avltree_traverser*);
void* traverser_replace(struct avltree_traverser*, void*);

/// Creates and returns a new table
///   with comparison function |compare| using parameter |param|
///   and memory allocator |allocator|.
///   Returns |NULL| if memory allocation failed.
avltree_t avltree_create(avltree_comparison_t compare, void* param,
                         struct gallocator* allocator)
{
    struct avltree_table* tree;
    assert(compare != NULL);
    if (allocator == NULL) {
        allocator = &gallocator_default;
    }
    tree = allocator->malloc(allocator, sizeof(*tree));
    if (tree == NULL) {
        return NULL;
    }
    tree->root = NULL;
    tree->compare = compare;
    tree->param = param;
    tree->allocator = allocator;
    tree->count = 0;
    tree->generation = 0;
    return(avltree_t)tree;
}

/// Search |tree| for an item matching |item|, and return it if found.
///   Otherwise return |NULL|.
void* avltree_find(const avltree_t avl, const void* item)
{
    const struct avltree_node* p;
    const struct avltree_table* tree = (const struct avltree_table*)avl;
    assert(tree != NULL && item != NULL);
    for (p = tree->root; p != NULL;) {
        int cmp = tree->compare(item, p->data, tree->param);
        if (cmp < 0) {
            p = p->link[0];
        } else if (cmp > 0) {
            p = p->link[1];
        } else { // |cmp == 0|
            return p->data;
        }
    }
    return NULL;
}

/// Inserts |item| into |tree| and returns a pointer to |item|'s address.
///   If a duplicate item is found in the tree,
///   returns a pointer to the duplicate without inserting |item|.
///   Returns |NULL| in case of memory allocation failure.
void** avltree_probe(avltree_t avl, void* item)
{
    struct avltree_table* tree = (struct avltree_table*)avl;
    struct avltree_node* y, *z; // Top node to update balance factor, and parent.
    struct avltree_node* p, *q; // Iterator, and parent.
    struct avltree_node* n;     // Newly inserted node.
    struct avltree_node* w;     // New root of rebalanced subtree.
    int dir;                    // Direction to descend.
    unsigned char da[AVLTREE_MAX_HEIGHT]; // Cached comparison results.
    int k = 0;                  // Number of cached results.
    assert(tree != NULL && item != NULL);
    z = (struct avltree_node*) &tree->root;
    y = tree->root;
    dir = 0;
    for (q = z, p = y; p != NULL; q = p, p = p->link[dir]) {
        int cmp = tree->compare(item, p->data, tree->param);
        if (cmp == 0) {
            return &p->data;
        }
        if (p->balance != 0) {
            z = q, y = p, k = 0;
        }
        da[k++] = dir = cmp > 0;
    }
    n = q->link[dir] =
            tree->allocator->malloc(tree->allocator, sizeof(*n));
    if (n == NULL) {
        return NULL;
    }
    tree->count++;
    n->data = item;
    n->link[0] = n->link[1] = NULL;
    n->balance = 0;
    if (y == NULL) {
        return &n->data;
    }
    for (p = y, k = 0; p != n; p = p->link[da[k]], k++)
        if (da[k] == 0) {
            p->balance--;
        } else {
            p->balance++;
        }
    if (y->balance == -2) {
        struct avltree_node* x = y->link[0];
        if (x->balance == -1) {
            w = x;
            y->link[0] = x->link[1];
            x->link[1] = y;
            x->balance = y->balance = 0;
        } else {
            assert(x->balance == +1);
            w = x->link[1];
            x->link[1] = w->link[0];
            w->link[0] = x;
            y->link[0] = w->link[1];
            w->link[1] = y;
            if (w->balance == -1) {
                x->balance = 0, y->balance = +1;
            } else if (w->balance == 0) {
                x->balance = y->balance = 0;
            } else { /* |w->balance == +1| */
                x->balance = -1, y->balance = 0;
            }
            w->balance = 0;
        }
    } else if (y->balance == +2) {
        struct avltree_node* x = y->link[1];
        if (x->balance == +1) {
            w = x;
            y->link[1] = x->link[0];
            x->link[0] = y;
            x->balance = y->balance = 0;
        } else {
            assert(x->balance == -1);
            w = x->link[0];
            x->link[0] = w->link[1];
            w->link[1] = x;
            y->link[1] = w->link[0];
            w->link[0] = y;
            if (w->balance == +1) {
                x->balance = 0, y->balance = -1;
            } else if (w->balance == 0) {
                x->balance = y->balance = 0;
            } else { /* |w->balance == -1| */
                x->balance = +1, y->balance = 0;
            }
            w->balance = 0;
        }
    } else {
        return &n->data;
    }
    z->link[y != z->link[0]] = w;
    tree->generation++;
    return &n->data;
}

/// Inserts |item| into |avl|.
///   Returns |NULL| if |item| was successfully inserted
///   or if a memory allocation error occurred.
///   Otherwise, returns the duplicate item.
void* avltree_insert(avltree_t avl, void* item)
{
    void** p = avltree_probe(avl, item);
    return p == NULL || *p == item ? NULL : *p;
}

/// Inserts |item| into |avl|, replacing any duplicate item.
///   Returns |NULL| if |item| was inserted without replacing a duplicate,
///   or if a memory allocation error occurred.
///   Otherwise, returns the item that was replaced.
void* avltree_replace(avltree_t avl, void* item)
{
    void** p = avltree_probe(avl, item);
    if (p == NULL || *p == item) {
        return NULL;
    } else {
        void* r = *p;
        *p = item;
        return r;
    }
}

/// Deletes from |avl| and returns an item matching |item|.
///   Returns a null pointer if no matching item found.
void* avltree_delete(avltree_t avl, const void* item)
{
    struct avltree_table* tree = (struct avltree_table*)avl;
    // Stack of nodes.
    struct avltree_node* pa[AVLTREE_MAX_HEIGHT];    // Nodes.
    unsigned char da[AVLTREE_MAX_HEIGHT];           // |link[]| indexes.
    int k;                                          // Stack pointer.
    struct avltree_node* p;     // Traverses tree to find node to delete.
    int cmp;                    // Result of comparison between |item| and |p|.
    assert(tree != NULL && item != NULL);
    k = 0;
    p = (struct avltree_node*)&tree->root;
    for (cmp = -1; cmp != 0;
            cmp = tree->compare(item, p->data, tree->param)) {
        int dir = cmp > 0;
        pa[k] = p;
        da[k++] = dir;
        p = p->link[dir];
        if (p == NULL) {
            return NULL;
        }
    }
    item = p->data;
    if (p->link[1] == NULL) {
        pa[k - 1]->link[da[k - 1]] = p->link[0];
    } else {
        struct avltree_node* r = p->link[1];
        if (r->link[0] == NULL) {
            r->link[0] = p->link[0];
            r->balance = p->balance;
            pa[k - 1]->link[da[k - 1]] = r;
            da[k] = 1;
            pa[k++] = r;
        } else {
            struct avltree_node* s;
            int j = k++;
            for (;;) {
                da[k] = 0;
                pa[k++] = r;
                s = r->link[0];
                if (s->link[0] == NULL) {
                    break;
                }
                r = s;
            }
            s->link[0] = p->link[0];
            r->link[0] = s->link[1];
            s->link[1] = p->link[1];
            s->balance = p->balance;
            pa[j - 1]->link[da[j - 1]] = s;
            da[j] = 1;
            pa[j] = s;
        }
    }
    tree->allocator->free(tree->allocator, p);
    assert(k > 0);
    while (--k > 0) {
        struct avltree_node* y = pa[k];
        if (da[k] == 0) {
            y->balance++;
            if (y->balance == +1) {
                break;
            } else if (y->balance == +2) {
                struct avltree_node* x = y->link[1];
                if (x->balance == -1) {
                    struct avltree_node* w;
                    assert(x->balance == -1);
                    w = x->link[0];
                    x->link[0] = w->link[1];
                    w->link[1] = x;
                    y->link[1] = w->link[0];
                    w->link[0] = y;
                    if (w->balance == +1) {
                        x->balance = 0, y->balance = -1;
                    } else if (w->balance == 0) {
                        x->balance = y->balance = 0;
                    } else { // |w->balance == -1|
                        x->balance = +1, y->balance = 0;
                    }
                    w->balance = 0;
                    pa[k - 1]->link[da[k - 1]] = w;
                } else {
                    y->link[1] = x->link[0];
                    x->link[0] = y;
                    pa[k - 1]->link[da[k - 1]] = x;
                    if (x->balance == 0) {
                        x->balance = -1;
                        y->balance = +1;
                        break;
                    } else {
                        x->balance = y->balance = 0;
                    }
                }
            }
        } else {
            y->balance--;
            if (y->balance == -1) {
                break;
            } else if (y->balance == -2) {
                struct avltree_node* x = y->link[0];
                if (x->balance == +1) {
                    struct avltree_node* w;
                    assert(x->balance == +1);
                    w = x->link[1];
                    x->link[1] = w->link[0];
                    w->link[0] = x;
                    y->link[0] = w->link[1];
                    w->link[1] = y;
                    if (w->balance == -1) {
                        x->balance = 0, y->balance = +1;
                    } else if (w->balance == 0) {
                        x->balance = y->balance = 0;
                    } else { // |w->balance == +1|
                        x->balance = -1, y->balance = 0;
                    }
                    w->balance = 0;
                    pa[k - 1]->link[da[k - 1]] = w;
                } else {
                    y->link[0] = x->link[1];
                    x->link[1] = y;
                    pa[k - 1]->link[da[k - 1]] = x;
                    if (x->balance == 0) {
                        x->balance = +1;
                        y->balance = -1;
                        break;
                    } else {
                        x->balance = y->balance = 0;
                    }
                }
            }
        }
    }
    tree->count--;
    tree->generation++;
    return (void*) item;
}

/// Refreshes the stack of parent pointers in |trav|
///   and updates its generation number.
static void trav_refresh(struct avltree_traverser* trav)
{
    assert(trav != NULL);
    trav->generation = trav->table->generation;
    if (trav->node != NULL) {
        avltree_comparison_t cmp = trav->table->compare;
        void* param = trav->table->param;
        struct avltree_node* node = trav->node;
        struct avltree_node* i;
        trav->height = 0;
        for (i = trav->table->root; i != node;) {
            assert(trav->height < AVLTREE_MAX_HEIGHT);
            assert(i != NULL);
            trav->stack[trav->height++] = i;
            i = i->link[cmp(node->data, i->data, param) > 0];
        }
    }
}

/// Initializes |trav| for use with |tree|
///   and selects the null node.
void traverser_init(struct avltree_traverser* trav, struct avltree_table* tree)
{
    trav->table = tree;
    trav->node = NULL;
    trav->height = 0;
    trav->generation = tree->generation;
}

/// Initializes |trav| for |tree|
///   and selects and returns a pointer to its least-valued item.
///   Returns |NULL| if |tree| contains no nodes.
void* traverser_first(struct avltree_traverser* trav)
{
    struct avltree_node* x;
    trav->height = 0;
    trav->generation = trav->table->generation;
    x = trav->table->root;
    if (x != NULL)
        while (x->link[0] != NULL) {
            assert(trav->height < AVLTREE_MAX_HEIGHT);
            trav->stack[trav->height++] = x;
            x = x->link[0];
        }
    trav->node = x;
    return x != NULL ? x->data : NULL;
}

/// Initializes |trav| for |tree|
///   and selects and returns a pointer to its greatest-valued item.
///   Returns |NULL| if |tree| contains no nodes.
void* traverser_last(struct avltree_traverser* trav)
{
    struct avltree_node* x;
    trav->height = 0;
    trav->generation = trav->table->generation;
    x = trav->table->root;
    if (x != NULL)
        while (x->link[1] != NULL) {
            assert(trav->height < AVLTREE_MAX_HEIGHT);
            trav->stack[trav->height++] = x;
            x = x->link[1];
        }
    trav->node = x;
    return x != NULL ? x->data : NULL;
}

/// Searches for |item| in |tree|.
///   If found, initializes |trav| to the item found and returns the item
///   as well.
///   If there is no matching item, initializes |trav| to the null item
///   and returns |NULL|.
void* traverser_find(struct avltree_traverser* trav, void* item)
{
    struct avltree_node* p, *q;
    trav->height = 0;
    trav->generation = trav->table->generation;
    for (p = trav->table->root; p != NULL; p = q) {
        int cmp = trav->table->compare(item, p->data, trav->table->param);
        if (cmp < 0) {
            q = p->link[0];
        } else if (cmp > 0) {
            q = p->link[1];
        } else { /* |cmp == 0| */
            trav->node = p;
            return p->data;
        }
        assert(trav->height < AVLTREE_MAX_HEIGHT);
        trav->stack[trav->height++] = p;
    }
    trav->height = 0;
    trav->node = NULL;
    return NULL;
}

/// Attempts to insert |item| into |tree|.
///   If |item| is inserted successfully, it is returned and |trav| is
///   initialized to its location.
///   If a duplicate is found, it is returned and |trav| is initialized to
///   its location.  No replacement of the item occurs.
///   If a memory allocation failure occurs, |NULL| is returned and |trav|
///   is initialized to the null item.
void* traverser_insert(struct avltree_traverser* trav, void* item)
{
    void** p;
    p = avltree_probe(trav->table, item);
    if (p != NULL) {
        trav->node =
            ((struct avltree_node*)
             ((char*) p - offsetof(struct avltree_node, data)));
        trav->generation = trav->table->generation - 1;
        return *p;
    } else {
        traverser_init(trav, trav->table);
        return NULL;
    }
}

/// Initializes |trav| to have the same current node as |src|.
void* traverser_copy(struct avltree_traverser* trav, const struct avltree_traverser* src)
{
    if (trav != src) {
        trav->table = src->table;
        trav->node = src->node;
        trav->generation = src->generation;
        if (trav->generation == trav->table->generation) {
            trav->height = src->height;
            memcpy(trav->stack, (const void*) src->stack,
                   sizeof(*trav->stack) * trav->height);
        }
    }
    return trav->node != NULL ? trav->node->data : NULL;
}

/// Returns the next data item in inorder
///   within the tree being traversed with |trav|,
///   or if there are no more data items returns |NULL|.
void* traverser_next(struct avltree_traverser* trav)
{
    struct avltree_node* x;
    if (trav->generation != trav->table->generation) {
        trav_refresh(trav);
    }
    x = trav->node;
    if (x == NULL) {
        return traverser_first(trav);
    } else if (x->link[1] != NULL) {
        assert(trav->height < AVLTREE_MAX_HEIGHT);
        trav->stack[trav->height++] = x;
        x = x->link[1];
        while (x->link[0] != NULL) {
            assert(trav->height < AVLTREE_MAX_HEIGHT);
            trav->stack[trav->height++] = x;
            x = x->link[0];
        }
    } else {
        struct avltree_node* y;
        do {
            if (trav->height == 0) {
                trav->node = NULL;
                return NULL;
            }
            y = x;
            x = trav->stack[--trav->height];
        } while (y == x->link[1]);
    }
    trav->node = x;
    return x->data;
}

/// Returns the previous data item in inorder
///   within the tree being traversed with |trav|,
///   or if there are no more data items returns |NULL|.
void* traverser_prev(struct avltree_traverser* trav)
{
    struct avltree_node* x;
    if (trav->generation != trav->table->generation) {
        trav_refresh(trav);
    }
    x = trav->node;
    if (x == NULL) {
        return traverser_last(trav);
    } else if (x->link[0] != NULL) {
        assert(trav->height < AVLTREE_MAX_HEIGHT);
        trav->stack[trav->height++] = x;
        x = x->link[0];
        while (x->link[1] != NULL) {
            assert(trav->height < AVLTREE_MAX_HEIGHT);
            trav->stack[trav->height++] = x;
            x = x->link[1];
        }
    } else {
        struct avltree_node* y;
        do {
            if (trav->height == 0) {
                trav->node = NULL;
                return NULL;
            }
            y = x;
            x = trav->stack[--trav->height];
        } while (y == x->link[0]);
    }
    trav->node = x;
    return x->data;
}

/// Returns |trav|'s current item.
void* traverser_cur(struct avltree_traverser* trav)
{
    return trav->node != NULL ? trav->node->data : NULL;
}

/// Replaces the current item in |trav| by |new| and returns the item replaced.
///   |trav| must not have the null item selected.
///   The new item must not upset the ordering of the tree.
void* traverser_replace(struct avltree_traverser* trav, void* new)
{
    void* old;
    old = trav->node->data;
    trav->node->data = new;
    return old;
}

/// Destroys |new| with |avltree_destroy (new, destroy)|,
///   first setting right links of nodes in |stack| within |new|
///   to null pointers to avoid touching uninitialized data.
static void copy_error_recovery(struct avltree_node** stack, int height,
                                struct avltree_table* new,
                                avltree_item_dtor_t destroy)
{
    assert(stack != NULL && height >= 0 && new != NULL);
    for (; height > 2; height -= 2) {
        stack[height - 1]->link[1] = NULL;
    }
    avltree_destroy((avltree_t)new, destroy);
}

/// Copies |org| to a newly created tree, which is returned.
///   If |copy != NULL|, each data item in |org| is first passed to |copy|,
///   and the return values are inserted into the tree,
///   with |NULL| return values taken as indications of failure.
///   On failure, destroys the partially created new tree,
///   applying |destroy|, if non-null, to each item in the new tree so far,
///   and returns |NULL|.
///   If |allocator != NULL|, it is used for allocation in the new tree.
///   Otherwise, the same allocator used for |org| is used.
avltree_t avltree_copy(const avltree_t avl,
                       avltree_item_copy_t copy,
                       avltree_item_ctor_t construct,
                       avltree_item_dtor_t destroy,
                       struct gallocator* allocator)
{
    struct avltree_table* org = (struct avltree_table*)avl;
    struct avltree_node* stack[2 * (AVLTREE_MAX_HEIGHT + 1)];
    int height = 0;
    struct avltree_table* new;
    const struct avltree_node* x;
    struct avltree_node* y;
    assert(org != NULL);
    new = avltree_create(org->compare, org->param,
                         allocator != NULL ? allocator : org->allocator);
    if (new == NULL) {
        return NULL;
    }
    new->count = org->count;
    if (new->count == 0) {
        return new;
    }
    x = (const struct avltree_node*)&org->root;
    y = (struct avltree_node*)&new->root;
    for (;;) {
        while (x->link[0] != NULL) {
            assert(height < 2 * (AVLTREE_MAX_HEIGHT + 1));
            y->link[0] =
                new->allocator->malloc(new->allocator, sizeof(*y->link[0]));
            if (y->link[0] == NULL) {
                if (y != (struct avltree_node*)&new->root) {
                    y->data = NULL;
                    y->link[1] = NULL;
                }
                copy_error_recovery(stack, height, new, destroy);
                return NULL;
            }
            stack[height++] = (struct avltree_node*) x;
            stack[height++] = y;
            x = x->link[0];
            y = y->link[0];
        }
        y->link[0] = NULL;
        for (;;) {
            y->balance = x->balance;
            if (copy == NULL) {
                y->data = x->data;
            } else {
                y->data = copy(x->data, org->param);
                if (y->data == NULL) {
                    y->link[1] = NULL;
                    copy_error_recovery(stack, height, new, destroy);
                    return NULL;
                }
            }
            if (x->link[1] != NULL) {
                y->link[1] =
                    new->allocator->malloc(new->allocator, sizeof(*y->link[1]));
                if (y->link[1] == NULL) {
                    copy_error_recovery(stack, height, new, destroy);
                    return NULL;
                }
                x = x->link[1];
                y = y->link[1];
                break;
            } else {
                y->link[1] = NULL;
            }
            if (height <= 2) {
                return new;
            }
            y = stack[--height];
            x = stack[--height];
        }
    }
}

/// Frees storage allocated for |avl|.
///   If |destroy != NULL|, applies it to each data item in inorder.
void avltree_destroy(avltree_t avl, avltree_item_dtor_t destroy)
{
    struct avltree_table* tree = (struct avltree_table*)avl;
    struct avltree_node* p;
    struct avltree_node* q;
    assert(tree != NULL);
    for (p = tree->root; p != NULL; p = q)
        if (p->link[0] == NULL) {
            q = p->link[1];
            if (destroy != NULL && p->data != NULL) {
                destroy(p->data, tree->param);
            }
            tree->allocator->free(tree->allocator, p);
        } else {
            q = p->link[0];
            p->link[0] = q->link[1];
            q->link[1] = p;
        }
    tree->allocator->free(tree->allocator, tree);
}

avltraverser_t avltraverser_create(avltree_t avl)
{
    struct avltree_table* tree = (struct avltree_table*)avl;
    assert(tree);
    assert(tree->allocator);
    assert(tree->allocator->malloc);
    struct avltree_traverser* trav = tree->allocator->malloc(tree->allocator, sizeof(*trav));
    if (trav)
        traverser_init(trav, tree);
    return (avltraverser_t)trav;
}

avltraverser_t avltraverser_create_on_buffer(avltree_t avl, void* buffer, size_t size)
{
    assert(avl);
    assert(buffer);
    assert(size > 0);
    if (size < sizeof(struct avltree_traverser))
        return NULL_AVLTRAVERSER;
    struct avltree_traverser* trav = (struct avltree_traverser*)buffer;
    traverser_init(trav, (struct avltree_table*)avl);
    trav->flags |= AVLTRAVERSER_FLAG_COB;
    return (avltraverser_t)trav;
}

void avltraverser_destroy(avltraverser_t trav)
{
    struct avltree_traverser* trv = (struct avltree_traverser*)trav;
    if (!(trv->flags & AVLTRAVERSER_FLAG_COB)) {
        assert(trv->table);
        assert(trv->table->allocator);
        assert(trv->table->allocator->free);
        trv->table->allocator->free(trv->table->allocator, trv);
    }
}

void* avltraverser_first(avltraverser_t trav)
{
    return traverser_first((struct avltree_traverser*)trav);
}

void* avltraverser_last(avltraverser_t trav)
{
    return traverser_last((struct avltree_traverser*)trav);
}

void* avltraverser_find(avltraverser_t trav, void* item)
{
    return traverser_find((struct avltree_traverser*)trav, item);
}

void* avltraverser_insert(avltraverser_t trav, void* item)
{
    return traverser_insert((struct avltree_traverser*)trav, item);
}

void* avltraverser_copy(avltraverser_t trav, const avltraverser_t src)
{
    return traverser_copy((struct avltree_traverser*)trav, (struct avltree_traverser*)src);
}

void* avltraverser_next(avltraverser_t trav)
{
    return traverser_next((struct avltree_traverser*)trav);
}

void* avltraverser_prev(avltraverser_t trav)
{
    return traverser_prev((struct avltree_traverser*)trav);
}

void* avltraverser_cur(avltraverser_t trav)
{
    return traverser_cur((struct avltree_traverser*)trav);
}

void* avltraverser_replace(avltraverser_t trav, void* item)
{
    return traverser_replace((struct avltree_traverser*)trav, item);
}

