/*
 * Copyright (C) 2012 Varphone Wong <varphone@qq.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public Licens
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-
 *
 */

#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include "gallocator.h"

void* gallocator_calloc(struct gallocator* allocator, size_t nmemb, size_t size)
{
    assert(allocator);
    assert(nmemb > 0);
    assert(size > 0);
    return calloc(nmemb, size);
}

/// Frees |ptr|.
/// @param [in]	allocator       A pointer to a gallocator.
/// @param [in] ptr             Pointer pointed to allocated memory block.
/// @return no value returns.
void gallocator_free(struct gallocator* allocator, void* ptr)
{
    assert(allocator);
    assert(ptr);
    free(ptr);
}

/// Allocates |size| bytes of space using |malloc()|.
/// @param [in] allocator		A pointer to a gallocator.
/// @param [in] size			Size of the memory block.
/// @return a null pointer if allocation fails.
void* gallocator_malloc(struct gallocator* allocator, size_t size)
{
    assert(allocator);
    assert(size > 0);
    return malloc(size);
}

void* gallocator_realloc(struct gallocator* allocator, void* ptr, size_t size)
{
    assert(allocator);
    assert(ptr);
    assert(size > 0);
    return realloc(ptr, size);
}

void* gallocator_zalloc(struct gallocator* allocator, size_t size)
{
    assert(allocator);
    assert(size > 0);
    return calloc(1, size);
}

/// Default memory allocator that uses |malloc()| and |free()|.
struct gallocator gallocator_default = {
    .calloc     = gallocator_calloc,
    .free       = gallocator_free,
    .malloc     = gallocator_malloc,
    .realloc    = gallocator_realloc,
    .zalloc     = gallocator_zalloc,
    .user       = 0,
};

void gallocator_init(struct gallocator* allocator, void* user)
{
    assert(allocator);
    *allocator = gallocator_default;
    allocator->user = user;
}

