/*
 * Copyright (C) 2012 Varphone Wong <varphone@qq.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public Licens
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-
 *
 */

#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif
#if defined(HAVE_WIN32) || defined(HAVE_WIN64)
#include <windows.h>
#else
#include <pthread.h>
#endif
#include "gallocator.h"
#include "gmutex.h"

struct gmutex {
    unsigned int            flags;
    struct gallocator*      allocator;
#if defined(HAVE_WIN32) || defined(HAVE_WIN64)
    CRITICAL_SECTION        ctx;
#else
    pthread_mutex_t         ctx;
#endif
};

#if defined(HAVE_WIN32) || defined(HAVE_WIN64)
#define __mutex_init(x)     InitializeCriticalSection(x)
#define __mutex_fini(x)     DeleteCriticalSection(x)
static int __mutex_lock(CRITICAL_SECTION* cs)
{
    EnterCriticalSection(cs);
    return 0;
}
static int __mutex_trylock(CRITICAL_SECTION* cs)
{
    return TryEnterCriticalSection(cs) ? 0 : -1;
}
static int __mutex_unlock(CRITICAL_SECTION* cs)
{
    LeaveCriticalSection(cs);
    return 0;
}
#else
#define __mutex_init(x)     pthread_mutex_init(x, 0)
#define __mutex_fini(x)     pthread_mutex_destroy(x)
#define __mutex_lock(x)     pthread_mutex_lock(x)
#define __mutex_trylock(x)  pthread_mutex_trylock(x)
#define __mutex_unlock(x)   pthread_mutex_unlock(x)
#endif

gmutex_t gmutex_create(struct gallocator* allocator)
{
    allocator = allocator ? allocator : &gallocator_default;
    struct gmutex* mutex = allocator->zalloc(allocator, sizeof(*mutex));
    if (mutex) {
        mutex->flags         = 0;
        mutex->allocator     = allocator;
        __mutex_init(&mutex->ctx);
    }
    return (gmutex_t)mutex;
}

gmutex_t gmutex_create_on_buffer(void* buffer, size_t size)
{
    assert(buffer);
    assert(size >= sizeof(struct gmutex));
    struct gmutex* mutex = (struct gmutex*)buffer;
    mutex->flags       |= GMUTEX_FLAG_COB;
    mutex->allocator    = 0;
    __mutex_init(&mutex->ctx);
    return (gmutex_t)mutex;
}

void gmutex_destroy(gmutex_t mutex)
{
    struct gmutex* mtx = (struct gmutex*)mutex;
    if (mtx) {
        __mutex_fini(&mtx->ctx);
        if (!(mtx->flags & GMUTEX_FLAG_COB)) {
            assert(mtx->allocator);
            mtx->allocator->free(mtx->allocator, mtx);
        }
    }
}

int gmutex_lock(gmutex_t mutex)
{
    assert(mutex);
    return __mutex_lock(&((struct gmutex*)mutex)->ctx);
}

int gmutex_trylock(gmutex_t mutex)
{
    assert(mutex);
    return __mutex_trylock(&((struct gmutex*)mutex)->ctx);
}

int gmutex_unlock(gmutex_t mutex)
{
    assert(mutex);
    return __mutex_unlock(&((struct gmutex*)mutex)->ctx);
}

