/*
 * Copyright (C) 2012 Varphone Wong <varphone@qq.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public Licens
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-
 *
 */
#ifdef HAVE_ASSERT_H
#include <assert.h>
#else
#define assert(x)   (void)(x)
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include "gallocator.h"
#include "spool.h"

/// String pool
struct spool {
    char* data;                     ///< Data buffer.
    unsigned int size;              ///< The size of buffer.
    unsigned int rpos;              ///< Current read pos of buffer.
    unsigned int wpos;              ///< Current write pos of buffer.
    unsigned int flags;             ///< Flags.
    struct gallocator* allocator;   ///< Generic memory allocator.
};

/// Create a string pool.
spool_t spool_create(unsigned int size, struct gallocator* allocator)
{
    assert(size > 0);
    if (!allocator)
        allocator = &gallocator_default;
    struct spool* p = allocator->calloc(allocator, sizeof(*p) + size, 1);
    if (p) {
        p->data         = (char*)(p) + sizeof(*p);
        p->size         = size;
        p->rpos         = 0;
        p->wpos         = 0;
        p->flags        = 0;
        p->allocator    = allocator;
    }
    return p;
}

#define SPOOL_COB_STUFF_SIZE    8   ///< Padding for 8 bytes stuff.

/// Create a string pool on preallocated buffer.
spool_t spool_create_on_buffer(char* buffer, unsigned int size)
{
    assert(buffer);
    assert(size > 0);
    struct spool* p = (struct spool*)buffer;
    if (p) {
        p->data         = buffer + sizeof(*p) + SPOOL_COB_STUFF_SIZE;
        p->size         = size - sizeof(*p) - SPOOL_COB_STUFF_SIZE;
        p->rpos         = 0;
        p->wpos         = 0;
        p->flags        = SPOOL_FLAG_COB;
        p->allocator    = NULL;
    }
    return p;
}

/// Destroy a string pool.
void spool_destroy(spool_t pool)
{
    assert(pool);
    if (!(spool_flags(pool) & SPOOL_FLAG_COB)) {
        struct gallocator* allocator = ((struct spool*)pool)->allocator;
        assert(allocator);
        allocator->free(allocator, pool);
    }
}

/// Get string pool data.
char* spool_data(spool_t pool)
{
    assert(pool);
    return ((struct spool*)pool)->data;
}

/// Get string pool current read ptr.
char* spool_current_rptr(spool_t pool)
{
    assert(pool);
    return ((struct spool*)pool)->data + ((struct spool*)pool)->rpos;
}

/// Get string pool current write ptr.
char* spool_current_wptr(spool_t pool)
{
    assert(pool);
    return ((struct spool*)pool)->data + ((struct spool*)pool)->wpos;
}

/// Get string pool data length.
unsigned int spool_length(spool_t pool)
{
    assert(pool);
    return ((struct spool*)pool)->wpos;
}

/// Get pool size.
unsigned int spool_size(spool_t pool)
{
    assert(pool);
    return ((struct spool*)pool)->size;
}

/// Get available size in pool.
unsigned int spool_available_size(spool_t pool)
{
    assert(pool);
    return ((struct spool*)pool)->size - ((struct spool*)pool)->wpos;
}

/// Get string pool flags.
unsigned int spool_flags(spool_t pool)
{
    assert(pool);
    return ((struct spool*)pool)->flags;
}

/// Clear string pool.
void spool_clear(spool_t pool)
{
    assert(pool);
    ((struct spool*)pool)->rpos = 0;
    ((struct spool*)pool)->wpos = 0;
    memset(((struct spool*)pool)->data, 0, ((struct spool*)pool)->size);
}

/// Append string with foramt.
int spool_putfmt(spool_t pool, const char* fmt, ...)
{
    assert(pool);
    assert(fmt);
    va_list ap;
    va_start(ap, fmt);
	size_t size = spool_available_size(pool);
	assert(size > 0);
    int count = vsnprintf(spool_current_wptr(pool), size, fmt, ap);
    va_end(ap);
    if (count < 0 || count > size)
        return -1;
    ((struct spool*)pool)->wpos += count;
    return count;
}

/// Add new line.
int spool_endl(spool_t pool)
{
    assert(pool);
	size_t size = spool_available_size(pool);
	assert(size > 0);
    int count = snprintf(spool_current_wptr(pool), size, "\n");
    if (count < 0 || count > size)
        return -1;
    ((struct spool*)pool)->wpos += count;
    return count;
}

