/*
 * Copyright (C) 2012 Varphone Wong <varphone@qq.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public Licens
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-
 *
 */

#ifndef LIBGCONTAINERS_SPOOL_H
#define LIBGCONTAINERS_SPOOL_H

#ifndef SPOOL_T_DEFINED
#define SPOOL_T_DEFINED

#define SPOOL_FLAG_COB         0x00000001      ///< Indicate the spool was create on preallocated buffer.
#define SPOOL_FLAG_EXPANDABLE  0x00000002      ///< Indicate the spool can expand capacity when needed.

// Forwards.
struct gallocator;

typedef void* spool_t;  ///< type of String pool.

/// Create a string pool.
/// @param [in] size    String pool capacity.
/// @return pool instance.
spool_t spool_create(unsigned int size, struct gallocator* allocator);

/// Create a string pool on preallocated buffer.
/// @param [in] buffer  Preallocated buffer.
/// @param [in] size    Buffer size in bytes.
/// @return pool instance.
spool_t spool_create_on_buffer(char* buffer, unsigned int size);

/// Destroy a string pool.
/// @param [in] pool    String pool instance.
/// @return no value returns.
void spool_destroy(spool_t pool);

/// Get string pool data.
/// @param [in] pool    String pool instance.
/// @return data ptr in pool.
char* spool_data(spool_t pool);

/// Get string pool current read ptr.
/// @param [in] pool    String pool instance.
/// @return current write ptr.
char*spool_current_rptr(spool_t pool);

/// Get string pool current write ptr.
/// @param [in] pool    String pool instance.
/// @return current write ptr.
char*spool_current_wptr(spool_t pool);

/// Get string pool data length.
/// @param [in] pool    String pool instance.
/// @return length in bytes.
unsigned int spool_length(spool_t pool);

/// Get pool size.
/// @param [in] pool    String pool instance.
/// @return size in btytes.
unsigned int spool_size(spool_t pool);

/// Get available size in pool.
/// @param [in] pool    String pool instance.
/// @return size in bytes.
unsigned int spool_available_size(spool_t pool);

/// Get string pool flags.
/// @param [in] pool	String pool instance.
/// @return flags of string pool.
unsigned int spool_flags(spool_t pool);

/// Clear string pool.
/// @param [in] pool    String pool instance.
/// @return no value returns.
void spool_clear(spool_t pool);

/// Append string with foramt.
/// @param [in] pool    String pool instance, see [#spool_create/#spool_create_on_buffer].
/// @param [in] fmt     String format, see [#printf]
/// @return appened string length.
int spool_putfmt(spool_t pool, const char* fmt, ...);

/// Add new line.
/// @param [in] pool    Pool instance.
/// @return bytes added.
int spool_endl(spool_t pool);
#endif

#endif

