/*
 * Copyright (C) 2012 Varphone Wong <varphone@qq.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public Licens
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-
 *
 */

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "vector.h"

struct vector {
    unsigned int size;
    unsigned int capacity;
    unsigned int flags;
    unsigned int element_size;
    void* options[VECTOR_OPTION_MAX];
    char* array;
};

#define vector_malloc(v)               ((vector_malloc_t)((v)->options[VECTOR_MALLOC]))
#define vector_realloc(v)              ((vector_realloc_t)((v)->options[VECTOR_REALLOC]))
#define vector_free(v)                 ((vector_free_t)((v)->options[VECTOR_FREE]))
#define vector_memset(v)               ((vector_memset_t)((v)->options[VECTOR_MEMSET]))
#define vector_element_copy(v)         ((vector_element_copy_t)((v)->options[VECTOR_ELEMENT_COPY]))
#define vector_element_ctor(v)         ((vector_element_ctor_t)((v)->options[VECTOR_ELEMENT_CTOR]))
#define vector_element_dtor(v)         ((vector_element_dtor_t)((v)->options[VECTOR_ELEMENT_DTOR]))
#define vector_element_actsize(v)      ((v)->element_size + VECTOR_ELEMENT_PADSIZE)
#define vector_element_offset(v, i)    (vector_element_actsize(v) * (i))

vector_t vector_create(unsigned int capacity, unsigned int element_size,
                       vector_malloc_t v_malloc, vector_realloc_t v_realloc,
                       vector_free_t v_free, vector_memset_t v_memset)
{
    if (!v_malloc)
        v_malloc = VECTOR_DEFAULT_MALLOC;
    if (!v_realloc)
        v_realloc = VECTOR_DEFAULT_REALLOC;
    if (!v_free)
        v_free = VECTOR_DEFAULT_FREE;
    if (!v_memset)
        v_memset = VECTOR_DEFAULT_MEMSET;
    struct vector* vec = v_malloc(sizeof(*vec));
    if (vec) {
        v_memset(vec, 0, sizeof(*vec));
        vec->size = 0;
        vec->capacity = capacity > 0 ? capacity : 10;
        vec->flags = VECTOR_FLAG_EXPANDABLE;
        vec->element_size = element_size;
        vec->options[VECTOR_MALLOC] = v_malloc;
        vec->options[VECTOR_REALLOC] = v_realloc;
        vec->options[VECTOR_FREE] = v_free;
        vec->options[VECTOR_MEMSET] = v_memset;
        vec->options[VECTOR_ELEMENT_COPY] = (vector_element_copy_t)VECTOR_DEFAULT_MEMCPY;
        vec->array = v_malloc(vec->element_size * vec->capacity + VECTOR_ELEMENT_PADSIZE * vec->capacity);
        if (!vec->array) {
            v_free(vec);
            vec = 0;
        }
    }
    return (vector_t)vec;
}

vector_t vector_create_on_buffer(unsigned int element_size, void* buffer, unsigned int buffer_size)
{
    struct vector* vec = (struct vector*)buffer;
    if (vec) {
        VECTOR_DEFAULT_MEMSET(buffer, 0, buffer_size);
        vec->size = 0;
        vec->capacity = (buffer_size - sizeof(*vec)) / (element_size + VECTOR_ELEMENT_PADSIZE);
        vec->flags = VECTOR_FLAG_COB;
        vec->element_size = element_size;
        vec->options[VECTOR_ELEMENT_COPY] = (vector_element_copy_t)VECTOR_DEFAULT_MEMCPY;
        vec->array = (char*)vec + sizeof(*vec);
    }
    return (vector_t)vec;
}

void vector_destroy(vector_t vec)
{
    assert(vec);
    struct vector* v = (struct vector*)vec;
    vector_clear(vec);
    if (!(v->flags & VECTOR_FLAG_COB)) {
        vector_free(v)(v->array);
        vector_free(v)(v);
    }
}

void* vector_at(vector_t vec, unsigned int index)
{
    assert(vec);
    struct vector* v = (struct vector*)vec;
    if (index >= v->size)
        return 0;
    if (v->flags & VECTOR_FLAG_VALUEISREF)
        return *(void**)(v->array + index * vector_element_actsize(v));
    return (void*)(v->array + index * vector_element_actsize(v));
}

int vector_assign(vector_t vec, unsigned int index, void* element)
{
    assert(vec);
    assert(element);
    struct vector* v = (struct vector*)vec;
    if (index > v->size)
        return VECTOR_ENOITEM;
    element = vector_element_ctor(v) ? vector_element_ctor(v)(element) : element;
    if (vector_element_copy(v) != (vector_element_copy_t)VECTOR_DEFAULT_MEMCPY) {
        vector_element_copy(v)(v->array + vector_element_offset(v, index), element,
                               v->element_size);
    } else {
        vector_element_copy(v)(v->array + vector_element_offset(v, index),
                               (v->flags & VECTOR_FLAG_VALUEISREF) ?
                               &element : element,
                               v->element_size);
    }
    return VECTOR_SUCCESS;
}

void vector_clear(vector_t vec)
{
    assert(vec);
    struct vector* v = (struct vector*)vec;
    if (v->flags & VECTOR_FLAG_FASTCLEAR)
        v->flags |= VECTOR_FLAG_NOMOVETAIL;
    for (int i = 0; i < v->capacity; i++)
        vector_erase(vec, i);
    if (v->flags & VECTOR_FLAG_FASTCLEAR)
        v->flags &= ~VECTOR_FLAG_NOMOVETAIL;
}

int vector_erase(vector_t vec, unsigned int index)
{
    assert(vec);
    struct vector* v = (struct vector*)vec;
    if (index >= v->size)
        return VECTOR_ENOITEM;
    if (vector_element_ctor(v))
        vector_element_ctor(v)((void*)(v->array + vector_element_offset(v, index)));
    if (!(v->flags & VECTOR_FLAG_NOMOVETAIL))
        VECTOR_DEFAULT_MEMMOVE(v->array + vector_element_offset(v, index),
                               v->array + vector_element_offset(v, index + 1),
                               (v->size - 1 - index) * vector_element_actsize(v));
    v->size--;
    return VECTOR_SUCCESS;
}

int vector_push_back(vector_t vec, void* element)
{
    assert(vec);
    assert(element);
    struct vector* v = (struct vector*)vec;
    if (v->size >= v->capacity) {
        if (v->flags & VECTOR_FLAG_EXPANDABLE) {
            if (vector_reserve(vec, v->capacity * 2))
                return VECTOR_ENOMEM;
        } else {
            return VECTOR_EOVERFLOW;
        }
    }
    vector_assign(vec, v->size, element);
    v->size++;
    return VECTOR_SUCCESS;
}

int vector_reserve(vector_t vec, unsigned int capacity)
{
    assert(vec);
    struct vector* v = (struct vector*)vec;
    void* ptr = vector_realloc(v)(v->array, capacity * vector_element_actsize(v));
    if (!ptr)
        return VECTOR_ENOMEM;
    v->array = ptr;
    v->capacity = capacity;
    return VECTOR_SUCCESS;
}

unsigned int vector_capacity(vector_t vec)
{
    assert(vec);
    return ((struct vector*)vec)->capacity;
}

unsigned int vector_size(vector_t vec)
{
    assert(vec);
    return ((struct vector*)vec)->size;
}

unsigned int vector_flags(vector_t vec)
{
    assert(vec);
    return ((struct vector*)vec)->flags;
}

unsigned int vector_set_flags(vector_t vec, unsigned int flags)
{
    assert(vec);
    return ((struct vector*)vec)->flags = flags;
}

void* vector_get_option(vector_t vec, int option)
{
    assert(vec);
    if (option < 0 || option > VECTOR_OPTION_MAX)
        return 0;
    return ((struct vector*)vec)->options[option];
}

int vector_set_option(vector_t vec, int option, void* value)
{
    assert(vec);
    if (option < 0 || option > VECTOR_OPTION_MAX)
        return VECTOR_ENOITEM;
    ((struct vector*)vec)->options[option] = value;
    return VECTOR_SUCCESS;
}

void vector_foreach(vector_t vec, vector_element_iter_t iter, void* ctx)
{
    assert(vec);
    assert(iter);
    struct vector* v = (struct vector*)vec;
    for (int i = 0; i < v->size; i++) {
        if (iter(vector_at(vec, i), i, ctx) == VECTOR_ITERCHANGED)
            i--;
    }
}

vector_t vector_from_string_list(vector_t vec, const char* str, const char* sep)
{
    int created = 0;
    assert(str);
    if (!vec) {
        vec = vector_create2(0, sizeof(char*));
        if (vec) {
            vector_add_flags(vec, VECTOR_FLAG_VALUEISREF);
            vector_set_option(vec, VECTOR_ELEMENT_CTOR, strdup);
            vector_set_option(vec, VECTOR_ELEMENT_DTOR, free);
        }
    }
    assert(vec);
    if (!sep) sep = ",";
    int total_len   = strlen(str);
    int sep_len     = strlen(sep);
    char* s         = strdup(str);
    char* p         = strstr(s, sep);
    // Check whether only one item.
    if (!p) {
        vector_push_back(vec, (void*)s);
        return vec;
    }
    // Split each items.
    while (p) {
        // Check the next pattern, skip if equal seperator.
        if (p == s)
            goto next;
        // Get the item length.
        int str_len = p - s;
        char tmp    = p[str_len];
        s[str_len]  = 0;
        vector_push_back(vec, (void*)s);
        s[str_len]  = tmp;
next:
        // Move to next item.
        s = p + sep_len;
        // Check string pointer.
        if ((s - p) > total_len)
            break;
        // Get next pattern.
        p = strstr(s, sep);
    }
    // Append the last one if exists.
    if (s - str < total_len)
        vector_push_back(vec, (void*)s);
    return vec;
}

