/*
 * Copyright (C) 2012 Varphone Wong <varphone@qq.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public Licens
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-
 *
 */

#ifndef LIBVECTOR_VECTOR_H
#define LIBVECTOR_VECTOR_H

#ifndef VECTOR_T_DEFINED
#define VECTOT_T_DEFINED

#define VECTOR_DEFAULT_FREE free
#define VECTOR_DEFAULT_MALLOC malloc
#define VECTOR_DEFAULT_MEMCPY memcpy
#define VECTOR_DEFAULT_MEMMOVE memmove
#define VECTOR_DEFAULT_MEMSET memset
#define VECTOR_DEFAULT_REALLOC realloc

#define VECTOR_ELEMENT_PADSIZE  0x04

#define VECTOR_FLAG_COB         0x00000001      ///< Indicate the vector was create on buffer
#define VECTOR_FLAG_EXPANDABLE  0x00000002      ///< Indicate the vector can expand capacity when needed
#define VECTOR_FLAG_VALUEISREF  0x00000004      ///< Indicate the items is reference.
#define VECTOR_FLAG_FASTCLEAR   0x00000008      ///< Skip elements arrange on clearing.
#define VECTOR_FLAG_NOMOVETAIL  0x00000010      ///< Skip memory move on erasing an element.

#define NULL_VECTOR		((vector_t)0)           ///< Null vector.

enum vector_error_type {
    VECTOR_EOVERFLOW = -4,
    VECTOR_ENOITEM = -3,
    VECTOR_ENOMEM = -2,
    VECTOR_EGENERIC = -1,
    VECTOR_SUCCESS = 0,
    VECTOR_ITERCHANGED = 1,
};

enum vector_option_type {
    VECTOR_MALLOC,
    VECTOR_REALLOC,
    VECTOR_FREE,
    VECTOR_MEMMOVE,
    VECTOR_MEMSET,
    VECTOR_ELEMENT_COPY,
    VECTOR_ELEMENT_CTOR,
    VECTOR_ELEMENT_DTOR,
    VECTOR_OPTION_MAX,
};

typedef void* vector_t;
typedef void* (*vector_malloc_t)(unsigned int);
typedef void* (*vector_realloc_t)(void*, unsigned int);
typedef void (*vector_free_t)(void*);
typedef void* (*vector_memset_t)(void*, int, unsigned int);
typedef void (*vector_element_copy_t)(void*, const void*, unsigned int);
typedef void* (*vector_element_ctor_t)(void*);
typedef void (*vector_element_dtor_t)(void*);
typedef int (*vector_element_iter_t)(void*, unsigned int, void*);

vector_t vector_create(unsigned int capacity, unsigned int element_size,
                       vector_malloc_t malloc, vector_realloc_t realloc,
                       vector_free_t free, vector_memset_t memset);
#define vector_create2(c, s)        vector_create(c, s, 0, 0, 0, 0)
vector_t vector_create_on_buffer(unsigned int element_size, void* buffer,
                                 unsigned int buffer_size);
void vector_destroy(vector_t vec);

#define vector_is_null(vec)		((vec) != NULL_VECTOR)

void* vector_at(vector_t vec, unsigned int index);
int vector_assign(vector_t vec, unsigned int index, void* element);
void vector_clear(vector_t vec);
int vector_erase(vector_t vec, unsigned int index);
int vector_push_back(vector_t vec, void* element);
int vector_reserve(vector_t vec, unsigned int capacity);

unsigned int vector_capacity(vector_t vec);
unsigned int vector_size(vector_t vec);

unsigned int vector_flags(vector_t vec);
unsigned int vector_set_flags(vector_t vec, unsigned int flags);
#define vector_add_flags(vec, flags)    vector_set_flags(vec, vector_flags(vec) | (flags))
#define vector_clr_flags(vec, flags)    vector_set_flags(vec, vector_flags(vec) & (~flags))

void* vector_get_option(vector_t vec, int option);

/// Set vection option.
/// @param [in] vec         Vector instance.
/// @param [in] option      Request option.
/// @param [in] value       Option value.
/// @retval VECTOR_SUCCESS  Success.
/// @retval Otherwise       Error occurred.
int vector_set_option(vector_t vec, int option, void* value);

/// Iterator each items in vector.
/// @param [in] vec         Vector instance.
/// @param [in] iter        Iterator callback.
/// @param [in] ctx         User context.
/// @return no value reurns.
void vector_foreach(vector_t vec, vector_element_iter_t iter, void* ctx);

/// Split string to new or exists vector.
/// @param [in] vec         An exists vector, this can be null, if null, a new vector will be create.
/// @param [in] str         A list formatted string.
/// @param [in] sep         Seperator of list item.
/// @return vector that contain the list items.
vector_t vector_from_string_list(vector_t vec, const char* str, const char* sep);

#endif

#endif

