#include "types.h"
#include "defs.h"
#include "param.h"
#include "memlayout.h"
#include "stat.h"
#include "mmu.h"
#include "x86.h"
#include "proc.h"
#include "spinlock.h"
#include "fs.h"
#include "file.h"
#include "fcntl.h"

char * createPageFileName(struct proc *p){
	int tens = p->pid / 10;
	int units =  p->pid % 10;
	
	p->pageFileName[0] = '.';
	p->pageFileName[1] = 'p';
	p->pageFileName[2] = 'a';
	p->pageFileName[3] = 'g';
	p->pageFileName[4] = 'e';
	if (tens == 0){
		p->pageFileName[5] = units + 48;
		p->pageFileName[6] = '\0';
		return p->pageFileName;
	} 
	
	p->pageFileName[5] = tens + 48;
	p->pageFileName[6] = units + 48; 
	p->pageFileName[7] = '\0';
	return p->pageFileName;
}

struct {
  struct spinlock lock;
  struct proc proc[NPROC];
} ptable;

static struct proc *initproc;

int nextpid = 1;
extern void forkret(void);
extern void trapret(void);

static void wakeup1(void *chan);

void
pinit(void)
{
  initlock(&ptable.lock, "ptable");
}



//PAGEBREAK: 32
// Look in the process table for an UNUSED proc.
// If found, change state to EMBRYO and initialize
// state required to run in the kernel.
// Otherwise return 0.
static struct proc*
allocproc(void)
{
  
  struct proc *p;
  char *sp;

  acquire(&ptable.lock);
  for(p = ptable.proc; p < &ptable.proc[NPROC]; p++)
    if(p->state == UNUSED)
      goto found;
  release(&ptable.lock);
  return 0;

found:
  p->state = EMBRYO;
  p->pid = nextpid++;
  release(&ptable.lock);

  // Allocate kernel stack.
  if((p->kstack = kalloc()) == 0){
    p->state = UNUSED;
    return 0;
  }
  sp = p->kstack + KSTACKSIZE;
  
  // Leave room for trap frame.
  sp -= sizeof *p->tf;
  p->tf = (struct trapframe*)sp;
  
  // Set up new context to start executing at forkret,
  // which returns to trapret.
  sp -= 4;
  *(uint*)sp = (uint)trapret;

  sp -= sizeof *p->context;
  p->context = (struct context*)sp;
  memset(p->context, 0, sizeof *p->context);
  p->context->eip = (uint)forkret;
  #ifndef NONE
	p->pageFaultCount = 0;
	p->totalPagedOutCount = 0;
	p->pagesInFileCount = 0;
	p->pagesInMemCount = 0;
	p->pageFile = -1;
	
  #endif
	return p;
  
}

//PAGEBREAK: 32
// Set up first user process.
void
userinit(void)
{
  struct proc *p;
  extern char _binary_initcode_start[], _binary_initcode_size[];
  
  p = allocproc();
  initproc = p;
  if((p->pgdir = setupkvm(kalloc)) == 0)
    panic("userinit: out of memory?");
  inituvm(p->pgdir, _binary_initcode_start, (int)_binary_initcode_size);
  p->sz = PGSIZE;
  memset(p->tf, 0, sizeof(*p->tf));
  p->tf->cs = (SEG_UCODE << 3) | DPL_USER;
  p->tf->ds = (SEG_UDATA << 3) | DPL_USER;
  p->tf->es = p->tf->ds;
  p->tf->ss = p->tf->ds;
  p->tf->eflags = FL_IF;
  p->tf->esp = PGSIZE;
  p->tf->eip = 0;  // beginning of initcode.S

  safestrcpy(p->name, "initcode", sizeof(p->name));
  p->cwd = namei("/");

  p->state = RUNNABLE;
  
  
}

// Grow current process's memory by n bytes.
// Return 0 on success, -1 on failure.
int
growproc(int n)
{
  uint sz;
  
  sz = proc->sz;
  if(n > 0){
    if((sz = allocuvm(proc->pgdir, sz, sz + n)) == 0)
      return -1;
  } else if(n < 0){
    if((sz = deallocuvm(proc->pgdir, sz, sz + n)) == 0)
      return -1;
  }
  proc->sz = sz;
  switchuvm(proc);
  return 0;
}

// Create a new process copying p as the parent.
// Sets up stack to return as if from system call.
// Caller must set state of returned proc to RUNNABLE.
int
fork(void)
{
  int i, pid;
  struct proc *np;
  // Allocate process.
  if((np = allocproc()) == 0)
    return -1;

  // Copy process state from p.
  if((np->pgdir = copyuvm(proc->pgdir, proc->sz)) == 0){
    kfree(np->kstack);
    np->kstack = 0;
    np->state = UNUSED;
    return -1;
  }
  np->sz = proc->sz;
  np->parent = proc;
  *np->tf = *proc->tf;

  // Clear %eax so that fork returns 0 in the child.
  np->tf->eax = 0;
  
  #ifndef NONE
  /* 1.1, hard copy of information */
  if (np->pageFile == -1)
		np->pageFile = open(createPageFileName(np), O_CREATE | O_RDWR);
  for(i = 0; i < NOFILE; i++)
    if(proc->ofile[i] && i !=proc->pageFile) //dont copy open pagefile
      np->ofile[i] = filedup(proc->ofile[i]);
	else
	  np->ofile[i] = 0;
	  
  // call copy of files:
  copyPageFile(proc->pageFile, np->pageFile);
  close(np->pageFile);		//proc does not need that page file, its his child's
  
  np->pagesInFileCount = proc->pagesInFileCount;
  np->pagesInMemCount = proc->pagesInFileCount;
  np->totalPagedOutCount = proc->totalPagedOutCount;
  np->pageFaultCount = proc->pageFaultCount;
  
  //deep copy father's page file to new page file
  //copies addresses from father to child
  for(i = 0; i < MAX_PSYC_PAGES; i++){
	np->pagesInMemory[i] = proc->pagesInMemory[i];
	
  }
  for(i = 0; i < MAX_PSYC_PAGES; i++){
  np->pagesInPageFile[i] = proc->pagesInPageFile[i];
  }
  
  #else 
  for(i = 0; i < NOFILE; i++)
    if(proc->ofile[i])
		np->ofile[i] = filedup(proc->ofile[i]);
  #endif
  
  /* 1.1 end hard copy */
  
  
  np->cwd = idup(proc->cwd);
 
  pid = np->pid;
  np->state = RUNNABLE;
  safestrcpy(np->name, proc->name, sizeof(proc->name));
  
  
  return pid;
}

// Exit the current process.  Does not return.
// An exited process remains in the zombie state
// until its parent calls wait() to find out it exited.
void
exit(void)
{
  struct proc *p;
  int fd;

  if(proc == initproc)
    panic("init exiting");
  
  #ifdef TRUE
  procdump();
  #endif
  
  
  // Close all open files.
  for(fd = 0; fd < NOFILE; fd++){
    if(proc->ofile[fd]){
      fileclose(proc->ofile[fd]);
      proc->ofile[fd] = 0;
    }
  }
  //delete pagefile fd
  proc->pageFile = -1;

  iput(proc->cwd);
  proc->cwd = 0;

  acquire(&ptable.lock);

  // Parent might be sleeping in wait().
  wakeup1(proc->parent);

  // Pass abandoned children to init.
  for(p = ptable.proc; p < &ptable.proc[NPROC]; p++){
    if(p->parent == proc){
      p->parent = initproc;
      if(p->state == ZOMBIE)
        wakeup1(initproc);
    }
  }

  // Jump into the scheduler, never to return.
  proc->state = ZOMBIE;
  sched();
  panic("zombie exit");
}

// Wait for a child process to exit and return its pid.
// Return -1 if this process has no children.
int
wait(void)
{
  struct proc *p;
  int havekids, pid;

  acquire(&ptable.lock);
  for(;;){
    // Scan through table looking for zombie children.
    havekids = 0;
    for(p = ptable.proc; p < &ptable.proc[NPROC]; p++){
      if(p->parent != proc)
        continue;
      havekids = 1;
      if(p->state == ZOMBIE){
        // Found one.
        pid = p->pid;
        kfree(p->kstack);
        p->kstack = 0;
        freevm(p->pgdir);
        p->state = UNUSED;
        p->pid = 0;
        p->parent = 0;
        p->name[0] = 0;
        p->killed = 0;
        release(&ptable.lock);
        return pid;
      }
    }

    // No point waiting if we don't have any children.
    if(!havekids || proc->killed){
      release(&ptable.lock);
      return -1;
    }

    // Wait for children to exit.  (See wakeup1 call in proc_exit.)
    sleep(proc, &ptable.lock);  //DOC: wait-sleep
  }
}

void
register_handler(sighandler_t sighandler)
{
  char* addr = uva2ka(proc->pgdir, (char*)proc->tf->esp);
  if ((proc->tf->esp & 0xFFF) == 0)
    panic("esp_offset == 0");

    /* open a new frame */
  *(int*)(addr + ((proc->tf->esp - 4) & 0xFFF))
          = proc->tf->eip;
  proc->tf->esp -= 4;

    /* update eip */
  proc->tf->eip = (uint)sighandler;
}


//PAGEBREAK: 42
// Per-CPU process scheduler.
// Each CPU calls scheduler() after setting itself up.
// Scheduler never returns.  It loops, doing:
//  - choose a process to run
//  - swtch to start running that process
//  - eventually that process transfers control
//      via swtch back to the scheduler.
void
scheduler(void)
{
  struct proc *p;

  for(;;){
    // Enable interrupts on this processor.
    sti();

    // Loop over process table looking for process to run.
    acquire(&ptable.lock);
    for(p = ptable.proc; p < &ptable.proc[NPROC]; p++){
      if(p->state != RUNNABLE)
        continue;

      // Switch to chosen process.  It is the process's job
      // to release ptable.lock and then reacquire it
      // before jumping back to us.
      proc = p;
      switchuvm(p);
      p->state = RUNNING;
      swtch(&cpu->scheduler, proc->context);
      switchkvm();

      // Process is done running for now.
      // It should have changed its p->state before coming back.
      proc = 0;
    }
    release(&ptable.lock);

  }
}

// Enter scheduler.  Must hold only ptable.lock
// and have changed proc->state.
void
sched(void)
{
  int intena;

  if(!holding(&ptable.lock))
    panic("sched ptable.lock");
  if(cpu->ncli != 1)
    panic("sched locks");
  if(proc->state == RUNNING)
    panic("sched running");
  if(readeflags()&FL_IF)
    panic("sched interruptible");
  intena = cpu->intena;
  swtch(&proc->context, cpu->scheduler);
  cpu->intena = intena;
}

// Give up the CPU for one scheduling round.
void
yield(void)
{
  acquire(&ptable.lock);  //DOC: yieldlock
  proc->state = RUNNABLE;
  sched();
  release(&ptable.lock);
}

// A fork child's very first scheduling by scheduler()
// will swtch here.  "Return" to user space.
void
forkret(void)
{
  static int first = 1;
  // Still holding ptable.lock from scheduler.
  release(&ptable.lock);

  if (first) {
    // Some initialization functions must be run in the context
    // of a regular process (e.g., they call sleep), and thus cannot 
    // be run from main().
    first = 0;
    initlog();
  }
  
  // Return to "caller", actually trapret (see allocproc).
}

// Atomically release lock and sleep on chan.
// Reacquires lock when awakened.
void
sleep(void *chan, struct spinlock *lk)
{
  if(proc == 0)
    panic("sleep");

  if(lk == 0)
    panic("sleep without lk");

  // Must acquire ptable.lock in order to
  // change p->state and then call sched.
  // Once we hold ptable.lock, we can be
  // guaranteed that we won't miss any wakeup
  // (wakeup runs with ptable.lock locked),
  // so it's okay to release lk.
  if(lk != &ptable.lock){  //DOC: sleeplock0
    acquire(&ptable.lock);  //DOC: sleeplock1
    release(lk);
  }

  // Go to sleep.
  proc->chan = chan;
  proc->state = SLEEPING;
  sched();

  // Tidy up.
  proc->chan = 0;

  // Reacquire original lock.
  if(lk != &ptable.lock){  //DOC: sleeplock2
    release(&ptable.lock);
    acquire(lk);
  }
}

//PAGEBREAK!
// Wake up all processes sleeping on chan.
// The ptable lock must be held.
static void
wakeup1(void *chan)
{
  struct proc *p;

  for(p = ptable.proc; p < &ptable.proc[NPROC]; p++)
    if(p->state == SLEEPING && p->chan == chan)
      p->state = RUNNABLE;
}

// Wake up all processes sleeping on chan.
void
wakeup(void *chan)
{
  acquire(&ptable.lock);
  wakeup1(chan);
  release(&ptable.lock);
}

// Kill the process with the given pid.
// Process won't exit until it returns
// to user space (see trap in trap.c).
int
kill(int pid)
{
  struct proc *p;

  acquire(&ptable.lock);
  for(p = ptable.proc; p < &ptable.proc[NPROC]; p++){
    if(p->pid == pid){
      p->killed = 1;
      // Wake process from sleep if necessary.
      if(p->state == SLEEPING)
        p->state = RUNNABLE;
      release(&ptable.lock);
      return 0;
    }
  }
  release(&ptable.lock);
  return -1;
}

//PAGEBREAK: 36
// Print a process listing to console.  For debugging.
// Runs when user types ^P on console.
// No lock to avoid wedging a stuck machine further.
void
procdump(void)
{
  static char *states[] = {
  [UNUSED]    "unused",
  [EMBRYO]    "embryo",
  [SLEEPING]  "sleep ",
  [RUNNABLE]  "runble",
  [RUNNING]   "run   ",
  [ZOMBIE]    "zombie"
  };
  int i;
  struct proc *p;
  char *state;
  uint pc[10];
  int numOfPagesInMem = 0;
  
  for(p = ptable.proc; p < &ptable.proc[NPROC]; p++){
    if(p->state == UNUSED)
      continue;
    if(p->state >= 0 && p->state < NELEM(states) && states[p->state])
      state = states[p->state];
    else
      state = "???";
    cprintf("%d %s %s", p->pid, state, p->name);
    cprintf(" %d %d %d %d",p->pagesInMemCount, p->pagesInFileCount,p->pageFaultCount, p->totalPagedOutCount);
    
    if(p->state == SLEEPING){
      getcallerpcs((uint*)p->context->ebp+2, pc);
      for(i=0; i<10 && pc[i] != 0; i++)
        cprintf(" %p", pc[i]);
    }
    numOfPagesInMem += p->pagesInMemCount;
    cprintf("\n");
  }
  cprintf("%d% free pages in the system\n", 100 * ((MAX_PSYC_PAGES * NPROC) - numOfPagesInMem)/(MAX_PSYC_PAGES * NPROC));
  
} 


/* file system operations: */


static struct inode*
create(char *path, short type, short major, short minor)
{
  uint off;
  struct inode *ip, *dp;
  char name[DIRSIZ];

  if((dp = nameiparent(path, name)) == 0)
    return 0;
  ilock(dp);

  if((ip = dirlookup(dp, name, &off)) != 0){
    iunlockput(dp);
    ilock(ip);
    if(type == T_FILE && ip->type == T_FILE)
      return ip;
    iunlockput(ip);
    return 0;
  }

  if((ip = ialloc(dp->dev, type)) == 0)
    panic("create: ialloc");

  ilock(ip);
  ip->major = major;
  ip->minor = minor;
  ip->nlink = 1;
  iupdate(ip);

  if(type == T_DIR){  // Create . and .. entries.
    dp->nlink++;  // for ".."
    iupdate(dp);
    // No ip->nlink++ for ".": avoid cyclic ref count.
    if(dirlink(ip, ".", ip->inum) < 0 || dirlink(ip, "..", dp->inum) < 0)
      panic("create dots");
  }

  if(dirlink(dp, name, ip->inum) < 0)
    panic("create: dirlink");

  iunlockput(dp);

  return ip;
}

static int
fdalloc(struct file *f)
{
  int fd;

  for(fd = 0; fd < NOFILE; fd++){
    if(proc->ofile[fd] == 0){
      proc->ofile[fd] = f;
      return fd;
    }
  }
  return -1;
}


int
open(char* path, int omode)
{
 // char *path;
  int fd; // omode;
  struct file *f;
  struct inode *ip;
  
  if(omode & O_CREATE){
    begin_trans();
    ip = create(path, T_FILE, 0, 0);
    commit_trans();
    if(ip == 0)
      return -1;
  } else {
    if((ip = namei(path)) == 0)
      return -1;
    ilock(ip);
    if(ip->type == T_DIR && omode != O_RDONLY){
      iunlockput(ip);
      return -1;
    }
  }

  if((f = filealloc()) == 0 || (fd = fdalloc(f)) < 0){
    if(f)
      fileclose(f);
    iunlockput(ip);
    return -1;
  }
  iunlock(ip);

  f->type = FD_INODE;
  f->ip = ip;
  f->off = 0;
  f->readable = !(omode & O_WRONLY);
  f->writable = (omode & O_WRONLY) || (omode & O_RDWR);
  return fd;
}

int
read(int fd, char *buf, int count)
{
  struct file *f;
  if(fd < 0 || fd >= NOFILE || (f=proc->ofile[fd]) == 0)
    return -1;
  return fileread(f, buf, count);
}

int
write(int fd, char *buf, int count)
{
  struct file *f;

  if(fd < 0 || fd >= NOFILE || (f=proc->ofile[fd]) == 0)
    return -1;
  return filewrite(f, buf, count);
}

int
close(int fd)
{
  struct file *f;
  
  if(fd < 0 || fd >= NOFILE || (f=proc->ofile[fd]) == 0)
    return -1;
  proc->ofile[fd] = 0;
  fileclose(f);
  return 0;
}

int
lseekPage(int fd, uint pageNumber)
{
	struct file *f;
	if(fd < 0 || fd >= NOFILE || (f=proc->ofile[fd]) == 0)
		return -1;
	f->off = PGSIZE * pageNumber;
	return 0;
}

//takes a page and stores it to a file
int
storePage(struct proc *p, uint pageNum){
	lseekPage(p->pageFile, pageNum);
	
//	write(p->pageFile, , PGSIZE);
	
	return 0;
}

int
copyPageFile(int fatherFD, int childFD){
	struct file *father;
	struct file *child;
	char buf[PGSIZE];
	int count=0;
	int i=0;
	if(fatherFD < 0 || fatherFD >= NOFILE || (father=proc->ofile[fatherFD]) == 0)
		return -1;
	if(childFD < 0 || childFD >= NOFILE || (child=proc->ofile[childFD]) == 0)
		return -1;
	lseekPage(fatherFD, 0);
	lseekPage(childFD, 0);
	for	(i=0; i< MAX_PSYC_PAGES; i++){
		if (read(fatherFD, buf, PGSIZE) == 0){
			break;
		} else {
			count++;
			write(childFD, buf, PGSIZE);
		}
	}
	return count;
}

void insertToMetaData(int pagesInMemCellNum){
	#ifdef NFU
	return;
	#endif
	//if it is FIFO:
	proc->metaDataFIFO.pages[proc->metaDataFIFO.counter] = pagesInMemCellNum;
	proc->metaDataFIFO.counter++;

	
	
}

int getPageToRemove(){
	#ifdef FIFO
	return proc->metaDataFIFO.pages[0];
	#endif
	
	#ifdef NFU
	int i;
	uint min = -1; //max UINT
	int minIndex = -1;
	for (i = 0; i < MAX_PAGES_IN_FILE(); i++){
		if (proc->pagesInMemory[i].aCounter < min && proc->pagesInMemory[i].isUsed == PAGE_USED){
			min = proc->pagesInMemory[i].aCounter;
			minIndex = i;
		}
	} 
	// no removing!!
	return minIndex;
	#endif
	
	#ifdef NONE
	return 0;
	#endif
}


int removeFromMetaData(){
	#ifdef FIFO
	int ans = proc->metaDataFIFO.pages[0];
	int i;
	
	for(i = 0; i< proc->metaDataFIFO.counter - 1; i++){
		proc->metaDataFIFO.pages[i] = proc->metaDataFIFO.pages[i+1];
	}
	
	proc->metaDataFIFO.counter--;
	return ans;
	#endif
	
	#ifdef NFU
	return getPageToRemove();
	#endif
	
	#ifdef NONE
	return 0;
	#endif
}

void initMetaData(struct proc *p){
	int i = 0;
	p->metaDataFIFO.counter = 0;
	for (; i< MAX_PAGES_IN_FILE(); i++)
		p->metaDataFIFO.pages[i] = 0;
	for (i = 0; i< MAX_PSYC_PAGES; i++)
		p->pagesInMemory[i].aCounter = 0;
	for (i = 0; i< MAX_PAGES_IN_FILE(); i++)
		p->pagesInPageFile[i].aCounter = 0;
	
}

void updateAll(void){
	struct proc *p;
	acquire(&ptable.lock);
	for(p = ptable.proc; p < &ptable.proc[NPROC]; p++){
		if (p->pid == 1 || p->pid == 2 || p->state != RUNNABLE || p->state != RUNNING)
			continue;
		updateNFUData(p);
	}
			
	release(&ptable.lock);
	
}
