/**
 * @brief     libdae source
 * @author    Thomas Atwood (tatwood.net)
 * @date      2011
 * @copyright unlicense / public domain
 ****************************************************************************/
#include "dae.h"
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define dae_GET_HEADER(pobj_) \
    ((dae_obj_header*) (((ptrdiff_t) pobj_)-((sizeof(dae_obj_header)+7)&~7)))

#define dae_GET_PTR(phdr_) \
    ((dae_obj_ptr)  (((ptrdiff_t) phdr_)+((sizeof(dae_obj_header)+7)&~7)))

enum dae_obj_memberdeftype_e
{
    dae_MEMBER_ATTRIB,
    dae_MEMBER_ELEM
};

enum dae_obj_flags_e
{
    dae_XSD_ANY = 1 << 0,
};

typedef enum dae_obj_memberdeftype_e dae_obj_memberdeftype;
typedef enum dae_obj_flags_e dae_obj_flags;

typedef struct dae_obj_memberdef_s dae_obj_memberdef;
typedef struct dae_obj_typedef_s dae_obj_typedef;
typedef struct dae_obj_vector_s dae_obj_vector;
typedef struct dae_obj_list_s dae_obj_list;
typedef struct dae_obj_header_s dae_obj_header;

struct dae_obj_memberdef_s
{
    const char* name;
    const char* deflt;
    unsigned offset;
    short objtypeid;
    short min;
    short max;
    short seq;
};

struct dae_obj_typedef_s
{
    const char* name;
    dae_obj_memberdef* attribs;
    dae_obj_memberdef* elems;
    unsigned flags;
    unsigned size;
    unsigned dataoffset;
    short objtypeid;
    short datatypeid;
    unsigned short numattribs;
    unsigned short numelems;
    short datamin;
    short datamax;
};

struct dae_obj_vector_s
{
    void* values;
    size_t size;
};

struct dae_obj_list_s
{
    dae_obj_header* head;
    dae_obj_header* tail;
};

struct dae_obj_header_s
{
    dae_obj_header* parent;
    dae_obj_list* parentlist;
    const dae_obj_memberdef* memberdef;
    const char* membername;
    int structindex;
    const dae_obj_typedef* def;
    dae_obj_list attribs;
    dae_obj_list elems;
    dae_obj_header* prev;
    dae_obj_header* next;
};

static void dae_free(
    void* ptr);

static char* dae_strdup(
    const char* str);

static dae_obj_ptr dae_create_obj(
    const dae_obj_typedef* def);

static dae_obj_ptr dae_add_obj(
    dae_obj_header* parenthdr,
    dae_obj_list* parentlist,
    const dae_obj_memberdef* memberdef,
    const char* membername,
    const dae_obj_typedef* childdef,
    size_t childsize);

static void dae_destroy_obj(
    dae_obj_header* hdr);

static size_t dae_convert_string_floats(
    const char* str,
    float* floats_out);

static size_t dae_convert_string_int8s(
    const char* str,
    char* ints_out);

static size_t dae_convert_string_int16s(
    const char* str,
    short* ints_out);

static size_t dae_convert_string_int32s(
    const char* str,
    int* ints_out);

static size_t dae_convert_string_uint8s(
    const char* str,
    unsigned char* ints_out);

static size_t dae_convert_string_uint32s(
    const char* str,
    unsigned* ints_out);

static size_t dae_convert_string_words(
    const char* str,
    char** strings_out);

static size_t dae_count_string_floats(
    const char* str);

static size_t dae_count_string_int8s(
    const char* str);

static size_t dae_count_string_int16s(
    const char* str);

static size_t dae_count_string_int32s(
    const char* str);

static size_t dae_count_string_uint8s(
    const char* str);

static size_t dae_count_string_uint32s(
    const char* str);

static size_t dae_count_string_words(
    const char* str);

static void dae_get_schema(
    dae_obj_typedef** types,
    unsigned* numtypes);

static dae_obj_typedef* dae_get_type(
    unsigned objtypeid);

static dae_obj_typedef* dae_find_type(
    const char* name);

//****************************************************************************
static void dae_build_schema(
    dae_obj_typedef** types_out,
    unsigned* numtypes_out)
{
    static dae_obj_typedef s_types[736];
    dae_obj_typedef* t = s_types;
    t->name = "xs:boolean";
    t->size = sizeof(int);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOLEAN;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:hexBinary";
    t->size = sizeof(unsigned char);
    t->flags= 0;
    t->objtypeid = dae_ID_HEXBINARY;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_HEX8;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:float";
    t->size = sizeof(float);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:byte";
    t->size = sizeof(char);
    t->flags= 0;
    t->objtypeid = dae_ID_BYTE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT8;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:short";
    t->size = sizeof(short);
    t->flags= 0;
    t->objtypeid = dae_ID_SHORT;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT16;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:int";
    t->size = sizeof(int);
    t->flags= 0;
    t->objtypeid = dae_ID_INT;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:string";
    t->size = sizeof(char*);
    t->flags= 0;
    t->objtypeid = dae_ID_STRING;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:unsignedByte";
    t->size = sizeof(unsigned char);
    t->flags= 0;
    t->objtypeid = dae_ID_UNSIGNEDBYTE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT8;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:unsignedInt";
    t->size = sizeof(unsigned int);
    t->flags= 0;
    t->objtypeid = dae_ID_UNSIGNEDINT;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:anyURI";
    t->size = sizeof(dae_anyURI);
    t->flags= 0;
    t->objtypeid = dae_ID_ANYURI;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:dateTime";
    t->size = sizeof(dae_dateTime);
    t->flags= 0;
    t->objtypeid = dae_ID_DATETIME;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:double";
    t->size = sizeof(dae_double);
    t->flags= 0;
    t->objtypeid = dae_ID_DOUBLE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:ID";
    t->size = sizeof(dae_ID);
    t->flags= 0;
    t->objtypeid = dae_ID_ID;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:integer";
    t->size = sizeof(dae_integer);
    t->flags= 0;
    t->objtypeid = dae_ID_INTEGER;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:long";
    t->size = sizeof(dae_long);
    t->flags= 0;
    t->objtypeid = dae_ID_LONG;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:Name";
    t->size = sizeof(dae_Name);
    t->flags= 0;
    t->objtypeid = dae_ID_NAME;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:NCName";
    t->size = sizeof(dae_NCName);
    t->flags= 0;
    t->objtypeid = dae_ID_NCNAME;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:NMTOKEN";
    t->size = sizeof(dae_NMTOKEN);
    t->flags= 0;
    t->objtypeid = dae_ID_NMTOKEN;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:nonNegativeInteger";
    t->size = sizeof(dae_nonNegativeInteger);
    t->flags= 0;
    t->objtypeid = dae_ID_NONNEGATIVEINTEGER;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:positiveInteger";
    t->size = sizeof(dae_positiveInteger);
    t->flags= 0;
    t->objtypeid = dae_ID_POSITIVEINTEGER;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:token";
    t->size = sizeof(dae_token);
    t->flags= 0;
    t->objtypeid = dae_ID_TOKEN;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "xs:unsignedLong";
    t->size = sizeof(dae_unsignedLong);
    t->flags= 0;
    t->objtypeid = dae_ID_UNSIGNEDLONG;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "math:math";
    t->size = sizeof(dae_math);
    t->flags= 0 | dae_XSD_ANY;
    t->objtypeid = dae_ID_MATH;
    t->datatypeid=dae_ID_INVALID;
    ++t;
    t->name = "float_type";
    t->size = sizeof(dae_float_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "int_type";
    t->size = sizeof(dae_int_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "uint_type";
    t->size = sizeof(dae_uint_type);
    t->flags= 0;
    t->objtypeid = dae_ID_UINT_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "sidref_type";
    t->size = sizeof(dae_sidref_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SIDREF_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "sid_type";
    t->size = sizeof(dae_sid_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SID_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "list_of_ints_type";
    t->size = sizeof(dae_list_of_ints_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIST_OF_INTS_TYPE;
    t->dataoffset=offsetof(dae_list_of_ints_type,data);
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=0;
    t->datamax=-1;
    ++t;
    t->name = "list_of_names_type";
    t->size = sizeof(dae_list_of_names_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIST_OF_NAMES_TYPE;
    t->dataoffset=offsetof(dae_list_of_names_type,data);
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=0;
    t->datamax=-1;
    ++t;
    t->name = "list_of_uints_type";
    t->size = sizeof(dae_list_of_uints_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIST_OF_UINTS_TYPE;
    t->dataoffset=offsetof(dae_list_of_uints_type,data);
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=0;
    t->datamax=-1;
    ++t;
    t->name = "bool2_type";
    t->size = sizeof(dae_bool2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=2;
    t->datamax=2;
    ++t;
    t->name = "bool3_type";
    t->size = sizeof(dae_bool3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=3;
    t->datamax=3;
    ++t;
    t->name = "bool4_type";
    t->size = sizeof(dae_bool4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=4;
    t->datamax=4;
    ++t;
    t->name = "bool2x2_type";
    t->size = sizeof(dae_bool2x2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL2X2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=4;
    t->datamax=4;
    ++t;
    t->name = "bool2x3_type";
    t->size = sizeof(dae_bool2x3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL2X3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=6;
    t->datamax=6;
    ++t;
    t->name = "bool2x4_type";
    t->size = sizeof(dae_bool2x4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL2X4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=8;
    t->datamax=8;
    ++t;
    t->name = "bool3x2_type";
    t->size = sizeof(dae_bool3x2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL3X2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=6;
    t->datamax=6;
    ++t;
    t->name = "bool3x3_type";
    t->size = sizeof(dae_bool3x3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL3X3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=9;
    t->datamax=9;
    ++t;
    t->name = "bool3x4_type";
    t->size = sizeof(dae_bool3x4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL3X4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=12;
    t->datamax=12;
    ++t;
    t->name = "bool4x2_type";
    t->size = sizeof(dae_bool4x2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL4X2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=8;
    t->datamax=8;
    ++t;
    t->name = "bool4x3_type";
    t->size = sizeof(dae_bool4x3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL4X3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=12;
    t->datamax=12;
    ++t;
    t->name = "bool4x4_type";
    t->size = sizeof(dae_bool4x4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL4X4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=16;
    t->datamax=16;
    ++t;
    t->name = "float2_type";
    t->size = sizeof(dae_float2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=2;
    t->datamax=2;
    ++t;
    t->name = "float3_type";
    t->size = sizeof(dae_float3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=3;
    t->datamax=3;
    ++t;
    t->name = "float4_type";
    t->size = sizeof(dae_float4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=4;
    t->datamax=4;
    ++t;
    t->name = "float2x2_type";
    t->size = sizeof(dae_float2x2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT2X2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=4;
    t->datamax=4;
    ++t;
    t->name = "float2x3_type";
    t->size = sizeof(dae_float2x3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT2X3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=6;
    t->datamax=6;
    ++t;
    t->name = "float2x4_type";
    t->size = sizeof(dae_float2x4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT2X4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=8;
    t->datamax=8;
    ++t;
    t->name = "float3x2_type";
    t->size = sizeof(dae_float3x2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT3X2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=6;
    t->datamax=6;
    ++t;
    t->name = "float3x3_type";
    t->size = sizeof(dae_float3x3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT3X3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=9;
    t->datamax=9;
    ++t;
    t->name = "float3x4_type";
    t->size = sizeof(dae_float3x4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT3X4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=12;
    t->datamax=12;
    ++t;
    t->name = "float4x2_type";
    t->size = sizeof(dae_float4x2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT4X2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=8;
    t->datamax=8;
    ++t;
    t->name = "float4x3_type";
    t->size = sizeof(dae_float4x3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT4X3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=12;
    t->datamax=12;
    ++t;
    t->name = "float4x4_type";
    t->size = sizeof(dae_float4x4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT4X4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=16;
    t->datamax=16;
    ++t;
    t->name = "int2_type";
    t->size = sizeof(dae_int2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=2;
    t->datamax=2;
    ++t;
    t->name = "int3_type";
    t->size = sizeof(dae_int3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=3;
    t->datamax=3;
    ++t;
    t->name = "int4_type";
    t->size = sizeof(dae_int4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=4;
    t->datamax=4;
    ++t;
    t->name = "int2x2_type";
    t->size = sizeof(dae_int2x2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT2X2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=4;
    t->datamax=4;
    ++t;
    t->name = "int2x3_type";
    t->size = sizeof(dae_int2x3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT2X3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=6;
    t->datamax=6;
    ++t;
    t->name = "int2x4_type";
    t->size = sizeof(dae_int2x4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT2X4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=8;
    t->datamax=8;
    ++t;
    t->name = "int3x2_type";
    t->size = sizeof(dae_int3x2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT3X2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=6;
    t->datamax=6;
    ++t;
    t->name = "int3x3_type";
    t->size = sizeof(dae_int3x3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT3X3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=9;
    t->datamax=9;
    ++t;
    t->name = "int3x4_type";
    t->size = sizeof(dae_int3x4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT3X4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=12;
    t->datamax=12;
    ++t;
    t->name = "int4x2_type";
    t->size = sizeof(dae_int4x2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT4X2_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=8;
    t->datamax=8;
    ++t;
    t->name = "int4x3_type";
    t->size = sizeof(dae_int4x3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT4X3_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=12;
    t->datamax=12;
    ++t;
    t->name = "int4x4_type";
    t->size = sizeof(dae_int4x4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT4X4_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=16;
    t->datamax=16;
    ++t;
    t->name = "digits_type";
    t->size = sizeof(dae_digits_type);
    t->flags= 0;
    t->objtypeid = dae_ID_DIGITS_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT8;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "magnitude_type";
    t->size = sizeof(dae_magnitude_type);
    t->flags= 0;
    t->objtypeid = dae_ID_MAGNITUDE_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_INT16;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "morph_method_enum";
    t->size = sizeof(dae_morph_method_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_MORPH_METHOD_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "node_enum";
    t->size = sizeof(dae_node_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_NODE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "sampler_behavior_enum";
    t->size = sizeof(dae_sampler_behavior_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_SAMPLER_BEHAVIOR_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "urifragment_type";
    t->size = sizeof(dae_urifragment_type);
    t->flags= 0;
    t->objtypeid = dae_ID_URIFRAGMENT_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "up_axis_enum";
    t->size = sizeof(dae_up_axis_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_UP_AXIS_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "version_enum";
    t->size = sizeof(dae_version_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_VERSION_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "input_local_type";
    t->size = sizeof(dae_input_local_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INPUT_LOCAL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="semantic";
        at->offset=offsetof(dae_input_local_type,at_semantic);
        at->objtypeid=dae_ID_NMTOKEN;
        at->min=1;
        at->max=1;
        ++at;
        at->name="source";
        at->offset=offsetof(dae_input_local_type,at_source);
        at->objtypeid=dae_ID_URIFRAGMENT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "input_local_offset_type";
    t->size = sizeof(dae_input_local_offset_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INPUT_LOCAL_OFFSET_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="offset";
        at->offset=offsetof(dae_input_local_offset_type,at_offset);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="semantic";
        at->offset=offsetof(dae_input_local_offset_type,at_semantic);
        at->objtypeid=dae_ID_NMTOKEN;
        at->min=1;
        at->max=1;
        ++at;
        at->name="source";
        at->offset=offsetof(dae_input_local_offset_type,at_source);
        at->objtypeid=dae_ID_URIFRAGMENT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="set";
        at->offset=offsetof(dae_input_local_offset_type,at_set);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    ++t;
    t->name = "targetable_float_type";
    t->size = sizeof(dae_targetable_float_type);
    t->flags= 0;
    t->objtypeid = dae_ID_TARGETABLE_FLOAT_TYPE;
    t->dataoffset=offsetof(dae_targetable_float_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_targetable_float_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "targetable_float3_type";
    t->size = sizeof(dae_targetable_float3_type);
    t->flags= 0;
    t->objtypeid = dae_ID_TARGETABLE_FLOAT3_TYPE;
    t->dataoffset=offsetof(dae_targetable_float3_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=3;
    t->datamax=3;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_targetable_float3_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "token_array_type";
    t->size = sizeof(dae_token_array_type);
    t->flags= 0;
    t->objtypeid = dae_ID_TOKEN_ARRAY_TYPE;
    t->dataoffset=offsetof(dae_token_array_type,data);
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=0;
    t->datamax=-1;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_token_array_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_token_array_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_token_array_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "idref_array_type";
    t->size = sizeof(dae_idref_array_type);
    t->flags= 0;
    t->objtypeid = dae_ID_IDREF_ARRAY_TYPE;
    t->dataoffset=offsetof(dae_idref_array_type,data);
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=0;
    t->datamax=-1;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_idref_array_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_idref_array_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_idref_array_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "name_array_type";
    t->size = sizeof(dae_name_array_type);
    t->flags= 0;
    t->objtypeid = dae_ID_NAME_ARRAY_TYPE;
    t->dataoffset=offsetof(dae_name_array_type,data);
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=0;
    t->datamax=-1;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_name_array_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_name_array_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_name_array_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "bool_array_type";
    t->size = sizeof(dae_bool_array_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOOL_ARRAY_TYPE;
    t->dataoffset=offsetof(dae_bool_array_type,data);
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=0;
    t->datamax=-1;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_bool_array_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_bool_array_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_bool_array_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "float_array_type";
    t->size = sizeof(dae_float_array_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FLOAT_ARRAY_TYPE;
    t->dataoffset=offsetof(dae_float_array_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=0;
    t->datamax=-1;
    {
        static dae_obj_memberdef attribs[5];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_float_array_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_float_array_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_float_array_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="digits";
        at->offset=offsetof(dae_float_array_type,at_digits);
        at->objtypeid=dae_ID_DIGITS_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="6";
        ++at;
        at->name="magnitude";
        at->offset=offsetof(dae_float_array_type,at_magnitude);
        at->objtypeid=dae_ID_MAGNITUDE_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="38";
        ++at;
        t->attribs = attribs;
        t->numattribs = 5;
    }
    ++t;
    t->name = "int_array_type";
    t->size = sizeof(dae_int_array_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INT_ARRAY_TYPE;
    t->dataoffset=offsetof(dae_int_array_type,data);
    t->datatypeid=dae_NATIVE_INT32;
    t->datamin=0;
    t->datamax=-1;
    {
        static dae_obj_memberdef attribs[5];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_int_array_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_int_array_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_int_array_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="minInclusive";
        at->offset=offsetof(dae_int_array_type,at_minInclusive);
        at->objtypeid=dae_ID_INTEGER;
        at->min=0;
        at->max=1;
        at->deflt="-2147483648";
        ++at;
        at->name="maxInclusive";
        at->offset=offsetof(dae_int_array_type,at_maxInclusive);
        at->objtypeid=dae_ID_INTEGER;
        at->min=0;
        at->max=1;
        at->deflt="2147483647";
        ++at;
        t->attribs = attribs;
        t->numattribs = 5;
    }
    ++t;
    t->name = "param_type";
    t->size = sizeof(dae_param_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PARAM_TYPE;
    t->dataoffset=offsetof(dae_param_type,data);
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="name";
        at->offset=offsetof(dae_param_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_param_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="semantic";
        at->offset=offsetof(dae_param_type,at_semantic);
        at->objtypeid=dae_ID_NMTOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="type";
        at->offset=offsetof(dae_param_type,at_type);
        at->objtypeid=dae_ID_NMTOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    ++t;
    t->name = "accessor_type";
    t->size = sizeof(dae_accessor_type);
    t->flags= 0;
    t->objtypeid = dae_ID_ACCESSOR_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="count";
        at->offset=offsetof(dae_accessor_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="offset";
        at->offset=offsetof(dae_accessor_type,at_offset);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="source";
        at->offset=offsetof(dae_accessor_type,at_source);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="stride";
        at->offset=offsetof(dae_accessor_type,at_stride);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="param";
        el->offset=offsetof(dae_accessor_type,el_param);
        el->objtypeid=dae_ID_PARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "p_type";
    t->size = sizeof(dae_p_type);
    t->flags= 0;
    t->objtypeid = dae_ID_P_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=0;
    t->datamax=-1;
    ++t;
    t->name = "lookat_type";
    t->size = sizeof(dae_lookat_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LOOKAT_TYPE;
    t->dataoffset=offsetof(dae_lookat_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=9;
    t->datamax=9;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_lookat_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "matrix_type";
    t->size = sizeof(dae_matrix_type);
    t->flags= 0;
    t->objtypeid = dae_ID_MATRIX_TYPE;
    t->dataoffset=offsetof(dae_matrix_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=16;
    t->datamax=16;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_matrix_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "scale_type";
    t->size = sizeof(dae_scale_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SCALE_TYPE;
    t->dataoffset=offsetof(dae_scale_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=3;
    t->datamax=3;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_scale_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "skew_type";
    t->size = sizeof(dae_skew_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SKEW_TYPE;
    t->dataoffset=offsetof(dae_skew_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=7;
    t->datamax=7;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_skew_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "translate_type";
    t->size = sizeof(dae_translate_type);
    t->flags= 0;
    t->objtypeid = dae_ID_TRANSLATE_TYPE;
    t->dataoffset=offsetof(dae_translate_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=3;
    t->datamax=3;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_translate_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "image_source_type_hex";
    t->size = sizeof(dae_image_source_type_hex);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_SOURCE_TYPE_HEX;
    t->dataoffset=offsetof(dae_image_source_type_hex,data);
    t->datatypeid=dae_NATIVE_HEX8;
    t->datamin=0;
    t->datamax=-1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="format";
        at->offset=offsetof(dae_image_source_type_hex,at_format);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "image_mips_type";
    t->size = sizeof(dae_image_mips_type);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_MIPS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="levels";
        at->offset=offsetof(dae_image_mips_type,at_levels);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        at->name="auto_generate";
        at->offset=offsetof(dae_image_mips_type,at_auto_generate);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "image_face_enum";
    t->size = sizeof(dae_image_face_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_FACE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "image_format_hint_channels_enum";
    t->size = sizeof(dae_image_format_hint_channels_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_FORMAT_HINT_CHANNELS_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "image_format_hint_precision_enum";
    t->size = sizeof(dae_image_format_hint_precision_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_FORMAT_HINT_PRECISION_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "image_format_hint_range_enum";
    t->size = sizeof(dae_image_format_hint_range_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_FORMAT_HINT_RANGE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "channel_type";
    t->size = sizeof(dae_channel_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CHANNEL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="source";
        at->offset=offsetof(dae_channel_type,at_source);
        at->objtypeid=dae_ID_URIFRAGMENT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="target";
        at->offset=offsetof(dae_channel_type,at_target);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "sampler_type";
    t->size = sizeof(dae_sampler_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SAMPLER_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_sampler_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="pre_behavior";
        at->offset=offsetof(dae_sampler_type,at_pre_behavior);
        at->objtypeid=dae_ID_SAMPLER_BEHAVIOR_ENUM;
        at->min=0;
        at->max=1;
        ++at;
        at->name="post_behavior";
        at->offset=offsetof(dae_sampler_type,at_post_behavior);
        at->objtypeid=dae_ID_SAMPLER_BEHAVIOR_ENUM;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_sampler_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "altitude_mode_enum";
    t->size = sizeof(dae_altitude_mode_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_ALTITUDE_MODE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "technique_type";
    t->size = sizeof(dae_technique_type);
    t->flags= 0 | dae_XSD_ANY;
    t->objtypeid = dae_ID_TECHNIQUE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="profile";
        at->offset=offsetof(dae_technique_type,at_profile);
        at->objtypeid=dae_ID_NMTOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "extra_type";
    t->size = sizeof(dae_extra_type);
    t->flags= 0;
    t->objtypeid = dae_ID_EXTRA_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_extra_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_extra_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="type";
        at->offset=offsetof(dae_extra_type,at_type);
        at->objtypeid=dae_ID_NMTOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_extra_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_extra_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "polylist_type";
    t->size = sizeof(dae_polylist_type);
    t->flags= 0;
    t->objtypeid = dae_ID_POLYLIST_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="name";
        at->offset=offsetof(dae_polylist_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_polylist_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="material";
        at->offset=offsetof(dae_polylist_type,at_material);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_polylist_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="vcount";
        el->offset=offsetof(dae_polylist_type,el_vcount);
        el->objtypeid=dae_ID_LIST_OF_UINTS_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_polylist_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_polylist_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "polygons_type";
    t->size = sizeof(dae_polygons_type);
    t->flags= 0;
    t->objtypeid = dae_ID_POLYGONS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="name";
        at->offset=offsetof(dae_polygons_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_polygons_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="material";
        at->offset=offsetof(dae_polygons_type,at_material);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_polygons_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_polygons_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="ph";
        el->offset=offsetof(dae_polygons_type,el_ph);
        el->objtypeid=dae_ID_POLYGONS_TYPE_PH;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_polygons_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "polygons_type_ph";
    t->size = sizeof(dae_polygons_type_ph);
    t->flags= 0;
    t->objtypeid = dae_ID_POLYGONS_TYPE_PH;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="p";
        el->offset=offsetof(dae_polygons_type_ph,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="h";
        el->offset=offsetof(dae_polygons_type_ph,el_h);
        el->objtypeid=dae_ID_LIST_OF_UINTS_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "linestrips_type";
    t->size = sizeof(dae_linestrips_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LINESTRIPS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="name";
        at->offset=offsetof(dae_linestrips_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_linestrips_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="material";
        at->offset=offsetof(dae_linestrips_type,at_material);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_linestrips_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_linestrips_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_linestrips_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "tristrips_type";
    t->size = sizeof(dae_tristrips_type);
    t->flags= 0;
    t->objtypeid = dae_ID_TRISTRIPS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="name";
        at->offset=offsetof(dae_tristrips_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_tristrips_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="material";
        at->offset=offsetof(dae_tristrips_type,at_material);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_tristrips_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_tristrips_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_tristrips_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "lines_type";
    t->size = sizeof(dae_lines_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LINES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="name";
        at->offset=offsetof(dae_lines_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_lines_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="material";
        at->offset=offsetof(dae_lines_type,at_material);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_lines_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_lines_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_lines_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "triangles_type";
    t->size = sizeof(dae_triangles_type);
    t->flags= 0;
    t->objtypeid = dae_ID_TRIANGLES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="name";
        at->offset=offsetof(dae_triangles_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_triangles_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="material";
        at->offset=offsetof(dae_triangles_type,at_material);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_triangles_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_triangles_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_triangles_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "trifans_type";
    t->size = sizeof(dae_trifans_type);
    t->flags= 0;
    t->objtypeid = dae_ID_TRIFANS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="name";
        at->offset=offsetof(dae_trifans_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_trifans_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="material";
        at->offset=offsetof(dae_trifans_type,at_material);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_trifans_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_trifans_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_trifans_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "vertices_type";
    t->size = sizeof(dae_vertices_type);
    t->flags= 0;
    t->objtypeid = dae_ID_VERTICES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_vertices_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_vertices_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_vertices_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_vertices_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "asset_type";
    t->size = sizeof(dae_asset_type);
    t->flags= 0;
    t->objtypeid = dae_ID_ASSET_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[11];
        dae_obj_memberdef* el = elems;
        el->name="contributor";
        el->offset=offsetof(dae_asset_type,el_contributor);
        el->objtypeid=dae_ID_ASSET_TYPE_CONTRIBUTOR;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="coverage";
        el->offset=offsetof(dae_asset_type,el_coverage);
        el->objtypeid=dae_ID_ASSET_TYPE_COVERAGE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="created";
        el->offset=offsetof(dae_asset_type,el_created);
        el->objtypeid=dae_ID_DATETIME;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="keywords";
        el->offset=offsetof(dae_asset_type,el_keywords);
        el->objtypeid=dae_ID_TOKEN;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="modified";
        el->offset=offsetof(dae_asset_type,el_modified);
        el->objtypeid=dae_ID_DATETIME;
        el->min=1;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="revision";
        el->offset=offsetof(dae_asset_type,el_revision);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="subject";
        el->offset=offsetof(dae_asset_type,el_subject);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="title";
        el->offset=offsetof(dae_asset_type,el_title);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=1;
        el->seq=7;
        ++el;
        el->name="unit";
        el->offset=offsetof(dae_asset_type,el_unit);
        el->objtypeid=dae_ID_ASSET_TYPE_UNIT;
        el->min=0;
        el->max=1;
        el->seq=8;
        ++el;
        el->name="up_axis";
        el->offset=offsetof(dae_asset_type,el_up_axis);
        el->objtypeid=dae_ID_UP_AXIS_ENUM;
        el->min=0;
        el->max=1;
        el->seq=9;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_asset_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=10;
        ++el;
        t->elems = elems;
        t->numelems = 11;
    }
    ++t;
    t->name = "asset_type_contributor";
    t->size = sizeof(dae_asset_type_contributor);
    t->flags= 0;
    t->objtypeid = dae_ID_ASSET_TYPE_CONTRIBUTOR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="author";
        el->offset=offsetof(dae_asset_type_contributor,el_author);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="author_email";
        el->offset=offsetof(dae_asset_type_contributor,el_author_email);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="author_website";
        el->offset=offsetof(dae_asset_type_contributor,el_author_website);
        el->objtypeid=dae_ID_ANYURI;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="authoring_tool";
        el->offset=offsetof(dae_asset_type_contributor,el_authoring_tool);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="comments";
        el->offset=offsetof(dae_asset_type_contributor,el_comments);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="copyright";
        el->offset=offsetof(dae_asset_type_contributor,el_copyright);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="source_data";
        el->offset=offsetof(dae_asset_type_contributor,el_source_data);
        el->objtypeid=dae_ID_ANYURI;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "asset_type_coverage";
    t->size = sizeof(dae_asset_type_coverage);
    t->flags= 0;
    t->objtypeid = dae_ID_ASSET_TYPE_COVERAGE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="geographic_location";
        el->offset=offsetof(dae_asset_type_coverage,el_geographic_location);
        el->objtypeid=dae_ID_ASSET_TYPE_COVERAGE_GEOGRAPHIC_LOCATION;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "asset_type_coverage_geographic_location";
    t->size = sizeof(dae_asset_type_coverage_geographic_location);
    t->flags= 0;
    t->objtypeid = dae_ID_ASSET_TYPE_COVERAGE_GEOGRAPHIC_LOCATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="longitude";
        el->offset=offsetof(dae_asset_type_coverage_geographic_location,el_longitude);
        el->objtypeid=dae_ID_FLOAT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="latitude";
        el->offset=offsetof(dae_asset_type_coverage_geographic_location,el_latitude);
        el->objtypeid=dae_ID_FLOAT;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="altitude";
        el->offset=offsetof(dae_asset_type_coverage_geographic_location,el_altitude);
        el->objtypeid=dae_ID_ASSET_TYPE_COVERAGE_GEOGRAPHIC_LOCATION_ALTITUDE;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "asset_type_coverage_geographic_location_altitude";
    t->size = sizeof(dae_asset_type_coverage_geographic_location_altitude);
    t->flags= 0;
    t->objtypeid = dae_ID_ASSET_TYPE_COVERAGE_GEOGRAPHIC_LOCATION_ALTITUDE;
    t->dataoffset=offsetof(dae_asset_type_coverage_geographic_location_altitude,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="mode";
        at->offset=offsetof(dae_asset_type_coverage_geographic_location_altitude,at_mode);
        at->objtypeid=dae_ID_ALTITUDE_MODE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="relativeToGround";
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "asset_type_unit";
    t->size = sizeof(dae_asset_type_unit);
    t->flags= 0;
    t->objtypeid = dae_ID_ASSET_TYPE_UNIT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="meter";
        at->offset=offsetof(dae_asset_type_unit,at_meter);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1.0";
        ++at;
        at->name="name";
        at->offset=offsetof(dae_asset_type_unit,at_name);
        at->objtypeid=dae_ID_NMTOKEN;
        at->min=0;
        at->max=1;
        at->deflt="meter";
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "image_type";
    t->size = sizeof(dae_image_type);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_image_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_image_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_image_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_image_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="renderable";
        el->offset=offsetof(dae_image_type,el_renderable);
        el->objtypeid=dae_ID_IMAGE_TYPE_RENDERABLE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="init_from";
        el->offset=offsetof(dae_image_type,el_init_from);
        el->objtypeid=dae_ID_IMAGE_TYPE_INIT_FROM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="create_2d";
        el->offset=offsetof(dae_image_type,el_create_2d);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_2D;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="create_3d";
        el->offset=offsetof(dae_image_type,el_create_3d);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_3D;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="create_cube";
        el->offset=offsetof(dae_image_type,el_create_cube);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_CUBE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_image_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "image_type_renderable";
    t->size = sizeof(dae_image_type_renderable);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_RENDERABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="share";
        at->offset=offsetof(dae_image_type_renderable,at_share);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "image_type_init_from";
    t->size = sizeof(dae_image_type_init_from);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_INIT_FROM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="mips_generate";
        at->offset=offsetof(dae_image_type_init_from,at_mips_generate);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="true";
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="ref";
        el->offset=offsetof(dae_image_type_init_from,el_ref);
        el->objtypeid=dae_ID_ANYURI;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="hex";
        el->offset=offsetof(dae_image_type_init_from,el_hex);
        el->objtypeid=dae_ID_IMAGE_SOURCE_TYPE_HEX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "image_type_create_2d";
    t->size = sizeof(dae_image_type_create_2d);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_2D;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="size_exact";
        el->offset=offsetof(dae_image_type_create_2d,el_size_exact);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_2D_SIZE_EXACT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="size_ratio";
        el->offset=offsetof(dae_image_type_create_2d,el_size_ratio);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_2D_SIZE_RATIO;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mips";
        el->offset=offsetof(dae_image_type_create_2d,el_mips);
        el->objtypeid=dae_ID_IMAGE_MIPS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="unnormalized";
        el->offset=offsetof(dae_image_type_create_2d,el_unnormalized);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_2D_UNNORMALIZED;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_image_type_create_2d,el_array);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_2D_ARRAY;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="format";
        el->offset=offsetof(dae_image_type_create_2d,el_format);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_2D_FORMAT;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="init_from";
        el->offset=offsetof(dae_image_type_create_2d,el_init_from);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_2D_INIT_FROM;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "image_type_create_2d_size_exact";
    t->size = sizeof(dae_image_type_create_2d_size_exact);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_2D_SIZE_EXACT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="width";
        at->offset=offsetof(dae_image_type_create_2d_size_exact,at_width);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        at->name="height";
        at->offset=offsetof(dae_image_type_create_2d_size_exact,at_height);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "image_type_create_2d_size_ratio";
    t->size = sizeof(dae_image_type_create_2d_size_ratio);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_2D_SIZE_RATIO;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="width";
        at->offset=offsetof(dae_image_type_create_2d_size_ratio,at_width);
        at->objtypeid=dae_ID_FLOAT;
        at->min=1;
        at->max=1;
        ++at;
        at->name="height";
        at->offset=offsetof(dae_image_type_create_2d_size_ratio,at_height);
        at->objtypeid=dae_ID_FLOAT;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "image_type_create_2d_unnormalized";
    t->size = sizeof(dae_image_type_create_2d_unnormalized);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_2D_UNNORMALIZED;
    t->datatypeid=dae_ID_INVALID;
    ++t;
    t->name = "image_type_create_2d_array";
    t->size = sizeof(dae_image_type_create_2d_array);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_2D_ARRAY;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="length";
        at->offset=offsetof(dae_image_type_create_2d_array,at_length);
        at->objtypeid=dae_ID_POSITIVEINTEGER;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "image_type_create_2d_format";
    t->size = sizeof(dae_image_type_create_2d_format);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_2D_FORMAT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="hint";
        el->offset=offsetof(dae_image_type_create_2d_format,el_hint);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_2D_FORMAT_HINT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="exact";
        el->offset=offsetof(dae_image_type_create_2d_format,el_exact);
        el->objtypeid=dae_ID_TOKEN;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "image_type_create_2d_format_hint";
    t->size = sizeof(dae_image_type_create_2d_format_hint);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_2D_FORMAT_HINT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="channels";
        at->offset=offsetof(dae_image_type_create_2d_format_hint,at_channels);
        at->objtypeid=dae_ID_IMAGE_FORMAT_HINT_CHANNELS_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        at->name="range";
        at->offset=offsetof(dae_image_type_create_2d_format_hint,at_range);
        at->objtypeid=dae_ID_IMAGE_FORMAT_HINT_RANGE_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        at->name="precision";
        at->offset=offsetof(dae_image_type_create_2d_format_hint,at_precision);
        at->objtypeid=dae_ID_IMAGE_FORMAT_HINT_PRECISION_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="DEFAULT";
        ++at;
        at->name="space";
        at->offset=offsetof(dae_image_type_create_2d_format_hint,at_space);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    ++t;
    t->name = "image_type_create_2d_init_from";
    t->size = sizeof(dae_image_type_create_2d_init_from);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_2D_INIT_FROM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="mip_index";
        at->offset=offsetof(dae_image_type_create_2d_init_from,at_mip_index);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        at->name="array_index";
        at->offset=offsetof(dae_image_type_create_2d_init_from,at_array_index);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="ref";
        el->offset=offsetof(dae_image_type_create_2d_init_from,el_ref);
        el->objtypeid=dae_ID_ANYURI;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="hex";
        el->offset=offsetof(dae_image_type_create_2d_init_from,el_hex);
        el->objtypeid=dae_ID_IMAGE_SOURCE_TYPE_HEX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "image_type_create_3d";
    t->size = sizeof(dae_image_type_create_3d);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_3D;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="size";
        el->offset=offsetof(dae_image_type_create_3d,el_size);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_3D_SIZE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mips";
        el->offset=offsetof(dae_image_type_create_3d,el_mips);
        el->objtypeid=dae_ID_IMAGE_MIPS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_image_type_create_3d,el_array);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_3D_ARRAY;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="format";
        el->offset=offsetof(dae_image_type_create_3d,el_format);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_3D_FORMAT;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="init_from";
        el->offset=offsetof(dae_image_type_create_3d,el_init_from);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_3D_INIT_FROM;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "image_type_create_3d_size";
    t->size = sizeof(dae_image_type_create_3d_size);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_3D_SIZE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="width";
        at->offset=offsetof(dae_image_type_create_3d_size,at_width);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        at->name="height";
        at->offset=offsetof(dae_image_type_create_3d_size,at_height);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        at->name="depth";
        at->offset=offsetof(dae_image_type_create_3d_size,at_depth);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "image_type_create_3d_array";
    t->size = sizeof(dae_image_type_create_3d_array);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_3D_ARRAY;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="length";
        at->offset=offsetof(dae_image_type_create_3d_array,at_length);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "image_type_create_3d_format";
    t->size = sizeof(dae_image_type_create_3d_format);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_3D_FORMAT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="hint";
        el->offset=offsetof(dae_image_type_create_3d_format,el_hint);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_3D_FORMAT_HINT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="exact";
        el->offset=offsetof(dae_image_type_create_3d_format,el_exact);
        el->objtypeid=dae_ID_TOKEN;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "image_type_create_3d_format_hint";
    t->size = sizeof(dae_image_type_create_3d_format_hint);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_3D_FORMAT_HINT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="channels";
        at->offset=offsetof(dae_image_type_create_3d_format_hint,at_channels);
        at->objtypeid=dae_ID_IMAGE_FORMAT_HINT_CHANNELS_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        at->name="range";
        at->offset=offsetof(dae_image_type_create_3d_format_hint,at_range);
        at->objtypeid=dae_ID_IMAGE_FORMAT_HINT_RANGE_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        at->name="precision";
        at->offset=offsetof(dae_image_type_create_3d_format_hint,at_precision);
        at->objtypeid=dae_ID_IMAGE_FORMAT_HINT_PRECISION_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="DEFAULT";
        ++at;
        at->name="space";
        at->offset=offsetof(dae_image_type_create_3d_format_hint,at_space);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    ++t;
    t->name = "image_type_create_3d_init_from";
    t->size = sizeof(dae_image_type_create_3d_init_from);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_3D_INIT_FROM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="depth";
        at->offset=offsetof(dae_image_type_create_3d_init_from,at_depth);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        at->name="mip_index";
        at->offset=offsetof(dae_image_type_create_3d_init_from,at_mip_index);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        at->name="array_index";
        at->offset=offsetof(dae_image_type_create_3d_init_from,at_array_index);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="ref";
        el->offset=offsetof(dae_image_type_create_3d_init_from,el_ref);
        el->objtypeid=dae_ID_ANYURI;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="hex";
        el->offset=offsetof(dae_image_type_create_3d_init_from,el_hex);
        el->objtypeid=dae_ID_IMAGE_SOURCE_TYPE_HEX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "image_type_create_cube";
    t->size = sizeof(dae_image_type_create_cube);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_CUBE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="size";
        el->offset=offsetof(dae_image_type_create_cube,el_size);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_CUBE_SIZE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mips";
        el->offset=offsetof(dae_image_type_create_cube,el_mips);
        el->objtypeid=dae_ID_IMAGE_MIPS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_image_type_create_cube,el_array);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_CUBE_ARRAY;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="format";
        el->offset=offsetof(dae_image_type_create_cube,el_format);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_CUBE_FORMAT;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="init_from";
        el->offset=offsetof(dae_image_type_create_cube,el_init_from);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_CUBE_INIT_FROM;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "image_type_create_cube_size";
    t->size = sizeof(dae_image_type_create_cube_size);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_CUBE_SIZE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="width";
        at->offset=offsetof(dae_image_type_create_cube_size,at_width);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "image_type_create_cube_array";
    t->size = sizeof(dae_image_type_create_cube_array);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_CUBE_ARRAY;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="length";
        at->offset=offsetof(dae_image_type_create_cube_array,at_length);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "image_type_create_cube_format";
    t->size = sizeof(dae_image_type_create_cube_format);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_CUBE_FORMAT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="hint";
        el->offset=offsetof(dae_image_type_create_cube_format,el_hint);
        el->objtypeid=dae_ID_IMAGE_TYPE_CREATE_CUBE_FORMAT_HINT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="exact";
        el->offset=offsetof(dae_image_type_create_cube_format,el_exact);
        el->objtypeid=dae_ID_TOKEN;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "image_type_create_cube_format_hint";
    t->size = sizeof(dae_image_type_create_cube_format_hint);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_CUBE_FORMAT_HINT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="channels";
        at->offset=offsetof(dae_image_type_create_cube_format_hint,at_channels);
        at->objtypeid=dae_ID_IMAGE_FORMAT_HINT_CHANNELS_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        at->name="range";
        at->offset=offsetof(dae_image_type_create_cube_format_hint,at_range);
        at->objtypeid=dae_ID_IMAGE_FORMAT_HINT_RANGE_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        at->name="precision";
        at->offset=offsetof(dae_image_type_create_cube_format_hint,at_precision);
        at->objtypeid=dae_ID_IMAGE_FORMAT_HINT_PRECISION_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="DEFAULT";
        ++at;
        at->name="space";
        at->offset=offsetof(dae_image_type_create_cube_format_hint,at_space);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    ++t;
    t->name = "image_type_create_cube_init_from";
    t->size = sizeof(dae_image_type_create_cube_init_from);
    t->flags= 0;
    t->objtypeid = dae_ID_IMAGE_TYPE_CREATE_CUBE_INIT_FROM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="face";
        at->offset=offsetof(dae_image_type_create_cube_init_from,at_face);
        at->objtypeid=dae_ID_IMAGE_FACE_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        at->name="mip_index";
        at->offset=offsetof(dae_image_type_create_cube_init_from,at_mip_index);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=1;
        at->max=1;
        ++at;
        at->name="array_index";
        at->offset=offsetof(dae_image_type_create_cube_init_from,at_array_index);
        at->objtypeid=dae_ID_UNSIGNEDINT;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="ref";
        el->offset=offsetof(dae_image_type_create_cube_init_from,el_ref);
        el->objtypeid=dae_ID_ANYURI;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="hex";
        el->offset=offsetof(dae_image_type_create_cube_init_from,el_hex);
        el->objtypeid=dae_ID_IMAGE_SOURCE_TYPE_HEX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "camera_type";
    t->size = sizeof(dae_camera_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CAMERA_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_camera_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_camera_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_camera_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="optics";
        el->offset=offsetof(dae_camera_type,el_optics);
        el->objtypeid=dae_ID_CAMERA_TYPE_OPTICS;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="imager";
        el->offset=offsetof(dae_camera_type,el_imager);
        el->objtypeid=dae_ID_CAMERA_TYPE_IMAGER;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_camera_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "camera_type_optics";
    t->size = sizeof(dae_camera_type_optics);
    t->flags= 0;
    t->objtypeid = dae_ID_CAMERA_TYPE_OPTICS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="technique_common";
        el->offset=offsetof(dae_camera_type_optics,el_technique_common);
        el->objtypeid=dae_ID_CAMERA_TYPE_OPTICS_TECHNIQUE_COMMON;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_camera_type_optics,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_camera_type_optics,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "camera_type_optics_technique_common";
    t->size = sizeof(dae_camera_type_optics_technique_common);
    t->flags= 0;
    t->objtypeid = dae_ID_CAMERA_TYPE_OPTICS_TECHNIQUE_COMMON;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="orthographic";
        el->offset=offsetof(dae_camera_type_optics_technique_common,el_orthographic);
        el->objtypeid=dae_ID_CAMERA_TYPE_OPTICS_TECHNIQUE_COMMON_ORTHOGRAPHIC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="perspective";
        el->offset=offsetof(dae_camera_type_optics_technique_common,el_perspective);
        el->objtypeid=dae_ID_CAMERA_TYPE_OPTICS_TECHNIQUE_COMMON_PERSPECTIVE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "camera_type_optics_technique_common_orthographic";
    t->size = sizeof(dae_camera_type_optics_technique_common_orthographic);
    t->flags= 0;
    t->objtypeid = dae_ID_CAMERA_TYPE_OPTICS_TECHNIQUE_COMMON_ORTHOGRAPHIC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="xmag";
        el->offset=offsetof(dae_camera_type_optics_technique_common_orthographic,el_xmag);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ymag";
        el->offset=offsetof(dae_camera_type_optics_technique_common_orthographic,el_ymag);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ymag";
        el->offset=offsetof(dae_camera_type_optics_technique_common_orthographic,el_ymag);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="aspect_ratio";
        el->offset=offsetof(dae_camera_type_optics_technique_common_orthographic,el_aspect_ratio);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="aspect_ratio";
        el->offset=offsetof(dae_camera_type_optics_technique_common_orthographic,el_aspect_ratio);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="znear";
        el->offset=offsetof(dae_camera_type_optics_technique_common_orthographic,el_znear);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="zfar";
        el->offset=offsetof(dae_camera_type_optics_technique_common_orthographic,el_zfar);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "camera_type_optics_technique_common_perspective";
    t->size = sizeof(dae_camera_type_optics_technique_common_perspective);
    t->flags= 0;
    t->objtypeid = dae_ID_CAMERA_TYPE_OPTICS_TECHNIQUE_COMMON_PERSPECTIVE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="xfov";
        el->offset=offsetof(dae_camera_type_optics_technique_common_perspective,el_xfov);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="yfov";
        el->offset=offsetof(dae_camera_type_optics_technique_common_perspective,el_yfov);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="yfov";
        el->offset=offsetof(dae_camera_type_optics_technique_common_perspective,el_yfov);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="aspect_ratio";
        el->offset=offsetof(dae_camera_type_optics_technique_common_perspective,el_aspect_ratio);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="aspect_ratio";
        el->offset=offsetof(dae_camera_type_optics_technique_common_perspective,el_aspect_ratio);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="znear";
        el->offset=offsetof(dae_camera_type_optics_technique_common_perspective,el_znear);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="zfar";
        el->offset=offsetof(dae_camera_type_optics_technique_common_perspective,el_zfar);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "camera_type_imager";
    t->size = sizeof(dae_camera_type_imager);
    t->flags= 0;
    t->objtypeid = dae_ID_CAMERA_TYPE_IMAGER;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="technique";
        el->offset=offsetof(dae_camera_type_imager,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_camera_type_imager,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "light_type";
    t->size = sizeof(dae_light_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIGHT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_light_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_light_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_light_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="technique_common";
        el->offset=offsetof(dae_light_type,el_technique_common);
        el->objtypeid=dae_ID_LIGHT_TYPE_TECHNIQUE_COMMON;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_light_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_light_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "light_type_technique_common";
    t->size = sizeof(dae_light_type_technique_common);
    t->flags= 0;
    t->objtypeid = dae_ID_LIGHT_TYPE_TECHNIQUE_COMMON;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="ambient";
        el->offset=offsetof(dae_light_type_technique_common,el_ambient);
        el->objtypeid=dae_ID_LIGHT_TYPE_TECHNIQUE_COMMON_AMBIENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="directional";
        el->offset=offsetof(dae_light_type_technique_common,el_directional);
        el->objtypeid=dae_ID_LIGHT_TYPE_TECHNIQUE_COMMON_DIRECTIONAL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point";
        el->offset=offsetof(dae_light_type_technique_common,el_point);
        el->objtypeid=dae_ID_LIGHT_TYPE_TECHNIQUE_COMMON_POINT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="spot";
        el->offset=offsetof(dae_light_type_technique_common,el_spot);
        el->objtypeid=dae_ID_LIGHT_TYPE_TECHNIQUE_COMMON_SPOT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "light_type_technique_common_ambient";
    t->size = sizeof(dae_light_type_technique_common_ambient);
    t->flags= 0;
    t->objtypeid = dae_ID_LIGHT_TYPE_TECHNIQUE_COMMON_AMBIENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="color";
        el->offset=offsetof(dae_light_type_technique_common_ambient,el_color);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "light_type_technique_common_directional";
    t->size = sizeof(dae_light_type_technique_common_directional);
    t->flags= 0;
    t->objtypeid = dae_ID_LIGHT_TYPE_TECHNIQUE_COMMON_DIRECTIONAL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="color";
        el->offset=offsetof(dae_light_type_technique_common_directional,el_color);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "light_type_technique_common_point";
    t->size = sizeof(dae_light_type_technique_common_point);
    t->flags= 0;
    t->objtypeid = dae_ID_LIGHT_TYPE_TECHNIQUE_COMMON_POINT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="color";
        el->offset=offsetof(dae_light_type_technique_common_point,el_color);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="constant_attenuation";
        el->offset=offsetof(dae_light_type_technique_common_point,el_constant_attenuation);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="linear_attenuation";
        el->offset=offsetof(dae_light_type_technique_common_point,el_linear_attenuation);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="quadratic_attenuation";
        el->offset=offsetof(dae_light_type_technique_common_point,el_quadratic_attenuation);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "light_type_technique_common_spot";
    t->size = sizeof(dae_light_type_technique_common_spot);
    t->flags= 0;
    t->objtypeid = dae_ID_LIGHT_TYPE_TECHNIQUE_COMMON_SPOT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[6];
        dae_obj_memberdef* el = elems;
        el->name="color";
        el->offset=offsetof(dae_light_type_technique_common_spot,el_color);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="constant_attenuation";
        el->offset=offsetof(dae_light_type_technique_common_spot,el_constant_attenuation);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="linear_attenuation";
        el->offset=offsetof(dae_light_type_technique_common_spot,el_linear_attenuation);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="quadratic_attenuation";
        el->offset=offsetof(dae_light_type_technique_common_spot,el_quadratic_attenuation);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="falloff_angle";
        el->offset=offsetof(dae_light_type_technique_common_spot,el_falloff_angle);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="falloff_exponent";
        el->offset=offsetof(dae_light_type_technique_common_spot,el_falloff_exponent);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        t->elems = elems;
        t->numelems = 6;
    }
    ++t;
    t->name = "instance_with_extra_type";
    t->size = sizeof(dae_instance_with_extra_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_WITH_EXTRA_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_with_extra_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_with_extra_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_with_extra_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="extra";
        el->offset=offsetof(dae_instance_with_extra_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "instance_image_type";
    t->size = sizeof(dae_instance_image_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_IMAGE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_image_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_image_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_image_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "instance_camera_type";
    t->size = sizeof(dae_instance_camera_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_CAMERA_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_camera_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_camera_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_camera_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "instance_force_field_type";
    t->size = sizeof(dae_instance_force_field_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_FORCE_FIELD_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_force_field_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_force_field_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_force_field_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "instance_light_type";
    t->size = sizeof(dae_instance_light_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_LIGHT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_light_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_light_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_light_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "instance_material_type";
    t->size = sizeof(dae_instance_material_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_MATERIAL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="symbol";
        at->offset=offsetof(dae_instance_material_type,at_symbol);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        at->name="target";
        at->offset=offsetof(dae_instance_material_type,at_target);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_material_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_material_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="bind";
        el->offset=offsetof(dae_instance_material_type,el_bind);
        el->objtypeid=dae_ID_INSTANCE_MATERIAL_TYPE_BIND;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="bind_vertex_input";
        el->offset=offsetof(dae_instance_material_type,el_bind_vertex_input);
        el->objtypeid=dae_ID_INSTANCE_MATERIAL_TYPE_BIND_VERTEX_INPUT;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_instance_material_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "instance_material_type_bind";
    t->size = sizeof(dae_instance_material_type_bind);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_MATERIAL_TYPE_BIND;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="semantic";
        at->offset=offsetof(dae_instance_material_type_bind,at_semantic);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        at->name="target";
        at->offset=offsetof(dae_instance_material_type_bind,at_target);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "instance_material_type_bind_vertex_input";
    t->size = sizeof(dae_instance_material_type_bind_vertex_input);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_MATERIAL_TYPE_BIND_VERTEX_INPUT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="semantic";
        at->offset=offsetof(dae_instance_material_type_bind_vertex_input,at_semantic);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        at->name="input_semantic";
        at->offset=offsetof(dae_instance_material_type_bind_vertex_input,at_input_semantic);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        at->name="input_set";
        at->offset=offsetof(dae_instance_material_type_bind_vertex_input,at_input_set);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "bind_material_type";
    t->size = sizeof(dae_bind_material_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BIND_MATERIAL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="param";
        el->offset=offsetof(dae_bind_material_type,el_param);
        el->objtypeid=dae_ID_PARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="technique_common";
        el->offset=offsetof(dae_bind_material_type,el_technique_common);
        el->objtypeid=dae_ID_BIND_MATERIAL_TYPE_TECHNIQUE_COMMON;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_bind_material_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_bind_material_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "bind_material_type_technique_common";
    t->size = sizeof(dae_bind_material_type_technique_common);
    t->flags= 0;
    t->objtypeid = dae_ID_BIND_MATERIAL_TYPE_TECHNIQUE_COMMON;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="instance_material";
        el->offset=offsetof(dae_bind_material_type_technique_common,el_instance_material);
        el->objtypeid=dae_ID_INSTANCE_MATERIAL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "instance_controller_type";
    t->size = sizeof(dae_instance_controller_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_CONTROLLER_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_controller_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_controller_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_controller_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="skeleton";
        el->offset=offsetof(dae_instance_controller_type,el_skeleton);
        el->objtypeid=dae_ID_ANYURI;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="bind_material";
        el->offset=offsetof(dae_instance_controller_type,el_bind_material);
        el->objtypeid=dae_ID_BIND_MATERIAL_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_instance_controller_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "instance_geometry_type";
    t->size = sizeof(dae_instance_geometry_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_GEOMETRY_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_geometry_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_geometry_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_geometry_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="bind_material";
        el->offset=offsetof(dae_instance_geometry_type,el_bind_material);
        el->objtypeid=dae_ID_BIND_MATERIAL_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_instance_geometry_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "instance_node_type";
    t->size = sizeof(dae_instance_node_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_NODE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="proxy";
        at->offset=offsetof(dae_instance_node_type,at_proxy);
        at->objtypeid=dae_ID_ANYURI;
        at->min=0;
        at->max=1;
        ++at;
        at->name="url";
        at->offset=offsetof(dae_instance_node_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_node_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_node_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="extra";
        el->offset=offsetof(dae_instance_node_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "instance_physics_material_type";
    t->size = sizeof(dae_instance_physics_material_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_PHYSICS_MATERIAL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_physics_material_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_physics_material_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_physics_material_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "instance_rigid_constraint_type";
    t->size = sizeof(dae_instance_rigid_constraint_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_RIGID_CONSTRAINT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="constraint";
        at->offset=offsetof(dae_instance_rigid_constraint_type,at_constraint);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_rigid_constraint_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_rigid_constraint_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="extra";
        el->offset=offsetof(dae_instance_rigid_constraint_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "library_cameras_type";
    t->size = sizeof(dae_library_cameras_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_CAMERAS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_cameras_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_cameras_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_cameras_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="camera";
        el->offset=offsetof(dae_library_cameras_type,el_camera);
        el->objtypeid=dae_ID_CAMERA_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_cameras_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "library_images_type";
    t->size = sizeof(dae_library_images_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_IMAGES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_images_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_images_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_images_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="image";
        el->offset=offsetof(dae_library_images_type,el_image);
        el->objtypeid=dae_ID_IMAGE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_images_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "library_lights_type";
    t->size = sizeof(dae_library_lights_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_LIGHTS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_lights_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_lights_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_lights_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light";
        el->offset=offsetof(dae_library_lights_type,el_light);
        el->objtypeid=dae_ID_LIGHT_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_lights_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "fx_color_type";
    t->size = sizeof(dae_fx_color_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COLOR_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=4;
    t->datamax=4;
    ++t;
    t->name = "fx_opaque_enum";
    t->size = sizeof(dae_fx_opaque_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_OPAQUE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "fx_sampler_wrap_enum";
    t->size = sizeof(dae_fx_sampler_wrap_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLER_WRAP_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "fx_sampler_min_filter_enum";
    t->size = sizeof(dae_fx_sampler_min_filter_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLER_MIN_FILTER_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "fx_sampler_mag_filter_enum";
    t->size = sizeof(dae_fx_sampler_mag_filter_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLER_MAG_FILTER_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "fx_sampler_mip_filter_enum";
    t->size = sizeof(dae_fx_sampler_mip_filter_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLER_MIP_FILTER_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "fx_setparam_group_sampler_states";
    t->size = sizeof(dae_fx_setparam_group_sampler_states);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SETPARAM_GROUP_SAMPLER_STATES;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[12];
        dae_obj_memberdef* el = elems;
        el->name="wrap_s";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_wrap_s);
        el->objtypeid=dae_ID_FX_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="wrap_t";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_wrap_t);
        el->objtypeid=dae_ID_FX_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="wrap_p";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_wrap_p);
        el->objtypeid=dae_ID_FX_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="minfilter";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_minfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MIN_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="magfilter";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_magfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MAG_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="mipfilter";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_mipfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MIP_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="border_color";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_border_color);
        el->objtypeid=dae_ID_FX_COLOR_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="mip_max_level";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_mip_max_level);
        el->objtypeid=dae_ID_UNSIGNEDBYTE;
        el->min=0;
        el->max=1;
        el->seq=7;
        ++el;
        el->name="mip_min_level";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_mip_min_level);
        el->objtypeid=dae_ID_UNSIGNEDBYTE;
        el->min=0;
        el->max=1;
        el->seq=8;
        ++el;
        el->name="mip_bias";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_mip_bias);
        el->objtypeid=dae_ID_FLOAT;
        el->min=0;
        el->max=1;
        el->seq=9;
        ++el;
        el->name="max_anisotropy";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_max_anisotropy);
        el->objtypeid=dae_ID_UNSIGNEDINT;
        el->min=0;
        el->max=1;
        el->seq=10;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_fx_setparam_group_sampler_states,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=11;
        ++el;
        t->elems = elems;
        t->numelems = 12;
    }
    ++t;
    t->name = "instance_effect_type";
    t->size = sizeof(dae_instance_effect_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_EFFECT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_effect_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_effect_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_effect_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="technique_hint";
        el->offset=offsetof(dae_instance_effect_type,el_technique_hint);
        el->objtypeid=dae_ID_INSTANCE_EFFECT_TYPE_TECHNIQUE_HINT;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="setparam";
        el->offset=offsetof(dae_instance_effect_type,el_setparam);
        el->objtypeid=dae_ID_INSTANCE_EFFECT_TYPE_SETPARAM;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_instance_effect_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "instance_effect_type_technique_hint";
    t->size = sizeof(dae_instance_effect_type_technique_hint);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_EFFECT_TYPE_TECHNIQUE_HINT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="platform";
        at->offset=offsetof(dae_instance_effect_type_technique_hint,at_platform);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="profile";
        at->offset=offsetof(dae_instance_effect_type_technique_hint,at_profile);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="ref";
        at->offset=offsetof(dae_instance_effect_type_technique_hint,at_ref);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "instance_effect_type_setparam";
    t->size = sizeof(dae_instance_effect_type_setparam);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_EFFECT_TYPE_SETPARAM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_instance_effect_type_setparam,at_ref);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[27];
        dae_obj_memberdef* el = elems;
        el->name="bool";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x1";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x3";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x4";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x1";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x2";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x4";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x1";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x2";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x3";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_enum);
        el->objtypeid=dae_ID_STRING;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler_image";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_sampler_image);
        el->objtypeid=dae_ID_INSTANCE_IMAGE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler_states";
        el->offset=offsetof(dae_instance_effect_type_setparam,el_sampler_states);
        el->objtypeid=dae_ID_FX_SETPARAM_GROUP_SAMPLER_STATES;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 27;
    }
    ++t;
    t->name = "material_type";
    t->size = sizeof(dae_material_type);
    t->flags= 0;
    t->objtypeid = dae_ID_MATERIAL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_material_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_material_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_material_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="instance_effect";
        el->offset=offsetof(dae_material_type,el_instance_effect);
        el->objtypeid=dae_ID_INSTANCE_EFFECT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_material_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "library_materials_type";
    t->size = sizeof(dae_library_materials_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_MATERIALS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_materials_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_materials_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_materials_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material";
        el->offset=offsetof(dae_library_materials_type,el_material);
        el->objtypeid=dae_ID_MATERIAL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_materials_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "fx_sampler_type";
    t->size = sizeof(dae_fx_sampler_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLER_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[13];
        dae_obj_memberdef* el = elems;
        el->name="instance_image";
        el->offset=offsetof(dae_fx_sampler_type,el_instance_image);
        el->objtypeid=dae_ID_INSTANCE_IMAGE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="wrap_s";
        el->offset=offsetof(dae_fx_sampler_type,el_wrap_s);
        el->objtypeid=dae_ID_FX_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="wrap_t";
        el->offset=offsetof(dae_fx_sampler_type,el_wrap_t);
        el->objtypeid=dae_ID_FX_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="wrap_p";
        el->offset=offsetof(dae_fx_sampler_type,el_wrap_p);
        el->objtypeid=dae_ID_FX_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="minfilter";
        el->offset=offsetof(dae_fx_sampler_type,el_minfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MIN_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="magfilter";
        el->offset=offsetof(dae_fx_sampler_type,el_magfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MAG_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="mipfilter";
        el->offset=offsetof(dae_fx_sampler_type,el_mipfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MIP_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="border_color";
        el->offset=offsetof(dae_fx_sampler_type,el_border_color);
        el->objtypeid=dae_ID_FX_COLOR_TYPE;
        el->min=0;
        el->max=1;
        el->seq=7;
        ++el;
        el->name="mip_max_level";
        el->offset=offsetof(dae_fx_sampler_type,el_mip_max_level);
        el->objtypeid=dae_ID_UNSIGNEDBYTE;
        el->min=0;
        el->max=1;
        el->seq=8;
        ++el;
        el->name="mip_min_level";
        el->offset=offsetof(dae_fx_sampler_type,el_mip_min_level);
        el->objtypeid=dae_ID_UNSIGNEDBYTE;
        el->min=0;
        el->max=1;
        el->seq=9;
        ++el;
        el->name="mip_bias";
        el->offset=offsetof(dae_fx_sampler_type,el_mip_bias);
        el->objtypeid=dae_ID_FLOAT;
        el->min=0;
        el->max=1;
        el->seq=10;
        ++el;
        el->name="max_anisotropy";
        el->offset=offsetof(dae_fx_sampler_type,el_max_anisotropy);
        el->objtypeid=dae_ID_UNSIGNEDINT;
        el->min=0;
        el->max=1;
        el->seq=11;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_fx_sampler_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=12;
        ++el;
        t->elems = elems;
        t->numelems = 13;
    }
    ++t;
    t->name = "fx_sampler1D_type";
    t->size = sizeof(dae_fx_sampler1D_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLER1D_TYPE;
    t->datatypeid=dae_ID_INVALID;
    ++t;
    t->name = "fx_sampler2D_type";
    t->size = sizeof(dae_fx_sampler2D_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLER2D_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[14];
        dae_obj_memberdef* el = elems;
        el->name="instance_image";
        el->offset=offsetof(dae_fx_sampler2D_type,el_instance_image);
        el->objtypeid=dae_ID_INSTANCE_IMAGE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="wrap_s";
        el->offset=offsetof(dae_fx_sampler2D_type,el_wrap_s);
        el->objtypeid=dae_ID_FX_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="wrap_t";
        el->offset=offsetof(dae_fx_sampler2D_type,el_wrap_t);
        el->objtypeid=dae_ID_FX_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="wrap_p";
        el->offset=offsetof(dae_fx_sampler2D_type,el_wrap_p);
        el->objtypeid=dae_ID_FX_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="minfilter";
        el->offset=offsetof(dae_fx_sampler2D_type,el_minfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MIN_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="magfilter";
        el->offset=offsetof(dae_fx_sampler2D_type,el_magfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MAG_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="mipfilter";
        el->offset=offsetof(dae_fx_sampler2D_type,el_mipfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MIP_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="border_color";
        el->offset=offsetof(dae_fx_sampler2D_type,el_border_color);
        el->objtypeid=dae_ID_FX_COLOR_TYPE;
        el->min=0;
        el->max=1;
        el->seq=7;
        ++el;
        el->name="mip_max_level";
        el->offset=offsetof(dae_fx_sampler2D_type,el_mip_max_level);
        el->objtypeid=dae_ID_UNSIGNEDBYTE;
        el->min=0;
        el->max=1;
        el->seq=8;
        ++el;
        el->name="mip_min_level";
        el->offset=offsetof(dae_fx_sampler2D_type,el_mip_min_level);
        el->objtypeid=dae_ID_UNSIGNEDBYTE;
        el->min=0;
        el->max=1;
        el->seq=9;
        ++el;
        el->name="mip_bias";
        el->offset=offsetof(dae_fx_sampler2D_type,el_mip_bias);
        el->objtypeid=dae_ID_FLOAT;
        el->min=0;
        el->max=1;
        el->seq=10;
        ++el;
        el->name="max_anisotropy";
        el->offset=offsetof(dae_fx_sampler2D_type,el_max_anisotropy);
        el->objtypeid=dae_ID_UNSIGNEDINT;
        el->min=0;
        el->max=1;
        el->seq=11;
        ++el;
        el->name="source";
        el->offset=offsetof(dae_fx_sampler2D_type,el_source);
        el->objtypeid=dae_ID_NCNAME;
        el->min=0;
        el->max=1;
        el->seq=12;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_fx_sampler2D_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=12;
        ++el;
        t->elems = elems;
        t->numelems = 14;
    }
    ++t;
    t->name = "fx_sampler3D_type";
    t->size = sizeof(dae_fx_sampler3D_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLER3D_TYPE;
    t->datatypeid=dae_ID_INVALID;
    ++t;
    t->name = "fx_samplerCUBE_type";
    t->size = sizeof(dae_fx_samplerCUBE_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLERCUBE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    ++t;
    t->name = "fx_samplerRECT_type";
    t->size = sizeof(dae_fx_samplerRECT_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLERRECT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    ++t;
    t->name = "fx_samplerDEPTH_type";
    t->size = sizeof(dae_fx_samplerDEPTH_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SAMPLERDEPTH_TYPE;
    t->datatypeid=dae_ID_INVALID;
    ++t;
    t->name = "fx_modifier_enum";
    t->size = sizeof(dae_fx_modifier_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_MODIFIER_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "fx_rendertarget_type";
    t->size = sizeof(dae_fx_rendertarget_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_RENDERTARGET_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_fx_rendertarget_type,at_index);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="mip";
        at->offset=offsetof(dae_fx_rendertarget_type,at_mip);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="face";
        at->offset=offsetof(dae_fx_rendertarget_type,at_face);
        at->objtypeid=dae_ID_IMAGE_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="POSITIVE_X";
        ++at;
        at->name="slice";
        at->offset=offsetof(dae_fx_rendertarget_type,at_slice);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="param";
        el->offset=offsetof(dae_fx_rendertarget_type,el_param);
        el->objtypeid=dae_ID_FX_RENDERTARGET_TYPE_PARAM;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="instance_image";
        el->offset=offsetof(dae_fx_rendertarget_type,el_instance_image);
        el->objtypeid=dae_ID_INSTANCE_IMAGE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "fx_rendertarget_type_param";
    t->size = sizeof(dae_fx_rendertarget_type_param);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_RENDERTARGET_TYPE_PARAM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_fx_rendertarget_type_param,at_ref);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_colortarget_type";
    t->size = sizeof(dae_fx_colortarget_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COLORTARGET_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_fx_colortarget_type,at_index);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="mip";
        at->offset=offsetof(dae_fx_colortarget_type,at_mip);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="face";
        at->offset=offsetof(dae_fx_colortarget_type,at_face);
        at->objtypeid=dae_ID_IMAGE_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="POSITIVE_X";
        ++at;
        at->name="slice";
        at->offset=offsetof(dae_fx_colortarget_type,at_slice);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="param";
        el->offset=offsetof(dae_fx_colortarget_type,el_param);
        el->objtypeid=dae_ID_FX_RENDERTARGET_TYPE_PARAM;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="instance_image";
        el->offset=offsetof(dae_fx_colortarget_type,el_instance_image);
        el->objtypeid=dae_ID_INSTANCE_IMAGE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "fx_depthtarget_type";
    t->size = sizeof(dae_fx_depthtarget_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_DEPTHTARGET_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_fx_depthtarget_type,at_index);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="mip";
        at->offset=offsetof(dae_fx_depthtarget_type,at_mip);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="face";
        at->offset=offsetof(dae_fx_depthtarget_type,at_face);
        at->objtypeid=dae_ID_IMAGE_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="POSITIVE_X";
        ++at;
        at->name="slice";
        at->offset=offsetof(dae_fx_depthtarget_type,at_slice);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="param";
        el->offset=offsetof(dae_fx_depthtarget_type,el_param);
        el->objtypeid=dae_ID_FX_RENDERTARGET_TYPE_PARAM;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="instance_image";
        el->offset=offsetof(dae_fx_depthtarget_type,el_instance_image);
        el->objtypeid=dae_ID_INSTANCE_IMAGE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "fx_stenciltarget_type";
    t->size = sizeof(dae_fx_stenciltarget_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_STENCILTARGET_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_fx_stenciltarget_type,at_index);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="mip";
        at->offset=offsetof(dae_fx_stenciltarget_type,at_mip);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="face";
        at->offset=offsetof(dae_fx_stenciltarget_type,at_face);
        at->objtypeid=dae_ID_IMAGE_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="POSITIVE_X";
        ++at;
        at->name="slice";
        at->offset=offsetof(dae_fx_stenciltarget_type,at_slice);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="param";
        el->offset=offsetof(dae_fx_stenciltarget_type,el_param);
        el->objtypeid=dae_ID_FX_RENDERTARGET_TYPE_PARAM;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="instance_image";
        el->offset=offsetof(dae_fx_stenciltarget_type,el_instance_image);
        el->objtypeid=dae_ID_INSTANCE_IMAGE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "fx_clearcolor_type";
    t->size = sizeof(dae_fx_clearcolor_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_CLEARCOLOR_TYPE;
    t->dataoffset=offsetof(dae_fx_clearcolor_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=4;
    t->datamax=4;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_fx_clearcolor_type,at_index);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_cleardepth_type";
    t->size = sizeof(dae_fx_cleardepth_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_CLEARDEPTH_TYPE;
    t->dataoffset=offsetof(dae_fx_cleardepth_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_fx_cleardepth_type,at_index);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_clearstencil_type";
    t->size = sizeof(dae_fx_clearstencil_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_CLEARSTENCIL_TYPE;
    t->dataoffset=offsetof(dae_fx_clearstencil_type,data);
    t->datatypeid=dae_NATIVE_INT8;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_fx_clearstencil_type,at_index);
        at->objtypeid=dae_ID_NONNEGATIVEINTEGER;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_draw_type";
    t->size = sizeof(dae_fx_draw_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_DRAW_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "fx_pipeline_stage_enum";
    t->size = sizeof(dae_fx_pipeline_stage_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_PIPELINE_STAGE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "fx_annotate_type";
    t->size = sizeof(dae_fx_annotate_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_ANNOTATE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="name";
        at->offset=offsetof(dae_fx_annotate_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[16];
        dae_obj_memberdef* el = elems;
        el->name="bool";
        el->offset=offsetof(dae_fx_annotate_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_fx_annotate_type,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_fx_annotate_type,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_fx_annotate_type,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_fx_annotate_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_fx_annotate_type,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_fx_annotate_type,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_fx_annotate_type,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_fx_annotate_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_fx_annotate_type,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_fx_annotate_type,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_fx_annotate_type,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_fx_annotate_type,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_fx_annotate_type,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_fx_annotate_type,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="string";
        el->offset=offsetof(dae_fx_annotate_type,el_string);
        el->objtypeid=dae_ID_STRING;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 16;
    }
    ++t;
    t->name = "fx_newparam_type";
    t->size = sizeof(dae_fx_newparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_NEWPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_fx_newparam_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[34];
        dae_obj_memberdef* el = elems;
        el->name="annotate";
        el->offset=offsetof(dae_fx_newparam_type,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="semantic";
        el->offset=offsetof(dae_fx_newparam_type,el_semantic);
        el->objtypeid=dae_ID_NCNAME;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="modifier";
        el->offset=offsetof(dae_fx_newparam_type,el_modifier);
        el->objtypeid=dae_ID_FX_MODIFIER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_fx_newparam_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_fx_newparam_type,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_fx_newparam_type,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_fx_newparam_type,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_fx_newparam_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_fx_newparam_type,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_fx_newparam_type,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_fx_newparam_type,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_fx_newparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_fx_newparam_type,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_fx_newparam_type,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_fx_newparam_type,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x1";
        el->offset=offsetof(dae_fx_newparam_type,el_float2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_fx_newparam_type,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x3";
        el->offset=offsetof(dae_fx_newparam_type,el_float2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x4";
        el->offset=offsetof(dae_fx_newparam_type,el_float2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x1";
        el->offset=offsetof(dae_fx_newparam_type,el_float3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x2";
        el->offset=offsetof(dae_fx_newparam_type,el_float3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_fx_newparam_type,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x4";
        el->offset=offsetof(dae_fx_newparam_type,el_float3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x1";
        el->offset=offsetof(dae_fx_newparam_type,el_float4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x2";
        el->offset=offsetof(dae_fx_newparam_type,el_float4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x3";
        el->offset=offsetof(dae_fx_newparam_type,el_float4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_fx_newparam_type,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler1D";
        el->offset=offsetof(dae_fx_newparam_type,el_sampler1D);
        el->objtypeid=dae_ID_FX_SAMPLER1D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_fx_newparam_type,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_fx_newparam_type,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_fx_newparam_type,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerRECT";
        el->offset=offsetof(dae_fx_newparam_type,el_samplerRECT);
        el->objtypeid=dae_ID_FX_SAMPLERRECT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_fx_newparam_type,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_fx_newparam_type,el_enum);
        el->objtypeid=dae_ID_STRING;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 34;
    }
    ++t;
    t->name = "fx_include_type";
    t->size = sizeof(dae_fx_include_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_INCLUDE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_fx_include_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="url";
        at->offset=offsetof(dae_fx_include_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "fx_code_type";
    t->size = sizeof(dae_fx_code_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_CODE_TYPE;
    t->dataoffset=offsetof(dae_fx_code_type,data);
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_fx_code_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_sources_type";
    t->size = sizeof(dae_fx_sources_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SOURCES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="inline";
        el->offset=offsetof(dae_fx_sources_type,el_inline);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="import";
        el->offset=offsetof(dae_fx_sources_type,el_import);
        el->objtypeid=dae_ID_FX_SOURCES_TYPE_IMPORT;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "fx_sources_type_import";
    t->size = sizeof(dae_fx_sources_type_import);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SOURCES_TYPE_IMPORT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_fx_sources_type_import,at_ref);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_target_type";
    t->size = sizeof(dae_fx_target_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_TARGET_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="platform";
        at->offset=offsetof(dae_fx_target_type,at_platform);
        at->objtypeid=dae_ID_STRING;
        at->min=1;
        at->max=1;
        ++at;
        at->name="target";
        at->offset=offsetof(dae_fx_target_type,at_target);
        at->objtypeid=dae_ID_STRING;
        at->min=0;
        at->max=1;
        ++at;
        at->name="options";
        at->offset=offsetof(dae_fx_target_type,at_options);
        at->objtypeid=dae_ID_STRING;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="binary";
        el->offset=offsetof(dae_fx_target_type,el_binary);
        el->objtypeid=dae_ID_FX_TARGET_TYPE_BINARY;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "fx_target_type_binary";
    t->size = sizeof(dae_fx_target_type_binary);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_TARGET_TYPE_BINARY;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="ref";
        el->offset=offsetof(dae_fx_target_type_binary,el_ref);
        el->objtypeid=dae_ID_ANYURI;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="hex";
        el->offset=offsetof(dae_fx_target_type_binary,el_hex);
        el->objtypeid=dae_ID_FX_TARGET_TYPE_BINARY_HEX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "fx_target_type_binary_hex";
    t->size = sizeof(dae_fx_target_type_binary_hex);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_TARGET_TYPE_BINARY_HEX;
    t->dataoffset=offsetof(dae_fx_target_type_binary_hex,data);
    t->datatypeid=dae_NATIVE_HEX8;
    t->datamin=0;
    t->datamax=-1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="format";
        at->offset=offsetof(dae_fx_target_type_binary_hex,at_format);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_common_float_or_param_type";
    t->size = sizeof(dae_fx_common_float_or_param_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="float";
        el->offset=offsetof(dae_fx_common_float_or_param_type,el_float);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE_FLOAT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_fx_common_float_or_param_type,el_param);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE_PARAM;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "fx_common_float_or_param_type_float";
    t->size = sizeof(dae_fx_common_float_or_param_type_float);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE_FLOAT;
    t->dataoffset=offsetof(dae_fx_common_float_or_param_type_float,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_fx_common_float_or_param_type_float,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_common_float_or_param_type_param";
    t->size = sizeof(dae_fx_common_float_or_param_type_param);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE_PARAM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_fx_common_float_or_param_type_param,at_ref);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_common_color_or_texture_type";
    t->size = sizeof(dae_fx_common_color_or_texture_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="color";
        el->offset=offsetof(dae_fx_common_color_or_texture_type,el_color);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE_COLOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_fx_common_color_or_texture_type,el_param);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE_PARAM;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture";
        el->offset=offsetof(dae_fx_common_color_or_texture_type,el_texture);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE_TEXTURE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "fx_common_color_or_texture_type_color";
    t->size = sizeof(dae_fx_common_color_or_texture_type_color);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE_COLOR;
    t->dataoffset=offsetof(dae_fx_common_color_or_texture_type_color,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=4;
    t->datamax=4;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_fx_common_color_or_texture_type_color,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_common_color_or_texture_type_param";
    t->size = sizeof(dae_fx_common_color_or_texture_type_param);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE_PARAM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_fx_common_color_or_texture_type_param,at_ref);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "fx_common_color_or_texture_type_texture";
    t->size = sizeof(dae_fx_common_color_or_texture_type_texture);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE_TEXTURE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="texture";
        at->offset=offsetof(dae_fx_common_color_or_texture_type_texture,at_texture);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        at->name="texcoord";
        at->offset=offsetof(dae_fx_common_color_or_texture_type_texture,at_texcoord);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="extra";
        el->offset=offsetof(dae_fx_common_color_or_texture_type_texture,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "fx_common_transparent_type";
    t->size = sizeof(dae_fx_common_transparent_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COMMON_TRANSPARENT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="opaque";
        at->offset=offsetof(dae_fx_common_transparent_type,at_opaque);
        at->objtypeid=dae_ID_FX_OPAQUE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="A_ONE";
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="color";
        el->offset=offsetof(dae_fx_common_transparent_type,el_color);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE_COLOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_fx_common_transparent_type,el_param);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE_PARAM;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture";
        el->offset=offsetof(dae_fx_common_transparent_type,el_texture);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE_TEXTURE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "fx_common_newparam_type";
    t->size = sizeof(dae_fx_common_newparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_COMMON_NEWPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_fx_common_newparam_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="semantic";
        el->offset=offsetof(dae_fx_common_newparam_type,el_semantic);
        el->objtypeid=dae_ID_NCNAME;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="surface";
        el->offset=offsetof(dae_fx_common_newparam_type,el_surface);
        el->objtypeid=dae_ID_FX_SURFACE_COMMON_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_fx_common_newparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_fx_common_newparam_type,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_fx_common_newparam_type,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_fx_common_newparam_type,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_fx_common_newparam_type,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "profile_common_type";
    t->size = sizeof(dae_profile_common_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_COMMON_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_common_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_common_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_profile_common_type,el_newparam);
        el->objtypeid=dae_ID_FX_COMMON_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_profile_common_type,el_technique);
        el->objtypeid=dae_ID_PROFILE_COMMON_TYPE_TECHNIQUE;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_common_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "profile_common_type_technique";
    t->size = sizeof(dae_profile_common_type_technique);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_COMMON_TYPE_TECHNIQUE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_common_type_technique,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_profile_common_type_technique,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[6];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_common_type_technique,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="constant";
        el->offset=offsetof(dae_profile_common_type_technique,el_constant);
        el->objtypeid=dae_ID_PROFILE_COMMON_TYPE_TECHNIQUE_CONSTANT;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="lambert";
        el->offset=offsetof(dae_profile_common_type_technique,el_lambert);
        el->objtypeid=dae_ID_PROFILE_COMMON_TYPE_TECHNIQUE_LAMBERT;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="phong";
        el->offset=offsetof(dae_profile_common_type_technique,el_phong);
        el->objtypeid=dae_ID_PROFILE_COMMON_TYPE_TECHNIQUE_PHONG;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="blinn";
        el->offset=offsetof(dae_profile_common_type_technique,el_blinn);
        el->objtypeid=dae_ID_PROFILE_COMMON_TYPE_TECHNIQUE_BLINN;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_common_type_technique,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 6;
    }
    ++t;
    t->name = "profile_common_type_technique_constant";
    t->size = sizeof(dae_profile_common_type_technique_constant);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_COMMON_TYPE_TECHNIQUE_CONSTANT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[6];
        dae_obj_memberdef* el = elems;
        el->name="emission";
        el->offset=offsetof(dae_profile_common_type_technique_constant,el_emission);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="reflective";
        el->offset=offsetof(dae_profile_common_type_technique_constant,el_reflective);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="reflectivity";
        el->offset=offsetof(dae_profile_common_type_technique_constant,el_reflectivity);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="transparent";
        el->offset=offsetof(dae_profile_common_type_technique_constant,el_transparent);
        el->objtypeid=dae_ID_FX_COMMON_TRANSPARENT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="transparency";
        el->offset=offsetof(dae_profile_common_type_technique_constant,el_transparency);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="index_of_refraction";
        el->offset=offsetof(dae_profile_common_type_technique_constant,el_index_of_refraction);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        t->elems = elems;
        t->numelems = 6;
    }
    ++t;
    t->name = "profile_common_type_technique_lambert";
    t->size = sizeof(dae_profile_common_type_technique_lambert);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_COMMON_TYPE_TECHNIQUE_LAMBERT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[8];
        dae_obj_memberdef* el = elems;
        el->name="emission";
        el->offset=offsetof(dae_profile_common_type_technique_lambert,el_emission);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ambient";
        el->offset=offsetof(dae_profile_common_type_technique_lambert,el_ambient);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="diffuse";
        el->offset=offsetof(dae_profile_common_type_technique_lambert,el_diffuse);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="reflective";
        el->offset=offsetof(dae_profile_common_type_technique_lambert,el_reflective);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="reflectivity";
        el->offset=offsetof(dae_profile_common_type_technique_lambert,el_reflectivity);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="transparent";
        el->offset=offsetof(dae_profile_common_type_technique_lambert,el_transparent);
        el->objtypeid=dae_ID_FX_COMMON_TRANSPARENT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="transparency";
        el->offset=offsetof(dae_profile_common_type_technique_lambert,el_transparency);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="index_of_refraction";
        el->offset=offsetof(dae_profile_common_type_technique_lambert,el_index_of_refraction);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=7;
        ++el;
        t->elems = elems;
        t->numelems = 8;
    }
    ++t;
    t->name = "profile_common_type_technique_phong";
    t->size = sizeof(dae_profile_common_type_technique_phong);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_COMMON_TYPE_TECHNIQUE_PHONG;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[10];
        dae_obj_memberdef* el = elems;
        el->name="emission";
        el->offset=offsetof(dae_profile_common_type_technique_phong,el_emission);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ambient";
        el->offset=offsetof(dae_profile_common_type_technique_phong,el_ambient);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="diffuse";
        el->offset=offsetof(dae_profile_common_type_technique_phong,el_diffuse);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="specular";
        el->offset=offsetof(dae_profile_common_type_technique_phong,el_specular);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="shininess";
        el->offset=offsetof(dae_profile_common_type_technique_phong,el_shininess);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="reflective";
        el->offset=offsetof(dae_profile_common_type_technique_phong,el_reflective);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="reflectivity";
        el->offset=offsetof(dae_profile_common_type_technique_phong,el_reflectivity);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="transparent";
        el->offset=offsetof(dae_profile_common_type_technique_phong,el_transparent);
        el->objtypeid=dae_ID_FX_COMMON_TRANSPARENT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=7;
        ++el;
        el->name="transparency";
        el->offset=offsetof(dae_profile_common_type_technique_phong,el_transparency);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=8;
        ++el;
        el->name="index_of_refraction";
        el->offset=offsetof(dae_profile_common_type_technique_phong,el_index_of_refraction);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=9;
        ++el;
        t->elems = elems;
        t->numelems = 10;
    }
    ++t;
    t->name = "profile_common_type_technique_blinn";
    t->size = sizeof(dae_profile_common_type_technique_blinn);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_COMMON_TYPE_TECHNIQUE_BLINN;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[10];
        dae_obj_memberdef* el = elems;
        el->name="emission";
        el->offset=offsetof(dae_profile_common_type_technique_blinn,el_emission);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ambient";
        el->offset=offsetof(dae_profile_common_type_technique_blinn,el_ambient);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="diffuse";
        el->offset=offsetof(dae_profile_common_type_technique_blinn,el_diffuse);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="specular";
        el->offset=offsetof(dae_profile_common_type_technique_blinn,el_specular);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="shininess";
        el->offset=offsetof(dae_profile_common_type_technique_blinn,el_shininess);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="reflective";
        el->offset=offsetof(dae_profile_common_type_technique_blinn,el_reflective);
        el->objtypeid=dae_ID_FX_COMMON_COLOR_OR_TEXTURE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="reflectivity";
        el->offset=offsetof(dae_profile_common_type_technique_blinn,el_reflectivity);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="transparent";
        el->offset=offsetof(dae_profile_common_type_technique_blinn,el_transparent);
        el->objtypeid=dae_ID_FX_COMMON_TRANSPARENT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=7;
        ++el;
        el->name="transparency";
        el->offset=offsetof(dae_profile_common_type_technique_blinn,el_transparency);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=8;
        ++el;
        el->name="index_of_refraction";
        el->offset=offsetof(dae_profile_common_type_technique_blinn,el_index_of_refraction);
        el->objtypeid=dae_ID_FX_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=9;
        ++el;
        t->elems = elems;
        t->numelems = 10;
    }
    ++t;
    t->name = "profile_bridge_type";
    t->size = sizeof(dae_profile_bridge_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_BRIDGE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_bridge_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="platform";
        at->offset=offsetof(dae_profile_bridge_type,at_platform);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="url";
        at->offset=offsetof(dae_profile_bridge_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_bridge_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_bridge_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_max_lights_index_type";
    t->size = sizeof(dae_gl_max_lights_index_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_max_clip_planes_index_type";
    t->size = sizeof(dae_gl_max_clip_planes_index_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_MAX_CLIP_PLANES_INDEX_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_max_texture_image_units_index_type";
    t->size = sizeof(dae_gl_max_texture_image_units_index_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_blend_enum";
    t->size = sizeof(dae_gl_blend_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_BLEND_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_face_enum";
    t->size = sizeof(dae_gl_face_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_FACE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_blend_equation_enum";
    t->size = sizeof(dae_gl_blend_equation_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_BLEND_EQUATION_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_func_enum";
    t->size = sizeof(dae_gl_func_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_FUNC_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_stencil_op_enum";
    t->size = sizeof(dae_gl_stencil_op_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_STENCIL_OP_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_material_enum";
    t->size = sizeof(dae_gl_material_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_MATERIAL_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_fog_enum";
    t->size = sizeof(dae_gl_fog_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_FOG_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_fog_coord_src_enum";
    t->size = sizeof(dae_gl_fog_coord_src_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_FOG_COORD_SRC_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_front_face_enum";
    t->size = sizeof(dae_gl_front_face_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_FRONT_FACE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_light_model_color_control_enum";
    t->size = sizeof(dae_gl_light_model_color_control_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_LIGHT_MODEL_COLOR_CONTROL_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_logic_op_enum";
    t->size = sizeof(dae_gl_logic_op_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_LOGIC_OP_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_polygon_mode_enum";
    t->size = sizeof(dae_gl_polygon_mode_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_POLYGON_MODE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_shade_model_enum";
    t->size = sizeof(dae_gl_shade_model_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_SHADE_MODEL_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_alpha_value_type";
    t->size = sizeof(dae_gl_alpha_value_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_ALPHA_VALUE_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_enumeration_type";
    t->size = sizeof(dae_gl_enumeration_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_ENUMERATION_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gl_pipeline_settings_group_alpha_func";
    t->size = sizeof(dae_gl_pipeline_settings_group_alpha_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="func";
        el->offset=offsetof(dae_gl_pipeline_settings_group_alpha_func,el_func);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="value";
        el->offset=offsetof(dae_gl_pipeline_settings_group_alpha_func,el_value);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC_VALUE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_alpha_func_func";
    t->size = sizeof(dae_gl_pipeline_settings_group_alpha_func_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_alpha_func_func,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_alpha_func_func,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_alpha_func_value";
    t->size = sizeof(dae_gl_pipeline_settings_group_alpha_func_value);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC_VALUE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_alpha_func_value,at_value);
        at->objtypeid=dae_ID_GL_ALPHA_VALUE_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0.0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_alpha_func_value,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_func";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="src";
        el->offset=offsetof(dae_gl_pipeline_settings_group_blend_func,el_src);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SRC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="dest";
        el->offset=offsetof(dae_gl_pipeline_settings_group_blend_func,el_dest);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_DEST;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_func_src";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_func_src);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SRC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_src,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ONE";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_src,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_func_dest";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_func_dest);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_DEST;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_dest,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ZERO";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_dest,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_func_separate";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_func_separate);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="src_rgb";
        el->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate,el_src_rgb);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_SRC_RGB;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="dest_rgb";
        el->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate,el_dest_rgb);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_DEST_RGB;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="src_alpha";
        el->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate,el_src_alpha);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_SRC_ALPHA;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="dest_alpha";
        el->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate,el_dest_alpha);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_DEST_ALPHA;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_func_separate_src_rgb";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_func_separate_src_rgb);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_SRC_RGB;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate_src_rgb,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ONE";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate_src_rgb,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_func_separate_dest_rgb";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_func_separate_dest_rgb);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_DEST_RGB;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate_dest_rgb,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ZERO";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate_dest_rgb,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_func_separate_src_alpha";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_func_separate_src_alpha);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_SRC_ALPHA;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate_src_alpha,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ONE";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate_src_alpha,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_func_separate_dest_alpha";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_func_separate_dest_alpha);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_DEST_ALPHA;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate_dest_alpha,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ZERO";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_func_separate_dest_alpha,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_equation";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_equation);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_equation,at_value);
        at->objtypeid=dae_ID_GL_BLEND_EQUATION_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FUNC_ADD";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_equation,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_equation_separate";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_equation_separate);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="rgb";
        el->offset=offsetof(dae_gl_pipeline_settings_group_blend_equation_separate,el_rgb);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE_RGB;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="alpha";
        el->offset=offsetof(dae_gl_pipeline_settings_group_blend_equation_separate,el_alpha);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE_ALPHA;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_equation_separate_rgb";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_equation_separate_rgb);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE_RGB;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_equation_separate_rgb,at_value);
        at->objtypeid=dae_ID_GL_BLEND_EQUATION_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FUNC_ADD";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_equation_separate_rgb,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_equation_separate_alpha";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_equation_separate_alpha);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE_ALPHA;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_equation_separate_alpha,at_value);
        at->objtypeid=dae_ID_GL_BLEND_EQUATION_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FUNC_ADD";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_equation_separate_alpha,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_color_material";
    t->size = sizeof(dae_gl_pipeline_settings_group_color_material);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="face";
        el->offset=offsetof(dae_gl_pipeline_settings_group_color_material,el_face);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mode";
        el->offset=offsetof(dae_gl_pipeline_settings_group_color_material,el_mode);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL_MODE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_color_material_face";
    t->size = sizeof(dae_gl_pipeline_settings_group_color_material_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_color_material_face,at_value);
        at->objtypeid=dae_ID_GL_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FRONT_AND_BACK";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_color_material_face,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_color_material_mode";
    t->size = sizeof(dae_gl_pipeline_settings_group_color_material_mode);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL_MODE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_color_material_mode,at_value);
        at->objtypeid=dae_ID_GL_MATERIAL_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="AMBIENT_AND_DIFFUSE";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_color_material_mode,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_cull_face";
    t->size = sizeof(dae_gl_pipeline_settings_group_cull_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_CULL_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_cull_face,at_value);
        at->objtypeid=dae_ID_GL_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="BACK";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_cull_face,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_depth_func";
    t->size = sizeof(dae_gl_pipeline_settings_group_depth_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_func,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_func,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_fog_mode";
    t->size = sizeof(dae_gl_pipeline_settings_group_fog_mode);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_MODE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_mode,at_value);
        at->objtypeid=dae_ID_GL_FOG_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="EXP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_mode,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_fog_coord_src";
    t->size = sizeof(dae_gl_pipeline_settings_group_fog_coord_src);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_COORD_SRC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_coord_src,at_value);
        at->objtypeid=dae_ID_GL_FOG_COORD_SRC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FOG_COORDINATE";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_coord_src,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_front_face";
    t->size = sizeof(dae_gl_pipeline_settings_group_front_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_FRONT_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_front_face,at_value);
        at->objtypeid=dae_ID_GL_FRONT_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="CCW";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_front_face,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_model_color_control";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_model_color_control);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_COLOR_CONTROL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_model_color_control,at_value);
        at->objtypeid=dae_ID_GL_LIGHT_MODEL_COLOR_CONTROL_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="SINGLE_COLOR";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_model_color_control,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_logic_op";
    t->size = sizeof(dae_gl_pipeline_settings_group_logic_op);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LOGIC_OP;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_logic_op,at_value);
        at->objtypeid=dae_ID_GL_LOGIC_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="COPY";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_logic_op,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_polygon_mode";
    t->size = sizeof(dae_gl_pipeline_settings_group_polygon_mode);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_MODE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="face";
        el->offset=offsetof(dae_gl_pipeline_settings_group_polygon_mode,el_face);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_MODE_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mode";
        el->offset=offsetof(dae_gl_pipeline_settings_group_polygon_mode,el_mode);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_MODE_MODE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_polygon_mode_face";
    t->size = sizeof(dae_gl_pipeline_settings_group_polygon_mode_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_MODE_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_mode_face,at_value);
        at->objtypeid=dae_ID_GL_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FRONT_AND_BACK";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_mode_face,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_polygon_mode_mode";
    t->size = sizeof(dae_gl_pipeline_settings_group_polygon_mode_mode);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_MODE_MODE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_mode_mode,at_value);
        at->objtypeid=dae_ID_GL_POLYGON_MODE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FILL";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_mode_mode,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_shade_model";
    t->size = sizeof(dae_gl_pipeline_settings_group_shade_model);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_SHADE_MODEL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_shade_model,at_value);
        at->objtypeid=dae_ID_GL_SHADE_MODEL_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="SMOOTH";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_shade_model,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_func";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="func";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func,el_func);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ref";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func,el_ref);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_REF;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="mask";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func,el_mask);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_MASK;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_func_func";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_func_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_func,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_func,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_func_ref";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_func_ref);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_REF;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_ref,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_ref,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_func_mask";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_func_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_mask,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="255";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_mask,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_op";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_op);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="fail";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op,el_fail);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_FAIL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="zfail";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op,el_zfail);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZFAIL;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="zpass";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op,el_zpass);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZPASS;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_op_fail";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_op_fail);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_FAIL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_fail,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_fail,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_op_zfail";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_op_zfail);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZFAIL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_zfail,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_zfail,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_op_zpass";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_op_zpass);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZPASS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_zpass,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_zpass,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_func_separate";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_func_separate);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="front";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate,el_front);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_FRONT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="back";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate,el_back);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_BACK;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="ref";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate,el_ref);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_REF;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="mask";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate,el_mask);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_MASK;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_func_separate_front";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_func_separate_front);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_FRONT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate_front,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate_front,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_func_separate_back";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_func_separate_back);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_BACK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate_back,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate_back,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_func_separate_ref";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_func_separate_ref);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_REF;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate_ref,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate_ref,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_func_separate_mask";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_func_separate_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate_mask,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="255";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_func_separate_mask,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_op_separate";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_op_separate);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="face";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate,el_face);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fail";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate,el_fail);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_FAIL;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="zfail";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate,el_zfail);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_ZFAIL;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="zpass";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate,el_zpass);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_ZPASS;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_op_separate_face";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_op_separate_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate_face,at_value);
        at->objtypeid=dae_ID_GL_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FRONT_AND_BACK";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate_face,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_op_separate_fail";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_op_separate_fail);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_FAIL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate_fail,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate_fail,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_op_separate_zfail";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_op_separate_zfail);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_ZFAIL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate_zfail,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate_zfail,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_op_separate_zpass";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_op_separate_zpass);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_ZPASS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate_zpass,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_op_separate_zpass,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_mask_separate";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_mask_separate);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="face";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_mask_separate,el_face);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mask";
        el->offset=offsetof(dae_gl_pipeline_settings_group_stencil_mask_separate,el_mask);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE_MASK;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_mask_separate_face";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_mask_separate_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_mask_separate_face,at_value);
        at->objtypeid=dae_ID_GL_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FRONT_AND_BACK";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_mask_separate_face,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_mask_separate_mask";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_mask_separate_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_mask_separate_mask,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="255";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_mask_separate_mask,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_enable,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_ambient";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_ambient);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_AMBIENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_ambient,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_ambient,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_ambient,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_diffuse";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_diffuse);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_DIFFUSE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_diffuse,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_diffuse,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_diffuse,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_specular";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_specular);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPECULAR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_specular,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_specular,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_specular,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_position";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_position);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_POSITION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_position,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 1 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_position,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_position,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_constant_attenuation";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_constant_attenuation);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_CONSTANT_ATTENUATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_constant_attenuation,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_constant_attenuation,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_constant_attenuation,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_linear_attenuation";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_linear_attenuation);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_LINEAR_ATTENUATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_linear_attenuation,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_linear_attenuation,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_linear_attenuation,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_quadratic_attenuation";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_quadratic_attenuation);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_QUADRATIC_ATTENUATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_quadratic_attenuation,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_quadratic_attenuation,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_quadratic_attenuation,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_spot_cutoff";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_spot_cutoff);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_CUTOFF;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_spot_cutoff,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="180";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_spot_cutoff,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_spot_cutoff,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_spot_direction";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_spot_direction);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_DIRECTION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_spot_direction,at_value);
        at->objtypeid=dae_ID_FLOAT3_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 -1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_spot_direction,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_spot_direction,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_spot_exponent";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_spot_exponent);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_EXPONENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_spot_exponent,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_spot_exponent,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_spot_exponent,at_index);
        at->objtypeid=dae_ID_GL_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_texture1D";
    t->size = sizeof(dae_gl_pipeline_settings_group_texture1D);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE1D;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture1D,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="value";
        el->offset=offsetof(dae_gl_pipeline_settings_group_texture1D,el_value);
        el->objtypeid=dae_ID_FX_SAMPLER1D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_gl_pipeline_settings_group_texture1D,el_param);
        el->objtypeid=dae_ID_NCNAME;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_texture2D";
    t->size = sizeof(dae_gl_pipeline_settings_group_texture2D);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE2D;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture2D,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="value";
        el->offset=offsetof(dae_gl_pipeline_settings_group_texture2D,el_value);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_gl_pipeline_settings_group_texture2D,el_param);
        el->objtypeid=dae_ID_NCNAME;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_texture3D";
    t->size = sizeof(dae_gl_pipeline_settings_group_texture3D);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE3D;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture3D,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="value";
        el->offset=offsetof(dae_gl_pipeline_settings_group_texture3D,el_value);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_gl_pipeline_settings_group_texture3D,el_param);
        el->objtypeid=dae_ID_NCNAME;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_textureCUBE";
    t->size = sizeof(dae_gl_pipeline_settings_group_textureCUBE);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURECUBE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureCUBE,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="value";
        el->offset=offsetof(dae_gl_pipeline_settings_group_textureCUBE,el_value);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_gl_pipeline_settings_group_textureCUBE,el_param);
        el->objtypeid=dae_ID_NCNAME;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_textureRECT";
    t->size = sizeof(dae_gl_pipeline_settings_group_textureRECT);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURERECT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureRECT,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="value";
        el->offset=offsetof(dae_gl_pipeline_settings_group_textureRECT,el_value);
        el->objtypeid=dae_ID_FX_SAMPLERRECT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_gl_pipeline_settings_group_textureRECT,el_param);
        el->objtypeid=dae_ID_NCNAME;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_textureDEPTH";
    t->size = sizeof(dae_gl_pipeline_settings_group_textureDEPTH);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTUREDEPTH;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureDEPTH,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="value";
        el->offset=offsetof(dae_gl_pipeline_settings_group_textureDEPTH,el_value);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_gl_pipeline_settings_group_textureDEPTH,el_param);
        el->objtypeid=dae_ID_NCNAME;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_texture1D_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_texture1D_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE1D_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture1D_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture1D_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture1D_enable,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_texture2D_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_texture2D_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE2D_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture2D_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture2D_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture2D_enable,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_texture3D_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_texture3D_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE3D_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture3D_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture3D_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture3D_enable,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_textureCUBE_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_textureCUBE_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURECUBE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureCUBE_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureCUBE_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureCUBE_enable,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_textureRECT_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_textureRECT_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURERECT_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureRECT_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureRECT_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureRECT_enable,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_textureDEPTH_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_textureDEPTH_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTUREDEPTH_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureDEPTH_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureDEPTH_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_textureDEPTH_enable,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_texture_env_color";
    t->size = sizeof(dae_gl_pipeline_settings_group_texture_env_color);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE_ENV_COLOR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture_env_color,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture_env_color,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture_env_color,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_texture_env_mode";
    t->size = sizeof(dae_gl_pipeline_settings_group_texture_env_mode);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE_ENV_MODE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture_env_mode,at_value);
        at->objtypeid=dae_ID_STRING;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture_env_mode,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_texture_env_mode,at_index);
        at->objtypeid=dae_ID_GL_MAX_TEXTURE_IMAGE_UNITS_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_clip_plane";
    t->size = sizeof(dae_gl_pipeline_settings_group_clip_plane);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_CLIP_PLANE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_clip_plane,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_clip_plane,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_clip_plane,at_index);
        at->objtypeid=dae_ID_GL_MAX_CLIP_PLANES_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_clip_plane_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_clip_plane_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_CLIP_PLANE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_clip_plane_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_clip_plane_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gl_pipeline_settings_group_clip_plane_enable,at_index);
        at->objtypeid=dae_ID_GL_MAX_CLIP_PLANES_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_color";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_color);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_COLOR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_color,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_color,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_color_mask";
    t->size = sizeof(dae_gl_pipeline_settings_group_color_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_color_mask,at_value);
        at->objtypeid=dae_ID_BOOL4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="true true true true";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_color_mask,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_depth_bounds";
    t->size = sizeof(dae_gl_pipeline_settings_group_depth_bounds);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_BOUNDS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_bounds,at_value);
        at->objtypeid=dae_ID_FLOAT2_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_bounds,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_depth_mask";
    t->size = sizeof(dae_gl_pipeline_settings_group_depth_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_mask,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="true";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_mask,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_depth_range";
    t->size = sizeof(dae_gl_pipeline_settings_group_depth_range);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_RANGE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_range,at_value);
        at->objtypeid=dae_ID_FLOAT2_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_range,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_fog_density";
    t->size = sizeof(dae_gl_pipeline_settings_group_fog_density);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_DENSITY;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_density,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_density,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_fog_start";
    t->size = sizeof(dae_gl_pipeline_settings_group_fog_start);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_START;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_start,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_start,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_fog_end";
    t->size = sizeof(dae_gl_pipeline_settings_group_fog_end);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_END;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_end,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_end,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_fog_color";
    t->size = sizeof(dae_gl_pipeline_settings_group_fog_color);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_COLOR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_color,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_color,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_model_ambient";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_model_ambient);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_AMBIENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_model_ambient,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0.2 0.2 0.2 1.0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_model_ambient,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_lighting_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_lighting_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHTING_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_lighting_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_lighting_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_line_stipple";
    t->size = sizeof(dae_gl_pipeline_settings_group_line_stipple);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_STIPPLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_line_stipple,at_value);
        at->objtypeid=dae_ID_INT2_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1 65536";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_line_stipple,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_line_width";
    t->size = sizeof(dae_gl_pipeline_settings_group_line_width);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_WIDTH;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_line_width,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_line_width,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_material_ambient";
    t->size = sizeof(dae_gl_pipeline_settings_group_material_ambient);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_AMBIENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_material_ambient,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0.2 0.2 0.2 1.0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_material_ambient,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_material_diffuse";
    t->size = sizeof(dae_gl_pipeline_settings_group_material_diffuse);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_DIFFUSE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_material_diffuse,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0.8 0.8 0.8 1.0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_material_diffuse,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_material_emission";
    t->size = sizeof(dae_gl_pipeline_settings_group_material_emission);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_EMISSION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_material_emission,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_material_emission,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_material_shininess";
    t->size = sizeof(dae_gl_pipeline_settings_group_material_shininess);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_SHININESS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_material_shininess,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_material_shininess,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_material_specular";
    t->size = sizeof(dae_gl_pipeline_settings_group_material_specular);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_SPECULAR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_material_specular,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_material_specular,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_model_view_matrix";
    t->size = sizeof(dae_gl_pipeline_settings_group_model_view_matrix);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_MODEL_VIEW_MATRIX;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_model_view_matrix,at_value);
        at->objtypeid=dae_ID_FLOAT4X4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_model_view_matrix,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_point_distance_attenuation";
    t->size = sizeof(dae_gl_pipeline_settings_group_point_distance_attenuation);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_DISTANCE_ATTENUATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_distance_attenuation,at_value);
        at->objtypeid=dae_ID_FLOAT3_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_distance_attenuation,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_point_fade_threshold_size";
    t->size = sizeof(dae_gl_pipeline_settings_group_point_fade_threshold_size);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_FADE_THRESHOLD_SIZE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_fade_threshold_size,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_fade_threshold_size,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_point_size";
    t->size = sizeof(dae_gl_pipeline_settings_group_point_size);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SIZE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_size,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_size,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_point_size_min";
    t->size = sizeof(dae_gl_pipeline_settings_group_point_size_min);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SIZE_MIN;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_size_min,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_size_min,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_point_size_max";
    t->size = sizeof(dae_gl_pipeline_settings_group_point_size_max);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SIZE_MAX;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_size_max,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_size_max,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_polygon_offset";
    t->size = sizeof(dae_gl_pipeline_settings_group_polygon_offset);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_offset,at_value);
        at->objtypeid=dae_ID_FLOAT2_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_offset,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_projection_matrix";
    t->size = sizeof(dae_gl_pipeline_settings_group_projection_matrix);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_PROJECTION_MATRIX;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_projection_matrix,at_value);
        at->objtypeid=dae_ID_FLOAT4X4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_projection_matrix,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_scissor";
    t->size = sizeof(dae_gl_pipeline_settings_group_scissor);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_SCISSOR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_scissor,at_value);
        at->objtypeid=dae_ID_INT4_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_scissor,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_mask";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_mask,at_value);
        at->objtypeid=dae_ID_INT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="4294967295";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_mask,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_alpha_test_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_alpha_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_ALPHA_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_alpha_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_alpha_test_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_blend_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_blend_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_blend_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_color_logic_op_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_color_logic_op_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_LOGIC_OP_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_color_logic_op_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_color_logic_op_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_color_material_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_color_material_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_color_material_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="true";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_color_material_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_cull_face_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_cull_face_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_CULL_FACE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_cull_face_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_cull_face_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_depth_bounds_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_depth_bounds_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_BOUNDS_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_bounds_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_bounds_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_depth_clamp_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_depth_clamp_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_CLAMP_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_clamp_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_clamp_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_depth_test_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_depth_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_depth_test_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_dither_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_dither_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_DITHER_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_dither_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="true";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_dither_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_fog_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_fog_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_fog_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_model_local_viewer_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_model_local_viewer_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_LOCAL_VIEWER_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_model_local_viewer_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_model_local_viewer_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_light_model_two_side_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_light_model_two_side_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_TWO_SIDE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_model_two_side_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_light_model_two_side_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_line_smooth_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_line_smooth_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_SMOOTH_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_line_smooth_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_line_smooth_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_line_stipple_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_line_stipple_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_STIPPLE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_line_stipple_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_line_stipple_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_logic_op_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_logic_op_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_LOGIC_OP_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_logic_op_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_logic_op_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_multisample_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_multisample_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_MULTISAMPLE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_multisample_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_multisample_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_normalize_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_normalize_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_NORMALIZE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_normalize_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_normalize_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_point_smooth_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_point_smooth_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SMOOTH_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_smooth_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_point_smooth_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_polygon_offset_fill_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_polygon_offset_fill_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_FILL_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_offset_fill_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_offset_fill_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_polygon_offset_line_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_polygon_offset_line_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_LINE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_offset_line_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_offset_line_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_polygon_offset_point_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_polygon_offset_point_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_POINT_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_offset_point_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_offset_point_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_polygon_smooth_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_polygon_smooth_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_SMOOTH_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_smooth_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_smooth_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_polygon_stipple_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_polygon_stipple_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_STIPPLE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_stipple_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_polygon_stipple_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_rescale_normal_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_rescale_normal_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_RESCALE_NORMAL_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_rescale_normal_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_rescale_normal_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_sample_alpha_to_coverage_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_sample_alpha_to_coverage_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_COVERAGE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_sample_alpha_to_coverage_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_sample_alpha_to_coverage_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_sample_alpha_to_one_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_sample_alpha_to_one_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_ONE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_sample_alpha_to_one_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_sample_alpha_to_one_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_sample_coverage_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_sample_coverage_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_sample_coverage_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_sample_coverage_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_scissor_test_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_scissor_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_SCISSOR_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_scissor_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_scissor_test_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gl_pipeline_settings_group_stencil_test_enable";
    t->size = sizeof(dae_gl_pipeline_settings_group_stencil_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gl_pipeline_settings_group_stencil_test_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_color";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_color);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_COLOR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_color,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_color,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_equation";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_equation);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_equation,at_value);
        at->objtypeid=dae_ID_GL_BLEND_EQUATION_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FUNC_ADD";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_equation,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_equation_separate";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_equation_separate);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="rgb";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_blend_equation_separate,el_rgb);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE_RGB;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="alpha";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_blend_equation_separate,el_alpha);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE_ALPHA;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_equation_separate_rgb";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_equation_separate_rgb);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE_RGB;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_equation_separate_rgb,at_value);
        at->objtypeid=dae_ID_GL_BLEND_EQUATION_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FUNC_ADD";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_equation_separate_rgb,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_equation_separate_alpha";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_equation_separate_alpha);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE_ALPHA;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_equation_separate_alpha,at_value);
        at->objtypeid=dae_ID_GL_BLEND_EQUATION_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FUNC_ADD";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_equation_separate_alpha,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_func";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="src";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func,el_src);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SRC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="dest";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func,el_dest);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_DEST;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_func_src";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_func_src);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SRC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_src,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ONE";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_src,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_func_dest";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_func_dest);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_DEST;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_dest,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ZERO";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_dest,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_func_separate";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_func_separate);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="src_rgb";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate,el_src_rgb);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_SRC_RGB;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="dest_rgb";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate,el_dest_rgb);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_DEST_RGB;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="src_alpha";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate,el_src_alpha);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_SRC_ALPHA;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="dest_alpha";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate,el_dest_alpha);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_DEST_ALPHA;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_func_separate_src_rgb";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_func_separate_src_rgb);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_SRC_RGB;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate_src_rgb,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ONE";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate_src_rgb,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_func_separate_dest_rgb";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_func_separate_dest_rgb);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_DEST_RGB;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate_dest_rgb,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ZERO";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate_dest_rgb,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_func_separate_src_alpha";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_func_separate_src_alpha);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_SRC_ALPHA;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate_src_alpha,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ONE";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate_src_alpha,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_func_separate_dest_alpha";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_func_separate_dest_alpha);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE_DEST_ALPHA;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate_dest_alpha,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ZERO";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_func_separate_dest_alpha,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_color_mask";
    t->size = sizeof(dae_gles2_pipeline_settings_group_color_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_COLOR_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_color_mask,at_value);
        at->objtypeid=dae_ID_BOOL4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="true true true true";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_color_mask,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_cull_face";
    t->size = sizeof(dae_gles2_pipeline_settings_group_cull_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_CULL_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_cull_face,at_value);
        at->objtypeid=dae_ID_GL_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="BACK";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_cull_face,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_depth_func";
    t->size = sizeof(dae_gles2_pipeline_settings_group_depth_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_DEPTH_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_depth_func,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_depth_func,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_depth_mask";
    t->size = sizeof(dae_gles2_pipeline_settings_group_depth_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_DEPTH_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_depth_mask,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="true";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_depth_mask,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_depth_range";
    t->size = sizeof(dae_gles2_pipeline_settings_group_depth_range);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_DEPTH_RANGE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_depth_range,at_value);
        at->objtypeid=dae_ID_FLOAT2_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_depth_range,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_front_face";
    t->size = sizeof(dae_gles2_pipeline_settings_group_front_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_FRONT_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_front_face,at_value);
        at->objtypeid=dae_ID_GL_FRONT_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="CCW";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_front_face,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_line_width";
    t->size = sizeof(dae_gles2_pipeline_settings_group_line_width);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_LINE_WIDTH;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_line_width,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_line_width,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_polygon_offset";
    t->size = sizeof(dae_gles2_pipeline_settings_group_polygon_offset);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_polygon_offset,at_value);
        at->objtypeid=dae_ID_FLOAT2_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_polygon_offset,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_point_size";
    t->size = sizeof(dae_gles2_pipeline_settings_group_point_size);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_POINT_SIZE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_point_size,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_point_size,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_sample_coverage";
    t->size = sizeof(dae_gles2_pipeline_settings_group_sample_coverage);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="value";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_sample_coverage,el_value);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_VALUE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="invert";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_sample_coverage,el_invert);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_INVERT;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_sample_coverage_value";
    t->size = sizeof(dae_gles2_pipeline_settings_group_sample_coverage_value);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_VALUE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_sample_coverage_value,at_value);
        at->objtypeid=dae_ID_FLOAT;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_sample_coverage_value,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_sample_coverage_invert";
    t->size = sizeof(dae_gles2_pipeline_settings_group_sample_coverage_invert);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_INVERT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_sample_coverage_invert,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_sample_coverage_invert,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_scissor";
    t->size = sizeof(dae_gles2_pipeline_settings_group_scissor);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SCISSOR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_scissor,at_value);
        at->objtypeid=dae_ID_INT4_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_scissor,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_func";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="func";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func,el_func);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ref";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func,el_ref);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_REF;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="mask";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func,el_mask);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_MASK;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_func_func";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_func_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_func,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_func,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_func_ref";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_func_ref);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_REF;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_ref,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_ref,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_func_mask";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_func_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_mask,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="255";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_mask,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_func_separate";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_func_separate);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="front";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate,el_front);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_FRONT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="back";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate,el_back);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_BACK;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="ref";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate,el_ref);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_REF;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="mask";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate,el_mask);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_MASK;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_func_separate_front";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_func_separate_front);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_FRONT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate_front,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate_front,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_func_separate_back";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_func_separate_back);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_BACK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate_back,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate_back,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_func_separate_ref";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_func_separate_ref);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_REF;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate_ref,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate_ref,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_func_separate_mask";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_func_separate_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate_mask,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="255";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_func_separate_mask,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_mask";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_mask,at_value);
        at->objtypeid=dae_ID_INT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="4294967295";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_mask,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_mask_separate";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_mask_separate);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="face";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_mask_separate,el_face);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mask";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_mask_separate,el_mask);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE_MASK;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_mask_separate_face";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_mask_separate_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_mask_separate_face,at_value);
        at->objtypeid=dae_ID_GL_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FRONT_AND_BACK";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_mask_separate_face,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_mask_separate_mask";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_mask_separate_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_mask_separate_mask,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="255";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_mask_separate_mask,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_op";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_op);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="fail";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op,el_fail);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_FAIL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="zfail";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op,el_zfail);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZFAIL;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="zpass";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op,el_zpass);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZPASS;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_op_fail";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_op_fail);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_FAIL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_fail,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_fail,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_op_zfail";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_op_zfail);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZFAIL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_zfail,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_zfail,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_op_zpass";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_op_zpass);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZPASS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_zpass,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_zpass,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_op_separate";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_op_separate);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="face";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate,el_face);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fail";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate,el_fail);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_FAIL;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="zfail";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate,el_zfail);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_ZFAIL;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="zpass";
        el->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate,el_zpass);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_ZPASS;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_op_separate_face";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_op_separate_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate_face,at_value);
        at->objtypeid=dae_ID_GL_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="FRONT_AND_BACK";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate_face,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_op_separate_fail";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_op_separate_fail);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_FAIL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate_fail,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate_fail,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_op_separate_zfail";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_op_separate_zfail);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_ZFAIL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate_zfail,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate_zfail,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_op_separate_zpass";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_op_separate_zpass);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE_ZPASS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate_zpass,at_value);
        at->objtypeid=dae_ID_GL_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_op_separate_zpass,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_blend_enable";
    t->size = sizeof(dae_gles2_pipeline_settings_group_blend_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_blend_enable,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_cull_face_enable";
    t->size = sizeof(dae_gles2_pipeline_settings_group_cull_face_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_CULL_FACE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_cull_face_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_cull_face_enable,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_depth_test_enable";
    t->size = sizeof(dae_gles2_pipeline_settings_group_depth_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_DEPTH_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_depth_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_depth_test_enable,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_dither_enable";
    t->size = sizeof(dae_gles2_pipeline_settings_group_dither_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_DITHER_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_dither_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="true";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_dither_enable,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_polygon_offset_fill_enable";
    t->size = sizeof(dae_gles2_pipeline_settings_group_polygon_offset_fill_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_FILL_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_polygon_offset_fill_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_polygon_offset_fill_enable,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_point_size_enable";
    t->size = sizeof(dae_gles2_pipeline_settings_group_point_size_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_POINT_SIZE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_point_size_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_point_size_enable,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_sample_alpha_to_coverage_enable";
    t->size = sizeof(dae_gles2_pipeline_settings_group_sample_alpha_to_coverage_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_COVERAGE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_sample_alpha_to_coverage_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_sample_alpha_to_coverage_enable,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_sample_coverage_enable";
    t->size = sizeof(dae_gles2_pipeline_settings_group_sample_coverage_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_sample_coverage_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_sample_coverage_enable,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_scissor_test_enable";
    t->size = sizeof(dae_gles2_pipeline_settings_group_scissor_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SCISSOR_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_scissor_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_scissor_test_enable,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_pipeline_settings_group_stencil_test_enable";
    t->size = sizeof(dae_gles2_pipeline_settings_group_stencil_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles2_pipeline_settings_group_stencil_test_enable,at_param);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles2_value_group_usertype";
    t->size = sizeof(dae_gles2_value_group_usertype);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_VALUE_GROUP_USERTYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="typename";
        at->offset=offsetof(dae_gles2_value_group_usertype,at_typename);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="setparam";
        el->offset=offsetof(dae_gles2_value_group_usertype,el_setparam);
        el->objtypeid=dae_ID_GLES2_VALUE_GROUP_USERTYPE_SETPARAM;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "gles2_value_group_usertype_setparam";
    t->size = sizeof(dae_gles2_value_group_usertype_setparam);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_VALUE_GROUP_USERTYPE_SETPARAM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_gles2_value_group_usertype_setparam,at_ref);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[22];
        dae_obj_memberdef* el = elems;
        el->name="bool";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bvec2";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_bvec2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bvec3";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_bvec3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bvec4";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_bvec4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="vec2";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_vec2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="vec3";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_vec3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="vec4";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_vec4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mat2";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_mat2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mat3";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_mat3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mat4";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_mat4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ivec2";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_ivec2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ivec3";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_ivec3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ivec4";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_ivec4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="usertype";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_usertype);
        el->objtypeid=dae_ID_GLES2_VALUE_GROUP_USERTYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_gles2_value_group_usertype_setparam,el_array);
        el->objtypeid=dae_ID_GLES2_VALUE_GROUP_ARRAY;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 22;
    }
    ++t;
    t->name = "gles2_value_group_array";
    t->size = sizeof(dae_gles2_value_group_array);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_VALUE_GROUP_ARRAY;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="length";
        at->offset=offsetof(dae_gles2_value_group_array,at_length);
        at->objtypeid=dae_ID_POSITIVEINTEGER;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[22];
        dae_obj_memberdef* el = elems;
        el->name="bool";
        el->offset=offsetof(dae_gles2_value_group_array,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bvec2";
        el->offset=offsetof(dae_gles2_value_group_array,el_bvec2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bvec3";
        el->offset=offsetof(dae_gles2_value_group_array,el_bvec3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bvec4";
        el->offset=offsetof(dae_gles2_value_group_array,el_bvec4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_gles2_value_group_array,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="vec2";
        el->offset=offsetof(dae_gles2_value_group_array,el_vec2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="vec3";
        el->offset=offsetof(dae_gles2_value_group_array,el_vec3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="vec4";
        el->offset=offsetof(dae_gles2_value_group_array,el_vec4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mat2";
        el->offset=offsetof(dae_gles2_value_group_array,el_mat2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mat3";
        el->offset=offsetof(dae_gles2_value_group_array,el_mat3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mat4";
        el->offset=offsetof(dae_gles2_value_group_array,el_mat4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_gles2_value_group_array,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ivec2";
        el->offset=offsetof(dae_gles2_value_group_array,el_ivec2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ivec3";
        el->offset=offsetof(dae_gles2_value_group_array,el_ivec3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ivec4";
        el->offset=offsetof(dae_gles2_value_group_array,el_ivec4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_gles2_value_group_array,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_gles2_value_group_array,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_gles2_value_group_array,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_gles2_value_group_array,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_gles2_value_group_array,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="usertype";
        el->offset=offsetof(dae_gles2_value_group_array,el_usertype);
        el->objtypeid=dae_ID_GLES2_VALUE_GROUP_USERTYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_gles2_value_group_array,el_array);
        el->objtypeid=dae_ID_GLES2_VALUE_GROUP_ARRAY;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 22;
    }
    ++t;
    t->name = "gles2_newparam_type";
    t->size = sizeof(dae_gles2_newparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_NEWPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_gles2_newparam_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[25];
        dae_obj_memberdef* el = elems;
        el->name="annotate";
        el->offset=offsetof(dae_gles2_newparam_type,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="semantic";
        el->offset=offsetof(dae_gles2_newparam_type,el_semantic);
        el->objtypeid=dae_ID_NCNAME;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="modifier";
        el->offset=offsetof(dae_gles2_newparam_type,el_modifier);
        el->objtypeid=dae_ID_FX_MODIFIER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_gles2_newparam_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bvec2";
        el->offset=offsetof(dae_gles2_newparam_type,el_bvec2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bvec3";
        el->offset=offsetof(dae_gles2_newparam_type,el_bvec3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bvec4";
        el->offset=offsetof(dae_gles2_newparam_type,el_bvec4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_gles2_newparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="vec2";
        el->offset=offsetof(dae_gles2_newparam_type,el_vec2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="vec3";
        el->offset=offsetof(dae_gles2_newparam_type,el_vec3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="vec4";
        el->offset=offsetof(dae_gles2_newparam_type,el_vec4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="mat2";
        el->offset=offsetof(dae_gles2_newparam_type,el_mat2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="mat3";
        el->offset=offsetof(dae_gles2_newparam_type,el_mat3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="mat4";
        el->offset=offsetof(dae_gles2_newparam_type,el_mat4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_gles2_newparam_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="ivec2";
        el->offset=offsetof(dae_gles2_newparam_type,el_ivec2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="ivec3";
        el->offset=offsetof(dae_gles2_newparam_type,el_ivec3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="ivec4";
        el->offset=offsetof(dae_gles2_newparam_type,el_ivec4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_gles2_newparam_type,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_gles2_newparam_type,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_gles2_newparam_type,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_gles2_newparam_type,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_gles2_newparam_type,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="usertype";
        el->offset=offsetof(dae_gles2_newparam_type,el_usertype);
        el->objtypeid=dae_ID_GLES2_VALUE_GROUP_USERTYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_gles2_newparam_type,el_array);
        el->objtypeid=dae_ID_GLES2_VALUE_GROUP_ARRAY;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 25;
    }
    ++t;
    t->name = "gles2_shader_type";
    t->size = sizeof(dae_gles2_shader_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_SHADER_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="stage";
        at->offset=offsetof(dae_gles2_shader_type,at_stage);
        at->objtypeid=dae_ID_FX_PIPELINE_STAGE_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="sources";
        el->offset=offsetof(dae_gles2_shader_type,el_sources);
        el->objtypeid=dae_ID_GLES2_SHADER_TYPE_SOURCES;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="compiler";
        el->offset=offsetof(dae_gles2_shader_type,el_compiler);
        el->objtypeid=dae_ID_FX_TARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_gles2_shader_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "gles2_shader_type_sources";
    t->size = sizeof(dae_gles2_shader_type_sources);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_SHADER_TYPE_SOURCES;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="entry";
        at->offset=offsetof(dae_gles2_shader_type_sources,at_entry);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        at->deflt="main";
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="inline";
        el->offset=offsetof(dae_gles2_shader_type_sources,el_inline);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="import";
        el->offset=offsetof(dae_gles2_shader_type_sources,el_import);
        el->objtypeid=dae_ID_FX_SOURCES_TYPE_IMPORT;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gles2_program_type";
    t->size = sizeof(dae_gles2_program_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PROGRAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="shader";
        el->offset=offsetof(dae_gles2_program_type,el_shader);
        el->objtypeid=dae_ID_GLES2_SHADER_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="linker";
        el->offset=offsetof(dae_gles2_program_type,el_linker);
        el->objtypeid=dae_ID_FX_TARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="bind_attribute";
        el->offset=offsetof(dae_gles2_program_type,el_bind_attribute);
        el->objtypeid=dae_ID_GLES2_PROGRAM_TYPE_BIND_ATTRIBUTE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="bind_uniform";
        el->offset=offsetof(dae_gles2_program_type,el_bind_uniform);
        el->objtypeid=dae_ID_GLES2_PROGRAM_TYPE_BIND_UNIFORM;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "gles2_program_type_bind_attribute";
    t->size = sizeof(dae_gles2_program_type_bind_attribute);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PROGRAM_TYPE_BIND_ATTRIBUTE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="symbol";
        at->offset=offsetof(dae_gles2_program_type_bind_attribute,at_symbol);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="semantic";
        el->offset=offsetof(dae_gles2_program_type_bind_attribute,el_semantic);
        el->objtypeid=dae_ID_TOKEN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "gles2_program_type_bind_uniform";
    t->size = sizeof(dae_gles2_program_type_bind_uniform);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PROGRAM_TYPE_BIND_UNIFORM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="symbol";
        at->offset=offsetof(dae_gles2_program_type_bind_uniform,at_symbol);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[23];
        dae_obj_memberdef* el = elems;
        el->name="param";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_param);
        el->objtypeid=dae_ID_GLES2_PROGRAM_TYPE_BIND_UNIFORM_PARAM;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bvec2";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_bvec2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bvec3";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_bvec3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bvec4";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_bvec4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="vec2";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_vec2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="vec3";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_vec3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="vec4";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_vec4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mat2";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_mat2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mat3";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_mat3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mat4";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_mat4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ivec2";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_ivec2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ivec3";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_ivec3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ivec4";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_ivec4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="usertype";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_usertype);
        el->objtypeid=dae_ID_GLES2_VALUE_GROUP_USERTYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_gles2_program_type_bind_uniform,el_array);
        el->objtypeid=dae_ID_GLES2_VALUE_GROUP_ARRAY;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 23;
    }
    ++t;
    t->name = "gles2_program_type_bind_uniform_param";
    t->size = sizeof(dae_gles2_program_type_bind_uniform_param);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PROGRAM_TYPE_BIND_UNIFORM_PARAM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_gles2_program_type_bind_uniform_param,at_ref);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "gles2_pass_type";
    t->size = sizeof(dae_gles2_pass_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PASS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_gles2_pass_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="annotate";
        el->offset=offsetof(dae_gles2_pass_type,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="states";
        el->offset=offsetof(dae_gles2_pass_type,el_states);
        el->objtypeid=dae_ID_GLES2_PASS_TYPE_STATES;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="program";
        el->offset=offsetof(dae_gles2_pass_type,el_program);
        el->objtypeid=dae_ID_GLES2_PROGRAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="evaluate";
        el->offset=offsetof(dae_gles2_pass_type,el_evaluate);
        el->objtypeid=dae_ID_GLES2_PASS_TYPE_EVALUATE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_gles2_pass_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "gles2_pass_type_states";
    t->size = sizeof(dae_gles2_pass_type_states);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PASS_TYPE_STATES;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[32];
        dae_obj_memberdef* el = elems;
        el->name="blend_color";
        el->offset=offsetof(dae_gles2_pass_type_states,el_blend_color);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_COLOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_equation";
        el->offset=offsetof(dae_gles2_pass_type_states,el_blend_equation);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_equation_separate";
        el->offset=offsetof(dae_gles2_pass_type_states,el_blend_equation_separate);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_func";
        el->offset=offsetof(dae_gles2_pass_type_states,el_blend_func);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_func_separate";
        el->offset=offsetof(dae_gles2_pass_type_states,el_blend_func_separate);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_mask";
        el->offset=offsetof(dae_gles2_pass_type_states,el_color_mask);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_COLOR_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="cull_face";
        el->offset=offsetof(dae_gles2_pass_type_states,el_cull_face);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_CULL_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_func";
        el->offset=offsetof(dae_gles2_pass_type_states,el_depth_func);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_DEPTH_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_mask";
        el->offset=offsetof(dae_gles2_pass_type_states,el_depth_mask);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_DEPTH_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_range";
        el->offset=offsetof(dae_gles2_pass_type_states,el_depth_range);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_DEPTH_RANGE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="front_face";
        el->offset=offsetof(dae_gles2_pass_type_states,el_front_face);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_FRONT_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_width";
        el->offset=offsetof(dae_gles2_pass_type_states,el_line_width);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_LINE_WIDTH;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset";
        el->offset=offsetof(dae_gles2_pass_type_states,el_polygon_offset);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size";
        el->offset=offsetof(dae_gles2_pass_type_states,el_point_size);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_POINT_SIZE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_coverage";
        el->offset=offsetof(dae_gles2_pass_type_states,el_sample_coverage);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="scissor";
        el->offset=offsetof(dae_gles2_pass_type_states,el_scissor);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SCISSOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_func";
        el->offset=offsetof(dae_gles2_pass_type_states,el_stencil_func);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_func_separate";
        el->offset=offsetof(dae_gles2_pass_type_states,el_stencil_func_separate);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_mask";
        el->offset=offsetof(dae_gles2_pass_type_states,el_stencil_mask);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_mask_separate";
        el->offset=offsetof(dae_gles2_pass_type_states,el_stencil_mask_separate);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_op";
        el->offset=offsetof(dae_gles2_pass_type_states,el_stencil_op);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_op_separate";
        el->offset=offsetof(dae_gles2_pass_type_states,el_stencil_op_separate);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_enable";
        el->offset=offsetof(dae_gles2_pass_type_states,el_blend_enable);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_BLEND_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="cull_face_enable";
        el->offset=offsetof(dae_gles2_pass_type_states,el_cull_face_enable);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_CULL_FACE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_test_enable";
        el->offset=offsetof(dae_gles2_pass_type_states,el_depth_test_enable);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_DEPTH_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="dither_enable";
        el->offset=offsetof(dae_gles2_pass_type_states,el_dither_enable);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_DITHER_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset_fill_enable";
        el->offset=offsetof(dae_gles2_pass_type_states,el_polygon_offset_fill_enable);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_FILL_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size_enable";
        el->offset=offsetof(dae_gles2_pass_type_states,el_point_size_enable);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_POINT_SIZE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_alpha_to_coverage_enable";
        el->offset=offsetof(dae_gles2_pass_type_states,el_sample_alpha_to_coverage_enable);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_COVERAGE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_coverage_enable";
        el->offset=offsetof(dae_gles2_pass_type_states,el_sample_coverage_enable);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="scissor_test_enable";
        el->offset=offsetof(dae_gles2_pass_type_states,el_scissor_test_enable);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_SCISSOR_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_test_enable";
        el->offset=offsetof(dae_gles2_pass_type_states,el_stencil_test_enable);
        el->objtypeid=dae_ID_GLES2_PIPELINE_SETTINGS_GROUP_STENCIL_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 32;
    }
    ++t;
    t->name = "gles2_pass_type_evaluate";
    t->size = sizeof(dae_gles2_pass_type_evaluate);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES2_PASS_TYPE_EVALUATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="color_target";
        el->offset=offsetof(dae_gles2_pass_type_evaluate,el_color_target);
        el->objtypeid=dae_ID_FX_COLORTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="depth_target";
        el->offset=offsetof(dae_gles2_pass_type_evaluate,el_depth_target);
        el->objtypeid=dae_ID_FX_DEPTHTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="stencil_target";
        el->offset=offsetof(dae_gles2_pass_type_evaluate,el_stencil_target);
        el->objtypeid=dae_ID_FX_STENCILTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="color_clear";
        el->offset=offsetof(dae_gles2_pass_type_evaluate,el_color_clear);
        el->objtypeid=dae_ID_FX_CLEARCOLOR_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="stencil_clear";
        el->offset=offsetof(dae_gles2_pass_type_evaluate,el_stencil_clear);
        el->objtypeid=dae_ID_FX_CLEARSTENCIL_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        el->name="depth_clear";
        el->offset=offsetof(dae_gles2_pass_type_evaluate,el_depth_clear);
        el->objtypeid=dae_ID_FX_CLEARDEPTH_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=5;
        ++el;
        el->name="draw";
        el->offset=offsetof(dae_gles2_pass_type_evaluate,el_draw);
        el->objtypeid=dae_ID_FX_DRAW_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "profile_gles2_type";
    t->size = sizeof(dae_profile_gles2_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLES2_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_gles2_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="language";
        at->offset=offsetof(dae_profile_gles2_type,at_language);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        at->name="platforms";
        at->offset=offsetof(dae_profile_gles2_type,at_platforms);
        at->objtypeid=dae_ID_LIST_OF_NAMES_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[6];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_gles2_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="include";
        el->offset=offsetof(dae_profile_gles2_type,el_include);
        el->objtypeid=dae_ID_FX_INCLUDE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="code";
        el->offset=offsetof(dae_profile_gles2_type,el_code);
        el->objtypeid=dae_ID_FX_CODE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_profile_gles2_type,el_newparam);
        el->objtypeid=dae_ID_GLES2_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_profile_gles2_type,el_technique);
        el->objtypeid=dae_ID_PROFILE_GLES2_TYPE_TECHNIQUE;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_gles2_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 6;
    }
    ++t;
    t->name = "profile_gles2_type_technique";
    t->size = sizeof(dae_profile_gles2_type_technique);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLES2_TYPE_TECHNIQUE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_gles2_type_technique,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_profile_gles2_type_technique,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_gles2_type_technique,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="annotate";
        el->offset=offsetof(dae_profile_gles2_type_technique,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="pass";
        el->offset=offsetof(dae_profile_gles2_type_technique,el_pass);
        el->objtypeid=dae_ID_GLES2_PASS_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_gles2_type_technique,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "glsl_array_type";
    t->size = sizeof(dae_glsl_array_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLSL_ARRAY_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="length";
        at->offset=offsetof(dae_glsl_array_type,at_length);
        at->objtypeid=dae_ID_POSITIVEINTEGER;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[23];
        dae_obj_memberdef* el = elems;
        el->name="bool";
        el->offset=offsetof(dae_glsl_array_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_glsl_array_type,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_glsl_array_type,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_glsl_array_type,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_glsl_array_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_glsl_array_type,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_glsl_array_type,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_glsl_array_type,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_glsl_array_type,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_glsl_array_type,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_glsl_array_type,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_glsl_array_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_glsl_array_type,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_glsl_array_type,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_glsl_array_type,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler1D";
        el->offset=offsetof(dae_glsl_array_type,el_sampler1D);
        el->objtypeid=dae_ID_FX_SAMPLER1D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_glsl_array_type,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_glsl_array_type,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_glsl_array_type,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerRECT";
        el->offset=offsetof(dae_glsl_array_type,el_samplerRECT);
        el->objtypeid=dae_ID_FX_SAMPLERRECT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_glsl_array_type,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_glsl_array_type,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_glsl_array_type,el_array);
        el->objtypeid=dae_ID_GLSL_ARRAY_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 23;
    }
    ++t;
    t->name = "glsl_newparam_type";
    t->size = sizeof(dae_glsl_newparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLSL_NEWPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_glsl_newparam_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[26];
        dae_obj_memberdef* el = elems;
        el->name="annotate";
        el->offset=offsetof(dae_glsl_newparam_type,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="semantic";
        el->offset=offsetof(dae_glsl_newparam_type,el_semantic);
        el->objtypeid=dae_ID_NCNAME;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="modifier";
        el->offset=offsetof(dae_glsl_newparam_type,el_modifier);
        el->objtypeid=dae_ID_FX_MODIFIER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_glsl_newparam_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_glsl_newparam_type,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_glsl_newparam_type,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_glsl_newparam_type,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_glsl_newparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_glsl_newparam_type,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_glsl_newparam_type,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_glsl_newparam_type,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_glsl_newparam_type,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_glsl_newparam_type,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_glsl_newparam_type,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_glsl_newparam_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_glsl_newparam_type,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_glsl_newparam_type,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_glsl_newparam_type,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler1D";
        el->offset=offsetof(dae_glsl_newparam_type,el_sampler1D);
        el->objtypeid=dae_ID_FX_SAMPLER1D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_glsl_newparam_type,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_glsl_newparam_type,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_glsl_newparam_type,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerRECT";
        el->offset=offsetof(dae_glsl_newparam_type,el_samplerRECT);
        el->objtypeid=dae_ID_FX_SAMPLERRECT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_glsl_newparam_type,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_glsl_newparam_type,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_glsl_newparam_type,el_array);
        el->objtypeid=dae_ID_GLSL_ARRAY_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 26;
    }
    ++t;
    t->name = "glsl_shader_type";
    t->size = sizeof(dae_glsl_shader_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLSL_SHADER_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="stage";
        at->offset=offsetof(dae_glsl_shader_type,at_stage);
        at->objtypeid=dae_ID_FX_PIPELINE_STAGE_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="sources";
        el->offset=offsetof(dae_glsl_shader_type,el_sources);
        el->objtypeid=dae_ID_FX_SOURCES_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_glsl_shader_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "glsl_program_type";
    t->size = sizeof(dae_glsl_program_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLSL_PROGRAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="shader";
        el->offset=offsetof(dae_glsl_program_type,el_shader);
        el->objtypeid=dae_ID_GLSL_SHADER_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="bind_attribute";
        el->offset=offsetof(dae_glsl_program_type,el_bind_attribute);
        el->objtypeid=dae_ID_GLSL_PROGRAM_TYPE_BIND_ATTRIBUTE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="bind_uniform";
        el->offset=offsetof(dae_glsl_program_type,el_bind_uniform);
        el->objtypeid=dae_ID_GLSL_PROGRAM_TYPE_BIND_UNIFORM;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "glsl_program_type_bind_attribute";
    t->size = sizeof(dae_glsl_program_type_bind_attribute);
    t->flags= 0;
    t->objtypeid = dae_ID_GLSL_PROGRAM_TYPE_BIND_ATTRIBUTE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="symbol";
        at->offset=offsetof(dae_glsl_program_type_bind_attribute,at_symbol);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="semantic";
        el->offset=offsetof(dae_glsl_program_type_bind_attribute,el_semantic);
        el->objtypeid=dae_ID_TOKEN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "glsl_program_type_bind_uniform";
    t->size = sizeof(dae_glsl_program_type_bind_uniform);
    t->flags= 0;
    t->objtypeid = dae_ID_GLSL_PROGRAM_TYPE_BIND_UNIFORM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="symbol";
        at->offset=offsetof(dae_glsl_program_type_bind_uniform,at_symbol);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[24];
        dae_obj_memberdef* el = elems;
        el->name="param";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_param);
        el->objtypeid=dae_ID_GLSL_PROGRAM_TYPE_BIND_UNIFORM_PARAM;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler1D";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_sampler1D);
        el->objtypeid=dae_ID_FX_SAMPLER1D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerRECT";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_samplerRECT);
        el->objtypeid=dae_ID_FX_SAMPLERRECT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_glsl_program_type_bind_uniform,el_array);
        el->objtypeid=dae_ID_GLSL_ARRAY_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 24;
    }
    ++t;
    t->name = "glsl_program_type_bind_uniform_param";
    t->size = sizeof(dae_glsl_program_type_bind_uniform_param);
    t->flags= 0;
    t->objtypeid = dae_ID_GLSL_PROGRAM_TYPE_BIND_UNIFORM_PARAM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_glsl_program_type_bind_uniform_param,at_ref);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "profile_glsl_type";
    t->size = sizeof(dae_profile_glsl_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLSL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_glsl_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="platform";
        at->offset=offsetof(dae_profile_glsl_type,at_platform);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        at->deflt="PC";
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[6];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_glsl_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="code";
        el->offset=offsetof(dae_profile_glsl_type,el_code);
        el->objtypeid=dae_ID_FX_CODE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="include";
        el->offset=offsetof(dae_profile_glsl_type,el_include);
        el->objtypeid=dae_ID_FX_INCLUDE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_profile_glsl_type,el_newparam);
        el->objtypeid=dae_ID_GLSL_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_profile_glsl_type,el_technique);
        el->objtypeid=dae_ID_PROFILE_GLSL_TYPE_TECHNIQUE;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_glsl_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 6;
    }
    ++t;
    t->name = "profile_glsl_type_technique";
    t->size = sizeof(dae_profile_glsl_type_technique);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLSL_TYPE_TECHNIQUE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_glsl_type_technique,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_profile_glsl_type_technique,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_glsl_type_technique,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="annotate";
        el->offset=offsetof(dae_profile_glsl_type_technique,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="pass";
        el->offset=offsetof(dae_profile_glsl_type_technique,el_pass);
        el->objtypeid=dae_ID_PROFILE_GLSL_TYPE_TECHNIQUE_PASS;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_glsl_type_technique,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "profile_glsl_type_technique_pass";
    t->size = sizeof(dae_profile_glsl_type_technique_pass);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLSL_TYPE_TECHNIQUE_PASS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_profile_glsl_type_technique_pass,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="annotate";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="states";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass,el_states);
        el->objtypeid=dae_ID_PROFILE_GLSL_TYPE_TECHNIQUE_PASS_STATES;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="program";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass,el_program);
        el->objtypeid=dae_ID_GLSL_PROGRAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="evaluate";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass,el_evaluate);
        el->objtypeid=dae_ID_PROFILE_GLSL_TYPE_TECHNIQUE_PASS_EVALUATE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "profile_glsl_type_technique_pass_states";
    t->size = sizeof(dae_profile_glsl_type_technique_pass_states);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLSL_TYPE_TECHNIQUE_PASS_STATES;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[104];
        dae_obj_memberdef* el = elems;
        el->name="alpha_func";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_alpha_func);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_func";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_blend_func);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_func_separate";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_blend_func_separate);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_equation";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_blend_equation);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_equation_separate";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_blend_equation_separate);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_material";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_color_material);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="cull_face";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_cull_face);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_CULL_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_func";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_depth_func);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_mode";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_fog_mode);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_MODE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_coord_src";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_fog_coord_src);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_COORD_SRC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="front_face";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_front_face);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FRONT_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_model_color_control";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_model_color_control);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_COLOR_CONTROL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="logic_op";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_logic_op);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LOGIC_OP;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_mode";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_polygon_mode);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_MODE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="shade_model";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_shade_model);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SHADE_MODEL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_func";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_stencil_func);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_op";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_stencil_op);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_func_separate";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_stencil_func_separate);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_op_separate";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_stencil_op_separate);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_mask_separate";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_stencil_mask_separate);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_ambient";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_ambient);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_AMBIENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_diffuse";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_diffuse);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_DIFFUSE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_specular";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_specular);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPECULAR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_position";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_position);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_POSITION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_constant_attenuation";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_constant_attenuation);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_CONSTANT_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_linear_attenuation";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_linear_attenuation);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_LINEAR_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_quadratic_attenuation";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_quadratic_attenuation);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_QUADRATIC_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_spot_cutoff";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_spot_cutoff);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_CUTOFF;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_spot_direction";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_spot_direction);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_DIRECTION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_spot_exponent";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_spot_exponent);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_EXPONENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture1D";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_texture1D);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE1D;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture2D";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_texture2D);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE2D;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture3D";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_texture3D);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE3D;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureCUBE";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_textureCUBE);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURECUBE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureRECT";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_textureRECT);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURERECT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureDEPTH";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_textureDEPTH);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTUREDEPTH;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture1D_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_texture1D_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE1D_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture2D_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_texture2D_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE2D_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture3D_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_texture3D_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE3D_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureCUBE_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_textureCUBE_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURECUBE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureRECT_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_textureRECT_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURERECT_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureDEPTH_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_textureDEPTH_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTUREDEPTH_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture_env_color";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_texture_env_color);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE_ENV_COLOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture_env_mode";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_texture_env_mode);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE_ENV_MODE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="clip_plane";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_clip_plane);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_CLIP_PLANE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="clip_plane_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_clip_plane_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_CLIP_PLANE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_color";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_blend_color);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_COLOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_mask";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_color_mask);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_bounds";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_depth_bounds);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_BOUNDS;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_mask";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_depth_mask);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_range";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_depth_range);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_RANGE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_density";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_fog_density);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_DENSITY;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_start";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_fog_start);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_START;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_end";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_fog_end);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_END;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_color";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_fog_color);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_COLOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_model_ambient";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_model_ambient);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_AMBIENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="lighting_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_lighting_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHTING_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_stipple";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_line_stipple);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_STIPPLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_width";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_line_width);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_WIDTH;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_ambient";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_material_ambient);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_AMBIENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_diffuse";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_material_diffuse);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_DIFFUSE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_emission";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_material_emission);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_EMISSION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_shininess";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_material_shininess);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_SHININESS;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_specular";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_material_specular);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_SPECULAR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="model_view_matrix";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_model_view_matrix);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MODEL_VIEW_MATRIX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_distance_attenuation";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_point_distance_attenuation);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_DISTANCE_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_fade_threshold_size";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_point_fade_threshold_size);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_FADE_THRESHOLD_SIZE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_point_size);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SIZE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size_min";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_point_size_min);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SIZE_MIN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size_max";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_point_size_max);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SIZE_MAX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_polygon_offset);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="projection_matrix";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_projection_matrix);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_PROJECTION_MATRIX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="scissor";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_scissor);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SCISSOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_mask";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_stencil_mask);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="alpha_test_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_alpha_test_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_ALPHA_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_blend_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_logic_op_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_color_logic_op_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_LOGIC_OP_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_material_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_color_material_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="cull_face_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_cull_face_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_CULL_FACE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_bounds_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_depth_bounds_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_BOUNDS_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_clamp_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_depth_clamp_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_CLAMP_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_test_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_depth_test_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="dither_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_dither_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DITHER_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_fog_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_model_local_viewer_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_model_local_viewer_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_LOCAL_VIEWER_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_model_two_side_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_light_model_two_side_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_TWO_SIDE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_smooth_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_line_smooth_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_SMOOTH_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_stipple_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_line_stipple_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_STIPPLE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="logic_op_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_logic_op_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LOGIC_OP_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="multisample_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_multisample_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MULTISAMPLE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="normalize_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_normalize_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_NORMALIZE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_smooth_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_point_smooth_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SMOOTH_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset_fill_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_polygon_offset_fill_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_FILL_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset_line_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_polygon_offset_line_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_LINE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset_point_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_polygon_offset_point_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_POINT_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_smooth_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_polygon_smooth_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_SMOOTH_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_stipple_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_polygon_stipple_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_STIPPLE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="rescale_normal_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_rescale_normal_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_RESCALE_NORMAL_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_alpha_to_coverage_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_sample_alpha_to_coverage_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_COVERAGE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_alpha_to_one_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_sample_alpha_to_one_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_ONE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_coverage_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_sample_coverage_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="scissor_test_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_scissor_test_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SCISSOR_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_test_enable";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_states,el_stencil_test_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 104;
    }
    ++t;
    t->name = "profile_glsl_type_technique_pass_evaluate";
    t->size = sizeof(dae_profile_glsl_type_technique_pass_evaluate);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLSL_TYPE_TECHNIQUE_PASS_EVALUATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="color_target";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_evaluate,el_color_target);
        el->objtypeid=dae_ID_FX_COLORTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="depth_target";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_evaluate,el_depth_target);
        el->objtypeid=dae_ID_FX_DEPTHTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="stencil_target";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_evaluate,el_stencil_target);
        el->objtypeid=dae_ID_FX_STENCILTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="color_clear";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_evaluate,el_color_clear);
        el->objtypeid=dae_ID_FX_CLEARCOLOR_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="depth_clear";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_evaluate,el_depth_clear);
        el->objtypeid=dae_ID_FX_CLEARDEPTH_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        el->name="stencil_clear";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_evaluate,el_stencil_clear);
        el->objtypeid=dae_ID_FX_CLEARSTENCIL_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=5;
        ++el;
        el->name="draw";
        el->offset=offsetof(dae_profile_glsl_type_technique_pass_evaluate,el_draw);
        el->objtypeid=dae_ID_FX_DRAW_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "cg_array_type";
    t->size = sizeof(dae_cg_array_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_ARRAY_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="length";
        at->offset=offsetof(dae_cg_array_type,at_length);
        at->objtypeid=dae_ID_POSITIVEINTEGER;
        at->min=1;
        at->max=1;
        ++at;
        at->name="resizable";
        at->offset=offsetof(dae_cg_array_type,at_resizable);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[90];
        dae_obj_memberdef* el = elems;
        el->name="bool";
        el->offset=offsetof(dae_cg_array_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_cg_array_type,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_cg_array_type,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_cg_array_type,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x1";
        el->offset=offsetof(dae_cg_array_type,el_bool2x1);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x2";
        el->offset=offsetof(dae_cg_array_type,el_bool2x2);
        el->objtypeid=dae_ID_BOOL2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x3";
        el->offset=offsetof(dae_cg_array_type,el_bool2x3);
        el->objtypeid=dae_ID_BOOL2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x4";
        el->offset=offsetof(dae_cg_array_type,el_bool2x4);
        el->objtypeid=dae_ID_BOOL2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x1";
        el->offset=offsetof(dae_cg_array_type,el_bool3x1);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x2";
        el->offset=offsetof(dae_cg_array_type,el_bool3x2);
        el->objtypeid=dae_ID_BOOL3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x3";
        el->offset=offsetof(dae_cg_array_type,el_bool3x3);
        el->objtypeid=dae_ID_BOOL3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x4";
        el->offset=offsetof(dae_cg_array_type,el_bool3x4);
        el->objtypeid=dae_ID_BOOL3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x1";
        el->offset=offsetof(dae_cg_array_type,el_bool4x1);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x2";
        el->offset=offsetof(dae_cg_array_type,el_bool4x2);
        el->objtypeid=dae_ID_BOOL4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x3";
        el->offset=offsetof(dae_cg_array_type,el_bool4x3);
        el->objtypeid=dae_ID_BOOL4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x4";
        el->offset=offsetof(dae_cg_array_type,el_bool4x4);
        el->objtypeid=dae_ID_BOOL4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_cg_array_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_cg_array_type,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_cg_array_type,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_cg_array_type,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x1";
        el->offset=offsetof(dae_cg_array_type,el_float2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_cg_array_type,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x3";
        el->offset=offsetof(dae_cg_array_type,el_float2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x4";
        el->offset=offsetof(dae_cg_array_type,el_float2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x1";
        el->offset=offsetof(dae_cg_array_type,el_float3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x2";
        el->offset=offsetof(dae_cg_array_type,el_float3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_cg_array_type,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x4";
        el->offset=offsetof(dae_cg_array_type,el_float3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x1";
        el->offset=offsetof(dae_cg_array_type,el_float4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x2";
        el->offset=offsetof(dae_cg_array_type,el_float4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x3";
        el->offset=offsetof(dae_cg_array_type,el_float4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_cg_array_type,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_cg_array_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_cg_array_type,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_cg_array_type,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_cg_array_type,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x1";
        el->offset=offsetof(dae_cg_array_type,el_int2x1);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x2";
        el->offset=offsetof(dae_cg_array_type,el_int2x2);
        el->objtypeid=dae_ID_INT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x3";
        el->offset=offsetof(dae_cg_array_type,el_int2x3);
        el->objtypeid=dae_ID_INT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x4";
        el->offset=offsetof(dae_cg_array_type,el_int2x4);
        el->objtypeid=dae_ID_INT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x1";
        el->offset=offsetof(dae_cg_array_type,el_int3x1);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x2";
        el->offset=offsetof(dae_cg_array_type,el_int3x2);
        el->objtypeid=dae_ID_INT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x3";
        el->offset=offsetof(dae_cg_array_type,el_int3x3);
        el->objtypeid=dae_ID_INT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x4";
        el->offset=offsetof(dae_cg_array_type,el_int3x4);
        el->objtypeid=dae_ID_INT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x1";
        el->offset=offsetof(dae_cg_array_type,el_int4x1);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x2";
        el->offset=offsetof(dae_cg_array_type,el_int4x2);
        el->objtypeid=dae_ID_INT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x3";
        el->offset=offsetof(dae_cg_array_type,el_int4x3);
        el->objtypeid=dae_ID_INT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x4";
        el->offset=offsetof(dae_cg_array_type,el_int4x4);
        el->objtypeid=dae_ID_INT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half";
        el->offset=offsetof(dae_cg_array_type,el_half);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2";
        el->offset=offsetof(dae_cg_array_type,el_half2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3";
        el->offset=offsetof(dae_cg_array_type,el_half3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4";
        el->offset=offsetof(dae_cg_array_type,el_half4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x1";
        el->offset=offsetof(dae_cg_array_type,el_half2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x2";
        el->offset=offsetof(dae_cg_array_type,el_half2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x3";
        el->offset=offsetof(dae_cg_array_type,el_half2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x4";
        el->offset=offsetof(dae_cg_array_type,el_half2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x1";
        el->offset=offsetof(dae_cg_array_type,el_half3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x2";
        el->offset=offsetof(dae_cg_array_type,el_half3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x3";
        el->offset=offsetof(dae_cg_array_type,el_half3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x4";
        el->offset=offsetof(dae_cg_array_type,el_half3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x1";
        el->offset=offsetof(dae_cg_array_type,el_half4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x2";
        el->offset=offsetof(dae_cg_array_type,el_half4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x3";
        el->offset=offsetof(dae_cg_array_type,el_half4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x4";
        el->offset=offsetof(dae_cg_array_type,el_half4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed";
        el->offset=offsetof(dae_cg_array_type,el_fixed);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2";
        el->offset=offsetof(dae_cg_array_type,el_fixed2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3";
        el->offset=offsetof(dae_cg_array_type,el_fixed3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4";
        el->offset=offsetof(dae_cg_array_type,el_fixed4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x1";
        el->offset=offsetof(dae_cg_array_type,el_fixed2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x2";
        el->offset=offsetof(dae_cg_array_type,el_fixed2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x3";
        el->offset=offsetof(dae_cg_array_type,el_fixed2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x4";
        el->offset=offsetof(dae_cg_array_type,el_fixed2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x1";
        el->offset=offsetof(dae_cg_array_type,el_fixed3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x2";
        el->offset=offsetof(dae_cg_array_type,el_fixed3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x3";
        el->offset=offsetof(dae_cg_array_type,el_fixed3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x4";
        el->offset=offsetof(dae_cg_array_type,el_fixed3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x1";
        el->offset=offsetof(dae_cg_array_type,el_fixed4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x2";
        el->offset=offsetof(dae_cg_array_type,el_fixed4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x3";
        el->offset=offsetof(dae_cg_array_type,el_fixed4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x4";
        el->offset=offsetof(dae_cg_array_type,el_fixed4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler1D";
        el->offset=offsetof(dae_cg_array_type,el_sampler1D);
        el->objtypeid=dae_ID_FX_SAMPLER1D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_cg_array_type,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_cg_array_type,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerRECT";
        el->offset=offsetof(dae_cg_array_type,el_samplerRECT);
        el->objtypeid=dae_ID_FX_SAMPLERRECT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_cg_array_type,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_cg_array_type,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="string";
        el->offset=offsetof(dae_cg_array_type,el_string);
        el->objtypeid=dae_ID_STRING;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_cg_array_type,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_cg_array_type,el_array);
        el->objtypeid=dae_ID_CG_ARRAY_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="usertype";
        el->offset=offsetof(dae_cg_array_type,el_usertype);
        el->objtypeid=dae_ID_CG_USER_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 90;
    }
    ++t;
    t->name = "cg_setparam_type";
    t->size = sizeof(dae_cg_setparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_SETPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_cg_setparam_type,at_ref);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[90];
        dae_obj_memberdef* el = elems;
        el->name="bool";
        el->offset=offsetof(dae_cg_setparam_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_cg_setparam_type,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_cg_setparam_type,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_cg_setparam_type,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x1";
        el->offset=offsetof(dae_cg_setparam_type,el_bool2x1);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x2";
        el->offset=offsetof(dae_cg_setparam_type,el_bool2x2);
        el->objtypeid=dae_ID_BOOL2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x3";
        el->offset=offsetof(dae_cg_setparam_type,el_bool2x3);
        el->objtypeid=dae_ID_BOOL2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x4";
        el->offset=offsetof(dae_cg_setparam_type,el_bool2x4);
        el->objtypeid=dae_ID_BOOL2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x1";
        el->offset=offsetof(dae_cg_setparam_type,el_bool3x1);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x2";
        el->offset=offsetof(dae_cg_setparam_type,el_bool3x2);
        el->objtypeid=dae_ID_BOOL3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x3";
        el->offset=offsetof(dae_cg_setparam_type,el_bool3x3);
        el->objtypeid=dae_ID_BOOL3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x4";
        el->offset=offsetof(dae_cg_setparam_type,el_bool3x4);
        el->objtypeid=dae_ID_BOOL3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x1";
        el->offset=offsetof(dae_cg_setparam_type,el_bool4x1);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x2";
        el->offset=offsetof(dae_cg_setparam_type,el_bool4x2);
        el->objtypeid=dae_ID_BOOL4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x3";
        el->offset=offsetof(dae_cg_setparam_type,el_bool4x3);
        el->objtypeid=dae_ID_BOOL4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x4";
        el->offset=offsetof(dae_cg_setparam_type,el_bool4x4);
        el->objtypeid=dae_ID_BOOL4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_cg_setparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_cg_setparam_type,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_cg_setparam_type,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_cg_setparam_type,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x1";
        el->offset=offsetof(dae_cg_setparam_type,el_float2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_cg_setparam_type,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x3";
        el->offset=offsetof(dae_cg_setparam_type,el_float2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x4";
        el->offset=offsetof(dae_cg_setparam_type,el_float2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x1";
        el->offset=offsetof(dae_cg_setparam_type,el_float3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x2";
        el->offset=offsetof(dae_cg_setparam_type,el_float3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_cg_setparam_type,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x4";
        el->offset=offsetof(dae_cg_setparam_type,el_float3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x1";
        el->offset=offsetof(dae_cg_setparam_type,el_float4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x2";
        el->offset=offsetof(dae_cg_setparam_type,el_float4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x3";
        el->offset=offsetof(dae_cg_setparam_type,el_float4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_cg_setparam_type,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_cg_setparam_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_cg_setparam_type,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_cg_setparam_type,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_cg_setparam_type,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x1";
        el->offset=offsetof(dae_cg_setparam_type,el_int2x1);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x2";
        el->offset=offsetof(dae_cg_setparam_type,el_int2x2);
        el->objtypeid=dae_ID_INT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x3";
        el->offset=offsetof(dae_cg_setparam_type,el_int2x3);
        el->objtypeid=dae_ID_INT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x4";
        el->offset=offsetof(dae_cg_setparam_type,el_int2x4);
        el->objtypeid=dae_ID_INT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x1";
        el->offset=offsetof(dae_cg_setparam_type,el_int3x1);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x2";
        el->offset=offsetof(dae_cg_setparam_type,el_int3x2);
        el->objtypeid=dae_ID_INT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x3";
        el->offset=offsetof(dae_cg_setparam_type,el_int3x3);
        el->objtypeid=dae_ID_INT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x4";
        el->offset=offsetof(dae_cg_setparam_type,el_int3x4);
        el->objtypeid=dae_ID_INT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x1";
        el->offset=offsetof(dae_cg_setparam_type,el_int4x1);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x2";
        el->offset=offsetof(dae_cg_setparam_type,el_int4x2);
        el->objtypeid=dae_ID_INT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x3";
        el->offset=offsetof(dae_cg_setparam_type,el_int4x3);
        el->objtypeid=dae_ID_INT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x4";
        el->offset=offsetof(dae_cg_setparam_type,el_int4x4);
        el->objtypeid=dae_ID_INT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half";
        el->offset=offsetof(dae_cg_setparam_type,el_half);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2";
        el->offset=offsetof(dae_cg_setparam_type,el_half2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3";
        el->offset=offsetof(dae_cg_setparam_type,el_half3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4";
        el->offset=offsetof(dae_cg_setparam_type,el_half4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x1";
        el->offset=offsetof(dae_cg_setparam_type,el_half2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x2";
        el->offset=offsetof(dae_cg_setparam_type,el_half2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x3";
        el->offset=offsetof(dae_cg_setparam_type,el_half2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x4";
        el->offset=offsetof(dae_cg_setparam_type,el_half2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x1";
        el->offset=offsetof(dae_cg_setparam_type,el_half3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x2";
        el->offset=offsetof(dae_cg_setparam_type,el_half3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x3";
        el->offset=offsetof(dae_cg_setparam_type,el_half3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x4";
        el->offset=offsetof(dae_cg_setparam_type,el_half3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x1";
        el->offset=offsetof(dae_cg_setparam_type,el_half4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x2";
        el->offset=offsetof(dae_cg_setparam_type,el_half4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x3";
        el->offset=offsetof(dae_cg_setparam_type,el_half4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x4";
        el->offset=offsetof(dae_cg_setparam_type,el_half4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x1";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x2";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x3";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x4";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x1";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x2";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x3";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x4";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x1";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x2";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x3";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x4";
        el->offset=offsetof(dae_cg_setparam_type,el_fixed4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler1D";
        el->offset=offsetof(dae_cg_setparam_type,el_sampler1D);
        el->objtypeid=dae_ID_FX_SAMPLER1D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_cg_setparam_type,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_cg_setparam_type,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerRECT";
        el->offset=offsetof(dae_cg_setparam_type,el_samplerRECT);
        el->objtypeid=dae_ID_FX_SAMPLERRECT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_cg_setparam_type,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_cg_setparam_type,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="string";
        el->offset=offsetof(dae_cg_setparam_type,el_string);
        el->objtypeid=dae_ID_STRING;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_cg_setparam_type,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_cg_setparam_type,el_array);
        el->objtypeid=dae_ID_CG_ARRAY_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="usertype";
        el->offset=offsetof(dae_cg_setparam_type,el_usertype);
        el->objtypeid=dae_ID_CG_USER_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 90;
    }
    ++t;
    t->name = "cg_user_type";
    t->size = sizeof(dae_cg_user_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_USER_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="typename";
        at->offset=offsetof(dae_cg_user_type,at_typename);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        at->name="source";
        at->offset=offsetof(dae_cg_user_type,at_source);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="setparam";
        el->offset=offsetof(dae_cg_user_type,el_setparam);
        el->objtypeid=dae_ID_CG_SETPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "cg_newparam_type";
    t->size = sizeof(dae_cg_newparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_NEWPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_cg_newparam_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[93];
        dae_obj_memberdef* el = elems;
        el->name="annotate";
        el->offset=offsetof(dae_cg_newparam_type,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="semantic";
        el->offset=offsetof(dae_cg_newparam_type,el_semantic);
        el->objtypeid=dae_ID_NCNAME;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="modifier";
        el->offset=offsetof(dae_cg_newparam_type,el_modifier);
        el->objtypeid=dae_ID_FX_MODIFIER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_cg_newparam_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_cg_newparam_type,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_cg_newparam_type,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_cg_newparam_type,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool2x1";
        el->offset=offsetof(dae_cg_newparam_type,el_bool2x1);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool2x2";
        el->offset=offsetof(dae_cg_newparam_type,el_bool2x2);
        el->objtypeid=dae_ID_BOOL2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool2x3";
        el->offset=offsetof(dae_cg_newparam_type,el_bool2x3);
        el->objtypeid=dae_ID_BOOL2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool2x4";
        el->offset=offsetof(dae_cg_newparam_type,el_bool2x4);
        el->objtypeid=dae_ID_BOOL2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool3x1";
        el->offset=offsetof(dae_cg_newparam_type,el_bool3x1);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool3x2";
        el->offset=offsetof(dae_cg_newparam_type,el_bool3x2);
        el->objtypeid=dae_ID_BOOL3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool3x3";
        el->offset=offsetof(dae_cg_newparam_type,el_bool3x3);
        el->objtypeid=dae_ID_BOOL3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool3x4";
        el->offset=offsetof(dae_cg_newparam_type,el_bool3x4);
        el->objtypeid=dae_ID_BOOL3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool4x1";
        el->offset=offsetof(dae_cg_newparam_type,el_bool4x1);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool4x2";
        el->offset=offsetof(dae_cg_newparam_type,el_bool4x2);
        el->objtypeid=dae_ID_BOOL4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool4x3";
        el->offset=offsetof(dae_cg_newparam_type,el_bool4x3);
        el->objtypeid=dae_ID_BOOL4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool4x4";
        el->offset=offsetof(dae_cg_newparam_type,el_bool4x4);
        el->objtypeid=dae_ID_BOOL4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_cg_newparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_cg_newparam_type,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_cg_newparam_type,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_cg_newparam_type,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x1";
        el->offset=offsetof(dae_cg_newparam_type,el_float2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_cg_newparam_type,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x3";
        el->offset=offsetof(dae_cg_newparam_type,el_float2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x4";
        el->offset=offsetof(dae_cg_newparam_type,el_float2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x1";
        el->offset=offsetof(dae_cg_newparam_type,el_float3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x2";
        el->offset=offsetof(dae_cg_newparam_type,el_float3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_cg_newparam_type,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x4";
        el->offset=offsetof(dae_cg_newparam_type,el_float3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x1";
        el->offset=offsetof(dae_cg_newparam_type,el_float4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x2";
        el->offset=offsetof(dae_cg_newparam_type,el_float4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x3";
        el->offset=offsetof(dae_cg_newparam_type,el_float4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_cg_newparam_type,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_cg_newparam_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_cg_newparam_type,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_cg_newparam_type,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_cg_newparam_type,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int2x1";
        el->offset=offsetof(dae_cg_newparam_type,el_int2x1);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int2x2";
        el->offset=offsetof(dae_cg_newparam_type,el_int2x2);
        el->objtypeid=dae_ID_INT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int2x3";
        el->offset=offsetof(dae_cg_newparam_type,el_int2x3);
        el->objtypeid=dae_ID_INT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int2x4";
        el->offset=offsetof(dae_cg_newparam_type,el_int2x4);
        el->objtypeid=dae_ID_INT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int3x1";
        el->offset=offsetof(dae_cg_newparam_type,el_int3x1);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int3x2";
        el->offset=offsetof(dae_cg_newparam_type,el_int3x2);
        el->objtypeid=dae_ID_INT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int3x3";
        el->offset=offsetof(dae_cg_newparam_type,el_int3x3);
        el->objtypeid=dae_ID_INT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int3x4";
        el->offset=offsetof(dae_cg_newparam_type,el_int3x4);
        el->objtypeid=dae_ID_INT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int4x1";
        el->offset=offsetof(dae_cg_newparam_type,el_int4x1);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int4x2";
        el->offset=offsetof(dae_cg_newparam_type,el_int4x2);
        el->objtypeid=dae_ID_INT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int4x3";
        el->offset=offsetof(dae_cg_newparam_type,el_int4x3);
        el->objtypeid=dae_ID_INT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int4x4";
        el->offset=offsetof(dae_cg_newparam_type,el_int4x4);
        el->objtypeid=dae_ID_INT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half";
        el->offset=offsetof(dae_cg_newparam_type,el_half);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half2";
        el->offset=offsetof(dae_cg_newparam_type,el_half2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half3";
        el->offset=offsetof(dae_cg_newparam_type,el_half3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half4";
        el->offset=offsetof(dae_cg_newparam_type,el_half4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half2x1";
        el->offset=offsetof(dae_cg_newparam_type,el_half2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half2x2";
        el->offset=offsetof(dae_cg_newparam_type,el_half2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half2x3";
        el->offset=offsetof(dae_cg_newparam_type,el_half2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half2x4";
        el->offset=offsetof(dae_cg_newparam_type,el_half2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half3x1";
        el->offset=offsetof(dae_cg_newparam_type,el_half3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half3x2";
        el->offset=offsetof(dae_cg_newparam_type,el_half3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half3x3";
        el->offset=offsetof(dae_cg_newparam_type,el_half3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half3x4";
        el->offset=offsetof(dae_cg_newparam_type,el_half3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half4x1";
        el->offset=offsetof(dae_cg_newparam_type,el_half4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half4x2";
        el->offset=offsetof(dae_cg_newparam_type,el_half4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half4x3";
        el->offset=offsetof(dae_cg_newparam_type,el_half4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="half4x4";
        el->offset=offsetof(dae_cg_newparam_type,el_half4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed2";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed3";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed4";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed2x1";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed2x2";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed2x3";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed2x4";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed3x1";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed3x2";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed3x3";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed3x4";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed4x1";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed4x2";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed4x3";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="fixed4x4";
        el->offset=offsetof(dae_cg_newparam_type,el_fixed4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler1D";
        el->offset=offsetof(dae_cg_newparam_type,el_sampler1D);
        el->objtypeid=dae_ID_FX_SAMPLER1D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_cg_newparam_type,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_cg_newparam_type,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerRECT";
        el->offset=offsetof(dae_cg_newparam_type,el_samplerRECT);
        el->objtypeid=dae_ID_FX_SAMPLERRECT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_cg_newparam_type,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_cg_newparam_type,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="string";
        el->offset=offsetof(dae_cg_newparam_type,el_string);
        el->objtypeid=dae_ID_STRING;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_cg_newparam_type,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_cg_newparam_type,el_array);
        el->objtypeid=dae_ID_CG_ARRAY_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="usertype";
        el->offset=offsetof(dae_cg_newparam_type,el_usertype);
        el->objtypeid=dae_ID_CG_USER_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 93;
    }
    ++t;
    t->name = "cg_pass_type";
    t->size = sizeof(dae_cg_pass_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_PASS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_cg_pass_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="annotate";
        el->offset=offsetof(dae_cg_pass_type,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="states";
        el->offset=offsetof(dae_cg_pass_type,el_states);
        el->objtypeid=dae_ID_CG_PASS_TYPE_STATES;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="program";
        el->offset=offsetof(dae_cg_pass_type,el_program);
        el->objtypeid=dae_ID_CG_PASS_TYPE_PROGRAM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="evaluate";
        el->offset=offsetof(dae_cg_pass_type,el_evaluate);
        el->objtypeid=dae_ID_CG_PASS_TYPE_EVALUATE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_cg_pass_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "cg_pass_type_states";
    t->size = sizeof(dae_cg_pass_type_states);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_PASS_TYPE_STATES;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[104];
        dae_obj_memberdef* el = elems;
        el->name="alpha_func";
        el->offset=offsetof(dae_cg_pass_type_states,el_alpha_func);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_func";
        el->offset=offsetof(dae_cg_pass_type_states,el_blend_func);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_func_separate";
        el->offset=offsetof(dae_cg_pass_type_states,el_blend_func_separate);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_equation";
        el->offset=offsetof(dae_cg_pass_type_states,el_blend_equation);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_equation_separate";
        el->offset=offsetof(dae_cg_pass_type_states,el_blend_equation_separate);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_EQUATION_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_material";
        el->offset=offsetof(dae_cg_pass_type_states,el_color_material);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="cull_face";
        el->offset=offsetof(dae_cg_pass_type_states,el_cull_face);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_CULL_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_func";
        el->offset=offsetof(dae_cg_pass_type_states,el_depth_func);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_mode";
        el->offset=offsetof(dae_cg_pass_type_states,el_fog_mode);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_MODE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_coord_src";
        el->offset=offsetof(dae_cg_pass_type_states,el_fog_coord_src);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_COORD_SRC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="front_face";
        el->offset=offsetof(dae_cg_pass_type_states,el_front_face);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FRONT_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_model_color_control";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_model_color_control);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_COLOR_CONTROL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="logic_op";
        el->offset=offsetof(dae_cg_pass_type_states,el_logic_op);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LOGIC_OP;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_mode";
        el->offset=offsetof(dae_cg_pass_type_states,el_polygon_mode);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_MODE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="shade_model";
        el->offset=offsetof(dae_cg_pass_type_states,el_shade_model);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SHADE_MODEL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_func";
        el->offset=offsetof(dae_cg_pass_type_states,el_stencil_func);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_op";
        el->offset=offsetof(dae_cg_pass_type_states,el_stencil_op);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_func_separate";
        el->offset=offsetof(dae_cg_pass_type_states,el_stencil_func_separate);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_op_separate";
        el->offset=offsetof(dae_cg_pass_type_states,el_stencil_op_separate);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_OP_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_mask_separate";
        el->offset=offsetof(dae_cg_pass_type_states,el_stencil_mask_separate);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_MASK_SEPARATE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_ambient";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_ambient);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_AMBIENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_diffuse";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_diffuse);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_DIFFUSE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_specular";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_specular);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPECULAR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_position";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_position);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_POSITION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_constant_attenuation";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_constant_attenuation);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_CONSTANT_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_linear_attenuation";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_linear_attenuation);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_LINEAR_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_quadratic_attenuation";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_quadratic_attenuation);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_QUADRATIC_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_spot_cutoff";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_spot_cutoff);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_CUTOFF;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_spot_direction";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_spot_direction);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_DIRECTION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_spot_exponent";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_spot_exponent);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_EXPONENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture1D";
        el->offset=offsetof(dae_cg_pass_type_states,el_texture1D);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE1D;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture2D";
        el->offset=offsetof(dae_cg_pass_type_states,el_texture2D);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE2D;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture3D";
        el->offset=offsetof(dae_cg_pass_type_states,el_texture3D);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE3D;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureCUBE";
        el->offset=offsetof(dae_cg_pass_type_states,el_textureCUBE);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURECUBE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureRECT";
        el->offset=offsetof(dae_cg_pass_type_states,el_textureRECT);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURERECT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureDEPTH";
        el->offset=offsetof(dae_cg_pass_type_states,el_textureDEPTH);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTUREDEPTH;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture1D_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_texture1D_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE1D_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture2D_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_texture2D_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE2D_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture3D_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_texture3D_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE3D_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureCUBE_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_textureCUBE_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURECUBE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureRECT_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_textureRECT_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURERECT_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="textureDEPTH_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_textureDEPTH_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTUREDEPTH_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture_env_color";
        el->offset=offsetof(dae_cg_pass_type_states,el_texture_env_color);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE_ENV_COLOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture_env_mode";
        el->offset=offsetof(dae_cg_pass_type_states,el_texture_env_mode);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_TEXTURE_ENV_MODE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="clip_plane";
        el->offset=offsetof(dae_cg_pass_type_states,el_clip_plane);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_CLIP_PLANE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="clip_plane_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_clip_plane_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_CLIP_PLANE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_color";
        el->offset=offsetof(dae_cg_pass_type_states,el_blend_color);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_COLOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_mask";
        el->offset=offsetof(dae_cg_pass_type_states,el_color_mask);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_bounds";
        el->offset=offsetof(dae_cg_pass_type_states,el_depth_bounds);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_BOUNDS;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_mask";
        el->offset=offsetof(dae_cg_pass_type_states,el_depth_mask);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_range";
        el->offset=offsetof(dae_cg_pass_type_states,el_depth_range);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_RANGE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_density";
        el->offset=offsetof(dae_cg_pass_type_states,el_fog_density);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_DENSITY;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_start";
        el->offset=offsetof(dae_cg_pass_type_states,el_fog_start);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_START;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_end";
        el->offset=offsetof(dae_cg_pass_type_states,el_fog_end);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_END;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_color";
        el->offset=offsetof(dae_cg_pass_type_states,el_fog_color);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_COLOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_model_ambient";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_model_ambient);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_AMBIENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="lighting_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_lighting_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHTING_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_stipple";
        el->offset=offsetof(dae_cg_pass_type_states,el_line_stipple);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_STIPPLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_width";
        el->offset=offsetof(dae_cg_pass_type_states,el_line_width);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_WIDTH;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_ambient";
        el->offset=offsetof(dae_cg_pass_type_states,el_material_ambient);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_AMBIENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_diffuse";
        el->offset=offsetof(dae_cg_pass_type_states,el_material_diffuse);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_DIFFUSE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_emission";
        el->offset=offsetof(dae_cg_pass_type_states,el_material_emission);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_EMISSION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_shininess";
        el->offset=offsetof(dae_cg_pass_type_states,el_material_shininess);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_SHININESS;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_specular";
        el->offset=offsetof(dae_cg_pass_type_states,el_material_specular);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MATERIAL_SPECULAR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="model_view_matrix";
        el->offset=offsetof(dae_cg_pass_type_states,el_model_view_matrix);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MODEL_VIEW_MATRIX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_distance_attenuation";
        el->offset=offsetof(dae_cg_pass_type_states,el_point_distance_attenuation);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_DISTANCE_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_fade_threshold_size";
        el->offset=offsetof(dae_cg_pass_type_states,el_point_fade_threshold_size);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_FADE_THRESHOLD_SIZE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size";
        el->offset=offsetof(dae_cg_pass_type_states,el_point_size);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SIZE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size_min";
        el->offset=offsetof(dae_cg_pass_type_states,el_point_size_min);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SIZE_MIN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size_max";
        el->offset=offsetof(dae_cg_pass_type_states,el_point_size_max);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SIZE_MAX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset";
        el->offset=offsetof(dae_cg_pass_type_states,el_polygon_offset);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="projection_matrix";
        el->offset=offsetof(dae_cg_pass_type_states,el_projection_matrix);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_PROJECTION_MATRIX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="scissor";
        el->offset=offsetof(dae_cg_pass_type_states,el_scissor);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SCISSOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_mask";
        el->offset=offsetof(dae_cg_pass_type_states,el_stencil_mask);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="alpha_test_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_alpha_test_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_ALPHA_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_blend_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_BLEND_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_logic_op_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_color_logic_op_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_LOGIC_OP_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_material_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_color_material_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="cull_face_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_cull_face_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_CULL_FACE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_bounds_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_depth_bounds_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_BOUNDS_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_clamp_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_depth_clamp_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_CLAMP_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_test_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_depth_test_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DEPTH_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="dither_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_dither_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_DITHER_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_fog_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_FOG_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_model_local_viewer_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_model_local_viewer_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_LOCAL_VIEWER_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_model_two_side_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_light_model_two_side_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_TWO_SIDE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_smooth_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_line_smooth_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_SMOOTH_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_stipple_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_line_stipple_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LINE_STIPPLE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="logic_op_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_logic_op_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_LOGIC_OP_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="multisample_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_multisample_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_MULTISAMPLE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="normalize_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_normalize_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_NORMALIZE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_smooth_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_point_smooth_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POINT_SMOOTH_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset_fill_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_polygon_offset_fill_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_FILL_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset_line_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_polygon_offset_line_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_LINE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset_point_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_polygon_offset_point_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_POINT_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_smooth_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_polygon_smooth_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_SMOOTH_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_stipple_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_polygon_stipple_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_POLYGON_STIPPLE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="rescale_normal_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_rescale_normal_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_RESCALE_NORMAL_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_alpha_to_coverage_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_sample_alpha_to_coverage_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_COVERAGE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_alpha_to_one_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_sample_alpha_to_one_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_ONE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_coverage_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_sample_coverage_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="scissor_test_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_scissor_test_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_SCISSOR_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_test_enable";
        el->offset=offsetof(dae_cg_pass_type_states,el_stencil_test_enable);
        el->objtypeid=dae_ID_GL_PIPELINE_SETTINGS_GROUP_STENCIL_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 104;
    }
    ++t;
    t->name = "cg_pass_type_program";
    t->size = sizeof(dae_cg_pass_type_program);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_PASS_TYPE_PROGRAM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="shader";
        el->offset=offsetof(dae_cg_pass_type_program,el_shader);
        el->objtypeid=dae_ID_CG_PASS_TYPE_PROGRAM_SHADER;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "cg_pass_type_program_shader";
    t->size = sizeof(dae_cg_pass_type_program_shader);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_PASS_TYPE_PROGRAM_SHADER;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="stage";
        at->offset=offsetof(dae_cg_pass_type_program_shader,at_stage);
        at->objtypeid=dae_ID_FX_PIPELINE_STAGE_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="sources";
        el->offset=offsetof(dae_cg_pass_type_program_shader,el_sources);
        el->objtypeid=dae_ID_CG_PASS_TYPE_PROGRAM_SHADER_SOURCES;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="compiler";
        el->offset=offsetof(dae_cg_pass_type_program_shader,el_compiler);
        el->objtypeid=dae_ID_FX_TARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="bind_uniform";
        el->offset=offsetof(dae_cg_pass_type_program_shader,el_bind_uniform);
        el->objtypeid=dae_ID_CG_PASS_TYPE_PROGRAM_SHADER_BIND_UNIFORM;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "cg_pass_type_program_shader_sources";
    t->size = sizeof(dae_cg_pass_type_program_shader_sources);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_PASS_TYPE_PROGRAM_SHADER_SOURCES;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="entry";
        at->offset=offsetof(dae_cg_pass_type_program_shader_sources,at_entry);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="inline";
        el->offset=offsetof(dae_cg_pass_type_program_shader_sources,el_inline);
        el->objtypeid=dae_ID_STRING;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="import";
        el->offset=offsetof(dae_cg_pass_type_program_shader_sources,el_import);
        el->objtypeid=dae_ID_FX_SOURCES_TYPE_IMPORT;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "cg_pass_type_program_shader_bind_uniform";
    t->size = sizeof(dae_cg_pass_type_program_shader_bind_uniform);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_PASS_TYPE_PROGRAM_SHADER_BIND_UNIFORM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="symbol";
        at->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,at_symbol);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[91];
        dae_obj_memberdef* el = elems;
        el->name="param";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_param);
        el->objtypeid=dae_ID_CG_PASS_TYPE_PROGRAM_SHADER_BIND_UNIFORM_PARAM;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool2x1);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool2x2);
        el->objtypeid=dae_ID_BOOL2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool2x3);
        el->objtypeid=dae_ID_BOOL2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool2x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool2x4);
        el->objtypeid=dae_ID_BOOL2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool3x1);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool3x2);
        el->objtypeid=dae_ID_BOOL3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool3x3);
        el->objtypeid=dae_ID_BOOL3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool3x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool3x4);
        el->objtypeid=dae_ID_BOOL3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool4x1);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool4x2);
        el->objtypeid=dae_ID_BOOL4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool4x3);
        el->objtypeid=dae_ID_BOOL4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool4x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_bool4x4);
        el->objtypeid=dae_ID_BOOL4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float2x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float3x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int2x1);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int2x2);
        el->objtypeid=dae_ID_INT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int2x3);
        el->objtypeid=dae_ID_INT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int2x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int2x4);
        el->objtypeid=dae_ID_INT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int3x1);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int3x2);
        el->objtypeid=dae_ID_INT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int3x3);
        el->objtypeid=dae_ID_INT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int3x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int3x4);
        el->objtypeid=dae_ID_INT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int4x1);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int4x2);
        el->objtypeid=dae_ID_INT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int4x3);
        el->objtypeid=dae_ID_INT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int4x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_int4x4);
        el->objtypeid=dae_ID_INT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half2x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half3x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="half4x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_half4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed2x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed3x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x1";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x2";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x3";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fixed4x4";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_fixed4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler1D";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_sampler1D);
        el->objtypeid=dae_ID_FX_SAMPLER1D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_sampler2D);
        el->objtypeid=dae_ID_FX_SAMPLER2D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sampler3D";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_sampler3D);
        el->objtypeid=dae_ID_FX_SAMPLER3D_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerRECT";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_samplerRECT);
        el->objtypeid=dae_ID_FX_SAMPLERRECT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerCUBE";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_samplerCUBE);
        el->objtypeid=dae_ID_FX_SAMPLERCUBE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="samplerDEPTH";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_samplerDEPTH);
        el->objtypeid=dae_ID_FX_SAMPLERDEPTH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="string";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_string);
        el->objtypeid=dae_ID_STRING;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_enum);
        el->objtypeid=dae_ID_GL_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="array";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_array);
        el->objtypeid=dae_ID_CG_ARRAY_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="usertype";
        el->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform,el_usertype);
        el->objtypeid=dae_ID_CG_USER_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 91;
    }
    ++t;
    t->name = "cg_pass_type_program_shader_bind_uniform_param";
    t->size = sizeof(dae_cg_pass_type_program_shader_bind_uniform_param);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_PASS_TYPE_PROGRAM_SHADER_BIND_UNIFORM_PARAM;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_cg_pass_type_program_shader_bind_uniform_param,at_ref);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "cg_pass_type_evaluate";
    t->size = sizeof(dae_cg_pass_type_evaluate);
    t->flags= 0;
    t->objtypeid = dae_ID_CG_PASS_TYPE_EVALUATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="color_target";
        el->offset=offsetof(dae_cg_pass_type_evaluate,el_color_target);
        el->objtypeid=dae_ID_FX_COLORTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="depth_target";
        el->offset=offsetof(dae_cg_pass_type_evaluate,el_depth_target);
        el->objtypeid=dae_ID_FX_DEPTHTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="stencil_target";
        el->offset=offsetof(dae_cg_pass_type_evaluate,el_stencil_target);
        el->objtypeid=dae_ID_FX_STENCILTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="color_clear";
        el->offset=offsetof(dae_cg_pass_type_evaluate,el_color_clear);
        el->objtypeid=dae_ID_FX_CLEARCOLOR_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="depth_clear";
        el->offset=offsetof(dae_cg_pass_type_evaluate,el_depth_clear);
        el->objtypeid=dae_ID_FX_CLEARDEPTH_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        el->name="stencil_clear";
        el->offset=offsetof(dae_cg_pass_type_evaluate,el_stencil_clear);
        el->objtypeid=dae_ID_FX_CLEARSTENCIL_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=5;
        ++el;
        el->name="draw";
        el->offset=offsetof(dae_cg_pass_type_evaluate,el_draw);
        el->objtypeid=dae_ID_FX_DRAW_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "profile_cg_type";
    t->size = sizeof(dae_profile_cg_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_CG_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_cg_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="platform";
        at->offset=offsetof(dae_profile_cg_type,at_platform);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        at->deflt="PC";
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[6];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_cg_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="code";
        el->offset=offsetof(dae_profile_cg_type,el_code);
        el->objtypeid=dae_ID_FX_CODE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="include";
        el->offset=offsetof(dae_profile_cg_type,el_include);
        el->objtypeid=dae_ID_FX_INCLUDE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_profile_cg_type,el_newparam);
        el->objtypeid=dae_ID_CG_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_profile_cg_type,el_technique);
        el->objtypeid=dae_ID_PROFILE_CG_TYPE_TECHNIQUE;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_cg_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 6;
    }
    ++t;
    t->name = "profile_cg_type_technique";
    t->size = sizeof(dae_profile_cg_type_technique);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_CG_TYPE_TECHNIQUE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_cg_type_technique,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_profile_cg_type_technique,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_cg_type_technique,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="annotate";
        el->offset=offsetof(dae_profile_cg_type_technique,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="pass";
        el->offset=offsetof(dae_profile_cg_type_technique,el_pass);
        el->objtypeid=dae_ID_CG_PASS_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_cg_type_technique,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "gles_max_lights_index_type";
    t->size = sizeof(dae_gles_max_lights_index_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_max_clip_planes_index_type";
    t->size = sizeof(dae_gles_max_clip_planes_index_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_MAX_CLIP_PLANES_INDEX_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_UINT32;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_texenv_mode_enum";
    t->size = sizeof(dae_gles_texenv_mode_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXENV_MODE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_texture_constant_type";
    t->size = sizeof(dae_gles_texture_constant_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXTURE_CONSTANT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_texture_constant_type,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_texture_constant_type,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_texenv_command_type";
    t->size = sizeof(dae_gles_texenv_command_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXENV_COMMAND_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="operator";
        at->offset=offsetof(dae_gles_texenv_command_type,at_operator);
        at->objtypeid=dae_ID_GLES_TEXENV_MODE_ENUM;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sampler";
        at->offset=offsetof(dae_gles_texenv_command_type,at_sampler);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="constant";
        el->offset=offsetof(dae_gles_texenv_command_type,el_constant);
        el->objtypeid=dae_ID_GLES_TEXTURE_CONSTANT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "gles_texcombiner_operator_rgb_enum";
    t->size = sizeof(dae_gles_texcombiner_operator_rgb_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXCOMBINER_OPERATOR_RGB_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_texcombiner_operator_alpha_enum";
    t->size = sizeof(dae_gles_texcombiner_operator_alpha_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXCOMBINER_OPERATOR_ALPHA_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_texcombiner_source_enum";
    t->size = sizeof(dae_gles_texcombiner_source_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXCOMBINER_SOURCE_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_texcombiner_operand_rgb_enum";
    t->size = sizeof(dae_gles_texcombiner_operand_rgb_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXCOMBINER_OPERAND_RGB_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_texcombiner_operand_alpha_enum";
    t->size = sizeof(dae_gles_texcombiner_operand_alpha_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXCOMBINER_OPERAND_ALPHA_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_texcombiner_argument_rgb_type";
    t->size = sizeof(dae_gles_texcombiner_argument_rgb_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXCOMBINER_ARGUMENT_RGB_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="source";
        at->offset=offsetof(dae_gles_texcombiner_argument_rgb_type,at_source);
        at->objtypeid=dae_ID_GLES_TEXCOMBINER_SOURCE_ENUM;
        at->min=0;
        at->max=1;
        ++at;
        at->name="operand";
        at->offset=offsetof(dae_gles_texcombiner_argument_rgb_type,at_operand);
        at->objtypeid=dae_ID_GLES_TEXCOMBINER_OPERAND_RGB_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="SRC_COLOR";
        ++at;
        at->name="sampler";
        at->offset=offsetof(dae_gles_texcombiner_argument_rgb_type,at_sampler);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_texcombiner_argument_alpha_type";
    t->size = sizeof(dae_gles_texcombiner_argument_alpha_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXCOMBINER_ARGUMENT_ALPHA_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="source";
        at->offset=offsetof(dae_gles_texcombiner_argument_alpha_type,at_source);
        at->objtypeid=dae_ID_GLES_TEXCOMBINER_SOURCE_ENUM;
        at->min=0;
        at->max=1;
        ++at;
        at->name="operand";
        at->offset=offsetof(dae_gles_texcombiner_argument_alpha_type,at_operand);
        at->objtypeid=dae_ID_GLES_TEXCOMBINER_OPERAND_ALPHA_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="SRC_ALPHA";
        ++at;
        at->name="sampler";
        at->offset=offsetof(dae_gles_texcombiner_argument_alpha_type,at_sampler);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_texcombiner_command_rgb_type";
    t->size = sizeof(dae_gles_texcombiner_command_rgb_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXCOMBINER_COMMAND_RGB_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="operator";
        at->offset=offsetof(dae_gles_texcombiner_command_rgb_type,at_operator);
        at->objtypeid=dae_ID_GLES_TEXCOMBINER_OPERATOR_RGB_ENUM;
        at->min=0;
        at->max=1;
        ++at;
        at->name="scale";
        at->offset=offsetof(dae_gles_texcombiner_command_rgb_type,at_scale);
        at->objtypeid=dae_ID_FLOAT;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="argument";
        el->offset=offsetof(dae_gles_texcombiner_command_rgb_type,el_argument);
        el->objtypeid=dae_ID_GLES_TEXCOMBINER_ARGUMENT_RGB_TYPE;
        el->min=1;
        el->max=3;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "gles_texcombiner_command_alpha_type";
    t->size = sizeof(dae_gles_texcombiner_command_alpha_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXCOMBINER_COMMAND_ALPHA_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="operator";
        at->offset=offsetof(dae_gles_texcombiner_command_alpha_type,at_operator);
        at->objtypeid=dae_ID_GLES_TEXCOMBINER_OPERATOR_ALPHA_ENUM;
        at->min=0;
        at->max=1;
        ++at;
        at->name="scale";
        at->offset=offsetof(dae_gles_texcombiner_command_alpha_type,at_scale);
        at->objtypeid=dae_ID_FLOAT;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="argument";
        el->offset=offsetof(dae_gles_texcombiner_command_alpha_type,el_argument);
        el->objtypeid=dae_ID_GLES_TEXCOMBINER_ARGUMENT_ALPHA_TYPE;
        el->min=1;
        el->max=3;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "gles_texcombiner_command_type";
    t->size = sizeof(dae_gles_texcombiner_command_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXCOMBINER_COMMAND_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="constant";
        el->offset=offsetof(dae_gles_texcombiner_command_type,el_constant);
        el->objtypeid=dae_ID_GLES_TEXTURE_CONSTANT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="RGB";
        el->offset=offsetof(dae_gles_texcombiner_command_type,el_RGB);
        el->objtypeid=dae_ID_GLES_TEXCOMBINER_COMMAND_RGB_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="alpha";
        el->offset=offsetof(dae_gles_texcombiner_command_type,el_alpha);
        el->objtypeid=dae_ID_GLES_TEXCOMBINER_COMMAND_ALPHA_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "gles_texture_pipeline_type";
    t->size = sizeof(dae_gles_texture_pipeline_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_TEXTURE_PIPELINE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_gles_texture_pipeline_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="texcombiner";
        el->offset=offsetof(dae_gles_texture_pipeline_type,el_texcombiner);
        el->objtypeid=dae_ID_GLES_TEXCOMBINER_COMMAND_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="texenv";
        el->offset=offsetof(dae_gles_texture_pipeline_type,el_texenv);
        el->objtypeid=dae_ID_GLES_TEXENV_COMMAND_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_gles_texture_pipeline_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "gles_sampler_wrap_enum";
    t->size = sizeof(dae_gles_sampler_wrap_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_SAMPLER_WRAP_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_sampler_type";
    t->size = sizeof(dae_gles_sampler_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_SAMPLER_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[10];
        dae_obj_memberdef* el = elems;
        el->name="instance_image";
        el->offset=offsetof(dae_gles_sampler_type,el_instance_image);
        el->objtypeid=dae_ID_INSTANCE_IMAGE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texcoord";
        el->offset=offsetof(dae_gles_sampler_type,el_texcoord);
        el->objtypeid=dae_ID_GLES_SAMPLER_TYPE_TEXCOORD;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="wrap_s";
        el->offset=offsetof(dae_gles_sampler_type,el_wrap_s);
        el->objtypeid=dae_ID_GLES_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="wrap_t";
        el->offset=offsetof(dae_gles_sampler_type,el_wrap_t);
        el->objtypeid=dae_ID_GLES_SAMPLER_WRAP_ENUM;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="minfilter";
        el->offset=offsetof(dae_gles_sampler_type,el_minfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MIN_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="magfilter";
        el->offset=offsetof(dae_gles_sampler_type,el_magfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MAG_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="mipfilter";
        el->offset=offsetof(dae_gles_sampler_type,el_mipfilter);
        el->objtypeid=dae_ID_FX_SAMPLER_MIP_FILTER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="mip_max_level";
        el->offset=offsetof(dae_gles_sampler_type,el_mip_max_level);
        el->objtypeid=dae_ID_UNSIGNEDBYTE;
        el->min=0;
        el->max=1;
        el->seq=7;
        ++el;
        el->name="mip_bias";
        el->offset=offsetof(dae_gles_sampler_type,el_mip_bias);
        el->objtypeid=dae_ID_FLOAT;
        el->min=0;
        el->max=1;
        el->seq=8;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_gles_sampler_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=9;
        ++el;
        t->elems = elems;
        t->numelems = 10;
    }
    ++t;
    t->name = "gles_sampler_type_texcoord";
    t->size = sizeof(dae_gles_sampler_type_texcoord);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_SAMPLER_TYPE_TEXCOORD;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="semantic";
        at->offset=offsetof(dae_gles_sampler_type_texcoord,at_semantic);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "gles_stencil_op_enum";
    t->size = sizeof(dae_gles_stencil_op_enum);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_STENCIL_OP_ENUM;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_enumeration_type";
    t->size = sizeof(dae_gles_enumeration_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_ENUMERATION_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "gles_pipeline_settings_group_alpha_func";
    t->size = sizeof(dae_gles_pipeline_settings_group_alpha_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="func";
        el->offset=offsetof(dae_gles_pipeline_settings_group_alpha_func,el_func);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="value";
        el->offset=offsetof(dae_gles_pipeline_settings_group_alpha_func,el_value);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC_VALUE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_alpha_func_func";
    t->size = sizeof(dae_gles_pipeline_settings_group_alpha_func_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_alpha_func_func,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_alpha_func_func,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_alpha_func_value";
    t->size = sizeof(dae_gles_pipeline_settings_group_alpha_func_value);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC_VALUE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_alpha_func_value,at_value);
        at->objtypeid=dae_ID_GL_ALPHA_VALUE_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0.0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_alpha_func_value,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_blend_func";
    t->size = sizeof(dae_gles_pipeline_settings_group_blend_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_BLEND_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="src";
        el->offset=offsetof(dae_gles_pipeline_settings_group_blend_func,el_src);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SRC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="dest";
        el->offset=offsetof(dae_gles_pipeline_settings_group_blend_func,el_dest);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_DEST;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_blend_func_src";
    t->size = sizeof(dae_gles_pipeline_settings_group_blend_func_src);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_SRC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_blend_func_src,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ONE";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_blend_func_src,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_blend_func_dest";
    t->size = sizeof(dae_gles_pipeline_settings_group_blend_func_dest);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_BLEND_FUNC_DEST;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_blend_func_dest,at_value);
        at->objtypeid=dae_ID_GL_BLEND_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ZERO";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_blend_func_dest,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_clip_plane";
    t->size = sizeof(dae_gles_pipeline_settings_group_clip_plane);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_CLIP_PLANE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_clip_plane,at_value);
        at->objtypeid=dae_ID_BOOL4_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_clip_plane,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_clip_plane,at_index);
        at->objtypeid=dae_ID_GLES_MAX_CLIP_PLANES_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_color_mask";
    t->size = sizeof(dae_gles_pipeline_settings_group_color_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_COLOR_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_color_mask,at_value);
        at->objtypeid=dae_ID_BOOL4_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_color_mask,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_cull_face";
    t->size = sizeof(dae_gles_pipeline_settings_group_cull_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_CULL_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_cull_face,at_value);
        at->objtypeid=dae_ID_GL_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="BACK";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_cull_face,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_depth_func";
    t->size = sizeof(dae_gles_pipeline_settings_group_depth_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_DEPTH_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_depth_func,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_depth_func,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_depth_mask";
    t->size = sizeof(dae_gles_pipeline_settings_group_depth_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_DEPTH_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_depth_mask,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_depth_mask,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_depth_range";
    t->size = sizeof(dae_gles_pipeline_settings_group_depth_range);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_DEPTH_RANGE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_depth_range,at_value);
        at->objtypeid=dae_ID_FLOAT2_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_depth_range,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_fog_color";
    t->size = sizeof(dae_gles_pipeline_settings_group_fog_color);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_COLOR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_color,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_color,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_fog_density";
    t->size = sizeof(dae_gles_pipeline_settings_group_fog_density);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_DENSITY;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_density,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_density,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_fog_mode";
    t->size = sizeof(dae_gles_pipeline_settings_group_fog_mode);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_MODE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_mode,at_value);
        at->objtypeid=dae_ID_GL_FOG_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="EXP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_mode,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_fog_start";
    t->size = sizeof(dae_gles_pipeline_settings_group_fog_start);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_START;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_start,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_start,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_fog_end";
    t->size = sizeof(dae_gles_pipeline_settings_group_fog_end);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_END;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_end,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_end,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_front_face";
    t->size = sizeof(dae_gles_pipeline_settings_group_front_face);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FRONT_FACE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_front_face,at_value);
        at->objtypeid=dae_ID_GL_FRONT_FACE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="CCW";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_front_face,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_logic_op";
    t->size = sizeof(dae_gles_pipeline_settings_group_logic_op);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LOGIC_OP;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_logic_op,at_value);
        at->objtypeid=dae_ID_GL_LOGIC_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="COPY";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_logic_op,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_ambient";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_ambient);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_AMBIENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_ambient,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_ambient,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_ambient,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_diffuse";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_diffuse);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_DIFFUSE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_diffuse,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_diffuse,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_diffuse,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_specular";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_specular);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_SPECULAR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_specular,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_specular,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_specular,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_position";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_position);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_POSITION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_position,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 1 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_position,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_position,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_constant_attenuation";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_constant_attenuation);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_CONSTANT_ATTENUATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_constant_attenuation,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_constant_attenuation,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_constant_attenuation,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_linear_attenuation";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_linear_attenuation);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_LINEAR_ATTENUATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_linear_attenuation,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_linear_attenuation,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_linear_attenuation,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_quadratic_attenuation";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_quadratic_attenuation);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_QUADRATIC_ATTENUATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_quadratic_attenuation,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_quadratic_attenuation,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_quadratic_attenuation,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_spot_cutoff";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_spot_cutoff);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_CUTOFF;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_spot_cutoff,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="180";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_spot_cutoff,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_spot_cutoff,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_spot_direction";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_spot_direction);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_DIRECTION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_spot_direction,at_value);
        at->objtypeid=dae_ID_FLOAT3_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 -1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_spot_direction,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_spot_direction,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_spot_exponent";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_spot_exponent);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_EXPONENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_spot_exponent,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_spot_exponent,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_spot_exponent,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_model_ambient";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_model_ambient);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_AMBIENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_model_ambient,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0.2 0.2 0.2 1.0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_model_ambient,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_line_width";
    t->size = sizeof(dae_gles_pipeline_settings_group_line_width);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LINE_WIDTH;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_line_width,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_line_width,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_material_ambient";
    t->size = sizeof(dae_gles_pipeline_settings_group_material_ambient);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MATERIAL_AMBIENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_material_ambient,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0.2 0.2 0.2 1.0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_material_ambient,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_material_diffuse";
    t->size = sizeof(dae_gles_pipeline_settings_group_material_diffuse);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MATERIAL_DIFFUSE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_material_diffuse,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0.8 0.8 0.8 1.0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_material_diffuse,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_material_emission";
    t->size = sizeof(dae_gles_pipeline_settings_group_material_emission);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MATERIAL_EMISSION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_material_emission,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_material_emission,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_material_shininess";
    t->size = sizeof(dae_gles_pipeline_settings_group_material_shininess);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MATERIAL_SHININESS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_material_shininess,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_material_shininess,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_material_specular";
    t->size = sizeof(dae_gles_pipeline_settings_group_material_specular);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MATERIAL_SPECULAR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_material_specular,at_value);
        at->objtypeid=dae_ID_FLOAT4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0 0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_material_specular,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_model_view_matrix";
    t->size = sizeof(dae_gles_pipeline_settings_group_model_view_matrix);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MODEL_VIEW_MATRIX;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_model_view_matrix,at_value);
        at->objtypeid=dae_ID_FLOAT4X4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_model_view_matrix,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_point_distance_attenuation";
    t->size = sizeof(dae_gles_pipeline_settings_group_point_distance_attenuation);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_DISTANCE_ATTENUATION;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_distance_attenuation,at_value);
        at->objtypeid=dae_ID_FLOAT3_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1 0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_distance_attenuation,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_point_fade_threshold_size";
    t->size = sizeof(dae_gles_pipeline_settings_group_point_fade_threshold_size);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_FADE_THRESHOLD_SIZE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_fade_threshold_size,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_fade_threshold_size,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_point_size";
    t->size = sizeof(dae_gles_pipeline_settings_group_point_size);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_SIZE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_size,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_size,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_point_size_min";
    t->size = sizeof(dae_gles_pipeline_settings_group_point_size_min);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_SIZE_MIN;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_size_min,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_size_min,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_point_size_max";
    t->size = sizeof(dae_gles_pipeline_settings_group_point_size_max);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_SIZE_MAX;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_size_max,at_value);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_size_max,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_polygon_offset";
    t->size = sizeof(dae_gles_pipeline_settings_group_polygon_offset);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_polygon_offset,at_value);
        at->objtypeid=dae_ID_FLOAT2_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0 0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_polygon_offset,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_projection_matrix";
    t->size = sizeof(dae_gles_pipeline_settings_group_projection_matrix);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_PROJECTION_MATRIX;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_projection_matrix,at_value);
        at->objtypeid=dae_ID_FLOAT4X4_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_projection_matrix,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_scissor";
    t->size = sizeof(dae_gles_pipeline_settings_group_scissor);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SCISSOR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_scissor,at_value);
        at->objtypeid=dae_ID_INT4_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_scissor,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_shade_model";
    t->size = sizeof(dae_gles_pipeline_settings_group_shade_model);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SHADE_MODEL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_shade_model,at_value);
        at->objtypeid=dae_ID_GL_SHADE_MODEL_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="SMOOTH";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_shade_model,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_stencil_func";
    t->size = sizeof(dae_gles_pipeline_settings_group_stencil_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="func";
        el->offset=offsetof(dae_gles_pipeline_settings_group_stencil_func,el_func);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ref";
        el->offset=offsetof(dae_gles_pipeline_settings_group_stencil_func,el_ref);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_REF;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="mask";
        el->offset=offsetof(dae_gles_pipeline_settings_group_stencil_func,el_mask);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_MASK;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_stencil_func_func";
    t->size = sizeof(dae_gles_pipeline_settings_group_stencil_func_func);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_FUNC;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_func_func,at_value);
        at->objtypeid=dae_ID_GL_FUNC_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="ALWAYS";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_func_func,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_stencil_func_ref";
    t->size = sizeof(dae_gles_pipeline_settings_group_stencil_func_ref);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_REF;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_func_ref,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="0";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_func_ref,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_stencil_func_mask";
    t->size = sizeof(dae_gles_pipeline_settings_group_stencil_func_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_func_mask,at_value);
        at->objtypeid=dae_ID_UNSIGNEDBYTE;
        at->min=0;
        at->max=1;
        at->deflt="255";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_func_mask,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_stencil_mask";
    t->size = sizeof(dae_gles_pipeline_settings_group_stencil_mask);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_MASK;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_mask,at_value);
        at->objtypeid=dae_ID_INT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="4294967295";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_mask,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_stencil_op";
    t->size = sizeof(dae_gles_pipeline_settings_group_stencil_op);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_OP;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="fail";
        el->offset=offsetof(dae_gles_pipeline_settings_group_stencil_op,el_fail);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_OP_FAIL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="zfail";
        el->offset=offsetof(dae_gles_pipeline_settings_group_stencil_op,el_zfail);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZFAIL;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="zpass";
        el->offset=offsetof(dae_gles_pipeline_settings_group_stencil_op,el_zpass);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZPASS;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_stencil_op_fail";
    t->size = sizeof(dae_gles_pipeline_settings_group_stencil_op_fail);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_OP_FAIL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_op_fail,at_value);
        at->objtypeid=dae_ID_GLES_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_op_fail,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_stencil_op_zfail";
    t->size = sizeof(dae_gles_pipeline_settings_group_stencil_op_zfail);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZFAIL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_op_zfail,at_value);
        at->objtypeid=dae_ID_GLES_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_op_zfail,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_stencil_op_zpass";
    t->size = sizeof(dae_gles_pipeline_settings_group_stencil_op_zpass);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_OP_ZPASS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_op_zpass,at_value);
        at->objtypeid=dae_ID_GLES_STENCIL_OP_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="KEEP";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_op_zpass,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_texture_pipeline";
    t->size = sizeof(dae_gles_pipeline_settings_group_texture_pipeline);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_TEXTURE_PIPELINE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="value";
        el->offset=offsetof(dae_gles_pipeline_settings_group_texture_pipeline,el_value);
        el->objtypeid=dae_ID_GLES_TEXTURE_PIPELINE_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_alpha_test_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_alpha_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_ALPHA_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_alpha_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_alpha_test_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_blend_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_blend_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_BLEND_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_blend_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_blend_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_clip_plane_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_clip_plane_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_CLIP_PLANE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_clip_plane_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_clip_plane_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_clip_plane_enable,at_index);
        at->objtypeid=dae_ID_GLES_MAX_CLIP_PLANES_INDEX_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_color_logic_op_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_color_logic_op_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_COLOR_LOGIC_OP_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_color_logic_op_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_color_logic_op_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_color_material_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_color_material_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_color_material_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="true";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_color_material_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_cull_face_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_cull_face_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_CULL_FACE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_cull_face_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_cull_face_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_depth_test_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_depth_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_DEPTH_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_depth_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_depth_test_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_dither_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_dither_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_DITHER_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_dither_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_dither_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_fog_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_fog_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_fog_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        at->name="index";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_enable,at_index);
        at->objtypeid=dae_ID_GLES_MAX_LIGHTS_INDEX_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_lighting_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_lighting_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHTING_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_lighting_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_lighting_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_light_model_two_side_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_light_model_two_side_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_TWO_SIDE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_model_two_side_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_light_model_two_side_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_line_smooth_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_line_smooth_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LINE_SMOOTH_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_line_smooth_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_line_smooth_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_multisample_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_multisample_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MULTISAMPLE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_multisample_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_multisample_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_normalize_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_normalize_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_NORMALIZE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_normalize_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_normalize_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_point_smooth_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_point_smooth_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_SMOOTH_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_smooth_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_point_smooth_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_polygon_offset_fill_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_polygon_offset_fill_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_FILL_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_polygon_offset_fill_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_polygon_offset_fill_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_rescale_normal_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_rescale_normal_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_RESCALE_NORMAL_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_rescale_normal_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_rescale_normal_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_sample_alpha_to_coverage_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_sample_alpha_to_coverage_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_COVERAGE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_sample_alpha_to_coverage_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_sample_alpha_to_coverage_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_sample_alpha_to_one_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_sample_alpha_to_one_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_ONE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_sample_alpha_to_one_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_sample_alpha_to_one_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_sample_coverage_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_sample_coverage_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_sample_coverage_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_sample_coverage_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_scissor_test_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_scissor_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SCISSOR_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_scissor_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_scissor_test_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_pipeline_settings_group_stencil_test_enable";
    t->size = sizeof(dae_gles_pipeline_settings_group_stencil_test_enable);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_TEST_ENABLE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="value";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_test_enable,at_value);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="param";
        at->offset=offsetof(dae_gles_pipeline_settings_group_stencil_test_enable,at_param);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "gles_newparam_type";
    t->size = sizeof(dae_gles_newparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GLES_NEWPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_gles_newparam_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[33];
        dae_obj_memberdef* el = elems;
        el->name="annotate";
        el->offset=offsetof(dae_gles_newparam_type,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="semantic";
        el->offset=offsetof(dae_gles_newparam_type,el_semantic);
        el->objtypeid=dae_ID_NCNAME;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="modifier";
        el->offset=offsetof(dae_gles_newparam_type,el_modifier);
        el->objtypeid=dae_ID_FX_MODIFIER_ENUM;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_gles_newparam_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool2";
        el->offset=offsetof(dae_gles_newparam_type,el_bool2);
        el->objtypeid=dae_ID_BOOL2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool3";
        el->offset=offsetof(dae_gles_newparam_type,el_bool3);
        el->objtypeid=dae_ID_BOOL3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="bool4";
        el->offset=offsetof(dae_gles_newparam_type,el_bool4);
        el->objtypeid=dae_ID_BOOL4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_gles_newparam_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int2";
        el->offset=offsetof(dae_gles_newparam_type,el_int2);
        el->objtypeid=dae_ID_INT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int3";
        el->offset=offsetof(dae_gles_newparam_type,el_int3);
        el->objtypeid=dae_ID_INT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="int4";
        el->offset=offsetof(dae_gles_newparam_type,el_int4);
        el->objtypeid=dae_ID_INT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_gles_newparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2";
        el->offset=offsetof(dae_gles_newparam_type,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3";
        el->offset=offsetof(dae_gles_newparam_type,el_float3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4";
        el->offset=offsetof(dae_gles_newparam_type,el_float4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float1x1";
        el->offset=offsetof(dae_gles_newparam_type,el_float1x1);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float1x2";
        el->offset=offsetof(dae_gles_newparam_type,el_float1x2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float1x3";
        el->offset=offsetof(dae_gles_newparam_type,el_float1x3);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float1x4";
        el->offset=offsetof(dae_gles_newparam_type,el_float1x4);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x1";
        el->offset=offsetof(dae_gles_newparam_type,el_float2x1);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x2";
        el->offset=offsetof(dae_gles_newparam_type,el_float2x2);
        el->objtypeid=dae_ID_FLOAT2X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x3";
        el->offset=offsetof(dae_gles_newparam_type,el_float2x3);
        el->objtypeid=dae_ID_FLOAT2X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float2x4";
        el->offset=offsetof(dae_gles_newparam_type,el_float2x4);
        el->objtypeid=dae_ID_FLOAT2X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x1";
        el->offset=offsetof(dae_gles_newparam_type,el_float3x1);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x2";
        el->offset=offsetof(dae_gles_newparam_type,el_float3x2);
        el->objtypeid=dae_ID_FLOAT3X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x3";
        el->offset=offsetof(dae_gles_newparam_type,el_float3x3);
        el->objtypeid=dae_ID_FLOAT3X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float3x4";
        el->offset=offsetof(dae_gles_newparam_type,el_float3x4);
        el->objtypeid=dae_ID_FLOAT3X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x1";
        el->offset=offsetof(dae_gles_newparam_type,el_float4x1);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x2";
        el->offset=offsetof(dae_gles_newparam_type,el_float4x2);
        el->objtypeid=dae_ID_FLOAT4X2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x3";
        el->offset=offsetof(dae_gles_newparam_type,el_float4x3);
        el->objtypeid=dae_ID_FLOAT4X3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="float4x4";
        el->offset=offsetof(dae_gles_newparam_type,el_float4x4);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sampler2D";
        el->offset=offsetof(dae_gles_newparam_type,el_sampler2D);
        el->objtypeid=dae_ID_GLES_SAMPLER_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="enum";
        el->offset=offsetof(dae_gles_newparam_type,el_enum);
        el->objtypeid=dae_ID_GLES_ENUMERATION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 33;
    }
    ++t;
    t->name = "profile_gles_type";
    t->size = sizeof(dae_profile_gles_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_gles_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="platform";
        at->offset=offsetof(dae_profile_gles_type,at_platform);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        at->deflt="PC";
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_gles_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_profile_gles_type,el_newparam);
        el->objtypeid=dae_ID_GLES_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_profile_gles_type,el_technique);
        el->objtypeid=dae_ID_PROFILE_GLES_TYPE_TECHNIQUE;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_gles_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "profile_gles_type_technique";
    t->size = sizeof(dae_profile_gles_type_technique);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLES_TYPE_TECHNIQUE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_profile_gles_type_technique,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_profile_gles_type_technique,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_profile_gles_type_technique,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="annotate";
        el->offset=offsetof(dae_profile_gles_type_technique,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="pass";
        el->offset=offsetof(dae_profile_gles_type_technique,el_pass);
        el->objtypeid=dae_ID_PROFILE_GLES_TYPE_TECHNIQUE_PASS;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_gles_type_technique,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "profile_gles_type_technique_pass";
    t->size = sizeof(dae_profile_gles_type_technique_pass);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLES_TYPE_TECHNIQUE_PASS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_profile_gles_type_technique_pass,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="annotate";
        el->offset=offsetof(dae_profile_gles_type_technique_pass,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="states";
        el->offset=offsetof(dae_profile_gles_type_technique_pass,el_states);
        el->objtypeid=dae_ID_PROFILE_GLES_TYPE_TECHNIQUE_PASS_STATES;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="evaluate";
        el->offset=offsetof(dae_profile_gles_type_technique_pass,el_evaluate);
        el->objtypeid=dae_ID_PROFILE_GLES_TYPE_TECHNIQUE_PASS_EVALUATE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_profile_gles_type_technique_pass,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "profile_gles_type_technique_pass_states";
    t->size = sizeof(dae_profile_gles_type_technique_pass_states);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLES_TYPE_TECHNIQUE_PASS_STATES;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[69];
        dae_obj_memberdef* el = elems;
        el->name="alpha_func";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_alpha_func);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_ALPHA_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_func";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_blend_func);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_BLEND_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="clip_plane";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_clip_plane);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_CLIP_PLANE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_mask";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_color_mask);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_COLOR_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="cull_face";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_cull_face);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_CULL_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_func";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_depth_func);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_DEPTH_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_mask";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_depth_mask);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_DEPTH_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_range";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_depth_range);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_DEPTH_RANGE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_color";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_fog_color);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_COLOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_density";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_fog_density);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_DENSITY;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_mode";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_fog_mode);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_MODE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_start";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_fog_start);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_START;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_end";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_fog_end);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_END;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="front_face";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_front_face);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FRONT_FACE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="logic_op";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_logic_op);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LOGIC_OP;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_ambient";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_ambient);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_AMBIENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_diffuse";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_diffuse);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_DIFFUSE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_specular";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_specular);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_SPECULAR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_position";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_position);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_POSITION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_constant_attenuation";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_constant_attenuation);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_CONSTANT_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_linear_attenuation";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_linear_attenuation);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_LINEAR_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_quadratic_attenuation";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_quadratic_attenuation);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_QUADRATIC_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_spot_cutoff";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_spot_cutoff);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_CUTOFF;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_spot_direction";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_spot_direction);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_DIRECTION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_spot_exponent";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_spot_exponent);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_SPOT_EXPONENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_model_ambient";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_model_ambient);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_AMBIENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_width";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_line_width);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LINE_WIDTH;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_ambient";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_material_ambient);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MATERIAL_AMBIENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_diffuse";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_material_diffuse);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MATERIAL_DIFFUSE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_emission";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_material_emission);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MATERIAL_EMISSION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_shininess";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_material_shininess);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MATERIAL_SHININESS;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="material_specular";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_material_specular);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MATERIAL_SPECULAR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="model_view_matrix";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_model_view_matrix);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MODEL_VIEW_MATRIX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_distance_attenuation";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_point_distance_attenuation);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_DISTANCE_ATTENUATION;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_fade_threshold_size";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_point_fade_threshold_size);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_FADE_THRESHOLD_SIZE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_point_size);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_SIZE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size_min";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_point_size_min);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_SIZE_MIN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_size_max";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_point_size_max);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_SIZE_MAX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_polygon_offset);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="projection_matrix";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_projection_matrix);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_PROJECTION_MATRIX;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="scissor";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_scissor);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SCISSOR;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="shade_model";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_shade_model);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SHADE_MODEL;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_func";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_stencil_func);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_FUNC;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_mask";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_stencil_mask);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_MASK;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_op";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_stencil_op);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_OP;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="texture_pipeline";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_texture_pipeline);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_TEXTURE_PIPELINE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="alpha_test_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_alpha_test_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_ALPHA_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="blend_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_blend_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_BLEND_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="clip_plane_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_clip_plane_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_CLIP_PLANE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_logic_op_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_color_logic_op_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_COLOR_LOGIC_OP_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="color_material_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_color_material_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_COLOR_MATERIAL_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="cull_face_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_cull_face_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_CULL_FACE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="depth_test_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_depth_test_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_DEPTH_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="dither_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_dither_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_DITHER_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="fog_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_fog_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_FOG_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="lighting_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_lighting_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHTING_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="light_model_two_side_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_light_model_two_side_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LIGHT_MODEL_TWO_SIDE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="line_smooth_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_line_smooth_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_LINE_SMOOTH_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="multisample_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_multisample_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_MULTISAMPLE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="normalize_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_normalize_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_NORMALIZE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="point_smooth_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_point_smooth_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POINT_SMOOTH_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="polygon_offset_fill_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_polygon_offset_fill_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_POLYGON_OFFSET_FILL_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="rescale_normal_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_rescale_normal_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_RESCALE_NORMAL_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_alpha_to_coverage_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_sample_alpha_to_coverage_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_COVERAGE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_alpha_to_one_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_sample_alpha_to_one_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SAMPLE_ALPHA_TO_ONE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sample_coverage_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_sample_coverage_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SAMPLE_COVERAGE_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="scissor_test_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_scissor_test_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_SCISSOR_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="stencil_test_enable";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_states,el_stencil_test_enable);
        el->objtypeid=dae_ID_GLES_PIPELINE_SETTINGS_GROUP_STENCIL_TEST_ENABLE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 69;
    }
    ++t;
    t->name = "profile_gles_type_technique_pass_evaluate";
    t->size = sizeof(dae_profile_gles_type_technique_pass_evaluate);
    t->flags= 0;
    t->objtypeid = dae_ID_PROFILE_GLES_TYPE_TECHNIQUE_PASS_EVALUATE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="color_target";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_evaluate,el_color_target);
        el->objtypeid=dae_ID_FX_COLORTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="depth_target";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_evaluate,el_depth_target);
        el->objtypeid=dae_ID_FX_DEPTHTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="stencil_target";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_evaluate,el_stencil_target);
        el->objtypeid=dae_ID_FX_STENCILTARGET_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="color_clear";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_evaluate,el_color_clear);
        el->objtypeid=dae_ID_FX_CLEARCOLOR_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="depth_clear";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_evaluate,el_depth_clear);
        el->objtypeid=dae_ID_FX_CLEARDEPTH_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        el->name="stencil_clear";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_evaluate,el_stencil_clear);
        el->objtypeid=dae_ID_FX_CLEARSTENCIL_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=5;
        ++el;
        el->name="draw";
        el->offset=offsetof(dae_profile_gles_type_technique_pass_evaluate,el_draw);
        el->objtypeid=dae_ID_FX_DRAW_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "effect_type";
    t->size = sizeof(dae_effect_type);
    t->flags= 0;
    t->objtypeid = dae_ID_EFFECT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_effect_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_effect_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[10];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_effect_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="annotate";
        el->offset=offsetof(dae_effect_type,el_annotate);
        el->objtypeid=dae_ID_FX_ANNOTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_effect_type,el_newparam);
        el->objtypeid=dae_ID_FX_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="profile_COMMON";
        el->offset=offsetof(dae_effect_type,el_profile_COMMON);
        el->objtypeid=dae_ID_PROFILE_COMMON_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="profile_BRIDGE";
        el->offset=offsetof(dae_effect_type,el_profile_BRIDGE);
        el->objtypeid=dae_ID_PROFILE_BRIDGE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="profile_GLES2";
        el->offset=offsetof(dae_effect_type,el_profile_GLES2);
        el->objtypeid=dae_ID_PROFILE_GLES2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="profile_GLSL";
        el->offset=offsetof(dae_effect_type,el_profile_GLSL);
        el->objtypeid=dae_ID_PROFILE_GLSL_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="profile_CG";
        el->offset=offsetof(dae_effect_type,el_profile_CG);
        el->objtypeid=dae_ID_PROFILE_CG_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="profile_GLES";
        el->offset=offsetof(dae_effect_type,el_profile_GLES);
        el->objtypeid=dae_ID_PROFILE_GLES_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_effect_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 10;
    }
    ++t;
    t->name = "library_effects_type";
    t->size = sizeof(dae_library_effects_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_EFFECTS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_effects_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_effects_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_effects_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="effect";
        el->offset=offsetof(dae_library_effects_type,el_effect);
        el->objtypeid=dae_ID_EFFECT_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_effects_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "box_type";
    t->size = sizeof(dae_box_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BOX_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="half_extents";
        el->offset=offsetof(dae_box_type,el_half_extents);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_box_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "plane_type";
    t->size = sizeof(dae_plane_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PLANE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="equation";
        el->offset=offsetof(dae_plane_type,el_equation);
        el->objtypeid=dae_ID_FLOAT4_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_plane_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "sphere_type";
    t->size = sizeof(dae_sphere_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SPHERE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="radius";
        el->offset=offsetof(dae_sphere_type,el_radius);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_sphere_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "cylinder_type";
    t->size = sizeof(dae_cylinder_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CYLINDER_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="height";
        el->offset=offsetof(dae_cylinder_type,el_height);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="radius";
        el->offset=offsetof(dae_cylinder_type,el_radius);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_cylinder_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "capsule_type";
    t->size = sizeof(dae_capsule_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CAPSULE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="height";
        el->offset=offsetof(dae_capsule_type,el_height);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="radius";
        el->offset=offsetof(dae_capsule_type,el_radius);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_capsule_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "force_field_type";
    t->size = sizeof(dae_force_field_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FORCE_FIELD_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_force_field_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_force_field_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_force_field_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_force_field_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_force_field_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "library_force_fields_type";
    t->size = sizeof(dae_library_force_fields_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_FORCE_FIELDS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_force_fields_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_force_fields_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_force_fields_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="force_field";
        el->offset=offsetof(dae_library_force_fields_type,el_force_field);
        el->objtypeid=dae_ID_FORCE_FIELD_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_force_fields_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "physics_material_type";
    t->size = sizeof(dae_physics_material_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PHYSICS_MATERIAL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_physics_material_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_physics_material_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_physics_material_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="technique_common";
        el->offset=offsetof(dae_physics_material_type,el_technique_common);
        el->objtypeid=dae_ID_PHYSICS_MATERIAL_TYPE_TECHNIQUE_COMMON;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_physics_material_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_physics_material_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "physics_material_type_technique_common";
    t->size = sizeof(dae_physics_material_type_technique_common);
    t->flags= 0;
    t->objtypeid = dae_ID_PHYSICS_MATERIAL_TYPE_TECHNIQUE_COMMON;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="dynamic_friction";
        el->offset=offsetof(dae_physics_material_type_technique_common,el_dynamic_friction);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="restitution";
        el->offset=offsetof(dae_physics_material_type_technique_common,el_restitution);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="static_friction";
        el->offset=offsetof(dae_physics_material_type_technique_common,el_static_friction);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "library_physics_materials_type";
    t->size = sizeof(dae_library_physics_materials_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_PHYSICS_MATERIALS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_physics_materials_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_physics_materials_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_physics_materials_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="physics_material";
        el->offset=offsetof(dae_library_physics_materials_type,el_physics_material);
        el->objtypeid=dae_ID_PHYSICS_MATERIAL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_physics_materials_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "edges_type";
    t->size = sizeof(dae_edges_type);
    t->flags= 0;
    t->objtypeid = dae_ID_EDGES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_edges_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_edges_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_edges_type,at_count);
        at->objtypeid=dae_ID_INT;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_edges_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=4;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_edges_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_edges_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "wires_type";
    t->size = sizeof(dae_wires_type);
    t->flags= 0;
    t->objtypeid = dae_ID_WIRES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_wires_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_wires_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_wires_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_wires_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=2;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="vcount";
        el->offset=offsetof(dae_wires_type,el_vcount);
        el->objtypeid=dae_ID_LIST_OF_UINTS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_wires_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_wires_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "faces_type";
    t->size = sizeof(dae_faces_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FACES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_faces_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_faces_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_faces_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_faces_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=3;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="vcount";
        el->offset=offsetof(dae_faces_type,el_vcount);
        el->objtypeid=dae_ID_LIST_OF_UINTS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_faces_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_faces_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "shells_type";
    t->size = sizeof(dae_shells_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SHELLS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_shells_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_shells_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_shells_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_shells_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=2;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="vcount";
        el->offset=offsetof(dae_shells_type,el_vcount);
        el->objtypeid=dae_ID_LIST_OF_UINTS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_shells_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_shells_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "solids_type";
    t->size = sizeof(dae_solids_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SOLIDS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_solids_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_solids_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_solids_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_solids_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=2;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="vcount";
        el->offset=offsetof(dae_solids_type,el_vcount);
        el->objtypeid=dae_ID_LIST_OF_UINTS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_solids_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_solids_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "line_type";
    t->size = sizeof(dae_line_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LINE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="origin";
        el->offset=offsetof(dae_line_type,el_origin);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="direction";
        el->offset=offsetof(dae_line_type,el_direction);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_line_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "circle_type";
    t->size = sizeof(dae_circle_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CIRCLE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="radius";
        el->offset=offsetof(dae_circle_type,el_radius);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_circle_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "ellipse_type";
    t->size = sizeof(dae_ellipse_type);
    t->flags= 0;
    t->objtypeid = dae_ID_ELLIPSE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="radius";
        el->offset=offsetof(dae_ellipse_type,el_radius);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_ellipse_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "parabola_type";
    t->size = sizeof(dae_parabola_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PARABOLA_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="focal";
        el->offset=offsetof(dae_parabola_type,el_focal);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_parabola_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "hyperbola_type";
    t->size = sizeof(dae_hyperbola_type);
    t->flags= 0;
    t->objtypeid = dae_ID_HYPERBOLA_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="radius";
        el->offset=offsetof(dae_hyperbola_type,el_radius);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_hyperbola_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "torus_type";
    t->size = sizeof(dae_torus_type);
    t->flags= 0;
    t->objtypeid = dae_ID_TORUS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="radius";
        el->offset=offsetof(dae_torus_type,el_radius);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_torus_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "pcurves_type";
    t->size = sizeof(dae_pcurves_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PCURVES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_pcurves_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_pcurves_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_pcurves_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_pcurves_type,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=3;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="vcount";
        el->offset=offsetof(dae_pcurves_type,el_vcount);
        el->objtypeid=dae_ID_LIST_OF_UINTS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="p";
        el->offset=offsetof(dae_pcurves_type,el_p);
        el->objtypeid=dae_ID_P_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_pcurves_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "sidref_array_type";
    t->size = sizeof(dae_sidref_array_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SIDREF_ARRAY_TYPE;
    t->dataoffset=offsetof(dae_sidref_array_type,data);
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=0;
    t->datamax=-1;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_sidref_array_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_sidref_array_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="count";
        at->offset=offsetof(dae_sidref_array_type,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "source_type";
    t->size = sizeof(dae_source_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SOURCE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_source_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_source_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[10];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_source_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="token_array";
        el->offset=offsetof(dae_source_type,el_token_array);
        el->objtypeid=dae_ID_TOKEN_ARRAY_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="IDREF_array";
        el->offset=offsetof(dae_source_type,el_IDREF_array);
        el->objtypeid=dae_ID_IDREF_ARRAY_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="Name_array";
        el->offset=offsetof(dae_source_type,el_Name_array);
        el->objtypeid=dae_ID_NAME_ARRAY_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="bool_array";
        el->offset=offsetof(dae_source_type,el_bool_array);
        el->objtypeid=dae_ID_BOOL_ARRAY_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="float_array";
        el->offset=offsetof(dae_source_type,el_float_array);
        el->objtypeid=dae_ID_FLOAT_ARRAY_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="int_array";
        el->offset=offsetof(dae_source_type,el_int_array);
        el->objtypeid=dae_ID_INT_ARRAY_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="SIDREF_array";
        el->offset=offsetof(dae_source_type,el_SIDREF_array);
        el->objtypeid=dae_ID_SIDREF_ARRAY_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="technique_common";
        el->offset=offsetof(dae_source_type,el_technique_common);
        el->objtypeid=dae_ID_SOURCE_TYPE_TECHNIQUE_COMMON;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_source_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 10;
    }
    ++t;
    t->name = "source_type_technique_common";
    t->size = sizeof(dae_source_type_technique_common);
    t->flags= 0;
    t->objtypeid = dae_ID_SOURCE_TYPE_TECHNIQUE_COMMON;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="accessor";
        el->offset=offsetof(dae_source_type_technique_common,el_accessor);
        el->objtypeid=dae_ID_ACCESSOR_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "animation_type";
    t->size = sizeof(dae_animation_type);
    t->flags= 0;
    t->objtypeid = dae_ID_ANIMATION_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_animation_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_animation_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[11];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_animation_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="source";
        el->offset=offsetof(dae_animation_type,el_source);
        el->objtypeid=dae_ID_SOURCE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="sampler";
        el->offset=offsetof(dae_animation_type,el_sampler);
        el->objtypeid=dae_ID_SAMPLER_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="animation";
        el->offset=offsetof(dae_animation_type,el_animation);
        el->objtypeid=dae_ID_ANIMATION_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="sampler";
        el->offset=offsetof(dae_animation_type,el_sampler);
        el->objtypeid=dae_ID_SAMPLER_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="animation";
        el->offset=offsetof(dae_animation_type,el_animation);
        el->objtypeid=dae_ID_ANIMATION_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="channel";
        el->offset=offsetof(dae_animation_type,el_channel);
        el->objtypeid=dae_ID_CHANNEL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="channel";
        el->offset=offsetof(dae_animation_type,el_channel);
        el->objtypeid=dae_ID_CHANNEL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="animation";
        el->offset=offsetof(dae_animation_type,el_animation);
        el->objtypeid=dae_ID_ANIMATION_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="animation";
        el->offset=offsetof(dae_animation_type,el_animation);
        el->objtypeid=dae_ID_ANIMATION_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_animation_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=5;
        ++el;
        t->elems = elems;
        t->numelems = 11;
    }
    ++t;
    t->name = "library_animations_type";
    t->size = sizeof(dae_library_animations_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_ANIMATIONS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_animations_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_animations_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_animations_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="animation";
        el->offset=offsetof(dae_library_animations_type,el_animation);
        el->objtypeid=dae_ID_ANIMATION_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_animations_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "skin_type";
    t->size = sizeof(dae_skin_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SKIN_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="source";
        at->offset=offsetof(dae_skin_type,at_source);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="bind_shape_matrix";
        el->offset=offsetof(dae_skin_type,el_bind_shape_matrix);
        el->objtypeid=dae_ID_FLOAT4X4_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="source";
        el->offset=offsetof(dae_skin_type,el_source);
        el->objtypeid=dae_ID_SOURCE_TYPE;
        el->min=3;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="joints";
        el->offset=offsetof(dae_skin_type,el_joints);
        el->objtypeid=dae_ID_SKIN_TYPE_JOINTS;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="vertex_weights";
        el->offset=offsetof(dae_skin_type,el_vertex_weights);
        el->objtypeid=dae_ID_SKIN_TYPE_VERTEX_WEIGHTS;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_skin_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "skin_type_joints";
    t->size = sizeof(dae_skin_type_joints);
    t->flags= 0;
    t->objtypeid = dae_ID_SKIN_TYPE_JOINTS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_skin_type_joints,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_TYPE;
        el->min=2;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_skin_type_joints,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "skin_type_vertex_weights";
    t->size = sizeof(dae_skin_type_vertex_weights);
    t->flags= 0;
    t->objtypeid = dae_ID_SKIN_TYPE_VERTEX_WEIGHTS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="count";
        at->offset=offsetof(dae_skin_type_vertex_weights,at_count);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_skin_type_vertex_weights,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_OFFSET_TYPE;
        el->min=2;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="vcount";
        el->offset=offsetof(dae_skin_type_vertex_weights,el_vcount);
        el->objtypeid=dae_ID_LIST_OF_UINTS_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="v";
        el->offset=offsetof(dae_skin_type_vertex_weights,el_v);
        el->objtypeid=dae_ID_LIST_OF_INTS_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_skin_type_vertex_weights,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "convex_mesh_type";
    t->size = sizeof(dae_convex_mesh_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CONVEX_MESH_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="convex_hull_of";
        at->offset=offsetof(dae_convex_mesh_type,at_convex_hull_of);
        at->objtypeid=dae_ID_ANYURI;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[10];
        dae_obj_memberdef* el = elems;
        el->name="source";
        el->offset=offsetof(dae_convex_mesh_type,el_source);
        el->objtypeid=dae_ID_SOURCE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="vertices";
        el->offset=offsetof(dae_convex_mesh_type,el_vertices);
        el->objtypeid=dae_ID_VERTICES_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="lines";
        el->offset=offsetof(dae_convex_mesh_type,el_lines);
        el->objtypeid=dae_ID_LINES_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="linestrips";
        el->offset=offsetof(dae_convex_mesh_type,el_linestrips);
        el->objtypeid=dae_ID_LINESTRIPS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="polygons";
        el->offset=offsetof(dae_convex_mesh_type,el_polygons);
        el->objtypeid=dae_ID_POLYGONS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="polylist";
        el->offset=offsetof(dae_convex_mesh_type,el_polylist);
        el->objtypeid=dae_ID_POLYLIST_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="triangles";
        el->offset=offsetof(dae_convex_mesh_type,el_triangles);
        el->objtypeid=dae_ID_TRIANGLES_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="trifans";
        el->offset=offsetof(dae_convex_mesh_type,el_trifans);
        el->objtypeid=dae_ID_TRIFANS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="tristrips";
        el->offset=offsetof(dae_convex_mesh_type,el_tristrips);
        el->objtypeid=dae_ID_TRISTRIPS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_convex_mesh_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 10;
    }
    ++t;
    t->name = "mesh_type";
    t->size = sizeof(dae_mesh_type);
    t->flags= 0;
    t->objtypeid = dae_ID_MESH_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[10];
        dae_obj_memberdef* el = elems;
        el->name="source";
        el->offset=offsetof(dae_mesh_type,el_source);
        el->objtypeid=dae_ID_SOURCE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="vertices";
        el->offset=offsetof(dae_mesh_type,el_vertices);
        el->objtypeid=dae_ID_VERTICES_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="lines";
        el->offset=offsetof(dae_mesh_type,el_lines);
        el->objtypeid=dae_ID_LINES_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="linestrips";
        el->offset=offsetof(dae_mesh_type,el_linestrips);
        el->objtypeid=dae_ID_LINESTRIPS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="polygons";
        el->offset=offsetof(dae_mesh_type,el_polygons);
        el->objtypeid=dae_ID_POLYGONS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="polylist";
        el->offset=offsetof(dae_mesh_type,el_polylist);
        el->objtypeid=dae_ID_POLYLIST_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="triangles";
        el->offset=offsetof(dae_mesh_type,el_triangles);
        el->objtypeid=dae_ID_TRIANGLES_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="trifans";
        el->offset=offsetof(dae_mesh_type,el_trifans);
        el->objtypeid=dae_ID_TRIFANS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="tristrips";
        el->offset=offsetof(dae_mesh_type,el_tristrips);
        el->objtypeid=dae_ID_TRISTRIPS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_mesh_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 10;
    }
    ++t;
    t->name = "spline_type";
    t->size = sizeof(dae_spline_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SPLINE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="closed";
        at->offset=offsetof(dae_spline_type,at_closed);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="source";
        el->offset=offsetof(dae_spline_type,el_source);
        el->objtypeid=dae_ID_SOURCE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="control_vertices";
        el->offset=offsetof(dae_spline_type,el_control_vertices);
        el->objtypeid=dae_ID_SPLINE_TYPE_CONTROL_VERTICES;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_spline_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "spline_type_control_vertices";
    t->size = sizeof(dae_spline_type_control_vertices);
    t->flags= 0;
    t->objtypeid = dae_ID_SPLINE_TYPE_CONTROL_VERTICES;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_spline_type_control_vertices,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_spline_type_control_vertices,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "nurbs_surface_type";
    t->size = sizeof(dae_nurbs_surface_type);
    t->flags= 0;
    t->objtypeid = dae_ID_NURBS_SURFACE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="degree_u";
        at->offset=offsetof(dae_nurbs_surface_type,at_degree_u);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="closed_u";
        at->offset=offsetof(dae_nurbs_surface_type,at_closed_u);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        at->name="degree_v";
        at->offset=offsetof(dae_nurbs_surface_type,at_degree_v);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="closed_v";
        at->offset=offsetof(dae_nurbs_surface_type,at_closed_v);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="source";
        el->offset=offsetof(dae_nurbs_surface_type,el_source);
        el->objtypeid=dae_ID_SOURCE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="control_vertices";
        el->offset=offsetof(dae_nurbs_surface_type,el_control_vertices);
        el->objtypeid=dae_ID_NURBS_SURFACE_TYPE_CONTROL_VERTICES;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_nurbs_surface_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "nurbs_surface_type_control_vertices";
    t->size = sizeof(dae_nurbs_surface_type_control_vertices);
    t->flags= 0;
    t->objtypeid = dae_ID_NURBS_SURFACE_TYPE_CONTROL_VERTICES;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_nurbs_surface_type_control_vertices,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_nurbs_surface_type_control_vertices,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "nurbs_type";
    t->size = sizeof(dae_nurbs_type);
    t->flags= 0;
    t->objtypeid = dae_ID_NURBS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="degree";
        at->offset=offsetof(dae_nurbs_type,at_degree);
        at->objtypeid=dae_ID_UINT_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="closed";
        at->offset=offsetof(dae_nurbs_type,at_closed);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="false";
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="source";
        el->offset=offsetof(dae_nurbs_type,el_source);
        el->objtypeid=dae_ID_SOURCE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="control_vertices";
        el->offset=offsetof(dae_nurbs_type,el_control_vertices);
        el->objtypeid=dae_ID_NURBS_TYPE_CONTROL_VERTICES;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_nurbs_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "nurbs_type_control_vertices";
    t->size = sizeof(dae_nurbs_type_control_vertices);
    t->flags= 0;
    t->objtypeid = dae_ID_NURBS_TYPE_CONTROL_VERTICES;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_nurbs_type_control_vertices,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_nurbs_type_control_vertices,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "morph_type";
    t->size = sizeof(dae_morph_type);
    t->flags= 0;
    t->objtypeid = dae_ID_MORPH_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="method";
        at->offset=offsetof(dae_morph_type,at_method);
        at->objtypeid=dae_ID_MORPH_METHOD_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="NORMALIZED";
        ++at;
        at->name="source";
        at->offset=offsetof(dae_morph_type,at_source);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="source";
        el->offset=offsetof(dae_morph_type,el_source);
        el->objtypeid=dae_ID_SOURCE_TYPE;
        el->min=2;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="targets";
        el->offset=offsetof(dae_morph_type,el_targets);
        el->objtypeid=dae_ID_MORPH_TYPE_TARGETS;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_morph_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "morph_type_targets";
    t->size = sizeof(dae_morph_type_targets);
    t->flags= 0;
    t->objtypeid = dae_ID_MORPH_TYPE_TARGETS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="input";
        el->offset=offsetof(dae_morph_type_targets,el_input);
        el->objtypeid=dae_ID_INPUT_LOCAL_TYPE;
        el->min=2;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_morph_type_targets,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "controller_type";
    t->size = sizeof(dae_controller_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CONTROLLER_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_controller_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_controller_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_controller_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="skin";
        el->offset=offsetof(dae_controller_type,el_skin);
        el->objtypeid=dae_ID_SKIN_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="morph";
        el->offset=offsetof(dae_controller_type,el_morph);
        el->objtypeid=dae_ID_MORPH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_controller_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "library_controllers_type";
    t->size = sizeof(dae_library_controllers_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_CONTROLLERS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_controllers_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_controllers_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_controllers_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="controller";
        el->offset=offsetof(dae_library_controllers_type,el_controller);
        el->objtypeid=dae_ID_CONTROLLER_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_controllers_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "origin_type";
    t->size = sizeof(dae_origin_type);
    t->flags= 0;
    t->objtypeid = dae_ID_ORIGIN_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=3;
    t->datamax=3;
    ++t;
    t->name = "orient_type";
    t->size = sizeof(dae_orient_type);
    t->flags= 0;
    t->objtypeid = dae_ID_ORIENT_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=4;
    t->datamax=4;
    ++t;
    t->name = "curve_type";
    t->size = sizeof(dae_curve_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CURVE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_curve_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_curve_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[8];
        dae_obj_memberdef* el = elems;
        el->name="line";
        el->offset=offsetof(dae_curve_type,el_line);
        el->objtypeid=dae_ID_LINE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="circle";
        el->offset=offsetof(dae_curve_type,el_circle);
        el->objtypeid=dae_ID_CIRCLE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="ellipse";
        el->offset=offsetof(dae_curve_type,el_ellipse);
        el->objtypeid=dae_ID_ELLIPSE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="parabola";
        el->offset=offsetof(dae_curve_type,el_parabola);
        el->objtypeid=dae_ID_PARABOLA_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="hyperbola";
        el->offset=offsetof(dae_curve_type,el_hyperbola);
        el->objtypeid=dae_ID_HYPERBOLA_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="nurbs";
        el->offset=offsetof(dae_curve_type,el_nurbs);
        el->objtypeid=dae_ID_NURBS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="orient";
        el->offset=offsetof(dae_curve_type,el_orient);
        el->objtypeid=dae_ID_ORIENT_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="origin";
        el->offset=offsetof(dae_curve_type,el_origin);
        el->objtypeid=dae_ID_ORIGIN_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 8;
    }
    ++t;
    t->name = "surface_curves_type";
    t->size = sizeof(dae_surface_curves_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SURFACE_CURVES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="curve";
        el->offset=offsetof(dae_surface_curves_type,el_curve);
        el->objtypeid=dae_ID_CURVE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_surface_curves_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "curves_type";
    t->size = sizeof(dae_curves_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CURVES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="curve";
        el->offset=offsetof(dae_curves_type,el_curve);
        el->objtypeid=dae_ID_CURVE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_curves_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "swept_surface_type";
    t->size = sizeof(dae_swept_surface_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SWEPT_SURFACE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="curve";
        el->offset=offsetof(dae_swept_surface_type,el_curve);
        el->objtypeid=dae_ID_CURVE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="direction";
        el->offset=offsetof(dae_swept_surface_type,el_direction);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="origin";
        el->offset=offsetof(dae_swept_surface_type,el_origin);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="axis";
        el->offset=offsetof(dae_swept_surface_type,el_axis);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_swept_surface_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "cone_type";
    t->size = sizeof(dae_cone_type);
    t->flags= 0;
    t->objtypeid = dae_ID_CONE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="radius";
        el->offset=offsetof(dae_cone_type,el_radius);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="angle";
        el->offset=offsetof(dae_cone_type,el_angle);
        el->objtypeid=dae_ID_FLOAT;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_cone_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "surface_type";
    t->size = sizeof(dae_surface_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SURFACE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_surface_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_surface_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[9];
        dae_obj_memberdef* el = elems;
        el->name="plane";
        el->offset=offsetof(dae_surface_type,el_plane);
        el->objtypeid=dae_ID_PLANE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="sphere";
        el->offset=offsetof(dae_surface_type,el_sphere);
        el->objtypeid=dae_ID_SPHERE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="torus";
        el->offset=offsetof(dae_surface_type,el_torus);
        el->objtypeid=dae_ID_TORUS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="swept_surface";
        el->offset=offsetof(dae_surface_type,el_swept_surface);
        el->objtypeid=dae_ID_SWEPT_SURFACE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="nurbs_surface";
        el->offset=offsetof(dae_surface_type,el_nurbs_surface);
        el->objtypeid=dae_ID_NURBS_SURFACE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="cone";
        el->offset=offsetof(dae_surface_type,el_cone);
        el->objtypeid=dae_ID_CONE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="cylinder";
        el->offset=offsetof(dae_surface_type,el_cylinder);
        el->objtypeid=dae_ID_SURFACE_TYPE_CYLINDER;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="orient";
        el->offset=offsetof(dae_surface_type,el_orient);
        el->objtypeid=dae_ID_ORIENT_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="origin";
        el->offset=offsetof(dae_surface_type,el_origin);
        el->objtypeid=dae_ID_ORIGIN_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 9;
    }
    ++t;
    t->name = "surface_type_cylinder";
    t->size = sizeof(dae_surface_type_cylinder);
    t->flags= 0;
    t->objtypeid = dae_ID_SURFACE_TYPE_CYLINDER;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="radius";
        el->offset=offsetof(dae_surface_type_cylinder,el_radius);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_surface_type_cylinder,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "surfaces_type";
    t->size = sizeof(dae_surfaces_type);
    t->flags= 0;
    t->objtypeid = dae_ID_SURFACES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="surface";
        el->offset=offsetof(dae_surfaces_type,el_surface);
        el->objtypeid=dae_ID_SURFACE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_surfaces_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "brep_type";
    t->size = sizeof(dae_brep_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BREP_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[12];
        dae_obj_memberdef* el = elems;
        el->name="curves";
        el->offset=offsetof(dae_brep_type,el_curves);
        el->objtypeid=dae_ID_CURVES_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="surface_curves";
        el->offset=offsetof(dae_brep_type,el_surface_curves);
        el->objtypeid=dae_ID_SURFACE_CURVES_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="surfaces";
        el->offset=offsetof(dae_brep_type,el_surfaces);
        el->objtypeid=dae_ID_SURFACES_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="source";
        el->offset=offsetof(dae_brep_type,el_source);
        el->objtypeid=dae_ID_SOURCE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="vertices";
        el->offset=offsetof(dae_brep_type,el_vertices);
        el->objtypeid=dae_ID_VERTICES_TYPE;
        el->min=1;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="edges";
        el->offset=offsetof(dae_brep_type,el_edges);
        el->objtypeid=dae_ID_EDGES_TYPE;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="wires";
        el->offset=offsetof(dae_brep_type,el_wires);
        el->objtypeid=dae_ID_WIRES_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="faces";
        el->offset=offsetof(dae_brep_type,el_faces);
        el->objtypeid=dae_ID_FACES_TYPE;
        el->min=0;
        el->max=1;
        el->seq=7;
        ++el;
        el->name="pcurves";
        el->offset=offsetof(dae_brep_type,el_pcurves);
        el->objtypeid=dae_ID_PCURVES_TYPE;
        el->min=0;
        el->max=1;
        el->seq=8;
        ++el;
        el->name="shells";
        el->offset=offsetof(dae_brep_type,el_shells);
        el->objtypeid=dae_ID_SHELLS_TYPE;
        el->min=0;
        el->max=1;
        el->seq=9;
        ++el;
        el->name="solids";
        el->offset=offsetof(dae_brep_type,el_solids);
        el->objtypeid=dae_ID_SOLIDS_TYPE;
        el->min=0;
        el->max=1;
        el->seq=10;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_brep_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=11;
        ++el;
        t->elems = elems;
        t->numelems = 12;
    }
    ++t;
    t->name = "geometry_type";
    t->size = sizeof(dae_geometry_type);
    t->flags= 0;
    t->objtypeid = dae_ID_GEOMETRY_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_geometry_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_geometry_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[6];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_geometry_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="convex_mesh";
        el->offset=offsetof(dae_geometry_type,el_convex_mesh);
        el->objtypeid=dae_ID_CONVEX_MESH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="mesh";
        el->offset=offsetof(dae_geometry_type,el_mesh);
        el->objtypeid=dae_ID_MESH_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="spline";
        el->offset=offsetof(dae_geometry_type,el_spline);
        el->objtypeid=dae_ID_SPLINE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="brep";
        el->offset=offsetof(dae_geometry_type,el_brep);
        el->objtypeid=dae_ID_BREP_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_geometry_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 6;
    }
    ++t;
    t->name = "library_geometries_type";
    t->size = sizeof(dae_library_geometries_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_GEOMETRIES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_geometries_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_geometries_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_geometries_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="geometry";
        el->offset=offsetof(dae_library_geometries_type,el_geometry);
        el->objtypeid=dae_ID_GEOMETRY_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_geometries_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "instance_joint_type";
    t->size = sizeof(dae_instance_joint_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_JOINT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_joint_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_joint_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_joint_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    ++t;
    t->name = "targetable_float4_type";
    t->size = sizeof(dae_targetable_float4_type);
    t->flags= 0;
    t->objtypeid = dae_ID_TARGETABLE_FLOAT4_TYPE;
    t->dataoffset=offsetof(dae_targetable_float4_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=4;
    t->datamax=4;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_targetable_float4_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "rotate_type";
    t->size = sizeof(dae_rotate_type);
    t->flags= 0;
    t->objtypeid = dae_ID_ROTATE_TYPE;
    t->dataoffset=offsetof(dae_rotate_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=4;
    t->datamax=4;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_rotate_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "link_type";
    t->size = sizeof(dae_link_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LINK_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_link_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_link_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="rotate";
        el->offset=offsetof(dae_link_type,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="translate";
        el->offset=offsetof(dae_link_type,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="attachment_full";
        el->offset=offsetof(dae_link_type,el_attachment_full);
        el->objtypeid=dae_ID_LINK_TYPE_ATTACHMENT_FULL;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="attachment_start";
        el->offset=offsetof(dae_link_type,el_attachment_start);
        el->objtypeid=dae_ID_LINK_TYPE_ATTACHMENT_START;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="attachment_end";
        el->offset=offsetof(dae_link_type,el_attachment_end);
        el->objtypeid=dae_ID_LINK_TYPE_ATTACHMENT_END;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "link_type_attachment_full";
    t->size = sizeof(dae_link_type_attachment_full);
    t->flags= 0;
    t->objtypeid = dae_ID_LINK_TYPE_ATTACHMENT_FULL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="joint";
        at->offset=offsetof(dae_link_type_attachment_full,at_joint);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="rotate";
        el->offset=offsetof(dae_link_type_attachment_full,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="translate";
        el->offset=offsetof(dae_link_type_attachment_full,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="link";
        el->offset=offsetof(dae_link_type_attachment_full,el_link);
        el->objtypeid=dae_ID_LINK_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "link_type_attachment_start";
    t->size = sizeof(dae_link_type_attachment_start);
    t->flags= 0;
    t->objtypeid = dae_ID_LINK_TYPE_ATTACHMENT_START;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="joint";
        at->offset=offsetof(dae_link_type_attachment_start,at_joint);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="rotate";
        el->offset=offsetof(dae_link_type_attachment_start,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="translate";
        el->offset=offsetof(dae_link_type_attachment_start,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "link_type_attachment_end";
    t->size = sizeof(dae_link_type_attachment_end);
    t->flags= 0;
    t->objtypeid = dae_ID_LINK_TYPE_ATTACHMENT_END;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="joint";
        at->offset=offsetof(dae_link_type_attachment_end,at_joint);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="rotate";
        el->offset=offsetof(dae_link_type_attachment_end,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="translate";
        el->offset=offsetof(dae_link_type_attachment_end,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "rigid_constraint_type";
    t->size = sizeof(dae_rigid_constraint_type);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_rigid_constraint_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_rigid_constraint_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="ref_attachment";
        el->offset=offsetof(dae_rigid_constraint_type,el_ref_attachment);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_REF_ATTACHMENT;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="attachment";
        el->offset=offsetof(dae_rigid_constraint_type,el_attachment);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_ATTACHMENT;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="technique_common";
        el->offset=offsetof(dae_rigid_constraint_type,el_technique_common);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_rigid_constraint_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_rigid_constraint_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "rigid_constraint_type_ref_attachment";
    t->size = sizeof(dae_rigid_constraint_type_ref_attachment);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_REF_ATTACHMENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="rigid_body";
        at->offset=offsetof(dae_rigid_constraint_type_ref_attachment,at_rigid_body);
        at->objtypeid=dae_ID_ANYURI;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="translate";
        el->offset=offsetof(dae_rigid_constraint_type_ref_attachment,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="rotate";
        el->offset=offsetof(dae_rigid_constraint_type_ref_attachment,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_rigid_constraint_type_ref_attachment,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "rigid_constraint_type_attachment";
    t->size = sizeof(dae_rigid_constraint_type_attachment);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_ATTACHMENT;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="rigid_body";
        at->offset=offsetof(dae_rigid_constraint_type_attachment,at_rigid_body);
        at->objtypeid=dae_ID_ANYURI;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="translate";
        el->offset=offsetof(dae_rigid_constraint_type_attachment,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="rotate";
        el->offset=offsetof(dae_rigid_constraint_type_attachment,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_rigid_constraint_type_attachment,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "rigid_constraint_type_technique_common";
    t->size = sizeof(dae_rigid_constraint_type_technique_common);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="enabled";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common,el_enabled);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_ENABLED;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="interpenetrate";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common,el_interpenetrate);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_INTERPENETRATE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="limits";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common,el_limits);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_LIMITS;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="spring";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common,el_spring);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_SPRING;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "rigid_constraint_type_technique_common_enabled";
    t->size = sizeof(dae_rigid_constraint_type_technique_common_enabled);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_ENABLED;
    t->dataoffset=offsetof(dae_rigid_constraint_type_technique_common_enabled,data);
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_rigid_constraint_type_technique_common_enabled,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "rigid_constraint_type_technique_common_interpenetrate";
    t->size = sizeof(dae_rigid_constraint_type_technique_common_interpenetrate);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_INTERPENETRATE;
    t->dataoffset=offsetof(dae_rigid_constraint_type_technique_common_interpenetrate,data);
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_rigid_constraint_type_technique_common_interpenetrate,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "rigid_constraint_type_technique_common_limits";
    t->size = sizeof(dae_rigid_constraint_type_technique_common_limits);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_LIMITS;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="swing_cone_and_twist";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_limits,el_swing_cone_and_twist);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_LIMITS_SWING_CONE_AND_TWIST;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="linear";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_limits,el_linear);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_LIMITS_LINEAR;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "rigid_constraint_type_technique_common_limits_swing_cone_and_twist";
    t->size = sizeof(dae_rigid_constraint_type_technique_common_limits_swing_cone_and_twist);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_LIMITS_SWING_CONE_AND_TWIST;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="min";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_limits_swing_cone_and_twist,el_min);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="max";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_limits_swing_cone_and_twist,el_max);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "rigid_constraint_type_technique_common_limits_linear";
    t->size = sizeof(dae_rigid_constraint_type_technique_common_limits_linear);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_LIMITS_LINEAR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="min";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_limits_linear,el_min);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="max";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_limits_linear,el_max);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "rigid_constraint_type_technique_common_spring";
    t->size = sizeof(dae_rigid_constraint_type_technique_common_spring);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_SPRING;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="angular";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_spring,el_angular);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_SPRING_ANGULAR;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="linear";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_spring,el_linear);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_SPRING_LINEAR;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "rigid_constraint_type_technique_common_spring_angular";
    t->size = sizeof(dae_rigid_constraint_type_technique_common_spring_angular);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_SPRING_ANGULAR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="stiffness";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_spring_angular,el_stiffness);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="damping";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_spring_angular,el_damping);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="target_value";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_spring_angular,el_target_value);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "rigid_constraint_type_technique_common_spring_linear";
    t->size = sizeof(dae_rigid_constraint_type_technique_common_spring_linear);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_CONSTRAINT_TYPE_TECHNIQUE_COMMON_SPRING_LINEAR;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="stiffness";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_spring_linear,el_stiffness);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="damping";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_spring_linear,el_damping);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="target_value";
        el->offset=offsetof(dae_rigid_constraint_type_technique_common_spring_linear,el_target_value);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "node_type";
    t->size = sizeof(dae_node_type);
    t->flags= 0;
    t->objtypeid = dae_ID_NODE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[5];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_node_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_node_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_node_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="type";
        at->offset=offsetof(dae_node_type,at_type);
        at->objtypeid=dae_ID_NODE_ENUM;
        at->min=0;
        at->max=1;
        at->deflt="NODE";
        ++at;
        at->name="layer";
        at->offset=offsetof(dae_node_type,at_layer);
        at->objtypeid=dae_ID_LIST_OF_NAMES_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 5;
    }
    {
        static dae_obj_memberdef elems[14];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_node_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="lookat";
        el->offset=offsetof(dae_node_type,el_lookat);
        el->objtypeid=dae_ID_LOOKAT_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="matrix";
        el->offset=offsetof(dae_node_type,el_matrix);
        el->objtypeid=dae_ID_MATRIX_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="rotate";
        el->offset=offsetof(dae_node_type,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="scale";
        el->offset=offsetof(dae_node_type,el_scale);
        el->objtypeid=dae_ID_SCALE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="skew";
        el->offset=offsetof(dae_node_type,el_skew);
        el->objtypeid=dae_ID_SKEW_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="translate";
        el->offset=offsetof(dae_node_type,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="instance_camera";
        el->offset=offsetof(dae_node_type,el_instance_camera);
        el->objtypeid=dae_ID_INSTANCE_CAMERA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="instance_controller";
        el->offset=offsetof(dae_node_type,el_instance_controller);
        el->objtypeid=dae_ID_INSTANCE_CONTROLLER_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="instance_geometry";
        el->offset=offsetof(dae_node_type,el_instance_geometry);
        el->objtypeid=dae_ID_INSTANCE_GEOMETRY_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="instance_light";
        el->offset=offsetof(dae_node_type,el_instance_light);
        el->objtypeid=dae_ID_INSTANCE_LIGHT_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        el->name="instance_node";
        el->offset=offsetof(dae_node_type,el_instance_node);
        el->objtypeid=dae_ID_INSTANCE_NODE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=5;
        ++el;
        el->name="node";
        el->offset=offsetof(dae_node_type,el_node);
        el->objtypeid=dae_ID_NODE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=6;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_node_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=7;
        ++el;
        t->elems = elems;
        t->numelems = 14;
    }
    ++t;
    t->name = "library_nodes_type";
    t->size = sizeof(dae_library_nodes_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_NODES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_nodes_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_nodes_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_nodes_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="node";
        el->offset=offsetof(dae_library_nodes_type,el_node);
        el->objtypeid=dae_ID_NODE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_nodes_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "visual_scene_type";
    t->size = sizeof(dae_visual_scene_type);
    t->flags= 0;
    t->objtypeid = dae_ID_VISUAL_SCENE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_visual_scene_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_visual_scene_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_visual_scene_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="node";
        el->offset=offsetof(dae_visual_scene_type,el_node);
        el->objtypeid=dae_ID_NODE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="evaluate_scene";
        el->offset=offsetof(dae_visual_scene_type,el_evaluate_scene);
        el->objtypeid=dae_ID_VISUAL_SCENE_TYPE_EVALUATE_SCENE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_visual_scene_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "visual_scene_type_evaluate_scene";
    t->size = sizeof(dae_visual_scene_type_evaluate_scene);
    t->flags= 0;
    t->objtypeid = dae_ID_VISUAL_SCENE_TYPE_EVALUATE_SCENE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="enable";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene,at_enable);
        at->objtypeid=dae_ID_BOOLEAN;
        at->min=0;
        at->max=1;
        at->deflt="true";
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_visual_scene_type_evaluate_scene,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="render";
        el->offset=offsetof(dae_visual_scene_type_evaluate_scene,el_render);
        el->objtypeid=dae_ID_VISUAL_SCENE_TYPE_EVALUATE_SCENE_RENDER;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_visual_scene_type_evaluate_scene,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "visual_scene_type_evaluate_scene_render";
    t->size = sizeof(dae_visual_scene_type_evaluate_scene_render);
    t->flags= 0;
    t->objtypeid = dae_ID_VISUAL_SCENE_TYPE_EVALUATE_SCENE_RENDER;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene_render,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene_render,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="camera_node";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene_render,at_camera_node);
        at->objtypeid=dae_ID_ANYURI;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="layer";
        el->offset=offsetof(dae_visual_scene_type_evaluate_scene_render,el_layer);
        el->objtypeid=dae_ID_NCNAME;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="instance_material";
        el->offset=offsetof(dae_visual_scene_type_evaluate_scene_render,el_instance_material);
        el->objtypeid=dae_ID_VISUAL_SCENE_TYPE_EVALUATE_SCENE_RENDER_INSTANCE_MATERIAL;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_visual_scene_type_evaluate_scene_render,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "visual_scene_type_evaluate_scene_render_instance_material";
    t->size = sizeof(dae_visual_scene_type_evaluate_scene_render_instance_material);
    t->flags= 0;
    t->objtypeid = dae_ID_VISUAL_SCENE_TYPE_EVALUATE_SCENE_RENDER_INSTANCE_MATERIAL;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene_render_instance_material,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="technique_override";
        el->offset=offsetof(dae_visual_scene_type_evaluate_scene_render_instance_material,el_technique_override);
        el->objtypeid=dae_ID_VISUAL_SCENE_TYPE_EVALUATE_SCENE_RENDER_INSTANCE_MATERIAL_TECHNIQUE_OVERRIDE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bind";
        el->offset=offsetof(dae_visual_scene_type_evaluate_scene_render_instance_material,el_bind);
        el->objtypeid=dae_ID_VISUAL_SCENE_TYPE_EVALUATE_SCENE_RENDER_INSTANCE_MATERIAL_BIND;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_visual_scene_type_evaluate_scene_render_instance_material,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "visual_scene_type_evaluate_scene_render_instance_material_technique_override";
    t->size = sizeof(dae_visual_scene_type_evaluate_scene_render_instance_material_technique_override);
    t->flags= 0;
    t->objtypeid = dae_ID_VISUAL_SCENE_TYPE_EVALUATE_SCENE_RENDER_INSTANCE_MATERIAL_TECHNIQUE_OVERRIDE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene_render_instance_material_technique_override,at_ref);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        at->name="pass";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene_render_instance_material_technique_override,at_pass);
        at->objtypeid=dae_ID_NCNAME;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "visual_scene_type_evaluate_scene_render_instance_material_bind";
    t->size = sizeof(dae_visual_scene_type_evaluate_scene_render_instance_material_bind);
    t->flags= 0;
    t->objtypeid = dae_ID_VISUAL_SCENE_TYPE_EVALUATE_SCENE_RENDER_INSTANCE_MATERIAL_BIND;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="semantic";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene_render_instance_material_bind,at_semantic);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        at->name="target";
        at->offset=offsetof(dae_visual_scene_type_evaluate_scene_render_instance_material_bind,at_target);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "library_visual_scenes_type";
    t->size = sizeof(dae_library_visual_scenes_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_VISUAL_SCENES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_visual_scenes_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_visual_scenes_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_visual_scenes_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="visual_scene";
        el->offset=offsetof(dae_library_visual_scenes_type,el_visual_scene);
        el->objtypeid=dae_ID_VISUAL_SCENE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_visual_scenes_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "rigid_body_type";
    t->size = sizeof(dae_rigid_body_type);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_BODY_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_rigid_body_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_rigid_body_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_rigid_body_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="technique_common";
        el->offset=offsetof(dae_rigid_body_type,el_technique_common);
        el->objtypeid=dae_ID_RIGID_BODY_TYPE_TECHNIQUE_COMMON;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_rigid_body_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_rigid_body_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "rigid_body_type_technique_common";
    t->size = sizeof(dae_rigid_body_type_technique_common);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_BODY_TYPE_TECHNIQUE_COMMON;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="dynamic";
        el->offset=offsetof(dae_rigid_body_type_technique_common,el_dynamic);
        el->objtypeid=dae_ID_RIGID_BODY_TYPE_TECHNIQUE_COMMON_DYNAMIC;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mass";
        el->offset=offsetof(dae_rigid_body_type_technique_common,el_mass);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="mass_frame";
        el->offset=offsetof(dae_rigid_body_type_technique_common,el_mass_frame);
        el->objtypeid=dae_ID_RIGID_BODY_TYPE_TECHNIQUE_COMMON_MASS_FRAME;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="inertia";
        el->offset=offsetof(dae_rigid_body_type_technique_common,el_inertia);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="instance_physics_material";
        el->offset=offsetof(dae_rigid_body_type_technique_common,el_instance_physics_material);
        el->objtypeid=dae_ID_INSTANCE_PHYSICS_MATERIAL_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="physics_material";
        el->offset=offsetof(dae_rigid_body_type_technique_common,el_physics_material);
        el->objtypeid=dae_ID_PHYSICS_MATERIAL_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="shape";
        el->offset=offsetof(dae_rigid_body_type_technique_common,el_shape);
        el->objtypeid=dae_ID_RIGID_BODY_TYPE_TECHNIQUE_COMMON_SHAPE;
        el->min=1;
        el->max=-1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "rigid_body_type_technique_common_dynamic";
    t->size = sizeof(dae_rigid_body_type_technique_common_dynamic);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_BODY_TYPE_TECHNIQUE_COMMON_DYNAMIC;
    t->dataoffset=offsetof(dae_rigid_body_type_technique_common_dynamic,data);
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_rigid_body_type_technique_common_dynamic,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "rigid_body_type_technique_common_mass_frame";
    t->size = sizeof(dae_rigid_body_type_technique_common_mass_frame);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_BODY_TYPE_TECHNIQUE_COMMON_MASS_FRAME;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="translate";
        el->offset=offsetof(dae_rigid_body_type_technique_common_mass_frame,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="rotate";
        el->offset=offsetof(dae_rigid_body_type_technique_common_mass_frame,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "rigid_body_type_technique_common_shape";
    t->size = sizeof(dae_rigid_body_type_technique_common_shape);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_BODY_TYPE_TECHNIQUE_COMMON_SHAPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[14];
        dae_obj_memberdef* el = elems;
        el->name="hollow";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_hollow);
        el->objtypeid=dae_ID_RIGID_BODY_TYPE_TECHNIQUE_COMMON_SHAPE_HOLLOW;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mass";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_mass);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="density";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_density);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="instance_physics_material";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_instance_physics_material);
        el->objtypeid=dae_ID_INSTANCE_PHYSICS_MATERIAL_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="physics_material";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_physics_material);
        el->objtypeid=dae_ID_PHYSICS_MATERIAL_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="instance_geometry";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_instance_geometry);
        el->objtypeid=dae_ID_INSTANCE_GEOMETRY_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="plane";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_plane);
        el->objtypeid=dae_ID_PLANE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="box";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_box);
        el->objtypeid=dae_ID_BOX_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sphere";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_sphere);
        el->objtypeid=dae_ID_SPHERE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="cylinder";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_cylinder);
        el->objtypeid=dae_ID_CYLINDER_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="capsule";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_capsule);
        el->objtypeid=dae_ID_CAPSULE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="translate";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="rotate";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_rigid_body_type_technique_common_shape,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 14;
    }
    ++t;
    t->name = "rigid_body_type_technique_common_shape_hollow";
    t->size = sizeof(dae_rigid_body_type_technique_common_shape_hollow);
    t->flags= 0;
    t->objtypeid = dae_ID_RIGID_BODY_TYPE_TECHNIQUE_COMMON_SHAPE_HOLLOW;
    t->dataoffset=offsetof(dae_rigid_body_type_technique_common_shape_hollow,data);
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_rigid_body_type_technique_common_shape_hollow,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "instance_rigid_body_type";
    t->size = sizeof(dae_instance_rigid_body_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_RIGID_BODY_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="body";
        at->offset=offsetof(dae_instance_rigid_body_type,at_body);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_rigid_body_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_rigid_body_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="target";
        at->offset=offsetof(dae_instance_rigid_body_type,at_target);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="technique_common";
        el->offset=offsetof(dae_instance_rigid_body_type,el_technique_common);
        el->objtypeid=dae_ID_INSTANCE_RIGID_BODY_TYPE_TECHNIQUE_COMMON;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_instance_rigid_body_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_instance_rigid_body_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "instance_rigid_body_type_technique_common";
    t->size = sizeof(dae_instance_rigid_body_type_technique_common);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_RIGID_BODY_TYPE_TECHNIQUE_COMMON;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[9];
        dae_obj_memberdef* el = elems;
        el->name="angular_velocity";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common,el_angular_velocity);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="velocity";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common,el_velocity);
        el->objtypeid=dae_ID_FLOAT3_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="dynamic";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common,el_dynamic);
        el->objtypeid=dae_ID_INSTANCE_RIGID_BODY_TYPE_TECHNIQUE_COMMON_DYNAMIC;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="mass";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common,el_mass);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="mass_frame";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common,el_mass_frame);
        el->objtypeid=dae_ID_INSTANCE_RIGID_BODY_TYPE_TECHNIQUE_COMMON_MASS_FRAME;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="inertia";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common,el_inertia);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="instance_physics_material";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common,el_instance_physics_material);
        el->objtypeid=dae_ID_INSTANCE_PHYSICS_MATERIAL_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="physics_material";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common,el_physics_material);
        el->objtypeid=dae_ID_PHYSICS_MATERIAL_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        el->name="shape";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common,el_shape);
        el->objtypeid=dae_ID_INSTANCE_RIGID_BODY_TYPE_TECHNIQUE_COMMON_SHAPE;
        el->min=0;
        el->max=-1;
        el->seq=6;
        ++el;
        t->elems = elems;
        t->numelems = 9;
    }
    ++t;
    t->name = "instance_rigid_body_type_technique_common_dynamic";
    t->size = sizeof(dae_instance_rigid_body_type_technique_common_dynamic);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_RIGID_BODY_TYPE_TECHNIQUE_COMMON_DYNAMIC;
    t->dataoffset=offsetof(dae_instance_rigid_body_type_technique_common_dynamic,data);
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_instance_rigid_body_type_technique_common_dynamic,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "instance_rigid_body_type_technique_common_mass_frame";
    t->size = sizeof(dae_instance_rigid_body_type_technique_common_mass_frame);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_RIGID_BODY_TYPE_TECHNIQUE_COMMON_MASS_FRAME;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="translate";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_mass_frame,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="rotate";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_mass_frame,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "instance_rigid_body_type_technique_common_shape";
    t->size = sizeof(dae_instance_rigid_body_type_technique_common_shape);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_RIGID_BODY_TYPE_TECHNIQUE_COMMON_SHAPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[14];
        dae_obj_memberdef* el = elems;
        el->name="hollow";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_hollow);
        el->objtypeid=dae_ID_INSTANCE_RIGID_BODY_TYPE_TECHNIQUE_COMMON_SHAPE_HOLLOW;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="mass";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_mass);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="density";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_density);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="instance_physics_material";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_instance_physics_material);
        el->objtypeid=dae_ID_INSTANCE_PHYSICS_MATERIAL_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="physics_material";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_physics_material);
        el->objtypeid=dae_ID_PHYSICS_MATERIAL_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="instance_geometry";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_instance_geometry);
        el->objtypeid=dae_ID_INSTANCE_GEOMETRY_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="plane";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_plane);
        el->objtypeid=dae_ID_PLANE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="box";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_box);
        el->objtypeid=dae_ID_BOX_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="sphere";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_sphere);
        el->objtypeid=dae_ID_SPHERE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="cylinder";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_cylinder);
        el->objtypeid=dae_ID_CYLINDER_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="capsule";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_capsule);
        el->objtypeid=dae_ID_CAPSULE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="translate";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="rotate";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 14;
    }
    ++t;
    t->name = "instance_rigid_body_type_technique_common_shape_hollow";
    t->size = sizeof(dae_instance_rigid_body_type_technique_common_shape_hollow);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_RIGID_BODY_TYPE_TECHNIQUE_COMMON_SHAPE_HOLLOW;
    t->dataoffset=offsetof(dae_instance_rigid_body_type_technique_common_shape_hollow,data);
    t->datatypeid=dae_NATIVE_BOOL32;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_instance_rigid_body_type_technique_common_shape_hollow,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "instance_physics_model_type";
    t->size = sizeof(dae_instance_physics_model_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_PHYSICS_MODEL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_physics_model_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_physics_model_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_physics_model_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="parent";
        at->offset=offsetof(dae_instance_physics_model_type,at_parent);
        at->objtypeid=dae_ID_ANYURI;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="instance_force_field";
        el->offset=offsetof(dae_instance_physics_model_type,el_instance_force_field);
        el->objtypeid=dae_ID_INSTANCE_FORCE_FIELD_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="instance_rigid_body";
        el->offset=offsetof(dae_instance_physics_model_type,el_instance_rigid_body);
        el->objtypeid=dae_ID_INSTANCE_RIGID_BODY_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="instance_rigid_constraint";
        el->offset=offsetof(dae_instance_physics_model_type,el_instance_rigid_constraint);
        el->objtypeid=dae_ID_INSTANCE_RIGID_CONSTRAINT_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_instance_physics_model_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "physics_model_type";
    t->size = sizeof(dae_physics_model_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PHYSICS_MODEL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_physics_model_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_physics_model_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_physics_model_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="rigid_body";
        el->offset=offsetof(dae_physics_model_type,el_rigid_body);
        el->objtypeid=dae_ID_RIGID_BODY_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="rigid_constraint";
        el->offset=offsetof(dae_physics_model_type,el_rigid_constraint);
        el->objtypeid=dae_ID_RIGID_CONSTRAINT_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="instance_physics_model";
        el->offset=offsetof(dae_physics_model_type,el_instance_physics_model);
        el->objtypeid=dae_ID_INSTANCE_PHYSICS_MODEL_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_physics_model_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "library_physics_models_type";
    t->size = sizeof(dae_library_physics_models_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_PHYSICS_MODELS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_physics_models_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_physics_models_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_physics_models_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="physics_model";
        el->offset=offsetof(dae_library_physics_models_type,el_physics_model);
        el->objtypeid=dae_ID_PHYSICS_MODEL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_physics_models_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "physics_scene_type";
    t->size = sizeof(dae_physics_scene_type);
    t->flags= 0;
    t->objtypeid = dae_ID_PHYSICS_SCENE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_physics_scene_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_physics_scene_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[6];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_physics_scene_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="instance_force_field";
        el->offset=offsetof(dae_physics_scene_type,el_instance_force_field);
        el->objtypeid=dae_ID_INSTANCE_FORCE_FIELD_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="instance_physics_model";
        el->offset=offsetof(dae_physics_scene_type,el_instance_physics_model);
        el->objtypeid=dae_ID_INSTANCE_PHYSICS_MODEL_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="technique_common";
        el->offset=offsetof(dae_physics_scene_type,el_technique_common);
        el->objtypeid=dae_ID_PHYSICS_SCENE_TYPE_TECHNIQUE_COMMON;
        el->min=1;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_physics_scene_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_physics_scene_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=5;
        ++el;
        t->elems = elems;
        t->numelems = 6;
    }
    ++t;
    t->name = "physics_scene_type_technique_common";
    t->size = sizeof(dae_physics_scene_type_technique_common);
    t->flags= 0;
    t->objtypeid = dae_ID_PHYSICS_SCENE_TYPE_TECHNIQUE_COMMON;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="gravity";
        el->offset=offsetof(dae_physics_scene_type_technique_common,el_gravity);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT3_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="time_step";
        el->offset=offsetof(dae_physics_scene_type_technique_common,el_time_step);
        el->objtypeid=dae_ID_TARGETABLE_FLOAT_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "library_physics_scenes_type";
    t->size = sizeof(dae_library_physics_scenes_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_PHYSICS_SCENES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_physics_scenes_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_physics_scenes_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_physics_scenes_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="physics_scene";
        el->offset=offsetof(dae_library_physics_scenes_type,el_physics_scene);
        el->objtypeid=dae_ID_PHYSICS_SCENE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_physics_scenes_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "kinematics_newparam_type";
    t->size = sizeof(dae_kinematics_newparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_NEWPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_kinematics_newparam_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="float";
        el->offset=offsetof(dae_kinematics_newparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_kinematics_newparam_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="SIDREF";
        el->offset=offsetof(dae_kinematics_newparam_type,el_SIDREF);
        el->objtypeid=dae_ID_SIDREF_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_kinematics_newparam_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "kinematics_frame_type";
    t->size = sizeof(dae_kinematics_frame_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_FRAME_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="link";
        at->offset=offsetof(dae_kinematics_frame_type,at_link);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="translate";
        el->offset=offsetof(dae_kinematics_frame_type,el_translate);
        el->objtypeid=dae_ID_TRANSLATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="rotate";
        el->offset=offsetof(dae_kinematics_frame_type,el_rotate);
        el->objtypeid=dae_ID_ROTATE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "common_param_type";
    t->size = sizeof(dae_common_param_type);
    t->flags= 0;
    t->objtypeid = dae_ID_COMMON_PARAM_TYPE;
    t->dataoffset=0;
    t->datatypeid=dae_NATIVE_STRING;
    t->datamin=1;
    t->datamax=1;
    ++t;
    t->name = "common_bool_or_param_type";
    t->size = sizeof(dae_common_bool_or_param_type);
    t->flags= 0;
    t->objtypeid = dae_ID_COMMON_BOOL_OR_PARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="bool";
        el->offset=offsetof(dae_common_bool_or_param_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_common_bool_or_param_type,el_param);
        el->objtypeid=dae_ID_COMMON_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "common_float_or_param_type";
    t->size = sizeof(dae_common_float_or_param_type);
    t->flags= 0;
    t->objtypeid = dae_ID_COMMON_FLOAT_OR_PARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="float";
        el->offset=offsetof(dae_common_float_or_param_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_common_float_or_param_type,el_param);
        el->objtypeid=dae_ID_COMMON_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "kinematics_limits_type";
    t->size = sizeof(dae_kinematics_limits_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_LIMITS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="min";
        el->offset=offsetof(dae_kinematics_limits_type,el_min);
        el->objtypeid=dae_ID_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="max";
        el->offset=offsetof(dae_kinematics_limits_type,el_max);
        el->objtypeid=dae_ID_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "common_sidref_or_param_type";
    t->size = sizeof(dae_common_sidref_or_param_type);
    t->flags= 0;
    t->objtypeid = dae_ID_COMMON_SIDREF_OR_PARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="SIDREF";
        el->offset=offsetof(dae_common_sidref_or_param_type,el_SIDREF);
        el->objtypeid=dae_ID_SIDREF_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_common_sidref_or_param_type,el_param);
        el->objtypeid=dae_ID_COMMON_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "bind_joint_axis_type";
    t->size = sizeof(dae_bind_joint_axis_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BIND_JOINT_AXIS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="target";
        at->offset=offsetof(dae_bind_joint_axis_type,at_target);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="axis";
        el->offset=offsetof(dae_bind_joint_axis_type,el_axis);
        el->objtypeid=dae_ID_COMMON_SIDREF_OR_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="value";
        el->offset=offsetof(dae_bind_joint_axis_type,el_value);
        el->objtypeid=dae_ID_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "bind_kinematics_model_type";
    t->size = sizeof(dae_bind_kinematics_model_type);
    t->flags= 0;
    t->objtypeid = dae_ID_BIND_KINEMATICS_MODEL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="node";
        at->offset=offsetof(dae_bind_kinematics_model_type,at_node);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="SIDREF";
        el->offset=offsetof(dae_bind_kinematics_model_type,el_SIDREF);
        el->objtypeid=dae_ID_SIDREF_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_bind_kinematics_model_type,el_param);
        el->objtypeid=dae_ID_COMMON_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "kinematics_connect_param_type";
    t->size = sizeof(dae_kinematics_connect_param_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_CONNECT_PARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_kinematics_connect_param_type,at_ref);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "kinematics_setparam_type";
    t->size = sizeof(dae_kinematics_setparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_SETPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_kinematics_setparam_type,at_ref);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="float";
        el->offset=offsetof(dae_kinematics_setparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_kinematics_setparam_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="SIDREF";
        el->offset=offsetof(dae_kinematics_setparam_type,el_SIDREF);
        el->objtypeid=dae_ID_SIDREF_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_kinematics_setparam_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="connect_param";
        el->offset=offsetof(dae_kinematics_setparam_type,el_connect_param);
        el->objtypeid=dae_ID_KINEMATICS_CONNECT_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "instance_kinematics_scene_type";
    t->size = sizeof(dae_instance_kinematics_scene_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_KINEMATICS_SCENE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_kinematics_scene_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_kinematics_scene_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_kinematics_scene_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[6];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_instance_kinematics_scene_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_instance_kinematics_scene_type,el_newparam);
        el->objtypeid=dae_ID_KINEMATICS_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="setparam";
        el->offset=offsetof(dae_instance_kinematics_scene_type,el_setparam);
        el->objtypeid=dae_ID_KINEMATICS_SETPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="bind_kinematics_model";
        el->offset=offsetof(dae_instance_kinematics_scene_type,el_bind_kinematics_model);
        el->objtypeid=dae_ID_BIND_KINEMATICS_MODEL_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="bind_joint_axis";
        el->offset=offsetof(dae_instance_kinematics_scene_type,el_bind_joint_axis);
        el->objtypeid=dae_ID_BIND_JOINT_AXIS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=4;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_instance_kinematics_scene_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=5;
        ++el;
        t->elems = elems;
        t->numelems = 6;
    }
    ++t;
    t->name = "kinematics_index_type";
    t->size = sizeof(dae_kinematics_index_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_INDEX_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="semantic";
        at->offset=offsetof(dae_kinematics_index_type,at_semantic);
        at->objtypeid=dae_ID_NMTOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="int";
        el->offset=offsetof(dae_kinematics_index_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_kinematics_index_type,el_param);
        el->objtypeid=dae_ID_COMMON_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "kinematics_param_type";
    t->size = sizeof(dae_kinematics_param_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_PARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_kinematics_param_type,at_ref);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    ++t;
    t->name = "kinematics_bind_type";
    t->size = sizeof(dae_kinematics_bind_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_BIND_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="symbol";
        at->offset=offsetof(dae_kinematics_bind_type,at_symbol);
        at->objtypeid=dae_ID_NCNAME;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="param";
        el->offset=offsetof(dae_kinematics_bind_type,el_param);
        el->objtypeid=dae_ID_KINEMATICS_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_kinematics_bind_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="float";
        el->offset=offsetof(dae_kinematics_bind_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_kinematics_bind_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="SIDREF";
        el->offset=offsetof(dae_kinematics_bind_type,el_SIDREF);
        el->objtypeid=dae_ID_SIDREF_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "motion_axis_info_type";
    t->size = sizeof(dae_motion_axis_info_type);
    t->flags= 0;
    t->objtypeid = dae_ID_MOTION_AXIS_INFO_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_motion_axis_info_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="axis";
        at->offset=offsetof(dae_motion_axis_info_type,at_axis);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_motion_axis_info_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="bind";
        el->offset=offsetof(dae_motion_axis_info_type,el_bind);
        el->objtypeid=dae_ID_KINEMATICS_BIND_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_motion_axis_info_type,el_newparam);
        el->objtypeid=dae_ID_KINEMATICS_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="setparam";
        el->offset=offsetof(dae_motion_axis_info_type,el_setparam);
        el->objtypeid=dae_ID_KINEMATICS_SETPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="speed";
        el->offset=offsetof(dae_motion_axis_info_type,el_speed);
        el->objtypeid=dae_ID_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="acceleration";
        el->offset=offsetof(dae_motion_axis_info_type,el_acceleration);
        el->objtypeid=dae_ID_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="deceleration";
        el->offset=offsetof(dae_motion_axis_info_type,el_deceleration);
        el->objtypeid=dae_ID_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="jerk";
        el->offset=offsetof(dae_motion_axis_info_type,el_jerk);
        el->objtypeid=dae_ID_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "instance_kinematics_model_type";
    t->size = sizeof(dae_instance_kinematics_model_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_KINEMATICS_MODEL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="url";
        at->offset=offsetof(dae_instance_kinematics_model_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_instance_kinematics_model_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_kinematics_model_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="bind";
        el->offset=offsetof(dae_instance_kinematics_model_type,el_bind);
        el->objtypeid=dae_ID_KINEMATICS_BIND_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_instance_kinematics_model_type,el_newparam);
        el->objtypeid=dae_ID_KINEMATICS_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="setparam";
        el->offset=offsetof(dae_instance_kinematics_model_type,el_setparam);
        el->objtypeid=dae_ID_KINEMATICS_SETPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_instance_kinematics_model_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "instance_articulated_system_type";
    t->size = sizeof(dae_instance_articulated_system_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_ARTICULATED_SYSTEM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_instance_articulated_system_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="url";
        at->offset=offsetof(dae_instance_articulated_system_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=1;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_articulated_system_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="bind";
        el->offset=offsetof(dae_instance_articulated_system_type,el_bind);
        el->objtypeid=dae_ID_KINEMATICS_BIND_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="setparam";
        el->offset=offsetof(dae_instance_articulated_system_type,el_setparam);
        el->objtypeid=dae_ID_KINEMATICS_SETPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_instance_articulated_system_type,el_newparam);
        el->objtypeid=dae_ID_KINEMATICS_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_instance_articulated_system_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "kinematics_scene_type";
    t->size = sizeof(dae_kinematics_scene_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_SCENE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_kinematics_scene_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_kinematics_scene_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_kinematics_scene_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="instance_kinematics_model";
        el->offset=offsetof(dae_kinematics_scene_type,el_instance_kinematics_model);
        el->objtypeid=dae_ID_INSTANCE_KINEMATICS_MODEL_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="instance_articulated_system";
        el->offset=offsetof(dae_kinematics_scene_type,el_instance_articulated_system);
        el->objtypeid=dae_ID_INSTANCE_ARTICULATED_SYSTEM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_kinematics_scene_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "library_kinematics_scenes_type";
    t->size = sizeof(dae_library_kinematics_scenes_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_KINEMATICS_SCENES_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_kinematics_scenes_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_kinematics_scenes_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_kinematics_scenes_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="kinematics_scene";
        el->offset=offsetof(dae_library_kinematics_scenes_type,el_kinematics_scene);
        el->objtypeid=dae_ID_KINEMATICS_SCENE_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_kinematics_scenes_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "common_float2_or_param_type";
    t->size = sizeof(dae_common_float2_or_param_type);
    t->flags= 0;
    t->objtypeid = dae_ID_COMMON_FLOAT2_OR_PARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="float2";
        el->offset=offsetof(dae_common_float2_or_param_type,el_float2);
        el->objtypeid=dae_ID_FLOAT2_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="param";
        el->offset=offsetof(dae_common_float2_or_param_type,el_param);
        el->objtypeid=dae_ID_COMMON_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "motion_effector_info_type";
    t->size = sizeof(dae_motion_effector_info_type);
    t->flags= 0;
    t->objtypeid = dae_ID_MOTION_EFFECTOR_INFO_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_motion_effector_info_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_motion_effector_info_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="bind";
        el->offset=offsetof(dae_motion_effector_info_type,el_bind);
        el->objtypeid=dae_ID_KINEMATICS_BIND_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="newparam";
        el->offset=offsetof(dae_motion_effector_info_type,el_newparam);
        el->objtypeid=dae_ID_KINEMATICS_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="setparam";
        el->offset=offsetof(dae_motion_effector_info_type,el_setparam);
        el->objtypeid=dae_ID_KINEMATICS_SETPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="speed";
        el->offset=offsetof(dae_motion_effector_info_type,el_speed);
        el->objtypeid=dae_ID_COMMON_FLOAT2_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="acceleration";
        el->offset=offsetof(dae_motion_effector_info_type,el_acceleration);
        el->objtypeid=dae_ID_COMMON_FLOAT2_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="deceleration";
        el->offset=offsetof(dae_motion_effector_info_type,el_deceleration);
        el->objtypeid=dae_ID_COMMON_FLOAT2_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=5;
        ++el;
        el->name="jerk";
        el->offset=offsetof(dae_motion_effector_info_type,el_jerk);
        el->objtypeid=dae_ID_COMMON_FLOAT2_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=6;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "motion_technique_type";
    t->size = sizeof(dae_motion_technique_type);
    t->flags= 0;
    t->objtypeid = dae_ID_MOTION_TECHNIQUE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="axis_info";
        el->offset=offsetof(dae_motion_technique_type,el_axis_info);
        el->objtypeid=dae_ID_MOTION_AXIS_INFO_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="effector_info";
        el->offset=offsetof(dae_motion_technique_type,el_effector_info);
        el->objtypeid=dae_ID_MOTION_EFFECTOR_INFO_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "motion_type";
    t->size = sizeof(dae_motion_type);
    t->flags= 0;
    t->objtypeid = dae_ID_MOTION_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="instance_articulated_system";
        el->offset=offsetof(dae_motion_type,el_instance_articulated_system);
        el->objtypeid=dae_ID_INSTANCE_ARTICULATED_SYSTEM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="technique_common";
        el->offset=offsetof(dae_motion_type,el_technique_common);
        el->objtypeid=dae_ID_MOTION_TECHNIQUE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_motion_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_motion_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "minmax_type";
    t->size = sizeof(dae_minmax_type);
    t->flags= 0;
    t->objtypeid = dae_ID_MINMAX_TYPE;
    t->dataoffset=offsetof(dae_minmax_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=1;
    t->datamax=1;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="name";
        at->offset=offsetof(dae_minmax_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_minmax_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "joint_limits_type";
    t->size = sizeof(dae_joint_limits_type);
    t->flags= 0;
    t->objtypeid = dae_ID_JOINT_LIMITS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="min";
        el->offset=offsetof(dae_joint_limits_type,el_min);
        el->objtypeid=dae_ID_MINMAX_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="max";
        el->offset=offsetof(dae_joint_limits_type,el_max);
        el->objtypeid=dae_ID_MINMAX_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "axis_type";
    t->size = sizeof(dae_axis_type);
    t->flags= 0;
    t->objtypeid = dae_ID_AXIS_TYPE;
    t->dataoffset=offsetof(dae_axis_type,data);
    t->datatypeid=dae_NATIVE_FLOAT;
    t->datamin=3;
    t->datamax=3;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_axis_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_axis_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    ++t;
    t->name = "axis_constraint_type";
    t->size = sizeof(dae_axis_constraint_type);
    t->flags= 0;
    t->objtypeid = dae_ID_AXIS_CONSTRAINT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_axis_constraint_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[2];
        dae_obj_memberdef* el = elems;
        el->name="axis";
        el->offset=offsetof(dae_axis_constraint_type,el_axis);
        el->objtypeid=dae_ID_AXIS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="limits";
        el->offset=offsetof(dae_axis_constraint_type,el_limits);
        el->objtypeid=dae_ID_JOINT_LIMITS_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 2;
    }
    ++t;
    t->name = "joint_type";
    t->size = sizeof(dae_joint_type);
    t->flags= 0;
    t->objtypeid = dae_ID_JOINT_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_joint_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_joint_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_joint_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="prismatic";
        el->offset=offsetof(dae_joint_type,el_prismatic);
        el->objtypeid=dae_ID_AXIS_CONSTRAINT_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="revolute";
        el->offset=offsetof(dae_joint_type,el_revolute);
        el->objtypeid=dae_ID_AXIS_CONSTRAINT_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_joint_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "library_joints_type";
    t->size = sizeof(dae_library_joints_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_JOINTS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_joints_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_joints_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_joints_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="joint";
        el->offset=offsetof(dae_library_joints_type,el_joint);
        el->objtypeid=dae_ID_JOINT_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_joints_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "formula_technique_type";
    t->size = sizeof(dae_formula_technique_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FORMULA_TECHNIQUE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="math:math";
        el->offset=offsetof(dae_formula_technique_type,el_math_math);
        el->objtypeid=dae_ID_MATH;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "formula_newparam_type";
    t->size = sizeof(dae_formula_newparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FORMULA_NEWPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_formula_newparam_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="float";
        el->offset=offsetof(dae_formula_newparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_formula_newparam_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="SIDREF";
        el->offset=offsetof(dae_formula_newparam_type,el_SIDREF);
        el->objtypeid=dae_ID_SIDREF_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_formula_newparam_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "formula_setparam_type";
    t->size = sizeof(dae_formula_setparam_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FORMULA_SETPARAM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="ref";
        at->offset=offsetof(dae_formula_setparam_type,at_ref);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="float";
        el->offset=offsetof(dae_formula_setparam_type,el_float);
        el->objtypeid=dae_ID_FLOAT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="int";
        el->offset=offsetof(dae_formula_setparam_type,el_int);
        el->objtypeid=dae_ID_INT_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="SIDREF";
        el->offset=offsetof(dae_formula_setparam_type,el_SIDREF);
        el->objtypeid=dae_ID_SIDREF_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="bool";
        el->offset=offsetof(dae_formula_setparam_type,el_bool);
        el->objtypeid=dae_ID_BOOLEAN;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="connect_param";
        el->offset=offsetof(dae_formula_setparam_type,el_connect_param);
        el->objtypeid=dae_ID_KINEMATICS_CONNECT_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "instance_formula_type";
    t->size = sizeof(dae_instance_formula_type);
    t->flags= 0;
    t->objtypeid = dae_ID_INSTANCE_FORMULA_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_instance_formula_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_instance_formula_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="url";
        at->offset=offsetof(dae_instance_formula_type,at_url);
        at->objtypeid=dae_ID_ANYURI;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="setparam";
        el->offset=offsetof(dae_instance_formula_type,el_setparam);
        el->objtypeid=dae_ID_FORMULA_SETPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    t->name = "animation_clip_type";
    t->size = sizeof(dae_animation_clip_type);
    t->flags= 0;
    t->objtypeid = dae_ID_ANIMATION_CLIP_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[4];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_animation_clip_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_animation_clip_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="start";
        at->offset=offsetof(dae_animation_clip_type,at_start);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        at->deflt="0.0";
        ++at;
        at->name="end";
        at->offset=offsetof(dae_animation_clip_type,at_end);
        at->objtypeid=dae_ID_FLOAT_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 4;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_animation_clip_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="instance_animation";
        el->offset=offsetof(dae_animation_clip_type,el_instance_animation);
        el->objtypeid=dae_ID_INSTANCE_WITH_EXTRA_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="instance_formula";
        el->offset=offsetof(dae_animation_clip_type,el_instance_formula);
        el->objtypeid=dae_ID_INSTANCE_FORMULA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_animation_clip_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "library_animation_clips_type";
    t->size = sizeof(dae_library_animation_clips_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_ANIMATION_CLIPS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_animation_clips_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_animation_clips_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_animation_clips_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="animation_clip";
        el->offset=offsetof(dae_library_animation_clips_type,el_animation_clip);
        el->objtypeid=dae_ID_ANIMATION_CLIP_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_animation_clips_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "formula_type";
    t->size = sizeof(dae_formula_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FORMULA_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_formula_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_formula_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="sid";
        at->offset=offsetof(dae_formula_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="newparam";
        el->offset=offsetof(dae_formula_type,el_newparam);
        el->objtypeid=dae_ID_FORMULA_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="target";
        el->offset=offsetof(dae_formula_type,el_target);
        el->objtypeid=dae_ID_COMMON_FLOAT_OR_PARAM_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="technique_common";
        el->offset=offsetof(dae_formula_type,el_technique_common);
        el->objtypeid=dae_ID_FORMULA_TECHNIQUE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_formula_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "kinematics_model_technique_type";
    t->size = sizeof(dae_kinematics_model_technique_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_MODEL_TECHNIQUE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[6];
        dae_obj_memberdef* el = elems;
        el->name="newparam";
        el->offset=offsetof(dae_kinematics_model_technique_type,el_newparam);
        el->objtypeid=dae_ID_KINEMATICS_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="instance_joint";
        el->offset=offsetof(dae_kinematics_model_technique_type,el_instance_joint);
        el->objtypeid=dae_ID_INSTANCE_JOINT_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="joint";
        el->offset=offsetof(dae_kinematics_model_technique_type,el_joint);
        el->objtypeid=dae_ID_JOINT_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="link";
        el->offset=offsetof(dae_kinematics_model_technique_type,el_link);
        el->objtypeid=dae_ID_LINK_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="formula";
        el->offset=offsetof(dae_kinematics_model_technique_type,el_formula);
        el->objtypeid=dae_ID_FORMULA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="instance_formula";
        el->offset=offsetof(dae_kinematics_model_technique_type,el_instance_formula);
        el->objtypeid=dae_ID_INSTANCE_FORMULA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 6;
    }
    ++t;
    t->name = "kinematics_model_type";
    t->size = sizeof(dae_kinematics_model_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_MODEL_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_kinematics_model_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_kinematics_model_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_kinematics_model_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="technique_common";
        el->offset=offsetof(dae_kinematics_model_type,el_technique_common);
        el->objtypeid=dae_ID_KINEMATICS_MODEL_TECHNIQUE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_kinematics_model_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_kinematics_model_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "library_kinematics_models_type";
    t->size = sizeof(dae_library_kinematics_models_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_KINEMATICS_MODELS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_kinematics_models_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_kinematics_models_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_kinematics_models_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="kinematics_model";
        el->offset=offsetof(dae_library_kinematics_models_type,el_kinematics_model);
        el->objtypeid=dae_ID_KINEMATICS_MODEL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_kinematics_models_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "kinematics_axis_info_type";
    t->size = sizeof(dae_kinematics_axis_info_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_AXIS_INFO_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[3];
        dae_obj_memberdef* at = attribs;
        at->name="sid";
        at->offset=offsetof(dae_kinematics_axis_info_type,at_sid);
        at->objtypeid=dae_ID_SID_TYPE;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_kinematics_axis_info_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        at->name="axis";
        at->offset=offsetof(dae_kinematics_axis_info_type,at_axis);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 3;
    }
    {
        static dae_obj_memberdef elems[7];
        dae_obj_memberdef* el = elems;
        el->name="newparam";
        el->offset=offsetof(dae_kinematics_axis_info_type,el_newparam);
        el->objtypeid=dae_ID_KINEMATICS_NEWPARAM_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="active";
        el->offset=offsetof(dae_kinematics_axis_info_type,el_active);
        el->objtypeid=dae_ID_COMMON_BOOL_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="locked";
        el->offset=offsetof(dae_kinematics_axis_info_type,el_locked);
        el->objtypeid=dae_ID_COMMON_BOOL_OR_PARAM_TYPE;
        el->min=0;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="index";
        el->offset=offsetof(dae_kinematics_axis_info_type,el_index);
        el->objtypeid=dae_ID_KINEMATICS_INDEX_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        el->name="limits";
        el->offset=offsetof(dae_kinematics_axis_info_type,el_limits);
        el->objtypeid=dae_ID_KINEMATICS_LIMITS_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        el->name="formula";
        el->offset=offsetof(dae_kinematics_axis_info_type,el_formula);
        el->objtypeid=dae_ID_FORMULA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=5;
        ++el;
        el->name="instance_formula";
        el->offset=offsetof(dae_kinematics_axis_info_type,el_instance_formula);
        el->objtypeid=dae_ID_INSTANCE_FORMULA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=5;
        ++el;
        t->elems = elems;
        t->numelems = 7;
    }
    ++t;
    t->name = "kinematics_technique_type";
    t->size = sizeof(dae_kinematics_technique_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_TECHNIQUE_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[5];
        dae_obj_memberdef* el = elems;
        el->name="axis_info";
        el->offset=offsetof(dae_kinematics_technique_type,el_axis_info);
        el->objtypeid=dae_ID_KINEMATICS_AXIS_INFO_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="frame_origin";
        el->offset=offsetof(dae_kinematics_technique_type,el_frame_origin);
        el->objtypeid=dae_ID_KINEMATICS_FRAME_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="frame_tip";
        el->offset=offsetof(dae_kinematics_technique_type,el_frame_tip);
        el->objtypeid=dae_ID_KINEMATICS_FRAME_TYPE;
        el->min=1;
        el->max=1;
        el->seq=2;
        ++el;
        el->name="frame_tcp";
        el->offset=offsetof(dae_kinematics_technique_type,el_frame_tcp);
        el->objtypeid=dae_ID_KINEMATICS_FRAME_TYPE;
        el->min=0;
        el->max=1;
        el->seq=3;
        ++el;
        el->name="frame_object";
        el->offset=offsetof(dae_kinematics_technique_type,el_frame_object);
        el->objtypeid=dae_ID_KINEMATICS_FRAME_TYPE;
        el->min=0;
        el->max=1;
        el->seq=4;
        ++el;
        t->elems = elems;
        t->numelems = 5;
    }
    ++t;
    t->name = "kinematics_type";
    t->size = sizeof(dae_kinematics_type);
    t->flags= 0;
    t->objtypeid = dae_ID_KINEMATICS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="instance_kinematics_model";
        el->offset=offsetof(dae_kinematics_type,el_instance_kinematics_model);
        el->objtypeid=dae_ID_INSTANCE_KINEMATICS_MODEL_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="technique_common";
        el->offset=offsetof(dae_kinematics_type,el_technique_common);
        el->objtypeid=dae_ID_KINEMATICS_TECHNIQUE_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="technique";
        el->offset=offsetof(dae_kinematics_type,el_technique);
        el->objtypeid=dae_ID_TECHNIQUE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_kinematics_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "articulated_system_type";
    t->size = sizeof(dae_articulated_system_type);
    t->flags= 0;
    t->objtypeid = dae_ID_ARTICULATED_SYSTEM_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_articulated_system_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_articulated_system_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_articulated_system_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="kinematics";
        el->offset=offsetof(dae_articulated_system_type,el_kinematics);
        el->objtypeid=dae_ID_KINEMATICS_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="motion";
        el->offset=offsetof(dae_articulated_system_type,el_motion);
        el->objtypeid=dae_ID_MOTION_TYPE;
        el->min=1;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_articulated_system_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "library_articulated_systems_type";
    t->size = sizeof(dae_library_articulated_systems_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_ARTICULATED_SYSTEMS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_articulated_systems_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_articulated_systems_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_articulated_systems_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="articulated_system";
        el->offset=offsetof(dae_library_articulated_systems_type,el_articulated_system);
        el->objtypeid=dae_ID_ARTICULATED_SYSTEM_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_articulated_systems_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "library_formulas_type";
    t->size = sizeof(dae_library_formulas_type);
    t->flags= 0;
    t->objtypeid = dae_ID_LIBRARY_FORMULAS_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="id";
        at->offset=offsetof(dae_library_formulas_type,at_id);
        at->objtypeid=dae_ID_ID;
        at->min=0;
        at->max=1;
        ++at;
        at->name="name";
        at->offset=offsetof(dae_library_formulas_type,at_name);
        at->objtypeid=dae_ID_TOKEN;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[3];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_library_formulas_type,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=0;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="formula";
        el->offset=offsetof(dae_library_formulas_type,el_formula);
        el->objtypeid=dae_ID_FORMULA_TYPE;
        el->min=1;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_library_formulas_type,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 3;
    }
    ++t;
    t->name = "COLLADA";
    t->size = sizeof(dae_COLLADA);
    t->flags= 0;
    t->objtypeid = dae_ID_COLLADA;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[2];
        dae_obj_memberdef* at = attribs;
        at->name="version";
        at->offset=offsetof(dae_COLLADA,at_version);
        at->objtypeid=dae_ID_VERSION_ENUM;
        at->min=1;
        at->max=1;
        ++at;
        at->name="xml:base";
        at->offset=offsetof(dae_COLLADA,at_xml_base);
        at->objtypeid=dae_ID_ANYURI;
        at->min=0;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 2;
    }
    {
        static dae_obj_memberdef elems[23];
        dae_obj_memberdef* el = elems;
        el->name="asset";
        el->offset=offsetof(dae_COLLADA,el_asset);
        el->objtypeid=dae_ID_ASSET_TYPE;
        el->min=1;
        el->max=1;
        el->seq=0;
        ++el;
        el->name="library_animations";
        el->offset=offsetof(dae_COLLADA,el_library_animations);
        el->objtypeid=dae_ID_LIBRARY_ANIMATIONS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_animation_clips";
        el->offset=offsetof(dae_COLLADA,el_library_animation_clips);
        el->objtypeid=dae_ID_LIBRARY_ANIMATION_CLIPS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_cameras";
        el->offset=offsetof(dae_COLLADA,el_library_cameras);
        el->objtypeid=dae_ID_LIBRARY_CAMERAS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_controllers";
        el->offset=offsetof(dae_COLLADA,el_library_controllers);
        el->objtypeid=dae_ID_LIBRARY_CONTROLLERS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_geometries";
        el->offset=offsetof(dae_COLLADA,el_library_geometries);
        el->objtypeid=dae_ID_LIBRARY_GEOMETRIES_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_effects";
        el->offset=offsetof(dae_COLLADA,el_library_effects);
        el->objtypeid=dae_ID_LIBRARY_EFFECTS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_force_fields";
        el->offset=offsetof(dae_COLLADA,el_library_force_fields);
        el->objtypeid=dae_ID_LIBRARY_FORCE_FIELDS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_images";
        el->offset=offsetof(dae_COLLADA,el_library_images);
        el->objtypeid=dae_ID_LIBRARY_IMAGES_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_lights";
        el->offset=offsetof(dae_COLLADA,el_library_lights);
        el->objtypeid=dae_ID_LIBRARY_LIGHTS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_materials";
        el->offset=offsetof(dae_COLLADA,el_library_materials);
        el->objtypeid=dae_ID_LIBRARY_MATERIALS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_nodes";
        el->offset=offsetof(dae_COLLADA,el_library_nodes);
        el->objtypeid=dae_ID_LIBRARY_NODES_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_physics_materials";
        el->offset=offsetof(dae_COLLADA,el_library_physics_materials);
        el->objtypeid=dae_ID_LIBRARY_PHYSICS_MATERIALS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_physics_models";
        el->offset=offsetof(dae_COLLADA,el_library_physics_models);
        el->objtypeid=dae_ID_LIBRARY_PHYSICS_MODELS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_physics_scenes";
        el->offset=offsetof(dae_COLLADA,el_library_physics_scenes);
        el->objtypeid=dae_ID_LIBRARY_PHYSICS_SCENES_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_visual_scenes";
        el->offset=offsetof(dae_COLLADA,el_library_visual_scenes);
        el->objtypeid=dae_ID_LIBRARY_VISUAL_SCENES_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_joints";
        el->offset=offsetof(dae_COLLADA,el_library_joints);
        el->objtypeid=dae_ID_LIBRARY_JOINTS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_kinematics_models";
        el->offset=offsetof(dae_COLLADA,el_library_kinematics_models);
        el->objtypeid=dae_ID_LIBRARY_KINEMATICS_MODELS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_articulated_systems";
        el->offset=offsetof(dae_COLLADA,el_library_articulated_systems);
        el->objtypeid=dae_ID_LIBRARY_ARTICULATED_SYSTEMS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_kinematics_scenes";
        el->offset=offsetof(dae_COLLADA,el_library_kinematics_scenes);
        el->objtypeid=dae_ID_LIBRARY_KINEMATICS_SCENES_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="library_formulas";
        el->offset=offsetof(dae_COLLADA,el_library_formulas);
        el->objtypeid=dae_ID_LIBRARY_FORMULAS_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=1;
        ++el;
        el->name="scene";
        el->offset=offsetof(dae_COLLADA,el_scene);
        el->objtypeid=dae_ID_COLLADA_SCENE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_COLLADA,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        t->elems = elems;
        t->numelems = 23;
    }
    ++t;
    t->name = "COLLADA_scene";
    t->size = sizeof(dae_COLLADA_scene);
    t->flags= 0;
    t->objtypeid = dae_ID_COLLADA_SCENE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef elems[4];
        dae_obj_memberdef* el = elems;
        el->name="instance_physics_scene";
        el->offset=offsetof(dae_COLLADA_scene,el_instance_physics_scene);
        el->objtypeid=dae_ID_INSTANCE_WITH_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        el->name="instance_visual_scene";
        el->offset=offsetof(dae_COLLADA_scene,el_instance_visual_scene);
        el->objtypeid=dae_ID_INSTANCE_WITH_EXTRA_TYPE;
        el->min=0;
        el->max=1;
        el->seq=1;
        ++el;
        el->name="instance_kinematics_scene";
        el->offset=offsetof(dae_COLLADA_scene,el_instance_kinematics_scene);
        el->objtypeid=dae_ID_INSTANCE_KINEMATICS_SCENE_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=2;
        ++el;
        el->name="extra";
        el->offset=offsetof(dae_COLLADA_scene,el_extra);
        el->objtypeid=dae_ID_EXTRA_TYPE;
        el->min=0;
        el->max=-1;
        el->seq=3;
        ++el;
        t->elems = elems;
        t->numelems = 4;
    }
    ++t;
    t->name = "fx_surface_common_type";
    t->size = sizeof(dae_fx_surface_common_type);
    t->flags= 0;
    t->objtypeid = dae_ID_FX_SURFACE_COMMON_TYPE;
    t->datatypeid=dae_ID_INVALID;
    {
        static dae_obj_memberdef attribs[1];
        dae_obj_memberdef* at = attribs;
        at->name="type";
        at->offset=offsetof(dae_fx_surface_common_type,at_type);
        at->objtypeid=dae_ID_TOKEN;
        at->min=1;
        at->max=1;
        ++at;
        t->attribs = attribs;
        t->numattribs = 1;
    }
    {
        static dae_obj_memberdef elems[1];
        dae_obj_memberdef* el = elems;
        el->name="init_from";
        el->offset=offsetof(dae_fx_surface_common_type,el_init_from);
        el->objtypeid=dae_ID_IMAGE_TYPE_INIT_FROM;
        el->min=0;
        el->max=-1;
        el->seq=0;
        ++el;
        t->elems = elems;
        t->numelems = 1;
    }
    ++t;
    *types_out = s_types;
    *numtypes_out = sizeof(s_types)/sizeof(*s_types);
}

//****************************************************************************
static void dae_free(void* ptr)
{
    free(ptr);
}

//****************************************************************************
static char* dae_strdup(
    const char* str)
{
    char* dup = NULL;
    if(str != NULL && *str != '\0')
    {
        size_t sz = strlen(str) + 1;
        dup = (char*) malloc(sz);
        memcpy(dup, str, sz);
    }
    return dup;
}

//****************************************************************************
static dae_obj_ptr dae_create_obj(
    const dae_obj_typedef* def)
{
    size_t objsize = (def != NULL) ? def->size : sizeof(char*);
    size_t bufsize = ((ptrdiff_t) dae_GET_PTR(0)) + objsize;
    dae_obj_header* hdr = (dae_obj_header*) malloc(bufsize);
    dae_obj_ptr obj = dae_GET_PTR(hdr);
    memset(hdr, 0, bufsize);
    hdr->structindex = -1;
    hdr->def = def;
    return obj;
}

//****************************************************************************
static dae_obj_ptr dae_add_obj(
    dae_obj_header* parenthdr,
    dae_obj_list* parentlist,
    const dae_obj_memberdef* memberdef,
    const char* membername,
    const dae_obj_typedef* childdef,
    size_t childsize)
{
    void* parentobj = dae_GET_PTR(parenthdr);
    void* childobj = dae_create_obj(childdef);
    dae_obj_header* childhdr = dae_GET_HEADER(childobj);
    // initialize header information
    childhdr->parent = parenthdr;
    childhdr->parentlist = parentlist;
    childhdr->memberdef = memberdef;
    // add element to parent's header
    childhdr->prev = parentlist->tail;
    childhdr->next = NULL;
    if(parentlist->tail != NULL)
    {
        parentlist->tail->next = childhdr;
    }
    else
    {
        parentlist->head = childhdr;
    }
    parentlist->tail = childhdr;
    // add element to parent's struct
    if(memberdef != NULL)
    {
        int offset = memberdef->offset;
        int max = memberdef->max;
        if(max == 1)
        {
            // single instance pointer
            void** mbr = (void**) (((ptrdiff_t) parentobj)+offset);
            if(*mbr == NULL)
            {
                childhdr->structindex = 0;
                *mbr = childobj;
            }
        }
        else if(max < 0)
        {
            // variable sized vector
            dae_obj_vector* vec;
            void** buf;
            size_t i;
            vec = (dae_obj_vector*) (((ptrdiff_t) parentobj)+offset);
            i = vec->size;
            childhdr->structindex = i;
            buf = (void**) realloc(vec->values, (i+1) * sizeof(*buf));
            buf[i] = childobj;
            vec->values = buf;
            ++vec->size;
        }
        else
        {
            // fixed length array
            void** mbr = (void**) (((ptrdiff_t) parentobj)+offset);
            void** mbritr = mbr;
            void** mbrend = mbritr + max;
            while(mbritr != mbrend)
            {
                if(*mbritr == NULL)
                {
                    childhdr->structindex = (int) (mbritr - mbr);
                    *mbritr = childobj;
                    break;
                }
                ++mbritr;
            }
        }
        // if member definition exists, re-use name string from definition
        childhdr->membername = memberdef->name;
    }
    else
    {
        // if no member definition exists, need to make a copy of name string
        childhdr->membername = dae_strdup(membername);
    }
    return (dae_obj_ptr) childobj;
}

//****************************************************************************
static void dae_destroy_obj(
    dae_obj_header* hdr)
{
    dae_obj_header* itr = hdr;
    while(1)
    {
        dae_obj_header* next;
        // move down to bottom of tree
        while(itr->elems.head != NULL)
        {
            itr = itr->elems.head;
        }
        while(itr->attribs.head != NULL)
        {
            itr = itr->attribs.head;
        }
        if(itr->next != NULL)
        {
            next = itr->next;
        }
        else
        {
            next = itr->parent;
            if(itr->parentlist != NULL)
            {
                itr->parentlist->head = NULL;
                itr->parentlist->tail = NULL;
            }
        }
        if(itr->attribs.head == NULL && itr->elems.head == NULL)
        {
            void* obj = dae_GET_PTR(itr);
            dae_obj_typeid datatype = dae_ID_INVALID;
            int datamax;
            size_t dataoff;
            if(itr->memberdef == NULL)
            {
                // if the object had a member definition, the member name
                // was derived from the def and does not need to be freed.
                // otherwise, it does
                dae_free((char*) itr->membername);
            }
            if(itr->def != NULL)
            {
                // clean up struct members
                const dae_obj_typedef* def = itr->def;
                const dae_obj_memberdef* mbritr = def->elems;
                const dae_obj_memberdef* mbrend = mbritr + def->numelems;
                while(mbritr != mbrend)
                {
                    if(mbritr->max < 0)
                    {
                        // if the struct member is a variable sized vector
                        // the buffer needs to be freed.
                        dae_obj_vector* vec;
                        size_t off;
                        off = mbritr->offset;
                        vec = (dae_obj_vector*) (((ptrdiff_t) obj) + off);
                        if(vec->values != NULL)
                        {
                            dae_free(vec->values);
                            // set buffer to NULL to prevent multiple free as
                            // multiple element definitions may reference it
                            vec->values = NULL;
                        }
                    }
                    ++mbritr;
                }
                datatype = (dae_obj_typeid) def->datatypeid;
                datamax = def->datamax;
                dataoff = def->dataoffset;
            }
            else
            {
                datatype = dae_ID_STRING;
                datamax = 1;
                dataoff = 0;
            }
            if(datatype != dae_ID_INVALID)
            {
                if(datamax == -1)
                {
                    // if it's a variable sized vector, free the buffer
                    dae_obj_vector* vec;
                    vec = (dae_obj_vector*) (((ptrdiff_t) obj) + dataoff);
                    if(datatype == dae_ID_STRING)
                    {
                        // if it's a vector of strings, all the strings need
                        // to be freed individually
                        char** sitr = (char**) vec->values;
                        char** send = sitr + vec->size;
                        while(sitr != send)
                        {
                            dae_free(*sitr);
                            ++sitr;
                        }
                    }
                    dae_free(vec->values);
                }
                else if(datatype == dae_ID_STRING)
                {
                    // single or fixed size array of strings
                    char** sitr = (char**) (((ptrdiff_t) obj) + dataoff);
                    char** send = sitr + datamax;
                    while(sitr != send)
                    {
                        dae_free(*sitr);
                        ++sitr;
                    }
                }
            }
            dae_free(itr);
            if(itr == hdr)
            {
                // if the freed item was the selected object, stop
                break;
            }
            itr = next;
        }
    }
}

//****************************************************************************
static size_t dae_convert_string_floats(
    const char* str,
    float* floats_out)
{
    // parses the string into an array of floats
    // the array must be large enough to hold the result of
    // dae_count_float_string
    size_t n = 0;
    while(*str != '\0')
    {
        float f = 0.0f;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // attempt to convert string word to float
        if(sscanf(str, "%f", &f) == 1)
        {
            floats_out[n] = f;
            ++n;
        }
        else
        {
            break;
        }
        // find the next set of whitespace
        while(*str != '\0' && !isspace(*str))
        {
            ++str;
        }
    }
    return n;
}

//****************************************************************************
static size_t dae_convert_string_int8s(
    const char* str,
    char* ints_out)
{
    // counts the number of ints that can be parsed from the string
    size_t n = 0;
    while(*str != '\0')
    {
        int i = 0;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // attempt to convert string word to int
        if(sscanf(str, "%i", &i) == 1)
        {
            ints_out[n] = (char) i;
            ++n;
        }
        else
        {
            break;
        }
        // find the next set of whitespace
        while(*str != '\0' && !isspace(*str))
        {
            ++str;
        }
    }
    return n;
}

//****************************************************************************
static size_t dae_convert_string_int16s(
    const char* str,
    short* ints_out)
{
    // counts the number of ints that can be parsed from the string
    size_t n = 0;
    while(*str != '\0')
    {
        int i = 0;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // attempt to convert string word to int
        if(sscanf(str, "%i", &i) == 1)
        {
            ints_out[n] = (short) i;
            ++n;
        }
        else
        {
            break;
        }
        // find the next set of whitespace
        while(*str != '\0' && !isspace(*str))
        {
            ++str;
        }
    }
    return n;
}

//****************************************************************************
static size_t dae_convert_string_int32s(
    const char* str,
    int* ints_out)
{
    // counts the number of ints that can be parsed from the string
    size_t n = 0;
    while(*str != '\0')
    {
        int i = 0;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // attempt to convert string word to int
        if(sscanf(str, "%i", &i) == 1)
        {
            ints_out[n] = i;
            ++n;
        }
        else
        {
            break;
        }
        // find the next set of whitespace
        while(*str != '\0' && !isspace(*str))
        {
            ++str;
        }
    }
    return n;
}

//****************************************************************************
static size_t dae_convert_string_uint8s(
    const char* str,
    unsigned char* ints_out)
{
    // counts the number of ints that can be parsed from the string
    size_t n = 0;
    while(*str != '\0')
    {
        unsigned i = 0;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // attempt to convert string word to int
        if(sscanf(str, "%i", &i) == 1)
        {
            ints_out[n] = (unsigned char) i;
            ++n;
        }
        else
        {
            break;
        }
        // find the next set of whitespace
        while(*str != '\0' && !isspace(*str))
        {
            ++str;
        }
    }
    return n;
}

//****************************************************************************
static size_t dae_convert_string_uint32s(
    const char* str,
    unsigned* ints_out)
{
    // counts the number of ints that can be parsed from the string
    size_t n = 0;
    while(*str != '\0')
    {
        unsigned i = 0;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // attempt to convert string word to int
        if(sscanf(str, "%u", &i) == 1)
        {
            ints_out[n] = i;
            ++n;
        }
        else
        {
            break;
        }
        // find the next set of whitespace
        while(*str != '\0' && !isspace(*str))
        {
            ++str;
        }
    }
    return n;
}

//****************************************************************************
static size_t dae_convert_string_words(
    const char* str,
    char** strings_out)
{
    size_t n = 0;
    while(*str != '\0')
    {
        size_t len = 0;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // determine length of trimmed word
        while(str[len] != '\0' && !isspace(str[len]))
        {
            ++len;
        }
        // duplicate trimmed word
        if(len > 0)
        {
            char* dup = (char*) malloc(len + 1);
            memcpy(dup, str, len);
            dup[len] = '\0';
            strings_out[n] = dup;
            ++n;
        }
        // find the next set of whitespace
        str += len;
    }
    return n;
}

//****************************************************************************
static size_t dae_count_string_floats(
    const char* str)
{
    // counts the number of floats that can be parsed from the string
    size_t n = 0;
    while(*str != '\0')
    {
        float f = 0.0f;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // attempt to convert string word to float
        if(sscanf(str, "%f", &f) == 1)
        {
            ++n;
        }
        else
        {
            break;
        }
        // find the next set of whitespace
        while(*str != '\0' && !isspace(*str))
        {
            ++str;
        }
    }
    return n;
}

//****************************************************************************
static size_t dae_count_string_int8s(
    const char* str)
{
    return dae_count_string_int32s(str);
}

//****************************************************************************
static size_t dae_count_string_int16s(
    const char* str)
{
    return dae_count_string_int32s(str);
}

//****************************************************************************
static size_t dae_count_string_int32s(
    const char* str)
{
    // counts the number of ints that can be parsed from the string
    size_t n = 0;
    while(*str != '\0')
    {
        int i = 0;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // attempt to convert string word to int
        if(sscanf(str, "%i", &i) == 1)
        {
            ++n;
        }
        else
        {
            break;
        }
        // find the next set of whitespace
        while(*str != '\0' && !isspace(*str))
        {
            ++str;
        }
    }
    return n;
}

//****************************************************************************
static size_t dae_count_string_uint8s(
    const char* str)
{
    return dae_count_string_uint32s(str);
}

//****************************************************************************
static size_t dae_count_string_uint32s(
    const char* str)
{
    // counts the number of ints that can be parsed from the string
    size_t n = 0;
    while(*str != '\0')
    {
        int i = 0;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // attempt to convert string word to int
        if(sscanf(str, "%u", &i) == 1)
        {
            ++n;
        }
        else
        {
            break;
        }
        // find the next set of whitespace
        while(*str != '\0' && !isspace(*str))
        {
            ++str;
        }
    }
    return n;
}

//****************************************************************************
static size_t dae_count_string_words(
    const char* str)
{
    size_t n = 0;
    while(*str != '\0')
    {
        size_t len = 0;
        // move past any leading whitespace
        while(*str != '\0' && isspace(*str)) 
        {
            ++str;
        }
        // determine length of trimmed word
        while(str[len] != '\0' && !isspace(str[len]))
        {
            ++len;
        }
        // count it
        if(len > 0)
        {
            ++n;
        }
        // find the next set of whitespace
        str += len;
    }
    return n;
}

//****************************************************************************
static void dae_get_schema(
    dae_obj_typedef** types,
    unsigned* numtypes)
{
    static dae_obj_typedef* s_types;
    static unsigned s_numtypes;
    if(s_types == NULL)
    {
        dae_build_schema(&s_types, &s_numtypes);
    }
    *types = s_types;
    *numtypes = s_numtypes;
}

//****************************************************************************
static dae_obj_typedef* dae_get_type(
    unsigned objtypeid)
{
    dae_obj_typedef* types;
    unsigned numtypes;
    dae_get_schema(&types, &numtypes);
    return (objtypeid < numtypes) ? types + objtypeid : NULL;
}

//****************************************************************************
static dae_obj_typedef* dae_find_type(
    const char* name)
{
    dae_obj_typedef* result = NULL;
    dae_obj_typedef* titr;
    dae_obj_typedef* tend;
    unsigned numtypes;
    dae_get_schema(&titr, &numtypes);
    tend = titr + numtypes;
    while(titr != tend)
    {
        if(!strcmp(titr->name, name))
        {
            result = titr;
            break;
        }
        ++titr;
    }
    return result;
}

//****************************************************************************
dae_obj_ptr dae_add_attrib(
    dae_obj_ptr parent,
    const char* name,
    const char* value)
{
    dae_obj_header* parenthdr = dae_GET_HEADER(parent);
    const dae_obj_typedef* parentdef = parenthdr->def;
    const dae_obj_memberdef* at = NULL;
    const dae_obj_typedef* def = NULL;
    dae_obj_ptr obj = NULL;
    if(parentdef != NULL)
    {
        // predefined content type
        // need to find a predefined attribute
        dae_obj_memberdef* atitr = parentdef->attribs;
        dae_obj_memberdef* atend = atitr + parentdef->numattribs;
        while(atitr != atend)
        {
            if(!strcmp(name, atitr->name))
            {
                at = atitr;
                def = dae_get_type(atitr->objtypeid);
                break;
            }
            ++atitr;
        }
    }
    else
    {
        // undefined type, assume all attributes are strings
        def = dae_get_type(dae_ID_STRING);
    }
    if(def != NULL)
    {
        obj = dae_add_obj(
            parenthdr,
            &parenthdr->attribs,
            at,
            name,
            def,
            def->size);
        if(value == NULL && at != NULL)
        {
            value = at->deflt;
        }
        if(value != NULL)
        {
            dae_set_string(obj, value);
        }
    }
    return obj;
}

//****************************************************************************
dae_obj_ptr dae_add_element(
    dae_obj_ptr parent,
    const char* name)
{
    dae_obj_header* parenthdr = dae_GET_HEADER(parent);
    const dae_obj_typedef* parentdef = parenthdr->def;
    const dae_obj_memberdef* el = NULL;
    const dae_obj_typedef* def = NULL;
    dae_obj_ptr obj = NULL;
    if(parentdef!=NULL && (parentdef->flags&dae_XSD_ANY)==0)
    {
        // predefined content type
        // need to find a predefined element
        dae_obj_memberdef* elitr = parentdef->elems;
        dae_obj_memberdef* elend = elitr + parentdef->numelems;
        while(elitr != elend)
        {
            if(!strcmp(name, elitr->name))
            {
                el = elitr;
                def = dae_get_type(elitr->objtypeid);
                break;
            }
            ++elitr;
        }
    }
    else
    {
        // any element content is allowed
        def = dae_find_type(name);
    }
    if(parentdef == NULL || parentdef->datamax == 0)
    {
        obj = dae_add_obj(
            parenthdr,
            &parenthdr->elems,
            el,
            name,
            def,
            (def != NULL) ? def->size : 0);
    }
    else
    {
        assert(0);
    }
    return obj;
}

//****************************************************************************
dae_COLLADA* dae_create()
{
    const dae_obj_typedef* def = dae_get_type(dae_ID_COLLADA);
    return (dae_COLLADA*) dae_create_obj(def);
}

//****************************************************************************
void dae_destroy(
    dae_COLLADA* doc)
{
    dae_destroy_obj(dae_GET_HEADER(doc));
}

//****************************************************************************
size_t dae_get_data(
    dae_obj_ptr obj,
    dae_native_typeid* type_out,
    void** data_out,
    size_t* datalen_out)
{
    size_t datalen = 0;
    const dae_obj_typedef* def = dae_GET_HEADER(obj)->def;
    if(def != NULL)
    {
        void* data = (void*) (((ptrdiff_t) obj) + def->dataoffset);
        if(def->datamax >= 0)
        {
            // fixed size array
            datalen = def->datamax;
        }
        else
        {
            // dynamically sized vector
            dae_obj_vector* vec = (dae_obj_vector*) data;
            data = vec->values;
            datalen = vec->size;
        }
        *type_out = (dae_native_typeid) def->datatypeid;
        *data_out = data;
        *datalen_out = datalen;
    }
    return datalen;
}

//****************************************************************************
dae_obj_ptr dae_get_first_attrib(
    dae_obj_ptr obj)
{
    dae_obj_header* childhdr = dae_GET_HEADER(obj)->attribs.head;
    return (childhdr != NULL) ? dae_GET_PTR(childhdr) : NULL;
}

//****************************************************************************
dae_obj_ptr dae_get_first_element(
    dae_obj_ptr obj)
{
    dae_obj_header* childhdr = dae_GET_HEADER(obj)->elems.head;
    return (childhdr != NULL) ? dae_GET_PTR(childhdr) : NULL;
}

//****************************************************************************
const char* dae_get_name(
    dae_obj_ptr obj)
{
    return dae_GET_HEADER(obj)->membername;
}

//****************************************************************************
dae_obj_ptr dae_get_next(
    dae_obj_ptr obj)
{
    dae_obj_header* nexthdr = dae_GET_HEADER(obj)->next;
    return (nexthdr != NULL) ? dae_GET_PTR(nexthdr) : NULL;
}

//****************************************************************************
dae_obj_ptr dae_get_parent(
    dae_obj_ptr child)
{
    dae_obj_header* parenthdr = dae_GET_HEADER(child)->parent;
    return (parenthdr != NULL) ? dae_GET_PTR(parenthdr) : NULL;
}

//****************************************************************************
dae_obj_typeid dae_get_typeid(
    dae_obj_ptr obj)
{
    const dae_obj_typedef* def = dae_GET_HEADER(obj)->def;
    return (def != NULL) ? (dae_obj_typeid) def->objtypeid : dae_ID_INVALID;
}

//****************************************************************************
void dae_set_string(
    dae_obj_ptr obj,
    const char* data)
{
    dae_obj_header* hdr = dae_GET_HEADER(obj);
    const dae_obj_typedef* def = hdr->def;
    dae_native_typeid datatype = dae_NATIVE_STRING;
    int dataoffset = -1;
    int max;
    if(def != NULL)
    {
        if(def->datatypeid != dae_ID_INVALID)
        {
            datatype = (dae_native_typeid) def->datatypeid;
            dataoffset = def->dataoffset;
            max = def->datamax;
        }
    }
    else
    {
        // if this is a generic element with no type definition, make it a
        // pointer to a string
        datatype = dae_NATIVE_STRING;
        dataoffset = 0;
        max = 1;
    }
    if(dataoffset >= 0)
    {
        if(max == 1)
        {
            void* p = (void*) (((ptrdiff_t) obj) + dataoffset);
            // trim leading space
            while(*data != '\0' && isspace(*data))
            {
                ++data;
            }
            switch(datatype)
            {
            case dae_NATIVE_BOOL32:
                // TODO
                break;
            case dae_NATIVE_HEX8:
                // TODO
                break;
            case dae_NATIVE_FLOAT:
                {
                    float f = 0;
                    if(sscanf(data, "%f", &f) == 1)
                    {
                        float* fp = (float*) p;
                        *fp = (float) f;
                    }
                }
                break;
            case dae_NATIVE_INT8:
                {
                    int i = 0;
                    if(sscanf(data, "%i", &i) == 1)
                    {
                        char* ip = (char*) p;
                        *ip = (char) i;
                    }
                }
                break;
            case dae_NATIVE_INT16:
                {
                    int i = 0;
                    if(sscanf(data, "%i", &i) == 1)
                    {
                        short* ip = (short*) p;
                        *ip = (short) i;
                    }
                }
                break;
            case dae_NATIVE_INT32:
                {
                    int i = 0;
                    if(sscanf(data, "%i", &i) == 1)
                    {
                        int* ip = (int*) p;
                        *ip = (int) i;
                    }
                }
                break;
            case dae_NATIVE_STRING:
                {
                    // TODO: free old pointer?
                    size_t len = 0;
                    // trim trailing space
                    while(data[len] != '\0' && !isspace(data[len]))
                    {
                        ++len;
                    }
                    if(len > 0)
                    {
                        char** sp = (char**) p;
                        char* dup = (char*) malloc(len + 1);
                        memcpy(dup, data, len);
                        dup[len] = '\0';
                        *sp = dup;
                    }
                }
                break;
            case dae_NATIVE_UINT8:
                {
                    unsigned i = 0;
                    if(sscanf(data, "%u", &i) == 1)
                    {
                        unsigned char* ip = (unsigned char*) p;
                        *ip = (unsigned char) i;
                    }
                }
                break;
            case dae_NATIVE_UINT32:
                {
                    unsigned i = 0;
                    if(sscanf(data, "%u", &i) == 1)
                    {
                        unsigned* ip = (unsigned *) p;
                        *ip = (unsigned) i;
                    }
                }
                break;
            default:
                assert(0);
                break;
            }
        }
        else if(max == -1)
        {
            dae_obj_vector* vec;
            vec = (dae_obj_vector*) (((ptrdiff_t) obj) + dataoffset);
            switch(datatype)
            {
            case dae_NATIVE_BOOL32:
                // TODO
                break;
            case dae_NATIVE_HEX8:
                // TODO
                break;
            case dae_NATIVE_FLOAT:
                {
                    size_t n = dae_count_string_floats(data);
                    if(n > 0)
                    {
                        float* fv = (float*) vec->values;
                        if(n > vec->size)
                        {
                            fv = (float*) realloc(fv, n*sizeof(*fv));
                            vec->values = fv;
                        }
                        vec->size = dae_convert_string_floats(data, fv);
                    }
                }
                break;
            case dae_NATIVE_INT8:
                {
                    size_t n = dae_count_string_int8s(data);
                    if(n > 0)
                    {
                        char* iv = (char*) vec->values;
                        if(n > vec->size)
                        {
                            iv = (char*) realloc(iv, n*sizeof(*iv));
                            vec->values = iv;
                        }
                        vec->size = dae_convert_string_int8s(data, iv);
                    }
                }
                break;
            case dae_NATIVE_INT16:
                {
                    size_t n = dae_count_string_int16s(data);
                    if(n > 0)
                    {
                        short* iv = (short*) vec->values;
                        if(n > vec->size)
                        {
                            iv = (short*) realloc(iv, n*sizeof(*iv));
                            vec->values = iv;
                        } 
                        vec->size = dae_convert_string_int16s(data, iv);
                    }
                }
                break;
            case dae_NATIVE_INT32:
                {
                    size_t n = dae_count_string_int32s(data);
                    if(n > 0)
                    {
                        int* iv = (int*) vec->values;
                        if(n > vec->size)
                        {
                            iv = (int*) realloc(iv, n*sizeof(*iv));
                            vec->values = iv;
                        }
                        vec->size = dae_convert_string_int32s(data, iv);
                    }
                }
                break;
            case dae_NATIVE_STRING:
                {
                    size_t n = dae_count_string_words(data);
                    if(n > 0)
                    {
                        // TODO: free old pointers?
                        char** sv = (char**) vec->values;
                        if(n > vec->size)
                        {
                            sv = (char**) realloc(sv, n*sizeof(*sv));
                            vec->values = sv;
                        }
                        vec->size = dae_convert_string_words(data, sv);
                    }
                }
                break;
            case dae_NATIVE_UINT8:
                {
                    size_t n = dae_count_string_uint8s(data);
                    if(n > 0)
                    {
                        unsigned char* iv = (unsigned char*) vec->values;
                        if(n > vec->size)
                        {
                            iv = (unsigned char*) realloc(iv, n*sizeof(*iv));
                            vec->values = iv;
                        }
                        vec->size = dae_convert_string_uint8s(data, iv);
                    }
                }
                break;
            case dae_NATIVE_UINT32:
                {
                    size_t n = dae_count_string_uint32s(data);
                    if(n > 0)
                    {
                        unsigned* iv = (unsigned*) vec->values;
                        if(n > vec->size)
                        {
                            iv = (unsigned*) realloc(iv, n*sizeof(*iv));
                            vec->values = iv;
                        }
                        vec->size = dae_convert_string_uint32s(data, iv);
                    }
                }
                break;
            default:
                assert(0);
                break;
            }
        }
        else
        {
            // fixed size array
            void* ar = (void*) (((ptrdiff_t) obj) + dataoffset);
            switch(datatype)
            {
            case dae_NATIVE_BOOL32:
                // TODO
                break;
            case dae_NATIVE_HEX8:
                // TODO
                break;
            case dae_NATIVE_FLOAT:
                {
                    int n = dae_count_string_floats(data);
                    if(n == max)
                    {
                        float* fv = (float*) ar;
                        dae_convert_string_floats(data, fv);
                    }
                }
                break;
            case dae_NATIVE_INT8:
                {
                    int n = dae_count_string_int8s(data);
                    if(n == max)
                    {
                        char* iv = (char*) ar;
                        dae_convert_string_int8s(data, iv);
                    }
                }
                break;
            case dae_NATIVE_INT16:
                {
                    int n = dae_count_string_int16s(data);
                    if(n == max)
                    {
                        short* iv = (short*) ar;
                        dae_convert_string_int16s(data, iv);
                    }
                }
                break;
            case dae_NATIVE_INT32:
                {
                    int n = dae_count_string_int32s(data);
                    if(n == max)
                    {
                        int* iv = (int*) ar;
                        dae_convert_string_int32s(data, iv);
                    }
                }
                break;
            case dae_NATIVE_STRING:
                {
                    int n = dae_count_string_words(data);
                    if(n == max)
                    {
                        // TODO: free old pointers?
                        char** sv = (char**) ar;
                        dae_convert_string_words(data, sv);
                    }
                }
                break;
            case dae_NATIVE_UINT8:
                {
                    int n = dae_count_string_uint8s(data);
                    if(n == max)
                    {
                        unsigned char* iv = (unsigned char*) ar;
                        dae_convert_string_uint8s(data, iv);
                    }
                }
                break;
            case dae_NATIVE_UINT32:
                {
                    int n = dae_count_string_uint32s(data);
                    if(n == max)
                    {
                        unsigned* iv = (unsigned*)ar;
                        dae_convert_string_uint32s(data, iv);
                    }
                }
                break;
            default:
                assert(0);
                break;
            }
        }
    }
}
